#!/usr/bin/env python
#
#  Copyright Euratom-CEA
#  Authors : 
#     Virginie Grandgirard (virginie.grandgirard@cea.fr)
#     Chantal Passeron (chantal.passeron@cea.fr)
#     Guillaume Latu (guillaume.latu@cea.fr)
#     Xavier Garbet (xavier.garbet@cea.fr)
#     Philippe Ghendrih (philippe.ghendrih@cea.fr)
#     Yanick Sarazin (yanick.sarazin@cea.fr)
#  
#  This code GYSELA (for GYrokinetic SEmi-LAgrangian) 
#  is a 5D gyrokinetic global full-f code for simulating 
#  the plasma turbulence in a tokamak.
#  
#  This software is governed by the CeCILL-B license 
#  under French law and abiding by the rules of distribution 
#  of free software.  You can  use, modify and redistribute 
#  the software under the terms of the CeCILL-B license as 
#  circulated by CEA, CNRS and INRIA at the following URL
#  "http://www.cecill.info".

# TODO:
#  * provide completion for path input (w. readline)
#  * accept a parameter to change log level
#  * save info to log for easy debugging

__author__  = 'Julien Bigot <julien.bigot@cea.fr>'
__version__ = '0.1-wip'
__date__    = '2013/05/19'
__license__ = 'CeCILL-B'
BUILD_TYPES = ['Release', 'Timed', 'Deterministic', 'Debug']

"""
A simple script that sets up a cmake build dir for GYSELA5D

Author: Julien Bigot

"""

import logging
from optparse import OptionParser
import os
import re
import shutil
import string
import subprocess
import sys


GYS_DIR = os.path.abspath(os.path.join(os.path.dirname(os.path.realpath(sys.argv[0])), '..'))

logging.basicConfig(format='*** %(levelname)s: %(message)s')
LOGGER = logging.getLogger()


def user_prompt(prompt, default=None):
    """
    Asks the user for a string value.
    
    Arguments:
     * prompt: string description of the value asked to the user
     * default: default value if the user does not provide any input
    
    Result: the user supplied value
      
    """
    
    if ( default != None ):
        prompt = '--> '+prompt+'? '+'['+default+']: '
    else:
        prompt = '--> '+prompt+'? '+': '
    result = raw_input(prompt)
    print
    if ( result != '' ):
        return result
    else:
        return default


def user_select(prompt, values, default_rank=0):
    """
    Asks the user for a choice amongst a set of possible values
    
    Arguments:
     * prompt: string description of the value asked to the user
     * values: an iterable of strings that describes the various possibilities
     * default: rank of the default value if the user does not provide any input
    
    Result: rank of the user choice in the values argument
      
    """
    
    rank = 0
    print '--> '+prompt+':'
    while ( rank < len(values) ):
        if ( rank == default_rank ):
            print rank,') '+values[rank]+' [*]'
        else:
            print rank,') '+values[rank]
        rank = rank+1
    result = -1
    while ( not ( result >=0 and result < len(values) ) ):
        result = raw_input('choice: ')
        if ( result == '' ):
            result = default_rank
        else:
            result = string.atoi(result)
    print
    return result


def user_yn(prompt, default=True):
    """
    Asks the user for a boolean values
    
    Arguments:
     * prompt: string description of the value asked to the user
     * default: value if the user does not provide any input
    
    Result: the user supplied value
      
    """
    
    prompt = '--> '+prompt+'? (y/n)'
    if (default):
        prompt = prompt+' [Y]: '
    else:
        prompt = prompt+' [N]: '
    result = raw_input(prompt)
    print
    while ( True ):
        if ( result == '' ):
            return default
        elif ( result.lower() == 'y' ):
            return True
        elif ( result.lower() == 'n' ):
            return False
        else:
            result = raw_input('Invalid choice (y/n): ')


def available_arch():
    """
    Returns the list of ARCH values for which there are predefined parameters available
    
    Result: a list containing all possible ARCH values
    
    """
    
    if hasattr(available_arch, 'result'):
        return available_arch.result
        
    available_arch.result = list([''])
    for arch in sorted(os.listdir(os.path.join(GYS_DIR, 'buildsystem', 'defaults'))):
        if ( re.match('\A(\w|-)+\.cmake\Z', arch) ):
            arch = re.sub('\.cmake\Z', '', arch)
            available_arch.result.append(arch)
    return available_arch.result


def check():
    """
    Checks if cmake is correctly installed
      
    """
    
    try:
        cmk = subprocess.Popen(['cmake', '--version'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        (cmake_version, cmkerr) = cmk.communicate()
        matched_version = re.match('\Acmake version ([0-9]+)\.([0-9]+)(\.([0-9]+))*\n\Z', cmake_version)
        if cmk.returncode != 0:
            LOGGER.error('unable to find cmake: %s', cmkerr)
            return False
        elif not matched_version:
            LOGGER.error('unable to parse cmake version: %s', cmake_version)
            return False
        elif int(matched_version.group(1)) < 2 or int(matched_version.group(2)) < 8:
            LOGGER.error('%s too low (>=2.8.0 required)', cmake_version)
            return False
    except OSError:
        LOGGER.error('unable to find cmake')
        exit
    
    return True
    

def check_build_dir(dir):
    
    if dir is None:
        return None
    
    dir = os.path.realpath(dir)
    
    if ( os.path.exists(dir) and not os.path.isdir(dir) ):
        LOGGER.error('%s exists and is not a directory', dir)
        sys.exit(os.EX_USAGE)
    
    if ( os.path.isdir(dir) and
        len(os.listdir(dir))>0 and
        not os.path.exists(os.path.join(dir, 'CMakeCache.txt')) ):
        LOGGER.error('directory %s is not empty', dir)
        sys.exit(os.EX_USAGE)
    
    return dir


def callback_jobs(option, opt_str, value, parser):
    if parser.rargs and parser.rargs[0].isdigit():
        val = int(parser.rargs[0])
        parser.rargs.pop(0)
    else:
        val = 0
    parser.values.jobs = val


def config():
    """
    Reads parameters sets on the command line.
    
    Result: a Namespace containing the following elements:
     -> arch(string): initial cache file
     -> build_dir(string): Gysela build directory
     -> build_type(string): Build type for GYSELA (Release, Timed, Deterministic, Debug)
     -> compile(boolean): whether make should be run after cmake
     -> configure(boolean): whether to run cmake
     -> full(boolean): whether to let the use configure all options
     -> interactive(boolean): whether to automatically choose values
     -> jobs(integer): parallelism level of the compilation, 0 for maximum (if compile == True)
     -> verbose(boolean): whether to compile in verbose mode
    
    """
    
    parser = OptionParser(description="A compilation script fot Gysela5D\n\nConfigs: "+string.join(BUILD_TYPES, ', '), version=__version__, usage='%prog [options...] [<config>]')
    parser.add_option('-A', '--arch',       choices=available_arch(), help='The architechture for which to select default values')
    parser.add_option('-d', '--directory',  dest='build_dir', metavar='DIR', help='Where to build')
    parser.add_option('-c', '--no-compile', action='store_false', dest='compile', help='Only configure, do not compile')
    parser.add_option('-f', '--full',       action='store_true', help='Configure the full set of available options by hand')
    parser.add_option('-i', '--interactive',dest='interactive', default=None, action='store_true', help='Force interactive mode')
    parser.add_option('-j', '--jobs',       action='callback', default=None, callback=callback_jobs, help='Parallelism to use')
    parser.add_option('-v', '--verbose',    action='store_true', help='make the compilation more verbose')
    (cfg, build_type) = parser.parse_args()
    
    if len(build_type) == 0:
        cfg.build_type = None
    elif len(build_type) == 1:
        cfg.build_type = build_type[0]
    else:
        parser.error('Too many arguments')
    
    if not hasattr(cfg, 'jobs'):
        cfg.jobs = None
    
    if cfg.full:
        cfg.interactive = True
    
    if cfg.interactive is None:
        cfg.interactive = False
    
    # build_type
    if cfg.build_type is None:
        build_type = 0
        if cfg.interactive:
            build_type = user_select('Type of build', BUILD_TYPES, build_type)
        build_type = BUILD_TYPES[build_type]
        cfg.build_type = build_type
    if cfg.build_type not in BUILD_TYPES:
        parser.error('Invalid build type "'+cfg.build_type+'", valid build types are: '+string.join(BUILD_TYPES, ' '))
    LOGGER.info('build_type='+str(cfg.build_type))
    
    # arch
    if cfg.arch is None:
        arch = 0
        try:
            arch = available_arch().index(os.environ.get('ARCH'))
        except ValueError:
            pass
        if cfg.interactive:
            arch = user_select('Predefined values for parameters', ['Global defaults values']+available_arch()[1:], arch)
        cfg.arch = available_arch()[arch]
    LOGGER.info('arch='+cfg.arch)
    
    # build_dir
    cfg.build_dir = check_build_dir(cfg.build_dir)
    if cfg.build_dir is None:
        cwd = os.path.realpath(os.getcwd())
        if ( os.path.commonprefix([cwd, GYS_DIR]) == GYS_DIR ):
            build_dir = os.path.join(GYS_DIR,'build')
        else:
            build_dir = os.path.join(cwd, 'gysela-build')
        if cfg.arch != '':
            build_dir = build_dir+'-'+cfg.arch
        build_dir = build_dir+'-'+cfg.build_type
        if cfg.interactive:
            build_dir = check_build_dir(user_prompt('build directory', build_dir))
        else:
            build_dir = check_build_dir(build_dir)
        cfg.build_dir = build_dir
    LOGGER.info('build_dir='+cfg.build_dir)
    
    # configure
    if os.path.exists(os.path.join(cfg.build_dir, 'Makefile')):
        cfg.configure = False
    else:
        cfg.configure = True
    LOGGER.info('configure='+str(cfg.configure))
    
    # full
    if cfg.configure and cfg.full is None:
        full = False
        if cfg.interactive:
            full = user_yn('Configure all options by hand', full)
        cfg.full = full
    if cfg.full is not None:
        LOGGER.info('full='+str(cfg.full))
    
    # compile
    if cfg.compile is None:
        compile = True
        if cfg.interactive:
            compile = user_yn('actually compile GYSELA', compile)
        cfg.compile = compile
    LOGGER.info('compile='+str(cfg.compile))
    
    # jobs
    if cfg.compile and cfg.jobs is None:
        jobs = 0
        if cfg.interactive:
            jobs = user_select('Number of jobs', ['unlimited', '1', '2', '4', '8', '16'], jobs)
            if jobs != 0:
                jobs = pow(2, jobs-1)
        cfg.jobs = jobs
    if cfg.jobs is not None:
        LOGGER.info('jobs='+str(cfg.jobs))
    
    return cfg


def cmake(cfg):
    """
    Runs cmake
    
    Arguments:
     * cfg: the user-specified configuration
    
    """
    
    args = list(['cmake'])
    
    if cfg.full:
        args.append('-i')
    
    if cfg.arch != '':
        args.append('-DCMAKE_TOOLCHAIN_FILE='+os.path.join(GYS_DIR, 'buildsystem', 'defaults', cfg.arch+'.cmake'))
    
    args.append('-DCMAKE_BUILD_TYPE='+cfg.build_type)
    
    args.append(os.path.join(GYS_DIR, 'buildsystem'))
    
    print '# '+string.join(args,' ')
    LOGGER.info('running cmake "'+string.join(args,'" "')+'"')
    result = subprocess.call(args)
    
    if ( result != 0 ):
        LOGGER.error('cmake failed, please send above information to Julien Bigot <julien.bigot@cea.fr>')
        sys.exit(1)


def make(cfg):
    """
    Runs make
    
    Arguments:
     * cfg: the user-specified configuration
    
    """
    
    if ( cfg.compile ):
        args = list(['make'])
        if cfg.jobs == 0:
            args.append('-j')
        elif cfg.jobs != 1:
            args.append('-j')
            args.append(str(cfg.jobs))
        if cfg.verbose:
            args.append('VERBOSE=1')
        args.append('install')
        print '# '+string.join(args)
        LOGGER.info('compiling gysela')
        #TODO: check if make supports the --no-print-directory option
        result = subprocess.call([args[0], '--no-print-directory']+args[1:])
        print
        if ( result != 0 ):
            print 'compilation failed, see log above for details'
    else:
        print 'you can now compile gysela by running "make install" in '+cfg.build_dir


def main():
    
    if not check():
        return
    
    cfg = config()
    
    if not os.path.isdir(cfg.build_dir):
        LOGGER.info('creating build directory')
        print '# mkdir '+cfg.build_dir
        os.mkdir(cfg.build_dir)
    print '# cd '+cfg.build_dir
    os.chdir(cfg.build_dir)
    
    if cfg.configure:
        print
        result = cmake(cfg)
    
    if cfg.compile:
        print
        make(cfg)


if ( __name__ == "__main__" ):
    main()


