import collections
from decimal import Decimal
import namelist
import re

__author__  = 'Julien Bigot <julien.bigot@cea.fr>'
__version__ = '0.1-wip'
__date__    = '2013/01/01'
__license__ = 'CECILL-B'

class GysData(namelist.Namelist):
    """ A parser for Gysela data files
    """

    def var(self, varName=None):
        """ Returns the value of a shell-style variable """
        if varName is not None:
            return self._var[varName]
        else:
            return self._var
    
    def _parse(self):
        """ Parses the whole data file """
        self._var = dict()
        self._nl()
        val = self._varAssign()
        while ( not self._err(val) ):
            self._sp()
            newline = self._nl()
            if self._err(newline): return val
            val = self._varAssign()
        nl = namelist.Namelist._parse(self)
        return self._or(val, nl)

    def _varAssign(self):
        """ Parses the assignment of a shell-style variable """
        startidx = self._index
        self._nl()
        self._sp()
        id = self._id()
        if self._err(id): return self._required(startidx, id)
        self._sp()
        eq = self._token('=', "'='")
        if self._err(eq): return self._required(startidx, eq)
        self._sp()
        val = self._spString()
        if self._err(val): return self._required(startidx, val)
        self._sp()
        self._var[id] = val

    def _int(self):
        """ Parses an integer value """
        tok = namelist.Namelist._int(self)
        if not self._err(tok): return int(tok)
        err = tok
        tok = self._token(r'\$[a-zA-Z_]+', '$variable')
        if not self._err(tok): return int(self._var[tok[1:]])
        err = self._or(err, tok)
        tok = self._token(r'\${[a-zA-Z_]+}', '${variable}')
        if not self._err(tok): return int(self._var[tok[2:-1]])
        err = self._or(err, tok)
        tok = self._token(r'\${[a-zA-Z_]+}', '${variable}')
        if not self._err(tok): return int(self._var[tok[2:-1]])
        err = self._or(err, tok)
        val = self._shMath()
        if not self._err(val): return val
        err = self._or(err, val)
        return err

    def _shMath(self):
        """ Parses shell maths """
        startidx = self._index
        tok = self._token(r'\$\(\(', '"$(("')
        if self._err(tok): return self._required(startidx, tok)
        self._sp()
        val = self._inShMath()
        if self._err(val): return self._required(startidx, val)
        tok = self._token(r'\)\)', '"))"')
        if self._err(tok): return self._required(startidx, tok)
        self._sp()
        return val

    def _inShMath(self):
        """ Parses a shell arithmetic expression """
        val = self._inShInt()
        if self._err(val): return val
        valididx = self._index
        op = self._token(r'[\-+]', 'arithmetic operator (+,-)')
        while not self._err(op):
            self._sp()
            right = self._inShInt()
            if self._err(right): return self._required(valididx, right)
            if op == '-': val -= right
            if op == '+': val += right
            self._sp()
            valididx = self._index
            op = self._token(r'[\-+]', 'arithmetic operator (+,-)')
        self._sp()
        return val

    def _inShInt(self):
        """ Parses a shell arithmetic integer """
        val = self._int()
        if not self._err(val): return val
        err = val
        tok = self._token(r'[a-zA-Z_]+', 'variable')
        if not self._err(tok): return int(self._var[tok])
        return self._or(err, tok)

    def _spString(self):
        """ Parses a space separated string """
        return self._token(r"[^ \t\n!#]*", 'shell value')
    
