from decimal import Decimal
import basicparser
from ordereddict import OrderedDict

__author__  = 'Julien Bigot <julien.bigot@cea.fr>'
__version__ = '0.1-wip'
__date__    = '2013/01/01'
__license__ = 'CECILL-B'

class Namelist(basicparser.BasicParser):
    """ A parser for Fortran namelists
    """
    
    def fort(self, secName=None, varName=None):
        if secName is not None and varName is not None:
            if not secName.upper() in self._fort:
                return None
            if not varName.upper() in self._fort[secName.upper()]:
                return None
            return self._fort[secName.upper()][varName.upper()]
        elif secName is not None:
            if not secName.upper() in self._fort:
                return None
            return self._fort[secName.upper()]
        else:
            return self._fort

    def _parse(self):
        self._fort = OrderedDict()
        val = self._namelist()
        while ( not self._err(val) ):
            val = self._namelist()
        self._nl()
        self._sp()
        return val

    def _namelist(self):
        startidx = self._index
        self._nl()
        self._sp()
        secName = self._token(r'&[a-zA-Z_]+', 'namelist title')
        if self._err(secName): return self._required(startidx, secName)
        secVal = OrderedDict()
        secName = secName[1:].upper()
        while not self._err(self._nlassign(secVal)):
            pass
        self._fort[secName]= secVal
        self._nl()
        secEnd = self._token('/', '"/"')
        if self._err(secEnd): return self._required(startidx, secEnd)
        self._sp()

    def _nlassign(self, res):
        startidx = self._index
        self._nl()
        self._sp()
        id = self._id()
        if self._err(id): return self._required(startidx, id)
        id = id.upper()
        self._nl()
        self._sp()
        eq = self._token('=', "'='")
        if self._err(eq): return self._required(startidx, eq)
        self._nl()
        self._sp()
        res[id] = list()
        val = self._nlvalsep()
        if self._err(val): return self._required(startidx, val)
        while not self._err(val):
            res[id].append(val)
            val = self._nlvalsep()

    def _nlvalsep(self):
        startidx = self._index
        val = self._nlval()
        if self._err(val): return self._required(startidx, val)
        sep = self._nlsep()
        if self._err(sep): return self._required(startidx, sep)
        return val

    def _nlval(self):
        real = self._real()
        if not self._err(real): return real
        val = self._int()
        if not self._err(val): return val
        qstr = self._qteString()
        if not self._err(qstr): return qstr
        boolean = self._bool()
        if not self._err(boolean): return boolean
        return self._or(val, qstr, real, boolean)

    def _nlsep(self):
        res = ''
        tok = self._sp()
        if not self._err(tok): res+=tok
        tok = self._nl()
        if not self._err(tok): res+=tok
        tok = self._token(',', '","')
        if not self._err(tok): res+=tok
        tok = self._sp()
        if not self._err(tok): res+=tok
        tok = self._nl()
        if not self._err(tok): res+=tok
        if res == '': return self._invalid('namelist separator')
        return res

    def _qteString(self):
        return self._token(r'"[^"]*"', 'quoted string')

    def _bool(self):
        res = self._token(r"(?i)(\.true\.|\.false\.)", 'boolean')
        if self._err(res): return res
        if res.upper() == '.TRUE.': return True
        if res.upper() == '.FALSE.': return False

    def _real(self):
        res = self._token(r"[+\-]?[0-9]+\.[0-9]*([eE][+\-]?[0-9]+)?", 'real')
        if self._err(res): return res
        return Decimal(res)

    def _nl(self):
        return self._token(r"(([ \t]|(/\*([^\*]|\*[^/])\*/)|(#[^\n]*$)|(![^\n]*$))*$\n?)+", 'spaces and newlines')

    def _sp(self):
        return self._token(r"([ \t]|(/\*([^\*]|\*[^/])\*/))+", 'space(s)')

    def _int(self):
        res = self._token(r'[+\-]? *[0-9]+', 'integer')
        if self._err(res): return res
        return int(res)

    def _id(self):
        return self._token(r"[a-zA-Z_][a-zA-Z0-9_]*", 'identifier')

    def _fortstr(self, val):
        """ dumps the FORTRAN representation of some data """
        if val is True: return '.TRUE.'
        if val is False: return '.FALSE.'
        if (isinstance(val, int)):
            return str(val)
        if (isinstance(val, Decimal)):
            if val % Decimal(1) == Decimal(0): return str(val)+'.'
            return str(val)
        if (isinstance(val, basestring)):
            return '"'+val+'"'
        raise NotImplementedError

    def __str__(self):
        return ''.join([ '&'+secName+'\n'+', '.join([
            varName+'='+', '.join(map(self._fortstr, varVal))
            for (varName, varVal) in self.fort(secName).iteritems()
        ])+'\n/\n'
        for secName in self.fort()])
