#!/bin/bash

#**************************************************************
#  Copyright Euratom-CEA
#  Authors : 
#     Virginie Grandgirard (virginie.grandgirard@cea.fr)
#     Chantal Passeron (chantal.passeron@cea.fr)
#     Guillaume Latu (guillaume.latu@cea.fr)
#     Xavier Garbet (xavier.garbet@cea.fr)
#     Philippe Ghendrih (philippe.ghendrih@cea.fr)
#     Yanick Sarazin (yanick.sarazin@cea.fr)
#  
#  This code GYSELA (for GYrokinetic SEmi-LAgrangian) 
#  is a 5D gyrokinetic global full-f code for simulating 
#  the plasma turbulence in a tokamak.
#  
#  This software is governed by the CeCILL-B license 
#  under French law and abiding by the rules of distribution 
#  of free software.  You can  use, modify and redistribute 
#  the software under the terms of the CeCILL-B license as 
#  circulated by CEA, CNRS and INRIA at the following URL
#  "http://www.cecill.info". 
#**************************************************************

# F: Fortran type
F_i="integer"
F_f="real(SKIND)"
F_d="real(RKIND)"
F_z="complex(CKIND)"

# S: type size in Bytes (octet)
S_i="4"
S_f="4"
S_d="8"
S_z="16"

# M: Format descriptor for I/O
M_i="I30"
M_f="E30.5"
M_d="E30.5"
M_z="(E15.5,E15.5)"

ALLTYPES="i f d z"

cat<<HEADER1_END

#include "config-F.inc"

module Debug_module
  
  use Io_Unit_module
  use mpiMod
  use prec_const
  
  implicit none
  
  private
  public :: cksum, print_cksum
  
  interface cksum
HEADER1_END
for T in ${ALLTYPES}; do # T: type 
  for D in $(seq 0 7); do # D: dimensions of array
    echo "    module procedure cksum_${T}${D}d"
  done
done
cat<<HEADER2_END

  endinterface cksum
  
  interface print_cksum
    module procedure &
HEADER2_END
SEP=''
for T in ${ALLTYPES}; do # T: type 
  for D in $(seq 0 7); do # D: dimensions of array
    echo "                   print_cksum_${T}${D}d, &"
  done
done
cat<<HEADER3_END
           print_cksum_c0d
  endinterface print_cksum
  
  integer:: cksum_unit = -1
  
contains
  
  subroutine unit_open ()
    
    integer:: rank, ierr
    character(len=17):: fname
    
    if ( cksum_unit == -1 ) then
      
      call MPI_Comm_rank(MPI_COMM_WORLD, rank, ierr)
      write(fname, '(A6,I7.7,A4)') 'cksum-', rank, '.out'
      call IU_open(cksum_unit, file=fname)
      write(cksum_unit, "(' === cksum-',I7.7,'.out === ')") rank
      
    endif
    
  endsubroutine unit_open
HEADER3_END
for T in ${ALLTYPES}; do # T: type 
	for D in $(seq 0 7); do # D: dimensions of array
		A="" # A: array descriptor
		C=1; # array count in elements
		CKS="data"
		CKS_M="$(eval "echo \${M_${T}}")" # M: Format descriptor for I/O
		if (( D>0 )); then
			A="(:"
			N=1
			C="size(data)"
			CKS="cksum(data)"
      CKS_M="A30"
			while (( N<D )); do
				A="${A},:"
				(( ++N ))
			done
			A="${A})"
		fi
		F="$(eval "echo \${F_${T}}")" # F: Fortran type
		S="$(eval "echo \${S_${T}}")" # S: type size in Bytes (octet)
		cat<<BODY2_END
  
  character(len=30) function cksum_${T}${D}d(data)
    
    ${F}, intent(in):: data${A}
    
    call fnv0_impl(data, ${S}, ${C}, cksum_${T}${D}d)
    
  endfunction cksum_${T}${D}d
  
  subroutine print_cksum_${T}${D}d(file, line, identifier, data)
    
    character(len=*), intent(in):: file
    integer, intent(in):: line
    character(len=*), intent(in):: identifier
    ${F}, intent(in):: data${A}
    
#ifdef ENABLE_CKSUM
    call unit_open()
    
    write(cksum_unit, "('# ',A40,1X,I6)") file, line
    write(cksum_unit, "(A50,' = ',${CKS_M})") identifier, ${CKS}
#endif
    
  endsubroutine print_cksum_${T}${D}d
BODY2_END
	done
done
cat<<FOOTER_END
  
  subroutine print_cksum_c0d(file, line, identifier, data)
    
    character(len=*), intent(in):: file
    integer, intent(in):: line
    character(len=*), intent(in):: identifier
    character(len=*), intent(in):: data
    
#ifdef ENABLE_CKSUM
    call unit_open()
    
    write(cksum_unit, "('# ',A40,1X,I6)") file, line
    write(cksum_unit, "(A50,' = ',A)") identifier, data
#endif
    
  endsubroutine print_cksum_c0d
  
endmodule Debug_module
FOOTER_END