!**************************************************************
!  Copyright Euratom-CEA
!  Authors : 
!     Virginie Grandgirard (virginie.grandgirard@cea.fr)
!     Chantal Passeron (chantal.passeron@cea.fr)
!     Guillaume Latu (guillaume.latu@cea.fr)
!     Xavier Garbet (xavier.garbet@cea.fr)
!     Philippe Ghendrih (philippe.ghendrih@cea.fr)
!     Yanick Sarazin (yanick.sarazin@cea.fr)
!  
!  This code GYSELA (for GYrokinetic SEmi-LAgrangian) 
!  is a 5D gyrokinetic global full-f code for simulating 
!  the plasma turbulence in a tokamak.
!  
!  This software is governed by the CeCILL-B license 
!  under French law and abiding by the rules of distribution 
!  of free software.  You can  use, modify and redistribute 
!  the software under the terms of the CeCILL-B license as 
!  circulated by CEA, CNRS and INRIA at the following URL
!  "http://www.cecill.info". 
!**************************************************************

#include "config-F.inc"

!-------------------------------------------------
! file : Pcross_section.F90
! date : 18/04/2005
!  used for saving cross_sections of the 
!  electric potential and the
!  distribution function in all directions
!------------------------------------------------
module Pcross_section_module
  use fdistribu5d_class, only: fdistribu5d
  use geometry_class,    only: geometry
  use globals,           only: DIAG_PHI2D
  use poisson_class,     only: poisson
  use prec_const
  use sources_module,    only: init_sources

  implicit none
  
  !-------------------------------------
  !  PRIVATE VARIABLES
  !-------------------------------------
  real(RKIND), dimension(:,:), pointer :: frtheta_loc_tmp
  real(RKIND), dimension(:,:), pointer :: frtheta_glob_tmp
  real(RKIND), dimension(:,:), pointer :: frvpar_loc_tmp
  real(RKIND), dimension(:,:), pointer :: frvpar_glob_tmp
  real(RKIND), dimension(:,:), pointer :: fthetavpar_loc_tmp
  real(RKIND), dimension(:,:), pointer :: fthetavpar_glob_tmp

  !-------------------------------------
  !  PUBLIC VARIABLES
  !-------------------------------------
  public
  ! -> Electric Potential in 2D for output saving 
  real(RKIND), dimension(:,:), pointer :: Phirth_diag
  real(RKIND), dimension(:,:), pointer :: Phirphi_diag
  real(RKIND), dimension(:,:), pointer :: Phithphi_diag
  ! -> distribution function in 2D for output saving
  real(RKIND), dimension(:,:), pointer :: frtheta_v0_diag
  real(RKIND), dimension(:,:), pointer :: frtheta_vmax_diag
  real(RKIND), dimension(:,:), pointer :: fphivpar_mu0_diag
  real(RKIND), dimension(:,:), pointer :: fphivpar_mumax_diag
  real(RKIND), dimension(:,:), pointer :: frvpar_mu0_diag
  real(RKIND), dimension(:,:), pointer :: frvpar_mumax_diag
  real(RKIND), dimension(:,:), pointer :: fthetavpar_mu0_diag
  real(RKIND), dimension(:,:), pointer :: fthetavpar_mumax_diag
  real(RKIND), dimension(:,:), pointer :: fvparmu_diag
  !--> for <f>_FS(vpar,mu) around rpeak (if diag_level.ge.6)
  integer                              :: ir_fFS, ir_fFS_p15 
  integer                              :: ir_fFS_p30, ir_fFS_p40 
  integer                              :: ir_fFS_m15, ir_fFS_m30 
  integer                              :: ir_fFS_m40
  real(RKIND), dimension(:,:), pointer :: fFSavg_irp_vparmu_diag
  real(RKIND), dimension(:,:), pointer :: fFSavg_irp_p15_vparmu_diag
  real(RKIND), dimension(:,:), pointer :: fFSavg_irp_p30_vparmu_diag
  real(RKIND), dimension(:,:), pointer :: fFSavg_irp_p40_vparmu_diag
  real(RKIND), dimension(:,:), pointer :: fFSavg_irp_m15_vparmu_diag
  real(RKIND), dimension(:,:), pointer :: fFSavg_irp_m30_vparmu_diag
  real(RKIND), dimension(:,:), pointer :: fFSavg_irp_m40_vparmu_diag
  ! -> for <f>_FS(vpar,mu) at the position of the heat source
  integer                              :: irS
  real(RKIND), dimension(:,:), pointer :: fFSavg_irS_vparmu_diag

  !******************************
  contains
  !******************************

  !-------------------------------------------------
  ! Constructor 
  !-------------------------------------------------
  subroutine new_Pcross_section ( geom )
    use globals,    only: cible_istart, cible_jstart, cible_mu, dom_r, &
                          dom_theta, istart_id, jstart_id, Nbproc_r, &
                          Nbproc_theta, Ntheta, Nvpar, plocal_id, &
                          atleast1_source              
    use wrap_alloc_module
    
    type(geometry), intent(in) :: geom

!bmodif
!VG!    !*** for f(r,theta) cross-sections ***
!VG!    call glob_allocate(frtheta_loc_tmp,istart,iend,jstart,jend)
!VG!    if (plocal_id.eq.cible_mu) then
!VG!      call glob_allocate(frtheta_glob_tmp,0,dom_r-1,0,&
!VG!        Nbproc_r*(Ntheta+1),'frtheta_glob_tmp')
!VG!    else
!VG!      call glob_allocate(frtheta_glob_tmp,0,1,0,1, &
!VG!        'frtheta_glob_tmp')
!VG!    end if
!VG!    !*** for f(r,vpar) cross-sections ***
!VG!    call glob_allocate(frvpar_loc_tmp,istart,iend,0,Nvpar)
!VG!    if (jstart_id.eq.cible_jstart) then
!VG!      call glob_allocate(frvpar_glob_tmp,0,dom_r-1,0,&
!VG!        Nbproc_r*(Nvpar+1),'frvpar_glob_tmp')
!VG!    else
!VG!      call glob_allocate(frvpar_glob_tmp,0,1,0,&
!VG!        1,'frvpar_glob_tmp')
!VG!    end if
    ! -> solution to avoid temporary the problem 
    !    of MPI_GATHER instruction
    !*** for f(r,theta) cross-sections ***
    call FTR_ALLOC_2_ZERO(frtheta_loc_tmp,0,dom_r-1,0,dom_theta-1)
    if (plocal_id.eq.cible_mu) then
      call FTR_ALLOC_2_ZERO(frtheta_glob_tmp,0,dom_r-1,0,Nbproc_r*(Ntheta+1))
    else
      call FTR_ALLOC_2_ZERO(frtheta_glob_tmp,0,1,0,1)
    end if
    !*** for f(r,vpar) cross-sections ***
    call FTR_ALLOC_2_ZERO(frvpar_loc_tmp,0,dom_r-1,0,Nvpar)
    if (jstart_id.eq.cible_jstart) then
      call FTR_ALLOC_2_ZERO(frvpar_glob_tmp,0,dom_r-1,0,Nbproc_r*(Nvpar+1))
    else
      call FTR_ALLOC_2_ZERO(frvpar_glob_tmp,0,1,0,1)
    end if
    !*** for f(theta,vpar) cross-sections ***
    call FTR_ALLOC_2_ZERO(fthetavpar_loc_tmp,0,dom_theta-1,0,Nvpar)
    if (istart_id.eq.cible_istart) then
      call FTR_ALLOC_2_ZERO(fthetavpar_glob_tmp,0,dom_theta-1,0,Nbproc_theta*(Nvpar+1))
    else
      call FTR_ALLOC_2_ZERO(fthetavpar_glob_tmp,0,1,0,1)
    end if
!emodif

    !*** allocation of arrays for output saving ***
    !-> for Phi2D saving 
    call Pcross_section_phi2d_diag_init(geom)
    !-> for f2D saving 
    call Pcross_section_f2d_diag_init(geom)
  end subroutine new_Pcross_section

  !> allocation of arrays for output saving for Phi2D saving
  subroutine Pcross_section_phi2d_diag_init ( geom )

    use wrap_alloc_module

    type(geometry), intent(in) :: geom

    call FTR_ALLOC_2_ZERO(Phirth_diag, 0, geom%Nr, 0, geom%Ntheta)
    call FTR_ALLOC_2_ZERO(Phirphi_diag, 0, geom%Nr, 0, geom%Nphi)
    call FTR_ALLOC_2_ZERO(Phithphi_diag, 0, geom%Ntheta, 0, geom%Nphi)

  end subroutine Pcross_section_phi2d_diag_init

  !> allocation of arrays for output saving for f2D saving
  subroutine Pcross_section_f2d_diag_init ( geom )

    use globals,          only: diag_level
    use globals,          only: atleast1_source           
    use wrap_alloc_module

    type(geometry), intent(in) :: geom

    call FTR_ALLOC_2_ZERO(frtheta_v0_diag, 0, geom%Nr, 0, geom%Ntheta)
    call FTR_ALLOC_2_ZERO(frtheta_vmax_diag, 0, geom%Nr, 0, geom%Ntheta)
    call FTR_ALLOC_2_ZERO(fphivpar_mu0_diag, 0, geom%Nphi, 0, geom%Nvpar)
    call FTR_ALLOC_2_ZERO(fphivpar_mumax_diag, 0, geom%Nphi, 0,geom%Nvpar)
    call FTR_ALLOC_2_ZERO(frvpar_mu0_diag, 0, geom%Nr, 0, geom%Nvpar)
    call FTR_ALLOC_2_ZERO(frvpar_mumax_diag, 0, geom%Nr, 0, geom%Nvpar)
    call FTR_ALLOC_2_ZERO(fthetavpar_mu0_diag, 0, geom%Ntheta, 0,geom%Nvpar)
    call FTR_ALLOC_2_ZERO(fthetavpar_mumax_diag, 0, geom%Ntheta, 0,geom%Nvpar)
    call FTR_ALLOC_2_ZERO(fvparmu_diag, 0, geom%Nvpar, 0, geom%Nmu)
    !-> for <f>_FS(vpar,mu) saving
    if (diag_level.ge.6) then
      call FTR_ALLOC_2_ZERO(fFSavg_irp_vparmu_diag, 0, geom%Nvpar, 0,geom%Nmu)
      call FTR_ALLOC_2_ZERO(fFSavg_irp_p15_vparmu_diag, 0, geom%Nvpar, 0,geom%Nmu)
      call FTR_ALLOC_2_ZERO(fFSavg_irp_p30_vparmu_diag, 0, geom%Nvpar, 0,geom%Nmu)
      call FTR_ALLOC_2_ZERO(fFSavg_irp_p40_vparmu_diag, 0, geom%Nvpar, 0,geom%Nmu)
      call FTR_ALLOC_2_ZERO(fFSavg_irp_m15_vparmu_diag, 0, geom%Nvpar, 0,geom%Nmu)
      call FTR_ALLOC_2_ZERO(fFSavg_irp_m30_vparmu_diag, 0, geom%Nvpar, 0,geom%Nmu)
      call FTR_ALLOC_2_ZERO(fFSavg_irp_m40_vparmu_diag, 0, geom%Nvpar, 0,geom%Nmu)
    end if
    if (atleast1_source) then
      call FTR_ALLOC_2_ZERO(fFSavg_irS_vparmu_diag, 0, geom%Nvpar, 0,geom%Nmu)
    end if

  end subroutine Pcross_section_f2d_diag_init


  !-------------------------------------------------
  ! Destructor 
  !-------------------------------------------------
  subroutine delete_Pcross_section()
    use globals,    only: diag_level, atleast1_source
    use wrap_alloc_module

    call FTR_DEALLOC(frtheta_loc_tmp)
    call FTR_DEALLOC(frtheta_glob_tmp)
    call FTR_DEALLOC(frvpar_loc_tmp) 
    call FTR_DEALLOC(frvpar_glob_tmp)
    call FTR_DEALLOC(fthetavpar_loc_tmp) 
    call FTR_DEALLOC(fthetavpar_glob_tmp)

    !*** deallocation of arrays for output saving ***
    !-> for Phi2D saving 
    call FTR_DEALLOC(Phirth_diag)
    call FTR_DEALLOC(Phirphi_diag)
    call FTR_DEALLOC(Phithphi_diag)
    !->  for f2D saving 
    call FTR_DEALLOC(frtheta_v0_diag)
    call FTR_DEALLOC(frtheta_vmax_diag)
    call FTR_DEALLOC(fphivpar_mu0_diag)
    call FTR_DEALLOC(fphivpar_mumax_diag)
    call FTR_DEALLOC(frvpar_mu0_diag)
    call FTR_DEALLOC(frvpar_mumax_diag)
    call FTR_DEALLOC(fthetavpar_mu0_diag)
    call FTR_DEALLOC(fthetavpar_mumax_diag)
    call FTR_DEALLOC(fvparmu_diag)
    !-> for <f>_FS(vpar,mu) saving
    if (diag_level.ge.6) then
      call FTR_DEALLOC(fFSavg_irp_vparmu_diag)
      call FTR_DEALLOC(fFSavg_irp_p15_vparmu_diag)
      call FTR_DEALLOC(fFSavg_irp_p30_vparmu_diag)
      call FTR_DEALLOC(fFSavg_irp_p40_vparmu_diag)
      call FTR_DEALLOC(fFSavg_irp_m15_vparmu_diag)
      call FTR_DEALLOC(fFSavg_irp_m30_vparmu_diag)
      call FTR_DEALLOC(fFSavg_irp_m40_vparmu_diag)
    end if
    !-> for \int f dtheta saving
    if (atleast1_source) &
      call FTR_DEALLOC(fFSavg_irS_vparmu_diag)
  end subroutine delete_Pcross_section


  !---------------------------------------------- 
  ! computation of the 2D cross-sections of Phi
  !---------------------------------------------- 
  subroutine compute_Phi2DCS(geom, poiss)

    use globals, only: pInSpecies_id, diag_targ

    type(geometry), intent(in) :: geom
    type(poisson) , intent(in) :: poiss
    
    call Pcross_section_Phi2dcs_preprocess(geom)
    call Pcross_section_Phi2dcs_transfer(geom, poiss)
    if ( pInSpecies_id .eq. diag_targ(DIAG_PHI2D) ) then
      call Pcross_section_Phi2dcs_postprocess(geom)
    end if
    
  end subroutine compute_Phi2DCS


  !-------------------------------------------------------------------
  subroutine Pcross_section_Phi2dcs_preprocess ( geom )

    use globals,    only: ir_Phi, itheta_Phi, iphi_Phi

    type(geometry), intent(in) :: geom

    !*** coordinates of the cross section ***
    ir_Phi      = int(geom%Nr/2)
    itheta_Phi  = 0
    iphi_Phi    = 0

  end subroutine Pcross_section_Phi2dcs_preprocess
  !-------------------------------------------------------------------

  !-------------------------------------------------------------------
  subroutine Pcross_section_Phi2dcs_transfer ( geom, poiss )

    use globals,    only: iphi_Phi, ir_Phi, itheta_Phi, pInSpecies_id, &
                          diag_targ, Rarray_NrNphi, &
                          Rarray_NthetaNvparphi, mpi_comm_moments
    use MPIutils_module, &
                    only: comm_gather2d, dom_mapphi, iphistart, &
                          moments_mapphi
                          
    use mpiMod

    type(geometry), intent(in) :: geom
    type(poisson) , intent(in) :: poiss

    integer :: ir, itheta, iphi
    integer :: nbelts, itag, ierr, reqr
    integer, dimension(MPI_STATUS_SIZE) :: status

    itag = 1000
    nbelts = (geom%Nr+1)*(geom%Ntheta+1)

    if (pInSpecies_id .eq. diag_targ(DIAG_PHI2D)) then
      Phirth_diag = 0._RKIND !TODO[GL] is this required?
      call MPI_irecv(Phirth_diag(0, 0), nbelts, MPI_REAL8, &
                     MPI_ANY_SOURCE, itag, mpi_comm_moments, reqr, &
                     ierr)
    end if
    
    if ( moments_mapphi(iphi_Phi) == pInSpecies_id ) then
      call MPI_ssend(poiss%Phi_NrNthPNphi_Cworld(0, 0, iphi_Phi), &
                     nbelts, MPI_REAL8, diag_targ(DIAG_PHI2D), itag, &
                     mpi_comm_moments, ierr)
    end if
    
    if (pInSpecies_id .eq. diag_targ(DIAG_PHI2D)) then
      call MPI_wait(reqr, status, ierr)
    end if

    if (iphistart .ne. -1) then
    
      nbelts=dom_mapphi*(geom%Nr+1)
      do iphi = iphistart,iphistart+dom_mapphi-1
        do ir = 0,geom%Nr
          Rarray_NrNphi(ir,iphi-iphistart) = &
            poiss%Phi_NrNthPNphi_Cworld(ir,itheta_Phi,iphi)
        end do
      end do
      Phirphi_diag = 0._RKIND !TODO[GL] is this required?
      call comm_gather2d(Rarray_NrNphi, nbelts, 0, dom_mapphi, &
                      diag_targ(DIAG_PHI2D), Phirphi_diag, &
                      mpi_comm_moments)
      
      nbelts=dom_mapphi*(geom%Ntheta+1)
      do iphi = iphistart,iphistart+dom_mapphi-1
        do itheta = 0,geom%Ntheta
          Rarray_NthetaNvparphi(itheta,iphi-iphistart) = &
            poiss%Phi_NrNthPNphi_Cworld(ir_Phi,itheta,iphi)
        end do
      end do
      Phithphi_diag = 0._RKIND !TODO[GL] is this required?
      call comm_gather2d(Rarray_NthetaNvparphi, nbelts, 0, dom_mapphi, &
                      diag_targ(DIAG_PHI2D), Phithphi_diag, &
                      mpi_comm_moments)
    
    end if
    
  end subroutine Pcross_section_Phi2dcs_transfer
  !-------------------------------------------------------------------

  !-------------------------------------------------------------------
  subroutine Pcross_section_Phi2dcs_postprocess ( geom )

    type(geometry), intent(in) :: geom

    integer :: ir, itheta

    do ir = 0, geom%Nr
      Phirphi_diag(ir,geom%Nphi) = Phirphi_diag(ir,0) 
    end do
    do itheta = 0, geom%Ntheta
      Phithphi_diag(itheta,geom%Nphi) = Phithphi_diag(itheta,0) 
    end do

  end subroutine Pcross_section_Phi2dcs_postprocess
  !-------------------------------------------------------------------


  !------------------------------------------------------------- 
  ! distribution function computation in r and theta
  !  directions f for fixed values of phi and vpar
  !  ( i.e for iphi=phig(iphi_fix) and ivpar=vparg(ivpar_fix) )
  !------------------------------------------------------------- 
  subroutine compute_frtheta(f,iphi_fix,ivpar_fix,frtheta)
    use globals,      only: cible_mu, dom_r, dom_theta, mpi_comm_mu, &
                            Nbproc_r, Nbproc_theta, plocal_id
    use MPIutils_module, &
                      only: comm_gather2D
    type(fdistribu5d)            , intent(in)    :: f
    integer                      , intent(in)    :: iphi_fix
    integer                      , intent(in)    :: ivpar_fix
    real(RKIND), dimension(0:,0:), intent(inout) :: frtheta
    
    integer :: ierr
    integer :: ir, itheta
    integer :: p_r, p_theta
    integer :: nbelts

!bmodif
!VG!    do itheta = f%jstart,f%jend
!VG!      do ir = f%istart,f%iend
!VG!        frtheta_loc_tmp(ir,itheta) = &
!VG!          f%values(ir,itheta,iphi_fix,ivpar_fix)
!VG!      enddo
!VG!    enddo
!VG!    call MPI_GATHER(frtheta_loc_tmp,1, &
!VG!      mpitype_globbloc,frtheta_glob_tmp,1, &
!VG!      mpitype_globbloc,cible_mu,mpi_comm_mu,ierr)

    ! -> solution to avoid temporary the problem 
    !   of MPI_GATHER instruction
    do itheta = f%jstart,f%jend
      do ir = f%istart,f%iend
        frtheta_loc_tmp(ir-f%istart,itheta-f%jstart) = &
          f%values(ir,itheta,iphi_fix,ivpar_fix)
      enddo
    enddo
    nbelts = dom_r * dom_theta
    call comm_gather2D(frtheta_loc_tmp,nbelts,0,dom_theta, &
      cible_mu,frtheta_glob_tmp,mpi_comm_mu) 
!emodif

    if (plocal_id .eq. cible_mu) then
      do p_theta = 0,Nbproc_theta-1
        do p_r = 0,Nbproc_r-1
          do itheta = 0,dom_theta -1
            do ir = 0,dom_r-1
              frtheta(p_r*dom_r+ir,p_theta*dom_theta+itheta) = &
                frtheta_glob_tmp(ir, &
                (p_theta*Nbproc_r+p_r)*dom_theta+itheta)
            enddo
          enddo
        enddo
      enddo
      !*** periodic condition in theta ***
      do ir = 0,f%n1
        frtheta(ir,f%n2) = frtheta(ir,0)
      end do
    end if
  end subroutine compute_frtheta
  
  
  !----------------------------------------------------------- 
  ! distribution function computation in phi, v parallel 
  !  directions for fixed values of r and theta
  !  (i.e. for r=rg(ir_fix), theta=thetag(itheta_fix))
  ! Rk : it supposes that the processor contains the 
  !      point (ir_fix,itheta_fix)
  !-----------------------------------------------------------
  subroutine compute_fphivpar(f,ir_fix,itheta_fix,fphivpar)
    type(fdistribu5d)            , intent(in)    :: f
    integer                      , intent(in)    :: ir_fix
    integer                      , intent(in)    :: itheta_fix
    real(RKIND), dimension(0:,0:), intent(inout) :: fphivpar

    integer :: iphi, ivpar

    fphivpar = 0._RKIND
    do ivpar = 0,f%n4
      do iphi = 0,f%n3-1
        fphivpar(iphi,ivpar) = &
          f%values(ir_fix,itheta_fix,iphi,ivpar)
      enddo
    enddo
    fphivpar(f%n3,0:f%n4) = fphivpar(0,0:f%n4)
  end subroutine compute_fphivpar


  !-----------------------------------------------------------
  ! distribution function computation in v parallel, mu
  !  directions for fixed values of r, theta and phi
  !  (i.e. for r=rg(ir_fix), theta=thetag(itheta_fix) and
  !   phi=phig(iphi_fix))
  ! Rk : it supposes that the processor contains the 
  !      point (ir_fix,itheta_fix)
  !-----------------------------------------------------------
  subroutine compute_fvparmu(f,ir_fix,itheta_fix,iphi_fix,fvparmu)
    use globals,    only: cible, iend, istart, jend, jstart, &
                          mpi_comm_species, mu_id, Nmu, Nvpar, &
                          Pinspecies_id
    use wrap_alloc_module
    use mpiMod
    
    type(fdistribu5d)            , intent(in)    :: f
    integer                      , intent(in)    :: ir_fix
    integer                      , intent(in)    :: itheta_fix
    integer                      , intent(in)    :: iphi_fix
    real(RKIND), dimension(0:,0:), intent(inout) :: fvparmu
    
    integer :: ierr
    integer :: ivpar, nb_send, imu, nb_reqr, itag_vparmu

    real(RKIND), dimension(0:Nvpar)               :: fvpar_tmp
    integer    , dimension(MPI_STATUS_SIZE,0:Nmu) :: status
    integer    , dimension(0:Nmu)                 :: reqr
    integer    :: myreqs, nb_reqs

    call Write_dyn_balise_open('CPU_compute_fvparmu') !R4
    nb_send = Nvpar+1
    nb_reqr = Nmu+1
    nb_reqs = 0
    if (pInSpecies_id .eq. cible) then
      do imu = 0,Nmu
        itag_vparmu = 6000 + imu
        call MPI_IRECV(fvparmu(0,imu),nb_send, &
          MPI_REAL8,MPI_ANY_SOURCE,itag_vparmu, &
          mpi_comm_species,reqr(imu),ierr)
      enddo
    endif

    if ( (ir_fix.ge.istart).and.(ir_fix.le.iend).and. &
      (itheta_fix.ge.jstart).and.(itheta_fix.le.jend) ) then
      do ivpar = 0,Nvpar
        fvpar_tmp(ivpar) = &
          f%values(ir_fix,itheta_fix,iphi_fix,ivpar)
      end do
      itag_vparmu = 6000 + mu_id
      call MPI_ISEND(fvpar_tmp,nb_send,MPI_REAL8,cible, &
        itag_vparmu,mpi_comm_species,myreqs,ierr)
      nb_reqs = 1
    end if
    if (pInSpecies_id.eq.cible) &
      call MPI_WAITALL(Nmu+1, reqr, status, ierr)
    if (nb_reqs .ne. 0) then
       call MPI_WAIT(myreqs, status(:,0), ierr)
    end if
    call Write_dyn_balise_close('CPU_compute_fvparmu') !R4
  end subroutine compute_fvparmu


  !-----------------------------------------------------------
  ! computation of (vpar,mu) cross-section of the 
  !  flux surface average of f, i.e:
  !    \int f Js dthetadphi / \int Js dtheta dphi
  !  for a fixed radial value (i.e. for r=rg(ir_fix))
  !-----------------------------------------------------------
  subroutine compute_fFSavg_vparmu(f,geom, &
    ir_fix,intfdthetadphi_vparmu)
    use globals, only : istart, iend, jstart, jend, &
      cible, mu_id, Nbproc_theta, Nvpar, Nmu, Nphi, &
      mpi_comm_row, istart_id
    use globals, only : pInSpecies_id, pSpecies_id, &
      mpi_comm_species
    use coord_system_class, only : jacobian_space, &
      intdthetadphi_Js
    use wrap_alloc_module
    use MPIutils_module, &
                    only: ppbarrier
    use mpiMod
    type(fdistribu5d) , intent(in)    :: f
    type(geometry)    , intent(in)    :: geom
    integer           , intent(in)    :: ir_fix
    real(RKIND), &
      dimension(0:,0:), intent(inout) :: intfdthetadphi_vparmu
    
    integer     :: ierr
    integer     :: itheta, ivpar, iphi
    integer     :: nb_send, imu, nb_reqr, itag_vparmu
    real(RKIND) :: intfdthetadphi
    !--> for flux surface average computation
    real(RKIND) :: jacob_space_tmp
    real(RKIND) :: intdthetadphi_irfix_Js

    real(RKIND), dimension(0:Nvpar) :: fFSavg_vpar_loc
    real(RKIND), dimension(0:Nvpar) :: fFSavg_vpar_tmp
    integer    , dimension(0:Nmu)   :: reqr
    integer    , dimension(MPI_STATUS_SIZE,0:Nmu) :: status
    
    call Write_dyn_balise_open('CPU_compute_fFSavg_vparmu') !R4
    ! There is a problem when successive
    ! calls are performed to the subroutine.
    ! The receiver 'cible' can receive too many messages.
    ! The MPI barrier split clearly the MPI send/recv
    ! from two different calls to this subroutine.
    call ppbarrier()

    nb_send = Nvpar+1
    nb_reqr = Nmu+1

    fFSavg_vpar_loc(0:Nvpar) = 0._RKIND
    fFSavg_vpar_tmp(0:Nvpar) = 0._RKIND

    !--> \int Js dtheta dphi at ir = ir_fix
    intdthetadphi_irfix_Js = intdthetadphi_Js(ir_fix)

    if (Nbproc_theta.eq.1) then
      if (pInSpecies_id .eq. cible) then
        do imu = 0,Nmu
          itag_vparmu = 8000 + imu + pSpecies_id*Nmu
          call MPI_IRECV(intfdthetadphi_vparmu(0,imu),nb_send, &
            MPI_REAL8,MPI_ANY_SOURCE,itag_vparmu, &
            mpi_comm_species,reqr(imu),ierr)
        enddo
      endif
      if ( (ir_fix.ge.istart).and.(ir_fix.le.iend) ) then
        do ivpar = 0,Nvpar
          intfdthetadphi = 0._RKIND
          do itheta = 0,geom%Ntheta
            jacob_space_tmp = jacobian_space(ir_fix,itheta)
            do iphi = 0,geom%Nphi
              intfdthetadphi = intfdthetadphi + &
                f%values(ir_fix,itheta,iphi,ivpar) * &
                geom%coeff_intdtheta(itheta)* &
                geom%coeff_intdphi(iphi)* &
                jacob_space_tmp
            end do
          end do
          fFSavg_vpar_tmp(ivpar) = intfdthetadphi / &
            intdthetadphi_irfix_Js
        end do
        itag_vparmu = 8000 + mu_id + pSpecies_id*Nmu
        call MPI_SSEND(fFSavg_vpar_tmp,nb_send,MPI_REAL8,cible, &
          itag_vparmu,mpi_comm_species,ierr)
      end if
    else  ! (several processors in theta direction)
      if ( (ir_fix.ge.istart) .and. (ir_fix.le.iend) ) then
        do ivpar = 0,Nvpar
          intfdthetadphi = 0._RKIND
          do itheta = jstart,jend
            jacob_space_tmp = jacobian_space(ir_fix,itheta)
            do iphi = 0,Nphi
              intfdthetadphi = intfdthetadphi + &
                f%values(ir_fix,itheta,iphi,ivpar) * &
                geom%coeff_intdtheta(itheta)* &
                geom%coeff_intdphi(iphi)* &
                jacob_space_tmp
            end do
          end do
          fFSavg_vpar_loc(ivpar) = intfdthetadphi / &
            intdthetadphi_irfix_Js
        end do
      end if
      
      call MY_Reduce(fFSavg_vpar_loc,fFSavg_vpar_tmp, &
        nb_send,MPI_REAL8,MPI_SUM,0,mpi_comm_row,ierr)

      if (pInSpecies_id .eq. cible) then
        do imu = 0,Nmu
        itag_vparmu = 8000 + imu + pSpecies_id*Nmu
        call MPI_IRECV(intfdthetadphi_vparmu(0,imu),nb_send, &
          MPI_REAL8,MPI_ANY_SOURCE,itag_vparmu, &
          mpi_comm_species,reqr(imu),ierr)
      enddo
    endif

      if (istart_id.eq.0) then
        if ( (ir_fix.ge.istart).and.(ir_fix.le.iend) ) then
          itag_vparmu = 8000 + mu_id + pSpecies_id*Nmu
          call MPI_SSEND(fFSavg_vpar_tmp,nb_send,MPI_REAL8,cible, &
            itag_vparmu,mpi_comm_species,ierr)
        end if
      end if
    end if    

    if (pInSpecies_id.eq.cible) then
      call MPI_WAITALL(Nmu+1, reqr, status, ierr)
    end if
    ! There is a problem when successive
    ! calls are performed to the subroutine.
    ! The receiver 'cible' can receive too many messages.
    ! The MPI barrier split clearly the MPI send/recv
    ! from two different calls to this subroutine.
    call ppbarrier()
    call Write_dyn_balise_close('CPU_compute_fFSavg_vparmu') !R4
  end subroutine compute_fFSavg_vparmu


  !----------------------------------------------------------- 
  ! distribution function computation in r, v parallel 
  !  directions for fixed values of theta and phi
  !  (i.e theta=thetag(itheta_fix), phi=phig(iphi_fix)
  ! Rk : it supposes that the processor contains the 
  !      point ir_fix
  !-----------------------------------------------------------
  subroutine compute_frvpar(f,itheta_fix,iphi_fix,frvpar)
    use globals, only : istart, iend, jstart, jend, &
      jstart_id, cible_jstart, Nbproc_r, mpi_comm_column, dom_r
    use MPIutils_module, &
                    only: comm_gather2D
    use mpiMod
    type(fdistribu5d)            , intent(in)    :: f
    integer                      , intent(in)    :: itheta_fix
    integer                      , intent(in)    :: iphi_fix
    real(RKIND), dimension(0:,0:), intent(inout) :: frvpar

    integer :: ierr
    integer :: ir, ivpar, p_r
    integer :: nbelts

!bmodif
!VG!    do ivpar = 0,f%n4
!VG!      do ir = f%istart,f%iend
!VG!        frvpar_loc_tmp(ir,ivpar) = &
!VG!          f%values(ir,itheta_fix,iphi_fix,ivpar)
!VG!      end do
!VG!    end do
!VG!    call MPI_GATHER(frvpar_loc_tmp,1, &
!VG!      mpitype_globbloc_rvpar,frvpar_glob_tmp,1, &
!VG!      mpitype_globbloc_rvpar,cible_jstart,mpi_comm_column,ierr)
    ! -> solution to avoid temporary the problem 
    !of MPI_GATHER instruction
    do ivpar = 0,f%n4
      do ir = f%istart,f%iend
        frvpar_loc_tmp(ir-f%istart,ivpar) = &
          f%values(ir,itheta_fix,iphi_fix,ivpar)
      end do
    end do
    nbelts = dom_r*(f%n4+1)
    call comm_gather2D(frvpar_loc_tmp,nbelts,0,f%n4+1, &
      cible_jstart,frvpar_glob_tmp,mpi_comm_column) 
!emodif

    if (jstart_id .eq. cible_jstart) then
      do ivpar = 0,f%n4
        do p_r = 0,Nbproc_r-1
          do ir = 0,dom_r-1
            frvpar(p_r*dom_r+ir,ivpar) = &
              frvpar_glob_tmp(ir,p_r*(f%n4+1)+ivpar)
          enddo
        enddo
      enddo
    end if
  end subroutine compute_frvpar


  !----------------------------------------------------------- 
  ! distribution function computation in theta, v parallel 
  !  directions for fixed values of r and phi
  !  (i.e r=rg(ir_fix), phi=phig(iphi_fix))
  ! Rk : it supposes that the processor contains the 
  !      point itheta_fix
  !-----------------------------------------------------------
  subroutine compute_fthetavpar(f,ir_fix,iphi_fix,fthetavpar)
    use globals, only : istart, iend, jstart, jend, &
      istart_id, cible_istart, Nbproc_theta, &
      mpi_comm_row, dom_theta
    use MPIutils_module, &
                    only: comm_gather2D
    use mpiMod
    type(fdistribu5d)            , intent(in)    :: f
    integer                      , intent(in)    :: ir_fix
    integer                      , intent(in)    :: iphi_fix
    real(RKIND), dimension(0:,0:), intent(inout) :: fthetavpar

    integer :: ierr
    integer :: itheta, ivpar, p_theta
    integer :: nbelts

    do ivpar = 0,f%n4
      do itheta = f%jstart,f%jend
        fthetavpar_loc_tmp(itheta-f%jstart,ivpar) = &
          f%values(ir_fix,itheta,iphi_fix,ivpar)
      end do
    end do
    nbelts = dom_theta*(f%n4+1)
    call comm_gather2D(fthetavpar_loc_tmp,nbelts,0,f%n4+1, &
      cible_istart,fthetavpar_glob_tmp,mpi_comm_row) 

    if (istart_id .eq. cible_istart) then
      do ivpar = 0,f%n4
        do p_theta = 0,Nbproc_theta-1
          do itheta = 0,dom_theta-1
            fthetavpar(p_theta*dom_theta+itheta,ivpar) = &
              fthetavpar_glob_tmp(itheta,p_theta*(f%n4+1)+ivpar)
          enddo
        enddo
        fthetavpar(f%n2,ivpar) = fthetavpar(0,ivpar)
      enddo
    end if
  end subroutine compute_fthetavpar


  !------------------------------------------------------------- 
  ! Saving of the three cross-sections of f:
  !  - f(r,theta)  for iphi, ivpar, imu fixed
  !  - f(phi,vpar) for ir, itheta, imu fixed
  !  - f(r,vpar)   for itheta, iphi, imu fixed
  !-------------------------------------------------------------
  subroutine compute_cross_section(f,geom, &
    ir_diag,itheta_diag,iphi_diag, &
    ivpar_diag,imu_diag,frtheta,fphivpar,frvpar,fthetavpar)
    use globals, only : Nbproc_r, Nbproc_theta, Nbproc_mu, &
      plocal_id, mu_id, istart_id, jstart_id, &
      cible, cible_mu, cible_istart, cible_jstart, &
      istart, iend, jstart, jend, &
      Rarray1_NrNtheta, Rarray_NthetaNvparphi, &
      Rarray_NphiNvpar, Rarray_NrNvpar
    use globals, only : mpi_comm_species, pInSpecies_id
    use wrap_alloc_module
    use mpiMod
    type(fdistribu5d)            , intent(in)  :: f
    type(geometry)               , intent(in)  :: geom
    integer                      , intent(in)  :: ir_diag
    integer                      , intent(in)  :: itheta_diag
    integer                      , intent(in)  :: iphi_diag
    integer                      , intent(in)  :: ivpar_diag
    integer                      , intent(in)  :: imu_diag
    real(RKIND), dimension(0:,0:), intent(out) :: frtheta
    real(RKIND), dimension(0:,0:), intent(out) :: fphivpar
    real(RKIND), dimension(0:,0:), intent(out) :: frvpar
    real(RKIND), dimension(0:,0:), intent(out) :: fthetavpar

    real(RKIND), dimension(:,:), pointer :: frtheta_proc
    real(RKIND), dimension(:,:), pointer :: fphivpar_proc
    real(RKIND), dimension(:,:), pointer :: frvpar_proc
    real(RKIND), dimension(:,:), pointer :: fthetavpar_proc

    integer :: ierr
    integer :: nb_send1
    integer :: nb_send2
    integer :: nb_send3
    integer :: nb_send4
    integer :: itag_rtheta
    integer :: itag_rvpar
    integer :: itag_thetavpar
    integer :: itag_phivpar
    integer, dimension(MPI_STATUS_SIZE,1:4) :: status
    integer, dimension(1:4)                 :: reqr

    call Write_dyn_balise_open('CPU_cross_section') !R4
    frtheta_proc    => Rarray1_NrNtheta
    nb_send1        = (geom%Nr+1)*(geom%Ntheta+1)
    fphivpar_proc   => Rarray_NphiNvpar
    nb_send2        = (geom%Nphi+1)*(geom%Nvpar+1)
    frvpar_proc     => Rarray_NrNvpar
    nb_send3        = (geom%Nr+1)*(geom%Nvpar+1)
    fthetavpar_proc => Rarray_NthetaNvparphi
    nb_send4        = (geom%Ntheta+1)*(geom%Nvpar+1)

    !*** initialization of the tags ***
    itag_rtheta    = 11
    itag_phivpar   = 12
    itag_rvpar     = 13
    itag_thetavpar = 14
    reqr           = 0

    if ((mu_id.eq.imu_diag)) then
      call compute_frtheta(f,iphi_diag,ivpar_diag,frtheta_proc)
      if ( (ir_diag.ge.istart).and.(ir_diag.le.iend).and. &
        (itheta_diag.ge.jstart).and.(itheta_diag.le.jend) ) then
        call compute_fphivpar(f,ir_diag,itheta_diag,fphivpar_proc)
      endif
      if ((itheta_diag.ge.jstart).and.(itheta_diag.le.jend)) then
        call compute_frvpar(f,itheta_diag,iphi_diag,frvpar_proc)
      endif
      if ((ir_diag.ge.istart).and.(ir_diag.le.iend)) then
        call compute_fthetavpar(f,ir_diag,iphi_diag,fthetavpar_proc)
      endif
    endif

    if (pInSpecies_id.eq.cible) then
      call MPI_IRECV(frtheta(0,0),nb_send1, &
        MPI_REAL8,MPI_ANY_SOURCE,itag_rtheta, &
        mpi_comm_species,reqr(1),ierr)
      call MPI_IRECV(fphivpar(0,0),nb_send2, &
        MPI_REAL8,MPI_ANY_SOURCE,itag_phivpar, &
        mpi_comm_species,reqr(2),ierr)
      call MPI_IRECV(frvpar(0,0),nb_send3, &
        MPI_REAL8,MPI_ANY_SOURCE,itag_rvpar, &
        mpi_comm_species,reqr(3),ierr)
      call MPI_IRECV(fthetavpar(0,0),nb_send4, &
        MPI_REAL8,MPI_ANY_SOURCE,itag_thetavpar, &
        mpi_comm_species,reqr(4),ierr)
    end if

    !*** cross-sections for mu=imu_diag ***
    if ((mu_id.eq.imu_diag)) then
       ! -> f(r,theta) for iphi, ivpar, imu fixed
       if (plocal_id.eq.cible_mu) then
         call MPI_SSEND(frtheta_proc(0,0),nb_send1, &
           MPI_REAL8,cible,itag_rtheta,mpi_comm_species,ierr)
       end if
       ! -> f(phi,vpar) for ir, itheta, imu fixed
       if ( (ir_diag.ge.istart).and.(ir_diag.le.iend).and. &
            (itheta_diag.ge.jstart).and.(itheta_diag.le.jend) ) then
         call MPI_SSEND(fphivpar_proc(0,0),nb_send2, &
           MPI_REAL8,cible,itag_phivpar,mpi_comm_species,ierr)
       end if
       ! -> f(r,vpar) for itheta, iphi, imu fixed
       if ((itheta_diag.ge.jstart).and.(itheta_diag.le.jend)) then
         if (jstart_id.eq.cible_jstart) then
           call MPI_SSEND(frvpar_proc(0,0),nb_send3, &
             MPI_REAL8,cible,itag_rvpar,mpi_comm_species,ierr)
         end if
       end if
       ! -> f(theta,vpar) for ir, iphi, imu fixed
       if ((ir_diag.ge.istart).and.(ir_diag.le.iend)) then
         if (istart_id.eq.cible_istart) then
           call MPI_SSEND(fthetavpar_proc(0,0),nb_send4, &
             MPI_REAL8,cible,itag_thetavpar,mpi_comm_species,ierr)
         end if
       end if
    end if

    if (pInSpecies_id.eq.cible) &
      call MPI_WAITALL(4, reqr, status, ierr)
    call Write_dyn_balise_close('CPU_cross_section') !R4
  end subroutine compute_cross_section


  !------------------------------------------------------------- 
  ! Saving of the cross-sections associated to imu=Nmu/4,
  ! corresponding to trapped particles.
  !-------------------------------------------------------------
  subroutine compute_ftrapped(f,geom)
    use globals, only : ir_f2D_trapped, itheta_f2D_trapped, &
      iphi_f2D_trapped, ivpar_f2D_trapped, imu_f2D_trapped
    use wrap_alloc_module

    type(fdistribu5d), intent(in) :: f
    type(geometry)   , intent(in) :: geom

    call Write_dyn_balise_open('CPU_compute_ftrapped') !R4
    !*** cross-sections for mu=max(mu)/2 ***
    ir_f2D_trapped     = int(geom%Nr/2)
    itheta_f2D_trapped = 0
    iphi_f2D_trapped   = int(geom%Nphi/3)
    ivpar_f2D_trapped  = geom%ivpar0
    imu_f2D_trapped    = int(geom%Nmu/2)
    call compute_cross_section(f,geom, &
      ir_f2D_trapped,itheta_f2D_trapped, &
      iphi_f2D_trapped,ivpar_f2D_trapped,imu_f2D_trapped, &
      frtheta_v0_diag,fphivpar_mumax_diag, &
      frvpar_mumax_diag,fthetavpar_mumax_diag)

    !*** f(v parallel, mu) cross-sections for theta=0 ***
    call compute_fvparmu(f,ir_f2D_trapped,itheta_f2D_trapped, &
      iphi_f2D_trapped,fvparmu_diag)
    call Write_dyn_balise_close('CPU_compute_ftrapped') !R4
  end subroutine compute_ftrapped


  !------------------------------------------------------------- 
  ! Saving of the cross-sections associated to mu=0.,
  ! corresponding to passing particles.
  !-------------------------------------------------------------
  subroutine compute_fpassing(f,geom)
    use globals     , only :     nb_vth0, &
      ir_f2D_passing, itheta_f2D_passing, iphi_f2D_passing, &
      ivpar_f2D_passing, imu_f2D_passing
    use utils_module, only : locate, locate_nonequidistant
    use wrap_alloc_module
    type(fdistribu5d), intent(in) :: f
    type(geometry)   , intent(in) :: geom

    real(RKIND) :: vpar_diag, mu_diag
    character(LEN=50) :: subp

    call Write_dyn_balise_open('CPU_compute_fpassing') !R4
    !*** cross-sections for mu=0. ***
    ir_f2D_passing     = int(geom%Nr/2)
    itheta_f2D_passing = 0
    iphi_f2D_passing   = int(geom%Nphi/3)
    vpar_diag          = 3._RKIND*geom%vparg(geom%Nvpar)/ &
      real(nb_vth0)    
    mu_diag            = max(0._RKIND,geom%mug(0))
    subp = " compute_fpassing "//char(0)
    call locate(vpar_diag,geom%vparg,geom%Nvpar, &
      geom%dvpar,subp,ivpar_f2D_passing)
    subp = " compute_fpassing "//char(0)
    call locate_nonequidistant(mu_diag,geom%mug,geom%Nmu, &
      subp,imu_f2D_passing)
    call compute_cross_section(f,geom, &
      ir_f2D_passing,itheta_f2D_passing, &
      iphi_f2D_passing,ivpar_f2D_passing,imu_f2D_passing, &
      frtheta_vmax_diag,fphivpar_mu0_diag, &
      frvpar_mu0_diag,fthetavpar_mu0_diag)
    call Write_dyn_balise_close('CPU_compute_fpassing') !R4
  end subroutine compute_fpassing


  !------------------------------------------------------------- 
  ! Saving of the cross-sections in (vpar,mu) of 
  !  - the distribution function (at ir=ir_fix, itheta=0 and
  !       iphi=iphi_fix)
  !  - \int f dtheta dphi for two different radial positions, 
  !    the first one corresponding to the radial location of
  !    the source (ir=ir_fixS) and the second one to 
  !    the radial location of the fast particle source 
  !    (ir=ir_fixSfp)
  !-------------------------------------------------------------
  subroutine compute_vparmu_crosssection(geom,init_srce,f)
    use globals, only : diag_level, atleast1_source               
    use wrap_alloc_module
    use utils_module, &
                    only: locate
    type(geometry)    , intent(in) :: geom
    type(init_sources), intent(in) :: init_srce        
    type(fdistribu5d) , intent(in) :: f

    integer     :: ir_fix, itheta_fix, iphi_fix
    integer     :: ir_fixS                          
    real(RKIND) :: rho_fFS
    
    character(LEN=50) :: subp

    call Write_dyn_balise_open('CPU_compute_vparmu_crosssection') !R4
    ir_fix     = int(geom%Nr/2)
    itheta_fix = 0
    iphi_fix   = int(geom%Nphi/3)

    !*** f(vpar, mu) cross-section for theta=0 ***
    call compute_fvparmu(f,ir_fix,itheta_fix, &
      iphi_fix,fvparmu_diag)

    !*** <f>_FS(vpar,mu) cross-sections  ***
    !***  for different radial position  ***
    if (diag_level.ge.6) then
      rho_fFS = 0.5_RKIND
      !--> radial location of the middle of the 
      !-->  simulation box
      subp = " compute_vparmu_crosssection "//char(0)
      call locate(geom%Lr*rho_fFS+geom%rg(0), &
        geom%rg,geom%Nr,geom%dr,subp,ir_fFS)
      ir_fFS  = ir_fFS + 1
      call compute_fFSavg_vparmu(f,geom,ir_fFS, &
        fFSavg_irp_vparmu_diag)

      !--> Drho = +- 0.15 around radial middle
      !    of the simulation box
      call locate(geom%Lr*(rho_fFS+0.15_RKIND) + &
        geom%rg(0),geom%rg,geom%Nr,geom%dr,subp,ir_fFS_p15)
      ir_fFS_p15  = ir_fFS_p15 + 1
      call compute_fFSavg_vparmu(f,geom,ir_fFS_p15, &
        fFSavg_irp_p15_vparmu_diag)

      call locate(geom%Lr*(rho_fFS-0.15_RKIND) + &
        geom%rg(0),geom%rg,geom%Nr,geom%dr,subp,ir_fFS_m15)
      ir_fFS_m15  = ir_fFS_m15 + 1
      call compute_fFSavg_vparmu(f,geom,ir_fFS_m15, &
        fFSavg_irp_m15_vparmu_diag)

      !--> Drho = +- 0.30 around radial middle
      !    of the simulation box
      call locate(geom%Lr*(rho_fFS+0.3_RKIND) + &
        geom%rg(0),geom%rg,geom%Nr,geom%dr,subp,ir_fFS_p30)
      ir_fFS_p30  = ir_fFS_p30 + 1
      call compute_fFSavg_vparmu(f,geom,ir_fFS_p30, &
        fFSavg_irp_p30_vparmu_diag)

      call locate(geom%Lr*(rho_fFS-0.3_RKIND) + &
        geom%rg(0),geom%rg,geom%Nr,geom%dr,subp,ir_fFS_m30)
      ir_fFS_m30  = ir_fFS_m30 + 1
      call compute_fFSavg_vparmu(f,geom,ir_fFS_m30, &
        fFSavg_irp_m30_vparmu_diag)

      !--> Drho = +- 0.40 around radial middle
      !    of the simulation box
      call locate(geom%Lr*(rho_fFS+0.4_RKIND) + &
        geom%rg(0),geom%rg,geom%Nr,geom%dr,subp,ir_fFS_p40)
      ir_fFS_p40  = ir_fFS_p40 + 1
      call compute_fFSavg_vparmu(f,geom,ir_fFS_p40, &
        fFSavg_irp_p40_vparmu_diag)

      call locate(geom%Lr*(rho_fFS-0.4_RKIND) + &
        geom%rg(0),geom%rg,geom%Nr,geom%dr,subp,ir_fFS_m40)
      ir_fFS_m40  = ir_fFS_m40 + 1
      call compute_fFSavg_vparmu(f,geom,ir_fFS_m40, &
        fFSavg_irp_m40_vparmu_diag)
    end if

    !*** (vpar, mu) cross-sections of  ***
    !*** the flux surface average of f ***
    if (atleast1_source) then
      !--> radial location of the source ?
      subp = " compute_vparmu_crosssection "//char(0)
      call locate(geom%Lr*init_srce%rhoS + &
        geom%rg(0),geom%rg,geom%Nr,geom%dr,subp,ir_fixS)
      ir_fixS = ir_fixS + 1
      call compute_fFSavg_vparmu(f,geom,ir_fixS, &
        fFSavg_irS_vparmu_diag)
      !--> keep the radial indexes for output saving 
      irS = ir_fixS
    end if
    call Write_dyn_balise_close('CPU_compute_vparmu_crosssection') !R4
  end subroutine compute_vparmu_crosssection

end module Pcross_section_module
