!**************************************************************
!  Copyright Euratom-CEA
!  Authors : 
!     Virginie Grandgirard (virginie.grandgirard@cea.fr)
!     Chantal Passeron (chantal.passeron@cea.fr)
!     Guillaume Latu (guillaume.latu@cea.fr)
!     Xavier Garbet (xavier.garbet@cea.fr)
!     Philippe Ghendrih (philippe.ghendrih@cea.fr)
!     Yanick Sarazin (yanick.sarazin@cea.fr)
!  
!  This code GYSELA (for GYrokinetic SEmi-LAgrangian) 
!  is a 5D gyrokinetic global full-f code for simulating 
!  the plasma turbulence in a tokamak.
!  
!  This software is governed by the CeCILL-B license 
!  under French law and abiding by the rules of distribution 
!  of free software.  You can  use, modify and redistribute 
!  the software under the terms of the CeCILL-B license as 
!  circulated by CEA, CNRS and INRIA at the following URL
!  "http://www.cecill.info". 
!**************************************************************

#include "config-F.inc"
#include "code_instrumentation/debug.inc"

!----------------------------------------------------------------- 
! file : diag3D.F90
! date : 18/04/2005
!  Computes and save the fluid moments of the distribution
!   function of particles:
!     -> J0M0      = \int J0.f Jv dvpar dmu  
!     -> J0M0_trap = \int_{trapped} J0.f Jv dvpar dmu  
!     -> J0M1      = \int J0.f vpar Jv dvpar dmu 
!     -> J0M2      = \int J0.f (0.5*vpar^2+mu B) Jv dvpar dmu 
!
! Rk: The trapped constraint is
!      vpar^2/mu < 2 ( B(r,theta=pi)-B(r,theta) )
!  
!  Computes and save the fluid moments of the distribution
!   function of guiding-centers:
!     -> M2_parallel = \int f*(0.5*vpar^2) Jv dvpar dmu 
!     -> M2_perp     = \int f*(mu*B) Jv dvpar dmu 
!
!  Rk: Jv is the jacobian in velocity, i.e : 
!   Jv = 2*pi*Bstar(r,theta)
!
!  Save the electric potential Phi(r,theta,phi)
!--------------------------------------------------------------- 
module diag3D_module
  use prec_const

  implicit none
  
  include "gyrokinetic/Bstar_inline.h"
  include "gyrokinetic/velocities_inline.h"
  !******************************
  contains
  !******************************
#include "gyrokinetic/Bstar_inline.F90"
#include "gyrokinetic/velocities_inline.F90"

  !--------------------------------------------------------
  ! Compute and save 3D fluid moments:
  !  - J0M0      (density of guiding-centers)
  !  - J0M0_trap (density of trapped guiding-centers)
  !  - J0M1
  !  - J0M2
  !  - M2_parallel
  !  - M2_perp
  ! Save 3D electric potential Phi
  !--------------------------------------------------------
  subroutine compute_save_diag3D(Siter_time, &
    SPhi3D_diag_OK,SPhi3D_nb_diag,SPhi3D_time_save, &
    SMoments3D_diag_OK,SMoments3D_nb_diag,SMoments3D_time_save, &
    geom,init_prof,init_magnet, &
    init_curr,J0,f,sJ0ftmp,sPhi)
! This directive of optimisation  must be enable on Poincare. 
! Without it, from the -O2 level of optimisation the program 
! cannot be executed until the end.
!dir$ optimize:1
    use globals
    use OMPutils_module
    use clock_module
    use integration_module, only : compute_vpar_integral_colloc, &
      compute_omp_vpar_integral_CS
    use gyroaverage_class
    use poisson_class
    use fdistribu5d_class
    use init_magnetic_class
    use init_current_class

    use wrap_alloc_module
#ifndef NO_HDF5
    use resu3D_saving_module
    use f5D_saving_module
#endif

    use Debug_module

    real(RKIND)        , intent(in)    :: Siter_time
    logical            , intent(in)    :: SPhi3D_diag_OK
    integer            , intent(in)    :: SPhi3D_nb_diag
    real(RKIND)        , intent(in)    :: SPhi3D_time_save
    logical            , intent(in)    :: SMoments3D_diag_OK
    integer            , intent(in)    :: SMoments3D_nb_diag
    real(RKIND)        , intent(in)    :: SMoments3D_time_save
    type(geometry)     , intent(in)    :: geom
    type(init_profile) , intent(in)    :: init_prof
    type(init_magnetic), intent(in)    :: init_magnet
    type(init_current) , intent(in)    :: init_curr
    type(J0operator)   , intent(inout) :: J0
    type(fdistribu5d)  , intent(inout) :: f
    type(fdistribu5d)  , intent(inout) :: sJ0ftmp
    real(RKIND), &
      dimension(:,:,:) , pointer       :: sPhi

    integer, parameter :: nbmoments = 8
    integer     :: ir, itheta, iphi, ivpar, imu, iphiloc
    real(RKIND) :: mu_value
    real(RKIND) :: ri, Bij
    real(RKIND) :: Bi_thetapi, trap_constraint
    real(RKIND) :: vpar, fval, J0fval, energy
    real(RKIND) :: coeff_intdmu
    real(RKIND) :: jacob_space_ij, jacob_vel_tmp

    real(RKIND),    dimension(:)    , pointer :: scoefvpar
    real(RKIND),    dimension(:)    , pointer :: fvperp2
    real(RKIND),    dimension(:)    , pointer :: fvpar2
    real(RKIND),    dimension(:)    , pointer :: J0f
    real(RKIND),    dimension(:)    , pointer :: J0f_trap
    real(RKIND),    dimension(:)    , pointer :: J0fv
    real(RKIND),    dimension(:)    , pointer :: J0fE

    real(RKIND),    dimension(:,:)  , pointer :: intfvperp2_dvpar
    real(RKIND),    dimension(:,:)  , pointer :: intfvpar2_dvpar
    real(RKIND),    dimension(:,:)  , pointer :: intJ0f_dvpar
    real(RKIND),    dimension(:,:)  , pointer :: intJ0f_trap_dvpar
    real(RKIND),    dimension(:,:)  , pointer :: intJ0fv_dvpar
    real(RKIND),    dimension(:,:)  , pointer :: intJ0fE_dvpar
    real(RKIND),    dimension(:,:)  , pointer :: J0f_rtheta
    complex(CKIND), dimension(:,:)  , pointer :: Acomp
    real(RKIND),    dimension(:,:,:), pointer :: M2par_loc
    real(RKIND),    dimension(:,:,:), pointer :: M2perp_loc
    real(RKIND),    dimension(:,:,:), pointer :: J0M0_loc
    real(RKIND),    dimension(:,:,:), pointer :: J0M0_trap_loc
    real(RKIND),    dimension(:,:,:), pointer :: J0M1_loc
    real(RKIND),    dimension(:,:,:), pointer :: J0M2_loc

    ! -> variable for parallelization
    integer :: no_operation
    integer :: ierr
    integer :: itag_phi, j
    integer :: i, nbreqr, local_id, dep, base
    integer :: global_id, startbuf
    integer :: l_istart, l_iend, l_jstart, l_jend
    integer :: nbelements, tid

    ! -> variables for time perfomance analysis
    integer(TIMEPREC) :: tdeb, tfin
    real(RKIND)       :: tcomm1, tcomm2, tsave1
    real(RKIND)       :: cexport, tivpar, timu
    ! -> variable used to export small 3D files in parallel
#ifdef LCHD    
    type(mfdstruct)   :: mfds
#endif    
#ifndef NO_HDF5
    call clck_time(tdeb)
    call clck_time(bclock_diag3D)

    !*** computing and saving of 3D moments ***
    if (SMoments3D_diag_OK)  then
      J0M0_loc      => Rarray1_NrNthetaDomphi
      J0M0_trap_loc => Rarray2_NrNthetaDomphi
      J0M1_loc      => Rarray3_NrNthetaDomphi
      J0M2_loc      => Rarray4_NrNthetaDomphi
      M2perp_loc    => Rarray5_NrNthetaDomphi
      M2par_loc     => Rarray6_NrNthetaDomphi

      do tid = 1, Nbthread
         call FTR_ALLOC_2_ZERO(J0%omp_diag(tid)%val,1,J0%N1,1,J0%N2)
         call FTR_ALLOC_2_ZERO(J0%omp_ldiag(tid)%val,1,J0%N1-1,1,J0%N2)
         call FTR_ALLOC_2_ZERO(J0%omp_udiag(tid)%val,1,J0%N1-1,1,J0%N2)
      enddo

      tid        = 1
      call FTR_ALLOC_2_ZERO(Romp1_1Nrp1_1Nthetap1(tid)%val,1,Nr+1,1,Ntheta+1)
      call FTR_ALLOC_2_ZERO(Comp1_1Nrp1_1Ntheta(tid)%val,1,Nr+1,1,Ntheta)
      Acomp      => Comp1_1Nrp1_1Ntheta(tid)%val
      J0f_rtheta => Romp1_1Nrp1_1Nthetap1(tid)%val

      select case ( transpose_version )
      case ( TRANSPOSE_SENDIRECV )
         call FTR_ALLOC_4_ZERO(f4D_send,0,dom_r-1,0,dom_theta-1,0,dom_phi-1,0,dom_vpar-1)
         call FTR_ALLOC_4_ZERO(f4D_recv,0,dom_r-1,0,dom_theta-1,0,dom_phi-1,0,dom_vpar-1)
      case ( TRANSPOSE_ALLTOALLW )
         ! nothing specific to allocate
      case default
         print *, pglobal_id, 'Invalid value for transpose_version', &
              transpose_version
         call MPI_abort(MPI_COMM_WORLD, 1, ierr)
      end select

      PRINT_CKSUM(f%values)

      !*** initialisation to 0 of the integrals ***
      call pptranspose_forward(f%values,timing=.false.)

      PRINT_CKSUM(f4D_transp)

      do ivpar = lstart,lend
         do iphi = kstart,kend
            do itheta = 0,f%n2
               do ir = 0,f%n1
                  J0f_rtheta(ir+1,itheta+1) = &
                       f4D_transp(ir,itheta,iphi,ivpar)
               end do
            end do
            !--> computation of J0.f
            call omp_compute_gyrofunction_2D(J0,geom, &
                 mu_id,Acomp,J0f_rtheta,tid)
            do itheta = 0,f%n2
               do ir = 0,f%n1
                  f4D_transp(ir,itheta,iphi,ivpar) = &
                       J0f_rtheta(ir+1,itheta+1)
               end do
            end do
         end do
      end do

      do tid = 1, Nbthread
         call FTR_DEALLOC(J0%omp_diag(tid)%val)
         call FTR_DEALLOC(J0%omp_ldiag(tid)%val)
         call FTR_DEALLOC(J0%omp_udiag(tid)%val)
      enddo

      tid = 1
      call FTR_DEALLOC(Romp1_1Nrp1_1Nthetap1(tid)%val)
      call FTR_DEALLOC(Comp1_1Nrp1_1Ntheta(tid)%val)

      !->  f4D_transp -> J0ftmp%values
      call pptranspose_backward(sJ0ftmp%values,timing=.false.)

      select case ( transpose_version )
      case ( TRANSPOSE_SENDIRECV )
         call FTR_DEALLOC(f4D_send)
         call FTR_DEALLOC(f4D_recv)
      case ( TRANSPOSE_ALLTOALLW )
         ! nothing specific to allocate
      case default
         print *, pglobal_id, 'Invalid value for transpose_version', &
              transpose_version
         call MPI_abort(MPI_COMM_WORLD, 1, ierr)
      end select
      
      mu_value  = f%mu_value

      !*** Parallel communication to prepare the      ***
      !***  working processors                        ***
      !***  (i.e processors having a phi value)       *** 
      !***  to receive the partial values of          ***
      !***  several moments of f                      ***
      !*** (RK: The processors are always prepared    ***
      !***   to the receiving before the sending)     ***
#ifdef _OPENMP
!$OMP PARALLEL default(none) &
!$OMP private(tid,ir,itheta,iphi,ivpar, &
!$OMP  scoefvpar,fvperp2,fvpar2,J0f,J0f_trap,J0fv,J0fE, &
!$OMP  ri,Bij,Bi_thetapi,trap_constraint, &
!$OMP  vpar,energy,fval,J0fval, &
!$OMP  jacob_space_ij,jacob_vel_tmp) &
!$OMP shared(sJ0ftmp,f,geom,init_curr,init_magnet,mu_value, &
!$OMP  jacobian_space,Romp1_0Nvpar,Romp2_0Nvpar, &
!$OMP  Romp3_0Nvpar,Romp4_0Nvpar,Romp5_0Nvpar,Romp6_0Nvpar, &
!$OMP  Romp_scoefvpar,Rarray_PNrPNthetaNphi_nbM8)
!$OMP BARRIER
      tid = 1+omp_get_thread_num()
#else
      tid = 1
#endif

      scoefvpar => Romp_scoefvpar(tid)%val
      J0f       => Romp1_0Nvpar(tid)%val
      J0f_trap  => Romp2_0Nvpar(tid)%val
      J0fv      => Romp3_0Nvpar(tid)%val
      J0fE      => Romp4_0Nvpar(tid)%val
      fvpar2    => Romp5_0Nvpar(tid)%val
      fvperp2   => Romp6_0Nvpar(tid)%val

      
!$OMP DO SCHEDULE(STATIC)
      do iphi = 0,f%n3-1
        do itheta = f%jstart,f%jend
          do ir = f%istart,f%iend
            Bij              = init_magnet%B_norm(ir,itheta)
            Bi_thetapi       = init_magnet%B_at_thetapi(ir)
            ri               = geom%rg(ir)
            jacob_space_ij   = jacobian_space(ir,itheta)
            do ivpar = 0,f%n4
              vpar            = geom%vparg(ivpar)
              !-> jacobian_velocity = 2*pi*Bstar
              call compute_jacobian_velocity(geom,init_magnet, &
                init_curr,ir,itheta,ivpar,jacob_vel_tmp)
              energy          = 0.5_RKIND*vpar*vpar+mu_value*Bij
              trap_constraint = sqrt(2._RKIND*mu_value * &
                (Bi_thetapi-Bij))
              fval            = f%values(ir,itheta,iphi,ivpar)
              J0fval          = sJ0ftmp%values(ir,itheta,iphi,ivpar)
              J0f(ivpar)      = J0fval*jacob_vel_tmp
              if (abs(vpar).lt.trap_constraint) then
                J0f_trap(ivpar) = J0f(ivpar)
              else
                J0f_trap(ivpar) = 0._RKIND
              end if
              J0fv(ivpar)     = J0fval*vpar*jacob_vel_tmp
              J0fE(ivpar)     = J0fval*energy*jacob_vel_tmp
              fvpar2(ivpar)   = fval*(0.5_RKIND*vpar*vpar) * &
                jacob_vel_tmp
              fvperp2(ivpar)  = fval*(mu_value*Bij)*jacob_vel_tmp
            end do

            !*** computation of the integrals in vpar direction ***
            !--> \int J0.f Jv dvpar 
            call compute_vpar_integral_colloc(J0f(0:f%n4),geom, &
              Rarray_PNrPNthetaNphi_nbM8(0,ir,itheta,iphi))
            !--> \int J0.f Jv dvpar for trapped particles
            call compute_vpar_integral_colloc(J0f_trap(0:f%n4), &
              geom,Rarray_PNrPNthetaNphi_nbM8(1,ir,itheta,iphi))
            !--> \int (J0.f)*vpar Jv dvpar
            call compute_vpar_integral_colloc(J0fv(0:f%n4),geom, &
              Rarray_PNrPNthetaNphi_nbM8(2,ir,itheta,iphi))
            !--> \int J0.f*(0.5*vpar^2+mu*B) Jv dvpar
            call compute_vpar_integral_colloc(J0fE(0:f%n4),geom, &
              Rarray_PNrPNthetaNphi_nbM8(3,ir,itheta,iphi))
            !--> \int f*(0.5*vpar^2) Jv dvpar
            call compute_vpar_integral_colloc(fvpar2(0:f%n4),geom, &
              Rarray_PNrPNthetaNphi_nbM8(4,ir,itheta,iphi))
            !--> \int f*(mu*B) Jv dvpar
            call compute_vpar_integral_colloc(fvperp2(0:f%n4), &
              geom,Rarray_PNrPNthetaNphi_nbM8(5,ir,itheta,iphi))
          enddo
        enddo
      end do
!$OMP END DO
!$OMP BARRIER
!$OMP END PARALLEL
      call clck_time(tfin)
      call clck_ldiff(tdeb,tfin,tivpar)

      call ppbarrier()
      call clck_time(tdeb)

      do tid = 1, Nbthread
         call FTR_ALLOC_2(Romp1_1Nrp1_1Nthetap1(tid)%val,1,Nr+1,1,Ntheta+1)
         call FTR_ALLOC_2(Romp2_1Nrp1_1Nthetap1(tid)%val,1,Nr+1,1,Ntheta+1)
         call FTR_ALLOC_2(Romp3_1Nrp1_1Nthetap1(tid)%val,1,Nr+1,1,Ntheta+1)
         call FTR_ALLOC_2(Romp4_1Nrp1_1Nthetap1(tid)%val,1,Nr+1,1,Ntheta+1)
         call FTR_ALLOC_2(Romp5_1Nrp1_1Nthetap1(tid)%val,1,Nr+1,1,Ntheta+1)
         call FTR_ALLOC_2(Romp6_1Nrp1_1Nthetap1(tid)%val,1,Nr+1,1,Ntheta+1)

         call FTR_ALLOC_2_ZERO(Comp1_1Nrp1_1Ntheta(tid)%val,1,Nr+1,1,Ntheta)
      enddo

      do tid = 0, MAX(Nmu,Nbthread)
         call FTR_ALLOC_2(Romp9_1Nrp1_1Nthetap1(tid)%val,1,Nr+1,1,Ntheta+1)
         call FTR_ALLOC_2(Romp10_1Nrp1_1Nthetap1(tid)%val,1,Nr+1,1,Ntheta+1)
         call FTR_ALLOC_2(Romp11_1Nrp1_1Nthetap1(tid)%val,1,Nr+1,1,Ntheta+1)
         call FTR_ALLOC_2(Romp12_1Nrp1_1Nthetap1(tid)%val,1,Nr+1,1,Ntheta+1)
         call FTR_ALLOC_2(Romp13_1Nrp1_1Nthetap1(tid)%val,1,Nr+1,1,Ntheta+1)
         call FTR_ALLOC_2(Romp14_1Nrp1_1Nthetap1(tid)%val,1,Nr+1,1,Ntheta+1)
      enddo

      
      call write_param("nbmoments", nbmoments)
      call FTR_ALLOC_1_ZERO(RRecvBufMPI_E,0,nbmoments*dom_r*dom_theta*Nbproc_loc*(Nmu+1)*dom_mapphi)

      nbreqr = 0
      if (iphistart .ne. -1) then
        do imu = 0, Nmu
          iphiloc = 0
          do iphi = 0,f%n3-1
            if (moments_mapphi(iphi) .eq. pInSpecies_id) then
              do local_id = 0,Nbproc_loc-1
                global_id = imu*Nbproc_loc + local_id
                startbuf  = dom_theta * &
                  (local_id+Nbproc_loc*(imu+(Nmu+1)*iphiloc)) 
                nbelements = nbmoments * dom_r * dom_theta
                itag_phi   = 20 + iphi + pSpecies_id*f%n3
                call MPI_IRECV( &
                  RRecvBufMPI_E(startbuf*nbmoments*dom_r), &
                  nbelements,MPI_REAL8,global_id,itag_phi, &
                  mpi_comm_species,moments_reqr(nbreqr),ierr)
                nbreqr = nbreqr + 1
              enddo
              iphiloc = iphiloc + 1
            endif
          enddo
        enddo
      endif
      call clck_time(tfin)
      call clck_ldiff(tdeb,tfin,tcomm1)

      !*** Parallel communication to send the five   ***
      !***  integrals in vpar                        ***
      !***  [ \int f*(vpar^2) Jv dvpar,              ***
      !***    \int f*(mu*B) Jv dvpar,                ***
      !***    \int (J0.f) Jv dvpar,                  ***
      !***    \int (J0.f)*vpar Jv dvpar,             ***
      !***    \int (J0.f)*(vpar^2+mu*B) Jv dvpar, ***
      !*** The parallelism is such that              ***
      !***   (r=block,theta=block,phi=*)             ***
      !*** RK : These 5 integrals are saved in an    ***
      !***      unique array                         ***
      call clck_time(tdeb)
      nbelements = nbmoments*dom_r*dom_theta
      do iphi = 0,f%n3-1
        itag_phi = 20 + iphi + pSpecies_id*f%n3
        call MPI_SEND( &
          Rarray_PNrPNthetaNphi_nbM8(0,istart,jstart,iphi), &
          nbelements,MPI_REAL8,moments_mapphi(iphi),itag_phi, &
          mpi_comm_species,ierr)
      enddo
      if (nbreqr .ne. 0) then
        call MPI_WAITALL(nbreqr, moments_reqr(0:nbreqr-1), &
                moments_status(:,0:nbreqr-1), ierr)
      end if
      call clck_time(tfin)
      call clck_ldiff(tdeb,tfin,tcomm2)

      !*** Reconstruction of the complete (r,theta) domain    ***
      !***  RK:- Parallelization in phi on all                ***
      !***       the processors                               ***
      !***     - iphistart.eq.-1 for some processors          ***
      !***       if the total number of processors is         ***
      !***       greater than the number of phi values (Nphi) ***
      call clck_time(tdeb)
      timu = 0.

      if (iphistart .ne. -1) then
#ifdef _OPENMP
!$OMP PARALLEL private(tid,iphi,itheta,ir,ivpar, &
!$OMP iphiloc,imu,coeff_intdmu,local_id,base,dep, &
!$OMP l_iend,l_istart,startbuf,i,j,l_jstart,l_jend,Acomp, &
!$OMP intJ0f_dvpar,intJ0f_trap_dvpar, &
!$OMP intJ0fv_dvpar,intJ0fE_dvpar, &
!$OMP intfvpar2_dvpar,intfvperp2_dvpar) default(shared) 
!$OMP BARRIER
        tid = 1+omp_get_thread_num()
#else
        tid = 1
#endif
        Acomp             => Comp1_1Nrp1_1Ntheta(tid)%val
        intJ0f_dvpar      => Romp1_1Nrp1_1Nthetap1(tid)%val
        intJ0f_trap_dvpar => Romp2_1Nrp1_1Nthetap1(tid)%val
        intJ0fv_dvpar     => Romp3_1Nrp1_1Nthetap1(tid)%val
        intJ0fE_dvpar     => Romp4_1Nrp1_1Nthetap1(tid)%val
        intfvpar2_dvpar   => Romp5_1Nrp1_1Nthetap1(tid)%val
        intfvperp2_dvpar  => Romp6_1Nrp1_1Nthetap1(tid)%val

        do iphi = 0,geom%Nphi-1
          if (moments_mapphi(iphi) .eq. pInSpecies_id) then
            iphiloc = iphi - iphistart
            do imu = 0,Nmu
              ! -> This if block replaces a OMP DO 
              !     SCHEDULE(static,1) directive
              !  (this for avoiding problem on BULL architecture)
              if (tid-1 .eq. mod(imu,Nbthread)) then
                coeff_intdmu = geom%coeff_intdmu(imu)
                do itheta = 0,f%n2
                  do ir = 0,f%n1
                    intJ0f_dvpar(1+ir,1+itheta)      = 0._RKIND
                    intJ0f_trap_dvpar(1+ir,1+itheta) = 0._RKIND
                    intJ0fv_dvpar(1+ir,1+itheta)     = 0._RKIND
                    intJ0fE_dvpar(1+ir,1+itheta)     = 0._RKIND
                    intfvpar2_dvpar(1+ir,1+itheta)   = 0._RKIND
                    intfvperp2_dvpar(1+ir,1+itheta)  = 0._RKIND
                  end do
                end do
                do local_id = 0,Nbproc_loc-1
                  base     = (local_id/Nbproc_r)
                  dep      = mod(local_id,Nbproc_r)              
                  l_istart = dep *  dom_r
                  l_iend   = l_istart + dom_r - 1
                  l_jstart = base * dom_theta
                  l_jend   = l_jstart + dom_theta - 1
                  startbuf = dom_theta * (local_id + &
                    Nbproc_loc * (imu + (Nmu+1) * iphiloc)) 
                  do itheta = l_jstart, l_jend
                    do ir = l_istart, l_iend
                      intJ0f_dvpar(1+ir,1+itheta)      = &
                        RRecvBufMPI_E( &
                        nbmoments*(ir-l_istart)+0+ &
                        (startbuf+itheta-l_jstart)*dom_r*nbmoments) 
                      intJ0f_trap_dvpar(1+ir,1+itheta) = &
                        RRecvBufMPI_E( &
                        nbmoments*(ir-l_istart)+1+ &
                        (startbuf+itheta-l_jstart)*dom_r*nbmoments) 
                      intJ0fv_dvpar(1+ir,1+itheta)     = &
                        RRecvBufMPI_E( &
                        nbmoments*(ir-l_istart)+2+ &
                        (startbuf+itheta-l_jstart)*dom_r*nbmoments)
                      intJ0fE_dvpar(1+ir,1+itheta)     = &
                        RRecvBufMPI_E( &
                        nbmoments*(ir-l_istart)+3+ &
                        (startbuf+itheta-l_jstart)*dom_r*nbmoments)
                      intfvpar2_dvpar(1+ir,1+itheta)   = &
                        RRecvBufMPI_E( &
                        nbmoments*(ir-l_istart)+4+ &
                        (startbuf+itheta-l_jstart)*dom_r*nbmoments)
                      intfvperp2_dvpar(1+ir,1+itheta)  = &
                        RRecvBufMPI_E( &
                        nbmoments*(ir-l_istart)+5+ &
                        (startbuf+itheta-l_jstart)*dom_r*nbmoments)
                    enddo
                  enddo
                enddo
                !*** periodic boundary conditions in theta ***
                do ir = 0,f%n1
                  intJ0f_dvpar(1+ir,1+f%n2)      = &
                    intJ0f_dvpar(1+ir,1+0)
                  intJ0f_trap_dvpar(1+ir,1+f%n2) = &
                    intJ0f_trap_dvpar(1+ir,1+0)
                  intJ0fv_dvpar(1+ir,1+f%n2)     = &
                    intJ0fv_dvpar(1+ir,1+0)
                  intJ0fE_dvpar(1+ir,1+f%n2)     = &
                    intJ0fE_dvpar(1+ir,1+0)
                  intfvpar2_dvpar(1+ir,1+f%n2)   = &
                    intfvpar2_dvpar(1+ir,1+0)
                  intfvperp2_dvpar(1+ir,1+f%n2)  = &
                    intfvperp2_dvpar(1+ir,1+0)
                end do

                !*** computation of the integrals in mu ***
                do itheta = 0,f%n2-1
                  do ir = 0,f%n1
                    !-> J0M0 = \int (\int_trap J0.f Jv dvpar) dmu
                    Romp9_1Nrp1_1Nthetap1(imu)%val(1+ir,1+itheta) &
                      = intJ0f_dvpar(1+ir,1+itheta)*coeff_intdmu 
                    !-> J0M0_trap = \int (\int J0.f Jv dvpar) dmu
                    !->  for trapped particles
                    Romp10_1Nrp1_1Nthetap1(imu)%val(1+ir,1+itheta) &
                      = intJ0f_trap_dvpar(1+ir,1+itheta) * &
                      coeff_intdmu 
                    !-> J0M1 = \int (\int J0.f*vpar Jv dvpar) dmu
                    Romp11_1Nrp1_1Nthetap1(imu)%val(1+ir,1+itheta) &
                      = intJ0fv_dvpar(1+ir,1+itheta)*coeff_intdmu
                    !-> J0M2 = \int (\int 
                    !        J0.f*(0.5*vpar^2+mu*B) Jv dvpar) dmu 
                    Romp12_1Nrp1_1Nthetap1(imu)%val(1+ir,1+itheta) &
                      = intJ0fE_dvpar(1+ir,1+itheta)*coeff_intdmu
                    !-> M2vpar = \int \int 
                    !            f*(0.5*vpar^2) Jv dvpar dmu 
                    Romp13_1Nrp1_1Nthetap1(imu)%val(1+ir,1+itheta) &
                      = intfvpar2_dvpar(1+ir,1+itheta)*coeff_intdmu
                    !-> M2perp = \int \int f*(mu*B) Jv dvpar dmu 
                    Romp14_1Nrp1_1Nthetap1(imu)%val(1+ir,1+itheta) &
                      = intfvperp2_dvpar(1+ir,1+itheta)*coeff_intdmu
                  end do
                end do
              end if
            end do
!$OMP BARRIER
!$OMP MASTER 
            do itheta = 0,geom%Ntheta
              do ir = 0,geom%Nr
                J0M0_loc(ir,itheta,iphi)      = 0._RKIND
                J0M0_trap_loc(ir,itheta,iphi) = 0._RKIND
                J0M1_loc(ir,itheta,iphi)      = 0._RKIND
                J0M2_loc(ir,itheta,iphi)      = 0._RKIND
                M2par_loc(ir,itheta,iphi)     = 0._RKIND
                M2perp_loc(ir,itheta,iphi)    = 0._RKIND
              end do
            end do

            !*** computation of the integrals in mu ***
            do j = 0,Nmu
              do itheta = 0,f%n2-1
                do ir = 0,f%n1
                  !-> J0M0 = \int (\int J0.f Jv dvpar) dmu
                  J0M0_loc(ir,itheta,iphi) = &
                    J0M0_loc(ir,itheta,iphi) + &
                    Romp9_1Nrp1_1Nthetap1(j)%val(1+ir,1+itheta)
                  !-> J0M0 = \int (\int J0.f Jv dvpar) dmu
                  !->  for trapped particle
                  J0M0_trap_loc(ir,itheta,iphi) = &
                    J0M0_trap_loc(ir,itheta,iphi) + &
                    Romp10_1Nrp1_1Nthetap1(j)%val(1+ir,1+itheta)
                  !-> J0M1 = \int (\int J0.f*vpar Jv dvpar) dmu
                  J0M1_loc(ir,itheta,iphi) = &
                    J0M1_loc(ir,itheta,iphi) + &
                    Romp11_1Nrp1_1Nthetap1(j)%val(1+ir,1+itheta)
                  ! -> J0M2 = \int [\int J0.f *       
                  !    (0.5*vpar^2+mu*B) Jv dvpar] dmu 
                  J0M2_loc(ir,itheta,iphi) = &
                    J0M2_loc(ir,itheta,iphi) + &
                    Romp12_1Nrp1_1Nthetap1(j)%val(1+ir,1+itheta)
                  !-> M2_par = \int [\int f * 
                  !                  (0.5*vpar^2) Jv dvpar] dmu
                  M2par_loc(ir,itheta,iphi) = &
                    M2par_loc(ir,itheta,iphi) + &
                    Romp13_1Nrp1_1Nthetap1(j)%val(1+ir,1+itheta)
                  !-> M2_perp = \int [\int f * (mu*B) Jv dvpar] dmu
                  M2perp_loc(ir,itheta,iphi) = &
                    M2perp_loc(ir,itheta,iphi) + &
                    Romp14_1Nrp1_1Nthetap1(j)%val(1+ir,1+itheta)
                end do
              end do
            end do

            !*** periodic conditions in theta ***
            do ir = 0,f%n1
              J0M0_loc(ir,f%n2,iphi)      = J0M0_loc(ir,0,iphi)
              J0M0_trap_loc(ir,f%n2,iphi) = J0M0_trap_loc(ir,0,iphi)
              J0M1_loc(ir,f%n2,iphi)      = J0M1_loc(ir,0,iphi)
              J0M2_loc(ir,f%n2,iphi)      = J0M2_loc(ir,0,iphi)
              M2par_loc(ir,f%n2,iphi)     = M2par_loc(ir,0,iphi)
              M2perp_loc(ir,f%n2,iphi)    = M2perp_loc(ir,0,iphi)
            end do
!$OMP END MASTER
          endif
          no_operation = 0
!$OMP BARRIER
        end do
!$OMP END PARALLEL
        call clck_time(tfin)
        call clck_ldiff(tdeb,tfin,timu)
        
        !*** distribution of the six moments  ***
        !***   J0M0(r=*,theta=*,phi=*),       ***
        !***   J0M0_trap(r=*,theta=*,phi=*),  ***
        !***   J0M1, J0M2, M2perp, M2vpar     ***
        call clck_time(tdeb)
        nbelements = dom_mapphi*(Nr+1)*(Ntheta+1)

        !*** Write on disk the six moments in parallel ***
#ifdef LCHD
        !*** Initialize the external library LCHD ***
        call HDF5_LCHD_init(mfds,geom,6)
        !*** Launch 6 channels of communication to scatter the six moments ***
        !*** onto multiple processors                                      ***
        call HDF5_LCHD_start(mfds,'J0M0',SMoments3D_nb_diag,J0M0_loc)  
        call HDF5_LCHD_start(mfds,'J0M0_trap',SMoments3D_nb_diag,J0M0_trap_loc)
        call HDF5_LCHD_start(mfds,'J0M1',SMoments3D_nb_diag,J0M1_loc)
        call HDF5_LCHD_start(mfds,'J0M2',SMoments3D_nb_diag,J0M2_loc)
        call HDF5_LCHD_start(mfds,'M2par',SMoments3D_nb_diag,M2par_loc)
        call HDF5_LCHD_start(mfds,'M2perp',SMoments3D_nb_diag,M2perp_loc)
        !*** Wait the end of the previous communications ***
        !*** and write several binary or HDF5 files      ***
        call HDF5_LCHD_waitall_and_write(mfds,SMoments3D_time_save,geom)
        !*** Close all opened files ***
        call HDF5_LCHD_close(mfds)
#else
        !*** Export the six moments on disk, one after the other ***
        call HDF5_parallel3D_saving('J0M0',SMoments3D_nb_diag, &
          SMoments3D_time_save,geom,J0M0_loc) 
        call HDF5_parallel3D_saving('J0M0_trap', &
          SMoments3D_nb_diag,SMoments3D_time_save, &
          geom,J0M0_trap_loc) 
        call HDF5_parallel3D_saving('J0M1',SMoments3D_nb_diag, &
          SMoments3D_time_save,geom,J0M1_loc)
        call HDF5_parallel3D_saving('J0M2',SMoments3D_nb_diag, &
          SMoments3D_time_save,geom,J0M2_loc)
        call HDF5_parallel3D_saving('M2par',SMoments3D_nb_diag, &
          SMoments3D_time_save,geom,M2par_loc)
        call HDF5_parallel3D_saving('M2perp',SMoments3D_nb_diag, &
          SMoments3D_time_save,geom,M2perp_loc)
#endif
      end if  ! if (iphistart .ne. -1) then

      do tid = 1, Nbthread
         call FTR_DEALLOC(Romp1_1Nrp1_1Nthetap1(tid)%val)
         call FTR_DEALLOC(Romp2_1Nrp1_1Nthetap1(tid)%val)
         call FTR_DEALLOC(Romp3_1Nrp1_1Nthetap1(tid)%val)
         call FTR_DEALLOC(Romp4_1Nrp1_1Nthetap1(tid)%val)
         call FTR_DEALLOC(Romp5_1Nrp1_1Nthetap1(tid)%val)
         call FTR_DEALLOC(Romp6_1Nrp1_1Nthetap1(tid)%val)

         call FTR_DEALLOC(Comp1_1Nrp1_1Ntheta(tid)%val)
      enddo

      do tid = 0, MAX(Nmu,Nbthread)
         call FTR_DEALLOC(Romp9_1Nrp1_1Nthetap1(tid)%val)
         call FTR_DEALLOC(Romp10_1Nrp1_1Nthetap1(tid)%val)
         call FTR_DEALLOC(Romp11_1Nrp1_1Nthetap1(tid)%val)
         call FTR_DEALLOC(Romp12_1Nrp1_1Nthetap1(tid)%val)
         call FTR_DEALLOC(Romp13_1Nrp1_1Nthetap1(tid)%val)
         call FTR_DEALLOC(Romp14_1Nrp1_1Nthetap1(tid)%val)
      enddo

      call FTR_DEALLOC(RRecvBufMPI_E)
	  
    end if  ! if (SMoments3D_diag_OK)

    !*** Saving of Phi3D ***
    if (SPhi3D_diag_OK) then
#ifdef LCHD
       call HDF5_LCHD_init(mfds,geom,1)
       call HDF5_LCHD_start(mfds,'Phi',SPhi3D_nb_diag,sPhi)
       call HDF5_LCHD_waitall_and_write(mfds,SPhi3D_time_save,geom)
       call HDF5_LCHD_close(mfds)
#else
       call HDF5_parallel3D_saving('Phi',SPhi3D_nb_diag, &
            SPhi3D_time_save,geom,sPhi)
#endif
    end if
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tsave1)
    
    call clck_time(eclock_diag3D)
    call clck_diff(bclock_diag3D,eclock_diag3D,&
      global_time_diag3D)

#ifdef TIMER    
    if (pglobal_id .eq. 0) then
      write(6,'(I4,A,F10.4,A,F10.4,A,&
       &F10.4,A,F10.4,A,F10.4,A,F10.4)') pglobal_id, &
        " Temps comp_save_3D ivpar ", tivpar," ivperp ", &
        timu," comm1 ",tcomm1, " comm2 ", tcomm2, &
        " savef ",tsave1
    end if
#endif
#endif !! #ifndef NO_HDF5
  end subroutine compute_save_diag3D
end module diag3D_module
