!**************************************************************
!  Copyright Euratom-CEA
!  Authors : 
!     Virginie Grandgirard (virginie.grandgirard@cea.fr)
!     Chantal Passeron (chantal.passeron@cea.fr)
!     Guillaume Latu (guillaume.latu@cea.fr)
!     Xavier Garbet (xavier.garbet@cea.fr)
!     Philippe Ghendrih (philippe.ghendrih@cea.fr)
!     Yanick Sarazin (yanick.sarazin@cea.fr)
!  
!  This code GYSELA (for GYrokinetic SEmi-LAgrangian) 
!  is a 5D gyrokinetic global full-f code for simulating 
!  the plasma turbulence in a tokamak.
!  
!  This software is governed by the CeCILL-B license 
!  under French law and abiding by the rules of distribution 
!  of free software.  You can  use, modify and redistribute 
!  the software under the terms of the CeCILL-B license as 
!  circulated by CEA, CNRS and INRIA at the following URL
!  "http://www.cecill.info". 
!**************************************************************

#include "config-F.inc"

!-------------------------------------------------
! file : diagnostics.F90
! date : 17/07/2000
!  used for all physic output computation
!------------------------------------------------
module diagnostics_module
  use prec_const
  use MPIutils_module
  use wrap_alloc_module
  use geometry_class
  use poisson_class
  implicit none

  type :: diagnostics
     !-> saving option or no of the different diagnostics
     logical :: Phi3D_saving
     logical :: Moments3D_saving
     logical :: f5D_saving
     logical :: Sp_saving
     !-> logicals equal to true if the time iteration
     !->   corresponds to a time saving
     logical :: diag_OK
     logical :: Phi3D_diag_OK
     logical :: Moments3D_diag_OK
     logical :: f5D_diag_OK
     logical :: Sp_diag_OK
     !-> item of the diagnostic saving
     integer :: nb_diag           
     integer :: Phi3D_nb_diag     
     integer :: Moments3D_nb_diag
     integer :: f5D_nb_diag
     integer :: Sp_nb_diag
     !-> elapsed time between two savings
     real(RKIND) :: dt_diag
     real(RKIND) :: Phi3D_dt_diag
     real(RKIND) :: Moments3D_dt_diag
     real(RKIND) :: f5D_dt_diag
     real(RKIND) :: Sp_dt_diag
     !-> time value of the next saving
     real(RKIND) :: time_save
     real(RKIND) :: Phi3D_time_save
     real(RKIND) :: Moments3D_time_save
     real(RKIND) :: f5D_time_save
     real(RKIND) :: Sp_time_save
  end type diagnostics

  !******************************
  contains
  !******************************

  !-----------------------------------------------------
  ! Initialises all the variables used to follow 
  !  the saving of all the different diagnostics 
  !-----------------------------------------------------
  subroutine new_diagnostics(SPhi3D_saving,SMoments3D_saving, &
    Sf5D_saving, SSp_saving, &
    Sdt_diag,SPhi3D_dt_diag, SMoments3D_dt_diag,Sf5D_dt_diag, &
    SSp_dt_diag, Sdiag)
    logical          , intent(in)    :: SPhi3D_saving
    logical          , intent(in)    :: SMoments3D_saving
    logical          , intent(in)    :: Sf5D_saving
    real(RKIND)      , intent(in)    :: Sdt_diag
    real(RKIND)      , intent(in)    :: SPhi3D_dt_diag
    real(RKIND)      , intent(in)    :: SMoments3D_dt_diag
    real(RKIND)      , intent(in)    :: Sf5D_dt_diag
    logical          , intent(in)    :: SSp_saving
    real(RKIND)      , intent(in)    :: SSp_dt_diag
    type(diagnostics), intent(inout) :: Sdiag
    
    !--> initialisation of the logical values
    Sdiag%Phi3D_saving     = SPhi3D_saving
    Sdiag%Moments3D_saving = SMoments3D_saving
    Sdiag%f5D_saving       = Sf5D_saving
    Sdiag%Sp_saving        = SSp_saving
    
    !--> initialisation of the elapsed times between two savings
    Sdiag%dt_diag           = Sdt_diag
    Sdiag%Phi3D_dt_diag     = SPhi3D_dt_diag
    Sdiag%Moments3D_dt_diag = SMoments3D_dt_diag
    Sdiag%f5D_dt_diag       = Sf5D_dt_diag
    Sdiag%Sp_dt_diag        = SSp_dt_diag
  end subroutine new_diagnostics


  !-----------------------------------------------------
  ! Compute the values of the next times of savings
  !  at the beginning of the run
  !-----------------------------------------------------
  subroutine init_diagnostics_time(Sinit_mode,Stime_save,Sdiag)
    use globals, only : restart
    logical          , intent(in)    :: Sinit_mode
    real(RKIND)      , intent(in)    :: Stime_save
    type(diagnostics), intent(inout) :: Sdiag

    Sdiag%diag_OK           = .false.
    Sdiag%Phi3D_diag_OK     = .false.
    Sdiag%Moments3D_diag_OK = .false.
    Sdiag%f5D_diag_OK       = .false.
    Sdiag%Sp_diag_OK        = .false.

    if (Sinit_mode) then
      !--> for general diagnostics
      Sdiag%diag_OK   = .true.
      Sdiag%time_save = Stime_save
      !--> for Phi3D diagnostics
      if (Sdiag%Phi3D_saving) &
        Sdiag%Phi3D_diag_OK   = .true.
      Sdiag%Phi3D_time_save = Stime_save
      !--> for Moments3D diagnostics
      if (Sdiag%Moments3D_saving) &
        Sdiag%Moments3D_diag_OK   = .true.
      Sdiag%Moments3D_time_save = Stime_save
      !--> for f5D diagnostics
      if (Sdiag%f5D_saving) &
        Sdiag%f5D_diag_OK   = .true.
      Sdiag%f5D_time_save = Stime_save
      if (Sdiag%Sp_saving) &
        Sdiag%Sp_diag_OK    = .true.
      Sdiag%Sp_time_save  = Stime_save
    else
      Sdiag%time_save           = Sdiag%dt_diag * &
        (floor(Stime_save/Sdiag%dt_diag)+1._RKIND)
      Sdiag%Phi3D_time_save     = Sdiag%Phi3D_dt_diag * &
        (floor(Stime_save/Sdiag%Phi3D_dt_diag)+1._RKIND)
      Sdiag%Moments3D_time_save = Sdiag%Moments3D_dt_diag * &
        (floor(Stime_save/Sdiag%Moments3D_dt_diag)+1._RKIND)
      Sdiag%f5D_time_save       = Sdiag%f5D_dt_diag * &
        (floor(Stime_save/Sdiag%f5D_dt_diag)+1._RKIND)
      Sdiag%Sp_time_save  = Sdiag%Sp_dt_diag * &
        (floor(Stime_save/Sdiag%Sp_dt_diag)+1._RKIND)    
    end if
  end subroutine init_diagnostics_time


  !-----------------------------------------------------
  ! Increment the diagnostic savings if required at the
  !  current time iteration
  !-----------------------------------------------------
  subroutine check_diagnostics(Siter_time,Sdiag)
    use utils_module, only : approx_equal
    use globals     , only : outputproc, uout_res, &
      file_name_res
    use Io_Unit_module, only: IU_open, IU_close
    real(RKIND)      , intent(in)    :: Siter_time
    type(diagnostics), intent(inout) :: Sdiag

    !*** open file results ***
    if (pglobal_id.eq.outputproc) then
      call IU_open(unit=uout_res,file=file_name_res,status='OLD',&
        position='APPEND', form='FORMATTED')
    end if

    Sdiag%diag_OK           = .false.
    Sdiag%Phi3D_diag_OK     = .false.
    Sdiag%Moments3D_diag_OK = .false.
    Sdiag%f5D_diag_OK       = .false.
    Sdiag%Sp_diag_OK        = .false.

    if (Siter_time.le.0._RKIND) then
      Sdiag%diag_OK   = .true.
      if (Sdiag%Phi3D_saving)     Sdiag%Phi3D_diag_OK     = .true.
      if (Sdiag%Moments3D_saving) Sdiag%Moments3D_diag_OK = .true.
      if (Sdiag%f5D_saving)       Sdiag%f5D_diag_OK       = .true.
      if (Sdiag%Sp_saving)        Sdiag%Sp_diag_OK        = .true.
    else
      !-> check if it is time for 01 - 1D and 2D diagnostic savings ?
      if (approx_equal(Siter_time,Sdiag%time_save)) &
        Sdiag%diag_OK   = .true.

      !-> check if it is time for Phi3D diagnostic saving ?
      if ( Sdiag%Phi3D_saving .and. &
        approx_equal(Siter_time,Sdiag%Phi3D_time_save)) then
        if (pglobal_id.eq.outputproc) &
          write (uout_res,*) '---> saving of Phi3D'
        Sdiag%Phi3D_diag_OK   = .true.
      end if

      !-> check if it is time for Moments3D diagnostic saving ?
      if ( Sdiag%Moments3D_saving .and. &
        approx_equal(Siter_time,Sdiag%Moments3D_time_save)) then
        if (pglobal_id.eq.outputproc) &
          write (uout_res,*) '---> saving of 3D moments'
        Sdiag%Moments3D_diag_OK   = .true.
      end if

      !-> check if it is time for f5D diagnostic saving ?
      if ( Sdiag%f5D_saving .and. &
        approx_equal(Siter_time,Sdiag%f5D_time_save)) then
        if (pglobal_id.eq.outputproc) &
          write (uout_res,*) '---> saving of f5D'
        Sdiag%f5D_diag_OK   = .true.
      end if

      !-> check if it is time for Spectral diagnostic saving ?
      if ( Sdiag%Sp_saving .and. &
        approx_equal(Siter_time,Sdiag%Sp_time_save)) then
        if (pglobal_id.eq.outputproc) &
          write (uout_res,*) '---> saving of spectral diag'
        Sdiag%Sp_diag_OK   = .true.
      end if
    end if

    !*** close file results ***
    if (pglobal_id.eq.outputproc) &
      call IU_close(uout_res)
  end subroutine check_diagnostics


  !-----------------------------------------------------
  ! Increment the diagnostic savings if required at the
  !  current time iteration
  !-----------------------------------------------------
  subroutine increment_diagnostics(Siter_time,Sdeltat,Sdiag)
    use utils_module, only : approx_equal
    real(RKIND)      , intent(in)    :: Siter_time
    real(RKIND)      , intent(in)    :: Sdeltat
    type(diagnostics), intent(inout) :: Sdiag

    !-> check if it is time for 01 - 1D and 2D diagnostic savings ?
    if (approx_equal(Siter_time,Sdiag%time_save)) then
      Sdiag%nb_diag   = Sdiag%nb_diag + 1
      Sdiag%time_save = Sdiag%time_save + Sdiag%dt_diag
    end if

    !-> check if it is time for Phi3D diagnostic saving ?
    if ( Sdiag%Phi3D_saving .and. &
      approx_equal(Siter_time,Sdiag%Phi3D_time_save)) then
      Sdiag%Phi3D_nb_diag   = Sdiag%Phi3D_nb_diag + 1
      Sdiag%Phi3D_time_save = Sdiag%Phi3D_time_save + &
        Sdiag%Phi3D_dt_diag
    end if

    !-> check if it is time for Moments3D diagnostic saving ?
    if ( Sdiag%Moments3D_saving .and. &
      approx_equal(Siter_time,Sdiag%Moments3D_time_save)) then
      Sdiag%Moments3D_nb_diag   = Sdiag%Moments3D_nb_diag + 1
      Sdiag%Moments3D_time_save = Sdiag%Moments3D_time_save + &
        Sdiag%Moments3D_dt_diag
    end if

    !-> check if it is time for f5D diagnostic saving ?
    if ( Sdiag%f5D_saving .and. &
      approx_equal(Siter_time,Sdiag%f5D_time_save)) then
      Sdiag%f5D_nb_diag   = Sdiag%f5D_nb_diag + 1
      Sdiag%f5D_time_save = Sdiag%f5D_time_save + &
        Sdiag%f5D_dt_diag
    end if

    !-> check if it is time for Spectral diagnostic saving ?
    if ( Sdiag%Sp_saving .and. &
      approx_equal(Siter_time,Sdiag%Sp_time_save)) then
      Sdiag%Sp_nb_diag   = Sdiag%Sp_nb_diag + 1
      Sdiag%Sp_time_save = Sdiag%Sp_time_save + &
        Sdiag%Sp_dt_diag
    end if
  end subroutine increment_diagnostics


  !---------------------------------------------- 
  ! array saving for all the diagnostics in time 
  !---------------------------------------------- 
  subroutine compute_diagnostics(Siter_time,Sdiag,Sdeltat, &
    geom,coord_sys,init_prof,init_magnet,init_curr &
    , init_srce &                                    
    , fn,fnp1,sJ0ftmp,Sfmu_eq,sJ0fmu_eq,poiss,J0)
    use coord_system_class
    use init_profile_class
    use init_magnetic_class
    use init_current_class
    use sources_module                               
    use gyroaverage_class
    use fdistribu5d_class
    use poisson_class
    use output_saving_module
    use diag3D_module, only : compute_save_diag3D
    use Pcross_section_module
    use physics_GC_module, only : compute_GC_moments
    use physics_part_module, only : compute_part_moments
#ifndef NO_HDF5
    use resu3D_saving_module
    use f5D_saving_module
#endif
    use spectra_module, only : diag_spectra          
    use clock_module
!R3 #include "code_instrumentation/r3_info.h" !R3
    real(RKIND)          , intent(in)    :: Siter_time
    type(diagnostics)    , intent(inout) :: Sdiag
    real(RKIND)          , intent(in)    :: Sdeltat
    type(geometry)       , intent(in)    :: geom
    type(coord_system)   , intent(in)    :: coord_sys
    type(init_profile)   , intent(in)    :: init_prof
    type(init_magnetic)  , intent(in)    :: init_magnet
    type(init_current)   , intent(in)    :: init_curr
    type(init_sources)   , intent(in)    :: init_srce  
    type(fdistribu5d)    , intent(inout) :: fn
    type(fdistribu5d)    , intent(inout) :: fnp1
    type(fdistribu5d)    , intent(inout) :: sJ0ftmp
    real(RKIND), &
      dimension(:,:,:)   , pointer       :: Sfmu_eq
    real(RKIND), &
      dimension(:,:,:)   , pointer       :: sJ0fmu_eq
    type(poisson)        , intent(inout) :: poiss
    type(J0operator)     , intent(inout) :: J0

    integer(TIMEPREC) :: tdeb,tfin
    real(RKIND)       :: tdiag(1:12)
    real(RKIND)       :: thdf5_diag(1:12)
#ifdef TIMER
    real(RKIND)       ::  tvect(32), tmin(32), tmax(32), tavg(32)
#endif

!R3 call r3_info_begin (r3_info_index_0, 'CPU_diagnostics') !R3
    call Write_dyn_balise_open('CPU_diagnostics') !R4

    !*****************************************************
    !*** Parallelisation of the output saving          ***
    !***  - HDF5_CL_saving                             *** 
    !***     --> performed by the proc. diag_targ(1)   ***
    !***  - HDF5_rprof_GC_saving                       ***
    !***     --> performed by the proc. diag_targ(2)   ***
    !***  - HDF5_rprof_part_saving                     ***
    !***     --> performed by the proc. diag_targ(3)   ***
    !***  - HDF5_Phi2D_saving                          ***
    !***     --> performed by the proc. diag_targ(4)   ***
    !***  - HDF5_f2D_saving                            ***
    !***     --> performed by the proc. diag_targ(5)   ***
    !***  - HDF5_collision_saving                      *** 
    !***     --> performed by the proc. diag_targ(6)   *** 
    !*****************************************************
    call ppbarrier_timer()
    call clck_time(tdeb)   

    call clck_time(bclock_diagnostics)
    call clck_time(bclock_physics)
    !*** computation of the moments associated  ***
    !***  to the guiding-center (GC)            *** 
    call compute_GC_moments(geom,coord_sys, &
      init_magnet,init_curr,fn,fnp1,Sfmu_eq, &
      dJ0Phidr_PNrPNthNphi,dJ0Phidtheta_PNrPNthNphi, &
      dJ0Phidphi_PNrPNthNphi,Sdeltat,Siter_time)
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tdiag(1))    
    tdeb = tfin

    ! F4D_transp must be allocated before compute_part_moments
    call FTR_ALLOC_4(f4D_transp,0,Nr,0,Ntheta,kstart,kend,lstart,lend)

    !*** computation of the moments associated ***
    !***   to the particles                    ***
    call compute_part_moments(geom,coord_sys, &
      init_magnet,init_curr,J0,fnp1,SJ0ftmp, &
      SJ0fmu_eq,poiss)

    call ppbarrier()
    call clck_time(eclock_physics)
    call clck_diff(bclock_physics,eclock_physics, &
      global_time_physics)

    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tdiag(2))    
    tdeb = tfin

    !*** computation of the cross-sections of ***
    !***  the distribution function and the   ***
    !***  potential                           ***
    call clck_time(bclock_crosssection)
    call compute_ftrapped(fnp1,geom)
    call ppbarrier()
    call compute_fpassing(fnp1,geom)
    call ppbarrier()
    call compute_vparmu_crosssection(geom,init_srce,fnp1)          

    call ppbarrier()
    call clck_time(eclock_crosssection)
    call clck_diff(bclock_crosssection,eclock_crosssection, &
      global_time_crosssection)
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tdiag(3))    
    tdeb = tfin

    !*****************************************************
    !*   0D-1D and 2D diagnostics saving in HDF5 format  *
    !*****************************************************
    call clck_time(bclock_diag2D)

    thdf5_diag(1:12) = 0.0
    call clck_time(tdeb)
    !-> Writing of "conservation_laws_<num_diag>.h5" files 
    if (diag_targ(1) .eq. pInSpecies_id) then
      call HDF5_CL_saving(geom,Sdiag%nb_diag,Sdiag%time_save)
    endif
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,thdf5_diag(1))
    tdeb = tfin

    !-> Writing of "rprof_GC_<num_diag>.h5" files
    if (diag_targ(2) .eq. pInSpecies_id) then
      call HDF5_rprof_GC_saving(geom,Sdiag%nb_diag,Sdiag%time_save)
    end if
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,thdf5_diag(2))
    tdeb = tfin

    !-> Writing of "rprof_part_<num_diag>.h5" files
    if (diag_targ(3) .eq. pInSpecies_id) then
      call HDF5_rprof_part_saving(geom,Sdiag%nb_diag,Sdiag%time_save)
    endif
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,thdf5_diag(3))
    tdeb = tfin

    !-> Writing of "Phi2D_<num_diag>.h5" files
    call compute_Phi2DCS(geom,poiss)
    if (diag_targ(4) .eq. pInSpecies_id) then
      call HDF5_Phi2D_saving(geom,Sdiag%nb_diag,Sdiag%time_save)  
    endif
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,thdf5_diag(4))
    tdeb = tfin

    !-> Writing of "f2D_<num_diag>.h5" files
    if (diag_targ(5) .eq. pInSpecies_id) then
      call HDF5_f2D_saving(geom,Sdiag%nb_diag,Sdiag%time_save)
    endif
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,thdf5_diag(5))
    tdeb = tfin

    !-> Writing of "collisions_<num_diag>.h5 files 
    if (collisions) then
      if ((diag_level.ge.5) .and. &
        (diag_targ(6) .eq. pInSpecies_id)) then
        call HDF5_collision_saving(geom,Sdiag%nb_diag,Sdiag%time_save)
      end if
    end if
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,thdf5_diag(6))

    call ppbarrier()
    call clck_time(eclock_diag2D)
    call clck_diff(bclock_diag2D,eclock_diag2D, &
      global_time_diag2D)
    call clck_time(tfin)
    call clck_ldiff(bclock_diag2D,eclock_diag2D, &
      tdiag(4))    

    !************************************************
    !*  3D and 5D diagnostics saving in HDF5 format *
    !************************************************
    call clck_time(tdeb)
#ifndef NO_HDF5
    if (Sdiag%nb_diag.ne.-1) then
      !*** Saving of 3D fluid moments and Phi3D(r,theta,phi) ***
      if ( Sdiag%Phi3D_diag_OK .or. Sdiag%Moments3D_diag_OK ) then
        call compute_save_diag3D(Siter_time, &
          Sdiag%Phi3D_diag_OK,Sdiag%Phi3D_nb_diag, &
          Sdiag%Phi3D_time_save,Sdiag%Moments3D_diag_OK, &
          Sdiag%Moments3D_nb_diag,Sdiag%Moments3D_time_save, &
          geom,init_prof,init_magnet,init_curr,J0, &
          fnp1,sJ0ftmp,poiss%Phi_NrNthPNphi_Cworld)
      end if

      !*** Saving of f(r,theta,phi,vpar,mu) ***
      if (Sdiag%f5D_diag_OK) &
        call HDF5_f5D_saving(Sdiag%f5D_nb_diag,Sdiag%time_save, &
        geom,fnp1%values)
    end if
    call ppbarrier()
#endif

    ! F4D_transp must be deallocated after compute_save_diag3D
    call FTR_DEALLOC(f4D_transp)
      
    !************************************************
    !*  Spectral diagnostics saving in HDF5 format  *
    !************************************************ 
    if (Sdiag%Sp_diag_OK) then
      call diag_spectra(geom,init_magnet,init_prof,fnp1,poiss,J0, &
        Sdiag%Sp_nb_diag,Sdiag%time_save)
    end if

    !******************************************
    !* output fdistrib values for several v// *
    !******************************************
    if (diagf) then
       call compute_fslice(geom,init_magnet,init_curr,fnp1,Sfmu_eq)
    end if

    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tdiag(5))    

    call ppbarrier_timer()
    call clck_time(eclock_diagnostics)
    call clck_diff(bclock_diagnostics,eclock_diagnostics, &
      global_time_diagnostics)
    call clck_ldiff(bclock_diagnostics, &
      eclock_diagnostics,tdiag(6))

#ifdef TIMER
    if (pglobal_id .eq. 0) then
      write(6,'(I4,A,7F13.6)') pglobal_id, &
        " Temps HDF5 diagnostics ", thdf5_diag(1:6)
    end if
    tvect(1:7) = (/ tdiag(1:5), tdiag(6)-SUM(tdiag(1:5)), tdiag(6) /)
    call pp_reduce_vector(tvect,7,tmin,tmax,tavg)
    if (plocal_id .eq. 0) then
       write(6,'(I4,A,6F11.4,A,F11.4)') pglobal_id, &
            " TMIN diagnostics ", tmin(1:5), &
            tmin(6)," tot ",tmin(7)
       write(6,'(I4,A,6F11.4,A,F11.4)') pglobal_id, &
            " TMAX diagnostics ", tmax(1:5), &
            tmax(6)," tot ",tmax(7)
       write(6,'(I4,A,6F11.4,A,F11.4)') pglobal_id, &
            " TAVG diagnostics ", tavg(1:5), &
            tavg(6)," tot ",tavg(7)
    end if
#endif

    !************************************
    !* Write the ASCII results          *
    !************************************
    call WriteScreen(Siter_time,Sdeltat,.true.)
    call WriteResults_CL(Siter_time)
    call WriteResults_rprof(Siter_time,geom,poiss)

    !************************************
    !* Diagnostic incrementation        *
    !************************************
    call increment_diagnostics(Siter_time,Sdeltat,Sdiag)
    call Write_dyn_balise_close('CPU_diagnostics') !R4
!R3 call r3_info_end (r3_info_index_0) !R3
  end subroutine compute_diagnostics


  !-----------------------------------------------------
  ! Used to write screen infos for ne and ni conservation
  !    at each iteration
  !-----------------------------------------------------
  subroutine WriteScreen(Siter_time,dt,diag_status)
    use globals, only : plocal_id, mu_id, &
      major_species, init_mode, &
      iter_glob, iter_run, nbiter, &
      nbelectrons_diag, nbions_diag, diag_targ, &
      uout_res, outputproc, file_name_res
    use Io_Unit_module, only: IU_open, IU_close
    
    real(RKIND), intent(in) :: Siter_time
    real(RKIND), intent(in) :: dt
    logical    , intent(in) :: diag_status
    
    integer :: uout
    integer :: line_count

    if ((.not.init_mode).and.(iter_run.eq.1)) then
      line_count = 0
    else
      line_count = modulo(iter_run,40)
    end if

    if (pglobal_id.eq.outputproc) then
      call IU_open(unit=uout_res,file=file_name_res,status='OLD',&
        position='APPEND', form='FORMATTED')

      if (line_count==0) then
        write(uout_res,'(3A36)') &
          '------------------------------------', &
          '------------------------------------', &
          '------------------------------------'
        write(uout_res,'(3A8,2A12,3A20)') 'GStep','LStep','NbStep',&
          'Time','TimeStep','ni','ne','ni/ne'
        write(6,'(3A36)') &
          '------------------------------------', &
          '------------------------------------', &
          '------------------------------------'
        write(6,'(3A8,2A12,3A20)') 'GStep','LStep','NbStep',&
          'Time','TimeStep','ni','ne','ni/ne'
      end if
      if (.not.(diag_status)) then
        write(uout_res,'(3I8,2(1pe12.3),3(1pe20.12))') &
          iter_glob, iter_run, nbiter, Siter_time, dt
        write(6,'(3I8,2(1pe12.3),3(1pe20.12))') &
          iter_glob, iter_run, nbiter, Siter_time, dt
      else
        if (nbelectrons_diag.ne.0._RKIND) then
          write(uout_res,'(3I8,2(1pe12.3),3(1pe20.12))') &
            iter_glob, iter_run, nbiter, Siter_time, dt, &
            nbions_diag, nbelectrons_diag, &
            (nbions_diag/nbelectrons_diag)
          write(6,'(3I8,2(1pe12.3),3(1pe20.12))') &
            iter_glob, iter_run, nbiter, Siter_time, dt, &
            nbions_diag, nbelectrons_diag, &
            (nbions_diag/nbelectrons_diag)
        else
          write(uout_res,'(3I8,2(1pe12.3),2(1pe20.12))') &
            iter_glob, iter_run, nbiter, Siter_time, dt, &
            nbions_diag, nbelectrons_diag
          write(6,'(3I8,2(1pe12.3),2(1pe20.12))') &
            iter_glob, iter_run, nbiter, Siter_time, dt, &
            nbions_diag, nbelectrons_diag
        end if
      end if
      call IU_close(uout_res)
    end if
  end subroutine WriteScreen


  !-----------------------------------------------------
  ! Used to write results on the conservation laws
  !    in gysela_CL.out file at each iteration
  !-----------------------------------------------------
  subroutine WriteResults_CL(Siter_time)
    use globals, only : pInSpecies_id, pSpecies_id, &
      mu_id, iter_glob, iter_run, nbiter, &
      nbelectrons_diag, nbions_diag, diag_targ
    use physics_GC_module  , only : L2norm_diag, &
      entropy_diag, Enkin_diag
    use physics_part_module, only : Enpot_diag
    use read_write_module  , only : uout_CL
    use Io_Unit_module     , only : IU_open, IU_close
    
    real(RKIND), intent(in) :: Siter_time

    character(LEN=20) :: file_name_CL
    character(LEN=3)  :: num_species

    write(num_species,'(A2,I1)') "sp", pSpecies_id+1
    if (diag_targ(1) .eq. pInSpecies_id) then
      if (pSpecies_id.eq.0) then
        write(file_name_CL,'(A)')"gysela_CL.out"//char(0)
      else
        write(file_name_CL,'(2A)') num_species, "/gysela_CL.out"//char(0)
      end if
      call IU_open(uout_CL, file = file_name_CL, status = 'UNKNOWN', &
        position = 'APPEND', form = 'FORMATTED')
      !-> title writting
      if (iter_glob.eq.-1) then
        write(uout_CL,'(A8,A12,7A20)') &
          'GStep','Time','ni','ne','ni/ne', &
          'Enkin','Enpot','Entropy','L2norm'
      end if
      !-> writting of the different 0D physical quantities
      write(uout_CL,'(I8,1pe12.3,7(1pe20.12))') &
        iter_glob, Siter_time, nbions_diag, nbelectrons_diag, &
        nbions_diag/nbelectrons_diag, Enkin_diag, Enpot_diag, &
        entropy_diag, L2norm_diag
      !-> file closing
      call IU_close(uout_CL)
    end if
  end subroutine WriteResults_CL


  !-----------------------------------------------------
  ! Used to write results 
  !    in gysela_rprof.out file at each iteration
  !-----------------------------------------------------
  subroutine WriteResults_rprof(Siter_time,geom,poiss)
    use globals, only : pInSpecies_id, pSpecies_id, mu_id, &
      iter_glob, iter_run, nbiter, &
      diag_level, diag_targ, Rarray1_Nr, plocal_id, &
      mpi_comm_species, distrib3D
    use physics_part_module, only : Qr_turb_diag 
    use poisson_class      , only : Phi00_diag
    use read_write_module  , only : uout_rprof
    use utils_module       , only : deriv1
    use integration_module , only : compute_intrdr_colloc
    use Io_Unit_module     , only: IU_open, IU_close

    use mpiMod
    real(RKIND)   , intent(in) :: Siter_time
    type(geometry), intent(in) :: geom
    type(poisson) , intent(in) :: poiss
    
    integer     :: ir, irp, itheta, iphi
    integer     :: ierr
    real(RKIND) :: Phi_tmp
    real(RKIND) :: intPhi2, intPhi2_loc
    real(RKIND) :: sqrt_intPhi2
    real(RKIND) :: max_Phi00
    real(RKIND) :: max_dPhi00dr
    real(RKIND) :: int_Qrdr
    real(RKIND), dimension(:), pointer :: dPhi00dr 

    character(LEN=100) :: file_name_rprof
    character(LEN=3)   :: num_species

!R3 #include "code_instrumentation/r3_info.h" !R3

    if (diag_level.ge.3) then
      !*** computation of                                   ***
      !***   sqrt(\int(Phi(rpeak,theta,phi)^2 dtheta dphi)) ***
      irp         = int(geom%Nr/2)
      intPhi2_loc = 0._RKIND
      if (distrib3D) then
        do iphi = 0,geom%Nphi-1
          if (moments_locphi(iphi).eq.plocal_id .and. mu_id .eq. 0) then
            do itheta = 0,geom%Ntheta
              Phi_tmp     = poiss%Phi_NrNthPNphi_Cmu(irp,itheta,iphi)
              intPhi2_loc = intPhi2_loc + Phi_tmp*Phi_tmp * &
                geom%coeff_intdtheta(itheta)*geom%coeff_intdphi(iphi)
            end do
          end if
        end do
        call MY_Reduce(intPhi2_loc,intPhi2,1,MPI_REAL8,&
          MPI_SUM,diag_targ(3),mpi_comm_species,ierr)
      else
        intPhi2 = 0._RKIND
        do iphi = 0,geom%Nphi
          do itheta = 0,geom%Ntheta
            Phi_tmp = poiss%Phi_NrNthPNphi_Cworld(irp,itheta,iphi)
            intPhi2 = intPhi2 + Phi_tmp*Phi_tmp * &
              geom%coeff_intdtheta(itheta)*geom%coeff_intdphi(iphi)
          end do
        end do
      end if
      
!R3 call r3_info_begin (r3_info_index_0, 'CPU_WriteResRprof') !R3
      call Write_dyn_balise_open('CPU_WriteResRprof') !R4
      if (diag_targ(3) .eq. pInSpecies_id) then
        sqrt_intPhi2 = sqrt(intPhi2)
        !*** computation of the maximal value of ***
        !***  the (0,0) mode of Phi              ***
        max_Phi00    = maxval(Phi00_diag)
        !*** computation of maximal velocity due to zonal flow ***
        dPhi00dr => Rarray1_Nr
        call deriv1(Phi00_diag(0:),dPhi00dr(0:),geom%Nr,geom%dr,0)
        max_dPhi00dr = maxval(dPhi00dr)
        !*** computation of \int Q(r) r dr ***
        call compute_intrdr_colloc(Qr_turb_diag(0:),geom,int_Qrdr)

        !-> opening of the file 'gysela_rprof.out'
        write(num_species,'(A2,I1)') "sp", pSpecies_id+1
        if (pSpecies_id.eq.0) then
          write(file_name_rprof,'(A)')"gysela_rprof.out"//char(0)
        else
          write(file_name_rprof,'(2A)') num_species, &
            "/gysela_rprof.out"//char(0)
        end if
        call IU_open(uout_rprof, file = file_name_rprof, &
          status = 'UNKNOWN', position = 'APPEND', &
          form = 'FORMATTED')
        !-> title writting
        if (iter_glob.eq.-1) then
          write(uout_rprof,'(A8,A12,4A20)') &
            'GStep','Time','sqrt(intPhi2)', &
            'max(Phi00)', 'max(dPhi00dr)','\intQ(r)rdr'
        end if
        !-> writting of the different 0D physical quantities
        write(uout_rprof,'(I8,1pe12.3,4(1pe20.12))') &
          iter_glob, Siter_time, sqrt_intPhi2, &
          max_Phi00, max_dPhi00dr, int_Qrdr
        !-> file closing
        call IU_close(uout_rprof)
      end if
      call Write_dyn_balise_close('CPU_WriteResRprof') !R4
!R3 call r3_info_end (r3_info_index_0) !R3
    end if
  end subroutine WriteResults_rprof

end module diagnostics_module
