!**************************************************************
!  Copyright Euratom-CEA
!  Authors : 
!     Virginie Grandgirard (virginie.grandgirard@cea.fr)
!     Chantal Passeron (chantal.passeron@cea.fr)
!     Guillaume Latu (guillaume.latu@cea.fr)
!     Xavier Garbet (xavier.garbet@cea.fr)
!     Philippe Ghendrih (philippe.ghendrih@cea.fr)
!     Yanick Sarazin (yanick.sarazin@cea.fr)
!  
!  This code GYSELA (for GYrokinetic SEmi-LAgrangian) 
!  is a 5D gyrokinetic global full-f code for simulating 
!  the plasma turbulence in a tokamak.
!  
!  This software is governed by the CeCILL-B license 
!  under French law and abiding by the rules of distribution 
!  of free software.  You can  use, modify and redistribute 
!  the software under the terms of the CeCILL-B license as 
!  circulated by CEA, CNRS and INRIA at the following URL
!  "http://www.cecill.info". 
!**************************************************************

#include "config-F.inc"

!-------------------------------------------------
! file : output_saving.F90
! date : 16/09/2005
!  used for saving the physic results
!------------------------------------------------
module output_saving_module
  use geometry_class
  use init_profile_class
  use init_magnetic_class
  use sources_module                         
  use globals
  use wrap_alloc_module
#ifndef NO_HDF5
  use HDF5
  use HDF5_module
#endif

  implicit none
  
  !******************************
  contains
  !******************************
      

  !***********************************************************
  !   SAVING DIRECTLY IN HDF5 FORMAT
  !***********************************************************
  !---------------------------------------------- 
  ! Output saving of the coordinate system
  !  . covariant components of the metric tensor
  !  . jacobian in space 
  !  -> directly in HDF5 binary format
  !----------------------------------------------   
  subroutine HDF5_coordsys_saving(geom,coord_sys)
    use coord_system_class
    type(geometry)     , intent(in) :: geom
    type(coord_system) , intent(in) :: coord_sys
#ifndef NO_HDF5
    integer           :: Nbr, Nbc, icr
    character(LEN=30) :: coordsys_file
    character(LEN=3)  :: num_species
    integer(HID_T)    :: file_id 
    integer           :: ierr

    write(num_species,'(A2,I1)') "sp", pSpecies_id+1
    if (pSpecies_id .eq. 0) then
      write(coordsys_file,'(A)') "coord_system.h5"
    else
      write(coordsys_file,'((2A))') num_species, "/coord_system.h5"
    end if

    call HDF5_create(coordsys_file,file_id,ierr)
    if (ierr.ne.0) then
      print*,'pglobal_id = ',pglobal_id, &
        ' ==> error for opening of ',coordsys_file
    end if

    !*** code version saving ***
    call HDF5_save(file_id,code_version, &
      'code_version'//char(0))

    !*** covariant components of the metric tensor ***
    Nbr = geom%Nr+1
    Nbc = geom%Ntheta+1
    call HDF5_save(file_id,coord_sys%g11,Nbr,Nbc, &
      'g11'//char(0))
    call HDF5_save(file_id,coord_sys%g12,Nbr,Nbc, &
      'g12'//char(0))
    call HDF5_save(file_id,coord_sys%g13,Nbr,Nbc, &
      'g13'//char(0))
    call HDF5_save(file_id,coord_sys%g21,Nbr,Nbc, &
      'g21'//char(0))
    call HDF5_save(file_id,coord_sys%g22,Nbr,Nbc, &
      'g22'//char(0))
    call HDF5_save(file_id,coord_sys%g23,Nbr,Nbc, &
      'g23'//char(0))
    call HDF5_save(file_id,coord_sys%g31,Nbr,Nbc, &
      'g31'//char(0))
    call HDF5_save(file_id,coord_sys%g32,Nbr,Nbc, &
      'g32'//char(0))
    call HDF5_save(file_id,coord_sys%g33,Nbr,Nbc, &
      'g33'//char(0))

    !*** jacobian in space ***
    call HDF5_save(file_id,jacobian_space,Nbr,Nbc, &
      'jacob_space'//char(0))

    !***  \int dtheta dphi Js with Js the jacobian in space ***
    call HDF5_save(file_id,intdthetadphi_Js(0:),Nbr, &
      'intdthetadphi_Js'//char(0))
    call HDF5_close(file_id)
#endif
  end subroutine HDF5_coordsys_saving


  !--------------------------------------------------
  ! Output saving of all /NAMELIST/ input parameters
  !--------------------------------------------------
  subroutine HDF5_data_saving
    use globals
    
#ifndef NO_HDF5
    character(LEN=100) :: data_file
    character(LEN=3)   :: num_species
    integer(HID_T)     :: file_id   
    integer            :: ispec, ierr
    
    ! Group identifiers
    integer(HID_T) :: grpname_id, grp_PARALLEL_id, grp_MESH_id, &
      grp_EQUIL_id, grp_COLLISION_id, grp_BUFFER_ID, grp_SOURCE_id, &
      grp_ALGORITHM_id, grp_TEST_id, grp_OUTPUT_id
    
    ! Group names
    character(LEN=4),  parameter :: grpname = "DATA"
    character(LEN=50), parameter :: &
      grp_PARALLEL  = "DATA/PARALLEL", &
      grp_MESH      = "DATA/MESH", &
      grp_EQUIL     = "DATA/EQUIL", &
      grp_COLLISION = "DATA/COLLISION", &
      grp_BUFFER    = "DATA/BUFFER", &
      grp_SOURCE    = "DATA/SOURCE", &
      grp_ALGORITHM = "DATA/ALGORITHM", &
      grp_TEST      = "DATA/TEST", &
      grp_OUTPUT    = "DATA/OUTPUT"
    
    !--------------------------------
    ! hdf5 output file: data_rxxx.h5
    !--------------------------------
    write(num_species,'(A2,I1)') "sp", pSpecies_id+1
    if (pSpecies_id .eq. 0) then
      write(data_file,'(A,'//numfmt_rst//',A)') &
        "data", nb_restart_diag, ".h5"       
    else
      write(data_file,'((2A),'//numfmt_rst//',A)') &
        num_species, "/data", nb_restart_diag, ".h5"       
    end if
    
    call HDF5_create(trim(data_file),file_id,ierr)
    if (ierr.ne.0) then
      print*,'pglobal_id = ',pglobal_id, &
        ' ==> error for opening of ',data_file
    end if
    
    !-------------------
    ! Create all groups
    !-------------------
    ! create group "DATA" in the root group using absolute name.
    call H5gcreate_f(file_id, grpname, grpname_id, ierr)
    ! create group "PARALLEL" in group "DATA" using absolute name.
    call H5gcreate_f(file_id, grp_PARALLEL, grp_PARALLEL_id, ierr)
    ! create group "MESH" in group "DATA" using absolute name.
    call H5gcreate_f(file_id, grp_MESH, grp_MESH_id, ierr)
    ! create group "EQUIL" in group "DATA" using absolute name.
    call H5gcreate_f(file_id, grp_EQUIL, grp_EQUIL_id, ierr)
    ! create group "COLLISION" in group "DATA" using absolute name.
    call H5gcreate_f(file_id, grp_COLLISION, grp_COLLISION_id, ierr)
    ! create group "BUFFER" in group "DATA" using absolute name.
    call H5gcreate_f(file_id, grp_BUFFER, grp_BUFFER_id, ierr)
    ! create group "SOURCE" in group "DATA" using absolute name.
    call H5gcreate_f(file_id, grp_SOURCE, grp_SOURCE_id, ierr)
    ! create group "ALGORITHM" in group "DATA" using absolute name.
    call H5gcreate_f(file_id, grp_ALGORITHM, grp_ALGORITHM_id, ierr)
    ! create group "TEST" in group "DATA" using absolute name.
    call H5gcreate_f(file_id, grp_TEST, grp_TEST_id, ierr)
    ! create group "OUTPUT" in group "DATA" using absolute name.
    call H5gcreate_f(file_id, grp_OUTPUT, grp_OUTPUT_id, ierr)

    !------------------------------------------
    !---> Parallel data of namelist /PARALLEL/ 
    !------------------------------------------
    call HDF5_save(grp_PARALLEL_id,Nspecies,"Nspecies"//char(0))
    call HDF5_save(grp_PARALLEL_id,Nbproc_r,"Nbproc_r"//char(0))
    call HDF5_save(grp_PARALLEL_id,Nbproc_theta, &
      "Nbproc_theta"//char(0))
    call HDF5_save(grp_PARALLEL_id,Nbproc_mu,"Nbproc_mu"//char(0))
    call HDF5_save(grp_PARALLEL_id,bloc_phi,"bloc_phi"//char(0))
    call HDF5_save(grp_PARALLEL_id,Nbthread,"Nbthread"//char(0))
    if (transpose4D) then
      call HDF5_save(grp_PARALLEL_id,"T","transpose4D"//char(0))
    else
      call HDF5_save(grp_PARALLEL_id,"F","transpose4D"//char(0))
    end if
    if (transp_velocity) then
      call HDF5_save(grp_PARALLEL_id,"T","transp_velocity"//char(0))
    else
      call HDF5_save(grp_PARALLEL_id,"F","transp_velocity"//char(0))
    end if
    if (large_platform) then
      call HDF5_save(grp_PARALLEL_id,"T","large_platform"//char(0))
    else
      call HDF5_save(grp_PARALLEL_id,"F","large_platform"//char(0))
    end if
    if (distrib3D) then
      call HDF5_save(grp_PARALLEL_id,"T","distrib3D"//char(0))
    else
      call HDF5_save(grp_PARALLEL_id,"F","distrib3D"//char(0))
    end if

    !--------------------------------------
    !---> Geometry data of namelist /MESH/  
    !--------------------------------------
    if (CylindricalGeometry) then
      call HDF5_save(grp_MESH_id,"T","CylindricalGeometry"//char(0))
    else
      call HDF5_save(grp_MESH_id,"F","CylindricalGeometry"//char(0))
    end if
    call HDF5_save(grp_MESH_id,Nr,"Nr"//char(0))
    call HDF5_save(grp_MESH_id,Ntheta,"Ntheta"//char(0))
    call HDF5_save(grp_MESH_id,Nphi,"Nphi"//char(0))
    call HDF5_save(grp_MESH_id,Nvpar,"Nvpar"//char(0))
    call HDF5_save(grp_MESH_id,Nmu,"Nmu"//char(0))
    call HDF5_save(grp_MESH_id,a,"a"//char(0))
    call HDF5_save(grp_MESH_id,rhomin,"rhomin"//char(0))
    call HDF5_save(grp_MESH_id,rhomax,"rhomax"//char(0))
    call HDF5_save(grp_MESH_id,Ltheta,"Ltheta"//char(0))
    call HDF5_save(grp_MESH_id,Lphi,"Lphi"//char(0))
    call HDF5_save(grp_MESH_id,aspect_ratio,"aspect_ratio"//char(0))
    call HDF5_save(grp_MESH_id,nb_vth0,"nb_vth0"//char(0))
    call HDF5_save(grp_MESH_id,Lmu,"Lmu"//char(0))
    call HDF5_save(grp_MESH_id,mumin,"mumin"//char(0))

    !------------------------------------------
    !---> Equilibrium data of namelist /EQUIL/
    !------------------------------------------
    if (adiabatic_electrons) then
      call HDF5_save(grp_EQUIL_id,"T","adiabatic_electrons"//char(0))
    else
      call HDF5_save(grp_EQUIL_id,"F","adiabatic_electrons"//char(0))
    end if
    call HDF5_save(grp_EQUIL_id,canonical_profile, &
      "canonical_profile"//char(0))
    call HDF5_save(grp_EQUIL_id,tau0,"tau0"//char(0))
    call HDF5_save(grp_EQUIL_id,A0,"A0"//char(0))
    call HDF5_save(grp_EQUIL_id,Z0,"Z0"//char(0))
    call HDF5_save(grp_EQUIL_id,As(0:),Nspecies,"As"//char(0))
    call HDF5_save(grp_EQUIL_id,Zs(0:),Nspecies,"Zs"//char(0))
    call HDF5_save(grp_EQUIL_id,species_frac(0:),Nspecies, &
      "species_frac"//char(0))
    call HDF5_save(grp_EQUIL_id,read_ns0,Nspecies, &
      "read_ns0"//char(0))
    call HDF5_save(grp_EQUIL_id,read_Ts0,Nspecies, &
      "read_Ts0"//char(0))
    if (read_Te0) then
      call HDF5_save(grp_EQUIL_id,"T","read_Te0"//char(0))
    else
      call HDF5_save(grp_EQUIL_id,"F","read_Te0"//char(0))
    end if
    if (read_q) then
      call HDF5_save(grp_EQUIL_id,"T","read_q"//char(0))
    else
      call HDF5_save(grp_EQUIL_id,"F","read_q"//char(0))
    end if
    if (read_Vpar0) then
      call HDF5_save(grp_EQUIL_id,"T","read_Vpar0"//char(0))
    else
      call HDF5_save(grp_EQUIL_id,"F","read_Vpar0"//char(0))
    end if
    call HDF5_save(grp_EQUIL_id,profile_choice,Nspecies, &
      "profile_choice"//char(0))
    call HDF5_save(grp_EQUIL_id,rpeak,"rpeak"//char(0))
    call HDF5_save(grp_EQUIL_id,kappa_ns0,Nspecies, &
      "kappa_ns0"//char(0))
    call HDF5_save(grp_EQUIL_id,kappa_Ts0,Nspecies, &
      "kappa_Ts0"//char(0))
    call HDF5_save(grp_EQUIL_id,deltar_ns0,Nspecies, &
      "deltar_ns0"//char(0))
    call HDF5_save(grp_EQUIL_id,deltar_Ts0,Nspecies, &
      "deltar_Ts0"//char(0))
    call HDF5_save(grp_EQUIL_id,q_profile,"q_profile"//char(0))
    call HDF5_save(grp_EQUIL_id,q_param1,"q_param1"//char(0))
    call HDF5_save(grp_EQUIL_id,q_param2,"q_param2"//char(0))
    call HDF5_save(grp_EQUIL_id,q_param3,"q_param3"//char(0))
    call HDF5_save(grp_EQUIL_id,q_param4,"q_param4"//char(0))
    if (magnetic_drift) then
      call HDF5_save(grp_EQUIL_id,"T","magnetic_drift"//char(0))
    else
      call HDF5_save(grp_EQUIL_id,"F","magnetic_drift"//char(0))
    end if
    call HDF5_save(grp_EQUIL_id,perturb_amplitude,Nspecies, &
      "perturb_amplitude"//char(0))
    call HDF5_save(grp_EQUIL_id,perturb_choice, &
      "perturb_choice"//char(0))
    call HDF5_save(grp_EQUIL_id,n,"n"//char(0))
    call HDF5_save(grp_EQUIL_id,m,"m"//char(0))
    call HDF5_save(grp_EQUIL_id,FFTpotential_filter_deltam, &
      "FFTpotential_filter_deltam"//char(0))
    call HDF5_save(grp_EQUIL_id,FFTpotential_filter, &
      "FFTpotential_filter"//char(0))
    if (TF_ripple) then
      call HDF5_save(grp_EQUIL_id,"T","TF_ripple"//char(0))
    else
      call HDF5_save(grp_EQUIL_id,"F","TF_ripple"//char(0))
    end if
    call HDF5_save(grp_EQUIL_id,delta_ripple,"delta_ripple"//char(0))
    call HDF5_save(grp_EQUIL_id,N_ripple,"N_ripple"//char(0))

    !-----------------------------------------------------
    !---> Collision data of namelist /COLLISION_OPERATOR/
    !-----------------------------------------------------
    if (collisions) then
      call HDF5_save(grp_COLLISION_id,"T","collisions"//char(0))
    else
      call HDF5_save(grp_COLLISION_id,"F","collisions"//char(0))
    end if
    call HDF5_save(grp_COLLISION_id,version_coll, &
      "version_coll"//char(0))
    if (energy_transfer) then
      call HDF5_save(grp_COLLISION_id,"T","energy_transfer"//char(0))
    else
      call HDF5_save(grp_COLLISION_id,"F","energy_transfer"//char(0))
    end if
    if (moment_transfer) then
      call HDF5_save(grp_COLLISION_id,"T","moment_transfer"//char(0))
    else
      call HDF5_save(grp_COLLISION_id,"F","moment_transfer"//char(0))
    end if
    call HDF5_save(grp_COLLISION_id,nustar,"nustar"//char(0))
    call HDF5_save(grp_COLLISION_id,coll_FMrefresh, &
      "coll_FMrefresh"//char(0))
    call HDF5_save(grp_COLLISION_id,coll_nbstep, &
      "coll_nbstep"//char(0))

    !--------------------------------------------------
    !---> Dissipation data of namelist /BUFFER_REGION/
    !--------------------------------------------------
    call HDF5_save(grp_BUFFER_id,coef_diffr,Nspecies, &
      "coef_diffr"//char(0))
    call HDF5_save(grp_BUFFER_id,coef_diffth,Nspecies, &
      "coef_diffth"//char(0))
    call HDF5_save(grp_BUFFER_id,coef_krook,"coef_krook"//char(0))
    call HDF5_save(grp_BUFFER_id,coef_nustar,"coef_nustar"//char(0))
    if (buffer_asymmetry) then
      call HDF5_save(grp_BUFFER_id,"T","buffer_asymmetry"//char(0))
    else
      call HDF5_save(grp_BUFFER_id,"F","buffer_asymmetry"//char(0))
    end if
    call HDF5_save(grp_BUFFER_id,buffer_location, &
      "buffer_location"//char(0))
    call HDF5_save(grp_BUFFER_id,buffer_stiffness, &
      "buffer_stiffness"//char(0))

    !---------------------------------------------
    !---> Flux driven system of namelist /SOURCE/
    !---------------------------------------------
    if (flux_driven) then
      call HDF5_save(grp_SOURCE_id,"T","flux_driven"//char(0))
    else
      call HDF5_save(grp_SOURCE_id,"F","flux_driven"//char(0))
    end if      
    !--> Source of heating
    call HDF5_save(grp_SOURCE_id,heat_source, &
      Nspecies,"heat_source"//char(0))
    call HDF5_save(grp_SOURCE_id,heat_version(0:), &
      Nspecies,"heat_version"//char(0))
    call HDF5_save(grp_SOURCE_id,Sce0(0:),Nspecies,"Sce0"//char(0))
    call HDF5_save(grp_SOURCE_id,TS(0:),Nspecies,"TS"//char(0))
    call HDF5_save(grp_SOURCE_id,rhoS(0:),Nspecies,"rhoS"//char(0))
    call HDF5_save(grp_SOURCE_id,LS(0:),Nspecies,"LS"//char(0))

    !--> Source of density
    call HDF5_save(grp_SOURCE_id,dens_source, &
      Nspecies,"dens_source"//char(0))
    call HDF5_save(grp_SOURCE_id,Sce_dens, &
      Nspecies,"Sce_dens"//char(0))
    call HDF5_save(grp_SOURCE_id,TS_dens, &
      Nspecies,"TS_dens"//char(0))
    call HDF5_save(grp_SOURCE_id,rhoS_dens, &
      Nspecies,"rhoS_dens"//char(0))
    call HDF5_save(grp_SOURCE_id,LS_dens, &
      Nspecies,"LS_dens"//char(0))

    !--> Source of momentum
    call HDF5_save(grp_SOURCE_id,mom_source, &
      Nspecies,"mom_source"//char(0))
    call HDF5_save(grp_SOURCE_id,Sce_mom, &
      Nspecies,"Sce_mom"//char(0))
    call HDF5_save(grp_SOURCE_id,TS_mom, &
      Nspecies,"TS_mom"//char(0))
    call HDF5_save(grp_SOURCE_id,rhoS_mom, &
      Nspecies,"rhoS_mom"//char(0))
    call HDF5_save(grp_SOURCE_id,LS_mom, &
      Nspecies,"LS_mom"//char(0))

    !--> Source of vorticity 
    call HDF5_save(grp_SOURCE_id,vor_source, &
      Nspecies,"vor_source"//char(0))
    call HDF5_save(grp_SOURCE_id,Sce_vor, &
      Nspecies,"Sce_vor"//char(0))
    call HDF5_save(grp_SOURCE_id,TS_vor, &
      Nspecies,"TS_vor"//char(0))
    call HDF5_save(grp_SOURCE_id,rhoS_vor, &
      Nspecies,"rhoS_vor"//char(0))
    call HDF5_save(grp_SOURCE_id,LS_vor, &
      Nspecies,"LS_vor"//char(0))

    !-->  Source of fast particles
    call HDF5_save(grp_SOURCE_id,fp_source, &
      Nspecies,"fp_source"//char(0))
    call HDF5_save(grp_SOURCE_id,Sce_fp, &
      Nspecies,"Sce_fp"//char(0))
    call HDF5_save(grp_SOURCE_id,Sce_fpplus, &
      Nspecies,"Sce_fpplus"//char(0))
    call HDF5_save(grp_SOURCE_id,TS_fp, &
      Nspecies,"TS_fp"//char(0))
    call HDF5_save(grp_SOURCE_id,v0_fp, &
      Nspecies,"v0_fp"//char(0))
    call HDF5_save(grp_SOURCE_id,rhoS_fp, &
      Nspecies,"rhoS_fp"//char(0))
    call HDF5_save(grp_SOURCE_id,LS_fp, &
      Nspecies,"LS_fp"//char(0))
    call HDF5_save(grp_SOURCE_id,alpha_trapp, &
      Nspecies,"alpha_trapp"//char(0))

    !--------------------------------------------
    !---> Algorithm data of namelist /ALGORITHM/
    !--------------------------------------------
    call HDF5_save(grp_ALGORITHM_id, &
      trim(Vlasov_scheme),"Vlasov_scheme"//char(0))
    call HDF5_save(grp_ALGORITHM_id,deltat,"deltat"//char(0))
    call HDF5_save(grp_ALGORITHM_id,nbiter,"nbiter"//char(0))
    call HDF5_save(grp_ALGORITHM_id,dt_diag,"dt_diag"//char(0))
    call HDF5_save(grp_ALGORITHM_id,limiter,"limiter"//char(0))
    call HDF5_save(grp_ALGORITHM_id,errmax,"errmax"//char(0))

    !----------------------------------
    !---> Test data of namelist /TEST/
    !----------------------------------
    if (memory_test) then
      call HDF5_save(grp_TEST_id,"T","memory_test"//char(0))
    else
      call HDF5_save(grp_TEST_id,"F","memory_test"//char(0))
    end if
    if (gyroaverage) then
      call HDF5_save(grp_TEST_id,"T","gyroaverage"//char(0))
    else
      call HDF5_save(grp_TEST_id,"F","gyroaverage"//char(0))
    end if
    if (integral_vperp) then
      call HDF5_save(grp_TEST_id,"T", &
        "integral_vperp"//char(0))
    else
      call HDF5_save(grp_TEST_id,"F", &
        "integral_vperp"//char(0))
    end if
    call HDF5_save(grp_TEST_id, &
      integration_scheme,"integration_scheme"//char(0))
    if (BstareqB) then
      call HDF5_save(grp_TEST_id,"T","BstareqB"//char(0))
    else
      call HDF5_save(grp_TEST_id,"F","BstareqB"//char(0))
    end if
    if (Rosenbluth_Hinton) then
      call HDF5_save(grp_TEST_id,"T", &
        "Rosenbluth_Hinton"//char(0))
    else
      call HDF5_save(grp_TEST_id,"F", &
        "Rosenbluth_Hinton"//char(0))
    end if
    if (solve_QN) then
      call HDF5_save(grp_TEST_id,"T","solve_QN"//char(0))
    else
      call HDF5_save(grp_TEST_id,"F","solve_QN"//char(0))
    end if
    call HDF5_save(grp_TEST_id, QN_version,"QN_version"//char(0))
    call HDF5_save(grp_BUFFER_id,QN_coef_polarization, &
      "QN_coef_polarization"//char(0))
    call HDF5_save(grp_TEST_id,QN_coef_Phi00, &
      "QN_coef_Phi00"//char(0))
    if (Phi00_BCrmin_Neumann) then
      call HDF5_save(grp_TEST_id,"T","Phi00_BCrmin_Neumann"//char(0))
    else
      call HDF5_save(grp_TEST_id,"F","Phi00_BCrmin_Neumann"//char(0))
    end if
    if (single_m) then
      call HDF5_save(grp_TEST_id,"T","single_m"//char(0))
    else
      call HDF5_save(grp_TEST_id,"F","single_m"//char(0))
    end if
    if (single_n) then
      call HDF5_save(grp_TEST_id,"T","single_n"//char(0))
    else
      call HDF5_save(grp_TEST_id,"F","single_n"//char(0))
    end if
    if (RHS_only) then
      call HDF5_save(grp_TEST_id,"T","RHS_only"//char(0))
    else
      call HDF5_save(grp_TEST_id,"F","RHS_only"//char(0))
    end if
    call HDF5_save(grp_TEST_id,feq_choice,"feq_choice"//char(0))
    call HDF5_save(grp_TEST_id,canonical_vpar0, &
      "canonical_vpar0"//char(0))
    if (hvpar_in_fperturb) then
      call HDF5_save(grp_TEST_id,"T","hvpar_in_fperturb"//char(0))
    else
      call HDF5_save(grp_TEST_id,"F","hvpar_in_fperturb"//char(0))
    end if
    if (adaptive_deltat) then
      call HDF5_save(grp_TEST_id,"T","adaptive_deltat"//char(0))
    else
      call HDF5_save(grp_TEST_id,"F","adaptive_deltat"//char(0))
    end if
    call HDF5_save(grp_TEST_id,advec2D_deltat, &
      "advec2D_deltat"//char(0))
    call HDF5_save(grp_TEST_id,reduced_deltat, &
      "reduced_deltat"//char(0))
    call HDF5_save(grp_TEST_id,reduced_begin,"reduced_begin"//char(0))
    call HDF5_save(grp_TEST_id,reduced_end,"reduced_end"//char(0))
    call HDF5_save(grp_TEST_id,hffiltertype, &
      "hffiltertype"//char(0))
    call HDF5_save(grp_TEST_id,hffilterfreq, &
      "hffilterfreq"//char(0))
    call HDF5_save(grp_TEST_id,hffilterfref, &
      "hffilterfref"//char(0))
    if (deltaF_in_diff) then
      call HDF5_save(grp_TEST_id,"T","deltaF_in_diff"//char(0))
    else
      call HDF5_save(grp_TEST_id,"F","deltaF_in_diff"//char(0))
    end if
    if (asynchrone_writing) then
      call HDF5_save(grp_TEST_id,"T","asynchrone_writing"//char(0))
    else
      call HDF5_save(grp_TEST_id,"F","asynchrone_writing"//char(0))
    end if
    call HDF5_save(grp_TEST_id,ask_for_checksum, &
      "ask_for_checksum"//char(0))
    call HDF5_save(grp_TEST_id,modulo_restart_iter, &
      "modulo_restart_iter"//char(0))

    !--------------------------------------
    !---> Output data of namelist /OUTPUT/
    !-------------------------------------
    if (integration_CS) then
      call HDF5_save(grp_OUTPUT_id,"T","integration_CS"//char(0))
    else
      call HDF5_save(grp_OUTPUT_id,"F","integration_CS"//char(0))
    end if
    call HDF5_save(grp_OUTPUT_ID,diag_level,"diag_level"//char(0))
    if (Phi3D_saving) then
      call HDF5_save(grp_OUTPUT_id,"T","Phi3D_saving"//char(0))
    else
      call HDF5_save(grp_OUTPUT_id,"F","Phi3D_saving"//char(0))
    end if
    if (Moments3D_saving) then
      call HDF5_save(grp_OUTPUT_id,"T","Moments3D_saving"//char(0))
    else
      call HDF5_save(grp_OUTPUT_id,"F","Moments3D_saving"//char(0))
    end if
    if (f5D_saving) then
      call HDF5_save(grp_OUTPUT_id,"T","f5D_saving"//char(0))
    else
      call HDF5_save(grp_OUTPUT_id,"F","f5D_saving"//char(0))
    end if
    call HDF5_save(grp_OUTPUT_id,Phi3D_dt_diag, &
      "Phi3D_dt_diag"//char(0))
    call HDF5_save(grp_OUTPUT_id,Moments3D_dt_diag, &
      "Moments3D_dt_diag"//char(0))
    call HDF5_save(grp_OUTPUT_id,f5D_dt_diag,"f5D_dt_diag"//char(0))
    if (rst_saving) then
      call HDF5_save(grp_OUTPUT_id,"T","rst_saving"//char(0))
    else
      call HDF5_save(grp_OUTPUT_id,"F","rst_saving"//char(0))
    end if
    call HDF5_save(grp_OUTPUT_ID,N_rdiags_Spectr, &
      "N_rdiags_Spectr"//char(0))
    if (Sp_saving) then
      call HDF5_save(grp_OUTPUT_id,"T","Sp_saving"//char(0))
    else
      call HDF5_save(grp_OUTPUT_id,"F","Sp_saving"//char(0))
    end if
    call HDF5_save(grp_OUTPUT_id,Sp_dt_diag,"Sp_dt_diag"//char(0))
    call HDF5_save(grp_OUTPUT_ID,doubleSP_m,nb_fulldiag_max, &
      "doubleSP_m"//char(0))
    call HDF5_save(grp_OUTPUT_ID,doubleSP_n,nb_fulldiag_max, &
      "doubleSP_n"//char(0))
    call HDF5_save(grp_OUTPUT_ID,r_diags,N_rdiags_Spectr, &
      "r_diags"//char(0))

    !-----------------------
    ! close the groups.
    !-----------------------
    call H5gclose_f(grp_PARALLEL_id,ierr)
    call H5gclose_f(grp_MESH_id,ierr)
    call H5gclose_f(grp_EQUIL_id,ierr)
    call H5gclose_f(grp_COLLISION_id,ierr)
    call H5gclose_f(grp_BUFFER_id,ierr)
    call H5gclose_f(grp_SOURCE_id,ierr)
    call H5gclose_f(grp_ALGORITHM_id,ierr)
    call H5gclose_f(grp_TEST_id,ierr)
    call H5gclose_f(grp_OUTPUT_id,ierr)
    call H5gclose_f(grpname_id,ierr)
    
    ! close the file.
    call HDF5_close(file_id)
#endif
  end subroutine HDF5_data_saving


  !---------------------------------------------- 
  ! Output saving of the initial state 
  !  -> directly in HDF5 binary format
  !----------------------------------------------   
  subroutine HDF5_init_saving(geom,init_prof,init_magnet &
    ,init_curr ,init_srce)
    use globals, only : numfmt_rst, nb_restart_diag
    use globals, only : A0, Z0, As, Zs, &
      species_frac, pSpecies_id
    use globals, only : TF_ripple, N_ripple, delta_ripple 
    use globals, only : atleast1_source                   
    use coord_system_class    , only : R
    use collisions_vpar_module, only : nustars            
    use krook_module          , only : nu_r               
    use diffusion_module      , only : diff_coeffDr, &
      diff_coeffDth
    
    type(geometry)     , intent(in) :: geom
    type(init_profile) , intent(in) :: init_prof
    type(init_magnetic), intent(in) :: init_magnet
    type(init_current) , intent(in) :: init_curr
    type(init_sources) , intent(in) &                     
                       , optional   :: init_srce          

#ifndef NO_HDF5
    integer            :: Nbr, Nbc, icr
    character(LEN=100) :: init_state_file
    character(LEN=3)   :: num_species
    integer(HID_T)     :: file_id   
    integer            :: ierr

    real(RKIND) :: rhostar
    real(RKIND) :: Sce0_tmp, Sce_dens_tmp       
    real(RKIND) :: Sce_mom_tmp, Sce_vor_tmp     

    write(num_species,'(A2,I1)') "sp", pSpecies_id+1
    if (pSpecies_id .eq. 0) then
      write(init_state_file,'(A,'//numfmt_rst//',A)') &
        "init_state", nb_restart_diag, ".h5"       
    else
      write(init_state_file,'((2A),'//numfmt_rst//',A)') &
        num_species, "/init_state", nb_restart_diag, ".h5"       
    end if

    call HDF5_create(trim(init_state_file),file_id,ierr)
    if (ierr.ne.0) then
      print*,'pglobal_id = ',pglobal_id, &
        ' ==> error for opening of ',init_state_file
    end if

    !*** code version saving ***
    call HDF5_save(file_id,code_version, &
      'code_version'//char(0))

    !*** normalized mass, charge and fraction of each species ***
    call HDF5_save(file_id,hat_As,Nspecies, &
      'hat_As'//char(0))
    call HDF5_save(file_id,hat_Zs,Nspecies, &
      'hat_Zs'//char(0))
    call HDF5_save(file_id,species_frac,Nspecies, &
      'species_frac'//char(0))
    if (collisions) then
      call HDF5_save(file_id,nustars,Nspecies, &
      'nustars'//char(0))
    endif   

    !*** initial geometry ****
    rhostar = 1._RKIND/a
    call HDF5_save(file_id,geom%Nr,'Nr'//char(0))
    call HDF5_save(file_id,geom%Ntheta,'Ntheta'//char(0))
    call HDF5_save(file_id,geom%Nphi,'Nphi'//char(0))
    call HDF5_save(file_id,geom%Nvpar,'Nvpar'//char(0))
    call HDF5_save(file_id,geom%Nmu,'Nmu'//char(0))
    call HDF5_save(file_id,rhostar,'rhostar'//char(0))
    call HDF5_save(file_id,nustar,'nustar'//char(0))  
    call HDF5_save(file_id,R0,'R0'//char(0))
    Nbc = geom%Nr+1
    call HDF5_save(file_id,geom%rg(0:),Nbc,'rg'//char(0))
    Nbc = geom%Ntheta+1
    call HDF5_save(file_id,geom%thetag(0:),Nbc, &
      'thetag'//char(0))
    Nbc = geom%Nphi+1 
    call HDF5_save(file_id,geom%phig(0:),Nbc, &
      'phig'//char(0))
    Nbc = geom%Nvpar+1
    call HDF5_save(file_id,geom%vparg(0:),Nbc, &
      'vparg'//char(0))
    Nbc = geom%Nmu+1
    call HDF5_save(file_id,geom%mug(0:),Nbc, &
      'mug'//char(0))

    !*** initial profiles ****
    Nbc = geom%Nr+1
    !--> for ion density profile (one profile per specie s)
    call HDF5_save(file_id,init_prof%inv_Lns0(pSpecies_id), &
      'inv_Lns0'//char(0))
    call HDF5_save(file_id, &
      init_prof%deltar_ns0(pSpecies_id),'deltar_ns0'//char(0))
    call HDF5_save(file_id, &
      init_prof%ns0(pSpecies_id,0:),Nbc,'ns0'//char(0))
    call HDF5_save(file_id, &
      init_prof%dns0dr(pSpecies_id,0:),Nbc,'dns0dr'//char(0))
    !--> for ion temperature profile (one profile per specie s)
    call HDF5_save(file_id, &
      init_prof%inv_LTs0(pSpecies_id),'inv_LTs0'//char(0))
    call HDF5_save(file_id, &
      init_prof%deltar_Ts0(pSpecies_id),'deltar_Ts0'//char(0))
    call HDF5_save(file_id, &
      init_prof%Ts0(pSpecies_id,0:),Nbc,'Ts0'//char(0))
    call HDF5_save(file_id, &
      init_prof%dlogTs0dr(pSpecies_id,0:),Nbc,'dlogTs0dr'//char(0))
    !--> for electron temperature profile
    call HDF5_save(file_id,init_prof%inv_LTe0, &
      'inv_LTe0'//char(0))
    call HDF5_save(file_id,init_prof%deltar_Te0, &
      'deltar_Te0'//char(0))
    call HDF5_save(file_id,init_prof%Te0,Nbc, &
      'Te0'//char(0))
    call HDF5_save(file_id,init_prof%dTe0dr,Nbc, &
      'dTe0dr'//char(0))
    !--> for electron density profile
    call HDF5_save(file_id,init_prof%ne0,Nbc, &
      'ne0'//char(0))
    !--> for safety factor profile
    call HDF5_save(file_id,init_prof%iota,Nbc, &
      'iota'//char(0))
    call HDF5_save(file_id,init_prof%shear,Nbc, &
      'shear'//char(0))
    call HDF5_save(file_id,init_prof%psi,Nbc, &
      'psi'//char(0))

    !*** initial profile of Vpar0(theta) for r=rmax ***
    Nbc = geom%Ntheta+1
    call HDF5_save(file_id,init_prof%Vpar0,Nbc,'Vpar0'//char(0))

    !*** Dissipation coefficient profiles ****
    Nbr = 1
    Nbc = geom%Nr+1
    call HDF5_save(file_id,nu_r(0:),Nbc,'coefNu'//char(0))
    call HDF5_save(file_id,diff_coeffDr(0:), &
      Nbc,'coefDr'//char(0))
    call HDF5_save(file_id,diff_coeffDth(0:), &
      Nbc,'coefDth'//char(0))

    if (atleast1_source) then
      !*** magnitude and radial profile of the sources ****
      Nbr = 1
      Nbc = geom%Nr+1
      !--> Heat source
      Sce0_tmp = init_srce%Sce0
      call HDF5_save(file_id,Sce0_tmp,'Sce0'//char(0))
      call HDF5_save(file_id, &
        init_srce%rshape(0:)*Sce0_tmp, &
        Nbc,'S_rshape'//char(0))
      call HDF5_save(file_id, &
        init_srce%dshapedr(0:)*Sce0_tmp, &
        Nbc,'S_dshapedr'//char(0))
      !--> Density source
      Sce_dens_tmp = init_srce%Sce_dens
      call HDF5_save(file_id,Sce_dens_tmp, &
        'Sce_dens'//char(0))
      call HDF5_save(file_id, &
        init_srce%rshape_dens(0:)*Sce_dens_tmp, &
        Nbc,'S_rshape_dens'//char(0))
      call HDF5_save(file_id, &
        init_srce%dshapedr_dens(0:)*Sce_dens_tmp, &
        Nbc,'S_dshapedr_dens'//char(0))
      !--> Moment source
      Sce_mom_tmp = init_srce%Sce_mom
      call HDF5_save(file_id,Sce_mom_tmp,'Sce_mom'//char(0))
      call HDF5_save(file_id, &
        init_srce%rshape_mom(0:)*Sce_mom_tmp, &
        Nbc,'S_rshape_mom'//char(0))
      call HDF5_save(file_id, &
        init_srce%dshapedr_mom(0:)*Sce_mom_tmp, &
        Nbc,'S_dshapedr_mom'//char(0))
      !--> Vorticity source
      Sce_vor_tmp = init_srce%Sce_vor
      call HDF5_save(file_id,Sce_vor_tmp,'Sce_vor'//char(0))
      call HDF5_save(file_id, &
        init_srce%rshape_vor(0:)*Sce_vor_tmp, &
        Nbc,'S_rshape_vor'//char(0))
      call HDF5_save(file_id, &
        init_srce%dshapedr_vor(0:)*Sce_vor_tmp, &
        Nbc,'S_dshapedr_vor'//char(0))
    end if
    !*** initial profile of R(r,theta) ***
    Nbr = geom%Nr+1
    Nbc = geom%Ntheta+1
    call HDF5_save(file_id,R,Nbr,Nbc,'R'//char(0))

    !*** initial profile of B(r,theta) and its derivatives ***
    Nbr = geom%Nr+1
    Nbc = geom%Ntheta+1
    call HDF5_save(file_id,init_magnet%B_norm, &
      Nbr,Nbc,'B'//char(0))
    call HDF5_save(file_id,init_magnet%dBdr, &
      Nbr,Nbc,'dBdr'//char(0))
    call HDF5_save(file_id,init_magnet%dBdtheta, &
      Nbr,Nbc,'dBdtheta'//char(0))

    if (TF_ripple) then
      !*** saving of the input data for TF ripple ***
      call HDF5_save(file_id,N_ripple,'N_ripple'//char(0))
      call HDF5_save(file_id,delta_ripple,'ripple_delta'//char(0))
    end if
    
    !*** saving of the n_eq(r) and dnseq_dr(r) profiles ***
    Nbr = 1
    Nbc = geom%Nr+1
    call HDF5_save(file_id,nseq_r,Nbc,'nseq'//char(0))
    call HDF5_save(file_id,dnseq_dr,Nbc,'dnseq_dr'//char(0))
    !*** saving of the Ts_eq(r) and TGC_eq(r) profiles ***
    Nbr = 1
    Nbc = geom%Nr+1
    call HDF5_save(file_id,Tseq_r,Nbc,'Tseq'//char(0))
    call HDF5_save(file_id,TGCseq_r,Nbc,'TGCseq'//char(0))

    !*** initial profile of mu0J_phi(r,theta) ***
    Nbr = geom%Nr+1
    Nbc = geom%Ntheta+1
    call HDF5_save(file_id,init_curr%mu0_Jphi, &
      Nbr,Nbc,'mu0_Jphi'//char(0))

    call HDF5_close(file_id)
#endif
  end subroutine HDF5_init_saving


  !---------------------------------------------- 
  ! Output saving of the initial state 
  !  -> in ASCII format
  !----------------------------------------------   
  subroutine ASCII_init1D_saving(geom, init_prof,init_magnet &
    ,init_srce)          
    use globals, only : Sce0                              
    use globals, only : numfmt_rst, nb_restart_diag
    use globals, only : pSpecies_id
    use coord_system_class, only : R
    use krook_module      , only : nu_r                  
    use diffusion_module  , only : diff_coeffDr, &
      diff_coeffDth
    use Io_Unit_module, only: IU_open, IU_close
    
    type(geometry)     , intent(in) :: geom
    type(init_profile) , intent(in) :: init_prof
    type(init_magnetic), intent(in) :: init_magnet
    type(init_sources) , intent(in) &                    
                       , optional   :: init_srce         

    integer            :: ir
    integer:: uinit
    character(LEN=100) :: init_state_file
    character(LEN=3)   :: num_species

    write(num_species,'(A2,I1)') "sp", pSpecies_id+1
    if (pSpecies_id .eq. 0) then
      write(init_state_file,'(A,'//numfmt_rst//',A)') &
        "init_state1D", nb_restart_diag, ".dat"
    else
      write(init_state_file,'((2A),'//numfmt_rst//',A)') &
        num_species, "/init_state1D", nb_restart_diag, ".dat"
    end if
    call IU_open(uinit, file = init_state_file, status = 'UNKNOWN', &
      form = 'FORMATTED')
    if (present (init_srce)) then
      write(uinit,'(A12,17A20)') &
        'rg','Te0','Ts0','ns0','dTe0dr', &
        'dlogTs0dr','dn0dr','nseq','dnseq_dr', &
        'Tseq','iota','shear','psi', &
        'S_rshape', 'S_dshapedr', &
        'coefNu','coefDr','coefDth'
    else
      write(uinit,'(A12,16A20)') &
        'rg','Te0','Ts0','ns0','dTe0dr', &
        'dlogTsdr','dn0dr','nseq','dnseq_dr', &
        'Tseq','iota','shear','psi','coefNu','coefDr', 'coefDth'
    end if                                             
    do ir = 0,geom%Nr 
      if (present (init_srce)) then
        write(uinit,'(1pe12.3,17(1pe20.12))') &
          geom%rg(ir), init_prof%Te0(ir), &
          init_prof%Ts0(pSpecies_id,ir), &
          init_prof%ns0(pSpecies_id,ir), &
          init_prof%dTe0dr(ir), &
          init_prof%dlogTs0dr(pSpecies_id,ir), &
          init_prof%dns0dr(pSpecies_id,ir), &
          nseq_r(ir), dnseq_dr(ir), &
          Tseq_r(ir), init_prof%iota(ir), &
          init_prof%shear(ir), init_prof%psi(ir), &
          init_srce%rshape(ir)*Sce0, &                   
          init_srce%dshapedr(ir)*Sce0, &                 
          nu_r(ir), &                                    
          diff_coeffDr(ir), diff_coeffDth(ir)
      else
        write(uinit,'(1pe12.3,14(1pe20.12))') &
          geom%rg(ir), init_prof%Te0(ir), &
          init_prof%Ts0(pSpecies_id,ir), &
          init_prof%ns0(pSpecies_id,ir), &
          init_prof%dTe0dr(ir), &
          init_prof%dlogTs0dr(pSpecies_id,ir), &
          init_prof%dns0dr(pSpecies_id,ir), &
          nseq_r(ir), dnseq_dr(ir), Tseq_r(ir), &
          init_prof%iota(ir), &
          init_prof%shear(ir), init_prof%psi(ir), &
          nu_r(ir), diff_coeffDr(ir)
      end if                                             
    end do
    call IU_close(uinit)
  end subroutine ASCII_init1D_saving


  !---------------------------------------------- 
  ! Output saving of conservation laws 
  !  -> directly in HDF5 binary format
  !----------------------------------------------   
  subroutine HDF5_CL_saving(geom,Snb_diag,Stime_diag)
    use globals       , only : nbelectrons_diag, nbions_diag, &
      pInSpecies_id
    use physics_GC_module  , only : L2norm_diag, &
      entropy_diag, Enkin_diag
    use physics_part_module, only : Enpot_diag
    type(geometry), intent(in) :: geom
    integer       , intent(in) :: Snb_diag
    real(RKIND)   , intent(in) :: Stime_diag
     
#ifndef NO_HDF5
    character(LEN=100) :: conservation_laws_file    
    character(LEN=3)   :: num_species
    integer(HID_T)     :: file_id   
    integer            :: icr, ierr
    real(RKIND)        :: rdiag_level

    write(num_species,'(A2,I1)') "sp", pSpecies_id+1
    if (Snb_diag.ne.-1) then
      if (pSpecies_id .eq. 0) then
        write(conservation_laws_file,'(A,'//numfmt//',A)') &
          "conservation_laws", Snb_diag, ".h5"
      else
        write(conservation_laws_file,'((2A),'//numfmt//',A)') &
          num_species, "/conservation_laws", Snb_diag, ".h5"
      end if
    else
      if (pSpecies_id .eq. 0) then
        conservation_laws_file = "conservation_laws_init.h5"
      else
        write(conservation_laws_file,'(2A)') num_species, &
          "/conservation_laws_init.h5"
      end if
    end if
    call HDF5_create(trim(conservation_laws_file),file_id,ierr)
    if (ierr.ne.0) then
      print*,'pglobal_id = ',pglobal_id, &
        ' ==> error for opening of ',conservation_laws_file
    end if

    !*** code version saving ***
    call HDF5_save(file_id,code_version, &
      'code_version'//char(0))

    !--> save : 'time_diag   ' 
    call HDF5_save(file_id,Stime_diag,'time_diag'//char(0))
    !--> save : 'deltat   ' 
    call HDF5_save(file_id,deltat,'deltat'//char(0))
    !--> save : 'diag_level   ' 
    rdiag_level = MIN(4.,float(diag_level))
    call HDF5_save(file_id,rdiag_level,'diag_level'//char(0))
    !--> save : 'nbions_diag  '
    call HDF5_save(file_id,nbions_diag,'nbions'//char(0))
    !--> save : 'nbelectrons_diag '
    call HDF5_save(file_id,nbelectrons_diag, &
      'nbelec'//char(0))
    if (diag_level.gt.1) then
      !--> save : 'Enkin_diag   '
    call HDF5_save(file_id,Enkin_diag,'Enkin'//char(0))
      !--> save : 'Enpot_diag   '
    call HDF5_save(file_id,Enpot_diag,'Enpot'//char(0))
      !--> save : 'entropy_diag   '
    call HDF5_save(file_id,entropy_diag,'entropy'//char(0))
      !--> save : 'L2norm_diag '
    call HDF5_save(file_id,L2norm_diag,'L2norm'//char(0))
    end if
    call HDF5_close(file_id)
#else
    call ASCII_CL_saving(geom,Snb_diag,Stime_diag)
#endif
  end subroutine HDF5_CL_saving


  !---------------------------------------------- 
  ! Output saving of conservation laws 
  !  -> in ASCII format
  !----------------------------------------------   
  subroutine ASCII_CL_saving(geom,Snb_diag,Stime_diag)  
    use globals, only : nbelectrons_diag, nbions_diag
    use physics_GC_module  , only : L2norm_diag, &
      entropy_diag, Enkin_diag
    use physics_part_module, only : Enpot_diag
    use Io_Unit_module     , only : IU_open, IU_close
    
    type(geometry), intent(in) :: geom
    integer       , intent(in) :: Snb_diag
    real(RKIND)   , intent(in) :: Stime_diag

    integer:: ucl
    character(LEN=100) :: conservation_laws_file    
    character(LEN=3)   :: num_species

    write(num_species,'(A2,I1)') "sp", pSpecies_id+1
    if (Snb_diag.ne.-1) then
      if (pSpecies_id .eq. 0) then
        write(conservation_laws_file,'(A,'//numfmt//',A)') &
          "conservation_laws", Snb_diag, ".dat"
      else
        write(conservation_laws_file,'((2A),'//numfmt//',A)') &
          num_species, "/conservation_laws", Snb_diag, ".dat"
      end if
    else
      if(pSpecies_id .eq. 0) then
        conservation_laws_file = "conservation_laws_init.dat"
      else
        write(conservation_laws_file,'(2A)') num_species, &
          "/conservation_laws_init.dat"
      end if
    end if
    call IU_open(ucl, file = conservation_laws_file, status = 'UNKNOWN', &
      form = 'FORMATTED')
    write(ucl,'(A12,6A20)') &
      'time_diag','nbions','nbelectrons','Enkin', &
      'Enpot','entropy','entropy','L2norm'
    write(ucl,'(1pe12.3,6(1pe20.12))') &
      Stime_diag, nbions_diag, nbelectrons_diag, Enkin_diag, &
      Enpot_diag, entropy_diag, L2norm_diag
    call IU_close(ucl)
  end subroutine ASCII_CL_saving


  !---------------------------------------------- 
  ! Output saving of radial profiles linked to 
  !  guiding-centers
  !  -> directly in HDF5 binary format
  !----------------------------------------------   
  subroutine HDF5_rprof_GC_saving(geom,Snb_diag,Stime_diag)  
    use globals, only : pSpecies_id
    use physics_GC_module
    type(geometry), intent(in) :: geom
    integer       , intent(in) :: Snb_diag
    real(RKIND)   , intent(in) :: Stime_diag
     
#ifndef NO_HDF5
    character(LEN=100) :: rprof_GC_file    
    character(LEN=3)   :: num_species
    integer(HID_T)     :: file_id   
    integer            :: Nbr, Nbc, icr
    real(RKIND)        :: rdiag_level
    integer            :: ierr

    write(num_species,'(A2,I1)') "sp", pSpecies_id+1
    if (Snb_diag.ne.-1) then
      if (pSpecies_id .eq. 0) then
        write(rprof_GC_file,'(A,'//numfmt//',A)') &
          "rprof_GC", Snb_diag, ".h5"
      else
        write(rprof_GC_file,'((2A),'//numfmt//',A)') &
          num_species, "/rprof_GC", Snb_diag, ".h5"
      end if
    else
      if(pSpecies_id .eq. 0) then
        rprof_GC_file = "rprof_GC_init.h5"
      else
        write(rprof_GC_file,'(2A)') num_species, "/rprof_GC_init.h5"
      end if
    end if
    call HDF5_create(trim(rprof_GC_file),file_id,ierr)
    if (ierr.ne.0) then
      print*,'pglobal_id = ',pglobal_id, &
        ' ==> error for opening of ',rprof_GC_file
    end if

    !*** code version saving ***
    call HDF5_save(file_id,code_version, &
      'code_version'//char(0))

    !--> save : 'time_diag   ' 
    Nbr   = 1
    Nbc   = 1
    call HDF5_save(file_id,Stime_diag,'time_diag'//char(0))
    !--> save : 'deltat   ' 
    call HDF5_save(file_id,deltat,'deltat'//char(0))
    !--> save : 'diag_level   ' 
    rdiag_level = MIN(4.,float(diag_level))
    call HDF5_save(file_id,rdiag_level,'diag_level'//char(0))
    Nbr   = 1
    Nbc   = geom%Nr+1
    !--> save : 'rg    '
    call HDF5_save(file_id,geom%rg,Nbc,'rg'//char(0))
    !--> save : 'nvpoloGCr_turb_diag '
    call HDF5_save(file_id,nvpoloGCr_turb_diag,Nbc, &
      'nvpoloGCr_turb'//char(0))
    !--> save : 'nvpoloGCr_neo_diag '
    call HDF5_save(file_id,nvpoloGCr_neo_diag,Nbc, &
      'nvpoloGCr_neo'//char(0))
    !--> save : 'nvpoloGCr_vpar_diag '
    call HDF5_save(file_id,nvpoloGCr_vpar_diag,Nbc,&
      'nvpoloGCr_vpar'//char(0))
    !--> save : 'niGCr_diag   '
    call HDF5_save(file_id,niGCr_diag,Nbc, &
      'niGCr'//char(0))
    !--> save : 'pressGCr_diag   '
    call HDF5_save(file_id,pressGCr_diag,Nbc, &
      'pressGCr'//char(0))
    !--> save : 'pressGCr_perp_diag   '
    call HDF5_save(file_id,pressGCr_perp_diag,Nbc, &
      'pressGCr_perp'//char(0))
    !--> save : 'pressGCr_par_diag   '
    call HDF5_save(file_id,pressGCr_par_diag,Nbc, &
      'pressGCr_par'//char(0))
    !--> save : 'LphiGCr_diag   '
    call HDF5_save(file_id,LphiGCr_diag,Nbc, &
      'LphiGCr'//char(0))
    !--> save : 'TpolGCr   '
    call HDF5_save(file_id,TpolGCr_diag,Nbc, &
      'TpolGCr'//char(0))
    !--> save : 'GammaGCr_turb_diag   '
    call HDF5_save(file_id,GammaGCr_turb_diag,Nbc, &
      'GammaGCr_turb'//char(0))
    !--> save : 'GammaGCr_neo_diag   '
    call HDF5_save(file_id,GammaGCr_neo_diag,Nbc, &
      'GammaGCr_neo'//char(0))
    !--> save : 'RSphiGCr_turb_diag   '
    call HDF5_save(file_id,RSphiGCr_turb_diag,Nbc, &
      'RSphiGCr_turb'//char(0))
    !--> save : 'RSphiGCr_neo_diag   '
    call HDF5_save(file_id,RSphiGCr_neo_diag,Nbc, &
      'RSphiGCr_neo'//char(0))
    !--> save : 'dpressGCr_dt_diag   '
    call HDF5_save(file_id,dpressGCr_dt_diag,Nbc, &
      'dpressGCr_dt'//char(0))
    !--> save : 'dLphiGCr_dt_diag   '
    call HDF5_save(file_id,dLphiGCr_dt_diag,Nbc, &
      'dLphiGCr_dt'//char(0))
    !--> save : 'QGCr_turb_diag   '
    call HDF5_save(file_id,QGCr_turb_diag,Nbc, &
      'QGCr_turb'//char(0))
    !--> save : 'QGCr_neo_diag   '
    call HDF5_save(file_id,QGCr_neo_diag,Nbc, &
      'QGCr_neo'//char(0))
    !--> save : 'QGCr_perp_turb_diag   '
    call HDF5_save(file_id,QGCr_perp_turb_diag, &
      Nbc,'QGCr_perp_turb'//char(0))
    !--> save : 'QGCr_perp_neo_diag   '
    call HDF5_save(file_id,QGCr_perp_neo_diag, &
      Nbc,'QGCr_perp_neo'//char(0))
    !--> save : 'QGCr_par_turb_diag   '
    call HDF5_save(file_id,QGCr_par_turb_diag,Nbc, &
      'QGCr_par_turb'//char(0))
    !--> save : 'QGCr_par_neo_diag   '
    call HDF5_save(file_id,QGCr_par_neo_diag,Nbc, &
      'QGCr_par_neo'//char(0))
    !--> save : 'QGCr_dtvpar_diag   '
    call HDF5_save(file_id,QGCr_dtvpar_diag,Nbc, &
      'QGCr_dtvpar'//char(0))
    !--> save : 'dpGCr_dt_perp_diag   '
    call HDF5_save(file_id,dpGCr_dt_perp_diag,Nbc, &
      'dpGCr_dt_perp'//char(0))
    !--> save : 'dpGCr_dt_par_diag   '
    call HDF5_save(file_id,dpGCr_dt_par_diag,Nbc, &
      'dpGCr_dt_par'//char(0))
    !--> save : 'SGCr_pperp_par_diag   '
    call HDF5_save(file_id,SGCr_pperp_par_diag,Nbc, &
      'SGCr_pperp_par'//char(0))
    !--> save : 'SGCr_pperp_turb_diag   '
    call HDF5_save(file_id,SGCr_pperp_turb_diag,Nbc, &
      'SGCr_pperp_turb'//char(0))
    !--> save : 'heatexGCr_turb_diag   '
    call HDF5_save(file_id,heatexGCr_turb_diag,Nbc, &
      'heatexGCr_turb_diag'//char(0))
    !--> save : 'heatexGCr_neo_diag   '
    call HDF5_save(file_id,heatexGCr_neo_diag,Nbc, &
      'heatexGCr_neo_diag'//char(0))
    !--> save : 'QGC_loc_diag'
    Nbr   = geom%Nr+1
    Nbc   = geom%Ntheta+1
    call HDF5_save(file_id,QGC_loc_diag(0:,0:),&
      Nbr,Nbc,'QGC_loc'//char(0))
    !--> save : 'RSphiGC_loc_diag'
    call HDF5_save(file_id,RSphiGC_loc_diag(0:,0:),&
      Nbr,Nbc,'RSphiGC_loc'//char(0))
    call HDF5_close(file_id)
#else
    call ASCII_rprof_GC_saving(geom,Snb_diag)
#endif
  end subroutine HDF5_rprof_GC_saving


  !---------------------------------------------- 
  ! Output saving of radial profiles linked to 
  !  particles
  !  -> directly in HDF5 binary format
  !----------------------------------------------   
  subroutine HDF5_rprof_part_saving(geom,Snb_diag,Stime_diag)  
    use globals, only : pSpecies_id
    use physics_part_module
    use poisson_class , only : Phi00_diag, &
      Phi_FSavg_diag
    type(geometry), intent(in) :: geom
    integer       , intent(in) :: Snb_diag
    real(RKIND)   , intent(in) :: Stime_diag
     
#ifndef NO_HDF5
    character(LEN=100) :: rprof_part_file    
    character(LEN=3)   :: num_species
    integer(HID_T)     :: file_id   
    integer            :: Nbr, Nbc, icr
    real(RKIND)        :: rdiag_level
    integer            :: ierr

    write(num_species,'(A2,I1)') "sp", pSpecies_id+1
    if (Snb_diag.ne.-1) then
      if (pSpecies_id .eq. 0) then
        write(rprof_part_file,'(A,'//numfmt//',A)') &
          "rprof_part", Snb_diag, ".h5"
      else
        write(rprof_part_file,'((2A),'//numfmt//',A)') &
          num_species, "/rprof_part", Snb_diag, ".h5"
      end if
    else
      if (pSpecies_id .eq. 0) then
        rprof_part_file = "rprof_part_init.h5"
      else
        write(rprof_part_file,'(2A)') num_species, "/rprof_part_init.h5"
      end if
    end if
    call HDF5_create(trim(rprof_part_file),file_id,ierr)
    if (ierr.ne.0) then
      print*,'pglobal_id = ',pglobal_id, &
        ' ==> error for opening of ',rprof_part_file
    end if

    !*** code version saving ***
    call HDF5_save(file_id,code_version, &
      'code_version'//char(0))

    !--> save : 'time_diag   ' 
    Nbr   = 1
    Nbc   = 1
    call HDF5_save(file_id,Stime_diag,'time_diag'//char(0))
    !--> save : 'deltat   ' 
    call HDF5_save(file_id,deltat,'deltat'//char(0))
    !--> save : 'diag_level   ' 
    rdiag_level = MIN(4.,float(diag_level))
    call HDF5_save(file_id,rdiag_level,'diag_level'//char(0))
    Nbr   = 1
    Nbc   = geom%Nr+1
    !--> save : 'rg    '
    call HDF5_save(file_id,geom%rg,Nbc,'rg'//char(0))
    !--> save : 'nvpolor_mag_diag '
    call HDF5_save(file_id,nvpolor_mag_diag,Nbc, &
      'nvpolor_mag'//char(0))
    !--> save : 'nir_diag    '
    call HDF5_save(file_id,nir_diag,Nbc,'nir'//char(0))
    !--> save : 'nir_trapped_diag    '
    call HDF5_save(file_id,nir_trapped_diag, &
      Nbc,'nir_trapped'//char(0))
    !--> save : 'nuparr_diag    '
    call HDF5_save(file_id,nuparr_diag,Nbc, &
      'nuparr'//char(0))
    !--> save : 'pressr_diag '
    call HDF5_save(file_id,pressr_diag,Nbc, &
      'pressr'//char(0))
    !--> save : 'Qr_turb_diag   '
    call HDF5_save(file_id,Qr_turb_diag,Nbc, &
      'Qr_turb'//char(0))
    !--> save : 'Qr_neo_diag   '
    call HDF5_save(file_id,Qr_neo_diag,Nbc, &
      'Qr_neo'//char(0))
    !--> save : 'Phi00_diag '
    call HDF5_save(file_id,Phi00_diag,Nbc, &
      'Phi00'//char(0))
    !--> save : 'Phi_FSavg_diag '
    call HDF5_save(file_id,Phi_FSavg_diag,Nbc, &
      'Phi_FSavg'//char(0))
    !--> save : 'Gammar_turb_diag   '
    call HDF5_save(file_id,Gammar_turb_diag,Nbc, &
      'Gammar_turb'//char(0))
    !--> save : 'Gammar_neo_diag   '
    call HDF5_save(file_id,Gammar_neo_diag,Nbc, &
      'Gammar_neo'//char(0))
    !--> save : 'RSthetar_diag   '
    call HDF5_save(file_id,RSthetar_diag,Nbc, &
      'RSthetar'//char(0))
    !--> save : 'RSparr_diag   '
    call HDF5_save(file_id,RSparr_diag,Nbc, &
      'RSparr'//char(0))
    call HDF5_close(file_id)
#else
    call ASCII_rprof_part_saving(geom,Snb_diag)
#endif
  end subroutine HDF5_rprof_part_saving


  !---------------------------------------------- 
  ! Output saving of radial profiles linked 
  !  to guiding-centers
  !  -> in ASCII format
  !----------------------------------------------   
  subroutine ASCII_rprof_GC_saving(geom,Snb_diag)  
    use globals, only : pSpecies_id
    use physics_GC_module
    use Io_Unit_module, only: IU_open, IU_close
    
    type(geometry), intent(in) :: geom
    integer       , intent(in) :: Snb_diag

    integer:: uprof
    character(LEN=100) :: rprof_GC_file
    character(LEN=3)   :: num_species
    integer            :: ir

    write(num_species,'(A2,I1)') "sp", pSpecies_id+1
    if (Snb_diag.ne.-1) then
      if (pSpecies_id .eq. 0) then
        write(rprof_GC_file,'(A,'//numfmt//',A)') &
          "rprof_GC", Snb_diag, ".dat"
      else
        write(rprof_GC_file,'((2A),'//numfmt//',A)') &
          num_species, "/rprof_GC", Snb_diag, ".dat"
      end if
    else
      if (pSpecies_id .eq. 0) then
        rprof_GC_file = "rprof_GC_init.dat"
      else
        write(rprof_GC_file,'(2A)') num_species, "/rprof_GC_init.dat"
      end if
    end if
    call IU_open(uprof, file = rprof_GC_file, status = 'UNKNOWN', &
      form = 'FORMATTED')
    write(uprof,'(A12,7A20)') &
      'rg', 'nvpoloGCr_turb', 'nvpoloGCr_neo', 'nvpoloGCr_vpar', &
      'QGCr_turb', 'QGCr_neo', 'niGCr', 'pressGCr'
    do ir = 0,geom%Nr
      write(uprof,'(1pe12.3,7(1pe20.12))') &
        geom%rg(ir), nvpoloGCr_turb_diag(ir), &
        nvpoloGCr_neo_diag(ir), nvpoloGCr_vpar_diag(ir), &
        QGCr_turb_diag(ir), QGCr_neo_diag(ir), &
        niGCr_diag(ir), pressGCr_diag(ir)
    end do
    call IU_close(uprof)
  end subroutine ASCII_rprof_GC_saving


  !---------------------------------------------- 
  ! Output saving of radial profiles linked 
  !  to particles
  !  -> in ASCII format
  !----------------------------------------------   
  subroutine ASCII_rprof_part_saving(geom,Snb_diag)  
    use globals, only : pSpecies_id
    use physics_part_module
    use poisson_class , only : Phi00_diag
    use Io_Unit_module, only: IU_open, IU_close

    type(geometry), intent(in) :: geom
    integer       , intent(in) :: Snb_diag

    integer:: uprof
    character(LEN=100) :: rprof_part_file    
    character(LEN=3)   :: num_species
    integer            :: ir

    write(num_species,'(A2,I1)') "sp", pSpecies_id+1
    if (Snb_diag.ne.-1) then
      if (pSpecies_id .eq. 0) then
        write(rprof_part_file,'(A,'//numfmt//',A)') &
          "rprof_part", Snb_diag, ".dat"
      else
        write(rprof_part_file,'((2A),'//numfmt//',A)') &
          num_species, "/rprof_part", Snb_diag, ".dat"
      end if
    else
      if (pSpecies_id .eq. 0) then
        rprof_part_file = "rprof_part_init.dat"
      else
        write(rprof_part_file,'(2A)') num_species, "/rprof_part_init.dat"
      end if
    end if
    call IU_open(uprof, file = rprof_part_file, status = 'UNKNOWN', &
      form = 'FORMATTED')
    write(uprof,'(A12,4A20)') &
      'rg', 'Phi00', 'nvpolor_mag','nir', 'nuparr'
    do ir = 0,geom%Nr
      write(uprof,'(1pe12.3,11(1pe20.12))') &
        geom%rg(ir), Phi00_diag(ir), &
        nvpolor_mag_diag(ir), nir_diag(ir), &
        nuparr_diag(ir)
    end do
    call IU_close(uprof)
  end subroutine ASCII_rprof_part_saving


  !---------------------------------------------- 
  ! Output saving of 2D cross-sections of Phi
  !  -> directly in HDF5 binary format
  !----------------------------------------------   
  subroutine HDF5_Phi2D_saving(geom,Snb_diag,Stime_diag)  
    use globals, only : pSpecies_id
    use Pcross_section_module, only : Phirth_diag, &
      Phirphi_diag, Phithphi_diag
    type(geometry), intent(in) :: geom
    integer       , intent(in) :: Snb_diag
    real(RKIND)   , intent(in) :: Stime_diag
     
#ifndef NO_HDF5
    character(LEN=100) :: Phi2D_file    
    character(LEN=3)   :: num_species
    integer(HID_T)     :: file_id        
    integer            :: Nbr, Nbc, icr
    integer            :: ierr

    write(num_species,'(A2,I1)') "sp", pSpecies_id+1
    if (Snb_diag.ne.-1) then
      if (pSpecies_id .eq. 0) then
        write(Phi2D_file,'(A,'//numfmt//',A)') &
          "Phi2D", Snb_diag, ".h5"
      else
        write(Phi2D_file,'((2A),'//numfmt//',A)') &
          num_species, "/Phi2D", Snb_diag, ".h5"
      end if
    else
      if (pSpecies_id .eq. 0) then
        Phi2D_file = "Phi2D_init.h5"
      else
        write(Phi2D_file,'(2A)') num_species, "/Phi2D_init.h5"
      end if
    end if
    call HDF5_create(trim(Phi2D_file),file_id,ierr)
    if (ierr.ne.0) then
      print*,'pglobal_id = ',pglobal_id, &
        ' ==> error for opening of ',Phi2D_file
    end if

    !*** code version saving ***
    call HDF5_save(file_id,code_version, &
      'code_version'//char(0))

    !--> save ir_Phi, itheta_Phi, iphi_Phi
    call HDF5_save(file_id,ir_Phi,'ir_Phi'//char(0))
    call HDF5_save(file_id,itheta_Phi, &
      'itheta_Phi'//char(0))
    call HDF5_save(file_id,iphi_Phi,'iphi_Phi'//char(0))

    !--> save : 'time_diag   ' 
    call HDF5_save(file_id,Stime_diag,'time_diag'//char(0))
    !--> save : 'deltat   ' 
    call HDF5_save(file_id,deltat,'deltat'//char(0))
    !--> save : 'Phirth_diag '
    Nbr    = geom%Nr+1
    Nbc    = geom%Ntheta+1
    call HDF5_save(file_id,Phirth_diag(0:,0:),Nbr,Nbc, &
      'Phirth'//char(0))
    !--> save : 'Phirphi_diag   '
    Nbr    = geom%Nr+1
    Nbc    = geom%Nphi+1
    call HDF5_save(file_id,Phirphi_diag(0:,0:),Nbr,Nbc, &
      'Phirphi'//char(0))
    !--> save : 'Phithphi_diag  '
    Nbr    = geom%Ntheta+1
    Nbc    = geom%Nphi+1
    call HDF5_save(file_id,Phithphi_diag(0:,0:),Nbr,Nbc, &
      'Phithphi'//char(0))

    call HDF5_close(file_id)
#endif
  end subroutine HDF5_Phi2D_saving


  !---------------------------------------------- 
  ! Output saving of 2D cross-sections of f
  !  -> directly in HDF5 binary format
  !----------------------------------------------   
  subroutine HDF5_f2D_saving(geom,Snb_diag,Stime_diag)
    use globals, only : pSpecies_id, diag_level
    use globals, only : ir_f2D_trapped, &
      itheta_f2D_trapped, iphi_f2D_trapped, &
      ivpar_f2D_trapped, imu_f2D_trapped, &
      ir_f2D_passing, itheta_f2D_passing, &
      iphi_f2D_passing, ivpar_f2D_passing, &
      imu_f2D_passing
    use Pcross_section_module, only : &
      frtheta_v0_diag, frtheta_vmax_diag, &
      fphivpar_mu0_diag, fphivpar_mumax_diag, &
      frvpar_mu0_diag, frvpar_mumax_diag, &
      fthetavpar_mu0_diag, fthetavpar_mumax_diag, &
      fvparmu_diag, ir_fFS, ir_fFS_p15, ir_fFS_p30, &
      ir_fFS_p40, ir_fFS_m15, ir_fFS_m30, ir_fFS_m40, & 
      fFSavg_irp_vparmu_diag, fFSavg_irp_p15_vparmu_diag, &  
      fFSavg_irp_p30_vparmu_diag, fFSavg_irp_p40_vparmu_diag, & 
      fFSavg_irp_m15_vparmu_diag, fFSavg_irp_m30_vparmu_diag, & 
      fFSavg_irp_m40_vparmu_diag                        
    use globals, only : atleast1_source
    use Pcross_section_module, only : irS, fFSavg_irS_vparmu_diag
    
    type(geometry), intent(in) :: geom
    integer       , intent(in) :: Snb_diag
    real(RKIND)   , intent(in) :: Stime_diag
     
#ifndef NO_HDF5
    character(LEN=100) :: f2D_file    
    character(LEN=3)   :: num_species
    integer(HID_T)     :: file_id        
    integer            :: Nbr, Nbc, icr
    integer            :: ierr

    write(num_species,'(A2,I1)') "sp", pSpecies_id+1
    if (Snb_diag.ne.-1) then
      if (pSpecies_id .eq. 0) then
        write(f2D_file,'(A,'//numfmt//',A)') &
          "f2D", Snb_diag, ".h5"
      else
        write(f2D_file,'((2A),'//numfmt//',A)') &
          num_species, "/f2D", Snb_diag, ".h5"
      end if
    else
      if (pSpecies_id .eq. 0) then
        f2D_file = "f2D_init.h5"
      else
        write(f2D_file,'(2A)') num_species, "/f2D_init.h5"
      end if
    end if
    call HDF5_create(trim(f2D_file),file_id,ierr)
    if (ierr.ne.0) then
      print*,'pglobal_id = ',pglobal_id, &
        ' ==> error for opening of ',f2D_file
    end if

    !*** code version saving ***
    call HDF5_save(file_id,code_version, &
      'code_version'//char(0))

    !--> save : ir_diag, itheta_diag, iphi_diag, ivpar_diag, &
    !    imu_diag for trapped and passing particules
    call HDF5_save(file_id,ir_f2D_trapped, &
      'ir_f2D_trapped'//char(0))
    call HDF5_save(file_id,itheta_f2D_trapped, &
      'itheta_f2D_trapped'//char(0))
    call HDF5_save(file_id,iphi_f2D_trapped, &
      'iphi_f2D_trapped'//char(0))
    call HDF5_save(file_id,ivpar_f2D_trapped, &
      'ivpar_f2D_trapped'//char(0))
    call HDF5_save(file_id,imu_f2D_trapped, &
      'imu_f2D_trapped'//char(0))
    call HDF5_save(file_id,ir_f2D_passing, &
      'ir_f2D_passing'//char(0))
    call HDF5_save(file_id,itheta_f2D_passing, &
      'itheta_f2D_passing'//char(0))
    call HDF5_save(file_id,iphi_f2D_passing, &
      'iphi_f2D_passing'//char(0))
    call HDF5_save(file_id,ivpar_f2D_passing, &
      'ivpar_f2D_passing'//char(0))
    call HDF5_save(file_id,imu_f2D_passing, &
      'imu_f2D_passing'//char(0))
    !--> save : 'time_diag   ' 
    call HDF5_save(file_id,Stime_diag,'time_diag'//char(0))
    !--> save : 'deltat   ' 
    call HDF5_save(file_id,deltat,'deltat'//char(0))
    !--> save : 'frth_v0_diag  '
    Nbr    = geom%Nr+1
    Nbc    = geom%Ntheta+1
    call HDF5_save(file_id,frtheta_v0_diag(0:,0:), &
      Nbr,Nbc,'frthv0'//char(0))
    !--> save : 'frth_vm_diag  '
    Nbr    = geom%Nr+1
    Nbc    = geom%Ntheta+1
    call HDF5_save(file_id,frtheta_vmax_diag(0:,0:), &
      Nbr,Nbc,'frthvm'//char(0))
    !--> save : 'fphivpar_mu0_diag '
    Nbr    = geom%Nphi+1
    Nbc    = geom%Nvpar+1
    call HDF5_save(file_id,fphivpar_mu0_diag(0:,0:), &
      Nbr,Nbc,'fphivpar_mu0'//char(0))
    !--> save : 'fphivpar_mumax_diag '
    Nbr    = geom%Nphi+1
    Nbc    = geom%Nvpar+1
    call HDF5_save(file_id,fphivpar_mumax_diag(0:,0:), &
      Nbr,Nbc,'fphivpar_mumax'//char(0))
    !--> save : 'frvpar_mu0_diag '
    Nbr    = geom%Nr+1
    Nbc    = geom%Nvpar+1
    call HDF5_save(file_id,frvpar_mu0_diag(0:,0:), &
      Nbr,Nbc,'frvpar_mu0'//char(0))
    !--> save : 'frvpar_mumax_diag '
    Nbr    = geom%Nr+1
    Nbc    = geom%Nvpar+1
    call HDF5_save(file_id,frvpar_mumax_diag(0:,0:), &
      Nbr,Nbc,'frvpar_mumax'//char(0))
    !--> save : 'fthetavpar_mu0_diag '
    Nbr    = geom%Ntheta+1
    Nbc    = geom%Nvpar+1
    call HDF5_save(file_id,fthetavpar_mu0_diag(0:,0:), &
      Nbr,Nbc,'fthvpar_mu0'//char(0))
    !--> save : 'fthetavpar_mumax_diag '
    Nbr    = geom%Ntheta+1
    Nbc    = geom%Nvpar+1
    call HDF5_save(file_id,fthetavpar_mumax_diag(0:,0:), &
      Nbr,Nbc,'fthvpar_mumax'//char(0))
    !--> save : 'fvparmu_diag '
    Nbr    = geom%Nvpar+1
    Nbc    = geom%Nmu+1
    call HDF5_save(file_id,fvparmu_diag(0:,0:), &
      Nbr,Nbc,'fvparmu'//char(0))
    if (diag_level.ge.6) then
      !--> save : 'fFSavg_irp_vparmu_diag '
      Nbr    = geom%Nvpar+1
      Nbc    = geom%Nmu+1
      call HDF5_save(file_id, &
        fFSavg_irp_vparmu_diag(0:,0:),Nbr,Nbc, &
        'fFSavg_irp_vparmu'//char(0))
      call HDF5_save(file_id,ir_fFS,'ir_fFS'//char(0))
      !--> save <f>_FS around ir_fFS +- 0.15%
      call HDF5_save(file_id,ir_fFS_m15,'ir_fFS_m15'//char(0))
      call HDF5_save(file_id, &
        fFSavg_irp_m15_vparmu_diag(0:,0:),Nbr,Nbc, &
        'fFSavg_irp_m15_vparmu'//char(0))
      call HDF5_save(file_id,ir_fFS_p15,'ir_fFS_p15'//char(0))
      call HDF5_save(file_id, &
        fFSavg_irp_p15_vparmu_diag(0:,0:),Nbr,Nbc, &
        'fFSavg_irp_p15_vparmu'//char(0))
      !--> save <f>_FS around ir_fFS +- 0.30%
      call HDF5_save(file_id,ir_fFS_m30,'ir_fFS_m30'//char(0))
      call HDF5_save(file_id, &
        fFSavg_irp_m30_vparmu_diag(0:,0:),Nbr,Nbc, &
        'fFSavg_irp_m30_vparmu'//char(0))
      call HDF5_save(file_id,ir_fFS_p30,'ir_fFS_p30'//char(0))
      call HDF5_save(file_id, &
        fFSavg_irp_p30_vparmu_diag(0:,0:),Nbr,Nbc, &
        'fFSavg_irp_p30_vparmu'//char(0))
      !--> save <f>_FS around ir_fFS +- 0.40%
      call HDF5_save(file_id,ir_fFS_m40,'ir_fFS_m40'//char(0))
      call HDF5_save(file_id, &
        fFSavg_irp_m40_vparmu_diag(0:,0:),Nbr,Nbc, &
        'fFSavg_irp_m40_vparmu'//char(0))
      call HDF5_save(file_id,ir_fFS_p40,'ir_fFS_p40'//char(0))
      call HDF5_save(file_id, &
        fFSavg_irp_p40_vparmu_diag(0:,0:),Nbr,Nbc, &
        'fFSavg_irp_p40_vparmu'//char(0))
    end if
    if (atleast1_source) then
      call HDF5_save(file_id,irS,'irS'//char(0))
      !--> save : 'fFSavg_irS_vparmu_diag '
      Nbr    = geom%Nvpar+1
      Nbc    = geom%Nmu+1
      call HDF5_save(file_id, &
        fFSavg_irS_vparmu_diag(0:,0:),Nbr,Nbc, &
        'fFSavg_irS_vparmu'//char(0))
    end if

    call HDF5_close(file_id)
#endif
  end subroutine HDF5_f2D_saving


  !--------------------------------------------------------
  ! Output saving of Dcoll_rvpar, Dsemi_rvpar, Vcoll_rvpar,
  ! collision_freqr_diag and envelope_r_diag
  !  -> directly in HDF5  binary format
  !--------------------------------------------------------
  subroutine HDF5_collision_saving(geom,Snb_diag,Stime_diag)
    use globals, only : pSpecies_id
    use collisions_vpar_module, only : &
      collision_freqr_diag, nustarr_diag, &
      Dcoll_rvpar, Vcoll_rvpar, envelope_r_diag, &
      Vmean_r_avg_diag, Tmean_r_avg_diag, &
      Vmean_r_diag, Tmean_r_diag, &
      Vmean_thetaphi_diag, Tmean_thetaphi_diag
    use species_transf_module, only : Ns_FSavg, &
      Vs_FSavg, Ts_FSavg
    type(geometry), intent(in) :: geom
    integer       , intent(in) :: Snb_diag
    real(RKIND)   , intent(in) :: Stime_diag

#ifndef NO_HDF5
    character(LEN=100) :: coll_file
    character(LEN=3)   :: num_species
    integer(HID_T)     :: file_id   
    integer            :: Nb1, Nbc, Nbr, icr
    integer            :: ierr

    write(num_species,'(A2,I1)') "sp", pSpecies_id+1
    if (Snb_diag.ne.-1) then
      if (pSpecies_id .eq. 0) then
        write(coll_file,'(A,'//numfmt//',A)') &
          "collisions", Snb_diag, ".h5"
      else
        write(coll_file,'((2A),'//numfmt//',A)') &
          num_species, "/collisions", Snb_diag, ".h5"
      end if
    else
      if  (pSpecies_id .eq. 0) then
        coll_file = "collisions_init.h5"
      else
        write(coll_file,'(2A)') num_species, "/collisions_init.h5"
      end if
    end if
    call HDF5_create(trim(coll_file),file_id,ierr)
    if (ierr.ne.0) then
      print*,'pglobal_id = ',pglobal_id, &
        ' ==> error for opening of ',coll_file
    end if

    !*** code version saving ***
    call HDF5_save(file_id,code_version, &
      'code_version'//char(0))

    !--> save : 'time_diag   ' 
    call HDF5_save(file_id,Stime_diag,'time_diag'//char(0))
    !--> save : nustar
    Nbr = 1
    Nbc = geom%Nr+1
    !--> save : 'envelope_r'
    call HDF5_save(file_id,envelope_r_diag,Nbc, &
      'envelope_r'//char(0))
    !--> save : 'collision_freq_r'
    call HDF5_save(file_id,collision_freqr_diag,Nbc, &
      'collision_freq_r'//char(0))
    !--> save : 'nustar_r'
    call HDF5_save(file_id,nustarr_diag,Nbc, &
      'nustar_r'//char(0))
    !--> save : 'Vmean_r_avg    '
    call HDF5_save(file_id,Vmean_r_avg_diag,Nbc, &
      'Vmean_r_avg'//char(0))
    !--> save : 'Tmean_r_avg    '
    call HDF5_save(file_id,Tmean_r_avg_diag,Nbc, &
      'Tmean_r_avg'//char(0))
    if (version_coll.eq.2) then
      !--> save : 'Vmean_r    '
      call HDF5_save(file_id,Vmean_r_diag,Nbc, &
        'Vmean_r'//char(0))
      !--> save : 'Tmean_r    '
      call HDF5_save(file_id,Tmean_r_diag,Nbc, &
        'Tmean_r'//char(0))
      Nbr = geom%Ntheta+1
      Nbc = geom%Nphi+1
      !--> save : 'Vmean_thetaphi    '
      call HDF5_save(file_id,Vmean_thetaphi_diag, &
        Nbr,Nbc,'Vmean_thetaphi'//char(0))
      !--> save : 'Tmean_thetaphi    '
      call HDF5_save(file_id,Tmean_thetaphi_diag, &
        Nbr,Nbc,'Tmean_thetaphi'//char(0))
    end if
    if ( (energy_transfer).or.(moment_transfer) ) then
      Nbr = Nspecies
      Nbc = geom%Nr+1
      !--> save :  'Ns_FSavg'
      call HDF5_save(file_id,Ns_FSavg, &
        Nbr,Nbc,'Ns_FSavg'//char(0))
      !--> save :  'Vs_FSavg'
      call HDF5_save(file_id,Vs_FSavg, &
        Nbr,Nbc,'Vs_FSavg'//char(0))
      !--> save :  'Ts_FSavg'
      call HDF5_save(file_id,Ts_FSavg, &
        Nbr,Nbc,'Ts_FSavg'//char(0))
    end if
    call HDF5_close(file_id)
#endif
  end subroutine HDF5_collision_saving
  
end module output_saving_module
