!**************************************************************
!  Copyright Euratom-CEA
!  Authors : 
!     Virginie Grandgirard (virginie.grandgirard@cea.fr)
!     Chantal Passeron (chantal.passeron@cea.fr)
!     Guillaume Latu (guillaume.latu@cea.fr)
!     Xavier Garbet (xavier.garbet@cea.fr)
!     Philippe Ghendrih (philippe.ghendrih@cea.fr)
!     Yanick Sarazin (yanick.sarazin@cea.fr)
!  
!  This code GYSELA (for GYrokinetic SEmi-LAgrangian) 
!  is a 5D gyrokinetic global full-f code for simulating 
!  the plasma turbulence in a tokamak.
!  
!  This software is governed by the CeCILL-B license 
!  under French law and abiding by the rules of distribution 
!  of free software.  You can  use, modify and redistribute 
!  the software under the terms of the CeCILL-B license as 
!  circulated by CEA, CNRS and INRIA at the following URL
!  "http://www.cecill.info". 
!**************************************************************

#include "config-F.inc"
#include "code_instrumentation/debug.inc"

!----------------------------------------------------------------- 
! file : physics_GC.F90
! date : 16/04/2010
!  Compute all the physics quantities dedicated
!   to the guiding-centers.
!    -> for the poloidal velocities
!      . nvpoloGCr_turb (due to ExB velocity)
!      . nvpoloGCr_neo (due to curvature)
!      . nvpoloGCr_vpar  (due to parallel velocity)
!    -> niGCr            : guiding-center density
!    -> pressGCr         : guiding-center pressure
!    -> pressGCr_perp    : guiding-center perpendicular pressure
!    -> pressGCr_par     : guiding-center parallel pressure
!    -> for toroidal angular momentum conservation
!      . LphiGCr         : toroidal angular momentum
!      . TpolGCr         : polarization term
!      . GammaGCr_turb   : guiding-center turbulent flux
!                          (due to ExB velocity)
!      . GammaGCr_neo    : guiding-center neoclassical flux
!      . RSphiGCr_turb   : toroidal Reynolds stress due to ExB
!      . RSphiGCr_neo    : neoclassical toroidal Reynolds stress
!    -> dpressGCr_dt     : time derivative of the guiding-center
!                          pressure: (P(tn+1)-P(tn))/dt
!    -> dLphiGCr_dt      : time derivative of the toroidal angular
!                           momentum: (Lphi(tn+1)-Lphi(tn))/dt
!    -> QGCr_turb        : guiding-center heat flux due to ExB
!    -> QGCr_neo         : guiding-center heat flux due to 
!                          curvature drift
!    -> QGCr_perp_turb   : guiding-center perpendicular heat flux
!                          due to ExB
!    -> QGCr_perp_neo    : guiding-center perpendicular neoclassical 
!                          heat flux
!    -> QGCr_par_turb    : guiding-center parallel heat flux due to ExB
!    -> QGCr_par_neo     : guiding-center parallel neoclassical heat flux 
!    -> QGCr_dtvpar      : guiding-center dtvpar term 
!                         (in parallel pressure balance)
!    -> dpGCr_dt_perp    : time derivative of the perpendicular pressure
!    -> dpGCr_dt_par     : time derivative of the parallel pressure
!    -> for transverse energy conservation 
!      . SGCr_Pperp_par  : parallel flux of perpendicular energy 
!      . SGCr_Pperp_turb : turbulent flux of perpendicular energy
!    -> for turbulent heating exchange
!      . heatexGCr_turb  : ExB part of the turbulent heating exchange
!      . heatexGCr_neo   : turbulent heating exchange due to 
!                           curvature drift
!    -> 2D diagnostics (r,theta)
!      . QGC_loc     : guiding-center heat flux due to ExB
!      . RSphiGC_loc : toroidal Reynolds stress due to ExB
!    -> entropy    
!    -> L2 norm    
!    -> kinetic energy
!
! Rk: All the previous radial profiles correspond to
!     flux surface average, i.e:
!       <.>_FS = \int . Js dtheta dphi /
!                \int Js dtheta dphi 
!      with Js the jacobian in space
!----------------------------------------------------------------- 
module physics_GC_module
  use clock_module, &
                  only: TIMEPREC
  use coord_system_class, &
                  only: coord_system
  use Debug_module
  use fdistribu5d_class, &
                  only: fdistribu5d
  use geometry_class, &
                  only: geometry
  use globals,    only: DIAG_CL, DIAG_RPROF_GC, DIAG_RPROF_PART
  use gyroaverage_class, &
                  only: J0operator
  use init_current_class, &
                  only: init_current
  use init_magnetic_class, &
                  only: init_magnetic
  use init_profile_class, &
                  only: init_profile
  use OMPutils_module, &
                  only: omp1dvector
  use poisson_class, &
                  only: poisson
  use prec_const
  
  use mpiMod

  implicit none

  !-------------------------------------
  !  PRIVATE VARIABLES
  !------------------------------------- 
  !-> variables for the guiding-center moments
  real(RKIND), dimension(:,:), pointer, private :: GC_Fmoments
#ifdef NO_MPI_INPLACE
  real(RKIND), dimension(:,:), pointer, private :: GC_Fmoments_loc
#endif

  !-------------------------------------
  !  PUBLIC VARIABLES
  !------------------------------------- 
  public
  !--> for guiding-center radial profiles
  real(RKIND), dimension(:), pointer, public :: nvpoloGCr_turb_diag
  real(RKIND), dimension(:), pointer, public :: nvpoloGCr_neo_diag
  real(RKIND), dimension(:), pointer, public :: nvpoloGCr_vpar_diag
  real(RKIND), dimension(:), pointer, public :: niGCr_diag
  real(RKIND), dimension(:), pointer, public :: pressGCr_diag
  real(RKIND), dimension(:), pointer, public :: pressGCr_perp_diag
  real(RKIND), dimension(:), pointer, public :: pressGCr_par_diag
  real(RKIND), dimension(:), pointer, public :: LphiGCr_diag
  real(RKIND), dimension(:), pointer, public :: TpolGCr_diag
  real(RKIND), dimension(:), pointer, public :: GammaGCr_turb_diag
  real(RKIND), dimension(:), pointer, public :: GammaGCr_neo_diag
  real(RKIND), dimension(:), pointer, public :: RSphiGCr_turb_diag
  real(RKIND), dimension(:), pointer, public :: RSphiGCr_neo_diag
  real(RKIND), dimension(:), pointer, public :: dpressGCr_dt_diag
  real(RKIND), dimension(:), pointer, public :: dLphiGCr_dt_diag
  real(RKIND), dimension(:), pointer, public :: QGCr_turb_diag
  real(RKIND), dimension(:), pointer, public :: QGCr_neo_diag
  real(RKIND), dimension(:), pointer, public :: QGCr_perp_turb_diag
  real(RKIND), dimension(:), pointer, public :: QGCr_perp_neo_diag
  real(RKIND), dimension(:), pointer, public :: QGCr_par_turb_diag
  real(RKIND), dimension(:), pointer, public :: QGCr_par_neo_diag
  real(RKIND), dimension(:), pointer, public :: QGCr_dtvpar_diag
  real(RKIND), dimension(:), pointer, public :: dpGCr_dt_perp_diag
  real(RKIND), dimension(:), pointer, public :: dpGCr_dt_par_diag
  real(RKIND), dimension(:), pointer, public :: SGCr_Pperp_par_diag
  real(RKIND), dimension(:), pointer, public :: SGCr_Pperp_turb_diag
  real(RKIND), dimension(:), pointer, public :: heatexGCr_turb_diag
  real(RKIND), dimension(:), pointer, public :: heatexGCr_neo_diag

  !--> for entropy and L2norm 
  real(RKIND), public :: entropy_diag
  real(RKIND), public :: L2norm_diag
  !--> for kinetic energy
  real(RKIND), public :: Enkin_diag

  !-> for local (r,theta) diagnostics
  real(RKIND), dimension(:,:), pointer, public :: QGC_loc_diag
  real(RKIND), dimension(:,:), pointer, public :: RSphiGC_loc_diag

  include "gyrokinetic/Bstar_inline.h"
  include "gyrokinetic/velocities_inline.h"
  !******************************
  contains
  !******************************
#include "gyrokinetic/Bstar_inline.F90"
#include "gyrokinetic/velocities_inline.F90"


  !-----------------------------------------------------
  ! Constructor
  !-----------------------------------------------------
  subroutine new_physics_GC(geom)

    type(geometry), intent(in) :: geom

    call Physics_GC_moments_init(geom)
  end subroutine new_physics_GC


  !-----------------------------------------------------
  ! Destructor
  !-----------------------------------------------------
  subroutine delete_physics_GC()
    call Physics_GC_moments_cleanup()
  end subroutine delete_physics_GC


  !--------------------------------------------------------------
  ! Computation of the fluid moments associated to the 
  !  guiding-centers, corresponding to 
  !   1) nvpoloGC_turb(r)   = <\int f*vE_GC.etheta d3v>_FS
  !   2) nvpoloGC_neo(r)    = <\int f*vD.etheta d3v>_FS
  !   3) nvpoloGC_vpar(r)   = <\int f*vpar bstar.etheta d3v>_FS
  ! where vE_GC is the guiding center ExB velocity
  !   4) niGC(r)            = <\int f d3v>_FS 
  !   5) pressGC(r)         = <\int f*E d3v>_FS 
  !   6) pressGC_perp(r)    = <\int f*mu*B d3v>_FS 
  !   7) pressGC_par(r)     = <\int f*v_par^2 d3v>_FS 
  !   8) LphiGC(r)          = <\int f*vpar R^2*B*gradphi/B d3v>_FS
  !   9) TpolGC(r)          = <\int f*dJ0Phi/dphi d3v>_FS
  !  10) GammaGC_turb(r)    = <\int f*vE_GC.gradr d3v>_FS
  !  11) GammaGC_neo(r)     = <\int f*vD_GC.gradr d3v>_FS
  !  12) RSphiGC_turb(r)    = <\int f*vE_GC.gradr*vpar* 
  !                           R^2*B*gradphi/B d3v>_FS
  !  13) RSphiGC_neo(r)     = <\int f*vD_GC.gradr*vpar*
  !                           R^2*B*gradphi/B d3v>_FS
  !  14) dpressGC_dt(r)     = (pressGC(tn+1)-pressGC(tn))/dt
  !  15) dLphiGC_dt(r)      = (LphiGC(tn+1)-LphiGC(tn))/dt
  !  16) QGC_turb(r)        = <\int f*E*vE_GC.gradr d3v>_FS
  !  17) QGC_neo(r)         = <\int f*E*vD_GC.gradr d3v>_FS 
  !  18) QGCr_perp_turb(r)  = <\int f*mu*B*vE_GC.gradr d3v>_FS
  !  19) QGCr_perp_neo(r)   = <\int f*mu*B*vD_GC.gradr d3v>_FS
  !  20) QGCr_par_turb(r)   = <\int f*0.5*vpar*vpar*
  !                           vE_GC.gradr d3v>_FS
  !  21) QGCr_par_neo(r)    = <\int f*0.5*vpar*vpar*
  !                           vD_GC.gradr d3v>_FS
  !  22) QGCr_dtvpar(r)     = <\int f*v_par*dv_par/dt>
  !  23) dpGCr_dt_perp(r)   = (pressGC_perp(tn+1) - 
  !                            pressGC_perp(tn))/dt
  !  24) dpGCr_dt_par(r)    = (pressGC_par(tn+1)-pressGC_par(tn))/dt
  !  25) SGCr_Pperp_par(r)  = <\int f*(vpar.gradpar_B/B)*mu*B>_FS
  !  26) SGCr_Pperp_turb(r) = <\int f*(vE_GC.gradpar_B/B)*mu*B>_FS
  !  27) heatexGCr_turb(r)  = -hatZs* 
  !                           <\int f*(vExB_GC.grad(J0.Phi) d3v>_FS 
  !  28) heatexGCr_neo(r)   = -hatZs *
  !                           <\int f*(vD_GC.grad(J0.Phi) d3v>_FS 
  !  - the 2D diagnostics:
  !  29) QGC_loc(r,theta)     = \int f*E*vE_GC.gradr d3v dphi / 
  !                              (2*pi)
  !  30) RSphiGC_loc(r,theta) = \int f*vE_GC.gradr*vpar* 
  !                              R^2*B*gradphi/B d3v dphi/(2*pi)
  !  - entropy    
  !  - L2 norm    
  !  - kinetic energy
  !
  ! Rk2: <.>_FS corresponds to the flux surface average,
  !      i.e <.>_FS = \int . Js dtheta dphi /
  !                   \int Js dtheta dphi 
  !      with Js the jacobian in space
  !
  ! Rk3: the poloidal velocity nvpolo corresponds to the
  !     sum of the three previous moments
  !     nvpolo(r) = nvpoloGC_turb(r) + nvpoloGC_neo(r) + 
  !                 nvpoloGC_vpar(r)
  !       NB: vD contains the magnetisation velocity
  !--------------------------------------------------------------
  subroutine compute_GC_moments(geom, coord_sys, init_magnet, &
    init_curr, fn, fnp1, Sfmu_eq, SdJ0Phidr, &
    SdJ0Phidtheta, SdJ0Phidphi, Sdt_diag, Siter_time)

    use globals,    only: diag_targ, pInSpecies_id
#ifdef TIMER
    use globals,    only: pglobal_id, plocal_id
    use clock_module, &
                    only: clck_time, clck_ldiff
    use MPIutils_module, &
                    only: pp_reduce_vector
#endif

    type(geometry)               , intent(in) :: geom
    type(coord_system)           , intent(in) :: coord_sys
    type(init_magnetic)          , intent(in) :: init_magnet
    type(init_current)           , intent(in) :: init_curr
    type(fdistribu5d)            , intent(in) :: fn
    type(fdistribu5d)            , intent(in) :: fnp1
    real(RKIND), dimension(:,:,:), pointer    :: Sfmu_eq
    real(RKIND), dimension(:,:,:), pointer    :: SdJ0Phidr
    real(RKIND), dimension(:,:,:), pointer    :: SdJ0Phidtheta
    real(RKIND), dimension(:,:,:), pointer    :: SdJ0Phidphi
    real(RKIND)                  , intent(in) :: Sdt_diag
    real(RKIND)                  , intent(in) :: Siter_time

    real(RKIND) :: Enkin, entropy, L2norm
    real(RKIND), dimension(:,:), pointer :: QGC_loc
    real(RKIND), dimension(:,:), pointer :: RSphiGC_loc

    ! -> variables for time perfomance analysis
#ifdef TIMER
    integer, parameter :: NB_TIMER = 3
    integer(TIMEPREC)  :: tdeb, tfin
    real(RKIND)        :: tvect(NB_TIMER), tmin(NB_TIMER), &
                          tmax(NB_TIMER), tavg(NB_TIMER)
#endif

#ifdef TIMER
    call clck_time(tdeb)
#endif
    call Physics_GC_moments_preprocess( &
      geom,coord_sys,init_magnet,init_curr, &
      fn,fnp1,Sfmu_eq,SdJ0Phidr, &
      SdJ0Phidtheta,SdJ0Phidphi,Enkin, &
      entropy,L2norm,QGC_loc,RSphiGC_loc)
#ifdef TIMER
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tvect(1))
#endif

#ifdef TIMER
    call clck_time(tdeb)
#endif
    call Physics_GC_moments_transfer( &
      geom,Enkin,entropy,L2norm, &
      QGC_loc,RSphiGC_loc)
    
#ifdef TIMER
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tvect(2))
#endif

#ifdef TIMER
    call clck_time(tdeb)
#endif
    if ( pInSpecies_id == diag_targ(DIAG_RPROF_GC) ) then
      call Physics_GC_moments_postprocess(geom,Sdt_diag,Siter_time)
    end if
#ifdef TIMER
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tvect(3))
#endif

#ifdef TIMER
    call pp_reduce_vector(tvect,NB_TIMER,tmin,tmax,tavg)
    if (plocal_id .eq. 0) then
      write(6,'(I4,A,F13.7,A,F13.7,A,F13.7)') pglobal_id, &
            " TMIN GC moments: comp ", tmin(1)," comm ", tmin(2), &
            " comp2 ", tmin(3)
      write(6,'(I4,A,F13.7,A,F13.7,A,F13.7)') pglobal_id, &
            " TMAX GC moments: comp ", tmax(1)," comm ", tmax(2), &
            " comp2 ", tmax(3)
      write(6,'(I4,A,F13.7,A,F13.7,A,F13.7)') pglobal_id, &
            " TAVG GC moments: comp ", tavg(1)," comm ", tavg(2), &
            " comp2 ", tavg(3)
    end if
#endif
  end subroutine compute_GC_moments


  !-----------------------------------------------------------------
  !> allocation of arrays for output saving of guiding-center radial
  !> profiles
  !-----------------------------------------------------------------
  subroutine Physics_GC_moments_init ( geom )
    use globals          , only: nbmoments_GC
    use wrap_alloc_module
    
    type(geometry), intent(in) :: geom
    
    !*** allocation for fluid moments computation ***
#ifdef NO_MPI_INPLACE
    call FTR_ALLOC_2_ZERO(GC_Fmoments_loc,0,geom%Nr,1,nbmoments_GC)
#endif
    call FTR_ALLOC_2_ZERO(GC_Fmoments,0,geom%Nr, 1,nbmoments_GC)

    call FTR_ALLOC_1_ZERO(nvpoloGCr_turb_diag,0,geom%Nr)
    call FTR_ALLOC_1_ZERO(nvpoloGCr_neo_diag,0,geom%Nr)
    call FTR_ALLOC_1_ZERO(nvpoloGCr_vpar_diag,0,geom%Nr)
    call FTR_ALLOC_1_ZERO(niGCr_diag,0,geom%Nr)
    call FTR_ALLOC_1_ZERO(pressGCr_diag,0,geom%Nr)
    call FTR_ALLOC_1_ZERO(pressGCr_perp_diag,0,geom%Nr)
    call FTR_ALLOC_1_ZERO(pressGCr_par_diag,0,geom%Nr)
    call FTR_ALLOC_1_ZERO(LphiGCr_diag,0,geom%Nr)
    PRINT_CKSUM(LphiGCr_diag)
    call FTR_ALLOC_1_ZERO(TpolGCr_diag,0,geom%Nr)
    call FTR_ALLOC_1_ZERO(GammaGCr_turb_diag,0,geom%Nr)
    call FTR_ALLOC_1_ZERO(GammaGCr_neo_diag,0,geom%Nr)
    call FTR_ALLOC_1_ZERO(RSphiGCr_turb_diag,0,geom%Nr)
    call FTR_ALLOC_1_ZERO(RSphiGCr_neo_diag,0,geom%Nr)
    call FTR_ALLOC_1_ZERO(dpressGCr_dt_diag,0,geom%Nr)
    call FTR_ALLOC_1_ZERO(dLphiGCr_dt_diag,0,geom%Nr)
    call FTR_ALLOC_1_ZERO(QGCr_turb_diag,0,geom%Nr)
    call FTR_ALLOC_1_ZERO(QGCr_neo_diag,0,geom%Nr)
    call FTR_ALLOC_1_ZERO(QGCr_perp_turb_diag,0,geom%Nr)
    call FTR_ALLOC_1_ZERO(QGCr_perp_neo_diag,0,geom%Nr)
    call FTR_ALLOC_1_ZERO(QGCr_par_turb_diag,0,geom%Nr)
    call FTR_ALLOC_1_ZERO(QGCr_par_neo_diag,0,geom%Nr)
    call FTR_ALLOC_1_ZERO(QGCr_dtvpar_diag,0,geom%Nr)
    call FTR_ALLOC_1_ZERO(dpGCr_dt_par_diag,0,geom%Nr)
    call FTR_ALLOC_1_ZERO(dpGCr_dt_perp_diag,0,geom%Nr)
    call FTR_ALLOC_1_ZERO(SGCr_Pperp_par_diag,0,geom%Nr)
    call FTR_ALLOC_1_ZERO(SGCr_Pperp_turb_diag,0,geom%Nr)
    call FTR_ALLOC_1_ZERO(heatexGCr_turb_diag,0,geom%Nr)
    call FTR_ALLOC_1_ZERO(heatexGCr_neo_diag,0,geom%Nr)
    call FTR_ALLOC_2_ZERO(QGC_loc_diag,0,geom%Nr,0,geom%Ntheta)
    call FTR_ALLOC_2_ZERO(RSphiGC_loc_diag,0,geom%Nr,0,geom%Ntheta)
  end subroutine Physics_GC_moments_init


  !--------------------------------------------------------------
  !> deallocation of arrays for output saving of 
  !> guiding-center radial profiles
  !--------------------------------------------------------------
  subroutine Physics_GC_moments_cleanup ()    
    use wrap_alloc_module

    !*** allocation for fluid moments computation ***
#ifdef NO_MPI_INPLACE
    call FTR_DEALLOC(GC_Fmoments_loc)
#endif
    call FTR_DEALLOC(GC_Fmoments)

    !*** deallocation of arrays for output saving of ***
    !***  guiding-center radial profiles             ***
    call FTR_DEALLOC(nvpoloGCr_turb_diag)
    call FTR_DEALLOC(nvpoloGCr_neo_diag)
    call FTR_DEALLOC(nvpoloGCr_vpar_diag)
    call FTR_DEALLOC(niGCr_diag)
    call FTR_DEALLOC(pressGCr_diag)
    call FTR_DEALLOC(pressGCr_perp_diag)
    call FTR_DEALLOC(pressGCr_par_diag)
    PRINT_CKSUM(LphiGCr_diag)
    call FTR_DEALLOC(LphiGCr_diag)
    call FTR_DEALLOC(TpolGCr_diag)
    call FTR_DEALLOC(GammaGCr_turb_diag)
    call FTR_DEALLOC(GammaGCr_neo_diag)
    call FTR_DEALLOC(RSphiGCr_turb_diag)
    call FTR_DEALLOC(RSphiGCr_neo_diag)
    call FTR_DEALLOC(dpressGCr_dt_diag)
    call FTR_DEALLOC(dLphiGCr_dt_diag)
    call FTR_DEALLOC(QGCr_turb_diag)
    call FTR_DEALLOC(QGCr_neo_diag)
    call FTR_DEALLOC(QGCr_perp_turb_diag)
    call FTR_DEALLOC(QGCr_perp_neo_diag)
    call FTR_DEALLOC(QGCr_par_turb_diag)
    call FTR_DEALLOC(QGCr_par_neo_diag)
    call FTR_DEALLOC(QGCr_dtvpar_diag)
    call FTR_DEALLOC(dpGCr_dt_par_diag)
    call FTR_DEALLOC(dpGCr_dt_perp_diag)
    call FTR_DEALLOC(SGCr_Pperp_par_diag)
    call FTR_DEALLOC(SGCr_Pperp_turb_diag)
    call FTR_DEALLOC(heatexGCr_turb_diag)
    call FTR_DEALLOC(heatexGCr_neo_diag)
    call FTR_DEALLOC(QGC_loc_diag)
    call FTR_DEALLOC(RSphiGC_loc_diag)
  end subroutine Physics_GC_moments_cleanup


  !--------------------------------------------------------------
  !> Parallel preprocessing for compute_GC_moments()
  !--------------------------------------------------------------
  subroutine Physics_GC_moments_preprocess ( geom, coord_sys, &
    init_magnet, init_curr, fn, fnp1, Sfmu_eq, &
    SdJ0Phidr, SdJ0Phidtheta, SdJ0Phidphi, Enkin, &
    entropy, L2norm, QGC_loc, RSphiGC_loc)
    use coord_system_class , only: jacobian_space, R
    use globals            , only: iend, istart, jend, jstart, &
      hat_As, hat_Zs, K_Bdrift, mu_id, nbmoments_GC, Nbthread, &
      pSpecies_id
    use globals            , only: Bstar_PNrPNthNvpar
    use init_magnetic_class, only: compute_jacobian_velocity
#ifdef _OPENMP
    use OMPutils_module    , only: omp_get_thread_num
#endif
    use OMPutils_module    , only: Romp1_0Nr, Romp2_0Nr, &
      Romp3_0Nr, Romp4_0Nr, Romp5_0Nr, Romp6_0Nr, Romp7_0Nr, &
      Romp8_0Nr, Romp1_1nbmom_0Nr, Romp1_0Nr_0Ntheta, &
      Romp2_0Nr_0Ntheta
    use wrap_alloc_module
    use globals        , only : Nbthread, Nr, Ntheta

    type(geometry)               , intent(in) :: geom
    type(coord_system)           , intent(in) :: coord_sys
    type(init_magnetic)          , intent(in) :: init_magnet
    type(init_current)           , intent(in) :: init_curr
    type(fdistribu5d)            , intent(in) :: fn
    type(fdistribu5d)            , intent(in) :: fnp1
    real(RKIND), dimension(:,:,:), pointer    :: Sfmu_eq
    real(RKIND), dimension(:,:,:), pointer    :: SdJ0Phidr 
    real(RKIND), dimension(:,:,:), pointer    :: SdJ0Phidtheta 
    real(RKIND), dimension(:,:,:), pointer    :: SdJ0Phidphi
    real(RKIND), intent(out) :: entropy
    real(RKIND), intent(out) :: L2norm
    real(RKIND), intent(out) :: Enkin
    real(RKIND), dimension(:,:), pointer :: QGC_loc
    real(RKIND), dimension(:,:), pointer :: RSphiGC_loc

    integer     :: imom, tid, ithread
    integer     :: ir, itheta, iphi, ivpar, imix
    integer     :: nbmoments
    real(RKIND) :: sqrt_g22, vpar, mu_value
    real(RKIND) :: Bij, energy, fnval, fnp1val
    real(RKIND) :: inv_sqrtAs, hatZs
    !-> variables for the poloidal velocities
    real(RKIND) :: bstar_gradtheta_tmp
    real(RKIND) :: vpar_etheta, vExB_etheta, vD_etheta
    real(RKIND), dimension(:), pointer :: vExB_gradtheta_1D
    real(RKIND), dimension(:), pointer :: vD_gradtheta_1D
    !-> variables for entropy and L2norm
    real(RKIND), dimension(1:Nbthread) :: entropy_thread 
    real(RKIND), dimension(1:Nbthread) :: L2norm_thread
    !-> variables for the kinetic energy
    real(RKIND) :: fmfeq_E
    real(RKIND), dimension(1:Nbthread) :: Enkin_thread 
    !-> variables for toroidal angular momentum
    real(RKIND) :: Rij, inv_Bij, B_gphi
    real(RKIND) :: vExB_gradr, vD_gradr
    real(RKIND), dimension(:), pointer :: vExB_gradr_1D
    real(RKIND), dimension(:), pointer :: vD_gradr_1D
    !-> variables for parallel and perp pressure
    real(RKIND) :: gradpar_B_tmp, gradpar_J0Phi_tmp
    real(RKIND) :: dBdr_tmp, dBdtheta_tmp, dvpar
    real(RKIND) :: vExB_gradB_tmp, vpar_gradparB_tmp
    real(RKIND), dimension(:), pointer :: gradpar_B_1D
    real(RKIND), dimension(:), pointer :: gradpar_J0Phi_1D
    !-> variables for turbulent heating transfer
    real(RKIND) :: vExB_gradtheta, vD_gradtheta
    real(RKIND) :: vExB_gradphi, vD_gradphi
    real(RKIND) :: dJ0Phidr_tmp, dJ0Phidtheta_tmp, dJ0Phidphi_tmp
    real(RKIND) :: gradJ0Phi_r, gradJ0Phi_theta, gradJ0Phi_phi
    real(RKIND) :: scalprod_vExB_gradPhi, scalprod_vD_gradPhi
    real(RKIND), dimension(:), pointer :: vExB_gradphi_1D
    real(RKIND), dimension(:), pointer :: vD_gradphi_1D 

    !-> variables for the flux surface average
    real(RKIND) :: jacob_space_tmp, jacob_vel_tmp
    real(RKIND) :: coeff_intdr_tmp
    real(RKIND) :: coeff_intdtheta_tmp, coeff_intdphi_tmp
    real(RKIND) :: coeff_intdvpar_tmp, coeff_intdmu_tmp
    real(RKIND) :: coeff_int_volume, coeff_int_velocities

    call Write_dyn_balise_open('CPU_diag_moments1') !R4
    
    hatZs      = hat_Zs(pSpecies_id)
    inv_sqrtAs = 1._RKIND/sqrt(hat_As(pSpecies_id))

    !-> BECAREFUL: if you want to add the computation of
    !    a guiding-center moment, do not forget to change
    !    nbmoments_GC in 'globals.F90' file
    nbmoments        = nbmoments_GC
    mu_value         = fnp1%mu_value
    coeff_intdmu_tmp = geom%coeff_intdmu(mu_id)

    !*** initialisation to 0 of the radial arrays ***
    do imom = 1,nbmoments
      do ir =  0,geom%Nr
#ifdef NO_MPI_INPLACE
        GC_Fmoments_loc(ir,imom) = 0._RKIND
#endif
        GC_Fmoments(ir,imom)  = 0._RKIND
      end do
    end do
    ! These arrays must be initialized to 0
    call FTR_ALLOC_2_ZERO(RSphiGC_loc, 0, geom%Nr, 0, geom%Ntheta)
    call FTR_ALLOC_2_ZERO(QGC_loc, 0, geom%Nr, 0, geom%Ntheta)

   do tid = 1, Nbthread
     call FTR_ALLOC_2(Romp1_0Nr_0Ntheta(tid)%val,0,Nr,0,Ntheta)
     call FTR_ALLOC_2(Romp2_0Nr_0Ntheta(tid)%val,0,Nr,0,Ntheta)
   enddo

    PRINT_CKSUM(istart)
    PRINT_CKSUM(iend)
    PRINT_CKSUM(jstart)
    PRINT_CKSUM(jend)
    PRINT_CKSUM(geom%Nvpar)
    PRINT_CKSUM(geom%Nphi)
    PRINT_CKSUM(coeff_intdmu_tmp)
    PRINT_CKSUM(fnp1%values)
    PRINT_CKSUM(geom%vparg)
    PRINT_CKSUM(R)
    PRINT_CKSUM(init_magnet%B_gradphi)
    PRINT_CKSUM(init_magnet%B_norm)
    PRINT_CKSUM(jacobian_space)
    PRINT_CKSUM(geom%coeff_intdtheta)
    PRINT_CKSUM(geom%coeff_intdphi)
    PRINT_CKSUM(geom%coeff_intdvpar)
    PRINT_CKSUM(Bstar_PNrPNthNvpar)
    PRINT_CKSUM(Nbthread)
    do ithread = 1, Nbthread
      PRINT_CKSUM(Romp1_1nbmom_0Nr(ithread)%val(8,istart:iend))
    enddo
#ifdef _OPENMP
!$OMP PARALLEL default(none) &
!$OMP private(tid,ir,itheta,iphi,ivpar,imom, &
!$OMP  Bij,sqrt_g22,vpar,energy,fnval,fnp1val,fmfeq_E, &
!$OMP  dBdr_tmp, dBdtheta_tmp, gradpar_B_tmp, gradpar_J0Phi_tmp, &
!$OMP  coeff_intdr_tmp,coeff_intdtheta_tmp,&
!$OMP  coeff_intdphi_tmp,coeff_intdvpar_tmp,coeff_int_velocities, &
!$OMP  coeff_int_volume,jacob_space_tmp,jacob_vel_tmp, &
!$OMP  bstar_gradtheta_tmp,vExB_gradtheta_1D,vD_gradtheta_1D, &
!$OMP  gradpar_J0Phi_1D,gradpar_B_1D,vExB_gradr_1D,vD_gradr_1D, &
!$OMP  vExB_gradphi_1D, vD_gradphi_1D, &
!$OMP  inv_Bij, Rij, B_gphi, vExB_gradr, vD_gradr, &
!$OMP  vExB_gradB_tmp, vpar_gradparB_tmp, &
!$OMP  vpar_etheta,vD_etheta,vExB_etheta,dvpar, &
!$OMP  vExB_gradphi, vD_gradphi, dJ0Phidr_tmp, &
!$OMP  dJ0Phidtheta_tmp,dJ0Phidphi_tmp, &
!$OMP  vExB_gradtheta,vD_gradtheta, &
!$OMP  gradJ0Phi_r,gradJ0Phi_theta,gradJ0Phi_phi, &
!$OMP  scalprod_vExB_gradPhi,scalprod_vD_gradPhi) &
!$OMP shared(geom,coord_sys,init_magnet,init_curr, &
!$OMP  fnp1,fn,R,K_Bdrift,jacobian_space, &
!$OMP  inv_sqrtAs,hatZs,mu_value, &
!$OMP  Romp1_0Nr,Romp2_0Nr,Romp3_0Nr,Romp4_0Nr,Romp5_0Nr, &
!$OMP  Romp6_0Nr,Romp7_0Nr,Romp8_0Nr,Romp1_1nbmom_0Nr, &
!$OMP  Romp1_0Nr_0Ntheta, Romp2_0Nr_0Ntheta, &
!$OMP  nbmoments,istart,iend,jstart,jend, &
!$OMP  entropy_thread,L2norm_thread,Enkin_thread, &
!$OMP  SdJ0Phidr,SdJ0Phidtheta,SdJ0Phidphi,Sfmu_eq, &
!$OMP  coeff_intdmu_tmp,imix)
!$OMP BARRIER
    tid = 1+omp_get_thread_num()
#else
    tid = 1
#endif

    vExB_gradtheta_1D => Romp1_0Nr(tid)%val
    vD_gradtheta_1D   => Romp2_0Nr(tid)%val
    vExB_gradr_1D     => Romp3_0Nr(tid)%val
    vD_gradr_1D       => Romp4_0Nr(tid)%val
    gradpar_B_1D      => Romp5_0Nr(tid)%val
    gradpar_J0Phi_1D  => Romp6_0Nr(tid)%val
    vExB_gradphi_1D   => Romp7_0Nr(tid)%val
    vD_gradphi_1D     => Romp8_0Nr(tid)%val

    !*** initialisation of the values to 0 ***
    do ir = 0,geom%Nr
      do imom = 1,nbmoments
        Romp1_1nbmom_0Nr(tid)%val(imom,ir) = 0._RKIND
      end do
    end do
    do itheta = jstart,jend
      do ir = istart,iend
        Romp1_0Nr_0Ntheta(tid)%val(ir,itheta) = 0._RKIND
        Romp2_0Nr_0Ntheta(tid)%val(ir,itheta) = 0._RKIND
      end do
    end do
    entropy_thread(tid) = 0._RKIND
    L2norm_thread(tid)  = 0._RKIND
    Enkin_thread(tid)   = 0._RKIND

!$OMP DO SCHEDULE(STATIC)
    do imix = 0, (geom%Nvpar+1)*(geom%Nphi)-1
      ! Variable imix encapsulates ivpar and iphi together 
      ! in order to parallelize on these two variables 
      ! simultaneously (kind of nested loops parallelization).
      ! Useful on BG/Q machine with many threads (turing).
      ivpar = mod(imix,geom%Nvpar+1)
      iphi  = (imix/(geom%Nvpar+1))
      vpar               = geom%vparg(ivpar)
      coeff_intdvpar_tmp = geom%coeff_intdvpar(ivpar)
      coeff_intdphi_tmp  = geom%coeff_intdphi(iphi)
      do itheta = jstart,jend 
        coeff_intdtheta_tmp = geom%coeff_intdtheta(itheta)
        !*** computation of the poloidal velocities    ***
        !-> computation of vExB_gradtheta(0:Nr)
        call compute_vExBgradtheta(geom,init_magnet,init_curr, &
          SdJ0Phidr,SdJ0Phidphi,istart,iend,itheta,itheta, &
          iphi,iphi,ivpar,ivpar,.false.,vExB_gradtheta_1D)
        !-> computation of vD_gradtheta(0:Nr)
        call compute_vDgradtheta(geom,init_magnet,init_curr, &
          istart,iend,itheta,itheta,iphi,iphi, &
          ivpar,ivpar,.false.,vD_gradtheta_1D)
        !-> computation of vExB_gradr(0:Nr)
        call compute_vExBgradr(geom,init_magnet,init_curr, &
          SdJ0Phidtheta,SdJ0Phidphi,istart,iend,itheta,itheta, &
          iphi,iphi,ivpar,ivpar,.false.,vExB_gradr_1D)
        !-> computation of vD_gradr(0:Nr)
        call compute_vDgradr(geom,init_magnet,init_curr, &
          istart,iend,itheta,itheta,iphi,iphi, &
          ivpar,ivpar,.false.,vD_gradr_1D)
        !-> computation of gradpar.B(0:Nr)
        call compute_gradpar_B(geom,init_magnet,init_curr, &
          istart,iend,itheta,itheta,iphi,iphi, &
          ivpar,ivpar,.false.,gradpar_B_1D)
        !-> computation of gradpar.Phi(0:Nr)
        call compute_gradpar_J0Phi(geom,init_magnet,init_curr, &
          istart,iend,itheta,itheta,iphi,iphi,ivpar,ivpar, &
          .false.,gradpar_J0Phi_1D)
        !-> computation of vExB_gradphi(0:Nr)
        call compute_vExBgradphi(geom,init_magnet,init_curr, &
          SdJ0Phidr,SdJ0Phidtheta,istart,iend,itheta,itheta, &
          iphi,iphi,ivpar,ivpar,.false.,vExB_gradphi_1D)
        !-> computation of vD_gradphi(0:Nr)
        call compute_vDgradphi(geom,init_magnet,init_curr, &
          istart,iend,itheta,itheta,ivpar,ivpar, &
          .false.,vD_gradphi_1D)
        do ir = istart,iend
          Bij          = init_magnet%B_norm(ir,itheta)
          dBdr_tmp     = init_magnet%dBdr(ir,itheta)
          dBdtheta_tmp = init_magnet%dBdtheta(ir,itheta)
          energy       = 0.5_RKIND*vpar*vpar+mu_value*Bij
          fnval        = fn%values(ir,itheta,iphi,ivpar)
          fnp1val      = fnp1%values(ir,itheta,iphi,ivpar)
          fmfeq_E      = (fnp1val-Sfmu_eq(ir,itheta,ivpar))*energy
          !-> jacobian_velocity = 2*pi*Bstar
          if (geom%Nmu .eq. 0) then
#ifdef NO_PRECOMP_BSTAR 
            call compute_Bstar(geom,mthis,init_curr, &
              ir,itheta,ivpar,jacob_vel_tmp)
#else
            call precomp_Bstar(ir,itheta,ivpar,jacob_vel_tmp)
#endif
          else
            call compute_jacobian_velocity(geom,init_magnet, &
              init_curr,ir,itheta,ivpar,jacob_vel_tmp)            
          end if
          !-> jacobian in space
          jacob_space_tmp = jacobian_space(ir,itheta)
          !-> compute the elements of volume in space and velocity
          coeff_int_volume = jacob_space_tmp * &
            coeff_intdtheta_tmp*coeff_intdphi_tmp * &
            jacob_vel_tmp*coeff_intdvpar_tmp*coeff_intdmu_tmp 
          coeff_int_velocities = jacob_vel_tmp * &
            coeff_intdvpar_tmp*coeff_intdmu_tmp
          !-> compute bstar.gradtheta
#ifdef NO_PRECOMP_BSTAR
          call compute_bstar_contravariant(geom, &
            init_magnet,init_curr,ir,itheta,ivpar, &
            Sbstar_gradx2=bstar_gradtheta_tmp)
#else
          call precomp_bstar_gradtheta(ir,itheta,ivpar, &
            bstar_gradtheta_tmp)
#endif
          !*** computation of the projection of the      ***
          !***  poloidal velocities on the unit poloidal ***
          !***  vector                                   ***
          sqrt_g22    = sqrt(coord_sys%g22(ir,itheta))
          !-> vpar.etheta = vpar*r*bstar_gradtheta 
          vpar_etheta = vpar*sqrt_g22*bstar_gradtheta_tmp
          !-> vD.etheta = r*vD_gradtheta 
          vD_etheta   = sqrt_g22*vD_gradtheta_1D(ir-istart)
          !-> vE.etheta = r*vExB_gradtheta
          vExB_etheta = sqrt_g22*vExB_gradtheta_1D(ir-istart)
          !*** Computations for toroidal angular momentum ***
          inv_Bij     = 1._RKIND/Bij
          B_gphi      = init_magnet%B_gradphi(ir,itheta)
          Rij         = R(ir,itheta)
          !-> vD.gradr
          vD_gradr    = vD_gradr_1D(ir-istart)
          !-> vExB.gradr
          vExB_gradr  = vExB_gradr_1D(ir-istart)
          !-> computations for dvpar
          gradpar_B_tmp     = gradpar_B_1D(ir-istart)
          gradpar_J0Phi_tmp = gradpar_J0Phi_1D(ir-istart)
          !-> computation of vExB.gradB
          vExB_gradB_tmp    = vExB_gradr_1D(ir-istart)* &
            dBdr_tmp + vExB_gradtheta_1D(ir-istart)*dBdtheta_tmp
          !-> computation of 1/B*vpar.gradB
          vpar_gradparB_tmp = vpar*gradpar_B_tmp*inv_Bij
          !-> computation of vpar displacement
          dvpar = - inv_sqrtAs*mu_value*gradpar_B_tmp - &
            (hatZs*inv_sqrtAs)*gradpar_J0Phi_tmp + &
            K_Bdrift*vpar*inv_Bij*vExB_gradB_tmp
          !-> computation of scalprod(vGC,gradPhi)
          !-> where vGC = vExB + vD
          !-> and the scalar product is defined as 
          !->   scalprod(vec_A,vec_B) = sum_{ij} g_{ij} A^i B^j
          !->  where A^i = A.grad_xi is the contravariant 
          !->  coordinate of A
          vExB_gradtheta       = vExB_gradtheta_1D(ir-istart)
          vExB_gradphi         = vExB_gradphi_1D(ir-istart)
          vD_gradtheta         = vD_gradtheta_1D(ir-istart)
          vD_gradphi           = vD_gradphi_1D(ir-istart)
          dJ0Phidr_tmp         = SdJ0Phidr(ir,itheta,iphi) 
          dJ0Phidtheta_tmp     = SdJ0Phidtheta(ir,itheta,iphi) 
          dJ0Phidphi_tmp       = SdJ0Phidphi(ir,itheta,iphi) 
          gradJ0Phi_r          = &
            (coord_sys%g11(ir,itheta)*dJ0Phidr_tmp + &
            coord_sys%g12(ir,itheta)*dJ0Phidtheta_tmp + &
            coord_sys%g13(ir,itheta)*dJ0Phidphi_tmp)
          gradJ0Phi_theta      = &
            (coord_sys%g21(ir,itheta)*dJ0Phidr_tmp + &
            coord_sys%g22(ir,itheta)*dJ0Phidtheta_tmp + &
            coord_sys%g23(ir,itheta)*dJ0Phidphi_tmp)
          gradJ0Phi_phi        = &
            (coord_sys%g31(ir,itheta)*dJ0Phidr_tmp + &
            coord_sys%g32(ir,itheta)*dJ0Phidtheta_tmp + &
            coord_sys%g33(ir,itheta)*dJ0Phidphi_tmp)
          scalprod_vExB_gradPhi = vExB_gradr*gradJ0Phi_r + &
            vExB_gradtheta*gradJ0Phi_theta + &
            vExB_gradphi*gradJ0Phi_phi
          scalprod_vD_gradPhi = vD_gradr*gradJ0Phi_r + &
            vD_gradtheta*gradJ0Phi_theta + &
            vD_gradphi*gradJ0Phi_phi
          !*** computation of the integrals in velocity and ***
          !***  in theta and phi                            ***
          !-> 1) compute nvpoloGCr_turb
          Romp1_1nbmom_0Nr(tid)%val(1,ir) = &
            Romp1_1nbmom_0Nr(tid)%val(1,ir) + &
            fnp1val*vExB_etheta*coeff_int_volume
          !-> 2) compute nvpoloGCr_neo
          Romp1_1nbmom_0Nr(tid)%val(2,ir) = &
            Romp1_1nbmom_0Nr(tid)%val(2,ir) + &
            fnp1val*vD_etheta*coeff_int_volume
          !-> 3) compute nvpoloGCr_vpar
          Romp1_1nbmom_0Nr(tid)%val(3,ir) = &
            Romp1_1nbmom_0Nr(tid)%val(3,ir) + &
            fnp1val*vpar_etheta*coeff_int_volume
          !-> 4) compute niGCr
          Romp1_1nbmom_0Nr(tid)%val(4,ir) = &
            Romp1_1nbmom_0Nr(tid)%val(4,ir) + &
            fnp1val*coeff_int_volume
          !-> 5) compute pressGCr_tnp1
          Romp1_1nbmom_0Nr(tid)%val(5,ir) = &
            Romp1_1nbmom_0Nr(tid)%val(5,ir) + &
            fnp1val*energy*coeff_int_volume
          !-> 6) compute pressGCr_perp
          Romp1_1nbmom_0Nr(tid)%val(6,ir) = &
            Romp1_1nbmom_0Nr(tid)%val(6,ir) + &
            fnp1val*mu_value*Bij*coeff_int_volume
          !-> 7) compute pressGCr_par
          Romp1_1nbmom_0Nr(tid)%val(7,ir) = &
            Romp1_1nbmom_0Nr(tid)%val(7,ir) + &
            fnp1val*vpar*vpar*coeff_int_volume
          !-> 8) compute LphiGCr
          Romp1_1nbmom_0Nr(tid)%val(8,ir) = &
            Romp1_1nbmom_0Nr(tid)%val(8,ir) + &
            fnp1val*vpar*Rij*Rij*B_gphi*inv_Bij * &
            coeff_int_volume
          !-> 9) compute TpolGCr
          Romp1_1nbmom_0Nr(tid)%val(9,ir) = &
            Romp1_1nbmom_0Nr(tid)%val(9,ir) + &
            fnp1val* SdJ0Phidphi(ir,itheta,iphi) * &
            coeff_int_volume
          !-> 10) compute GammaGCr_turb
          Romp1_1nbmom_0Nr(tid)%val(10,ir) = &
            Romp1_1nbmom_0Nr(tid)%val(10,ir) + &
            fnp1val*vExB_gradr*coeff_int_volume
          !-> 11) compute GammaGCr_neo
          Romp1_1nbmom_0Nr(tid)%val(11,ir) = &
            Romp1_1nbmom_0Nr(tid)%val(11,ir) + &
            fnp1val*vD_gradr*coeff_int_volume
          !-> 12) compute RSphiGCr_turb
          Romp1_1nbmom_0Nr(tid)%val(12,ir) = &
            Romp1_1nbmom_0Nr(tid)%val(12,ir) + &
            fnp1val*vpar*Rij*Rij*B_gphi*inv_Bij * &
            vExB_gradr*coeff_int_volume
          !-> 13) compute RSphiGCr_neo
          Romp1_1nbmom_0Nr(tid)%val(13,ir) = &
            Romp1_1nbmom_0Nr(tid)%val(13,ir) + &
            fnp1val*vpar*Rij*Rij*B_gphi*inv_Bij * &
            vD_gradr*coeff_int_volume
          !-> 14) compute pressGCr_tn (for dpressGCr_dt)
          Romp1_1nbmom_0Nr(tid)%val(14,ir) = &
            Romp1_1nbmom_0Nr(tid)%val(14,ir) + &
            fnval*energy*coeff_int_volume
          !-> 15) compute LphiGCr_tn (for dLphiGCr_dt)
          Romp1_1nbmom_0Nr(tid)%val(15,ir) = &
            Romp1_1nbmom_0Nr(tid)%val(15,ir) + &
            fnval*vpar*Rij*Rij*B_gphi*inv_Bij * &
            coeff_int_volume
          !-> 16) compute QGCr_turb
          Romp1_1nbmom_0Nr(tid)%val(16,ir) = &
            Romp1_1nbmom_0Nr(tid)%val(16,ir) + &
            fnp1val*vExB_gradr*energy* &
            coeff_int_volume
          !-> 17) compute QGCr_neo
          Romp1_1nbmom_0Nr(tid)%val(17,ir) = &
            Romp1_1nbmom_0Nr(tid)%val(17,ir) + &
            fnp1val*vD_gradr*energy* &
            coeff_int_volume
          !-> 18) compute QGCr_perp_turb
          Romp1_1nbmom_0Nr(tid)%val(18,ir) = &
            Romp1_1nbmom_0Nr(tid)%val(18,ir) + &
            fnp1val*vExB_gradr*mu_value*Bij* &
            coeff_int_volume
          !-> 19) compute QGCr_perp_neo
          Romp1_1nbmom_0Nr(tid)%val(19,ir) = &
            Romp1_1nbmom_0Nr(tid)%val(19,ir) + &
            fnp1val*vD_gradr*mu_value*Bij* &
            coeff_int_volume
          !-> 20) compute QGCr_par_turb
          Romp1_1nbmom_0Nr(tid)%val(20,ir) = &
            Romp1_1nbmom_0Nr(tid)%val(20,ir) + &
            fnp1val*vExB_gradr*vpar*vpar* &
            0.5_RKIND*coeff_int_volume
          !-> 21) compute QGCr_par_neo
          Romp1_1nbmom_0Nr(tid)%val(21,ir) = &
            Romp1_1nbmom_0Nr(tid)%val(21,ir) + &
            fnp1val*vD_gradr*vpar*vpar* &
            0.5_RKIND*coeff_int_volume
          !-> 22) compute Par_press_dtvpar
          Romp1_1nbmom_0Nr(tid)%val(22,ir) = &
            Romp1_1nbmom_0Nr(tid)%val(22,ir) + &
            fnp1val*vpar*dvpar* &
            coeff_int_volume
          !-> 23) compute press_perp_tn (for dpGCr_dt_perp)
          Romp1_1nbmom_0Nr(tid)%val(23,ir) = &
            Romp1_1nbmom_0Nr(tid)%val(23,ir) + &
            fnval*mu_value*Bij*coeff_int_volume
          !-> 24) compute press_par_tn (for dpGCr_dt_par)
          Romp1_1nbmom_0Nr(tid)%val(24,ir) = &
            Romp1_1nbmom_0Nr(tid)%val(24,ir) + &
            fnval*vpar*vpar*coeff_int_volume
          !-> 25) compute SGCr_Pperp_par
          Romp1_1nbmom_0Nr(tid)%val(25,ir) = &
            Romp1_1nbmom_0Nr(tid)%val(25,ir) + &
            fnp1val*mu_value*Bij*vpar_gradparB_tmp* &
            inv_Bij*coeff_int_volume
          !-> 26) compute SGCr_Pperp_turb
          Romp1_1nbmom_0Nr(tid)%val(26,ir) = &
            Romp1_1nbmom_0Nr(tid)%val(26,ir) + &
            fnp1val*mu_value*Bij* &
            vExB_gradB_tmp*inv_Bij* &
            coeff_int_volume
          !-> 27) compute heatexGCr_turb 
          Romp1_1nbmom_0Nr(tid)%val(27,ir) = &
            Romp1_1nbmom_0Nr(tid)%val(27,ir) + &
            fnp1val*scalprod_vExB_gradPhi*coeff_int_volume
          !-> 28) compute heatexGCr_neo 
          Romp1_1nbmom_0Nr(tid)%val(28,ir) = &
            Romp1_1nbmom_0Nr(tid)%val(28,ir) + &
            fnp1val*scalprod_vD_gradPhi*coeff_int_volume
          !-> 29) compute QGC_loc
          Romp1_0Nr_0Ntheta(tid)%val(ir,itheta) = &
            Romp1_0Nr_0Ntheta(tid)%val(ir,itheta) + &
            fnp1val*vExB_gradr*energy* &
            coeff_intdphi_tmp*coeff_int_velocities
          !-> 30) compute RSphiGC_loc
          Romp2_0Nr_0Ntheta(tid)%val(ir,itheta) = &
            Romp2_0Nr_0Ntheta(tid)%val(ir,itheta) + &
            fnp1val*vExB_gradr*vpar*Rij*Rij*B_gphi*inv_Bij* &
              coeff_intdphi_tmp*coeff_int_velocities
          !-> compute entropy
          coeff_intdr_tmp = geom%coeff_intdr(ir)
          if (fnp1val .ne. 0._RKIND) then
            entropy_thread(tid) = entropy_thread(tid) - &
              fnp1val*log(abs(fnp1val))*coeff_int_volume * &
              coeff_intdr_tmp
          end if
          !-> compute L2norm
          L2norm_thread(tid)  = L2norm_thread(tid)  + &
            fnp1val*fnp1val*coeff_int_volume * &
            coeff_intdr_tmp
          !->  compute Enkin
          Enkin_thread(tid)  = Enkin_thread(tid)  + &
            fmfeq_E*coeff_int_volume * &
            coeff_intdr_tmp
        end do ! end of do ir
      end do   ! end of do itheta
    end do     ! end of do imix (iphi and ivpar)
!$OMP END DO
!$OMP BARRIER
!$OMP END PARALLEL
    PRINT_CKSUM(Nbthread)
    do ithread = 1, Nbthread
      PRINT_CKSUM(Romp1_1nbmom_0Nr(ithread)%val(8,istart:iend))
    enddo
    !*** fill an unique array containing all the       ***
    !***  radial profiles of the fluid moments         ***
    !*** Rq: each processor fills the part istart:iend ***
    !***     and all the others values are put equal   ***
    !***     to 0                                      ***
    entropy = 0._RKIND
    L2norm  = 0._RKIND
    Enkin   = 0._RKIND
    do ithread = 1,Nbthread
      do ir = istart,iend
        do imom = 1,nbmoments
#ifdef NO_MPI_INPLACE
          GC_Fmoments_loc(ir,imom) = GC_Fmoments_loc(ir,imom) + &
            Romp1_1nbmom_0Nr(ithread)%val(imom,ir)
#else
          GC_Fmoments(ir,imom) = GC_Fmoments(ir,imom) + &
            Romp1_1nbmom_0Nr(ithread)%val(imom,ir)
#endif
        end do
      end do
      entropy = entropy + entropy_thread(ithread)
      L2norm  = L2norm + L2norm_thread(ithread) 
      Enkin   = Enkin + Enkin_thread(ithread) 
      do itheta = jstart,jend
        do ir = istart,iend
          QGC_loc(ir,itheta)     = QGC_loc(ir,itheta) + &
            Romp1_0Nr_0Ntheta(ithread)%val(ir,itheta)
          RSphiGC_loc(ir,itheta) = RSphiGC_loc(ir,itheta) + &
            Romp2_0Nr_0Ntheta(ithread)%val(ir,itheta)
        end do
      end do
    end do
#ifdef NO_MPI_INPLACE
    PRINT_CKSUM(GC_Fmoments_loc(istart:iend,1:nbmoments))
#else
    PRINT_CKSUM(GC_Fmoments(istart:iend,1:nbmoments))
#endif

   do tid = 1, Nbthread
     call FTR_DEALLOC(Romp1_0Nr_0Ntheta(tid)%val)
     call FTR_DEALLOC(Romp2_0Nr_0Ntheta(tid)%val)
   enddo

    call Write_dyn_balise_close('CPU_diag_moments1') !R4
  end subroutine Physics_GC_moments_preprocess


  !--------------------------------------------------------------
  !> Data transfer between parallel preprocessing and local
  !> postprocessing for compute_GC_moments()
  !>
  !> Beware, that except for geom, all other parameters are 
  !> only used on sending processes, so nothing usefull has 
  !> to be passed if the current process isn't sending
  !--------------------------------------------------------------
  subroutine Physics_GC_moments_transfer ( geom, Enkin, entropy, &
    L2norm, QGC_loc, RSphiGC_loc )
    
    use globals          , only: diag_targ, mpi_comm_species, &
      nbmoments_GC, Nbproc_species, pInSpecies_id
    use wrap_alloc_module

    type(geometry), intent(in) :: geom
    real(RKIND),    intent(in) :: Enkin
    real(RKIND),    intent(in) :: entropy
    real(RKIND),    intent(in) :: L2norm
    real(RKIND), dimension(:,:), pointer :: QGC_loc
    real(RKIND), dimension(:,:), pointer :: RSphiGC_loc

    integer :: ierr, imom, ir, itheta, nbelements, nbelements2
    
    PRINT_CKSUM(geom%Nr)
    PRINT_CKSUM(nbmoments_GC)
    PRINT_CKSUM(DIAG_RPROF_GC)
#ifdef NO_MPI_INPLACE
    PRINT_CKSUM(GC_Fmoments_loc)
#else
    PRINT_CKSUM(GC_Fmoments)
#endif
    
    if (Nbproc_species.ne.1) then
      nbelements  = nbmoments_GC*(geom%Nr+1)
      nbelements2 = (geom%Nr+1)*(geom%Ntheta+1)
      PRINT_CKSUM(nbelements)
#ifdef NO_MPI_INPLACE
      call MY_Reduce(GC_Fmoments_loc, GC_Fmoments, nbelements, &
        MPI_REAL8, MPI_SUM, diag_targ(DIAG_RPROF_GC), &
        mpi_comm_species, ierr)
#else
      if ( pInSpecies_id == diag_targ(DIAG_RPROF_GC) ) then
        call MY_Reduce(MPI_IN_PLACE, GC_Fmoments, nbelements, &
          MPI_REAL8, MPI_SUM, diag_targ(DIAG_RPROF_GC), &
          mpi_comm_species, ierr)
      else
        call MY_Reduce(GC_Fmoments, MPI_BOTTOM, nbelements, &
          MPI_REAL8, MPI_SUM, diag_targ(DIAG_RPROF_GC), &
          mpi_comm_species, ierr)
      endif
#endif
      call MY_Reduce(entropy, entropy_diag, 1, MPI_REAL8, MPI_SUM, &
        diag_targ(DIAG_CL), mpi_comm_species, ierr)
      call MY_Reduce(L2norm, L2norm_diag, 1, MPI_REAL8, MPI_SUM, &
        diag_targ(DIAG_CL), mpi_comm_species, ierr)
      call MY_Reduce(Enkin, Enkin_diag, 1, MPI_REAL8, MPI_SUM,&
        diag_targ(DIAG_CL), mpi_comm_species, ierr)
      call MY_Reduce(QGC_loc, QGC_loc_diag, nbelements2, &
        MPI_REAL8, MPI_SUM, diag_targ(DIAG_RPROF_GC), &
        mpi_comm_species, ierr)
      call MY_Reduce(RSphiGC_loc, RSphiGC_loc_diag, nbelements2, &
        MPI_REAL8, MPI_SUM, diag_targ(DIAG_RPROF_GC), &
        mpi_comm_species, ierr)
    else
#ifdef NO_MPI_INPLACE
      do imom = 1, nbmoments_GC
        do ir = 0,geom%Nr
          GC_Fmoments(ir,imom) = GC_Fmoments_loc(ir,imom)
        end do
      end do
#endif
      entropy_diag = entropy
      L2norm_diag  = L2norm
      Enkin_diag   = Enkin
      do itheta = 0,geom%Ntheta-1
        do ir = 0,geom%Nr
           QGC_loc_diag(ir,itheta)     = QGC_loc(ir,itheta)
           RSphiGC_loc_diag(ir,itheta) = RSphiGC_loc(ir,itheta)
        end do
      end do
      
      PRINT_CKSUM(GC_Fmoments(0:geom%Nr,1:nbmoments_GC))
    end if
    
    call FTR_DEALLOC(QGC_loc)
    call FTR_DEALLOC(RSphiGC_loc)    
  end subroutine Physics_GC_moments_transfer


  !-------------------------------------------------------
  !> Local postprocessing for compute_GC_moments()
  !-------------------------------------------------------
  subroutine Physics_GC_moments_postprocess ( geom, Sdt_diag, &
    Siter_time )
    use geometry_class    , only: ipow
    use globals           , only: reduced_deltat, reduced_begin, &
      reduced_end, hat_As, hat_Zs, pSpecies_id
    use coord_system_class, only: intdthetadphi_Js

    type(geometry), intent(in) :: geom
    real(RKIND),    intent(in) :: Sdt_diag
    real(RKIND),    intent(in) :: Siter_time

    real(RKIND) :: hatZs
    real(RKIND) :: coeff_pressr
    integer     :: nbadapt
    integer     :: ir, itheta
    real(RKIND) :: intdthetadphi_Js_tmp

    hatZs = hat_Zs(pSpecies_id)
    ! computation of the flux surface average for the computation of
    ! the pressure and its time derivative
    coeff_pressr = 1._RKIND/(0.5_RKIND+float(ipow))
    if ((reduced_deltat .ne. 0._RKIND) .and. &
      (Siter_time .ge. reduced_begin) .and. &
      (Siter_time .le. reduced_end)) then
      nbadapt    = MAX(1, nint(Sdt_diag / reduced_deltat))
    else
      nbadapt    = 1
    end if
    PRINT_CKSUM(geom%Nr)
    PRINT_CKSUM(intdthetadphi_Js(0:geom%Nr))
    PRINT_CKSUM(GC_Fmoments(0:geom%Nr,1:28))
    do ir = 0,geom%Nr
      intdthetadphi_Js_tmp    = intdthetadphi_Js(ir)
      nvpoloGCr_turb_diag(ir) = &
        GC_Fmoments(ir,1)/intdthetadphi_Js_tmp
      nvpoloGCr_neo_diag(ir)  = &
        GC_Fmoments(ir,2)/intdthetadphi_Js_tmp
      nvpoloGCr_vpar_diag(ir) = &
        GC_Fmoments(ir,3)/intdthetadphi_Js_tmp
      niGCr_diag(ir)          = &
        GC_Fmoments(ir,4)/intdthetadphi_Js_tmp
      pressGCr_diag(ir)       = coeff_pressr* &
        GC_Fmoments(ir,5)/intdthetadphi_Js_tmp
      pressGCr_perp_diag(ir)  = &
        GC_Fmoments(ir,6)/intdthetadphi_Js_tmp
      pressGCr_par_diag(ir)   = &
        GC_Fmoments(ir,7)/intdthetadphi_Js_tmp
      LphiGCr_diag(ir)        = &
        GC_Fmoments(ir,8)/intdthetadphi_Js_tmp
      TpolGCr_diag(ir)        = &
        GC_Fmoments(ir,9)/intdthetadphi_Js_tmp
      GammaGCr_turb_diag(ir)  = &
        GC_Fmoments(ir,10)/intdthetadphi_Js_tmp
      GammaGCr_neo_diag(ir)   = &
        GC_Fmoments(ir,11)/intdthetadphi_Js_tmp
      RSphiGCr_turb_diag(ir)  = &
        GC_Fmoments(ir,12)/intdthetadphi_Js_tmp
      RSphiGCr_neo_diag(ir)   = &
        GC_Fmoments(ir,13)/intdthetadphi_Js_tmp
      dpressGCr_dt_diag(ir)   = coeff_pressr* &
        (GC_Fmoments(ir,5)-GC_Fmoments(ir,14))/ &
        (intdthetadphi_Js_tmp*Sdt_diag) *nbadapt
      dLphiGCr_dt_diag(ir)    = &
        (GC_Fmoments(ir,8)-GC_Fmoments(ir,15))/ &
        (intdthetadphi_Js_tmp*Sdt_diag) *nbadapt
      QGCr_turb_diag(ir)      = &
        GC_Fmoments(ir,16)/intdthetadphi_Js_tmp
      QGCr_neo_diag(ir)       = &
        GC_Fmoments(ir,17)/intdthetadphi_Js_tmp
      QGCr_perp_turb_diag(ir) = &
        GC_Fmoments(ir,18)/intdthetadphi_Js_tmp
      QGCr_perp_neo_diag(ir)  = &
        GC_Fmoments(ir,19)/intdthetadphi_Js_tmp
      QGCr_par_turb_diag(ir)  = &
        GC_Fmoments(ir,20)/intdthetadphi_Js_tmp
      QGCr_par_neo_diag(ir)   = &
        GC_Fmoments(ir,21)/intdthetadphi_Js_tmp
      QGCr_dtvpar_diag(ir)    = &
        GC_Fmoments(ir,22)/intdthetadphi_Js_tmp
      dpGCr_dt_perp_diag(ir)    = &
        (GC_Fmoments(ir,6)-GC_Fmoments(ir,23))/ &
        (intdthetadphi_Js_tmp*Sdt_diag) *nbadapt
      dpGCr_dt_par_diag(ir)    = &
        (GC_Fmoments(ir,7)-GC_Fmoments(ir,24))/ &
        (intdthetadphi_Js_tmp*Sdt_diag) *nbadapt
      SGCr_Pperp_par_diag(ir)  = &
        GC_Fmoments(ir,25)/intdthetadphi_Js_tmp
      SGCr_Pperp_turb_diag(ir) = &
        GC_Fmoments(ir,26)/intdthetadphi_Js_tmp
      heatexGCr_turb_diag(ir)  = - hatZs* &
        GC_Fmoments(ir,27)/intdthetadphi_Js_tmp
      heatexGCr_neo_diag(ir)   = - hatZs* &
        GC_Fmoments(ir,28)/intdthetadphi_Js_tmp
    end do
    PRINT_CKSUM(LphiGCr_diag(0:geom%Nr))
    do itheta = 0,geom%Ntheta-1
      do ir = 0,geom%Nr
        QGC_loc_diag(ir,itheta)     = &
          QGC_loc_diag(ir,itheta)/geom%Lphi
        RSphiGC_loc_diag(ir,itheta) = &
          RSphiGC_loc_diag(ir,itheta)/geom%Lphi
      end do
    end do
    do ir = 0,geom%Nr
      QGC_loc_diag(ir,geom%Ntheta)     = QGC_loc_diag(ir,0)
      RSphiGC_loc_diag(ir,geom%Ntheta) = RSphiGC_loc_diag(ir,0)
    end do
  end subroutine Physics_GC_moments_postprocess
end module physics_GC_module


