!**************************************************************
!  Copyright Euratom-CEA
!  Authors : 
!     Virginie Grandgirard (virginie.grandgirard@cea.fr)
!     Chantal Passeron (chantal.passeron@cea.fr)
!     Guillaume Latu (guillaume.latu@cea.fr)
!     Xavier Garbet (xavier.garbet@cea.fr)
!     Philippe Ghendrih (philippe.ghendrih@cea.fr)
!     Yanick Sarazin (yanick.sarazin@cea.fr)
!  
!  This code GYSELA (for GYrokinetic SEmi-LAgrangian) 
!  is a 5D gyrokinetic global full-f code for simulating 
!  the plasma turbulence in a tokamak.
!  
!  This software is governed by the CeCILL-B license 
!  under French law and abiding by the rules of distribution 
!  of free software.  You can  use, modify and redistribute 
!  the software under the terms of the CeCILL-B license as 
!  circulated by CEA, CNRS and INRIA at the following URL
!  "http://www.cecill.info". 
!**************************************************************

#include "config-F.inc"
#include "code_instrumentation/debug.inc"

!----------------------------------------------------------------- 
! file : physics_part.F90
! date : 16/04/2010
!  Compute all the physics quantities 
!   dedicated to the quantities associated to the particles.
! - Particles quantities: 
!    -> nvpolor_mag : poloidal velocity due to the magnetisation
!    -> nir         : ion density
!    -> nuparr      : parallel velocity
!    -> pressr      : pressure
!    -> Qr_turb     : turbulent heat flux (due to ExB velocity)
!    -> Qr_neo      : neoclassical heat flux 
!                     (due to curvature drift)
!    -> Gammar_turb : turbulent flux (due to ExB velocity)
!    -> Gammar_neo  : neoclassical flux (due to curvature drift)
!    -> RSthetar    : poloidal Reynold stress
!    -> RSparr      : parallel Reynold stress
!    -> potential energy
!
! Rk: All the previous radial profiles correspond to
!     flux surface average, i.e:
!       <.>_FS = \int . Js dtheta dphi /
!                \int Js dtheta dphi 
!      with Js the jacobian in space
!----------------------------------------------------------------- 
module physics_part_module
  use clock_module, &
                  only: TIMEPREC
  use coord_system_class, &
                  only: coord_system
  use Debug_module
  use fdistribu5d_class, &
                  only: fdistribu5d
  use geometry_class, &
                  only: geometry
  use globals,    only: DIAG_CL, DIAG_RPROF_GC, DIAG_RPROF_PART
  use gyroaverage_class, &
                  only: J0operator
  use init_current_class, &
                  only: init_current
  use init_magnetic_class, &
                  only: init_magnetic
  use init_profile_class, &
                  only: init_profile
  use OMPutils_module, &
                  only: omp1dvector
  use poisson_class, &
                  only: poisson
  use prec_const
  
  use mpiMod

  implicit none

  !-------------------------------------
  !  PRIVATE VARIABLES
  !------------------------------------- 
  !-> variables for the particle moments
  real(RKIND), dimension(:,:), pointer, private :: part_Fmoments
#ifdef NO_MPI_INPLACE
  real(RKIND), dimension(:,:), pointer, private :: part_Fmoments_loc
#endif

  !-------------------------------------
  !  PUBLIC VARIABLES
  !------------------------------------- 
  public
  !--> for particle radial profiles
  real(RKIND), dimension(:), pointer, public :: nvpolor_mag_diag
  real(RKIND), dimension(:), pointer, public :: nir_diag
  real(RKIND), dimension(:), pointer, public :: nir_trapped_diag
  real(RKIND), dimension(:), pointer, public :: nuparr_diag
  real(RKIND), dimension(:), pointer, public :: pressr_diag
  real(RKIND), dimension(:), pointer, public :: Qr_turb_diag
  real(RKIND), dimension(:), pointer, public :: Qr_neo_diag
  real(RKIND), dimension(:), pointer, public :: Gammar_turb_diag
  real(RKIND), dimension(:), pointer, public :: Gammar_neo_diag
  real(RKIND), dimension(:), pointer, public :: RSthetar_diag
  real(RKIND), dimension(:), pointer, public :: RSparr_diag

  !--> for potential energy
  real(RKIND), public :: Enpot_diag

  include "gyrokinetic/Bstar_inline.h"
  include "gyrokinetic/velocities_inline.h"
  !******************************
  contains
  !******************************
#include "gyrokinetic/Bstar_inline.F90"
#include "gyrokinetic/velocities_inline.F90"


  !-----------------------------------------------------
  ! Constructor
  !-----------------------------------------------------
  subroutine new_physics_part(geom)

    type(geometry), intent(in) :: geom

    call Physics_part_moments_init(geom)
  end subroutine new_physics_part


  !-----------------------------------------------------
  ! Destructor
  !-----------------------------------------------------
  subroutine delete_physics_part()

    call Physics_part_moments_cleanup()
  end subroutine delete_physics_part


  !-------------------------------------------------------------
  !> allocation of arrays for output saving of particle radial
  !> profiles
  !-------------------------------------------------------------
  subroutine Physics_part_moments_init ( geom )
    use globals          , only: nbmoments_part
    use wrap_alloc_module

    type(geometry), intent(in) :: geom

    !*** allocation for fluid moments computation ***
#ifdef NO_MPI_INPLACE
    call FTR_ALLOC_2_ZERO(part_Fmoments_loc,0,geom%Nr,1,nbmoments_part)
#endif
    call FTR_ALLOC_2_ZERO(part_Fmoments,0,geom%Nr,1,nbmoments_part)

    call FTR_ALLOC_1_ZERO(nvpolor_mag_diag,0,geom%Nr)
    call FTR_ALLOC_1_ZERO(nir_diag,0,geom%Nr)
    call FTR_ALLOC_1_ZERO(nir_trapped_diag,0,geom%Nr)
    call FTR_ALLOC_1_ZERO(nuparr_diag,0,geom%Nr)
    call FTR_ALLOC_1_ZERO(pressr_diag,0,geom%Nr)
    call FTR_ALLOC_1_ZERO(Qr_turb_diag,0,geom%Nr)
    call FTR_ALLOC_1_ZERO(Qr_neo_diag,0,geom%Nr)
    call FTR_ALLOC_1_ZERO(Gammar_turb_diag,0,geom%Nr)
    call FTR_ALLOC_1_ZERO(Gammar_neo_diag,0,geom%Nr)
    call FTR_ALLOC_1_ZERO(RSthetar_diag,0,geom%Nr)
    call FTR_ALLOC_1_ZERO(RSparr_diag,0,geom%Nr)
  end subroutine Physics_part_moments_init


  !--------------------------------------------------------------
  !> deallocation of arrays for output saving of particle radial
  !> profiles
  !--------------------------------------------------------------
  subroutine Physics_part_moments_cleanup ()
    use wrap_alloc_module

    !*** allocation for fluid moments computation ***
#ifdef NO_MPI_INPLACE
    call FTR_DEALLOC(part_Fmoments_loc)
#endif
    call FTR_DEALLOC(part_Fmoments)

    !*** deallocation of arrays for output saving of ***
    !***  particle radial profiles                   ***
    call FTR_DEALLOC(nvpolor_mag_diag)
    call FTR_DEALLOC(nir_diag)
    call FTR_DEALLOC(nir_trapped_diag)
    call FTR_DEALLOC(nuparr_diag)
    call FTR_DEALLOC(pressr_diag)
    call FTR_DEALLOC(Qr_turb_diag)
    call FTR_DEALLOC(Qr_neo_diag)
    call FTR_DEALLOC(Gammar_turb_diag)
    call FTR_DEALLOC(Gammar_neo_diag)
    call FTR_DEALLOC(RSthetar_diag)
    call FTR_DEALLOC(RSparr_diag)
  end subroutine Physics_part_moments_cleanup


  !----------------------------------------------------------------
  ! Computation of the fluid moments associated to the 
  !  particles, corresponding to 
  !   1) nvpolor_mag(r) = -<rot(M).etheta>_FS
  !   2) Ni(r)          = <\int J0.f d3v>_FS
  !   3) Ni_trapped(r)  = <\int_{trapped} J0.f d3v>_FS
  !   4) nVpar(r)       = <\int J0.f*vpar d3v>_FS
  !   5) P(r)           = <\int J0.f*E d3v>_FS
  !   6) Q_turb(r)      = <\int J0.f*E*vE.er d3v>_FS
  !   7) Q_neo(r)       = <\int J0.f*E*vD.er d3v>_FS 
  !   8) Gamma_turb(r)  = <\int J0.f*vE.er d3v>_FS
  !   9) Gamma_neo(r)   = <\int J0.f*vD.er d3v>_FS
  !  10) RStheta(r)     = <\int J0.(f*vExB.gradr) d3v>_FS
  !  11) RSpar(r)       = <\int J0.f vtilde_r vpar d3v>_FS 
  !  - potential energy
  !
  ! Rk2: <.>_FS corresponds to the flux surface average,
  !      i.e <.>_FS = \int . Js dtheta dphi /
  !                   \int Js dtheta dphi 
  !      with Js the jacobian in space
  !
  ! Rk2: Compute specifically the magnetisation component in nvpolo
  !    -> nvpolor_mag = nvpolo - nvpolor_GuidingCentre
  !  . nvpolor_mag = -rot(M).etheta                   
  !               = -r*rot(M).gradtheta              
  !     with rot(M).gradtheta = -1/Js d/dr(M_phi)     
  !      where Js is the jacobian in space and        
  !      M_phi is the toroidal covariant component of 
  !       M = -\int (\int J0.f vec_b Jv dvpar)*mu dmu 
  !     i.e M_phi = - b_phi *                         
  !          \int (\int J0.f vec_b Jv dvpar)*mu dmu     
  !
  ! Rk3: The trapped constraint is
  !      vpar^2/mu < 2 ( B(r,theta=pi)-B(r,theta) )
  !----------------------------------------------------------------
  subroutine compute_part_moments(geom, coord_sys, init_magnet, &
    init_curr, J0, fnp1, sJ0ftmp, sJ0fmu_eq, poiss)

    use globals         , only: nbmoments_part, &
      Rarray1_PNrPNthetaNphi, Rarray2_PNrPNthetaNphi, &
      Rarray3_PNrPNthetaNphi, Rarray4_PNrPNthetaNphi, distrib3D
    use globals         , only: istart_buf, iend_buf, kstart, kend
    ! used if the 3D data are distributed (distrib = .true.)
    use globals          , only: Rarray5_NrNthetaPKNphi, &
      Rarray6_NrNthetaPKNphi, &
      Rarray7_NrNthetaPKNphi
    ! used if the 3D data are not distributed (distrib = .false.)
    use globals          , only: Rarray1_NrNthetaNphi, &
      Rarray2_NrNthetaNphi
#ifdef TIMER
    use globals          , only: pglobal_id, plocal_id
    use clock_module     , only: clck_time, clck_ldiff
    use MPIutils_module  , only: pp_reduce_vector
#endif
    use wrap_alloc_module
    
    use MPIutils_module  , only: ppbarrier_timer
    use efield_module    , only: comp_dphidx
    use utils_module     , only: deriv1
 
    use Debug_module, only: cksum
    type(geometry)               , intent(in)    :: geom
    type(coord_system)           , intent(in)    :: coord_sys
    type(init_magnetic)          , intent(in)    :: init_magnet
    type(init_current)           , intent(in)    :: init_curr
    type(J0operator)             , intent(inout) :: J0
    type(fdistribu5d)            , intent(in)    :: fnp1
    type(fdistribu5d)            , intent(inout) :: sJ0ftmp
    real(RKIND), dimension(:,:,:), pointer       :: sJ0fmu_eq
    type(poisson)                , intent(in)    :: poiss

    integer     :: imom, ir
    !-> variables for the velocities projected on r
    real(RKIND), dimension(:,:,:), pointer :: dPhidr
    real(RKIND), dimension(:,:,:), pointer :: dPhidtheta
    real(RKIND), dimension(:,:,:), pointer :: dPhidphi
    real(RKIND), dimension(:,:,:), pointer :: Phi_PNrPNthetaNphi
    real(RKIND), dimension(:,:,:), pointer :: transpdPhidr
    real(RKIND), dimension(:,:,:), pointer :: transpdPhidtheta
    real(RKIND), dimension(:,:,:), pointer :: transpdPhidphi
    !-> variables for potential energy
    real(RKIND) :: Enpot

    integer :: istart_clamp, iend_clamp
#ifdef TIMER
    ! -> variables for time perfomance analysis
    integer(TIMEPREC) :: tdeb, tfin
    real(RKIND)       :: tdPhidx, tgyro, tcomm, tcomp1, tcomp2
    real(RKIND)       ::  tvect(32), tmin(32), tmax(32), tavg(32)
#endif

    call Write_dyn_balise_open('CPU_diag_moments2') !R4

    istart_clamp = max(0,istart_buf)
    iend_clamp   = min(geom%Nr,iend_buf)
	
    call write_expr("istart_clamp", "max(0,istart_buf)")
    call write_expr("iend_clamp"  , "min(geom%Nr,iend_buf)")

    !*** initialisation to 0 of the radial arrays ***
    do imom = 1,nbmoments_part
      do ir =  0,geom%Nr
#ifdef NO_MPI_INPLACE
        part_Fmoments_loc(ir,imom)  = 0._RKIND
#endif
        part_Fmoments(ir,imom) = 0._RKIND
      end do
    end do

    if (distrib3D) then
      call FTR_ALLOC_3_ZERO(dPhidr,istart_clamp,iend_clamp,0,geom%Ntheta,0,geom%Nphi)
      call FTR_ALLOC_3_ZERO(dPhidtheta,istart_clamp,iend_clamp,0,geom%Ntheta,0,geom%Nphi)
      call FTR_ALLOC_3_ZERO(dPhidphi,istart_clamp,iend_clamp,0,geom%Ntheta,0,geom%Nphi)
      call FTR_ALLOC_3_ZERO(Phi_PNrPNthetaNphi,istart_clamp,iend_clamp,0,geom%Ntheta,0,geom%Nphi)
      call FTR_ALLOC_3_ZERO(transpdPhidr,0,geom%Nr,0,geom%Ntheta,kstart,kend)
      call FTR_ALLOC_3_ZERO(transpdPhidtheta,0,geom%Nr,0,geom%Ntheta,kstart,kend)
      call FTR_ALLOC_3_ZERO(transpdPhidphi,0,geom%Nr,0,geom%Ntheta,kstart,kend)
    else
      call FTR_ALLOC_3_ZERO(dPhidr,istart_clamp,iend_clamp,0,geom%Ntheta,0,geom%Nphi)
      call FTR_ALLOC_3_ZERO(dPhidtheta,istart_clamp,iend_clamp,0,geom%Ntheta,0,geom%Nphi)
      call FTR_ALLOC_3_ZERO(dPhidphi,istart_clamp,iend_clamp,0,geom%Ntheta,0,geom%Nphi)
      Phi_PNrPNthetaNphi => poiss%Phi_NrNthPNphi_Cworld

      transpdPhidr       => NULL()
      call FTR_ALLOC_3_ZERO(transpdPhidtheta,0,geom%Nr,0,geom%Ntheta,kstart,kend)
      call FTR_ALLOC_3_ZERO(transpdPhidphi,0,geom%Nr,0,geom%Ntheta,kstart,kend)
    end if

    !*** computation of the derivatives of Phi
#ifdef TIMER
    call clck_time(tdeb)
    call ppbarrier_timer()
#endif
    call comp_dPhidx(geom, poiss, Phi_PNrPNthetaNphi, dPhidr, &
                    dPhidtheta, dPhidphi, transpdPhidr, &
                    transpdPhidtheta, transpdPhidphi)
#ifdef TIMER
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tdPhidx)
#endif

#ifdef TIMER
    call clck_time(tdeb)
#endif
    call Physics_gyro_preprocess(geom, init_magnet, init_curr, J0, &
      fnp1, sJ0ftmp, transpdPhidtheta, transpdPhidphi)
#ifdef TIMER
    call ppbarrier_timer()
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tgyro)
#endif
    call Physics_part_moments_preprocess( &
      geom, coord_sys, init_magnet, init_curr, &
      fnp1, sJ0ftmp, sJ0fmu_eq, Enpot, &
      dPhidr, dPhidtheta, dPhidphi, Phi_PNrPNthetaNphi)
#ifdef TIMER
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tcomp1)
#endif

#ifdef TIMER
    call clck_time(tdeb)
#endif
    call Physics_part_moments_transfer(geom, Enpot)
#ifdef TIMER
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tcomm)
#endif

#ifdef TIMER
    call clck_time(tdeb)
#endif
    call Physics_part_moments_postprocess(geom, coord_sys)
    call ppbarrier_timer()

    if (distrib3D) then
      call FTR_DEALLOC(dPhidr)
      call FTR_DEALLOC(dPhidtheta)
      call FTR_DEALLOC(dPhidphi)
      call FTR_DEALLOC(Phi_PNrPNthetaNphi)
      call FTR_DEALLOC(transpdPhidr)
      call FTR_DEALLOC(transpdPhidtheta)
      call FTR_DEALLOC(transpdPhidphi)
    else
      call FTR_DEALLOC(dPhidr)
      call FTR_DEALLOC(dPhidtheta)
      call FTR_DEALLOC(dPhidphi)

      call FTR_DEALLOC(transpdPhidtheta)
      call FTR_DEALLOC(transpdPhidphi)
   endif

#ifdef TIMER
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tcomp2)
#endif

#ifdef TIMER
    tvect(1:5) = (/ tdPhidx, tgyro, tcomp1, tcomm, tcomp2 /)
    call pp_reduce_vector(tvect,5,tmin,tmax,tavg)
    if (plocal_id .eq. 0) then
      write(6,'(I4,A,F13.7,A,F13.7,A,F13.7,A,F13.7,A,F13.7)') &
            pglobal_id," TMIN part moments: dPhidx ",tmin(1), &
            " J0f ",tmin(2)," comp1 ",tmin(3), &
            " comm ",tmin(4)," comp2 ",tmin(5)
      write(6,'(I4,A,F13.7,A,F13.7,A,F13.7,A,F13.7,A,F13.7)') &
            pglobal_id," TMAX part moments: dPhidx ",tmax(1), &
            " J0f ",tmax(2)," comp1 ",tmax(3), &
            " comm ",tmax(4)," comp2 ",tmax(5)
      write(6,'(I4,A,F13.7,A,F13.7,A,F13.7,A,F13.7,A,F13.7)') &
            pglobal_id," TAVG part moments: dPhidx ",tavg(1), &
            " J0f ",tavg(2)," comp1 ",tavg(3), &
            " comm ",tavg(4)," comp2 ",tavg(5)
    end if
#endif

    call Write_dyn_balise_close('CPU_diag_moments2') !R4
  end subroutine compute_part_moments


  !------------------------------------------------------
  !> Parallel preprocessing for compute_part_moments()
  !> computation of J0.f
  !------------------------------------------------------
  subroutine Physics_gyro_preprocess ( &
    geom, init_magnet, init_curr, &
    J0, fnp1, sJ0ftmp, transpdPhidtheta, transpdPhidphi )
    use coord_system_class , only: jacobian_space
    ! use globals            , only: kstart, kend, lstart, lend, &
    !   mu_id, Nbthread, transp_Bstar
    ! use globals            , only: Nr, Ntheta
    use gyroaverage_class  , only: omp_compute_gyrofunction_2D
    use init_magnetic_class, only: compute_jacobian_velocity, &
      compute_noprecomp_jacobian_velocity
    use MPIutils_module    , only: f4D_transp, &
      pptranspose_forward, pptranspose_backward, f4D_send, f4D_recv
#ifdef _OPENMP
    use OMPutils_module    , only: omp_get_thread_num
#endif
    use OMPutils_module    , only: Comp1_1Nrp1_1Ntheta, &
      Romp1_1Nrp1_1Nthetap1, Romp1_0Nr, Romp2_1Nrp1_1Nthetap1, &
      Romp5_0Nr
    use wrap_alloc_module

    ! use globals, only : Nbproc_species, pglobal_id, transpose_version, &
    !      TRANSPOSE_SENDIRECV, TRANSPOSE_ALLTOALLW 
    use globals

    type(geometry)               , intent(in)    :: geom
    type(init_magnetic)          , intent(in)    :: init_magnet
    type(init_current)           , intent(in)    :: init_curr
    type(J0operator)             , intent(inout) :: J0
    type(fdistribu5d)            , intent(in)    :: fnp1
    type(fdistribu5d)            , intent(inout) :: sJ0ftmp

    !-> variables for the velocities projected on r
    real(RKIND), dimension(:,:,:), pointer :: transpdPhidtheta
    real(RKIND), dimension(:,:,:), pointer :: transpdPhidphi
    real(RKIND), dimension(:),     pointer :: vExB_gradr_1D

    integer     :: tid, ithread, ir, itheta, iphi, imix, ivpar
    real(RKIND) :: fval

    !-> variables for J0.f
    complex(CKIND), dimension(:,:), pointer :: Acomp
    real(RKIND), dimension(:,:), pointer    :: J0f_rtheta
    real(RKIND), dimension(:,:), pointer    :: J0_fvExBgradr      

    !-> variables for the flux surface average
    real(RKIND) :: jacob_space_tmp, jacob_vel_tmp
    real(RKIND) :: coeff_intdtheta_tmp
    real(RKIND) :: coeff_intdphi_tmp, coeff_intdvpar_tmp
    real(RKIND) :: coeff_intdmu_tmp
    real(RKIND) :: coeff_int_volume

    !-> variables for the Reynold Stress
    real(RKIND) :: RStheta_tmp
    type(omp1dvector), dimension(:), pointer :: RStheta

    integer :: i_proc, ierr

    do tid = 1, Nbthread
       call FTR_ALLOC_2(Romp1_1Nrp1_1Nthetap1(tid)%val,1,Nr+1,1,Ntheta+1)
       call FTR_ALLOC_2(Romp2_1Nrp1_1Nthetap1(tid)%val,1,Nr+1,1,Ntheta+1)

       call FTR_ALLOC_2_ZERO(Comp1_1Nrp1_1Ntheta(tid)%val,1,Nr+1,1,Ntheta)
    enddo

    do tid = 1, Nbthread
       call FTR_ALLOC_2(J0%omp_diag(tid)%val,1,J0%N1,1,J0%N2)
       call FTR_ALLOC_2(J0%omp_ldiag(tid)%val,1,J0%N1-1,1,J0%N2)
       call FTR_ALLOC_2(J0%omp_udiag(tid)%val,1,J0%N1-1,1,J0%N2)
    enddo

    coeff_intdmu_tmp = geom%coeff_intdmu(mu_id)

    select case ( transpose_version )
    case ( TRANSPOSE_SENDIRECV )
       call FTR_ALLOC_4_ZERO(f4D_send,0,dom_r-1,0,dom_theta-1,0,dom_phi-1,0,dom_vpar-1)
       call FTR_ALLOC_4_ZERO(f4D_recv,0,dom_r-1,0,dom_theta-1,0,dom_phi-1,0,dom_vpar-1)
    case ( TRANSPOSE_ALLTOALLW )
       ! nothing specific to allocate
    case default
       print *, pglobal_id, 'Invalid value for transpose_version', &
            transpose_version
       call MPI_abort(MPI_COMM_WORLD, 1, ierr)
    end select
    
    PRINT_CKSUM(fnp1%values)

    !-> transposition of f to have all the informations in (r,theta)
    call pptranspose_forward(fnp1%values,timing=.false.)

    PRINT_CKSUM(f4D_transp)
 
    RStheta => Romp5_0Nr

#ifdef _OPENMP
!$OMP PARALLEL default(none) &
!$OMP private(tid,Acomp,J0f_rtheta, &
!$OMP  vExB_gradr_1D,J0_fvExBgradr, &
!$OMP  ir,itheta,iphi,ivpar,imix,fval, &
!$OMP  coeff_intdvpar_tmp,coeff_intdphi_tmp, &
!$OMP  coeff_intdtheta_tmp,coeff_int_volume, &
!$OMP  jacob_space_tmp,jacob_vel_tmp) &
!$OMP shared (Comp1_1Nrp1_1Ntheta, &
!$OMP  Romp1_1Nrp1_1Nthetap1, &
!$OMP  Romp2_1Nrp1_1Nthetap1, &
!$OMP  Romp1_0Nr, Romp5_0Nr, &
!$OMP  transp_Bstar, & 
!$OMP  RStheta, &
!$OMP  kstart,kend,lstart,lend,mu_id, &
!$OMP  fnp1,f4D_transp,jacobian_space,geom, &
!$OMP  transpdPhidtheta,transpdPhidphi, &
!$OMP  init_magnet,&
!$OMP  init_curr,J0,coeff_intdmu_tmp,pglobal_id)
!$OMP BARRIER
    tid = 1+omp_get_thread_num()
#else
    tid = 1
#endif
    Acomp          => Comp1_1Nrp1_1Ntheta(tid)%val
    J0f_rtheta     => Romp1_1Nrp1_1Nthetap1(tid)%val
    J0_fvExBgradr  => Romp2_1Nrp1_1Nthetap1(tid)%val
    vExB_gradr_1D  => Romp1_0Nr(tid)%val
    do ir = 0,fnp1%n1
      RStheta(tid)%val(ir) = 0._RKIND
    end do
!$OMP DO SCHEDULE(STATIC)
    do imix = 0, (lend-lstart+1)*(kend-kstart+1)-1
      ! Variable imix encapsulates ivpar and iphi together 
      ! in order to parallelize on these two variables 
      ! simultaneously (kind of nested loops parallelization).
      ! Useful on BG/Q machine with many threads (turing).
      ivpar = lstart + mod(imix,lend-lstart+1)
      iphi  = kstart + (imix/(lend-lstart+1))

      coeff_intdvpar_tmp  = geom%coeff_intdvpar(ivpar)
      coeff_intdphi_tmp   = geom%coeff_intdphi(iphi)
      
      do itheta = 0,fnp1%n2-1
        !-> computation of vExB_gradr(0:Nr)             
        call compute_vExBgradr(geom, &                 
          init_magnet,init_curr, &                     
          transpdPhidtheta,transpdPhidphi,0,geom%Nr, & 
          itheta,itheta,iphi,iphi, &                   
          ivpar,ivpar,.true.,vExB_gradr_1D)            
        do ir = 0,fnp1%n1
          fval = f4D_transp(ir,itheta,iphi,ivpar)
          J0f_rtheta(ir+1,itheta+1)    = fval
          J0_fvExBgradr(ir+1,itheta+1) = fval*vExB_gradr_1D(ir)
        end do
      end do
      ! Boundary condition in theta direction
      do ir = 0,fnp1%n1
         J0f_rtheta(ir+1,fnp1%n2+1)    = J0f_rtheta(ir+1,1)
         J0_fvExBgradr(ir+1,fnp1%n2+1) = J0_fvExBgradr(ir+1,1)
      end do

      !--> computation of J0.f
      call omp_compute_gyrofunction_2D(J0,geom, &
        mu_id,Acomp,J0f_rtheta,tid)
      !--> computation of J0.(f.vExB_gradr)
      call omp_compute_gyrofunction_2D(J0,geom, &
        mu_id,Acomp,J0_fvExBgradr,tid)
      !-> \int J0.(f*vExB.gradr) d3v (for RStheta)
      do itheta = 0,geom%Ntheta
        coeff_intdtheta_tmp = geom%coeff_intdtheta(itheta)
        do ir = 0,geom%Nr
          !-> jacobian_velocity = 2*pi*Bstar
          if (transp_Bstar) then
            call compute_noprecomp_jacobian_velocity(geom, &
              init_magnet,init_curr,ir,itheta,ivpar, &
              jacob_vel_tmp)            
          else
            call compute_jacobian_velocity(geom,init_magnet, &
              init_curr,ir,itheta,ivpar,jacob_vel_tmp)            
          end if
          !-> jacobian in space
          jacob_space_tmp    = jacobian_space(ir,itheta)
          !-> compute the elements of volume in space and velocity
          coeff_int_volume   = jacob_space_tmp * &
            coeff_intdtheta_tmp*coeff_intdphi_tmp * &
            jacob_vel_tmp*coeff_intdvpar_tmp*coeff_intdmu_tmp 
          RStheta(tid)%val(ir) = RStheta(tid)%val(ir) + &
            J0_fvExBgradr(ir+1,itheta+1) * coeff_int_volume
        end do
      end do
      do itheta = 0,fnp1%n2
        do ir = 0,fnp1%n1
          f4D_transp(ir,itheta,iphi,ivpar) = &
            J0f_rtheta(ir+1,itheta+1)
        end do
      end do
    end do
!$OMP END DO
!$OMP END PARALLEL
    !->  f4D_transp -> J0ftmp%values
    call pptranspose_backward(sJ0ftmp%values,timing=.false.)

    select case ( transpose_version )
    case ( TRANSPOSE_SENDIRECV )
       call FTR_DEALLOC(f4D_send)
       call FTR_DEALLOC(f4D_recv)
    case ( TRANSPOSE_ALLTOALLW )
       ! nothing specific to allocate
    case default
       print *, pglobal_id, 'Invalid value for transpose_version', &
            transpose_version
       call MPI_abort(MPI_COMM_WORLD, 1, ierr)
    end select

    do ir = 0, geom%Nr
      RStheta_tmp = 0._RKIND
      do ithread = 1,Nbthread
        RStheta_tmp = RStheta_tmp + &
          RStheta(ithread)%val(ir)
      end do
#ifdef NO_MPI_INPLACE
      part_Fmoments_loc(ir,10) = RStheta_tmp
#else
      part_Fmoments(ir,10) = RStheta_tmp
#endif
    end do

   do tid = 1, Nbthread
     call FTR_DEALLOC(J0%omp_diag(tid)%val)
     call FTR_DEALLOC(J0%omp_ldiag(tid)%val)
     call FTR_DEALLOC(J0%omp_udiag(tid)%val)
   enddo

    do tid = 1, Nbthread
       call FTR_DEALLOC(Romp1_1Nrp1_1Nthetap1(tid)%val)
       call FTR_DEALLOC(Romp2_1Nrp1_1Nthetap1(tid)%val)

       call FTR_DEALLOC(Comp1_1Nrp1_1Ntheta(tid)%val)
    enddo
  end subroutine Physics_gyro_preprocess


  !-----------------------------------------------------
  !> Parallel preprocessing for compute_part_moments()
  !-----------------------------------------------------
  subroutine Physics_part_moments_preprocess (geom, coord_sys, &
    init_magnet, init_curr, fnp1, sJ0ftmp, sJ0fmu_eq, &
    Enpot, dPhidr, dPhidtheta, dPhidphi, &
    Phi_PNrPNthetaNphi)

    use coord_system_class , only: jacobian_space
    use globals            , only: distrib3D, &
      istart, iend, jstart, jend, mu_id, nbmoments_part, Nbthread
    use init_magnetic_class, only: compute_jacobian_velocity, &
      compute_noprecomp_jacobian_velocity
#ifdef _OPENMP
    use OMPutils_module    , only: omp_get_thread_num
#endif
    use OMPutils_module    , only: Romp1_0Nr, Romp1_1nbmom_0Nr, &
      Romp2_0Nr, Romp3_0Nr, Romp4_0Nr

    type(geometry)               , intent(in)    :: geom
    type(coord_system)           , intent(in)    :: coord_sys
    type(init_magnetic)          , intent(in)    :: init_magnet
    type(init_current)           , intent(in)    :: init_curr
    type(fdistribu5d)            , intent(in)    :: fnp1
    type(fdistribu5d)            , intent(inout) :: sJ0ftmp
    real(RKIND), dimension(:,:,:), pointer       :: sJ0fmu_eq
    !> Potential energy result
    real(RKIND),                   intent(out)   :: Enpot
    !-> variables for the velocities projected on r
    real(RKIND), dimension(:,:,:), pointer :: dPhidr
    real(RKIND), dimension(:,:,:), pointer :: dPhidtheta
    real(RKIND), dimension(:,:,:), pointer :: dPhidphi
    real(RKIND), dimension(:,:,:), pointer :: Phi_PNrPNthetaNphi

    integer     :: imom, tid, ithread
    integer     :: ir, itheta, iphi, ivpar, imix
    integer     :: nbmoments
    real(RKIND) :: sqrt_g22, vpar, mu_value, Bij
    real(RKIND) :: Bi_thetapi, trap_constraint
    real(RKIND) :: J0f, J0fmu_eq, J0fv
    !-> variables for the flux surface average
    real(RKIND) :: jacob_space_tmp, jacob_vel_tmp
    real(RKIND) :: coeff_intdr_tmp, coeff_intdtheta_tmp
    real(RKIND) :: coeff_intdphi_tmp, coeff_intdvpar_tmp
    real(RKIND) :: coeff_intdmu_tmp
    real(RKIND) :: coeff_int_velocities, coeff_int_volume
    !-> variables for the velocities projected on theta
    real(RKIND) :: vExB_etheta, vD_etheta
    real(RKIND), dimension(:)    , pointer :: vExB_gradtheta_1D
    real(RKIND), dimension(:)    , pointer :: vD_gradtheta_1D
    !-> variables for the velocities projected on r
    real(RKIND) :: vExB_er, vD_er
    real(RKIND) :: energy, J0fE
    real(RKIND), dimension(:)    , pointer :: vD_gradr_1D
    real(RKIND), dimension(:)    , pointer :: vExB_gradr_1D

    !-> variables for the theta component of the magnetisation
    real(RKIND)                        :: bphi
    !-> variables for potential energy
    real(RKIND), dimension(1:Nbthread) :: Enpot_thread

    !*** computation of the moments associated to the particles ***
    !-> BECAREFUL: if you want to add the computation of
    !    a particle moment, do not forget to change
    !    nbmoments_part in 'globals.F90' file
    nbmoments        = nbmoments_part
    mu_value         = fnp1%mu_value
    coeff_intdmu_tmp = geom%coeff_intdmu(mu_id)

#ifdef _OPENMP
!$OMP PARALLEL default(none) &
!$OMP private(tid,ir,itheta,iphi,ivpar,imom, &
!$OMP  Bij,sqrt_g22,vpar,J0f,J0fmu_eq,J0fv, &
!$OMP  energy,J0fE,vExB_er,vD_er, &
!$OMP  vExB_gradr_1D,vD_gradr_1D, &
!$OMP  vExB_gradtheta_1D,vD_gradtheta_1D, &
!$OMP  vExB_etheta,vD_etheta, &
!$OMP  Bi_thetapi,trap_constraint, &
!$OMP  coeff_intdr_tmp,coeff_intdtheta_tmp, &
!$OMP  coeff_intdphi_tmp,coeff_intdvpar_tmp, &
!$OMP  coeff_int_velocities,coeff_int_volume, &
!$OMP  jacob_space_tmp,jacob_vel_tmp,bphi) &
!$OMP shared(geom,coord_sys,init_magnet,init_curr, &
!$OMP  Romp1_0Nr,Romp2_0Nr,distrib3D, &
!$OMP  Romp3_0Nr,Romp4_0Nr,Romp1_1nbmom_0Nr, &
!$OMP  istart,iend,jstart,jend,nbmoments, &
!$OMP  sJ0ftmp,sJ0fmu_eq,mu_value,jacobian_space, &
!$OMP  dPhidr, &
!$OMP  dPhidtheta,dPhidphi, &
!$OMP  coeff_intdmu_tmp,Enpot_thread, &
!$OMP  Phi_PNrPNthetaNphi)
!$OMP BARRIER
    tid = 1+omp_get_thread_num()
#else
    tid = 1
#endif

    vExB_gradr_1D     => Romp1_0Nr(tid)%val
    vD_gradr_1D       => Romp2_0Nr(tid)%val

    vExB_gradtheta_1D => Romp3_0Nr(tid)%val
    vD_gradtheta_1D   => Romp4_0Nr(tid)%val

    !*** initialisation of the values to 0 ***
    do imom = 1,nbmoments
      do ir = 0,geom%Nr
        Romp1_1nbmom_0Nr(tid)%val(imom,ir) = 0._RKIND
      end do
    end do
    Enpot_thread(tid) = 0._RKIND

!$OMP DO SCHEDULE(STATIC)
    do imix = 0, (geom%Nvpar+1)*(geom%Nphi)-1
      ! Variable imix encapsulates ivpar and iphi together 
      ! in order to parallelize on these two variables 
      ! simultaneously (kind of nested loops parallelization).
      ! Useful on BG/Q machine with many threads (turing).
      ivpar = mod(imix,geom%Nvpar+1)
      iphi  = (imix/(geom%Nvpar+1))
      vpar                = geom%vparg(ivpar)
      coeff_intdvpar_tmp  = geom%coeff_intdvpar(ivpar)
      coeff_intdphi_tmp   = geom%coeff_intdphi(iphi)
      do itheta = jstart,jend
        coeff_intdtheta_tmp = geom%coeff_intdtheta(itheta)
        !*** computation of the radial velocities    ***
        !-> computation of vExB_gradr(0:Nr)
        call compute_vExBgradr(geom,init_magnet,init_curr, &
          dPhidtheta,dPhidphi,istart,iend,itheta,itheta, &
          iphi,iphi,ivpar,ivpar,.false.,vExB_gradr_1D)
        !-> computation of vD_gradr(0:Nr)
        call compute_vDgradr(geom,init_magnet,init_curr, &
          istart,iend,itheta,itheta,iphi,iphi, &
          ivpar,ivpar,.false.,vD_gradr_1D)
        !*** computation of the poloidal velocities    ***
        !-> computation of vExB_gradtheta(0:Nr)
        call compute_vExBgradtheta(geom, &
          init_magnet,init_curr, &
          dPhidr,dPhidphi,istart,iend,itheta,itheta, &
          iphi,iphi,ivpar,ivpar,.false.,vExB_gradtheta_1D)
        !-> computation of vD_gradtheta(0:Nr)
        call compute_vDgradtheta(geom,init_magnet,init_curr, &
          istart,iend,itheta,itheta,iphi,iphi, &
          ivpar,ivpar,.false.,vD_gradtheta_1D)
        do ir = istart,iend
          coeff_intdr_tmp = geom%coeff_intdr(ir)
          Bij             = init_magnet%B_norm(ir,itheta)
          Bi_thetapi      = init_magnet%B_at_thetapi(ir)
          J0f             = sJ0ftmp%values(ir,itheta,iphi,ivpar)
          J0fmu_eq        = sJ0fmu_eq(ir,itheta,ivpar)
          J0fv            = J0f*vpar
          energy          = 0.5_RKIND*vpar*vpar+mu_value*Bij
          trap_constraint = sqrt(2._RKIND*mu_value * &
            (Bi_thetapi-Bij))
          J0fE            = J0f*energy
          !-> jacobian_velocity = 2*pi*Bstar
          if (geom%Nmu .eq. 0) then
#ifdef NO_PRECOMP_BSTAR 
            call compute_Bstar(geom,mthis,init_curr, &
              ir,itheta,ivpar,jacob_vel_tmp)
#else
            call precomp_Bstar(ir,itheta,ivpar,jacob_vel_tmp)
#endif
          else
            if (distrib3D) then
              call compute_noprecomp_jacobian_velocity( &
                geom,init_magnet,init_curr, &
                ir,itheta,ivpar,jacob_vel_tmp)            
            else
              call compute_jacobian_velocity(geom,init_magnet, &
                init_curr,ir,itheta,ivpar,jacob_vel_tmp)            
            end if
          end if
          !-> jacobian in space
          jacob_space_tmp = jacobian_space(ir,itheta)
          !-> compute the elements of velocity
          coeff_int_velocities = jacob_vel_tmp * &
            coeff_intdvpar_tmp*coeff_intdmu_tmp 
          !-> compute the elements in the total phase space
          coeff_int_volume = jacob_space_tmp * &
            coeff_intdtheta_tmp*coeff_intdphi_tmp * &
            coeff_int_velocities
          !*** computation of the projection of the   ***
          !***  velocities on the unit radial         ***
          !***  vector                                ***
          !-> vE.er = vExB_gradr
          vExB_er = vExB_gradr_1D(ir-istart)
          !-> vD.er = vD_gradr 
          vD_er   = vD_gradr_1D(ir-istart)
          !*** computation of the projection of the   ***
          !***  radial velocities on the unit         ***
          !***  poloidal vector                       ***
          sqrt_g22 = sqrt(coord_sys%g22(ir,itheta)) 
          !-> vE.etheta = sqrt(g_{22})*vExB_gradtheta
          vExB_etheta = sqrt_g22*vExB_gradtheta_1D(ir-istart)
          !-> vD.etheta = sqrt(g_{22})*vD_gradtheta
          vD_etheta = sqrt_g22*vD_gradtheta_1D(ir-istart)
          
          !*** computation of b_phi = B_phi/B ***
          bphi = init_magnet%Bphi(ir,itheta)/Bij

          !*** computation of the integrals in velocity and ***
          !***  in theta and phi                            ***
          !-> 1) \int M_phi dtheta dphi (for nvpolor_mag)
          !->  where M_phi = - \int (\int J0.f b_phi 
          !->                 Jv dvpar)*mu dmu            
          Romp1_1nbmom_0Nr(tid)%val(1,ir) = &
            Romp1_1nbmom_0Nr(tid)%val(1,ir) - &
            J0f*mu_value*bphi * &
            coeff_int_velocities * &
            coeff_intdtheta_tmp*coeff_intdphi_tmp
          !-> 2) \int J0.f d3v  (for Ni)
          Romp1_1nbmom_0Nr(tid)%val(2,ir) = &
            Romp1_1nbmom_0Nr(tid)%val(2,ir) + &
            J0f*coeff_int_volume
          !-> 3) \int_trapped J0.f d3v  (for Ni_trapped)
          if (abs(vpar).lt.trap_constraint) then
            Romp1_1nbmom_0Nr(tid)%val(3,ir) = &
              Romp1_1nbmom_0Nr(tid)%val(3,ir) + &
              J0f*coeff_int_volume
          else
            Romp1_1nbmom_0Nr(tid)%val(3,ir) = 0._RKIND
          end if
          !-> 4) \int J0.f*vpar d3v (for nVpar)
          Romp1_1nbmom_0Nr(tid)%val(4,ir) = &
            Romp1_1nbmom_0Nr(tid)%val(4,ir) + &
            J0fv*coeff_int_volume
          !-> 5) \int J0.f*E d3v (for P)
          Romp1_1nbmom_0Nr(tid)%val(5,ir) = &
            Romp1_1nbmom_0Nr(tid)%val(5,ir) + &
            J0fE*coeff_int_volume
          !-> 6) \int J0.f*E*vE.er d3v (for Q_turb)
          Romp1_1nbmom_0Nr(tid)%val(6,ir) = &
            Romp1_1nbmom_0Nr(tid)%val(6,ir) + &
            J0fE*vExB_er*coeff_int_volume
          !-> 7) \int J0.f*E*vD.er d3v (for Q_neo)
          Romp1_1nbmom_0Nr(tid)%val(7,ir) = &
            Romp1_1nbmom_0Nr(tid)%val(7,ir) + &
            J0fE*vD_er*coeff_int_volume
          !-> 8) \int J0.f*vE.er d3v (for Gamma_turb)
          Romp1_1nbmom_0Nr(tid)%val(8,ir) = &
            Romp1_1nbmom_0Nr(tid)%val(8,ir) + &
            J0f*vExB_er*coeff_int_volume
          !-> 9) \int J0.f*vD.er d3v (for Gamma_neo)
          Romp1_1nbmom_0Nr(tid)%val(9,ir) = &
            Romp1_1nbmom_0Nr(tid)%val(9,ir) + &
            J0f*vD_er*coeff_int_volume
          !-> 10) Romp1_1nbmom_0Nr(tid)%val(10,ir) 
          !->    is already computed
          !-> 11) \int J0.f vtilde_r vpar d3v (for RSpar)
          Romp1_1nbmom_0Nr(tid)%val(11,ir) = &
            Romp1_1nbmom_0Nr(tid)%val(11,ir) + &
            vExB_er*J0fv*coeff_int_volume
          !-> compute \int JO.f Phi d3v
          Enpot_thread(tid) = Enpot_thread(tid) + &
            0.5_RKIND*(J0f-J0fmu_eq) * &
            Phi_PNrPNthetaNphi(ir,itheta,iphi) * &
            coeff_int_volume*coeff_intdr_tmp            
        end do ! end of do ir
      end do   ! end of do itheta
    end do     ! end of do imix (iphi + ivpar)
!$OMP END DO
!$OMP BARRIER
!$OMP END PARALLEL

    !*** fill an unique array containing all the       ***
    !***  radial profiles of the fluid moments         ***
    !*** Rq: each processor fills the part istart:iend ***
    !***     and all the others values are put equal   ***
    !***     to 0                                      ***
    Enpot = 0._RKIND
    do ithread = 1,Nbthread
      do ir = istart,iend
        do imom = 1,nbmoments
          if (imom .ne. 10) then
#ifdef NO_MPI_INPLACE
            part_Fmoments_loc(ir,imom) = &
              part_Fmoments_loc(ir,imom) + & 
              Romp1_1nbmom_0Nr(ithread)%val(imom,ir)
#else
            part_Fmoments(ir,imom) = &
              part_Fmoments(ir,imom) + & 
              Romp1_1nbmom_0Nr(ithread)%val(imom,ir)
#endif
          end if
        end do
      end do
      Enpot = Enpot + Enpot_thread(ithread) 
    end do
  end subroutine Physics_part_moments_preprocess


  !----------------------------------------------------------
  !> Data transfer between parallel preprocessing and local
  !> postprocessing for compute_part_moments()
  !----------------------------------------------------------
  subroutine Physics_part_moments_transfer ( geom, Enpot )

    use globals,    only: nbmoments_part, Nbproc_species, &
                          mpi_comm_species, diag_targ, pInSpecies_id

    type(geometry), intent(in):: geom
    !> potential energy
    real(RKIND),    intent(in):: Enpot
    
    integer     :: nbelements, nbmoments, ierr, imom, ir

    !*** computation of the moments associated to the particles ***
    !-> BECAREFUL: if you want to add the computation of
    !    a particle moment, do not forget to change
    !    nbmoments_part in 'globals.F90' file
    nbmoments = nbmoments_part

    if ( Nbproc_species /= 1 ) then
      nbelements = nbmoments*(geom%Nr+1)
#ifdef NO_MPI_INPLACE
      call MY_Reduce(part_Fmoments_loc, part_Fmoments, nbelements, &
        MPI_REAL8, MPI_SUM, diag_targ(DIAG_RPROF_PART), &
        mpi_comm_species, ierr)
#else
      if ( pInSpecies_id == diag_targ(DIAG_RPROF_PART) ) then
        call MY_Reduce(MPI_IN_PLACE, part_Fmoments, nbelements, &
          MPI_REAL8, MPI_SUM, diag_targ(DIAG_RPROF_PART), &
          mpi_comm_species, ierr)
      else
        call MY_Reduce(part_Fmoments, MPI_BOTTOM, nbelements, &
          MPI_REAL8, MPI_SUM, diag_targ(DIAG_RPROF_PART), &
          mpi_comm_species, ierr)
      endif
#endif
      call MY_Reduce(Enpot, Enpot_diag, 1, &
        MPI_REAL8, MPI_SUM, diag_targ(DIAG_CL), &
        mpi_comm_species, ierr)
    else
#ifdef NO_MPI_INPLACE
      do imom = 1,nbmoments
        do ir = 0,geom%Nr
          part_Fmoments(ir,imom) = part_Fmoments_loc(ir,imom)
        end do
      end do
#endif
      Enpot_diag = Enpot
    end if
  end subroutine Physics_part_moments_transfer


  !-----------------------------------------------------
  !> Local postprocessing for compute_part_moments()
  !-----------------------------------------------------
  subroutine Physics_part_moments_postprocess ( geom, coord_sys )

    use coord_system_class, only: intdthetadphi_Js
    use geometry_class    , only: ipow
    use globals           , only: hat_As, Rarray1_Nr, pSpecies_id
    use utils_module      , only: deriv1

    type(geometry)               , intent(in)    :: geom
    type(coord_system)           , intent(in)    :: coord_sys

    integer     :: ir
    real(RKIND) :: sqrt_g22
    real(RKIND) :: intdthetadphi_Js_tmp
    real(RKIND), dimension(:), pointer :: intMphi_dthetadphi
    real(RKIND) :: coeff_pressr

    intMphi_dthetadphi => Rarray1_Nr

    coeff_pressr = 1._RKIND/(0.5_RKIND+float(ipow))
    do ir = 0,geom%Nr
      intdthetadphi_Js_tmp   = intdthetadphi_Js(ir)
      !-> 1) compute \int M_phi dtheta dphi 
      intMphi_dthetadphi(ir) = part_Fmoments(ir,1)
      !-> 2) compute Ni(r) = <\int J0.f d3v>_FS
      nir_diag(ir)           = & 
        part_Fmoments(ir,2)/intdthetadphi_Js_tmp
      !-> 3) compute Ni_trapped(r) = <\int_trapped J0.f d3v>_FS
      nir_trapped_diag(ir)   = & 
        part_Fmoments(ir,3)/intdthetadphi_Js_tmp
      !-> 4) compute nVpar(r) = <\int J0.f*vpar d3v>_FS
      nuparr_diag(ir)        = &
        part_Fmoments(ir,4)/intdthetadphi_Js_tmp
      !-> 5) compute P(r) = <\int J0.f*E d3v>_FS
      pressr_diag(ir)        = coeff_pressr * &
        part_Fmoments(ir,5)/intdthetadphi_Js_tmp
      !-> 6) compute Q_turb(r) = <\int J0.f*E*vE.er d3v>_FS
      Qr_turb_diag(ir)       = &
        part_Fmoments(ir,6)/intdthetadphi_Js_tmp
      !-> 7) compute Q_neo(r) = <\int J0.f*E*vD.er d3v>_FS 
      Qr_neo_diag(ir)        = &
        part_Fmoments(ir,7)/intdthetadphi_Js_tmp
      !-> 8) compute Gamma_turb(r) = <\int J0.f*vE.er d3v>_FS
      Gammar_turb_diag(ir)   = &
        part_Fmoments(ir,8)/intdthetadphi_Js_tmp
      !-> 9) compute Gamma_neo(r) = <\int J0.f*vD.er d3v>_FS
      Gammar_neo_diag(ir)    = &
        part_Fmoments(ir,9)/intdthetadphi_Js_tmp
      !-> 10) compute RStheta(r) =  
      !->               <\int J0.(f*vExB.gradr) d3v>_FS
      RSthetar_diag(ir)      = &
        part_Fmoments(ir,10)/intdthetadphi_Js_tmp
      !-> 11) compute RSpar(r) = <\int J0.f vtilde_r vpar d3v>_FS
      RSparr_diag(ir)        = &
        part_Fmoments(ir,11)/intdthetadphi_Js_tmp
    end do
 
    !*** Computation of the flux surface average of the  ***
    !***  theta component of the magnetisation, i.e:     ***
    !***  nvpolor_mag = -<rot(M).etheta>_FS              ***
    !***              = -r*<rot(M).gradtheta>_FS         ***
    !***              = r d/dr(\int M_phi dtheta dphi) / ***
    !***                     \int Js dtheta dphi         ***
    !***  because rot(M).gradtheta = -1/Js d/dr(M_phi)   ***
    !***   with Js is the jacobian in space and          ***
    !-> A(r) = d/dr (\int M_phi dtheta dphi)
    call deriv1(intMphi_dthetadphi,nvpolor_mag_diag, &
      geom%Nr,geom%dr,0)
    !-> nvpolor_mag(r) = r*A(r) / \int Js dtheta dphi
    do ir = 0,geom%Nr
      intdthetadphi_Js_tmp = intdthetadphi_Js(ir)
      !---> ATTENTION (no dependence of sqrt_g22 in theta ?)
      sqrt_g22             = sqrt(coord_sys%g22(ir,0))
      nvpolor_mag_diag(ir) = -sqrt_g22 * &
        nvpolor_mag_diag(ir)/intdthetadphi_Js_tmp
    end do
  end subroutine Physics_part_moments_postprocess
end module physics_part_module


