!**************************************************************
!  Copyright Euratom-CEA
!  Authors : 
!     Virginie Grandgirard (virginie.grandgirard@cea.fr)
!     Chantal Passeron (chantal.passeron@cea.fr)
!     Guillaume Latu (guillaume.latu@cea.fr)
!     Xavier Garbet (xavier.garbet@cea.fr)
!     Philippe Ghendrih (philippe.ghendrih@cea.fr)
!     Yanick Sarazin (yanick.sarazin@cea.fr)
!  
!  This code GYSELA (for GYrokinetic SEmi-LAgrangian) 
!  is a 5D gyrokinetic global full-f code for simulating 
!  the plasma turbulence in a tokamak.
!  
!  This software is governed by the CeCILL-B license 
!  under French law and abiding by the rules of distribution 
!  of free software.  You can  use, modify and redistribute 
!  the software under the terms of the CeCILL-B license as 
!  circulated by CEA, CNRS and INRIA at the following URL
!  "http://www.cecill.info". 
!**************************************************************

#include "config-F.inc"

!-----------------------------------------------------
! file : resu3D_saving.F90
! date : 25/09/2008
!  Saving of the 3D results in HDF5 format, like :
!   - the electrostatic potential Phi(r,theta,phi) 
!   - the fluid moments : J0M0(r,theta,phi) and
!       J0M2(r,theta,phi)
!  Rk : The 3D arrays are supposed parallelized 
!       and each processor saves its local part
!-----------------------------------------------------
module resu3D_saving_module
  use prec_const
#ifndef NO_HDF5
  use HDF5
  use HDF5_module
#endif

  use MpiMod
  implicit none

  !----------------------------------------------------------------------------!
  ! Data structure that is used to store internal information of the LCHD      !
  ! library to save 3D results as multiple files.                              !
  !----------------------------------------------------------------------------!
#ifdef LCHD    
  include "mfdiag_inc.h" 
  integer, parameter :: FNAME_LENGTH = 16
  type :: mfdstruct
     integer, dimension(:)  , pointer :: mom_id ! Request id , dims 0:Nquant-1
     integer, dimension(3)  :: mom_tot_size, mom_loc_size, mom_tot_size_rp1
     integer, dimension(:,:), pointer :: Mranks   !0:Nquant-1,0:Nbfiles-1
     integer, dimension(:)  , pointer :: nbdiag  !0:Nquant-1
     integer :: Nbfiles, Nquant, Nquantmax
     character(FNAME_LENGTH), dimension(:), pointer :: fnames !0:Nquant-1
     integer :: situ0post1 = 1  ! Compression location: 0-In situ compression, 1- Compression as a post-processing step or no compression
     integer(MFD_POINTER) mfdobj
  end type mfdstruct
#else
  type :: mfdstruct
     integer :: fakeint
  end type mfdstruct
#endif ! LCHD

  !******************************
  contains
  !******************************
    
    !----------------------------------------------------- 
    ! create the name of the HDF5 file on each processor 
    !   name+"_d<num_diag>"+"_p<proc_diag>".h5
    ! ( used for 3D storage)
    !-----------------------------------------------------
    function create_filename3D(name,diag_num,proc_num)
      use globals, only : numfmt
      character(LEN=*), intent(in)           :: name
      integer         , intent(in)           :: diag_num
      integer         , intent(in), optional :: proc_num

      character(LEN=50) :: create_filename3D
      character(LEN=30) :: diag_name, proc_name    

      write(diag_name,'('//numfmt//')') diag_num
      if (.not.present(proc_num)) then
         create_filename3D = trim(name)//&
              trim(diag_name)//".h5"
      else
         proc_name  = "_p    "
         write(proc_name(3:6),'(i4.4)') proc_num
         create_filename3D = trim(name)//trim(diag_name)&
              //trim(proc_name)//".h5"
      end if
    end function create_filename3D

    !------------------------------------------------------------- 
    ! Saves 3D results associated to the diagnostic "var3D_name"
    ! (where the 3D array resu3D is distributed on several processors)
    !-------------------------------------------------------------
    subroutine HDF5_parallel3D_saving(var3D_name,Snb_diag,Stime_diag,&
         geom,resu3D)
      use globals
      use MPIutils_module
      use geometry_class
      character*(*)                , intent(in) :: var3D_name
      integer                      , intent(in) :: Snb_diag
      real(RKIND)                  , intent(in) :: Stime_diag
      type(geometry)               , intent(in) :: geom
      real(RKIND), dimension(:,:,:), pointer    :: resu3D ! intent(in)
      ! Local variable
      real(RKIND), dimension(:,:,:), pointer    :: glob3D
#ifndef NO_HDF5
      integer(HID_T)    :: file_id   ! file identifier
      character(LEN=70) :: var3D_name_tmp, file_tmp
      character(LEN=70) :: var3D_file_name, file_name
      character(LEN=3)  :: num_species
      integer           :: len_varname, error, ierr
      integer           :: nbelts, ir, itheta, iphi, Nrp1, Nthp1

      if (iphistart .ne. -1) then
         Nthp1  = geom%Ntheta+1
         Nrp1   = geom%Nr+1
         nbelts = (geom%Nphi+1) * (geom%Nr+1) * (geom%Ntheta+1)
         allocate(glob3D(0:geom%Nr,0:geom%Ntheta,0:geom%Nphi))
         glob3D = 0._RKIND
         nbelts = dom_mapphi * (geom%Nr+1) * (geom%Ntheta+1)

         call MPI_GATHER(resu3D(0,0,iphistart),nbelts,MPI_REAL8, &
              glob3D(0,0,0),nbelts,MPI_REAL8,0,mpi_comm_moments,ierr)

         !*** periodic boundary conditions ***
         do itheta = 0, geom%Ntheta
            do ir = 0, geom%Nr
               glob3D(ir, itheta, geom%Nphi) = glob3D(ir, itheta, 0)
            end do
         end do
         do iphi = 0, geom%Nphi
            do ir = 0, geom%Nr
               glob3D(ir, geom%Ntheta, iphi) = glob3D(ir, 0, iphi)
            end do
         end do
         if (pInSpecies_id .eq. 0) then
            var3D_name_tmp = trim(var3D_name)//'_3D'
            !*** saving of the local part of resu3D ***
            write(num_species,'(A2,I1)') "sp", pSpecies_id+1
            if (pSpecies_id .eq. 0) then
               var3D_file_name = trim(var3D_name)//'_3D'
            else
               var3D_file_name = trim(num_species)//"/"//trim(var3D_name)//'_3D'
            end if
            file_name = create_filename3D(var3D_file_name,Snb_diag)
            call HDF5_create(trim(file_name),file_id,error)
            if (error .ne. 0) then
               print *, "error in creating ", file_name
            else
               call HDF5_save(file_id,Stime_diag,"time_diag")
               call HDF5_save(file_id,geom%Nr,"Nr")
               call HDF5_save(file_id,geom%Ntheta,"Ntheta")
               call HDF5_save(file_id,geom%Nphi,"Nphi")
               call HDF5_save(file_id,R0,"R0")
               call HDF5_save(file_id,geom%rg(0:),geom%Nr+1,"rg")
               call HDF5_save(file_id,geom%thetag(0:), &
                    geom%Ntheta+1,"thetag")
               call HDF5_save(file_id,geom%phig(0:), &
                    geom%Nphi+1,"phig")
               call HDF5_save(file_id,glob3D,geom%Nr+1, &
                    geom%Ntheta+1,geom%Nphi+1,trim(var3D_name_tmp))
               call HDF5_close(file_id)
            end if
         end if
         !*** Array deallocation ***
         deallocate(glob3D)
      end if
#endif
    end subroutine HDF5_parallel3D_saving


    !---------------------------------------------------------------------!
    ! Initialize the LCHD library to save 3D results as multiple files    !
    ! and to avoid large memory consumption on the export node.           !
    ! Parameters:                                                         !
    ! mfds: this data struct. will be filled/used internally by LCHD lib. !
    ! geom: data struct. geometry                                         !
    ! nquant: the number of 3D fields that we want to store on disks now  !
    !---------------------------------------------------------------------!
    subroutine HDF5_LCHD_init(mfds,geom,nquant)
      use globals
      use geometry_class
      use MPIutils_module
#ifdef LCHD    
      include "mfdiag_inc.h" 
#endif
      type(mfdstruct), intent(inout) :: mfds
      type(geometry),  intent(in)    :: geom
      integer       ,  intent(in)    :: nquant

#ifdef LCHD    
      integer :: nbfiles         ! # of files that will be generated for each 3D field
      integer :: npf             ! # of nodes that send info to each writing node
      integer :: Nbcompn         ! # of computing nodes
      integer :: iMranks, iq, in ! Counters
      integer :: distn           ! Distance among writing nodes
      real(8) :: tw1, tw2        ! Time before writing, time after writing
      real(8) :: twmax           ! Max writing time
      integer :: Nbfilest        ! # of files in total, for all the quantities
      ! Mapping function: 0: computing node, 1: non-computing nodes   
      integer, dimension(0:Nbproc_loc-1) :: freenodes 
      integer :: mfd_err         ! Error flag
      integer :: ifreen, Nbfreen, idistn
      integer :: nbelts, ir, itheta, iphi, Nrp1, Nthp1, ierr, i, j

      Nbcompn  = min(geom%nphi,nbproc_loc)-1
      ! nbfiles: Number of files to be created for each 3D quantity".
      ! This number is chosen to be less than the number Nphi and Nbproc_loc
      nbfiles  = max((Nbcompn+1)/32, 1)
      Npf      = (Nbcompn+1)/Nbfiles
      Nbfilest = Nquant*Nbfiles

      ! Allocation of data that store mapping and communication stuff 
      allocate(mfds%mom_id(0:Nquant-1))
      allocate(mfds%Mranks(0:Nquant-1,0:Nbfiles-1))
      allocate(mfds%fnames(0:Nquant-1))
      allocate(mfds%nbdiag(0:Nquant-1))
      do i = 0, Nquant-1
         do j = 1, FNAME_LENGTH
            mfds%fnames(i)(j:j) = char(0)
         end do
      end do

      !------------------
      ! NODE ASSIGNATION 
      !------------------
      ! The first possible option will be taken:
      ! 1) Enough non-computing nodes for exporting all the files from all 
      !    the variables simultaneously
      ! 2) Enough nodes for exporting all the files from all the variables 
      !    simultaneously
      ! 3) Files will be exported sequentially using first nodes of the 
      !    MPI communicator
      ! -------------------------------------------------------------------
      ! 1) Enough non-computing nodes for exporting all simultaneously
      if (Nbproc_loc.gt.(Nbcompn+Nbfilest)) then
         freenodes = 1
         do iMranks = 0, Nbcompn, 1
            freenodes(moments_mapphi(iMranks)) = 0
         enddo
         iq = 0
         iMranks = 0
         do in = 0, Nbproc_loc-1, 1
            if (freenodes(in).eq.1) then
               mfds%Mranks(iq,iMranks) = in
               if (iMranks.lt.(Nbfiles-1)) then
                  iMranks = iMranks+1
               else
                  iMranks = 0
                  if (iq.lt.(Nquant-1)) then
                     iq = iq+1
                  else
                     exit
                  endif
               endif
            endif
         enddo
         ! 2) Enough nodes for exporting all the files from compute nodes
      elseif (Nbproc_loc.ge.Nbfilest) then 
         iq = 0
         iMranks = 0
         do in = 0, Nbproc_loc-1, 1
            mfds%Mranks(iq,iMranks) = in
            if (iMranks.lt.(Nbfiles-1)) then
               iMranks = iMranks+1
            else
               iMranks = 0
               if (iq.lt.(Nquant-1)) then
                  iq = iq+1
               else
                  exit 
               endif
            endif
         enddo
         ! 3) Files will be exported sequentially using first nodes 
      else
         do iMranks = 0, Nbfiles-1, 1
            do iq = 0, Nquant-1, 1
               mfds%Mranks(iq, iMranks)= iMranks*npf
               if (mfds%Mranks(iq, iMranks) .ge. Nbproc_loc) then
                  write(*, "(A)") "Internal error, check resu3D_saving.F90"
                  STOP
               end if
            enddo
         enddo
      endif

      ! Size of the global 3D data structures
      mfds%mom_tot_size(1)     = geom%Nr
      mfds%mom_tot_size(2)     = geom%Ntheta
      mfds%mom_tot_size(3)     = geom%Nphi
      ! Size of the local 3D data structures to write on disk
      mfds%mom_tot_size_rp1(1) = geom%Nr+1
      mfds%mom_tot_size_rp1(2) = geom%Ntheta
      mfds%mom_tot_size_rp1(3) = geom%Nphi/Nbfiles
      ! Size of the local 3D data structures that are stored in each MPI process
      mfds%mom_loc_size(1)     = geom%Nr
      mfds%mom_loc_size(2)     = geom%Ntheta
      mfds%mom_loc_size(3)     = dom_mapphi
      ! Number of files per 3D data structure that will be exported
      mfds%nbfiles             = nbfiles
      ! Number of quantities (number of different 3D data structures)
      mfds%nquantmax           = nquant
      ! Current number of quantities recorded
      mfds%nquant              = 0
      call mfd_start(mfds%mfdobj, mfd_err)
#endif
    end subroutine HDF5_LCHD_init
    
    !---------------------------------------------------------------------!
    ! Start communications of the LCHD library to gather 3D quantities    !
    ! on a set of MPI process (#of MPI process = #files * #quantities)    !
    ! Parameters:                                                         !
    ! mfds: this data struct. will be filled/used internally by LCHD lib. !
    ! var3D_name: string defining the prefix used for the file name       !
    ! nbdiag: iteration number (part of the suffix of the file name)      !
    ! resu3D: 3D fields to export (domain decomp along \varphi, 3rd dim)  !
    !---------------------------------------------------------------------!
    subroutine HDF5_LCHD_start(mfds,var3D_name,nbdiag,resu3D)
      use globals
      use MPIutils_module
      use geometry_class       
#ifdef LCHD    
      include "mfdiag_inc.h" 
#endif

      type(mfdstruct),            intent(inout) :: mfds
      character*(*),                 intent(in) :: var3D_name
      integer,                       intent(in) :: nbdiag
      real(RKIND), dimension(:,:,:), pointer    :: resu3D ! intent(in)

#ifdef LCHD    
      integer :: mfd_err
      integer :: index_s3D

      index_s3D = mfds%nquant
      if (index_s3D .ge. mfds%nquantmax) then
         write(*, "(A)") "Internal error, check resu3D_saving.F90"
         STOP
      end if
      mfds%nbdiag(index_s3D) = nbdiag
      mfds%fnames(index_s3D) = trim(adjustl(var3D_name))//char(0)

      ! Create a new request in LCHD lib.(new quantity to store)
      call mfd_create_request(mfds%mfdobj, mfds%mom_id(index_s3D), mfd_err)

      ! Set up the MPI processes that will participate to communications
      ! and afterwards, the storing of files
      call mfd_commpernode(mfds%mom_tot_size, mfds%mfdobj, mfds%mom_id(index_s3D), &
           moments_mapphi, mfds%Mranks(index_s3D,:), &
           mfds%nbfiles, mpi_comm_species, mfd_err)

      ! Start up communications (non-bloquant communications, we have
      ! to wait for termination in another step: mf_array_wait)
      call mfd_array_comm(var3D_name, resu3D, mfds%mom_tot_size, 3, mfds%mfdobj, &
           mfds%mom_id(index_s3D), moments_mapphi, mfds%mom_loc_size, &
           mfds%Mranks(0,:), mfds%nbfiles, mpi_comm_species, mfd_err, nbdiag, &
           mfds%situ0post1)

      mfds%nquant = mfds%nquant + 1
#endif /* LCHD */

    end subroutine HDF5_LCHD_start


    !---------------------------------------------------------------------!
    ! Wait for all pending communications of the LCHD library and also    !
    ! write files in parallel.                                            !
    ! Parameters:                                                         !
    ! mfds: this data struct. will be filled/used internally by LCHD lib. !
    ! timestamp: simulation physics time, to be written in the files      !
    ! nbdiag: iteration number (part of the suffix of the file name)      !
    ! resu3D: 3D fields to export (domain decomp along \varphi, 3rd dim)  !
    ! geom: data struct. geometry                                         !
    !---------------------------------------------------------------------!
    subroutine HDF5_LCHD_waitall_and_write(mfds,timestamp,geom)
      use globals
      use MPIutils_module
      use geometry_class       
#ifdef LCHD    
      include "mfdiag_inc.h" 
#endif

      type(mfdstruct),            intent(inout) :: mfds
      real(RKIND),                   intent(in) :: timestamp
      type(geometry),                intent(in) :: geom

#ifdef LCHD    
      integer mfd_err
      integer index_s3D

      do index_s3D = 0, mfds%nquant-1
         if (plocal_id == mfds%Mranks(index_s3D,0)) then
            call mfd_doublesave_ds("phig", geom%phig, geom%Nphi+1, mfds%mfdobj, mfds%mom_id(index_s3D), mfd_err)
            call mfd_doublesave_ds("thetag", geom%thetag, geom%Ntheta+1, mfds%mfdobj, mfds%mom_id(index_s3D), mfd_err)
            call mfd_doublesave_ds("rg", geom%rg, geom%Nr+1, mfds%mfdobj, mfds%mom_id(index_s3D), mfd_err)
            call mfd_doublesave_ds("time_diag", timestamp, 1, mfds%mfdobj, mfds%mom_id(index_s3D), mfd_err)
            call mfd_doublesave_ds("R0", R0, 1, mfds%mfdobj, mfds%mom_id(index_s3D), mfd_err)
            call mfd_intsave_ds("Nr", geom%Nr, 1, mfds%mfdobj, mfds%mom_id(index_s3D), mfd_err)
            call mfd_intsave_ds("Ntheta", geom%Ntheta, 1, mfds%mfdobj, mfds%mom_id(index_s3D), mfd_err)
            call mfd_intsave_ds("Nphi", geom%Nphi, 1, mfds%mfdobj, mfds%mom_id(index_s3D), mfd_err)
            call mfd_intsave_ds("Nbfiles", mfds%nbfiles, 1, mfds%mfdobj, mfds%mom_id(index_s3D), mfd_err)
         endif
         call mfd_array_wait(mfds%mfdobj, mfds%mom_id(index_s3D))
         call mfd_array_write(mfds%fnames(index_s3D), NULL(), mfds%mom_tot_size_rp1, 3, mfds%mfdobj, mfds%mom_id(index_s3D), &
              moments_mapphi, mfds%mom_loc_size, mfds%Mranks(0,:), mfds%nbfiles, mpi_comm_species, mfd_err, mfds%nbdiag(index_s3D), &
              mfds%situ0post1)
      end do
#endif

    end subroutine HDF5_LCHD_waitall_and_write

    !---------------------------------------------------------------------!
    ! End of all file writings of the LCHD library and close files.            !
    ! mfds: this data struct. will be filled/used internally by LCHD lib. !
    !---------------------------------------------------------------------!
    subroutine HDF5_LCHD_close(mfds)
#ifdef LCHD    
      include "mfdiag_inc.h" 
#endif
      type(mfdstruct), intent(inout) :: mfds
#ifdef LCHD    
      integer :: mfd_err         ! Error flag
      deallocate(mfds%mom_id)
      deallocate(mfds%Mranks)
      deallocate(mfds%fnames)
      deallocate(mfds%nbdiag)

      call mfd_end(mfds%mfdobj, mfd_err)    
#endif
    end subroutine HDF5_LCHD_close

end module resu3D_saving_module
