!**************************************************************
!  Copyright Euratom-CEA
!  Authors : 
!     Virginie Grandgirard (virginie.grandgirard@cea.fr)
!     Chantal Passeron (chantal.passeron@cea.fr)
!     Guillaume Latu (guillaume.latu@cea.fr)
!     Xavier Garbet (xavier.garbet@cea.fr)
!     Philippe Ghendrih (philippe.ghendrih@cea.fr)
!     Yanick Sarazin (yanick.sarazin@cea.fr)
!  
!  This code GYSELA (for GYrokinetic SEmi-LAgrangian) 
!  is a 5D gyrokinetic global full-f code for simulating 
!  the plasma turbulence in a tokamak.
!  
!  This software is governed by the CeCILL-B license 
!  under French law and abiding by the rules of distribution 
!  of free software.  You can  use, modify and redistribute 
!  the software under the terms of the CeCILL-B license as 
!  circulated by CEA, CNRS and INRIA at the following URL
!  "http://www.cecill.info". 
!**************************************************************

#include "config-F.inc"

!------------------------------------------------------------
! file : spectra.F90
! date : 05/13/2011
!--------------------------------------------------------------
module spectra_module
  use prec_const
  use OMPutils_module, only : omp1dvector, omp1cvector, omp2cvector
  use gyroaverage_class
  use geometry_class
  use init_profile_class
  use init_magnetic_class
  use fdistribu5d_class
  use fft_module
  use wrap_alloc_module
  use fft_NRF90_module
  use integration_module
  use clock_module
  use MPIutils_module
  use poisson_class
  
  implicit none

  integer     :: ndiags_glob, nb_fulldiag
  !--> Boolean for computing the spectral diagnostic for 
  !--> a mpi proc, discrimination in the radial direction
  integer, dimension(:)  , pointer  :: cumul_r
  integer, dimension(:,:), pointer  :: ir_map
  integer                           :: pid_computing

  private
  integer :: nb_terms 
  !--> Temporary arrays
  real(RKIND), dimension(:,:,:), pointer  :: Rarray1_NdiagsNthetaNphi
  real(RKIND), dimension(:,:,:), pointer  :: Rarray2_NdiagsNthetaNphi
  complex(RKIND)                          :: vg_tmp, drfmp1, drfmm1, drfmp2
  complex(RKIND), dimension(:)  , pointer :: vpartmp1,vpartmp2,vpartmp3
  complex(RKIND), dimension(:,:), pointer :: dvparFmm1,dvparFmp1 
  complex(RKIND), dimension(:,:,:), pointer :: dvparF_mn
  integer       , dimension(:)  , pointer :: pos_rdiags, i_rdiags
  real(RKIND)   , dimension(:,:), pointer :: F_fluxS_avg
  !--> FFT wave vector in theta and phi direction
  real(RKIND), dimension(:)  , pointer :: SP_kth, kth_output
  real(RKIND), dimension(:)  , pointer :: SP_kphi, kphi_output
  integer    , dimension(:)  , pointer :: reorg_n,reorg_m
  !--> Arrays for FFT2D
  type(omp1cvector), dimension(:), pointer :: omp_rhs
  type(omp2cvector), dimension(:), pointer :: omp_data_i
  type(omp2cvector), dimension(:), pointer :: omp_tmp
  !--> Arrays for transposition
  real(RKIND), dimension(:,:,:,:), pointer :: f4D_Sdiag_transp

  public :: init_spectra_diag, del_spectra_diag, diag_spectra, ndiags_glob

  !******************************
  contains
  !******************************

  !---------------------------------------------------------------
  ! Constructor
  !---------------------------------------------------------------
  subroutine init_spectra_diag(geom)
    use globals
    use wrap_alloc_module

    type(geometry)     , intent(in)    :: geom
    integer :: i,ir,indx!loop indices
    integer :: N1, N2, N3, N2m1, N3m1
    integer :: tid, pid_r

    N1   = geom%Nr+1       !size(geom%rg)
    N2   = geom%Ntheta+1   !size(geom%thetag)
    N2m1 = N2-1
    N3   = geom%Nphi+1     !size(geom%phig)
    N3m1 = N3-1

    !*** Tables for communications between radial processors:
    !*** the radial proc that will receive all the radii to 
    !*** compute is labeled by pid_computing
    call FTR_ALLOC_1_ZERO(cumul_r,0,Nbproc_r-1)
    cumul_r(0:Nbproc_r-1) = 0
    ndiags_glob = 0
    pid_computing = Nbproc_r
    do i = 1, N_rdiags_Spectr
      if (r_diags(i) .ne. -1) then
        pid_r = r_diags(i)/dom_r
        pid_computing = min(pid_computing,pid_r)
        cumul_r(pid_r) = cumul_r(pid_r) + 1
      end if
    end do
    ndiags_glob = sum(cumul_r)

    call FTR_ALLOC_2_ZERO(ir_map,1,maxval(cumul_r(0:Nbproc_r-1)),0,Nbproc_r-1)
    ir_map(:,:) = -1
    cumul_r(0:Nbproc_r-1) = 0
    do i = 1, ndiags_glob
      pid_r = r_diags(i)/dom_r
      cumul_r(pid_r) = cumul_r(pid_r) + 1
      if (pid_r*dom_r.gt.(r_diags(i)-2)) then
        if (pglobal_id.eq.0) &
          print*,'WARNING: position r_diags ',r_diags(i), &
          'was too close to the boundary ', pid_r*dom_r, &
          ' ==> it has been moved to position = ', pid_r*dom_r+2
        r_diags(i) = pid_r*dom_r+2
      end if
      if ((r_diags(i)+2).ge.(pid_r+1)*dom_r) then
        if (pglobal_id.eq.0) &
          print*,'WARNING: position r_diags ',r_diags(i), &
          'was too close to the boundary', (pid_r+1)*dom_r-1, &
          ' ==> it has been moved to position = ', (pid_r+1)*dom_r-3
        r_diags(i) = (pid_r+1)*dom_r-3
      end if
      ir_map(cumul_r(pid_r),pid_r) = r_diags(i)
    end do

print*,pglobal_id,'==> ir ', istart,iend
if (pglobal_id.eq.0) &
  print*,pglobal_id,'==> r_diags', r_diags

    !*** pos_rdiags has the positions of the computed
    !*** radii, plus 2 points before and two points after
    !*** for the computation of radial derivatives
    call FTR_ALLOC_1_ZERO(pos_rdiags,1,ndiags_glob*5)
    !*** i_rdiags has the positions of the computed
    !*** radii in pos_rdiags
    call FTR_ALLOC_1_ZERO(i_rdiags,1,ndiags_glob)
    do i = 1,ndiags_glob
      pos_rdiags(5*(i-1)+1) = r_diags(i)-2
      pos_rdiags(5*(i-1)+2) = r_diags(i)-1
      pos_rdiags(5*(i-1)+3) = r_diags(i)
      pos_rdiags(5*(i-1)+4) = r_diags(i)+1
      pos_rdiags(5*(i-1)+5) = r_diags(i)+2
      i_rdiags(i)           = 5*(i-1)+3
    end do

    !*** find out the modes for which all the diags are
    !*** calculated
    nb_fulldiag = 0
    do i = 1, nb_fulldiag_max
      if ((doubleSP_m(i) .ne. -1).and.&
        (doubleSP_n(i) .ne. -1)) then
        nb_fulldiag = nb_fulldiag + 1
      end if
    end do

    nb_terms = 8 + nb_fulldiag*6

    !*** Temporary arrays for derivatives in vpar direction
    call FTR_ALLOC_1_ZERO(vpartmp1,0,geom%Nvpar)
    call FTR_ALLOC_1_ZERO(vpartmp2,0,geom%Nvpar)
    call FTR_ALLOC_1_ZERO(vpartmp3,0,geom%Nvpar)
    call FTR_ALLOC_2_ZERO(dvparFmm1,1,ndiags_glob,0,geom%Nvpar)
    call FTR_ALLOC_2_ZERO(dvparFmp1,1,ndiags_glob,0,geom%Nvpar)
    call FTR_ALLOC_3_ZERO(dvparF_mn,1,ndiags_glob,0,geom%Nvpar,1,nb_fulldiag)

    call FTR_ALLOC_2_ZERO(F_fluxS_avg,1,ndiags_glob,0,geom%Nvpar)

    call FTR_ALLOC_3_ZERO(Rarray1_NdiagsNthetaNphi,1,nb_terms*ndiags_glob,1,geom%Ntheta,1,geom%Nphi)
    call FTR_ALLOC_3_ZERO(Rarray2_NdiagsNthetaNphi,1,nb_terms*ndiags_glob,1,geom%Ntheta,1,geom%Nphi)

    call FTR_ALLOC_1_ZERO(SP_kth,1,geom%Ntheta)
    call FTR_ALLOC_1_ZERO(SP_kphi,1,geom%Nphi)
    call FTR_ALLOC_1_ZERO(kth_output,1,geom%Ntheta)
    call FTR_ALLOC_1_ZERO(kphi_output,1,geom%Nphi)
    call FTR_ALLOC_1_ZERO(reorg_m,1,geom%Ntheta)
    call FTR_ALLOC_1_ZERO(reorg_n,1,geom%Nphi)
    
    allocate(omp_rhs(1:Nbthread))

    call write_param("Nbthread", Nbthread)
    call write_param("N2m1", N2m1)
    call write_param("N3m1", N3m1)

    allocate(omp_tmp(1:Nbthread))
    allocate(omp_data_i(1:Nbthread))

    do tid = 1,Nbthread
      call FTR_ALLOC_1_ZERO(omp_rhs(tid)%val,1,N1)
    enddo

    !*** For transpong the distriubtion function
    call FTR_ALLOC_4_ZERO(f4D_Sdiag_transp,1,5*ndiags_glob,0,Ntheta,0,Nphi,0,Nvpar)
  end subroutine init_spectra_diag


  !---------------------------------------------------------------
  ! Destructor
  !---------------------------------------------------------------
  subroutine del_spectra_diag
    use globals
    use wrap_alloc_module
    implicit none

    integer :: tid

    call FTR_DEALLOC(ir_map)
    call FTR_DEALLOC(cumul_r)
    call FTR_DEALLOC(pos_rdiags)
    call FTR_DEALLOC(i_rdiags)
    call FTR_DEALLOC(vpartmp1)
    call FTR_DEALLOC(vpartmp2)
    call FTR_DEALLOC(vpartmp3)
    call FTR_DEALLOC(dvparFmm1)
    call FTR_DEALLOC(dvparFmp1)
    call FTR_DEALLOC(dvparF_mn)
    call FTR_DEALLOC(F_fluxS_avg)
    call FTR_DEALLOC(Rarray1_NdiagsNthetaNphi)
    call FTR_DEALLOC(Rarray2_NdiagsNthetaNphi)
    call FTR_DEALLOC(SP_kth)
    call FTR_DEALLOC(SP_kphi)
    call FTR_DEALLOC(kth_output)
    call FTR_DEALLOC(kphi_output)
    call FTR_DEALLOC(reorg_m)
    call FTR_DEALLOC(reorg_n)
    do tid = 1,Nbthread
      call FTR_DEALLOC(omp_rhs(tid)%val)
    enddo
    deallocate(omp_rhs)
    call FTR_DEALLOC(f4D_Sdiag_transp)
  end subroutine del_spectra_diag
  

  !---------------------------------------------------------------
  ! Compute the Fourier transforms of F and Phibar needed for the 
  ! spectral diagnostic
  !
  ! Ins     : geom, poiss, J0
  ! In/Outs : FFT_F, FFT_Phi
  ! Rmks    : - FFT_F contains the Fourier transform of the 
  !             transposed distribution function 
  !           - FFT_Phi contains the Fourier transform of J0Phi
  !           - The FFT is only performed on the useful radii
  !---------------------------------------------------------------
  subroutine compute_FFTs(geom, FFT_F, poiss, J0, FFT_Phi, pid_computing)
    use globals, only : Nbthread, Rarray1_NrNthetaNphi, &
      mu_id, r_diags, plocal_id, distrib3D, mpi_comm_mu, mpi_comm_moments 
#ifdef _OPENMP
    use OMPutils_module, only : omp_get_thread_num, &
         Comp1_1Nrp1_1Ntheta
#endif
    use mpiMod

    type(geometry)                        , intent(in)    :: geom
    type(poisson)                         , intent(in)    :: poiss
    type(J0operator)                      , intent(inout) :: J0
    complex(CKIND), dimension(:,:,:,:)    , pointer       :: FFT_F
    complex(CKIND), dimension(:,:,:)      , pointer       :: FFT_Phi

    !*** local variables ***   
    integer         :: i, j, k, l                    ! loop indices
    integer         :: ir, itheta, iphi              ! loop indices
    integer         :: N1, N2, N3, N4, N2m1, N3m1    ! domain dimensions
    integer         :: stopit               ! error flags
    integer         :: tid, isign
    integer         :: pid_computing
    real(RKIND), dimension(:,:,:), pointer :: gyroPhi

    !*** for distrib3D use ***
    integer         :: nbelts, ierr
    logical         :: ifgyro2D
    real(RKIND),dimension(:,:,:), pointer    :: phi3D_tmp
    real(RKIND),dimension(:,:,:), pointer    :: J0_intdv_PhiNrNthetaPNPhi
    real(RKIND),dimension(:,:), pointer      :: J0_intdv_Phirtheta
    real(RKIND),dimension(:,:,:), pointer    :: J0_intdv_PhiDrNthetaPNPhi
    real(RKIND),dimension(:,:,:), pointer    :: gyroPhi_Diagr
    complex(CKIND), dimension(:,:) , pointer :: Acomp
#ifdef TIMER
    integer(TIMEPREC) :: tdeb, tfin, tth1, tth2
    real(RKIND)       :: tbar(12), tcmp(12), tcom(12), tthd
    call clck_time(tdeb)
    tcmp(1:4) = 0.
    tcom(1:4) = 0.
#endif

    !*** initialize local variables ***
    N2   = geom%Ntheta+1
    N3   = geom%Nphi+1
    N4   = size(geom%vparg)
    N2m1 = N2-1
    N3m1 = N3-1

#ifdef TIMER
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tcmp(1))
    tdeb = tfin
#endif

    do tid = 1, Nbthread
       call FTR_ALLOC_2(J0%omp_diag(tid)%val,1,J0%N1,1,J0%N2)
       call FTR_ALLOC_2(J0%omp_ldiag(tid)%val,1,J0%N1-1,1,J0%N2)
       call FTR_ALLOC_2(J0%omp_udiag(tid)%val,1,J0%N1-1,1,J0%N2)
    enddo

!**** Here we need to make the 2d gyroaverage of the 
!**** local Phi(ir=block,itheta=*,iphi=block)
!**** Then we would need to send and receive the whole J0Phi 
!**** function for phi(ir=rdiags*5,itheta=*,iphi=*)
!**** then we can do the FFTs.

    !**** test parameters
    ifgyro2D = .true. !.true to compute not distrib3D case with gyro2D
    !**** If we use distrib3D, we need to gather the info on Phi
    if (distrib3D) then
       !*** we compute the 2D gyroaverage locally before gathering 
       !*** all phi on one proc
       tid=1
       call FTR_ALLOC_2_ZERO(Comp1_1Nrp1_1Ntheta(tid)%val, 1, Nr+1, 1, Ntheta)
       Acomp => Comp1_1Nrp1_1Ntheta(tid)%val
       !*** Allocation of temporary variable for J0_Phi(r,th,phi=iphi)
       !*** And J0_Phi(rdiag,th,phi=block)
       if (ilocstart .ne. -1) then
          call FTR_ALLOC_3(J0_intdv_PhiDrNthetaPNPhi,1,5*ndiags_glob,0,geom%Ntheta,ilocstart,ilocstart+dom_locphi-1)
          call FTR_ALLOC_2(J0_intdv_Phirtheta,1,geom%Nr+1,1,geom%Ntheta+1)
          !*** Potential is copied into an 2D array for gyroaverage
          do iphi = ilocstart, ilocstart+dom_locphi-1
             do ir=0,geom%Nr
                do itheta=0,geom%Ntheta
                   J0_intdv_Phirtheta(ir+1,itheta+1)= &
                        poiss%Phi_NrNthPNphi_Cmu(ir,itheta,iphi)
                end do
             end do
             !*** We compute 2D gyroaverage
             call omp_compute_gyrofunction_2D(J0,geom,mu_id, &
                  Acomp,J0_intdv_Phirtheta,tid)
             !*** result is copied into an array of size(rdiag,Ntheta,Nphi)
             do ir=1,5*ndiags_glob
                do itheta=0,geom%Ntheta
                   J0_intdv_PhiDrNthetaPNPhi(ir,itheta,iphi)= &
                        J0_intdv_Phirtheta(pos_rdiags(ir)+1,itheta+1)
                end do
             end do
          end do
       end if
       !*** Now J0_Phi is in J0_intdv_PhiNrNthetaPNPhi(r,th,phi=block)
       call FTR_DEALLOC(J0_intdv_Phirtheta)
       !*** Todo : choose spectral diag radii only
       call FTR_ALLOC_3_ZERO(gyroPhi_Diagr,1,5*ndiags_glob,0,geom%Ntheta,0,geom%Nphi-1)
       if(ilocstart .ne.-1) then
          !*** test with ilocstart et Cmu
          nbelts = dom_locphi * (5*ndiags_glob) * (geom%Ntheta+1)
          call MPI_GATHER(J0_intdv_PhiDrNthetaPNPhi(1,0,ilocstart), &
               nbelts,MPI_REAL8,gyroPhi_Diagr(1,0,0),nbelts,MPI_REAL8, &
               pid_computing,mpi_comm_mu,ierr)
       end if
       call FTR_DEALLOC(Comp1_1Nrp1_1Ntheta(tid)%val)
       call FTR_DEALLOC(J0_intdv_PhiDrNthetaPNPhi)
       !*** Now all J0_Phi(5*rdiags) is in gyroPhi_Diagr(5*rdiags,th,phi)
    else
       !*** If distrib3D is set to false, 3D gyroaverage is computed
       !*** If distrib3D is set to true, Rarray1_NrNthetaNphi is empty
       if (plocal_id .eq. pid_computing) then
          !*** poiss%Phi_NrNthPNphi_Cworld start with ir=0 
          ! gyroPhi => Rarray1_NrNthetaNphi
          call FTR_ALLOC_3(gyroPhi,0,geom%Nr,0,geom%Ntheta,0,geom%Nphi)
          if (.not. ifgyro2D) then !*** original gyro3D
             call compute_gyrofunction_3D(J0,geom,mu_id, &
                  poiss%Phi_NrNthPNphi_Cworld,gyroPhi)
          else
             !*** check 3d gyroaverage with phi loop 2D gyroaverage
             call FTR_ALLOC_2(J0_intdv_Phirtheta,1,geom%Nr+1,1,geom%Ntheta+1)
             tid=1
             call FTR_ALLOC_2_ZERO(Comp1_1Nrp1_1Ntheta(tid)%val, 1, Nr+1, 1, Ntheta)
             Acomp => Comp1_1Nrp1_1Ntheta(tid)%val
             !*** Potential is copied into an 2D array for gyroaverage
             do iphi = 0,geom%Nphi-1
                do ir=0,geom%Nr
                   do itheta=0,geom%Ntheta
                      J0_intdv_Phirtheta(ir+1,itheta+1)= &
                           poiss%Phi_NrNthPNphi_Cworld(ir,itheta,iphi)
                   end do
                end do
                !*** We compute 2D gyroaverage
                call omp_compute_gyrofunction_2D(J0,geom,mu_id, &
                     Acomp,J0_intdv_Phirtheta,tid)
                !*** result is copied into an array of size(rdiag,Ntheta,Nphi)
                do ir=0,geom%Nr
                   do itheta=0,geom%Ntheta
                      gyroPhi(ir,itheta,iphi)= &
                           J0_intdv_Phirtheta(ir+1,itheta+1)
                   end do
                end do
             end do
             call FTR_DEALLOC(Comp1_1Nrp1_1Ntheta(tid)%val)
             call FTR_DEALLOC(J0_intdv_Phirtheta)
          end if
       end if
       !**** from here on, gyroPhi contains the info for all radii -> to change
    end if

    do tid = 1, Nbthread
       call FTR_DEALLOC(J0%omp_diag(tid)%val)
       call FTR_DEALLOC(J0%omp_ldiag(tid)%val)
       call FTR_DEALLOC(J0%omp_udiag(tid)%val)
    enddo
         
    do tid = 1, Nbthread
       call FTR_ALLOC_2(omp_data_i(tid)%val,1,N2m1,1,N3m1)
       call FTR_ALLOC_2(omp_tmp(tid)%val,1,N3m1,1,N2m1)
    enddo

    !**** The remaining of the action is done 
    !**** by only one proc for each mu
    if (plocal_id .eq. pid_computing) then
       !*** perform FFT 2D in theta and phi  ***
       isign  = 1
       stopit = 0    
!$OMP PARALLEL private(tid,k,j,i,l) &
!$OMP firstprivate(isign) default(shared)
!$OMP BARRIER
#ifdef _OPENMP
       tid = 1+omp_get_thread_num()
#else
       tid = 1
#endif

!$OMP DO SCHEDULE(STATIC)
       !*** Compute the selected radii plus the 2 preceding and the 
       !*** 2 following ones for the computation of radial derivatives
       do i = 1,5*ndiags_glob
          do l = 0,N4-1
             do k = 1,N3m1
                do j = 1,N2m1
                   omp_data_i(tid)%val(j,k) = f4D_Sdiag_transp(i,j-1,k-1,l)
                end do
             end do
             call fourrow(omp_data_i(tid)%val,isign)
             do k = 1,N3m1
                do j = 1,N2m1
                   omp_tmp(tid)%val(k,j) = omp_data_i(tid)%val(j,k)
                end do
             end do
             call fourrow(omp_tmp(tid)%val,isign)
             do k = 1,N3m1
                do j = 1,N2m1
                   FFT_F(i,l,j,k) = omp_tmp(tid)%val(k,j)
                end do
             end do
          enddo
          
          if (distrib3D) then          
             do k = 1,N3m1
                do j = 1,N2m1
                   omp_data_i(tid)%val(j,k) = gyroPhi_Diagr(i,j-1,k-1)
                end do
             end do
          else !*** if not distrib3D
             do k = 1,N3m1
                do j = 1,N2m1
                   omp_data_i(tid)%val(j,k) = gyroPhi(pos_rdiags(i),j-1,k-1)
                end do
             end do
          end if
          call fourrow(omp_data_i(tid)%val,isign)
          do k = 1,N3m1
             do j = 1,N2m1
                omp_tmp(tid)%val(k,j) = omp_data_i(tid)%val(j,k)
             end do
          end do
          call fourrow(omp_tmp(tid)%val,isign)
          do k = 1,N3m1
             do j = 1,N2m1
                FFT_Phi(i,j,k) = omp_tmp(tid)%val(k,j)
             end do
          end do
       enddo
!$OMP END DO
!$OMP BARRIER
!$OMP END PARALLEL

       !*** DEAL WITH CONVENTION ISSUES WITH FOURROW
       !*** AND THE WAY I WROTE THE CALCULATIONS
       FFT_F   = conjg(FFT_F)/(N3m1*N2m1)
       FFT_Phi = conjg(FFT_Phi)/(N3m1*N2m1)
       
       do tid = 1, Nbthread
          call FTR_DEALLOC(omp_data_i(tid)%val)
          call FTR_DEALLOC(omp_tmp(tid)%val)
       enddo

       if (distrib3D) then
          call FTR_DEALLOC(gyroPhi_Diagr)
       else
          call FTR_DEALLOC(gyroPhi)
       end if
    end if   ! if (plocal_id .eq. pid_computing)

#ifdef TIMER
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tcmp(4))
    tdeb = tfin
    write(6,'(I4,A,4F13.6)') pglobal_id, &
      " Temps FFT2D on 4D array, tcomp ",tcmp(1:4)
    call clck_ldiff(tth1,tfin,tthd)
    write(6,'(I4,A,F13.6)') pglobal_id, &
      " Temps FFT2D on 4D array,  tcomp ",tthd
#endif
  end subroutine compute_FFTs

  
  !---------------------------------------------------------------
  ! Routine for spectra HDF5 file saving
  !---------------------------------------------------------------
  subroutine HDF5_spectra_saving(geom,resu,diag_num,Stime_diag)
    use globals, only : numfmt, r_diags, doubleSP_m, doubleSP_n
    use geometry_class
#ifndef NO_HDF5
    use HDF5
    use HDF5_module
#endif
    integer                      , intent(in) :: diag_num
    real(RKIND)                  , intent(in) :: Stime_diag
    type(geometry)               , intent(in) :: geom
    real(RKIND), dimension(:,:,:), pointer    :: resu
    integer                                   :: idiag_loc

#ifndef NO_HDF5
    character(LEN=50) :: file_name, diag_name
    integer(HID_T)    :: file_id   ! file identifier
    
    if (diag_num .ne.-1) then
      write (diag_name,'('//numfmt//')') diag_num
      file_name = "spectra"//trim(diag_name)//".h5"
    else
      file_name = "spectra_init.h5"
    end if

    call HDF5_create(trim(file_name),file_id)
    call HDF5_save(file_id,Stime_diag,"time_diag")
    call HDF5_save(file_id,ndiags_glob,"N_rdiags")
    call HDF5_save(file_id,geom%Ntheta,"Nktheta")
    call HDF5_save(file_id,geom%Nphi,"Nkphi")
    call HDF5_save(file_id,doubleSP_m(1:nb_fulldiag),&
      nb_fulldiag,"m_diag")
    call HDF5_save(file_id,doubleSP_n(1:nb_fulldiag),&
      nb_fulldiag,"n_diag")
    call HDF5_save(file_id,r_diags(1:ndiags_glob),&
      ndiags_glob,"r_diags")
    call HDF5_save(file_id,kth_output(1:geom%Ntheta),&
      geom%Ntheta,'ktheta')
    call HDF5_save(file_id,kphi_output(1:geom%Nphi),&
      geom%Nphi,'kphi')
    call HDF5_save(file_id,geom%coeff_intdvpar(0:geom%Nvpar),&
      geom%Nvpar+1,'coeff_intdvpar')
    call HDF5_save(file_id,geom%coeff_intdmu(0:geom%Nmu),&
      geom%Nmu+1,'coeff_intdmu')
    idiag_loc = 0
    call HDF5_save(file_id, &
      resu(idiag_loc*ndiags_glob+1:(idiag_loc+1)*ndiags_glob,:,:), &
      ndiags_glob, geom%Ntheta, geom%Nphi, "Entropy_v1")
    idiag_loc = idiag_loc+1
    call HDF5_save(file_id, &
      resu(idiag_loc*ndiags_glob+1:(idiag_loc+1)*ndiags_glob,:,:), &
      ndiags_glob,geom%Ntheta, geom%Nphi, "Entropy_v2")
    idiag_loc = idiag_loc+1
    call HDF5_save(file_id, &
      resu(idiag_loc*ndiags_glob+1:(idiag_loc+1)*ndiags_glob,:,:),&
      ndiags_glob,geom%Ntheta, geom%Nphi, "Entropy_v3")
    idiag_loc = idiag_loc+1
    call HDF5_save(file_id, &
      resu(idiag_loc*ndiags_glob+1:(idiag_loc+1)*ndiags_glob,:,:),&
      ndiags_glob,geom%Ntheta, geom%Nphi, "Epot")
    idiag_loc = idiag_loc+1
    call HDF5_save(file_id, &
      resu(idiag_loc*ndiags_glob+1:(idiag_loc+1)*ndiags_glob,:,:),&
      ndiags_glob,geom%Ntheta, geom%Nphi, "Curvature_Drift_mm1")
    idiag_loc = idiag_loc+1
    call HDF5_save(file_id, &
      resu(idiag_loc*ndiags_glob+1:(idiag_loc+1)*ndiags_glob,:,:),&
      ndiags_glob,geom%Ntheta, geom%Nphi, "Curvature_Drift_mp1")
    idiag_loc = idiag_loc+1
    call HDF5_save(file_id, &
      resu(idiag_loc*ndiags_glob+1:(idiag_loc+1)*ndiags_glob,:,:),&
      ndiags_glob,geom%Ntheta, geom%Nphi, "Linear_Velocity_mm1")
    idiag_loc = idiag_loc+1
    call HDF5_save(file_id, &
      resu(idiag_loc*ndiags_glob+1:(idiag_loc+1)*ndiags_glob,:,:),&
      ndiags_glob,geom%Ntheta, geom%Nphi, "Linear_Velocity_mp1")
    idiag_loc = idiag_loc+1
    call HDF5_save(file_id, &
      resu(idiag_loc*ndiags_glob+1:(idiag_loc+nb_fulldiag)*ndiags_glob,:,:),&
      ndiags_glob*nb_fulldiag,geom%Ntheta, geom%Nphi, "ExB_m")
    call HDF5_save(file_id, &
      resu((idiag_loc+nb_fulldiag)*ndiags_glob+1:(idiag_loc+2*nb_fulldiag)*ndiags_glob,:,:),&
      ndiags_glob*nb_fulldiag,geom%Ntheta, geom%Nphi, "ExB_mm1")
    call HDF5_save(file_id, &
      resu((idiag_loc+2*nb_fulldiag)*ndiags_glob+1:(idiag_loc+3*nb_fulldiag)*ndiags_glob,:,:),&
      ndiags_glob*nb_fulldiag,geom%Ntheta, geom%Nphi, "ExB_mp1")
    call HDF5_save(file_id, &
      resu((idiag_loc+3*nb_fulldiag)*ndiags_glob+1:(idiag_loc+4*nb_fulldiag)*ndiags_glob,:,:),&
      ndiags_glob*nb_fulldiag,geom%Ntheta, geom%Nphi, "GradParPhi")
    call HDF5_save(file_id, &
      resu((idiag_loc+4*nb_fulldiag)*ndiags_glob+1:(idiag_loc+5*nb_fulldiag)*ndiags_glob,:,:),&
      ndiags_glob*nb_fulldiag,geom%Ntheta, geom%Nphi, "VeGradB_mm1")
    call HDF5_save(file_id, &
      resu((idiag_loc+5*nb_fulldiag)*ndiags_glob+1:(idiag_loc+6*nb_fulldiag)*ndiags_glob,:,:),&
      ndiags_glob*nb_fulldiag,geom%Ntheta, geom%Nphi, "VeGradB_mp1")
    call HDF5_close(file_id)
#endif
  end subroutine HDF5_spectra_saving
  

  !---------------------------------------------------------------
  ! Main routine for the spectral diagnostic
  !---------------------------------------------------------------
  subroutine diag_spectra(geom,init_magnet,init_prof,f,poiss,J0, &
    Snb_diag,Stime_diag)
    use globals, only : Nbthread, istart_id, jstart_id, R0, mu_id, &
      r_diags, plocal_id, Nbproc_mu, distrib3D, &
      Carray_1Nrp1Nvpar1Ntheta1Nphi, Carray_1Nrp11Ntheta1Nphi, uout_res, &
      istart, iend, mpi_comm_intermu, doubleSP_m, doubleSP_n
    use fft_module, only : NR_fft_vector
    use utils_module, only : cderiv1
    use coord_system_class, only : jacobian_space, intdthetadphi_Js 
    use mpiMod
    implicit none

    type(geometry)     , intent(in)    :: geom
    type(init_magnetic), intent(in)    :: init_magnet
    type(init_profile) , intent(in)    :: init_prof
    type(fdistribu5d)  , intent(in)    :: f
    type(poisson)      , intent(in)    :: poiss
    type(J0operator)   , intent(inout) :: J0
    integer            , intent(in)    :: Snb_diag
    real(RKIND)        , intent(in)    :: Stime_diag

    !**** Local Variables
    complex(CKIND), dimension(:,:,:,:), pointer :: FFT_F
    complex(CKIND), dimension(:,:,:)  , pointer :: FFT_PhiBar
    real(RKIND)     :: coeff_intdvpar_tmp, coeff_intdmu_tmp, denom
    integer         :: itheta,iphi,ivpar,iphi_indx,ir,itheta_f_indx,iphi_f_indx
    integer         :: imode,idiag_loc
    integer         :: nbelements, ierr, mypid, itheta_p1, itheta_m1,m_nl_diag
    integer         :: n_nl_diag, itheta_out, iphi_out, iphi_out_tmp
    integer         :: itheta_out_transf, iphi_out_transf
    real(RKIND)     :: intdS_H_tmp, intdthetadphi_Js_tmp
    real(RKIND)     :: coeff_intdtheta, coeff_intdphi
    real(RKIND)     :: jacob_space_ij
    complex(CKIND)  :: FFT_F_tmp, FFT_F_Mp1_tmp, FFT_F_Mm1_tmp, &
      FFT_F_Mp1_rp1_tmp, FFT_F_Mp1_rp2_tmp, FFT_F_Mm1_rp1_tmp, &
      FFT_F_Mm1_rp2_tmp, FFT_F_Mp1_rm1_tmp, FFT_F_Mp1_rm2_tmp, &
      FFT_F_Mm1_rm1_tmp, FFT_F_Mm1_rm2_tmp, & 
      FFT_F_rp1_mn_tmp, &
      FFT_F_rp2_mn_tmp, FFT_F_rm1_mn_tmp, FFT_F_rm2_mn_tmp, &
      FFT_F_mn_tmp
    complex(CKIND)  :: FFT_P_tmp, FFT_P_rp1_tmp, FFT_P_rp2_tmp, &
      FFT_P_rm1_tmp, FFT_P_rm2_tmp, FFT_P_Mp1_tmp, FFT_P_Mm1_tmp, &
      FFT_P_Mp1_rp1_tmp, FFT_P_Mp1_rp2_tmp, FFT_P_Mm1_rp1_tmp, &
      FFT_P_Mm1_rp2_tmp, FFT_P_Mp1_rm1_tmp, FFT_P_Mp1_rm2_tmp, &
      FFT_P_Mm1_rm1_tmp, FFT_P_Mm1_rm2_tmp
    complex(CKIND)  :: VeGradB
    real(RKIND)     :: zeta_tmp, iota_tmp, rg_tmp, rg_p1_tmp, &
      rg_p2_tmp, rg_m1_tmp, rg_m2_tmp , vpar_tmp, mu_tmp, m_tmp,&
      n_tmp,coef_nl
    !**** For the Diagnostics
    real(RKIND)   , dimension(:,:,:), pointer   :: Diags
    real(RKIND)   , dimension(:,:,:), pointer   :: Output

    call clck_time(bclock_diagSP)

    !**** communication of f in theta and r ***
    call distribute_Sdiag(f%values,f4D_Sdiag_transp, &
      ndiags_glob, cumul_r, pos_rdiags, ir_map, pid_computing)

    !*** Begin Changes for distrib3D
    if (distrib3D) then
       !*** Version with distrib3D
       FFT_F   => Carray_1Nrp1Nvpar1Ntheta1Nphi
       FFT_PhiBar => Carray_1Nrp11Ntheta1Nphi !*** Size array : 1,5*ndiags_glob,1,Ntheta,1,Nphi
       call compute_FFTs(geom,FFT_F, poiss, J0, FFT_PhiBar, pid_computing)
       !**** After FFTs, all info is on one proc for each value of nu
    else
       !*** No distrib3D version
       !*** At this point, we only need one proc for each mu to continue
       if (plocal_id .eq. pid_computing) then
          
          !**** Pointers for Fourier transforms
          FFT_F   => Carray_1Nrp1Nvpar1Ntheta1Nphi 
          FFT_PhiBar => Carray_1Nrp11Ntheta1Nphi 
          
          !**** Fourier transforms
          call compute_FFTs(geom,FFT_F, poiss, J0, FFT_PhiBar, pid_computing)
      end if
      !**** After FFTs, all info is on one proc for each value of nu
    end if

    if (plocal_id .eq. pid_computing) then
      !*** End changes for distrib3D

      !*** compute kth for the FFT in theta direction ***
      call NR_fft_vector(geom%Ntheta,geom%dtheta,SP_kth)
      !*** compute kphi for the FFT in phi direction ***
      call NR_fft_vector(geom%Nphi,geom%dphi,SP_kphi)
      !*** compute the look-up tables for the output indexes
      !*** WARNING : the tables are written to reverse the 
      !*** organization made by fftw, if another fft is used
      !*** then you will have to modify that
      do iphi = 1,geom%Nphi          
        if (iphi .le. geom%Nphi/2+1) then
          reorg_n(iphi) = geom%Nphi/2 + iphi - 1
        else
          reorg_n(iphi) = -geom%Nphi/2 + iphi - 1
        end if
      end do
      do itheta = 1,geom%Ntheta
        if (itheta .le. geom%Ntheta/2+1) then
          reorg_m(itheta) = geom%Ntheta/2+ itheta - 1 
        else
          reorg_m(itheta) = -geom%Ntheta/2 + itheta - 1 
        end if
      end do
      !*** reorganize also the ktheta and kphi arrays
      !*** (only for the output files)
      do itheta = 1,geom%Ntheta
        kth_output(reorg_m(itheta))=SP_kth(itheta)
      end do
      do iphi = 1,geom%Nphi
        kphi_output(reorg_n(iphi))=nint(SP_kphi(iphi))
      end do

      !*** compute the flux surface average of F
      do ir = 1,ndiags_glob
        intdS_H_tmp          = 0._RKIND
        intdthetadphi_Js_tmp = intdthetadphi_Js(ir)
        do ivpar = 0, geom%Nvpar
          do iphi = 0,geom%Nphi
            coeff_intdphi =  geom%coeff_intdphi(iphi)
            do itheta = 0,geom%Ntheta
              jacob_space_ij  = jacobian_space(ir,itheta)
              coeff_intdtheta = geom%coeff_intdtheta(itheta)
              intdS_H_tmp     = intdS_H_tmp + &
                f4D_Sdiag_transp(i_rdiags(ir),itheta,iphi,ivpar)* &
                jacob_space_ij * &
                coeff_intdphi*coeff_intdtheta
            end do
          end do
          F_fluxS_avg(ir,ivpar) = intdS_H_tmp/intdthetadphi_Js_tmp
        end do
      end do

!DEB_TO_DEL
      F_fluxS_avg = 1._RKIND
!END_TO_DEL

      Diags => Rarray2_NdiagsNthetaNphi
      Diags = 0._RKIND

      !**** Computations 
      !*** Perform integration in vpar and prepare mu integration
      mu_tmp      = f%mu_value
      do iphi = 1,geom%Nphi
        n_tmp = SP_kphi(iphi)
        iphi_out = reorg_n(iphi)
        do itheta = 1,geom%Ntheta
          itheta_out = reorg_m(itheta)
          m_tmp       = SP_kth(itheta)
          itheta_p1   = minloc(abs(m_tmp+1._RKIND-SP_kth),dim=1)            
          itheta_m1   = minloc(abs(m_tmp-1._RKIND-SP_kth),dim=1)

          ! Compute the derivative with respect to vpar
          do ir = 1,ndiags_glob
            ! For linear parallel velocity advection
            if (SP_kth(itheta_m1).ne.m_tmp-1) then
              vpartmp1(0:geom%Nvpar) = 0._RKIND
            else
              vpartmp1(0:geom%Nvpar) = -FFT_F(i_rdiags(ir),0:geom%Nvpar,itheta_m1,iphi)
            end if
            if (SP_kth(itheta_p1).ne.m_tmp+1) then
              vpartmp2(0:geom%Nvpar) = 0._RKIND              
            else
              vpartmp2(0:geom%Nvpar) = FFT_F(i_rdiags(ir),0:geom%Nvpar,itheta_p1,iphi)
            end if
            call cderiv1(vpartmp1,vpartmp3,geom%Nvpar, &
              abs(geom%vparg(2)-geom%vparg(1)),0)
            dvparFmm1(ir,0:geom%Nvpar) = vpartmp3
            call cderiv1(vpartmp2,vpartmp3,geom%Nvpar, &
              abs(geom%vparg(2)-geom%vparg(1)),0)
            dvparFmp1(ir,0:geom%Nvpar) = vpartmp3

            ! For (m,n) NL parallel velocity terms
            do imode = 1, nb_fulldiag
              itheta_f_indx = minloc(abs(doubleSP_m(imode)-m_tmp-SP_kth),dim=1)
              iphi_f_indx   = minloc(abs(doubleSP_n(imode)-n_tmp-SP_kphi),dim=1)
              if ((SP_kth(itheta_f_indx).ne.doubleSP_m(imode)-m_tmp).or.&
                (nint(SP_kphi(iphi_f_indx)).ne.doubleSP_n(imode)-nint(n_tmp))) then
                dvparF_mn(ir,0:geom%Nvpar,imode) = 0._RKIND
              else
                vpartmp1(0:geom%Nvpar) = FFT_F(i_rdiags(ir),0:geom%Nvpar, &
                  itheta_f_indx,iphi_f_indx)
                call cderiv1(vpartmp1,vpartmp2,geom%Nvpar, &
                  abs(geom%vparg(2)-geom%vparg(1)),0)
                dvparF_mn(ir,0:geom%Nvpar,imode) = vpartmp2
              end if
            end do
          end do

          do ir = 1,ndiags_glob
            FFT_P_tmp         = FFT_PhiBar(i_rdiags(ir),itheta,iphi)
            FFT_P_rp1_tmp     = FFT_PhiBar(i_rdiags(ir)+1,itheta,iphi)
            FFT_P_rm1_tmp     = FFT_PhiBar(i_rdiags(ir)-1,itheta,iphi)
            FFT_P_rp2_tmp     = FFT_PhiBar(i_rdiags(ir)+2,itheta,iphi)
            FFT_P_rm2_tmp     = FFT_PhiBar(i_rdiags(ir)-2,itheta,iphi)

            do ivpar = 0,geom%Nvpar
              coeff_intdvpar_tmp = geom%coeff_intdvpar(ivpar)

              ! Compute temp parameters to make the code easier to read
              FFT_F_tmp = FFT_F(i_rdiags(ir),ivpar,itheta,iphi)

              if (SP_kth(itheta_p1).ne.m_tmp+1._RKIND) then
                FFT_F_Mp1_tmp     = 0._RKIND
                FFT_F_Mp1_rp1_tmp = 0._RKIND
                FFT_F_Mp1_rp2_tmp = 0._RKIND
                FFT_F_Mp1_rm1_tmp = 0._RKIND
                FFT_F_Mp1_rm2_tmp = 0._RKIND
                FFT_P_Mp1_tmp     = 0._RKIND
                FFT_P_Mp1_rp2_tmp = 0._RKIND
                FFT_P_Mp1_rp1_tmp = 0._RKIND
                FFT_P_Mp1_rm1_tmp = 0._RKIND
                FFT_P_Mp1_rm2_tmp = 0._RKIND
              else
                FFT_F_Mp1_tmp     = FFT_F(i_rdiags(ir),ivpar,itheta_p1,iphi)
                FFT_F_Mp1_rp1_tmp = FFT_F(i_rdiags(ir)+1,ivpar,itheta_p1,iphi)
                FFT_F_Mp1_rp2_tmp = FFT_F(i_rdiags(ir)+2,ivpar,itheta_p1,iphi)
                FFT_F_Mp1_rm1_tmp = FFT_F(i_rdiags(ir)-1,ivpar,itheta_p1,iphi)
                FFT_F_Mp1_rm2_tmp = FFT_F(i_rdiags(ir)-2,ivpar,itheta_p1,iphi)
                FFT_P_Mp1_tmp     = FFT_PhiBar(i_rdiags(ir),itheta_p1,iphi)
                FFT_P_Mp1_rp2_tmp = FFT_PhiBar(i_rdiags(ir)+2,itheta_p1,iphi)
                FFT_P_Mp1_rp1_tmp = FFT_PhiBar(i_rdiags(ir)+1,itheta_p1,iphi)
                FFT_P_Mp1_rm1_tmp = FFT_PhiBar(i_rdiags(ir)-1,itheta_p1,iphi)
                FFT_P_Mp1_rm2_tmp = FFT_PhiBar(i_rdiags(ir)-2,itheta_p1,iphi)
              end if
              if (SP_kth(itheta_m1).ne.m_tmp-1._RKIND) then
                FFT_F_Mm1_tmp     = 0._RKIND
                FFT_F_Mm1_rp1_tmp = 0._RKIND
                FFT_F_Mm1_rp2_tmp = 0._RKIND
                FFT_F_Mm1_rm1_tmp = 0._RKIND
                FFT_F_Mm1_rm2_tmp = 0._RKIND
                FFT_P_Mm1_tmp     = 0._RKIND
                FFT_P_Mm1_rp2_tmp = 0._RKIND
                FFT_P_Mm1_rp1_tmp = 0._RKIND
                FFT_P_Mm1_rm1_tmp = 0._RKIND
                FFT_P_Mm1_rm2_tmp = 0._RKIND
              else
                FFT_F_Mm1_tmp     = FFT_F(i_rdiags(ir),ivpar,itheta_m1,iphi)
                FFT_F_Mm1_rp1_tmp = FFT_F(i_rdiags(ir)+1,ivpar,itheta_m1,iphi)
                FFT_F_Mm1_rp2_tmp = FFT_F(i_rdiags(ir)+2,ivpar,itheta_m1,iphi)
                FFT_F_Mm1_rm1_tmp = FFT_F(i_rdiags(ir)-1,ivpar,itheta_m1,iphi)
                FFT_F_Mm1_rm2_tmp = FFT_F(i_rdiags(ir)-2,ivpar,itheta_m1,iphi)
                FFT_P_Mm1_tmp     = FFT_PhiBar(i_rdiags(ir),itheta_m1,iphi)
                FFT_P_Mm1_rp2_tmp = FFT_PhiBar(i_rdiags(ir)+2,itheta_m1,iphi)
                FFT_P_Mm1_rp1_tmp = FFT_PhiBar(i_rdiags(ir)+1,itheta_m1,iphi)
                FFT_P_Mm1_rm1_tmp = FFT_PhiBar(i_rdiags(ir)-1,itheta_m1,iphi)
                FFT_P_Mm1_rm2_tmp = FFT_PhiBar(i_rdiags(ir)-2,itheta_m1,iphi)
              end if

              zeta_tmp    = init_magnet%zeta(r_diags(ir))
              iota_tmp    = init_prof%iota(r_diags(ir))
              rg_tmp      = geom%rg(r_diags(ir))
              rg_p1_tmp   = geom%rg(r_diags(ir)+1)
              rg_m1_tmp   = geom%rg(r_diags(ir)-1)
              rg_p2_tmp   = geom%rg(r_diags(ir)+2)
              rg_m2_tmp   = geom%rg(r_diags(ir)-2)
              vpar_tmp    = geom%vparg(ivpar)

              !*** Compute the diagnostics now
              
              idiag_loc = 0
              ! Entropy with B = B0
              Diags(idiag_loc*ndiags_glob+ir,itheta_out,iphi_out) = &
                Diags(idiag_loc*ndiags_glob+ir,itheta_out,iphi_out) + &
                sqrt(1._RKIND+zeta_tmp**2) * coeff_intdvpar_tmp * &
                real(FFT_F_tmp * conjg(FFT_F_tmp)) / &
                (2._RKIND*F_fluxS_avg(ir,ivpar))
              
              idiag_loc = idiag_loc+1
              ! Entropy with B on one f
              Diags(idiag_loc*ndiags_glob+ir,itheta_out,iphi_out) = &
                Diags(idiag_loc*ndiags_glob+ir,itheta_out,iphi_out) + &
                sqrt(1._RKIND+zeta_tmp**2) * coeff_intdvpar_tmp * &
                real( FFT_F_tmp * conjg(FFT_F_tmp - &
                rg_tmp/(2._RKIND*R0) * (FFT_F_Mp1_tmp + &
                FFT_F_Mm1_tmp) ))/(2._RKIND*F_fluxS_avg(ir,ivpar))
            
              idiag_loc = idiag_loc+1
              ! Entropy with sqrt(B) on each f
              Diags(idiag_loc*ndiags_glob+ir,itheta_out,iphi_out) = &
                Diags(idiag_loc*ndiags_glob+ir,itheta_out,iphi_out) + &
                sqrt(1._RKIND+zeta_tmp**2) * coeff_intdvpar_tmp * &
                real( (FFT_F_tmp - rg_tmp/(4._RKIND*R0) * (&
                FFT_F_Mp1_tmp + FFT_F_Mm1_tmp )) * &
                conjg(FFT_F_tmp  - rg_tmp/(4._RKIND*R0) * (&
                FFT_F_Mp1_tmp + FFT_F_Mm1_tmp )) )/&
                (2._RKIND*F_fluxS_avg(ir,ivpar))
              
              idiag_loc = idiag_loc+1
              !Potential Energy
              Diags(idiag_loc*ndiags_glob+ir,itheta_out,iphi_out) = &
                Diags(idiag_loc*ndiags_glob+ir,itheta_out,iphi_out) + &
                coeff_intdvpar_tmp*real( FFT_P_tmp * conjg(FFT_P_tmp))

              idiag_loc = idiag_loc+1
              ! Advection through curvature drift (m-1)
              vg_tmp = (vpar_tmp**2/((1._RKIND+zeta_tmp**2)*R0) + &
                mu_tmp/(R0*sqrt(1._RKIND+zeta_tmp**2)))
              drfmm1 = 2._RKIND*(FFT_F_Mm1_rp1_tmp - &
                FFT_F_Mm1_rm1_tmp + &
                (FFT_F_Mm1_rm2_tmp - &
                FFT_F_Mm1_rp2_tmp )/8._RKIND ) / &
                (3._RKIND*(geom%rg(2)-geom%rg(1)))
              Diags(idiag_loc*ndiags_glob+ir,itheta_out,iphi_out) = &
                Diags(idiag_loc*ndiags_glob+ir,itheta_out,iphi_out) - &
                sqrt(1._RKIND+zeta_tmp**2) * &
                coeff_intdvpar_tmp*vg_tmp/(2._RKIND)* &
                real(ci*( ((m_tmp-1._RKIND)*FFT_F_Mm1_tmp)/rg_tmp - &
                drfmm1)*conjg(FFT_F_tmp) )/ &
                F_fluxS_avg(ir,ivpar)

              idiag_loc = idiag_loc+1
              ! Advection through curvature drift (m+1)
              drfmp1 = 2._RKIND*(FFT_F_Mp1_rp1_tmp - &
                FFT_F_Mp1_rm1_tmp + &
                (FFT_F_Mp1_rm2_tmp - &
                FFT_F_Mp1_rp2_tmp )/8._RKIND ) / &
                (3._RKIND*(geom%rg(2)-geom%rg(1)))
              Diags(idiag_loc*ndiags_glob+ir,itheta_out,iphi_out) = &
                Diags(idiag_loc*ndiags_glob+ir,itheta_out,iphi_out) - &
                sqrt(1._RKIND+zeta_tmp**2) * &
                coeff_intdvpar_tmp*vg_tmp/(2._RKIND)* &
                real(ci*( ((m_tmp+1._RKIND)*FFT_F_Mp1_tmp)/rg_tmp + &
                drfmp1)*conjg(FFT_F_tmp) )/ &
                F_fluxS_avg(ir,ivpar)
              
              idiag_loc = idiag_loc+1
              ! Linear velocity term (m-1)
              Diags(idiag_loc*ndiags_glob+ir,itheta_out,iphi_out) = &
                Diags(idiag_loc*ndiags_glob+ir,itheta_out,iphi_out) - &
                sqrt(1._RKIND+zeta_tmp**2) * &
                coeff_intdvpar_tmp*mu_tmp*rg_tmp*iota_tmp/ &
                (2._RKIND*R0**2)*real ( &
                ci*dvparFmm1(ir,ivpar)*conjg(FFT_F_tmp))/ &
                F_fluxS_avg(ir,ivpar)

              idiag_loc = idiag_loc+1
              ! Linear velocity term (m+1)
              Diags(idiag_loc*ndiags_glob+ir,itheta_out,iphi_out) = &
                Diags(idiag_loc*ndiags_glob+ir,itheta_out,iphi_out) - &
                sqrt(1._RKIND+zeta_tmp**2) * &
                coeff_intdvpar_tmp*mu_tmp*rg_tmp*iota_tmp/ &
                (2._RKIND*R0**2)*real ( &
                ci*dvparFmp1(ir,ivpar)*conjg(FFT_F_tmp))/ &
                F_fluxS_avg(ir,ivpar)

              !*** Start the loop for the 2D transfer maps
              !*** that are only calculated for few modes
              do imode = 1,nb_fulldiag

                itheta_f_indx = minloc(abs(doubleSP_m(imode)-m_tmp-SP_kth),dim=1)
                iphi_f_indx   = minloc(abs(doubleSP_n(imode)-n_tmp-SP_kphi),dim=1)
                itheta_out_transf = reorg_m(itheta_f_indx)
                iphi_out_transf   = reorg_n(iphi_f_indx)
                if ((SP_kth(itheta_f_indx).ne.doubleSP_m(imode)-m_tmp).or.&
                  (nint(SP_kphi(iphi_f_indx)).ne.doubleSP_n(imode)-nint(n_tmp))) then
                  FFT_F_mn_tmp     = 0._RKIND
                  FFT_F_rp1_mn_tmp = 0._RKIND
                  FFT_F_rm1_mn_tmp = 0._RKIND
                  FFT_F_rp2_mn_tmp = 0._RKIND
                  FFT_F_rm2_mn_tmp = 0._RKIND
                else
                  FFT_F_mn_tmp     = FFT_F(i_rdiags(ir),ivpar, &
                    itheta_f_indx,iphi_f_indx)
                  FFT_F_rp1_mn_tmp = FFT_F(i_rdiags(ir)+1,ivpar, &
                    itheta_f_indx,iphi_f_indx)
                  FFT_F_rm1_mn_tmp = FFT_F(i_rdiags(ir)-1,ivpar, &
                    itheta_f_indx,iphi_f_indx)
                  FFT_F_rp2_mn_tmp = FFT_F(i_rdiags(ir)+2,ivpar, &
                    itheta_f_indx,iphi_f_indx)
                  FFT_F_rm2_mn_tmp = FFT_F(i_rdiags(ir)-2,ivpar, &
                    itheta_f_indx,iphi_f_indx)
                end if
                m_nl_diag   = minloc(abs(doubleSP_m(imode)-SP_kth),dim=1)
                coef_nl     = 1._RKIND
                if (SP_kth(m_nl_diag).ne.doubleSP_m(imode)) coef_nl=0._RKIND
                n_nl_diag   = minloc(abs(doubleSP_n(imode)-SP_kphi),dim=1)
                if (nint(SP_kphi(n_nl_diag)).ne.doubleSP_n(imode)) coef_nl=0._RKIND

                ! ExB advection for (m,n) mode, phi_(m,n)
                drfmm1 = 2._RKIND * ( FFT_F_rp1_mn_tmp - FFT_F_rm1_mn_tmp + ( &
                  FFT_F_rm2_mn_tmp - FFT_F_rp2_mn_tmp)/8._RKIND) / &
                  (3._RKIND*(geom%rg(2)-geom%rg(1)))
                drfmp1 = 2._RKIND * ( FFT_P_rp1_tmp - FFT_P_rm1_tmp + ( &
                  FFT_P_rm2_tmp - FFT_P_rp2_tmp)/8._RKIND) / &
                  (3._RKIND*(geom%rg(2)-geom%rg(1)))
                Diags((idiag_loc+imode)*ndiags_glob+ir,itheta_out_transf,iphi_out_transf) = &
                  Diags((idiag_loc+imode)*ndiags_glob+ir,itheta_out_transf,iphi_out_transf) + &
                  coeff_intdvpar_tmp/ &
                  (rg_tmp*sqrt(1._RKIND+zeta_tmp**2)) * &
                  real(ci*( (doubleSP_m(imode)-m_tmp) * &
                  drfmp1 * FFT_F_mn_tmp - m_tmp*FFT_P_tmp*drfmm1) * &
                  conjg(FFT_F(i_rdiags(ir),ivpar,m_nl_diag, &
                  n_nl_diag)))*coef_nl/ &
                  F_fluxS_avg(ir,ivpar)

                ! ExB advection for (m,n) mode, phi_(m-1,n)
                drfmp2 = 2._RKIND * ( &
                  (FFT_P_Mm1_rp1_tmp) - &
                  (FFT_P_Mm1_rm1_tmp) + ( &
                  (FFT_P_Mm1_rm2_tmp) - &
                  (FFT_P_Mm1_rp2_tmp))/8._RKIND) / &
                  (3._RKIND*(geom%rg(2)-geom%rg(1)))
                Diags((idiag_loc+nb_fulldiag+imode)*ndiags_glob+ir, &
                  itheta_out_transf,iphi_out_transf) = &
                  Diags((idiag_loc+nb_fulldiag+imode)*ndiags_glob+ir, &
                  itheta_out_transf,iphi_out_transf) + &
                  coeff_intdvpar_tmp/ &
                  (rg_tmp*sqrt(1._RKIND+zeta_tmp**2)) * &
                  real(ci*( (doubleSP_m(imode)-m_tmp) * &
                  rg_tmp*drfmp2/(2._RKIND*R0) * FFT_F_mn_tmp -  &
                  rg_tmp*(m_tmp-1._RKIND)*FFT_P_Mm1_tmp/(2_RKIND*R0)*drfmm1) * &
                  conjg(FFT_F(i_rdiags(ir),ivpar,m_nl_diag, &
                  n_nl_diag)))*coef_nl/ &
                  F_fluxS_avg(ir,ivpar)

                ! ExB advection for (m,n) mode, phi_(m+1,n)
                drfmp2 = 2._RKIND * ( &
                  (FFT_P_Mp1_rp1_tmp) - &
                  (FFT_P_Mp1_rm1_tmp) + ( &
                  (FFT_P_Mp1_rm2_tmp) - &
                  (FFT_P_Mp1_rp2_tmp))/8._RKIND) / &
                  (3._RKIND*(geom%rg(2)-geom%rg(1)))
                Diags((idiag_loc+2*nb_fulldiag+imode)*ndiags_glob+ir, &
                  itheta_out_transf,iphi_out_transf) = &
                  Diags((idiag_loc+2*nb_fulldiag+imode)*ndiags_glob+ir, &
                  itheta_out_transf,iphi_out_transf) + &
                  coeff_intdvpar_tmp/ &
                  (rg_tmp*sqrt(1._RKIND+zeta_tmp**2)) * &
                  real(ci*( (doubleSP_m(imode)-m_tmp) * &
                  rg_tmp*drfmp2/(2._RKIND*R0) * FFT_F_mn_tmp - &
                  rg_tmp*(m_tmp+1._RKIND)*FFT_P_Mp1_tmp/(2_RKIND*R0)*drfmm1) * &
                  conjg(FFT_F(i_rdiags(ir),ivpar,m_nl_diag, &
                  n_nl_diag)))*coef_nl/ &
                  F_fluxS_avg(ir,ivpar)

                ! GradparPhi advection for (m,n) mode
                Diags((idiag_loc+3*nb_fulldiag+imode)*ndiags_glob+ir, &
                  itheta_out_transf,iphi_out_transf) = &
                  Diags((idiag_loc+3*nb_fulldiag+imode)*ndiags_glob+ir, &
                  itheta_out_transf,iphi_out_transf) - &
                  coeff_intdvpar_tmp/R0*(n_tmp+m_tmp*iota_tmp)*&
                  real(ci*FFT_P_tmp*dvparF_mn(ir,ivpar,imode) * &
                  conjg(FFT_F(i_rdiags(ir),ivpar,m_nl_diag, &
                  n_nl_diag)))*coef_nl/ &
                  F_fluxS_avg(ir,ivpar)

                ! VeGradB advection for (m,n) mode, phi_(m-1)
                drfmm1 = 2._RKIND*(FFT_P_Mm1_rp1_tmp - &
                  FFT_P_Mm1_rm1_tmp + &
                  (FFT_P_Mm1_rm2_tmp - &
                  FFT_P_Mm1_rp2_tmp )/8._RKIND ) / &
                  (3._RKIND*(geom%rg(2)-geom%rg(1)))
                VeGradB = ci /(2._RKIND*R0 * sqrt(1._RKIND+zeta_tmp**2)) * ( &
                  ((m_tmp-1._RKIND)*FFT_P_Mm1_tmp)/rg_tmp - drfmm1)
                Diags((idiag_loc+4*nb_fulldiag+imode)*ndiags_glob+ir, &
                  itheta_out_transf,iphi_out_transf) = &
                  Diags((idiag_loc+4*nb_fulldiag+imode)*ndiags_glob+ir, &
                  itheta_out_transf,iphi_out_transf) + &
                  coeff_intdvpar_tmp*real(VeGradB*vpar_tmp* &
                  dvparF_mn(ir,ivpar,imode) * &
                  conjg(FFT_F(i_rdiags(ir),ivpar,m_nl_diag, &
                  n_nl_diag)))*coef_nl/ &
                  F_fluxS_avg(ir,ivpar)

                ! VeGradB advection for (m,n) mode, phi_(m+1)
                drfmp1 = 2._RKIND*(FFT_P_Mp1_rp1_tmp - &
                  FFT_P_Mp1_rm1_tmp + &
                  (FFT_P_Mp1_rm2_tmp - &
                  FFT_P_Mp1_rp2_tmp )/8._RKIND ) / &
                  (3._RKIND*(geom%rg(2)-geom%rg(1)))
                VeGradB = ci /(2._RKIND*R0 * sqrt(1._RKIND+zeta_tmp**2)) * ( &
                  ((m_tmp+1._RKIND)*FFT_P_Mp1_tmp)/rg_tmp + drfmp1)
                Diags((idiag_loc+5*nb_fulldiag+imode)*ndiags_glob+ir, &
                  itheta_out_transf,iphi_out_transf) = &
                  Diags((idiag_loc+5*nb_fulldiag+imode)*ndiags_glob+ir, &
                  itheta_out_transf,iphi_out_transf) + &
                  coeff_intdvpar_tmp*real(VeGradB*vpar_tmp* &
                  dvparF_mn(ir,ivpar,imode) * &
                  conjg(FFT_F(i_rdiags(ir),ivpar,m_nl_diag, &
                  n_nl_diag)))*coef_nl/ &
                  F_fluxS_avg(ir,ivpar)

              end do

            end do
          end do
        end do
      end do


      Output => Rarray1_NdiagsNthetaNphi
      Output = 0._RKIND
      !*** Perform integration in mu
      if (Nbproc_mu .gt. 1) then
        !*** Divide to get the correct jacobian,
        !*** remember that B is already taken into account 
        !*** in the spectral calculations
        denom = sum(geom%coeff_intdmu)*sum(geom%coeff_intdvpar)
        do iphi = 1,geom%Nphi
          do itheta = 1,geom%Ntheta
            do ir = 1,nb_terms*ndiags_glob
              Diags(ir,itheta,iphi) = Diags(ir,itheta,iphi) * &
                geom%coeff_intdmu(mu_id) / denom
            end do
          end do
        end do
        nbelements = nb_terms * ndiags_glob * (geom%Ntheta) * (geom%Nphi)
        call MY_Reduce(Diags(1,1,1), Output(1,1,1), nbelements, MPI_REAL8, &
          MPI_SUM,0,mpi_comm_intermu,ierr)
      else 
        denom = sum(geom%coeff_intdvpar)
        do iphi = 1,geom%Nphi
          do itheta = 1,geom%Ntheta
            do ir = 1,nb_terms*ndiags_glob
              Diags(ir,itheta,iphi) = Diags(ir,itheta,iphi)/denom
            end do
          end do
        end do
        Output = Diags
      end if

      !*** Write in File
      if (mu_id.eq.0) then
        call HDF5_spectra_saving(geom,Output,Snb_diag,Stime_diag)
      end if
    end if

    call clck_time(eclock_diagSP)
    call clck_diff(bclock_diagSP,eclock_diagSP,&
      global_time_diagSP)
  end subroutine diag_spectra
  
end module spectra_module

