!**************************************************************
!  Copyright Euratom-CEA
!  Authors : 
!     Virginie Grandgirard (virginie.grandgirard@cea.fr)
!     Chantal Passeron (chantal.passeron@cea.fr)
!     Guillaume Latu (guillaume.latu@cea.fr)
!     Xavier Garbet (xavier.garbet@cea.fr)
!     Philippe Ghendrih (philippe.ghendrih@cea.fr)
!     Yanick Sarazin (yanick.sarazin@cea.fr)
!  
!  This code GYSELA (for GYrokinetic SEmi-LAgrangian) 
!  is a 5D gyrokinetic global full-f code for simulating 
!  the plasma turbulence in a tokamak.
!  
!  This software is governed by the CeCILL-B license 
!  under French law and abiding by the rules of distribution 
!  of free software.  You can  use, modify and redistribute 
!  the software under the terms of the CeCILL-B license as 
!  circulated by CEA, CNRS and INRIA at the following URL
!  "http://www.cecill.info". 
!**************************************************************

#include "config-F.inc"

!-------------------------------------------------------
! file : fequil4d.F90
! date : 19/05/2006
! - computation of the initial contribution of the 
!  equilibrium part of the distribution function :
!   feq(r,theta,vpar) for the value of mu equal to mu_id
!-------------------------------------------------------
module fequil4d_module
  use prec_const
  use MPIutils_module
  use wrap_alloc_module
  use spline1d_class
  use geometry_class
  use init_profile_class
  use init_magnetic_class
  
  use wrap_alloc_module

  implicit none      

  integer        , private :: istart_fmu, iend_fmu
  integer        , private :: jstart_fmu, jend_fmu
  integer        , private :: f4d_BCr_left, f4d_BCr_right
  integer        , private :: f4d_BCvpar_left, f4d_BCvpar_right
  type(nspline1d), private :: nspline1d_r
  type(pspline1d), private :: pspline1d_theta
  type(nspline1d), private :: nspline1d_vpar
  real(RKIND)    , private :: minBnorm, T_h, n_h

  include "gyrokinetic/Bstar_inline.h"
  include "gyrokinetic/velocities_inline.h"
  !******************************
  contains
  !******************************
#include "gyrokinetic/Bstar_inline.F90"
#include "gyrokinetic/velocities_inline.F90"

  !*********************************************************
  ! EQUILIBRIUM PART OF THE DISTRIBUTION FUNCTION
  !*********************************************************
  !----------------------------------------------------------
  ! allocate the arrays associated to the equilibrium part
  !----------------------------------------------------------
  subroutine allocate_equilibrium_part(geom, &
    init_prof,init_magnet,Sfmu_eq,SJ0fmu_eq, &
    Snseq_r,Sdnseq_dr,STseq_r,STGCseq_r)
    use globals, only : BC_Hermite, istart, iend, &
      jstart, jend, hffilterfref, hffilterfreq
    use globals, only : pSpecies_id, coef_diffth, & 
         deltaF_in_advec, avoidtaylor
    type(geometry)               , intent(in) :: geom
    type(init_profile)           , intent(in) :: init_prof
    type(init_magnetic)          , intent(in) :: init_magnet
    real(RKIND), dimension(:,:,:), pointer    :: Sfmu_eq
    real(RKIND), dimension(:,:,:), pointer    :: SJ0fmu_eq
    real(RKIND), dimension(:)    , pointer    :: Snseq_r
    real(RKIND), dimension(:)    , pointer    :: Sdnseq_dr
    real(RKIND), dimension(:)    , pointer    :: STseq_r
    real(RKIND), dimension(:)    , pointer    :: STGCseq_r

!baoter
    istart_fmu = istart
    iend_fmu   = iend
    jstart_fmu = jstart
    jend_fmu   = jend

    call write_expr("istart_fmu", "istart")
    call write_expr("iend_fmu", "iend")
    call write_expr("jstart_fmu", "jstart")
    call write_expr("jend_fmu", "jend")
    if ((hffilterfref.eq.1).and.(hffilterfreq.gt.0)) then
      istart_fmu = 0
      iend_fmu   = geom%Nr
      jstart_fmu = 0
      jend_fmu   = geom%Ntheta
      call write_param("istart_fmu", 0)
      call write_expr("iend_fmu", "geom%Nr")
      call write_param("jstart_fmu", 0)
      call write_expr("jend_fmu", "geom%Ntheta")
    end if
    if (abs(coef_diffth(pSpecies_id)).gt.1.e-12) then  
      istart_fmu = 0
      iend_fmu   = geom%Nr
      jstart_fmu = 0
      jend_fmu   = geom%Ntheta
      call write_param("istart_fmu", 0)
      call write_expr("iend_fmu", "geom%Nr")
      call write_param("jstart_fmu", 0)
      call write_expr("jend_fmu", "geom%Ntheta")
    end if
    if (deltaF_in_advec .and. avoidtaylor) then  
      istart_fmu = 0
      iend_fmu   = geom%Nr
      jstart_fmu = 0
      jend_fmu   = geom%Ntheta
      call write_param("istart_fmu", 0)
      call write_expr("iend_fmu", "geom%Nr")
      call write_param("jstart_fmu", 0)
      call write_expr("jend_fmu", "geom%Ntheta")
    end if
!eaoter
    f4d_BCr_left     = BC_Hermite
    f4d_BCr_right    = BC_Hermite
    f4d_BCvpar_left  = BC_Hermite
    f4d_BCvpar_right = BC_Hermite

    T_h      = minval(init_prof%Ts0(pSpecies_id,:))
    n_h      = maxval(init_prof%ns0(pSpecies_id,:))
    minBnorm = init_magnet%Bmin

    call write_expr("n", "geom%Nr")
    call new_spline1d_natural(nspline1d_r,geom%Nr,geom%dr)
    call write_expr("n", "geom%Ntheta")
    call new_spline1d_period(pspline1d_theta, &
      geom%Ntheta,geom%dtheta)
    call write_expr("n", "geom%Nvpar")
    call new_spline1d_natural(nspline1d_vpar,geom%Nvpar,geom%dvpar)

    call FTR_ALLOC_3_ZERO(Sfmu_eq,istart_fmu,iend_fmu,jstart_fmu,jend_fmu,0,geom%Nvpar)
    call FTR_ALLOC_3_ZERO(SJ0fmu_eq,istart,iend,jstart,jend,0,geom%Nvpar)
    call FTR_ALLOC_1_ZERO(Snseq_r,0,geom%Nr)
    call FTR_ALLOC_1_ZERO(Sdnseq_dr,0,geom%Nr)
    call FTR_ALLOC_1_ZERO(STseq_r,0,geom%Nr)
    call FTR_ALLOC_1_ZERO(STGCseq_r,0,geom%Nr)
  end subroutine allocate_equilibrium_part


  !---------------------------------------------------------
  ! delete the arrays associated to the equilibrium part
  !---------------------------------------------------------
  subroutine deallocate_equilibrium_part(Sfmu_eq,SJ0fmu_eq, &
    Snseq_r,Sdnseq_dr,STseq_r,STGCseq_r)
    real(RKIND), dimension(:,:,:), pointer :: Sfmu_eq 
    real(RKIND), dimension(:,:,:), pointer :: SJ0fmu_eq
    real(RKIND), dimension(:)    , pointer :: Snseq_r 
    real(RKIND), dimension(:)    , pointer :: Sdnseq_dr
    real(RKIND), dimension(:)    , pointer :: STseq_r
    real(RKIND), dimension(:)    , pointer :: STGCseq_r

    call del_spline1d_natural(nspline1d_r)
    call del_spline1d_period(pspline1d_theta)
    call del_spline1d_natural(nspline1d_vpar)

    call FTR_DEALLOC(Sfmu_eq)
    call FTR_DEALLOC(SJ0fmu_eq)
    call FTR_DEALLOC(Snseq_r)
    call FTR_DEALLOC(Sdnseq_dr)
    call FTR_DEALLOC(STseq_r)
    call FTR_DEALLOC(STGCseq_r)
  end subroutine deallocate_equilibrium_part


  !******************************************************
  !*** Definition of the normalized mass (hat_As) and ***
  !***  normalized charge (hat_Zs)                    *** 
  !***  for all species                               ***
  !***  Rk: Mass and charge are normalized to the     ***
  !***      first specie                              ***
  !***  In : A0 = mass for normalisation              ***
  !***       Z0 = charge for normalisation            ***
  !***       As = mass of s species                   ***
  !***       Zs = charge of s species                 ***
  !***       species_frac = fraction of s species     ***
  !*** Out : hat_As = normalized mass of specie s     ***
  !***       hat_Zs = normalized charge of specie s   ***
  !******************************************************
  subroutine init_mass_charge(A0,Z0,As,Zs,species_frac, &
    hat_As,hat_Zs)
    use globals, only : pglobal_id, Nspecies, &
      adiabatic_electrons
!R3 #include "r3_info.h" !R3
    real(RKIND)                         , intent(in)  :: A0, Z0
    real(RKIND), dimension(0:Nspecies-1), intent(in)  :: As
    real(RKIND), dimension(0:Nspecies-1), intent(in)  :: Zs
    real(RKIND), dimension(0:Nspecies-1), intent(in)  :: species_frac
    real(RKIND), dimension(0:Nspecies-1), intent(out) :: hat_As
    real(RKIND), dimension(0:Nspecies-1), intent(out) :: hat_Zs

    integer     :: ispec
    real(RKIND) :: sum_all_species

!R3 call r3_info_begin (r3_info_index_1, 'Init_mass_charge') !R3
    call Write_dyn_balise_open('Init_mass_charge') !R4
    do ispec = 0,Nspecies-1
      hat_As(ispec) = As(ispec)/A0
      hat_Zs(ispec) = Zs(ispec)/Z0
    end do

    !*** Test for the coherence of the species fraction      ***
    !*** -> With adiabatic electrons,                        ***
    !***      sum_s cs Zs = 1 must be satisfied, where s     ***
    !***    correspond only to ion species                   ***
    !**  -> With kinetic electrons,                          ***
    !***      sum_s cs Zs = 0 must be satisfied, where s     ***
    !***    correspond to all species, including electrons   ***
    !*** Note : for electrons hat_Zs = -1                    ***
    sum_all_species = 0._RKIND
    do ispec = 0,Nspecies-1
      sum_all_species = sum_all_species + &
        Zs(ispec)*species_frac(ispec)
    end do
    if (adiabatic_electrons) then
      if (sum_all_species.ne.1._RKIND) then
        if (pglobal_id.eq.0) &
          print*,' BECAREFUL sum_s species_frac*Zs=1 ', &
          'is not satisfied because equal to ',sum_all_species
        call ppexit()
        call exit(-1)
      end if
    else  ! <- kinetic electrons
      if (sum_all_species.ne.0._RKIND) then
        if (pglobal_id.eq.0) &
          print*,' BECAREFUL sum_s species_frac*Zs=0 ', &
          'is not satisfied because equal to ',sum_all_species
        call ppexit()
        call exit(-1)
      end if
    end if
!R3 call r3_info_end (r3_info_index_1) !R3
    call Write_dyn_balise_close('Init_mass_charge') !R4
  end subroutine init_mass_charge


  !-----------------------------------------------------------
  ! Computation of rbar = rp-(q_rp/rp)*[psi(r)-psi(rp)]
  !    -(q_rp/rp)*[vpar*R-vpar0*R0*H(energy-mu*Bmax)] 
  !  where . rp     = r(peak)
  !        . psi(r) = -\int_0^r r/q(r) dr
  !        . vpar0  = sign(vpar) * 
  !                   sqrt(vpar^2+2*mu*[B(r,theta)-Bmin])
  !        . Bmin   = min(B)
  ! with H(energy-mu*Bmax) is close to the heaviside 
  !  ( => tanh for having a smoothness transition between 
  !     trapped and passing particles)
  !  Rk : H is equal to 0 for the trapped particles
  !-----------------------------------------------------------
  subroutine compute_rbar(geom,ir,itheta,ivpar,imu, &
    init_prof,init_magnet,rbar,psibar)
    use globals, only : R0, a, q_param1, q_param2, q_param3, &
      feq_choice, canonical_vpar0, q_profile
    type(geometry)     , intent(in)  :: geom
    integer            , intent(in)  :: ir, itheta, ivpar, imu
    type(init_profile) , intent(in)  :: init_prof
    type(init_magnetic), intent(in)  :: init_magnet
    real(RKIND)        , intent(out) :: rbar, psibar

    real(RKIND) :: rp, q_rp, psi_rp
    real(RKIND) :: vparl, Bmin, Bmax, Bij, vpar0, mum
    real(RKIND) :: energy, part_type, H_passing, v_tmp, sign_vpar

    rp        = init_prof%rpeak
    q_rp      = init_prof%q_rp 
    psi_rp    = init_prof%psi_rp
    ! -> vpar0 = sign(vpar)*sqrt(vpar^2+2*mu*[B(r,theta)-Bmin])
    Bmin      = init_magnet%Bmin
    Bmax      = init_magnet%Bmax
    Bij       = init_magnet%B_norm(ir,itheta)
    vparl     = geom%vparg(ivpar)
    mum       = geom%mug(imu)
    energy    = 0.5_RKIND*vparl**2 + mum*Bij
    sign_vpar = sign(1._RKIND,vparl)
    
    !*** computation of vpar0 ***
    select case (canonical_vpar0)
    case(0)
      ! ->  vpar0 = 0 
      vpar0     = 0._RKIND
      H_passing = 0._RKIND
    case(1)  
      ! -> vpar0 = sign(vpar)*sqrt(2*[E-mu*Bmin])*H(E-mu*Bmax)
      part_type = energy-mum*Bmax
      if (part_type.le.0._RKIND) then
        vpar0     = 0._RKIND
        H_passing = 0._RKIND
      else
        v_tmp     = sqrt(2._RKIND*(energy-mum*Bmin))
        vpar0     = sign_vpar*v_tmp
        H_passing = 1._RKIND
      end if
    case(2)  
      ! -> vpar0 = sign(vpar)*sqrt(2*[E-mu*Bmax])*H(E-mu*Bmax)
      part_type = energy-mum*Bmax
      if (part_type.le.0._RKIND) then
        vpar0     = 0._RKIND
        H_passing = 0._RKIND
      else
        v_tmp     = sqrt(2._RKIND*(energy-mum*Bmax))
        vpar0     = sign_vpar*v_tmp
        H_passing = 1._RKIND
      end if
    case(3)
      ! -> vpar0 = sign(vpar)*sqrt(2*[E-mu*B0])*H(E-mu*Bmax)
      part_type = energy-mum*Bmax
      if (part_type.le.0._RKIND) then
        vpar0     = 0._RKIND
        H_passing = 0._RKIND
      else
        v_tmp     = sqrt(2._RKIND*(energy-mum*1._RKIND))
        vpar0     = sign_vpar*v_tmp
        H_passing = (tanh((part_type-3._RKIND*geom%dvpar) / &
          (geom%dvpar)) + 1._RKIND) * 0.5_RKIND 
      end if
    case(4)
      ! -> vpar0 = sign(vpar)*sqrt(2*[E-mu*B0])*H(E-mu*B0)
      part_type = energy-mum*1._RKIND
      if (part_type.le.0._RKIND) then
        vpar0     = 0._RKIND
        H_passing = 0._RKIND
      else
        v_tmp     = sqrt(2._RKIND*(energy-mum*1._RKIND))
        vpar0     = sign_vpar*v_tmp
        H_passing = 1._RKIND
      end if
    case default
      if (pglobal_id.eq.0) then
        print*, 'option canonical_vpar0 = ', &
          canonical_vpar0, ' not treated'
      end if
    end select

    psibar = init_prof%psi(ir)+R0*(vparl/Bij-H_passing*vpar0)
#ifdef OPTRBAR
    rbar = rp-(q_rp/rp) * &
        (init_prof%psi(ir)-psi_rp+R0*vparl/Bij-H_passing*vpar0*R0)
    rbar = MIN(MAX(rbar, geom%rg(0)),geom%rg(Nr))
#else
    if ((q_param3.ne.2._RKIND).or.(q_profile.ne.2)) then
       ! -> rbar = rp-(q_rp/rp)*[psi(r)-psi(rp)- 
       !    R0*vpar/(B(ir,itheta)-vpar0*R0)] 
       rbar = rp-(q_rp/rp) * &
         (init_prof%psi(ir)-psi_rp+R0*vparl/Bij-H_passing*vpar0*R0)
    else
      ! in the case of a parobolic profile for q:
      !  q=q0+deltaq*(r/a)^2 the expression of r can be 
      !  deduced directly from psi(r) so the correction is 
      !  made on psi: psibar = psi + corr and then 
      !  rbar is obtained with psibar
      ! -> psibar = psi-R0*vpar/(q*B(ir,itheta))-vpar0*R0/q)] 
      ! -> rbar = sqrt(q0a^2/deltaq * 
      !           (exp(-2._RKIND*deltaq*psibar/a^2)-1))
      if (psibar.lt.0) then
        rbar = sqrt(q_param1*a**2/q_param2 * &
          (exp(-2._RKIND*q_param2*psibar/a**2)-1))      
      else
        !-> in this case rbar < rmin so is put equal to rmin 
        rbar = geom%rg(0)
      end if
    end if
#endif
  end subroutine compute_rbar
  

  !------------------------------------------------------------
  ! Computation of the EP distribution function
  !------------------------------------------------------------  
  subroutine get_ffast(geom,ir,itheta,ivpar,imu, &
    init_prof,init_magnet,ffast)
    use globals, only : pSpecies_id, fast_part, &
      bump_on_tail, fp_rshape, &
      fp_lambda, fp_mom, fp_momgauss, fp_energy, &
      Eb_fp, Emin_fp, n_fp, dn_fp, E_critfp, rho_0fp, &
      deltarhofp, Lambda_0fp, deltaLambdafp, psi_0fp, &
      deltapsifp, fp_mom, a, R0, zetabar    
    type(geometry)     , intent(in)  :: geom
    integer            , intent(in)  :: ir, itheta, ivpar, imu
    type(init_profile) , intent(in)  :: init_prof
    type(init_magnetic), intent(in)  :: init_magnet
    real(RKIND)        , intent(out) :: ffast
    
    real(RKIND) :: rho, energy, Lambda
    real(RKIND) :: vpar_mod, denom_h, feq_tmp
    real(RKIND) :: rshape, pphishape, lambdashape
    real(RKIND) :: energyshape
    real(RKIND) :: Pphi, dpsi, psi0, major_Radius
    
    if (.not.fast_part(pSpecies_id)) then
      ffast = 0._RKIND
    else
      energy = 0.5_RKIND*geom%vparg(ivpar)**2 + &
        geom%mug(imu)*init_magnet%B_norm(ir,itheta)
      denom_h = sqrt(2._RKIND*PI*T_h)*(2._RKIND*PI*T_h)**ipow
      feq_tmp = n_h/denom_h * exp(-energy/T_h)
      if (fp_rshape(pSpecies_id)) then
        rho    = (geom%rg(ir) - geom%rg(0)) / &
          (geom%rg(Nr) - geom%rg(0))
        rshape = -0.5_RKIND* &
          (tanh((rho-rho_0fp(pSpecies_id)- &
          3._RKIND*deltarhofp(pSpecies_id))/ &
          deltarhofp(pSpecies_id)) + &
          tanh((rho_0fp(pSpecies_id)- &
          3._RKIND*deltarhofp(pSpecies_id)-rho)/ &
          deltarhofp(pSpecies_id)))
      else
        rshape = 1._RKIND
      end if
      if (bump_on_tail(pSpecies_id)) then
        if (geom%vparg(ivpar) .ge. 0) then
          vpar_mod = sqrt(2._RKIND* &
            (energy-geom%mug(imu)*minBnorm))
        else
          vpar_mod = -1._RKIND*sqrt(2._RKIND* &
            (energy-geom%mug(imu)*minBnorm))
        end if
        ffast = rshape * &
          n_h*n_fp(pSpecies_id)/denom_h*exp(-energy/T_h)* &
          (exp(-zetabar(pSpecies_id)**2/(2._RKIND*T_h))* &
          cosh(zetabar(pSpecies_id)*vpar_mod/T_h) + &
          dn_fp(pSpecies_id)*exp(-zetabar(pSpecies_id)**2/ &
          (2._RKIND*T_h))*exp(zetabar(pSpecies_id)*vpar_mod/T_h))
      else
        if (fp_mom(pSpecies_id)) then
          major_Radius = R0 + geom%rg(ir)*geom%cos_theta(itheta)
          Pphi = init_prof%psi(ir) + &
            geom%vparg(ivpar)*(major_Radius**2._RKIND)*&
            init_magnet%B_gradphi(ir,itheta)/&
            init_magnet%B_norm(ir,itheta)
          dpsi = -deltapsifp(pSpecies_id)* &
            abs(init_prof%psimax - init_prof%psimin)
          if (.not.fp_momgauss(pSpecies_id)) then
            pphishape = exp(-Pphi/dpsi)
          else
            psi0 = psi_0fp(pSpecies_id)*init_prof%psi(Nr)
            pphishape = exp(-(Pphi-psi0)**2/dpsi**2)
          end if
        else
          pphishape = 1._RKIND
        end if
        if (fp_lambda(pSpecies_id)) then
          Lambda = geom%mug(imu)/energy
          lambdashape = exp(-(Lambda-Lambda_0fp(pSpecies_id))**2/ &
            deltaLambdafp(pSpecies_id)**2)
        else
          lambdashape = 1._RKIND
        end if
        if (fp_energy(pSpecies_id)) then
          energyshape = 0.25_RKIND*&
            (1._RKIND+tanh(1.5_RKIND*(energy+Emin_fp(pSpecies_id))))*&
            (1._RKIND-tanh(1.5_RKIND*(energy-Eb_fp(pSpecies_id))))
        else
          energyshape = 1._RKIND
        end if
        ffast = n_fp(pSpecies_id) / &
          (energy**1.5_RKIND+E_critfp(pSpecies_id)**1.5_RKIND)* &
          rshape*pphishape*lambdashape*energyshape
      end if
    end if
  end subroutine get_ffast


  !------------------------------------------------------------
  ! Computation of the equilibrium part feq of the 
  !   distribution function at the point (ir,itheta,ivpar,imu):
  !    feq(rbar,theta,vpar,mu) = ns(rbar) / 
  !        (2*PI*Ts(rbar))**(3/2) *
  !        exp(-1/Ts(rbar)*(1/2*vpar**2+mu*B(r,theta)))
  !  with rbar = (q_rp/rp)*[psi(r)-R0*vpar/(B(ir,itheta)] 
  !  where rp = r(peak)
  !------------------------------------------------------------
  subroutine compute_feq_value(geom,ir,itheta,ivpar,imu, &
    init_prof,init_magnet,feq)
    use globals, only : Rarray1_Nr, &
      canonical_profile, feq_choice, &
      min_err_rbar, max_err_rbar 
    use globals, only : pSpecies_id, &
      buffer_location, buffer_stiffness, &
      buffer_asymmetry, rflux_driven
    use utils_module
    type(geometry)     , intent(in)  :: geom
    integer            , intent(in)  :: ir, itheta, ivpar, imu
    type(init_profile) , intent(in)  :: init_prof
    type(init_magnetic), intent(in)  :: init_magnet
    real(RKIND)        , intent(out) :: feq

    integer     :: ipos, ibase
    integer     :: ipsimin, ipsimax, irp
    real(RKIND) :: dTidr_rp
    real(RKIND) :: denom, Ts_r, Ts_rp, ns_rp
    real(RKIND) :: rbar, Ts_rbar, ns_rbar
    real(RKIND) :: psibar, psibar_minus_psi, &
      test_inside, deltar, psimin, psimax
    real(RKIND) :: energy, float_Nmu
    real(RKIND), dimension(-1:2) :: sbase
    real(RKIND) :: Vpar0_th, Vpar0_rshape, Vpar0_rth
    real(RKIND) :: mask, cst, rho, asymmetric

    select case (canonical_profile)
    case(0)
      Ts_r   = init_prof%Ts0(pSpecies_id,ir)

      !Compute mean parallel velocity
      Vpar0_th = init_prof%Vpar0(itheta)
      rho      = (geom%rg(ir)-geom%rg(0))/geom%Lr
      if (buffer_asymmetry) then
        asymmetric = 1._RKIND-rflux_driven/2._RKIND
      else
        asymmetric = 1._RKIND
      end if
      ! use buffer shape for parallel velocity
      ! (0 near rmin, 1 near rmax)
      Vpar0_rshape = HF + &
        HF*(tanh((rho-HF)/buffer_stiffness))
      cst  = 1.e-7_RKIND
      mask = 0.5_RKIND * &
        (1._RKIND+sign(1._RKIND,Vpar0_rshape-cst))
      Vpar0_rshape = mask*Vpar0_rshape
      Vpar0_rth    = Vpar0_th * Vpar0_rshape

      denom  = sqrt(2._RKIND*PI*Ts_r) * &
        (2._RKIND*PI*Ts_r)**ipow
      !*** compute the energy as E = 1/2 vpar^2 + mu*B         ***
      !*** Rk: In the 4D case for one unique value of mu.ne.0  ***
      !***      we take into account only the 1/2 vpar^2 part  ***
      !***  Becareful, in this case the equilibrium            ***
      !***   distribution is no more a function of the         ***
      !***   invariants, but this is done to avoid the fact    ***
      !***   that exp(mu*B) is not cancelled when performing   ***
      !***   the velocity space integral                       ***
      float_Nmu = geom%Nmu
      energy    = 0.5_RKIND*(geom%vparg(ivpar)-Vpar0_rth)**2 + &
        min(float_Nmu,1._RKIND)*geom%mug(imu) * &
        init_magnet%B_norm(ir,itheta)
      feq       = init_prof%ns0(pSpecies_id,ir)/denom * &
        exp(-energy/Ts_r)
    case(1)
      !*** initialisation of the Maxwellian with psibar ***  
      call compute_rbar(geom,ir,itheta,ivpar,imu, &
        init_prof,init_magnet,rbar,psibar)
      min_err_rbar = min(min_err_rbar,abs(rbar-geom%rg(ir)))
      max_err_rbar = max(max_err_rbar,abs(rbar-geom%rg(ir)))
      !*** computation of Ts(rbar) and ns(rbar) ***
      !***  by cubic spline interpolation       ***
      ! ->  ipos = position in rg of the first knot 
      !      at the left of rbar
      ipos = floor( (rbar-geom%rg(0))/geom%dr)
      if (ipos.le.0) then
        Ts_rbar = init_prof%Ts0(pSpecies_id,0)
        ns_rbar = init_prof%ns0(pSpecies_id,0)
      else
        if (ipos.gt.geom%Nr) then
          Ts_rbar = init_prof%Ts0(pSpecies_id,geom%Nr)
          ns_rbar = init_prof%ns0(pSpecies_id,geom%Nr)
        else
          ! -> cubic spline interpolation
          if (ipos.eq.geom%Nr) ipos = geom%Nr-1
          call spline_basis(geom%rg(ipos),rbar, &
            geom%rg(ipos+1),geom%dr,sbase)
          Ts_rbar = 0._RKIND
          ns_rbar = 0._RKIND
          do ibase = -1,2
            Ts_rbar = Ts_rbar + &
              init_prof%Ts0_coef1d(pSpecies_id,ipos+ibase) * &
              sbase(ibase)
            ns_rbar = ns_rbar + &
              init_prof%ns0_coef1d(pSpecies_id,ipos+ibase) * &
              sbase(ibase)
          enddo
        end if
      end if
    case(2)
      !*** initialisation of the Maxwellian with psibar ***  
      call compute_rbar(geom,ir,itheta,ivpar,imu, &
        init_prof,init_magnet,rbar,psibar)
      min_err_rbar = min(min_err_rbar,abs(rbar-geom%rg(ir)))
      max_err_rbar = max(max_err_rbar,abs(rbar-geom%rg(ir)))
      psimin  = init_prof%psimin
      psimax  = init_prof%psimax
      ipsimin = init_prof%ipsimin
      ipsimax = init_prof%ipsimax
      test_inside = psibar-psimin
      if (test_inside.le.0._RKIND) then
        deltar  = -test_inside/(geom%rg(ipsimin) * &
          init_prof%iota(ipsimin))
        Ts_rbar = init_prof%Ts0(pSpecies_id,ipsimin)+deltar * &
          init_prof%dTs0dr(pSpecies_id,ipsimin)
        ns_rbar = init_prof%ns0(pSpecies_id,ipsimin)+deltar * &
          init_prof%dns0dr(pSpecies_id,ipsimin)
        Ts_rbar = max(Ts_rbar,0._RKIND)
        ns_rbar = max(ns_rbar,0._RKIND)
      else
        test_inside = psimax-psibar
        if (test_inside.le.0._RKIND) then
          deltar  = test_inside/(geom%rg(ipsimax) * &
            init_prof%iota(ipsimax))
          if (abs(deltar).le.geom%rg(0)) then
            Ts_rbar = init_prof%Ts0(pSpecies_id,ipsimax) + &
              deltar*init_prof%dTs0dr(pSpecies_id,ipsimax)
            ns_rbar = init_prof%ns0(pSpecies_id,ipsimax) + &
              deltar*init_prof%dns0dr(pSpecies_id,ipsimax)
          else if (abs(deltar).le.2._RKIND*geom%rg(0)) then
            Ts_rbar = init_prof%Ts0(pSpecies_id,ipsimax)- &
              (2._RKIND*geom%rg(0)-abs(deltar)) * &
              init_prof%dTs0dr(pSpecies_id,ipsimax)
            ns_rbar = init_prof%ns0(pSpecies_id,ipsimax)- &
              (2._RKIND*geom%rg(0)-abs(deltar)) * &
              init_prof%dns0dr(pSpecies_id,ipsimax)
          else
            ipos = floor((abs(deltar)-2._RKIND*geom%rg(0)) / &
              geom%dr)
            if (ipos.ge.geom%Nr) then
              Ts_rbar = init_prof%Ts0(pSpecies_id,ipsimin)
              ns_rbar = init_prof%ns0(pSpecies_id,ipsimin)
            else
              Ts_rbar = init_prof%Ts0(pSpecies_id,ipos)
              ns_rbar = init_prof%ns0(pSpecies_id,ipos)
            end if
          end if
        else
          !*** Looking for the nearest psi value
          !*** interpolation (2nd order Taylor expansion)
          test_inside = abs(psibar-psimin)
          do ipos = 0,geom%Nr
            psibar_minus_psi = psibar-init_prof%psi(ipos)
            if (abs(psibar_minus_psi).le.test_inside) then
              deltar  = -psibar_minus_psi / &
                (geom%rg(ipos)*init_prof%iota(ipos))
              Ts_rbar = init_prof%Ts0(pSpecies_id,ipos) + &
                deltar*init_prof%dTs0dr(pSpecies_id,ipos) + &
                deltar*deltar * &
                init_prof%d2Ts0dr2(pSpecies_id,ipos)/2._RKIND
              ns_rbar = init_prof%ns0(pSpecies_id,ipos) + &
                deltar*init_prof%dns0dr(pSpecies_id,ipos) + &
                deltar*deltar * &
                init_prof%d2ns0dr2(pSpecies_id,ipos)/2._RKIND
              test_inside = abs(psibar_minus_psi)
            end if
          end do
        end if
      end if
    end select

    float_Nmu = geom%Nmu
    if (canonical_profile.ne.0) then
      !*** computation of feq depending on rbar ***
      denom  = sqrt(2._RKIND*PI*Ts_rbar) * &
        (2._RKIND*PI*Ts_rbar)**ipow
      !*** compute the energy as E = 1/2 vpar^2 + mu*B         ***
      !*** Rk: In the 4D case for one unique value of mu.ne.0  ***
      !***      we take into account only the 1/2 vpar^2 part  ***
      !***  Becareful, in this case the equilibrium            ***
      !***   distribution is no more a function of the         ***
      !***   invariants, but this is done to avoid the fact    ***
      !***   that exp(mu*B) is not cancelled when performing   ***
      !***   the velocity space integral                       ***
      energy = 0.5_RKIND*geom%vparg(ivpar)**2 + &
        min(float_Nmu,1._RKIND)*geom%mug(imu) * &
        init_magnet%B_norm(ir,itheta) 
      Ts_rp  = init_prof%Ts0_rp(pSpecies_id)
      ns_rp  = init_prof%ns0_rp(pSpecies_id)
      select case (feq_choice)
      case (1,5)
        feq = ns_rbar/denom * exp(-energy/Ts_rbar)
      case (2,4,6)
        feq = exp(-(rbar/init_prof%rpeak)**2)
      case (3)
        feq = exp(-energy)
      case (7)
        denom  = sqrt(2._RKIND*PI*Ts_rp)*(2._RKIND*PI*Ts_rp)**ipow
        feq = (ns_rbar+(energy/Ts_rp-0.5_RKIND-float(ipow)) &
          *ns_rbar*(Ts_rbar-Ts_rp)/Ts_rp)/denom*exp(-energy/Ts_rp)
      case (8)
        denom  = sqrt(2._RKIND*PI*Ts_rp)*(2._RKIND*PI*Ts_rp)**ipow
        feq    = ns_rbar*exp(-energy/Ts_rp)/denom
      end select
    else
      !*** equilibrium used for comparison with TIM4D code ***
      if (feq_choice.eq.9) then
        irp      = int(geom%Nr/2)
        dTidr_rp = (init_prof%Ts0(pSpecies_id,irp-1) - &
          init_prof%Ts0(pSpecies_id,irp+1)) / &
          (geom%rg(irp+1)-geom%rg(irp-1))
        Ts_r     = init_prof%Ts0(pSpecies_id,irp)
        denom    = sqrt(2._RKIND*PI*Ts_r) * &
          (2._RKIND*PI*Ts_r)**ipow
        energy = 0.5_RKIND*geom%vparg(ivpar)**2 + &
          min(float_Nmu,1._RKIND)*geom%mug(imu) * &
          init_magnet%B_norm(ir,itheta)
        feq  = init_prof%ns0(pSpecies_id,ir)/denom * & 
          exp(-energy/Ts_r)*(1._RKIND+(energy/Ts_r - & 
          0.5_RKIND-float(ipow))*dTidr_rp*geom%rg(ir))
      end if
    end if
    if ( abs(feq).le.1.e-30_RKIND ) feq = 1.e-30_RKIND
  end subroutine compute_feq_value


  !-------------------------------------------------------------
  ! Computation of the equilibrium terms (global variables):
  !    . fs_eq(r,theta,vpar,mu) the equilibrium part of f
  !    . J0.fs_eq the gyroaveraged fs_eq
  !    . ns_eq(r,theta) = \int J0.fs_eq dv
  !        with dv = Jv dvpar dmu
  !        where Jv the Jacobian in velocity space 
  !        is equal to:   
  !          Jv = 2*pi*Bstar(r,theta,vpar) in 5D case
  !             = 1                        in 4D case
  !    . P_eq(r,theta) = press_coeff*<J0M2_eq(r,theta)>_theta
  !        where J0M2_eq(r,theta) = \int J0.fs_eq Epars dv
  !        with Epars      = 0.5*(vpar-u_eq)^2+mu*B
  !        and  press_coef = 2/3 if Nmu.ne.0        
  !                        = 2   otherwise 
  !    . u_eq = \int J0.fs_eq vpar dv / ns_eq = 0
  !    . T_eq(r,theta) = P_eq(r,theta)/ns_eq(r,theta)
  !
  ! Let us define for all function g,
  !   <g>_theta = \int g Js dtheta / \int Js dtheta
  !     with Js the jacobian in space
  ! Then,
  !    . ns_eq(r)  = <ns_eq(r,theta)>_theta
  !    . T_eq(r)   = <T_eq(r,theta)>_theta
  !    . nbions_eq = \int J0.fs_eq Js dr dtheta dv
  !                = \int ns_eq dr                   
  !-------------------------------------------------------------
  subroutine compute_equilibrium_part(geom,J0, &
    init_prof,init_magnet,init_curr,Sfs_eq,SJ0fs_eq, &
    Snseq_r,Sdnseq_dr,STseq_r,Snbions_eq,STGCseq_r)
    use globals, only : pglobal_id, Nbproc_mu, mu_id, Nbthread, &
      mpi_comm_intermu, min_err_rbar, max_err_rbar, &
      istart, iend, jstart, jend, a, q_param3, q_profile, &
      Rarray1_Nr, Rarray2_Nr, Rarray3_Nr, &
      Rarray4_Nr, Rarray5_Nr, Rarray6_Nr, Rarray7_Nr, &
      Rarray1_NrNtheta, Rarray2_NrNtheta, &
      fast_part
#ifdef _OPENMP
    use OMPutils_module, only : omp_get_thread_num
#endif
    use OMPutils_module, only : Romp1_0Nr, Romp2_0Nr, &
      Romp3_0Nr, Romp4_0Nr
    use utils_module, only : deriv1
    use gyroaverage_class
    use mpiMod
!R3 #include "r3_info.h" !R3
    type(geometry)               , intent(in)    :: geom
    type(J0operator)             , intent(inout) :: J0
    type(init_profile)           , intent(in)    :: init_prof
    type(init_magnetic)          , intent(in)    :: init_magnet
    type(init_current)           , intent(in)    :: init_curr
    real(RKIND), dimension(:,:,:), pointer       :: Sfs_eq
    real(RKIND), dimension(:,:,:), pointer       :: SJ0fs_eq
    real(RKIND), dimension(0:)   , intent(out)   :: Snseq_r 
    real(RKIND), dimension(0:)   , intent(out)   :: Sdnseq_dr
    real(RKIND), dimension(0:)   , intent(out)   :: STseq_r 
    real(RKIND)                  , intent(out)   :: Snbions_eq
    real(RKIND), dimension(0:)   , intent(out)   :: STGCseq_r  

    integer     :: ierr
    integer     :: ithread, tid
    integer     :: ir, itheta, ivpar
    real(RKIND) :: fs_eq_tmp , ffast, J0fs_eq_tmp 
    real(RKIND) :: vparl2, B_ij, energy
    real(RKIND) :: inv_intdtheta_Js
    real(RKIND) :: jacob_space_tmp, jacob_vel_tmp
    real(RKIND) :: coeff_intdr_tmp, coeff_intdtheta_tmp
    real(RKIND) :: coeff_intdvpar_tmp, coeff_intdmu_tmp
    real(RKIND) :: int_Jsdtheta_dV_tmp
    real(RKIND) :: nseq_tmp, Pseq_tmp, press_coeff
    real(RKIND) :: nGCseq_tmp, PGCseq_tmp

    real(RKIND), dimension(:,:), pointer :: fs_eq_rtheta
    real(RKIND), dimension(:,:), pointer :: J0fs_eq_rtheta
    real(RKIND), dimension(:)  , pointer :: nseq_r_loc
    real(RKIND), dimension(:)  , pointer :: Pseq_r_loc
    real(RKIND), dimension(:)  , pointer :: Pseq_r
    real(RKIND), dimension(:)  , pointer :: nGCseq_r_loc
    real(RKIND), dimension(:)  , pointer :: PGCseq_r_loc
    real(RKIND), dimension(:)  , pointer :: nGCseq_r
    real(RKIND), dimension(:)  , pointer :: PGCseq_r

!R3 call r3_info_begin (r3_info_index_1, 'Init_comp_equil') !R3
    call Write_dyn_balise_open('Init_comp_equil') !R4

   do tid = 1, Nbthread
      call FTR_ALLOC_2(J0%omp_diag(tid)%val,1,J0%N1,1,J0%N2)
      call FTR_ALLOC_2(J0%omp_ldiag(tid)%val,1,J0%N1-1,1,J0%N2)
      call FTR_ALLOC_2(J0%omp_udiag(tid)%val,1,J0%N1-1,1,J0%N2)
   enddo

    if (pglobal_id.eq.0) then
      write(6,'(2A45)') &
        '---------------------------------------------', &
        '---------------------------------------------'
      write(6,*) '--> EQUILIBRIUM COMPUTATION :'
      if ((q_param3.ne.2._RKIND).or.(q_profile.ne.2)) then        
        write(6,*) '-->   - WARNING : the q profile is ', &
          'not parabolic, so rbar is '
        write(6,*) '-->               computed with an ', &
          'approximation at first order of psibar'
      else
        write(6,*) '-->   - rbar is the analytical ',&
          'expression of psibar'
      end if
    end if

    fs_eq_rtheta   => Rarray1_NrNtheta
    J0fs_eq_rtheta => Rarray2_NrNtheta
    nseq_r_loc     => Rarray1_Nr
    Pseq_r_loc     => Rarray2_Nr
    Pseq_r         => Rarray3_Nr
    nGCseq_r_loc   => Rarray4_Nr
    PGCseq_r_loc   => Rarray5_Nr
    nGCseq_r       => Rarray6_Nr
    PGCseq_r       => Rarray7_Nr

    !*** initialisation to 0 ***
    do ir = 0,geom%Nr
      do ithread = 1,Nbthread
        Romp1_0Nr(ithread)%val(ir) = 0._RKIND
        Romp2_0Nr(ithread)%val(ir) = 0._RKIND
        Romp3_0Nr(ithread)%val(ir) = 0._RKIND
        Romp4_0Nr(ithread)%val(ir) = 0._RKIND
      end do
      nseq_r_loc(ir) = 0._RKIND
      Pseq_r_loc(ir) = 0._RKIND
    end do

    min_err_rbar     = a
    max_err_rbar     = 0._RKIND
    coeff_intdmu_tmp = geom%coeff_intdmu(mu_id)

    do ivpar = 0,geom%Nvpar
      vparl2             = geom%vparg(ivpar)**2
      coeff_intdvpar_tmp = geom%coeff_intdvpar(ivpar)
      !*** computation of fs_eq ***
!$OMP PARALLEL default(shared) private(itheta,ir,fs_eq_tmp,ffast)
!$OMP DO schedule(static)
      do itheta = 0,geom%Ntheta
        do ir = 0,geom%Nr
          call compute_feq_value(geom,ir,itheta,ivpar,mu_id, &
            init_prof,init_magnet,fs_eq_tmp)
          call get_ffast(geom,ir,itheta,ivpar,mu_id, &
            init_prof,init_magnet,ffast)
          fs_eq_rtheta(ir,itheta) = fs_eq_tmp + ffast
        end do
      end do
!$OMP END DO

      !*** save fs_eq(istart:iend,jstart:jend,0:Nvpar) ***
!$OMP DO schedule(static)
      do itheta = jstart_fmu,jend_fmu
        do ir = istart_fmu,iend_fmu
          Sfs_eq(ir,itheta,ivpar) = fs_eq_rtheta(ir,itheta)
        end do
      end do
!$OMP END DO
!$OMP END PARALLEL

      !*** computation of J0.fs_eq ***
      call compute_gyrofunction_2D(J0,geom,mu_id, &
        fs_eq_rtheta,J0fs_eq_rtheta)
!$OMP PARALLEL default(none) &
!$OMP private(ir,itheta) &
!$OMP shared(istart,iend,jstart,jend,ivpar, &
!$OMP  SJ0fs_eq,J0fs_eq_rtheta)
!$OMP DO schedule(static)
      do itheta = jstart,jend
        do ir = istart,iend
          SJ0fs_eq(ir,itheta,ivpar) = J0fs_eq_rtheta(ir,itheta)
        end do
      end do
!$OMP END DO
!$OMP END PARALLEL

      !********************************************************
      !*** computation of the radial equilibrium profiles:  ***
      !***   . n_eq(r) = \int J0.fseq Js dtheta dV          ***
      !***   . P_eq(r) = \int J0.fseq Epars Js dtheta dV    ***
      !***   . nGC_eq(r) = \int fseq Js dtheta dV           ***
      !***   . PGC_eq(r) = \int fseq Epars Js dtheta dV     ***
      !********************************************************
#ifdef _OPENMP
!$OMP PARALLEL default(none) &
!$OMP private(tid,ir,itheta, &
!$OMP  J0fs_eq_tmp,fs_eq_tmp,B_ij,energy, &
!$OMP  jacob_space_tmp,jacob_vel_tmp,coeff_intdtheta_tmp, &
!$OMP  int_Jsdtheta_dV_tmp) &
!$OMP shared(geom,init_magnet,init_curr, &
!$OMP  istart,iend,jstart,jend,ivpar, &
!$OMP  Romp1_0Nr,Romp2_0Nr,Romp3_0Nr,Romp4_0Nr, &
!$OMP  coeff_intdvpar_tmp,coeff_intdmu_tmp, &
!$OMP  J0fs_eq_rtheta,fs_eq_rtheta,jacobian_space,vparl2,mu_id)
!$OMP BARRIER
    tid = 1+omp_get_thread_num()
#else
    tid = 1
#endif
!$OMP DO schedule(static)
      do ir = 0,geom%Nr
        do itheta = 0,geom%Ntheta
          coeff_intdtheta_tmp = geom%coeff_intdtheta(itheta)
          !-> Energy
          B_ij   = init_magnet%B_norm(ir,itheta)
          energy = 0.5_RKIND*vparl2 + geom%mug(mu_id)*B_ij
          !-> jacobian space
          jacob_space_tmp     = jacobian_space(ir,itheta)
          !-> jacobian_velocity = 2*pi*Bstar
          call compute_noprecomp_jacobian_velocity( &
            geom,init_magnet,init_curr,ir,itheta,ivpar, &
            jacob_vel_tmp)
          int_Jsdtheta_dV_tmp = jacob_space_tmp * &
            coeff_intdtheta_tmp * jacob_vel_tmp * &
            coeff_intdvpar_tmp*coeff_intdmu_tmp 
          !--> J0.fs_eq
          J0fs_eq_tmp = J0fs_eq_rtheta(ir,itheta)
          !--> \int J0.fs_eq dv (saved in Romp1)
          Romp1_0Nr(tid)%val(ir) = Romp1_0Nr(tid)%val(ir) + &
            J0fs_eq_tmp * int_Jsdtheta_dV_tmp
          !--> \int[J0.fs_eq*(0.5*vpar^2+mu*B(r,theta))]dvpar (saved in Romp2)
          Romp2_0Nr(tid)%val(ir) = Romp2_0Nr(tid)%val(ir) + &
            J0fs_eq_tmp * energy * int_Jsdtheta_dV_tmp
          !--> fs_eq
          fs_eq_tmp = fs_eq_rtheta(ir,itheta)
          !--> \int fs_eq dv (saved in Romp3)
          Romp3_0Nr(tid)%val(ir) = Romp3_0Nr(tid)%val(ir) + &
            fs_eq_tmp * int_Jsdtheta_dV_tmp
          !--> \int[fs_eq*(0.5*vpar^2+mu*B(r,theta))]dvpar (saved in Romp4)
          Romp4_0Nr(tid)%val(ir) = Romp4_0Nr(tid)%val(ir) + &
            fs_eq_tmp * energy * int_Jsdtheta_dV_tmp
        end do
      end do
!$OMP END DO
!$OMP END PARALLEL
    end do
    
    !*** sum over the thread number ***
    do ir = 0,geom%Nr
      nseq_r_loc(ir)   = 0._RKIND
      Pseq_r_loc(ir)   = 0._RKIND
      nGCseq_r_loc(ir) = 0._RKIND
      PGCseq_r_loc(ir) = 0._RKIND
      do ithread = 1,Nbthread
        nseq_r_loc(ir)   = nseq_r_loc(ir) + Romp1_0Nr(ithread)%val(ir)
        Pseq_r_loc(ir)   = Pseq_r_loc(ir) + Romp2_0Nr(ithread)%val(ir)
        nGCseq_r_loc(ir) = nGCseq_r_loc(ir) + Romp3_0Nr(ithread)%val(ir)
        PGCseq_r_loc(ir) = PGCseq_r_loc(ir) + Romp4_0Nr(ithread)%val(ir)
      end do
    end do

    !*** integration in mu direction for computing ***
    !***  \int Jv d3v J0(sqrt(2*mu)).fs_eq         ***
    if (Nbproc_mu.ne.1) then
      call MY_Allreduce(nseq_r_loc,Snseq_r,(geom%Nr+1), &
        MPI_REAL8,MPI_SUM,mpi_comm_intermu,ierr)
      call MY_Allreduce(Pseq_r_loc,Pseq_r,(geom%Nr+1), &
        MPI_REAL8,MPI_SUM,mpi_comm_intermu,ierr)
      call MY_Allreduce(nGCseq_r_loc,nGCseq_r,(geom%Nr+1), &
        MPI_REAL8,MPI_SUM,mpi_comm_intermu,ierr)
      call MY_Allreduce(PGCseq_r_loc,PGCseq_r,(geom%Nr+1), &
        MPI_REAL8,MPI_SUM,mpi_comm_intermu,ierr)
    else
      do ir = 0,geom%Nr
        Snseq_r(ir)  = nseq_r_loc(ir)
        Pseq_r(ir)   = Pseq_r_loc(ir)
        nGCseq_r(ir) = nGCseq_r_loc(ir)
        PGCseq_r(ir) = PGCseq_r_loc(ir)
      end do
    end if

    !********************************************************
    !*** Computation of the radial equilibrium profiles:  ***
    !***   . n_eq(r) = <ns_eq(r,theta)>_theta             ***
    !***   . P_eq(r) = <P_eq(r,theta)>_theta              ***
    !***   . T_eq(r) = P_eq(r)/n_eq(r)                    ***
    !***  and the same kind of computation to compute the ***
    !**   guiding-center temperature TGC_eq               ***
    !*** Computation of                                   ***
    !***   . nbions_eq = \int J0.fs_eq Js dr dtheta dv    *** 
    !***               = \int n_eq dr                     ***
    !********************************************************
    press_coeff = 1._RKIND/(0.5_RKIND+float(ipow))
    Snbions_eq  = 0._RKIND
    do ir = 0,geom%Nr
      coeff_intdr_tmp  = geom%coeff_intdr(ir)
      Snbions_eq       = Snbions_eq + &
        Snseq_r(ir)*coeff_intdr_tmp*geom%Lphi
      inv_intdtheta_Js = 1._RKIND/intdtheta_Js(ir)
      !--> for particles
      nseq_tmp    = Snseq_r(ir)*inv_intdtheta_Js
      Pseq_tmp    = press_coeff*Pseq_r(ir)*inv_intdtheta_Js
      Snseq_r(ir) = nseq_tmp
      STseq_r(ir) = Pseq_tmp/nseq_tmp
      !--> for guiding-centers
      nGCseq_tmp    = nGCseq_r(ir)*inv_intdtheta_Js
      PGCseq_tmp    = press_coeff*PGCseq_r(ir)*inv_intdtheta_Js
      STGCseq_r(ir) = PGCseq_tmp/nGCseq_tmp
    end do

    !*** dn_eq/dr ***
    call deriv1(Snseq_r,Sdnseq_dr,geom%Nr,geom%dr,0)
    
    if (pglobal_id.eq.0) then
      print*,' min err(r-rbar) = ', min_err_rbar
      print*,' max err(r-rbar) = ', max_err_rbar
      write(6,'(2A45)') &
        '---------------------------------------------', &
        '---------------------------------------------'
    end if

   do tid = 1, Nbthread
   call FTR_DEALLOC(J0%omp_diag(tid)%val)
   call FTR_DEALLOC(J0%omp_ldiag(tid)%val)
   call FTR_DEALLOC(J0%omp_udiag(tid)%val)
   enddo

!R3 call r3_info_end (r3_info_index_1) !R3
    call Write_dyn_balise_close('Init_comp_equil') !R4
  end subroutine compute_equilibrium_part
end module fequil4d_module


