!**************************************************************
!  Copyright Euratom-CEA
!  Authors : 
!     Virginie Grandgirard (virginie.grandgirard@cea.fr)
!     Chantal Passeron (chantal.passeron@cea.fr)
!     Guillaume Latu (guillaume.latu@cea.fr)
!     Xavier Garbet (xavier.garbet@cea.fr)
!     Philippe Ghendrih (philippe.ghendrih@cea.fr)
!     Yanick Sarazin (yanick.sarazin@cea.fr)
!  
!  This code GYSELA (for GYrokinetic SEmi-LAgrangian) 
!  is a 5D gyrokinetic global full-f code for simulating 
!  the plasma turbulence in a tokamak.
!  
!  This software is governed by the CeCILL-B license 
!  under French law and abiding by the rules of distribution 
!  of free software.  You can  use, modify and redistribute 
!  the software under the terms of the CeCILL-B license as 
!  circulated by CEA, CNRS and INRIA at the following URL
!  "http://www.cecill.info". 
!**************************************************************

#include "config-F.inc"

!-------------------------------------------------------
! file : init_magnetic.F90
! date : 21/06/2000
!  Initialization of the magnetic configuration
!   vec_B = B0 R0/R(r,theta)*[zeta(r)*e_theta+e_phi]
!  with zeta(r) = r/(q R0)
!  and R(r,theta) = R0+h(r)*cos(theta) 
!  (Rk : h(r) = r but should be a profile vanishing
!    at the boundaries) 
!-------------------------------------------------------
module init_magnetic_class
  use prec_const
  use globals, only : TF_ripple  
  use wrap_alloc_module
  use geometry_class
  use coord_system_class, only : inv_R
  use init_current_class

  implicit none

  !*** init_magnet type definition ***
  type :: init_magnetic
    !-> covariant coordinates of B
    real(RKIND), dimension(:,:), pointer :: Br
    real(RKIND), dimension(:,:), pointer :: Btheta
    real(RKIND), dimension(:,:), pointer :: Bphi
    !-> contravariant coordinates of B
    real(RKIND), dimension(:,:), pointer :: B_gradr
    real(RKIND), dimension(:,:), pointer :: B_gradtheta
    real(RKIND), dimension(:,:), pointer :: B_gradphi
    !-> norm of B
    real(RKIND), dimension(:,:), pointer :: B_norm
    real(RKIND), dimension(:)  , pointer :: zeta
    real(RKIND), dimension(:)  , pointer :: b0
    !-> B(r,theta) = pi
    real(RKIND), dimension(:)  , pointer :: B_at_thetapi
    !-> Bmin and Bmax
    real(RKIND)                          :: Bmin
    real(RKIND)                          :: Bmax
    !-> partial derivatives of B(r,theta)
    real(RKIND), dimension(:,:), pointer :: dBdr
    real(RKIND), dimension(:,:), pointer :: dBdtheta
    !-> TF ripple (constant in r-theta)
    real(RKIND), dimension(:)  , pointer :: dBpardphi
  end type init_magnetic

  type :: precomp_traj
     real(RKIND) :: dt
     real(RKIND), dimension(:,:,:), pointer :: rfoot
     real(RKIND), dimension(:,:,:), pointer :: thfoot
  end type precomp_traj

  type(precomp_traj), dimension(:), pointer :: precompadv

  integer,     parameter :: NBPREC = 2

  include "gyrokinetic/Bstar_inline.h"
  !******************************
  contains
  !******************************
#include "gyrokinetic/Bstar_inline.F90"
     
  !---------------------------------------------------- 
  ! Constructor for magnetic field configuration 
  !----------------------------------------------------
  subroutine new_magnet_config(mthis,geom)
    use globals
    type(init_magnetic), intent(out) :: mthis
    type(geometry)     , intent(in)  :: geom

    integer :: i

    !-> allocation for B covariant coordinates and B norm
    call FTR_ALLOC_2_ZERO(mthis%B_norm,0,geom%Nr,0,geom%Ntheta)
    call FTR_ALLOC_2_ZERO(mthis%Br,0,geom%Nr,0,geom%Ntheta)
    call FTR_ALLOC_2_ZERO(mthis%Btheta,0,geom%Nr,0,geom%Ntheta)
    call FTR_ALLOC_2_ZERO(mthis%Bphi,0,geom%Nr,0,geom%Ntheta)
    call FTR_ALLOC_1_ZERO(mthis%zeta,0,geom%Nr)
    call FTR_ALLOC_1_ZERO(mthis%b0,0,geom%Nr)
    !-> allocation for the partial derivatives of B
    call FTR_ALLOC_2_ZERO(mthis%dBdr,0,geom%Nr,0,geom%Ntheta)
    call FTR_ALLOC_2_ZERO(mthis%dBdtheta,0,geom%Nr,0,geom%Ntheta)
    !-> allocation for the contravariant components of B
    call FTR_ALLOC_2_ZERO(mthis%B_gradr,0,geom%Nr,0,geom%Ntheta)
    call FTR_ALLOC_2_ZERO(mthis%B_gradtheta,0,geom%Nr,0,geom%Ntheta)
    call FTR_ALLOC_2_ZERO(mthis%B_gradphi,0,geom%Nr,0,geom%Ntheta)
    call FTR_ALLOC_1_ZERO(mthis%B_at_thetapi,0,geom%Nr)
    !-> allocation for the computation of Bstar and
    !   the contravariant components of bstar (bstar=Bstar/B)
    if (.not.transp_Bstar) then
      call FTR_ALLOC_3_ZERO(Bstar_PNrPNthNvpar,0,geom%Nr,0,geom%Ntheta,0,geom%Nvpar)
      call FTR_ALLOC_3_ZERO(bstar_gradr_PNrPNthNvpar,0,geom%Nr,0,geom%Ntheta,0,geom%Nvpar)
      call FTR_ALLOC_3_ZERO(bstar_gradtheta_PNrPNthNvpar,0,geom%Nr,0,geom%Ntheta,0,geom%Nvpar)
      call FTR_ALLOC_3_ZERO(bstar_gradphi_PNrPNthNvpar,0,geom%Nr,0,geom%Ntheta,0,geom%Nvpar)
      Bstar_NrNthPNvpar           => Bstar_PNrPNthNvpar
      bstar_gradr_NrNthPNvpar     => bstar_gradr_PNrPNthNvpar
      bstar_gradtheta_NrNthPNvpar => bstar_gradtheta_PNrPNthNvpar
      bstar_gradphi_NrNthPNvpar   => bstar_gradphi_PNrPNthNvpar
    else
      call FTR_ALLOC_3_ZERO(Bstar_PNrPNthNvpar,istart_buf,iend_buf,0,geom%Ntheta,0,geom%Nvpar)
      call FTR_ALLOC_3_ZERO(bstar_gradr_PNrPNthNvpar,istart_buf,iend_buf,0,geom%Ntheta,0,geom%Nvpar)
      call FTR_ALLOC_3_ZERO(bstar_gradtheta_PNrPNthNvpar,istart_buf,iend_buf,0,geom%Ntheta,0,geom%Nvpar)
      call FTR_ALLOC_3_ZERO(bstar_gradphi_PNrPNthNvpar,istart_buf,iend_buf,0,geom%Ntheta,0,geom%Nvpar)
      call FTR_ALLOC_3_ZERO(Bstar_NrNthPNvpar,0,geom%Nr,0,geom%Ntheta,lstart,lend)
      call FTR_ALLOC_3_ZERO(bstar_gradr_NrNthPNvpar,0,geom%Nr,0,geom%Ntheta,lstart,lend)
      call FTR_ALLOC_3_ZERO(bstar_gradtheta_NrNthPNvpar,0,geom%Nr,0,geom%Ntheta,lstart,lend)
      call FTR_ALLOC_3_ZERO(bstar_gradphi_NrNthPNvpar,0,geom%Nr,0,geom%Ntheta,lstart,lend)
    end if
    if (avoidtaylor) then
      allocate(precompadv(1:NBPREC))
      do i = 1, NBPREC
        precompadv(i)%dt = 0.
        call FTR_ALLOC_3_ZERO(precompadv(i)%rfoot,0,Nr,0,Ntheta-1,lstart,lend)
        call FTR_ALLOC_3_ZERO(precompadv(i)%thfoot,0,Nr,0,Ntheta-1,lstart,lend)
      end do
    end if
    if (TF_ripple) then
      !-> allocation for TF ripple
      call FTR_ALLOC_1_ZERO(mthis%dBpardphi,0,geom%Nphi)
    end if
  end subroutine new_magnet_config
  

  !---------------------------------------------------- 
  ! destructor for magnetic field configuration 
  !---------------------------------------------------- 
  subroutine del_magnet_config(mthis)
    use globals
    type(init_magnetic), intent(inout) :: mthis

    integer :: i

    !-> deallocation of B covariant coordinates and B norm
    call FTR_DEALLOC(mthis%B_norm)
    call FTR_DEALLOC(mthis%B_at_thetapi)
    call FTR_DEALLOC(mthis%Br)
    call FTR_DEALLOC(mthis%Btheta)
    call FTR_DEALLOC(mthis%Bphi)
    call FTR_DEALLOC(mthis%zeta)
    call FTR_DEALLOC(mthis%b0)
    !-> deallocation of the partial derivatives of B
    call FTR_DEALLOC(mthis%dBdr)
    call FTR_DEALLOC(mthis%dBdtheta)
    !-> deallocation of the contravariant components of B
    call FTR_DEALLOC(mthis%B_gradr)
    call FTR_DEALLOC(mthis%B_gradtheta)
    call FTR_DEALLOC(mthis%B_gradphi)
    !-> deallocation of the computation of Bstar and
    !   the contravariant components of bstar (bstar=Bstar/B)
    call FTR_DEALLOC(Bstar_PNrPNthNvpar)
    call FTR_DEALLOC(bstar_gradr_PNrPNthNvpar)
    call FTR_DEALLOC(bstar_gradtheta_PNrPNthNvpar)
    call FTR_DEALLOC(bstar_gradphi_PNrPNthNvpar)
    if (transp_Bstar) then
      call FTR_DEALLOC(Bstar_NrNthPNvpar)
      call FTR_DEALLOC(bstar_gradr_NrNthPNvpar)
      call FTR_DEALLOC(bstar_gradtheta_NrNthPNvpar)
      call FTR_DEALLOC(bstar_gradphi_NrNthPNvpar)
    end if
    if (avoidtaylor) then
      do i = 1, NBPREC
        call FTR_DEALLOC(precompadv(i)%rfoot)
        call FTR_DEALLOC(precompadv(i)%thfoot)
      end do
      deallocate(precompadv)
    end if
    if (TF_ripple) then
      !-> deallocation for TF ripple
      call FTR_DEALLOC(mthis%dBpardphi)
    end if
  end subroutine del_magnet_config


  !--------------------------------------------------------
  ! - initialization of zeta(r) = r/(q R0)
  !--------------------------------------------------------
  subroutine init_zeta(mthis,geom,init_prof)
    use globals, only : R0
    use init_profile_class
    type(init_magnetic), intent(inout) :: mthis  
    type(geometry)     , intent(in)    :: geom
    type(init_profile) , intent(in)    :: init_prof

    integer     :: ir
    real(RKIND) :: zeta_ri

    !*** initialisation of zeta(r) = r/(q R0) ***
    do ir = 0,geom%Nr
      zeta_ri        = geom%rg(ir)*init_prof%iota(ir)/R0
      mthis%zeta(ir) = zeta_ri
    end do
  end subroutine init_zeta


  !---------------------------------------------------- 
  ! - initialization of the covariant coordinates of
  ! the magnetic field, i.e:
  !   . Br(r,theta)     = 0
  !   . Btheta(r,theta) = B0*R0*r/R*zeta(r)
  !   . Bphi(r,theta)   = B0*R0
  ! because the magnetic field is defined as:
  !   . vec_B = B0R0/R*(zeta*vec_etheta+vec_ephi)
  !       with zeta(r) = r/(q R0)
  !---------------------------------------------------- 
  subroutine init_Bcovariants(mthis,geom)
    use globals, only : R0, magnetic_drift
    type(init_magnetic), intent(inout) :: mthis  
    type(geometry)     , intent(in)    :: geom

    integer     :: ir, itheta
    real(RKIND) :: ri, inv_Rij, zeta_ri

    if (magnetic_drift) then
      do itheta = 0,geom%Ntheta
        do ir = 0,geom%Nr
          ri                      = geom%rg(ir)
          inv_Rij                 = inv_R(ir,itheta)
          zeta_ri                 = mthis%zeta(ir)
          mthis%Br(ir,itheta)     = 0._RKIND
          mthis%Btheta(ir,itheta) = R0*inv_Rij*ri*zeta_ri 
          mthis%Bphi(ir,itheta)   = R0
        end do
      end do
    else
      do itheta = 0,geom%Ntheta
        do ir = 0,geom%Nr
          mthis%Br(ir,itheta)     = 0._RKIND
          mthis%Btheta(ir,itheta) = 0._RKIND
          mthis%Bphi(ir,itheta)   = R0
        end do
      end do
    end if
  end subroutine init_Bcovariants


  !------------------------------------------------
  ! - initialization of the partial derivatives
  !   of the magnetic field 
  !    . dBdr(r,theta)
  !    . dBdtheta(r,theta)
  !------------------------------------------------
  subroutine init_Bderivatives(mthis,geom,init_prof)
    use globals     , only : magnetic_drift
    use utils_module, only : deriv1
    use init_profile_class
    type(init_magnetic), intent(inout) :: mthis  
    type(geometry)     , intent(in)    :: geom
    type(init_profile) , intent(in)    :: init_prof
    
    integer     :: ir, itheta
    real(RKIND) :: ri, hri, dhri_dr, Rij
    real(RKIND), dimension(0:geom%Nr)     :: Br_tmp
    real(RKIND), dimension(0:geom%Nr)     :: dBdr_tmp
    real(RKIND), dimension(0:geom%Ntheta) :: Btheta_tmp
    real(RKIND), dimension(0:geom%Ntheta) :: dBdtheta_tmp

    !*** Initialization of B ***
    !-> computation of dB/dr
    do itheta = 0,geom%Ntheta
      do ir = 0,geom%Nr
        Br_tmp(ir) = mthis%B_norm(ir,itheta)
      end do
      call deriv1(Br_tmp,dBdr_tmp,geom%Nr,geom%dr,0)
      do ir = 0,geom%Nr
        mthis%dBdr(ir,itheta) = dBdr_tmp(ir)
      end do
    end do
    !-> computation of dB/dtheta
    do ir = 0,geom%Nr
      do itheta = 0,geom%Ntheta
        Btheta_tmp(itheta) = mthis%B_norm(ir,itheta)
      end do
      call deriv1(Btheta_tmp(0:Ntheta-1),dBdtheta_tmp(0:Ntheta-1),&
        geom%Ntheta-1,geom%dtheta,1)
      do itheta = 0,geom%Ntheta-1
        mthis%dBdtheta(ir,itheta) = dBdtheta_tmp(itheta)
      end do
      mthis%dBdtheta(ir,geom%Ntheta) = mthis%dBdtheta(ir,0)
    end do
  end subroutine init_Bderivatives


  !---------------------------------------------------- 
  ! Initialization of the magnetic configuration
  !  - initialisation of covariant components,
  !  - computation of the contravariant components,
  !  - computation of the norm,
  !  - computation of the partial derivatives and
  !
  ! Computation of jacobian in the velocity space
  !----------------------------------------------------
  subroutine init_magnetic_config(mthis,geom, &
    coord_sys,init_prof,init_curr)
    use globals     , only : memory_test, major_species
    use utils_module, only : locate
    use init_profile_class
    use coord_system_class
!R3 #include "r3_info.h" !R3
    type(init_magnetic), intent(inout) :: mthis
    type(geometry)     , intent(in)    :: geom
    type(coord_system) , intent(in)    :: coord_sys 
    type(init_profile) , intent(in)    :: init_prof
    type(init_current) , intent(in)    :: init_curr

    integer           :: itheta_pi, ir
    real(RKIND)       :: Ts_max, mu_max, err_threshold
    character(LEN=50) :: subp

!R3 call r3_info_begin (r3_info_index_1, 'Init_magnetic') !R3
    call Write_dyn_balise_open('Init_magnetic') !R4

    !*** allocation of the arrays ***
    call new_magnet_config(mthis,geom)
    if (.not. memory_test) then
      !*** initialisation of the magnetic field B ***
      !-> initialisation of the covariant components of B
      call init_zeta(mthis,geom,init_prof)
      call init_Bcovariants(mthis,geom)
      !-> computation of the norm of B
      call compute_norm(coord_sys,mthis%Br, &
        mthis%Btheta,mthis%Bphi,mthis%B_norm)
      !-> computation of the contravariant components of the
      !->  unit vector vec_b (with vec_b= vec_B/B)
      call compute_contravariant_vector(coord_sys, &
        mthis%Br,mthis%Btheta,mthis%Bphi, &
        mthis%B_gradr,mthis%B_gradtheta,mthis%B_gradphi)
      !-> computation of the partial derivatives of B
      call init_Bderivatives(mthis,geom,init_prof)
      if (TF_ripple) then
        !-> compute TF ripple
        call init_dBpar(mthis,geom)
      end if
      !-> computation of the min and the max of B
      mthis%Bmin = minval(mthis%B_norm)
      mthis%Bmax = maxval(mthis%B_norm)
      !-> cheking of mu_max 
      if (geom%Nmu.ne.0) then
        err_threshold = 1.e-5_RKIND
        Ts_max        = init_prof%Ts0(major_species,0)
        mu_max        = -Ts_max * &
          mthis%Bmin*log(err_threshold)
        if (geom%mug(geom%Nmu).lt.mu_max) then
          if (pglobal_id.eq.0) then
            print*,'---> REMARK : max(mu) = ', &
              geom%mug(geom%Nmu),' < ',mu_max
          end if
        end if
      end if

      !*** initialisation of B(r,theta=pi) values ***
      subp = " init_magnet_config "//char(0)
      call locate(PI,geom%thetag,geom%Ntheta,geom%dtheta, &
        subp,itheta_pi)
      do ir = 0,geom%Nr
        mthis%B_at_thetapi(ir) = mthis%B_norm(ir,itheta_pi)
      end do
    endif
    call Write_dyn_balise_close('Init_magnetic') !R4
!R3 call r3_info_end (r3_info_index_1) !R3
  end subroutine init_magnetic_config


  !---------------------------------------------------------
  ! Initialisation of a perturbed magnetic field dBpar
  !---------------------------------------------------------
  subroutine init_dBpar(mthis,geom)
    use globals, only : delta_ripple, N_ripple
    type(init_magnetic), intent(inout) :: mthis  
    type(geometry)     , intent(in)    :: geom

    integer :: iphi
    
    do iphi = 0, geom%Nphi
      ! Compute only the phi-derivative of the perturbation
      ! (constant ripple amplitude in r-theta)
      mthis%dBpardphi(iphi) = N_ripple * &
        delta_ripple*sin(N_ripple*geom%dphi*iphi)
    end do
  end subroutine init_dBpar


  !*********************************************************
  ! Computation of quantities associated to Bstar
  !*********************************************************
  !---------------------------------------------------------
  ! Computation of the scalar product mu0*vec_J.vec_b where
  !  the scalar product is defined as 
  !    vec_J.vec_b = 1/B*(J_gradx1*B_gradx1 + 
  !                  J_gradx2*B_gradx2 + J_gradx3*B_gradx3) 
  !
  ! Rk: This scalar product appears in the 
  !     definition of Bstar as:
  !  Bstar = B_norm + 
  !     (sqrt{hat_As}/{hat_Zs})*vpar*mu0*vec_J.vec_b/B_norm
  !---------------------------------------------------------
  subroutine compute_scalprod_mu0Jb(geom,mthis,init_curr, &
    ir,itheta,Sscalprod_mu0Jb)
    type(geometry)     , intent(in)  :: geom
    type(init_magnetic), intent(in)  :: mthis
    type(init_current) , intent(in)  :: init_curr
    integer            , intent(in)  :: ir
    integer            , intent(in)  :: itheta
    real(RKIND)        , intent(out) :: Sscalprod_mu0Jb
    
    real(RKIND) :: Bnorm_ij
    
    !*** compute the scalar product of mu0*vec_J ***
    !***  and vec_b                              ***
    !-> B norm
    Bnorm_ij = mthis%B_norm(ir,itheta) 
    !-> mu0*vec_J.vec_B
    Sscalprod_mu0Jb = &
      mthis%B_gradr(ir,itheta) * &
      init_curr%mu0_J_gradr(ir,itheta) + &
      mthis%B_gradtheta(ir,itheta) * &
      init_curr%mu0_J_gradtheta(ir,itheta) + &
      mthis%B_gradphi(ir,itheta) * &
      init_curr%mu0_J_gradphi(ir,itheta)
    !-> mu0*vec_J.vec_b = mu0*vec_J.vec_B/B
    Sscalprod_mu0Jb = Sscalprod_mu0Jb/Bnorm_ij
  end subroutine compute_scalprod_mu0Jb


  !-------------------------------------------------------
  ! Computation of Bstar 
  !   at the mesh point (ir,itheta,ivpar), as :
  !  Bstar = B_norm + 
  !     (sqrt{hat_As}/hat_Zs)*vpar*mu0*vec_J.vec_b/B_norm
  !-------------------------------------------------------
  subroutine compute_Bstar(geom,mthis,init_curr, &
    ir,itheta,ivpar,SBstar)
    use globals, only : pSpecies_id, hat_As, hat_Zs
    type(geometry)     , intent(in)  :: geom
    type(init_magnetic), intent(in)  :: mthis
    type(init_current) , intent(in)  :: init_curr
    integer            , intent(in)  :: ir
    integer            , intent(in)  :: itheta
    integer            , intent(in)  :: ivpar
    real(RKIND)        , intent(out) :: SBstar

    real(RKIND) :: vparl, Bnorm_ij
    real(RKIND) :: scalprod_mu0Jb_tmp
    real(RKIND) :: hatZs, sqrt_hatAs

    hatZs      = hat_Zs(pSpecies_id)
    sqrt_hatAs = sqrt(hat_As(pSpecies_id))
    vparl      = geom%vparg(ivpar)
    !-> B norm
    Bnorm_ij = mthis%B_norm(ir,itheta) 
    !-> computation of the scalar product of vec_J and vec_b
    !->  as vec_J.vec_b = 1/B*(J_gradx1*B_gradx1 + 
    !->  J_gradx2*B_gradx2 + J_gradx3*B_gradx3) 
    call compute_scalprod_mu0Jb(geom,mthis,init_curr, &
      ir,itheta,scalprod_mu0Jb_tmp)
    !-> computation of Bstar
    SBstar = Bnorm_ij + (sqrt_hatAs/hatZs) * &
      vparl*scalprod_mu0Jb_tmp/Bnorm_ij
  end subroutine compute_Bstar


  !---------------------------------------------------- 
  ! Computation of the jacobian in the velocity 
  !  space, which is equal to :
  !    . 1 in the 4D case 
  !    . 2*pi*Bstar(r,theta,vpar) in the 5D case
  !----------------------------------------------------   
  subroutine compute_jacobian_velocity(geom,mthis, &
    init_curr,ir,itheta,ivpar,Sjacobian_velocity)
    use globals, only : mumin
    type(geometry)     , intent(in)  :: geom
    type(init_magnetic), intent(in)  :: mthis
    type(init_current) , intent(in)  :: init_curr
    integer            , intent(in)  :: ir
    integer            , intent(in)  :: itheta
    integer            , intent(in)  :: ivpar
    real(RKIND)        , intent(out) :: Sjacobian_velocity

    real(RKIND) :: Bstar_tmp

    if (geom%Nmu.ne.0) then
#ifdef NO_PRECOMP_BSTAR 
      call compute_Bstar(geom,mthis,init_curr, &
        ir,itheta,ivpar,Bstar_tmp)
#else
      call precomp_Bstar(ir,itheta,ivpar,Bstar_tmp)
#endif
      Sjacobian_velocity = TWOPI*Bstar_tmp
    else
      !-> 4D case
      Sjacobian_velocity = 1._RKIND
    end if
  end subroutine compute_jacobian_velocity


  !---------------------------------------------------- 
  ! Computation of the jacobian in the velocity 
  !  space, which is equal to :
  !    . 1 in the 4D case 
  !    . 2*pi*Bstar(r,theta,vpar) in the 5D case
  !----------------------------------------------------   
  subroutine compute_noprecomp_jacobian_velocity(geom,mthis, &
    init_curr,ir,itheta,ivpar,Sjacobian_velocity)
    use globals, only : mumin
    type(geometry)     , intent(in)  :: geom
    type(init_magnetic), intent(in)  :: mthis
    type(init_current) , intent(in)  :: init_curr
    integer            , intent(in)  :: ir
    integer            , intent(in)  :: itheta
    integer            , intent(in)  :: ivpar
    real(RKIND)        , intent(out) :: Sjacobian_velocity

    real(RKIND) :: Bstar_tmp

    if (geom%Nmu.ne.0) then
      call compute_Bstar(geom,mthis,init_curr, &
        ir,itheta,ivpar,Bstar_tmp)
      Sjacobian_velocity = TWOPI*Bstar_tmp
    else
      !-> 4D case
      Sjacobian_velocity = 1._RKIND
    end if
  end subroutine compute_noprecomp_jacobian_velocity
  

  !-------------------------------------------------------
  !  Computes the contravariant components of bstar
  !  (i.e bstar_gradxi) are defined as :
  !    bstar_gradxi = 1/Bstar* [B_gradxi + 
  !       (sqrt{hat_As}/hat_Zs)*vpar*mu0*J_gradxi/B_norm]
  !-------------------------------------------------------
  subroutine compute_bstar_contravariant(geom, &
    mthis,init_curr,ir,itheta,ivpar, &
    Sbstar_gradx1,Sbstar_gradx2,Sbstar_gradx3)
    use globals, only : pSpecies_id, hat_As, hat_Zs
    type(geometry)       , intent(in)  :: geom
    type(init_magnetic)  , intent(in)  :: mthis
    type(init_current)   , intent(in)  :: init_curr
    integer              , intent(in)  :: ir
    integer              , intent(in)  :: itheta
    integer              , intent(in)  :: ivpar
    real(RKIND), optional, intent(out) :: Sbstar_gradx1
    real(RKIND), optional, intent(out) :: Sbstar_gradx2
    real(RKIND), optional, intent(out) :: Sbstar_gradx3
    
    real(RKIND) :: vparl
    real(RKIND) :: Bij, Bstar_ijl
    real(RKIND) :: sqrt_hatAs, hatZs

    vparl      = geom%vparg(ivpar)
    Bij        = mthis%B_norm(ir,itheta)
    hatZs      = hat_Zs(pSpecies_id)
    sqrt_hatAs = sqrt(hat_As(pSpecies_id))
    call compute_Bstar(geom,mthis,init_curr, &
      ir,itheta,ivpar,Bstar_ijl)
    !-> bstar_gradxi = 1/Bstar* [ b_gradxi + 
    !    (sqrt{hat_As}/hat_Zs)*vpar*mu0*J_gradxi/B_norm ]
    if (present(Sbstar_gradx1)) then
      Sbstar_gradx1 = ( mthis%B_gradr(ir,itheta) + &
        (sqrt_hatAs/hatZs)*vparl* &
        init_curr%mu0_J_gradr(ir,itheta)/Bij )/Bstar_ijl
    end if
    if (present(Sbstar_gradx2)) then
      Sbstar_gradx2 = ( mthis%B_gradtheta(ir,itheta) + &
        (sqrt_hatAs/hatZs)*vparl* &
        init_curr%mu0_J_gradtheta(ir,itheta)/Bij )/Bstar_ijl
    end if
    if (present(Sbstar_gradx3)) then
      Sbstar_gradx3 = ( mthis%B_gradphi(ir,itheta) + &
        (sqrt_hatAs/hatZs)*vparl* &
        init_curr%mu0_J_gradphi(ir,itheta)/Bij )/Bstar_ijl
    end if
  end subroutine compute_bstar_contravariant


  !-------------------------------------------------------
  !  Computes and save in 3D arrays:
  !    - Bstar and
  !    - the contravariant components of bstar
  ! (where bstar=Bstar/B)
  !-------------------------------------------------------
  subroutine init_precompute_Bstar(init_magnet, &
    init_curr,geom,coord_sys,init_prof)
    use globals
    use init_profile_class
!R3 #include "r3_info.h" !R3
    type(init_magnetic), intent(in) :: init_magnet
    type(init_current) , intent(in) :: init_curr
    type(geometry)     , intent(in) :: geom
    type(coord_system) , intent(in) :: coord_sys 
    type(init_profile) , intent(in) :: init_prof
    
    integer     :: ir, itheta, ivpar
    real(RKIND) :: ri, bstar_ijl
    real(RKIND) :: bstar_gradr_tmp, bstar_gradth_tmp
    real(RKIND) :: bstar_gradphi_tmp

!R3 call r3_info_begin (r3_info_index_1, 'Init_precomp_b*') !R3
    call Write_dyn_balise_open('Init_precomp_b*') !R4
    !*** initialization of the covariant coordinates ***
    if (.not.memory_test) then
      do ivpar = 0, geom%Nvpar
        do itheta = 0, geom%Ntheta
          do ir = 0, geom%Nr
            call compute_Bstar(geom,init_magnet,init_curr, &
              ir,itheta,ivpar,Bstar_ijl)
            call compute_bstar_contravariant(geom, &
              init_magnet,init_curr,ir,itheta,ivpar, &
              Sbstar_gradx1=bstar_gradr_tmp, &
              Sbstar_gradx2=bstar_gradth_tmp, &
              Sbstar_gradx3=bstar_gradphi_tmp)
            if (.not. transp_Bstar) then
              Bstar_PNrPNthNvpar(ir,itheta,ivpar)           = &
                Bstar_ijl
              bstar_gradr_PNrPNthNvpar(ir,itheta,ivpar)     = &
                bstar_gradr_tmp
              bstar_gradtheta_PNrPNthNvpar(ir,itheta,ivpar) = &
                bstar_gradth_tmp
              bstar_gradphi_PNrPNthNvpar(ir,itheta,ivpar)   = &
                bstar_gradphi_tmp
            else
              if (ir.ge.istart_buf .and. ir.le.iend_buf) then
                Bstar_PNrPNthNvpar(ir,itheta,ivpar)           = &
                  Bstar_ijl
                bstar_gradr_PNrPNthNvpar(ir,itheta,ivpar)     = &
                  bstar_gradr_tmp
                bstar_gradtheta_PNrPNthNvpar(ir,itheta,ivpar) = &
                  bstar_gradth_tmp
                bstar_gradphi_PNrPNthNvpar(ir,itheta,ivpar)   = &
                  bstar_gradphi_tmp
              endif
              if (ivpar.ge.lstart .and. ivpar.le.lend) then
                Bstar_NrNthPNvpar(ir,itheta,ivpar)           = &
                  Bstar_ijl
                bstar_gradr_NrNthPNvpar(ir,itheta,ivpar)     = &
                  bstar_gradr_tmp
                bstar_gradtheta_NrNthPNvpar(ir,itheta,ivpar) = &
                  bstar_gradth_tmp
                bstar_gradphi_NrNthPNvpar(ir,itheta,ivpar)   = &
                  bstar_gradphi_tmp
              end if
            end if
          end do
        end do
      end do
    end if
    call Write_dyn_balise_close('Init_precomp_b*') !R4
!R3 call r3_info_end (r3_info_index_1) !R3
  end subroutine init_precompute_bstar


  !-------------------------------------------------------
  ! 
  !-------------------------------------------------------
  subroutine create_precomp_fields(geom,mag,pt_precomp)
    use globals, only : lstart, lend, mpi_comm_mu, plocal_id, &
      pglobal_id, mu_id
    use MPIutils_module
#ifndef NO_HDF5
    use HDF5
    use HDF5_module
#endif
    use mpiMod
    type(geometry)     , intent(in) :: geom
    type(init_magnetic), intent(in) :: mag
    type(precomp_traj) , pointer    :: pt_precomp

#ifndef NO_HDF5
    integer           :: Nbr, Nbc, icr
    character(LEN=50) :: init_state_file
    character(LEN=30) :: strn
    integer(HID_T)    :: file_id   
    integer           :: ierr, j, ivpar, itheta, ir
    integer           :: dim(1:3), nbelts
    real(RKIND)       :: rhostar
    type(precomp_traj) :: read_tmp, pt_read 
    
    call FTR_ALLOC_3_ZERO(read_tmp%rfoot,0,geom%Nr,0,geom%Ntheta,0,geom%Nvpar)
    call FTR_ALLOC_3_ZERO(read_tmp%thfoot,0,geom%Nr,0,geom%Ntheta,0,geom%Nvpar)
    call FTR_ALLOC_3_ZERO(pt_read%rfoot,0,geom%Nr,0,geom%Ntheta,0,geom%Nvpar)
    call FTR_ALLOC_3_ZERO(pt_read%thfoot,0,geom%Nr,0,geom%Ntheta,0,geom%Nvpar)
    do ivpar = 0,geom%Nvpar
      do itheta = 0,geom%Ntheta
        do ir = 0,geom%Nr
          read_tmp%rfoot (ir,itheta,ivpar) = -1.e30_8
          read_tmp%thfoot(ir,itheta,ivpar) = -1.e30_8
        end do
      end do
    end do
    do ivpar = lstart,lend
      do itheta = 0,geom%Ntheta-1
        do ir = 0,geom%Nr
          read_tmp%rfoot(ir,itheta,ivpar) = &
            pt_precomp%rfoot(ir,itheta,ivpar)
          read_tmp%thfoot(ir,itheta,ivpar) = &
            pt_precomp%thfoot(ir,itheta,ivpar)
        end do
      end do
    end do
    nbelts = (geom%Nr+1)*(geom%Ntheta+1)*(geom%Nvpar+1)
    CALL MY_Reduce(read_tmp%rfoot(0,0,0),pt_read%rfoot(0,0,0),&
      nbelts,MPI_REAL8,MPI_MAX,0,MPI_COMM_MU,ierr)
    CALL MY_Reduce(read_tmp%thfoot(0,0,0),pt_read%thfoot(0,0,0),&
      nbelts,MPI_REAL8,MPI_MAX,0,MPI_COMM_MU,ierr)
    pt_read%dt = pt_precomp%dt
    call ppbarrier()
    if (plocal_id .eq. 0) then
      print *,"REDUCE done", pglobal_id, pt_precomp%dt
      write(strn,'(A,I10.10)') "_",nint(pt_precomp%dt*1000)
      
     write(init_state_file,'(A,I2.2,A11,A)') &
       "precomp_",mu_id,strn,".h5"
     call HDF5_create(trim(init_state_file),file_id,ierr)
     if (ierr.ne.0) then
       print*,'pglobal_id = ',pglobal_id, &
         ' ==> error for opening of ',init_state_file
     end if
     call HDF5_save(file_id,geom%Nr,'Nr'//char(0))
     call HDF5_save(file_id,geom%Ntheta,'Ntheta'//char(0))
     call HDF5_save(file_id,geom%Nvpar,'Nvpar'//char(0))
     
     dim(1) = geom%Nr+1
     dim(2) = geom%Ntheta+1
     dim(3) = geom%Nvpar+1
     
     call HDF5_save(file_id,pt_read%dt,'dt'//char(0))
     call HDF5_save(file_id,pt_read%rfoot, &
       dim(1),dim(2),dim(3),'rfoot'//char(0))
     call HDF5_save(file_id,pt_read%thfoot, &
       dim(1),dim(2),dim(3),'thfoot'//char(0))
     call HDF5_close(file_id)
   end if
   call FTR_DEALLOC(read_tmp%rfoot)
   call FTR_DEALLOC(read_tmp%thfoot)
   call FTR_DEALLOC(pt_read%rfoot)
   call FTR_DEALLOC(pt_read%thfoot)
#endif
  end subroutine create_precomp_fields


  !---------------------------------------------- 
  ! Output saving of the initial state 
  !  -> directly in HDF5 binary format
  !----------------------------------------------   
  subroutine read_precomp_fields(geom,mag,dt,pt_precomp,error)
    use globals, only : lstart, lend, mpi_comm_mu, plocal_id, &
      pglobal_id, mu_id
    use MPIutils_module
#ifndef NO_HDF5
    use HDF5
    use HDF5_module
    use mpiMod
#endif
    type(geometry)     , intent(in)  :: geom
    type(init_magnetic), intent(in)  :: mag
    real(RKIND)        , intent(in)  :: dt
    type(precomp_traj) , pointer     :: pt_precomp
    integer            , intent(out) :: error

#ifndef NO_HDF5
    integer           :: Nbr, Nbc, icr
    character(LEN=50) :: init_state_file
    character(LEN=30) :: strn
    integer(HID_T)    :: file_id   
    integer           :: j
    real(RKIND)       :: rhostar
    integer           :: lnr, lntheta, lnvpar, nbelts
    integer           :: ir, itheta, ivpar, ierr
    type(precomp_traj), target  :: read_tmp
    
    LOGICAL :: file_exists

    call ppbarrier()
    write(strn,'(A,I10.10)') "_",nint(dt*1000)
    write(init_state_file,'(A,I2.2,A11,A)') &
      "precomp_",mu_id,strn,".h5"
    if (plocal_id .eq. 0) then
      INQUIRE(FILE=trim(init_state_file), EXIST=file_exists) 
      if (.not. file_exists) then
        error = -1
      else
        call HDF5_open(trim(init_state_file),file_id,error)
        if (error .ne. 0) then
          write(*,'(A,A,A)')"FILE ", init_state_file, " not found"
        else
          write(*,'(A,A,A)')"FILE ", init_state_file, " found"
          call HDF5_load(file_id,lNr,'Nr')
          call HDF5_load(file_id,lNtheta,'Ntheta')
          call HDF5_load(file_id,lNvpar,'Nvpar')
          if (lNr .ne. geom%Nr .or. lNvpar .ne. geom%Nvpar) then
            print *, "Problem, lNr.ne.Nr .or. lNvpar.ne.Nvpar"
            print *, "ptest1 Nr", pglobal_id, lNr, lNtheta, lNvpar
            print *, "ptest1 Now", pglobal_id, geom%Nr, geom%Ntheta, geom%Nvpar
            STOP
          end if
          if (geom%Ntheta .ne. lNtheta) then
            print *, "ptest2 Nr", pglobal_id, lNr, lNtheta, lNvpar
            print *, "ptest2 Now", pglobal_id, geom%Nr, geom%Ntheta, geom%Nvpar
            STOP
          end if
        end if
      end if
    end if
    CALL MPI_BCAST(error,1,MPI_INTEGER,0,MPI_COMM_MU,ierr)
    if (error .eq. 0) then
      write(*,'(A,A,A)')"FILE ", init_state_file, " read"
      allocate(read_tmp%rfoot(0:geom%Nr,0:geom%Ntheta,0:geom%Nvpar))
      allocate(read_tmp%thfoot(0:geom%Nr,0:geom%Ntheta,0:geom%Nvpar))
      if (plocal_id .eq. 0) then
        call HDF5_load(file_id,read_tmp%dt,&
          'dt'//char(0))
        call HDF5_load(file_id,read_tmp%rfoot, &
          'rfoot'//char(0))
        call HDF5_load(file_id,read_tmp%thfoot, &
          'thfoot'//char(0))
        call HDF5_close(file_id)
      end if
      nbelts = (geom%Nr+1)*(geom%Ntheta+1)*(geom%Nvpar+1)
      CALL MPI_BCAST(read_tmp%rfoot,nbelts,MPI_REAL8,0,MPI_COMM_MU,ierr)
      CALL MPI_BCAST(read_tmp%thfoot,nbelts,MPI_REAL8,0,MPI_COMM_MU,ierr)
      CALL MPI_BCAST(read_tmp%dt,1,MPI_REAL8,0,MPI_COMM_MU,ierr)
      do ivpar = lstart,lend
        do itheta = 0,geom%Ntheta-1
          do ir = 0,geom%Nr
            pt_precomp%rfoot(ir,itheta,ivpar)  = read_tmp%rfoot(ir,itheta,ivpar)
            pt_precomp%thfoot(ir,itheta,ivpar) = read_tmp%thfoot(ir,itheta,ivpar)
          end do
        end do
      end do
      pt_precomp%dt = read_tmp%dt
      deallocate(read_tmp%rfoot,read_tmp%thfoot)
    end if
    call ppbarrier()
#endif
  end subroutine read_precomp_fields

end module init_magnetic_class
