!**************************************************************
!  Copyright Euratom-CEA
!  Authors : 
!     Virginie Grandgirard (virginie.grandgirard@cea.fr)
!     Chantal Passeron (chantal.passeron@cea.fr)
!     Guillaume Latu (guillaume.latu@cea.fr)
!     Xavier Garbet (xavier.garbet@cea.fr)
!     Philippe Ghendrih (philippe.ghendrih@cea.fr)
!     Yanick Sarazin (yanick.sarazin@cea.fr)
!  
!  This code GYSELA (for GYrokinetic SEmi-LAgrangian) 
!  is a 5D gyrokinetic global full-f code for simulating 
!  the plasma turbulence in a tokamak.
!  
!  This software is governed by the CeCILL-B license 
!  under French law and abiding by the rules of distribution 
!  of free software.  You can  use, modify and redistribute 
!  the software under the terms of the CeCILL-B license as 
!  circulated by CEA, CNRS and INRIA at the following URL
!  "http://www.cecill.info". 
!**************************************************************

#include "config-F.inc"

!------------------------------------------------
! file : init_profile.F90
! date : 21/06/2000
!  Initialization of the density and temperature
!  profiles
!------------------------------------------------
module init_profile_class
  use globals, only : profile_choice, q_profile
  use prec_const
  use wrap_alloc_module
  use geometry_class
  use spline1d_class
  use utils_module
#ifdef LIBM_MODULE_REQUIRED
  use libm
#endif
  implicit none

  !*** init_profile type definition ***
  type :: init_profile
    integer                              :: BCr_left, BCr_right
    real(RKIND)                          :: tau0
    !--> for profile definition (one profile per specie s)
    real(RKIND), dimension(:,:), pointer :: func_tanh
    real(RKIND)                          :: rpeak
    !--> for ion density profile (one profile per specie s)
    real(RKIND), dimension(:)  , pointer :: deltar_ns0
    real(RKIND), dimension(:)  , pointer :: inv_Lns0
    real(RKIND), dimension(:,:), pointer :: ns0
    real(RKIND), dimension(:,:), pointer :: ns0_coef1d
    real(RKIND), dimension(:)  , pointer :: ns0_rp
    real(RKIND), dimension(:,:), pointer :: dns0dr
    real(RKIND), dimension(:,:), pointer :: d2ns0dr2
    !--> for ion temperature profile (one profile per specie s)
    real(RKIND), dimension(:)  , pointer :: deltar_Ts0
    real(RKIND), dimension(:)  , pointer :: inv_LTs0
    real(RKIND), dimension(:,:), pointer :: Ts0
    real(RKIND), dimension(:,:), pointer :: Ts0_coef1d
    real(RKIND), dimension(:)  , pointer :: Ts0_rp
    real(RKIND), dimension(:,:), pointer :: dlogTs0dr
    real(RKIND), dimension(:,:), pointer :: dTs0dr
    real(RKIND), dimension(:,:), pointer :: d2Ts0dr2
    !--> for electron temperature profile
    real(RKIND)                          :: deltar_Te0
    real(RKIND)                          :: inv_LTe0
    real(RKIND), dimension(:)  , pointer :: Te0
    real(RKIND), dimension(:)  , pointer :: Te0_coef1d
    real(RKIND), dimension(:)  , pointer :: dTe0dr
    !--> for electron density profile
    real(RKIND), dimension(:)  , pointer :: ne0
    !--> for safety factor profile
    real(RKIND), dimension(:)  , pointer :: iota, shear, psi
    real(RKIND), dimension(:)  , pointer :: Vpar0
    real(RKIND)                          :: q_rp
    real(RKIND)                          :: psi_rp
    real(RKIND)                          :: psimin, psimax
    integer                              :: ipsimin, ipsimax
    !--> for buffer_region
    real(RKIND), dimension(:)  , pointer :: buffer_r
    !--> for cubic spline interpolation
    type(nspline1d) :: nspline1d_r
  end type init_profile

  !******************************
  contains
  !******************************
    
  !---------------------------------------- 
  ! sech = cosh^-1 definition
  !---------------------------------------- 
  function sech(x)
    real(RKIND), intent(in) :: x   
    real(RKIND) :: sech
    
    sech = 1._RKIND/cosh(x)
  end function sech  

     
  !---------------------------------------------------- 
  ! Constructor : Memory allocation for density 
  ! and temperature profiles
  !----------------------------------------------------
  subroutine new_init_profile(ithis,geom)
    use globals, only : memory_test, BC_Hermite, &
      tau0, rpeak, deltar_ns0, deltar_Te0, deltar_Ts0, &
      kappa_ns0, kappa_Te0, kappa_Ts0, R0, &
      Nspecies
    type(init_profile), intent(out) :: ithis
    type(geometry)    , intent(in)  :: geom
    
    integer     :: ispec
    integer     :: Nr_loc, Ntheta_loc, Nvpar_loc
    real(RKIND) :: dr_loc, a_norm
    
    Nr_loc     = geom%Nr
    Ntheta_loc = geom%Ntheta
    Nvpar_loc  = geom%Nvpar
    dr_loc     = geom%dr

    call write_expr("Nr_loc", "geom%Nr")
    call write_expr("Ntheta_loc", "geom%Ntheta")
    call write_expr("Nvpar_loc", "geom%Nvpar")
    call write_expr("dr_loc", "geom%dr")

    call write_param("Nspecies", Nspecies)

    !*** memory allocation ***
    !--> for profile definition (one profile per specie s)
    call FTR_ALLOC_2_ZERO(ithis%func_tanh,0,Nspecies-1,0,Nr_loc)
    !--> for ion density profile (one profile per specie s)
    call FTR_ALLOC_1_ZERO(ithis%deltar_ns0,0,Nspecies-1)
    call FTR_ALLOC_1_ZERO(ithis%inv_Lns0,0,Nspecies-1)
    call FTR_ALLOC_2_ZERO(ithis%ns0,0,Nspecies-1,0,Nr_loc)
    call FTR_ALLOC_2_ZERO(ithis%ns0_coef1d,0,Nspecies-1,-1,Nr_loc+1)
    call FTR_ALLOC_1_ZERO(ithis%ns0_rp,0,Nspecies-1)
    call FTR_ALLOC_2_ZERO(ithis%dns0dr,0,Nspecies-1,0,Nr_loc)
    call FTR_ALLOC_2_ZERO(ithis%d2ns0dr2,0,Nspecies-1,0,Nr_loc)
    !--> for ion temperature profile (one profile per specie s)
    call FTR_ALLOC_1_ZERO(ithis%deltar_Ts0,0,Nspecies-1)
    call FTR_ALLOC_1_ZERO(ithis%inv_LTs0,0,Nspecies-1)
    call FTR_ALLOC_2_ZERO(ithis%Ts0,0,Nspecies-1,0,Nr_loc)      
    call FTR_ALLOC_2_ZERO(ithis%Ts0_coef1d,0,Nspecies-1,-1,Nr_loc+1)
    call FTR_ALLOC_1_ZERO(ithis%Ts0_rp,0,Nspecies-1)
    call FTR_ALLOC_2_ZERO(ithis%dlogTs0dr,0,Nspecies-1,0,Nr_loc) 
    call FTR_ALLOC_2_ZERO(ithis%dTs0dr,0,Nspecies-1,0,Nr_loc)
    call FTR_ALLOC_2_ZERO(ithis%d2Ts0dr2,0,Nspecies-1,0,Nr_loc)
    !--> for electron temperature profile
    call FTR_ALLOC_1_ZERO(ithis%Te0,0,Nr_loc)
    call FTR_ALLOC_1_ZERO(ithis%Te0_coef1d,-1,Nr_loc+1)
    call FTR_ALLOC_1_ZERO(ithis%dTe0dr,0,Nr_loc)
    !--> for electron density profile
    call FTR_ALLOC_1_ZERO(ithis%ne0,0,Nr_loc)
    !--> for safety factor profile
    call FTR_ALLOC_1_ZERO(ithis%iota,0,Nr_loc)
    call FTR_ALLOC_1_ZERO(ithis%shear,0,Nr_loc)
    call FTR_ALLOC_1_ZERO(ithis%psi,0,Nr_loc)
    !--> for buffer region profile
    call FTR_ALLOC_1_ZERO(ithis%buffer_r,0,Nr_loc)
    !--> for parallel velocity profile
    call FTR_ALLOC_1_ZERO(ithis%Vpar0,0,Ntheta_loc)

    !*** variable initialization ***
    if (.not.memory_test) then
      a_norm           = geom%rg(Nr_loc)-geom%rg(0)
      ithis%tau0       = tau0
      ithis%rpeak      = geom%rg(0)+rpeak*a_norm
      ithis%deltar_Te0 = deltar_Te0*a_norm
      ithis%inv_LTe0   = kappa_Te0/R0
      do ispec = 0,Nspecies-1
        ithis%deltar_ns0(ispec) = deltar_ns0(ispec)*a_norm
        ithis%deltar_Ts0(ispec) = deltar_Ts0(ispec)*a_norm
        ! -> kappa = R/L 
        ithis%inv_Lns0(ispec)   = kappa_ns0(ispec)/R0        
        ithis%inv_LTs0(ispec)   = kappa_Ts0(ispec)/R0
      end do
    end if

    !*** initialization of the boundary conditions ***
    ithis%BCr_left  = BC_Hermite
    ithis%BCr_right = BC_Hermite

    !*** initialization of the spline in r direction ***
    call write_expr("n", "Nr_loc")
    call new_spline1d_natural(ithis%nspline1d_r,Nr_loc,dr_loc)
  end subroutine new_init_profile
  

  !---------------------------------------------------- 
  ! destructor
  !---------------------------------------------------- 
  subroutine del_init_profile(ithis)
    type(init_profile), intent(inout) :: ithis

    !--> for profile definition (one profile per specie s)
    call FTR_DEALLOC(ithis%func_tanh)
    !--> for ion density profile (one profile per specie s)
    call FTR_DEALLOC(ithis%deltar_ns0)
    call FTR_DEALLOC(ithis%inv_Lns0)
    call FTR_DEALLOC(ithis%ns0)
    call FTR_DEALLOC(ithis%ns0_coef1d)
    call FTR_DEALLOC(ithis%ns0_rp)
    call FTR_DEALLOC(ithis%dns0dr)    
    call FTR_DEALLOC(ithis%d2ns0dr2)
    !--> for ion temperature profile (one profile per specie s)
    call FTR_DEALLOC(ithis%deltar_Ts0)
    call FTR_DEALLOC(ithis%inv_LTs0)
    call FTR_DEALLOC(ithis%Ts0)
    call FTR_DEALLOC(ithis%Ts0_coef1d)
    call FTR_DEALLOC(ithis%Ts0_rp)
    call FTR_DEALLOC(ithis%dlogTs0dr)
    call FTR_DEALLOC(ithis%dTs0dr)      
    call FTR_DEALLOC(ithis%d2Ts0dr2)
    !--> for electron temperature profile
    call FTR_DEALLOC(ithis%Te0)
    call FTR_DEALLOC(ithis%Te0_coef1d) 
    call FTR_DEALLOC(ithis%dTe0dr) 
    !--> for electron density profile
    call FTR_DEALLOC(ithis%ne0)
    !--> for safety factor profile
    call FTR_DEALLOC(ithis%iota)
    call FTR_DEALLOC(ithis%shear)
    call FTR_DEALLOC(ithis%psi)
    !--> for buffer region profile
    call FTR_DEALLOC(ithis%buffer_r)
    !--> for parallel velocity profile
    call FTR_DEALLOC(ithis%Vpar0)
    !--> for cubic spline interpolation
    call del_spline1d_natural(ithis%nspline1d_r)
  end subroutine del_init_profile


  !**************************************************************
  !  CHOICE OF INITIAL PROFILES AS FUNCTION OF HYPERBOLIC
  !   COSINUS :
  !  -> EX: 1/T(r) dT(r)/dr = -(1/LT)*func(cosh^-2)
  !**************************************************************
  !------------------------------------------------------------- 
  !  ns0 density profile initialization
  !  this profile is given by the density gradient 
  !  profile, defined by 3 parameters : 
  !    - kappa_ns, rpeak and deltar
  !  and
  !    1/ns0(r) dns0(r)/dr = -(1/Lns0)*cosh^-2(r-rpeak/deltar)
  !  with (1/Ln) = kappa_ns0/R
  !------------------------------------------------------------- 
  subroutine init_ns0_1(geom,rpeak,inv_Lns0,deltar_ns0,ns0)
    type(geometry)            , intent(in)  :: geom
    real(RKIND)               , intent(in)  :: rpeak
    real(RKIND)               , intent(in)  :: inv_Lns0
    real(RKIND)               , intent(in)  :: deltar_ns0
    real(RKIND), dimension(0:), intent(out) :: ns0        ! [0:Nr]

    real(RKIND) :: r, dr_loc, rth, tmp
    integer     :: ir, Nr_loc
    real(RKIND) :: ns0norm_tmp
    
    !*** local variable initialization ***
    Nr_loc = geom%Nr
    dr_loc = geom%dr
    
    !*** compute ns0 solution of :                            ***
    !***  2/(ns0(r)+ns0(r-1))*(ns0(r)-ns0(r-1))/dr            ***
    !***                  = -(1/LTs0)*cosh^-2(r-rpeak/deltar) ***
    ns0(0) = 10._RKIND**19
    do ir = 1,Nr_loc
      r       = geom%rg(ir-1)
      rth     = r + dr_loc*0.5_RKIND
      tmp     = -inv_Lns0 * &
        sech((rth-rpeak)/deltar_ns0)**2
      tmp     = 0.5_RKIND*dr_loc*tmp
      ns0(ir) = (1._RKIND+tmp)/(1._RKIND-tmp)*ns0(ir-1)
    enddo
    
    !*** normalisation of the density at int(ns0(r)rdr)/int(rdr) ***
    ! -> computation of int(ns0(r)rdr)
    ns0norm_tmp = 0._RKIND
    do ir = 1,Nr_loc-1
      ns0norm_tmp = ns0norm_tmp + ns0(ir)*geom%rg(ir)
    enddo
    ns0norm_tmp = ns0norm_tmp + 0.5_RKIND* &
      (ns0(0)*geom%rg(0) + ns0(Nr_loc)*geom%rg(Nr_loc))
    ! -> division by int(rdr)
    ns0norm_tmp = ns0norm_tmp*2._RKIND*dr_loc/ & 
      (geom%rg(Nr_loc)**2-geom%rg(0)**2)

    ns0(0:Nr_loc) = ns0(0:Nr_loc)/ns0norm_tmp
  end subroutine init_ns0_1


  !-------------------------------------------------------- 
  ! derivative of ionic density profile
  !  dns0(r)/dr = -ns0(r)*(1/Ln)*cosh^-2(r-rpeak/deltar)
  !--------------------------------------------------------     
  subroutine init_dns0dr_1(geom,rpeak,inv_Lns0,deltar_ns0,ns0,dns0dr)
    type(geometry)            , intent(in)  :: geom
    real(RKIND)               , intent(in)  :: rpeak
    real(RKIND)               , intent(in)  :: inv_Lns0
    real(RKIND)               , intent(in)  :: deltar_ns0
    real(RKIND), dimension(0:), intent(in)  :: ns0         ! [0:Nr]
    real(RKIND), dimension(0:), intent(out) :: dns0dr      ! [0:Nr]

    real(RKIND) :: r     
    integer     :: ir, Nr_loc
        
    !*** local variable initialization ***
    Nr_loc = geom%Nr
        
    !*** temperature gradient profile for electron ***
    do ir = 0,Nr_loc
      r          = geom%rg(ir)
      dns0dr(ir) = -ns0(ir)* &
        inv_Lns0*sech((r-rpeak)/deltar_ns0)**2
    end do
  end subroutine init_dns0dr_1


  !------------------------------------------------------------------ 
  ! ionic temperature profile initialization
  !  this profile is given by the temperature gradient 
  !  profile, defined by 3 parameters : 
  !    - kappa_Ts0, rpeak and deltar
  !  and
  !    1/Ts0(r) dTs0(r)/dr = -kappa_Ts0*cosh^-2(r-rpeak/deltar)
  !  with (1/LTs0) = kappa_Ts0/R
  !------------------------------------------------------------------ 
  subroutine init_Ts0_1(geom,rpeak,inv_LTs0,deltar_Ts0,Ts0)
    type(geometry)            , intent(in)  :: geom
    real(RKIND)               , intent(in)  :: rpeak
    real(RKIND)               , intent(in)  :: inv_LTs0
    real(RKIND)               , intent(in)  :: deltar_Ts0
    real(RKIND), dimension(0:), intent(out) :: Ts0          ! [0:Nr]
    
    real(RKIND) :: r, dr_loc, rth, tmp, w0, w1, Ts0_norm
    integer     :: ir, Nr_loc
    
    !*** local variable initialization ***
    Nr_loc = geom%Nr
    dr_loc = geom%dr
        
    !*** compute Ts0 solution of :                            ***
    !***  2/(Ts0(r)+Ts0(r-1))*(Ts0(r)-Ts0(r-1))/dr            ***
    !***                  = -(1/LTs0)*cosh^-2(r-rpeak/deltar) ***
    Ts0(0) = 1.3_RKIND    !keV
    do ir = 1,Nr_loc
      r      = geom%rg(ir-1)
      rth    = r+ dr_loc*0.5_RKIND
      tmp    = -0.5_RKIND*dr_loc*inv_LTs0 &
        *sech((rth-rpeak)/deltar_Ts0)**2
      Ts0(ir) = (1._RKIND+tmp)/(1._RKIND-tmp)*Ts0(ir-1)
    enddo
    
    !*** normalisation of the temperature to 1 at r=rpeak ***
    ir            = int((rpeak-geom%rg(0))/dr_loc)
    w1            = (rpeak-geom%rg(ir))/dr_loc
    w0            = 1._RKIND-w1
    Ts0_norm      = w0*Ts0(ir)+w1*Ts0(ir+1)
    Ts0(0:Nr_loc) = (Ts0(0:Nr_loc)/Ts0_norm)
  end subroutine init_Ts0_1  
  
  
  !----------------------------------------------------------- 
  ! ionic temperature gradient profile initialization
  !  this profile is by 3 parameters : 
  !    - kappa_Ts0, rpeak and deltar
  !  and
  !    1/Ts0(r) dTs0(r)/dr = -(1/LTs0)*cosh^-2(r-rpeak/deltar)
  !  with (1/LTs0) = kappa_Ts0/R
  !-----------------------------------------------------------   
  subroutine init_dlogTs0dr_1(geom,rpeak,inv_LTs0,deltar_Ts0, &
    dlogTs0dr)
    type(geometry)            , intent(in)  :: geom
    real(RKIND)               , intent(in)  :: rpeak
    real(RKIND)               , intent(in)  :: inv_LTs0
    real(RKIND)               , intent(in)  :: deltar_Ts0
    real(RKIND), dimension(0:), intent(out) :: dlogTs0dr    ! [0:Nr]
    
    real(RKIND) :: r, dr_loc 
    integer     :: ir, Nr_loc
    
    !*** local variable initialization ***
    Nr_loc = geom%Nr
    dr_loc = geom%dr
    
    !*** temperature gradient profile for ion ***
    do ir = 0,Nr_loc
      r             = geom%rg(ir)
      dlogTs0dr(ir) = -inv_LTs0* &
        sech((r-rpeak)/deltar_Ts0)**2
    end do
  end subroutine init_dlogTs0dr_1


  !------------------------------------------------------------ 
  ! electronic temperature profile initialization
  !  this profile is given by the temperature gradient 
  !  profile, defined by 3 parameters : 
  !    - kappa_Te0, rpeak and deltar
  !  and
  !    1/Te0(r) dTe0(r)/dr = -invLTe0*cosh^-2(r-rpeak/deltar)
  !  with (1/LTe0) = kappa_Te0/R
  !------------------------------------------------------------ 
  subroutine init_Te0_1(geom,rpeak,inv_LTe0,deltar_Te0,tau0,Ts0_rmin,Te0)
    type(geometry)            , intent(in)  :: geom
    real(RKIND)               , intent(in)  :: rpeak
    real(RKIND)               , intent(in)  :: inv_LTe0
    real(RKIND)               , intent(in)  :: deltar_Te0
    real(RKIND)               , intent(in)  :: tau0
    real(RKIND)               , intent(in)  :: Ts0_rmin
    real(RKIND), dimension(0:), intent(out) :: Te0          ! [0:Nr]

    real(RKIND) :: r, dr_loc, rth, tmp, w0, w1
    real(RKIND) :: Te0_norm
    integer     :: ir, Nr_loc
        
    !*** local variable initialization ***
    Nr_loc = geom%Nr
    dr_loc = geom%dr
            
    !*** compute Te0 solution of :                            ***
    !***  2/(Te0(r)+Te0(r-1))*(Te0(r)-Te0(r-1))/dr            ***
    !***                  = -(1/LTe0)*cosh^-2(r-rpeak/deltar) ***
    Te0(0) = Ts0_rmin       !Rk: Ts0_rmin = Ts0(0)
    do ir = 1,Nr_loc
      r       = geom%rg(ir-1)
      rth     = r+ dr_loc*0.5_RKIND
      tmp     = -0.5_RKIND*dr_loc*inv_LTe0 * &
        sech((rth-rpeak)/deltar_Te0)**2
      Te0(ir) = (1._RKIND+tmp)/(1._RKIND-tmp)*Te0(ir-1)
    enddo
    
    !*** normalisation of the temperature to ***
    !*** Te0(0) (Te0(0) = Ts0(0)/tau0)       *** 
    ir            = int((rpeak-geom%rg(0))/dr_loc)
    w1            = (rpeak-geom%rg(ir))/dr_loc
    w0            = 1._RKIND-w1
    Te0_norm      = w0*Te0(ir)+w1*Te0(ir+1)
    Te0(0:Nr_loc) = (Te0(0:Nr_loc)/Te0_norm)/tau0
  end subroutine init_Te0_1


  !-------------------------------------------------------- 
  ! derivative of electronic temperature profile 
  !  dTe0(r)/dr = -Te0(r)*(1/LTe0)*cosh^-2(r-rpeak/deltar)
  !--------------------------------------------------------     
  subroutine init_dTe0dr_1(geom,rpeak,inv_LTe0,deltar_Te0,Te0,dTe0dr)
    type(geometry)            , intent(in)  :: geom
    real(RKIND)               , intent(in)  :: rpeak
    real(RKIND)               , intent(in)  :: inv_LTe0
    real(RKIND)               , intent(in)  :: deltar_Te0
    real(RKIND), dimension(0:), intent(in)  :: Te0         ! [0:Nr]
    real(RKIND), dimension(0:), intent(out) :: dTe0dr      ! [0:Nr]

    real(RKIND) :: r     
    integer     :: ir, Nr_loc
        
    !*** local variable initialization ***
    Nr_loc = geom%Nr
        
    !*** temperature gradient profile for electron ***
    do ir = 0,Nr_loc
      r          = geom%rg(ir)
      dTe0dr(ir) = -Te0(ir) * &
        inv_LTe0*sech((r-rpeak)/deltar_Te0)**2
    end do
  end subroutine init_dTe0dr_1
       
       

  !**************************************************************
  !  CHOICE OF INITIAL PROFILES SUCH THAT DN/DT 
  !   AND D(NT)/DT ARE CONSTANT
  !**************************************************************
  !-------------------------------------------------------- 
  ! computation of func(tanh) = [tanh(0.5*(rg-rmin-rbuff)) 
  !                - tanh(0.5*(rg-rmax+rbuff))]/max(hr) 
  !-------------------------------------------------------- 
  subroutine init_func_tanh_1(geom,func_tanh)
    type(geometry)            , intent(in)  :: geom
    real(RKIND), dimension(0:), intent(out) :: func_tanh   ! [0:Nr]

    integer     :: ir, Nr_loc
    real(RKIND) :: ri, dr_loc
    real(RKIND) :: rbuff, rmin, rmax, max_func
    real(RKIND) :: func_tanh_tmp
        
    !*** local variable initialization ***
    Nr_loc = geom%Nr
    dr_loc = geom%dr
    
    !*** computation of func(tanh) and its maximum ***
    rbuff    = 0.25_RKIND*Nr_loc*dr_loc
    rmin     = geom%rg(0)
    rmax     = geom%rg(Nr_loc)
    max_func = 0._RKIND
    do ir = 0,Nr_loc
      ri = geom%rg(ir)
      func_tanh_tmp = tanh(0.5_RKIND*(ri-rmin-rbuff)) - &
        tanh(0.5_RKIND*(ri-rmax+rbuff))
      func_tanh(ir) = func_tanh_tmp
      max_func      = max(func_tanh_tmp,max_func)
    end do
    func_tanh = func_tanh/max_func
  end subroutine init_func_tanh_1


  !**************************************************************
  !  CHOICE OF INITIAL PROFILES AS FUNCTION OF HYPERBOLIC
  !   COSINUS FOR BENCHMARK WITH ORB5
  !  -> EX: 1/T(r) dT(r)/dr = -(1/LT)*
  !       (-1+cosh^-2((r-rmin)/deltar)+cosh^-2((r-rmax)/deltar))
  !**************************************************************
  !-------------------------------------------------------- 
  ! computation of func(tanh) = [tanh(0.5*(rg-rmin-rbuff1)) 
  !                - tanh(0.5*(rg-rmax+rbuff2))]/max(hr) 
  ! REMARK: In this case the tangent profile is divided
  !         by the radial mesh step, to be consistent
  !         for invariance test comparisons
  !-------------------------------------------------------- 
  subroutine init_func_tanh_2(geom,func_tanh)
    type(geometry)            , intent(in)  :: geom
    real(RKIND), dimension(0:), intent(out) :: func_tanh   ! [0:Nr]

    integer     :: ir, Nr_loc
    real(RKIND) :: ri, Lr_loc
    real(RKIND) :: rmin, rmax, rp
    real(RKIND) :: rbuff1, rbuff2
    real(RKIND) :: func_tanh_tmp, max_func
        
    !*** local variable initialization ***
    Nr_loc = geom%Nr
    Lr_loc = geom%Lr
    
    !*** computation of func(tanh) and its maximum ***
    rbuff1   = 0.2_RKIND*Lr_loc 
    rbuff2   = 0.2_RKIND*Lr_loc 
    rmin     = geom%rg(0)
    rmax     = geom%rg(Nr_loc)
    rp       = geom%rg(Nr_loc/2)
    max_func = 0._RKIND
    do ir = 0,Nr_loc
      ri            = geom%rg(ir)
      func_tanh_tmp = tanh(0.5_RKIND*(ri-rmin-rbuff1)/geom%dr) - &
        tanh(0.5_RKIND*(ri-rmax+rbuff2)/geom%dr)
      func_tanh(ir) = func_tanh_tmp
      max_func      = max(func_tanh_tmp,max_func)
    end do
    func_tanh = func_tanh/(max_func)
  end subroutine init_func_tanh_2


  !-------------------------------------------------------------- 
  !  ns0 density profile initialization
  !  this profile is given by the density gradient 
  !  profile, defined by 3 parameters : 
  !    - kappa_ns, rpeak and deltar
  !  and
  !    1/ns0(r) dns0(r)/dr = -(1/Ln)*func_sech(r)*func_tanh(r)
  !  with (1/Lns0) = kappa_ns/R
  !  and where 
  !   func_sech(r) = -1+cosh^-2((r-rmin)/deltar) + 
  !      cosh^-2((r-rmax)/deltar)
  !-------------------------------------------------------------- 
  subroutine init_ns0_2(geom,rpeak,inv_Lns0,deltar_ns0,func_tanh,ns0)
    type(geometry)            , intent(in)  :: geom
    real(RKIND)               , intent(in)  :: rpeak
    real(RKIND)               , intent(in)  :: inv_Lns0
    real(RKIND)               , intent(in)  :: deltar_ns0
    real(RKIND), dimension(0:), intent(in)  :: func_tanh   ! [0:Nr]
    real(RKIND), dimension(0:), intent(out) :: ns0         ! [0:Nr]

    real(RKIND) :: r, dr_loc, rth, tmp
    integer     :: ir, Nr_loc
    real(RKIND) :: ns0norm_tmp, ns0norm
    real(RKIND) :: Lr_loc, rmin, rmax, func_sech
    
    !*** local variable initialization ***
    Nr_loc = geom%Nr
    dr_loc = geom%dr
    Lr_loc = geom%Lr
    rmin   = geom%rg(0)     !+0.1_RKIND*Lr_loc
    rmax   = geom%rg(Nr_loc)!-0.1_RKIND*Lr_loc
    
    !*** compute ns0 solution of :                ***
    !***  2/(ns0(r)+ns0(r-1))*(ns0(r)-ns0(r-1))/dr   ***
    !***                  = -(1/Ln)*func_sech(r) ***
    ns0(0) = 10._RKIND**19
    do ir = 1,Nr_loc
      r         = geom%rg(ir-1)
      rth       = r + dr_loc*0.5_RKIND
      func_sech = -1._RKIND + &
        sech((rth-rmin)/deltar_ns0)**2 + &
        sech((rth-rmax)/deltar_ns0)**2
      tmp       = 0.5_RKIND*dr_loc * &
        inv_Lns0*func_sech*func_tanh(ir)
      ns0(ir)    = (1._RKIND+tmp)/(1._RKIND-tmp)*ns0(ir-1)
    enddo
    
    !*** normalisation of the density at int(ns0(r)rdr)/int(rdr) ***
    ! -> computation of int(ns0(r)rdr)
    ns0norm_tmp = 0._RKIND
    do ir = 1,Nr_loc-1
      ns0norm_tmp = ns0norm_tmp + ns0(ir)*geom%rg(ir)
    enddo
    ns0norm_tmp = ns0norm_tmp + 0.5_RKIND* &
      (ns0(0)*geom%rg(0) + ns0(Nr_loc)*geom%rg(Nr_loc))
    ! -> division by int(rdr)
    ns0norm_tmp = ns0norm_tmp*2._RKIND*dr_loc / & 
      (geom%rg(Nr_loc)**2-geom%rg(0)**2)

    ns0norm       = ns0norm_tmp
    ns0(0:Nr_loc) = ns0(0:Nr_loc)/ns0norm
  end subroutine init_ns0_2


  !---------------------------------------------------------------
  ! derivative of ionic density profile
  !  dns0(r)/dr = -ns0(r)*(1/Ln)*
  !       (-1+cosh^-2((r-rmin)/deltar)+cosh^-2((r-rmax)/deltar))
  !---------------------------------------------------------------
  subroutine init_dns0dr_2(geom,rpeak,inv_Lns0,deltar_ns0, &
    func_tanh,ns0,dns0dr)
    type(geometry)            , intent(in)  :: geom
    real(RKIND)               , intent(in)  :: rpeak
    real(RKIND)               , intent(in)  :: inv_Lns0
    real(RKIND)               , intent(in)  :: deltar_ns0
    real(RKIND), dimension(0:), intent(in)  :: func_tanh   ! [0:Nr]
    real(RKIND), dimension(0:), intent(in)  :: ns0         ! [0:Nr]
    real(RKIND), dimension(0:), intent(out) :: dns0dr      ! [0:Nr]

    integer     :: ir, Nr_loc
    real(RKIND) :: Lr_loc, r, rmin, rmax, func_sech
        
    !*** local variable initialization ***
    Nr_loc = geom%Nr
    Lr_loc = geom%Lr
    rmin   = geom%rg(0)     !+0.1_RKIND*Lr_loc
    rmax   = geom%rg(Nr_loc)!-0.1_RKIND*Lr_loc
        
    !*** temperature gradient profile for electron ***
    do ir = 0,Nr_loc
      r         = geom%rg(ir)
      func_sech = -1._RKIND + &
        sech((r-rmin)/deltar_ns0)**2 + &
        sech((r-rmax)/deltar_ns0)**2      
      dns0dr(ir) = ns0(ir)* &
        inv_Lns0*func_sech*func_tanh(ir)
    end do
  end subroutine init_dns0dr_2


  !-------------------------------------------------------------
  ! ionic temperature profile initialization
  !  this profile is given by the temperature gradient 
  !  profile, defined by 3 parameters : 
  !    - kappa_Ts0, rpeak and deltar
  !  and
  !    1/Ts0(r) dTs0(r)/dr = -invLTs0*func_sech(r)*func_tanh(r)
  !  with (1/LTs0) = kappa_Ts0/R
  !  and where
  !   func_sech(r) = (-1+cosh^-2((r-rmin)/deltar) + 
  !      cosh^-2((r-rmax)/deltar))
  !-------------------------------------------------------------
  subroutine init_Ts0_2(geom,rpeak,invLTs0,deltar_Ts0,func_tanh,Ts0)
    type(geometry)            , intent(in)  :: geom
    real(RKIND)               , intent(in)  :: rpeak
    real(RKIND)               , intent(in)  :: invLTs0
    real(RKIND)               , intent(in)  :: deltar_Ts0
    real(RKIND), dimension(0:), intent(in)  :: func_tanh   ! [0:Nr]
    real(RKIND), dimension(0:), intent(out) :: Ts0         ! [0:Nr]
    
    integer     :: ir, Nr_loc
    real(RKIND) :: Lr_loc, r, dr_loc
    real(RKIND) :: rth, tmp, w0, w1, Ts0_norm
    real(RKIND) :: rmin, rmax, func_sech
    
    !*** local variable initialization ***
    Nr_loc = geom%Nr
    dr_loc = geom%dr
    Lr_loc = geom%Lr
    rmin   = geom%rg(0)     !+0.1_RKIND*Lr_loc
    rmax   = geom%rg(Nr_loc)!-0.1_RKIND*Lr_loc

    !*** compute Ts0 solution of :                            ***
    !***  2/(Ts0(r)+Ts0(r-1))*(Ts0(r)-Ts0(r-1))/dr               ***
    !***                  = -(1/LTs0)*cosh^-2(r-rpeak/deltar) ***
    Ts0(0) = 2._RKIND    !keV
    do ir = 1,Nr_loc
      r          = geom%rg(ir-1)
      rth        = r + dr_loc*0.5_RKIND
      func_sech  = -1._RKIND + &
        sech((rth-rmin)/(deltar_Ts0))**2 + &
        sech((rth-rmax)/deltar_Ts0)**2      
      tmp        = 0.5_RKIND*dr_loc * &
        invLTs0*func_sech*func_tanh(ir)
      Ts0(ir)     = (1._RKIND+tmp)/(1._RKIND-tmp)*Ts0(ir-1)
    enddo
    
    !*** normalisation of the temperature to 1 at r=rpeak ***
    ir            = int((rpeak-geom%rg(0))/dr_loc)
    w1            = (rpeak-geom%rg(ir))/dr_loc
    w0            = 1._RKIND-w1
    Ts0_norm      = w0*Ts0(ir)+w1*Ts0(ir+1)
    Ts0(0:Nr_loc) = (Ts0(0:Nr_loc)/Ts0_norm)
  end subroutine init_Ts0_2
  
  
  !-------------------------------------------------------------
  ! ionic temperature gradient profile initialization
  !  this profile is by 3 parameters : 
  !    - kappa_Ts0, rpeak and deltar
  !  and
  !    1/Ts0(r) dTs0(r)/dr = -(1/LTs0)*func_sech(r)*func_tanh(r)
  !  with (1/LTs0) = kappa_Ts0/R
  !  and where
  !   func_sech(r) = (-1+cosh^-2((r-rmin)/deltar) + 
  !     cosh^-2((r-rmax)/deltar))
  !-------------------------------------------------------------
  subroutine init_dlogTs0dr_2(geom,rpeak,invLTs0,deltar_Ts0,func_tanh, &
    dlogTs0dr)
    type(geometry)            , intent(in)  :: geom
    real(RKIND)               , intent(in)  :: rpeak
    real(RKIND)               , intent(in)  :: invLTs0
    real(RKIND)               , intent(in)  :: deltar_Ts0
    real(RKIND), dimension(0:), intent(in)  :: func_tanh   ! [0:Nr]
    real(RKIND), dimension(0:), intent(out) :: dlogTs0dr   ! [0:Nr]
    
    integer     :: ir, Nr_loc
    real(RKIND) :: Lr_loc, r, rmin, rmax, func_sech

    !*** local variable initialization ***
    Nr_loc = geom%Nr
    Lr_loc = geom%Lr
    rmin   = geom%rg(0)     !+0.1_RKIND*Lr_loc
    rmax   = geom%rg(Nr_loc)!-0.1_RKIND*Lr_loc

    do ir = 0,Nr_loc
      r            = geom%rg(ir)
      func_sech    = -1._RKIND + &
        sech((r-rmin)/(deltar_Ts0))**2 + &
        sech((r-rmax)/deltar_Ts0)**2
      dlogTs0dr(ir) = invLTs0 * &
        func_sech*func_tanh(ir) 
    end do
  end subroutine init_dlogTs0dr_2


  !------------------------------------------------------------
  ! electronic temperature profile initialization
  !  this profile is given by the temperature gradient 
  !  profile, defined by 3 parameters : 
  !    - kappa_Te0, rpeak and deltar
  !  and
  !    1/Te0(r) dTe0(r)/dr = -invLTe0*func_sech(r)
  !  with (1/LTe0) = kappa_Te0/R
  !  and where
  !   func_sech(r) = (-1+cosh^-2((r-rmin)/deltar) + 
  !    cosh^-2((r-rmax)/deltar))
  !------------------------------------------------------------
  subroutine init_Te0_2(geom,rpeak,invLTe0,deltar_Te0, &
    tau0,Ts0_rmin,func_tanh,Te0)
    type(geometry)            , intent(in)  :: geom
    real(RKIND)               , intent(in)  :: rpeak
    real(RKIND)               , intent(in)  :: invLTe0
    real(RKIND)               , intent(in)  :: deltar_Te0
    real(RKIND)               , intent(in)  :: tau0
    real(RKIND)               , intent(in)  :: Ts0_rmin
    real(RKIND), dimension(0:), intent(in)  :: func_tanh  ! [0:Nr]
    real(RKIND), dimension(0:), intent(out) :: Te0         ! [0:Nr]

    integer     :: ir, Nr_loc
    real(RKIND) :: r, dr_loc, rth, tmp, w0, w1
    real(RKIND) :: Lr_loc, rmin, rmax, func_sech
    real(RKIND) :: Te0_norm
        
    !*** local variable initialization ***
    Nr_loc = geom%Nr
    dr_loc = geom%dr
    Lr_loc = geom%Lr
    rmin   = geom%rg(0)     !+0.1_RKIND*Lr_loc
    rmax   = geom%rg(Nr_loc)!-0.1_RKIND*Lr_loc
            
    !*** compute Te0 solution of :                  ***
    !***  2/(Te0(r)+Te0(r-1))*(Te0(r)-Te0(r-1))/dr  ***
    !***                  = -(1/LTe0)*func_sech(r)  ***
    Te0(0) = Ts0_rmin    !Rk: Ts0_rmin = Ts0(0)
    do ir = 1,Nr_loc
      r          = geom%rg(ir-1)
      rth        = r+ dr_loc*0.5_RKIND
      func_sech  = -1._RKIND + &
        sech((rth-rmin)/deltar_Te0)**2 + &
        sech((rth-rmax)/deltar_Te0)**2
      tmp        = 0.5_RKIND*dr_loc * &
        invLTe0*func_sech*func_tanh(ir)
      Te0(ir)    = (1._RKIND+tmp)/(1._RKIND-tmp)*Te0(ir-1)
    enddo
    
    !*** normalisation of the temperature to 1 at r=rpeak ***
    ir            = int((rpeak-geom%rg(0))/dr_loc)
    w1            = (rpeak-geom%rg(ir))/dr_loc
    w0            = 1._RKIND-w1
    Te0_norm      = w0*Te0(ir)+w1*Te0(ir+1)
    Te0(0:Nr_loc) = (Te0(0:Nr_loc)/Te0_norm)/tau0  
  end subroutine init_Te0_2


  !--------------------------------------------------------------
  ! derivative of electronic temperature profile 
  !  dTe0(r)/dr = -Te0(r)*(1/LTe0)*
  !       (-1+cosh^-2((r-rmin)/deltar)+cosh^-2((r-rmax)/deltar))
  !--------------------------------------------------------------
  subroutine init_dTe0dr_2(geom,rpeak,invLTe0,deltar_Te0, &
    func_tanh,Te0,dTe0dr)
    type(geometry)            , intent(in)  :: geom
    real(RKIND)               , intent(in)  :: rpeak
    real(RKIND)               , intent(in)  :: invLTe0
    real(RKIND)               , intent(in)  :: deltar_Te0
    real(RKIND), dimension(0:), intent(in)  :: func_tanh   ! [0:Nr]
    real(RKIND), dimension(0:), intent(in)  :: Te0         ! [0:Nr]
    real(RKIND), dimension(0:), intent(out) :: dTe0dr      ! [0:Nr]

    real(RKIND) :: Lr_loc, r, rmin, rmax, func_sech
    integer     :: ir, Nr_loc
        
    !*** local variable initialization ***
    Nr_loc = geom%Nr
    Lr_loc = geom%Lr
    rmin   = geom%rg(0)     !+0.1_RKIND*Lr_loc
    rmax   = geom%rg(Nr_loc)!-0.1_RKIND*Lr_loc
        
    !*** temperature gradient profile for electron ***
    do ir = 0,Nr_loc
      r          = geom%rg(ir)
      func_sech  = -1._RKIND + &
        sech((r-rmin)/deltar_Te0)**2 + &
        sech((r-rmax)/deltar_Te0)**2
      dTe0dr(ir) = Te0(ir) * &
        invLTe0*func_sech*func_tanh(ir)
    end do
  end subroutine init_dTe0dr_2
              

  !**************************************************************
  !  CHOICE OF INITIAL PROFILES WITH FLAT LOGARITHMIC GRADIENTS
  !  OUTSIDE THE BUFFER REGIONS
  !  -> EX: 1/T(r) dT(r)/dr = -(1/LT)*func_tanh_3
  !**************************************************************
  !--------------------------------------------------------
  ! computation of func(tanh) = 0.5*[tanh((rg-rmin-rbuff)/lambda)
  !                - tanh((rg-rmax+rbuff)/lambda)]
  !--------------------------------------------------------
  subroutine init_func_tanh_3(geom, &
    deltar_Ts0,a,coef_diffr,buffer_location,func_tanh)
    type(geometry)            , intent(in)  :: geom
    real(RKIND)               , intent(in)  :: deltar_Ts0
    real(RKIND)               , intent(in)  :: a
    real(RKIND)               , intent(in)  :: coef_diffr
    real(RKIND)               , intent(in)  :: buffer_location
    real(RKIND), dimension(0:), intent(out) :: func_tanh   ! [0:Nr]

    integer     :: ir, Nr_loc
    real(RKIND) :: ri, Lr_loc
    real(RKIND) :: rmin, rmax
    real(RKIND) :: rbuff, lambda
    real(RKIND) :: func_tanh_tmp, max_func

    !*** local variable initialization ***
    Nr_loc = geom%Nr
    Lr_loc = geom%Lr

    !*** computation of func(tanh) and its maximum ***
    if (coef_diffr.ne.0._RKIND) then
      rbuff = 1.5_RKIND*buffer_location*Lr_loc
    else
      rbuff = 0.15_RKIND*Lr_loc
    end if
    rmin     = geom%rg(0)
    rmax     = geom%rg(Nr_loc)
    max_func = 0._RKIND
    lambda   = deltar_Ts0*Lr_loc
    do ir = 0,Nr_loc
      ri = geom%rg(ir)
      func_tanh_tmp = 0.5_RKIND* &
        (tanh((ri-rmin-rbuff)/lambda) - &
        tanh((ri-rmax+rbuff)/lambda))
      func_tanh(ir) = func_tanh_tmp * &
        (1._RKIND+0.7_RKIND*(0.5_RKIND-ri/a))
    end do
    func_tanh = 2._RKIND*func_tanh / &
      (func_tanh((Nr_loc-1)/2)+func_tanh((Nr_loc+1)/2))
  end subroutine init_func_tanh_3


  !--------------------------------------------------
  !  ns0 density profile initialization
  !  this profile is given by the density gradient 
  !  profile, defined by 3 parameters : 
  !    - kappa_ns, rpeak and deltar
  !  and
  !    1/ns0(r) dns0(r)/dr = -(1/Lns0)*func_tanh_3(r)
  !  with (1/Ln) = kappa_ns/R
  !--------------------------------------------------
  subroutine init_ns0_3(geom,inv_Lns0,func_tanh,ns0)
    type(geometry)            , intent(in)  :: geom
    real(RKIND)               , intent(in)  :: inv_Lns0
    real(RKIND), dimension(0:), intent(in)  :: func_tanh   ! [0:Nr]
    real(RKIND), dimension(0:), intent(out) :: ns0         ! [0:Nr]

    real(RKIND) :: r, dr_loc, rth, tmp
    integer     :: ir, Nr_loc
    real(RKIND) :: ns0norm_tmp, ns0norm
    
    !*** local variable initialization ***
    Nr_loc = geom%Nr
    dr_loc = geom%dr
    
    !*** compute ns0 solution of :                ***
    !***  2/(ns0(r)+ns0(r-1))*(ns0(r)-ns0(r-1))/dr   ***
    !***                  = -(1/Ln)*func_tanh_3(r) ***
    ns0(0) = 10._RKIND**(19)
    do ir = 1,Nr_loc
      tmp    = -0.5_RKIND*dr_loc*inv_Lns0 * &
        func_tanh(ir)
      ns0(ir) = (1._RKIND+tmp)/(1._RKIND-tmp)*ns0(ir-1)
    enddo
    
    !*** normalisation of the density at int(ns0(r)rdr)/int(rdr) ***
    ! -> computation of int(ns0(r)rdr)
    ns0norm_tmp = 0._RKIND
    do ir = 1,Nr_loc-1
      ns0norm_tmp = ns0norm_tmp + ns0(ir)*geom%rg(ir)
    enddo
    ns0norm_tmp = ns0norm_tmp + 0.5_RKIND* &
      (ns0(0)*geom%rg(0) + ns0(Nr_loc)*geom%rg(Nr_loc))
    ! -> division by int(rdr)
    ns0norm_tmp = ns0norm_tmp*2._RKIND*dr_loc/ & 
      (geom%rg(Nr_loc)**2-geom%rg(0)**2)

    ns0norm       = ns0norm_tmp
    ns0(0:Nr_loc) = ns0(0:Nr_loc)/ns0norm
  end subroutine init_ns0_3


  !-------------------------------------------------
  ! derivative of ionic density profile
  !  dns0(r)/dr = -ns0(r)*(1/Ln)*func_tanh_3(r)
  !-------------------------------------------------
  subroutine init_dns0dr_3(geom,inv_Lns0,func_tanh,ns0,dns0dr)
    type(geometry)            , intent(in)  :: geom
    real(RKIND)               , intent(in)  :: inv_Lns0
    real(RKIND), dimension(0:), intent(in)  :: func_tanh   ! [0:Nr]
    real(RKIND), dimension(0:), intent(in)  :: ns0         ! [0:Nr]
    real(RKIND), dimension(0:), intent(out) :: dns0dr      ! [0:Nr]

    integer :: ir, Nr_loc
        
    !*** local variable initialization ***
    Nr_loc = geom%Nr
        
    !*** temperature gradient profile for electron ***
    do ir = 0,Nr_loc
      dns0dr(ir) = -ns0(ir)*inv_Lns0*func_tanh(ir)
    end do
  end subroutine init_dns0dr_3

       
  !------------------------------------------------------
  ! ionic temperature profile initialization
  !  this profile is given by the temperature gradient 
  !  profile, defined by 3 parameters : 
  !    - kappa_Ts0, rpeak and deltar
  !  and
  !    1/Ts0(r) dTs0(r)/dr = -invLTs0*func_tanh_3(r)
  !  with (1/LTs0) = kappa_Ts0/R
  !------------------------------------------------------
  subroutine init_Ts0_3(geom,rpeak,invLTs0,func_tanh,Ts0)
    use globals, only : R0
    use globals, only : pSpecies_id, rflux_driven
    type(geometry)            , intent(in)  :: geom
    real(RKIND)               , intent(in)  :: rpeak
    real(RKIND)               , intent(in)  :: invLTs0
    real(RKIND), dimension(0:), intent(in)  :: func_tanh   ! [0:Nr]
    real(RKIND), dimension(0:), intent(out) :: Ts0         ! [0:Nr]
    
    integer     :: ir, Nr_loc
    real(RKIND) :: dr_loc
    real(RKIND) :: invLTs0_r, invLTs0_min, shift
    real(RKIND) :: tmp, w0, w1, Ts0_norm
    
    !*** local variable initialization ***
    Nr_loc = geom%Nr
    dr_loc = geom%dr

    !*** Define minimum R/LT in buffer regions: ***
    !*** R/LT_min = 0 if Flux Driven, 1 otherwise ***
    invLTs0_min = (1._RKIND-rflux_driven)*1._RKIND/R0

    !*** compute Ts0 solution of :                   ***
    !***  2/(Ts0(r)+Ts0(r-1))*(Ts0(r)-Ts0(r-1))/dr   ***
    !***                  = -(1/LTs0)*func_tanh(r)   ***
    Ts0(0) = 2._RKIND    !keV
    shift = invLTs0/(invLTs0/invLTs0_min-1._RKIND)
    do ir = 1,Nr_loc
      invLTs0_r = (invLTs0*func_tanh(ir)+shift)/(invLTs0+shift)*invLTs0
      tmp       = -0.5_RKIND*dr_loc*invLTs0_r
      Ts0(ir)   = (1._RKIND+tmp)/(1._RKIND-tmp)*Ts0(ir-1)
    enddo
    
    !*** normalisation of the temperature to 1 at r=rpeak ***
    ir            = int((rpeak-geom%rg(0))/dr_loc)
    w1            = (rpeak-geom%rg(ir))/dr_loc
    w0            = 1._RKIND-w1
    Ts0_norm      = w0*Ts0(ir)+w1*Ts0(ir+1)
    Ts0(0:Nr_loc) = (Ts0(0:Nr_loc)/Ts0_norm)
  end subroutine init_Ts0_3
  
  
  !-------------------------------------------------------
  ! ionic temperature gradient profile initialization
  !  this profile is by 3 parameters : 
  !    - kappa_Ts0, rpeak and deltar
  !  and
  !    1/Ts0(r) dTs0(r)/dr = -(1/LTs0)*func_tanh6(r)
  !  with (1/LTs0) = kappa_Ts0/R
  !-------------------------------------------------------
  subroutine init_dlogTs0dr_3(geom,invLTs0,func_tanh,dlogTs0dr)
    use globals, only : R0
    use globals, only : pSpecies_id, rflux_driven
    type(geometry)            , intent(in)  :: geom
    real(RKIND)               , intent(in)  :: invLTs0
    real(RKIND), dimension(0:), intent(in)  :: func_tanh  ! [0:Nr]
    real(RKIND), dimension(0:), intent(out) :: dlogTs0dr  ! [0:Nr]
    
    real(RKIND) :: invLTs0_r, invLTs0_min, shift
    integer     :: ir, Nr_loc

    !*** local variable initialization ***
    Nr_loc = geom%Nr

    !*** Define minimum R/LT in buffer regions: ***
    !*** R/LT_min = 0 if Flux Driven, 1 otherwise ***
    invLTs0_min = (1._RKIND-rflux_driven)*1._RKIND/R0

    shift = invLTs0/(invLTs0/invLTs0_min-1._RKIND)
    do ir = 0,Nr_loc
      invLTs0_r     = (invLTs0*func_tanh(ir)+shift)/(invLTs0+shift)*invLTs0
      dlogTs0dr(ir) = -invLTs0_r 
    end do
  end subroutine init_dlogTs0dr_3


  !-------------------------------------------------------
  ! electronic temperature profile initialization
  !  this profile is given by the temperature gradient 
  !  profile, defined by 3 parameters : 
  !    - kappa_Te0, rpeak and deltar
  !  and
  !    1/Te0(r) dTe0(r)/dr = -invLTe0*func_tanh_3(r)
  !  with (1/LTe0) = kappa_Te0/R
  !-------------------------------------------------------
  subroutine init_Te0_3(geom,rpeak,invLTe0,tau0,Ts0_rmin,func_tanh,Te0)
    type(geometry)            , intent(in)  :: geom
    real(RKIND)               , intent(in)  :: rpeak
    real(RKIND)               , intent(in)  :: invLTe0
    real(RKIND)               , intent(in)  :: tau0
    real(RKIND)               , intent(in)  :: Ts0_rmin
    real(RKIND), dimension(0:), intent(in)  :: func_tanh   ! [0:Nr]
    real(RKIND), dimension(0:), intent(out) :: Te0         ! [0:Nr]

    integer     :: ir, Nr_loc
    real(RKIND) :: dr_loc, rth, tmp, w0, w1
    real(RKIND) :: Te0_norm
        
    !*** local variable initialization ***
    Nr_loc = geom%Nr
    dr_loc = geom%dr
            
    !*** compute Te0 solution of :                 ***
    !***  2/(Te0(r)+Te0(r-1))*(Te0(r)-Te0(r-1))/dr    ***
    !***                  = -(1/LTe0)*func_tanh_3(r) ***
    Te0(0) = Ts0_rmin  !Rk: Ts0_rmin = Ts0(0)
    do ir = 1,Nr_loc
      tmp     = -0.5_RKIND*dr_loc * &
        invLTe0*func_tanh(ir)
      Te0(ir) = (1._RKIND+tmp)/(1._RKIND-tmp)*Te0(ir-1)
    enddo
    
    !*** normalisation of the temperature to 1 at r=rpeak ***
    ir            = int((rpeak-geom%rg(0))/dr_loc)
    w1            = (rpeak-geom%rg(ir))/dr_loc
    w0            = 1._RKIND-w1
    Te0_norm      = w0*Te0(ir)+w1*Te0(ir+1)
    Te0(0:Nr_loc) = (Te0(0:Nr_loc)/Te0_norm)/tau0 
  end subroutine init_Te0_3


  !----------------------------------------------------
  ! derivative of electronic temperature profile 
  !  dTe0(r)/dr = -Te0(r)*(1/LTe0)*func_tanh(r)
  !----------------------------------------------------
  subroutine init_dTe0dr_3(geom,invLTe0,func_tanh,Te0,dTe0dr)
    type(geometry)            , intent(in)  :: geom
    real(RKIND)               , intent(in)  :: invLTe0
    real(RKIND), dimension(0:), intent(in)  :: func_tanh   ! [0:Nr]
    real(RKIND), dimension(0:), intent(in)  :: Te0         ! [0:Nr]
    real(RKIND), dimension(0:), intent(out) :: dTe0dr      ! [0:Nr]

    integer :: ir, Nr_loc
        
    !*** local variable initialization ***
    Nr_loc = geom%Nr
        
    !*** temperature gradient profile for electron ***
    do ir = 0,Nr_loc
      dTe0dr(ir) = -Te0(ir)*invLTe0*func_tanh(ir)
    end do
  end subroutine init_dTe0dr_3


  !----------------------------------------------------------- 
  ! computation of func(tanh) = [tanh(0.5*(rg-rmin-rbuff1)) 
  !                - tanh(0.5*(rg-rmax+rbuff2))]/max(hr) 
  ! REMARK: This profile is the same than the one defined
  !         in the subroutine 'init_func_tanh_2' except
  !         the fact that it is not normalized by the
  !         radial mesh size.
  ! REMARK 2: This profile has the advantage of being 
  !           smoother than the one defined in 
  !           'init_func_tanh_2', so it must be preferred
  !           for simulations, but becarefule it cannot 
  !           be used for invariance test comparisons.
  !---------------------------------------------------------- 
  subroutine init_func_tanh_4(geom,func_tanh)
    type(geometry)            , intent(in)  :: geom
    real(RKIND), dimension(0:), intent(out) :: func_tanh   ! [0:Nr]

    integer     :: ir, Nr_loc
    real(RKIND) :: ri, Lr_loc
    real(RKIND) :: rmin, rmax, rp
    real(RKIND) :: rbuff1, rbuff2
    real(RKIND) :: func_tanh_tmp, max_func
        
    !*** local variable initialization ***
    Nr_loc = geom%Nr
    Lr_loc = geom%Lr
    
    !*** computation of func(tanh) and its maximum ***
    rbuff1   = 0.2_RKIND*Lr_loc 
    rbuff2   = 0.2_RKIND*Lr_loc 
    rmin     = geom%rg(0)
    rmax     = geom%rg(Nr_loc)
    rp       = geom%rg(Nr_loc/2)
    max_func = 0._RKIND
    do ir = 0,Nr_loc
      ri = geom%rg(ir)
      func_tanh_tmp = tanh(0.5_RKIND*(ri-rmin-rbuff1)) - &
        tanh(0.5_RKIND*(ri-rmax+rbuff2))
      func_tanh(ir) = func_tanh_tmp
      max_func      = max(func_tanh_tmp,max_func)
    end do
    func_tanh = func_tanh/(max_func)
  end subroutine init_func_tanh_4


  !-----------------------------------------------------------
  ! Initialisation of the hyperbolic tangent 
  !   used for profile definition
  !-----------------------------------------------------------
  subroutine init_func_tanh(geom,Sprofile_choice, &
    deltar_Ts0,a,coef_diffr,buffer_location,func_tanh)
    use globals, only : 
    type(geometry)            , intent(in)  :: geom
    integer                   , intent(in)  :: Sprofile_choice
    real(RKIND)               , intent(in)  :: deltar_Ts0
    real(RKIND)               , intent(in)  :: a
    real(RKIND)               , intent(in)  :: coef_diffr
    real(RKIND)               , intent(in)  :: buffer_location
    real(RKIND), dimension(0:), intent(out) :: func_tanh   ! [0:Nr]

    select case (Sprofile_choice)
    case (1)
      call init_func_tanh_1(geom,func_tanh)
    !*** Required for invariance test comparisons ***
    case (2)
      call init_func_tanh_2(geom,func_tanh)
    case (3)
      call init_func_tanh_3(geom,deltar_Ts0,a,coef_diffr, &
    buffer_location,func_tanh)
    !*** Same profile than case (2) but without ***
    !***  normalization by the radial mesh step ***
    case (4)
      call init_func_tanh_4(geom,func_tanh)
    case default
      if (pglobal_id.eq.0) then
        print*, 'profile_choice = ', Sprofile_choice, &
          ' is not a proper choice'
        stop
      end if
    end select
  end subroutine init_func_tanh


  !-----------------------------------------------------------
  ! Initialisation of ion density profile
  !  + computation of the first and second derivatives
  !-----------------------------------------------------------
  subroutine init_ns0(geom,Sdensity_filename, &
    Sprofile_choice,rpeak,inv_Lns0,deltar_ns0,func_tanh, &
    ns0,dns0dr,d2ns0dr2)
    use globals            , only : pSpecies_id
    use utils_module       , only : deriv1
    use read_profile_module, only : read_ascii_profile
    type(geometry)            , intent(in)  :: geom
    character(LEN=50),optional, intent(in)  :: Sdensity_filename
    integer                   , intent(in)  :: Sprofile_choice
    real(RKIND)               , intent(in)  :: rpeak
    real(RKIND)               , intent(in)  :: inv_Lns0
    real(RKIND)               , intent(in)  :: deltar_ns0
    real(RKIND), dimension(0:), intent(in)  :: func_tanh    ! [0:Nr]
    real(RKIND), dimension(0:), intent(out) :: ns0          ! [0:Nr]
    real(RKIND), dimension(0:), intent(out) :: dns0dr       ! [0:Nr]
    real(RKIND), dimension(0:), intent(out) :: d2ns0dr2     ! [0:Nr]

    if (.not.present(Sdensity_filename)) then
      select case (Sprofile_choice)
      case (1)
        call init_ns0_1(geom,rpeak,inv_Lns0,deltar_ns0,ns0)
        call init_dns0dr_1(geom,rpeak,inv_Lns0,deltar_ns0,ns0,dns0dr)
      !*** Required for invariance test comparisons ***
      case (2)
        call init_ns0_2(geom,rpeak,inv_Lns0,deltar_ns0,func_tanh,ns0)
        call init_dns0dr_2(geom,rpeak,inv_Lns0,deltar_ns0,func_tanh, &
          ns0,dns0dr)
      case (3)
        call init_ns0_3(geom,inv_Lns0,func_tanh,ns0)
        call init_dns0dr_3(geom,inv_Lns0,func_tanh,ns0,dns0dr)
      !*** Same profile than case (2) but without  ***
      !***  normalization by the radial mesh step, ***
      !***  for the hyperbolic tangent shape       ***
      !*** In this case, the profile is smoother   ***
      !***  than in case 2.                        ***
      case (4)
        call init_ns0_2(geom,rpeak,inv_Lns0,deltar_ns0,func_tanh,ns0)
        call init_dns0dr_2(geom,rpeak,inv_Lns0,deltar_ns0,func_tanh, &
          ns0,dns0dr)
      end select
    else
      call read_ascii_profile(Sdensity_filename,geom,ns0)
      call deriv1(ns0,dns0dr,geom%Nr,geom%dr,0)
    end if
    !*** computation of the second derivative of ns0 ***
    call deriv1(dns0dr,d2ns0dr2,geom%Nr,geom%dr,0)
  end subroutine init_ns0


  !-----------------------------------------------------------
  ! Initialisation of electron temperature profile
  !  + computation of the first and second derivatives
  !-----------------------------------------------------------
  subroutine init_Te0(geom,Sread_Te0,Sprofile_choice, &
    rpeak,invLTe0,deltar_Te0,tau0,Ts0_rmin,func_tanh,Te0,dTe0dr)
    use utils_module       , only : deriv1
    use read_profile_module, only : read_ascii_profile
    type(geometry)            , intent(in)  :: geom
    logical                   , intent(in)  :: Sread_Te0
    integer                   , intent(in)  :: Sprofile_choice
    real(RKIND)               , intent(in)  :: rpeak
    real(RKIND)               , intent(in)  :: invLTe0
    real(RKIND)               , intent(in)  :: deltar_Te0
    real(RKIND)               , intent(in)  :: tau0
    real(RKIND)               , intent(in)  :: Ts0_rmin
    real(RKIND), dimension(0:), intent(in)  :: func_tanh   ! [0:Nr]
    real(RKIND), dimension(0:), intent(out) :: Te0         ! [0:Nr]
    real(RKIND), dimension(0:), intent(out) :: dTe0dr      ! [0:Nr]

    if (.not.Sread_Te0) then
      select case (Sprofile_choice)
      case (1)
        call init_Te0_1(geom,rpeak,invLTe0,deltar_Te0,tau0,Ts0_rmin,Te0)
        call init_dTe0dr_1(geom,rpeak,invLTe0,deltar_Te0,Te0,dTe0dr)
      !*** Required for invariance test comparisons ***
      case (2)
        call init_Te0_2(geom,rpeak,invLTe0,deltar_Te0,tau0,Ts0_rmin,func_tanh,Te0)
        call init_dTe0dr_2(geom,rpeak,invLTe0,deltar_Te0,func_tanh,Te0,dTe0dr)
      case (3)
        call init_Te0_3(geom,rpeak,invLTe0,tau0,Ts0_rmin,func_tanh,Te0)
        call init_dTe0dr_3(geom,invLTe0,func_tanh,Te0,dTe0dr)
      !*** Same profile than case (2) but without  ***
      !***  normalization by the radial mesh step, ***
      !***  for the hyperbolic tangent shape       ***
      !*** In this case, the profile is smoother   ***
      !***  than in case 2.                        ***
      case (4)
        call init_Te0_2(geom,rpeak,invLTe0,deltar_Te0,tau0,Ts0_rmin,func_tanh,Te0)
        call init_dTe0dr_2(geom,rpeak,invLTe0,deltar_Te0,func_tanh,Te0,dTe0dr)
      end select
    else
      call read_ascii_profile('Te0.dat',geom,Te0)
      call deriv1(Te0,dTe0dr,geom%Nr,geom%dr,0)
    end if
  end subroutine init_Te0


  !-----------------------------------------------------------
  ! Initialisation of ion temperature profile of species s
  !  + computation of the first and second derivatives
  !-----------------------------------------------------------
  subroutine init_Ts0(geom,Sread_Ts0,Sprofile_choice, &
    rpeak,invLTs0,deltar_Ts0,func_tanh,Ts0,dlogTs0dr,dTs0dr,d2Ts0dr2)
    use globals            , only : pSpecies_id
    use utils_module       , only : deriv1
    use read_profile_module, only : read_ascii_profile
    type(geometry)            , intent(in)  :: geom
    logical                   , intent(in)  :: Sread_Ts0 
    integer                   , intent(in)  :: Sprofile_choice
    real(RKIND)               , intent(in)  :: rpeak
    real(RKIND)               , intent(in)  :: invLTs0
    real(RKIND)               , intent(in)  :: deltar_Ts0
    real(RKIND), dimension(0:), intent(in)  :: func_tanh    ! [0:Nr]
    real(RKIND), dimension(0:), intent(out) :: Ts0          ! [0:Nr]
    real(RKIND), dimension(0:), intent(out) :: dlogTs0dr    ! [0:Nr]
    real(RKIND), dimension(0:), intent(out) :: dTs0dr       ! [0:Nr]
    real(RKIND), dimension(0:), intent(out) :: d2Ts0dr2     ! [0:Nr]

    integer           :: ir_loc
    character(LEN=20) :: temperature_filename

    if (.not.Sread_Ts0) then
      select case (Sprofile_choice)
      case (1)
        call init_Ts0_1(geom,rpeak,invLTs0,deltar_Ts0,Ts0)
        call init_dlogTs0dr_1(geom,rpeak,invLTs0,deltar_Ts0,dlogTs0dr)
      !*** Required for invariance test comparisons ***
      case (2)
        call init_Ts0_2(geom,rpeak,invLTs0,deltar_Ts0,func_tanh,Ts0)
        call init_dlogTs0dr_2(geom,rpeak,invLTs0,deltar_Ts0,func_tanh,dlogTs0dr)
      case (3)
        call init_Ts0_3(geom,rpeak,invLTs0,func_tanh,Ts0)
        call init_dlogTs0dr_3(geom,invLTs0,func_tanh,dlogTs0dr)
      !*** Same profile than case (2) but without  ***
      !***  normalization by the radial mesh step, ***
      !***  for the hyperbolic tangent shape       ***
      !*** In this case, the profile is smoother   ***
      !***  than in case 2.                        ***
      case (4)
        call init_Ts0_2(geom,rpeak,invLTs0,deltar_Ts0,func_tanh,Ts0)
        call init_dlogTs0dr_2(geom,rpeak,invLTs0,deltar_Ts0,func_tanh,dlogTs0dr)
      end select
      !-> computation of dTs0/dr
      do ir_loc = 0,geom%Nr
        dTs0dr(ir_loc) = dlogTs0dr(ir_loc)*Ts0(ir_loc)
      end do
    else
      write(temperature_filename,'(A,I1,A)') "Ts0_sp", pSpecies_id, ".dat"
      call read_ascii_profile(temperature_filename,geom,Ts0)
      call deriv1(Ts0,dTs0dr,geom%Nr,geom%dr,0)
      !-> computation of dlogTs0/dr
      do ir_loc = 0,geom%Nr
        dlogTs0dr(ir_loc) = dTs0dr(ir_loc)/Ts0(ir_loc)
      end do
    end if
    !*** computation of the second derivative of Ts0 ***
    call deriv1(dTs0dr,d2Ts0dr2,geom%Nr,geom%dr,0)
  end subroutine init_Ts0


  !----------------------------------------------------
  ! Initialization of the radial profile of the
  !  buffer region
  !----------------------------------------------------
  subroutine init_buffer_region(ithis,geom)
    use globals, only : pSpecies_id, rflux_driven, buffer_asymmetry                                  
    use globals, only : & 
      buffer_location, buffer_stiffness, &
      deltat, pglobal_id, cible, Nbproc_r, &
      numfmt_rst, nb_restart_diag
    use Io_Unit_module, only: IU_open, IU_close
    
    type(init_profile), intent(inout)  :: ithis
    type(geometry)    , intent(in)     :: geom

    real(RKIND)        :: rho, Lr_loc, mask, rho_sub
    real(RKIND)        :: buffer_stiffness_max
    real(RKIND)        :: cst, frac, test
    real(RKIND)        :: asymmetric                     
    integer            :: ir, Nr_loc, Nr_proc
    logical            :: bufferOK
    character(LEN=100) :: parameter_file
    integer:: udata

    real(RKIND), dimension(0:geom%Nr) :: Bshape

    !*** local variable initialization ***
    Nr_loc  = geom%Nr
    Nr_proc = (Nr_loc+1)/Nbproc_r   
    Lr_loc  = geom%Lr
    rho_sub = (geom%rg(Nr_proc-1)-geom%rg(0))/Lr_loc
    !*** asymmetric=0.5 if flux_driven and buffer_asymmetry=true,
    !*** asymmetric=1.0 otherwise
    !*** => leads to larger buffer region at the inner boundary
    !***    when the source is active
    if (buffer_asymmetry) then                           
      asymmetric = 1._RKIND - rflux_driven/2._RKIND      
    else                                                 
      asymmetric = 1._RKIND
    end if                                               
    cst  = 1.e-7_RKIND
    frac = 0.4_RKIND
    !*** compute max buffer_stiffness such that Bshape vanishes
    !*** at rho_sub if buffer_location=frac*rho_sub
    buffer_stiffness_max = (1._RKIND-frac)*rho_sub/atanh(1._RKIND-cst)
    !*** re-define buffer_stiffness if buffer_stiffness>buffer_stiffness_max:
    !*** test=0 if buffer_stiffness>buffer_stiffness_max, 1 otherwise
    test = sign(1._RKIND,buffer_stiffness_max-buffer_stiffness)
    test = 0.5_RKIND*(test+abs(test))
    if (Nbproc_r.gt.1) then
      buffer_stiffness = test*buffer_stiffness + &
        (1._RKIND-test)*buffer_stiffness_max
      if (pglobal_id.eq.0) then
        write(6,*) '----------------------------------------------'
        write(6,*) 'WARNING: buffer region too large radially'
        write(6,*) '         new buffer_stiffness = ',buffer_stiffness
        write(6,*) '----------------------------------------------'
        write(parameter_file,'(A,'//numfmt_rst//',A)') &
          "gysela_data", nb_restart_diag, ".out"
        call IU_open(udata,file = parameter_file,status = 'UNKNOWN', &
          position = 'APPEND', form = 'FORMATTED')
        write(udata,*) 'WARNING: buffer region too large radially'
        write(udata,*) '         new buffer_stiffness = ',buffer_stiffness
        call IU_close(udata)
      end if
    end if
    !*** initialisation of the profiles of the
    !*** dissipative coefficients (diffusion & Krook)
    !*** for buffer regions at radial boundaries
    if (buffer_stiffness.ne.0._RKIND) then
      bufferOK = .FALSE.
      do while (.not.bufferOK)
        do ir = 0,Nr_loc
          rho        = (geom%rg(ir)-geom%rg(0))/Lr_loc
          Bshape(ir) = ON + &
            HF*(tanh((rho-(ON-asymmetric*buffer_location)) / &
            buffer_stiffness) - &
            tanh((rho-buffer_location)/buffer_stiffness))
          mask       = 0.5_RKIND * &
            (1._RKIND+sign(1._RKIND,Bshape(ir)-cst))
          Bshape(ir) = mask*Bshape(ir)
        end do
        if ((Nbproc_r.gt.1).and.(Bshape(Nr_proc-1).ne.0._RKIND)) &
          then
          buffer_location = buffer_location*0.9_RKIND
          if(pglobal_id.eq.0) then
            write(6,*) '-------------------------------------------'
            write(6,*) 'WARNING: buffer region too large radially'
            write(6,*) '         new buffer_location = ',buffer_location
            write(6,*) '-------------------------------------------'
            write(parameter_file,'(A,'//numfmt_rst//',A)') &
              "gysela_data", nb_restart_diag, ".out"
            call IU_open(udata,file = parameter_file,status = 'UNKNOWN', &
              position = 'APPEND',form = 'FORMATTED')
            write(udata,*) 'WARNING: buffer region too', &
              ' large radially'
            write(udata,*) '         new buffer_location = ', &
              buffer_location
            call IU_close(udata)
          end if
        else
          bufferOK = .TRUE.
        end if
      end do
    else
      do ir = 0,Nr_loc
        Bshape(ir) = 0._RKIND
      end do
    end if
    ithis%buffer_r(0:geom%Nr) = Bshape(0:geom%Nr)
  end subroutine init_buffer_region


  !---------------------------------------------- 
  ! compute the 1D-spline coefficient of Te0
  ! ( used for cubic spline interpolation )
  ! ( for example in energy computation )
  !---------------------------------------------- 
  subroutine compute_Te0coef1d(geom,BCr_left,BCr_right, &
    Te0,dTe0dr,nspline1d_r,Te0_coef1d)
    type(geometry)             , intent(in)    :: geom
    integer                    , intent(in)    :: BCr_left
    integer                    , intent(in)    :: BCr_right
    real(RKIND), dimension(0:) , intent(in)    :: Te0          ! [0:Nr]
    real(RKIND), dimension(0:) , intent(in)    :: dTe0dr       ! [0:Nr]
    type(nspline1d)            , intent(inout) :: nspline1d_r
    real(RKIND), dimension(-1:), intent(out)   :: Te0_coef1d   ! [-1:Nr+1]

    real(RKIND)                 :: dr_loc 
    integer                     :: Nr_loc
    real(RKIND), dimension(0:1) :: deriv_Te0

    !*** local variable initialization ***
    Nr_loc = geom%Nr
    dr_loc = geom%dr

    !*** compute the cubic spline coefficient ***
    deriv_Te0(0) = dTe0dr(Nr_loc)
    deriv_Te0(1) = dTe0dr(0)
    call natural_spline_coef(nspline1d_r,Te0,&
      BCr_left,BCr_right,deriv_Te0)
    Te0_coef1d   = nspline1d_r%scoef
  end subroutine compute_Te0coef1d


  !---------------------------------------------- 
  ! compute the 1D-spline coefficient of Te0
  ! ( used for cubic spline interpolation )
  ! ( for example in energy computation )
  !---------------------------------------------- 
  subroutine compute_Ts0coef1d(geom,BCr_left,BCr_right, &
    Ts0,dlogTs0dr,nspline1d_r,Ts0_coef1d)
    type(geometry)             , intent(in)    :: geom
    integer                    , intent(in)    :: BCr_left
    integer                    , intent(in)    :: BCr_right
    real(RKIND), dimension(0:) , intent(in)    :: Ts0          ! [0:Nr]
    real(RKIND), dimension(0:) , intent(in)    :: dlogTs0dr    ! [0:Nr]
    type(nspline1d)            , intent(inout) :: nspline1d_r
    real(RKIND), dimension(-1:), intent(out)   :: Ts0_coef1d   ! [-1:Nr+1]

    real(RKIND)                 :: dr_loc 
    integer                     :: Nr_loc
    real(RKIND), dimension(0:1) :: deriv_Ts0

    !*** local variable initialization ***
    Nr_loc = geom%Nr
    dr_loc = geom%dr

    !*** compute the cubic spline coefficient ***
    deriv_Ts0(0) = dlogTs0dr(Nr_loc)*Ts0(Nr_loc)
    deriv_Ts0(1) = dlogTs0dr(0)*Ts0(0)
    call natural_spline_coef(nspline1d_r,Ts0,BCr_left,BCr_right,deriv_Ts0)
    Ts0_coef1d   = nspline1d_r%scoef
  end subroutine compute_Ts0coef1d


  !---------------------------------------------- 
  ! compute the 1D-spline coefficient of Te0
  ! ( used for cubic spline interpolation )
  ! ( for example in energy computation )
  !---------------------------------------------- 
  subroutine compute_ns0coef1d(geom,BCr_left,BCr_right, &
    ns0,dns0dr,nspline1d_r,ns0_coef1d)
    type(geometry)             , intent(in)    :: geom
    integer                    , intent(in)    :: BCr_left
    integer                    , intent(in)    :: BCr_right
    real(RKIND), dimension(0:) , intent(in)    :: ns0          ! [0:Nr]
    real(RKIND), dimension(0:) , intent(in)    :: dns0dr       ! [0:Nr]
    type(nspline1d)            , intent(inout) :: nspline1d_r
    real(RKIND), dimension(-1:), intent(out)   :: ns0_coef1d   ! [-1:Nr+1]
    
    real(RKIND)                 :: dr_loc 
    integer                     :: Nr_loc
    real(RKIND), dimension(0:1) :: deriv_ns0
    
    !*** local variable initialization ***
    Nr_loc = geom%Nr
    dr_loc = geom%dr

    !*** compute the cubic spline coefficient ***
    deriv_ns0(0) = dns0dr(Nr_loc)
    deriv_ns0(1) = dns0dr(0) 
    call natural_spline_coef(nspline1d_r,ns0,BCr_left,BCr_right,deriv_ns0)
    ns0_coef1d   = nspline1d_r%scoef
  end subroutine compute_ns0coef1d


  !-------------------------------------------------------- 
  ! Compute iota the inverse of the safety factor
  !  iota(r) = Kshear * 1/q(r) with
  !     q(r) defined by q_profile between 0 and 3:
  !     0: q(r) = infty (cylindrical case)
  !     1: q(r) = q_param1
  !     2: q(r) = q_param1 + q_param2*exp(q_param3*log(ri/a))
  !     3: q(r) = reversed shear from Garbet et al. PoP 2001
  !  and Kshear = 1 if the magnetic shear is taken into 
  !   account and 0 otherwise
  !-------------------------------------------------------- 
  subroutine compute_iota(ithis,geom)
    use globals, only : q_profile, read_q, &
      q_param1, q_param2, q_param3, q_param4, a
    use read_profile_module, only : read_ascii_profile
    type(init_profile), intent(out) :: ithis
    type(geometry)    , intent(in)  :: geom 

    integer     :: ir, Nr_loc    
    real(RKIND) :: ri, qr
    real(RKIND) :: rhomin2, rhomin4, rhomin6, &
      var, rho, rho2, C2, C3

    !*** local variable initialization ***
    Nr_loc = geom%Nr

    if (read_q) then
      !-> reading of the q profile
      call read_ascii_profile('safety_factor.dat',geom,ithis%iota)
      !-> computation of 1/q
      do ir = 0,Nr_loc
        ithis%iota(ir) = 1._RKIND/ithis%iota(ir)
      end do
    else      
      select case (q_profile)
      case (0) !-> iota = 0._RKIND 
        do ir = 0,Nr_loc
          ithis%iota(ir) = 0._RKIND
        end do

      case (1) !-> flat case
        do ir =0,Nr_loc
          ithis%iota(ir) = 1._RKIND/q_param1
        end do
        
      case (2) !-> parabolic case
        do ir = 0,Nr_loc
          ri = geom%rg(ir)
          qr = q_param1    
          if (ri.ne.0._RKIND) &
            qr = qr + q_param2*exp(q_param3*log(ri/a))
          ithis%iota(ir) = 1._RKIND/qr
        end do

      case (3) !-> Reversed shear, Garbet et al. PoP2001 version
        rhomin2 = q_param4*q_param4
        rhomin4 = rhomin2*rhomin2
        rhomin6 = rhomin4*rhomin2
        var     = rhomin4*(1-rhomin2)*(1-rhomin2)
        C2      = (rhomin6*(q_param3-q_param2) + &
          (1-rhomin2)**3*(q_param1-q_param2))/var
        C3      = (rhomin4*(q_param3-q_param2) - &
          (1-rhomin2)**2*(q_param1-q_param2))/var
        do ir = 0,Nr_loc
          ri   = geom%rg(ir)
          rho  = ri/a
          rho2 = rho*rho
          qr   = q_param2 + C2*(rho2-rhomin2)**2 + &
            C3*(rho2-rhomin2)**3
          ithis%iota(ir) = 1._RKIND/qr
        end do

      case default
        if (pglobal_id.eq.0) then 
          print*,'---> The q_profile = ',q_profile,'is not supported'
          print*,'---> q is set to parabolic (q_profile = 2)'
        end if
        do ir = 0,Nr_loc
          ri = geom%rg(ir)
          qr = q_param1    
          if (ri.ne.0._RKIND) &
            qr = qr + q_param2*exp(q_param3*log(ri/a))
          ithis%iota(ir) = 1._RKIND/qr
        end do
        
      end select
    end if
  end subroutine compute_iota


  !------------------------------------------------------------ 
  ! Compute the magnetic shear
  !   s(r) = r/q dq/dr
  !     with q(r) defined by q_profile between 0 and 3:
  !     0: q(r) = infty (cylindrical case)
  !     1: q(r) = q_param1
  !     2: q(r) = q_param1 + q_param2*exp(q_param3*log(ri/a))
  !     3: q(r) = reversed shear from Garbet et al. PoP 2001
  !
  ! Rk : if q_profile=0, then iota = 0. so s(r) = 0
  !------------------------------------------------------------ 
  subroutine compute_magnetic_shear(ithis,geom)
    use globals, only : q_param1, q_param2, q_param3, q_param4, &
      read_q, q_profile, a
    type(init_profile), intent(out) :: ithis
    type(geometry)    , intent(in)  :: geom 

    integer     :: ir, Nr_loc    
    real(RKIND) :: ri
    real(RKIND) :: rhomin2, rhomin4, rhomin6, &
      var, rho, rho2, C2, C3, dqdr

    real(RKIND), dimension(0:geom%Nr) :: q_array

    !*** local variable initialization ***
    Nr_loc = geom%Nr

    if (read_q) then
      !-> reading of the q profile
      ! We compute numerically the shear
      do ir = 0,Nr_loc
        q_array(ir) = 1._RKIND/ithis%iota(ir)
      end do
      call deriv1(q_array,ithis%shear,geom%Nr,geom%dr,0)
      do ir = 0,Nr_loc
        ithis%shear(ir) = ithis%shear(ir)*ithis%iota(ir) * &
          geom%rg(ir)
      end do
    else
      select case (q_profile)
      case (1) !-> flat case
        do ir = 0,Nr_loc
          ithis%shear(ir)=0._RKIND
        end do
      case (2) !-> parabolic case
        do ir = 0,Nr_loc
          ri              = geom%rg(ir)
          ithis%shear(ir) = q_param2*q_param3*ithis%iota(ir) * &
            exp(q_param3*log(ri/a))
        end do
      case (3) !-> Reversed shear, Garbet et al. PoP2001 version
        rhomin2 = q_param4*q_param4
        rhomin4 = rhomin2*rhomin2
        rhomin6 = rhomin4*rhomin2
        var     = rhomin4*(1-rhomin2)*(1-rhomin2)
        C2      = (rhomin6*(q_param3-q_param2)+ &
          (1-rhomin2)**3*(q_param1-q_param2))/var
        C3      = (rhomin4*(q_param3-q_param2)- &
          (1-rhomin2)**2*(q_param1-q_param2))/var
        do ir = 0,Nr_loc
          ri   = geom%rg(ir)
          rho  = ri/a
          rho2 = rho*rho
          dqdr = 4._RKIND*C2*(rho2-rhomin2)*rho/a + &
            6._RKIND*C3*(rho2-rhomin2)**2*rho/a
          ithis%shear(ir) = ri*ithis%iota(ir) * dqdr
        end do
      case default
        if (pglobal_id.eq.0) then 
          print*,'---> The q_profile = ',q_profile,'is not supported'
          print*,'---> q is set to parabolic (q_profile = 2)'
        end if
        do ir = 0,Nr_loc
          ri              = geom%rg(ir)
          ithis%shear(ir) = q_param2*q_param3*ithis%iota(ir) * &
            exp(q_param3*log(ri/a))
        end do
      end select
    end if
  end subroutine compute_magnetic_shear


  !---------------------------------------------------------------
  ! Computation of psi(r) = -\int_0^r r/q(r) dr using the formula:
  !    psi(r) = -\int_0^r0 r/q(r) dr - \int_r0^r r/q(r) dr
  ! Rk : if q_profile=0, then iota = 0. so psi(r) = 0
  !---------------------------------------------------------------
  subroutine compute_psi(ithis,geom)
    type(init_profile), intent(inout) :: ithis
    type(geometry)    , intent(in)    :: geom
    
    integer                           :: ir
    real(RKIND)                       :: ri, iota_ri
    real(RKIND)                       :: psi_r
    real(RKIND), dimension(0:geom%Nr) :: H_r

    ! -> H(r) = -r/q(r)
    H_r = -geom%rg*ithis%iota
    ! -> psi(r0) = -\int_0^r0 H(r) dr = [H(r0)+H(0)]*(r0-0)/2
    ithis%psi(0) = H_r(0)*geom%rg(0)*0.5_RKIND 

    ! -> computation of psi(r) = -\int_r0^r H(r) dr
    !  with psi(ri) = psi(ri-1) + \int_ri-1^ri H(r) dr
    do ir = 1,geom%Nr
      ithis%psi(ir) = ithis%psi(ir-1) + &
        0.5_RKIND*(H_r(ir)+H_r(ir-1))*geom%dr
    end do

    !-> computation of psimin = min(psi) and psimax = max(psi)
    !->  and the corresponding indices ipsimin and ipsimax
    if ( (maxval(ithis%psi)+minval(ithis%psi)) .ne. 0._RKIND ) then 
      ithis%psimin = 1.e+15_RKIND
      ithis%psimax = -1.e+15_RKIND
      do ir = 0,geom%Nr
        ithis%psimin = min(ithis%psimin,ithis%psi(ir))
        ithis%psimax = max(ithis%psimax,ithis%psi(ir))
        if (ithis%psi(ir).eq.ithis%psimin) ithis%ipsimin = ir
        if (ithis%psi(ir).eq.ithis%psimax) ithis%ipsimax = ir
      end do
    else
      ithis%ipsimin = 0
      ithis%ipsimax = 0
    end if
  end subroutine compute_psi


  !---------------------------------------------------------- 
  ! Computation of a profile at the point r = rpeak
  !---------------------------------------------------------
  subroutine compute_value_atrpeak(ithis,geom,fct,fct_atrpeak)
    use globals     , only : Nr
    use utils_module, only : locate
    type(init_profile)             , intent(inout) :: ithis
    type(geometry)                 , intent(in)    :: geom
    real(RKIND)   , dimension(0:Nr), intent(in)    :: fct
    real(RKIND)                    , intent(out)   :: fct_atrpeak

    integer                      :: i, ipos
    real(RKIND), dimension(-1:2) :: sbase_fct
    character(LEN=50)            :: subp

    !*** computation of the cubic spline coefficients ***
    call natural_spline_coef(ithis%nspline1d_r,fct, &
      ithis%BCr_left,ithis%BCr_right)
    !*** interpolation at r=rpeak ***
    subp = " compute_value_atrpeak "//char(0)
    call locate(ithis%rpeak,geom%rg,geom%Nr,geom%dr, &
      subp,ipos)
    call spline_basis(geom%rg(ipos),ithis%rpeak, &
      geom%rg(ipos+1),geom%dr,sbase_fct)
    fct_atrpeak = 0._RKIND
    do i = -1,2
      fct_atrpeak = fct_atrpeak + &
        ithis%nspline1d_r%scoef(ipos+i)*sbase_fct(i)
    enddo
  end subroutine compute_value_atrpeak


  !----------------------------------------- 
  ! Initialization of the parallel velocity profile
  !-----------------------------------------

  subroutine init_Vpar0(ithis,geom)
    use globals, only : read_Vpar0
    use read_profile_module, only : read_ascii_theta_profile
    type(init_profile), intent(inout) :: ithis
    type(geometry)    , intent(in)    :: geom 

    integer :: itheta

    if (read_Vpar0) then
      call read_ascii_theta_profile('Vpar0.dat',geom,ithis%Vpar0)
    else
      do itheta = 0, geom%Ntheta
        ithis%Vpar0(itheta) = 0._RKIND
      end do
    end if
  end subroutine init_Vpar0



  !----------------------------------------- 
  ! Initialization of all the profiles  
  !-----------------------------------------
  subroutine init_all_profile(ithis,geom)
    use globals, only : pglobal_id
    use globals, only : a, coef_diffr,buffer_location, &
      read_ns0, read_Ts0, read_Te0, profile_choice
    use globals, only : Nspecies, major_species, &
      Z0, hat_Zs, species_frac, adiabatic_electrons
    use read_profile_module
    type(init_profile), intent(out) :: ithis
    type(geometry)    , intent(in)  :: geom 
    
    integer           :: ispec, ir
    real(RKIND)       :: iota_rp, dt
    real(RKIND)       :: cs, hatZs
    character(LEN=50) :: density_filename

    !*** profile object construction ***
    call new_init_profile(ithis,geom)

    if (.not.memory_test) then
      !*** initialization of the buffer region profile ***
      call init_buffer_region(ithis,geom)

      do ispec = 0,Nspecies-1
        !*** initialisation of the hyperbolic tangent used ***
        !***   for profile definition                      ***
        call init_func_tanh(geom,profile_choice(ispec), &  
          ithis%deltar_Ts0(ispec),a,coef_diffr(ispec), &   
          buffer_location,ithis%func_tanh(ispec,0:))       
        !*** initialisation of the ion density profile ***
        !--> ns0, dns0dr and d2ns0dr2
        if (.not.read_ns0(ispec)) then          
          call init_ns0(geom=geom, &
            Sprofile_choice=profile_choice(ispec), &
            rpeak=ithis%rpeak,inv_Lns0=ithis%inv_Lns0(ispec), &
            deltar_ns0=ithis%deltar_ns0(ispec), &
            func_tanh=ithis%func_tanh(ispec,0:), &
            ns0=ithis%ns0(ispec,0:),dns0dr=ithis%dns0dr(ispec,0:), &
            d2ns0dr2=ithis%d2ns0dr2(ispec,0:))
        else
          write(density_filename,'(A,I1,A)') "ns0_sp",ispec, ".dat"
          call init_ns0(geom=geom, &
            Sdensity_filename=density_filename, &
            Sprofile_choice=profile_choice(ispec), &
            rpeak=ithis%rpeak,inv_Lns0=ithis%inv_Lns0(ispec), &
            deltar_ns0=ithis%deltar_ns0(ispec), &
            func_tanh=ithis%func_tanh(ispec,0:), &
            ns0=ithis%ns0(ispec,0:),dns0dr=ithis%dns0dr(ispec,0:), &
            d2ns0dr2=ithis%d2ns0dr2(ispec,0:))
        end if
        !----> ns0(r) = cs * ns0(r) with cs the concentration of species s
        ithis%ns0(ispec,0:)      = species_frac(ispec) * &
          ithis%ns0(ispec,0:)
        ithis%dns0dr(ispec,0:)   = species_frac(ispec) * &
          ithis%dns0dr(ispec,0:)
        ithis%d2ns0dr2(ispec,0:) = species_frac(ispec) * &
          ithis%d2ns0dr2(ispec,0:)
        !--> cubic spline for ns0
        call compute_ns0coef1d(geom,ithis%BCr_left,ithis%BCr_right, &
          ithis%ns0(ispec,0:),ithis%dns0dr(ispec,0:), &
          ithis%nspline1d_r,ithis%ns0_coef1d(ispec,-1:))
        !--> value at rpeak
        call compute_value_atrpeak(ithis,geom, &
          ithis%ns0(ispec,0:),ithis%ns0_rp(ispec))
        !*** initialisation of the ion temperature profile ***
        !--> Ts0, dlogTs0/dr, dTs0/dr and d2Ts0dr2
        call init_Ts0(geom,read_Ts0(ispec),profile_choice(ispec), &
          ithis%rpeak,ithis%inv_LTs0(ispec),ithis%deltar_Ts0(ispec), &
          ithis%func_tanh(ispec,0:),ithis%Ts0(ispec,0:), &
          ithis%dlogTs0dr(ispec,0:),ithis%dTs0dr(ispec,0:), &
          ithis%d2Ts0dr2(ispec,0:))
        !-->  cubic spline coefficient for Ts0 
        call compute_Ts0coef1d(geom,ithis%BCr_left,ithis%BCr_right, &
          ithis%Ts0(ispec,0:),ithis%dlogTs0dr(ispec,0:), &
          ithis%nspline1d_r,ithis%Ts0_coef1d(ispec,-1:))
        !--> value at rpeak
        call compute_value_atrpeak(ithis,geom,ithis%Ts0(ispec,0:), &
          ithis%Ts0_rp(ispec))
      end do

      !*** initialisation of the electron temperature profile ***
      !--> Te0, dTe0dr
      !-----> Rk: Ts0_rmin and the hyperbolic tangent 
      !----->       are taken from the first specie 
      call init_Te0(geom,read_Te0,profile_choice(0), &
        ithis%rpeak,ithis%inv_LTe0,ithis%deltar_Te0, &
        ithis%tau0,ithis%Ts0(major_species,0), &
        ithis%func_tanh(major_species,0:), &
        ithis%Te0(0:),ithis%dTe0dr(0:))

      !-->  cubic spline coefficient for Te0
      call compute_Te0coef1d(geom,ithis%BCr_left,ithis%BCr_right, &
        ithis%Te0(0:),ithis%dTe0dr(0:), &
        ithis%nspline1d_r,ithis%Te0_coef1d(-1:))
      
      !*** initialisation of the electron density profile ***
      !--> ne0 = sum_s Z0*hat_Zs*ns0 = sum_s Zs ns0
      if (adiabatic_electrons) then
        do ispec = 0,Nspecies-1
          hatZs = hat_Zs(ispec)
          do ir = 0,geom%Nr
            ithis%ne0(ir) = ithis%ne0(ir) + &
              Z0*hatZs*ithis%ns0(ispec,ir)
          end do
        end do
      end if

      !*** initialisation of the parallel velocity (theta) profile ***
      call init_Vpar0(ithis,geom)

      !*** iota profile initialization ***
      call compute_iota(ithis,geom)
      call compute_magnetic_shear(ithis,geom)
      call compute_psi(ithis,geom)

      !*** initialization of the values at r=rpeak ***
      ! -> psi(rp) with rp = r(peak)
      call compute_value_atrpeak(ithis,geom,ithis%psi,ithis%psi_rp)
      ! -> q(rp) with rp = r(peak)
      call compute_value_atrpeak(ithis,geom,ithis%iota,iota_rp)
      if (iota_rp.ne.0._RKIND) then
        ithis%q_rp = 1._RKIND/iota_rp
      else 
        ithis%q_rp = 0._RKIND
      end if
    end if
  end subroutine init_all_profile
end module init_profile_class
