!**************************************************************
!  Copyright Euratom-CEA
!  Authors : 
!     Virginie Grandgirard (virginie.grandgirard@cea.fr)
!     Chantal Passeron (chantal.passeron@cea.fr)
!     Guillaume Latu (guillaume.latu@cea.fr)
!     Xavier Garbet (xavier.garbet@cea.fr)
!     Philippe Ghendrih (philippe.ghendrih@cea.fr)
!     Yanick Sarazin (yanick.sarazin@cea.fr)
!  
!  This code GYSELA (for GYrokinetic SEmi-LAgrangian) 
!  is a 5D gyrokinetic global full-f code for simulating 
!  the plasma turbulence in a tokamak.
!  
!  This software is governed by the CeCILL-B license 
!  under French law and abiding by the rules of distribution 
!  of free software.  You can  use, modify and redistribute 
!  the software under the terms of the CeCILL-B license as 
!  circulated by CEA, CNRS and INRIA at the following URL
!  "http://www.cecill.info". 
!**************************************************************

#include "config-F.inc"

!----------------------------------------
! file : globals.F90
! date : 16/09/2005
! global variable definition
!----------------------------------------
module globals
  use prec_const
  
  implicit none

  !*************************************************************
  !   GYSELA VERSION
  !*************************************************************
  !
  ! -> version 1.01: First code versioning
  !   - date: 2013-04-09
  !       
  ! -> version 1.02: Correction of Tseq_r computation 
  !   - date: 2013-04-10
  !      Comparaison with 1.01 version is 
  !      no more possible when collisions are taken 
  !      into account because :
  !      (i)  the computation of Tseq_r has been corrected
  !      (ii) the temperature involved for collision operator 
  !            is no more the one associated to the particles
  !            but TGCseq_r associated to the guiding-centers
  ! -> version 1.03: Modification of the fluid moment computation
  !   - date: 2013-08-23 
  !      The fluid moments associated to the guiding-centers and
  !      the particles have been modified to compute the 
  !      conservation laws in presence of several species.
  !      Rk: These changes have no impact on hydrogen but on
  !          all other species.
  !*************************************************************
  real(RKIND), parameter :: code_version = 1.03


  !**********************************************
  !   GLOBAL VARIABLES FOR MPI PARALLELISATION
  !********************************************** 
  integer, parameter :: cible        = 0    
  integer, parameter :: cible_mu     = 0
  integer, parameter :: cible_jstart = 0
  integer, parameter :: cible_istart = 0
  ! Maximum Number of species
  integer :: MAXNSPECIES = 3
  ! Local space domain size in r (i) and theta (j)
  integer :: istart, iend, inumber
  integer :: jstart, jend, jnumber
  ! Local space domain size in phi (k) and vpar (l)
  integer :: kstart, kend, knumber
  integer :: lstart, lend, lnumber
  ! Local space domain size in  r (i) and theta (j) 
  !  with an extra ghost zone : buf
  integer  :: jstart_buf, istart_buf, jend_buf, iend_buf
  ! Size of the local patch (domain decomposition) in 
  ! directions r, theta, phi and vpar
  integer  :: dom_r, dom_theta, dom_phi, dom_vpar
  ! Identity of the processor
  integer  :: pglobal_id
  ! Identity of the species treated by this processor
  integer  :: pSpecies_id
  ! Communicator with all processors having the same pSpecies_id
  integer  :: mpi_comm_species
  ! Identity of the processor in the mpi_comm_species communicator
  integer  :: pInSpecies_id
  ! Number of procs in an mpi_comm_species communicator
  integer  :: Nbproc_species
  ! Communicator holding corresponding PEs in different species
  integer  :: MPI_COMM_CROSS_SPECIES
  ! Number of processors
  integer  :: Nbproc_phi
  integer  :: Nbproc_vpar 
  integer  :: Nbproc_tot
  ! Size of the ghost zone
  integer, parameter :: bufsize = 3
  ! Size of the stencil 
  integer, parameter :: stencil = 10
  ! Index of the local mu value
  integer  :: mu_id
  ! Communicator shared between processors 
  !   that owns the same mu value
  integer  :: mpi_comm_mu
  ! Communicator shared between processors 
  !   that have the same plocal_id
  integer  :: mpi_comm_intermu
  ! Communicator with all the processors having the same mu_id 
  !  and the same jstart
  integer  :: mpi_comm_column
  ! Communicator with all the processors having the same mu_id 
  !  and the same istart
  integer  :: mpi_comm_row
  ! Communicator with all the processors managing a phi value
  integer  :: mpi_comm_moments
  ! Identity of processor in the 'mpi_comm_mu' communicator
  integer  :: plocal_id
  ! Identity of processor in the 'mpi_comm_column' communicator
  integer  :: jstart_id
  ! Identity of processor in the 'mpi_comm_row' communicator
  integer  :: istart_id
  ! Number of procs in 'mpi_comm_mu' communicator
  integer  :: Nbproc_loc
  integer  :: Nbpoints_muid

  !> Indicate if diagnostics should be parallel or not
  logical  :: diag_para

  !> Number of possible parallel process for diagnostic writing
  integer, parameter :: NBRP  = 6

  !> index of the rank of the target for conservation law diagnostics
  !> in diag_targ
  integer, parameter :: DIAG_CL = 1
  !> index of the rank of the target for radial profiles linked to
  !> guiding center diagnostics in diag_targ
  integer, parameter :: DIAG_RPROF_GC = 2
  !> index of the rank of the target for radial profiles linked to
  !> particles diagnostics in diag_targ
  integer, parameter :: DIAG_RPROF_PART = 3
  !> index of the rank of the target for 2D cross-sections of Phi
  !> diagnostics in diag_targ
  integer, parameter :: DIAG_PHI2D = 4
  !> index of the rank of the target for 2D cross-sections of f
  !> diagnostics in diag_targ
  integer, parameter :: DIAG_F2D = 5
  !> index of the rank of the target for Dcoll_rvpar, Dsemi_rvpar,
  !> Vcoll_rvpar, collision_freqr_diag and envelope_r_diag diagnostics
  !> in diag_targ
  integer, parameter :: DIAG_COLLISION = 6
  !> rank of the processto whom the data must be sent for above
  !> diagnostics
  integer, save      :: diag_targ(1:NBRP)

  ! Number of digits that ends the filenames of diagnostics
  integer      , save :: nbdigits   = 5
  character(20), save :: numfmt     = "'_d',i5.5"
  character(20), save :: numfmt_rst = "'_r',i3.3"
  ! Identity of processor responsible for the output file "gysela_res.out"
  integer  :: outputproc
  integer  :: iter_run
  ! h5_lock variable
  integer, save :: h5_lock     = 0
  integer, save :: stop_gysela = 0

  !********************************************
  !   GLOBAL VARIABLES FOR VERSION ENVIRONMENT
  !********************************************
  character(LEN=100), save ::  compile_time    = ""
  character(LEN=100), save ::  compile_user    = ""
  character(LEN=100), save ::  compile_machine = ""
  character(LEN=100), save ::  f_compiler      = ""
  character(LEN=100), save ::  f_compile_flags = ""
  character(LEN=100), save ::  c_compiler      = ""
  character(LEN=100), save ::  c_compile_flags = ""
  character(LEN=100), save ::  compile_libs    = ""
  character(LEN=100), save ::  HDF5_version    = ""

  !********************************************
  !   GLOBAL VARIABLES FOR INPUT DATA
  !********************************************
  integer, parameter :: Nspecies_max = 2

  !-------------------------
  !--->  Parallel data 
  !-------------------------
  namelist /PARALLEL/ Nspecies, &
    Nbproc_r, Nbproc_theta, Nbproc_mu, &
    bloc_phi, Nbthread, transpose4D, &
    large_platform, distrib3D

  integer, save :: Nspecies        = 1
  integer, save :: Nbproc_r        = 1
  integer, save :: Nbproc_theta    = 1
  integer, save :: Nbproc_mu       = 1
  integer, save :: bloc_phi        = 1
  integer, save :: Nbthread        = 1
  logical, save :: transpose4D     = .false.
  logical, save :: transp_velocity = .false.
  ! Indicate if special parallelization is employed for rho_i computation
  logical, save :: large_platform  = .true.
  ! Indicate if the 3D data are distributed
  logical, save :: distrib3D       = .true.

  !-------------------------
  !---> Geometry data 
  !-------------------------
  namelist /MESH/ CylindricalGeometry, &
    Nr, Ntheta, Nphi, Nvpar, Nmu, a, rhomin, rhomax, &
    Ltheta, Lphi, aspect_ratio, nb_vth0, Lmu, mumin

  logical    , save :: CylindricalGeometry = .false.
  integer    , save :: Nr                  = 50
  integer    , save :: Ntheta              = 50
  integer    , save :: Nphi                = 50
  integer    , save :: Nvpar               = 50
  integer    , save :: Nmu                 = 0
  real(RKIND), save :: a                   = 1._RKIND
  !-> a*rhomin < r < a*rhomax              
  real(RKIND), save :: rhomin              = 0._RKIND
  !-> a*rhomin < r < a*rhomax              
  real(RKIND), save :: rhomax              = 1._RKIND
  real(RKIND), save :: &
    Ltheta         = 6.283185307179586476925286766559005768394_RKIND
  real(RKIND), save :: &
    Lphi           = 6.283185307179586476925286766559005768394_RKIND
  real(RKIND), save :: aspect_ratio        = 3._RKIND
  real(RKIND), save :: nb_vth0             = 3._RKIND
  real(RKIND), save :: Lmu                 = 0._RKIND
  real(RKIND), save :: mumin               = 0._RKIND

  !*** Temporary value                     
  integer    , save :: imu0                = 0


  !-------------------------
  !---> Equilibrium data
  !-------------------------
  namelist /EQUIL/ adiabatic_electrons, canonical_profile, &
    tau0, A0, Z0, As, Zs, species_frac, & 
    read_ns0, read_Ts0, read_Te0, read_q, read_Vpar0, &
    profile_choice, rpeak, kappa_ns0, kappa_Ts0, kappa_Te0, &
    deltar_ns0, deltar_Ts0, deltar_Te0, &
    q_profile, q_param1, q_param2, q_param3, q_param4, &
    magnetic_drift, perturb_amplitude, perturb_choice, n, m, &
    FFTpotential_filter_deltam, FFTpotential_filter
  namelist /EQUIL/ TF_ripple, delta_ripple, N_ripple  

  logical    , save :: adiabatic_electrons = .true.
  integer    , save :: canonical_profile   = 0
  real(RKIND), save :: tau0                = 1._RKIND
  !---> charge and mass for normalisation
  real(RKIND), save :: A0                  = 1._RKIND
  real(RKIND), save :: Z0                  = 1._RKIND
  !---> charge and mass of each species
  real(RKIND), dimension(0:Nspecies_max) &
             , save :: As                  = 1._RKIND
  real(RKIND), dimension(0:Nspecies_max) &
             , save :: Zs                  = 1._RKIND
  !---> concentration of each species
  real(RKIND), dimension(0:Nspecies_max) &
             , save :: species_frac        = (/ 1._RKIND,0._RKIND,0._RKIND /)
  ! -> for temperature and density profiles
  !  = 1: 1/T(r) dT(r)/dr = -kappaT*func(cosh^-2)
  !  = 2: 1/T(r) dT(r)/dr = -kappaT*func(tanh)
  logical, dimension(0:Nspecies_max) &
             , save :: read_ns0            = .false.
  logical, dimension(0:Nspecies_max) &
             , save :: read_Ts0            = .false.
  logical    , save :: read_Te0            = .false.
  logical    , save :: read_q              = .false.
  logical    , save :: read_Vpar0          = .false.
  integer, dimension(0:Nspecies_max) &
             , save :: profile_choice      = 1
  real(RKIND), save :: rpeak               = 0.5_RKIND
  real(RKIND), dimension(0:Nspecies_max) &
             , save :: kappa_ns0           = 0._RKIND
  real(RKIND), dimension(0:Nspecies_max) &
             , save :: kappa_Ts0           = 6._RKIND
  real(RKIND), save :: kappa_Te0           = 0._RKIND
  real(RKIND), dimension(0:Nspecies_max) &
             , save :: deltar_ns0          = 0.33_RKIND
  real(RKIND), save :: deltar_Te0          = 0.33_RKIND
  real(RKIND), dimension(0:Nspecies_max) &
             , save :: deltar_Ts0          = 0.33_RKIND
  integer    , save :: q_profile           = 2
  real(RKIND), save :: q_param1            = 1._RKIND
  real(RKIND), save :: q_param2            = 2.7858_RKIND
  real(RKIND), save :: q_param3            = 2.8_RKIND
  real(RKIND), save :: q_param4            = 1._RKIND
  !-> The case magnetic_drift=.false. corresponds to a 
  !     temporary test case where the effects of the
  !     curvature of the magnetic field are not 
  !     taken into account in the advection equations
  logical    , save :: magnetic_drift      = .true.  
  !  -> initial perturbation
  real(RKIND), dimension(0:Nspecies_max) &
             , save :: perturb_amplitude   = 0.001_RKIND
  integer    , save :: m                   = 10
  integer    , save :: n                   = 3
  ! -> Initialization of the perturbation of 
  !      the distribution function
  !    = 1 if initialization with the mode (m,n)
  !    = 2 if initialization with several 
  !        modes 1 < im < m and 1 < in < n 
  !    = 3 if initialization of several in m for one 
  !       mode in n (1 < m and m > n*q)         
  integer    , save :: perturb_choice      = 2   
  real(RKIND), &
        save :: FFTpotential_filter_deltam = 5._RKIND
  integer    , save :: FFTpotential_filter = 0
  ! TF ripple
  logical    , save :: TF_ripple           = .false.
  real(RKIND), save :: delta_ripple        = 0._RKIND
  integer    , save :: N_ripple            = 0


  !-------------------------
  !---> Collision data
  !-------------------------
  namelist /COLLISION_OPERATOR/ collisions, version_coll, &
    energy_transfer, moment_transfer, nustar, &
    coll_FMrefresh, coll_nbstep

  logical    , save :: collisions      = .false.
  !****************************************************************
  ! Choice of the collision operator, according to
  !  the value of 'version_coll':
  !   - = 1 : case where the parallel velocity is forced equal to 0
  !           (developed by G. Dif-Pradalier) 
  !   - = 2 : case where there is no more constrainst 
  !           on the parallel velocity. The collision operator 
  !           is refreshed every 'coll_FMrefresh' iterations, 
  !           by computing the mean temperature (Tmean) and 
  !           the mean velocity (Vmean) which depend on 
  !           (r,theta,phi). (developed by J. Abiteboul)
  !   - = 3 : case 2 above with additionally nustar and the collision
  !           frequency having the correct radial dependence (so far
  !           only correct for ion-ion collisions).
  !           (developed by G. Dif-Pradalier)
  !****************************************************************
  integer    , save :: version_coll    = 1
  logical    , save :: energy_transfer = .false.
  logical    , save :: moment_transfer = .false.
  real(RKIND), save :: nustar          = 0._RKIND
  integer    , save :: coll_FMrefresh  = 0
  integer    , save :: coll_nbstep     = 1


  !-------------------------
  !---> Dissipation data
  !-------------------------
  namelist /BUFFER_REGION/ coef_diffr, coef_diffth, &
    coef_krook, coef_nustar, buffer_asymmetry, &                                  
    buffer_location, buffer_stiffness

  real(RKIND), dimension(0:Nspecies_max) &
             , save :: coef_diffr       = 0._RKIND
  real(RKIND), dimension(0:Nspecies_max) &
             , save :: coef_diffth      = 0._RKIND
  real(RKIND), save :: coef_krook       = 0._RKIND
  real(RKIND), save :: coef_nustar      = 0._RKIND
  logical    , save :: buffer_asymmetry = .false. 
  real(RKIND), save :: buffer_location  = 0._RKIND
  real(RKIND), save :: buffer_stiffness = 1._RKIND


  !-------------------------
  !---> Flux driven system 
  !-------------------------
  namelist /SOURCE/ flux_driven, heat_source, &
    heat_version, Sce0, TS, rhoS, LS, &
    fp_source, fast_part, Sce_fp, Sce_fpplus, &
    TS_fp, v0_fp, rhoS_fp, LS_fp, alpha_trapp, &
    dens_source, Sce_dens, TS_dens, rhoS_dens, LS_dens, &
    mom_source, Sce_mom, TS_mom, rhoS_mom, LS_mom, &
    vor_source, Sce_vor, TS_vor, rhoS_vor, LS_vor

  logical    , save :: flux_driven         = .false.
  logical    , dimension(0:Nspecies_max) &
             , save :: heat_source         = .false.
  integer    , dimension(0:Nspecies_max) &
             , save :: heat_version        = 1
  real(RKIND), dimension(0:Nspecies_max) &
             , save :: Sce0                = 0.01_RKIND
  real(RKIND), dimension(0:Nspecies_max) &
             , save :: TS                  = 1._RKIND
  real(RKIND), dimension(0:Nspecies_max) &
             , save :: rhoS                = 0.1_RKIND
  real(RKIND), dimension(0:Nspecies_max) &
             , save :: LS                  = 0.01_RKIND

  !----------------------------------------------
  ! Source of density :
  !    Actually this source is only           
  !    designed as a source of impurity which 
  !    must be compensated by a negative      
  !    source of the majority species such    
  !    that:                                  
  !       sum_s Zs \int S_ns d3v = 0          
  !                                           
  !    Rk: Actually this option is only       
  !        available for 2 species, i.e       
  !        the major species + one impurity   
  !    Therefore, considering s as the major  
  !    species and s' as the impurity:        
  !     1) The shape and temperature of both  
  !        density sources must be the same   
  !     2) The source intensities are linked  
  !        by S0_ns = -(Zs'/Zs) S0_ns'        
  !-----------------------------------------------
  logical    , dimension(0:Nspecies_max) &
             , save :: dens_source            = .false.
  real(RKIND), dimension(0:Nspecies_max) &
             , save :: Sce_dens               = 0.01_RKIND
  real(RKIND), dimension(0:Nspecies_max) &
             , save :: TS_dens                = 1._RKIND
  real(RKIND), dimension(0:Nspecies_max) &
             , save :: rhoS_dens              = 0.1_RKIND
  real(RKIND), dimension(0:Nspecies_max) &
             , save :: LS_dens                = 0.01_RKIND

  !-------------------------
  ! Source of momentum
  !-------------------------
  logical    , dimension(0:Nspecies_max) &
             , save :: mom_source            = .false.
  real(RKIND), dimension(0:Nspecies_max) &
             , save :: Sce_mom               = 0.01_RKIND
  real(RKIND), dimension(0:Nspecies_max) &
             , save :: TS_mom                = 1._RKIND
  real(RKIND), dimension(0:Nspecies_max) &
             , save :: rhoS_mom              = 0.1_RKIND
  real(RKIND), dimension(0:Nspecies_max) &
             , save :: LS_mom                = 0.01_RKIND

  !-------------------------
  ! Source of vorticity 
  !-------------------------
  logical    , dimension(0:Nspecies_max) &
             , save :: vor_source            = .false.
  real(RKIND), dimension(0:Nspecies_max) &
             , save :: Sce_vor               = 0.01_RKIND
  real(RKIND), dimension(0:Nspecies_max) &
             , save :: TS_vor                = 1._RKIND
  real(RKIND), dimension(0:Nspecies_max) &
             , save :: rhoS_vor              = 0.1_RKIND
  real(RKIND), dimension(0:Nspecies_max) &
             , save :: LS_vor                = 0.01_RKIND

  !-------------------------
  ! Source of fast particles
  !-------------------------
  logical    , dimension(0:Nspecies_max) &
             , save :: fp_source             = .false.
  logical    , dimension(0:Nspecies_max) &
             , save :: fast_part             = .false.
  real(RKIND), dimension(0:Nspecies_max) &
             , save :: Sce_fp                = 0._RKIND
  real(RKIND), dimension(0:Nspecies_max) &
             , save :: Sce_fpplus            = 0.5_RKIND
  real(RKIND), dimension(0:Nspecies_max) &
             , save :: TS_fp                 = 0.5_RKIND
  real(RKIND), dimension(0:Nspecies_max) &
             , save :: v0_fp                 = 2._RKIND
  real(RKIND), dimension(0:Nspecies_max) &
             , save :: rhoS_fp               = 0.1_RKIND
  real(RKIND), dimension(0:Nspecies_max) &
             , save :: LS_fp                 = 0.01_RKIND
  real(RKIND), dimension(0:Nspecies_max) &
             , save :: alpha_trapp           = 0.5_RKIND

  !-----------------------------
  !---> Initial EP distribution 
  !-----------------------------
  namelist /FAST_PARTICLES/ bump_on_tail, fp_rshape, &
    fp_lambda, fp_mom, fp_energy, fp_momgauss, &
    Eb_fp, Emin_fp, n_fp, dn_fp, E_critfp, rho_0fp, &
    deltarhofp, Lambda_0fp, deltaLambdafp, psi_0fp, &
    deltapsifp, zetabar

  logical    , dimension(0:Nspecies_max) &
             , save :: bump_on_tail          = .true.
  logical    , dimension(0:Nspecies_max) &
             , save :: fp_rshape             = .false.
  logical    , dimension(0:Nspecies_max) &
             , save :: fp_lambda             = .false.
  logical    , dimension(0:Nspecies_max) &
             , save :: fp_mom                = .false.
  logical    , dimension(0:Nspecies_max) &
             , save :: fp_energy             = .false.
  logical    , dimension(0:Nspecies_max) &
             , save :: fp_momgauss           = .false.
  real(RKIND), dimension(0:Nspecies_max) &
             , save :: Eb_fp                 = 32._RKIND
  real(RKIND), dimension(0:Nspecies_max) &
             , save :: Emin_fp               = 0._RKIND
  real(RKIND), dimension(0:Nspecies_max) &
             , save :: n_fp                  = 0.017_RKIND
  real(RKIND), dimension(0:Nspecies_max) &
             , save :: dn_fp                 = 0._RKIND
  real(RKIND), dimension(0:Nspecies_max) &
             , save :: E_critfp              = 25._RKIND
  real(RKIND), dimension(0:Nspecies_max) &
             , save :: rho_0fp               = 0.5_RKIND
  real(RKIND), dimension(0:Nspecies_max) &
             , save :: deltarhofp            = 0.05_RKIND
  real(RKIND), dimension(0:Nspecies_max) &
             , save :: Lambda_0fp            = 0.5_RKIND
  real(RKIND), dimension(0:Nspecies_max) &
             , save :: deltaLambdafp         = 0.2_RKIND
  real(RKIND), dimension(0:Nspecies_max) &
             , save :: psi_0fp               = 0._RKIND
  real(RKIND), dimension(0:Nspecies_max) &
             , save :: deltapsifp            = 0.3_RKIND
  real(RKIND), dimension(0:Nspecies_max) &
             , save :: zetabar               = 4._RKIND

  !-------------------------
  !---> Algorithm data
  !-------------------------
  namelist /ALGORITHM/ Vlasov_scheme, deltat,&
    nbiter, dt_diag
  namelist /ALGORITHM/ limiter                          

  !-> Vlasov_scheme :
  !   . BSL   = Backward Semi-Lagrangian scheme
  !   . PSM1D = Parabolic Spline Method with splitting (r,theta) 
  !   . PSM2D = Parabolic Spline Method with direct 
  !             (r,theta) and divergence-free 
  !   . PPM1D = Piecewise Parabolic Method with 
  !             splitting (r,theta) 
  !   . PPM2D = Piecewise Parabolic Method with 
  !             direct (r,theta) and divergence-free  
  character(LEN=5), save :: Vlasov_scheme = 'BSL' 
  !-> time algo specifications
  real(RKIND)     , save :: deltat        = 0.1_RKIND
  integer         , save :: nbiter        = 1
  real(RKIND)     , save :: dt_diag       = 0.1_RKIND
  !-> limiter choice (0=no, 1=entropic, 2=Slope Limited Spline)
  integer         , save :: limiter       = 0
  real(RKIND)     , save :: errmax        = 1.e-6_RKIND

  !-------------------------
  !---> Test data
  !-------------------------
  namelist /TEST/ &
    !--> Tests for general numerics
    integral_vperp, integration_scheme, &
    !--> Tests for QN solver
    solve_QN, QN_version, QN_comm_version, &
    QN_coef_polarization, QN_coef_Phi00, &
    Phi00_BCrmin_Neumann, QN_species_impact, &
    !--> Tests for Vlasov solver
    adaptive_deltat, advec2D_deltat, &
    reduced_deltat, reduced_begin, reduced_end, &
    hffiltertype, hffilterfreq, hffilterfref, &
    deltaF_in_diff, avoidtaylor, deltaF_in_advec, &
    split_linear, clampv, cancel_nonlin, diagf, &
    !--> Tests for RHS of Boltzmann equation
    RHS_only, &
    !--> Tests for equilibrium definition
    feq_choice, canonical_vpar0, hvpar_in_fperturb, &
    !--> Tests with simplified physics
    gyroaverage, BstareqB, single_m, single_n, &
    Rosenbluth_Hinton, &
    !--> Tests for computer sciences
    memory_test, asynchrone_writing, ask_for_checksum, &
    modulo_restart_iter, checkoverflow, cmpr, &
    transpose_version

  !--> Tests for general numerics
  logical    , save :: integral_vperp       = .false.
  integer    , save :: integration_scheme   = 2
  !--> Tests for QN solver
  logical    , save :: solve_QN             = .true. 
  integer    , save :: QN_version           = 1         
  integer    , save :: QN_comm_version      = 1         
  real(RKIND), save :: QN_coef_polarization = 1._RKIND
  integer    , save :: QN_coef_Phi00        = 1
  logical    , save :: Phi00_BCrmin_Neumann = .false.
  real(RKIND), dimension(0:Nspecies_max) &
             , save :: QN_species_impact    = 1._RKIND
  !--> Tests for Vlasov solver
  logical    , save :: adaptive_deltat      = .false.
  real(RKIND), save :: advec2D_deltat       = 0._RKIND
  real(RKIND), save :: reduced_deltat       = 0._RKIND
  real(RKIND), save :: reduced_begin        = -1e30_RKIND
  real(RKIND), save :: reduced_end          = -1e30_RKIND
  integer    , save :: hffiltertype         = 0  ! (0=no,1=FFT,2=wavelet)
  integer    , save :: hffilterfreq         = 0  ! frequency
  integer    , save :: hffilterfref         = 1  ! ref function type
  logical    , save :: deltaF_in_diff       = .true.
  logical    , save :: avoidtaylor          = .false.
  logical    , save :: deltaF_in_advec      = .false.
  logical    , save :: split_linear         = .false.
  logical    , save :: clampv               = .false.
  logical    , save :: cancel_nonlin        = .false.
  logical    , save :: diagf                = .false.
  !--> Tests for RHS of Boltzmann equation
  logical    , save :: RHS_only             = .false.
  !--> Tests for equilibrium definition
  integer    , save :: feq_choice           = 1
  integer    , save :: canonical_vpar0      = 1
  logical    , save :: hvpar_in_fperturb    = .true.
  !--> Tests with simplified physics
  logical    , save :: gyroaverage          = .true.
  logical    , save :: BstareqB             = .false.
  logical    , save :: single_m             = .false.
  logical    , save :: single_n             = .false.
  logical    , save :: Rosenbluth_Hinton    = .false.  
  !--> Tests for computer sciences
  logical    , save :: memory_test          = .false.
  logical    , save :: asynchrone_writing   = .false.
  integer    , save :: ask_for_checksum     = 0
  integer    , save :: modulo_restart_iter  = -1
  logical    , save :: checkoverflow        = .false.
  !> compression level for HDF5 data (-1 for none)
  integer    , save :: cmpr                 = DFLT_HDF5_COMPRESSION

  integer, parameter :: TRANSPOSE_SENDIRECV  = 0
  integer, parameter :: TRANSPOSE_ALLTOALLW  = 1
  !> implementation to use for the transposition.
  !>
  !> Might be one of TRANSPOSE_SENDIRECV or TRANSPOSE_ALLTOALLW
  integer :: transpose_version = DFLT_F4D_TRANSPOSE


  !-------------------------
  !---> Output data
  !-------------------------
  namelist /OUTPUT/ integration_CS, diag_level, &
    Phi3D_saving, Moments3D_saving, f5D_saving, &
    CalviExp_saving, Phi3D_dt_diag, Moments3D_dt_diag, &
    f5D_dt_diag, CalviExp_dt_diag, &
    Sp_saving, Sp_dt_diag, r_diags, &    
    doubleSP_m, doubleSP_n, &           
    rst_saving, enable_dyn_trace, &
    iter_beg_write_trace, iter_end_write_trace

  logical    , save :: integration_CS       = .true.
  integer    , save :: diag_level           = 3
  logical    , save :: Phi3D_saving         = .false.
  logical    , save :: Moments3D_saving     = .false.
  logical    , save :: f5D_saving           = .false.
  !TODO: kept for compatibility with old data files, remove when possible
  logical    , save :: CalviExp_saving      = .false.
  real(RKIND), save :: Phi3D_dt_diag        = 0.1
  real(RKIND), save :: Moments3D_dt_diag    = 0.1
  real(RKIND), save :: f5D_dt_diag          = 0.1
  !TODO: kept for compatibility with old data files, remove when possible
  real(RKIND), save :: CalviExp_dt_diag     = 0.1
  logical    , save :: rst_saving           = .true.
#ifdef TIMER
  logical    , save :: enable_dyn_trace     = .false.
#else
  logical    , save :: enable_dyn_trace     = .true.
#endif
  logical    , save :: mem_dyn_write        = .true.
  integer    , save :: iter_beg_write_trace = 1
  integer    , save :: iter_end_write_trace = 3

  integer    , parameter :: N_rdiags_Spectr = 10
  integer    , parameter :: nb_fulldiag_max = 50
  logical    , save      :: Sp_saving       = .false.
  real(RKIND), save      :: Sp_dt_diag      = 0.1
  integer, &
    dimension(nb_fulldiag_max), save  :: doubleSP_m      = -1
  integer, &
    dimension(nb_fulldiag_max), save  :: doubleSP_n      = -1 
  integer, &
    dimension(1:N_rdiags_Spectr), save :: r_diags = -1


  !********************************************
  !   GLOBAL VARIABLES 
  !********************************************
  !--> Expansion factor for the parallelization of mu 
  !-->   integrals for rho_i computation
  integer :: Nmuscal = 1
  integer :: Nmuloc  = 1

  !--> for normalized charge and mass of species
  real(RKIND), dimension(0:Nspecies_max) :: hat_As
  real(RKIND), dimension(0:Nspecies_max) :: hat_Zs
  !--> majoritary species
  integer :: major_species

  ! Indicate if special parallelization is employed for Bstar3D data
  logical :: transp_Bstar 

  !--> for restart and diagnostics
  logical, save               :: conservative_scheme = .false. 
  logical, &
    dimension(0:Nspecies_max) :: restart
  logical                     :: init_mode
  character(LEN=20)           :: file_name_res       = 'gysela_res.out'
  integer, save               :: uout_res            = 111
  integer, save               :: nbiter_prev         = 0 
  integer, save               :: nb_restart          = 0
  integer, save               :: num_restart         = 0
  integer, save               :: nb_restart_diag     = 0
  integer, parameter          :: nbmoments_GC        = 28
  integer, parameter          :: nbmoments_part      = 11


  !********************************************
  !  BOUNDARY CONDITIONS
  !   = 0 if Dirichlet condition (func=0)
  !   = 1 if Neumann (func'=0)
  !   = 2 if Hermite conditions 
  !       (func'=approximation of the derivate)
  !********************************************
  integer, parameter :: BC_Hermite    = 2
  integer, parameter :: Phi_BCr_left  = BC_Hermite
  integer, parameter :: Phi_BCr_right = BC_Hermite

  ! physical data
  real(RKIND) :: R0            !(major radius)
  logical     :: atleast1_source
  logical     :: atleast1_fp_srce
  !-->  rflux_driven = 1 if flux_driven=TRUE, = 0 otherwise
  real(RKIND) :: rflux_driven

  ! Iteration following
  integer :: iter_glob

  ! Input arguments
  integer     , save :: num_args = 1
  character(LEN=100) :: string   

  ! used for equilbrium part 
  !--> associated to particles
  real(RKIND), dimension(:)  , pointer :: nseq_r
  real(RKIND), dimension(:)  , pointer :: dnseq_dr
  real(RKIND), dimension(:)  , pointer :: Tseq_r
  real(RKIND)                          :: nbions_eq
  !--> associated to guiding-centers
  real(RKIND), dimension(:)  , pointer :: TGCseq_r

  ! used for number of electrons and ions 
  real(RKIND) :: nbelectrons_diag, nbions_diag

  ! min and max error (r-rbar(r,theta,vpar,mu))
  real(RKIND) :: min_err_rbar, max_err_rbar

  ! position for the output saving
  integer     :: ir_save, itheta_save, iphi_save

  ! position for the output f2D saving
  ! for trapped particles
  integer     :: ir_f2D_trapped, itheta_f2D_trapped, &
    iphi_f2D_trapped, ivpar_f2D_trapped, imu_f2D_trapped

  ! position for the output f2D saving
  ! for passing particles
  integer     :: ir_f2D_passing, itheta_f2D_passing, &
    iphi_f2D_passing, ivpar_f2D_passing, imu_f2D_passing

  ! position for the output Phi2D saving
  integer     :: ir_Phi, itheta_Phi, iphi_Phi

  ! start and final position in vpar
  integer     :: vstart, vend

  ! For V and T computation
  real(RKIND), dimension(:,:,:), pointer :: Vmean_loc
  real(RKIND), dimension(:,:,:), pointer :: Tmean_loc
  !--> if restart_save needed
  real(RKIND), dimension(:,:,:), pointer :: Vmean_loc_restart_save
  real(RKIND), dimension(:,:,:), pointer :: Tmean_loc_restart_save

  ! For precomputation of Bstar and contravariant 
  ! components of bstar
  real(RKIND), dimension(:,:,:), pointer :: Bstar_PNrPNthNvpar
  real(RKIND), dimension(:,:,:), pointer :: bstar_gradr_PNrPNthNvpar
  real(RKIND), dimension(:,:,:), pointer :: bstar_gradtheta_PNrPNthNvpar
  real(RKIND), dimension(:,:,:), pointer :: bstar_gradphi_PNrPNthNvpar

  real(RKIND), dimension(:,:,:), pointer :: Bstar_NrNthPNvpar
  real(RKIND), dimension(:,:,:), pointer :: bstar_gradr_NrNthPNvpar
  real(RKIND), dimension(:,:,:), pointer :: bstar_gradtheta_NrNthPNvpar
  real(RKIND), dimension(:,:,:), pointer :: bstar_gradphi_NrNthPNvpar

  ! Partial derivatives of J0.Phi(r,theta,phi)
  real(RKIND), dimension(:,:,:), public, pointer :: dJ0Phidr_PNrPNthNphi
  real(RKIND), dimension(:,:,:), public, pointer :: dJ0Phidtheta_PNrPNthNphi
  real(RKIND), dimension(:,:,:), public, pointer :: dJ0Phidphi_PNrPNthNphi
  real(RKIND), dimension(:,:,:), public, pointer :: dJ0Phidr_NrNthPNphi
  real(RKIND), dimension(:,:,:), public, pointer :: dJ0Phidtheta_NrNthPNphi
  real(RKIND), dimension(:,:,:), public, pointer :: dJ0Phidphi_NrNthPNphi

  real(RKIND), dimension(:,:,:), pointer, public :: J0Phi_PNrPNthNphi
  real(RKIND), dimension(:,:,:), pointer, public :: Phi_NthNphiPNr

  ! Coefficients according to the case treated 
  real(RKIND) :: K_Bdrift


  !********************************************
  !   GLOBAL VARIABLES FOR DISPLACEMENT STATISTICS
  !********************************************
  integer, dimension(:), pointer :: nbpart_rleft_iter
  integer, dimension(:), pointer :: nbpart_rright_iter
  integer, dimension(:), pointer :: nbpart_thleft_iter
  integer, dimension(:), pointer :: nbpart_thright_iter
  integer, dimension(:), pointer :: nbpart_rleft_muid
  integer, dimension(:), pointer :: nbpart_rright_muid
  integer, dimension(:), pointer :: nbpart_thleft_muid
  integer, dimension(:), pointer :: nbpart_thright_muid

  !********************************************
  !   GLOBAL VARIABLES FOR USEFUL ARRAYS
  !********************************************
  !*** -> for QN parallel solver (version 2 et 3) ***
  real(RKIND), &
    dimension(:,:,:,:), pointer :: Rarray_PNrPNthetaNphi_nbM2
  real(RKIND), &
    dimension(:,:,:,:), pointer :: Rarray_send_nbM5
  !*** -> Buffers to send or receive data transmitted   *** 
  !*** with MPI.                                        ***

  real(RKIND), dimension(:), pointer :: RSendBufMPI_A
  real(RKIND), dimension(:), pointer :: RSendBufMPI_B
  real(RKIND), dimension(:), pointer :: RSendBufMPI_C
  real(RKIND), dimension(:), pointer :: RSendBufMPI_D
  real(RKIND), dimension(:), pointer :: RRecvBufMPI_A
  real(RKIND), dimension(:), pointer :: RRecvBufMPI_B
  real(RKIND), dimension(:), pointer :: RRecvBufMPI_C
  real(RKIND), dimension(:), pointer :: RRecvBufMPI_D
  real(RKIND), dimension(:), pointer :: RRecvBufMPI_E

  real(RKIND), &
    dimension(:,:)    , pointer :: Rarray_PNr_nbM5
  real(RKIND), &
    dimension(:,:)    , pointer :: Rarray_Nr_nbM5
  real(RKIND), &
    dimension(:,:)    , pointer :: Rarray_PNthetaNphi_nbM2
  real(RKIND), &
    dimension(:,:)    , pointer :: Rarray_NthetaNphi_nbM2
  real(RKIND), &
    dimension(:,:)    , pointer :: Rarray_PNr_nbM6
  real(RKIND), &
    dimension(:,:)    , pointer :: Rarray_Nr_nbM6
  real(RKIND), &
    dimension(:,:,:,:), pointer :: Rarray_PNrPNthetaNphi_nbM8

  !*** real array allocations ***
  real(RKIND), dimension(:)    , pointer :: Rarray1_Nr
  real(RKIND), dimension(:)    , pointer :: Rarray2_Nr
  real(RKIND), dimension(:)    , pointer :: Rarray3_Nr
  real(RKIND), dimension(:)    , pointer :: Rarray4_Nr
  real(RKIND), dimension(:)    , pointer :: Rarray5_Nr
  real(RKIND), dimension(:)    , pointer :: Rarray6_Nr
  real(RKIND), dimension(:)    , pointer :: Rarray7_Nr
  real(RKIND), dimension(:)    , pointer :: Rarray1_Ntheta
  real(RKIND), dimension(:)    , pointer :: Rarray2_Ntheta
  real(RKIND), dimension(:)    , pointer :: Rarray1_Nphi
  real(RKIND), dimension(:)    , pointer :: Rarray2_Nphi
  real(RKIND), dimension(:)    , pointer :: Rarray1_Nvpar
  real(RKIND), dimension(:)    , pointer :: Rarray2_Nvpar
  real(RKIND), dimension(:)    , pointer :: Rarray3_Nvpar
  real(RKIND), dimension(:)    , pointer :: Rarray4_Nvpar
  real(RKIND), dimension(:)    , pointer :: Rarray5_Nvpar
  real(RKIND), dimension(:)    , pointer :: Rarray6_Nvpar
  real(RKIND), dimension(:)    , pointer :: Rarray7_Nvpar
  real(RKIND), dimension(:)    , pointer :: Rarray_m1Nphip1
  real(RKIND), dimension(:)    , pointer :: Rarray_m1Nvparp1
  real(RKIND), dimension(:,:)  , pointer :: Rarray_1Ntheta
  real(RKIND), dimension(:,:)  , pointer :: Rarray1_NrNtheta
  real(RKIND), dimension(:,:)  , pointer :: Rarray2_NrNtheta
  real(RKIND), dimension(:,:)  , pointer :: Rarray3_NrNtheta
  real(RKIND), dimension(:,:)  , pointer :: Rarray4_NrNtheta
  real(RKIND), dimension(:,:)  , pointer :: Rarray5_NrNtheta
  real(RKIND), dimension(:,:)  , pointer :: Rarray6_NrNtheta
  real(RKIND), dimension(:,:)  , pointer :: Rarray7_NrNtheta
  real(RKIND), dimension(:,:)  , pointer :: Rarray_NrNphi
  real(RKIND), dimension(:,:)  , pointer :: Rarray2_NrNphi
  real(RKIND), dimension(:,:)  , pointer :: Rarray_NrNvpar
  real(RKIND), dimension(:,:)  , pointer :: Rarray_NphiNvpar
  real(RKIND), dimension(:,:)  , pointer :: Rarray_NthetaNvparphi
  real(RKIND), dimension(:,:)  , pointer :: Rarray_NvparNmu
  real(RKIND), dimension(:,:)  , pointer :: Rarray1_m1Nrp1m1Nthetap1
  real(RKIND), dimension(:,:)  , pointer :: Rarray2_m1Nrp1m1Nthetap1
  real(RKIND), dimension(:,:)  , pointer :: Rarray3_m1Nrp1m1Nthetap1
  real(RKIND), dimension(:,:)  , pointer :: Rarray_m1Nphip1m1Nvparp1
  real(RKIND), dimension(:,:,:), pointer :: Rarray1_NrNthetaDomphi
  real(RKIND), dimension(:,:,:), pointer :: Rarray2_NrNthetaDomphi
  real(RKIND), dimension(:,:,:), pointer :: Rarray3_NrNthetaDomphi
  real(RKIND), dimension(:,:,:), pointer :: Rarray4_NrNthetaDomphi
  real(RKIND), dimension(:,:,:), pointer :: Rarray5_NrNthetaDomphi
  real(RKIND), dimension(:,:,:), pointer :: Rarray6_NrNthetaDomphi
  real(RKIND), dimension(:,:,:), pointer :: Rarray1_NrNthetaParphi
  real(RKIND), dimension(:,:,:), pointer :: Rarray1_PNrPNthetaNphi
  real(RKIND), dimension(:,:,:), pointer :: Rarray2_PNrPNthetaNphi
  real(RKIND), dimension(:,:,:), pointer :: Rarray3_PNrPNthetaNphi
  real(RKIND), dimension(:,:,:), pointer :: Rarray4_PNrPNthetaNphi
  complex(RKIND), &
    dimension(:,:,:), pointer :: Carray_1Nrp11Ntheta1Nphi

  !**** Arrays for spectral diagnostic
  complex(RKIND), &
    dimension(:,:,:,:), pointer :: Carray_1Nrp1Nvpar1Ntheta1Nphi

  !*** arrays only used when the 3D data are distributed ***
  !***   (distrib3D=.true.)                              ***
  real(RKIND), dimension(:,:,:), pointer :: Rarray1_NrNthetaPNmu
  real(RKIND), dimension(:,:,:), pointer :: Rarray1_NrNthetaPNphi
  real(RKIND), dimension(:,:,:), pointer :: Rarray2_NrNthetaPNphi
  real(RKIND), dimension(:,:,:), pointer :: Rarray3_NrNthetaPNphi
  real(RKIND), dimension(:,:,:), pointer :: Rarray5_NrNthetaPKNphi
  real(RKIND), dimension(:,:,:), pointer :: Rarray6_NrNthetaPKNphi
  real(RKIND), dimension(:,:,:), pointer :: Rarray7_NrNthetaPKNphi

  !*** arrays only used when the 3D data are not distributed ***
  !***  (distrib3D=.false.)                                  ***
  real(RKIND), dimension(:,:,:), pointer :: Rarray1_NrNthetaNphi
  real(RKIND), dimension(:,:,:), pointer :: Rarray2_NrNthetaNphi
  !--> used when Nspecies.gt.1
  real(RKIND), dimension(:,:,:), pointer :: Rarray1_NrNthetaNphi_Nspec
end module globals
