!**************************************************************
!  Copyright Euratom-CEA
!  Authors : 
!     Virginie Grandgirard (virginie.grandgirard@cea.fr)
!     Chantal Passeron (chantal.passeron@cea.fr)
!     Guillaume Latu (guillaume.latu@cea.fr)
!     Xavier Garbet (xavier.garbet@cea.fr)
!     Philippe Ghendrih (philippe.ghendrih@cea.fr)
!     Yanick Sarazin (yanick.sarazin@cea.fr)
!  
!  This code GYSELA (for GYrokinetic SEmi-LAgrangian) 
!  is a 5D gyrokinetic global full-f code for simulating 
!  the plasma turbulence in a tokamak.
!  
!  This software is governed by the CeCILL-B license 
!  under French law and abiding by the rules of distribution 
!  of free software.  You can  use, modify and redistribute 
!  the software under the terms of the CeCILL-B license as 
!  circulated by CEA, CNRS and INRIA at the following URL
!  "http://www.cecill.info". 
!**************************************************************

#include "config-F.inc"
#include "system_tools/iounit.inc"

!-------------------------------------------------------
! file : fdistribu5d.F90
! date : 16/09/2005
! - 5D distribution function which is assumed
!  - no-periodic in the first direction   (r),
!  - periodic in the second direction     (theta),
!  - periodic in the third direction      (phi),
!  - no-periodic in the fourth direction  (vparallel),
!  - no-periodic in the fifth direction   (mu).
!-------------------------------------------------------
module fdistribu5d_class
  use prec_const
  use MPIutils_module
  use wrap_alloc_module
  use spline1d_class
  use geometry_class
  use init_profile_class
  use local_spline_module
  use fequil4d_module

  use wrap_alloc_module
  use spline1d_class
  implicit none
  public :: new_f5D, del_f5D

  !*** f(r,theta,phi,vpar) parametrized by mu_val ***
  type :: fdistribu5d
    integer     :: muid
    real(RKIND) :: mu_value
    integer     :: n1, n2, n3, n4
    integer     :: istart, istart_buf, iend, iend_buf
    integer     :: jstart, jstart_buf, jend, jend_buf
    integer     :: istart_modif, iend_modif
    integer     :: rstart_modif, rend_modif
    real(RKIND) :: h1, h2, h3, h4

    !*** spline initialization ***
    integer                                  :: BCr_left
    integer                                  :: BCr_right
    integer                                  :: BCvpar_left
    integer                                  :: BCvpar_right
    type(nspline1d)                          :: nspline1d_r
    type(nspline1d), dimension (:), pointer  :: nspline1d_vpar
    type(pspline1d)                          :: pspline1d_theta
    type(pspline1d)                          :: pspline1d_phi
    type(splinehh) , dimension(:), pointer   :: hhspline2d_rtheta
    type(splinehh) , dimension(:), pointer   :: hhspline2d_vparphi
    type(splinehh) , dimension(:), pointer   :: uxspline2d_rtheta
    type(splinehh) , dimension(:), pointer   :: uyspline2d_rtheta

    ! values (4D array parametrized by a value of mu)
    real(RKIND), dimension(:,:,:,:), pointer :: values
  end type fdistribu5d

  !******************************
  contains
  !******************************       

  !------------------------------------------------------------
  ! Check for any NaN value in the distribution function
  !------------------------------------------------------------
  subroutine check_overflow_4D(f,message)
    use globals, only : istart, iend, jstart, jend, &
      pSpecies_id, checkoverflow
#ifdef _OPENMP
    use OMPutils_module, only : omp_get_thread_num
#endif
    use mpiMod    
    type(fdistribu5d), intent(inout) :: f
    character(len=*) , intent(in)    :: message


    integer     :: ir, itheta, iphi, ivpar, ierr
    logical     :: nan_exists, nan_glob
    real(RKIND) :: pinf, minf

    if (checkoverflow) then
      pinf       = huge(1._RKIND)
      pinf       = 2._RKIND * pinf
      minf       = -pinf
      nan_exists = .false.

#ifdef _OPENMP
!$OMP PARALLEL private(ir,itheta,iphi,ivpar) reduction(.or.:nan_exists) &
!$OMP default(shared)
!$OMP BARRIER
#endif

!$OMP DO SCHEDULE(static)
      do ivpar = 0,f%n4
        do iphi = 0,f%n3-1
          do itheta = jstart,jend 
            do ir = f%istart_modif,f%iend_modif
              nan_exists = nan_exists .or. &
                (f%values(ir,itheta,iphi,ivpar) .ne. f%values(ir,itheta,iphi,ivpar))
              nan_exists = nan_exists .or. &
                (f%values(ir,itheta,iphi,ivpar) .eq. pinf)
              nan_exists = nan_exists .or. &
                (f%values(ir,itheta,iphi,ivpar) .eq. minf)
            end do
          end do
        end do
      end do
!$OMP END DO
!$OMP END PARALLEL
      nan_glob = .false.
      call MY_Allreduce(nan_exists,nan_glob,1, &
        MPI_LOGICAL,MPI_LOR,MPI_COMM_WORLD,ierr)
      if (pglobal_id .eq. 0) then
        write(*,'(I4,A30,1X,L1)') pglobal_id, message//"#overflow ", nan_glob
      end if
      if (nan_glob) then
        do ivpar = 0,f%n4
          do iphi = 0,f%n3-1
            do itheta = jstart,jend 
              do ir = f%istart_modif,f%iend_modif
                nan_exists = &
                  (f%values(ir,itheta,iphi,ivpar) .ne. f%values(ir,itheta,iphi,ivpar))
                nan_exists = nan_exists .or. &
                  (f%values(ir,itheta,iphi,ivpar) .eq. pinf)
                nan_exists = nan_exists .or. &
                  (f%values(ir,itheta,iphi,ivpar) .eq. minf)
                if (nan_exists) then
                  write(*,'(I5,1X,4(A,1X,I5,1X),F18.7)') pglobal_id, message//"Pb located at ir=",&
                    ir,"itheta=",itheta,"iphi=",iphi,"ivpar=",ivpar, f%values(ir,itheta,iphi,ivpar)
                end if
              end do
            end do
          end do
        end do
        call ppexit()
      end if
      call ppbarrier()     
    end if
  end subroutine check_overflow_4D


  !------------------------------------------------------------
  ! Check for any NaN value in the distribution function
  !------------------------------------------------------------
  subroutine check_overflow_1D(vect,zstart,zend,message,optvect)
    use globals, only : pglobal_id, checkoverflow
#ifdef _OPENMP
    use OMPutils_module, only : omp_get_thread_num
#endif
    use mpiMod    
    real(RKIND)          , dimension(:), intent(in) :: vect
    integer                            , intent(in) :: zstart, zend
    character(len=*)                   , intent(in) :: message
    real(RKIND), dimension(:), optional, intent(in) :: optvect

    logical     :: nan_exists, nan_glob
    integer     :: count, ierr
    real(RKIND) :: pinf, minf

    if (checkoverflow) then
      pinf       = huge(1._RKIND)
      pinf       = 2._RKIND * pinf
      minf       = -pinf
      nan_exists = .false.
      
      do count = zstart,zend
        nan_exists = nan_exists .or. &
          (vect(count) .ne. vect(count))
        nan_exists = nan_exists .or. &
          (vect(count) .eq. pinf)
        nan_exists = nan_exists .or. &
          (vect(count) .eq. minf)
      end do
      
      nan_glob = .false.
      call MY_Allreduce(nan_exists,nan_glob,1, &
        MPI_LOGICAL,MPI_LOR,MPI_COMM_WORLD,ierr)
      if (pglobal_id .eq. 0 .and. nan_glob) then
        write(*,'(I4,A,1X,L1)') pglobal_id, &
          adjustl(trim(message))//"#overflow ", nan_glob
      end if
      
      if (nan_glob) then
        do count = zstart, zend
          nan_exists = nan_exists .or. &
            (vect(count) .ne. vect(count))
          nan_exists = nan_exists .or. &
            (vect(count) .eq. pinf)
          nan_exists = nan_exists .or. &
            (vect(count) .eq. minf)
          if (nan_exists) then
            if (present(optvect)) then
              write(*,'(I5,1X,1(A,1X,I5,1X),2G18.7)') pglobal_id, &
                adjustl(trim(message))//"Pb located at index=",&
                count, vect(count), optvect(count)
            else
              write(*,'(I5,1X,1(A,1X,I5,1X),G18.7)') pglobal_id, &
                adjustl(trim(message))//"Pb located at index=",&
                count, vect(count)
            end if
            FLUSH_STDOUT
          end if
        end do
        STOP "overflow 1D"
      end if
    end if
  end subroutine check_overflow_1D


  !-------------------------------------------------------
  ! check if the point must be ignore for the computation
  !  in the radial direction
  !------------------------------------------------------- 
  subroutine ignore_r_boundary(ir,bound_r)
    use globals, only : Nr, pSpecies_id, flux_driven           
    integer, intent(in)  :: ir
    logical, intent(out) :: bound_r 
    
    bound_r = .false.
    if (ir.eq.0) then
      if (.not.flux_driven) bound_r = .true.        
    end if
    if (ir.eq.Nr) bound_r = .true.
  end subroutine ignore_r_boundary


  !----------------------------------------------------------
  ! Constructor of the 5D distribution function 
  !   f(r,theta,phi,vpar,mu)=feq(1+delta f)
  !   where feq is the equilibrium part and delta f the 
  !   perturbated part.
  !   This function is parallelized on all the processors.
  !----------------------------------------------------------
  subroutine new_f5d(fthis,geom)
    use globals, only : mu_id, pSpecies_id, &
      istart, istart_buf, iend, iend_buf, &
      jstart, jstart_buf, jend, jend_buf, &
      dom_r, dom_theta, Nbthread, m, n, &
      BC_Hermite, transpose4D , Nr, Ntheta, &
      Nphi, Nvpar
    use globals, only : flux_driven, conservative_scheme
!R3 #include "r3_info.h" !R3
    type(fdistribu5d), intent(out) :: fthis
    type(geometry)   , intent(in)  :: geom

    real(RKIND) :: rmin_loc, thetamin_loc
    integer     :: ithread
    integer     :: min_i1, min_i2, max_i1, max_i2
    integer     :: min_i3, max_i3, min_i4, max_i4

!R3 call r3_info_begin (r3_info_index_1, 'Init_newf5d') !R3
    call Write_dyn_balise_open('Init_newf5d') !R4
    !*** dimension initialization ***
    fthis%n1         = geom%Nr
    fthis%n2         = geom%Ntheta
    fthis%n3         = geom%Nphi
    fthis%n4         = geom%Nvpar   
    fthis%h1         = geom%dr
    fthis%h2         = geom%dtheta
    fthis%h3         = geom%dphi
    fthis%h4         = geom%dvpar
    fthis%istart     = istart
    fthis%istart_buf = istart_buf
    fthis%jstart     = jstart
    fthis%jstart_buf = jstart_buf
    fthis%iend       = iend
    fthis%iend_buf   = iend_buf
    fthis%jend       = jend
    fthis%jend_buf   = jend_buf
    fthis%muid       = mu_id
    fthis%mu_value   = 0
	
    call write_expr("fthis%istart_buf", "istart_buf")
    call write_expr("fthis%jstart_buf", "jstart_buf")
    call write_expr("fthis%iend_buf"  , "iend_buf")
    call write_expr("fthis%jend_buf"  , "jend_buf")
    call write_expr("geom%Nr"         , "Nr")
    call write_expr("geom%Ntheta"     , "Ntheta")
    call write_expr("geom%Nphi"       , "Nphi")
    call write_expr("geom%Nvpar"      , "Nvpar")
    call write_expr("fthis%n1"        , "geom%Nr")
    call write_expr("fthis%n2"        , "geom%Ntheta")    
    call write_expr("fthis%n3"        , "geom%Nphi")
    call write_expr("fthis%n4"        , "geom%Nvpar")

    if (.not.memory_test) then
      fthis%mu_value   = geom%mug(mu_id)
    end if

    !*** boundary conditions initialization ***
    if (flux_driven) then
      fthis%istart_modif = fthis%istart
      fthis%iend_modif   = min(geom%Nr-1,fthis%iend)
      fthis%rstart_modif = 0
      fthis%rend_modif   = geom%Nr-1
      ! Neumann boundary conditions at r = rmin
      fthis%BCr_left     = 1
      fthis%BCr_right    = BC_Hermite
    else
      fthis%istart_modif = max(1,fthis%istart)
      fthis%iend_modif   = min(geom%Nr-1,fthis%iend)
      fthis%rstart_modif = 1
      fthis%rend_modif   = geom%Nr-1
      fthis%BCr_left     = BC_Hermite
      fthis%BCr_right    = BC_Hermite
    end if
    ! Neumann boundary conditions in vpar direction
    fthis%BCvpar_left    = 1
    fthis%BCvpar_right   = 1
	
    !*** array allocation for distribution function values ***
    call FTR_ALLOC_4(fthis%values,fthis%istart_buf,fthis%iend_buf,fthis%jstart_buf,fthis%jend_buf,0,fthis%n3,0,fthis%n4)

    !*** spline initialization ***
    call write_expr("n", "fthis%n1")
    call new_spline1d_natural(fthis%nspline1d_r,fthis%n1,fthis%h1)
    call write_expr("n", "fthis%n2")
    call new_spline1d_period(fthis%pspline1d_theta, &
      fthis%n2,fthis%h2)
    call write_expr("n", "fthis%n3")
    call new_spline1d_period(fthis%pspline1d_phi,fthis%n3,fthis%h3)
    allocate(fthis%nspline1d_vpar(1:Nbthread))

    if (transpose4D) then
       if (conservative_scheme) then
          min_i1 = -1
          max_i1 = (fthis%n1)+2
          min_i2 = -1
          max_i2 = (fthis%n2)+1

          call write_param("min_i1", -1)
          call write_expr("max_i1" , "(fthis%n1)+2")
          call write_param("min_i2", -1)
          call write_expr("max_i2" , "(fthis%n2)+1")

          min_i3 = -1
          max_i3 = (fthis%n3)+2
          min_i4 = -1
          max_i4 = (fthis%n4)+1

          call write_param("min_i3", -1)
          call write_expr("max_i3" , "(fthis%n3)+2")
          call write_param("min_i4", -1)
          call write_expr("max_i4" , "(fthis%n4)+1")
       else
          min_i1 = -2
          max_i1 = (fthis%n1+2)+1
          min_i2 = -2
          max_i2 = (fthis%n2+1)+1

          call write_param("min_i1", -2)
          call write_expr("max_i1" , "(fthis%n1+2)+1")
          call write_param("min_i2", -2)
          call write_expr("max_i2" , "(fthis%n2+1)+1")
       endif
    else
       if (conservative_scheme) then
          min_i1 = -1
          max_i1 = (dom_r+1)+2
          min_i2 = -1
          max_i2 = (dom_theta+1)+1

          call write_param("min_i1", -1)
          call write_expr("max_i1" , "(dom_r+1)+2")
          call write_param("min_i2", -1)
          call write_expr("max_i2" , "(dom_theta+1)+1")
       else
          min_i1 = -2
          max_i1 = (dom_r+1)+1
          min_i2 = -2
          max_i2 = (dom_theta+1)+1

          call write_param("min_i1", -2)
          call write_expr("max_i1" , "(dom_r+1)+1")
          call write_param("min_i2", -2)
          call write_expr("max_i2" , "(dom_theta+1)+1")
       endif
    endif

    allocate(fthis%hhspline2d_rtheta(1:Nbthread))

    if (transpose4D.and.conservative_scheme) then
      allocate(fthis%hhspline2d_vparphi(1:Nbthread))
    else
      allocate(fthis%uxspline2d_rtheta(1:Nbthread))
      allocate(fthis%uyspline2d_rtheta(1:Nbthread))
    endif

    do ithread = 1,Nbthread
      call write_expr("n", "fthis%n4")

      call new_spline1d_natural(fthis%nspline1d_vpar(ithread), &
        fthis%n4,fthis%h4)
      if (transpose4D) then
        rmin_loc     = geom%rmin + (-1)*fthis%h1
        thetamin_loc = geom%thetamin + (-1)*fthis%h2
        if (conservative_scheme) then
          call write_expr("nx", "fthis%n1")
          call new_splinehh(fthis%hhspline2d_rtheta(ithread), &
            fthis%n1,fthis%n2,geom%rmin,geom%thetamin, &
            fthis%h1,fthis%h2)
          call write_expr("nx", "fthis%n4")
          call new_splinehh(fthis%hhspline2d_vparphi(ithread), &
            fthis%n4,fthis%n3,geom%vparmin,geom%phimin, &
            fthis%h4,fthis%h3)
        else
          call write_expr("nx", "fthis%n1+2")
          call write_expr("ny", "fthis%n2+1")
          call new_splinehh(fthis%hhspline2d_rtheta(ithread), &
            fthis%n1+2,fthis%n2+1,rmin_loc,thetamin_loc, &
            fthis%h1,fthis%h2)
          call new_splinehh(fthis%uxspline2d_rtheta(ithread), &
            fthis%n1+2,fthis%n2+1,rmin_loc,thetamin_loc, &
            fthis%h1,fthis%h2)
          call new_splinehh(fthis%uyspline2d_rtheta(ithread), &
            fthis%n1+2,fthis%n2+1,rmin_loc,thetamin_loc, &
            fthis%h1,fthis%h2)
        end if                                           
      else    ! if (transpose4D) then
        rmin_loc     = geom%rmin + (istart-1)*fthis%h1
        thetamin_loc = geom%thetamin + (jstart-1)*fthis%h2
        call write_expr("nx", "dom_r+1")
        call write_expr("ny", "dom_theta+1")
        call new_splinehh(fthis%hhspline2d_rtheta(ithread), &
          dom_r+1,dom_theta+1,rmin_loc,thetamin_loc, &
          fthis%h1,fthis%h2)
        call new_splinehh(fthis%uxspline2d_rtheta(ithread), &
          dom_r+1,dom_theta+1,rmin_loc,thetamin_loc, &
          fthis%h1,fthis%h2)
        call new_splinehh(fthis%uyspline2d_rtheta(ithread), &
          dom_r+1,dom_theta+1,rmin_loc,thetamin_loc, &
          fthis%h1,fthis%h2)
      end if
    enddo

!R3 call r3_info_end(r3_info_index_1) !R3
    call Write_dyn_balise_close('Init_newf5d') !R4
  end subroutine new_f5d

  
  !--------------------------------- 
  ! 5D spline destruction 
  !---------------------------------
  subroutine del_f5d(fthis)
    use globals, only : Nbthread
    use globals, only : conservative_scheme, transpose4D
    type(fdistribu5d), intent(inout) :: fthis

    integer :: ithread

    call del_spline1d_natural(fthis%nspline1d_r)
    call del_spline1d_period(fthis%pspline1d_theta)
    call del_spline1d_period(fthis%pspline1d_phi)
    do ithread = 1,Nbthread
      call del_spline1d_natural(fthis%nspline1d_vpar(ithread))
      call del_splinehh(fthis%hhspline2d_rtheta(ithread))
      if (conservative_scheme) then
        call del_splinehh(fthis%hhspline2d_vparphi(ithread))
      else
        call del_splinehh(fthis%uxspline2d_rtheta(ithread))
        call del_splinehh(fthis%uyspline2d_rtheta(ithread))
      end if                                             
    enddo
    deallocate(fthis%nspline1d_vpar)
    if (transpose4D.and.conservative_scheme) then
      deallocate(fthis%hhspline2d_rtheta)
      deallocate(fthis%hhspline2d_vparphi)
    else
      deallocate(fthis%hhspline2d_rtheta)
      deallocate(fthis%uxspline2d_rtheta)
      deallocate(fthis%uyspline2d_rtheta)
   endif
  end subroutine del_f5d


  !************************************************
  ! AT THE FIRST STEP THE DISTRIBUTION FUNCTION 
  !  IS GIVEN BY THE INITIAL CONDITIONS
  !************************************************
  !------------------------------------------------------------
  ! Initialisation of the perturbed part (delta f) of the 
  !   distribution function
  !   delta_f(theta,phi) = perturb_amplitude * 
  !                        cos(k*phi+m*theta)) or
  !         = sum_{m,n} perturb_amplitude * 
  !           cos(k*phi+m*theta+pert_{m,n})
  !  where pert_{m,n} has random values
  !------------------------------------------------------------
  subroutine init_fperturb(geom,init_prof,Sfperturb_thetaphi,Srandom_phase)
    use globals, only : pSpecies_id, perturb_choice, m, n, &
      perturb_amplitude, Rosenbluth_Hinton
    type(geometry)               , intent(in) :: geom
    type(init_profile)           , intent(in) :: init_prof
    real(RKIND), dimension(:,:)  , pointer    :: Sfperturb_thetaphi
    real(RKIND), dimension(:,:)  , pointer    :: Srandom_phase
    
    real(RKIND) :: toroidal_ratio, fperturb_tmp
    real(RKIND) :: perturb_amplitude_spec
    integer     :: itheta, iphi
    integer     :: m_max, n_max, in, im, in_min, in_max

    !*** f initialisation *** 
    toroidal_ratio         = TWOPI/geom%Lphi
    perturb_amplitude_spec = perturb_amplitude(pSpecies_id)
    select case (perturb_choice)
    case (1)
      ! -> initialization with the mode (m,n)
      do iphi = 0,geom%Nphi
        do itheta = 0,geom%Ntheta      
          Sfperturb_thetaphi(itheta,iphi) = &
            perturb_amplitude_spec* &
            cos(m*geom%thetag(itheta) - &
            n*geom%phig(iphi)*toroidal_ratio)
        end do
      end do
    case (2)
      ! -> initialization with several modes 
      !    (1 < im < m) and (1 < in < n) 
      m_max = m
      n_max = n
      do iphi = 0,geom%Nphi
        do itheta = 0,geom%Ntheta
          fperturb_tmp = 0._RKIND
          do in = 1,n_max
            do im = 1,m_max
              fperturb_tmp = fperturb_tmp + &
                cos(im*geom%thetag(itheta) - &
                in*geom%phig(iphi)*toroidal_ratio + &
                Srandom_phase(im,in))
            end do
          end do
          fperturb_tmp = perturb_amplitude_spec * &
            fperturb_tmp / real(m_max*n_max)
          if ( abs(fperturb_tmp).le.1.e-30_RKIND ) &
            fperturb_tmp = 1.e-30_RKIND
          Sfperturb_thetaphi(itheta,iphi) = fperturb_tmp
        end do
      end do
    case (3)
      ! -> initialization with several m modes 
      !     for one unique n mode
      in    = 1
      m_max = m
      do iphi = 0,geom%Nphi
        do itheta = 0,geom%Ntheta
          fperturb_tmp = 0._RKIND        
          do im = 1,m_max
            fperturb_tmp = fperturb_tmp + &
              cos(im*geom%thetag(itheta) - &
              n*geom%phig(iphi)*toroidal_ratio + &
              Srandom_phase(im,in))
          end do
          fperturb_tmp = perturb_amplitude_spec*fperturb_tmp
          if ( abs(fperturb_tmp).le.1.e-30_RKIND ) &
            fperturb_tmp = 1.e-30_RKIND
          Sfperturb_thetaphi(itheta,iphi) = fperturb_tmp
        end do
      end do
    case (4)
      ! -> initialization with several n modes 
      !     for one unique m mode
      im    = 1
      n_max = n
      do iphi = 0,geom%Nphi
        do itheta = 0,geom%Ntheta
          fperturb_tmp = 0._RKIND        
          do in = 1,n_max
            fperturb_tmp = fperturb_tmp + &
              cos(m*geom%thetag(itheta) - &
              in*geom%phig(iphi)*toroidal_ratio + &
              Srandom_phase(im,in))
          end do
          fperturb_tmp = perturb_amplitude_spec*fperturb_tmp
          if ( abs(fperturb_tmp).le.1.e-30_RKIND ) &
            fperturb_tmp = 1.e-30_RKIND
          Sfperturb_thetaphi(itheta,iphi) = fperturb_tmp
        end do
      end do
    case (5)
      ! -> initialization with m=1,n=0 (GAM symmetry)
      do iphi = 0,geom%Nphi
        do itheta = 0,geom%Ntheta
          fperturb_tmp = sin(geom%thetag(itheta))
          fperturb_tmp = perturb_amplitude_spec*fperturb_tmp
          if ( abs(fperturb_tmp).le.1.e-30_RKIND ) &
            fperturb_tmp = 1.e-30_RKIND
          Sfperturb_thetaphi(itheta,iphi) = fperturb_tmp
        end do
      end do
    case (6)
      ! -> initialization with m=1,n=0 (GAM symmetry) and m=0,n=0 (RH)
      do iphi = 0,geom%Nphi
        do itheta = 0,geom%Ntheta
          fperturb_tmp = sin(geom%thetag(itheta))
          fperturb_tmp = perturb_amplitude_spec*(1._RKIND + fperturb_tmp)
          if ( abs(fperturb_tmp).le.1.e-30_RKIND ) &
            fperturb_tmp = 1.e-30_RKIND
          Sfperturb_thetaphi(itheta,iphi) = fperturb_tmp
        end do
      end do
    case (7)
      ! -> initialization with several modes 
      !    (1 < im < m) and (0 < in < n) 
      m_max = m
      n_max = n
      do iphi = 0,geom%Nphi
        do itheta = 0,geom%Ntheta
          fperturb_tmp = 0._RKIND
          do in = 0,n_max
            do im = 1,m_max
              fperturb_tmp = fperturb_tmp + &
                cos(im*geom%thetag(itheta) - &
                in*geom%phig(iphi)*toroidal_ratio + &
                Srandom_phase(im,in))
            end do
          end do
          fperturb_tmp = perturb_amplitude_spec*fperturb_tmp / &
            real(m_max*n_max)
          if ( abs(fperturb_tmp).le.1.e-30_RKIND ) &
            fperturb_tmp = 1.e-30_RKIND
          Sfperturb_thetaphi(itheta,iphi) = fperturb_tmp
        end do
      end do
    case default
      if (pglobal_id.eq.0) then
        print*, 'perturb_choice = ', &
          perturb_choice, ' must be in between 1 and 8'
        stop
      end if
    end select

  end subroutine init_fperturb


  !------------------------------------------------------------- 
  ! Initialisation of the distribution function
  !  f(r,theta,phi,vpar,mu) = 
  !    feq(r,theta,vpar,mu)*(1+delta f(r,theta,phi,vpar))
  !  where delta f = g(r)*h(vpar)*delta p(theta,phi) with
  !   - g(r) is defined with polynomial functions and
  !   - h(vpar) is defined with a Gaussian shape
  !  such that g and h vanishes at both radial boundaries.
  !   - delta p = perturb_amplitude*cos(k*phi+m*theta)) or
  !             = sum_{m,n} perturb_amplitude * 
  !               cos(k*phi+m*theta+pert_{m,n})
  !  where pert_{m,n} has random values
  !-------------------------------------------------------------
  subroutine f_initialisation(fthis,geom,init_prof,Sfmu_eq)
    use globals, only : mpi_comm_species, &
      pSpecies_id, pInSpecies_id,mu_id, &
      m, n, perturb_amplitude, &
      Rosenbluth_Hinton, hvpar_in_fperturb, &
      deltar_Ts0, profile_choice, uout_res, &
      Rarray1_Nr, Rarray1_Nvpar, &
      outputproc
    use utils_module, only : init_random_seed
    use mpiMod
!R3 #include "r3_info.h" !R3
    type(fdistribu5d)            , intent(inout) :: fthis
    type(geometry)               , intent(in)    :: geom
    type(init_profile)           , intent(in)    :: init_prof
    real(RKIND), dimension(:,:,:), pointer       :: Sfmu_eq

    ! for the initial perturbation
    real(RKIND), dimension(:,:), pointer :: fperturb_thetaphi
    real(RKIND), dimension(:,:), pointer :: random_phase    

    integer     :: i1, i2, i3, i4, ir, ivpar
    integer     :: i, j, m_max, n_max, ierr
    real(RKIND) :: rmin, rmax, vparmin, vparmax
    real(RKIND) :: max_gr_pol, rbuff, lambda, gr_pol_tmp, hvpar_tmp
    real(RKIND) :: Lr_loc, Lvpar_loc
    real(RKIND) :: feq_value, fperturb_tmp
    real(RKIND), dimension(:)  , pointer :: gr_pol, hvpar

    !*** allocation for the initial perturbation ***
    call write_param("n", n)
    call write_param("m", m)
    call FTR_ALLOC_2_ZERO(random_phase,1,m,1,n)
    call FTR_ALLOC_2_ZERO(fperturb_thetaphi,0,fthis%n2,0,fthis%n3)

    if ((.not.Rosenbluth_Hinton).and.((n .eq. 0) .or. (m .eq. 0))) then
      !*** there is normally no perturbation, i.e. perturb_amplitude = 0
      if ((perturb_amplitude(pSpecies_id) .ne. 0._RKIND) .and. &
        (pglobal_id .eq. outputproc)) then
        write (uout_res,*) '---> perturb_amplitude is set to 0 ', &
          'as n or m is equal to 0' 
        write (uout_res,*) ' '
      end if
      !*** computation of the initial values for f ***
!$OMP PARALLEL private(i1,i2,i3,i4,feq_value) default(shared)
!$OMP DO SCHEDULE(static)
    do i4 = 0,fthis%n4
      do i3 = 0,fthis%n3
        do i2 = fthis%jstart,fthis%jend
          do i1 = fthis%istart,fthis%iend
            feq_value  = Sfmu_eq(i1,i2,i4)
            fthis%values(i1,i2,i3,i4) = feq_value
          end do
        end do
      end do
    end do
!$OMP END DO
!$OMP END PARALLEL
      
  else
      gr_pol => Rarray1_Nr
      hvpar  => Rarray1_Nvpar

      !*** computation of the initial perturbation delta f *** 
!R3 call r3_info_begin (r3_info_index_0, 'CPU_f_initialisation') !R3
      call Write_dyn_balise_open('CPU_f_initialisation') !R4
      !-> computation of g(r)
      Lr_loc = geom%Lr
      rmin   = geom%rg(0)
      rmax   = geom%rg(geom%Nr)
      if (Rosenbluth_Hinton) then
        do ir = 0,geom%Nr
          gr_pol(ir) = &
            sin(TWOPI*(geom%rg(ir)-rmin)/(2._RKIND*Lr_loc))
        end do
      else
        if (profile_choice(pSpecies_id).ne.3) then
          max_gr_pol = (0.5_RKIND*Lr_loc)**6 * &
            (rmin-rmax+0.5_RKIND*Lr_loc)**6
          do ir = 0,geom%Nr
            gr_pol(ir) = (geom%rg(ir)-rmin)**6 * &
              (geom%rg(ir)-rmax)**6/max_gr_pol
          end do
        else
          lambda = deltar_Ts0(pSpecies_id)*Lr_loc
          rbuff  = 0.15_RKIND*Lr_loc
          do ir = 0,geom%Nr
            gr_pol(ir) = 0.5_RKIND * &
              (tanh((geom%rg(ir)-rmin-rbuff)/lambda) - &
              tanh((geom%rg(ir)-rmax+rbuff)/lambda))
          end do
        end if
      end if
      
      !-> computation of h(vpar)
      if (hvpar_in_fperturb) then
        vparmin    = geom%vparg(0)
        vparmax    = geom%vparg(geom%Nvpar)
        Lvpar_loc  = geom%Lvpar
        do ivpar = 0,geom%Nvpar
          hvpar(ivpar) = exp(-((geom%vparg(ivpar)-&
            (vparmin+Lvpar_loc/2._RKIND))/(0.1*Lvpar_loc))**2)
        end do
      else
        do ivpar = 0,geom%Nvpar
          hvpar(ivpar) = 1._RKIND
        end do
      end if
      
      !-> computation of the random phase
      m_max = m 
      n_max = n
!SJA All species get the same random phase at this stage in development
      call init_random_seed()
      if (pinspecies_id .eq. 0) then
        do j = 1,n_max
          do i = 1,m_max
            call random_number(random_phase(i,j))
          end do
        end do
      end if
      call MPI_BCAST(random_phase(1,1),n_max*m_max, &
        MPI_REAL8,0,mpi_comm_species,ierr)
      random_phase = TWOPI * random_phase

      !-> computation of the perturbation delta_p(theta,phi)
      call init_fperturb(geom,init_prof,fperturb_thetaphi,random_phase)
      
      !*** computation of the initial values for f ***
!$OMP PARALLEL private(i1,i2,i3,i4,feq_value,gr_pol_tmp,hvpar_tmp, &
!$OMP  fperturb_tmp) default(shared)
!$OMP DO SCHEDULE(static)
      do i4 = 0,fthis%n4
        hvpar_tmp = hvpar(i4)
        do i3 = 0,fthis%n3
          do i2 = fthis%jstart,fthis%jend
            fperturb_tmp = fperturb_thetaphi(i2,i3)
            do i1 = fthis%istart,fthis%iend
              gr_pol_tmp = gr_pol(i1)
              feq_value  = Sfmu_eq(i1,i2,i4)
              fthis%values(i1,i2,i3,i4) = feq_value*(1._RKIND + &
                fperturb_tmp*gr_pol_tmp*hvpar_tmp)
            end do
          end do
        end do
      end do
!$OMP END DO
!$OMP END PARALLEL
      call Write_dyn_balise_close('CPU_f_initialisation') !R4
!R3 call r3_info_end (r3_info_index_0) !R3
    end if
    call FTR_DEALLOC(random_phase)
    call FTR_DEALLOC(fperturb_thetaphi)
  end subroutine f_initialisation
  

  !************************************************
  ! CUBIC SPLINE INTERPOLATION
  !************************************************
  !------------------------------------------------------- 
  ! Compute the 2D cubic spline in the directions r and
  !  theta. These spline coefficients are used for the
  !  shift in (r,theta)
  !-------------------------------------------------------     
  subroutine compute_spline_rtheta(fthis,rhs,scoef2d)
    use globals, only : Rarray1_m1Nrp1m1Nthetap1
    type(fdistribu5d)              , intent(inout) :: fthis
    real(RKIND), dimension(0:,0:)  , intent(in)    :: rhs
    real(RKIND), dimension(-1:,-1:), intent(inout) :: scoef2d

    real(RKIND), dimension(:,:), pointer :: gamma    
    integer                              :: i1, i2

    scoef2d = 0._RKIND

    !*** solving in r direction with Dirichlet conditions ***
    gamma => Rarray1_m1Nrp1m1Nthetap1

    do i2 = 0,fthis%n2
      call natural_spline_coef(fthis%nspline1d_r,rhs(0:,i2),&
        fthis%BCr_left,fthis%BCr_right)
      do i1 = -1,fthis%n1+1
        gamma(i1,i2) = fthis%nspline1d_r%scoef(i1)
      end do
    end do
    
    !*** solving in theta direction with ***
    !***  periodic boundary conditions      ***
    do i1=-1,fthis%n1+1
      call period_spline_coef(fthis%pspline1d_theta,gamma(i1,0:))
      do i2 = -1,fthis%n2+1
        scoef2d(i1,i2) = fthis%pspline1d_theta%scoef(i2)
      end do
    end do
  end subroutine compute_spline_rtheta


  !------------------------------------------------------- 
  ! Compute the 2D cubic spline in the directions r and
  !  theta. These spline coefficients are used for the
  !  shift in (r,theta)
  !-------------------------------------------------------     
  subroutine compute_fslice(geom,init_magnet,init_curr,f,Sfmu_eq)
    use globals
    use Io_Unit_module, only: IU_open, IU_close
    use mpiMod
    
    type(geometry)     , intent(in) :: geom
    type(fdistribu5d)  , intent(in) :: f
    real(RKIND), &     
      dimension(:,:,:) , pointer    :: Sfmu_eq
    type(init_magnetic), intent(in) :: init_magnet
    type(init_current) , intent(in) :: init_curr

    real(RKIND)       :: jacob_vel_tmp
    character(LEN=40) :: file_name_out
    integer           :: ir, itheta, iphi, ivpar
    integer           :: uout = 2 
    integer           :: istep
    integer           :: ierr

    call FTR_ALLOC_4(f4D_transp,0,Nr,0,Ntheta,kstart,kend,lstart,lend)

    select case ( transpose_version )
    case ( TRANSPOSE_SENDIRECV )
       call FTR_ALLOC_4_ZERO(f4D_send,0,dom_r-1,0,dom_theta-1,0,dom_phi-1,0,dom_vpar-1)
       call FTR_ALLOC_4_ZERO(f4D_recv,0,dom_r-1,0,dom_theta-1,0,dom_phi-1,0,dom_vpar-1)
    case ( TRANSPOSE_ALLTOALLW )
       ! nothing specific to allocate
    case default
       print *, pglobal_id, 'Invalid value for transpose_version', &
            transpose_version
       call MPI_abort(MPI_COMM_WORLD, 1, ierr)
    end select
    call pptranspose_forward(f%values,timing=.true.)

    istep = (Nvpar+1)/8
    do ivpar = 2*istep, 6*istep, istep
      if (ivpar .ge. lstart .and. ivpar .le. lend) then
        iphi = 0
        if (iphi .ge. kstart .and. iphi .le. kend) then
          write(file_name_out,'(A3,I3.3,A1,I2.2,A1,I5.5,A4)')"d_",ivpar,&
            "_",mu_id,"_",iter_glob,".out"//char(0)
          call IU_open(uout, file = file_name_out, status = 'UNKNOWN', form = 'FORMATTED')
          do itheta = 0, geom%Ntheta-1
            do ir = 1, geom%Nr-1
              call compute_Bstar(geom,init_magnet, &
                init_curr,ir,itheta,ivpar,jacob_vel_tmp)
              if (iter_glob .eq. 0) then
                write(uout,'(2I5,3E16.8)') ir,itheta, &
                  f4D_transp(ir,itheta,iphi,ivpar), &
                  jacob_vel_tmp,jacobian_space(ir,itheta)
              else
                write(uout,'(2I5,1E16.8)') ir,itheta, &
                  f4D_transp(ir,itheta,iphi,ivpar)
              end if
            end do
            write(uout,'(A)')''
          end do
          call IU_close (uout)
        end if
      end if
    end do
	
    call FTR_DEALLOC(f4D_transp)
    select case ( transpose_version )
    case ( TRANSPOSE_SENDIRECV )
       call FTR_DEALLOC(f4D_send)
       call FTR_DEALLOC(f4D_recv)
    case ( TRANSPOSE_ALLTOALLW )
       ! nothing specific to allocate
    case default
       print *, pglobal_id, 'Invalid value for transpose_version', &
            transpose_version
       call MPI_abort(MPI_COMM_WORLD, 1, ierr)
    end select
  end subroutine compute_fslice
  
end module fdistribu5d_class
