!**************************************************************
!  Copyright Euratom-CEA
!  Authors : 
!     Virginie Grandgirard (virginie.grandgirard@cea.fr)
!     Chantal Passeron (chantal.passeron@cea.fr)
!     Guillaume Latu (guillaume.latu@cea.fr)
!     Xavier Garbet (xavier.garbet@cea.fr)
!     Philippe Ghendrih (philippe.ghendrih@cea.fr)
!     Yanick Sarazin (yanick.sarazin@cea.fr)
!  
!  This code GYSELA (for GYrokinetic SEmi-LAgrangian) 
!  is a 5D gyrokinetic global full-f code for simulating 
!  the plasma turbulence in a tokamak.
!  
!  This software is governed by the CeCILL-B license 
!  under French law and abiding by the rules of distribution 
!  of free software.  You can  use, modify and redistribute 
!  the software under the terms of the CeCILL-B license as 
!  circulated by CEA, CNRS and INRIA at the following URL
!  "http://www.cecill.info". 
!**************************************************************

#include "config-F.inc"

!------------------------------------------------------
! file : velocities_edges.F90
! date : 15/02/2010
! Rk : all the quantities are computed using the
!  cubic spline expression
! ( Module used in the case of the Parabolic Spline
!   Method (PSM) scheme )
!------------------------------------------------------
module velocities_edges_module
  use prec_const
  use geometry_class
  use spline1d_class
  use local_spline_module
  use init_profile_class
  use init_magnetic_class
  use spline1d_class

  implicit none

  !-------------------------------------
  !  PUBLIC VARIABLES
  !------------------------------------- 
  ! used for computation of Phi at the edges in PSM algorithm
  real(RKIND), dimension(:,:,:), pointer, &
    public :: Phiedges_rtheta
  real(RKIND), dimension(:,:,:), pointer, &
    public :: Phiedges_thetaphi
  real(RKIND), dimension(:,:,:), pointer, &
    public :: Phiedges_rphi

  include "gyrokinetic/Bstar_inline.h"
  include "gyrokinetic/velocities_inline.h"
  !******************************
  contains
  !******************************
#include "gyrokinetic/Bstar_inline.F90"
#include "gyrokinetic/velocities_inline.F90"

  !-------------------------------------------------
  ! Constructor 
  !-------------------------------------------------
  subroutine new_velocities_edges(geom)
    use wrap_alloc_module
    type(geometry)     , intent(in) :: geom
#ifdef HAS_CONSERV
    call FTR_ALLOC_3_ZERO(Phiedges_rtheta,0,geom%Nr+1,0,geom%Ntheta+1,0,geom%Nphi+1)
    call FTR_ALLOC_3_ZERO(Phiedges_thetaphi,0,geom%Nr+1,0,geom%Ntheta+1,0,geom%Nphi+1)
    call FTR_ALLOC_3_ZERO(Phiedges_rphi,0,geom%Nr+1,0,geom%Ntheta+1,0,geom%Nphi+1)
#endif 
    ! HAS_CONSERV
  end subroutine new_velocities_edges


  !-------------------------------------------------
  ! Destructor 
  !-------------------------------------------------
  subroutine del_velocities_edges()
    use wrap_alloc_module

#ifdef HAS_CONSERV
    call FTR_DEALLOC(Phiedges_rtheta)
    call FTR_DEALLOC(Phiedges_thetaphi)
    call FTR_DEALLOC(Phiedges_rphi)
#endif 
    ! HAS_CONSERV
  end subroutine del_velocities_edges


  !--------------------------------------------------------
  ! Computation of the values of Phi at 
  !  edges knowing the cell centered values of Phi 
  !  i.e compute the global values of :
  !     . Phiedges_rtheta
  !     . Phiedges_rphi
  !     . Phiedges_thetaphi
  !  (used for the PSM method)
  !--------------------------------------------------------
  subroutine comp_Phi_edges(geom,sPhi)
    use globals, only : Nr, Ntheta, Nphi
#ifdef _OPENMP
    use OMPutils_module, only : omp_get_thread_num
#endif
    use OMPutils_module, only : Romp1_0Nr, Romp2_0Nr, &
      Romp1_0Ntheta, Romp1_0Nphi
    type(geometry)       , intent(in)    :: geom
    real(RKIND), &
      dimension(0:,0:,0:), intent(in)    :: sPhi

#ifdef HAS_CONSERV
    integer :: tid
    integer :: ir, itheta, iphi

#ifdef _OPENMP
!$OMP PARALLEL private(tid,ir,itheta,iphi) default(shared)
!$OMP BARRIER
    tid = 1+omp_get_thread_num()
#else
    tid = 1
#endif

    !*** computation of Phi at edges using  *** 
    !***  a fourth order reconstruction     ***
    !-> Phiedges_rtheta = Phi(nodes_r,nodes_theta,edges_phi)
!$OMP DO SCHEDULE(static)
    do iphi = 0,Nphi
      do ir = 0,Nr+1
        do itheta = 1,Ntheta
          Phiedges_rtheta(ir,itheta,iphi) = 0.25_RKIND * &
            (sPhi(min(ir,Nr),itheta,iphi) + &
            sPhi(max(ir-1,0),itheta,iphi) + &
            sPhi(min(ir,Nr),itheta-1,iphi) + &
            sPhi(max(ir-1,0),itheta-1,iphi))
          Phiedges_rtheta(ir,itheta,iphi) = &
            9._RKIND*Phiedges_rtheta(ir,itheta,iphi) - &
            2._RKIND*0.125_RKIND * &
            (sPhi(min(ir+1,Nr),itheta,iphi) + &
            sPhi(max(ir-2,0),itheta,iphi) + &
            sPhi(min(ir+1,Nr),itheta-1,iphi) + &
            sPhi(max(ir-2,0),itheta-1,iphi) + &
            sPhi(min(ir,Nr),mod(itheta+1,Ntheta),iphi) + &
            sPhi(max(ir-1,0),mod(itheta+1,Ntheta),iphi) + &
            sPhi(min(ir,Nr),mod(itheta-2+Ntheta,Ntheta),iphi) + &
            sPhi(max(ir-1,0),mod(itheta-2+Ntheta,Ntheta),iphi) )
          Phiedges_rtheta(ir,itheta,iphi) = &
            Phiedges_rtheta(ir,itheta,iphi)/7._RKIND
        enddo
        Phiedges_rtheta(ir,0,iphi)        = &
          Phiedges_rtheta(ir,Ntheta,iphi)
        Phiedges_rtheta(ir,Ntheta+1,iphi) = &
          Phiedges_rtheta(ir,1,iphi)
      enddo
    enddo
!$OMP END DO
!$OMP BARRIER

    !-> Phiedges_thetaphi = Phi(edges_r,nodes_theta,nodes_phi)
!$OMP DO SCHEDULE(static)
    do ir = 0,Nr
      do iphi = 1,Nphi
        do itheta = 1,Ntheta
          Phiedges_thetaphi(ir,itheta,iphi) = &
            0.25_RKIND*(sPhi(ir,itheta,iphi) + &
            sPhi(ir,itheta,iphi-1) + &
            sPhi(ir,itheta-1,iphi) + &
            sPhi(ir,itheta-1,iphi-1))
          Phiedges_thetaphi(ir,itheta,iphi) = &
            9._RKIND*Phiedges_thetaphi(ir,itheta,iphi) - &
            2._RKIND*0.125_RKIND * &
            (sPhi(ir,itheta  ,mod(iphi+1,Nphi)) + &
            sPhi(ir,itheta  ,mod(iphi-2+Nphi,Nphi)) + &
            sPhi(ir,itheta-1,mod(iphi+1,Nphi)) + &
            sPhi(ir,itheta-1,mod(iphi-2+Nphi,Nphi)) + &
            sPhi(ir,mod(itheta+1,Ntheta),iphi) + &
            sPhi(ir,mod(itheta+1,Ntheta),iphi-1) + &
            sPhi(ir,mod(itheta-2+Ntheta,Ntheta),iphi) + &
            sPhi(ir,mod(itheta-2+Ntheta,Ntheta),iphi-1) )
          Phiedges_thetaphi(ir,itheta,iphi) = &
            Phiedges_thetaphi(ir,itheta,iphi)/7._RKIND
        enddo
        !---> periodicity in theta
        Phiedges_thetaphi(ir,0,iphi)        = &
          Phiedges_thetaphi(ir,Ntheta,iphi)
        Phiedges_thetaphi(ir,Ntheta+1,iphi) = &
          Phiedges_thetaphi(ir,1,iphi)
      enddo
      !---> periodicity in phi
      Phiedges_thetaphi(ir,:,0)      = &
        Phiedges_thetaphi(ir,:,Nphi)
      Phiedges_thetaphi(ir,:,Nphi+1) = &
        Phiedges_thetaphi(ir,:,1)
    enddo
!$OMP END DO
!$OMP BARRIER

    !-> Phiedges_rphi = Phi(nodes_r,edges_theta,nodes_phi)
!$OMP DO SCHEDULE(static)
    do iphi = 1,Nphi
      do itheta = 0,Ntheta
        do ir = 0,Nr+1
          Phiedges_rphi(ir,itheta,iphi) = &
            0.25_RKIND*(sPhi(min(ir,Nr),itheta,iphi) + &
            sPhi(max(ir-1,0),itheta,iphi) + &
            sPhi(min(ir,Nr) ,itheta,iphi-1) + &
            sPhi(max(ir-1,0),itheta,iphi-1))
          Phiedges_rphi(ir,itheta,iphi) = &
            9._RKIND*Phiedges_rphi(ir,itheta,iphi) - &
            2._RKIND*0.125_RKIND * &
            (sPhi(min(ir+1,Nr),itheta,iphi) + &
            sPhi(max(ir-2,0)  ,itheta,iphi) + &
            sPhi(min(ir+1,Nr) ,itheta,iphi-1) + &
            sPhi(max(ir-2,0)  ,itheta,iphi-1) + &
            sPhi(min(ir,Nr)   ,itheta,mod(iphi+1,Nphi)) + &
            sPhi(max(ir-1,0)  ,itheta,mod(iphi+1,Nphi)) + &
            sPhi(min(ir,Nr)   ,itheta,mod(iphi-2+Nphi,Nphi)) + &
            sPhi(max(ir-1,0)  ,itheta,mod(iphi-2+Nphi,Nphi)))
          Phiedges_rphi(ir,itheta,iphi) = &
            Phiedges_rphi(ir,itheta,iphi)/7._RKIND
        enddo
      enddo
    enddo
!$OMP END DO
!$OMP BARRIER
!$OMP END PARALLEL
    Phiedges_rphi(:,:,0)      = &
      Phiedges_rphi(:,:,Nphi)
    Phiedges_rphi(:,:,Nphi+1) = &
      Phiedges_rphi(:,:,1)
#endif 
    ! HAS_CONSERV
  end subroutine comp_Phi_edges


  !--------------------------------------------------------------
  !  Computation of vExB_gradr by using the expression:
  !     vExB_gradr = 1/Bstar * [Phi,r]
  !  where the Poisson bracket [Phi,r] corresponds to
  !    [Phi,r] = 1/(Js B)*( Btheta*dPhi/dphi
  !                 - Bphi*dPhi/dtheta )
  !  where Js is the jacobian in space
  ! 
  !  Rk : This subroutine can be called for computing the
  !   ExB drift of the guiding centers or of the particles :
  !    -> Phi (for guiding center) in the advections
  !    -> Phi (for particles) in the computation of the fluxes
  !  (used for PSM scheme)
  !--------------------------------------------------------------
  subroutine comp_vExBgradr_edges(geom,init_magnet,init_curr, &
    sPhiedges_rtheta,sPhiedges_rphi, &
    begin_dim1,end_dim1,begin_dim2,end_dim2, &
    begin_dim3,end_dim3,begin_dim4,end_dim4,SvExB_gradr)
    type(geometry)       , intent(in)    :: geom
    type(init_magnetic)  , intent(in)    :: init_magnet
    type(init_current)   , intent(in)    :: init_curr
    real(RKIND), &
      dimension(0:,0:,0:), intent(in)    :: sPhiedges_rtheta
    real(RKIND), &
      dimension(0:,0:,0:), intent(in)    :: sPhiedges_rphi
    integer              , intent(in)    :: begin_dim1
    integer              , intent(in)    :: end_dim1
    integer              , intent(in)    :: begin_dim2
    integer              , intent(in)    :: end_dim2
    integer              , intent(in)    :: begin_dim3
    integer              , intent(in)    :: end_dim3
    integer              , intent(in)    :: begin_dim4
    integer              , intent(in)    :: end_dim4
    real(RKIND), &
      dimension(0:)      , intent(inout) :: SvExB_gradr

#ifdef HAS_CONSERV
    integer     :: ir, itheta, iphi, ivpar
    integer     :: ileft, iright
    integer     :: icount
    real(RKIND) :: B_edge
    real(RKIND) :: Bstar_edge, Bstar_left, Bstar_right
    real(RKIND) :: J_edge
    real(RKIND) :: dPhidtheta_gradr, dPhidphi_gradr
    real(RKIND) :: Btheta_edge, Bphi_edge
    real(RKIND) :: PoissBrack_Phi_r

    !*** computation of vExB_gradr as:                ***
    !***   vExB_gradr = 1/Bstar * [Phi,r]             ***
    !*** with [Phi,r] = 1/(Js B)*( Btheta*dPhi/dphi   ***
    !***                - Bphi*dPhi/dtheta )          ***
    icount = 0
    do ivpar = begin_dim4,end_dim4
      do iphi = begin_dim3,end_dim3
        do itheta = begin_dim2,end_dim2
          do ir = begin_dim1,end_dim1
            ileft  = max(ir-1,0)
            iright = min(ir,Nr) 
            B_edge = 0.5_RKIND*( &
              init_magnet%B_norm(ileft,itheta) + &
              init_magnet%B_norm(iright,itheta) )
#ifdef NO_PRECOMP_BSTAR 
            call compute_Bstar(geom,init_magnet,init_curr, &
              ileft,itheta,ivpar,Bstar_left)
            call compute_Bstar(geom,init_magnet,init_curr, &
              iright,itheta,ivpar,Bstar_right)
#else
            call precomp_Bstar(ileft,itheta,ivpar, &
              Bstar_left)
            call precomp_Bstar(iright,itheta,ivpar, &
              Bstar_right)
#endif
            Bstar_edge = 0.5_RKIND*(Bstar_left+Bstar_right)
            !-> derivatives of Phi
            dPhidtheta_gradr = &
              (sPhiedges_rtheta(ir,itheta+1,iphi) - &
              sPhiedges_rtheta(ir,itheta,iphi)) / &
              geom%dtheta 
            dPhidphi_gradr   = &
              (sPhiedges_rphi(ir,itheta,iphi+1) - &
              sPhiedges_rphi(ir,itheta,iphi)) / &
              geom%dphi             
            !-> Btheta and Bphi
            Btheta_edge = 0.5_RKIND * & 
              ( init_magnet%Btheta(ileft,itheta) + &
              init_magnet%Btheta(iright,itheta))
            Bphi_edge   = 0.5_RKIND * &
              ( init_magnet%Bphi(ileft,itheta) + &
              init_magnet%Bphi(iright,itheta) )
            !-> [Phi,r] = 1/(Js B) * 
            !     ( Btheta*dPhi/dphi- Bphi*dPhi/dtheta ) 
            J_edge           = 0.5_RKIND * &
              ( jacobian_space(ileft,itheta) + &
              jacobian_space(iright,itheta) )
            PoissBrack_Phi_r = 1._RKIND/(J_edge*B_edge) * &
              (Btheta_edge*dPhidphi_gradr - &
              Bphi_edge*dPhidtheta_gradr)
            !-> vEXB.gradr = 1/Bstar * [Phi,r]
            SvExB_gradr(icount) = PoissBrack_Phi_r/Bstar_edge
            icount              = icount+1
          end do
        enddo
      end do
    end do
#endif 
    ! HAS_CONSERV
  end subroutine comp_vExBgradr_edges


  !-----------------------------------------------------------
  !  Computation of vExB_gradtheta
  !     vExB_gradtheta = 1/Bstar * [Phi,theta]
  !  where the Poisson bracket [Phi,theta] corresponds to
  !    [Phi,theta] = 1/(Js B)*( -Br*dPhi/dphi
  !                  + Bphi*dPhi/dr )
  !  where Js is the jacobian in space
  !
  !  Rk : This subroutine can be called for computing the
  !   ExB drift of the guiding centers or of the particles :
  !    -> Phi (for guiding center) in the advections
  !    -> Phi (for particles) in the computation of the fluxes
  !  (used for PSM scheme)
  !-----------------------------------------------------------
  subroutine comp_vExBgradtheta_edges(geom, &
    init_magnet,init_curr, &
    sPhiedges_rtheta,sPhiedges_thetaphi,begin_dim1,end_dim1, &
    begin_dim2,end_dim2,begin_dim3,end_dim3, &
    begin_dim4,end_dim4,SvExB_gradtheta)
    type(geometry)       , intent(in)    :: geom
    type(init_magnetic)  , intent(in)    :: init_magnet
    type(init_current)   , intent(in)    :: init_curr
    real(RKIND), &
      dimension(0:,0:,0:), intent(in)    :: sPhiedges_rtheta
    real(RKIND), &
      dimension(0:,0:,0:), intent(in)    :: sPhiedges_thetaphi
    integer              , intent(in)    :: begin_dim1
    integer              , intent(in)    :: end_dim1
    integer              , intent(in)    :: begin_dim2
    integer              , intent(in)    :: end_dim2
    integer              , intent(in)    :: begin_dim3
    integer              , intent(in)    :: end_dim3
    integer              , intent(in)    :: begin_dim4
    integer              , intent(in)    :: end_dim4
    real(RKIND), &
      dimension(0:)      , intent(inout) :: SvExB_gradtheta

#ifdef HAS_CONSERV
    integer     :: ir, itheta, iphi, ivpar
    integer     :: ileft, iright
    integer     :: icount
    real(RKIND) :: B_edge
    real(RKIND) :: Bstar_edge, Bstar_left, Bstar_right
    real(RKIND) :: J_edge
    real(RKIND) :: dPhidr_gradtheta, dPhidphi_gradtheta
    real(RKIND) :: Br_edge, Bphi_edge
    real(RKIND) :: PoissBrack_Phi_theta

    !*** computation of vExB_gradtheta as:         ***
    !***   vExB_gradtheta = 1/Bstar * [Phi,theta]  ***
    !*** with                                      ***    
    !***   [Phi,r] = 1/(Js B)*( -Br*dPhi/dphi      ***
    !***             + Bphi*dPhi/dr )              ***
    icount = 0
    do ivpar = begin_dim4,end_dim4
      do iphi = begin_dim3,end_dim3
        do itheta = begin_dim2,end_dim2
          ileft = itheta-1
          if (ileft<0) ileft = Ntheta-1
          iright = itheta
          do ir = begin_dim1,end_dim1
            B_edge = 0.5_RKIND*( &
              init_magnet%B_norm(ir,ileft) + &
              init_magnet%B_norm(ir,iright) )
#ifdef NO_PRECOMP_BSTAR 
              call compute_Bstar(geom,init_magnet,init_curr, &
                ir,ileft,ivpar,Bstar_left)
              call compute_Bstar(geom,init_magnet,init_curr, &
                ir,iright,ivpar,Bstar_right)
#else
              call precomp_Bstar(ir,ileft,ivpar,Bstar_left)
              call precomp_Bstar(ir,iright,ivpar,Bstar_right)
#endif
            Bstar_edge = 0.5_RKIND*(Bstar_left+Bstar_right)
            !-> derivatives of Phi
            dPhidr_gradtheta   = &
              (sPhiedges_rtheta(ir+1,itheta,iphi) - &
              sPhiedges_rtheta(ir,itheta,iphi))/geom%dr
            dPhidphi_gradtheta = &
              (sPhiedges_thetaphi(ir,itheta,iphi+1) - &
              sPhiedges_thetaphi(ir,itheta,iphi)) / &
              geom%dphi
            !-> Br and Bphi
            Br_edge   = 0.5_RKIND * &
              ( init_magnet%Br(ir,ileft) + &
              init_magnet%Br(ir,iright) )
            Bphi_edge = 0.5_RKIND * &
              ( init_magnet%Bphi(ir,ileft) + &
              init_magnet%Bphi(ir,iright) )
            !-> [Phi,theta] = 1/(Js B) *
            !      ( -Br*dPhi/dphi+ Bphi*dPhi/dr )
            J_edge           = 0.5_RKIND * &
              ( jacobian_space(ir,ileft) + &
              jacobian_space(ir,iright) )
            PoissBrack_Phi_theta = &
              1._RKIND/(J_edge*B_edge) * &
              (-Br_edge*dPhidphi_gradtheta + &
              Bphi_edge*dPhidr_gradtheta)
            !-> vEXB.gradtheta = 1/Bstar * [Phi,theta]
            SvExB_gradtheta(icount) = PoissBrack_Phi_theta / &
              Bstar_edge
            icount                  = icount+1
          end do
        enddo
      end do
    end do
#endif 
    ! HAS_CONSERV
  end subroutine comp_vExBgradtheta_edges


  !--------------------------------------------------------
  !  Computation of vExB_gradphi
  !     vExB_gradphi = 1/Bstar * [Phi,phi]
  !  where the Poisson bracket [Phi,phi] corresponds to
  !    [Phi,phi] = 1/(Js B)*( Br*dPhi/dtheta
  !                  - Btheta*dPhi/dr )
  !  where Js is the jacobian in space
  !
  !  Rk : This subroutine can be called for computing the
  !   ExB drift of the guiding centers or of the particles :
  !    -> Phi (for guiding center) in the advections
  !    -> Phi (for particles) in the computation of the fluxes
  !  (used for PSM scheme)
  !--------------------------------------------------------
  subroutine comp_vExBgradphi_edges(geom, &
    init_magnet,init_curr, &
    sPhiedges_rphi,sPhiedges_thetaphi,begin_dim1,end_dim1, &
    begin_dim2,end_dim2,begin_dim3,end_dim3, &
    begin_dim4,end_dim4,SvExB_gradphi)
    type(geometry)       , intent(in)    :: geom
    type(init_magnetic)  , intent(in)    :: init_magnet
    type(init_current)   , intent(in)    :: init_curr
    real(RKIND), &
      dimension(0:,0:,0:), intent(in)    :: sPhiedges_rphi
    real(RKIND), &
      dimension(0:,0:,0:), intent(in)    :: sPhiedges_thetaphi
    integer              , intent(in)    :: begin_dim1
    integer              , intent(in)    :: end_dim1
    integer              , intent(in)    :: begin_dim2
    integer              , intent(in)    :: end_dim2
    integer              , intent(in)    :: begin_dim3
    integer              , intent(in)    :: end_dim3
    integer              , intent(in)    :: begin_dim4
    integer              , intent(in)    :: end_dim4
    real(RKIND), &
      dimension(0:)      , intent(inout) :: SvExB_gradphi

#ifdef HAS_CONSERV
    integer     :: ir, itheta, iphi, ivpar
    integer     :: icount
    real(RKIND) :: B_edge
    real(RKIND) :: Bstar_edge
    real(RKIND) :: J_edge
    real(RKIND) :: dPhidr_gradphi, dPhidtheta_gradphi
    real(RKIND) :: Br_edge, Btheta_edge
    real(RKIND) :: PoissBrack_Phi_phi

    !*** computation of vExB_gradphi as:              ***
    !***   vExB_gradphi = 1/Bstar * [Phi,phi]         ***
    !*** with [Phi,phi] = 1/(Js B)*( Br*dPhi/dtheta   ***
    !***                   - Btheta*dPhi/dr )         ***
    icount = 0
    do ivpar = begin_dim4,end_dim4
      do iphi = begin_dim3,end_dim3
        do itheta = begin_dim2,end_dim2
          do ir = begin_dim1,end_dim1
            B_edge = init_magnet%B_norm(ir,itheta)
#ifdef NO_PRECOMP_BSTAR 
              call compute_Bstar(geom,init_magnet,init_curr, &
                ir,itheta,ivpar,Bstar_edge)
#else
              call precomp_Bstar(ir,itheta,ivpar,Bstar_edge)
#endif
            !-> derivatives of Phi
            dPhidr_gradphi     = &
              (sPhiedges_rphi(ir+1,itheta,iphi) - &
              sPhiedges_rphi(ir,itheta,iphi))/geom%dr
            dPhidtheta_gradphi = &
              (sPhiedges_thetaphi(ir,itheta+1,iphi) - &
              sPhiedges_thetaphi(ir,itheta,iphi)) / &
              geom%dtheta
            !-> Br and Btheta
            Br_edge     = init_magnet%Br(ir,itheta)
            Btheta_edge = init_magnet%Btheta(ir,itheta)
            !-> [Phi,phi] = 1/(Js B)*( Br*dPhi/dtheta
            !                - Btheta*dPhi/dr 
            J_edge             = jacobian_space(ir,itheta)
            PoissBrack_Phi_phi = &
              1._RKIND/(J_edge*B_edge) * &
              (Br_edge*dPhidtheta_gradphi - &
              Btheta_edge*dPhidr_gradphi)
            !-> vEXB.gradphi = 1/Bstar * [Phi,phi]
            SvExB_gradphi(icount) = PoissBrack_Phi_phi/Bstar_edge
            icount                = icount+1
          end do
        enddo
      end do
    end do
#endif 
    ! HAS_CONSERV
  end subroutine comp_vExBgradphi_edges



  !*******************************************************
  !* COMMON UTILITIES
  !*******************************************************
  !--------------------------------------------------------
  ! Compute Hermite spline coefficients
  !   rhs(:,1:ntx) contains values of function to be 
  !   interpolated on the full patch and rhs(:0) and 
  !   rhs(:,nx+1) contains the values of the
  !   derivatives of the function at points x_0 and 
  !   x_N respectively.
  ! The transposed function is being used for better 
  !   data locality
  !--------------------------------------------------------
  subroutine hermite2_1D_r(hsthis,rhs,is,n)
    use globals, only : bufsize
    type(splinehh)              , intent(inout) :: hsthis
    real(RKIND), dimension(is:n), intent(inout) :: rhs
    integer                     , intent(in)    :: is 
    integer                     , intent(in)    :: n

#ifdef HAS_CONSERV
    !*** local variables ***
    real(RKIND), dimension(is:n) :: p
    integer     :: i, j, k 
    real(RKIND) :: a,b,c

        
    !*** Second derivative ***
    a  =  6._RKIND/hsthis%dx**2
    b  = -2._RKIND*a
    c  =  a

    j = 1
    !-> Sweep down 
    hsthis%aux(is,j) = rhs(is)
    do i = is+1,n-1
      hsthis%aux(i,j) = rhs(i) - &
        hsthis%lmatx(i-1) * hsthis%aux(i-1,j)
    end do
    hsthis%aux(n,j) = rhs(n) - &
      hsthis%lmatx(n-1) * hsthis%aux(n-1,j) - &
      hsthis%lmatx(n)   * hsthis%aux(n-2,j)
    
    !-> Sweep up 
    hsthis%bcoef(n,j) = hsthis%aux(n,j)/hsthis%umatx(n)
    do i = n-1,is+2,-1
      hsthis%bcoef(i,j) = (hsthis%aux(i,j) - &
        hsthis%bcoef(i+1,j))/hsthis%umatx(i)
    end do
    hsthis%bcoef(is+1,j) = (hsthis%aux(is+1,j) - &
      (1._RKIND-c/a)*hsthis%bcoef(is+2,j))/hsthis%umatx(is+1)
    hsthis%bcoef(is,j)   = (hsthis%aux(is,j) - &
      b*hsthis%bcoef(is+1,j)-c*hsthis%bcoef(is+2,j)) / & 
      hsthis%umatx(is)

!baoter ??
    ! To be in agreement with interpol1d_r, to be suppressed
    hsthis%bcoef(:,j) = hsthis%bcoef(:,j)*6._RKIND
!eaoter
#endif 
    ! HAS_CONSERV
  end subroutine hermite2_1D_r


  !-----------------------------------------------------------
  ! Compute Hermite spline coefficients
  !   rhs(:,1:ntx) contains values of function to be 
  !   interpolated on the full patch and rhs(:0) and 
  !   rhs(:,nx+1) contains the values of the
  !   derivatives of the function at points x_0 and 
  !   x_N respectively.
  ! The transposed function is being used for better 
  !   data locality
  ! -> Hermite splines in the radial direction
  !-----------------------------------------------------------
  subroutine hermite1D_r(hsthis,rhs,is,n)
    use globals, only : bufsize
    type(splinehh)             , intent(inout) :: hsthis
    real(RKIND), dimension(is:), intent(inout) :: rhs
    integer                    , intent(in)    :: is
    integer                    , intent(in)    :: n

#ifdef HAS_CONSERV
    !*** local variables ***
    integer     :: i, j, k 
    real(RKIND) :: threeOverh, hOverThree, twoh
    
    !**************************************
    ! Rk: The size of the arrays are
    !  - System  (is:n)
    !  - lmatx   (is+2:n)
    !**************************************
    
    !*** initializations for the x direction ***
    threeOverh = 3._RKIND/hsthis%dx
    hOverThree = hsthis%dx/3._RKIND
    twoh       = 2._RKIND*hsthis%dx

    j = 1
    !-> Sweep down 
    hsthis%aux(is,j)   = rhs(is)
    hsthis%aux(is+1,j) = rhs(is+1) + &
      hOverThree * hsthis%aux(is,j)
    do i = is+2,n-1
      hsthis%aux(i,j) = rhs(i) - &
        hsthis%lmatx(i) * hsthis%aux(i-1,j)
    end do
    hsthis%aux(n,j) = rhs(n) + &
      threeOverh * hsthis%lmatx(n-1) * hsthis%aux(n-2,j) - &
      threeOverh * hsthis%lmatx(n)   * hsthis%aux(n-1,j)
    
    !-> Sweep up 
    hsthis%bcoef(n,j)   = twoh*hsthis%aux(n,j)/hsthis%umatx(3)
    hsthis%bcoef(n-1,j) = (6._RKIND*hsthis%aux(n-1,j) - &
      hsthis%bcoef(n,j))/hsthis%umatx(2)
    do i = n-2,is+2,-1
      hsthis%bcoef(i,j) = (6._RKIND*hsthis%aux(i,j) - &
        hsthis%bcoef(i+1,j))*hsthis%lmatx(i+1)
    end do
    hsthis%bcoef(is+1,j) = (6._RKIND*hsthis%aux(is+1,j) - &
      2._RKIND*hsthis%bcoef(is+2,j))*hsthis%lmatx(is+2)
    hsthis%bcoef(is,j)   = hsthis%bcoef(is+2,j) - &
      twoh*hsthis%aux(is,j)
#endif 
    ! HAS_CONSERV
  end subroutine hermite1D_r


  !-----------------------------------------------------------
  ! Compute Hermite spline coefficients
  !   rhs(:,1:ntx) contains values of function to be 
  !   interpolated on the full patch and rhs(:0) and 
  !   rhs(:,nx+1) contains the values of the
  !   derivatives of the function at points x_0 and 
  !   x_N respectively.
  ! The transposed function is being used for better 
  !   data locality
  ! -> Hermite splines in the poloidal direction
  !-----------------------------------------------------------
  subroutine hermite1D_theta(hsthis,rhs)
    use globals, only : bufsize
    type(splinehh)  , intent(inout) :: hsthis
    real(RKIND), &
      dimension(-1:), intent(inout) :: rhs

#ifdef HAS_CONSERV
    !*** local variables ***
    integer     :: i, j, k 
    integer     :: ny
    real(RKIND) :: threeOverh, hOverThree, twoh
    
   !*** initializations for the y direction ***
    ny         = hsthis%ny
    threeOverh = 3._RKIND/hsthis%dy
    hOverThree = hsthis%dy/3._RKIND
    twoh       = 2._RKIND*hsthis%dy

    i = 1
    !-> Sweep down 
    hsthis%auy(i,-1) = rhs(-1)
    hsthis%auy(i,0)  = rhs(0) + hOverThree * hsthis%auy(i,-1)
    do j = 1,ny
      hsthis%auy(i,j) = rhs(j) - hsthis%lmaty(j) * hsthis%auy(i,j-1)
    end do
    hsthis%auy(i,ny+1) = rhs(ny+1) + &
      threeOverh * hsthis%lmaty(ny)   * hsthis%auy(i,ny-1) - &
      threeOverh * hsthis%lmaty(ny+1) * hsthis%auy(i,ny)
    
    !-> Sweep up 
    hsthis%bcoef(i,ny+1) = twoh*hsthis%auy(i,ny+1)/hsthis%umaty(3)
    hsthis%bcoef(i,ny)   = (6._RKIND*hsthis%auy(i,ny) - &
      hsthis%bcoef(i,ny+1))/hsthis%umaty(2)
    do j = ny-1,1,-1
      hsthis%bcoef(i,j) = (6._RKIND*hsthis%auy(i,j) - &
        hsthis%bcoef(i,j+1))*hsthis%lmaty(j+1)
    end do
    hsthis%bcoef(i,0)  = (6._RKIND*hsthis%auy(i,0) - &
      2._RKIND*hsthis%bcoef(i,1))*hsthis%lmaty(1)
    hsthis%bcoef(i,-1) = hsthis%bcoef(i,1) - twoh*hsthis%auy(i,-1)
#endif 
    ! HAS_CONSERV
  end subroutine hermite1D_theta


  !--------------------------------------------- 
  !  function interpolation by using the 
  !  1D-spline coefficients calculated 
  !  in r direction
  !   - coef1d contains the 1D cubic splines 
  !  in r direction
  ! (f must be divided according to (r,theta) direction ???)
  !---------------------------------------------
  subroutine interpol1d_r(n1,h1,coef1d,x1,xstar1,finterpol)
    integer                    , intent(in)  :: n1
    real(RKIND)                , intent(in)  :: h1
    real(RKIND), dimension(:,:), pointer     :: coef1d
    real(RKIND), dimension(:)  , pointer     :: x1
    real(RKIND)                , intent(in)  :: xstar1
    real(RKIND)                , intent(out) :: finterpol
    
#ifdef HAS_CONSERV
    integer :: j
    integer :: ipos1
    integer :: l
    real(RKIND), dimension(-1:2) :: sbase1
    character(LEN=50), parameter :: subp = " interpol1d_r "//char(0)
    
    !*** local variable initialisation ***
    j = 1

    !*** array position location ***
    call locate(xstar1,x1,n1,h1,subp,ipos1)

    !*** calculation of cubic spline basis ***
    call spline_basis(x1(ipos1),xstar1,x1(ipos1+1),h1,sbase1)

    !*** computation of f(i,j,k,x1*) ***
    finterpol = 0._RKIND
    finterpol = finterpol + (coef1d(ipos1-1,j)*sbase1(-1) + &
      coef1d(ipos1,j)*sbase1(0) +  coef1d(ipos1+1,j)*sbase1(1)  + &
      coef1d(ipos1+2,j)*sbase1(2))*.166666666666666667_RKIND
#else  
    ! HAS_CONSERV
    finterpol = 0._RKIND
#endif 
    ! HAS_CONSERV
  end subroutine interpol1d_r


  !--------------------------------------------- 
  !  function interpolation by using the 
  !  1D-spline coefficients calculated 
  !  in theta direction
  !   - coef1d contains the 1D cubic splines 
  !  in theta direction
  ! (f must be divided according to (r,theta) direction)
  !---------------------------------------------
  subroutine interpol1d_theta(n2,h2,coef1d,x2,xstar2,finterpol)
    integer                    , intent(in)  :: n2
    real(RKIND)                , intent(in)  :: h2
    real(RKIND), dimension(:,:), pointer     :: coef1d
    real(RKIND), dimension(:)  , pointer     :: x2
    real(RKIND)                , intent(in)  :: xstar2
    real(RKIND)                , intent(out) :: finterpol
    
#ifdef HAS_CONSERV
    integer :: i, ipos2, k
    real(RKIND), dimension(-1:2) :: sbase2
    character(LEN=50), parameter :: subp = " interpol1d_theta "//char(0)

    i = 1
    !*** array position location ***
    call locate(xstar2,x2,n2,h2,subp,ipos2)

    !*** calculation of cubic spline basis ***
    call spline_basis(x2(ipos2),xstar2,x2(ipos2+1),h2,sbase2)

    !*** computation of f(i,x2*,phi,l) ***
    finterpol = 0._RKIND
    finterpol = finterpol + (coef1d(i,ipos2-1)*sbase2(-1) + &
      coef1d(i,ipos2)*sbase2(0) +  coef1d(i,ipos2+1)*sbase2(1) + &
      coef1d(i,ipos2+2)*sbase2(2))*.166666666666666667_RKIND
#else  
    ! HAS_CONSERV
    finterpol = 0._RKIND
#endif  
    ! HAS_CONSERV
  end subroutine interpol1d_theta

end module velocities_edges_module
