!**************************************************************
!  Copyright Euratom-CEA
!  Authors : 
!     Virginie Grandgirard (virginie.grandgirard@cea.fr)
!     Chantal Passeron (chantal.passeron@cea.fr)
!     Guillaume Latu (guillaume.latu@cea.fr)
!     Xavier Garbet (xavier.garbet@cea.fr)
!     Philippe Ghendrih (philippe.ghendrih@cea.fr)
!     Yanick Sarazin (yanick.sarazin@cea.fr)
!  
!  This code GYSELA (for GYrokinetic SEmi-LAgrangian) 
!  is a 5D gyrokinetic global full-f code for simulating 
!  the plasma turbulence in a tokamak.
!  
!  This software is governed by the CeCILL-B license 
!  under French law and abiding by the rules of distribution 
!  of free software.  You can  use, modify and redistribute 
!  the software under the terms of the CeCILL-B license as 
!  circulated by CEA, CNRS and INRIA at the following URL
!  "http://www.cecill.info". 
!**************************************************************

#include "config-F.inc"
#include "code_instrumentation/debug.inc"

!-------------------------------------------------------
! file : gys_alloc.F90
! date : 24/11/2010
!  - allocation of all the arrays required for a 
!  GYSELA run
!  - deallocation of all the arrays required for a 
!  GYSELA run
!  - all initialisations required
!  - all deleting at the end of a GYSELA run
!-------------------------------------------------------
module gys_alloc_module
    use globals
    use MPIutils_module        , only : ppinit_comm, &    
      ppexit, ppdeallocate
    use OMPutils_module        , only : init_OMP, &
      del_OMP
    use read_write_module      , only : testif_restart, &
      Write_parameters
    use geometry_class         , only : geometry, &
      del_geometry
    use coord_system_class     , only : coord_system, &
      init_coordinate_system, del_coord_system
    use init_profile_class     , only : init_profile, &
      del_init_profile
    use init_magnetic_class    , only : init_magnetic, &
      init_magnetic_config, init_precompute_Bstar, &
      del_magnet_config
    use init_current_class     , only : init_current, &
      init_plasma_current, del_init_current
    use sources_module          , only : &              
      init_sources, new_init_sources, &                 
      init_source_properties, del_init_sources          
    use fdistribu5d_class      , only : fdistribu5d, &
      new_f5D, del_f5D
    use gyroaverage_class      , only : J0operator, &
      new_J0operator, del_J0operator
    use poisson_class          , only : poisson, &
      new_poisson, del_poisson
    use QNsolver_module        , only : new_QNsolver, del_QNsolver 
    use vlasov_solving_module  , only : &
      init_mesh_profile
    use velocities_edges_module, only : &                
      new_velocities_edges, del_velocities_edges         
    use physics_GC_module      , only : new_physics_GC, &
      delete_physics_GC
    use physics_part_module    , only : new_physics_part, &
      delete_physics_part
    use efield_module          , only : new_efield, &
      delete_efield
    use integration_module     , only : new_integration
    use Pcross_section_module  , only : &
      new_Pcross_section, delete_Pcross_section
    use fequil4d_module        , only : &
      allocate_equilibrium_part, &
      compute_equilibrium_part, &
      deallocate_equilibrium_part, init_mass_charge
    use diffusion_module       , only : &
      new_diffusion, del_diffusion
    use collisions_vpar_module , only : &                
      new_collisions_vpar, delete_collisions_vpar        
    use species_transf_module  , only : &                
      new_species_transfer, delete_species_transfer      
#ifndef NO_HDF5
    use f5D_saving_module      , only : &
      new_f5D_saving, delete_f5D_saving
#endif
    use output_saving_module   , only : &
      HDF5_coordsys_saving, HDF5_init_saving, &
      ASCII_init1D_saving
    use diagnostics_module     , only : &
      diagnostics, new_diagnostics
    use integration_module     , only : &
      delete_integration
    use spectra_module         , only : &                
      init_spectra_diag, del_spectra_diag, ndiags_glob   

    use wrap_alloc_module

  implicit none

  !******************************
  contains
  !******************************       

  !-----------------------------------------------
  ! temporary global arrays allocation
  !-----------------------------------------------
  subroutine allocate_temporary_arrays(geom)
    use globals
    use geometry_class
    use MPIutils_module

    implicit none
!R3 #include "code_instrumentation/r3_info.h" !R3
    type(geometry), intent(in)  :: geom
    ! Size of the 1D array RRecvBufMPI_shared computed by accumulation
    integer :: bigtmp_size    
    ! Size of the 1D array RSendBufMPI_shared
    integer :: send_nbM3_size 
    integer :: nbmoments, i, istart_clamp, iend_clamp

!R3 call r3_info_begin (r3_info_index_1, 'Init_alloc_tmp') !R3
    call Write_dyn_balise_open('Init_alloc_tmp') !R4
    bigtmp_size = 0
    nbmoments   = 3
    call write_param("nbmoments", nbmoments)

    bigtmp_size = MAX(bigtmp_size, nbmoments * (dom_r+2*bufsize) * &
      (dom_theta+2*bufsize) * MAX(Nphi,Nbproc_loc))
    call write_expr("bigtmp_size", "nbmoments * (dom_r+2*bufsize) * (dom_theta+2*bufsize) * MAX(Nphi,Nbproc_loc)")
    ! efield.F90 => distribute Phi derivatives
    bigtmp_size = MAX(bigtmp_size, nbmoments * (geom%Nr+1) * &
      (geom%Ntheta+1)*knumber) 
    call write_expr("bigtmp_size", "MAX(bigtmp_size, nbmoments * (geom%Nr+1) * (geom%Ntheta+1)*knumber)")
    ! efield.F90 => distribute Phi derivatives

    !*** arrays for parallel QN solver ***
    nbmoments = 2
    call write_param("nbmoments", nbmoments)
    call FTR_ALLOC_4_ZERO(Rarray_PNrPNthetaNphi_nbM2,0,(nbmoments-1),istart,iend,jstart,jend,0,Nphi-1)
    !*** arrays for collision operator computation ***
    if (collisions) then
      if (version_coll.eq.1)  then
        nbmoments = 6
        call write_param("nbmoments", nbmoments)
        call FTR_ALLOC_2_ZERO(Rarray_PNr_nbM6,0,nbmoments-1,istart,iend)
        call FTR_ALLOC_2_ZERO(Rarray_Nr_nbM6,0,nbmoments-1,0,geom%Nr)
      end if
      if ((version_coll.eq.2).or.(version_coll.eq.3)) then
        nbmoments = 5
        call write_param("nbmoments", nbmoments)
        call FTR_ALLOC_4_ZERO(Rarray_send_nbM5,0,(nbmoments-1),istart,iend,jstart,jend,0,Nphi-1)
        call FTR_ALLOC_2_ZERO(Rarray_PNr_nbM5,0,nbmoments-1,istart,iend)
        call FTR_ALLOC_2_ZERO(Rarray_Nr_nbM5,0,nbmoments-1,0,geom%Nr)
        nbmoments = 2
        call write_param("nbmoments", nbmoments)
        call FTR_ALLOC_2_ZERO(Rarray_PNthetaNphi_nbM2,0,nbmoments*dom_theta-1,0,Nphi)
        call FTR_ALLOC_2_ZERO(Rarray_NthetaNphi_nbM2,0,nbmoments*dom_theta-1,0,(Nphi+1)*Nbproc_theta-1)
      end if
    end if
    !*** arrays for fluid moment computations ***
    nbmoments = 8
    call write_param("nbmoments", nbmoments)
    call FTR_ALLOC_4_ZERO(Rarray_PNrPNthetaNphi_nbM8,0,(nbmoments-1),istart,iend,jstart,jend,0,Nphi-1)

    !*** array allocations for gyroaveraged    ***
    !*** real array allocations ***
    !*** -> array(0:Nr) ***
    call FTR_ALLOC_1_ZERO(Rarray1_Nr,0,geom%Nr)
    call FTR_ALLOC_1_ZERO(Rarray2_Nr,0,geom%Nr)
    call FTR_ALLOC_1_ZERO(Rarray3_Nr,0,geom%Nr)
    call FTR_ALLOC_1_ZERO(Rarray4_Nr,0,geom%Nr)
    call FTR_ALLOC_1_ZERO(Rarray5_Nr,0,geom%Nr)
    call FTR_ALLOC_1_ZERO(Rarray6_Nr,0,geom%Nr)
    call FTR_ALLOC_1_ZERO(Rarray7_Nr,0,geom%Nr)
    !*** -> array(0:Ntheta) ***
    call FTR_ALLOC_1_ZERO(Rarray1_Ntheta,0,geom%Ntheta)
    call FTR_ALLOC_1_ZERO(Rarray2_Ntheta,0,geom%Ntheta)
    !*** -> array(0:Nphi) ***
    call FTR_ALLOC_1_ZERO(Rarray1_Nphi,0,geom%Nphi)
    call FTR_ALLOC_1_ZERO(Rarray2_Nphi,0,geom%Nphi)
    !*** -> array(0:Nvpar) ***
    call FTR_ALLOC_1_ZERO(Rarray1_Nvpar,0,geom%Nvpar)
    call FTR_ALLOC_1_ZERO(Rarray2_Nvpar,0,geom%Nvpar)
    call FTR_ALLOC_1_ZERO(Rarray3_Nvpar,0,geom%Nvpar)
    call FTR_ALLOC_1_ZERO(Rarray4_Nvpar,0,geom%Nvpar)
    call FTR_ALLOC_1_ZERO(Rarray5_Nvpar,0,geom%Nvpar)
    call FTR_ALLOC_1_ZERO(Rarray6_Nvpar,0,geom%Nvpar)
    call FTR_ALLOC_1_ZERO(Rarray7_Nvpar,0,geom%Nvpar)
    !*** -> array(-1:Nphi+1) ***
    call FTR_ALLOC_1_ZERO(Rarray_m1Nphip1,-1,geom%Nphi+1)
    !*** -> array(-1:Nvpar+1) ***
    call FTR_ALLOC_1_ZERO(Rarray_m1Nvparp1,-1,geom%Nvpar+1)
    !*** -> array(0:1,0:Ntheta) ***
    call FTR_ALLOC_2_ZERO(Rarray_1Ntheta,0,1,0,geom%Ntheta)
    !*** -> array(0:Nr,0:Ntheta) ***
    call FTR_ALLOC_2_ZERO(Rarray1_NrNtheta,0,geom%Nr,0,geom%Ntheta)
    call FTR_ALLOC_2_ZERO(Rarray2_NrNtheta,0,geom%Nr,0,geom%Ntheta)
    call FTR_ALLOC_2_ZERO(Rarray3_NrNtheta,0,geom%Nr,0,geom%Ntheta)
    call FTR_ALLOC_2_ZERO(Rarray4_NrNtheta,0,geom%Nr,0,geom%Ntheta)
    call FTR_ALLOC_2_ZERO(Rarray5_NrNtheta,0,geom%Nr,0,geom%Ntheta)
    call FTR_ALLOC_2_ZERO(Rarray6_NrNtheta,0,geom%Nr,0,geom%Ntheta)
    call FTR_ALLOC_2_ZERO(Rarray7_NrNtheta,0,geom%Nr,0,geom%Ntheta)
    !*** -> array(0:Nr,0:Nphi) ***
    call FTR_ALLOC_2_ZERO(Rarray_NrNphi,0,geom%Nr,0,geom%Nphi)
    call FTR_ALLOC_2_ZERO(Rarray2_NrNphi,0,geom%Nr,0,geom%Nphi)
    !*** -> array(0:Nr,0:Nvpar) ***
    call FTR_ALLOC_2_ZERO(Rarray_NrNvpar,0,geom%Nr,0,geom%Nvpar)
    !*** -> array(0:Ntheta,0:Nvpar) ***
    call FTR_ALLOC_2_ZERO(Rarray_NthetaNvparphi,0,geom%Ntheta,0,MAX(geom%Nphi,geom%Nvpar))
    !*** -> array(0:Nphi,0:Nvpar) ***
    call FTR_ALLOC_2_ZERO(Rarray_NphiNvpar,0,geom%Nphi,0,geom%Nvpar)
    !*** -> array(0:Nvpar,0:Nmu) ***
    call FTR_ALLOC_2_ZERO(Rarray_NvparNmu,0,geom%Nvpar,0,geom%Nmu)
    !*** -> array(-1:Nr+1,-1:Ntheta+1) ***
    call FTR_ALLOC_2_ZERO(Rarray1_m1Nrp1m1Nthetap1,-1,geom%Nr+1,-1,geom%Ntheta+1)
    call FTR_ALLOC_2_ZERO(Rarray2_m1Nrp1m1Nthetap1,-1,geom%Nr+1,-1,geom%Ntheta+1)
    call FTR_ALLOC_2_ZERO(Rarray3_m1Nrp1m1Nthetap1,-1,geom%Nr+1,-1,geom%Ntheta+1)
    !*** -> array(-1:Nphi+1,-1:Nvpar+1) ***
    call FTR_ALLOC_2_ZERO(Rarray_m1Nphip1m1Nvparp1,-1,geom%Nphi+1,-1,geom%Nvpar+1)
    !*** -> array(0:Nr,0:Ntheta,iphistart:iphistart+dom_mapphi-1) ***
    call FTR_ALLOC_3_ZERO(Rarray1_NrNthetaDomphi,0,geom%Nr,0,geom%Ntheta,iphistart,iphistart+dom_mapphi-1)
    call FTR_ALLOC_3_ZERO(Rarray2_NrNthetaDomphi,0,geom%Nr,0,geom%Ntheta,iphistart,iphistart+dom_mapphi-1)
    call FTR_ALLOC_3_ZERO(Rarray3_NrNthetaDomphi,0,geom%Nr,0,geom%Ntheta,iphistart,iphistart+dom_mapphi-1)
    call FTR_ALLOC_3_ZERO(Rarray4_NrNthetaDomphi,0,geom%Nr,0,geom%Ntheta,iphistart,iphistart+dom_mapphi-1)
    call FTR_ALLOC_3_ZERO(Rarray5_NrNthetaDomphi,0,geom%Nr,0,geom%Ntheta,iphistart,iphistart+dom_mapphi-1)
    call FTR_ALLOC_3_ZERO(Rarray6_NrNthetaDomphi,0,geom%Nr,0,geom%Ntheta,iphistart,iphistart+dom_mapphi-1)
    call FTR_ALLOC_3_ZERO(Rarray1_NrNthetaParphi,0,geom%Nr,0,geom%Ntheta,ilargestart,ilargeend)
    !*** -> array(istart:iend,jstart:jend,0:Nphi) ***
    !=====TODO REDUCE interval 0->geom%Ntheta
    istart_clamp = max(0,istart_buf)
    iend_clamp   = min(geom%Nr,iend_buf)

    if (distrib3D) then
      call FTR_ALLOC_3_ZERO(Rarray1_NrNthetaPNmu,0,geom%Nr,0,geom%Ntheta,0,Nmuscal-1)
      call FTR_ALLOC_3_ZERO(Rarray1_NrNthetaPNphi,0,geom%Nr,0,geom%Ntheta,ilocstart,ilocstart+dom_locphi-1)
      call FTR_ALLOC_3_ZERO(Rarray2_NrNthetaPNphi,0,geom%Nr,0,geom%Ntheta,ilocstart,ilocstart+dom_locphi-1)
      call FTR_ALLOC_3_ZERO(Rarray3_NrNthetaPNphi,0,geom%Nr,0,geom%Ntheta,ilocstart,ilocstart+dom_locphi-1)
      Rarray1_NrNthetaNphi       => NULL()
      Rarray2_NrNthetaNphi       => NULL()
      Rarray1_NrNthetaNphi_Nspec => NULL()
    else
      !*** -> array(0:Nr,0:Ntheta,0:Nphi) ***
       call FTR_ALLOC_3_ZERO(Rarray2_NrNthetaNphi,0,geom%Nr,0,geom%Ntheta,0,geom%Nphi)
       if (Nspecies.gt.1) then
          call FTR_ALLOC_3_ZERO(Rarray1_NrNthetaNphi_Nspec,0,geom%Nr,0,geom%Ntheta,0,geom%Nphi)
       else
          Rarray1_NrNthetaNphi_Nspec => NULL()
       end if
       Rarray1_NrNthetaPNmu   => NULL()
       Rarray1_NrNthetaPNphi  => NULL()
       Rarray2_NrNthetaPNphi  => NULL()
       Rarray3_NrNthetaPNphi  => NULL()
    end if
  
    !*** complex array allocation ***
    !*** -> array(1:Nr+1,1:Ntheta,1:Nphi) ***
    if (QN_version.eq.1) then      
      call FTR_ALLOC_3_ZERO(Carray_1Nrp11Ntheta1Nphi,1,geom%Nr+1,1,geom%Ntheta,1,geom%Nphi)
    end if                         

    if (Sp_saving) then
      if (QN_version .ne. 1) then
        if (pglobal_id .eq. 0) print *, "ndiags_glob", ndiags_glob
        call FTR_ALLOC_3_ZERO(Carray_1Nrp11Ntheta1Nphi,1,5*ndiags_glob,1,geom%Ntheta,1,geom%Nphi)
      end if
      call FTR_ALLOC_4_ZERO(Carray_1Nrp1Nvpar1Ntheta1Nphi,1,5*ndiags_glob,0,geom%Nvpar,1,geom%Ntheta,1,geom%Nphi)
    end if

    call write_param("nbiter", nbiter)
    !*** array for counting the number of particles out of domain ***
    call FTR_ALLOC_1_ZERO(nbpart_rleft_iter,0,Nmu)
    call FTR_ALLOC_1_ZERO(nbpart_rright_iter,0,Nmu)
    call FTR_ALLOC_1_ZERO(nbpart_thleft_iter,0,Nmu)
    call FTR_ALLOC_1_ZERO(nbpart_thright_iter,0,Nmu)
    if (plocal_id.eq.0) then
      call FTR_ALLOC_1_ZERO(nbpart_rleft_muid,0,nbiter)
      call FTR_ALLOC_1_ZERO(nbpart_rright_muid,0,nbiter)
      call FTR_ALLOC_1_ZERO(nbpart_thleft_muid,0,nbiter)
      call FTR_ALLOC_1_ZERO(nbpart_thright_muid,0,nbiter)
    end if
    if ((FFTpotential_filter.eq.1).or. &
        (FFTpotential_filter.eq.2).or. &
        (single_n).or.(single_m)) then
       call FTR_ALLOC_3_ZERO(Phi_NthNphiPNr,0,geom%Ntheta-1,0,geom%Nphi-1,irstart,irstart+dom_mapr_loc-1)
    end if

!R3 call r3_info_end (r3_info_index_1) !R3
    call Write_dyn_balise_close('Init_alloc_tmp') !R4
  end subroutine allocate_temporary_arrays


  !-----------------------------------------------
  ! temporary global arrays deallocation
  !-----------------------------------------------
  subroutine deallocate_temporary_arrays
    use globals
    implicit none

    integer :: i

    !*** array deallocation for parallel QN solver ***
    call FTR_DEALLOC(Rarray_PNrPNthetaNphi_nbM2)
    !*** array deallocation for collision operator ***
    if (collisions) then
      if (version_coll.eq.1) then
        call FTR_DEALLOC(Rarray_PNr_nbM6)
        call FTR_DEALLOC(Rarray_Nr_nbM6)
      end if
      if ((version_coll.eq.2).or.(version_coll.eq.3)) then
        call FTR_DEALLOC(Rarray_send_nbM5)
        call FTR_DEALLOC(Rarray_PNr_nbM5)
        call FTR_DEALLOC(Rarray_Nr_nbM5)
        call FTR_DEALLOC(Rarray_PNthetaNphi_nbM2)
        call FTR_DEALLOC(Rarray_NthetaNphi_nbM2)
      end if
    end if
    !*** array deallocation for fluid moments ***
    call FTR_DEALLOC(Rarray_PNrPNthetaNphi_nbM8)
    !*** real array deallocations ***
    call FTR_DEALLOC(Rarray1_Nr)
    call FTR_DEALLOC(Rarray2_Nr)
    call FTR_DEALLOC(Rarray3_Nr)
    call FTR_DEALLOC(Rarray4_Nr)
    call FTR_DEALLOC(Rarray5_Nr)
    call FTR_DEALLOC(Rarray6_Nr)
    call FTR_DEALLOC(Rarray7_Nr)
    call FTR_DEALLOC(Rarray1_Ntheta)
    call FTR_DEALLOC(Rarray2_Ntheta)
    call FTR_DEALLOC(Rarray1_Nphi)
    call FTR_DEALLOC(Rarray2_Nphi)
    call FTR_DEALLOC(Rarray1_Nvpar)
    call FTR_DEALLOC(Rarray2_Nvpar)
    call FTR_DEALLOC(Rarray3_Nvpar)
    call FTR_DEALLOC(Rarray4_Nvpar)
    call FTR_DEALLOC(Rarray5_Nvpar)
    call FTR_DEALLOC(Rarray6_Nvpar)
    call FTR_DEALLOC(Rarray7_Nvpar)
    call FTR_DEALLOC(Rarray_m1Nphip1)
    call FTR_DEALLOC(Rarray_m1Nvparp1)
    call FTR_DEALLOC(Rarray_1Ntheta)
    call FTR_DEALLOC(Rarray1_NrNtheta)
    call FTR_DEALLOC(Rarray2_NrNtheta)
    call FTR_DEALLOC(Rarray3_NrNtheta)
    call FTR_DEALLOC(Rarray4_NrNtheta)
    call FTR_DEALLOC(Rarray5_NrNtheta)
    call FTR_DEALLOC(Rarray6_NrNtheta)
    call FTR_DEALLOC(Rarray7_NrNtheta)
    call FTR_DEALLOC(Rarray_NrNphi)
    call FTR_DEALLOC(Rarray2_NrNphi)
    call FTR_DEALLOC(Rarray_NrNvpar)
    call FTR_DEALLOC(Rarray_NthetaNvparphi)
    call FTR_DEALLOC(Rarray_NphiNvpar)
    call FTR_DEALLOC(Rarray_NvparNmu)
    call FTR_DEALLOC(Rarray1_m1Nrp1m1Nthetap1)
    call FTR_DEALLOC(Rarray2_m1Nrp1m1Nthetap1)
    call FTR_DEALLOC(Rarray3_m1Nrp1m1Nthetap1)
    call FTR_DEALLOC(Rarray_m1Nphip1m1Nvparp1)
    call FTR_DEALLOC(Rarray1_NrNthetaDomphi)
    call FTR_DEALLOC(Rarray2_NrNthetaDomphi)
    call FTR_DEALLOC(Rarray3_NrNthetaDomphi)
    call FTR_DEALLOC(Rarray4_NrNthetaDomphi)
    call FTR_DEALLOC(Rarray5_NrNthetaDomphi)
    call FTR_DEALLOC(Rarray6_NrNthetaDomphi)
    call FTR_DEALLOC(Rarray1_NrNthetaParphi)
    call FTR_DEALLOC(Rarray1_PNrPNthetaNphi)
    call FTR_DEALLOC(Rarray2_PNrPNthetaNphi)
    call FTR_DEALLOC(Rarray3_PNrPNthetaNphi)
    call FTR_DEALLOC(Rarray4_PNrPNthetaNphi)

    if (Sp_saving) then
      if (QN_version.ne.1) then
        call FTR_DEALLOC(Carray_1Nrp11Ntheta1Nphi)
      end if
      call FTR_DEALLOC(Carray_1Nrp1Nvpar1Ntheta1Nphi)
    end if

    !*** complex array deallocations ***
    if (QN_version.eq.1) &                     
      call FTR_DEALLOC(Carray_1Nrp11Ntheta1Nphi)

    if (distrib3D) then
      call FTR_DEALLOC(Rarray1_NrNthetaPNmu)
      call FTR_DEALLOC(Rarray1_NrNthetaPNphi)
      call FTR_DEALLOC(Rarray2_NrNthetaPNphi)
      call FTR_DEALLOC(Rarray3_NrNthetaPNphi)
    else
      call FTR_DEALLOC(Rarray2_NrNthetaNphi)
      if(Nspecies.gt.1)then
        call FTR_DEALLOC(Rarray1_NrNthetaNphi_Nspec)
      end if
    end if

    !*** array for counting the number of particles out of domain ***
    call FTR_DEALLOC(nbpart_rleft_iter)
    call FTR_DEALLOC(nbpart_rright_iter)
    call FTR_DEALLOC(nbpart_thleft_iter)
    call FTR_DEALLOC(nbpart_thright_iter)
    if (plocal_id.eq.0) then
      call FTR_DEALLOC(nbpart_rleft_muid)
      call FTR_DEALLOC(nbpart_rright_muid)
      call FTR_DEALLOC(nbpart_thleft_muid)
      call FTR_DEALLOC(nbpart_thright_muid)
    end if
    if ((FFTpotential_filter.eq.1).or. &
        (FFTpotential_filter.eq.2).or. &
        (single_n).or.(single_m)) then
       call FTR_DEALLOC(Phi_NthNphiPNr)
    end if
  end subroutine deallocate_temporary_arrays


  !**************************************************
  !  INITIALISATION OF GYSELA
  !**************************************************
  subroutine init_gysela(geom,coord_sys, &
    init_prof,init_magnet,init_curr, init_srce, &                                  
    fna,fnb,f5d_restart_save,fmu_equil, &
    J0,J0fmu_equil,poiss,init_time,diag, &
    Sinit_mode,sEntransf_read)
    use clock_module, only : clck_time, clck_diff, &
      bclock_init, eclock_init, &
      global_time_init, initialization_time
    use krook_module                 
    use diffusion_module
    use sources_module
    use globals              , only : read_Vpar0, &
      jstart, jend, iend
    use globals              , only : istart_buf, iend_buf
    implicit none

!R3 #include "code_instrumentation/r3_info.h" !R3
    type(geometry)           , intent(out)  :: geom
    type(coord_system)       , intent(out)  :: coord_sys
    type(init_profile)       , intent(out)  :: init_prof
    type(init_magnetic)      , intent(out)  :: init_magnet
    type(init_current)       , intent(out)  :: init_curr
    type(init_sources)       , intent(out)  :: init_srce 
    type(fdistribu5d), target, intent(out)  :: fna
    type(fdistribu5d), target, intent(out)  :: fnb
    type(fdistribu5d), target, intent(out)  :: f5d_restart_save
    real(RKIND), &
      dimension(:,:,:)       , pointer       :: fmu_equil
    type(J0operator)         , intent(out)   :: J0
    real(RKIND), &
      dimension(:,:,:)       , pointer       :: J0fmu_equil
    type(poisson)            , intent(out)   :: poiss
    real(RKIND)              , intent(inout) :: init_time
    type(diagnostics)        , intent(out)   :: diag
    logical                  , intent(out)   :: Sinit_mode
    logical                  , intent(out)   :: sEntransf_read

    integer     :: itheta, ivpar
    real(RKIND) :: fmu_eq_tmp
    integer :: istart_clamp, iend_clamp

    !*** creation of the file for the dynamic memory size saving ***
    call new_dyn_allocate()

#ifdef CHECK_MEM_LEAK
    !*** creation of the file for the trace of memory leak ***
    call new_trace_mem_leak()
#endif

    !*** initialization of the variables for CPU time computation ***
    call initialization_time()

!R3 call r3_info_begin (r3_info_index_0, 'Init_gysela') !R3
    call Write_dyn_balise_open('Init_gysela') !R4

    !*** input file reading ***
    call read_input()

    !*** creation of the file for the memory size saving ***
    ! call new_allocate()

    !*** initialisation of the communicators ***
    call ppinit_comm()

    !*** allocation of the OpenMP arrays ***
    call init_OMP()

    !------------------------------------------------
    ! ARRAY ALLOCATION AND INITIALIZATION
    !------------------------------------------------
    call clck_time(bclock_init)
    iter_run = 0

    !*** charge and mass initialization   ***
    call init_mass_charge(A0,Z0,As,Zs,species_frac,hat_As,hat_Zs)

    !*** mesh and profile initializations ***
    call init_mesh_profile(geom,init_prof)

    !*** initialization of dissipation coefficient profiles ***
    call write_expr("sNr", "geom%Nr")
    call new_diffusion(geom%Nr)

    if (.not.memory_test) then
      call init_dissip_coeff(geom,init_prof)
      call init_diffus_matrix(geom)
    end if

    !*** initialization of the coordinate system ***
    call init_coordinate_system(coord_sys,geom)

    !*** initialization of magnetic configuration ***
    call init_magnetic_config(init_magnet,geom, &
      coord_sys,init_prof,init_curr)

    !*** initialization of the plasma current ***
    call init_plasma_current(init_curr,geom,coord_sys, &
      init_prof,init_magnet%zeta)

    call init_precompute_Bstar(init_magnet,init_curr, &
      geom,coord_sys,init_prof)

    !*** initialization of the Krook operator ***
    call init_Krook(geom%Nr,coef_krook,init_prof%buffer_r)

    !*** initialization of the velocities ***
    if (conservative_scheme) &
      call new_velocities_edges(geom)

    !*** allocation of the distribution functions ***
    call new_f5D(fna,geom)
    call new_f5D(fnb,geom)
#ifndef NO_HDF5
    if (asynchrone_writing) &
      call new_f5D(f5d_restart_save,geom)
#endif

    !*** For spectra diagnostic
    if (Sp_saving) then
      call init_spectra_diag(geom)
    end if

    !*** temporary array allocations ***
    call allocate_temporary_arrays(geom) 

    !*** J0(Phi) allocation ***
    call new_J0operator(J0,geom)

    !*** source initialization ***
    if (.not.memory_test) then
      if (atleast1_source) &
        call init_source_properties(init_srce,init_prof, &
        init_magnet,init_curr,geom,J0)
    end if

    !*** Phi and Efield allocation ***
    call new_poisson(poiss,geom,init_prof)
    if (QN_version.ge.4) &
      call new_QNsolver(geom,init_prof)

    !*** allocation for physic value computations ***
    call new_physics_GC(geom)
    call new_physics_part(geom)
    call new_efield(geom)

    istart_clamp = max(0,istart_buf)
    iend_clamp   = min(geom%Nr,iend_buf)
	
    call new_integration(geom)
    call new_Pcross_section(geom)

    !*** allocation and computation of the equilibrium part ***
    call allocate_equilibrium_part(geom, &
      init_prof,init_magnet,fmu_equil,J0fmu_equil, &
      nseq_r,dnseq_dr,Tseq_r,TGCseq_r)
    if (.not.memory_test) then
      call compute_equilibrium_part(geom,J0,init_prof, &
        init_magnet,init_curr,fmu_equil,J0fmu_equil, &
        nseq_r,dnseq_dr,Tseq_r,nbions_eq,TGCseq_r)
    end if

    !*** collision initialization (Crank-Nicolson matrix) ***
    if (collisions) then
      call new_collisions_vpar(geom,init_prof,init_magnet)
      if ( (energy_transfer).or.(moment_transfer) ) &
        call new_species_transfer(geom%Nr)
    end if

    !*** allocation for the HDF5 file saving ***
#ifndef NO_HDF5
    if (f5D_saving) call new_f5D_saving()
#endif
    !*** initialization of the distribution function ***
    !***   at the first time                         ***
    call init_fdistribution(geom,init_prof, &
      init_time,diag,fmu_equil,fnb, &
      Sinit_mode,sEntransf_read)

    !*** change values at r=rmax ***
    if ((read_Vpar0).and.(iend.eq.geom%Nr)) then
      do ivpar=0,Nvpar
        do itheta = jstart, jend
          call compute_feq_value(geom,iend,itheta,ivpar,mu_id, &
            init_prof,init_magnet,fmu_eq_tmp)
          fnb%values(iend,itheta,:,ivpar) = fmu_eq_tmp
        end do
      end do
    end if

    !*** initialization of the diagnostics ***
    call init_diagnostics(Sinit_mode, &
      geom,coord_sys,init_prof, &
      init_magnet,init_curr,init_srce,J0,diag)

    call clck_time(eclock_init)
    call clck_diff(bclock_init,eclock_init, &
      global_time_init)

    call Write_dyn_balise_close('Init_gysela') !R4
!R3 call r3_info_end(r3_info_index_0) !R3
  end subroutine init_gysela


  !---------------------------------------------------
  ! Initialization of the distribution function
  !---------------------------------------------------
  subroutine init_fdistribution(geom,init_prof, &
    Stime_init,Sdiag,Sfmu_equil,Sfn, &
    Sinit_mode,sEntransf_read)
    use globals, only : restart, pSpecies_id, Nspecies, &
      nb_restart, nb_restart_diag, num_restart, iter_glob      
    use diagnostics_module
    use fdistribu5d_class
    use read_write_module    , only : init_fn_rst
    use vlasov_solving_module, only : initialisation_fstart
    use MPIutils_module
    use mpiMod
    use Io_Unit_module, only: IU_open, IU_close
    
    implicit none
    
    type(geometry)    , intent(in)    :: geom
    type(init_profile), intent(in)    :: init_prof
    real(RKIND)       , intent(inout) :: Stime_init
    type(diagnostics) , intent(inout) :: Sdiag
    real(RKIND), &
      dimension(:,:,:), pointer       :: Sfmu_equil
    type(fdistribu5d) , intent(out)   :: Sfn
    logical           , intent(out)   :: Sinit_mode
    logical           , intent(out)   :: sEntransf_read

    integer           :: ierr, ispec
    character(LEN=50) :: restart_file
    real(RKIND)       :: time_init_spec
    logical           :: diff_init_species

    !*** Test if the simulation is a restart case      ***
    !*** This test is done separately for each species ***
    call testif_restart(Sinit_mode,diff_init_species)

    if (.not.memory_test) then
      !*** Initialisation of the distribution function ***
      if (restart(pSpecies_id)) then
        call init_fn_rst(Sfn, &
          Sdiag%nb_diag,Sdiag%Phi3D_nb_diag, &
          Sdiag%Moments3D_nb_diag,Sdiag%f5D_nb_diag, &
          Sdiag%SP_nb_diag,time_init_spec,num_restart, &
          nb_restart,nbiter_prev,sEntransf_read)
        nb_restart = nb_restart+1
      else
        call initialisation_fstart(geom,init_prof,Sfmu_equil,Sfn)
        !--> time infos
        num_restart             = 0
        nb_restart              = 0
        nbiter_prev             = 0
        time_init_spec          = 0._RKIND
        !---> diagnostics infos
        Sdiag%nb_diag           = 0
        Sdiag%Phi3D_nb_diag     = 0
        Sdiag%Moments3D_nb_diag = 0
        Sdiag%f5D_nb_diag       = 0
        Sdiag%SP_nb_diag        = 0
      end if
    endif
    !*** If one species is added the diagnostics ***
    !***  are all reinitialized to 0             ***
    if (Sinit_mode) then
      num_restart             = 0
      nb_restart              = 0
      nbiter_prev             = 0
      !---> diagnostics infos
      Sdiag%nb_diag           = 0
      Sdiag%Phi3D_nb_diag     = 0
      Sdiag%Moments3D_nb_diag = 0
      Sdiag%f5D_nb_diag       = 0
      Sdiag%SP_nb_diag        = 0
    end if
    if (diff_init_species) then
      call MY_Allreduce(time_init_spec,Stime_init,1,MPI_REAL8, &
        MPI_MAX,MPI_COMM_WORLD,ierr)
    else    
      Stime_init = time_init_spec
    end if

    nb_restart_diag = nb_restart
    iter_glob       = nbiter_prev

    !*** Writting in the ASCII output file ***
    if (pglobal_id.eq.outputproc) then
      call IU_open(unit=uout_res,file=file_name_res,status='OLD',&
        position='APPEND', form='FORMATTED')
      do ispec = 0,Nspecies-1
        if (.not.restart(ispec)) then
          write(uout_res,'(A,A,I1.1)') '---> Distribution function ', &
            'initialisation : mode NO RESTART for species ', &
            ispec
        else
          write(restart_file(1:),'(A,I1.1,A,I1.1,A)') &
            "gys_rst.sp",ispec,".n",num_restart, &
            ".fn.p<num_proc>.h5"
          write(uout_res,'(A,A,A,A,I1.1)') &
            '---> Distribution function ', &
            'initialisation : mode RESTART with ', &
            restart_file(1:len_trim(restart_file)), &
            ' for species ', ispec
        end if
      end do
      if (asynchrone_writing) then
        write(uout_res,'(A)') '---> Asynchrone writting'
      else
        write(uout_res,'(A)') '---> Synchrone writting'
      endif
      call IU_close(uout_res)
    end if
  end subroutine init_fdistribution


  !---------------------------------------------------
  ! Initialization of the diagnostics
  !---------------------------------------------------
  subroutine init_diagnostics(Sinit_mode, &
    geom,coord_sys,init_prof, &
    init_magnet,init_curr, init_srce,J0,diag)
    use globals     , only : memory_test, restart, &
      pInSpecies_id, atleast1_source
    use output_saving_module
    logical                  , intent(in)    :: Sinit_mode
    type(geometry)           , intent(in)    :: geom
    type(coord_system)       , intent(in)    :: coord_sys
    type(init_profile)       , intent(in)    :: init_prof
    type(init_magnetic)      , intent(in)    :: init_magnet
    type(init_current)       , intent(in)    :: init_curr
    type(init_sources)       , intent(in)    :: init_srce 
    type(J0operator)         , intent(inout) :: J0
    type(diagnostics)        , intent(out)   :: diag

    integer     :: status

    call Write_dyn_balise_open('init_diagnostics') !R4

    !*** write the nb_restart in gysela_dyn_memory.out ***
    ! call Write_dyn_number(int(nb_restart,kind=8), "nb_restart")

    !*** input parameter saving ***
    call Write_parameters()
#ifndef NO_HDF5
    if (pInSpecies_id.eq.0) &
      call HDF5_data_saving()
#endif

    !*** initialisation of the diagnostic saving ***
    call new_diagnostics(Phi3D_saving,Moments3D_saving,f5D_saving,Sp_saving,&
      dt_diag,Phi3D_dt_diag,Moments3D_dt_diag,f5D_dt_diag,Sp_dt_diag,diag)

    !*** Saving of the initial state ***
    if (.not.memory_test) then
      if (Sinit_mode.and.(pInSpecies_id.eq.0)) then
        call HDF5_coordsys_saving(geom,coord_sys)
      end if
      if (atleast1_source) then
        if (pInSpecies_id.eq.0) then
#ifndef NO_HDF5
          call HDF5_init_saving(geom,init_prof, &
            init_magnet,init_curr,init_srce)
#else
          call ASCII_init1D_saving(geom,init_prof, &
            init_magnet,init_srce)
#endif
        end if
        call diag_sources(init_srce,geom,init_magnet,init_curr,J0)
      else
        if (pInSpecies_id.eq.0) then
#ifndef NO_HDF5
          call HDF5_init_saving(geom,init_prof, &
            init_magnet,init_curr)
#else
          call ASCII_init1D_saving(geom,init_prof, &
            init_magnet)
#endif
        end if
      end if                 
    end if

    if (memory_test) then
      call ppexit()
      status = -1
      if (pglobal_id.eq.0) print *,' STOP ! memory_test=.true. '
      call exit(status)
    end if
!R3 call r3_info_end(r3_info_index_0) !R3
    call Write_dyn_balise_close('init_diagnostics') !R4
  end subroutine init_diagnostics


  !*************************************************
  ! RESTART WRITTING
  !************************************************* 
  subroutine restart_writting_gysela(diag,Siter_time, &
    pfn,pfnp1,pfn_restart_save)
    use read_write_module
    use advec2D_BSL_module
    use mpiMod
    type(diagnostics), intent(in) :: diag
    real(RKIND)      , intent(in) :: Siter_time
    type(fdistribu5d), pointer    :: pfn
    type(fdistribu5d), pointer    :: pfnp1
    type(fdistribu5d), pointer    :: pfn_restart_save

    integer :: ierr, i, j, k, l, h5_lock_tmp

    nbiter_prev = iter_glob  
    call MPI_BCAST(num_restart,1,MPI_INTEGER, &
      0,MPI_COMM_WORLD,ierr)

    if (asynchrone_writing) then
      call MY_Allreduce(h5_lock,h5_lock_tmp,1, &
        MPI_INTEGER,MPI_MAX,MPI_COMM_WORLD,ierr)
      if (h5_lock_tmp == 0) then
        call write_f5d_rst("fn",Siter_time,diag%nb_diag, &
          diag%Phi3D_nb_diag,diag%Moments3D_nb_diag, &
          diag%f5D_nb_diag,diag%Sp_nb_diag, &                    
          pfnp1,pfn_restart_save)
      endif
    else
      call write_f5d_rst("fn",Siter_time,diag%nb_diag, &
        diag%Phi3D_nb_diag,diag%Moments3D_nb_diag, &
        diag%f5D_nb_diag,diag%Sp_nb_diag, &    
        pfnp1,NULL())
    endif

    !*** Global barrier in order to guaranty that ***
    !*** restart files are well written           *** 
    call ppbarrier()

    !*** writting of the files 'file_list.out' ***
    !***  (used for the restart script)        ***
    call end_restart()

    !*** write the number of particles out of domain ***
    if (plocal_id.eq.0) then
      call print_particles_outofdomain()
    end if
  end subroutine restart_writting_gysela


  !**************************************************
  !  END OF GYSELA
  !**************************************************
  subroutine delete_gysela(geom,coord_sys, &
    init_prof,init_magnet,init_curr,init_srce, &
    fna,fnb,f5d_restart_save, &
    fmu_equil,J0,J0fmu_equil,poiss)
    use krook_module
    use Io_Unit_module, only: IU_open, IU_close
    
    implicit none
    
    type(geometry)               , intent(inout) :: geom
    type(coord_system)           , intent(inout) :: coord_sys
    type(init_profile)           , intent(inout) :: init_prof
    type(init_magnetic)          , intent(inout) :: init_magnet
    type(init_current)           , intent(inout) :: init_curr
    type(init_sources)           , intent(inout) :: init_srce 
    type(fdistribu5d)    , target, intent(inout) :: fna
    type(fdistribu5d)    , target, intent(inout) :: fnb
    type(fdistribu5d)    , target, intent(inout) :: f5d_restart_save
    real(RKIND), dimension(:,:,:), pointer       :: fmu_equil
    type(J0operator)             , intent(inout) :: J0
    real(RKIND), dimension(:,:,:), pointer       :: J0fmu_equil
    type(poisson)                , intent(inout) :: poiss

    call Write_dyn_balise_open('Delete_gysela') !R4

    call FTR_DEALLOC(fna%values)
    call FTR_DEALLOC(fnb%values)
#ifndef NOHDF5
    if (asynchrone_writing) &
      call FTR_DEALLOC(f5d_restart_save%values)
#endif
	
    !*** array deallocations ***
    !-> deallocate arrays used for OpenMP parallelization
    call del_OMP()
    call del_geometry(geom)
    call del_init_profile(init_prof)
    call del_diffusion()
    call del_coord_system(coord_sys)
    call del_magnet_config(init_magnet)
    call del_init_current(init_curr)
    if (atleast1_source)  then
       call del_init_sources(init_srce)
    endif
    if (conservative_scheme) then
       call del_velocities_edges()
    endif
    call deallocate_equilibrium_part(fmu_equil, &
      J0fmu_equil,nseq_r,dnseq_dr,Tseq_r,TGCseq_r)
    !-> deallocate distribution functions
    call del_f5D(fna)
    call del_f5D(fnb)
#ifndef NO_HDF5
    if (asynchrone_writing) &
      call del_f5D(f5d_restart_save)
#endif

    !-> deallocate poisson + gyroaverage and collision operators
    call del_poisson(poiss)
    if (QN_version.ge.4) call del_QNsolver()    
    call del_J0operator(J0)
    if (collisions) then
      call delete_collisions_vpar()
      if ( (energy_transfer).or.(moment_transfer) ) & 
        call delete_species_transfer() 
    end if    
    !-> deallocate physical quantities 
    call delete_integration()
    call delete_physics_GC()
    call delete_physics_part()
    call delete_efield()
    call delete_Pcross_section()
    !-> deallocate 5D arrays
#ifndef NO_HDF5
    if (f5D_saving) call delete_f5D_saving()
#endif

    !-> deallocate Spectral arrays
    if (Sp_saving) then
      call del_spectra_diag()
    end if

    !-> deallocate temporary arrays
    call delete_Krook()

    !-> deallocate temporary arrays
    call deallocate_temporary_arrays()
    !-> deallocate parallel buffers 
    call ppdeallocate()

    call Write_dyn_balise_close('Delete_gysela') !R4

    call print_dyn_memory_size()
  end subroutine delete_gysela
end module gys_alloc_module

