!**************************************************************
!  Copyright Euratom-CEA
!  Authors : 
!     Virginie Grandgirard (virginie.grandgirard@cea.fr)
!     Chantal Passeron (chantal.passeron@cea.fr)
!     Guillaume Latu (guillaume.latu@cea.fr)
!     Xavier Garbet (xavier.garbet@cea.fr)
!     Philippe Ghendrih (philippe.ghendrih@cea.fr)
!     Yanick Sarazin (yanick.sarazin@cea.fr)
!  
!  This code GYSELA (for GYrokinetic SEmi-LAgrangian) 
!  is a 5D gyrokinetic global full-f code for simulating 
!  the plasma turbulence in a tokamak.
!  
!  This software is governed by the CeCILL-B license 
!  under French law and abiding by the rules of distribution 
!  of free software.  You can  use, modify and redistribute 
!  the software under the terms of the CeCILL-B license as 
!  circulated by CEA, CNRS and INRIA at the following URL
!  "http://www.cecill.info". 
!**************************************************************

#include "config-F.inc"

!---------------------------------------------------------------
! file : gysela.F90
! date : 28/02/2002
! main program
!                      GYROKINETIC CODE
!  - f(r,theta,phi,vpar,mu) : 5D ionic distribution function
!  - Phi(r,theta,phi)       : 3D electric potential
!  Solving of a 5D gyroaverage vlasov equation coupled to a 3D 
!   quasi-neutrality equation
!---------------------------------------------------------------
program gysela
  use globals  
  use clock_module
  use OMPutils_module
  use MPIutils_module
  use geometry_class
  use init_profile_class
  use coord_system_class
  use init_magnetic_class
  use init_current_class
  use sources_module
  use gys_alloc_module
  use Boltz_solving_module 
  use diagnostics_module
  use read_write_module
#ifndef NO_HDF5
  use resu3D_saving_module
#endif
  use version_module, only: write_version

  implicit none

#ifdef NO_HDF5
  use mpiMod
#endif
  
  !*** variable definition ***
  !-> mesh
  type(geometry)              :: geom
  !-> initial profiles
  type(init_profile)          :: init_prof
  !-> definition of the coordinate system
  type(coord_system)          :: coord_sys
  !-> initial plasma current
  type(init_current)          :: init_curr
  !-> initial magnetic configuration   
  type(init_magnetic)         :: init_magnet
  !-> initial source 
  type(init_sources)          :: init_srce
  !-> distribution function f(tn-1)
  type(fdistribu5d), pointer :: pfnm1
  !-> distribution function f(tn)
  type(fdistribu5d), pointer :: pfn
  !-> distribution function f(tn+1)
  type(fdistribu5d), pointer :: pfnp1
  !-> saving f(tn) for asynchrone writing
  type(fdistribu5d), pointer :: pfn_restart_save
  !-> distribution function 1
  type(fdistribu5d), target  :: fna
  !-> distribution function 2
  type(fdistribu5d), target  :: fnb
  !-> distribution function 3
  type(fdistribu5d), target  :: f5d_restart_save
  !-> J0.f
  type(fdistribu5d), pointer :: J0ftmp
  !-> gyroaveraged distribution function
  type(fdistribu5d)          :: gyrofn
  type(poisson)              :: poiss
  !-> equilibrium distribution functions 
  real(RKIND), dimension(:,:,:), pointer :: fmu_equil
  real(RKIND), dimension(:,:,:), pointer :: J0fmu_equil
  
  !-> for time following 
  !---> initial time of the run
  real(RKIND)       :: init_time   
  !---> time relative to the current iteration
  real(RKIND), save :: iter_time
  type(diagnostics) :: diag

  ! gyroaverage operator
  type(J0operator) :: J0

  ! used for energy transfer operator
  logical :: Entransf_read

  ! used for output
  integer     :: Nbt
  real(RKIND) :: max_memory
  integer     :: ierr_ce, ierr

  ! used for interrupt the run
  logical :: fexist, run
  integer :: status, moditer

  fexist = .false.
  mem_dyn_write = .true.
  !*** initialisation for the parallelisation ***
  call ppinit()

!R3 file_r3_open = .false. !R3
!R3 call r3_info_init () !R3


  !*** write version info ***
  call write_version()


  !*************************************************
  !  INITIALISATION 
  !*************************************************
  call init_gysela(geom,coord_sys, &
    init_prof,init_magnet,init_curr,init_srce, &
    fna,fnb,f5d_restart_save,fmu_equil, &
    J0,J0fmu_equil,poiss,init_time,diag, &
    init_mode,Entransf_read)

  !*************************************************
  ! FIRST STEP
  !*************************************************
  pfn    => fna
  pfnp1  => fnb
  J0ftmp => pfn
  if (asynchrone_writing) &
    pfn_restart_save => f5d_restart_save

  !*** First step ***
  call step0_predcorr(geom,coord_sys, &
    init_prof,init_magnet,init_curr,init_srce, &
    diag,deltat,pfnp1,J0ftmp,fmu_equil,J0fmu_equil, &
    poiss,J0,Entransf_read)

  !*** saving at the time t=0  ***
  if (init_mode) then
    call init_diagnostics_time(init_mode,init_time,diag)
    call check_diagnostics(init_time,diag)
    call compute_diagnostics(init_time,diag,deltat, &
      geom,coord_sys,init_prof,init_magnet,init_curr,init_srce, &
      pfnp1,pfnp1,J0ftmp,fmu_equil,J0fmu_equil,poiss,J0)
  else
    call init_diagnostics_time(init_mode,init_time,diag)
  end if

  call ppbarrier()
  iter_time = init_time


  !*************************************************
  ! GLOBAL SOLVING BEGINNING
  !*************************************************
!R3 call r3_info_print (-2, -2, 'INITIALIZATION') !R3

  iter_run  = 1
  run       = .true.
  status    = 0
  do while (run)
    iter_glob = iter_glob + 1
    iter_time = iter_time + deltat
    mem_dyn_write = (iter_beg_write_trace.le.iter_glob.and.&
         iter_glob.le.iter_end_write_trace)

    call predcorr(geom, &
      coord_sys,init_prof, &
      init_magnet,init_curr,init_srce, &
      deltat,iter_time,diag%time_save, &
      pfn,pfnp1,fmu_equil,J0,poiss)

    !*** diagnostics in the case iter_time = time_save ***
    call check_diagnostics(iter_time,diag)
    if (diag%diag_OK) then
      !*** saving at the time t = tn+1 ***
      J0ftmp => pfn
      call compute_diagnostics(iter_time,diag,deltat, &
        geom,coord_sys,init_prof,init_magnet,init_curr,init_srce, &
        pfn,pfnp1,J0ftmp,fmu_equil,J0fmu_equil,poiss,J0)
      if (pglobal_id.eq.0) then
        inquire(file='gysela.stop', exist=fexist)
        if (fexist) then
          write(6,*) ' '
          write(6,'(A)') &
            ' ===> The signal file "gysela.stop" exist '
          write(6,'(A,I6,A,E12.5)') &
            ' ===> So! stop the run at iter_glob = ', &
            iter_glob,' iter_time = ',iter_time
        end if
      end if
      call MPI_BCAST(fexist,1,MPI_LOGICAL,0,MPI_COMM_WORLD,ierr)
      if (fexist) then
        run    = .false.
        status = 1
      end if
    else
      call WriteScreen(iter_time,deltat,.false.)
    endif
!R3    if ((iter_run .eq. 1).and.(.not.restart)) then     !R3
!R3      call r3_info_print (-3, -2, 'ITERATION    1')    !R3
!R3    else                                               !R3
!R3      call r3_info_print (iter_glob, -2, 'ITERATION')  !R3
!R3    end if                                             !R3

    iter_run = iter_run+1
    if (iter_run.gt.nbiter) run = .false.

   !*************************************************
   ! RESTART WRITTING
   !*************************************************
!R3 call r3_info_summary () !R3
    moditer = mod(iter_run, modulo_restart_iter)
    if ((moditer .eq. 0).and.(iter_run.le.nbiter).and.(.not.fexist)) then
      call restart_writting_gysela(diag,iter_time,pfn,pfnp1, &
        pfn_restart_save)
    elseif ((iter_run.gt.nbiter).or.fexist) then
      if ((fexist).and.(pglobal_id.eq.0)) then
        write(6,*) ' '
        write(6,'(A)') &
          ' ===> writing last restart files before stop...'
      end if
      if (h5_lock == 1) call thread_restart_barrier()
      asynchrone_writing = .false.
      call restart_writting_gysela(diag,iter_time,pfn,pfnp1, &
        pfn_restart_save)
    endif

    if (stop_gysela .ne. 0) then
      if (pglobal_id.eq.0) then
        write(6,*) ' '
        write(6,'(A)') &
          ' ===> The signal file "gysela.stop" exist '
        write(6,'(A,I6,A,E12.5)') &
          ' ===> So! stop the run at iter_glob = ', &
          iter_glob,' iter_time = ',iter_time
      end if
      stop
    endif
  enddo

  mem_dyn_write = .true.

  !*** write infos on the CPU time ***
  call write_time()

  !*************************************************
  ! DELETING
  !************************************************* 
  call delete_gysela(geom,coord_sys,init_prof, &
    init_magnet,init_curr,init_srce, &
    fna,fnb,f5d_restart_save,fmu_equil, &
    J0,J0fmu_equil,poiss)

  !*** end of the parallelisation ***
  call ppexit()
  if (.not.run) call exit(status)
end program gysela
