!**************************************************************
!  Copyright Euratom-CEA
!  Authors : 
!     Virginie Grandgirard (virginie.grandgirard@cea.fr)
!     Chantal Passeron (chantal.passeron@cea.fr)
!     Guillaume Latu (guillaume.latu@cea.fr)
!     Xavier Garbet (xavier.garbet@cea.fr)
!     Philippe Ghendrih (philippe.ghendrih@cea.fr)
!     Yanick Sarazin (yanick.sarazin@cea.fr)
!  
!  This code GYSELA (for GYrokinetic SEmi-LAgrangian) 
!  is a 5D gyrokinetic global full-f code for simulating 
!  the plasma turbulence in a tokamak.
!  
!  This software is governed by the CeCILL-B license 
!  under French law and abiding by the rules of distribution 
!  of free software.  You can  use, modify and redistribute 
!  the software under the terms of the CeCILL-B license as 
!  circulated by CEA, CNRS and INRIA at the following URL
!  "http://www.cecill.info". 
!**************************************************************

#include "config-F.inc"

!-------------------------------------------------
! file : geometry.F90
! date : 16/09/2005
! 5d geometry definition (r,theta,phi,vparallel,mu)
!--------------------------------------------------
module geometry_class
  use prec_const
  use wrap_alloc_module

  use wrap_alloc_module
  implicit none
  
  !*** geometry type definition ***
  type :: geometry
    !->  origin coordinates
    real(RKIND) :: rmin, thetamin, phimin, vparmin, mumin
    !-> grid step size
    real(RKIND) :: dr, dtheta, dphi, dvpar
    !-> grid length            
    real(RKIND) :: Lr, Ltheta, Lphi, Lvpar, Lmu
    !-> number of grid points
    integer     :: Nr, Ntheta, Nphi, Nvpar, Nmu
    ! coordinates of the points in the 5 directions
    real(RKIND), dimension(:), pointer :: rg
    real(RKIND), dimension(:), pointer :: thetag
    real(RKIND), dimension(:), pointer :: phig
    real(RKIND), dimension(:), pointer :: vparg
    real(RKIND), dimension(:), pointer :: mug
    ! coefficients for the integral in all directions 
    !   with trapeze method
    real(RKIND), dimension(:), pointer :: coeff_intdr
    real(RKIND), dimension(:), pointer :: coeff_intdtheta
    real(RKIND), dimension(:), pointer :: coeff_intdphi
    real(RKIND), dimension(:), pointer :: coeff_intdvpar
    real(RKIND), dimension(:), pointer :: coeff_intdmu
    ! array for cos(theta) and sin(theta)
    real(RKIND), dimension(:), pointer :: cos_theta, sin_theta
    ! index in vparg of vparg = 0.
    integer :: ivpar0
  end type geometry

  ! power for the integral computation
  !(ipow=0 if Nmu=0 and Lmu=0.,ipow=1 otherwise)
  integer, public :: ipow
    
  !******************************
  contains
  !******************************

  !----------------------------------------------------
  ! initialisation of the space coordinates (r,theta,phi)
  !----------------------------------------------------  
  subroutine new_space_coord(geom,rmin,thetamin,phimin, &
    Lr,Ltheta,Lphi,Nr,Ntheta,Nphi)
    use globals, only : memory_test
    type(geometry) , intent(out)   :: geom
    real(RKIND)    , intent(in)    :: rmin, thetamin, phimin 
    real(RKIND)    , intent(in)    :: Lr, Ltheta, Lphi       
    integer        , intent(in)    :: Nr, Ntheta, Nphi       
    
    integer :: i1, i2, i3
    
    !*** variables initialisation ***
    !-> minimal values
    geom%rmin     = rmin
    geom%thetamin = thetamin
    geom%phimin   = phimin
    !-> length
    geom%Lr       = Lr
    geom%Ltheta   = Ltheta
    geom%Lphi     = Lphi
    !-> number of points
    geom%Nr       = Nr
    geom%Ntheta   = Ntheta
    geom%Nphi     = Nphi
    !-> grid step size
    geom%dr       = Lr/float(Nr)
    geom%dtheta   = Ltheta/float(Ntheta)
    geom%dphi     = Lphi/float(Nphi)
    
    call write_expr("geom%Nr", "Nr")
    call write_expr("geom%Ntheta", "Ntheta")
    call write_expr("geom%Nphi", "Nphi")
    call write_expr("geom%dr", "Lr/float(Nr)")
    
    !*** coordinate array allocation ***
    call FTR_ALLOC_1_ZERO(geom%rg,0,Nr)
    call FTR_ALLOC_1_ZERO(geom%thetag,0,Ntheta)
    call FTR_ALLOC_1_ZERO(geom%phig,0,Nphi)
    call FTR_ALLOC_1_ZERO(geom%cos_theta,0,Ntheta)
    call FTR_ALLOC_1_ZERO(geom%sin_theta,0,Ntheta)
            
    if (.not.memory_test) then
      !*** coordinate array initialisation ***    
      do i1 = 0,Nr
        geom%rg(i1) = geom%rmin + i1*geom%dr
      enddo      

      do i2 = 0,Ntheta
        geom%thetag(i2) = geom%thetamin + i2*geom%dtheta
      enddo
      do i3 = 0,Nphi
        geom%phig(i3) = geom%phimin + i3*geom%dphi
      enddo

      !*** initialization of the array for cosinus and sinus ***
      do i2 = 0,Ntheta
        geom%cos_theta(i2) = cos(geom%thetag(i2))
        geom%sin_theta(i2) = sin(geom%thetag(i2))
      end do
    end if

    !*** initialization of the coefficients for the ***
    !*** integral computation in r, theta, phi      ***
    call compute_coeff_intspace(geom)
  end subroutine new_space_coord
  

  !-------------------------------------------------------
  ! initialisation of the velocity coordinates 
  !   -> parallel velocity
  !   -> perpendicular velocity
  !-------------------------------------------------------     
  subroutine new_velocity_coord(geom,vparmin,mumin, &
    Lvpar,Lmu,Nvpar,Nmu)
    use globals, only : memory_test, nb_vth0, integral_vperp
    use utils_module, only : locate
    type(geometry), intent(inout) :: geom
    real(RKIND)   , intent(in)    :: vparmin, mumin  
    real(RKIND)   , intent(in)    :: Lvpar, Lmu      
    integer       , intent(in)      :: Nvpar, Nmu      

    real(RKIND) :: dmu
    real(RKIND) :: vperp, dvperp, vperp_min, vperp_max
    integer     :: i4, i5
    character(LEN=50), &
      parameter :: subp = " new_velocity_coord "//char(0)
    
    !*** variables initialization ***
    !-> minimal values
    geom%vparmin = vparmin
    geom%mumin   = mumin
    !-> length
    geom%Lvpar   = Lvpar
    geom%Lmu     = Lmu
    !-> number of points
    geom%Nvpar   = Nvpar
    geom%Nmu     = Nmu
    !-> grid step size
    geom%dvpar   = Lvpar/float(Nvpar)
    
    call write_expr("geom%Nvpar", "Nvpar")
    call write_expr("geom%Nmu", "Nmu")
    
	!*** coordinate array allocation ***
    call FTR_ALLOC_1_ZERO(geom%vparg,0,Nvpar)
    call FTR_ALLOC_1_ZERO(geom%mug,0,Nmu)
         
    if (.not.memory_test) then
      !*** parallel velocity array initialization ***  
      do i4 = 0,Nvpar
        geom%vparg(i4) = vparmin+i4*geom%dvpar
      enddo
      if (Nmu.ne.0) then
        if (integral_vperp) then
          vperp_min = sqrt(2._RKIND*mumin)
          vperp_max = sqrt(2._RKIND*(Lmu+mumin))
          dvperp    = (vperp_max-vperp_min)/float(Nmu)
          do i5 = 0,Nmu
            vperp         = vperp_min+i5*dvperp
            geom%mug(i5) = vperp*vperp/2._RKIND
          end do
        else
          dmu = geom%Lmu/float(geom%Nmu)
          do i5 = 0,Nmu
            geom%mug(i5) = geom%mumin+i5*dmu
          enddo
        end if
      else
        if (integral_vperp) then
          vperp_min    = sqrt(2._RKIND*mumin)
          geom%mug(0) = vperp_min*vperp_min/2._RKIND
        else
          geom%mug(0) = geom%mumin
        end if
      end if
      !*** find the index ivpar0 corresponding to vpar = 0 ***
      call locate(0._RKIND,geom%vparg,geom%Nvpar, &
        geom%dvpar,subp,geom%ivpar0)
    end if

    !*** initialization of the coefficients for the ***
    !*** integral computation in vpar and mu        ***
    call compute_coeff_intvelocity(geom)

    !*** power for the integral computation ***
    ipow = 1
    if ((geom%Nmu.eq.0).and.(geom%Lmu.eq.0)) ipow = 0
  end subroutine new_velocity_coord

  
  !-----------------------------------------------------------
  ! Computation of the coefficients required for a
  !  numerical integration in the three spatial directions 
  !  (r,theta,phi), using one of the following methods :
  !    1) trapeze method
  !    2) Simpson method
  !    3) Villarceau-Boole method
  !    4) Hardy method
  !  and storage in the arrays "coeff_intdr", "coeff_intdtheta"
  !   and "coeff_intdphi"
  !------------------------------------------------------------
  subroutine compute_coeff_intspace(geom)
    use globals, only : memory_test, integration_scheme
    use utils_module
    type(geometry), intent(inout) :: geom

    !*** array allocation ***
    call FTR_ALLOC_1_ZERO(geom%coeff_intdr,0,geom%Nr)
    call FTR_ALLOC_1_ZERO(geom%coeff_intdtheta,0,geom%Ntheta)
    call FTR_ALLOC_1_ZERO(geom%coeff_intdphi,0,geom%Nphi)

    if (.not.memory_test) then
      select case (integration_scheme)
      case(0)
        call compute_rectangle_coeff(geom%Nr,geom%dr, &
          .false.,geom%coeff_intdr)
        call compute_rectangle_coeff(geom%Ntheta,geom%dtheta, &
          .true.,geom%coeff_intdtheta)
        call compute_rectangle_coeff(geom%Nphi,geom%dphi, &
          .true.,geom%coeff_intdphi)
      case(1)
        call compute_trapeze_coeff(geom%Nr,geom%dr, &
          .false.,geom%coeff_intdr)
        call compute_trapeze_coeff(geom%Ntheta,geom%dtheta, &
          .true.,geom%coeff_intdtheta)
        call compute_trapeze_coeff(geom%Nphi,geom%dphi, &
          .true.,geom%coeff_intdphi)
      case(2)
        call compute_simpson_coeff(geom%Nr,geom%dr, &
          .false.,geom%coeff_intdr)
        call compute_simpson_coeff(geom%Ntheta,geom%dtheta, &
          .true.,geom%coeff_intdtheta)
        call compute_simpson_coeff(geom%Nphi,geom%dphi, &
          .true.,geom%coeff_intdphi)
      case(3)
        call compute_villarceau_coeff(geom%Nr,geom%dr, &
          geom%coeff_intdr)
        call compute_villarceau_coeff(geom%Ntheta,geom%dtheta, &
          geom%coeff_intdtheta)
        call compute_villarceau_coeff(geom%Nphi,geom%dphi, &
          geom%coeff_intdphi)
      case(4)
        call compute_hardy_coeff(geom%Nr,geom%dr, &
          geom%coeff_intdr)
        call compute_hardy_coeff(geom%Ntheta,geom%dtheta, &
          geom%coeff_intdtheta)
        call compute_hardy_coeff(geom%Nphi,geom%dphi, &
          geom%coeff_intdphi)
      end select
    end if
  end subroutine compute_coeff_intspace


  !------------------------------------------------------------
  ! Computation of the coefficients required for a
  !  numerical integration in the both velocity 
  !  directions (vpar,mu) using one of the following methods :
  !    1) trapeze method
  !    2) Simpson method
  !    3) Villarceau-Boole method
  !    4) Hardy method
  !  and storage in the arrays "coeff_intdvpar", "coeff_intdmu"
  !------------------------------------------------------------
  subroutine compute_coeff_intvelocity(geom)
    use globals, only : memory_test, integral_vperp, &
      integration_scheme
    use utils_module
    type(geometry), intent(inout) :: geom

    real(RKIND), dimension(0:geom%Nmu) :: integr_coeff
    real(RKIND) :: dmu
    real(RKIND) :: vperp, dvperp, vperp_min, vperp_max
    integer     :: imu

    !*** array allocation ***
    call FTR_ALLOC_1_ZERO(geom%coeff_intdvpar,0,geom%Nvpar)
    call FTR_ALLOC_1_ZERO(geom%coeff_intdmu,0,geom%Nmu)

    if (.not.memory_test) then
      select case (integration_scheme)
      case(0)
        call compute_rectangle_coeff(geom%Nvpar,geom%dvpar, &
          .false.,geom%coeff_intdvpar)
      case(1)
        call compute_trapeze_coeff(geom%Nvpar,geom%dvpar, &
          .false.,geom%coeff_intdvpar)
      case(2)
        call compute_simpson_coeff(geom%Nvpar,geom%dvpar, &
          .false.,geom%coeff_intdvpar)
      case(3)
        call compute_villarceau_coeff(geom%Nvpar,geom%dvpar, &
          geom%coeff_intdvpar)
      case(4)
        call compute_hardy_coeff(geom%Nvpar,geom%dvpar, &
          geom%coeff_intdvpar)
      end select
      !*** weights for integral in mu ***
      if (geom%Nmu.eq.0) then
        geom%coeff_intdmu(0) = 1._RKIND
      else
        if (integral_vperp) then
          vperp_min = sqrt(2._RKIND*geom%mumin)
          vperp_max = sqrt(2._RKIND*(geom%Lmu+geom%mumin))
          dvperp    = (vperp_max-vperp_min)/float(geom%Nmu)
          if (integration_scheme.eq.1) then
            ! trapeze
            integr_coeff = 0._RKIND
            do imu = 0,geom%Nmu-1
              integr_coeff(imu)   = integr_coeff(imu)   + 1._RKIND
              integr_coeff(imu+1) = integr_coeff(imu+1) + 1._RKIND
            end do
            integr_coeff = dvperp*integr_coeff/2._RKIND
          else
            ! simpson
            integr_coeff(0)         = 1._RKIND
            integr_coeff(geom%Nmu) = 1._RKIND
            do imu = 0,geom%Nmu-1
              if (mod(imu,2).eq.0) then
                integr_coeff(imu) = 2._RKIND
              else
                integr_coeff(imu) = 4._RKIND
              end if
            end do
            integr_coeff = dvperp*integr_coeff/3._RKIND
          end if
          do imu = 0,geom%Nmu
            vperp                   = sqrt(2._RKIND*geom%mug(imu))
            geom%coeff_intdmu(imu) = vperp*integr_coeff(imu)
          end do
        else
          dmu = geom%Lmu/float(geom%Nmu)
          call compute_simpson_coeff(geom%Nmu,dmu, &
            .false.,geom%coeff_intdmu)
        end if
      end if
    end if
  end subroutine compute_coeff_intvelocity


  !----------------------------------------------------- 
  ! geometry destructor
  !-----------------------------------------------------   
  subroutine del_geometry(geom)
    type(geometry), intent(inout) :: geom

    ! -> mesh grid
    call FTR_DEALLOC(geom%rg)
    call FTR_DEALLOC(geom%thetag)
    call FTR_DEALLOC(geom%phig)
    call FTR_DEALLOC(geom%vparg)
    call FTR_DEALLOC(geom%mug)
    ! -> integration coefficients
    call FTR_DEALLOC(geom%coeff_intdr)
    call FTR_DEALLOC(geom%coeff_intdtheta)
    call FTR_DEALLOC(geom%coeff_intdphi)
    call FTR_DEALLOC(geom%coeff_intdvpar)
    call FTR_DEALLOC(geom%coeff_intdmu)
    ! -> cos(theta) and sin(theta)
    call FTR_DEALLOC(geom%cos_theta)
    call FTR_DEALLOC(geom%sin_theta)
  end subroutine del_geometry

end module geometry_class
