!**************************************************************
!  Copyright Euratom-CEA
!  Authors : 
!     Virginie Grandgirard (virginie.grandgirard@cea.fr)
!     Chantal Passeron (chantal.passeron@cea.fr)
!     Guillaume Latu (guillaume.latu@cea.fr)
!     Xavier Garbet (xavier.garbet@cea.fr)
!     Philippe Ghendrih (philippe.ghendrih@cea.fr)
!     Yanick Sarazin (yanick.sarazin@cea.fr)
!  
!  This code GYSELA (for GYrokinetic SEmi-LAgrangian) 
!  is a 5D gyrokinetic global full-f code for simulating 
!  the plasma turbulence in a tokamak.
!  
!  This software is governed by the CeCILL-B license 
!  under French law and abiding by the rules of distribution 
!  of free software.  You can  use, modify and redistribute 
!  the software under the terms of the CeCILL-B license as 
!  circulated by CEA, CNRS and INRIA at the following URL
!  "http://www.cecill.info". 
!**************************************************************

#include "config-F.inc"
#include "code_instrumentation/debug.inc"

!-------------------------------------------------------
! file : MPIutils.F90
! date : 20/03/2003
!  subroutines used for the MPI parallelization
! => Developped by G. LATU 
!---------------------------------------------
module MPIutils_module
  use prec_const
  use globals
  use wrap_alloc_module
  use mpiMod
  implicit none

  private
  public :: pp_reduce_scalar, pp_reduce_vector
  public :: pptranspose_backward, pptranspose_forward
  public :: pptranspvel_backward, pptranspvel_forward
  public :: distribute_Sdiag                    
  public :: ppinit, ppinit_comm, ppdeallocate
  public :: ppexit, ppbarrier, ppbarrier_timer
  public :: comm_irecv, comm_send
  public :: comm_wait_recv, comm_exchange_neighbours
  public :: comm_gather2D

  ! Index of the processors in the neiborhood : 
  !   North, South, West ...
  integer :: NN, SS, WW, EE, NW, NE, SW, SE
  ! MPI types to send and receive blocs of 
  !   the distribution function
  integer :: mpitype_WE, mpitype_NS, mpitype_corner
  ! MPI Receiving buffers
  real(RKIND), dimension (:,:), pointer, public :: rbufWW 
  real(RKIND), dimension (:,:), pointer, public :: rbufNN
  real(RKIND), dimension (:,:), pointer, public :: rbufSS
  real(RKIND), dimension (:,:), pointer, public :: rbufEE
  real(RKIND), dimension (:,:), pointer, public :: rbufNW 
  real(RKIND), dimension (:,:), pointer, public :: rbufNE
  real(RKIND), dimension (:,:), pointer, public :: rbufSW
  real(RKIND), dimension (:,:), pointer, public :: rbufSE
  ! MPI Sending buffers
  real(RKIND), dimension (:,:), pointer, public :: sbufWW 
  real(RKIND), dimension (:,:), pointer, public :: sbufNN
  real(RKIND), dimension (:,:), pointer, public :: sbufSS
  real(RKIND), dimension (:,:), pointer, public :: sbufEE
  real(RKIND), dimension (:,:), pointer, public :: sbufNW 
  real(RKIND), dimension (:,:), pointer, public :: sbufNE
  real(RKIND), dimension (:,:), pointer, public :: sbufSW
  real(RKIND), dimension (:,:), pointer, public :: sbufSE
  real(RKIND), dimension (:,:), pointer, public :: fillderiv
  ! request pending for isends
  integer, dimension(1:8) :: reqs
  ! request pending for irecvs
  integer, dimension(1:8) :: reqr
  ! status for irecvs
  integer, dimension(MPI_STATUS_SIZE,1:8) :: status
  ! for testing purpose
!baoter
  integer    , public :: mpitype_locbloc, mpitype_globbloc
  integer    , public :: mpitype_globbloc_rvpar
  real(RKIND), public, dimension(:,:,:)    , pointer :: tmprhs2
  real(RKIND), public, dimension(:,:,:,:)  , pointer :: f4D_transp
  real(RKIND), public, dimension(:,:,:,:)  , pointer :: f4D_send
  real(RKIND), public, dimension(:,:,:,:)  , pointer :: f4D_recv
  real(RKIND), public, dimension(:,:,:,:)  , pointer :: f4D_Sdiag_send
  real(RKIND), public, dimension(:,:,:,:)  , pointer :: f4D_Sdiag_recv
  real(RKIND), public, &
    dimension(:,:,:,:,:), pointer :: f5D_transp
  real(RKIND), public, dimension(:)        , pointer :: f1D_send
  real(RKIND), public, dimension(:)        , pointer :: f1D_recv
!eaoter
  ! used when the 3D data are distributed
  integer                  , save , public :: dom_locphi   = 1
  integer                  , save , public :: dom_mapr     = 1
  integer                  , save , public :: dom_mapr_loc = 1
  integer                  , save , public :: irstart      = 1
  integer                  , save , public :: ilocstart   
  integer                  , save , public :: ilargestart  
  integer                  , save , public :: ilargeend    
  integer, dimension(:)  , pointer, public :: mapir  
  integer, dimension(:,:,:), pointer, &
       public                              :: scal_map
  integer, dimension(:)  , pointer, public :: moments_locphi

  ! for moment parallelization
  !> first phi handled on this process
  integer                         , public :: iphistart
  !> size of the phi domain handled by each process
  integer                         , public :: dom_mapphi
  integer                         , public :: dom_bisphi
  integer                         , public :: collstart, collend
  !> process id of the process handling the given phi
  integer, dimension(:)  , pointer, public :: moments_mapphi
  integer, dimension(:)  , pointer, public :: moments_reqs
  integer, dimension(:)  , pointer, public :: moments_reqr
  integer, dimension(:)  , pointer, public :: bcast_reqs
  integer, dimension(:)  , pointer, public :: bcast_reqr
  integer, dimension(:,:), pointer, public :: moments_status
  integer                                  :: moments_status_size

  character*(MPI_MAX_PROCESSOR_NAME), dimension(:), pointer :: vhost
  real(RKIND), dimension(:), pointer :: vtime
  !******************************
  contains
  !******************************   
  
  subroutine get_iphistart(Nbproc_species, dom_mapphi, Nphi, pInSpecies_id, iphistart)
    integer, intent(in)  :: Nbproc_species, dom_mapphi, Nphi, pInSpecies_id
    integer, intent(out) :: iphistart

    integer :: iphi, iproc, i

    iphi      = 0
    iphistart = -1
    do iproc = 0, Nbproc_species-1
       do i = 1, dom_mapphi
          if (iphi .le. Nphi-1) then
             if ( (iphistart.eq.-1) .and. (iproc.eq.pInSpecies_id) ) then
                iphistart = iphi
             endif
             iphi = iphi + 1
          endif
       enddo
    enddo
  end subroutine get_iphistart


  subroutine get_ilocstart(Nbproc_loc, dom_locphi, Nphi, plocal_id, ilocstart)
    integer, intent(in)  :: Nbproc_loc, dom_locphi, Nphi, plocal_id
    integer, intent(out) :: ilocstart

    integer :: iphi, iproc, i

    iphi      = 0
    ilocstart = -1
    do iproc = 0,Nbproc_loc-1
       do i = 1,dom_locphi
          if (iphi .le. Nphi-1) then
             if ( (ilocstart.eq.-1) .and. (iproc.eq.plocal_id) ) then
                ilocstart = iphi
             endif
             iphi = iphi + 1
          endif
       enddo
    enddo
  end subroutine get_ilocstart

  subroutine get_moments_mapphi(Nbproc_species, dom_mapphi, Nphi, moments_mapphi)
    integer,intent(in)  :: Nbproc_species, dom_mapphi, Nphi
    integer,dimension(0:Nphi-1),intent(out):: moments_mapphi

    integer iphi, iproc, i

    iphi      = 0
    do iproc = 0, Nbproc_species-1
       do i = 1, dom_mapphi
          if (iphi .le. Nphi-1) then
             moments_mapphi(iphi) = iproc
             iphi = iphi + 1
          endif
       enddo
    enddo
  end subroutine get_moments_mapphi

  subroutine get_ilargestart(Nphi, Nmuscal, moments_mapphi, pInSpecies_id, ilargestart)
    integer,               intent(in)  :: Nphi, Nmuscal, pInSpecies_id
    integer,dimension(0:Nphi-1),intent(in):: moments_mapphi
    integer,               intent(out) :: ilargestart

    integer :: iphi, i
	
    ilargestart = -1
    do iphi = 0, Nphi-1
       do i = 0, Nmuscal-1
          if (Nmuscal*moments_mapphi(iphi)+i .eq. pInSpecies_id) then
             if (ilargestart .eq. -1) ilargestart = iphi 
          end if
       end do
    end do
  end subroutine get_ilargestart

  subroutine get_ilargeend(Nphi, Nmuscal, moments_mapphi, pInSpecies_id, ilargeend)
    integer,               intent(in)  :: Nphi, Nmuscal, pInSpecies_id
    integer,dimension(0:Nphi-1),intent(in):: moments_mapphi
    integer,               intent(out) :: ilargeend

    integer :: iphi, i

    ilargeend = -1
    do iphi = 0, Nphi-1
       do i = 0, Nmuscal-1
          if (Nmuscal*moments_mapphi(iphi)+i .eq. pInSpecies_id) then
             ilargeend = iphi
          end if
       end do
    end do
  end subroutine get_ilargeend

  subroutine get_nbproc_phi(inbproc_loc, Nbproc_loc, Nphi, bloc_phi, Nbproc_phi)
    integer, intent(in)  :: inbproc_loc, Nbproc_loc, Nphi, bloc_phi
    integer, intent(out) :: Nbproc_phi

    integer :: i, j, k, dom_phi

    do k = 1, MAX(4*inbproc_loc,Nbproc_loc)
       j = Nbproc_loc / k
       i = j * k
       dom_phi  = Nphi/k
       if ((i .eq. Nbproc_loc) .and. (mod(dom_phi,bloc_phi) .eq. 0) .and. (Nphi.ge.k)) then
          Nbproc_phi  = k
       end if
    end do
  end subroutine get_nbproc_phi

  subroutine get_nbproc_vpar(inbproc_loc, Nbproc_loc, Nphi, bloc_phi, Nbproc_vpar)
    integer, intent(in)  :: inbproc_loc, Nbproc_loc, Nphi, bloc_phi
    integer, intent(out) :: Nbproc_vpar

    integer :: i, j, k, dom_phi

    do k = 1, MAX(4*inbproc_loc,Nbproc_loc)
       j = Nbproc_loc / k
       i = j * k
       dom_phi  = Nphi/k
       if ((i .eq. Nbproc_loc) .and. (mod(dom_phi,bloc_phi) .eq. 0) .and. (Nphi.ge.k)) then
          Nbproc_vpar = j
       end if
    end do
  end subroutine get_nbproc_vpar

  subroutine pp_reduce_scalar(mytime,tmin,idmin,hostmin,tmax,idmax,hostmax)
    real(RKIND) , intent(in)   :: mytime
    real(RKIND) , intent(out)  :: tmin, tmax
    integer     , intent(out)  :: idmin, idmax
    character(*), intent(out)  :: hostmin, hostmax
    real(RKIND) :: avgtime
    integer     :: idx, ierr
    call MPI_Gather(mytime, 1, MPI_REAL8, vtime, &
      1, MPI_REAL8, 0, MPI_COMM_MU, ierr);
    avgtime = SUM(vtime(1:Nbproc_loc))/Nbproc_loc
    if (plocal_id .eq. 0) then
      idmin  = minloc(vtime,1)
      tmin = vtime(idmin)
      hostmin = vhost(idmin)
      idmax  = maxloc(vtime,1)
      tmax = vtime(idmax)
      hostmax = vhost(idmax)
    end if
  end subroutine pp_reduce_scalar

  subroutine pp_reduce_vector(mytime,mysize,tmin,tmax,tavg)
    real(RKIND) , dimension(:), intent(in)   :: mytime
    integer                   , intent(in)   :: mysize
    real(RKIND) , dimension(:), intent(out)  :: tmin, tmax, tavg

    real(RKIND) , dimension(:), allocatable  :: gathtime
    integer i, k, ierr
    allocate(gathtime(mysize*Nbproc_loc))
    call MPI_Gather(mytime, mysize, MPI_REAL8, gathtime, &
      mysize, MPI_REAL8, 0, MPI_COMM_MU, ierr);
    if (plocal_id .eq. 0) then
       do k = 1, mysize
          tmin(k) = gathtime(k)
          tmax(k) = gathtime(k)          
          tavg(k) = gathtime(k)          
       end do
       do i = 1, Nbproc_loc-1
          do k = 1, mysize
             tmin(k) = MIN(tmin(k),gathtime(i*mysize+k))
             tmax(k) = MAX(tmax(k),gathtime(i*mysize+k)) 
             tavg(k) = tavg(k) + gathtime(i*mysize+k)
          end do
       end do
       do k = 1, mysize
          tavg(k) = tavg(k)/Nbproc_loc
       end do
    else
       tmin(:)=-1
       tmax(:)=-1
       tavg(:)=-1
    end if
    deallocate(gathtime)
  end subroutine pp_reduce_vector

  !-----------------------------------------------------
  ! global parallel initialization 
  !-----------------------------------------------------
  subroutine ppinit
    use globals, only : Nbproc_tot, pglobal_id, diag_para, &
      nbrp, diag_targ

    integer :: i, ierr, obt

#ifdef MPI2
    call MPI_INIT_THREAD(MPI_THREAD_MULTIPLE, obt, ierr)
#else
    call MPI_INIT(ierr)
#endif
#ifdef DETERMINISTIC_MPI
    call DET_MPI_Init(MPI_IN_PLACE)
#endif
    call MPI_COMM_SIZE(MPI_COMM_WORLD,Nbproc_tot,ierr)
    call MPI_COMM_RANK(MPI_COMM_WORLD,pglobal_id,ierr)

    !*** define if the diagnostics must be done parallely or ***
    !***  sequentially                                       ***
    diag_para = (Nbproc_tot .ge. 8*MAXNSPECIES) 

    !*** Defines a set of processor targets used ***
    !*** for diagnostic computation and savings  ***
    do i = 1,nbrp
      diag_targ(i) = cible
    end do
    if (diag_para) then
      outputproc = cible
      if (pglobal_id.eq.outputproc) then
        print*, ' '
        print*, '---> Physical computation : PARALLEL '
        print*, ' '
      end if

      diag_targ(1) = 1
      diag_targ(2) = 2
      diag_targ(3) = 3
      diag_targ(4) = 4
      !***  HDF5_f2D_saving                              ***
      !***  performed by the proc. diag_targ(5) = cible  ***
      diag_targ(5) = cible
      diag_targ(6) = 5
    else
      outputproc = 0
      if (pglobal_id.eq.outputproc) then
        print*, ' '
        print*, '---> Physical computation : SEQUENTIAL '
        print*, ' '
      end if
    end if
  end subroutine ppinit
  

  !-----------------------------------------------------
  ! initialization of the communicators
  !-----------------------------------------------------
  subroutine ppinit_comm
    use globals, only : memory_test
!R3 #include "code_instrumentation/r3_info.h" !R3
    integer           :: base, dep, size, manage_phivalue
    integer           :: stride
    character(LEN=11) :: string
    character(LEN=80) :: message
    !-> variables for moment parallelisation
    integer           :: i, iphi, iproc, Nbt, ierr, j, k
    integer           :: inbproc_loc, scal1, scal2
    integer           :: loc_mustart, loc_muend
    integer           :: Nmutg, resultlen
    real*8            :: ratio
    character (MPI_MAX_PROCESSOR_NAME) :: name

!R3 call r3_info_begin (r3_info_index_1, 'Init_comm') !R3
    call Write_dyn_balise_open('Init_comm') !R4

#ifdef _OPENMP
    call omp_set_dynamic(.false.)
    call omp_set_num_threads(Nbthread)
#else 
    Nbthread = 1
#endif

    call write_param("Nbthread", Nbthread)

    !*** partitioning in r and theta directions ***
    dom_r     = (Nr+1)/Nbproc_r
    dom_theta = Ntheta/Nbproc_theta

    call write_param("Nr", Nr)
    call write_param("Nbproc_r", Nbproc_r)
    call write_param("Ntheta", Ntheta)
    call write_param("Nbproc_theta", Nbproc_theta)
    call write_expr("dom_r", "(Nr+1)/Nbproc_r")
    call write_expr("dom_theta", "Ntheta/Nbproc_theta")
    
    !-> the size of the surrouding area must be lower or equal
    !   than the domain size (in r direction or theta direction)
    if ((dom_r < bufsize) .or. (dom_theta < bufsize)) then 
      print *, "(dom_theta or dom_r) < bufsize"
      call ppexit
      stop
    endif

    if ((dom_r < stencil+2) .or. (dom_theta < stencil+2)) then 
      print *, "(dom_theta or dom_r) < points needed", &
        " for the derivative"
      call ppexit
      stop
    endif

    !*** partitioning in phi and vpar
!baoter (temporary used for fluid moment computation even if 
!         transpose4D=.false.)
!VG!    if (transpose4D) then
    Nbproc_loc = Nbproc_r*Nbproc_theta
    inbproc_loc = sqrt(1.*Nbproc_loc)
    Nbproc_phi  = 1
    Nbproc_vpar = Nbproc_loc

    call get_nbproc_phi(inbproc_loc, Nbproc_loc, Nphi, bloc_phi, Nbproc_phi)
    call get_nbproc_vpar(inbproc_loc, Nbproc_loc, Nphi, bloc_phi, Nbproc_vpar)

    call write_expr("Nbproc_loc", "Nbproc_r*Nbproc_theta")
    call write_expr("inbproc_loc", "sqrt(1.*Nbproc_loc)")
    call write_param("Nphi", Nphi)
    call write_param("bloc_phi", bloc_phi)
    call write_expr("Nbproc_phi", "get_nbproc_phi(inbproc_loc, Nbproc_loc, Nphi, bloc_phi)")
    call write_expr("Nbproc_vpar", "get_nbproc_vpar(inbproc_loc, Nbproc_loc, Nphi, bloc_phi)")
    !OLDVERSION    Nbproc_phi  = MIN(Nphi/bloc_phi, Nbproc_r*Nbproc_theta)
    !OLDVERSION    Nbproc_vpar = (Nbproc_r*Nbproc_theta) / Nbproc_phi
    if (pglobal_id .eq. 0) then
       print *, "Nbprocs phi vpar ", Nbproc_phi, Nbproc_vpar
    endif
    if ( (Nbproc_r*Nbproc_theta) .ne. &
         (Nbproc_phi*Nbproc_vpar)) then
       print *, '(Nbproc_r*Nbproc_theta) .ne. ', &
            '(Nbproc_phi*Nbproc_vpar)'
       print *, 'Nproc_phi = ', Nbproc_phi, &
            'Nproc_vpar = ', Nbproc_vpar
       call ppexit
       stop
    end if
    dom_phi  = Nphi/Nbproc_phi
    call write_expr("dom_phi", "Nphi/Nbproc_phi")
    if ( mod(dom_phi,bloc_phi) .ne. 0 ) then
       print *, " mod(dom_phi,bloc_phi) .ne. 0 "
       print *, "dom_phi = ", dom_phi, "bloc_phi = ", bloc_phi
       call ppexit
       stop
    end if
    dom_vpar = ceiling((Nvpar+1)/float(Nbproc_vpar))

    call write_param("Nvpar", Nvpar)
    call write_expr("dom_vpar", "ceiling((Nvpar+1)/float(Nbproc_vpar))")
    !VG!    end if
!eaoter

    !*** partitioning in function     ***
    !*** of the number of species     ***
    pSpecies_id = floor(Nspecies*pglobal_id/float(Nbproc_tot))

    call MPI_COMM_SPLIT(MPI_COMM_WORLD, pSpecies_id, &
      pglobal_id,mpi_comm_species,ierr)
    call MPI_COMM_SIZE(mpi_comm_species,Nbproc_species,ierr)
    call MPI_COMM_RANK(mpi_comm_species,pInSpecies_id,ierr)

    call MPI_COMM_SPLIT(MPI_COMM_WORLD,pInSpecies_id, &
      pSpecies_id,MPI_COMM_CROSS_SPECIES,ierr)

    stride  = Nbproc_species / Nbproc_mu
    call write_param("Nbproc_mu", Nbproc_mu)
    call write_expr("Nbproc_species", "(Nbproc_r-1+1)*(Nbproc_theta-1+1)*(Nbproc_mu-1+1)")
    mu_id   = pInSpecies_id / stride

    if (pInSpecies_id .lt. Nphi) then 
      manage_phivalue = 1
    else
      manage_phivalue = 0
    endif
    !*** Split the global group of processors ***
    !-> communicator with all the processors managing a phi 
    !-> value in the moments computation
    call MPI_COMM_SPLIT(mpi_comm_species,manage_phivalue, &
      pInSpecies_id,mpi_comm_moments,ierr)

    !*** Split the global group of processors into 'Nbproc_mu' ***
    !-> communicator with all the processors having the same mu
    call MPI_COMM_SPLIT(mpi_comm_species,mu_id,pInSpecies_id, &
      mpi_comm_mu,ierr)
    !-> get the processor plocal_identity into the new 
    !    mpi_comm_mu communicator
    call MPI_COMM_SIZE(mpi_comm_mu,Nbproc_loc,ierr)
    call MPI_COMM_RANK(mpi_comm_mu,plocal_id,ierr)
    !-> communicator with all the processors having 
    !    the same plocal_id
    call MPI_COMM_SPLIT(mpi_comm_species,plocal_id,mu_id, &
      mpi_comm_intermu,ierr)
    Nbpoints_muid = (Nr+1)*(Ntheta+1)*(Nphi+1)*(Nvpar+1)

    !*** Definition of the blocks on each processors ***
    !-> local coordinates
    base       = (plocal_id/Nbproc_r)
    dep        = mod(plocal_id,Nbproc_r)              
    call write_param("plocal_id", plocal_id)
    call write_expr("base", "(plocal_id/Nbproc_phi)")
    call write_expr("dep", "mod(plocal_id,Nbproc_phi)")
    istart     = dep *  dom_r
    iend       = istart + dom_r - 1
    inumber    = iend - istart + 1
    jstart     = base * dom_theta
    jend       = jstart + dom_theta - 1
    jnumber    = jend - jstart + 1
    istart_buf = istart - bufsize
    iend_buf   = iend + bufsize
    jstart_buf = jstart - bufsize 
    jend_buf   = jend + bufsize

    call write_param("bufsize", bufsize)
    call write_expr("istart"    , "dep *  dom_r")
    call write_expr("iend"      , "istart + dom_r - 1")
    call write_expr("jstart"    , "base * dom_theta")
    call write_expr("jend"      , "jstart + dom_theta - 1")
    call write_expr("istart_buf", "istart - bufsize")
    call write_expr("iend_buf"  , "iend + bufsize")
    call write_expr("jstart_buf", "jstart - bufsize")
    call write_expr("jend_buf"  , "jend + bufsize")

!baoter (temporary used for fluid moment computation
!        even if transpose4D=false)
!VG!    if (transpose4D) then
    base       = (plocal_id/Nbproc_phi)
    dep        = mod(plocal_id,Nbproc_phi)
    kstart     = dep *  dom_phi
    call write_expr("kstart", "dep*dom_phi")
    kend       = kstart + dom_phi - 1
    call write_expr("kend", "kstart+dom_phi-1")
    knumber    = kend - kstart + 1
    call write_expr("knumber", "kend - kstart + 1")
    lstart     = (base * (Nvpar+1)) / Nbproc_vpar
    call write_expr("lstart", "(base*(Nvpar+1))/Nbproc_vpar")
    lend       = ((base+1) * (Nvpar+1)) / Nbproc_vpar - 1
    call write_expr("lend", "((base+1)*(Nvpar+1))/Nbproc_vpar-1")
    lnumber    = lend - lstart + 1
    call write_param("Nmu", Nmu)

    select case ( transpose_version )
    case ( TRANSPOSE_SENDIRECV )
       ! f4D_send and f4D_recv are dynamically allocated from now on
    case ( TRANSPOSE_ALLTOALLW )
       ! nothing specific to allocate
    case default
       print *, pglobal_id, 'Invalid value for transpose_version', &
            transpose_version
       call MPI_abort(MPI_COMM_WORLD, 1, ierr)
    end select
!DEB_PRIV_SPECTR
    if (Sp_saving) then
       call FTR_ALLOC_4_ZERO(f4D_Sdiag_send,0,dom_theta-1,0,Nphi,0,Nvpar,0,dom_r-1)
       call FTR_ALLOC_4_ZERO(f4D_Sdiag_recv,0,dom_theta-1,0,Nphi,0,Nvpar,0,dom_r-1)
    end if
!END_PRIV_SPECTR

!VG!    end if
!eaoter
    
    if (transp_velocity) then
       dom_bisphi = (Nphi / (Nmu+1))
       if ((Nmu+1)*dom_bisphi .ne. Nphi) dom_bisphi = dom_bisphi+1
       collstart = min(mu_id * dom_bisphi, Nphi-1)
       collend   = min(collstart + dom_bisphi-1, Nphi-1)
       call FTR_ALLOC_5_ZERO(f5D_transp,istart,iend,jstart,jend,collstart,collend,0,Nvpar,0,Nmu)
       call FTR_ALLOC_1_ZERO(f1D_send,0,dom_r*dom_theta*(Nvpar+1)*dom_bisphi-1)
       call FTR_ALLOC_1_ZERO(f1D_recv,0,dom_r*dom_theta*(Nvpar+1)*dom_bisphi-1)
    end if

    !-> communicator with all the processors having the same 
    !   mu_id and the same jstart
    call MPI_COMM_SPLIT(mpi_comm_mu, jstart, plocal_id, &
      mpi_comm_column, ierr)
    call MPI_COMM_RANK(mpi_comm_column, jstart_id,ierr)

    !-> communicator with all the processors having the same
    !   mu_id and the same istart
    call MPI_COMM_SPLIT(mpi_comm_mu, istart, plocal_id, &
      mpi_comm_row, ierr)
    call MPI_COMM_RANK(mpi_comm_row, istart_id,ierr)

    !*** Define the local identity of the eight neighbours of ***
    !***  the local processor 'plocal_id'                     ***
    WW   = mod(plocal_id+Nbproc_loc-Nbproc_r, Nbproc_loc) ! West
    EE   = mod(plocal_id+Nbproc_r, Nbproc_loc)            ! East
    base = (plocal_id/Nbproc_r)*Nbproc_r           
    dep  = mod(plocal_id+1,Nbproc_r)              
    SS   = mod(base+dep, Nbproc_loc)                  ! South
    dep  = mod(Nbproc_r+plocal_id-1,Nbproc_r)             
    NN   = mod(base+dep, Nbproc_loc)                  ! North 
    NW   = mod(NN+Nbproc_loc-Nbproc_r, Nbproc_loc)    ! North-West
    NE   = mod(NN+Nbproc_r, Nbproc_loc)               ! North-East
    SW   = mod(SS+Nbproc_loc-Nbproc_r, Nbproc_loc)    ! South-West
    SE   = mod(SS+Nbproc_r, Nbproc_loc)               ! South-East

    !*** Define three MPI types needed to transmit blocks ***
    !***   of data to processors in the neighborhood      ***
    !***     mpitype_WE     : block to send to processors ***
    !***                      on the West or East         ***
    !***     mpitype_NS     : block to send to processors ***
    !***                      on the North or South       ***
    !***     mpitype_corner : block to send to processors ***
    !***                      on NE, NW, SW, SE           ***
    !-> mpitype_WE
    if (transpose4D) then
      stride = (3 * (Nr+1) + 6) 
      call write_expr("stride", "3 * (Nr+1) + 6")
    else
      stride = (3 * dom_r + 6) 
      call write_expr("stride", "3 * dom_r + 6")
    end if
    size = stride * 3*bloc_phi
    call mpi_type_vector(size, 1, 1, MPI_REAL8, mpitype_WE, ierr)
    call mpi_type_commit(mpitype_WE, ierr)
    call FTR_ALLOC_2_ZERO(rbufWW,0,stride-1,0,3*bloc_phi-1)
    call FTR_ALLOC_2_ZERO(rbufEE,0,stride-1,0,3*bloc_phi-1)
    call FTR_ALLOC_2_ZERO(sbufWW,0,stride-1,0,3*bloc_phi-1)
    call FTR_ALLOC_2_ZERO(sbufEE,0,stride-1,0,3*bloc_phi-1)
    !-> mpitype_NS
    if (transpose4D) then
      stride = (3 * Ntheta + 6) 
      call write_expr("stride", "3 * Ntheta + 6")
    else
      stride = (3 * dom_theta + 6) 
      call write_expr("stride", "3 * dom_theta + 6")
    end if
    size   = stride * 3*bloc_phi
    call mpi_type_vector(size, 1, 1, MPI_REAL8, mpitype_NS, ierr)
    call mpi_type_commit(mpitype_NS, ierr)
    call FTR_ALLOC_2_ZERO(rbufNN,0,stride-1,0,3*bloc_phi-1)
    call FTR_ALLOC_2_ZERO(rbufSS,0,stride-1,0,3*bloc_phi-1)
    call FTR_ALLOC_2_ZERO(sbufNN,0,stride-1,0,3*bloc_phi-1)
    call FTR_ALLOC_2_ZERO(sbufSS,0,stride-1,0,3*bloc_phi-1)
    !-> mpitype_corner
    stride = 9
    call write_param("stride", stride)
    size   = stride * 3*bloc_phi
    call mpi_type_vector(size, 1, 1, MPI_REAL8, &
      mpitype_corner, ierr)
    call mpi_type_commit(mpitype_corner, ierr)
    call FTR_ALLOC_2_ZERO(rbufNW,0,stride-1,0,3*bloc_phi-1)
    call FTR_ALLOC_2_ZERO(rbufSW,0,stride-1,0,3*bloc_phi-1)
    call FTR_ALLOC_2_ZERO(rbufNE,0,stride-1,0,3*bloc_phi-1)
    call FTR_ALLOC_2_ZERO(rbufSE,0,stride-1,0,3*bloc_phi-1)
    call FTR_ALLOC_2_ZERO(sbufNW,0,stride-1,0,3*bloc_phi-1)
    call FTR_ALLOC_2_ZERO(sbufSW,0,stride-1,0,3*bloc_phi-1)
    call FTR_ALLOC_2_ZERO(sbufNE,0,stride-1,0,3*bloc_phi-1)
    call FTR_ALLOC_2_ZERO(sbufSE,0,stride-1,0,3*bloc_phi-1)
    call FTR_ALLOC_2_ZERO(fillderiv,1,12,0,3*bloc_phi-1)

!baoter    
    stride = Nr+1 ! Because it starts at 0 and ends at Nr
    call MPI_TYPE_VECTOR( dom_theta, dom_r, stride,&
      MPI_REAL8, mpitype_locbloc, ierr )
    call MPI_TYPE_COMMIT( mpitype_locbloc, ierr )
    
    stride = dom_r 
    call MPI_TYPE_VECTOR( dom_theta, dom_r, stride,&
      MPI_REAL8, mpitype_globbloc, ierr )
    call MPI_TYPE_COMMIT( mpitype_globbloc, ierr )

    stride = dom_r 
    call MPI_TYPE_VECTOR( (Nvpar+1), dom_r, stride,&
      MPI_REAL8, mpitype_globbloc_rvpar, ierr )
    call MPI_TYPE_COMMIT( mpitype_globbloc_rvpar, ierr )
!eaoter

    !*****************************************************
    ! FOR MOMENTS
    !*****************************************************
    allocate(moments_mapphi(0:Nphi-1))
    allocate(moments_locphi(0:Nphi-1))
    dom_mapphi = ceiling(1.0*Nphi/Nbproc_species)
    call write_expr("dom_mapphi", "ceiling(1.0*Nphi/Nbproc_species)")
    if ((dom_mapphi.gt.1).and.(dom_mapphi*Nbproc_species.ne.Nphi)) &
      then
      print*, 'dom_mapphi = ',dom_mapphi
      print*, 'dom_mapphi*Nbproc_tot = ',dom_mapphi*Nbproc_species
      print *, "condition not checked : dom_mapphi*Nbproc_tot .ne. Nphi"
      STOP 
    endif
    dom_locphi = ceiling(1.0*Nphi/Nbproc_loc)
    call write_expr("dom_locphi" , "ceiling(1.0*Nphi/Nbproc_loc)")
    if ((dom_locphi.gt.1).and.(dom_locphi*Nbproc_loc.ne.Nphi)) &
      then
      print*, 'dom_locphi = ',dom_locphi
      print*, 'dom_locphi*Nbproc_loc = ',dom_locphi*Nbproc_loc
      print *,"condition not checked : dom_locphi*Nbproc_loc .ne. Nphi"
      STOP
    endif

    !--> domain decomposition in phi direction 
    !-->  for poisson parallelisation and 
    !-->  diagnostics parallelisation
    iphistart = -1
    call get_iphistart(Nbproc_species, dom_mapphi, Nphi, pInSpecies_id, iphistart)
    call get_moments_mapphi(Nbproc_species, dom_mapphi, Nphi, moments_mapphi)
	
    call write_array("moments_mapphi", "integer;0;Nphi-1;get_moments_mapphi(Nbproc_species, dom_mapphi, Nphi)")
	! Useful to the compute of iphistart
    call write_param("pInSpecies_id", pInSpecies_id)
    call write_expr("iphistart", "get_iphistart(Nbproc_species, dom_mapphi, Nphi, pInSpecies_id)")
	
    !--> domain decomposition in phi direction 
    !-->  inside 'mpi_comm_mu' communicator used
    !-->  in poisson parallelisation on large HPC platform 
    iphi      = 0
    do iproc = 0,Nbproc_loc-1
      do i = 1,dom_locphi
        if (iphi .le. Nphi-1) then
          moments_locphi(iphi) = iproc
          iphi = iphi + 1
        endif
      enddo
    enddo
    
    call get_ilocstart(Nbproc_loc, dom_locphi, Nphi, plocal_id, ilocstart)
    call write_expr("ilocstart", "get_ilocstart(Nbproc_loc, dom_locphi, Nphi, plocal_id)")
	
    !--> domain decomposition of the integrals in mu direction,
    !--> useful for the rho computation of the poisson solver
    if (large_platform) then
      if (mod(Nbthread,2) .eq. 1 .and. Nbthread .ne. 1) then
        print*, 'This number of threads is ', & 
          ' not supported (should be an even number) ', Nbthread
        STOP "condition not checked"
      end if
      ! Nmutg = Power of two just above Nmu
      Nmutg = 2**ceiling(log(dfloat(Nmu+1))/log(2._8))
      call write_expr("Nmutg", "2**ceiling(log(dfloat(Nmu+1))/log(2.0))")
      if (Nbproc_species .gt. Nphi) then
        ! ratio = Maximum number of cores that one can 
        !         allocate to compute one iphi value
        ratio = floor(dfloat(Nbproc_species)/dfloat(Nphi))
        call write_expr("ratio", "floor(dfloat(Nbproc_species)/dfloat(Nphi))")
        ! scal1 = Power of two just below ratio variable.
        !         This restriction improve the numerical accuracy
        !         as far as parallelism is employed.
        scal1 = 2**floor(log(ratio)/log(2._8))!nint(ratio))!
        call write_expr("scal1", "2**floor(log(ratio)/log(2.0))")
      else
        scal1 = 1
        call write_param("scal1", scal1)
      end if
      ! scal2 = Maximal number of cores achievable for the 
      !         finest computation grain we know
      scal2 = MAX(1,2*Nmutg/Nbthread)
      call write_expr("scal2", "MAX(1,2*Nmutg/Nbthread)")
      ! Nmuscal integrates the two constraints of scal1 and scal2
      ! targeting the use of the maximal number of cores
      Nmuscal = MIN(scal1,scal2)
      call write_expr("Nmuscal", "MIN(scal1,scal2)")
      ! Compute Nmuloc, the number of Nmu values that one process
      ! will be in charge of. A power of two is valuable because
      ! it guarantees that numerical results of poisson solver
      ! is extremely robust.
      if (Nmuscal .gt. 1) then
        Nmuloc = ceiling(dfloat(Nmutg)/dfloat(Nmuscal))
        call write_expr("Nmuloc", "ceiling(dfloat(Nmutg)/dfloat(Nmuscal))")
        !!TO AVOID   Nmuloc = ceiling(dfloat(Nmu+1)/dfloat(Nmuscal))
        Nmuscal = ceiling(dfloat(Nmu+1)/dfloat(Nmuloc))
        call write_expr("Nmuscal", "ceiling(dfloat(Nmu+1)/dfloat(Nmuloc))")
      else
        Nmuscal = 1
        call write_param("Nmuscal", Nmuscal)
        Nmuloc = Nmu + 1 
        call write_expr("Nmuloc", "Nmu+1")
      end if
      if (pglobal_id .eq. 0) then
        print '(6(1X,A,I5))', "Config  Nphi", Nphi, "Nbp", &
          Nbproc_species, "Nscal", Nmuscal, &
          "Nmuloc", Nmuloc, "|   scal1", scal1, "scal2", scal2
      endif
    else
      Nmuscal = 1
      call write_param("Nmuscal", 1)
      Nmuloc  = Nmu + 1
      call write_expr("Nmuloc", "Nmu+1")
    end if

    if (Nmuscal .ne. 1) then
      if (dom_mapphi .ne. 1) then
        print *, "Strange thing happens: Nmuscal =", Nmuscal, &
          "dom_mapphi = ", dom_mapphi
        STOP
      end if
    end if

    allocate(scal_map(0:2,0:Nmuscal-1,0:Nphi-1))
    scal_map(:,:,:) = -1
    do iphi = 0, Nphi-1
      loc_mustart = 0
      do i = 0, Nmuscal-1
        loc_muend = MIN(loc_mustart + Nmuloc - 1, Nmu)
        if (loc_mustart .gt. loc_muend) then
          print *, "Big problem in establishing scal_map", &
            loc_mustart, loc_muend, Nmuscal
          STOP
        end if
        scal_map(0:2,i,iphi) = &
          (/ Nmuscal*moments_mapphi(iphi)+i, loc_mustart, loc_muend /)
        loc_mustart = loc_mustart + Nmuloc
      end do
    end do
	
    call get_ilargestart(Nphi, Nmuscal, moments_mapphi, pInSpecies_id, ilargestart)
    call get_ilargeend(Nphi, Nmuscal, moments_mapphi, pInSpecies_id, ilargeend)
	
    call write_expr("ilargestart", "get_ilargestart(Nphi, Nmuscal, moments_mapphi, pInSpecies_id)")
    call write_expr("ilargeend", "get_ilargeend(Nphi, Nmuscal, moments_mapphi, pInSpecies_id)")
    dom_mapr = MAX(nbthread,ceiling(real(Nr+1)/Nbproc_species))
    call write_expr("dom_mapr", "MAX(nbthread,ceiling(real(Nr+1)/Nbproc_species))")
	
    call FTR_ALLOC_1_ZERO(moments_reqs,0,MAX(Nphi-1,Nr))
    call FTR_ALLOC_1_ZERO(moments_reqr,0,MAX(dom_mapr*Nphi,Nbproc_loc*(Nmu+1)*dom_mapphi))
    call FTR_ALLOC_1_ZERO(bcast_reqs,0,MAX(3*dom_locphi*Nbproc_loc,dom_mapphi*Nbproc_species-1))
    call FTR_ALLOC_1_ZERO(bcast_reqr,0,MAX(Nbproc_species,MAX(Nphi,3*dom_phi)))
    moments_status_size = &
      max(max(max(max(Nphi-1,Nbproc_loc*(Nmu+1)*dom_mapphi), &
      3*dom_phi),Nr),dom_mapr*Nphi)
    call write_expr("moments_status_size" , "max(max(max(max(Nphi-1,Nbproc_loc*(Nmu+1)*dom_mapphi), 3*dom_phi),Nr),dom_mapr*Nphi)" )
    call write_param("mpi_status_size", mpi_status_size)
    call FTR_ALLOC_2_ZERO(moments_status,1,MPI_STATUS_SIZE,0,moments_status_size)

    call FTR_ALLOC_1_ZERO(mapir,0,Nr)
    if (.not.memory_test) then
      dom_mapr_loc = 0
      k            = 0
      irstart      = -1
      do iproc = 0, Nbproc_species-1
        do i = 1, dom_mapr
          if (k .le. Nr) then
            mapir(k) = iproc
            if ( (irstart.eq.-1) .and. (iproc.eq.pInSpecies_id) ) then
              irstart = k
            endif
            if (iproc.eq.pglobal_id) then
              dom_mapr_loc = dom_mapr_loc + 1
            end if
            k = k + 1
          endif
        enddo
      enddo
    end if

    if (dom_mapr_loc .gt. dom_mapr) then
      print *, "Precondition failed in the computation of dom_mapr :", &
        dom_mapr, dom_mapr_loc
      STOP
    end if

    allocate(vtime(Nbproc_loc))
    allocate(vhost(Nbproc_loc))
    call MPI_GET_PROCESSOR_NAME(name,resultlen,ierr)
    call MPI_Gather(name, MPI_MAX_PROCESSOR_NAME, MPI_CHARACTER, vhost, &
      MPI_MAX_PROCESSOR_NAME, MPI_CHARACTER, 0, MPI_COMM_MU, ierr);
#ifdef TIMER
    write (*,'(A,1X,I8,1X,I8,1X,A64)')"Proc.",pglobal_id,MPI_MAX_PROCESSOR_NAME,name
#endif
    call ppbarrier()
#ifndef TIMER
    if (pglobal_id .eq. 0) then
      do j = 0, 2
        write (*,'(A)',ADVANCE='NO')"smap"
        do iphi = 0, Nphi-1
          do i = 0, Nmuscal-1
            write (*,'(I4)',ADVANCE='NO')scal_map(j,i,iphi)
          end do
        end do
        write (*,'(A)')""
      end do
    end if
#endif
    call Write_dyn_balise_close('Init_comm') !R4
!R3 call r3_info_end(r3_info_index_1) !R3
  end subroutine ppinit_comm


  !---------------------------------------------------------------
  ! Transposition of a 4D structure (r,theta,phi,vpar) 
  !  distributed in r and theta into a 4D structure distributed
  !  in phi and vpar, i.e:
  !   fval(istart:iend,jstart:jend,0:Nphi,0:Nvpar)
  !   -> fval_transp(0:Nr,0:Ntheta,kstart:kend,lstart:lend)
  !---------------------------------------------------------------
  subroutine pptranspose_forward(fval,timing)
!R3 #include "code_instrumentation/r3_info.h" !R3
    use clock_module
    real*8, dimension (istart_buf:iend_buf,&
      jstart_buf:jend_buf,0:Nphi,0:Nvpar), intent(in) :: fval
    logical                              , intent(in) :: timing
    integer :: ir, itheta, iphi, ivpar, zz, k, q, tag, ierr, i
    integer :: base, dep
    integer :: dististart, distiend, distjstart, distjend
    integer :: distkstart, distlstart, distlend
    integer :: distlnumber
    integer :: offset, sendsize, send_id, recv_id
    integer :: transp_req(0:Nbproc_loc-1), transp_sreq(0:Nbproc_loc-1)
    integer :: transp_stat(MPI_STATUS_SIZE,0:Nbproc_loc-1)
    integer :: fval_lbound(4), fval_shape(4)
    integer :: f4D_transp_lbound(4), f4D_transp_shape(4)
    integer :: send_type(0:Nbproc_loc-1), recv_type(0:Nbproc_loc-1)
    integer :: count_vec(0:Nbproc_loc-1), disp_vec(0:Nbproc_loc-1)
#ifdef TIMER
    real(RKIND)                        :: tbar1, tbar2, tcalc, tmin, tmax
    real(RKIND)                        :: ttransp
    integer                            :: idmin, idmax
    character(12)                      :: hmin, hmax
    integer(TIMEPREC)                  :: t1, t2
#endif

!R3 call r3_info_begin (r3_info_index_1, 'Comm_transp_forw') !R3
    call Write_dyn_balise_open('Comm_transp_forw') !R4

    if (timing) then
      call ppbarrier()
      call clck_time(bclock_transp_adv2D)
#ifdef TIMER
      call clck_time(t1)
#endif
    end if
    sendsize = dom_vpar * dom_phi * dom_theta * dom_r
    fval_shape = shape(fval)
    fval_lbound = lbound(fval)
    f4D_transp_lbound = lbound(f4D_transp)
    f4D_transp_shape = shape(f4D_transp)
    do offset = 0,Nbproc_loc-1
      send_id     = mod(plocal_id + offset, Nbproc_loc)
      recv_id     = mod(plocal_id - offset + Nbproc_loc, Nbproc_loc)
      
      ! send shape (or reshaping)
      base        = (send_id/Nbproc_phi)
      dep         = mod(send_id,Nbproc_phi)
      distkstart  = dep *  dom_phi
      distlstart  = (base * (Nvpar+1)) / Nbproc_vpar
      distlend    = ((base+1) * (Nvpar+1)) / Nbproc_vpar - 1
      distlnumber = distlend - distlstart + 1
      select case ( transpose_version )
      case ( TRANSPOSE_SENDIRECV )
        do ivpar = 0,distlnumber-1
          do iphi = 0,dom_phi-1
            do itheta = 0,dom_theta-1
              do ir = 0,dom_r-1
                f4D_send(ir, itheta, iphi, ivpar) = &
                                fval(istart+ir, jstart+itheta, &
                                distkstart+iphi, distlstart+ivpar)
              end do
            end do
          end do
        end do
      case ( TRANSPOSE_ALLTOALLW )
        call MPI_Type_create_subarray(4, fval_shape, &
                        (/ dom_r, dom_theta, dom_phi, distlnumber /), &
                        (/ istart, jstart, distkstart, distlstart /) - &
                                         fval_lbound, &
                        MPI_ORDER_FORTRAN, MPI_REAL8, &
                        send_type(send_id), ierr)
        call MPI_Type_commit(send_type(send_id), ierr)
      case default
        print *, pglobal_id, 'Invalid value for transpose_version', &
                        transpose_version
        call MPI_abort(MPI_COMM_WORLD, 1, ierr)
      end select

      ! data exchange (reshaping vesrions)
      tag     = offset + 10
      select case ( transpose_version )
      case ( TRANSPOSE_SENDIRECV )
        call MPI_irecv(f4D_recv, sendsize, MPI_REAL8, recv_id, &
                       tag, mpi_comm_mu, transp_req(offset), ierr)
        call MPI_send(f4D_send, sendsize, MPI_REAL8, send_id, &
                      tag, mpi_comm_mu, ierr)
        call MPI_wait(transp_req(offset), status(:,1), ierr)
      case ( TRANSPOSE_ALLTOALLW )
        ! nothing to do here
      case default
        print *, pglobal_id, 'Invalid value for transpose_version', &
                        transpose_version
        call MPI_abort(MPI_COMM_WORLD, 1, ierr)
      end select
      
      ! receive shape (or reshaping)
      base       = (recv_id/Nbproc_r)
      dep        = mod(recv_id,Nbproc_r)              
      dististart = dep *  dom_r
      distjstart = base * dom_theta
      select case ( transpose_version )
      case ( TRANSPOSE_SENDIRECV )
        do ivpar = 0,lnumber-1
          do iphi = 0,knumber-1
            do itheta = 0,dom_theta-1
              do ir = 0,dom_r-1
                f4D_transp(dististart+ir, distjstart+itheta, &
                                kstart+iphi, lstart+ivpar) = &
                                f4D_recv(ir, itheta, iphi, ivpar) 
              end do
            end do
          end do
        end do
      case ( TRANSPOSE_ALLTOALLW )
        call MPI_Type_create_subarray(4, f4D_transp_shape, &
                        (/ dom_r, dom_theta, knumber, lnumber /), &
                        (/ dististart, distjstart, kstart, lstart /) - &
                                         f4D_transp_lbound, &
                        MPI_ORDER_FORTRAN, MPI_REAL8, &
                        recv_type(recv_id), ierr)
        call MPI_Type_commit(recv_type(recv_id), ierr)
      case default
        print *, pglobal_id, 'Invalid value for transpose_version', &
                        transpose_version
        call MPI_abort(MPI_COMM_WORLD, 1, ierr)
      end select
    end do
    
    ! data exchange (non reshaped versions)
    select case ( transpose_version )
    case ( TRANSPOSE_SENDIRECV )
      ! nothing to do here
    case ( TRANSPOSE_ALLTOALLW )
      count_vec = 1
      disp_vec = 0
      call MPI_alltoallw(fval, count_vec, disp_vec, send_type, &
                      f4D_transp, count_vec, disp_vec, recv_type, &
                      mpi_comm_mu, ierr)
      do i = 0, Nbproc_loc-1
        call MPI_Type_free(send_type(i), ierr)
        call MPI_Type_free(recv_type(i), ierr)
      end do
    case default
      print *, pglobal_id, 'Invalid value for transpose_version', &
                        transpose_version
      call MPI_abort(MPI_COMM_WORLD, 1, ierr)
    end select
    
    if (timing) then
#ifdef TIMER
      call clck_time(t2)
      call clck_ldiff(t1,t2,tcalc)
      call pp_reduce_scalar(tcalc,tmin,idmin,hmin,tmax,idmax,hmax)
#endif      

      call ppbarrier()
      call clck_time(eclock_transp_adv2D)
      call clck_diff(bclock_transp_adv2D,eclock_transp_adv2D, &
        global_time_transp_adv2D)
#ifdef TIMER
      call clck_ldiff(bclock_transp_adv2D,eclock_transp_adv2D, &
        ttransp)
      if (pglobal_id .eq. 0) then
        write(6,'(I4,A,F13.7)') &
          pglobal_id, " Temps transp", ttransp
      end if
      if (plocal_id .eq. 0) then
        write(6,'(I6,2(1X,A,1X,I6,1X,F13.7))') pglobal_id, hmin, idmin, tmin, hmax, idmax, tmax 
      end if
#endif
    end if
    call Write_dyn_balise_close('Comm_transp_forw') !R4
!R3 call r3_info_end(r3_info_index_1) !R3
  end subroutine pptranspose_forward


  !---------------------------------------------------------------
  ! Transposition of a 4D structure (r,theta,phi,vpar) 
  !  distributed in phi and vpar into a 4D structure distributed
  !  in r and theta, i.e:
  !   fval_transp(0:Nr,0:Ntheta,kstart:kend,lstart:lend)
  !   -> fval(istart:iend,jstart:jend,0:Nphi,0:Nvpar)
  !
  ! ( Rk : Opposite transposition of the one performed by 
  !   pptranspose_backward )
  !---------------------------------------------------------------
  subroutine pptranspose_backward(fval,timing)
!R3 #include "code_instrumentation/r3_info.h" !R3
    use clock_module

    real*8, dimension (istart_buf:iend_buf, &
      jstart_buf:jend_buf,0:Nphi,0:Nvpar), intent(inout) :: fval
    logical                              , intent(in)    :: timing

    integer :: ir, itheta, iphi, ivpar, zz, k, q, tag, ierr, i
    integer :: base, dep
    integer :: dististart, distiend, distjstart, distjend
    integer :: distkstart, distkend, distlstart, distlend
    integer :: distknumber, distlnumber
    integer :: offset, sendsize, send_id, recv_id
    integer :: transp_req(0:Nbproc_loc-1)
    integer :: transp_stat(MPI_STATUS_SIZE,0:Nbproc_loc-1)
    integer :: fval_lbound(4), fval_shape(4)
    integer :: f4D_transp_lbound(4), f4D_transp_shape(4)
    integer :: send_type(0:Nbproc_loc-1), recv_type(0:Nbproc_loc-1)
    integer :: count_vec(0:Nbproc_loc-1), disp_vec(0:Nbproc_loc-1)
#ifdef TIMER
    real(RKIND)                        :: tbar1, tbar2, tcalc, tmin, tmax
    real(RKIND)                        :: ttransp
    integer                            :: idmin, idmax
    character(12)                      :: hmin, hmax
    integer(TIMEPREC)                  :: t1, t2
#endif

    if (timing) then
      call ppbarrier()
      call clck_time(bclock_transp_adv2D)
#ifdef TIMER
      call clck_time(t1)
#endif
    end if
!R3 call r3_info_begin (r3_info_index_1, 'Comm_transp_back') !R3
    call Write_dyn_balise_open('Comm_transp_back') !R4
    sendsize = dom_vpar * dom_phi * dom_theta * dom_r
    fval_shape = shape(fval)
    fval_lbound = lbound(fval)
    f4D_transp_lbound = lbound(f4D_transp)
    f4D_transp_shape = shape(f4D_transp)
    do offset = 0,Nbproc_loc-1
      send_id    = mod(plocal_id + offset, Nbproc_loc)
      recv_id    = mod(plocal_id - offset + Nbproc_loc, Nbproc_loc)
      
      ! send shape 
      base       = (send_id/Nbproc_r)
      dep        = mod(send_id,Nbproc_r)              
      dististart = dep *  dom_r
      distiend   = dististart + dom_r - 1
      distjstart = base * dom_theta
      distjend   = distjstart + dom_theta - 1
      select case ( transpose_version )
      case ( TRANSPOSE_SENDIRECV )
        do ivpar = 0,lnumber-1
          do iphi = 0,knumber-1
            do itheta = 0,dom_theta-1
              do ir = 0,dom_r-1
                f4D_send(ir,itheta,iphi,ivpar)  = & 
                  f4D_transp(dististart+ir,distjstart+itheta,&
                  kstart+iphi,lstart+ivpar) 

              end do
            end do
          end do
        end do
      case ( TRANSPOSE_ALLTOALLW )
        call MPI_Type_create_subarray(4, f4D_transp_shape, &
                        (/ dom_r, dom_theta, knumber, lnumber /), &
                        (/ dististart, distjstart, kstart, lstart /) - &
                                         f4D_transp_lbound, &
                        MPI_ORDER_FORTRAN, MPI_REAL8, &
                        send_type(send_id), ierr)
        call MPI_Type_commit(send_type(send_id), ierr)
      case default
        print *, pglobal_id, 'Invalid value for transpose_version', &
                        transpose_version
        call MPI_abort(MPI_COMM_WORLD, 1, ierr)
      end select

      ! data exchange (p2p case)
      select case ( transpose_version )
      case ( TRANSPOSE_SENDIRECV )
        tag     = offset + 10
        call MPI_irecv(f4D_recv, sendsize, MPI_REAL8, recv_id, &
                        tag, mpi_comm_mu, transp_req(offset), ierr)
        call MPI_send(f4D_send, sendsize, MPI_REAL8, send_id, &
                        tag, mpi_comm_mu, ierr)
        call MPI_wait(transp_req(offset), status(:,1), ierr)
      case ( TRANSPOSE_ALLTOALLW )
        ! nothing to do here
      case default
        print *, pglobal_id, 'Invalid value for transpose_version', &
                        transpose_version
        call MPI_abort(MPI_COMM_WORLD, 1, ierr)
      end select

      ! receive data shape
      base        = (recv_id/Nbproc_phi)
      dep         = mod(recv_id,Nbproc_phi)
      distkstart  = dep * dom_phi
      distlstart  = (base * (Nvpar+1)) / Nbproc_vpar
      distlend    = ((base+1) * (Nvpar+1)) / Nbproc_vpar - 1
      distlnumber = distlend - distlstart + 1
      select case ( transpose_version )
      case ( TRANSPOSE_SENDIRECV )
        do ivpar = 0,distlnumber-1
          do iphi = 0,dom_phi-1
            do itheta = 0,dom_theta-1
              do ir = 0,dom_r-1
                fval(istart+ir,jstart+itheta, &
                  distkstart+iphi,distlstart+ivpar) = &
                  f4D_recv(ir,itheta,iphi,ivpar)
              end do
            end do
          end do
        end do
      case ( TRANSPOSE_ALLTOALLW )
        call MPI_Type_create_subarray(4, fval_shape, &
                        (/ dom_r, dom_theta, dom_phi, distlnumber /), &
                        (/ istart, jstart, distkstart, distlstart /) - &
                                         fval_lbound, &
                        MPI_ORDER_FORTRAN, MPI_REAL8, &
                        recv_type(recv_id), ierr)
        call MPI_Type_commit(recv_type(recv_id), ierr)
      case default
        print *, pglobal_id, 'Invalid value for transpose_version', &
                        transpose_version
        call MPI_abort(MPI_COMM_WORLD, 1, ierr)
      end select
    end do
    
    ! data exchange (non reshaped versions)
    select case ( transpose_version )
    case ( TRANSPOSE_SENDIRECV )
      ! nothing to do here
    case ( TRANSPOSE_ALLTOALLW )
      count_vec = 1
      disp_vec = 0
      call MPI_alltoallw(f4D_transp, count_vec, disp_vec, send_type, &
                      fval, count_vec, disp_vec, recv_type, &
                      mpi_comm_mu, ierr)
      do i = 0, Nbproc_loc-1
        call MPI_Type_free(send_type(i), ierr)
        call MPI_Type_free(recv_type(i), ierr)
      end do
    case default
      print *, pglobal_id, 'Invalid value for transpose_version', &
                        transpose_version
      call MPI_abort(MPI_COMM_WORLD, 1, ierr)
    end select
    
    if (timing) then
#ifdef TIMER
      call clck_time(t2)
      call clck_ldiff(t1,t2,tcalc)
      call pp_reduce_scalar(tcalc,tmin,idmin,hmin,tmax,idmax,hmax)
#endif      
      call ppbarrier()
      call clck_time(eclock_transp_adv2D)
      call clck_diff(bclock_transp_adv2D,eclock_transp_adv2D, &
        global_time_transp_adv2D)
#ifdef TIMER
      call clck_ldiff(bclock_transp_adv2D,eclock_transp_adv2D, &
        ttransp)
      if (pglobal_id .eq. 0) then
        write(6,'(I4,A,F13.7)') &
          pglobal_id, " Temps transp", ttransp
      end if
      if (plocal_id .eq. 0) then
        write(6,'(I6,2(1X,A,1X,I6,1X,F13.7))') pglobal_id, hmin, idmin, tmin, hmax, idmax, tmax 
    end if
#endif
    end if
    
    call Write_dyn_balise_close('Comm_transp_back') !R4
!R3 call r3_info_end(r3_info_index_1) !R3
  end subroutine pptranspose_backward

  !---------------------------------------------------------------
  ! Transposition of a 4D structure (r,theta,phi,vpar) 
  !  distributed in r and theta into a 4D structure distributed
  !  in r, i.e:
  !   fval(istart:iend,jstart:jend,0:Nphi,0:Nvpar)
  !   -> fval_transp(pos_rdiags,0:Ntheta,0:Nphi,0:Nvpar)
  !---------------------------------------------------------------
  subroutine distribute_Sdiag(fval,f4D_Sdiag_transp,&
    ndiags_glob,cumul_r,pos_rdiags,ir_map, pid_computing)
    
    real*8, dimension (istart_buf:iend_buf,&
      jstart_buf:jend_buf,0:Nphi,0:Nvpar), intent(in) :: fval
    real*8, dimension (1:,0:,0:,0:), intent(inout)    :: f4D_Sdiag_transp
    integer, intent(in)                               :: pid_computing
    integer, intent(in)                               :: ndiags_glob
    integer, dimension(0:)   , intent(in)             :: cumul_r
    integer, dimension(1:,0:), intent(in)             :: ir_map
    integer, dimension(:)    , intent(in)             :: pos_rdiags
    
    integer :: ir, itheta, iphi, ivpar, zz, k, q, tag, ierr
    integer :: base, dep
    integer :: dististart, distiend, distjstart, distjend
    integer :: distkstart, distkend, distlstart, distlend
    integer :: distknumber, distlnumber
    integer :: offset, sendsize, recvsize, send_id, recv_id
    integer :: pid_r, pid_r_recv, pid
      
!R3 #include "code_instrumentation/r3_info.h" !R3
    
    pid_r = mod(plocal_id,Nbproc_r)
    if (cumul_r(pid_r) .ne. 0) then
      if (plocal_id .eq. pid_computing) then
        do pid = 0,Nbproc_loc-1
          tag = 10 + pid
          pid_r_recv = mod(pid,Nbproc_r)
          if ((cumul_r(pid_r_recv) .ne. 0).and.&
            (pid.ne.pid_computing)) then
            recvsize = (Nvpar+1) * (Nphi+1) * dom_theta * cumul_r(pid_r_recv) * 5
            call MPI_RECV(f4D_Sdiag_recv(0,0,0,0),&
              recvsize,MPI_REAL8,pid, &
              tag,mpi_comm_mu,status,ierr)
          else
            dististart = 5*(sum(cumul_r(0:pid_r_recv))-cumul_r(pid_r_recv))
            do ivpar = 0,Nvpar
              do iphi = 0,Nphi
                do itheta = 0,dom_theta-1
                  do ir = 1, 5*cumul_r(pid_r_recv)
                    f4D_Sdiag_recv(itheta,iphi,ivpar,ir-1) = &
                      fval(pos_rdiags(dististart+ir),jstart+itheta, &
                      iphi,ivpar)
                  end do
                end do
              end do
            end do
          end if
          !*** Update Table
          base       = (pid/Nbproc_r)
          distjstart = base * dom_theta
          dististart = 5*(sum(cumul_r(0:pid_r_recv))-cumul_r(pid_r_recv))
          do ivpar = 0,Nvpar
            do iphi = 0,Nphi
              do itheta = 0,dom_theta-1
                do ir = 1,5*cumul_r(pid_r_recv)
                  f4D_Sdiag_transp(dististart+ir,distjstart+itheta, &
                    iphi,ivpar) = &
                    f4D_Sdiag_recv(itheta,iphi,ivpar,ir-1) 
                end do
              end do
            end do
          end do
        end do !end loop on procs
      else
        sendsize = (Nvpar+1) * (Nphi+1) * dom_theta * cumul_r(pid_r) * 5
        dististart = 5*(sum(cumul_r(0:pid_r))-cumul_r(pid_r))
        do ivpar = 0,Nvpar
          do iphi = 0,Nphi
            do itheta = 0,dom_theta-1
              do ir = 1, 5*cumul_r(pid_r)
                f4D_Sdiag_send(itheta,iphi,ivpar,ir-1) = &
                  fval(pos_rdiags(dististart+ir),jstart+itheta, &
                  iphi,ivpar)
              end do
            end do
          end do
        end do
        tag = 10 + plocal_id
        call MPI_SEND(f4D_Sdiag_send(0,0,0,0), &
          sendsize,MPI_REAL8,pid_computing, &
          tag,mpi_comm_mu,ierr)
      end if
    end if
  end subroutine distribute_Sdiag

  !---------------------------------------------------------------
  ! Transposition of a 5D structure (r,theta,phi,vpar) 
  !  distributed in (r, theta, mu) into a 5D structure distributed
  !  in (r, theta, phi), i.e:
  !   fval(istart:iend,jstart:jend,0:Nphi,0:Nvpar,mu_id)
  !   -> fval_transp(istart:iend,jstart:jend,
  !                 cstart:cend,0:Nvpar,0:Nmu)
  !---------------------------------------------------------------
  subroutine pptranspvel_forward(fval)
    real*8, dimension (istart_buf:iend_buf,&
      jstart_buf:jend_buf,0:Nphi,0:Nvpar), intent(in) :: fval

    integer :: ir, itheta, iphi, ivpar, zz, k, q, tag, ierr
    integer :: indx
    integer :: base, dep
    integer :: dististart, distiend, distjstart, distjend
    integer :: offset, sendsize, send_id, recv_id
    integer :: iproc, mudist
!R3 #include "code_instrumentation/r3_info.h" !R3

!R3 call r3_info_begin (r3_info_index_1, 'Comm_transpvel_forw') !R3
    call Write_dyn_balise_open('Comm_transpvel_forw') !R4
    sendsize = dom_bisphi * dom_theta * dom_r * (Nvpar+1)

    do offset = 0, Nmu
      mudist      = mod(mu_id + offset, Nmu+1)
      send_id     = plocal_id + mudist * Nbproc_loc
      base        = min(Nphi-1, mudist * dom_bisphi)
      if (base .eq. mudist * dom_bisphi) then
        do ivpar = 0, Nvpar
          do iphi = 0, min(Nphi-1-base,dom_bisphi-1)
            do itheta = 0, dom_theta-1
              do ir = 0, dom_r-1
                indx = ir + dom_r * &
                  (itheta+dom_theta*(iphi+dom_bisphi*ivpar))
                f1D_send(indx) = &
                  fval(istart+ir,jstart+itheta, &
                  base+iphi,ivpar)
              end do
            end do
          end do
        end do
      end if
      mudist      = mod(Nmu+1 + mu_id - offset, Nmu+1)
      recv_id     = plocal_id + mudist * Nbproc_loc
      tag     = offset + 10
      call MPI_IRECV(f1D_recv,sendsize,MPI_REAL8,recv_id, &
        tag,mpi_comm_species,reqr(1),ierr)
      call MPI_SEND(f1D_send,sendsize,MPI_REAL8,send_id, &
        tag,mpi_comm_species,ierr)
      call MPI_WAIT(reqr(1),status(:,1),ierr)

      if (collstart .eq. (mu_id * dom_bisphi)) then
        do ivpar = 0, Nvpar
          do iphi = 0, collend-collstart
            do itheta = 0, dom_theta-1
              do ir = 0, dom_r-1
                indx = ir + dom_r * &
                  (itheta+dom_theta*(iphi+dom_bisphi*ivpar))
                f5D_transp(istart+ir,jstart+itheta, &
                  collstart+iphi,ivpar,mudist) = &
                  f1D_recv(indx) 
              end do
            end do
          end do
        end do
      end if
    end do
    call Write_dyn_balise_close('Comm_transpvel_forw') !R4
!R3 call r3_info_end(r3_info_index_1) !R3
  end subroutine pptranspvel_forward


  !---------------------------------------------------------------
  ! Transposition of a 5D structure (r,theta,phi,vpar) 
  !  distributed in (r, theta, phi) into a 5D structure distributed
  !  in (r, theta, mu), i.e:
  !  fval_transp(istart:iend,jstart:jend,cstart:cend,0:Nvpar,0:Nmu)
  !    -> fval(istart:iend,jstart:jend,0:Nphi,0:Nvpar,mu_id)
  !---------------------------------------------------------------
  subroutine pptranspvel_backward(fval)
    real*8, dimension (istart_buf:iend_buf,&
      jstart_buf:jend_buf,0:Nphi,0:Nvpar), intent(out) :: fval

    integer :: ir, itheta, iphi, ivpar, zz, k, q, tag, ierr
    integer :: indx
    integer :: base, dep
    integer :: dististart, distiend, distjstart, distjend
    integer :: offset, sendsize, send_id, recv_id
    integer :: iproc, mudist
!R3 #include "code_instrumentation/r3_info.h" !R3

!R3 call r3_info_begin (r3_info_index_1, 'Comm_transpvel_back') !R3
    call Write_dyn_balise_open('Comm_transpvel_back') !R4
    sendsize = dom_bisphi * dom_theta * dom_r * (Nvpar+1)

    do offset = 0, Nmu
      mudist      = mod(mu_id + offset, Nmu+1)
      send_id     = plocal_id + mudist * Nbproc_loc

      if (collstart .eq. mu_id * dom_bisphi) then
        do ivpar = 0, Nvpar
          do iphi = 0, collend-collstart
            do itheta = 0, dom_theta-1
              do ir = 0, dom_r-1
                indx = ir + dom_r * &
                  (itheta+dom_theta*(iphi+dom_bisphi*ivpar))
                f1D_send(indx) = &
                  f5D_transp(istart+ir,jstart+itheta, &
                  collstart+iphi,ivpar,mudist) 
              end do
            end do
          end do
        end do
      end if
      mudist  = mod(Nmu+1 + mu_id - offset, Nmu+1)
      recv_id = plocal_id + mudist * Nbproc_loc
      tag     = offset + 10
      call MPI_IRECV(f1D_recv,sendsize,MPI_REAL8,recv_id, &
        tag,mpi_comm_species,reqr(1),ierr)
      call MPI_SEND(f1D_send,sendsize,MPI_REAL8,send_id, &
        tag,mpi_comm_species,ierr)
      call MPI_WAIT(reqr(1),status(:,1),ierr)

      base        = min(Nphi-1, mudist * dom_bisphi)
      if (base .eq. mudist * dom_bisphi) then
        do ivpar = 0, Nvpar
          do iphi = 0, min(Nphi-1-base,dom_bisphi-1)
            do itheta = 0, dom_theta-1
              do ir = 0, dom_r-1
                indx = ir + dom_r * &
                  (itheta+dom_theta*(iphi+dom_bisphi*ivpar))
                fval(istart+ir,jstart+itheta, &
                  base+iphi,ivpar) = f1D_recv(indx)
              end do
            end do
          end do
        end do
      end if
    end do
    call Write_dyn_balise_close('Comm_transpvel_back') !R4
!R3 call r3_info_end (r3_info_index_1) !R3
  end subroutine pptranspvel_backward


  !-----------------------------------------------------
  ! Delete all the arrays associated to the parallel
  !  buffers
  !-----------------------------------------------------
  subroutine ppdeallocate

    deallocate(moments_mapphi)
    deallocate(moments_locphi)
    deallocate(scal_map)
    deallocate(vtime)
    deallocate(vhost)

    call FTR_DEALLOC(rbufWW)
    call FTR_DEALLOC(rbufEE)
    call FTR_DEALLOC(sbufWW)
    call FTR_DEALLOC(sbufEE)
    call FTR_DEALLOC(rbufNN)
    call FTR_DEALLOC(rbufSS)
    call FTR_DEALLOC(sbufNN)
    call FTR_DEALLOC(sbufSS)
    call FTR_DEALLOC(rbufNW)
    call FTR_DEALLOC(rbufSW)
    call FTR_DEALLOC(rbufNE)
    call FTR_DEALLOC(rbufSE)
    call FTR_DEALLOC(sbufNW)
    call FTR_DEALLOC(sbufSW)
    call FTR_DEALLOC(sbufNE)
    call FTR_DEALLOC(sbufSE)
    call FTR_DEALLOC(fillderiv)
!baoter
    if (Sp_saving) then
      call FTR_DEALLOC(f4D_Sdiag_send)
      call FTR_DEALLOC(f4D_Sdiag_recv)
    end if
    if (transp_velocity) then
      call FTR_DEALLOC(f5D_transp)
      call FTR_DEALLOC(f1D_send)
      call FTR_DEALLOC(f1D_recv)
    end if
!eaoter
    call FTR_DEALLOC(moments_reqs)
    call FTR_DEALLOC(moments_reqr)
    call FTR_DEALLOC(bcast_reqs)
    call FTR_DEALLOC(bcast_reqr)
    call FTR_DEALLOC(moments_status)
    call FTR_DEALLOC(mapir)
  end subroutine ppdeallocate


  !-----------------------------------------------------
  ! send all data on the local processors 
  !  (using MPI_SSEND directive)
  !-----------------------------------------------------
  subroutine comm_send 
    integer :: ierr

    call MPI_SEND(sbufWW,1,mpitype_WE,WW,10,mpi_comm_mu,ierr)
    call MPI_SEND(sbufEE,1,mpitype_WE,EE,11,mpi_comm_mu,ierr)
    call MPI_SEND(sbufNW,1,mpitype_corner,NW,20,mpi_comm_mu,ierr)
    call MPI_SEND(sbufNE,1,mpitype_corner,NE,21,mpi_comm_mu,ierr)
    call MPI_SEND(sbufSE,1,mpitype_corner,SE,23,mpi_comm_mu,ierr)
    call MPI_SEND(sbufSW,1,mpitype_corner,SW,22,mpi_comm_mu,ierr)
    call MPI_SEND(sbufSS,1,mpitype_NS,SS,13,mpi_comm_mu,ierr)
    call MPI_SEND(sbufNN,1,mpitype_NS,NN,12,mpi_comm_mu,ierr)
  end subroutine comm_send


  !-----------------------------------------------------------------
  ! receive all data on the local processor sended by other procs.
  !-----------------------------------------------------------------
  subroutine comm_irecv 
    integer :: ierr

   call MPI_IRECV(rbufEE,1,mpitype_WE,EE,10,mpi_comm_mu, &
     reqr(1),ierr)
   call MPI_IRECV(rbufWW,1,mpitype_WE,WW,11,mpi_comm_mu, &
     reqr(2),ierr)
   call MPI_IRECV(rbufSE,1,mpitype_corner,SE,20,mpi_comm_mu, &
     reqr(3),ierr)
   call MPI_IRECV(rbufSW,1,mpitype_corner,SW,21,mpi_comm_mu, &
     reqr(4),ierr)
   call MPI_IRECV(rbufNW,1,mpitype_corner,NW,23,mpi_comm_mu, &
     reqr(5),ierr)
   call MPI_IRECV(rbufNE,1,mpitype_corner,NE,22,mpi_comm_mu, &
     reqr(6),ierr)
   call MPI_IRECV(rbufNN,1,mpitype_NS,NN,13,mpi_comm_mu, &
     reqr(7),ierr)
   call MPI_IRECV(rbufSS,1,mpitype_NS,SS,12,mpi_comm_mu, &
     reqr(8),ierr)
 end subroutine comm_irecv

 !----------------------------------------------------------
 ! Exchange data synchronously between neighbours.
 ! Needed to run on large number of cores on Curie machine.
 ! Macro OLDCOMM allows one to switch from new to old
 ! version of communication scheme.s
 !----------------------------------------------------------
  subroutine comm_exchange_neighbours
    integer :: ierr

   call MPI_IRECV(rbufEE,1,mpitype_WE,EE,10,mpi_comm_mu, &
     reqr(1),ierr)
   call MPI_SEND(sbufWW,1,mpitype_WE,WW,10,mpi_comm_mu,ierr)
!   call MPI_WAIT(reqr(1),status,ierr)

   call MPI_IRECV(rbufWW,1,mpitype_WE,WW,11,mpi_comm_mu, &
     reqr(2),ierr)
   call MPI_SEND(sbufEE,1,mpitype_WE,EE,11,mpi_comm_mu,ierr)
!   call MPI_WAIT(reqr(2),status,ierr)
   call MPI_WAITALL(2,reqr(1:2),status,ierr)

   call MPI_IRECV(rbufSE,1,mpitype_corner,SE,20,mpi_comm_mu, &
     reqr(3),ierr)
   call MPI_SEND(sbufNW,1,mpitype_corner,NW,20,mpi_comm_mu,ierr)
!   call MPI_WAIT(reqr(3),status,ierr)

   call MPI_IRECV(rbufNW,1,mpitype_corner,NW,23,mpi_comm_mu, &
     reqr(4),ierr)
   call MPI_SEND(sbufSE,1,mpitype_corner,SE,23,mpi_comm_mu,ierr)
!   call MPI_WAIT(reqr(4),status,ierr)
   call MPI_WAITALL(2,reqr(3:4),status,ierr)

   call MPI_IRECV(rbufSW,1,mpitype_corner,SW,21,mpi_comm_mu, &
     reqr(5),ierr)
   call MPI_SEND(sbufNE,1,mpitype_corner,NE,21,mpi_comm_mu,ierr)
!   call MPI_WAIT(reqr(5),status,ierr)

   call MPI_IRECV(rbufNE,1,mpitype_corner,NE,22,mpi_comm_mu, &
     reqr(6),ierr)
   call MPI_SEND(sbufSW,1,mpitype_corner,SW,22,mpi_comm_mu,ierr)
!   call MPI_WAIT(reqr(6),status,ierr)
   call MPI_WAITALL(2,reqr(5:6),status,ierr)

   call MPI_IRECV(rbufNN,1,mpitype_NS,NN,13,mpi_comm_mu, &
     reqr(7),ierr)
   call MPI_SEND(sbufSS,1,mpitype_NS,SS,13,mpi_comm_mu,ierr)
!   call MPI_WAIT(reqr(7),status,ierr)

   call MPI_IRECV(rbufSS,1,mpitype_NS,SS,12,mpi_comm_mu, &
     reqr(8),ierr)
   call MPI_SEND(sbufNN,1,mpitype_NS,NN,12,mpi_comm_mu,ierr)
!   call MPI_WAIT(reqr(8),status,ierr)
   call MPI_WAITALL(2, reqr(7:8), status, ierr)

 end subroutine comm_exchange_neighbours


!  subroutine comm_exchange_neighbours
!    integer :: ierr
!
!   call MPI_IRECV(rbufEE,1,mpitype_WE,EE,10,mpi_comm_mu, &
!     reqr(1),ierr)
!   call MPI_SEND(sbufWW,1,mpitype_WE,WW,10,mpi_comm_mu,ierr)
!   call MPI_WAIT(reqr(1),status,ierr)
!
!   call MPI_IRECV(rbufWW,1,mpitype_WE,WW,11,mpi_comm_mu, &
!     reqr(2),ierr)
!   call MPI_SEND(sbufEE,1,mpitype_WE,EE,11,mpi_comm_mu,ierr)
!   call MPI_WAIT(reqr(2),status,ierr)
!
!   call MPI_IRECV(rbufSE,1,mpitype_corner,SE,20,mpi_comm_mu, &
!     reqr(3),ierr)
!   call MPI_SEND(sbufNW,1,mpitype_corner,NW,20,mpi_comm_mu,ierr)
!   call MPI_WAIT(reqr(3),status,ierr)
!
!   call MPI_IRECV(rbufSW,1,mpitype_corner,SW,21,mpi_comm_mu, &
!     reqr(4),ierr)
!   call MPI_SEND(sbufNE,1,mpitype_corner,NE,21,mpi_comm_mu,ierr)
!   call MPI_WAIT(reqr(4),status,ierr)
!
!   call MPI_IRECV(rbufNW,1,mpitype_corner,NW,23,mpi_comm_mu, &
!     reqr(5),ierr)
!   call MPI_SEND(sbufSE,1,mpitype_corner,SE,23,mpi_comm_mu,ierr)
!   call MPI_WAIT(reqr(5),status,ierr)
!
!   call MPI_IRECV(rbufNE,1,mpitype_corner,NE,22,mpi_comm_mu, &
!     reqr(6),ierr)
!   call MPI_SEND(sbufSW,1,mpitype_corner,SW,22,mpi_comm_mu,ierr)
!   call MPI_WAIT(reqr(6),status,ierr)
!
!   call MPI_IRECV(rbufNN,1,mpitype_NS,NN,13,mpi_comm_mu, &
!     reqr(7),ierr)
!   call MPI_SEND(sbufSS,1,mpitype_NS,SS,13,mpi_comm_mu,ierr)
!   call MPI_WAIT(reqr(7),status,ierr)
!
!   call MPI_IRECV(rbufSS,1,mpitype_NS,SS,12,mpi_comm_mu, &
!     reqr(8),ierr)
!   call MPI_SEND(sbufNN,1,mpitype_NS,NN,12,mpi_comm_mu,ierr)
!   call MPI_WAIT(reqr(8),status,ierr)
!
! end subroutine comm_exchange_neighbours



  !-----------------------------------------------------
  ! used for test
  !-----------------------------------------------------
  subroutine comm_wait_recv
    integer :: ierr

    call MPI_WAITALL(8,reqr,status, ierr)
  end subroutine comm_wait_recv


  !-----------------------------------------------------
  ! exit 
  !-----------------------------------------------------
  subroutine ppexit
    integer :: ierr
    
    call MPI_BARRIER(MPI_COMM_WORLD,ierr)
    call MPI_FINALIZE(ierr)
  end subroutine ppexit
  
  
  !-----------------------------------------------------
  ! MPI_BARRIER
  !-----------------------------------------------------  
  subroutine ppbarrier
    integer :: ierr
    
    call MPI_BARRIER(MPI_COMM_WORLD,ierr)
  end subroutine ppbarrier


  !-----------------------------------------------------
  ! MPI_BARRIER
  !-----------------------------------------------------  
  subroutine ppbarrier_timer
    integer :: ierr
    
#ifdef TIMER
    call MPI_BARRIER(MPI_COMM_WORLD,ierr)
#endif
  end subroutine ppbarrier_timer


  !-----------------------------------------------------
  ! definition of the generic MPI directive MPI_GATHER
  !  directly with the simple MPI directive MPI_ISEND
  !   and MPI_RECV for a 2D structure
  !-----------------------------------------------------  
  subroutine comm_gather2D(buf_source,nbelts,start,stride, &
    cible,buf_cible,comm) 
    real(RKIND), dimension(0:,0:), intent(in)  :: buf_source
    integer                      , intent(in)  :: nbelts
    integer                      , intent(in)  :: start
    integer                      , intent(in)  :: stride
    integer                      , intent(in)  :: cible
    real(RKIND), dimension(0:,0:), intent(out) :: buf_cible
    integer                      , intent(in)  :: comm

    integer :: gath_reqs, tag, nbreqr, ierr, pid, curs, nbp, mypid
    integer, &
      dimension(0:Nbproc_tot-1)                   :: gath_reqr
    integer, &
      dimension(1:MPI_STATUS_SIZE,0:Nbproc_tot-1) :: gath_status

    call MPI_COMM_SIZE(comm,nbp,ierr)
    call MPI_COMM_RANK(comm,mypid,ierr)
    tag = 4001 
    if (mypid .eq. cible) then
      nbreqr = 0
      do pid = 0,nbp-1
        curs = pid*stride
        call MPI_IRECV(buf_cible(0,curs),nbelts,MPI_REAL8,&
          pid,tag,comm,gath_reqr(pid),ierr)
      enddo
    endif
    call MPI_ISEND(buf_source(0,start),nbelts,MPI_REAL8, &
      cible,tag,comm, gath_reqs, ierr)

    if (mypid .eq. cible) &
      call MPI_WAITALL(nbp, gath_reqr, gath_status, ierr)
    call MPI_WAIT(gath_reqs, gath_status(:,1), ierr)
  end subroutine comm_gather2D

end module MPIutils_module
