// *******************************************************************
//  Copyright Euratom-CEA
//  Authors : 
//     Virginie Grandgirard (virginie.grandgirard@cea.fr)
//     Chantal Passeron (chantal.passeron@cea.fr)
//     Guillaume Latu (guillaume.latu@cea.fr)
//     Xavier Garbet (xavier.garbet@cea.fr)
//     Philippe Ghendrih (philippe.ghendrih@cea.fr)
//     Yanick Sarazin (yanick.sarazin@cea.fr)
//  
//  This code GYSELA (for GYrokinetic SEmi-LAgrangian) 
//  is a 5D gyrokinetic global full-f code for simulating 
//  the plasma turbulence in a tokamak.
//  
//  This software is governed by the CeCILL-B license 
//  under French law and abiding by the rules of distribution 
//  of free software.  You can  use, modify and redistribute 
//  the software under the terms of the CeCILL-B license as 
//  circulated by CEA, CNRS and INRIA at the following URL
//  "http://www.cecill.info". 
// *******************************************************************

#include "config-c.h"

#include <math.h>
#include <mpi.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>

void* FORTRAN_MPI_IN_PLACE = 0;

inline double max(double v1, double v2)
{
	if ( v1 > v2 ) {
		return v1;
	} else {
		return v2;
	}
}

inline double min(double v1, double v2)
{
	if ( v1 < v2 ) {
		return v1;
	} else {
		return v2;
	}
}

/** Initialize the DET_MPI module
 * 
 * Since MPI_IN_PLACE value is different in C and Fortran and that there is no standard way to access the Fortran value from C,
 * we need to provide it in the Inititialization phase.
 * 
 * \param fortran_in_place Fortran value for the MPI_IN_PLACE constant
 */
void FORTSYM(det_mpi_init, DET_MPI_INIT)( void* fortran_mpi_in_place )
{
	FORTRAN_MPI_IN_PLACE = fortran_mpi_in_place;
}

/** Reduce operation that behaves in a deterministic fashion.
 * 
 * All parameters are the same as those of MPI_Reduce.
 */
inline int DET_Reduce(void* sendbuf, void* recvbuf, int count, MPI_Datatype datatype, MPI_Op op, int root, MPI_Comm comm)
{
	int result = MPI_SUCCESS;
  void* middlebuf = 0;
	
	if (datatype == MPI_INTEGER || datatype == MPI_LOGICAL ) { // MPI_Reduce is deterministic for those
		result = MPI_Reduce(sendbuf, recvbuf, count, datatype, op, root, comm);
		goto RETURN_NOFREE;
	}
	
	if ( FORTRAN_MPI_IN_PLACE == 0 ) {
		printf("*** Fatal error: using deterministic while not initialized\n");
		exit(1);
	}
	
	int elem_size;
	result = MPI_Type_size(datatype, &elem_size);
	if ( result != MPI_SUCCESS ) {
		printf("*** Error: Unable to determine type size in deterministic reduce\n");
		goto RETURN_NOFREE;
	}
	
	int comm_size;
	result = MPI_Comm_size(comm, &comm_size);
	if ( result != MPI_SUCCESS ) {
		printf("*** Error: Unable to determine comm size in deterministic reduce\n");
		goto RETURN_NOFREE;
	}
	
	int comm_rank;
	result = MPI_Comm_rank(comm, &comm_rank);
	if ( result != MPI_SUCCESS ) {
		printf("*** Error: Unable to determine comm rank in deterministic reduce\n");
		goto RETURN_NOFREE;
	}
	
	if ( comm_rank == root ) {
		middlebuf = malloc(elem_size * count * comm_size);
		if ( !middlebuf ) {
			printf("*** Error: Unable to allocate buffer in deterministic reduce\n");
			exit(1);
		}
	}
	
	if ( comm_rank == root && sendbuf == MPI_IN_PLACE ) {
		result = MPI_Gather(recvbuf, count, datatype, middlebuf, count, datatype, root, comm);
	} else {
		result = MPI_Gather(sendbuf, count, datatype, middlebuf, count, datatype, root, comm);
	}
	if ( result != MPI_SUCCESS ) {
		printf("*** Error: Unable to gather data in deterministic reduce\n");
		goto RETURN_FREE;
	}

	if ( comm_rank == root ) {
		memcpy(recvbuf, middlebuf, (elem_size * count));
		int rr;
		for( rr = 1; rr<comm_size; ++rr ) {
#if defined(MPI_VERSION) && defined(MPI_SUBVERSION) && (MPI_VERSION>2) && (MPI_SUBVERSION>2)
			MPI_Reduce_local(middlebuf+elem_size*count*rr, recvbuf, count, datatype, op);
#else
			int ii;
			for( ii = 0; ii<count; ++ii ) {
				int jj = ii+count*rr;
				if (op == MPI_MIN) {
					if (datatype == MPI_REAL4) {
						((float*)recvbuf)[ii] = min(((float*)recvbuf)[ii], ((float*)middlebuf)[jj]);
					} else if (datatype == MPI_REAL8) {
						((double*)recvbuf)[ii] = min(((double*)recvbuf)[ii], ((double*)middlebuf)[jj]);
					} else {
						exit(2);
					}
				} else if (op == MPI_MAX) {
					if (datatype == MPI_REAL4) {
						((float*)recvbuf)[ii] = max(((float*)recvbuf)[ii], ((float*)middlebuf)[jj]);
					} else if (datatype == MPI_REAL8) {
						((double*)recvbuf)[ii] = max(((double*)recvbuf)[ii], ((double*)middlebuf)[jj]);
					} else {
						exit(2);
					}
				} else if (op == MPI_SUM) {
					if (datatype == MPI_REAL4) {
						((float*)recvbuf)[ii] += ((float*)middlebuf)[jj];
					} else if (datatype == MPI_REAL8) {
						((double*)recvbuf)[ii] += ((double*)middlebuf)[jj];
					} else {
						exit(2);
					}
				} else if (op == MPI_PROD) {
					if (datatype == MPI_REAL4) {
						((float*)recvbuf)[ii] *= ((float*)middlebuf)[jj];
					} else if (datatype == MPI_REAL8) {
						((double*)recvbuf)[ii] *= ((double*)middlebuf)[jj];
					} else {
						exit(2);
					}
				} else {
					printf("*** Fatal Error: Unsupported MPI operation in deterministic reduce\n");
					exit(1);
				}
			}
#endif
		}
	}
	
RETURN_FREE:
	if ( comm_rank == root ) {
		free(middlebuf);
	}
	
RETURN_NOFREE:
	return result;
}

/** Fortran wrapper for DET_Reduce
 */
void FORTSYM(det_reduce, DET_REDUCE)(void* sendbuf, void* recvbuf, MPI_Fint* count, MPI_Fint* datatype, MPI_Fint* op, MPI_Fint* root, MPI_Fint* comm, MPI_Fint* ierror)
{
	if ( sendbuf == FORTRAN_MPI_IN_PLACE ) sendbuf = MPI_IN_PLACE;
	*ierror = DET_Reduce(sendbuf, recvbuf, *count, MPI_Type_f2c(*datatype), MPI_Op_f2c(*op), *root, MPI_Comm_f2c(*comm));
}

/** Reduce operation that forwards to MPI_Reduce or DET_Reduce depending on whether DETERMINISTIC_MPI is defined.
 */
void FORTSYM(my_reduce, MY_REDUCE)(void* sendbuf, void* recvbuf, MPI_Fint* count, MPI_Fint* datatype, MPI_Fint* op, MPI_Fint* root, MPI_Fint* comm, MPI_Fint* ierror)
{
	if ( sendbuf == FORTRAN_MPI_IN_PLACE ) sendbuf = MPI_IN_PLACE;
#if VERBOSITY > 0
	MPI_Comm ccomm = MPI_Comm_f2c(*comm);
	int comm_rank;
	*ierror = MPI_Comm_rank(ccomm, &comm_rank);
	if ( (*ierror) != MPI_SUCCESS ) {
		printf("*** Error: Unable to determine comm rank in my reduce\n");
		return;
	}
	if ( comm_rank == 0 ) {
		printf("*** Warning: using MY_Reduce through the wrapper function instead of the MACRO\n");
	}
#endif
#ifndef DETERMINISTIC_MPI
	*ierror = MPI_Reduce(sendbuf, recvbuf, *count, MPI_Type_f2c(*datatype), MPI_Op_f2c(*op), *root, MPI_Comm_f2c(*comm));
#else
	*ierror = DET_Reduce(sendbuf, recvbuf, *count, MPI_Type_f2c(*datatype), MPI_Op_f2c(*op), *root, MPI_Comm_f2c(*comm));
#endif
}

/** AllReduce operation that behaves in a deterministic fashion.
 * 
 * All parameters are the same as those of MPI_Allreduce.
 */
inline int DET_Allreduce(void* sendbuf, void* recvbuf, int count, MPI_Datatype datatype, MPI_Op op, MPI_Comm comm)
{
	int result = MPI_SUCCESS;
	
	result = DET_Reduce(sendbuf, recvbuf, count, datatype, op, 0, comm);
	if ( result != MPI_SUCCESS ) {
		printf("*** Error: Unable to reduce data in deterministic allreduce\n");
		return result;
	}
	
	result = MPI_Bcast(recvbuf, count, datatype, 0, comm);
	if ( result != MPI_SUCCESS ) {
		printf("*** Error: Unable to broadcast data in deterministic allreduce\n");
		return result;
	}
	
	return result;
}

/** Fortran wrapper for DET_Allreduce.
 */
void FORTSYM(det_allreduce, DET_ALLREDUCE)(void* sendbuf, void* recvbuf, MPI_Fint* count, MPI_Fint* datatype, MPI_Fint* op, MPI_Fint* comm, MPI_Fint* ierror)
{
	if ( sendbuf == FORTRAN_MPI_IN_PLACE ) sendbuf = MPI_IN_PLACE;
	*ierror = DET_Allreduce(sendbuf, recvbuf, *count, MPI_Type_f2c(*datatype), MPI_Op_f2c(*op), MPI_Comm_f2c(*comm));
}

/** AllReduce operation that forwards to MPI_Reduce or DET_Reduce depending on whether DETERMINISTIC_MPI is defined.
 */
void FORTSYM(my_allreduce, MY_ALLREDUCE)(void* sendbuf, void* recvbuf, MPI_Fint* count, MPI_Fint* datatype, MPI_Fint* op, MPI_Fint* root, MPI_Fint* comm, MPI_Fint* ierror)
{
	if ( sendbuf == FORTRAN_MPI_IN_PLACE ) sendbuf = MPI_IN_PLACE;
#if VERBOSITY > 0
	int comm_rank;
	*ierror = MPI_Comm_rank(MPI_Comm_f2c(*comm), &comm_rank);
	if ( (*ierror) != MPI_SUCCESS ) {
		printf("*** Error: Unable to determine comm rank in my allreduce\n");
		return;
	}
	if ( comm_rank == 0 ) {
		printf("*** Warning: using MY_Allreduce through the wrapper function instead of the MACRO\n");
	}
#endif
#ifndef DETERMINISTIC_MPI
  *ierror = MPI_Allreduce(sendbuf, recvbuf, *count, MPI_Type_f2c(*datatype), MPI_Op_f2c(*op), MPI_Comm_f2c(*comm));
#else
	*ierror = DET_Allreduce(sendbuf, recvbuf, *count, MPI_Type_f2c(*datatype), MPI_Op_f2c(*op), MPI_Comm_f2c(*comm));
#endif
}
