!**************************************************************
!  Copyright Euratom-CEA
!  Authors : 
!     Virginie Grandgirard (virginie.grandgirard@cea.fr)
!     Chantal Passeron (chantal.passeron@cea.fr)
!     Guillaume Latu (guillaume.latu@cea.fr)
!     Xavier Garbet (xavier.garbet@cea.fr)
!     Philippe Ghendrih (philippe.ghendrih@cea.fr)
!     Yanick Sarazin (yanick.sarazin@cea.fr)
!  
!  This code GYSELA (for GYrokinetic SEmi-LAgrangian) 
!  is a 5D gyrokinetic global full-f code for simulating 
!  the plasma turbulence in a tokamak.
!  
!  This software is governed by the CeCILL-B license 
!  under French law and abiding by the rules of distribution 
!  of free software.  You can  use, modify and redistribute 
!  the software under the terms of the CeCILL-B license as 
!  circulated by CEA, CNRS and INRIA at the following URL
!  "http://www.cecill.info". 
!**************************************************************

#include "config-F.inc"

!------------------------------------------------------------
! file : QN_solver.F90
! date : 14/01/2010
! . Compute the electrostatic potential Phi(r,theta,phi)
!    by solving the following quasi-neutrality equation: 
!   L.Phi + 1/(Z0^2*Te)[Phi-<Phi>_{flux surf}] = rho
!       
!  with the differential operator L defined as:
!   L.Phi = -1/ne0(r) sum_s hat_As * 
!            nabla_perp.[ns0(r) nabla_perp Phi] 
!         = -1/ne0(r) sum_s hat_As * ns0(r) *
!           { d^2(Phi)/dr^2 + [1/r+1/ns0 dns0/dr]*dPhi/dr
!          +1/r^2  d^2(Phi)/dth^2 }
!   and the RHS defined as:
!     rho = 1/ne0(r)*sum_s hat_Zs*[nGs-nGs_eq]
!   where nGs(r,theta,phi) = \int dmu \int Jv J0.f_s dvpar
!     and nGs_eq(r,theta)  = \int dmu \int Jv J0.feq_s dvpar     
!  Jv being the jacobian in velocity, f_s and feq_s 
!    being respectively the distribution function and 
!    equilibrium distribution function for specie s. 
!
! The flux surface average of Phi is defined by
!   <Phi>_{flux surf} = \int Phi Js(r,theta) dtheta dphi /
!                       \int Js(r,theta) dtheta dphi
!  where Js corresponds to the Jacobian in space  
!
! Rk:The boundary conditions for Phi are periodic in theta 
!    direction and Dirichlet conditions in r direction
!------------------------------------------------------------
module QNsolver_module
  use clock_module
  use prec_const
  use spline1d_class
  use OMPutils_module, only : omp2cvector
  use poisson_class
  implicit none

  private
  !-> coefficient for Neumann boundary condition
  real(RKIND) :: coeff_Neumann

  !-> FFT wave vector in theta and phi direction
  real(RKIND), dimension(:), pointer :: QN_kth
  real(RKIND), dimension(:), pointer :: QN_kphi

  !-> tridiagonal matrix
  !--> terms of the subdiagonal
  real(RKIND), dimension(:)  , pointer :: L_ldiag
  !--> terms of the diagonal
  real(RKIND), dimension(:)  , pointer :: L_diag
  real(RKIND), dimension(:)  , pointer :: L_invZ02Te_diag
  real(RKIND), dimension(:,:), pointer :: Lm_invZ02Te_diag
  !--> terms of the superdiagonal
  real(RKIND), dimension(:)  , pointer :: L_udiag
  
  !-> splines
  type(pspline1d), private :: QN_pspline1d_theta
  type(pspline1d), private :: QN_pspline1d_phi

  !-> temporary array for FFT2D
  type(omp2cvector), dimension (:), pointer :: Comp_data_i
  type(omp2cvector), dimension (:), pointer :: Comp_tmp2D

  public :: new_QNsolver, del_QNsolver, solve_quasineutrality
  
#ifdef TIMER
  real(RKIND), private  ::  tcmm(12), tbar(12), tcmp(12)
  real(RKIND), private  ::  tvect(32), tmin(32), tmax(32), tavg(32)
#endif  
  include "gyrokinetic/Bstar_inline.h"
  !******************************
  contains
  !******************************
#include "gyrokinetic/Bstar_inline.F90"


  !----------------------------------------------------- 
  ! constructor for the quasi-neutrality solver
  !-----------------------------------------------------   
  subroutine new_QNsolver(geom,init_prof)
    use globals, only : Nbthread, memory_test, QN_comm_version, &
                    pSpecies_id, pInSpecies_id, Nbproc_species, Nphi, &
                    Phi00_BCrmin_Neumann
    use geometry_class
    use init_profile_class
!R3 #include "code_instrumentation/r3_info.h" !R3
    type(geometry)    , intent(in) :: geom
    type(init_profile), intent(in) :: init_prof

    integer :: tid
    integer :: N1, N2, N3
!R3 call r3_info_begin (r3_info_index_1, 'Init_QNsolver') !R3
    call Write_dyn_balise_open('Init_QNsolver') !R4
    !*** allocation and construction of the matrices ***
    N1 = (geom%Nr+1)-2
    N2 = (geom%Ntheta+1)-1
    N3 = (geom%Nphi+1)-1
    if (Phi00_BCrmin_Neumann) then
      coeff_Neumann = 1._RKIND
    else
      coeff_Neumann = 0._RKIND
    end if
    call write_expr("N1", "(geom%Nr+1)-2")
    call write_expr("N2", "(geom%Ntheta+1)-1")
    call write_expr("N3", "(geom%Nphi+1)-1")
    call allocate_QNmatrices(N1,N2,N3)
    if (.not.memory_test) &
      call construct_QNmatrices(N1,N2,N3,geom,init_prof)

    !*** construction of the splines required to compute   ***
    !***  the integral in theta and phi with cubic splines ***
    call write_expr("n", "geom%Ntheta")
    call new_spline1d_period(QN_pspline1d_theta, &
         geom%Ntheta,geom%dtheta)
    call write_expr("n", "geom%Nphi")
    call new_spline1d_period(QN_pspline1d_phi, &
         geom%Nphi,geom%dphi)

    !*** allocation of arrays for OMP parallelisation ***
    allocate(Comp_tmp2D(1:Nbthread))
    allocate(Comp_data_i(1:Nbthread))

    if ( QN_comm_version==3 .and. NPhi>Nbproc_species ) then
       if ( pInSpecies_id == 0 ) then
          print *, pSpecies_id, 'WARNING: QN_comm_version==3 .and.'&
               &'NPhi > Nbproc_species, faling back to 2', &
               NPhi, Nbproc_species
       end if
       QN_comm_version = 2
    end if

    call Write_dyn_balise_close('Init_QNsolver') !R4
!R3 call r3_info_end (r3_info_index_1) !R3
  end subroutine new_QNsolver


  !----------------------------------------------------- 
  ! destructor for the quasi-neutrality solver
  !-----------------------------------------------------   
  subroutine del_QNsolver()

    !-> deallocation of the matrix diagonals
    call deallocate_QNmatrices()

    !-> deallocation of the cubic splines
    call del_spline1d_period(QN_pspline1d_theta)
    call del_spline1d_period(QN_pspline1d_phi)
    !-> deallocation of arrays for OMP parallelisation
    deallocate(Comp_tmp2D)
    deallocate(Comp_data_i)
  end subroutine del_QNsolver
  

  !----------------------------------------------------- 
  ! Array allocation for the tridiagonal matrices
  !-----------------------------------------------------   
  subroutine allocate_QNmatrices(N1,N2,N3)
    use wrap_alloc_module
    integer, intent(in) :: N1
    integer, intent(in) :: N2
    integer, intent(in) :: N3

    !-> for the FFT wave vector in theta and phi direction
    call FTR_ALLOC_1_ZERO(QN_kth,1,N2)
    call FTR_ALLOC_1_ZERO(QN_kphi,1,N3)

    !-> for the tridiagonal matrices
    call FTR_ALLOC_1_ZERO(L_ldiag,1,N1)
    call FTR_ALLOC_1_ZERO(L_diag,1,N1)
    call FTR_ALLOC_1_ZERO(L_invZ02Te_diag,1,N1)
    call FTR_ALLOC_2_ZERO(Lm_invZ02Te_diag,1,N1,1,N2)
    call FTR_ALLOC_1_ZERO(L_udiag,1,N1)
  end subroutine allocate_QNmatrices


  !----------------------------------------------------- 
  ! Array deallocation of the tridiagonal matrices
  !-----------------------------------------------------   
  subroutine deallocate_QNmatrices()
    use wrap_alloc_module

    !-> for the FFT wave vector in theta and phi direction
    call FTR_DEALLOC(QN_kth)
    call FTR_DEALLOC(QN_kphi)

    !-> for the tridiagonal matrices
    call FTR_DEALLOC(L_ldiag)
    call FTR_DEALLOC(L_diag)
    call FTR_DEALLOC(L_invZ02Te_diag)
    call FTR_DEALLOC(Lm_invZ02Te_diag)
    call FTR_DEALLOC(L_udiag)
  end subroutine deallocate_QNmatrices


  !----------------------------------------------------- 
  ! Construction of the tridiagonal matrices
  !  requires for the solving of the quasi-neutrality 
  !  equation 
  ! Rk : Due to the fact that finite differences of 
  !  second order are used to approximate the
  !  polarization term given by: 
  !   -1/ne0(r) sum_s hat_As * ns0(r) *
  !     { d^2(Phi)/dr^2 + [1/r+1/ns0 dns0/dr]*dPhi/dr
  !       + 1/r^2  d^2(Phi)/dth^2 }
  !-----------------------------------------------------   
  subroutine construct_QNmatrices(N1,N2,N3,geom,init_prof)
    use globals, only : Z0, QN_coef_polarization, &
      Nspecies, hat_As, QN_species_impact
    use geometry_class
    use init_profile_class
    use fft_module, only : NR_fft_vector
    integer           , intent(in) :: N1
    integer           , intent(in) :: N2
    integer           , intent(in) :: N3
    type(geometry)    , intent(in) :: geom
    type(init_profile), intent(in) :: init_prof
    
    integer     :: ispec, i1, i2
    integer     :: indx_grid, ir
    real(RKIND) :: hatAs, ns0_on_ne0, QN_spec_impact
    real(RKIND) :: inv_dr2, inv_Z02Te
    real(RKIND) :: ri, alpha_ri, beta_ri
    real(RKIND) :: Ldiag_tmp
    real(RKIND) :: Z02

    !*** compute kth for the FFT in theta direction ***
    call NR_fft_vector(N2,geom%dtheta,QN_kth)
    !*** compute kphi for the FFT in phi direction ***
    call NR_fft_vector(N3,geom%dphi,QN_kphi)

    indx_grid = 0  ! (because geom%rg(0:Nr))
    inv_dr2   = 1._RKIND/(geom%dr*geom%dr)
    Z02       = Z0*Z0

    !*****************************************************
    !*** initialisation of the terms of three          ***
    !***   diagonals of the matrix where               ***
    !*** alpha(r) = sum_s hat_As*(ns0(r)/ne0(r) *      ***
    !***             (1/r+1/ns0(r)*dns0(r)/dr)         ***
    !*** beta(r)  = sum_s hat_As*(ns0(r)/ne0(r)        ***
    !*****************************************************
    do i1 = 1,N1
      ir          = indx_grid+i1
      ri          = geom%rg(ir)
      alpha_ri    = 0._RKIND
      beta_ri     = 0._RKIND
      do ispec = 0,Nspecies-1
        hatAs          = hat_As(ispec)
        QN_spec_impact = QN_species_impact(ispec)
        ns0_on_ne0     = init_prof%ns0(ispec,ir)/init_prof%ne0(ir) 
        alpha_ri       = alpha_ri + &
          QN_spec_impact*hatAs*ns0_on_ne0 * &
        (1._RKIND/ri + init_prof%dns0dr(ispec,ir)/init_prof%ns0(ispec,ir))
        beta_ri    = beta_ri + QN_spec_impact*hatAs*ns0_on_ne0 
      end do
      inv_Z02Te   = 1._RKIND/(Z02*init_prof%Te0(ir))
      L_ldiag(i1) = - QN_coef_polarization * &
        (beta_ri*inv_dr2 - alpha_ri/(2._RKIND*geom%dr)) 
      L_udiag(i1) = - QN_coef_polarization * &
        (beta_ri*inv_dr2 + alpha_ri/(2._RKIND*geom%dr))
      Ldiag_tmp   = QN_coef_polarization*beta_ri*(2._RKIND*inv_dr2)
      L_diag(i1)  = Ldiag_tmp 
      L_invZ02Te_diag(i1) = Ldiag_tmp + inv_Z02Te
      do i2 = 1,N2
        Lm_invZ02Te_diag(i1,i2) = L_invZ02Te_diag(i1) + &
          QN_coef_polarization*beta_ri*((QN_kth(i2)*QN_kth(i2))/(ri*ri))
      end do
    end do
    L_diag(1)   = L_diag(1) + coeff_Neumann*L_ldiag(1)
    L_ldiag(1)  = 0._RKIND
    L_udiag(N1) = 0._RKIND
  end subroutine construct_QNmatrices
  

  !----------------------------------------------------------- 
  ! Modified version of Thomas algorithm to solve a 
  !  tridiagonal system of the form A.X = B
  !  where A is the tridiagonal NxN matrix while 
  !  X and B are NxNRHS matrices.
  ! 
  ! -> Factorization
  !     . gamma(0)=0 and l(1)=0
  !     . for i=1,...,N
  !         gamma(i) = (d(i)-l(i)*gamma(i-1)*u(i-1))^{-1}
  ! -> Forward substitution (Ly=b)
  !     . y(1)=gamma(1)*b(1)
  !     . for i=2,...,N
  !         y(i)=gamma(i)*(b(i)-l(i)*y(i-1))
  ! -> Backward substitution (Ux=y)
  !     . x(N)=y(N)
  !     . for i=N-1,...,1
  !         x(i)=y(i)-gamma(i)*u(i)*x(i+1)
  !
  ! Arguments
  ! =========
  !  . N     : (input) INTEGER 
  !            The order of the matrix A. N>=0.
  !  . ldiag : (input) REAL(RKIND), dimension(N) array
  !            Defined as [0,l1,...,LN] where li 
  !            are the (N-1) subdiagonal elements of
  !            the tridiagonal matrix 
  !  . diag  : (input) REAL(RKIND) array, dimension(N)
  !            The N diagonal elements of the tridiagonal
  !            matrix A
  !  . udiag : (input) REAL(RKIND) array, dimension(N) 
  !            Defined as [u1,...,u_{N-1}] where ui 
  !            are the (N-1) superdiagonal elements of
  !            the tridiagonal matrix A
  !  . rhs   : (input/output) COMPLEX(CKIND) array, dimension(N)
  !            On entry, the right hand side vectors B for 
  !            the system of linear equations.
  !            On exit, the solution vextors, X
  !----------------------------------------------------------- 
  subroutine modThomas_c(N,ldiag,diag,udiag,rhs,err)
    integer                      , intent(in)    :: N
    real(RKIND)   , dimension(1:), intent(in)    :: udiag
    real(RKIND)   , dimension(1:), intent(in)    :: ldiag
    real(RKIND)   , dimension(1:), intent(in)    :: diag
    complex(CKIND), dimension(1:), intent(inout) :: rhs
    integer                      , intent(out)   :: err
    
    integer :: i
    complex(CKIND), dimension(1:N) :: gamma
    complex(CKIND), dimension(1:N) :: y

    err = 0
    if (diag(1).eq.0) then
      err = 1
      return
    end if

    !-> Factorisation + Forward substitution (Ly=rhs)
    gamma(1) = 1._RKIND/diag(1)
    y(1)     = gamma(1)*rhs(1)
    do i = 2,N
      gamma(i) = 1._RKIND / &
        (diag(i)-ldiag(i)*gamma(i-1)*udiag(i-1))
      y(i)     = gamma(i)*(rhs(i)-ldiag(i)*y(i-1))
    end do
    !-> Backward substitution (Ux=y)
    rhs(N) = y(N)
    do i = N-1,1,-1
      rhs(i) = y(i) - gamma(i)*udiag(i)*rhs(i+1)
    end do
  end subroutine modThomas_c


  !----------------------------------------------------------- 
  ! Modified version of Thomas algorithm to solve a 
  !  tridiagonal system of the form A.X = B
  !  where A is the tridiagonal NxN matrix while 
  !  X and B are NxNRHS matrices.
  ! 
  ! -> Factorization
  !     . gamma(0)=0 and l(1)=0
  !     . for i=1,...,N
  !         gamma(i) = (d(i)-l(i)*gamma(i-1)*u(i-1))^{-1}
  ! -> Forward substitution (Ly=b)
  !     . y(1)=gamma(1)*b(1)
  !     . for i=2,...,N
  !         y(i)=gamma(i)*(b(i)-l(i)*y(i-1))
  ! -> Backward substitution (Ux=y)
  !     . x(N)=y(N)
  !     . for i=N-1,...,1
  !         x(i)=y(i)-gamma(i)*u(i)*x(i+1)
  !
  ! Arguments
  ! =========
  !  . N     : (input) INTEGER 
  !            The order of the matrix A. N>=0.
  !  . ldiag : (input) REAL(RKIND), dimension(N) array
  !            Defined as [0,l1,...,LN] where li 
  !            are the (N-1) subdiagonal elements of
  !            the tridiagonal matrix 
  !  . diag  : (input) REAL(RKIND) array, dimension(N)
  !            The N diagonal elements of the tridiagonal
  !            matrix A
  !  . udiag : (input) REAL(RKIND) array, dimension(N) 
  !            Defined as [u1,...,u_{N-1}] where ui 
  !            are the (N-1) superdiagonal elements of
  !            the tridiagonal matrix A
  !  . rhs   : (input/output) REAL(RKIND) array, dimension(N)
  !            On entry, the right hand side vectors B for 
  !            the system of linear equations.
  !            On exit, the solution vextors, X
  !----------------------------------------------------------- 
  subroutine modThomas_d(N,ldiag,diag,udiag,rhs,err)
    integer                   , intent(in)    :: N
    real(RKIND), dimension(1:), intent(in)    :: udiag
    real(RKIND), dimension(1:), intent(in)    :: ldiag
    real(RKIND), dimension(1:), intent(in)    :: diag
    real(RKIND), dimension(1:), intent(inout) :: rhs
    integer                   , intent(out)   :: err
    
    integer     :: i
    real(RKIND) :: denom_tmp
    real(RKIND), dimension(1:N) :: gamma
    real(RKIND), dimension(1:N) :: y

    err = 0
    if (diag(1).eq.0) then
      err = 1
      return
    end if

    !-> Factorisation + Forward substitution (Ly=rhs)
    gamma(1) = 1._RKIND/diag(1)
    y(1)     = gamma(1)*rhs(1)
    do i = 2,N
      denom_tmp = (diag(i)-ldiag(i)*gamma(i-1)*udiag(i-1))
      if (abs(denom_tmp).le.1.e-12) then
        print*,'Problem denom_tmp = ',denom_tmp
        err = 1
        return
      end if
      gamma(i) = 1._RKIND / denom_tmp 
      y(i)     = gamma(i)*(rhs(i)-ldiag(i)*y(i-1))
    end do
    !-> Backward substitution (Ux=y)
    rhs(N) = y(N)
    do i = N-1,1,-1
      rhs(i) = y(i) - gamma(i)*udiag(i)*rhs(i+1)
    end do
  end subroutine modThomas_d


  !------------------------------------------------------------   
  ! Compute rho the RHS of the quasi-neutrality 
  !  equation, i.e:
  !    rho(r,theta,phi) = sum_s rhos(r,theta,phi)
  !  where for each specie s
  !    rhos(r,theta,phi) = (1/ne0(r))* hat_Zs * 
  !                        (nGs(r,theta,phi)-nGs_eq(r,theta))
  !  where nGs = \int J0(sqrt{2*mu}).f_s Jv  dvpar dmu
  !            = \int J0(sqrt{2*mu}).[\int f_s Jv dvpar] dmu 
  !  with  nGs_eq(r,theta) = \int J0.feq_s(r,theta,vpar,mu) 
  !                          Jv(r,theta,vpar)dmu dvpar
  !  Jv being the Jacobian in velocity space :
  !   . Jv(r,theta,vpar) = 2*pi*Bstar in the 5D case
  !                      = 1          in the 4D case
  !  f_s being the distribution function of specie s and
  !
  ! Rk : rhos is computed as the sum of 2 integrals:
  !   Is =  (1/ne0(r))*hat_Zs* \int dmu
  !            (I0*B_norm + (sqrt{hat_As}/hat_Zs)* 
  !            I1*mu0*vec_J.vec_b/B_norm)
  !   where I0 = J0.(\int f_s dvpar) and          
  !         I1 = J0.(\int f_s*vpar dvpar)
  !   this by using the expresion of Bstar
  !    Bstar = B_norm + (sqrt{hat_As}/hat_Zs) * 
  !            vpar*mu0*vec_J.vec_b/B_norm
  !------------------------------------------------------------
  subroutine compute_rho(geom,J0,f,Sfmu_eq, &
    init_prof,init_magnet,init_curr,QN_rhs)
    use globals
#ifdef _OPENMP
    use OMPutils_module, only : omp_get_thread_num
#endif
    use OMPutils_module, only : Romp1_0Nr, Romp2_0Nr, Romp3_0Nr, &
      Romp1_1Nrp1, Romp2_1Nrp1, Romp3_1Nrp1, Romp4_1Nrp1, &
      Comp1_1Nrp1_1Ntheta, Romp10_1Nrp1_1Nthetap1, &
      Romp1_1Nrp1_1Nthetap1, Romp2_1Nrp1_1Nthetap1, &
      Romp1_0Nvpar, Romp2_0Nvpar, &
      Romp_scoefvpar, Comp_0Nr, Comp2_1Nrp1_1Ntheta,&
      Romp3_1Nrp1_1Nthetap1, Romp4_1Nrp1_1Nthetap1
    use geometry_class
    use gyroaverage_class
    use fdistribu5d_class
    use init_profile_class
    use init_magnetic_class
    use init_current_class
    use integration_module, only : &
      compute_omp_vpar_integral_CS, compute_vpar_integral_colloc
    use mpiMod
    type(geometry)               , intent(in)    :: geom
    type(J0operator)             , intent(inout) :: J0
    type(fdistribu5d)            , intent(in)    :: f
    real(RKIND), dimension(:,:,:), pointer       :: Sfmu_eq
    type(init_profile)           , intent(in)    :: init_prof
    type(init_magnetic)          , intent(in)    :: init_magnet
    type(init_current)           , intent(in)    :: init_curr
    real(RKIND), dimension(:,:,:), pointer       :: QN_rhs

#ifdef HAS_QN_V4
    integer, parameter :: nbmoments = 2
    real(RKIND)   , dimension(:)    , pointer :: fmu_1D
    real(RKIND)   , dimension(:)    , pointer :: fmu_v_1D
    real(RKIND)   , dimension(:)    , pointer :: scoefvpar
    real(RKIND)   , dimension(:,:)  , pointer :: J0_intdv_f
    real(RKIND)   , dimension(:,:)  , pointer :: J0_intvdv_f
    real(RKIND)   , dimension(:,:)  , pointer :: vscalprod_mu0Jb_ij
    real(RKIND)   , dimension(:,:)  , pointer :: vBnorm_ij
    real(RKIND)   , dimension(:,:,:), pointer :: M0_loc
    complex(CKIND), dimension(:,:)  , pointer :: Acomp
    real(RKIND)   , dimension(:)    , pointer :: ne0

    !-> used for 2 species case
    real(RKIND) :: QN_spec_impact
    real(RKIND) :: hatZs, sqrtAs_on_Zs
    real(RKIND), dimension(:,:,:), pointer :: QN_rhs_tmp

    integer :: ierr, no_operation
    integer :: ir, itheta, iphi, ivpar, imu, iphiloc
    integer :: isign, midn2, N1, i, j, k, err

    logical     :: case4D
    real(RKIND) :: fmu_tmp, jac_v
    real(RKIND) :: Bnorm_ij, scalprod_mu0Jb_ij
    real(RKIND) :: coeff_intdmu

    ! -> variable for parallelization
    integer     :: nbreqr, nbbcast_r, nbbcast_s
    integer     :: local_id, dep, base, global_id, startbuf
    integer     :: l_istart, l_iend, l_jstart, l_jend
    integer     :: iproc, itag_phi, nbelements, tid
    
    !-> used for collective version
    integer, dimension(0:Nbproc_species-1) :: &
                    send_count_vec, send_disp_vec, &
                    recv_count_vec, recv_disp_vec
    integer :: Rarray_PNrPNthetaNphi_nbM2_shape(4)

#ifdef TIMER
    integer(TIMEPREC) :: tdeb, tfin, tth1, tth2
    tcmp(1:6) = 0.
    tcmm(1:6) = 0.
    tbar(1:6) = 0.

    call clck_time(tdeb)
    tth1 = tdeb
#endif

    do tid = 1, Nbthread
       call FTR_ALLOC_2(Romp1_1Nrp1_1Nthetap1(tid)%val,1,Nr+1,1,Ntheta+1)
       call FTR_ALLOC_2(Romp2_1Nrp1_1Nthetap1(tid)%val,1,Nr+1,1,Ntheta+1)
       call FTR_ALLOC_2(Romp3_1Nrp1_1Nthetap1(tid)%val,1,Nr+1,1,Ntheta+1)
       call FTR_ALLOC_2(Romp4_1Nrp1_1Nthetap1(tid)%val,1,Nr+1,1,Ntheta+1)

       call FTR_ALLOC_2_ZERO(Comp1_1Nrp1_1Ntheta(tid)%val,1,Nr+1,1,Ntheta)
    enddo

    do tid = 0, MAX(Nmu,Nbthread)
       call FTR_ALLOC_2(Romp10_1Nrp1_1Nthetap1(tid)%val,1,Nr+1,1,Ntheta+1)
    enddo

   call FTR_ALLOC_1_ZERO(RRecvBufMPI_A,0,nbmoments*dom_r*dom_theta*Nbproc_loc*(Nmu+1)*dom_mapphi)

    case4D = .false.
    if (geom%Nmu.eq.0) case4D = .true.

    !--> Species impact in QN solver (for test)
    QN_spec_impact = QN_species_impact(pSpecies_id)

    !--> compute sqrt{hat_As}/hat_Zs
    hatZs        = hat_Zs(pSpecies_id)
    sqrtAs_on_Zs = sqrt(hat_As(pSpecies_id))/hatZs

    !*** initialize local variables ***
    ne0    => init_prof%ne0
    M0_loc => Rarray1_NrNthetaDomphi

    !*****************************************************
    !*** Computation of the RHS as:                    ***
    !***  rho(r,theta,phi) = sum_s rhos(r,theta,phi)   ***
    !*** where for each specie s                       ***
    !***  rhos(r,theta,phi)= (1/ne0(r))*hat_Zs *       ***
    !***          (nGs(r,theta,phi)-nGs_eq(r,theta))   ***
    !*** with Jv = 2*pi*Bstar being the jacobian       ***
    !*** in velocity and                               *** 
    !***   nGs_eq = \int J0.feq_s Jv dvpar dmu and     ***
    !***   nGs    = \int J0.f_s Jv dvpar dmu           ***
    !***                                               ***
    !*** Rk: We use the fact that                      ***
    !***     rhos = (1/ne0(r))*hat_Zs *                ***
    !***            \int dmu (I0*B_norm +              ***
    !***              I1*(sqrt{hat_As}/hat_Zs)*        ***
    !***              mu0*vec_J.vec_b/(e*B_norm))      ***
    !***   where I0 = J0.(\int (f_s-feq_s) dvpar) and  ***
    !***         I1 = J0.(\int (f_s-feq_s)*vpar dvpar) *** 
    !*** This by using the fact that                   ***
    !***    Bstar = B_norm + (sqrt{hat_As}/hat_Zs)*    ***
    !***            vpar*mu0*vec_J.vec_b/B_norm        ***
    !******************************************************
!$OMP PARALLEL default(none) &
!$OMP private (tid,imu,iphi,itheta,ir,ivpar, &
!$OMP  scoefvpar,fmu_tmp,fmu_1D,fmu_v_1D,jac_v) &
!$OMP shared(f,integration_CS,Romp1_0Nvpar, &
!$OMP  Romp2_0Nvpar,Rarray_PNrPNthetaNphi_nbM2, &
!$OMP  Romp_scoefvpar,Sfmu_eq,geom)
#ifdef _OPENMP
    tid = 1+omp_get_thread_num()
#else
    tid = 1
#endif
    scoefvpar => Romp_scoefvpar(tid)%val
    fmu_1D    => Romp1_0Nvpar(tid)%val
    fmu_v_1D  => Romp2_0Nvpar(tid)%val

!$OMP BARRIER
!$OMP DO SCHEDULE(STATIC)
    do iphi = 0,f%n3-1
      do itheta = f%jstart,f%jend
        do ir = f%istart,f%iend
          do ivpar = 0,f%n4
            if (geom%Nmu .eq. 0) then
              call precomp_Bstar(ir,itheta,ivpar,jac_v)
            else
              jac_v = 1
            end if
            fmu_tmp         = jac_v*(f%values(ir,itheta,iphi,ivpar) - &
              Sfmu_eq(ir,itheta,ivpar))
            fmu_1D(ivpar)   = fmu_tmp
            fmu_v_1D(ivpar) = fmu_tmp*geom%vparg(ivpar)
          end do
          !*** computation of the integrals in vpar direction ***
          if (integration_CS) then
            !-> \int (f-feq) dvpar
            call compute_omp_vpar_integral_CS(f%n4,f%h4,fmu_1D, &
              f%BCvpar_left,f%BCvpar_right, &
              f%nspline1d_vpar(tid),scoefvpar, &
              Rarray_PNrPNthetaNphi_nbM2(0,ir,itheta,iphi))
            !-> \int (f-feq) vpar dvpar
            call compute_omp_vpar_integral_CS(f%n4,f%h4,fmu_v_1D, &
              f%BCvpar_left,f%BCvpar_right, &
              f%nspline1d_vpar(tid),scoefvpar, &
              Rarray_PNrPNthetaNphi_nbM2(1,ir,itheta,iphi))
          else
            !-> \int (f-feq) dvpar
            call compute_vpar_integral_colloc(fmu_1D(0:f%n4), &
              geom,Rarray_PNrPNthetaNphi_nbM2(0,ir,itheta,iphi))
            !-> \int (f-feq) vpar dvpar
            call compute_vpar_integral_colloc(fmu_v_1D(0:f%n4), &
              geom,Rarray_PNrPNthetaNphi_nbM2(1,ir,itheta,iphi))
          end if
        enddo
      enddo
    end do
!$OMP END DO
!$OMP BARRIER
!$OMP END PARALLEL    

    !-> message receive
    if (iphistart .ne. -1) then
      nbreqr  = 0
      iphiloc = 0
      do iphi = 0,f%n3-1
        if (moments_mapphi(iphi).eq.pInSpecies_id) then
          do imu = 0, Nmu
            do local_id = 0,Nbproc_loc-1
              global_id  = imu*Nbproc_loc + local_id
              startbuf   = dom_theta * (local_id + Nbproc_loc * &
                (imu + (Nmu+1) * iphiloc)) 
              itag_phi   = 4000 + iphi + pSpecies_id*f%n3
              nbelements = nbmoments * dom_r * dom_theta
              select case (QN_comm_version)
              case (1,2)
                call MPI_IRECV( &
                  RRecvBufMPI_A(startbuf*dom_r*nbmoments), &
                  nbelements,MPI_REAL8,global_id,itag_phi, &
                  mpi_comm_species,moments_reqr(nbreqr),ierr)
              case (3)
                recv_count_vec(global_id) = nbelements
                recv_disp_vec(global_id) = &
                                startbuf*dom_r*nbmoments
              case default
                if (pglobal_id.eq.0) then
                    print*, 'Parameter QN_comm_version ',QN_comm_version,' is not set at a known value.'
                    stop
                end if
              end select
              nbreqr = nbreqr + 1
            enddo
          enddo
          iphiloc = iphiloc + 1
        endif
      enddo
    end if
    Rarray_PNrPNthetaNphi_nbM2_shape = shape(Rarray_PNrPNthetaNphi_nbM2)
    !-> message sending 
    do iphi = 0,f%n3-1
      itag_phi   = 4000 + iphi + pSpecies_id*f%n3
      nbelements = nbmoments*dom_r*dom_theta
      select case (QN_comm_version)
      case (1,2)
        call MPI_SEND(& 
          Rarray_PNrPNthetaNphi_nbM2(0,istart,jstart,iphi),&
          nbelements,MPI_REAL8,moments_mapphi(iphi),itag_phi,&
          mpi_comm_species,ierr)
      case (3)
        send_count_vec(moments_mapphi(iphi)) = nbelements
        send_disp_vec(moments_mapphi(iphi)) = &
                        0+ &
                        istart*Rarray_PNrPNthetaNphi_nbM2_shape(1)+ &
                        jstart*Rarray_PNrPNthetaNphi_nbM2_shape(2)+ &
                        iphi*Rarray_PNrPNthetaNphi_nbM2_shape(3)
      case default
        if (pglobal_id.eq.0) then
            print*, 'Parameter QN_comm_version ',QN_comm_version,' is not set at a known value.'
            stop
        end if
      end select
    end do
    select case (QN_comm_version)
    case (1,2)
      if ( iphistart .ne. -1 ) then
        call MPI_WAITALL(nbreqr, moments_reqr(0:nbreqr-1), &
                moments_status(:,0:nbreqr-1), ierr)
      end if
    case (3)
      if ( nbreqr /= nbmoments .and. pglobal_id==0 ) then
        print*, 'Cannot use alltoall, nbreqr /= nbmoments', &
                        nbreqr, nbmoments
        stop
      end if
      call MPI_alltoallv(Rarray_PNrPNthetaNphi_nbM2, send_count_vec, &
                      send_disp_vec, MPI_REAL8, RRecvBufMPI_A, &
                      recv_count_vec, recv_disp_vec, MPI_REAL8, &
                      mpi_comm_species, ierr)
    end select

#ifdef TIMER
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tcmm(1))
    tdeb = tfin
#endif

    if (iphistart .ne. -1) then
       do tid = 1, Nbthread
          call FTR_ALLOC_2(J0%omp_diag(tid)%val,1,J0%N1,1,J0%N2)
          call FTR_ALLOC_2(J0%omp_ldiag(tid)%val,1,J0%N1-1,1,J0%N2)
          call FTR_ALLOC_2(J0%omp_udiag(tid)%val,1,J0%N1-1,1,J0%N2)
      enddo
!$OMP PARALLEL default(none) &
!$OMP private(nbbcast_s,tid,imu,iphi,vscalprod_mu0Jb_ij, &
!$OMP  vBnorm_ij,coeff_intdmu,J0_intdv_f,J0_intvdv_f, &
!$OMP  nbelements,local_id,base,dep,l_istart, &
!$OMP  l_iend,l_jstart,l_jend,startbuf,itheta,ir, &
!$OMP  iphiloc,Bnorm_ij,scalprod_mu0Jb_ij,Acomp,isign,&
!$OMP  j,i,err,midn2,no_operation) &
!$OMP shared(geom,J0,f,init_magnet,init_curr,ne0, &
!$OMP  Ntheta,Nmu,moments_mapphi,pglobal_id, &
!$OMP  iphistart,Nbproc_loc,Nbproc_r,dom_r,dom_theta, &
!$OMP  M0_loc,case4D, &
!$OMP  pInSpecies_id,sqrtAs_on_Zs,hatZs,QN_spec_impact, &
!$OMP  Romp1_1Nrp1_1Nthetap1,Romp2_1Nrp1_1Nthetap1, &
!$OMP  Romp3_1Nrp1_1Nthetap1,Romp4_1Nrp1_1Nthetap1, &
!$OMP  Romp10_1Nrp1_1Nthetap1,RRecvBufMPI_A, &
!$OMP  Comp1_1Nrp1_1Ntheta)
!$OMP BARRIER
#ifdef _OPENMP
      tid = 1+omp_get_thread_num()
#else
      tid = 1
#endif
      J0_intdv_f         => Romp1_1Nrp1_1Nthetap1(tid)%val
      J0_intvdv_f        => Romp2_1Nrp1_1Nthetap1(tid)%val
      Acomp              => Comp1_1Nrp1_1Ntheta(tid)%val
      vscalprod_mu0Jb_ij => Romp3_1Nrp1_1Nthetap1(tid)%val
      vBnorm_ij          => Romp4_1Nrp1_1Nthetap1(tid)%val
      do itheta = 0, geom%Ntheta
        do ir = 0, geom%Nr
          call compute_scalprod_mu0Jb(geom,init_magnet, &
            init_curr,ir,itheta,vscalprod_mu0Jb_ij(1+ir,1+itheta))
          vBnorm_ij(1+ir,1+itheta) = init_magnet%B_norm(ir,itheta) 
        end do
      end do
      err   = 0
      isign = 1      
      midn2 = geom%Ntheta/2
      
      do iphi = 0,geom%Nphi-1
        if (moments_mapphi(iphi) .eq. pInSpecies_id) then
!$OMP BARRIER
          iphiloc = iphi - iphistart
          !*** Parallelization in variable imu in order ***
          !***  to speed up computation                 ***
          !*** whenever Nmu is large                    ***
!$OMP DO SCHEDULE(STATIC,1)
          do imu = 0,Nmu
            coeff_intdmu = geom%coeff_intdmu(imu)
            do itheta = 0,geom%Ntheta
              do ir = 0,geom%Nr
                J0_intdv_f(1+ir,1+itheta)  = 0._RKIND
                J0_intvdv_f(1+ir,1+itheta) = 0._RKIND
              end do
            end do
            do local_id = 0,Nbproc_loc-1
              base     = (local_id/Nbproc_r)
              dep      = mod(local_id,Nbproc_r)              
              l_istart = dep *  dom_r
              l_iend   = l_istart + dom_r - 1
              l_jstart = base * dom_theta
              l_jend   = l_jstart + dom_theta - 1
              startbuf = dom_theta * (local_id + Nbproc_loc * &
                (imu + (Nmu+1) * iphiloc)) 
              !*** J0_intdv_f  = (\int f dvpar) and  ***
              !*** J0_intvdv_f = (\int f vpar dvpar) *** 
              do itheta = l_jstart,l_jend
                do ir = l_istart,l_iend
                  J0_intdv_f(1+ir,1+itheta)  = &
                    RRecvBufMPI_A( &
                    nbmoments*(ir-l_istart)+0+ &
                    (startbuf+itheta-l_jstart)*dom_r*nbmoments) 
                  J0_intvdv_f(1+ir,1+itheta) = &
                    RRecvBufMPI_A( &
                    nbmoments*(ir-l_istart)+1+ &
                    (startbuf+itheta-l_jstart)*dom_r*nbmoments) 
                enddo
              enddo
            enddo
            !*** computation of the gyroaverages ***             
            !-> J0.(\int (f-feq) dvpar)
            call omp_compute_gyrofunction_2D(J0,geom,imu, &
              Acomp,J0_intdv_f,tid)
            if (.not.case4D) then
              !-> J0.(\int (f-feq) vpar dvpar)
              call omp_compute_gyrofunction_2D(J0,geom,imu, &
                Acomp,J0_intvdv_f,tid)
            end if
            !******************************************************
            !*** computation of the integrals in mu             ***
            !***  rhos is computed as the sum of 2 integrals:   ***
            !***   rhos =  (1/ne0(r))*hat_Zs*                   ***
            !***            \int dmu (I0*B_norm +               ***
            !***            sqrt{hat_As}/hat_Zs *               ***
            !***            I1*mu0*vec_J.vec_b/B_norm)          ***
            !***   where I0 = J0.(\int (f_s-feq_s) dvpar) and   ***
            !***         I1 = J0.(\int (f_s-feq_s)*vpar dvpar)  ***
            !***   this by using the expresion of Bstar         ***
            !***    Bstar = B_norm + sqrt{hat_As}/hat_Zs *      ***
            !***            vpar*mu0*vec_J.vec_b/B_norm         ***
            !******************************************************
            !->  Sum the contribution into       
            !    Romp10_1Nrp1_1Nthetap1(tid)%val belonging 
            !    to thread 'tid'      
            if (.not.case4D) then
              do itheta = 0,f%n2-1
                 do ir = 0,f%n1
                  Bnorm_ij = vBnorm_ij(1+ir,1+itheta) 
                  Romp10_1Nrp1_1Nthetap1(imu)%val(1+ir,1+itheta) = &
                    TWOPI*coeff_intdmu * &
                    (J0_intdv_f(1+ir,1+itheta)*Bnorm_ij + &
                    J0_intvdv_f(1+ir,1+itheta)*sqrtAs_on_Zs * &
                    vscalprod_mu0Jb_ij(1+ir,1+itheta)/Bnorm_ij)
                end do
              end do
            else
              do itheta = 0,f%n2-1
                do ir = 0,f%n1
                  Romp10_1Nrp1_1Nthetap1(imu)%val(1+ir,1+itheta) = &
                    coeff_intdmu * J0_intdv_f(1+ir,1+itheta)
                end do
              end do
            end if
          end do
!$OMP END DO
!$OMP BARRIER
!$OMP MASTER
          do itheta = 0,geom%Ntheta
            do ir = 0,geom%Nr
              M0_loc(ir,itheta,iphi) = 0._RKIND
            end do
          end do
          !*** compute rhos = (1/ne0(r))*hat_Zs*[nGs-nGs_eq] ***
          !***  AND compute the averaged value of rho        ***
          !***  over theta in mrho                           ***
          do itheta = 0,f%n2-1
            do ir = 0,f%n1
              !*** Sum the contribution of each thread   ***
              !*** 'Romp10_1Nrp1_1Nthetap1(tid)%val'     ***
              !***  into M0_loc data structure           *** 
              do j = 0,Nmu
                M0_loc(ir,itheta,iphi) = M0_loc(ir,itheta,iphi) + &
                  Romp10_1Nrp1_1Nthetap1(j)%val(1+ir,1+itheta)
              end do
              M0_loc(ir,itheta,iphi) = QN_spec_impact*hatZs* &
                M0_loc(ir,itheta,iphi)/ne0(ir)
            end do
          end do
          !*** periodicity in theta ***
          do ir = 0,f%n1
            M0_loc(ir,Ntheta,iphi) = M0_loc(ir,0,iphi)
          end do
!$OMP END MASTER
        end if
        no_operation = 0
!$OMP BARRIER
      end do
!$OMP BARRIER
!$OMP END PARALLEL
   do tid = 1, Nbthread
      call FTR_DEALLOC(J0%omp_diag(tid)%val)
      call FTR_DEALLOC(J0%omp_ldiag(tid)%val)
      call FTR_DEALLOC(J0%omp_udiag(tid)%val)
   enddo
    end if

    !*** distribution of rho ***
    if (iphistart .ne. -1) then
      do iphi = 0,geom%Nphi-1
        if (moments_mapphi(iphi) .eq. pInSpecies_id) then
          !*** restore periodic conditions in theta ***
          do ir = 0,geom%Nr
            M0_loc(ir,f%n2,iphi) = M0_loc(ir,0,iphi)
          end do
        end if
      end do
    end if
#ifdef TIMER
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tcmp(1))
    tdeb = tfin
#endif

    !*** distribution of rho ***
    !-> message receive
    nbbcast_r = 0
    do iphi = 0,f%n3-1
      itag_phi   = 9000 + iphi + pSpecies_id*f%n3
      nbelements = (Nr+1)*(Ntheta+1)
      call MPI_IRECV(QN_rhs(0,0,iphi),nbelements,MPI_REAL8, &
        moments_mapphi(iphi),itag_phi,mpi_comm_species, &
        bcast_reqr(nbbcast_r),ierr)
      nbbcast_r = nbbcast_r + 1
    enddo
    !-> message sending
    if (iphistart .ne. -1) then
      nbbcast_s = 0
      do iphi = 0,geom%Nphi-1
        if (moments_mapphi(iphi) .eq. pInSpecies_id) then
          nbelements = (Nr+1)*(Ntheta+1)
          itag_phi   = 9000 + iphi + pSpecies_id*f%n3
          do iproc = 0,Nbproc_species-1
            call MPI_SEND(M0_loc(0,0,iphi),nbelements,MPI_REAL8, &
              iproc,itag_phi,mpi_comm_species,ierr)
            nbbcast_s = nbbcast_s+1
          enddo
        endif
      end do
    endif
    call MPI_WAITALL(nbbcast_r, bcast_reqr(0:nbbcast_r-1), &
            moments_status(:,0:nbbcast_r-1), ierr)

    !*** Computation of rho = sum_s rhos ***
!baoter (question : est-ce-que QN_rhs_tmp est utile ?)
    if (Nspecies.gt.1) then
      QN_rhs_tmp =>  Rarray1_NrNthetaNphi_nspec
      Qn_rhs_tmp = QN_rhs
      nbelements = size(QN_rhs)
      call MY_Allreduce(QN_rhs_tmp,QN_rhs, &
        nbelements,MPI_REAL8,MPI_SUM,MPI_COMM_CROSS_SPECIES,ierr)
    endif

    !*** periodicity in phi ***
    QN_rhs(0:geom%Nr,0:geom%Ntheta,geom%Nphi) = &
      QN_rhs(0:geom%Nr,0:geom%Ntheta,0)

#ifdef TIMER
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tcmm(2))
    tdeb = tfin
#endif

    do tid = 1, Nbthread
       call FTR_DEALLOC(Romp1_1Nrp1_1Nthetap1(tid)%val)
       call FTR_DEALLOC(Romp2_1Nrp1_1Nthetap1(tid)%val)
       call FTR_DEALLOC(Romp3_1Nrp1_1Nthetap1(tid)%val)
       call FTR_DEALLOC(Romp4_1Nrp1_1Nthetap1(tid)%val)

       call FTR_DEALLOC(Comp1_1Nrp1_1Ntheta(tid)%val)
    enddo

    do tid = 0, MAX(Nmu,Nbthread)
       call FTR_DEALLOC(Romp10_1Nrp1_1Nthetap1(tid)%val)
    enddo

   call FTR_DEALLOC(RRecvBufMPI_A)
#endif 
    ! HAS_QN_V4
  end subroutine compute_rho



  !----------------------------------------------------------- 
  ! Solve quasi-neutrality equation
  !      L.Phi + 1/(Zi*Te)[Phi-<Phi>_{flux surf}] = rho
  !   with the differential operator L defined as:
  !     L.Phi = -1/ne0(r) sum_s hat_As * ns0(r) *
  !             { d^2(Phi)/dr^2 + [1/r+1/ns0 dns0/dr]*dPhi/dr
  !             + 1/r^2  d^2(Phi)/dth^2 }
  !   and the RHS defined as:
  !     rho = 1/ne0(r)*sum_s hat_Zs*[nGs-nGs_eq]
  !   where nGs(r,theta,phi) = \int dmu \int Jv J0.f_s dvpar
  !     and nGs_eq(r,theta)  = \int dmu \int Jv J0.feq_s dvpar     
  ! . Jv being the jacobian in velocity, f_s and feq_s 
  !    being respectively the distribution function and 
  !    equilibrium distribution function for species s.
  !
  ! The flux surface average of Phi is defined by
  !   <Phi>_{flux surf} = \int Phi Js(r,theta) dtheta dphi /
  !                       \int Js(r,theta) dtheta dphi
  !  where Js corresponds to the Jacobian in space  
  !-----------------------------------------------------------
  subroutine solve_QNequation(geom,J0,f,Sfmu_eq, &
    init_prof,init_magnet,init_curr,Phi)
    use globals, only : plocal_id, QN_coef_Phi00, &
      FFTpotential_filter, single_n, single_m, &
      Rarray1_Nr, Rarray2_Nr, Rarray3_Nr, &
      Rarray1_NrNthetaNphi, Rarray2_NrNthetaNphi, &
      nbelectrons_diag, nbions_diag
#ifdef _OPENMP
    use OMPutils_module, only : omp_get_thread_num
#endif
    use OMPutils_module, only : Comp1_1Nrp1_1Ntheta, &
      Comp2_1Nrp1_1Ntheta, Romp1_1Nrp1, Romp2_1Nrp1, &
      Romp3_1Nrp1, Romp4_1Nrp1
    use geometry_class
    use gyroaverage_class
    use fdistribu5d_class
    use init_profile_class
    use init_magnetic_class
    use init_current_class
    use fft_NRF90_module
    use integration_module, only : compute_thetaphi_average, &
      compute_fluxsurf_average
    use poisson_class     , only : Phi00_diag, &
      Phi_FSavg_diag, compute_nbparticles
    use globals, only : Nbthread
    type(geometry)               , intent(in)    :: geom
    type(J0operator)             , intent(inout) :: J0
    type(fdistribu5d)            , intent(in)    :: f
    real(RKIND), dimension(:,:,:), pointer       :: Sfmu_eq
    type(init_profile)           , intent(in)    :: init_prof
    type(init_magnetic)          , intent(in)    :: init_magnet
    type(init_current)           , intent(in)    :: init_curr
    real(RKIND), dimension(:,:,:), pointer       :: Phi

#ifdef HAS_QN_V4
    real(RKIND)   , dimension(:,:,:), pointer :: rhoi
    real(RKIND)   , dimension(:)    , pointer :: rhoi_TPavg
    complex(CKIND), dimension(:,:)  , pointer :: fft_rhoi 
    real(RKIND)   , dimension(:,:,:), pointer :: Phitilde
    real(RKIND)   , dimension(:)    , pointer :: Phitilde_Sflux_avg
    real(RKIND)   , dimension(:)    , pointer :: Phi_Sflux_avg
    real(RKIND)   , dimension(:)    , pointer :: ldiag
    real(RKIND)   , dimension(:)    , pointer :: diag
    real(RKIND)   , dimension(:)    , pointer :: udiag
    complex(CKIND), dimension(:)    , pointer :: Crhs_tmp 
    real(RKIND)   , dimension(:)    , pointer :: rhs_tmp 

    !--> local variables
    integer     :: tid
    integer     :: indx_grid
    integer     :: ir, itheta, iphi
    integer     :: i1, ikth, N1, N2
    integer     :: LU_err
    logical     :: stopit
    real(RKIND) :: inv_N2, LinvZ02Te_Phitilde_tmp
#ifdef TIMER
    integer(TIMEPREC) :: tdeb, tfin, tth1, tth2
    real(RKIND)       :: tthd

    call clck_time(tdeb)
    tth1 = tdeb

    call ppbarrier()
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tbar(1))
    tdeb = tfin
#endif

    call FTR_ALLOC_3(Rarray1_NrNthetaNphi,0,geom%Nr,0,geom%Ntheta,0,geom%Nphi)

    N1        = (geom%Nr+1)-2
    N2        = (geom%Ntheta+1)-1
    inv_N2    = 1._RKIND/real(N2)
    indx_grid = 0   ! The grid arrays begin with the 0 index

    rhs_tmp            => Romp1_1Nrp1(1)%val         !used (1:Nr)
    rhoi               => Rarray1_NrNthetaNphi
    rhoi_TPavg         => Rarray1_Nr
    Phitilde           => Rarray2_NrNthetaNphi
    Phitilde_Sflux_avg => Rarray2_Nr
    Phi_Sflux_avg      => Rarray3_Nr

    !--------------------------------------------------------
    !-> step 1 : compute rho = 1/ne0*sum_s Zs*[nGs-nGs_eq]
    !--------------------------------------------------------
    call compute_rho(geom,J0,f,Sfmu_eq, &
      init_prof,init_magnet,init_curr,rhoi)

#ifdef TIMER
    call clck_time(tdeb)
#endif

    !---------------------------------------------------------
    !-> step 2 : solve the equation:
    !     L.Phitilde + 1/(Zi Te) Phitilde = 
    !                   rhoi - <rhoi>_{theta,phi}
    !   where Phitilde = Phi-Phibar and where
    !   <.>_{theta,phi} = 1/(Ltheta*Lphi) \int . dtheta dphi 
    ! Rk : The boundary conditions are Dirichlet conditions
    !---------------------------------------------------------
    !----> compute rhoi_TPavg = <rhoi>_{theta,phi}
    rhoi_TPavg = 0._RKIND
    call compute_thetaphi_average(rhoi,geom, &
      QN_pspline1d_theta,QN_pspline1d_phi,rhoi_TPavg)

    do tid = 1, Nbthread
       call FTR_ALLOC_2_ZERO(Comp1_1Nrp1_1Ntheta(tid)%val,1,Nr+1,1,Ntheta)
       call FTR_ALLOC_2_ZERO(Comp2_1Nrp1_1Ntheta(tid)%val,1,Nr+1,1,Ntheta)
    enddo

    stopit = .false.
!$OMP PARALLEL default(none) &
!$OMP private(tid,ir,itheta,iphi,i1,ikth, &
!$OMP  fft_rhoi,Crhs_tmp,LU_err,ldiag,diag,udiag) &
!$OMP shared(Comp1_1Nrp1_1Ntheta,Comp2_1Nrp1_1Ntheta, &
!$OMP  Romp2_1Nrp1,Romp3_1Nrp1,Romp4_1Nrp1, &
!$OMP  stopit,geom,indx_grid,N1,N2,inv_N2, &
!$OMP  L_diag,L_ldiag,L_udiag,LinvZ02Te_Phitilde_tmp, &
!$OMP  L_invZ02Te_diag,Lm_invZ02Te_diag, &
!$OMP  rhoi,rhoi_TPavg,Phitilde,QN_coef_Phi00, &
!$OMP  QN_pspline1d_theta,QN_pspline1d_phi, &
!$OMP  Phi_Sflux_avg,Phitilde_Sflux_avg, &
!$OMP  Phi_FSavg_diag,Phi,rhs_tmp,coeff_neumann)
#ifdef _OPENMP
    tid = 1+omp_get_thread_num()
#else
    tid = 1
#endif
    fft_rhoi => Comp1_1Nrp1_1Ntheta(tid)%val        !used (1:Nr-1,1:Ntheta)
    Crhs_tmp => Comp2_1Nrp1_1Ntheta(tid)%val(1:,1)  !used (1:Nr)
    ldiag    => Romp2_1Nrp1(tid)%val
    diag     => Romp3_1Nrp1(tid)%val
    udiag    => Romp4_1Nrp1(tid)%val

!$OMP DO SCHEDULE(STATIC,1)
    !----> compute Phitilde = Phi-Phibar by solving
    !  A.Phitilde = b with A = (L+1/(Zi Te)) and 
    !   b = rhoi - <rhoi>_{theta,phi} 
    do iphi = 0,geom%Nphi-1
      !----> compute rhoi-rhoi_TPavg
      do ikth = 1,N2
        itheta = ikth + (indx_grid-1)
        do i1 = 1,N1
          ir                = indx_grid + i1 
          fft_rhoi(i1,ikth) = rhoi(ir,itheta,iphi) - &
            rhoi_TPavg(ir)
        end do
      end do

      !----> compute the poloidal Fourier transform 
      !      FFT_m(rhoi-rhoi_TPavg)
      call fourrow(fft_rhoi,1)

      !----> solve in the Fourier space
      !      (L+ 1/(Zi Te)) Phitilde^m = FFT_m(rhoi-rhoi_TPavg)
      do ikth = 1,(N2/2+1)
        !----> for each positive poloidal Fourier mode
        do i1 = 1,N1
          ldiag(i1)    = L_ldiag(i1)
          diag(i1)     = Lm_invZ02Te_diag(i1,ikth)
          udiag(i1)    = L_udiag(i1)
          Crhs_tmp(i1) = fft_rhoi(i1,ikth)
        end do
        call modThomas_c(N1,ldiag,diag,udiag,Crhs_tmp,LU_err)
        if (LU_err.ne.0) then
          print *,'problem in modThomas_c solving for ikth = ', &
            ikth
!$OMP CRITICAL
          stopit = .true.
!$OMP END CRITICAL
        endif
        do i1 = 1,N1
          fft_rhoi(i1,ikth) = Crhs_tmp(i1)
        end do
        !----> compute the negative part by using the 
        !      fact that it is equivalent to the conjugate value
        !      because the result is real
        if ( (ikth.ne.1).and.(ikth.ne.(N2/2+1)) ) then
          do i1 = 1,N1
            fft_rhoi(i1,N2+2-ikth) = conjg(Crhs_tmp(i1))
          end do
        end if
      end do

      !----> compute the inverse FFT
      call fourrow(fft_rhoi,-1)
      do ikth = 1,N2
        itheta = ikth+indx_grid-1
        do i1 = 1,N1
          ir = i1+indx_grid
          Phitilde(ir,itheta,iphi) = real(fft_rhoi(i1,ikth))*inv_N2
        end do
        !------> Dirichlet boundary conditions in radial direction
        Phitilde(indx_grid,itheta,iphi) = 0._RKIND
        Phitilde(N1+1,itheta,iphi)      = 0._RKIND
      end do
      !----> Periodic boundary conditions in theta
      do ir = 0,geom%Nr
        Phitilde(ir,geom%Ntheta,iphi) = Phitilde(ir,indx_grid,iphi)
      end do
    end do
!$OMP END DO
!$OMP BARRIER

!$OMP MASTER
    !---> Periodic boundary conditions in phi
    do itheta = indx_grid,geom%Ntheta
      do ir = indx_grid,geom%Nr
        Phitilde(ir,itheta,geom%Nphi) = Phitilde(ir,itheta,0)
      end do
    end do

    !----------------------------------------------------------
    !-> step 3 : compute <Phitilde>_{flux surf} the 
    !    flux surface average of Phitilde = (Phi-Phibar), i.e
    !    <Phitilde>_{flux surf} = 
    !         \int Phitilde Js(r,theta) dtheta dphi /
    !         \int Js(r,theta) dtheta dphi
    !----------------------------------------------------------
    call compute_fluxsurf_average(Phitilde,geom, &
      QN_pspline1d_theta,QN_pspline1d_phi,Phitilde_Sflux_avg)

    !----------------------------------------------------------
    !-> step 4 : compute the right side vector
    !    <rhoi(ir)>_{theta,phi} + 
    !        (L+1/(Zi*Te))*<Phi-Phibar>_{flux surf}
    !----------------------------------------------------------
    do i1 = 1,N1
      ir                     = indx_grid + i1
      LinvZ02Te_Phitilde_tmp = &
        L_ldiag(i1)*Phitilde_Sflux_avg(ir-1) + &
        L_invZ02Te_diag(i1)*Phitilde_Sflux_avg(ir) + &
        L_udiag(i1)*Phitilde_Sflux_avg(ir+1)
      rhs_tmp(i1)           = rhoi_TPavg(ir) + &
        LinvZ02Te_Phitilde_tmp 
    end do

    !-------------------------------------------------------------
    !-> step 5 : compute <Phi>_{flux surf} by solving
    !    (L + (1-lambda)*1/(Zi Te)) <Phi>_{flux surf} = 
    !    rhoi_TPavg(ir) + (L+1/(Zi*Te))*<Phi-Phibar>_{flux surf}
    !-------------------------------------------------------------
    if (QN_coef_Phi00==1) then
      !--> case lambda = 1
      call modThomas_d(N1,L_ldiag,L_diag,L_udiag,rhs_tmp,LU_err)
    else
      !--> case lambda = 0
      call modThomas_d(N1,L_ldiag,L_invZ02Te_diag,L_udiag,rhs_tmp,LU_err)
    end if
    if (LU_err.ne.0) then
      print *,'problem in modThomas_d solving'
    end if
    do i1 = 1,N1
      ir                = indx_grid + i1
      Phi_Sflux_avg(ir) = rhs_tmp(i1)  
    end do

    !----> Dirichlet or Neumann boundary condition
    Phi_Sflux_avg(indx_grid) = coeff_Neumann * &
         Phi_Sflux_avg(indx_grid+1)  
    Phi_Sflux_avg(N1+1)      = 0._RKIND

    !*** Saving of flux surface average of Phi ***
    do ir = indx_grid,geom%Nr
      Phi_FSavg_diag(ir) = Phi_Sflux_avg(ir) 
    end do
!$OMP END MASTER
!$OMP BARRIER

    !----------------------------------------------------------
    !-> step 6 : compute Phi(r,theta,phi) as
    !    Phi(r,theta,phi) = Phitilde(r,theta,phi) 
    !      - <Phitilde>_{flux surf} + <Phi>_{flux surf}
    !----------------------------------------------------------
!$OMP DO SCHEDULE(STATIC)
    do iphi = 0,geom%Nphi-1
      do itheta = 0,geom%Ntheta-1
        do ir = 0,geom%Nr
          Phi(ir,itheta,iphi) = Phitilde(ir,itheta,iphi) - &
            Phitilde_Sflux_avg(ir) + Phi_Sflux_avg(ir)
        end do
      end do
    end do
!$OMP END DO
!$OMP BARRIER
!$OMP END PARALLEL

    do tid = 1, Nbthread
       call FTR_DEALLOC(Comp2_1Nrp1_1Ntheta(tid)%val)
       call FTR_DEALLOC(Comp1_1Nrp1_1Ntheta(tid)%val)
    enddo

#ifdef TIMER
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tcmp(1))
    tdeb = tfin
#endif

    !*** filter applied on the electric potentiel if required ***
    if ((FFTpotential_filter.eq.1) .or. &
        (FFTpotential_filter.eq.2) .or. &
        (single_n).or.(single_m)) then
      call QNsolver_filter(Phi,geom,init_prof)
    end if

    !----> Periodic boundary conditions in theta
    do iphi = 0,geom%Nphi-1
      do ir = 0,geom%Nr
        Phi(ir,geom%Ntheta,iphi) = Phi(ir,indx_grid,iphi)
      end do
    end do

    !-> periodic boundary condition in phi direction
    do itheta = 0,geom%Ntheta
      do ir = 0,geom%Nr
        Phi(ir,itheta,geom%Nphi) = Phi(ir,itheta,0)
      end do
    end do

    !----------------------------------------------------------
    !-> Compute the Fourier (0,0) mode of Phi
    !    Phi00 = 1/(Ltheta*Lphi)\int Phi dtheta dphi 
    !----------------------------------------------------------
    call compute_thetaphi_average(Phi,geom, &
      QN_pspline1d_theta,QN_pspline1d_phi,Phi00_diag)
    
#ifdef TIMER
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tcmp(2))
    tdeb = tfin
#endif

    !**********************************************
    !*** computation of the number of electrons ***
    !***  and the number of ions                ***
    !**********************************************
    call compute_nbparticles(geom,init_prof,rhoi, &
      Phi,Phi_Sflux_avg,nbelectrons_diag,nbions_diag)

    call ppbarrier() 

    call FTR_DEALLOC(Rarray1_NrNthetaNphi)

#ifdef TIMER
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tcmp(3))
#endif

#ifdef TIMER
    tvect(1:12) = (/ tcmm(1:5), tcmp(1:5), tbar(1:2) /)
    call pp_reduce_vector(tvect,12,tmin,tmax,tavg)
    if (plocal_id .eq. 0) then
      write(6,'(I4,A,5F10.6,A,5F10.6,A,2F10.6)') pglobal_id, &
        " TMIN poisson5A, tcomm ", tmin(1:5), &
        " tcomp ",tmin(6:10)," tbarrier ",tmin(11:12)
      write(6,'(I4,A,5F10.6,A,5F10.6,A,2F10.6)') pglobal_id, &
        " TMAX poisson5A, tcomm ", tmax(1:5), &
        " tcomp ",tmax(6:10)," tbarrier ",tmax(11:12)
      write(6,'(I4,A,5F10.6,A,5F10.6,A,2F10.6)') pglobal_id, &
        " TAVG poisson5A, tcomm ", tavg(1:5), &
        " tcomp ",tavg(6:10)," tbarrier ",tavg(11:12)
    end if
    call clck_ldiff(tth1,tfin,tthd)
    if (pglobal_id .eq. 0) then
      write(6,'(I4,A,F13.6)') pglobal_id, &
        " Temps poisson5B,  tcomp ",tthd
    end if
#endif
    if (stopit) then
      print*, 'solve_QNequation : problem in Poisson solving'
      stop
    end if
#endif 
    ! HAS_QN_V4
  end subroutine solve_QNequation



  !------------------------------------------------------------   
  ! Compute rho the RHS of the quasi-neutrality 
  !  equation, i.e:
  !    rho(r,theta,phi) = sum_s rhos(r,theta,phi)
  !  where for each specie s
  !    rhos(r,theta,phi) = (1/ne0(r))*hat_Zs * 
  !                        (nGs(r,theta,phi)-nGs_eq(r,theta))
  !  where nGs = \int J0(sqrt{2*mu}).f_s Jv  dvpar dmu
  !            = \int J0(sqrt{2*mu}).[\int f_s Jv dvpar] dmu 
  !  with  nGs_eq(r,theta) = \int J0.feq_s(r,theta,vpar,mu) 
  !                          Jv(r,theta,vpar)dmu dvpar
  !  Jv being the Jacobian in velocity space :
  !   . Jv(r,theta,vpar) = 2*pi*Bstar in the 5D case
  !                      = 1          in the 4D case
  !  . f_s being the distribution function of specie s and
  !
  ! Rk : rhos is computed as the sum of 2 integrals:
  !   Is =  (1/ne0(r))*hat_Zs* \int dmu
  !            (I0*B_norm + (sqrt{hat_As}/hat_Zs)* 
  !            I1*mu0*vec_J.vec_b/B_norm)
  !   where I0 = J0.(\int f_s dvpar) and          
  !         I1 = J0.(\int f_s*vpar dvpar)
  !   this by using the expresion of Bstar
  !    Bstar = B_norm + (sqrt{hat_As}/hat_Zs) * 
  !            vpar*mu0*vec_J.vec_b/B_norm
  !
  ! Rk2 : this subroutine is the same than the previous
  !   one 'compute_rho' but it takes into account the
  !   fact that the electrostatic potential is distributed
  !   on several processors.
  !------------------------------------------------------------
  subroutine compute_rho_distrib3D(geom,J0,f,Sfmu_eq, &
    init_prof,init_magnet,init_curr,QN_rhs,QN_distribrhs)
    use globals, only : mumin, integration_CS, &
      Nbproc_r, Nbproc_loc, Nphi, &
      Nmu, Nmuscal, mu_id, mpi_comm_intermu, &
      dom_r, dom_theta, istart, iend, jstart, jend, &
      RRecvBufMPI_A, Rarray_PNrPNthetaNphi_nbM2, &
      Rarray1_NrNthetaPNmu, Rarray3_NrNthetaDomphi, &
      QN_comm_version, QN_species_impact
    use globals, only : hat_As, hat_Zs, &
      mpi_comm_species, mpi_comm_cross_species, &
      pInSpecies_id, Nbproc_species, pSpecies_id, &
      Nspecies
#ifdef _OPENMP
    use OMPutils_module, only : omp_get_thread_num
#endif
    use OMPutils_module, only : Romp_scoefvpar, &
      Romp1_0Nvpar, Romp2_0Nvpar, Romp3_1Nrp1_1Nthetap1, &
      Romp4_1Nrp1_1Nthetap1, Romp11_1Nrp1_1Nthetap1, &
      Romp12_1Nrp1_1Nthetap1, Comp1_1Nrp1_1Ntheta
    use geometry_class
    use gyroaverage_class
    use fdistribu5d_class
    use init_profile_class
    use init_magnetic_class
    use init_current_class
    use integration_module, only : &
      compute_omp_vpar_integral_CS, compute_vpar_integral_colloc
    use mpiMod
    use globals        , only : Nbthread
    type(geometry)               , intent(in)    :: geom
    type(J0operator)             , intent(inout) :: J0
    type(fdistribu5d)            , intent(in)    :: f
    real(RKIND), dimension(:,:,:), pointer       :: Sfmu_eq
    type(init_profile)           , intent(in)    :: init_prof
    type(init_magnetic)          , intent(in)    :: init_magnet
    type(init_current)           , intent(in)    :: init_curr
    !-> size (0:Nr,0:Ntheta,iphistart,iphistart+dom_mapphi-1)
    real(RKIND), dimension(:,:,:), pointer    :: QN_rhs
    real(RKIND), dimension(:,:,:), pointer    :: QN_distribrhs

#ifdef HAS_QN_V5
    integer, parameter :: nbmoments = 2
    real(RKIND)   , dimension(:)    , pointer :: fmu_1D
    real(RKIND)   , dimension(:)    , pointer :: fmu_v_1D
    real(RKIND)   , dimension(:)    , pointer :: scoefvpar
    real(RKIND)   , dimension(:,:)  , pointer :: scalprod_mu0Jb
    real(RKIND)   , dimension(:,:)  , pointer :: vBnorm_ij
    real(RKIND)   , dimension(:,:)  , pointer :: varray1
    real(RKIND)   , dimension(:,:)  , pointer :: varray2
    real(RKIND)   , dimension(:,:)  , pointer :: varray3
    complex(CKIND), dimension(:)    , pointer :: crhs
    complex(CKIND), dimension(:,:)  , pointer :: Acomp
    real(RKIND)   , dimension(:)    , pointer :: ne0

    integer     :: ir, itheta, iphi, ivpar, imu
    integer     :: ierr
    real(RKIND) :: fmu_tmp, jac_v
    real(RKIND) :: Bnorm_ij, scalprod_mu0Jb_ij
    real(RKIND) :: coeff_intdmu
    logical     :: case4D

    ! -> variables for parallelization
    integer :: tid
    integer :: iphiloc, bmu
    integer :: inc, imucontrib, maxinc
    integer :: subtask, nbsubtask, itask
    integer :: no_operation
    integer :: nbreqr, nbbcast_r, nbbcast_s, nbreqs
    integer :: local_id, dep, base, global_id, startbuf
    integer :: l_istart, l_iend, l_jstart, l_jend
    integer :: iproc, itag_phi, nbelements

    !-> used for 2 species case
    real(RKIND) :: QN_spec_impact
    real(RKIND) :: hatZs, sqrtAs_on_Zs, mysqrtAs_on_Zs
    real(RKIND), dimension(:,:,:), pointer :: QN_rhs_tmp
    
    !-> used for collective version
    integer, dimension(0:Nbproc_species-1) :: &
                    send_count_vec, send_disp_vec, &
                    recv_count_vec, recv_disp_vec
    integer :: Rarray_PNrPNthetaNphi_nbM2_shape(4)
    
#ifdef TIMER
    integer(TIMEPREC) :: tdeb, tfin, tth1, tth2
    real(RKIND)       :: tthd
    tcmp(1:6) = 0.
    tcmm(1:6) = 0.
    tbar(1:6) = 0.

    call clck_time(tdeb)
    tth1 = tdeb
#endif

    !--> check if 4D or 5D
    case4D = .false.
    if (geom%Nmu.eq.0) case4D = .true.

    !--> Species impact in QN solver (for test)
    QN_spec_impact = QN_species_impact(pSpecies_id)

    !--> compute sqrt{hat_As}/hat_Zs
    hatZs        = hat_Zs(pSpecies_id)
    sqrtAs_on_Zs = sqrt(hat_As(pSpecies_id))/hatZs

    !*** initialize local variables ***
    ne0 => init_prof%ne0

    if (Nmuscal .eq. 1 .and. iphistart .ne. -1)  then
      if ((iphistart .ne. ilargestart) .or. &
        (iphistart+dom_mapphi-1 .ne. ilargeend) ) then
        print *, "Big problem in definition of ilargestart or ilargeend"
        STOP
      end if
    end if

#ifdef TIMER
    call ppbarrier()
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tbar(1))
    tdeb = tfin
#endif

!$OMP PARALLEL default(none) &
!$OMP private(tid,imu,iphi,itheta,ir,ivpar, &
!$OMP  scoefvpar,fmu_tmp,fmu_1D,fmu_v_1D,jac_v) &
!$OMP shared(f,Sfmu_eq,integration_CS, &
!$OMP  geom,Rarray_PNrPNthetaNphi_nbM2, &
!$OMP  Romp_scoefvpar,Romp1_0Nvpar,Romp2_0Nvpar)
#ifdef _OPENMP
    tid = 1+omp_get_thread_num()
#else
    tid = 1
#endif
    scoefvpar => Romp_scoefvpar(tid)%val
    fmu_1D    => Romp1_0Nvpar(tid)%val
    fmu_v_1D  => Romp2_0Nvpar(tid)%val

!$OMP BARRIER
!$OMP DO SCHEDULE(STATIC)
    !*** computation of the integrals ***
    !**   in vpar direction           ***
    !***    \int (f-feq) dvpar and    ***
    !***    \int (f-feq)*vpar dvpar   ***
    do iphi = 0,f%n3-1
      do itheta = f%jstart,f%jend
        do ir = f%istart,f%iend
          do ivpar = 0,f%n4
            if (geom%Nmu .eq. 0) then
              call precomp_Bstar(ir,itheta,ivpar,jac_v)
            else
              jac_v = 1
            end if
            fmu_tmp         = jac_v*(f%values(ir,itheta,iphi,ivpar) - &
              Sfmu_eq(ir,itheta,ivpar))
            fmu_1D(ivpar)   = fmu_tmp
            fmu_v_1D(ivpar) = fmu_tmp*geom%vparg(ivpar)
          end do
          if (integration_CS) then
            !-> \int (f-feq) dvpar
            call compute_omp_vpar_integral_CS(f%n4,f%h4,fmu_1D, &
              f%BCvpar_left,f%BCvpar_right, &
              f%nspline1d_vpar(tid),scoefvpar, &
              Rarray_PNrPNthetaNphi_nbM2(0,ir,itheta,iphi))
            !-> \int (f-feq) vpar dvpar
            call compute_omp_vpar_integral_CS(f%n4,f%h4,fmu_v_1D, &
              f%BCvpar_left,f%BCvpar_right, &
              f%nspline1d_vpar(tid),scoefvpar, &
              Rarray_PNrPNthetaNphi_nbM2(1,ir,itheta,iphi))
          else
            !-> \int (f-feq) dvpar
            call compute_vpar_integral_colloc(fmu_1D(0:f%n4), &
              geom,Rarray_PNrPNthetaNphi_nbM2(0,ir,itheta,iphi))
            !-> \int (f-feq) vpar dvpar
            call compute_vpar_integral_colloc(fmu_v_1D(0:f%n4), &
              geom,Rarray_PNrPNthetaNphi_nbM2(1,ir,itheta,iphi))
          end if
        enddo
      enddo
    end do
!$OMP END DO
!$OMP BARRIER
!$OMP END PARALLEL

#ifdef TIMER
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tcmp(1))
    tdeb = tfin
#endif

    call write_param("nbmoments", nbmoments)
    call FTR_ALLOC_1_ZERO(RRecvBufMPI_A,0,nbmoments*dom_r*dom_theta*Nbproc_loc*(Nmu+1)*dom_mapphi)

    !*** store the 2 previous integrals in the big 1D array ***
    !-> message receive / send 
    Rarray_PNrPNthetaNphi_nbM2_shape = shape(Rarray_PNrPNthetaNphi_nbM2)
    recv_count_vec = 0
    recv_disp_vec = 0
    send_count_vec = 0
    send_disp_vec = 0
    nbreqr  = 0
    nbreqs  = 0
    iphiloc = 0
    moments_reqs = 0
    moments_reqr = 0
    do iphi = 0,f%n3-1
      do bmu = 0,Nmuscal-1

        itag_phi   = 4000 + iphi
        if (scal_map(0,bmu,iphi) .eq. pInSpecies_id) then
          do imu = scal_map(1,bmu,iphi),scal_map(2,bmu,iphi)
            do local_id = 0,Nbproc_loc-1
              global_id  = imu*Nbproc_loc + local_id
              startbuf   = dom_theta * (local_id + Nbproc_loc * &
                (imu + (Nmu+1) * iphiloc)) 
              nbelements = nbmoments * dom_r * dom_theta
              select case (QN_comm_version)
              case (1,2)
                call MPI_IRECV( &
                  RRecvBufMPI_A(startbuf*dom_r*nbmoments), &
                  nbelements,MPI_REAL8,global_id,itag_phi, &
                  mpi_comm_species,moments_reqr(nbreqr),ierr)
              case (3)
                recv_count_vec(global_id) = nbelements
                recv_disp_vec(global_id) = &
                                startbuf*dom_r*nbmoments
              case default
                if (pglobal_id.eq.0) then
                    print*, 'Parameter QN_comm_version ',QN_comm_version,' is not set at a known value.'
                    stop
                end if
              end select
              nbreqr = nbreqr + 1
            enddo
          enddo
          iphiloc = iphiloc + 1
        end if

        itag_phi = 4000 + iphi
        if (mu_id .ge. scal_map(1,bmu,iphi) .and. &
          mu_id .le. scal_map(2,bmu,iphi)) then
          nbelements = nbmoments*dom_r*dom_theta
          select case (QN_comm_version)
          case (1)
            call MPI_ISEND(& 
              Rarray_PNrPNthetaNphi_nbM2(0,istart,jstart,iphi), &
              nbelements,MPI_REAL8,scal_map(0,bmu,iphi),itag_phi, &
              mpi_comm_species,moments_reqs(nbreqs),ierr)
          case (2)
            call MPI_SEND(& 
              Rarray_PNrPNthetaNphi_nbM2(0,istart,jstart,iphi), &
              nbelements,MPI_REAL8,scal_map(0,bmu,iphi),itag_phi, &
              mpi_comm_species,ierr)
          case (3)
            send_count_vec(scal_map(0,bmu,iphi)) = nbelements
            send_disp_vec(scal_map(0,bmu,iphi)) = &
                            0+&
                            istart*Rarray_PNrPNthetaNphi_nbM2_shape(1)+ &
                            jstart*Rarray_PNrPNthetaNphi_nbM2_shape(2)+ &
                            iphi*Rarray_PNrPNthetaNphi_nbM2_shape(3)
          end select
          nbreqs = nbreqs + 1
        end if
      end do
    end do

    select case (QN_comm_version)
    case (1)
       if (nbreqr .ne. 0) &
            call MPI_WAITALL(nbreqr, moments_reqr(0:nbreqr-1), &
                  moments_status(:,0:nbreqr-1), ierr)
       if (nbreqs .ne. 0) &
            call MPI_WAITALL(nbreqs, moments_reqs(0:nbreqs-1), &
                  moments_status(:,0:nbreqs-1), ierr)
    case (2)
       if (nbreqr .ne. 0) &
            call MPI_WAITALL(nbreqr, moments_reqr(0:nbreqr-1), &
                  moments_status(:,0:nbreqr-1), ierr)
    case (3)
      if ( nbreqr /= Nbproc_species .and. pglobal_id==0 ) then
        print*, 'Cannot use alltoall, nbreqr /= Nbproc_species', &
                        nbreqr, Nbproc_species
        call MPI_abort(mpi_comm_species, nbreqr, ierr)
      end if
      call MPI_alltoallv(Rarray_PNrPNthetaNphi_nbM2, send_count_vec, &
                      send_disp_vec, MPI_REAL8, RRecvBufMPI_A, &
                      recv_count_vec, recv_disp_vec, MPI_REAL8, &
                      mpi_comm_species, ierr)
    end select

#ifdef TIMER
    call ppbarrier()
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tcmm(1))
    tdeb = tfin
#endif

    do tid = 1, Nbthread
       call FTR_ALLOC_2(Romp3_1Nrp1_1Nthetap1(tid)%val,1,Nr+1,1,Ntheta+1)
       call FTR_ALLOC_2(Romp4_1Nrp1_1Nthetap1(tid)%val,1,Nr+1,1,Ntheta+1)

       call FTR_ALLOC_2_ZERO(Comp1_1Nrp1_1Ntheta(tid)%val,1,Nr+1,1,Ntheta)
    enddo

    do tid = 0, MAX(Nmu,Nbthread)
       call FTR_ALLOC_2(Romp11_1Nrp1_1Nthetap1(tid)%val,1,Nr+1,1,Ntheta+1)
       call FTR_ALLOC_2(Romp12_1Nrp1_1Nthetap1(tid)%val,1,Nr+1,1,Ntheta+1)
    enddo

    do tid = 1, Nbthread
       call FTR_ALLOC_2(J0%omp_diag(tid)%val,1,J0%N1,1,J0%N2)
       call FTR_ALLOC_2(J0%omp_ldiag(tid)%val,1,J0%N1-1,1,J0%N2)
       call FTR_ALLOC_2(J0%omp_udiag(tid)%val,1,J0%N1-1,1,J0%N2)
    enddo
    
!$OMP PARALLEL default(none) &
!$OMP private(tid,nbbcast_s,imucontrib, mysqrtAs_on_Zs, &
!$OMP  imu,iphi,coeff_intdmu,no_operation, &
!$OMP  nbelements,local_id,base,dep,l_istart, &
!$OMP  l_iend,l_jstart,l_jend,startbuf,itheta,ir, &
!$OMP  scalprod_mu0Jb,scalprod_mu0Jb_ij,vBnorm_ij, &
!$OMP  iphiloc,Bnorm_ij,Acomp,varray1,varray2,varray3,&
!$OMP  inc,maxinc,bmu,subtask,nbsubtask,itask) & 
!$OMP shared(geom,J0,f,ne0,Nmu,dom_r,dom_theta, &
!$OMP  pInSpecies_id,sqrtAs_on_Zs,hatZs,QN_spec_impact, &
!$OMP  init_magnet,init_curr,case4D, & 
!$OMP  Nmuscal,scal_map,Nbproc_loc,Nbproc_r, &
!$OMP  Comp1_1Nrp1_1Ntheta,Romp11_1Nrp1_1Nthetap1, &
!$OMP  Romp3_1Nrp1_1Nthetap1,Romp4_1Nrp1_1Nthetap1, &
!$OMP  Romp12_1Nrp1_1Nthetap1,RRecvBufMPI_A,QN_rhs, &
!$OMP  QN_distribrhs)
!$OMP BARRIER
#ifdef _OPENMP
    tid = 1+omp_get_thread_num()
#else
    tid = 1
#endif
    Acomp          => Comp1_1Nrp1_1Ntheta(tid)%val
    scalprod_mu0Jb => Romp3_1Nrp1_1Nthetap1(tid)%val
    mysqrtAs_on_Zs = sqrtAs_on_Zs

    vBnorm_ij => Romp4_1Nrp1_1Nthetap1(tid)%val
    do itheta = 0, geom%Ntheta
      do ir = 0, geom%Nr
        call compute_scalprod_mu0Jb(geom,init_magnet, &
          init_curr,ir,itheta,scalprod_mu0Jb(1+ir,1+itheta))
        vBnorm_ij(1+ir,1+itheta) = init_magnet%B_norm(ir,itheta) 
      end do
    end do

    !*** Computation of                      ***
    !***   I0 = J0.(\int (f-feq) dvpar) and  ***
    !***   I1 = J0.(\int (f-feq)*vpar dvpar) ***     
    iphiloc = 0
    do iphi = 0,geom%Nphi-1
      do bmu = 0,Nmuscal-1
        if (scal_map(0,bmu,iphi) .eq. pInSpecies_id) then
!$OMP BARRIER
          nbsubtask = (scal_map(2,bmu,iphi)-scal_map(1,bmu,iphi)+1)*2
          !*** Parallelization in variable imu in order ***
          !***  to speed up computation                 ***
          !*** whenever Nmu is large                    ***
!$OMP DO SCHEDULE(STATIC,1)
          do subtask = 0,nbsubtask-1
            imu     = scal_map(1,bmu,iphi) + (subtask/2)
            varray1 => Romp11_1Nrp1_1Nthetap1(imu)%val
            varray2 => Romp12_1Nrp1_1Nthetap1(imu)%val
            itask   = mod(subtask,2)
            if (itask .eq. 0) then
              !*** computation of J0_intdv_f  = J0.(\int f dvpar) ***
              !-> reconstruction of \int f dvpar
              do itheta = 0,geom%Ntheta
                do ir = 0,geom%Nr
                  varray1(1+ir,1+itheta) = 0._RKIND
                end do
              end do
              do local_id = 0,Nbproc_loc-1
                base     = (local_id/Nbproc_r)
                dep      = mod(local_id,Nbproc_r)              
                l_istart = dep *  dom_r
                l_iend   = l_istart + dom_r - 1
                l_jstart = base * dom_theta
                l_jend   = l_jstart + dom_theta - 1
                startbuf = dom_theta * (local_id + Nbproc_loc * &
                  (imu + (Nmu+1) * iphiloc)) 
                do itheta = l_jstart,l_jend
                  do ir = l_istart,l_iend
                    varray1(1+ir,1+itheta) = &
                      RRecvBufMPI_A(nbmoments*(ir-l_istart)+0 + &
                      (startbuf+itheta-l_jstart)*nbmoments*dom_r)
                  enddo
                enddo
              enddo
              !-> gyroaverage of \int f dvpar
              call omp_compute_gyrofunction_2D(J0,geom,imu, &
                Acomp,varray1,tid)
            else
              !*** computation of J0_intvdv_f = J0.(\int f vpar dvpar) ***
              !-> reconstruction of \int f vpar dvpar
              do itheta = 0,geom%Ntheta
                do ir = 0,geom%Nr
                  varray2(1+ir,1+itheta) = 0._RKIND
                end do
              end do
              do local_id = 0,Nbproc_loc-1
                base     = (local_id/Nbproc_r)
                dep      = mod(local_id,Nbproc_r)              
                l_istart = dep *  dom_r
                l_iend   = l_istart + dom_r - 1
                l_jstart = base * dom_theta
                l_jend   = l_jstart + dom_theta - 1
                startbuf = dom_theta * (local_id + Nbproc_loc * &
                  (imu + (Nmu+1) * iphiloc)) 
                do itheta = l_jstart,l_jend
                  do ir = l_istart,l_iend
                    varray2(1+ir,1+itheta) = &
                      RRecvBufMPI_A(nbmoments*(ir-l_istart)+1 + &
                      (startbuf+itheta-l_jstart)*dom_r*nbmoments) 
                  enddo
                enddo
              enddo
              if (.not.case4D) then
                !-> gyroaverage of \int f vpar dvpar
                call omp_compute_gyrofunction_2D(J0,geom,imu, &
                  Acomp,varray2,tid)
              end if
            end if
          end do
!$OMP END DO
!$OMP BARRIER

          !******************************************************
          !*** computation of the integrals in mu             ***
          !***  rhos is computed as the sum of 2 integrals:   ***
          !***   rhos =  (1/ne0(r))*hat_Zs *                  ***
          !***            \int dmu (I0*B_norm +               ***
          !***            sqrt{hat_As}/hat_Zs *               ***
          !***            I1*mu0*vec_J.vec_b/B_norm)          ***
          !***   where I0 = J0.(\int (f_s-feq_s) dvpar) and   ***
          !***         I1 = J0.(\int (f_s-feq_s)*vpar dvpar)  ***
          !***   this by using the expresion of Bstar         ***
          !***    Bstar = B_norm + sqrt{hat_As}/hat_Zs *      ***
          !***            vpar*mu0*vec_J.vec_b/B_norm         ***
          !******************************************************
          nbsubtask = (scal_map(2,bmu,iphi)-scal_map(1,bmu,iphi)+1)
          !--> Parallelization in variable imu in order 
          !-->   to speed up computation                 
          !-->   whenever Nmu is large                    
!$OMP DO SCHEDULE(STATIC,1)
          do subtask = 0,nbsubtask-1
            imu          = scal_map(1,bmu,iphi) + subtask
            varray1      => Romp11_1Nrp1_1Nthetap1(imu)%val
            varray2      => Romp12_1Nrp1_1Nthetap1(imu)%val
            coeff_intdmu = geom%coeff_intdmu(imu)
            if (.not.case4D) then
              do itheta = 0,f%n2-1
                do ir = 0,f%n1
                  Bnorm_ij = vBnorm_ij(1+ir,1+itheta) 
                  scalprod_mu0Jb_ij = scalprod_mu0Jb(1+ir,1+itheta) 
                  varray1(1+ir,1+itheta) = &
                    TWOPI*coeff_intdmu * &
                    (varray1(1+ir,1+itheta)*Bnorm_ij + &
                    varray2(1+ir,1+itheta)*mysqrtAs_on_Zs * &
                    scalprod_mu0Jb_ij/Bnorm_ij)
                end do
              end do
            else
              do itheta = 0,f%n2-1
                do ir = 0,f%n1
                  varray1(1+ir,1+itheta) = &
                    coeff_intdmu*varray1(1+ir,1+itheta)
                end do
              end do
            end if
          end do
!$OMP END DO
!$OMP BARRIER

          !--> Add the contribution of all mu values
          inc    = 1
          maxinc = scal_map(2,bmu,iphi)-scal_map(1,bmu,iphi)+1
          do while (inc .lt. maxinc)
!$OMP DO SCHEDULE(STATIC,1)
            do imu = scal_map(1,bmu,iphi),scal_map(2,bmu,iphi), 2*inc
              varray1    => Romp11_1Nrp1_1Nthetap1(imu)%val
              varray2    => Romp12_1Nrp1_1Nthetap1(imu)%val
              imucontrib = imu+inc
              if (imucontrib .le. scal_map(2,bmu,iphi)) then
                varray3 => Romp11_1Nrp1_1Nthetap1(imucontrib)%val
                do itheta = 0,geom%Ntheta-1
                  do ir = 0,geom%Nr
                    varray1(1+ir,1+itheta) = varray1(1+ir,1+itheta) + &
                      varray3(1+ir,1+itheta)
                  end do
                end do
              end if
            end do
!$OMP END DO
            inc = inc*2
!$OMP BARRIER
         end do
!$OMP BARRIER
!$OMP MASTER
         if (Nmuscal .ne. 1) then
           do itheta = 0,f%n2-1
             do ir = 0,f%n1
               QN_distribrhs(ir,itheta,iphi) = QN_spec_impact*hatZs * &
                 Romp11_1Nrp1_1Nthetap1( &
                 scal_map(1,bmu,iphi))%val(1+ir,1+itheta)/ne0(ir)
             end do
           end do
         else
           do itheta = 0,f%n2-1
             do ir = 0,f%n1
               QN_rhs(ir,itheta,iphi) = QN_spec_impact*hatZs * &
                 Romp11_1Nrp1_1Nthetap1( &
                 scal_map(1,bmu,iphi))%val(1+ir,1+itheta)/ne0(ir)
             end do
           end do
         end if
!$OMP END MASTER
          iphiloc = iphiloc + 1
        end if
        no_operation = 0
!$OMP BARRIER
      end do
    end do
!$OMP END PARALLEL

    call FTR_DEALLOC(RRecvBufMPI_A)

    do tid = 1, Nbthread
       call FTR_DEALLOC(Romp3_1Nrp1_1Nthetap1(tid)%val)
       call FTR_DEALLOC(Romp4_1Nrp1_1Nthetap1(tid)%val)

       call FTR_DEALLOC(Comp1_1Nrp1_1Ntheta(tid)%val)
    end do

    do tid = 0, MAX(Nmu,Nbthread)
       call FTR_DEALLOC(Romp11_1Nrp1_1Nthetap1(tid)%val)
       call FTR_DEALLOC(Romp12_1Nrp1_1Nthetap1(tid)%val)
    enddo


    do tid = 1, Nbthread
       call FTR_DEALLOC(J0%omp_diag(tid)%val)
       call FTR_DEALLOC(J0%omp_ldiag(tid)%val)
       call FTR_DEALLOC(J0%omp_udiag(tid)%val)
    enddo

#ifdef TIMER
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tcmp(2))
    tdeb = tfin
#endif

    !*** 
    if (Nmuscal .ne. 1) then
      nbbcast_r = 0
      nbbcast_s = 0
      
      !-> message receive
      if (iphistart .ne. -1) then
        nbelements = (Nr+1)*(Ntheta+1)
        iphi = iphistart
        do bmu = 0,Nmuscal-1
          itag_phi   = 9000 + bmu
          call MPI_IRECV(Rarray1_NrNthetaPNmu(0,0,bmu),nbelements, &
            MPI_REAL8,scal_map(0,bmu,iphi),itag_phi,mpi_comm_species, &
            bcast_reqr(nbbcast_r),ierr)
          nbbcast_r = nbbcast_r + 1
        enddo
      end if
      !-> message sending
      do iphi = 0,geom%Nphi-1
        do bmu = 0,Nmuscal-1
          if (scal_map(0,bmu,iphi) .eq. pInSpecies_id) then
            nbelements = (Nr+1)*(Ntheta+1)
            itag_phi   = 9000 + bmu
            call MPI_SEND(QN_distribrhs(0,0,iphi),nbelements,MPI_REAL8, &
              moments_mapphi(iphi),itag_phi,mpi_comm_species, &
              ierr)
            nbbcast_s = nbbcast_s+1
          end if
        enddo
      end do
      if (iphistart .ne. -1) then
        if (nbbcast_r .ne. Nmuscal) then
          print *, "precond failed (nbbcast_r .ne. Nmuscal)", &
            nbbcast_r, Nmuscal
          STOP
        end if
        call MPI_WAITALL(nbbcast_r, bcast_reqr(0:nbbcast_r-1), &
                moments_status(:,0:nbbcast_r-1), ierr)
        tid = 1
        inc = 1
        !-> The sum is accumulated in Rarray1_NrNthetaPNmu(*,*,0)
        maxinc = Nmuscal
        do while (inc .lt. maxinc)
          do imu = 0,Nmuscal-1,2*inc
            imucontrib = imu+inc
            if (imucontrib .le. Nmuscal-1) then
              ! if (imucontrib.le.Nmuscal-1) then we have also 
              ! (imu.le.Nmuscal-1)
              do itheta = 0,geom%Ntheta-1
                do ir = 0,geom%Nr
                  Rarray1_NrNthetaPNmu(ir,itheta,imu) = &
                    Rarray1_NrNthetaPNmu(ir,itheta,imu) + &
                    Rarray1_NrNthetaPNmu(ir,itheta,imucontrib)
                end do
              end do
            end if
          end do
          inc = inc*2
        end do
        iphi = iphistart
        do itheta = 0,Ntheta-1
          do ir = 0,Nr
            QN_rhs(ir,itheta,iphi) = &
              Rarray1_NrNthetaPNmu(ir,itheta,0)
          end do
        end do  
      end if
    end if

    !*** Computation of rho = sum_s rhos ***
    if (Nspecies.gt.1) then
!baoter (question : est-ce-que QN_rhs_tmp est utile ?)
      QN_rhs_tmp =>  Rarray3_NrNthetaDomphi
      Qn_rhs_tmp = QN_rhs
      nbelements = size(QN_rhs)
      call MY_Allreduce(QN_rhs_tmp,QN_rhs, &
        nbelements,MPI_REAL8,MPI_SUM,MPI_COMM_CROSS_SPECIES,ierr)
    endif

#ifdef TIMER
!    call ppbarrier()
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tcmm(2))
    tdeb = tfin
#endif
#endif 
    !HAS_QN_V5
  end subroutine compute_rho_distrib3D


  !----------------------------------------------------------- 
  ! Solve quasi-neutrality equation
  !      L.Phi + 1/(Zi*Te)[Phi-<Phi>_{flux surf}] = rho
  !   with the differential operator L defined as:
  !   L.Phi = -1/ne0(r) sum_s hat_As * ns0(r) *
  !           { d^2(Phi)/dr^2 + [1/r+1/ns0 dns0/dr]*dPhi/dr
  !             + 1/r^2  d^2(Phi)/dth^2 }
  !   and the RHS defined as:
  !     rho = 1/ne0(r)*sum_s hat_Zs*[nGs-nGs_eq]
  !   where nGs(r,theta,phi) = \int dmu \int Jv J0.f_s dvpar
  !     and nGs_eq(r,theta)  = \int dmu \int Jv J0.feq_s dvpar     
  ! . Jv being the jacobian in velocity, 
  ! . f_s and feq_s being respectively the distribution 
  !    function and equilibrium distribution function 
  !    for species s.
  !
  ! The flux surface average of Phi is defined by
  !   <Phi>_{flux surf} = \int Phi Js(r,theta) dtheta dphi /
  !                       \int Js(r,theta) dtheta dphi
  !  where Js corresponds to the Jacobian in space  
  !
  ! Rk : this subroutine is the same than the previous
  !   one 'solve_QNequation' but it takes into account the
  !   fact that the electrostatic potential is distributed
  !   on several processors.
  !------------------------------------------------------------
  subroutine solve_QNequation_distrib3D(geom,J0,f,Sfmu_eq, &
    init_prof,init_magnet,init_curr,poiss)
    use globals, only : plocal_id, QN_coef_Phi00, &
      FFTpotential_filter, single_n, single_m, &
      Rarray1_Nr, Rarray2_Nr, Rarray3_Nr, &
      Rarray2_NrNthetaNphi, &
      nbelectrons_diag, nbions_diag, &
      Nphi, mpi_comm_moments, Nbproc_loc, &
      Phi00_BCrmin_Neumann, integration_CS, &
      Rarray_NrNphi, Rarray2_NrNphi, &
      Rarray1_NrNthetaDomphi, Rarray2_NrNthetaDomphi, &
      Rarray3_NrNthetaDomphi, Rarray1_NrNthetaParphi
    use globals, only : mpi_comm_species, &
      pInSpecies_id, Nbproc_species, pSpecies_id
#ifdef _OPENMP
    use OMPutils_module, only : omp_get_thread_num
#endif
    use OMPutils_module, only : Comp_0Nr, &
      Comp1_1Nrp1_1Ntheta, Comp2_1Nrp1_1Ntheta, &
      Romp1_1Nrp1, Romp2_1Nrp1, Romp3_1Nrp1, Romp4_1Nrp1, &
      Romp1_0Nr, Romp2_0Nr, Romp3_0Nr, Romp5_0Nr, Romp6_0Nr
    use geometry_class
    use gyroaverage_class
    use fdistribu5d_class
    use init_profile_class
    use init_magnetic_class
    use init_current_class
    use fft_NRF90_module
    use mpiMod
    use globals, only : Nbthread
    type(geometry)               , intent(in)    :: geom
    type(J0operator)             , intent(inout) :: J0
    type(fdistribu5d)            , intent(in)    :: f
    real(RKIND), dimension(:,:,:), pointer       :: Sfmu_eq
    type(init_profile)           , intent(in)    :: init_prof
    type(init_magnetic)          , intent(in)    :: init_magnet
    type(init_current)           , intent(in)    :: init_curr
    type(poisson)                , intent(inout) :: poiss

#ifdef HAS_QN_V5
    real(RKIND), dimension(:,:,:), pointer :: rhoi
    real(RKIND), dimension(:,:,:), pointer :: Phitilde
    real(RKIND), dimension(:,:,:), pointer :: Phi
    real(RKIND), dimension(:,:,:), pointer :: rhsdistrib

    integer     :: ierr, N1, N2
    integer     :: ir, itheta, iphi, ivpar, imu
    real(RKIND) :: coeff_intdth, coeff_intdthdphi, coeff_intdphi

    ! -> variable for parallelization
    integer :: tid, nbelements
    integer :: nbreqr, nbbcast_r, nbbcast_s
    integer :: iproc, itag_phi, local_id, procloop, maxproc
    
    real(RKIND)   , dimension(:)  , pointer :: ldiag
    real(RKIND)   , dimension(:)  , pointer :: udiag
    real(RKIND)   , dimension(:)  , pointer :: diag
    real(RKIND)   , dimension(:)  , pointer :: rhs_tmp
    real(RKIND)   , dimension(:)  , pointer :: Phi_TPavg
    real(RKIND)   , dimension(:)  , pointer :: rhoi_TPavg
    real(RKIND)   , dimension(:)  , pointer :: Phi_Sflux_avg
    real(RKIND)   , dimension(:)  , pointer :: rhoi_Tavg_loc
    real(RKIND)   , dimension(:)  , pointer :: Phitilde_Sflux_avg
    complex(CKIND), dimension(:)  , pointer :: Crhs_tmp
    complex(CKIND), dimension(:,:), pointer :: fft_rho 
    real(RKIND)   , dimension(:,:), pointer :: rhoi_Tavg 

    integer     :: ikth, indx_grid, i1
    integer     :: i, j, k, LU_err
    logical     :: stopit
    real(RKIND) :: icoef, inv_N2, LinvZ02Te_Phitilde_tmp
    real(RKIND) :: Phitilde_JsTavg

#ifdef TIMER
    integer(TIMEPREC) :: tdeb, tfin, tth1, tth2
    real(RKIND)       :: tthd

    call clck_time(tdeb)
    tth1 = tdeb

#endif

    !*** initialize local variables ***
    N1        = (geom%Nr+1)-2
    N2        = (geom%Ntheta+1)-1
    inv_N2    = 1._RKIND/real(N2)
    indx_grid = 0   ! The grid arrays begin with the 0 index

    !*** initialize local variables ***
    stopit             = .false.
    tid                = 1
    rhoi_TPavg         => Romp1_0Nr(tid)%val
    Phi_TPavg          => Romp2_0Nr(tid)%val
    Phitilde_Sflux_avg => Romp5_0Nr(tid)%val
    Phi_Sflux_avg      => Romp6_0Nr(tid)%val
    rhoi_Tavg          => Rarray2_NrNphi
    rhoi               => Rarray1_NrNthetaDomphi
    Phitilde           => Rarray2_NrNthetaDomphi
    Phi                => Rarray3_NrNthetaDomphi
    !*** rhsdistrib is a larger data structure than rhoi, Phi, Phitilde
    rhsdistrib         => Rarray1_NrNthetaParphi

    !-----------------------------------------------------------
    !-> step 1 : compute rho = 1/ne0*sum_s Zs*[nGs-nGs_eq]
    !-----------------------------------------------------------
    call compute_rho_distrib3D(geom,J0,f,Sfmu_eq, &
      init_prof,init_magnet,init_curr,rhoi,rhsdistrib)

#ifdef TIMER
    call clck_time(tdeb)
#endif

    !---------------------------------------------------------
    !-> step 2 : solve the equation:
    !     L.Phitilde + 1/(Zi Te) Phitilde = 
    !                   rhoi - <rhoi>_{theta,phi}
    !   where Phitilde = Phi-Phibar and where
    !   <.>_{theta,phi} = 1/(Ltheta*Lphi) \int . dtheta dphi 
    ! Rk : The boundary conditions are Dirichlet conditions
    !---------------------------------------------------------
    if (iphistart .ne. -1) then
!$OMP PARALLEL default(none) &
!$OMP private(rhoi_Tavg_loc,tid, &
!$OMP  ir,itheta,iphi)  &
!$OMP shared(f,iphistart,dom_mapphi, &
!$OMP  Romp3_0Nr,rhoi)
!$OMP BARRIER
#ifdef _OPENMP
      tid = 1+omp_get_thread_num()
#else
      tid = 1
#endif
      rhoi_Tavg_loc => Romp3_0Nr(tid)%val

      !----> compute rhoi_TPavg = <rhoi>_{theta,phi}
      !-> Computation of \int rhoi dtheta for a parallel
      !->  domain in phi direction, stored in rhoi_Tavg_loc 
!$OMP DO SCHEDULE(static)
      do iphi = iphistart,iphistart+dom_mapphi-1
        do ir = 0,f%n1
          rhoi_Tavg_loc(ir) = 0._RKIND
        end do
        do itheta = 0,f%n2-1
          do ir = 0,f%n1
            rhoi_Tavg_loc(ir) = rhoi_Tavg_loc(ir) + rhoi(ir,itheta,iphi)
          end do
        end do
        do ir = 0,f%n1
          rhoi(ir,f%n2,iphi) = rhoi_Tavg_loc(ir)          
        end do
      end do
!$OMP END DO
!$OMP BARRIER
!$OMP END PARALLEL
    end if

#ifdef TIMER
!    call ppbarrier()
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tcmp(3))
    tdeb = tfin
#endif

    !-> distribution of \int rhoi dtheta to obtain all
    !->  rhoi_Tavg(0:Nr,0:Nphi)
    if (iphistart .ne. -1) then
#ifdef OLDCOMM
      nbbcast_s = 0
      do iphi = 0,f%n3-1
        itag_phi   = 7000 + iphi
        nbelements = (Nr+1)
        call MPI_IRECV(rhoi_Tavg(0,iphi),nbelements,MPI_REAL8, &
          moments_mapphi(iphi),itag_phi,mpi_comm_species, &
          bcast_reqr(iphi),ierr)
      end do
      do iphi = 0,f%n3-1
        if (moments_mapphi(iphi) .eq. pInSpecies_id) then
          nbelements = (Nr+1)
          itag_phi   = 7000 + iphi
          maxproc = MIN(Nbproc_species,Nphi)
          do procloop = 0,maxproc-1
            iproc = mod(procloop+pInSpecies_id,maxproc)
            call MPI_SEND(rhoi(0,f%n2,iphi),nbelements, &
              MPI_REAL8,iproc,itag_phi,mpi_comm_species,ierr)
            nbbcast_s = nbbcast_s + 1
          enddo
        endif
      end do
      call MPI_WAITALL(f%n3,bcast_reqr(0), &
              moments_status(:,0:nbbcast_r-1), ierr)
#else
      if (pInSpecies_id .eq. 0) then
        do iphi = 0,f%n3-1
          itag_phi   = 7000 + iphi
          nbelements = (Nr+1)
          call MPI_IRECV(rhoi_Tavg(0,iphi),nbelements,MPI_REAL8, &
            moments_mapphi(iphi),itag_phi,mpi_comm_species, &
            bcast_reqr(iphi),ierr)
        end do
      end if
      do iphi = 0,f%n3-1
        if (moments_mapphi(iphi) .eq. pInSpecies_id) then
          nbelements = (Nr+1)
          itag_phi   = 7000 + iphi
          iproc = 0
          call MPI_SEND(rhoi(0,f%n2,iphi),nbelements, &
            MPI_REAL8,iproc,itag_phi,mpi_comm_species,ierr)
        end if
      enddo
      if (pInSpecies_id .eq. 0) then
        call MPI_WAITALL(f%n3, bcast_reqr(0:f%n3-1), &
                moments_status(:,0:f%n3-1), ierr)
      end if
      nbelements = (Nr+1)*(f%n3)
      call MPI_BCAST(rhoi_Tavg(0,0),nbelements,MPI_REAL8, &
        0,mpi_comm_moments,ierr)
#endif
    endif

    call ppbarrier()
#ifdef TIMER
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tcmm(3))
    tdeb = tfin
#endif
    
    !---> Computation of <rhoi>_theta,phi equal to 
    !--->  \int rhoi dtheta dphi 
    if (iphistart .ne. -1) then
      do ir = 0,geom%Nr
        rhoi_TPavg(ir) = 0._RKIND
      end do
      do iphi = 0,geom%Nphi-1
        do ir = 0,geom%Nr
          rhoi_TPavg(ir) = rhoi_TPavg(ir) + rhoi_Tavg(ir,iphi)
        end do
      enddo
      coeff_intdthdphi = 1._RKIND/(geom%Nphi*geom%Ntheta)
      do ir = 0,geom%Nr
        rhoi_TPavg(ir)  = rhoi_TPavg(ir)*coeff_intdthdphi
      end do
    end if
     
    !*** Restore the periodic boundary conditions ***
    !***  in theta direction for rho              ***
    if (iphistart .ne. -1) then
      do iphi = 0,geom%Nphi-1
        if (moments_mapphi(iphi) .eq. pInSpecies_id) then
          !*** restore periodic conditions in theta ***
          do ir = 0,geom%Nr
            rhoi(ir,f%n2,iphi) = rhoi(ir,0,iphi)
          end do
        end if
      end do

      do tid = 1, Nbthread
         call FTR_ALLOC_2_ZERO(Comp1_1Nrp1_1Ntheta(tid)%val,1,Nr+1,1,Ntheta)
         call FTR_ALLOC_2_ZERO(Comp2_1Nrp1_1Ntheta(tid)%val,1,Nr+1,1,Ntheta)
      enddo

!$OMP PARALLEL default(none) &
!$OMP private(nbbcast_s,tid, &
!$OMP  ir,itheta,iphi,imu, &
!$OMP  ldiag,udiag,diag, &
!$OMP  Crhs_tmp,fft_rho, &
!$OMP  i1,ikth,LU_err,Phitilde_JsTavg) &
!$OMP shared(geom,f,poiss,jacobian_space, &
!$OMP  indx_grid,N1,N2,inv_N2,stopit, &
!$OMP  moments_mapphi,pInSpecies_id, &
!$OMP  QN_coef_Phi00,rhoi,Phitilde,rhoi_TPavg, &
!$OMP  L_ldiag,Lm_invZ02Te_diag,L_udiag, &
!$OMP  Romp2_1Nrp1,Romp3_1Nrp1, &
!$OMP  Romp4_1Nrp1,Romp3_0Nr, &
!$OMP  Comp1_1Nrp1_1Ntheta,Comp2_1Nrp1_1Ntheta)
!$OMP BARRIER
#ifdef _OPENMP
      tid = 1+omp_get_thread_num()
#else
      tid = 1
#endif
      ldiag    => Romp2_1Nrp1(tid)%val
      diag     => Romp3_1Nrp1(tid)%val
      udiag    => Romp4_1Nrp1(tid)%val
      fft_rho  => Comp1_1Nrp1_1Ntheta(tid)%val
      Crhs_tmp => Comp2_1Nrp1_1Ntheta(tid)%val(1:,1)  !used (1:Nr)
      
      LU_err = 0

!$OMP DO SCHEDULE(STATIC,1)
      !----> compute Phitilde = Phi-Phibar by solving
      !  A.Phitilde = b with A = (L+1/(Zi Te)) and 
      !   b = rhoi - <rhoi>_{theta,phi} 
      do iphi = 0,geom%Nphi-1
        if (moments_mapphi(iphi) .eq. pInSpecies_id) then
          !----> compute rhoi-rhoi_TPavg
          do ikth = 1,N2
            itheta = ikth + (indx_grid-1)
            do i1 = 1,N1
              ir               = indx_grid + i1 
              fft_rho(i1,ikth) = rhoi(ir,itheta,iphi) - &
                rhoi_TPavg(ir)
            end do
          end do
          !----> compute the poloidal Fourier transform 
          !      FFT_m(rhoi-rhoi_TPavg)
          call fourrow(fft_rho,1)

          !----> solve in the Fourier space
          !      (L+ 1/(Zi Te)) Phitilde^m = FFT_m(rhoi-rhoi_TPavg)
          do ikth = 1,(N2/2+1)
            !----> for each positive poloidal Fourier mode
            do i1 = 1,N1
              ldiag(i1)    = L_ldiag(i1)
              diag(i1)     = Lm_invZ02Te_diag(i1,ikth)
              udiag(i1)    = L_udiag(i1)
              Crhs_tmp(i1) = fft_rho(i1,ikth)
            end do
            call modThomas_c(N1,ldiag,diag,udiag,Crhs_tmp,LU_err)
            if (LU_err.ne.0) then
!$OMP CRITICAL
              print *,'problem in modThomas_c solving for ikth = ', &
                ikth
              stopit = .true.
!$OMP END CRITICAL
            endif
            do i1 = 1,N1
              fft_rho(i1,ikth) = Crhs_tmp(i1)
            end do
            !----> compute the negative part by using the 
            !      fact that it is equivalent to the conjugate value
            !      because the result is real
            if ( (ikth.ne.1).and.(ikth.ne.(N2/2+1)) ) then
              do i1 = 1,N1
                fft_rho(i1,N2+2-ikth) = conjg(Crhs_tmp(i1))
              end do
            end if
          end do
          
          !----> compute the inverse FFT
          call fourrow(fft_rho,-1)
          do ikth = 1,N2
            itheta = ikth+indx_grid-1
            do i1 = 1,N1
              ir = i1+indx_grid
              Phitilde(ir,itheta,iphi) = real(fft_rho(i1,ikth))*inv_N2
            end do
            !------> Dirichlet boundary conditions in radial direction
            Phitilde(indx_grid,itheta,iphi) = 0._RKIND
            Phitilde(N1+1,itheta,iphi)      = 0._RKIND
          end do
       
          !---> compute \int Phitilde Js dtheta where 
          !--->   Phitilde = Phi- <Phi>_{theta,phi} and store it
          !---> at the end of the 3D array
          do ir = 0,geom%Nr                         
            Phitilde_JsTavg = 0._RKIND                         
            do itheta = 0,geom%Ntheta-1             
              Phitilde_JsTavg = Phitilde_JsTavg + &
                Phitilde(ir,itheta,iphi)*jacobian_space(ir,itheta)
            end do 
            Phitilde(ir,geom%Ntheta,iphi) = Phitilde_JsTavg
          end do
        endif
      end do
!$OMP END DO
!$OMP BARRIER
!$OMP END PARALLEL

      do tid = 1, Nbthread
         call FTR_DEALLOC(Comp1_1Nrp1_1Ntheta(tid)%val)
         call FTR_DEALLOC(Comp2_1Nrp1_1Ntheta(tid)%val)
      enddo
    endif

#ifdef TIMER
!    call ppbarrier()
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tcmp(4))
    tdeb = tfin
#endif

    !----------------------------------------------------------
    !-> step 3 : compute <Phitilde>_{flux surf} the 
    !    flux surface average of Phitilde, i.e
    !     <Phitilde>_{flux surf} = 
    !           \int Phitilde Js(r,theta) dtheta dphi /
    !           \int Js(r,theta) dtheta dphi
    !  where Phitilde = Phi - <Phi>_{theta,phi}
    !----------------------------------------------------------
    if (iphistart .ne. -1) then
      do ir = 0,geom%Nr
        do iphi = 0,geom%Nphi-1
          if (moments_mapphi(iphi) .eq. pInSpecies_id) then
            rhoi_Tavg(ir,iphi) = Phitilde(ir,geom%Ntheta,iphi) 
          end if
        enddo
       end do
       ir = (geom%Nr+1)*dom_mapphi
       call MPI_GATHER(rhoi_Tavg(0,iphistart),ir,MPI_REAL8, &
         Rarray_NrNphi(0,0),ir,MPI_REAL8, &
         0,mpi_comm_moments,ierr)
       if (pInSpecies_id .eq. 0) then
         coeff_intdphi = geom%dphi
         coeff_intdth = geom%dtheta
         do ir = 0,geom%Nr
           icoef = coeff_intdphi*coeff_intdth/intdthetadphi_Js(ir)
           Phitilde_Sflux_avg(ir) = 0._RKIND
           do iphi = 0,geom%Nphi-1
             Phitilde_Sflux_avg(ir) = Phitilde_Sflux_avg(ir) + &
               Rarray_NrNphi(ir,iphi)
           enddo
           Phitilde_Sflux_avg(ir) = Phitilde_Sflux_avg(ir)*icoef
         end do
       end if
     end if
    
    nbelements = (geom%Nr+1)
    call MPI_BCAST(Phitilde_Sflux_avg,nbelements,MPI_REAL8, &
      0,mpi_comm_species,ierr)
    call ppbarrier()

#ifdef TIMER
!    call ppbarrier()
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tcmm(4))
    tdeb = tfin
#endif
    
    !----------------------------------------------------------
    !-> step 4 : compute the right side vector
    !    <rhoi(ir)>_{theta,phi} + 
    !        (L+1/(Zi*Te))*<Phi-Phibar>_{flux surf}
    !----------------------------------------------------------
    if (iphistart .ne. -1) then
      rhs_tmp => Romp1_1Nrp1(1)%val         !used (1:Nr)
      do i1 = 1,N1
        ir                     = indx_grid + i1
        LinvZ02Te_Phitilde_tmp = &
          L_ldiag(i1)*Phitilde_Sflux_avg(ir-1) + &
          L_invZ02Te_diag(i1)*Phitilde_Sflux_avg(ir) + &
          L_udiag(i1)*Phitilde_Sflux_avg(ir+1)
        rhs_tmp(i1)           = rhoi_TPavg(ir) + &
          LinvZ02Te_Phitilde_tmp 
      end do
    end if

    !-------------------------------------------------------------
    !-> step 5 : compute <Phi>_{flux surf} by solving
    !    (L + (1-lambda)*1/(Zi Te)) <Phi>_{flux surf} = 
    !    rhoi_TPavg(ir) + (L+1/(Zi*Te))*<Phi-Phibar>_{flux surf}
    !-------------------------------------------------------------
    if (iphistart .ne. -1) then
      if (QN_coef_Phi00==1) then
        !--> case lambda = 1
        call modThomas_d(N1,L_ldiag,L_diag,L_udiag,rhs_tmp,LU_err)
      else
        !--> case lambda = 0
        call modThomas_d(N1,L_ldiag,L_invZ02Te_diag,L_udiag,rhs_tmp,LU_err)
      end if
      if (LU_err.ne.0) then
        print *,'problem in modThomas_d solving'
      end if
      do i1 = 1,N1
        ir                = indx_grid + i1
        Phi_Sflux_avg(ir) = rhs_tmp(i1)  
      end do

      !----> Dirichlet or Neumann boundary condition
      Phi_Sflux_avg(indx_grid) = coeff_Neumann * &
           Phi_Sflux_avg(indx_grid+1)  
      Phi_Sflux_avg(N1+1)      = 0._RKIND

      !*** Saving of flux surface average of Phi ***
      do ir = indx_grid,geom%Nr
        Phi_FSavg_diag(ir) = Phi_Sflux_avg(ir) 
        Phi_TPavg(ir)      = Phi_Sflux_avg(ir) - Phitilde_Sflux_avg(ir)
        Phi00_diag(ir)     = Phi_TPavg(ir)
      end do
    end if

    !----------------------------------------------------------
    !-> step 6 : compute Phi(r,theta,phi) as
    !    Phi(r,theta,phi) = Phitilde(r,theta,phi) 
    !      - <Phitilde>_{flux surf} + <Phi>_{flux surf}
    !----------------------------------------------------------
#ifdef _OPENMP
!$OMP PARALLEL default(none) &
!$OMP private(tid,ir,itheta,iphi) &
!$OMP shared(iphistart,geom,moments_mapphi, &
!$OMP  pInSpecies_id,Phitilde,Phi,Phi_TPavg)
!$OMP BARRIER
    tid = 1+omp_get_thread_num()
#else
    tid = 1
#endif

    if (iphistart .ne. -1) then
!$OMP DO SCHEDULE(STATIC)
      do iphi = 0,geom%Nphi-1
        if (moments_mapphi(iphi) .eq. pInSpecies_id) then
          do itheta = 0,geom%Ntheta-1
            do ir = 0,geom%Nr
              Phi(ir,itheta,iphi) = Phitilde(ir,itheta,iphi) + &
                Phi_TPavg(ir)
            end do
          end do
        end if
      end do
!$OMP END DO
    end if
!$OMP BARRIER
!$OMP END PARALLEL

    !*** filter applied on the electric potentiel if required ***
    if ((FFTpotential_filter.eq.1).or. &
        (FFTpotential_filter.eq.2).or. &
        (single_n).or.(single_m)) then
      call QNsolver_distrib_filter(poiss,Phi,geom,init_prof)
    end if

#ifdef TIMER
    call clck_time(tfin)
    call clck_diff(tdeb,tfin,tcmp(1))
    tdeb = tfin
#endif


    !**********************************************
    !*** computation of the number of electrons ***
    !***  and the number of ions                ***
    !**********************************************
    call compute_nbparticles(geom,init_prof,rhoi, &
      Phi,Phi_Sflux_avg,nbelectrons_diag,nbions_diag)

#ifdef TIMER
    call clck_time(tfin)
    call clck_diff(tdeb,tfin,tcmp(3))
    tdeb = tfin

#endif

    !********************************
    !*      distribution of Phi     *
    !********************************
    !-> message receive
    nbbcast_r = 0
    nbelements = (Nr+1)*(Ntheta+1)
    do iphi = 0,f%n3-1
      itag_phi = 9000 + iphi
      if (moments_locphi(iphi) .eq. plocal_id) then
        call MPI_IRECV(poiss%Phi_NrNthPNphi_Cmu(0,0,iphi), &
          nbelements,MPI_REAL8,moments_mapphi(iphi), &
          itag_phi,mpi_comm_species,bcast_reqr(nbbcast_r),ierr)
        nbbcast_r = nbbcast_r + 1
      end if
    enddo
    !-> message sending
    if (iphistart .ne. -1) then
      nbbcast_s = 0
      do iphi = 0,geom%Nphi-1
        if (moments_mapphi(iphi) .eq. pInSpecies_id) then
          nbelements = (Nr+1)*(Ntheta+1)
          itag_phi   = 9000 + iphi
          do procloop = 0,Nbproc_species-1
            iproc = mod(pInSpecies_id+procloop,Nbproc_species)
            local_id = mod(iproc,Nbproc_loc)
            if (moments_locphi(iphi) .eq. local_id) then
              call MPI_SEND(Phi(0,0,iphi),nbelements,MPI_REAL8, &
                iproc,itag_phi,mpi_comm_species, &
                ierr)
              nbbcast_s = nbbcast_s+1
            end if
          enddo
        endif
      end do
    endif
    call MPI_WAITALL(nbbcast_r, bcast_reqr(0:nbbcast_r-1), &
            moments_status(:,0:nbbcast_r-1), ierr)

    !*** duplicate periodic value ***
    do iphi = 0,geom%Nphi-1
      if (moments_mapphi(iphi) .eq. pInSpecies_id) then
        do itheta = 0, geom%Ntheta-1
          do ir = 0,geom%Nr
            poiss%Phi_NrNthPNphi_Cworld(ir,itheta,iphi) = &
              Phi(ir,itheta,iphi)
          end do
        end do
        do ir = 0,geom%Nr
          poiss%Phi_NrNthPNphi_Cworld(ir,geom%Ntheta,iphi) = &
            Phi(ir,0,iphi)
        end do
      end if
      if (moments_locphi(iphi) .eq. plocal_id) then
        do ir = 0,geom%Nr
          poiss%Phi_NrNthPNphi_Cmu(ir,geom%Ntheta,iphi) = &
            poiss%Phi_NrNthPNphi_Cmu(ir,0,iphi)
        end do
      end if
    end do

    call ppbarrier() 
#ifdef TIMER
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tcmm(5))
    tdeb = tfin
#endif

#ifdef TIMER
    tvect(1:12) = (/ tcmm(1:5), tcmp(1:5), tbar(1:2) /)
    call pp_reduce_vector(tvect,12,tmin,tmax,tavg)
    if (plocal_id .eq. 0) then
      write(6,'(I4,A,5F10.6,A,5F10.6,A,2F10.6)') pglobal_id, &
        " TMIN poisson5A, tcomm ", tmin(1:5), &
        " tcomp ",tmin(6:10)," tbarrier ",tmin(11:12)
      write(6,'(I4,A,5F10.6,A,5F10.6,A,2F10.6)') pglobal_id, &
        " TMAX poisson5A, tcomm ", tmax(1:5), &
        " tcomp ",tmax(6:10)," tbarrier ",tmax(11:12)
      write(6,'(I4,A,5F10.6,A,5F10.6,A,2F10.6)') pglobal_id, &
        " TAVG poisson5A, tcomm ", tavg(1:5), &
        " tcomp ",tavg(6:10)," tbarrier ",tavg(11:12)
    end if
    call clck_ldiff(tth1,tfin,tthd)
    if (pglobal_id .eq. 0) then
      write(6,'(I4,A,F13.6)') pglobal_id, &
        " Temps poisson5B,  tcomp ",tthd
    end if

#endif

    if (stopit) then
      print*, 'solve_QNequation_distrib3D : problem in Poisson solving'
      stop
    end if
#endif 
    ! HAS_QN_V5
  end subroutine solve_QNequation_distrib3D


  !----------------------------------------------------------
  ! Adding of filter in the Fourier space if required:
  !  - if filter choice = 1 a diagonal filter 
  !     [ m = n*q(r) +- delta_m ] is applied, i.e:
  !       with m0 = n*q(r), then FFT_PHI is put equal to 0
  !        if abs(m-m0)>delta_m                           
  !  Rk : this test is equivalent to                 
  !   ktheta = kphi*q(r)*(Lphi/Ltheta) +- 2*pi*delta_m/Ltheta
  !  with ktheta = 2*pi*m/Ltheta               
  !   and kphi   = 2*pi*n/Lphi                     
  !---------------------------------------------------------- 
  subroutine QNsolver_filter(Phi,geom,init_prof)
    use globals, only : FFTpotential_filter, single_n, single_m, &
      FFTpotential_filter_deltam, m, n
#ifdef _OPENMP
    use OMPutils_module, only : omp_get_thread_num
#endif
    use geometry_class
    use init_profile_class    
    use fft_NRF90_module
    use globals        , only : Nbthread
    real(RKIND), dimension(0:,0:,0:), intent(inout) :: Phi
    type(geometry)                  , intent(in)    :: geom
    type(init_profile)              , intent(in)    :: init_prof

#ifdef HAS_QN_FILTER
    integer :: tid, isign
    integer :: i, j, k

    !*** variables for the diagonal filter ***
    integer     :: N2, N3
    real(RKIND) :: ktheta, ktheta0
    real(RKIND) :: ktheta_min, ktheta_max 
    real(RKIND) :: kphi
    real(RKIND) :: coef_fftinv
    real(RKIND) :: mask
    real(RKIND) :: qmin, qmax, qloc
    complex(CKIND),dimension(:,:), pointer :: FFT2D_Phi

    N2          = (geom%Ntheta+1)-1 
    N3          = (geom%Nphi+1)-1
    ktheta_min  = minval(QN_kth)
    ktheta_max  = maxval(QN_kth)
    coef_fftinv = 1._RKIND/(N2*N3)

    call write_expr("N2", "(geom%Ntheta+1)-1")
    call write_expr("N3", "(geom%Nphi+1)-1")

    do tid = 1, Nbthread
       call FTR_ALLOC_2_ZERO(Comp_tmp2D(tid)%val,1,N3,1,N2)
       call FTR_ALLOC_2_ZERO(Comp_data_i(tid)%val,1,N3,1,N2)
    enddo

!$OMP PARALLEL private(isign,tid,k,j,i,mask, &
!$OMP  ktheta,ktheta0,kphi,FFT2D_Phi,qmin,qmax,qloc) &
!$OMP  default(shared)
!$OMP BARRIER
#ifdef _OPENMP
    tid = 1+omp_get_thread_num()
#else
    tid = 1
#endif 
    FFT2D_Phi => Comp_tmp2D(tid)%val   ! FFT2D_Phi(k,j)

!$OMP DO SCHEDULE(STATIC)
    do i = 1,geom%Nr+1 
      !*** Perform 2D FFT of Phi ***
      isign = 1
      do k = 1,N3
        do j = 1,N2
          Comp_data_i(tid)%val(j,k) = Phi(i-1,j-1,k-1)
        end do
      end do
      call fourrow(Comp_data_i(tid)%val,isign)
      do k = 1,N3
        do j = 1,N2
          FFT2D_Phi(k,j) = Comp_data_i(tid)%val(j,k)
        end do
      end do
      call fourrow(FFT2D_Phi,isign)

      !*** diagonal filter ***
      if (FFTpotential_filter.eq.1) then
        do j = 1,N2
          ktheta = QN_kth(j)
          do k = 1,N3
            kphi    = QN_kphi(k)
            ktheta0 = -kphi*geom%Lphi / &
              (init_prof%iota(i-1)*geom%Ltheta)
            ktheta0 = max(ktheta0,ktheta_min)
            ktheta0 = min(ktheta0,ktheta_max)
            if ( abs(ktheta-ktheta0).gt.FFTpotential_filter_deltam ) then
              FFT2D_Phi(k,j) = 0._CKIND
            end if
          end do
        end do
      else if (FFTpotential_filter.eq.2) then
          qmax = maxval(1._RKIND/init_prof%iota)
          qmin = minval(1._RKIND/init_prof%iota)
          do j = 1,N2
            ktheta = QN_kth(j)
            do k = 1,N3
              kphi    = QN_kphi(k)
              if (kphi .eq. 0) then
                if (ktheta .ne. 0) FFT2D_Phi(k,j) = 0._CKIND
              else
                qloc = - ktheta/kphi
                if ((qloc .gt. qmax).or.(qloc.lt.qmin)) then
                  FFT2D_Phi(k,j) = 0._CKIND
                end if
              end if
            end do
          end do
      end if
      !*** the single n mode is retained ***
      if (single_n) then
        do k = 1,(N3+1)/2
          mask =  ON - min(ON,abs(ON*(k-1-n)))
          do j = 1,N2
            FFT2D_Phi(k,j) = mask*FFT2D_Phi(k,j)
          end do
        end do
        do k = (N3+1)/2+1,N3
          mask =  ON - min(ON,abs(ON*(N3-k+1-n)))
          do j = 1,N2
            FFT2D_Phi(k,j) = mask*FFT2D_Phi(k,j)
          end do
        end do
      end if
      !*** the single m mode is retained ***
      if (single_m) then
        do j = 1,(N2+1)/2
          mask =  ON - min(ON,abs(ON*(j-1-m)))
          do k = 1,N3
            FFT2D_Phi(k,j) = mask*FFT2D_Phi(k,j)
          end do
        end do
        do j = (N2+1)/2+1,N2
          mask =  ON - min(ON,abs(ON*(N3-j+1-m)))
          do k = 1,N3
            FFT2D_Phi(k,j) = mask*FFT2D_Phi(k,j)
          end do
        end do
      end if

      !*** Perform 2D inverse FFT of Phi ***
      isign = -1
      do k = 1,N3
        do j = 1,N2
          Comp_data_i(tid)%val(j,k) = FFT2D_Phi(k,j)
        end do
      end do
      call fourrow(Comp_data_i(tid)%val,isign)
      do k = 1,N3
        do j = 1,N2
          FFT2D_Phi(k,j) = Comp_data_i(tid)%val(j,k)
        end do
      end do
      call fourrow(FFT2D_Phi,isign)
      do k = 0,geom%Nphi-1
        do j = 0,geom%Ntheta-1
          Phi(i-1,j,k) = &
            dreal(FFT2D_Phi(k+1,j+1))*coef_fftinv
        enddo
      enddo
    enddo
!$OMP END DO
!$OMP BARRIER
!$OMP END PARALLEL

   do tid = 1, Nbthread
     call FTR_DEALLOC(Comp_tmp2D(tid)%val)
     call FTR_DEALLOC(Comp_data_i(tid)%val)
   enddo
#endif 
    ! HAS_QN_FILTER
  end subroutine QNsolver_filter


  !----------------------------------------------------------
  ! Adding of filter in the Fourier space if required:
  !  - if filter choice = 1 a diagonal filter 
  !     [ m = n*q(r) +- delta_m ] is applied, i.e:
  !       with m0 = n*q(r), then FFT_PHI is put equal to 0
  !        if abs(m-m0)>delta_m                           
  !       Rk : this test is equivalent to                 
  !        ktheta = kphi*q(r)*(Lphi/Ltheta) +-            
  !                  2*pi*delta_m/Ltheta                   
  !         with ktheta = 2*pi*m/Ltheta and               
  !              kphi   = 2*pi*n/Lphi                     
  !---------------------------------------------------------- 
  subroutine QNsolver_distrib_filter(pthis,Phi,geom,init_prof)
    use globals, only : RSendBufMPI_D, RRecvBufMPI_B, nbthread, &
      Phi_NthNphiPNr, single_m, single_n, FFTpotential_filter, &
      pInSpecies_id, Nbproc_species, mpi_comm_species, n, m, &
      FFtpotential_filter_deltam
#ifdef _OPENMP
    use OMPutils_module, only : omp_get_thread_num
#endif
    use geometry_class
    use init_profile_class    
    use fft_NRF90_module
    use MPIutils_module, only : irstart, mapir, dom_mapr_loc, &
      iphistart, dom_mapphi, moments_reqr, moments_reqs, &
      moments_status
    use mpiMod   
    use globals, only : Nr, Ntheta, Nbthread
    use MPIutils_module, only : dom_mapr
    type(poisson)      , intent(inout) :: pthis
    real(RKIND),    dimension(:,:,:), pointer       :: Phi
    type(geometry)                  , intent(in)    :: geom
    type(init_profile)              , intent(in)    :: init_prof
    
#ifdef HAS_QN_FILTER
    integer :: tid, isign
    integer :: i, j, k, ir

    !*** variables for the diagonal filter ***
    integer     :: N2, N3, maxir
    real(RKIND) :: ktheta, ktheta0
    real(RKIND) :: ktheta_min, ktheta_max 
    real(RKIND) :: kphi
    real(RKIND) :: coef_fftinv
    real(RKIND) :: mask
    real(RKIND) :: qmin, qmax, qloc
    complex(CKIND),dimension(:,:), pointer :: FFT2D_Phi
    integer iproc
    real(RKIND), dimension(:), pointer :: tmp1d
    integer nbreqr, nbreqs, startbuf, itag, nbelements, ierr
    integer iphi, itheta, distiphistart

    call FTR_ALLOC_1_ZERO(RRecvBufMPI_B,0,geom%Ntheta*geom%Nphi*dom_mapr)
    call FTR_ALLOC_1_ZERO(RSendBufMPI_D,0,(Nr+1) * (Ntheta+1) * dom_mapphi)

    N2          = (geom%Ntheta+1)-1 
    N3          = (geom%Nphi+1)-1
    ktheta_min  = minval(pthis%kth)
    ktheta_max  = maxval(pthis%kth)
    coef_fftinv = 1._RKIND/(N2*N3)

    call write_expr("N2", "(geom%Ntheta+1)-1")
    call write_expr("N3", "(geom%Nphi+1)-1")

    do tid = 1, Nbthread
       call FTR_ALLOC_2_ZERO(Comp_tmp2D(tid)%val,1,N3,1,N2)
       call FTR_ALLOC_2_ZERO(Comp_data_i(tid)%val,1,N3,1,N2)
    enddo

    tmp1d => RSendBufMPI_D
    
    ! Parallel algorithm for FFT filtering
    !  1) Send Phi(ir=block,itheta=*,iphi=block), 
    !      receive tmp1d(ir=block,itheta=*,iphi=*)
    !  2) compute FFT2D in (theta,phi), apply the filter and backward FFT2D
    !  3) Send tmp1d(ir=block,itheta=*,iphi=*) 
    !      receive Phi(ir=block,itheta=*,iphi=block)
    
    nbreqr = 0
    nbreqs = 0
    if (irstart .ne. -1) then
      do ir = 0, geom%Nr
        if (mapir(ir) .eq. pInSpecies_id) then
          do iproc  = 0, MIN(Nbproc_species,geom%Nphi)-1
            startbuf   = ((ir-irstart)*geom%Nphi + iproc*dom_mapphi) * &
              geom%Ntheta 
            itag       = ir*Nbproc_species+iproc
            nbelements = geom%Ntheta*dom_mapphi 
            call MPI_IRECV(&
              RRecvBufMPI_B(startbuf), &
              nbelements,MPI_REAL8,iproc,itag, &
              mpi_comm_species,moments_reqr(nbreqr),ierr)
            nbreqr = nbreqr + 1
          enddo
        end if
      enddo
    end if
    if (iphistart .ne. -1) then
      do ir = 0, geom%Nr
        do iphi = iphistart, iphistart+dom_mapphi-1
          do itheta = 0, geom%Ntheta-1
            tmp1d(itheta+geom%Ntheta*((iphi-iphistart)+dom_mapphi*ir)) = &
              Phi(ir,itheta,iphi)
          end do
        end do
      end do
    end if
    if (iphistart .ne. -1) then
      do ir = 0, geom%Nr
        itag       = ir*Nbproc_species+pInSpecies_id
        nbelements = geom%Ntheta*dom_mapphi 
        call MPI_ISEND(&
          tmp1d(geom%Ntheta*dom_mapphi*ir), &
          nbelements,MPI_REAL8,mapir(ir),itag, &
          mpi_comm_species,moments_reqs(nbreqs),ierr)
        nbreqs = nbreqs + 1
      enddo
    end if
    if (nbreqr .ne. 0) &
      call MPI_WAITALL(nbreqr, moments_reqr(0:nbreqr-1), &
              moments_status(:,0:nbreqr-1), ierr)
    if (nbreqs .ne. 0) &
      call MPI_WAITALL(nbreqs, moments_reqs(0:nbreqs-1), &
              moments_status(:,0:nbreqs-1),ierr)
    do ir = 0, geom%Nr
      if (mapir(ir) .eq. pInSpecies_id) then
        do iproc  = 0, MIN(Nbproc_species,geom%Nphi)-1
          startbuf   = ((ir-irstart)*geom%Nphi + iproc*dom_mapphi) * &
            geom%Ntheta 
          nbelements = geom%Ntheta*dom_mapphi 
          distiphistart = iproc*dom_mapphi
          do iphi = distiphistart, distiphistart+dom_mapphi-1
            do itheta = 0, geom%Ntheta-1
              Phi_NthNphiPNr(itheta,iphi,ir) = &
                RRecvBufMPI_B(startbuf+itheta+geom%Ntheta * &
                (iphi-distiphistart))
            end do
          end do
        enddo
      end if
    enddo
    if (irstart .ne. -1) then
!$OMP PARALLEL private(isign,tid,k,j,i,ir,mask, &
!$OMP  ktheta,ktheta0,kphi,FFT2D_Phi,qmin,qmax,qloc) &
!$OMP  default(shared)
!$OMP BARRIER
#ifdef _OPENMP
      tid = 1+omp_get_thread_num()
#else
      tid = 1
#endif 
      FFT2D_Phi => Comp_tmp2D(tid)%val   ! FFT2D_Phi(k,j)
       
!$OMP DO SCHEDULE(STATIC,1)
      do ir = irstart,irstart+dom_mapr_loc-1
        !*** Perform 2D FFT of Phi ***
        isign = 1
        do k = 1,N3
          do j = 1,N2
            Comp_data_i(tid)%val(j,k) = Phi_NthNphiPNr(j-1,k-1,ir)
          end do
        end do
        call fourrow(Comp_data_i(tid)%val,isign)
        do k = 1,N3
          do j = 1,N2
            FFT2D_Phi(k,j) = Comp_data_i(tid)%val(j,k)
          end do
        end do
        call fourrow(FFT2D_Phi,isign)
        
        !*** diagonal filter ***
        if (FFTpotential_filter.eq.1) then
          do j = 1,N2
            ktheta = pthis%kth(j)
            do k = 1,N3
              kphi    = pthis%kphi(k)
              ktheta0 = -kphi*geom%Lphi / &
                (init_prof%iota(ir)*geom%Ltheta)
              ktheta0 = max(ktheta0,ktheta_min)
              ktheta0 = min(ktheta0,ktheta_max)
              if ( abs(ktheta-ktheta0).gt.FFTpotential_filter_deltam ) then
                FFT2D_Phi(k,j) = 0._CKIND
              end if
            end do
          end do
        else if (FFTpotential_filter.eq.2) then
          qmax = maxval(1._RKIND/init_prof%iota)
          qmin = minval(1._RKIND/init_prof%iota)
          do j = 1,N2
            ktheta = pthis%kth(j)
            do k = 1,N3
              kphi    = pthis%kphi(k)
              if (kphi .eq. 0) then
                if (ktheta .ne. 0) FFT2D_Phi(k,j) = 0._CKIND
              else
                qloc = - ktheta/kphi
                if ((qloc .gt. qmax).or.(qloc.lt.qmin)) then
                  FFT2D_Phi(k,j) = 0._CKIND
                end if
              end if
            end do
          end do
        end if
        !*** the single n mode is retained ***
        if (single_n) then
          do k = 1,(N3+1)/2
            mask =  ON - min(ON,abs(ON*(k-1-n)))
            do j = 1,N2
              FFT2D_Phi(k,j) = mask*FFT2D_Phi(k,j)
            end do
          end do
          do k = (N3+1)/2+1,N3
            mask =  ON - min(ON,abs(ON*(N3-k+1-n)))
            do j = 1,N2
              FFT2D_Phi(k,j) = mask*FFT2D_Phi(k,j)
            end do
          end do
        end if
        !*** the single m mode is retained ***
        if (single_m) then
          do j = 1,(N2+1)/2
            mask =  ON - min(ON,abs(ON*(j-1-m)))
            do k = 1,N3
              FFT2D_Phi(k,j) = mask*FFT2D_Phi(k,j)
            end do
          end do
          do j = (N2+1)/2+1,N2
            mask =  ON - min(ON,abs(ON*(N2-j+1-m)))
            do k = 1,N3
              FFT2D_Phi(k,j) = mask*FFT2D_Phi(k,j)
            end do
          end do
        end if
        !*** Perform 2D inverse FFT of Phi ***
        isign = -1
        call fourrow(FFT2D_Phi,isign)
        do k = 1,N3
          do j = 1,N2
            Comp_data_i(tid)%val(j,k) = FFT2D_Phi(k,j)
          end do
        end do
        call fourrow(Comp_data_i(tid)%val,isign)
        
        do k = 1,N3
          do j = 1,N2
            Phi_NthNphiPNr(j-1,k-1,ir) = &
              dreal(Comp_data_i(tid)%val(j,k))*coef_fftinv
          end do
        end do
      enddo
!$OMP END DO
!$OMP BARRIER
!$OMP END PARALLEL
    end if
    Phi(:,:,:) = 0._RKIND
    if (irstart .ne. -1) then
      do ir = 0, geom%Nr
        if (mapir(ir) .eq. pInSpecies_id) then
          do iproc  = 0, MIN(Nbproc_species,geom%Nphi)-1
            startbuf   = ((ir-irstart)*geom%Nphi + iproc*dom_mapphi) * &
              geom%Ntheta 
            nbelements = geom%Ntheta*dom_mapphi 
            distiphistart = iproc*dom_mapphi
            do iphi = distiphistart, distiphistart+dom_mapphi-1
              do itheta = 0, geom%Ntheta-1
                RRecvBufMPI_B(startbuf+itheta+geom%Ntheta * &
                  (iphi-distiphistart)) = &
                  Phi_NthNphiPNr(itheta,iphi,ir)
              end do
            end do
          enddo
        end if
      enddo
    end if
    
    nbreqr = 0
    nbreqs = 0
    if (iphistart .ne. -1) then
      do ir = 0, geom%Nr
        itag       = ir*Nbproc_species+pInSpecies_id
        nbelements = geom%Ntheta*dom_mapphi 
        call MPI_IRECV(&
          tmp1d(geom%Ntheta*dom_mapphi*ir), &
          nbelements,MPI_REAL8,mapir(ir),itag, &
          mpi_comm_species,moments_reqr(nbreqr),ierr)
        nbreqr = nbreqr + 1
      enddo
    end if
    
    if (irstart .ne. -1) then
      do ir = 0, geom%Nr
        if (mapir(ir) .eq. pInSpecies_id) then
          do iproc  = 0, MIN(Nbproc_species,geom%Nphi)-1
            startbuf   = ((ir-irstart)*geom%Nphi + iproc*dom_mapphi) * &
              geom%Ntheta 
            itag       = ir*Nbproc_species+iproc
            nbelements = geom%Ntheta*dom_mapphi 
            call MPI_ISEND(&
              RRecvBufMPI_B(startbuf), &
              nbelements,MPI_REAL8,iproc,itag, &
              mpi_comm_species,moments_reqs(nbreqs),ierr)
            nbreqs = nbreqs + 1
          enddo
        end if
      enddo
    end if
    if (nbreqr .ne. 0) &
      call MPI_WAITALL(nbreqr, moments_reqr(0:nbreqr-1), &
            moments_status(:,0:nbreqr-1), ierr)
    if (nbreqs .ne. 0) &
      call MPI_WAITALL(nbreqs, moments_reqs(0:nbreqs-1), &
            moments_status(:,0:nbreqs-1), ierr)
    if (iphistart .ne. -1) then
      do ir = 0, geom%Nr
        do iphi = iphistart, iphistart+dom_mapphi-1
          do itheta = 0, geom%Ntheta-1
            Phi(ir,itheta,iphi) = &
              tmp1d(itheta+geom%Ntheta*((iphi-iphistart)+dom_mapphi*ir))
          end do
        end do
      end do
    end if

   call FTR_DEALLOC(RRecvBufMPI_B)
   call FTR_DEALLOC(RSendBufMPI_D)
   do tid = 1, Nbthread
     call FTR_DEALLOC(Comp_tmp2D(tid)%val)
     call FTR_DEALLOC(Comp_data_i(tid)%val)
   enddo

#endif 
    ! HAS_QN_FILTER
  end subroutine QNsolver_distrib_filter


  !----------------------------------------------------- 
  ! Solve quasi-neutrality equation
  !----------------------------------------------------- 
  subroutine solve_quasineutrality(geom,J0,f,Sfmu_eq, &
    init_prof,init_magnet,init_curr,poiss)
    use globals, only : solve_QN, QN_version, distrib3D, iter_glob
    use geometry_class
    use gyroaverage_class
    use fdistribu5d_class
    use init_profile_class
    use init_magnetic_class
    use init_current_class
    use clock_module
    type(geometry)                  , intent(in)    :: geom
    type(J0operator)                , intent(inout) :: J0
    type(fdistribu5d)               , intent(in)    :: f
    real(RKIND),    dimension(:,:,:), pointer       :: Sfmu_eq
    type(init_profile)              , intent(in)    :: init_prof
    type(init_magnetic)             , intent(in)    :: init_magnet
    type(init_current)              , intent(in)    :: init_curr
    type(poisson)                   , intent(inout) :: poiss

!R3 #include "code_instrumentation/r3_info.h" !R3
!R3 call r3_info_begin (r3_info_index_0, 'CPU_poisson')      !R3
    call Write_dyn_balise_open('CPU_quasineutrality') !R4

#ifdef TIMER
    call ppbarrier()
#endif 
    call clck_time(bclock_poisson)

    if (solve_QN) then
      !*** solver chosen in function of the value of lambda ***
      select case(QN_version)
      case (4)
        call solve_QNequation(geom,J0,f,Sfmu_eq, &
          init_prof,init_magnet,init_curr, &
          poiss%Phi_NrNthPNphi_Cworld)
      case(5)
        !-> only available if the 3D datas are distributed
        call solve_QNequation_distrib3D(geom,J0,f,Sfmu_eq, &
          init_prof,init_magnet,init_curr,poiss)
      end select
    else
      if (distrib3D) then
        poiss%Phi_NrNthPNphi_Cworld = 0._RKIND
        poiss%Phi_NrNthPNphi_Cmu    = 0._RKIND
      else
        poiss%Phi_NrNthPNphi_Cworld = 0._RKIND
      end if
    end if
    
    call clck_time(eclock_poisson)  
    if (iter_glob .eq. -1) then
      call clck_diff(bclock_poisson,eclock_poisson, &
        global_time_init)
    else
      call clck_diff(bclock_poisson,eclock_poisson, &
        global_time_poisson)
    end if
    call Write_dyn_balise_close('CPU_quasineutrality') !R4
!R3 call r3_info_end (r3_info_index_0) !R3
  end subroutine solve_quasineutrality
end module QNsolver_module
