!**************************************************************
!  Copyright Euratom-CEA
!  Authors : 
!     Virginie Grandgirard (virginie.grandgirard@cea.fr)
!     Chantal Passeron (chantal.passeron@cea.fr)
!     Guillaume Latu (guillaume.latu@cea.fr)
!     Xavier Garbet (xavier.garbet@cea.fr)
!     Philippe Ghendrih (philippe.ghendrih@cea.fr)
!     Yanick Sarazin (yanick.sarazin@cea.fr)
!  
!  This code GYSELA (for GYrokinetic SEmi-LAgrangian) 
!  is a 5D gyrokinetic global full-f code for simulating 
!  the plasma turbulence in a tokamak.
!  
!  This software is governed by the CeCILL-B license 
!  under French law and abiding by the rules of distribution 
!  of free software.  You can  use, modify and redistribute 
!  the software under the terms of the CeCILL-B license as 
!  circulated by CEA, CNRS and INRIA at the following URL
!  "http://www.cecill.info". 
!**************************************************************

#include "config-F.inc"

!#define OLDCOMM
!---------------------------------------------
! file : efield.F90
! date : 26/06/2000
!  computation of the electric potential and 
!  the three components of the electric field
!  which is given by the quasi-neutrality
!  equation
! (all the quantities are computed using the
!  cubic spline expression)
!---------------------------------------------
module efield_module
  use prec_const
  use globals
  use MPIutils_module
  use clock_module
  use wrap_alloc_module
  use geometry_class
  use spline1d_class
  use poisson_class

  use wrap_alloc_module
  implicit none

  type(nspline1d), private :: E_nspline1d_r
  type(pspline1d), private :: E_pspline1d_theta
  type(pspline1d), private :: E_pspline1d_phi
     
  !******************************
  contains
  !******************************
  
  
  !-------------------------------------------------
  ! Constructor 
  !-------------------------------------------------
  subroutine new_efield(geom)
    type(geometry), intent(in) :: geom
    integer :: istart_clamp, iend_clamp

    call write_expr("n", "geom%Nr")
    call new_spline1d_natural(E_nspline1d_r,geom%Nr,geom%dr)
    call write_expr("n", "geom%Ntheta")
    call new_spline1d_period(E_pspline1d_theta, &
      geom%Ntheta,geom%dtheta)
    call write_expr("n", "geom%Nphi")
    call new_spline1d_period(E_pspline1d_phi,geom%Nphi,geom%dphi)

    istart_clamp = max(0,istart_buf)
    iend_clamp   = min(geom%Nr,iend_buf)
	
    call write_expr("istart_clamp", "max(0,istart_buf)")
    call write_expr("iend_clamp"  , "min(geom%Nr,iend_buf)")	
	
    if (distrib3D) then
       call FTR_ALLOC_3(dJ0Phidr_NrNthPNphi,0,geom%Nr,0,geom%Ntheta,kstart,kend)
       call FTR_ALLOC_3(dJ0Phidtheta_NrNthPNphi,0,geom%Nr,0,geom%Ntheta,kstart,kend)
       call FTR_ALLOC_3(dJ0Phidphi_NrNthPNphi,0,geom%Nr,0,geom%Ntheta,kstart,kend)
       
       call FTR_ALLOC_3(dJ0Phidr_PNrPNthNphi,istart_clamp,iend_clamp,0,geom%Ntheta,0,geom%Nphi)
       call FTR_ALLOC_3(dJ0Phidtheta_PNrPNthNphi,istart_clamp,iend_clamp,0,geom%Ntheta,0,geom%Nphi)
       call FTR_ALLOC_3(dJ0Phidphi_PNrPNthNphi,istart_clamp,iend_clamp,0,geom%Ntheta,0,geom%Nphi)
    else
       call FTR_ALLOC_3(dJ0Phidr_PNrPNthNphi,0,geom%Nr,0,geom%Ntheta,0,geom%Nphi)
       call FTR_ALLOC_3(dJ0Phidtheta_PNrPNthNphi,0,geom%Nr,0,geom%Ntheta,0,geom%Nphi)
       call FTR_ALLOC_3(dJ0Phidphi_PNrPNthNphi,0,geom%Nr,0,geom%Ntheta,0,geom%Nphi)

       dJ0Phidr_NrNthPNphi     => NULL()
       dJ0Phidtheta_NrNthPNphi => NULL()
       dJ0Phidphi_NrNthPNphi   => NULL()
    end if
  end subroutine new_efield


  !-------------------------------------------------
  ! Destructor 
  !-------------------------------------------------
  subroutine delete_efield()
    call del_spline1d_natural(E_nspline1d_r)
    call del_spline1d_period(E_pspline1d_theta)
    call del_spline1d_period(E_pspline1d_phi)

    call FTR_DEALLOC(dJ0Phidr_PNrPNthNphi)
    call FTR_DEALLOC(dJ0Phidtheta_PNrPNthNphi)
    call FTR_DEALLOC(dJ0Phidphi_PNrPNthNphi)

    if (distrib3D) then
       call FTR_DEALLOC(dJ0Phidr_NrNthPNphi)
       call FTR_DEALLOC(dJ0Phidtheta_NrNthPNphi)
       call FTR_DEALLOC(dJ0Phidphi_NrNthPNphi)
    endif
  end subroutine delete_efield


  !-------------------------------------------------
  ! Computation of the radial derivate of the electric
  ! potential d(Phi)/dr by using the spline 
  ! interpolation coefficients 
  !-------------------------------------------------    
  subroutine comp_dPhidr_loc(geom,Phi, &
    begin_dim1,end_dim1,begin_dim2,end_dim2, &
    begin_dim3,end_dim3,dPhidr)
    use globals, only : Phi_BCr_left,Phi_BCr_right
#ifdef _OPENMP
    use OMPutils_module, only : omp_get_thread_num
#endif
    use OMPutils_module, only : Romp1_0Nr, Romp2_0Nr, &
      Romp_scoefr
    type(geometry)       , intent(in) :: geom
    real(RKIND), &
      dimension(:,:,:)   , pointer    :: Phi
    integer              , intent(in) :: begin_dim1
    integer              , intent(in) :: end_dim1
    integer              , intent(in) :: begin_dim2
    integer              , intent(in) :: end_dim2
    integer              , intent(in) :: begin_dim3
    integer              , intent(in) :: end_dim3
    real(RKIND), &
      dimension(:,:,:)   , pointer    :: dPhidr
    
    integer                            :: tid, ir, itheta, iphi
    real(RKIND), dimension(:), pointer :: scoefr, rhsr, drhs_dr

    !*** dPhi/dr computation ***  
#ifdef _OPENMP
!$OMP PARALLEL private(ir,itheta,iphi,&
!$OMP rhsr,scoefr,drhs_dr,tid) default(shared)
!$OMP BARRIER
    tid = 1+omp_get_thread_num()
#else
    tid = 1
#endif
    rhsr    => Romp1_0Nr(tid)%val
    scoefr  => Romp_scoefr(tid)%val
    drhs_dr => Romp2_0Nr(tid)%val

!$OMP DO SCHEDULE(static)
    do iphi = begin_dim3,end_dim3
      do itheta = begin_dim2,end_dim2
        do ir = 0,geom%Nr
          rhsr(ir) = Phi(ir,itheta,iphi)
        end do
        call natural_omp_deriv1d(E_nspline1d_r,geom%Nr,geom%dr, &
          Phi_BCr_left,Phi_BCr_right,rhsr,scoefr,drhs_dr)
        do ir = begin_dim1,end_dim1
          dPhidr(ir,itheta,iphi) = drhs_dr(ir) 
        end do
      enddo
    end do
!$OMP END DO
!$OMP BARRIER
!$OMP END PARALLEL
  end subroutine comp_dPhidr_loc
    

  !-------------------------------------------------
  ! Computation of the derivate in theta of the 
  ! electric potential d(Phi)/dtheta by using  
  ! the spline interpolation coefficients
  !-------------------------------------------------    
  subroutine comp_dPhidtheta_loc(geom,Phi, &
    begin_dim1,end_dim1,begin_dim2,end_dim2, &
    begin_dim3,end_dim3,dPhidth)
#ifdef _OPENMP
    use OMPutils_module, only : omp_get_thread_num
#endif
    use OMPutils_module, only : Romp1_0Ntheta, Romp_scoeftheta
    type(geometry)       , intent(in) :: geom
    real(RKIND), &
      dimension(:,:,:)   , pointer    :: Phi
    integer              , intent(in) :: begin_dim1
    integer              , intent(in) :: end_dim1
    integer              , intent(in) :: begin_dim2
    integer              , intent(in) :: end_dim2
    integer              , intent(in) :: begin_dim3
    integer              , intent(in) :: end_dim3
    real(RKIND), &
         dimension(:,:,:), pointer    :: dPhidth
    
    integer                            :: ir, itheta, iphi, tid
    real(RKIND), dimension(:), pointer :: scoeftheta, rhstheta


    ! *** dPhi/dtheta computation ***  
#ifdef _OPENMP
!$OMP PARALLEL private(ir,itheta,iphi,&
!$OMP rhstheta,scoeftheta,tid) default(shared)
!$OMP BARRIER
    tid = 1+omp_get_thread_num()
#else
    tid = 1
#endif
    scoeftheta => Romp_scoeftheta(tid)%val
    rhstheta   => Romp1_0Ntheta(tid)%val

!$OMP DO SCHEDULE(static)
    do iphi = begin_dim3,end_dim3
      do ir = begin_dim1,end_dim1
        do itheta = 0,geom%Ntheta
          rhstheta(itheta) = Phi(ir,itheta,iphi)
        end do
        call period_omp_deriv1d(E_pspline1d_theta, &
          geom%Ntheta,geom%dtheta,rhstheta,scoeftheta)

        do itheta = begin_dim2,end_dim2
          dPhidth(ir,itheta,iphi) = rhstheta(itheta) 
        end do
      enddo
    end do
!$OMP END DO
!$OMP BARRIER
!$OMP END PARALLEL
  end subroutine comp_dPhidtheta_loc
    

  !-------------------------------------------------
  ! Computation of the derivate in phi of the 
  ! electric potential d(Phi)/dphi by using  
  ! the spline interpolation coefficients
  !-------------------------------------------------    
  subroutine comp_dPhidphi_loc(geom,Phi, &
    begin_dim1,end_dim1,begin_dim2,end_dim2, &
    begin_dim3,end_dim3,dPhidphi)
    use globals        , only : Rarray1_Nphi, Rarray2_Nphi
#ifdef _OPENMP
    use OMPutils_module, only : omp_get_thread_num
#endif
    type(geometry)       , intent(in) :: geom
    real(RKIND), &
      dimension(:,:,:)   , pointer    :: Phi
    integer              , intent(in) :: begin_dim1
    integer              , intent(in) :: end_dim1
    integer              , intent(in) :: begin_dim2
    integer              , intent(in) :: end_dim2
    integer              , intent(in) :: begin_dim3
    integer              , intent(in) :: end_dim3
    real(RKIND), &
         dimension(:,:,:), pointer    :: dPhidphi
    
    integer                            :: ir, itheta, iphi
    real(RKIND), dimension(:), pointer :: Phiphi_tmp, dPhidphi_tmp

    Phiphi_tmp   => Rarray1_Nphi
    dPhidphi_tmp => Rarray2_Nphi

    ! *** dPhi/dphi computation ***  
    do itheta = begin_dim2,end_dim2
      do ir = begin_dim1,end_dim1
        do iphi = 0,geom%Nphi
          Phiphi_tmp(iphi) = Phi(ir,itheta,iphi)
        end do
        call period_deriv1d(E_pspline1d_phi,geom%Nphi,geom%dphi, &
          Phiphi_tmp,dPhidphi_tmp)
        do iphi = begin_dim3,end_dim3
          dPhidphi(ir,itheta,iphi) = dPhidphi_tmp(iphi) 
        end do
      enddo
    end do
  end subroutine comp_dPhidphi_loc


  !----------------------------------------------------------
  ! Computation of the three partial derivatives of Phi, i.e
  !  dPhi/dr, dPhi/dtheta and dPhi/dphi
  !  (used for the BSL method)
  !  Rk : used if the 3D data are not distributed, i.e
  !       (distrib3D = .false.)
  !----------------------------------------------------------
  subroutine comp_dPhidx_sequential(geom,sPhi, &
    SdPhidr,SdPhidtheta, SdPhidphi)
    use globals        , only : Phi_BCr_left, Phi_BCr_right, &
      iter_glob
#ifdef _OPENMP
    use OMPutils_module, only : omp_get_thread_num
#endif
    use OMPutils_module, only : Romp1_0Nr, Romp2_0Nr, &
      Romp1_0Ntheta, Romp1_0Nphi, &
      Romp_scoefr, Romp_scoeftheta, Romp_scoefphi
    use geometry_class
    type(geometry)       , intent(in)  :: geom
    real(RKIND), &
      dimension(0:,0:,0:), intent(in)  :: sPhi
    real(RKIND), &
      dimension(0:,0:,0:), intent(out) :: SdPhidr
    real(RKIND), &
      dimension(0:,0:,0:), intent(out) :: SdPhidtheta
    real(RKIND), &
      dimension(0:,0:,0:), intent(out) :: SdPhidphi

    integer :: tid
    integer :: ir, itheta, iphi

    real(RKIND), dimension(:), pointer :: rhsr
    real(RKIND), dimension(:), pointer :: drhs_dr
    real(RKIND), dimension(:), pointer :: rhstheta
    real(RKIND), dimension(:), pointer :: rhsphi
    real(RKIND), dimension(:), pointer :: scoefr
    real(RKIND), dimension(:), pointer :: scoeftheta
    real(RKIND), dimension(:), pointer :: scoefphi

#ifdef TIMER
    integer(TIMEPREC) :: tdeb, tfin, tth1, tth2
    real(RKIND)       :: tcmm(12), tbar(12), tcmp(12), tthd
    real(RKIND)       :: tvect(32), tmin(32), tmax(32), tavg(32)

    tcmp(1:4) = 0._RKIND
    tcmm(1:4) = 0._RKIND
    tbar(1:4) = 0._RKIND

    call clck_time(tdeb)
    tth1 = tdeb
    call ppbarrier()
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tbar(1))
    tdeb = tfin
    call clck_ldiff(bclock_compJ0,eclock_compJ0,tcmp(1))
#endif

    call clck_time(bclock_compJ0)

#ifdef _OPENMP
!$OMP PARALLEL private(ir,itheta,iphi, &
!$OMP rhsr,drhs_dr,rhstheta,rhsphi, &
!$OMP scoefr,scoeftheta,scoefphi,tid) default(shared)
!$OMP BARRIER
    tid = 1+omp_get_thread_num()
#else
    tid = 1
#endif
    rhsr       => Romp1_0Nr(tid)%val
    drhs_dr    => Romp2_0Nr(tid)%val
    rhstheta   => Romp1_0Ntheta(tid)%val
    rhsphi     => Romp1_0Nphi(tid)%val
    scoefr     => Romp_scoefr(tid)%val
    scoeftheta => Romp_scoeftheta(tid)%val
    scoefphi   => Romp_scoefphi(tid)%val

    !*** computation of dPhi/dr ***
!$OMP DO SCHEDULE(static)
    do iphi = 0,geom%Nphi
      do itheta = 0,geom%Ntheta
        do ir = 0,geom%Nr
          rhsr(ir) = sPhi(ir,itheta,iphi)
        end do
        !-> computation of dPhi/dr (saved in rhsr)
        call natural_omp_deriv1d(E_nspline1d_r, &
          geom%Nr,geom%dr,Phi_BCr_left,Phi_BCr_right, &
          rhsr,scoefr,drhs_dr)
        do ir = 0,geom%Nr
          SdPhidr(ir,itheta,iphi) = drhs_dr(ir)
        end do
      enddo
    end do
!$OMP END DO
!$OMP BARRIER

    !*** computation of dPhi/dtheta ***
!$OMP DO SCHEDULE(static)
    do iphi = 0,geom%Nphi
      do ir = 0,geom%Nr
        do itheta = 0,geom%Ntheta
          rhstheta(itheta) = sPhi(ir,itheta,iphi)
        end do
        !-> computation of dPhi/dtheta (saved in rhstheta)
        call period_omp_deriv1d(E_pspline1d_theta, &
          geom%Ntheta,geom%dtheta,rhstheta,scoeftheta)
        do itheta = 0,geom%Ntheta
          SdPhidtheta(ir,itheta,iphi) = rhstheta(itheta)
        end do
      enddo
    end do
!$OMP END DO
!$OMP BARRIER

!$OMP MASTER
    !*** computation of dPhi/dphi ***
    do itheta = 0,geom%Ntheta
      do ir = 0,geom%Nr
        do iphi = 0,geom%Nphi
          rhsphi(iphi) = sPhi(ir,itheta,iphi)
        end do
        !-> computation of dPhi/dphi (saved in rhsphi)
        call period_omp_deriv1d(E_pspline1d_phi, &
          geom%Nphi,geom%dphi,rhsphi,scoefphi)
        do iphi = 0,geom%Nphi
          SdPhidphi(ir,itheta,iphi) = rhsphi(iphi)
        end do
      enddo
    end do
!$OMP END MASTER
!$OMP BARRIER
!$OMP END PARALLEL
    call ppbarrier()
#ifdef TIMER
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tcmp(2))
    tdeb = tfin

    tvect(1:6) = (/ tcmm(1:3), tcmp(1:2), tbar(1) /)
    call pp_reduce_vector(tvect,6,tmin,tmax,tavg)
    if (plocal_id .eq. 0) then
      write(6,'(I4,A,3F13.6,A,2F13.6,A,F13.6)') pglobal_id, &
        " TMIN efieldA, tcomm ", tmin(1:3), &
        " tcomp ",tmin(4:5)," tbarrier ",tmin(6)
      write(6,'(I4,A,3F13.6,A,2F13.6,A,F13.6)') pglobal_id, &
        " TMAX efieldA, tcomm ", tmax(1:3), &
        " tcomp ",tmax(4:5)," tbarrier ",tmax(6)
      write(6,'(I4,A,3F13.6,A,2F13.6,A,F13.6)') pglobal_id, &
        " TAVG efieldA, tcomm ", tavg(1:3), &
        " tcomp ",tavg(4:5)," tbarrier ",tavg(6)
    end if

    if (pglobal_id .eq. 0) then
      call clck_ldiff(tth1,tfin,tthd)
      write(6,'(I4,A,F13.6)') pglobal_id, &
        " Temps efieldB,  total ",tthd
    endif
#endif
    call clck_time(eclock_compJ0)
    if (iter_glob .eq. -1) then
      call clck_diff(bclock_compJ0,eclock_compJ0,global_time_init)
    else
      call clck_diff(bclock_compJ0,eclock_compJ0,global_time_compJ0)
    end if
  end subroutine comp_dPhidx_sequential


  !-------------------------------------------------------------
  ! Computation of the three partial derivatives of Phi, i.e
  !  dPhi/dr, dPhi/dtheta and dPhi/dphi
  !  (used for the BSL method)
  ! Rk : same than 'comp_dPhidx' but using the fact 
  !      that the electric potential is distributed
  !      inside mpi_comm_mu world
  ! - SdPhidr   : (out) dPhi/dr   if gyroaverage=.false.
  !                     dJ0Phi/dr if gyroaverage=.true.
  ! - SdPhidth  : (out) dPhi/dtheta   if gyroaverage=.false.
  !                     dJ0Phi/dtheta if gyroaverage=.true.
  ! - SdPhidphi : (out) dPhi/dphi   if gyroaverage=.false.
  !                     dJ0Phi/dphi if gyroaverage=.true.
  ! - StranspdPhidr : (out) data structures that contain 
  !     the radial derivative of the electric potential
  !     with domain decomposition along phi direction
  ! - StranspdPhidth : (out) data structures that contain 
  !     the poloidal derivative of the electric potential
  !     with domain decomposition along phi direction
  ! - StranspdPhidphi : (out) data structures that contain 
  !     the toroidal derivative of the electric potential
  !     with domain decomposition along phi direction
  !
  !  Rk : used if the 3D data are distributed, i.e
  !       (distrib3D = .true.)
  !-------------------------------------------------------------
  subroutine comp_dPhidx_parallel(geom,poiss, &
    SdPhidr,SdPhidtheta, SdPhidphi, &
    StranspdPhidr,StranspdPhidtheta, StranspdPhidphi, &
    J0,imu,sPhi_PNrPNthetaNphi,fill_transp_data)
! Optimisation directive of intel compiler (ifort)
!dir$ optimize:1
#ifdef _OPENMP
    use OMPutils_module, only : omp_get_thread_num
#endif
    use OMPutils_module, only : Romp1_0Nr, Romp2_0Nr, &
      Romp1_0Ntheta, Romp1_0Nphi, &
      Romp_scoefr, Romp_scoeftheta, Romp_scoefphi, &
      Romp1_1Nrp1_1Nthetap1, Comp1_1Nrp1_1Ntheta
    use gyroaverage_class
    use globals
    use mpiMod

    type(geometry)               , intent(in) :: geom
    type(poisson)                , intent(in) :: poiss
    real(RKIND), dimension(:,:,:), pointer    :: SdPhidr            
    real(RKIND), dimension(:,:,:), pointer    :: SdPhidtheta        
    real(RKIND), dimension(:,:,:), pointer    :: SdPhidphi
    real(RKIND), dimension(:,:,:), pointer    :: StranspdPhidr      
    real(RKIND), dimension(:,:,:), pointer    :: StranspdPhidtheta  
    real(RKIND), dimension(:,:,:), pointer    :: StranspdPhidphi    
    type(J0operator), intent(inout), optional :: J0
    integer         , intent(in) , optional   :: imu
    real(RKIND), dimension(:,:,:), &
      pointer, optional                       :: sPhi_PNrPNthetaNphi 
    logical         , intent(in) , optional   :: fill_transp_data

    integer :: tid
    integer :: ir, itheta, iphi
    integer :: kith
    logical :: Phi_present, J0_present
    real(RKIND), dimension(:)  , pointer :: rhsr
    real(RKIND), dimension(:)  , pointer :: drhs_dr
    real(RKIND), dimension(:)  , pointer :: rhstheta
    real(RKIND), dimension(:)  , pointer :: rhsphi
    real(RKIND), dimension(:)  , pointer :: scoefr
    real(RKIND), dimension(:)  , pointer :: scoeftheta
    real(RKIND), dimension(:)  , pointer :: scoefphi
    real(RKIND), dimension(:,:), pointer :: rhs2d

    complex(CKIND), dimension(:,:), pointer :: Acomp

    real(RKIND), dimension(:,:,:), pointer :: pdPhidr     
    real(RKIND), dimension(:,:,:), pointer :: pdPhidtheta 
    real(RKIND), dimension(:,:,:), pointer :: pdPhidphi   

    integer, parameter :: nbmoments = 3
    integer :: nbelts, ierr, j
    integer :: loc_istart, loc_iend, loc_jstart, loc_jend
    integer :: nbreqr, nbreqs
    integer :: loc_kstart, loc_kend
    integer :: iphiloc, itag_phi, startbuf
    integer :: local_id, base, dep, blociphi
    integer :: procloop, tgtid, iphitmp

    integer :: bigtmp_size

#ifdef TIMER
    integer(TIMEPREC) :: tdeb, tfin, tth1, tth2
    real(RKIND)       :: tcmm(12), tbar(12), tcmp(12), tthd
    real(RKIND)       :: tvect(32), tmin(32), tmax(32), tavg(32)

    tcmp(1:4) = 0.
    tcmm(1:4) = 0.
    tbar(1:4) = 0.

    call clck_time(tdeb)
    tth1 = tdeb
    call ppbarrier()
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tbar(1))
    tdeb = tfin
#endif
    
    call write_param("nbmoments", nbmoments)
    ! Initialization of MPI buffer to 0
    call FTR_ALLOC_1_ZERO(RSendBufMPI_A,0,nbmoments * dom_locphi * Nbproc_loc *(dom_r+2*bufsize) * (dom_theta+2*bufsize))

    do tid = 1, Nbthread
       call FTR_ALLOC_2(Romp1_1Nrp1_1Nthetap1(tid)%val,1,Nr+1,1,Ntheta+1)
       
       call FTR_ALLOC_2_ZERO(Comp1_1Nrp1_1Ntheta(tid)%val,1,Nr+1,1,Ntheta)
    enddo
	
    call clck_time(bclock_compJ0)
    Phi_present = present(sPhi_PNrPNthetaNphi)
    J0_present  = present(J0)
    pdPhidr     => Rarray1_NrNthetaPNphi
    pdPhidtheta => Rarray2_NrNthetaPNphi
    pdPhidphi   => Rarray3_NrNthetaPNphi
    tid = 1

#ifdef _OPENMP
!$OMP PARALLEL private(ir,itheta,iphi, &
!$OMP rhsr,drhs_dr,rhstheta,rhsphi,kith, &
!$OMP scoefr,scoeftheta,scoefphi,tid,rhs2d,Acomp,local_id, &
!$OMP startbuf,itag_phi,base,dep,loc_istart,loc_jstart, &
!$OMP loc_iend,loc_jend,j,nbelts) default(shared)
!$OMP BARRIER
    tid = 1+omp_get_thread_num()
#else
    tid = 1
#endif
    rhsr       => Romp1_0Nr(tid)%val
    rhs2d      => Romp1_1Nrp1_1Nthetap1(tid)%val
    Acomp      => Comp1_1Nrp1_1Ntheta(tid)%val
    drhs_dr    => Romp2_0Nr(tid)%val
    rhstheta   => Romp1_0Ntheta(tid)%val
    rhsphi     => Romp1_0Nphi(tid)%val
    scoefr     => Romp_scoefr(tid)%val
    scoeftheta => Romp_scoeftheta(tid)%val
    scoefphi   => Romp_scoefphi(tid)%val
    ! It seems that SdPhidr(:,:geom%Nphi), SdPhidtheta(:,:geom%Nphi)
    ! SdPhidphi(:,:geom%Nphi) are useless values

    nbelts = nbmoments * (dom_r + 2*bufsize) * &
      (dom_theta + 2*bufsize) * dom_locphi

    !*** computation of dPhi/dr ***
!$OMP DO SCHEDULE(static,1)
    do iphi = 0,geom%Nphi-1
      if (moments_locphi(iphi).eq.plocal_id) then
        do itheta = 0,geom%Ntheta
          do ir = 0,geom%Nr
            ! poiss%Phi_NrNthPNphi_Cmu has a parallel 
            !  domain decomposition in phi inside the
            !  mpi communicator mpi_comm_mu
            rhs2d(1+ir,1+itheta) = &
              poiss%Phi_NrNthPNphi_Cmu(ir,itheta,iphi)
          end do
        end do
        if (J0_present) then
          call omp_compute_gyrofunction_2D(J0,geom,imu, &
            Acomp,rhs2d,tid)
        end if
        do itheta = 0,geom%Ntheta
          do ir = 0,geom%Nr
            pdPhidphi(ir,itheta,iphi) = rhs2d(1+ir,1+itheta)
          end do
        end do
        do itheta = 0,geom%Ntheta
          do ir = 0,geom%Nr
            rhsr(ir) = rhs2d(1+ir,1+itheta)
          end do
          !-> computation of dPhi/dr (saved in rhsr)
          call natural_omp_deriv1d(E_nspline1d_r, &
            geom%Nr,geom%dr,Phi_BCr_left,Phi_BCr_right, &
            rhsr,scoefr,drhs_dr)
          do ir = 0, geom%Nr
            pdPhidr(ir,itheta,iphi) = drhs_dr(ir)
          end do
        enddo
        do ir = 0,geom%Nr
          do itheta = 0,geom%Ntheta
            rhstheta(itheta) = rhs2d(1+ir,1+itheta)
          end do
          !-> computation of dPhi/dtheta (saved in rhstheta)
          call period_omp_deriv1d(E_pspline1d_theta, &
            geom%Ntheta,geom%dtheta,rhstheta,scoeftheta)
          do itheta = 0, geom%Ntheta
            pdPhidtheta(ir,itheta,iphi) = rhstheta(itheta)
          end do
        enddo
        
        do local_id = 0, Nbproc_loc-1
          startbuf   = nbelts * local_id
          base       = (local_id/Nbproc_r)
          dep        = mod(local_id,Nbproc_r)              
          loc_istart = dep *  dom_r - bufsize
          loc_iend   = loc_istart + dom_r - 1 + 2*bufsize
          loc_jstart = base * dom_theta - bufsize
          loc_jend   = loc_jstart + dom_theta - 1 + 2*bufsize
          j          = (iphi-ilocstart) * nbmoments * &
            (dom_r+2*bufsize) * (dom_theta+2*bufsize)
          do kith = loc_jstart, loc_jend
            itheta = mod(geom%Ntheta+kith,geom%Ntheta)
            do ir = loc_istart, loc_iend
              if ((ir .ge. 0) .and. (ir .le. geom%Nr)) then
                RSendBufMPI_A(startbuf+j+0) = &
                  pdPhidr(ir,itheta,iphi)
                RSendBufMPI_A(startbuf+j+1) = &
                  pdPhidtheta(ir,itheta,iphi)
                ! At this time pdPhidphi stores J0Phi
                RSendBufMPI_A(startbuf+j+2) = &
                  pdPhidphi(ir,itheta,iphi)
              end if
              j = j + nbmoments
            end do
          end do
        end do
      end if
    end do
!$OMP END DO
!$OMP BARRIER
!$OMP END PARALLEL

    do tid = 1, Nbthread
       call FTR_DEALLOC(Romp1_1Nrp1_1Nthetap1(tid)%val)

       call FTR_DEALLOC(Comp1_1Nrp1_1Ntheta(tid)%val)
    enddo

#ifdef TIMER
    call ppbarrier()
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tcmp(1))
    tdeb = tfin
#endif

    bigtmp_size = 0
    bigtmp_size = MAX(bigtmp_size, nbmoments * (dom_r+2*bufsize) * &
      (dom_theta+2*bufsize) * MAX(Nphi,Nbproc_loc))
    bigtmp_size = MAX(bigtmp_size, nbmoments * (geom%Nr+1) * &
      (geom%Ntheta+1)*knumber)

    call FTR_ALLOC_1_ZERO(RRecvBufMPI_D,0,bigtmp_size)
    nbelts = nbmoments * (dom_r + 2*bufsize) * &
         (dom_theta + 2*bufsize) * dom_locphi
    !---------
    !-> message receive 
    nbreqr  = 0
    iphiloc = 0
    do iphi = 0,geom%Nphi-1, dom_locphi
      itag_phi = 2000 + iphi
      startbuf = nbelts * iphiloc
      local_id = moments_locphi(iphi)
      call MPI_IRECV(RRecvBufMPI_D(startbuf), &
        nbelts,MPI_REAL8,local_id,itag_phi, &
        mpi_comm_mu,bcast_reqr(nbreqr),ierr)
      nbreqr  = nbreqr + 1
      iphiloc = iphiloc + 1
    end do

    !-> message send
    if (ilocstart .ne. -1) then
      nbreqs  = 0
      do procloop = 0, Nbproc_loc-1
        local_id = mod(procloop+plocal_id,Nbproc_loc)
        startbuf = nbelts * local_id
        itag_phi = 2000 + ilocstart
        call MPI_SEND(RSendBufMPI_A(startbuf), &
          nbelts,MPI_REAL8,local_id,itag_phi, &
          mpi_comm_mu,ierr)
        nbreqs = nbreqs + 1
      enddo
    endif

    call MPI_WAITALL(nbreqr, bcast_reqr(0:nbreqr-1), &
            moments_status(:,0:nbreqr-1), ierr)
    call FTR_DEALLOC(RSendBufMPI_A)
#ifdef TIMER
    call ppbarrier()
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tcmm(1))
    tdeb = tfin
#endif

    call FTR_ALLOC_1_ZERO(RSendBufMPI_B,0,dom_r * dom_theta * (Nphi+1))
    !---------
#ifdef _OPENMP
!$OMP PARALLEL private(ir,itheta,iphi,kith,tgtid, &
!$OMP rhsphi,scoefphi,tid,iphiloc,blociphi, &
!$OMP startbuf,j,nbelts) default(shared)
!$OMP BARRIER
    tid = 1+omp_get_thread_num()
#else
    tid = 1
#endif

    nbelts = nbmoments * (dom_r + 2*bufsize) * &
      (dom_theta + 2*bufsize) * dom_locphi

!$OMP DO SCHEDULE(static,1)
    do iphiloc = 0, (geom%Nphi-1)/dom_locphi
      blociphi = iphiloc * dom_locphi
      startbuf = nbelts * iphiloc
      j        = 0
      do iphi = blociphi, blociphi+dom_locphi-1
        do kith = jstart_buf, jend_buf
          itheta = mod(geom%Ntheta+kith,geom%Ntheta)
          do ir = istart_buf, iend_buf
            if ((ir .ge. 0) .and. (ir .le. geom%Nr)) then
              sdPhidr(ir,itheta,iphi)     = &
                RRecvBufMPI_D(startbuf+j+0)
              sdPhidtheta(ir,itheta,iphi) = &
                RRecvBufMPI_D(startbuf+j+1)
              sdPhidphi(ir,itheta,iphi)   = &
                RRecvBufMPI_D(startbuf+j+2)
              if (Phi_present) then
                sPhi_PNrPNthetaNphi(ir,itheta,iphi) = &
                  RRecvBufMPI_D(startbuf+j+2)
              end if
            end if
            j = j + nbmoments
          end do
        end do
      end do
    end do
!$OMP END DO

    scoefphi   => Romp_scoefphi(tid)%val
    rhsphi     => Romp1_0Nphi(tid)%val
    !*** computation of dPhi/dphi ***
    do kith = jstart_buf, jend_buf
      itheta = mod(geom%Ntheta+kith,geom%Ntheta)
      tgtid  = 1 + mod(geom%Ntheta+kith,nbthread)
      ! Parallelization avoiding concurrent accesses in 
      !  writing RRecvBufMPI_D
      if (tgtid .eq. tid) then
        do ir = istart_buf, iend_buf
          if ((ir .ge. 0) .and. (ir .le. geom%Nr)) then
            do iphi = 0,geom%Nphi-1
              rhsphi(iphi) = sdPhidphi(ir,itheta,iphi)
            end do
            rhsphi(geom%Nphi) = rhsphi(0)
            !-> computation of dPhi/dphi (saved in rhsphi)
            call period_omp_deriv1d(E_pspline1d_phi, &
              geom%Nphi,geom%dphi,rhsphi,scoefphi)
            do iphi = 0,geom%Nphi-1 
              RRecvBufMPI_D(iphi+geom%Nphi*((ir-istart_buf) + &
                (dom_r+2*bufsize)*(kith-jstart_buf))) = rhsphi(iphi)
            end do
          end if
        enddo
      end if
    end do
!$OMP BARRIER
    ! Parallelization avoiding false sharing writing Sdphidphi
    do iphi = 0,geom%Nphi-1
      tgtid = 1 + mod(iphi,nbthread)
      if (tgtid .eq. tid) then
        do kith = jstart_buf, jend_buf
          itheta = mod(geom%Ntheta+kith,geom%Ntheta)
          do ir = istart_buf, iend_buf
            if ((ir .ge. 0) .and. (ir .le. geom%Nr)) then
              SdPhidphi(ir,itheta,iphi) = &
                RRecvBufMPI_D(iphi+geom%Nphi*((ir-istart_buf) + &
                (dom_r+2*bufsize)*(kith-jstart_buf)))
            end if
          end do
        end do
      end if
    end do
!$OMP BARRIER

    if (fill_transp_data) then
      nbelts = dom_r * dom_theta * dom_locphi
!$OMP DO SCHEDULE(static,1)
      do iphiloc = 0, (geom%Nphi-1)/dom_locphi
        blociphi = iphiloc*dom_locphi
        startbuf = nbelts * iphiloc
        j = 0
        do iphi = blociphi, blociphi+dom_locphi-1
          do itheta = jstart, jend
            do ir = istart, iend
              RSendBufMPI_B(startbuf+j) = SdPhidphi(ir,itheta,iphi)
              j = j + 1
            end do
          end do
        end do
      end do
!$OMP END DO
    end if
!$OMP BARRIER
!$OMP END PARALLEL


#ifdef TIMER
    call ppbarrier()
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tcmp(2))
    tdeb = tfin
#endif
    if (fill_transp_data) then
      !---------
      nbelts = dom_r * dom_theta * dom_locphi

      !-> message receive
      nbreqr  = 0
      if (ilocstart .ne. -1) then
        do local_id = 0, Nbproc_loc-1
          itag_phi = 3000 + local_id
          startbuf = nbelts * local_id
          call MPI_IRECV(RRecvBufMPI_D(startbuf), &
            nbelts,MPI_REAL8,local_id,itag_phi, &
            mpi_comm_mu,bcast_reqr(nbreqr),ierr)
          nbreqr = nbreqr + 1
        end do
      end if

      !-> message send 
      nbreqs  = 0
      iphiloc = 0
      do blociphi = 0,geom%Nphi-1, dom_locphi
        itag_phi = 3000 + plocal_id
        startbuf = nbelts * iphiloc
        local_id = moments_locphi(blociphi)
        call MPI_SEND(RSendBufMPI_B(startbuf), &
          nbelts,MPI_REAL8,local_id,itag_phi, &
          mpi_comm_mu,ierr)
        nbreqs = nbreqs + 1
        iphiloc = iphiloc + 1
      enddo

      call MPI_WAITALL(nbreqr, bcast_reqr(0:nbreqr-1), &
              moments_status(:,0:nbreqr-1), ierr)

      if (ilocstart .ne. -1) then
#ifdef _OPENMP
!$OMP PARALLEL private(ir,itheta,iphi, &
!$OMP rhsr,drhs_dr,rhstheta,rhsphi, &
!$OMP scoefr,scoeftheta,scoefphi,tid,rhs2d,Acomp,local_id, &
!$OMP startbuf,itag_phi,base,dep,loc_istart,loc_jstart, &
!$OMP loc_iend,loc_jend,j,nbelts) default(shared)
!$OMP BARRIER
        tid = 1+omp_get_thread_num()
#else
        tid = 1
#endif
        nbelts = dom_r * dom_theta * dom_locphi
!$OMP DO SCHEDULE(static,1)
        do local_id = 0, Nbproc_loc-1
          startbuf   = nbelts * local_id
          base       = (local_id/Nbproc_r)
          dep        = mod(local_id,Nbproc_r)              
          loc_istart = dep *  dom_r
          loc_iend   = loc_istart + dom_r - 1
          loc_jstart = base * dom_theta
          loc_jend   = loc_jstart + dom_theta - 1
          j          = 0
          do iphi = ilocstart, ilocstart+dom_locphi-1
            do itheta = loc_jstart, loc_jend
              do ir = loc_istart, loc_iend
                pdPhidphi(ir,itheta,iphi) = RRecvBufMPI_D(startbuf+j)
                j = j + 1
              end do
            end do
          end do
        end do
!$OMP END DO
!$OMP END PARALLEL
      end if

      ! Each proc. owns pdPhidphi, pdPhidr, pdPhidtheta.
      ! Now, we have to redistribute iphi-slices of these data 
      ! structures inside our mpi_comm_mu communicator.
      ! This will give us StranspdPhidr, StranspdPhidtheta, 
      ! StranspdPhidphi.
      
#ifdef OLDCOMM
      !-> message receive
      nbreqr = 0
      nbelts = (Nr+1)*(Ntheta+1)
      do iphi = kstart,kend
        itag_phi = 1000 + 3*iphi
        call MPI_IRECV(StranspdPhidr(0,0,iphi),nbelts,MPI_REAL8, &
          moments_locphi(iphi),itag_phi,mpi_comm_mu, &
          bcast_reqr(nbreqr),ierr)
        nbreqr = nbreqr + 1
        itag_phi = 1000 + 3*iphi+1
        call MPI_IRECV(StranspdPhidtheta(0,0,iphi), &
          nbelts,MPI_REAL8,moments_locphi(iphi), &
          itag_phi,mpi_comm_mu,bcast_reqr(nbreqr),ierr)
        nbreqr = nbreqr + 1
        itag_phi = 1000 + 3*iphi+2
        call MPI_IRECV(StranspdPhidphi(0,0,iphi),nbelts,MPI_REAL8, &
          moments_locphi(iphi),itag_phi,mpi_comm_mu, &
          bcast_reqr(nbreqr),ierr)
        nbreqr = nbreqr + 1
      enddo
      
#ifdef TIMER
      call ppbarrier()
      call clck_time(tfin)
      call clck_ldiff(tdeb,tfin,tcmm(2))
      tdeb = tfin
#endif

      !-> message sending
      nbreqs = 0
      if (ilocstart .ne. -1) then
        nbelts = (Nr+1)*(Ntheta+1)
        do iphi = ilocstart, ilocstart+dom_locphi-1
          do procloop = 0, Nbproc_loc-1
            local_id   = mod(procloop+plocal_id,Nbproc_loc)
            base       = (local_id/Nbproc_phi)
            dep        = mod(local_id,Nbproc_phi)
            loc_kstart = dep *  dom_phi
            loc_kend   = loc_kstart + dom_phi - 1
            if (loc_kstart .le. iphi .and. iphi .le. loc_kend) then
              itag_phi = 1000 + 3*iphi
              call MPI_ISEND(pdPhidr(0,0,iphi),nbelts,MPI_REAL8, &
                local_id,itag_phi,mpi_comm_mu, &
                bcast_reqs(nbreqs),ierr)
              nbreqs   = nbreqs+1
              itag_phi = 1000 + 3*iphi + 1
              call MPI_ISEND(pdPhidtheta(0,0,iphi), &
                nbelts,MPI_REAL8,local_id,itag_phi, &
                mpi_comm_mu,bcast_reqs(nbreqs),ierr)
              nbreqs   = nbreqs+1
              itag_phi = 1000 + 3*iphi + 2
              call MPI_ISEND(pdPhidphi(0,0,iphi),nbelts,MPI_REAL8, &
                local_id,itag_phi,mpi_comm_mu, bcast_reqs(nbreqs), &
                ierr)
              nbreqs = nbreqs+1
            end if
          enddo
        end do
      endif
      if (nbreqr .ne. 0) then
        call MPI_WAITALL(nbreqr,bcast_reqr(0), &
          moments_status(1,0),ierr)
      end if
      if (nbreqs .ne. 0) then
        call MPI_WAITALL(nbreqs,bcast_reqs(0), &
          moments_status(1,0),ierr)
      end if

#else

      call write_param("nbmoments", nbmoments)
      call FTR_ALLOC_1_ZERO(RSendBufMPI_C, 0,(Nr+1) * (Ntheta+1) * nbmoments * dom_locphi)
      nbreqr = 0
      !-> message receive
      nbelts = nbmoments * (Nr+1) * (Ntheta+1)
      do iphi = kstart,kend
         itag_phi = 1000 + iphi
         startbuf = (iphi - kstart) * nbelts
         call MPI_IRECV(RRecvBufMPI_D(startbuf),nbelts,&
              MPI_REAL8,moments_locphi(iphi),itag_phi,mpi_comm_mu, &
              bcast_reqr(nbreqr),ierr)
         nbreqr = nbreqr + 1
      enddo

      if (ilocstart .ne. -1) then
         nbelts = 3*(Nr+1)*(Ntheta+1)
         do iphi = ilocstart, ilocstart+dom_locphi-1
            iphitmp = iphi - ilocstart
            do itheta = 0, geom%Ntheta-1
               do ir = 0, geom%Nr
                  startbuf = iphitmp*nbelts + 3*(ir+(geom%Nr+1)*itheta)
                  RSendBufMPI_C(startbuf+0)= pdPhidr(ir,itheta,iphi)
                  RSendBufMPI_C(startbuf+1)= pdPhidtheta(ir,itheta,iphi)
                  RSendBufMPI_C(startbuf+2)= pdPhidphi(ir,itheta,iphi)
               end do
            end do
         end do
      end if

#ifdef TIMER
      call ppbarrier()
      call clck_time(tfin)
      call clck_ldiff(tdeb,tfin,tcmm(2))
      tdeb = tfin
#endif

      !-> message sending
      nbreqs = 0
      if (ilocstart .ne. -1) then
        nbelts = 3*(Nr+1)*(Ntheta+1)
        do iphi = ilocstart, ilocstart+dom_locphi-1
          iphitmp = iphi - ilocstart
          do procloop = 0, Nbproc_loc-1
            local_id   = mod(procloop+plocal_id,Nbproc_loc)
            base       = (local_id/Nbproc_phi)
            dep        = mod(local_id,Nbproc_phi)
            loc_kstart = dep *  dom_phi
            loc_kend   = loc_kstart + dom_phi - 1
            if (loc_kstart .le. iphi .and. iphi .le. loc_kend) then
              itag_phi = 1000 + iphi
              startbuf = iphitmp*nbelts
              call MPI_ISEND(RSendBufMPI_C(startbuf), &
                nbelts,MPI_REAL8,local_id,itag_phi, &
                mpi_comm_mu,bcast_reqs(nbreqs),ierr)
              nbreqs   = nbreqs+1
            end if
          enddo
        end do
      endif
      if (nbreqr .ne. 0) then
        call MPI_WAITALL(nbreqr, bcast_reqr(0:nbreqr-1), &
                moments_status(:,0:nbreqr-1), ierr)
      end if

      nbelts = 3*(Nr+1)*(Ntheta+1)
      do iphi = kstart, kend
        iphitmp = iphi - kstart
        do itheta = 0, geom%Ntheta-1
          do ir = 0, geom%Nr
            startbuf = iphitmp*nbelts + 3*(ir+(geom%Nr+1)*itheta)
            StranspdPhidr(ir,itheta,iphi)     = &
              RRecvBufMPI_D(startbuf+0) 
            StranspdPhidtheta(ir,itheta,iphi) = &
              RRecvBufMPI_D(startbuf+1) 
            StranspdPhidphi(ir,itheta,iphi)   = &
              RRecvBufMPI_D(startbuf+2) 
          end do
        end do
      end do

      if (nbreqs .ne. 0) then
        call MPI_WAITALL(nbreqs, bcast_reqs(0:nbreqr-1), &
                moments_status(:,0:nbreqr-1), ierr)
      end if
   call FTR_DEALLOC(RSendBufMPI_C)

#endif

#ifdef TIMER
      call ppbarrier()
      call clck_time(tfin)
      call clck_ldiff(tdeb,tfin,tcmm(3))
      tdeb = tfin
#endif
   end if
    call ppbarrier() 

   call FTR_DEALLOC(RSendBufMPI_B)
   call FTR_DEALLOC(RRecvBufMPI_D)

#ifdef TIMER
    tvect(1:6) = (/ tcmm(1:3), tcmp(1:2), tbar(1) /)
    call pp_reduce_vector(tvect,6,tmin,tmax,tavg)
    if (plocal_id .eq. 0) then
      write(6,'(I4,A,3F13.6,A,2F13.6,A,F13.6)') pglobal_id, &
        " TMIN efieldA, tcomm ", tmin(1:3), &
        " tcomp ",tmin(4:5)," tbarrier ",tmin(6)
      write(6,'(I4,A,3F13.6,A,2F13.6,A,F13.6)') pglobal_id, &
        " TMAX efieldA, tcomm ", tmax(1:3), &
        " tcomp ",tmax(4:5)," tbarrier ",tmax(6)
      write(6,'(I4,A,3F13.6,A,2F13.6,A,F13.6)') pglobal_id, &
        " TAVG efieldA, tcomm ", tavg(1:3), &
        " tcomp ",tavg(4:5)," tbarrier ",tavg(6)
    end if
    if (pglobal_id .eq. 0) then
      call clck_ldiff(tth1,tfin,tthd)
      write(6,'(I4,A,F13.6)') pglobal_id, &
        " Temps efieldB,  total ",tthd
    end if
#endif

    call clck_time(eclock_compJ0)
    if (iter_glob .eq. -1) then
      call clck_diff(bclock_compJ0,eclock_compJ0,global_time_init)
    else
      call clck_diff(bclock_compJ0,eclock_compJ0,global_time_compJ0)
    end if
  end subroutine comp_dPhidx_parallel


  !----------------------------------------------------------------
  ! Compute the three derivatives of Phi 
  !  - poiss : (in)  
  !     poiss%Phi_NrNthPNphi_Cworld(0:Nr,0:Ntheta,
  !                                 iphistart:iphistart+dom_mapphi)
  !     poiss%Phi_NrNthPNphi_Cmu(0:Nr,0:Ntheta,
  !                              ilocstart:ilocstart+dom_locphi)
  !  - sPhi_PNrPNthetaNphi : (out) Phi(istart:iend,jstart:jend,0:Nphi)
  !  - SdPhidr     : (out) dPhi/dr(istart:iend,jstart:jend,0:Nphi)
  !  - SdPhidtheta : (out) dPhi/dtheta(istart:iend,jstart:jend,0:Nphi)
  !  - SdPhidphi   : (out) dPhi/dphi(istart:iend,jstart:jend,0:Nphi)
  !  - StranspdPhidr     : (out) dPhi/dr(0:Nr,0:Ntheta,kstart:kend)
  !  - StranspdPhidtheta: (out) dPhi/dtheta(0:Nr,0:Ntheta,kstart:kend)
  !  - StranspdPhidphi  : (out) dPhi/dphi(0:Nr,0:Ntheta,kstart:kend)
  !----------------------------------------------------------------
  subroutine comp_dPhidx(geom,poiss,sPhi_PNrPNthetaNphi, &
    SdPhidr,SdPhidtheta,SdPhidphi, &
    StranspdPhidr,StranspdPhidtheta,StranspdPhidphi)
    use mpiMod
    type(geometry)               , intent(in) :: geom
    type(poisson)                , intent(in) :: poiss
    real(RKIND), dimension(:,:,:), pointer    :: sPhi_PNrPNthetaNphi 
    real(RKIND), dimension(:,:,:), pointer    :: SdPhidr             
    real(RKIND), dimension(:,:,:), pointer    :: SdPhidtheta        
    real(RKIND), dimension(:,:,:), pointer    :: SdPhidphi           
    real(RKIND), dimension(:,:,:), pointer    :: StranspdPhidr      
    real(RKIND), dimension(:,:,:), pointer    :: StranspdPhidtheta  
    real(RKIND), dimension(:,:,:), pointer    :: StranspdPhidphi 

    if (.not.distrib3D) then
      call comp_dPhidr_loc(geom,poiss%Phi_NrNthPNphi_Cworld, &
        istart,iend,jstart,jend,0,geom%Nphi,SdPhidr)
      call comp_dPhidtheta_loc(geom,poiss%Phi_NrNthPNphi_Cworld, &
        istart,iend,jstart,jend,0,geom%Nphi,SdPhidtheta)
      call comp_dPhidphi_loc(geom,poiss%Phi_NrNthPNphi_Cworld, &
        istart,iend,jstart,jend,0,geom%Nphi,SdPhidphi)
      call comp_dPhidtheta_loc(geom,poiss%Phi_NrNthPNphi_Cworld, &
        0,geom%Nr,0,geom%Ntheta,kstart,kend,StranspdPhidtheta)
      call comp_dPhidphi_loc(geom,poiss%Phi_NrNthPNphi_Cworld, &
        0,geom%Nr,0,geom%Ntheta,kstart,kend,StranspdPhidphi)
    else
      call comp_dPhidx_parallel(geom,poiss, &
        SdPhidr,SdPhidtheta,SdPhidphi, &
        StranspdPhidr,StranspdPhidtheta,StranspdPhidphi, &
        sPhi_PNrPNthetaNphi=sPhi_PNrPNthetaNphi, &
        fill_transp_data=.true.)
    end if
  end subroutine comp_dPhidx


  !-------------------------------------------------------------
  ! Compute the gyroaverage of Phi and its three derivatives 
  !-------------------------------------------------------------
  subroutine comp_dJ0Phidx(J0,geom,imu,poiss, &
    SdJ0Phidr,SdJ0Phidtheta, SdJ0Phidphi, &
    StranspdJ0Phidr,StranspdJ0Phidtheta, StranspdJ0Phidphi)
    use velocities_edges_module , only : comp_Phi_edges 
    use gyroaverage_class
    use globals, only : transpose4D
    ! only used if distrib3D = .false.
    use globals, only : Rarray1_NrNthetaNphi 
    use mpiMod
    type(J0operator)             , intent(inout) :: J0
    type(geometry)               , intent(in)    :: geom
    integer                      , intent(in)    :: imu
    type(poisson)                , intent(in)    :: poiss
    !--> intent(out) variables
    real(RKIND), dimension(:,:,:), pointer    :: SdJ0Phidr
    real(RKIND), dimension(:,:,:), pointer    :: SdJ0Phidtheta 
    real(RKIND), dimension(:,:,:), pointer    :: SdJ0Phidphi
    real(RKIND), dimension(:,:,:), pointer    :: StranspdJ0Phidr
    real(RKIND), dimension(:,:,:), pointer    :: StranspdJ0Phidtheta
    real(RKIND), dimension(:,:,:), pointer    :: StranspdJ0Phidphi

    !*** Rk : variable only used in the case '.not.distrib3D' ***
    real(RKIND), dimension(:,:,:), pointer :: gyroPhi

    integer :: tid

    call Write_dyn_balise_open('compute_dJ0Phidx') !R4

    do tid = 1, Nbthread
       call FTR_ALLOC_2(J0%omp_diag(tid)%val,1,J0%N1,1,J0%N2)
       call FTR_ALLOC_2(J0%omp_ldiag(tid)%val,1,J0%N1-1,1,J0%N2)
       call FTR_ALLOC_2(J0%omp_udiag(tid)%val,1,J0%N1-1,1,J0%N2)
    enddo
    
    if (distrib3D) then
      call comp_dPhidx_parallel(geom,poiss, &
        SdJ0Phidr,SdJ0Phidtheta,SdJ0Phidphi, &
        StranspdJ0Phidr,StranspdJ0Phidtheta,StranspdJ0Phidphi, &
        J0,imu,fill_transp_data=transpose4D)
    else
      call FTR_ALLOC_3(Rarray1_NrNthetaNphi,0,geom%Nr,0,geom%Ntheta,0,geom%Nphi)
      gyroPhi => Rarray1_NrNthetaNphi
      call compute_gyrofunction_3D(J0,geom,imu, &
        poiss%Phi_NrNthPNphi_Cworld,gyroPhi)
      call comp_dPhidx_sequential(geom,gyroPhi,SdJ0Phidr, &
        SdJ0Phidtheta,SdJ0Phidphi)
      if (conservative_scheme) then
         call comp_Phi_edges(geom,gyroPhi)
      end if
      call FTR_DEALLOC(Rarray1_NrNthetaNphi)
    end if

    do tid = 1, Nbthread
       call FTR_DEALLOC(J0%omp_diag(tid)%val)
       call FTR_DEALLOC(J0%omp_ldiag(tid)%val)
       call FTR_DEALLOC(J0%omp_udiag(tid)%val)
    enddo

    call Write_dyn_balise_close('compute_dJ0Phidx') !R4
  end subroutine comp_dJ0Phidx
end module efield_module
