!**************************************************************
!  Copyright Euratom-CEA
!  Authors : 
!     Virginie Grandgirard (virginie.grandgirard@cea.fr)
!     Chantal Passeron (chantal.passeron@cea.fr)
!     Guillaume Latu (guillaume.latu@cea.fr)
!     Xavier Garbet (xavier.garbet@cea.fr)
!     Philippe Ghendrih (philippe.ghendrih@cea.fr)
!     Yanick Sarazin (yanick.sarazin@cea.fr)
!  
!  This code GYSELA (for GYrokinetic SEmi-LAgrangian) 
!  is a 5D gyrokinetic global full-f code for simulating 
!  the plasma turbulence in a tokamak.
!  
!  This software is governed by the CeCILL-B license 
!  under French law and abiding by the rules of distribution 
!  of free software.  You can  use, modify and redistribute 
!  the software under the terms of the CeCILL-B license as 
!  circulated by CEA, CNRS and INRIA at the following URL
!  "http://www.cecill.info". 
!**************************************************************

#include "config-F.inc"

!------------------------------------------------------------
! file : poisson.F90
! date : 08/04/2001
! . used to solve the Poisson equation
!    -[d^2(Phi)/dr^2+(1/r)*dPhi/dr+(1/r^2)*d^2(Phi)/dth^2] 
!        = rho(phi,r,th)
!   where Phi is considered with periodic conditions in
!   theta direction and Dirichlet conditions in r direction
!--------------------------------------------------------------
module poisson_class
  use prec_const
  use init_profile_class
  use init_magnetic_class
  use fdistribu5d_class
  use fft_module
  use wrap_alloc_module
  use fft_NRF90_module
  use integration_module
  use clock_module
  use wrap_alloc_module
  
  implicit none

  private
  public :: new_poisson, del_poisson, solve_poisson, &
    compute_nbparticles

  type, public :: poisson
     real(RKIND)   , dimension(:,:,:), pointer :: Phi_NrNthPNphi_Cworld
     !-> Rk : 'Phi_NrNthPNphi_Cmu' only used if the 3D data are distributed
     real(RKIND)   , dimension(:,:,:), pointer :: Phi_NrNthPNphi_Cmu       
     real(RKIND)   , dimension(:)    , pointer :: kth, kphi
     !-> vector used for the LU factorisation 
     !->  of the system for (m,n) = (0,0)
     ! ---> diagonal terms of matrix
     complex(CKIND), dimension(:), pointer :: L_diag    
     complex(CKIND), dimension(:), pointer :: L_invZ02Te_diag 
     ! ---> lower diagonal terms of matrix
     complex(CKIND), dimension(:), pointer :: L_ldiag
     ! ---> upper diagonal terms of matrix
     complex(CKIND), dimension(:), pointer :: L_udiag   
     ! -> vector used for the LU factorisation 
     ! ->  of the system for (m,n) != (0,0)
     ! ---> diagonal terms of matrix
     complex(CKIND), dimension(:,:), pointer :: diag
     ! ---> lower diagonal terms of matrix
     complex(CKIND), dimension(:,:), pointer :: ldiag
     ! ---> upper diagonal terms of matrix        
     complex(CKIND), dimension(:,:), pointer :: udiag
  end type poisson

  type(omp1cvector), dimension (:), pointer :: omp_rhs
  type(omp2cvector), dimension (:), pointer :: omp_data_i
  type(omp2cvector), dimension (:), pointer :: omp_tmp

  real(RKIND), dimension(:), pointer, public :: Phi00_diag
  real(RKIND), dimension(:), pointer, public :: Phi_FSavg_diag

  include "gyrokinetic/Bstar_inline.h"
  !******************************
  contains
  !******************************
#include "gyrokinetic/Bstar_inline.F90"

  !----------------------------------------------------- 
  ! constructor for the poisson type
  !-----------------------------------------------------   
  subroutine new_poisson(pthis,geom,init_prof)
    use globals, only : memory_test
    type(poisson)     , intent(out) :: pthis
    type(geometry)    , intent(in)  :: geom 
    type(init_profile), intent(in)  :: init_prof

    !*** local variables ***
    integer     :: i, j, k           ! loop index
    integer     :: N2m1, N3m1        ! dimension in theta and phi
    integer     :: min_m, max_m, min_n, max_n
    integer     :: tid

    !*** set local variables ***
    N2m1  = geom%Ntheta           !size(geom%thetag)-1
    N3m1  = geom%Nphi             !size(geom%phig)-1

    !*** memory allocation ***
    call array_allocation(pthis,geom)

    if (.not.memory_test) then
      !*** compute kth for the FFT in theta direction ***
      call NR_fft_vector(N2m1,geom%dtheta,pthis%kth)
      
      !*** compute kphi for the FFT in phi direction ***
      call NR_fft_vector(N3m1,geom%dphi,pthis%kphi)

      !*** assemble and factorize poisson matrix for all k ***
      call LU_factorisation(pthis,init_prof,geom)
    end if
  end subroutine new_poisson


  !----------------------------------------------------- 
  ! Memory allocation (used in the constructor)
  !-----------------------------------------------------   
  subroutine array_allocation(pthis,geom)
    use globals, only : Nbthread, plocal_id, cible, &
      distrib3D
    type(poisson) , intent(inout) :: pthis
    type(geometry), intent(in)    :: geom  

    integer :: N1, N2, N3, N2m1, N3m1
    integer :: tid

    N1   = geom%Nr+1       !size(geom%rg)
    N2   = geom%Ntheta+1   !size(geom%thetag)
    N2m1 = N2-1
    N3   = geom%Nphi+1     !size(geom%phig)
    N3m1 = N3-1
	
    call write_expr("N1"  , "geom%Nr+1")
    call write_expr("N2"  , "geom%Ntheta+1")
    call write_expr("N2m1", "N2-1")
    call write_expr("N3"  , "geom%Nphi+1")
    call write_expr("N3m1", "N3-1")

    call FTR_ALLOC_1_ZERO(pthis%kth,1,N2m1)
    call FTR_ALLOC_1_ZERO(pthis%kphi,1,N3m1)

    if (distrib3D) then
      ! -> first array for the distributed solution : Phi 
      if (iphistart .ne. -1) then
        call FTR_ALLOC_3_ZERO(pthis%Phi_NrNthPNphi_Cworld,0,geom%Nr,0,geom%Ntheta,iphistart,iphistart+dom_mapphi-1)
      end if
      ! -> second array for the distributed solution : Phi 
      if (ilocstart .ne. -1) then
        call FTR_ALLOC_3_ZERO(pthis%Phi_NrNthPNphi_Cmu,0,geom%Nr,0,geom%Ntheta,ilocstart,ilocstart+dom_locphi-1)
      end if
    else
      call FTR_ALLOC_3_ZERO(pthis%Phi_NrNthPNphi_Cworld,0,geom%Nr,0,geom%Ntheta,0,geom%Nphi)
      pthis%Phi_NrNthPNphi_Cmu => NULL()
    end if
    ! -> arrays for LU factorisation and 
    ! ->   solving with (m,n) = (0,0) 
    call FTR_ALLOC_1_ZERO(pthis%L_diag,1,N1-2)
    call FTR_ALLOC_1_ZERO(pthis%L_invZ02Te_diag,1,N1-2)
    call FTR_ALLOC_1_ZERO(pthis%L_ldiag,1,N1-3)
    call FTR_ALLOC_1_ZERO(pthis%L_udiag,1,N1-3)

    ! -> arrays for LU factorisation and solving
    !    with (m,n) != (0,0)
    call FTR_ALLOC_2_ZERO(pthis%diag,1,N2m1,1,N1-2)
    call FTR_ALLOC_2_ZERO(pthis%ldiag,1,N2m1,1,N1-3)
    call FTR_ALLOC_2_ZERO(pthis%udiag,1,N2m1,1,N1-3)

    !*** allocation of arrays for OMP parallelisation ***
    allocate(omp_tmp(1:Nbthread))
    allocate(omp_data_i(1:Nbthread))
    allocate(omp_rhs(1:Nbthread))

    do tid = 1,Nbthread
       call FTR_ALLOC_1_ZERO(omp_rhs(tid)%val,1,N1)
    enddo

    !*** allocation of arrays for output saving ***
    !-> for radial profile saving of the Phi (0,0) Fourier mode
    call FTR_ALLOC_1_ZERO(Phi00_diag,0,geom%Nr)
    !-> for radial profile saving of the flux surface 
    !->  average of Phi
    call FTR_ALLOC_1_ZERO(Phi_FSavg_diag,0,geom%Nr)
  end subroutine array_allocation


  !----------------------------------------------------- 
  ! LU factorisation (used in the constructor)
  !-----------------------------------------------------   
  subroutine LU_factorisation(pthis,init_prof,geom)
    use globals, only : Z0, QN_coef_polarization, &
      QN_coef_Phi00, Phi00_BCrmin_Neumann, &
      Nspecies, hat_As, &
      Rarray1_Nr, Rarray2_Nr
    type(poisson)     , intent(inout) :: pthis
    type(init_profile), intent(in)    :: init_prof
    type(geometry)    , intent(in)    :: geom

    integer     :: ispec, i, j
    integer     :: N1, N2, N2m1
    real(RKIND) :: lambda, inv_Z02Te, inv_dr2
    real(RKIND) :: ri, rip1, coef_ri, coef_rip1
    real(RKIND) :: coef_r1, lr1, coef1_lr1, coef1_lr1_00
    real(RKIND) :: rN1m2, coef_rN1m2, crN1m2, coef1_crN1m2_00
    real(RKIND) :: ldiag_tmp, ddiag_tmp, udiag_tmp
    real(RKIND) :: hatAs, ns0_on_ne0
    real(RKIND) :: Z02
    real(RKIND), dimension(:), pointer :: alpha_ri, beta_ri

    !*** set local variables ***
    N1   = size(geom%rg)
    N2   = size(geom%thetag)
    N2m1 = N2-1

    alpha_ri => Rarray1_Nr
    beta_ri  => Rarray2_Nr
    !*** Choice between Dirichlet or Neumann conditions ***
    !***  for the mode (0,0)                            ***
    if (Phi00_BCrmin_Neumann) then
      coef1_lr1       = 0._RKIND      
      coef1_lr1_00    = 1._RKIND   ! Neumann   at r=rmin
      coef1_crN1m2_00 = 0._RKIND   ! Dirichlet at r=rmax
    else  
      !-> Neumann conditions are imposed on the axis 
      !    in the case rmin=0, otherwise Dirichlet    
      !    conditions are imposed                     
      if (geom%rg(0).ne.0_RKIND) then
        coef1_lr1 = 0._RKIND
      else
        coef1_lr1 = 1._RKIND
      end if
      coef1_lr1_00    = 0._RKIND
      coef1_crN1m2_00 = 0._RKIND
    end if

    !*** Initialisation of lambda ***
    !***  lambda = 0 if QN_coef_Phi00 = 0 ***
    !***  lambda = 1 if QN_coef_Phi00 = 1  ***
    if (QN_coef_Phi00.eq.1) then
      lambda = 1._RKIND
    else
      lambda = 0._RKIND
    end if

    !*** assemble the system matrices ***
    !-> Compute the coefficients
    !->   alpha(r) = sum_s hat_As*(ns0(r)/ne0(r) *  
    !->              (1/r+1/ns0(r)*dns0(r)/dr)         
    !->   beta(r)  = sum_s hat_As*(ns0(r)/ne0(r)
    do i = 1,N1-2
      ri          = geom%rg(i)
      alpha_ri(i) = 0._RKIND
      beta_ri(i)  = 0._RKIND
      do ispec = 0,Nspecies-1
        hatAs        = hat_As(ispec)
        ns0_on_ne0   = init_prof%ns0(ispec,i)/init_prof%ne0(i) 
        alpha_ri(i)  = alpha_ri(i) + hatAs*ns0_on_ne0 * &
          (1._RKIND/ri + init_prof%dns0dr(ispec,i)/init_prof%ns0(ispec,i))
        beta_ri(i)   = beta_ri(i) + hatAs*ns0_on_ne0
      end do
    end do
    !-> initialisation of the lower and upper diagonal 
    !->  of the matrix
    Z02     = Z0*Z0
    inv_dr2 = 1._RKIND/(geom%dr*geom%dr)
    do i = 1,N1-3
      ri         = geom%rg(i)
      rip1       = geom%rg(i+1)
      coef_ri    = alpha_ri(i)/(2._RKIND*geom%dr)
      coef_rip1  = alpha_ri(i+1)/(2._RKIND*geom%dr)
      udiag_tmp  = -1._RKIND*QN_coef_polarization * &
        (beta_ri(i)*inv_dr2 + coef_ri)
      ldiag_tmp  = -1._RKIND*QN_coef_polarization * &
        (beta_ri(i+1)*inv_dr2 - coef_rip1)
      ! -> for (m,n) = (0,0)
      pthis%L_udiag(i) = udiag_tmp
      pthis%L_ldiag(i) = ldiag_tmp
      do j = 1,N2m1
        ! -> forall (m,n) != (0,0)
        pthis%udiag(j,i) = udiag_tmp
        pthis%ldiag(j,i) = ldiag_tmp
      end do
    end do
!baoter ATTENTION A POURSUIVRE
    !-> initialisation of the diagonal of the matrix
    do i = 1,N1-3
      ri        = geom%rg(i)
      inv_Z02Te = 1._RKIND/(Z02*init_prof%Te0(i)) 
      ddiag_tmp = QN_coef_polarization * &
        2._RKIND*inv_dr2*beta_ri(i)
      ! -> for (m,n) = (0,0)
      pthis%L_diag(i)          = ddiag_tmp 
      pthis%L_invZ02Te_diag(i) = ddiag_tmp + &
        inv_Z02Te
      do j = 1,N2m1
        ! -> for (m,n) != (0,0)
        pthis%diag(j,i) = ddiag_tmp + &
          QN_coef_polarization*beta_ri(i) * & 
          pthis%kth(j)*pthis%kth(j)/(ri*ri) + &
          inv_Z02Te
      end do
    end do

    !*** boundary conditions at r = rmin ***
    !***  (first matrix row -> i = 1)    ***
    coef_r1  = alpha_ri(1)/(2._RKIND*geom%dr)
    lr1      = -1._RKIND*(beta_ri(1)*inv_dr2 - coef_r1)
    !-> for (m,n) != (0,0)
    do j = 1,N2m1
      pthis%diag(j,1) = QN_coef_polarization * &
        (coef1_lr1*lr1) + pthis%diag(j,1)
    end do
    !-> for (m,n) = (0,0)
    pthis%L_diag(1)          = QN_coef_polarization * &
      (coef1_lr1_00*lr1) + pthis%L_diag(1)
    pthis%L_invZ02Te_diag(1) = QN_coef_polarization * &
      (coef1_lr1_00*lr1) + pthis%L_invZ02Te_diag(1)

    !*** boundary conditions at r = rmax       ***
    !***  (last matrix row -> i = N1-2 = Nr-1) ***
    !***  case i = N1-2                        ***
    rN1m2       = geom%rg(N1-2)
    inv_Z02Te   = 1._RKIND/(Z02*init_prof%Te0(N1-2))
    coef_rN1m2  = alpha_ri(N1-2)/(2._RKIND*geom%dr)
    crN1m2      = -1._RKIND*(beta_ri(N1-2)*inv_dr2 + coef_rN1m2)
    ddiag_tmp   = QN_coef_polarization*beta_ri(N1-2) * &
      2._RKIND * inv_dr2
    do j = 1,N2m1
      ! -> for (m,n) != (0,0)
      pthis%diag(j,N1-2) = ddiag_tmp + &
        QN_coef_polarization*beta_ri(N1-2) * &
        pthis%kth(j)*pthis%kth(j)/(rN1m2*rN1m2) + &
        inv_Z02Te
    end do
    ! -> for (m,n) = (0,0)
    pthis%L_diag(N1-2)          = QN_coef_polarization * &
      coef1_crN1m2_00*crN1m2 + ddiag_tmp 
    pthis%L_invZ02Te_diag(N1-2) = QN_coef_polarization * &
      coef1_crN1m2_00*crN1m2 + ddiag_tmp + inv_Z02Te
  end subroutine LU_factorisation


  !----------------------------------------------------- 
  ! destructor for the poisson type
  !-----------------------------------------------------   
  subroutine del_poisson(pthis)
    use globals, only : Nbthread, plocal_id, cible, &
      distrib3D
    type(poisson), intent(inout) :: pthis

    integer :: tid

    call FTR_DEALLOC(pthis%kth)
    call FTR_DEALLOC(pthis%kphi)
    if (distrib3D) then
       if (iphistart .ne. -1) then
          call FTR_DEALLOC(pthis%Phi_NrNthPNphi_Cworld)
       end if
       if (ilocstart .ne. -1) then
          call FTR_DEALLOC(pthis%Phi_NrNthPNphi_Cmu)
       end if
    else
       call FTR_DEALLOC(pthis%Phi_NrNthPNphi_Cworld)
    end if

    !*** Matrix for (m,n) = (0,0) ***
    call FTR_DEALLOC(pthis%L_diag)
    call FTR_DEALLOC(pthis%L_invZ02Te_diag)
    call FTR_DEALLOC(pthis%L_ldiag)
    call FTR_DEALLOC(pthis%L_udiag)

    !*** Matrix for (m,n) != (0,0) ***
    call FTR_DEALLOC(pthis%diag)
    call FTR_DEALLOC(pthis%ldiag)
    call FTR_DEALLOC(pthis%udiag)

    do tid = 1,Nbthread
      call FTR_DEALLOC(omp_rhs(tid)%val)
    enddo

    !*** deallocation of arrays for OMP parallelisation ***
    deallocate(omp_tmp)
    deallocate(omp_data_i)
    deallocate(omp_rhs)

    !*** deallocation of arrays for output saving ***
    call FTR_DEALLOC(Phi00_diag)
    call FTR_DEALLOC(Phi_FSavg_diag)
  end subroutine del_poisson


  !*****************************************************
  ! PARALLEL VERSION
  !*****************************************************
  !-----------------------------------
  ! Thomas algorithm for LU inversion
  !-----------------------------------
  subroutine thomas_c(N,ud,ld,diag,rhs,err) 
    integer                     , intent(in)  :: N
    real(RKIND)   , dimension(:), pointer     :: ud
    real(RKIND)   , dimension(:), pointer     :: ld
    real(RKIND)   , dimension(:), pointer     :: diag
    complex(CKIND), dimension(:), pointer     :: rhs
    integer                     , intent(out) :: err

    integer :: nm1, i, i1

    nm1 = N-1
    err = 0
    do i = 1,nm1
       i1 = i+1
       if (diag(i) .eq. 0._RKIND) then
          err = 1
          return
       endif
       ld(i)    = ld(i)/diag(i)
       diag(i1) = diag(i1) - ld(i)*ud(i)
       rhs(i1)  = rhs(i1) - ld(i)*rhs(i)
    end do

    rhs(N) = rhs(N)/diag(N)

    do i = nm1,1,-1
       rhs(i) = (rhs(i)-ud(i)*rhs(i+1))/diag(i)
    end do
  end subroutine thomas_c


  !----------------------------------------------
  ! Thomas algorithm for LU inversion
  !  (the same than before but using a temporary
  !  array 'tmpv' to keep the initial values of
  !  'ld' and 'diag')
  !  => case where 'rhs' is real 
  !----------------------------------------------
  subroutine lusolver(N,ud,ld,diag,rhs,tmpv,err) 
    integer                  , intent(in)  :: N
    real(RKIND), dimension(:), pointer     :: ud
    real(RKIND), dimension(:), pointer     :: ld
    real(RKIND), dimension(:), pointer     :: diag
    real(RKIND), dimension(:), pointer     :: rhs
    real(RKIND), dimension(:), pointer     :: tmpv
    integer                  , intent(out) :: err

    integer     :: nm1, i
    real(RKIND) :: denom_tmp
    
    nm1     = N-1
    err     = 0
    tmpv(1) = 1._RKIND/diag(1)
    rhs(1)  = rhs(1)*tmpv(1)
    do i = 1,nm1
      denom_tmp = (diag(i+1) - ld(i)*ud(i)*tmpv(i))
      if (abs(denom_tmp).le.1.e-12) then
        print*,'Problem denom_tmp = ',denom_tmp
        err = 1
        return
      end if
      tmpv(i+1) = 1._RKIND/denom_tmp
      rhs(i+1) = (rhs(i+1) - ld(i)*rhs(i))*tmpv(i+1)
    end do

    do i = nm1,1,-1
       rhs(i)= (rhs(i)-ud(i)*tmpv(i)*rhs(i+1))
    end do
  end subroutine lusolver


  !----------------------------------------------
  ! Thomas algorithm for LU inversion
  !  (the same than before but using a temporary
  !  array 'tmpv' to keep the initial values of
  !  'ld' and 'diag')
  !  => case where 'rhs' is complex 
  !----------------------------------------------
  subroutine lusolvec(N,ud,ld,diag,rhs,tmpv,err) 
    integer                     , intent(in)  :: N
    real(RKIND),    dimension(:), pointer     :: ud
    real(RKIND),    dimension(:), pointer     :: ld
    real(RKIND),    dimension(:), pointer     :: diag
    complex(RKIND), dimension(:), pointer     :: rhs
    real(RKIND),    dimension(:), pointer     :: tmpv
    integer                     , intent(out) :: err

    integer :: nm1, i
    
    nm1     = N-1
    err     = 0
    tmpv(1) = 1._RKIND/diag(1)
    rhs(1)  = rhs(1)*tmpv(1)
    do i = 1,nm1
       tmpv(i+1) = 1._RKIND/(diag(i+1) - ld(i)*ud(i)*tmpv(i))
       if (tmpv(i+1) .eq. 0._RKIND) then
          err = 1
          return
       endif
       rhs(i+1) = (rhs(i+1) - ld(i)*rhs(i))*tmpv(i+1)
    end do
    do i = nm1,1,-1
       rhs(i)= (rhs(i)-ud(i)*tmpv(i)*rhs(i+1))
    end do
  end subroutine lusolvec


  !---------------------------------------------------------------
  ! Compute rho the RHS of the quasi-neutrality 
  !  equation, i.e:
  !    rhos(r,theta,phi) = (1/ne0(r))*hat_Zs * 
  !                        (nGs(r,theta,phi)-nGs_eq(r,theta))
  !  where nGs = \int J0(sqrt{2*mu}).f_s Jv  dvpar dmu
  !            = \int J0(sqrt{2*mu}).[\int f_s Jv dvpar] dmu 
  !  with  nGs_eq(r,theta) = \int J0.feq_s(r,theta,vpar,mu) 
  !                          Jv(r,theta,vpar)dmu dvpar
  !  Jv being the Jacobian in velocity space :
  !   . Jv(r,theta,vpar) = 2*pi*Bstar in the 5D case
  !                      = 1          in the 4D case
  !  f_s being the distribution function of specie s.
  !
  ! Rk : rhos is computed as the sum of 2 integrals:
  !   Is =  (1/ne0(r))*hat_Zs* \int dmu
  !            (I0*B_norm + (sqrt{hat_As}/hat_Zs)* 
  !            I1*mu0*vec_J.vec_b/B_norm)
  !   where I0 = J0.(\int f_s dvpar) and          
  !         I1 = J0.(\int f_s*vpar dvpar)
  !   this by using the expresion of Bstar
  !    Bstar = B_norm + (sqrt{hat_As}/hat_Zs) * 
  !            vpar*mu0*vec_J.vec_b/B_norm
  !
  ! Rk: BECAREFUL -> the case of several species is not 
  !     treated here
  !------------------------------------------------------------
  subroutine compute_rho_FFT2Dhybrid(J0,f,Sfmu_eq, geom, &
    init_prof,init_magnet,init_curr,rho)
    use globals, only : Nbproc_r, Nbproc_loc, Nbproc_mu, &
      dom_r, dom_theta, Nr, Ntheta, Nphi, Nmu, mumin, &
      istart, iend, jstart, jend, mu_id, mpi_comm_intermu, &
      Rarray1_NrNtheta, Rarray1_NrNthetaDomphi, &
      Rarray_PNrPNthetaNphi_nbM2, &
      integration_CS, Nbthread
    use globals, only : hat_As, hat_Zs, &
      mpi_comm_species, pInSpecies_id, Nbproc_species, pSpecies_id, &
      RRecvBufMPI_A
#ifdef _OPENMP
    use OMPutils_module, only : omp_get_thread_num
#endif
    use OMPutils_module, only : Romp_scoefvpar, &
      Romp1_0Nvpar, Romp2_0Nvpar, &
      Romp1_1Nrp1_1Nthetap1, Romp2_1Nrp1_1Nthetap1, &
      Comp1_1Nrp1_1Ntheta
    use gyroaverage_class
    use mpiMod
    type(J0operator)             , intent(inout) :: J0
    type(fdistribu5d)            , intent(in)    :: f
    real(RKIND), dimension(:,:,:), pointer       :: Sfmu_eq
    type(geometry)               , intent(in)    :: geom
    type(init_profile)           , intent(in)    :: init_prof
    type(init_magnetic)          , intent(in)    :: init_magnet
    type(init_current)           , intent(in)    :: init_curr
    real(RKIND), dimension(:,:,:), pointer       :: rho

#ifdef HAS_QN_V1
    integer, parameter :: nbmoments = 2
    real(RKIND)   , dimension(:)    , pointer :: fmu_1D
    real(RKIND)   , dimension(:)    , pointer :: fmu_v_1D
    real(RKIND)   , dimension(:)    , pointer :: scoefvpar
    real(RKIND)   , dimension(:,:)  , pointer :: J0_intdv_f
    real(RKIND)   , dimension(:,:)  , pointer :: J0_intvdv_f
    real(RKIND)   , dimension(:,:,:), pointer :: M0_loc
    complex(CKIND), dimension(:)    , pointer :: crhs
    complex(CKIND), dimension(:,:)  , pointer :: Acomp

    integer     :: ierr
    integer     :: ir, itheta, iphi, ivpar, imu, iphiloc
    logical     :: case4D
    real(RKIND) :: fmu_tmp
    real(RKIND) :: Bnorm_ij, scalprod_mu0Jb_ij
    real(RKIND) :: coeff_intdmu
    real(RKIND) :: hatZs, sqrtAs_on_Zs

    ! -> variable for parallelization
    integer :: iproc, nbreqr, nbbcast_r, nbbcast_s
    integer :: local_id, dep, base, global_id, startbuf
    integer :: l_istart, l_iend, l_jstart, l_jend
    integer :: itag_phi, nbelements, tid

#ifdef TIMER
    integer(TIMEPREC) :: tdeb, tfin, tth1, tth2
    real(RKIND)       :: tbar(12), tcmp(12), tcom(12), tthd
    call clck_time(tdeb)
    tcmp(1:4) = 0.
    tcom(1:4) = 0.
#endif

    do tid = 1, Nbthread
       call FTR_ALLOC_2(Romp1_1Nrp1_1Nthetap1(tid)%val,1,Nr+1,1,Ntheta+1)
       call FTR_ALLOC_2(Romp2_1Nrp1_1Nthetap1(tid)%val,1,Nr+1,1,Ntheta+1)

       call FTR_ALLOC_2_ZERO(Comp1_1Nrp1_1Ntheta(tid)%val,1,Nr+1,1,Ntheta)
    enddo

   call FTR_ALLOC_1_ZERO(RRecvBufMPI_A,0,nbmoments*dom_r*dom_theta*Nbproc_loc*(Nmu+1)*dom_mapphi)

    case4D = .false.
    if (geom%Nmu.eq.0) case4D = .true.

    !--> compute sqrt{hat_As}/hat_Zs
    hatZs        = hat_Zs(pSpecies_id)
    sqrtAs_on_Zs = sqrt(hat_As(pSpecies_id))/hatZs

    M0_loc => Rarray1_NrNthetaDomphi

    !*** initialization ***
    do iphi = iphistart,iphistart+dom_mapphi-1
      do itheta = 0,geom%Ntheta
        do ir = 0,geom%Nr
          M0_loc(ir,itheta,iphi) = 0._RKIND
        end do
      end do
    end do

    !*** computation ***
!$OMP PARALLEL private(tid,iphi,itheta,ir,ivpar,&
!$OMP scoefvpar,fmu_tmp,fmu_1D,fmu_v_1D)  default(shared)
#ifdef _OPENMP
    tid = 1+omp_get_thread_num()
#else
    tid = 1
#endif
    scoefvpar => Romp_scoefvpar(tid)%val
    fmu_1D    => Romp1_0Nvpar(tid)%val
    fmu_v_1D  => Romp2_0Nvpar(tid)%val
!$OMP DO SCHEDULE(STATIC)
    do iphi = 0,f%n3-1
      do itheta = f%jstart,f%jend
        do ir = f%istart,f%iend
          do ivpar = 0,f%n4
            fmu_tmp         = f%values(ir,itheta,iphi,ivpar) - &
              Sfmu_eq(ir,itheta,ivpar)
            fmu_1D(ivpar)   = fmu_tmp
            fmu_v_1D(ivpar) = fmu_tmp*geom%vparg(ivpar)
          end do
          !*** computation of the integrals in vpar direction ***
          if (integration_CS) then
            !-> \int (f-feq) dvpar
            call compute_omp_vpar_integral_CS(f%n4,f%h4,fmu_1D, &
              f%BCvpar_left,f%BCvpar_right, &
              f%nspline1d_vpar(tid),scoefvpar, &
              Rarray_PNrPNthetaNphi_nbM2(0,ir,itheta,iphi))
            !-> \int (f-feq) vpar dvpar
            call compute_omp_vpar_integral_CS(f%n4,f%h4,fmu_v_1D, &
              f%BCvpar_left,f%BCvpar_right, &
              f%nspline1d_vpar(tid),scoefvpar, &
              Rarray_PNrPNthetaNphi_nbM2(1,ir,itheta,iphi))
          else
            !-> \int (f-feq) dvpar
            call compute_vpar_integral_colloc(fmu_1D(0:f%n4), &
              geom,Rarray_PNrPNthetaNphi_nbM2(0,ir,itheta,iphi))
            !-> \int (f-feq) vpar dvpar
            call compute_vpar_integral_colloc(fmu_v_1D(0:f%n4), &
              geom,Rarray_PNrPNthetaNphi_nbM2(1,ir,itheta,iphi))
          end if
        enddo
      enddo
    end do
!$OMP END DO
!$OMP BARRIER
!$OMP END PARALLEL

#ifdef TIMER
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tcmp(1))
    tdeb = tfin

    call ppbarrier()

    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tbar(1))
    tdeb = tfin
#endif

    !*** distribution of \int (f-feq) dvpar and ***
    !***  \int (f-feq) vpar dvpar               ***
    !-> message receive
    if (iphistart .ne. -1) then
      nbreqr  = 0
      iphiloc = 0
      do iphi = 0,f%n3-1
        if (moments_mapphi(iphi).eq.pInSpecies_id) then
          do imu = 0, Nmu
            do local_id = 0, Nbproc_loc-1
              global_id  = imu*Nbproc_loc + local_id
              startbuf   = dom_theta * (local_id + Nbproc_loc * &
                (imu + (Nmu+1) * iphiloc)) 
              itag_phi   = 4000 + iphi + pSpecies_id*f%n3
              nbelements = nbmoments * dom_r * dom_theta
              call MPI_IRECV(&
                RRecvBufMPI_A(startbuf*dom_r*nbmoments), &
                nbelements,MPI_REAL8,global_id,itag_phi, &
                mpi_comm_species,moments_reqr(nbreqr),ierr)
              nbreqr = nbreqr + 1
            enddo
          enddo
          iphiloc = iphiloc + 1
        endif
      enddo
    endif

    !-> message sending 
    do iphi = 0,f%n3-1
      itag_phi   = 4000 + iphi + pSpecies_id*f%n3
      nbelements = nbmoments*dom_r*dom_theta
      call MPI_SEND( &
        Rarray_PNrPNthetaNphi_nbM2(0,istart,jstart,iphi),&
        nbelements,MPI_REAL8,moments_mapphi(iphi),itag_phi,&
        mpi_comm_species,ierr)
    end do
    if (iphistart .ne. -1) &
      call MPI_WAITALL(nbreqr, moments_reqr, &
              moments_status, ierr)

#ifdef TIMER
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tcom(1))
    tdeb = tfin
#endif

    if (iphistart .ne. -1) then
   do tid = 1, Nbthread
   call FTR_ALLOC_2(J0%omp_diag(tid)%val,1,J0%N1,1,J0%N2)
   call FTR_ALLOC_2(J0%omp_ldiag(tid)%val,1,J0%N1-1,1,J0%N2)
   call FTR_ALLOC_2(J0%omp_udiag(tid)%val,1,J0%N1-1,1,J0%N2)
   enddo
!$OMP PARALLEL private(tid,imu,iphi,coeff_intdmu, &
!$OMP J0_intdv_f,J0_intvdv_f, &
!$OMP Bnorm_ij,scalprod_mu0Jb_ij, &
!$OMP nbelements,local_id,base,dep, &
!$OMP l_istart,l_iend,l_jstart,l_jend,startbuf,&
!$OMP itheta,ir,iphiloc,Acomp)  default(shared)
!$OMP BARRIER
#ifdef _OPENMP
      tid = 1+omp_get_thread_num()
#else
      tid = 1
#endif
      J0_intdv_f  => Romp1_1Nrp1_1Nthetap1(tid)%val
      J0_intvdv_f => Romp2_1Nrp1_1Nthetap1(tid)%val
      Acomp       => Comp1_1Nrp1_1Ntheta(tid)%val
!$OMP DO SCHEDULE(STATIC,1)
      do iphi = 0,geom%Nphi-1
        if (moments_mapphi(iphi) .eq. pInSpecies_id) then
          iphiloc = iphi - iphistart
          do imu = 0,Nmu
            coeff_intdmu = geom%coeff_intdmu(imu)
            do itheta = 1,geom%Ntheta+1
              do ir = 1,geom%Nr+1
                J0_intdv_f(ir,itheta)  = 0._RKIND
                J0_intvdv_f(ir,itheta) = 0._RKIND
              end do
            end do
            do local_id = 0,Nbproc_loc-1
              base     = (local_id/Nbproc_r)
              dep      = mod(local_id,Nbproc_r)              
              l_istart = dep *  dom_r
              l_iend   = l_istart + dom_r - 1
              l_jstart = base * dom_theta
              l_jend   = l_jstart + dom_theta - 1
              startbuf = dom_theta * (local_id + Nbproc_loc * &
                (imu + (Nmu+1) * iphiloc)) 
              !*** J0_intdv_f  = (\int (f-feq) dvpar) and  ***
              !*** J0_intvdv_f = (\int (f-feq) vpar dvpar) *** 
              do itheta = l_jstart, l_jend
                do ir = l_istart, l_iend
                  J0_intdv_f(1+ir,1+itheta) = &
                    RRecvBufMPI_A( &
                    nbmoments*(ir-l_istart)+0+ &
                    (startbuf+itheta-l_jstart)*dom_r*nbmoments) 
                  J0_intvdv_f(1+ir,1+itheta) = &
                    RRecvBufMPI_A( &
                    nbmoments*(ir-l_istart)+1+ &
                    (startbuf+itheta-l_jstart)*dom_r*nbmoments) 
                enddo
              enddo
            enddo
            !*** compute of the gyroaverages ****             
            !-> J0.(\int (f-feq) dvpar) 
            call omp_compute_gyrofunction_2D(J0,geom,imu, &
              Acomp,J0_intdv_f,tid)
            if (.not.case4D) then
              !-> J0.(\int (f-feq) vpar dvpar)
              call omp_compute_gyrofunction_2D(J0,geom,imu, &
                Acomp,J0_intvdv_f,tid)
            end if
            !*****************************************************
            !*** Computation of the RHS as:                    ***
            !***  rhos(r,theta,phi)= (1/ne0(r))*hat_Zs *       ***
            !***          (nGs(r,theta,phi)-nGs_eq(r,theta))   ***
            !*** with Jv = 2*pi*Bstar being the jacobian       ***
            !*** in velocity and                               *** 
            !***   nGs_eq = \int J0.feq_s Jv dvpar dmu and     ***
            !***   nGs    = \int J0.f_s Jv dvpar dmu           ***
            !***                                               ***
            !*** Rk: We use the fact that                      ***
            !***     rhos = (1/ne0(r))*hat_Zs *                ***
            !***            \int dmu (I0*B_norm +              ***
            !***              I1*(sqrt{hat_As}/hat_Zs)*        ***
            !***              mu0*vec_J.vec_b/(e*B_norm))      ***
            !***   where I0 = J0.(\int (f_s-feq_s) dvpar) and  ***
            !***         I1 = J0.(\int (f_s-feq_s)*vpar dvpar) *** 
            !*** This by using the fact that                   ***
            !***    Bstar = B_norm + (sqrt{hat_As}/hat_Zs)*    ***
            !***            vpar*mu0*vec_J.vec_b/B_norm        ***
            !*****************************************************
            if (.not.case4D) then
              do itheta = 0,f%n2-1
                do ir = 0,f%n1
                  Bnorm_ij = init_magnet%B_norm(ir,itheta) 
                  call compute_scalprod_mu0Jb(geom,init_magnet, &
                    init_curr,ir,itheta,scalprod_mu0Jb_ij)
                  M0_loc(ir,itheta,iphi) = &
                    M0_loc(ir,itheta,iphi) + &
                    TWOPI*coeff_intdmu * &
                    (J0_intdv_f(1+ir,1+itheta)*Bnorm_ij + &
                    J0_intvdv_f(1+ir,1+itheta)*sqrtAs_on_Zs * &
                    scalprod_mu0Jb_ij/Bnorm_ij)
                end do
              end do
            else
              do itheta = 0,f%n2-1
                do ir = 0,f%n1
                  M0_loc(ir,itheta,iphi) = &
                    M0_loc(ir,itheta,iphi) + &
                    coeff_intdmu*J0_intdv_f(1+ir,1+itheta)
                end do
              end do
            end if
          end do
          !*** periodic conditions in theta ***
          M0_loc(0:f%n1,f%n2,iphi) = M0_loc(0:f%n1,0,iphi)
        endif
      end do
!$OMP END DO
!$OMP BARRIER
!$OMP END PARALLEL
   do tid = 1, Nbthread
     call FTR_DEALLOC(J0%omp_diag(tid)%val)
     call FTR_DEALLOC(J0%omp_ldiag(tid)%val)
     call FTR_DEALLOC(J0%omp_udiag(tid)%val)
   enddo

    end if
      
#ifdef TIMER
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tcmp(2))
    call ppbarrier()
    tdeb = tfin
#endif
      
    !**** distribution of rho ***
    !-> message receive
    nbbcast_r = 0
    do iphi = 0,f%n3-1
      itag_phi   = 7000 + iphi + pSpecies_id*f%n3
      nbelements = (Nr+1)*(Ntheta+1)
      call MPI_IRECV(rho(0,0,iphi),nbelements,MPI_REAL8,&
        moments_mapphi(iphi),itag_phi, &
        mpi_comm_species,bcast_reqr(nbbcast_r),ierr)
      nbbcast_r = nbbcast_r + 1
    enddo

    !-> message sending
    if (iphistart .ne. -1) then
      nbbcast_s = 0
      do iphi = 0,geom%Nphi-1
        if (moments_mapphi(iphi) .eq. pInSpecies_id) then
          nbelements = (Nr+1)*(Ntheta+1)
          itag_phi   = 7000 + iphi + pSpecies_id*f%n3
          do iproc = 0,Nbproc_species-1
            call MPI_SEND(M0_loc(0,0,iphi),nbelements,MPI_REAL8,&
              iproc,itag_phi,mpi_comm_species,ierr)
            nbbcast_s = nbbcast_s+1
          enddo
        endif
      end do
    endif
    call MPI_WAITALL(nbbcast_r, bcast_reqr(0:nbbcast_r), &
            moments_status, ierr)
    
    !*** periodic conditions in phi ***
    do itheta = 0,f%n2
      do ir = 0,f%n1
        rho(ir,itheta,f%n3) = rho(ir,itheta,0)
      end do
    end do
    
    !*** compute rhos = (1/ne0(r))*hat_Zs*[nGs-nGs_eq] ***
    do iphi = 0,geom%Nphi
      do itheta = 0,geom%Ntheta
        do ir = 0,geom%Nr
          rho(ir,itheta,iphi) = hatZs* &
            rho(ir,itheta,iphi)/init_prof%ne0(ir)
        end do
      end do
    end do

    do tid = 1, Nbthread
       call FTR_DEALLOC(Romp1_1Nrp1_1Nthetap1(tid)%val)
       call FTR_DEALLOC(Romp2_1Nrp1_1Nthetap1(tid)%val)

       call FTR_DEALLOC(Comp1_1Nrp1_1Ntheta(tid)%val)
    enddo

   call FTR_DEALLOC(RRecvBufMPI_A)

#ifdef TIMER
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tcom(2))
    tdeb = tfin

    if (pglobal_id .eq. 0) then
      write(6,'(I4,A,3F13.6,A,4F13.6,A,F13.6)') pglobal_id, &
        " Temps poisson1A, tcomm ",tcom(1:3)," tcomp ",tcmp(1:4), &
        " tbarrier ",tbar(1)
    end if
#endif
#endif 
    ! HAS_QN_V1
  end subroutine compute_rho_FFT2Dhybrid


  !-------------------------------------------------------
  !  4th order Poisson solver with periodic 
  !  boundary conditions in r and Dirichlet 0 boundary 
  !  conditions in th. phi is considered as a parameter.
  !-------------------------------------------------------
  subroutine solve_poisson_FFT2Dhybrid(pthis,geom, &
    init_prof,init_magnet,init_curr,J0,f,Sfmu_eq)
    use globals, only : mpi_comm_mu, plocal_id, &
      Rarray1_NrNthetaNphi, Carray_1Nrp11Ntheta1Nphi, &
      m, n, single_m, single_n, &
      FFTpotential_filter, FFTpotential_filter_deltam, &
      QN_coef_Phi00, iter_glob, Phi00_BCrmin_Neumann, &
      nbelectrons_diag, nbions_diag
#ifdef _OPENMP
    use OMPutils_module, only : omp_get_thread_num
#endif
    use OMPutils_module, only : Romp1_1Nrp1, Romp2_1Nrp1, &
      Romp3_1Nrp1
    use gyroaverage_class
    use mpiMod
    use globals, only: Nbthread
    type(poisson)      , intent(inout) :: pthis
    type(geometry)     , intent(in)    :: geom
    type(init_profile) , intent(in)    :: init_prof
    type(init_magnetic), intent(in)    :: init_magnet
    type(init_current) , intent(in)    :: init_curr
    type(J0operator)   , intent(inout) :: J0
    type(fdistribu5d)  , intent(in)    :: f
    real(RKIND), &
      dimension(:,:,:) , pointer       :: Sfmu_eq

#ifdef HAS_QN_V1
    !*** local variables ***
    integer     :: ierr
    integer     :: i, j, k                   ! loop indices
    integer     :: ir, itheta, iphi          ! loop indices
    integer     :: N1, N2, N3, N2m1, N3m1    ! domain dimensions
    integer     :: stopit, err               ! error flags
    real(RKIND) :: mask, coef_fftinv
    integer     :: tid, isign

    real(RKIND)   , dimension(:,:,:), pointer :: rho
    complex(CKIND), dimension(:,:,:), pointer :: FFT_Phi
    real(RKIND)   , dimension(:)    , pointer :: ldiag, udiag, diag

    !*** variables for the diagonal filter ***
    real(RKIND) :: ktheta, kphi, ktheta0, ktheta_min, ktheta_max
    real(RKIND) :: qmax, qmin, qloc

#ifdef TIMER
    integer(TIMEPREC) :: tdeb, tfin, tth1, tth2
    real(RKIND)       :: tbar(12), tcmp(12), tcom(12), tthd
    call clck_time(tdeb)
    tcmp(1:4) = 0.
    tcom(1:4) = 0.
#endif
    
    !*** initialize local variables ***
    N1   = geom%Nr+1
    N2   = geom%Ntheta+1
    N3   = geom%Nphi+1
    N2m1 = N2-1
    N3m1 = N3-1

    call write_expr("N1"  , "geom%Nr+1")
    call write_expr("N2"  , "geom%Ntheta+1")
    call write_expr("N2m1", "N2-1")
    call write_expr("N3"  , "geom%Nphi+1")
    call write_expr("N3m1", "N3-1")

    do tid = 1, Nbthread
       call FTR_ALLOC_2(omp_data_i(tid)%val,1,N2m1,1,N3m1)
       call FTR_ALLOC_2(omp_tmp(tid)%val,1,N3m1,1,N2m1)
    enddo

    call FTR_ALLOC_3(Rarray1_NrNthetaNphi,0,geom%Nr,0,geom%Ntheta,0,geom%Nphi)
   
    rho     => Rarray1_NrNthetaNphi
    FFT_Phi => Carray_1Nrp11Ntheta1Nphi 
    
    !*** compute the RHS of Poisson equation       ***
    !***  rho = hat_Zs*1/ne0(\int(J0(f)-J0(feq)dv) ***
    call compute_rho_FFT2Dhybrid(J0,f,Sfmu_eq,geom, &
      init_prof,init_magnet,init_curr,rho)

#ifdef TIMER
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tcmp(1))
    tdeb = tfin
#endif

    !*** perform FFT 2D in theta and phi  ***
    isign  = 1
    stopit = 0    
!$OMP PARALLEL private(tid,ldiag,udiag,diag,k,j,i,err,mask, &
!$OMP ktheta,kphi,ktheta0,ktheta_min,ktheta_max,qmin,qmax,qloc) &
!$OMP firstprivate(isign) default(shared)
!$OMP BARRIER
#ifdef _OPENMP
    tid = 1+omp_get_thread_num()
#else
    tid = 1
#endif
    ldiag => Romp1_1Nrp1(tid)%val
    udiag => Romp2_1Nrp1(tid)%val
    diag  => Romp3_1Nrp1(tid)%val

!$OMP DO SCHEDULE(STATIC)
    do i = 1,N1 
      do k = 1,N3m1
        do j = 1,N2m1
          omp_data_i(tid)%val(j,k) = rho(i-1,j-1,k-1)
        end do
      end do
      call fourrow(omp_data_i(tid)%val,isign)
      do k = 1,N3m1
        do j = 1,N2m1
          omp_tmp(tid)%val(k,j) = omp_data_i(tid)%val(j,k)
        end do
      end do
      call fourrow(omp_tmp(tid)%val,isign)
      do k = 1,N3m1
        do j = 1,N2m1
          FFT_Phi(i,j,k) = omp_tmp(tid)%val(k,j)
        end do
      end do
    enddo
!$OMP END DO
!$OMP BARRIER
    
!$OMP DO SCHEDULE(STATIC)
    do k = 1,N3m1
      !*** solving of tridiagonal systems L Phi^{m,n}=rho^{m,n} ***
      do j = 1,N2m1
        omp_rhs(tid)%val(1:N1-2) = FFT_Phi(2:N1-1,j,k)
        !*** solve the LU system with LAPACK routine ***
        if ((j.eq.1).and.(k.eq.1)) then
          do i = 1,N1-3
            ldiag(i) = pthis%L_ldiag(i)
            udiag(i) = pthis%L_udiag(i)
          end do
          if (QN_coef_Phi00==1) then
            !--> Case lambda = 1
            do i = 1,N1-2
              diag(i) = pthis%L_diag(i)
            end do
          else
            !--> Case lambda = 0
            do i = 1,N1-2
              diag(i) = pthis%L_invZ02Te_diag(i)
            end do
          end if
          call thomas_c(N1-2,udiag,ldiag,diag,omp_rhs(tid)%val,err)
        else
          do i = 1,N1-3
            ldiag(i) = pthis%ldiag(j,i)
            udiag(i) = pthis%udiag(j,i)
          end do
          do i = 1,N1-2
            diag(i)  = pthis%diag(j,i)
          end do
          call thomas_c(N1-2,udiag,ldiag,diag,omp_rhs(tid)%val,err)
        end if
        if (err.ne.0) then
          print *, 'solve_poisson_par: ', &
            'problem in Poisson solving for k = ', k
!$OMP CRITICAL
          stopit = err
!$OMP END CRITICAL
        endif
        !*** copy solution in FFT_Phi ***
        do i = 2,N1-1
          FFT_Phi(i,j,k) = omp_rhs(tid)%val(i-1)
        end do

        !*** boundary conditions ***
        if (Phi00_BCrmin_Neumann) then
          if (j.eq.1) then 
            !-> treatment of the mode (0,0)
            !---> Neumann at r=rmin
            FFT_Phi(1,j,k)  = FFT_Phi(2,j,k)
            !---> Dirichlet at r=rmax
            FFT_Phi(N1,j,k) = 0._CKIND
          else
            !-> treatment of the modes (m,n).ne.0
            !---> Dirichlet at r=rmin
            FFT_Phi(1,j,k)  = 0._CKIND
            !---> Dirichlet at r=rmax 
            FFT_Phi(N1,j,k) = 0._CKIND
          end if
        else
          if (j.eq.1) then
            !-> treatment of the mode (0,0)
            if (geom%rg(0).ne.0) then 
              ! Dirichlet at r=rmin if rmin.ne.0
              FFT_Phi(1,j,k) = 0._CKIND
            else
              ! Neumann   at r=rmin if rmin=0
              FFT_Phi(1,j,k) = FFT_Phi(2,j,k)
            end if
            ! Dirichlet at r=rmax
            FFT_Phi(N1,j,k) = 0._CKIND
          else
            !-> treatment of the modes (m,n).ne.0
            ! Dirichlet at r=rmin
            FFT_Phi(1,j,k)  = 0._CKIND
            ! Dirichlet at r=rmin
            FFT_Phi(N1,j,k) = 0._CKIND
          end if
        end if
      end do
    end do
!$OMP END DO

!$OMP BARRIER
!$OMP MASTER
    !*** Diagonal filter [ m = n*q(r) +- delta_m ]          ***
    !*** with m0 = n*q(r), then FFT_PHI is put equal to 0   ***
    !***  if abs(m-m0)>delta_m                              ***
    !*** Rk : this test is equivalent to                    ***
    !***  ktheta = kphi*q(r)*(Lphi/Ltheta)                  ***
    !***           +- 2*pi*delta_m/Ltheta                   ***
    !***   with ktheta = 2*pi*m/Ltheta and                  ***
    !***        kphi = 2*pi*n/Lphi                          ***
    if (FFTpotential_filter.eq.1) then
      ktheta_min = minval(pthis%kth)
      ktheta_max = maxval(pthis%kth)
      do k = 1,N3m1
        kphi = pthis%kphi(k)
        do j = 1,N2m1
          ktheta = pthis%kth(j)
          do i = 1,N1
            ktheta0 = -kphi*geom%Lphi / &
              (init_prof%iota(i-1)*geom%Ltheta)
            ktheta0 = max(ktheta0,ktheta_min)
            ktheta0 = min(ktheta0,ktheta_max)
            if ( abs(ktheta-ktheta0).gt.FFTpotential_filter_deltam ) then
              FFT_Phi(i,j,k) = 0._CKIND
            end if
          end do
        end do
      end do

    !*** Non res. modes filter           ***
    else if (FFTpotential_filter .eq. 2) then
!AS!      ktheta_min = minval(pthis%kth)
!AS!      ktheta_max = maxval(pthis%kth)
      qmax = maxval(1._RKIND/init_prof%iota)
      qmin = minval(1._RKIND/init_prof%iota)
      do k = 1,N3m1
        kphi = pthis%kphi(k)
        do j = 1,N2m1
          ktheta = pthis%kth(j)
          if (kphi .eq. 0) then
            if (ktheta .ne. 0) FFT_Phi(i,j,k) = 0._CKIND
          else
            qloc = - ktheta/kphi
            if ((qloc .gt. qmax).or.(qloc .lt. qmin)) then
              do i = 1,N1
                FFT_Phi(i,j,k) = 0._CKIND
              end do
            end if
          end if
        end do
      end do
    end if

    !*** the single n mode is retained ***
    if (single_n) then
      do k = 1,N3/2
        mask =  ON - min(ON,abs(ON*(k-1-n)))
        do j = 1,N2m1
          do i = 1,N1
            FFT_Phi(i,j,k) = mask*FFT_Phi(i,j,k)
          end do
        end do
      end do
      do k = N3/2+1,N3m1
        mask =  ON - min(ON,abs(ON*(N3m1-k+1-n)))
        do j = 1,N2m1
          do i = 1,N1
            FFT_Phi(i,j,k) = mask*FFT_Phi(i,j,k)
          end do
        end do
      end do
    end if

    !*** the single m mode is retained ***
    if (single_m) then
      do j = 1,N2/2
        mask =  ON - min(ON,abs(ON*(j-1-m)))
        do k = 1,N3m1
          do i = 1,N1
            FFT_Phi(i,j,k) = mask*FFT_Phi(i,j,k)
          end do
        end do
      end do
      do j = N2/2+1,N2m1
        mask =  ON - min(ON,abs(ON*(N2m1-j+1-m)))
        do k = 1,N3m1
          do i = 1,N1
            FFT_Phi(i,j,k) = mask*FFT_Phi(i,j,k)
          end do
        end do
      end do
    end if

    !*** Saving of the mode (0,0) (zonal flows) ***
    do i = 0,geom%Nr
      Phi00_diag(i) = real(FFT_Phi(i+1,1,1))/(N2m1*N3m1)
    end do
!$OMP END MASTER
!$OMP BARRIER

    !*** Perform inverse FFT 2D of the system solution ***
    isign       = -1
    coef_fftinv = 1._RKIND/(geom%Ntheta*geom%Nphi)
!$OMP DO SCHEDULE(STATIC)
    do i = 1,N1 
      do k = 1,N3m1
        do j = 1,N2m1
          omp_data_i(tid)%val(j,k) = FFT_Phi(i,j,k)
        end do
      end do
      call fourrow(omp_data_i(tid)%val,isign)
      do k = 1,N3m1
        do j = 1,N2m1
          omp_tmp(tid)%val(k,j) = omp_data_i(tid)%val(j,k)
        end do
      end do
      call fourrow(omp_tmp(tid)%val,isign)
      do k = 0,geom%Nphi-1
        do j = 0,geom%Ntheta-1
          pthis%Phi_NrNthPNphi_Cworld(i-1,j,k) = &
            dreal(omp_tmp(tid)%val(k+1,j+1))*coef_fftinv
        enddo
      enddo

      !*** duplicate periodic value ***
      do itheta = 0,geom%Ntheta-1
        pthis%Phi_NrNthPNphi_Cworld(i-1,itheta,geom%Nphi) = &
          pthis%Phi_NrNthPNphi_Cworld(i-1,itheta,0)
      end do
      do iphi = 0,geom%Nphi
        pthis%Phi_NrNthPNphi_Cworld(i-1,geom%Ntheta,iphi) = &
          pthis%Phi_NrNthPNphi_Cworld(i-1,0,iphi)
      end do
    end do
!$OMP END DO
!$OMP BARRIER
!$OMP END PARALLEL
    if (stopit.ne.0) then
      print*,'solve_poisson : problem in Poisson solving for k = '
      stop
    end if

    !----------------------------------------------------------
    !-> Compute the flux surface average of Phi
    !    <Phi>_{flux surf} = \int Phi Js(r,theta) dtheta dphi /
    !         \int Js(r,theta) dtheta dphi
    !----------------------------------------------------------
    call compute_fluxsurf_average(H=pthis%Phi_NrNthPNphi_Cworld, &
      geom=geom,H_Sflux_avg=Phi_FSavg_diag)

    !**********************************************
    !*** computation of the number of electrons ***
    !***  and the number of ions                ***
    !**********************************************
    call compute_nbparticles(geom,init_prof,rho, &
      pthis%Phi_NrNthPNphi_Cworld,Phi00_diag, &
      nbelectrons_diag,nbions_diag)

   do tid = 1, Nbthread
      call FTR_DEALLOC(omp_data_i(tid)%val)
      call FTR_DEALLOC(omp_tmp(tid)%val)
   enddo

    call FTR_DEALLOC(Rarray1_NrNthetaNphi)

#ifdef TIMER
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tcmp(2))
    tdeb = tfin
    if (pglobal_id .eq. 0) then
      write(6,'(I4,A,2F13.6)') pglobal_id, &
        " Temps poisson1B,  tcomp ",tcmp(1:2)
    end if
#endif
#endif 
    !HAS_QN_V1
  end subroutine solve_poisson_FFT2Dhybrid



  !---------------------------------------------------------------
  ! Compute rho the RHS of the quasi-neutrality 
  !  equation, i.e:
  !    rhos(r,theta,phi) = (1/ne0(r))*hat_Zs * 
  !                        (nGs(r,theta,phi)-nGs_eq(r,theta))
  !  where nGs = \int J0(sqrt{2*mu}).f_s Jv  dvpar dmu
  !            = \int J0(sqrt{2*mu}).[\int f_s Jv dvpar] dmu 
  !  with  nGs_eq(r,theta) = \int J0.feq_s(r,theta,vpar,mu) 
  !                          Jv(r,theta,vpar)dmu dvpar
  !  Jv being the Jacobian in velocity space :
  !   . Jv(r,theta,vpar) = 2*pi*Bstar in the 5D case
  !                      = 1          in the 4D case
  !  f_s being the distribution function of specie s.
  !
  ! Rk : rhos = (1/ne0(r)) is computed as the sum of 2 integrals:
  !   Is =  (1/ne0(r))*hat_Zs* \int dmu
  !            (I0*B_norm + (sqrt{hat_As}/hat_Zs)* 
  !            I1*mu0*vec_J.vec_b/B_norm)
  !   where I0 = J0.(\int f_s dvpar) and          
  !         I1 = J0.(\int f_s*vpar dvpar)
  !   this by using the expresion of Bstar
  !    Bstar = B_norm + (sqrt{hat_As}/hat_Zs) * 
  !            vpar*mu0*vec_J.vec_b/B_norm
  !
  ! Rk: BECAREFUL -> the case of several species is not 
  !     treated here
  !------------------------------------------------------------
  subroutine compute_rho_FFT1Dhybrid(pthis,J0,f,Sfmu_eq,geom, &
    init_prof,init_magnet,init_curr,rho,orho,ophi)
    use globals, only : Nbproc_r, Nbproc_loc, Nbproc_mu, &
      dom_r, dom_theta, Nr, Ntheta, Nphi, Nmu, mumin, &
      istart, iend, jstart, jend, mu_id, mpi_comm_intermu, &
      Rarray1_NrNtheta, Rarray1_NrNthetaDomphi, &
      Rarray_PNrPNthetaNphi_nbM2, &
      Phi00_BCrmin_Neumann, QN_coef_Phi00, integration_CS
    use globals, only : hat_As, hat_Zs, &
      mpi_comm_species, pInSpecies_id, Nbproc_species, pSpecies_id, &
      RRecvBufMPI_A
#ifdef _OPENMP
    use OMPutils_module, only : omp_get_thread_num
#endif
    use OMPutils_module, only : Romp_scoefvpar, &
      Romp1_0Nr, Comp1_1Nrp1_1Ntheta, &
      Romp1_1Nrp1_1Nthetap1, Romp2_1Nrp1_1Nthetap1, &
      Romp1_1Nrp1, Romp2_1Nrp1, Romp3_1Nrp1, Romp4_1Nrp1, &
      Romp1_0Nvpar, Romp2_0Nvpar
    use gyroaverage_class
    use mpiMod
    use globals, only: Nbthread
    type(poisson)            , intent(in)    :: pthis
    type(J0operator)         , intent(inout) :: J0
    type(fdistribu5d)        , intent(in)    :: f
    real(RKIND), &
      dimension(:,:,:)       , pointer    :: Sfmu_eq
    type(geometry)           , intent(in) :: geom
    type(init_profile)       , intent(in) :: init_prof
    type(init_magnetic)      , intent(in) :: init_magnet
    type(init_current)       , intent(in) :: init_curr
    real(RKIND), dimension(:), pointer    :: orho
    real(RKIND), dimension(:), pointer    :: ophi
    real(RKIND), dimension(:,:,:), pointer :: rho

#ifdef HAS_QN_V2
    real(RKIND), dimension(:)    , pointer :: ldiag
    real(RKIND), dimension(:)    , pointer :: udiag
    real(RKIND), dimension(:)    , pointer :: diag
    real(RKIND), dimension(:)    , pointer :: tmpv
    real(RKIND), dimension(:)    , pointer :: mrho

    integer, parameter :: nbmoments = 2
    real(RKIND)   , dimension(:)    , pointer :: fmu_1D
    real(RKIND)   , dimension(:)    , pointer :: fmu_v_1D
    real(RKIND)   , dimension(:)    , pointer :: scoefvpar
    real(RKIND)   , dimension(:,:)  , pointer :: J0_intdv_f
    real(RKIND)   , dimension(:,:)  , pointer :: J0_intvdv_f
    real(RKIND)   , dimension(:,:,:), pointer :: M0_loc
    complex(CKIND), dimension(:)    , pointer :: crhs
    complex(CKIND), dimension(:,:)  , pointer :: Acomp

    integer     :: ierr
    integer     :: ir, itheta, iphi, ivpar, imu, iphiloc
    integer     :: i, N1
    logical     :: case4D
    real(RKIND) :: fmu_tmp
    real(RKIND) :: Bnorm_ij, scalprod_mu0Jb_ij
    real(RKIND) :: coeff_intdmu
    real(RKIND) :: vpar, coeff_dthdphi
    real(RKIND) :: hatZs, sqrtAs_on_Zs

    ! -> variable for parallelization
    integer     :: iproc, nbreqr, nbbcast_r, nbbcast_s
    integer     :: local_id, dep, base, global_id, startbuf
    integer     :: l_istart, l_iend, l_jstart, l_jend
    integer     :: nbelements, itag_phi, tid

#ifdef TIMER
    integer(TIMEPREC) :: tdeb, tfin, tth1, tth2
    real(RKIND)       :: tbar(12), tcmp(12), tcom(12), tthd
    call clck_time(tdeb)
    tcmp(1:4) = 0.
    tcom(1:4) = 0.
#endif

    do tid = 1, Nbthread
       call FTR_ALLOC_2(Romp1_1Nrp1_1Nthetap1(tid)%val,1,Nr+1,1,Ntheta+1)
       call FTR_ALLOC_2(Romp2_1Nrp1_1Nthetap1(tid)%val,1,Nr+1,1,Ntheta+1)

       call FTR_ALLOC_2_ZERO(Comp1_1Nrp1_1Ntheta(tid)%val,1,Nr+1,1,Ntheta)
    enddo

   call FTR_ALLOC_1_ZERO(RRecvBufMPI_A,0,nbmoments*dom_r*dom_theta*Nbproc_loc*(Nmu+1)*dom_mapphi)

    case4D = .false.
    if (geom%Nmu.eq.0) case4D = .true.

    !--> compute sqrt{hat_As}/hat_Zs
    hatZs        = hat_Zs(pSpecies_id)
    sqrtAs_on_Zs = sqrt(hat_As(pSpecies_id))/hatZs

    M0_loc => Rarray1_NrNthetaDomphi

!$OMP PARALLEL private(tid,imu,iphi,itheta,ir,ivpar,vpar,&
!$OMP scoefvpar,fmu_tmp,fmu_1D,fmu_v_1D)  default(shared)
#ifdef _OPENMP
    tid = 1+omp_get_thread_num()
#else
    tid = 1
#endif
    scoefvpar => Romp_scoefvpar(tid)%val
    fmu_1D    => Romp1_0Nvpar(tid)%val
    fmu_v_1D  => Romp2_0Nvpar(tid)%val

    !********************************************
    !***  Computation of the two integrals:   ***
    !***    . \int f dvpar                    ***
    !***    . \int f*vpar dvpar               ***
    !********************************************
!$OMP DO SCHEDULE(STATIC)
    do iphi = 0,f%n3-1
      do itheta = f%jstart,f%jend
        do ir = f%istart,f%iend
          do ivpar = 0,f%n4
            fmu_tmp         = f%values(ir,itheta,iphi,ivpar) - &
              Sfmu_eq(ir,itheta,ivpar)
            fmu_1D(ivpar)   = fmu_tmp
            fmu_v_1D(ivpar) = fmu_tmp*geom%vparg(ivpar)
          end do
          !*** computation of the integrals in vpar direction ***
          if (integration_CS) then
            !-> \int f dvpar
            call compute_omp_vpar_integral_CS(f%n4,f%h4,fmu_1D, &
              f%BCvpar_left,f%BCvpar_right, &
              f%nspline1d_vpar(tid),scoefvpar, &
              Rarray_PNrPNthetaNphi_nbM2(0,ir,itheta,iphi))
            !-> \int f vpar dvpar
            call compute_omp_vpar_integral_CS(f%n4,f%h4,fmu_v_1D, &
              f%BCvpar_left,f%BCvpar_right, &
              f%nspline1d_vpar(tid),scoefvpar, &
              Rarray_PNrPNthetaNphi_nbM2(1,ir,itheta,iphi))
          else
            !-> \int f dvpar
            call compute_vpar_integral_colloc(fmu_1D(0:f%n4), &
              geom,Rarray_PNrPNthetaNphi_nbM2(0,ir,itheta,iphi))
            !-> \int f vpar dvpar
            call compute_vpar_integral_colloc(fmu_v_1D(0:f%n4), &
              geom,Rarray_PNrPNthetaNphi_nbM2(1,ir,itheta,iphi))
          end if
        enddo
      enddo
    end do
!$OMP END DO
!$OMP BARRIER
!$OMP END PARALLEL

#ifdef TIMER
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tcmp(1))
    tdeb = tfin
#endif

    call ppbarrier()

#ifdef TIMER
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tbar(1))
    tdeb = tfin
#endif

    !*** distribution of \int f dvpar and \int f vpar dvpar ***
    !-> message receive
    if (iphistart .ne. -1) then
      nbreqr  = 0
      iphiloc = 0
      do iphi = 0,f%n3-1
        if (moments_mapphi(iphi).eq.pInSpecies_id) then
          do imu = 0,Nmu
            do local_id = 0,Nbproc_loc-1
              global_id  = imu*Nbproc_loc + local_id
              startbuf   = dom_theta * &
                (local_id + Nbproc_loc * (imu + (Nmu+1) * iphiloc)) 
              itag_phi   = 4000 + iphi + pSpecies_id*f%n3
              nbelements = nbmoments * dom_r * dom_theta
              call MPI_IRECV( &
                RRecvBufMPI_A(startbuf*dom_r*nbmoments), &
                nbelements,MPI_REAL8,global_id,itag_phi, &
                mpi_comm_species,moments_reqr(nbreqr),ierr)
              nbreqr = nbreqr + 1
            enddo
          enddo
          iphiloc = iphiloc + 1
        endif
      enddo
    endif

    !-> message sending 
    do iphi = 0,f%n3-1
      itag_phi   = 4000 + iphi + pSpecies_id*f%n3
      nbelements = nbmoments*dom_r*dom_theta
      call MPI_SEND( &
        Rarray_PNrPNthetaNphi_nbM2(0,istart,jstart,iphi),&
        nbelements,MPI_REAL8,moments_mapphi(iphi),itag_phi,&
        mpi_comm_species,ierr)
    end do
    if (iphistart .ne. -1) &
      call MPI_WAITALL(nbreqr, moments_reqr, &
              moments_status, ierr)

#ifdef TIMER
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tcom(1))
    tdeb = tfin
#endif

    if (iphistart .ne. -1) then
   do tid = 1, Nbthread
     call FTR_ALLOC_2(J0%omp_diag(tid)%val,1,J0%N1,1,J0%N2)
     call FTR_ALLOC_2(J0%omp_ldiag(tid)%val,1,J0%N1-1,1,J0%N2)
     call FTR_ALLOC_2(J0%omp_udiag(tid)%val,1,J0%N1-1,1,J0%N2)
   enddo
!$OMP PARALLEL private(mrho,tid,imu,iphi, &
!$OMP coeff_intdmu,J0_intdv_f,J0_intvdv_f,nbelements, &
!$OMP Bnorm_ij,scalprod_mu0Jb_ij, &
!$OMP local_id,base,dep,l_istart,l_iend,l_jstart,l_jend,startbuf, &
!$OMP itheta,ir,iphiloc,Acomp) default(shared)
!$OMP BARRIER
#ifdef _OPENMP
      tid = 1+omp_get_thread_num()
#else
      tid = 1
#endif
      mrho        => Romp1_0Nr(tid)%val
      J0_intdv_f  => Romp1_1Nrp1_1Nthetap1(tid)%val
      J0_intvdv_f => Romp2_1Nrp1_1Nthetap1(tid)%val
      Acomp       => Comp1_1Nrp1_1Ntheta(tid)%val

      !***********************************************
      !***  Computation of the gyroaverage applied ***
      !***   to the two integrals \int f dvpar and ***
      !***   \int f*vpar dvpar, i.e:               ***
      !***      . J0.(\int f dvpar)                ***
      !***      . J0.(\int f*vpar dvpar)           ***
      !***********************************************
!$OMP DO SCHEDULE(STATIC,1)
      do iphi = 0,geom%Nphi-1
        if (moments_mapphi(iphi) .eq. pInSpecies_id) then
          do itheta = 0,geom%Ntheta
            do ir = 0,geom%Nr
              M0_loc(ir,itheta,iphi) = 0._RKIND
            end do
          end do
          iphiloc = iphi - iphistart
          do imu = 0,Nmu
            coeff_intdmu = geom%coeff_intdmu(imu)
            do itheta = 1,geom%Ntheta+1
              do ir = 1,geom%Nr+1
                J0_intdv_f(ir,itheta)  = 0._RKIND
                J0_intvdv_f(ir,itheta) = 0._RKIND
              end do
            end do
            do local_id = 0,Nbproc_loc-1
              base     = (local_id/Nbproc_r)
              dep      = mod(local_id,Nbproc_r)              
              l_istart = dep *  dom_r
              l_iend   = l_istart + dom_r - 1
              l_jstart = base * dom_theta
              l_jend   = l_jstart + dom_theta - 1
              startbuf = dom_theta * &
                (local_id + Nbproc_loc * (imu + (Nmu+1) * iphiloc)) 
              !*** J0_intdv_f  = (\int f dvpar) and  ***
              !*** J0_intvdv_f = (\int f vpar dvpar) *** 
              do itheta = l_jstart, l_jend
                do ir = l_istart, l_iend
                  J0_intdv_f(1+ir,1+itheta)  = &
                    RRecvBufMPI_A( &
                    nbmoments*(ir-l_istart)+0+ &
                    (startbuf+itheta-l_jstart)*dom_r*nbmoments) 
                  J0_intvdv_f(1+ir,1+itheta) = &
                    RRecvBufMPI_A( &
                    nbmoments*(ir-l_istart)+1+ &
                    (startbuf+itheta-l_jstart)*dom_r*nbmoments) 
                enddo
              enddo
            enddo
            !*** compute of the gyroaverages ****             
            !-> J0.(\int f dvpar)
            call omp_compute_gyrofunction_2D(J0,geom,imu, &
              Acomp,J0_intdv_f,tid)
            if (.not.case4D) then
              !-> J0.(\int f vpar dvpar)
              call omp_compute_gyrofunction_2D(J0,geom,imu, &
                Acomp,J0_intvdv_f,tid)
            end if
            !*****************************************************
            !*** Computation of the RHS as:                    ***
            !***  rhos(r,theta,phi)= (1/ne0(r))*hat_Zs *       ***
            !***          (nGs(r,theta,phi)-nGs_eq(r,theta))   ***
            !*** with Jv = 2*pi*Bstar being the jacobian       ***
            !*** in velocity and                               *** 
            !***   nGs_eq = \int J0.feq_s Jv dvpar dmu and     ***
            !***   nGs    = \int J0.f_s Jv dvpar dmu           ***
            !***                                               ***
            !*** Rk: We use the fact that                      ***
            !***     rhos = (1/ne0(r))*hat_Zs *                ***
            !***            \int dmu (I0*B_norm +              ***
            !***              I1*(sqrt{hat_As}/hat_Zs)*        ***
            !***              mu0*vec_J.vec_b/(e*B_norm))      ***
            !***   where I0 = J0.(\int (f_s-feq_s) dvpar) and  ***
            !***         I1 = J0.(\int (f_s-feq_s)*vpar dvpar) *** 
            !*** This by using the fact that                   ***
            !***    Bstar = B_norm + (sqrt{hat_As}/hat_Zs)*    ***
            !***            vpar*mu0*vec_J.vec_b/B_norm        ***
            !******************************************************
            if (.not.case4D) then
              do itheta = 0,f%n2-1
                do ir = 0,f%n1
                  Bnorm_ij = init_magnet%B_norm(ir,itheta) 
                  call compute_scalprod_mu0Jb(geom,init_magnet, &
                    init_curr,ir,itheta,scalprod_mu0Jb_ij)
                  M0_loc(ir,itheta,iphi) = &
                    M0_loc(ir,itheta,iphi) + &
                    TWOPI*coeff_intdmu * &
                    (J0_intdv_f(1+ir,1+itheta)*Bnorm_ij + &
                    J0_intvdv_f(1+ir,1+itheta)*sqrtAs_on_Zs * &
                    scalprod_mu0Jb_ij/Bnorm_ij)
                end do
              end do
            else
              do itheta = 0,f%n2-1
                do ir = 0,f%n1
                  M0_loc(ir,itheta,iphi) = &
                    M0_loc(ir,itheta,iphi) + coeff_intdmu * &
                    J0_intdv_f(1+ir,1+itheta)
                end do
              end do
            end if
          end do
          
          !*** compute rhos = 1/ne0(r)*hat_Zs*[nGs-nGs_eq] ***
          mrho(0:f%n1) = 0._RKIND
          do itheta = 0,f%n2-1
            do ir = 0,f%n1
              M0_loc(ir,itheta,iphi) = hatZs* &
                M0_loc(ir,itheta,iphi)/init_prof%ne0(ir)          
              mrho(ir) = mrho(ir) + M0_loc(ir,itheta,iphi)
            end do
          end do
          !*** Store the average of rho over theta into   ***
          !***  the boundaries of M0_loc in order to send ***
          !***  it to other processors                    ***
          M0_loc(0:f%n1,f%n2,iphi) = mrho(0:f%n1)
        endif
      end do
!$OMP END DO
!$OMP BARRIER
!$OMP END PARALLEL
      do tid = 1, Nbthread
        call FTR_DEALLOC(J0%omp_diag(tid)%val)
        call FTR_DEALLOC(J0%omp_ldiag(tid)%val)
        call FTR_DEALLOC(J0%omp_udiag(tid)%val)

        ! TODO : check these two loc_drop_
        call FTR_DEALLOC(Romp1_1Nrp1_1Nthetap1(tid)%val)
        call FTR_DEALLOC(Romp2_1Nrp1_1Nthetap1(tid)%val)
      enddo
    endif

#ifdef TIMER
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tcmp(2))
    call ppbarrier()
    tdeb = tfin
#endif

    !*** distibution of rho ***
    !-> message receive
    nbbcast_r = 0
    do iphi = 0,f%n3-1
      itag_phi   = 7000 + iphi + pSpecies_id*f%n3
      nbelements = (Nr+1)*(Ntheta+1)
      call MPI_IRECV(rho(0,0,iphi),nbelements,MPI_REAL8, &
        moments_mapphi(iphi),itag_phi,mpi_comm_species, &
        bcast_reqr(nbbcast_r),ierr)
      nbbcast_r = nbbcast_r + 1
    enddo

    !-> message sending
    if (iphistart .ne. -1) then
      nbbcast_s = 0
      do iphi = 0,geom%Nphi-1
        if (moments_mapphi(iphi) .eq. pInSpecies_id) then 
          nbelements = (Nr+1)*(Ntheta+1)
          itag_phi   = 7000 + iphi + pSpecies_id*f%n3
          do iproc = 0,Nbproc_species-1
            call MPI_SEND(M0_loc(0,0,iphi),nbelements,MPI_REAL8, &
              iproc,itag_phi,mpi_comm_species,ierr)
            nbbcast_s = nbbcast_s+1
          enddo
        endif
      end do
    endif
    call MPI_WAITALL(nbbcast_r, bcast_reqr, &
            moments_status, ierr)

    do tid = 1, Nbthread
       call FTR_DEALLOC(Comp1_1Nrp1_1Ntheta(tid)%val)
    enddo

   call FTR_DEALLOC(RRecvBufMPI_A)
#ifdef TIMER
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tcom(2))
    tdeb = tfin
#endif

    orho(0:f%n1) = 0._RKIND
    do iphi = 0,geom%Nphi-1
      !*** accumulate the averaged value of rho ***
      !***  over theta (packed into rho)        ***
      orho(0:f%n1) = orho(0:f%n1) + rho(0:f%n1,f%n2,iphi)
      !*** restore periodic conditions in theta ***
      rho(0:f%n1,f%n2,iphi) = rho(0:f%n1,0,iphi)
    enddo
    coeff_dthdphi   = 1._RKIND/(geom%Nphi*geom%Ntheta)
    orho(0:geom%Nr) = orho(0:geom%Nr)*coeff_dthdphi

    !*** periodic conditions in phi ***
    do itheta = 0,f%n2
      do ir = 0,f%n1
        rho(ir,itheta,f%n3) = rho(ir,itheta,0)
      end do
    end do
    
    ! === phi solve
    tid = 1
    ldiag => Romp1_1Nrp1(tid)%val
    udiag => Romp2_1Nrp1(tid)%val
    diag  => Romp3_1Nrp1(tid)%val
    tmpv  => Romp4_1Nrp1(tid)%val

    N1             = f%n1+1
    do i = 1,N1-3
      ldiag(i) = pthis%L_ldiag(i) 
      udiag(i) = pthis%L_udiag(i)
    end do
    if (QN_coef_Phi00==1) then
      !--> Case lambda = 1
      do i = 1,N1-2
        diag(i) = pthis%L_diag(i) 
        ophi(i) = orho(i)
      end do
    else
      !--> Case lambda = 0
      do i = 1,N1-2
        diag(i) = pthis%L_invZ02Te_diag(i) 
        ophi(i) = orho(i)
      end do
    end if
    call lusolver(N1-2,udiag,ldiag,diag,ophi,tmpv,ierr) 
    if (Phi00_BCrmin_Neumann) then
      ophi(0)    = ophi(1)
      ophi(N1-1) = 0._RKIND
    else
      if (geom%rg(0).eq.0_RKIND) then 
        ophi(0) = ophi(1)
      else
        ophi(0) = 0._RKIND
      end if
      ophi(N1-1) = 0._RKIND
    end if

#ifdef TIMER
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tcmp(3))
    tdeb = tfin

    if (pglobal_id .eq. 0) then
      write(6,'(I4,A,3F13.6,A,4F13.6,A,F13.6)') pglobal_id, &
        " Temps poisson2A, tcomm ", tcom(1:3)," tcomp ",tcmp(1:4), &
        " tbarrier ",tbar(1)
    end if
#endif
#endif 
    ! HAS_QN_V2
  end subroutine compute_rho_FFT1Dhybrid

  !------------------------------------------------------
  !  4th order Poisson solver with periodic boundary 
  !  conditions in r and Dirichlet 0 boundary conditions 
  !  in theta. 
  !  phi is considered as a parameter.
  !------------------------------------------------------
  subroutine solve_poisson_FFT1Dhybrid(pthis,geom, &
    init_prof,init_magnet,init_curr,J0,f,Sfmu_eq)
    use globals, only : Nbthread, mpi_comm_mu, plocal_id, &
      iter_glob, QN_coef_Phi00, Z0, &
      FFTpotential_filter, FFTpotential_filter_deltam, &
      nbelectrons_diag, nbions_diag, Rarray1_NrNthetaNphi
#ifdef _OPENMP
    use OMPutils_module, only : omp_get_thread_num
#endif
    use OMPutils_module, only : Romp1_0Nr, Romp2_0Nr, &
      Romp3_0Nr, Romp4_0Nr, Romp5_0Nr, &
      Romp1_1Nrp1, Romp2_1Nrp1, Romp3_1Nrp1, Romp4_1Nrp1, &
      Comp1_1Nrp1_1Ntheta, Comp2_1Nrp1_1Ntheta
    use gyroaverage_class
    use mpiMod
    type(poisson)      , intent(inout) :: pthis
    type(geometry)     , intent(in)    :: geom
    ! used to compute rho expression (equation RHS)
    type(init_profile) , intent(in)    :: init_prof
    type(init_magnetic), intent(in)    :: init_magnet
    type(init_current) , intent(in)    :: init_curr
    type(J0operator)   , intent(inout) :: J0
    type(fdistribu5d)  , intent(in)    :: f
    real(RKIND), &
      dimension(:,:,:) , pointer       :: Sfmu_eq

#ifdef HAS_QN_V2
    !*** local variables ***
    integer     :: ierr
    integer     :: i, j, k                   ! loop indices
    integer     :: ir, itheta, iphi          ! loop indices
    integer     :: N1, N2, N3, N2m1, N3m1    ! domain dimensions
    integer     :: err                       ! error flags
    logical     :: stopit
    real(RKIND) :: mask, coef_fftinv
    real(RKIND) :: icoef
    integer     :: tid
    integer     :: isign, midn2

    real(RKIND)      , dimension(:,:,:), pointer :: rho
    real(RKIND)      , dimension(:)    , pointer :: ldiag
    real(RKIND)      , dimension(:)    , pointer :: udiag
    real(RKIND)      , dimension(:)    , pointer :: diag
    real(RKIND)      , dimension(:)    , pointer :: tmpv
    real(RKIND)      , dimension(:)    , pointer :: mrho
    real(RKIND)      , dimension(:)    , pointer :: mphi
    real(RKIND)      , dimension(:)    , pointer :: ophi
    real(RKIND)      , dimension(:)    , pointer :: orho
    complex(CKIND)   , dimension(:)    , pointer :: prhs
    complex(CKIND)   , dimension(:,:)  , pointer :: fft_rho 
    type(omp1dvector), dimension(:)    , pointer :: Phi00

    !*** variables for the diagonal filter ***
    real(RKIND)                            :: ktheta
    real(RKIND)                            :: ktheta0
    real(RKIND)                            :: ktheta_min
    real(RKIND)                            :: ktheta_max 
    real(RKIND)                            :: kphi
    complex(CKIND),dimension(:,:), pointer :: FFT2D_Phi
    real(RKIND)                            :: qmin, qmax, qloc
#ifdef TIMER
    integer(TIMEPREC) :: tdeb, tfin, tth1, tth2
    real(RKIND)       :: tbar(12), tcmp(12), tthd
    tcmp(1:4) = 0.
#endif

    !*** initialize local variables ***
    N1   = size(geom%rg)
    N2   = size(geom%thetag)
    N3   = size(geom%phig)
    N2m1 = N2-1
    N3m1 = N3-1

    call FTR_ALLOC_3(Rarray1_NrNthetaNphi,0,geom%Nr,0,geom%Ntheta,0,geom%Nphi)

    tid = 1
    rho   => Rarray1_NrNthetaNphi
    orho  => Romp1_0Nr(tid)%val
    ophi  => Romp2_0Nr(tid)%val
    Phi00 => Romp5_0Nr

#ifdef TIMER
    do ir = 1,4
      tcmp(ir) = 0
      tbar(ir) = 0
    end do
#endif

#ifdef TIMER
    call clck_time(tdeb)
#endif

    !*** compute the RHS of Poisson equation   ***
    !***   rhos = 1/ne0(r)*hat_Zs*[nGs-nGs_eq] ***
    call compute_rho_FFT1Dhybrid(pthis,J0,f,Sfmu_eq, &
      geom,init_prof,init_magnet,init_curr,rho,orho,ophi)

#ifdef TIMER
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tcmp(1))
    tdeb = tfin
#endif

    do tid = 1, Nbthread
       call FTR_ALLOC_2_ZERO(Comp1_1Nrp1_1Ntheta(tid)%val,1,Nr+1,1,Ntheta)
       call FTR_ALLOC_2_ZERO(Comp2_1Nrp1_1Ntheta(tid)%val,1,Nr+1,1,Ntheta)
       call FTR_ALLOC_2(omp_data_i(tid)%val,1,N2m1,1,N3m1)
       call FTR_ALLOC_2(omp_tmp(tid)%val,1,N3m1,1,N2m1)
    enddo
    
    !*** perform FFT 1D in theta  ***
    stopit = .false.
!$OMP PARALLEL private(isign,tid,ldiag,udiag,diag, &
!$OMP  tmpv,mrho,mphi,prhs,fft_rho,k,j,i,err,midn2,icoef,ir, &
!$OMP  ktheta,ktheta0,ktheta_min,ktheta_max,kphi, &
!$OMP  coef_fftinv,FFT2D_Phi,qmin,qmax,qloc) &
!$OMP default(shared)
!$OMP BARRIER
#ifdef _OPENMP
    tid = 1+omp_get_thread_num()
#else
    tid = 1
#endif 

    ldiag   => Romp1_1Nrp1(tid)%val
    udiag   => Romp2_1Nrp1(tid)%val
    diag    => Romp3_1Nrp1(tid)%val
    tmpv    => Romp4_1Nrp1(tid)%val
    prhs    => Comp1_1Nrp1_1Ntheta(tid)%val(1:,1)
    mphi    => Romp3_0Nr(tid)%val
    mrho    => Romp4_0Nr(tid)%val
    fft_rho => Comp2_1Nrp1_1Ntheta(tid)%val
    err   = 0
    midn2 = N2m1/2
    icoef = 1.0_RKIND/N2m1
    mphi  = ophi
    do ir = 0,geom%Nr
      mrho(ir)           = orho(ir)
      Phi00(tid)%val(ir) = 0._RKIND
    end do
!$OMP BARRIER
!$OMP DO SCHEDULE(STATIC)
    do k = 1,N3m1  ! N3 -> phi
      do j = 1, N2m1
        do i = 1,N1
          fft_rho(i,j) = rho(i-1,j-1,k-1)-mrho(i-1)
        end do
      end do
      call fourrow(fft_rho,1)
      
      do j=1,midn2+1 ! N2 -> theta
        do i = 1,N1-3
          ldiag(i) = pthis%ldiag(j,i) 
          udiag(i) = pthis%udiag(j,i) 
        end do
        do i = 1,N1-2
          diag(i) = pthis%diag (j,i) 
          prhs(i) = fft_rho(i+1,j) 
        end do
        call  lusolvec(N1-2,udiag,ldiag,diag,prhs,tmpv,err)
        if (err.ne.0) then
!$OMP CRITICAL
          stopit = .true.
!$OMP END CRITICAL
        endif
        do i = 1,N1-2
          fft_rho(i+1,j) = prhs(i)
          if ((j .ne. 1) .and. (j.ne.midn2+1)) then
            fft_rho(i+1,N2m1+2-j) = conjg(prhs(i))
          endif
        end do

        !*** Dirichlet boundary conditions at r=rmin ***
        fft_rho(1,j) = 0._CKIND
        if (j .ne. 1) fft_rho(1,N2m1+2-j) = 0._CKIND
        !*** vanishing Dirichlet boundary conditions at r=rmax ***
        fft_rho(N1,j) = 0._CKIND
        if (j .ne. 1) fft_rho(N1,N2m1+2-j) = 0._CKIND
      enddo

      do i = 1,N1
        fft_rho(i,1) = fft_rho(i,1) + geom%Ntheta * mphi(i-1)
      end do
      
      !*** Saving of the mode (0,0) (zonal flows) ***
      if (QN_coef_Phi00.eq.1) then 
         do i = 0,geom%Nr
            Phi00(tid)%val(i) = Phi00(tid)%val(i) + &
                 real(fft_rho(i+1,1))/(N2m1*N3m1)
         end do
      endif

      call fourrow(fft_rho,-1)

      ! Copy of the results by adding or not
      !  the average of Phi on (theta,phi)
      do j = 1,N2m1
        do i = 1,N1
          pthis%Phi_NrNthPNphi_Cworld(i-1,j-1,k-1) = &
            icoef*real(fft_rho(i,j))
        enddo
      end do
    end do
!$OMP END DO
!$OMP BARRIER
!$OMP MASTER
    do ir = 0,geom%Nr
      Phi00_diag(ir) = 0._RKIND
      do j = 1,Nbthread
        Phi00_diag(ir) = Phi00_diag(ir) + Phi00(j)%val(ir)
      end do
    enddo
!$OMP END MASTER
!$OMP BARRIER

    !*********************************************************
    !*** Diagonal filter [ m = n*q(r) +- delta_m ]         ***
    !*** with m0 = n*q(r), then FFT_PHI is put equal to 0  ***
    !***  if abs(m-m0)>delta_m                             ***
    !*** Rk : this test is equivalent to                   ***
    !***  ktheta = kphi*q(r)*(Lphi/Ltheta)                 ***
    !***           +- 2*pi*delta_m/Ltheta                  ***
    !***   with ktheta = 2*pi*m/Ltheta and                 ***
    !***        kphi   = 2*pi*n/Lphi                       ***
    !*********************************************************
    if ((FFTpotential_filter.eq.1).or.(FFTpotential_filter.eq.2)) then
      ktheta_min  = minval(pthis%kth)
      ktheta_max  = maxval(pthis%kth)
      FFT2D_Phi   => omp_tmp(tid)%val   ! FFT2D_Phi(k,j)
      coef_fftinv = 1._RKIND/(geom%Ntheta*geom%Nphi)

!$OMP DO SCHEDULE(STATIC)
      do i = 1,N1 
        !*** Perform 2D FFT of Phi ***
        isign = 1
        do k = 1,N3m1
          do j = 1,N2m1
            omp_data_i(tid)%val(j,k) = &
              pthis%Phi_NrNthPNphi_Cworld(i-1,j-1,k-1)
          end do
        end do
        call fourrow(omp_data_i(tid)%val,isign)
        do k = 1,N3m1
          do j = 1,N2m1
            FFT2D_Phi(k,j) = omp_data_i(tid)%val(j,k)
          end do
        end do
        call fourrow(FFT2D_Phi,isign)
        if (FFTpotential_filter.eq.1) then
          !*** diagonal filter ***
          do j = 1,N2m1
            ktheta = pthis%kth(j)
            do k = 1,N3m1
              kphi    = pthis%kphi(k)
              ktheta0 = -kphi*geom%Lphi / &
                (init_prof%iota(i-1)*geom%Ltheta)
              ktheta0 = max(ktheta0,ktheta_min)
              ktheta0 = min(ktheta0,ktheta_max)
              if ( abs(ktheta-ktheta0).gt.FFTpotential_filter_deltam ) then
                FFT2D_Phi(k,j) = 0._CKIND
              end if
            end do
          end do
        else if (FFTpotential_filter.eq.2) then
          !*** Non res. modes filter ***
          qmax = maxval(1._RKIND/init_prof%iota)
          qmin = minval(1._RKIND/init_prof%iota)
          do j = 1,N2m1
            ktheta = pthis%kth(j)
            do k = 1,N3m1
              kphi    = pthis%kphi(k)
              if (kphi .eq. 0) then
                if (ktheta .ne. 0) FFT2D_Phi(k,j) = 0._CKIND
              else
                qloc = - ktheta/kphi
                if ((qloc .gt. qmax).or.(qloc.lt.qmin)) then
                  FFT2D_Phi(k,j) = 0._CKIND
                end if
              end if
            end do
          end do
        end if
        !*** Perform 2D inverse FFT of Phi ***
        isign       = -1
        do k = 1,N3m1
          do j = 1,N2m1
            omp_data_i(tid)%val(j,k) = FFT2D_Phi(k,j)
          end do
        end do
        call fourrow(omp_data_i(tid)%val,isign)
        do k = 1,N3m1
          do j = 1,N2m1
            FFT2D_Phi(k,j) = omp_data_i(tid)%val(j,k)
          end do
        end do
        call fourrow(FFT2D_Phi,isign)
        do k = 0,geom%Nphi-1
          do j = 0,geom%Ntheta-1
            pthis%Phi_NrNthPNphi_Cworld(i-1,j,k) = &
              dreal(FFT2D_Phi(k+1,j+1)) * coef_fftinv
          enddo
        enddo
      enddo
!$OMP END DO
    end if
!$OMP BARRIER
!$OMP END PARALLEL

    !*** duplicate periodic value ***
    do i = 1,N1 
      do itheta = 0,geom%Ntheta-1
        pthis%Phi_NrNthPNphi_Cworld(i-1,itheta,geom%Nphi) = &
          pthis%Phi_NrNthPNphi_Cworld(i-1,itheta,0)
      end do
      do iphi = 0,geom%Nphi
        pthis%Phi_NrNthPNphi_Cworld(i-1,geom%Ntheta,iphi) = &
          pthis%Phi_NrNthPNphi_Cworld(i-1,0,iphi)
      end do
    end do

    !**********************************************
    !*** computation of the number of electrons ***
    !***  and the number of ions                ***
    !**********************************************
    call compute_nbparticles(geom,init_prof,rho, &
      pthis%Phi_NrNthPNphi_Cworld,Phi00_diag, &
      nbelectrons_diag,nbions_diag)

    do tid = 1, Nbthread
       call FTR_DEALLOC(Comp1_1Nrp1_1Ntheta(tid)%val)
       call FTR_DEALLOC(Comp2_1Nrp1_1Ntheta(tid)%val)
       call FTR_DEALLOC(omp_data_i(tid)%val)
    enddo

    call FTR_DEALLOC(Rarray1_NrNthetaNphi)

#ifdef TIMER
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tcmp(2))
    tdeb = tfin
    if (pglobal_id .eq. 0) then
      write(6,'(I4,A,2F13.6)') pglobal_id, &
        " Temps poisson2B,  tcomp ",tcmp(1:2)
    end if
#endif
#endif 
    ! HAS_QN_V2    
  end subroutine solve_poisson_FFT1Dhybrid


  !---------------------------------------------------------------
  !  4th order Poisson solver with periodic boundary 
  !  conditions in r and Dirichlet 0 boundary conditions 
  !  in theta. phi is considered as a parameter.
  !  
  ! Compute rho the RHS of the quasi-neutrality 
  !  equation, i.e:
  !    rhos(r,theta,phi) = (1/ne0(r))*hat_Zs * 
  !                        (nGs(r,theta,phi)-nGs_eq(r,theta))
  !  where nGs = \int J0(sqrt{2*mu}).f_s Jv  dvpar dmu
  !            = \int J0(sqrt{2*mu}).[\int f_s Jv dvpar] dmu 
  !  with  nGs_eq(r,theta) = \int J0.feq_s(r,theta,vpar,mu) 
  !                          Jv(r,theta,vpar)dmu dvpar
  !  Jv being the Jacobian in velocity space :
  !   . Jv(r,theta,vpar) = 2*pi*Bstar in the 5D case
  !                      = 1          in the 4D case
  !  f_s being the distribution function of specie s.
  !
  ! Rk : rhos is computed as the sum of 2 integrals:
  !   Is =  (1/ne0(r))*hat_Zs* \int dmu
  !            (I0*B_norm + (sqrt{hat_As}/hat_Zs)* 
  !            I1*mu0*vec_J.vec_b/B_norm)
  !   where I0 = J0.(\int f_s dvpar) and          
  !         I1 = J0.(\int f_s*vpar dvpar)
  !   this by using the expresion of Bstar
  !    Bstar = B_norm + (sqrt{hat_As}/hat_Zs) * 
  !            vpar*mu0*vec_J.vec_b/B_norm
  !
  ! Rk: BECAREFUL -> the case of several species is not 
  !     treated here
  !------------------------------------------------------------
  subroutine solve_poisson_FFT1DMPI(pthis,geom, &
    init_prof,init_magnet,init_curr,J0,f,Sfmu_eq)
    use globals, only : Nbproc_r, Nbproc_loc, Nbproc_mu, &
      dom_r, dom_theta, Nr, Ntheta, Nphi, Nmu, mumin, &
      istart, iend, jstart, jend, mu_id, mpi_comm_intermu, &
      plocal_id, m, n, Z0, QN_coef_Phi00, &
      FFTpotential_filter, FFTpotential_filter_deltam, &
      single_m, single_n, &
      nbelectrons_diag, nbions_diag, &
      Rarray1_NrNtheta, Rarray2_NrNphi, &
      Rarray1_NrNthetaNphi, Rarray1_NrNthetaDomphi, &
      Rarray_PNrPNthetaNphi_nbM2, &
      Phi00_BCrmin_Neumann, integration_CS
    use globals, only : hat_As, hat_Zs, &
      mpi_comm_species, pInSpecies_id, Nbproc_species, pSpecies_id, &
      RRecvBufMPI_A
#ifdef _OPENMP
    use OMPutils_module, only : omp_get_thread_num
#endif
    use OMPutils_module, only : Romp1_0Nr, Romp2_0Nr, Romp3_0Nr, &
      Romp1_1Nrp1, Romp2_1Nrp1, Romp3_1Nrp1, Romp4_1Nrp1, &
      Comp1_1Nrp1_1Ntheta, Romp10_1Nrp1_1Nthetap1, &
      Romp1_1Nrp1_1Nthetap1, Romp2_1Nrp1_1Nthetap1, &
      Romp1_0Nvpar, Romp2_0Nvpar, &
      Romp_scoefvpar, Comp_0Nr, Comp2_1Nrp1_1Ntheta
    use gyroaverage_class
    use mpiMod
    use globals, only: Nbthread
    type(poisson)      , intent(inout) :: pthis
    type(geometry)     , intent(in)    :: geom
    type(init_profile) , intent(in)    :: init_prof
    type(init_magnetic), intent(in)    :: init_magnet
    type(init_current) , intent(in)    :: init_curr
    type(J0operator)   , intent(inout) :: J0
    type(fdistribu5d)  , intent(in)    :: f
    real(RKIND), &
      dimension(:,:,:) , pointer       :: Sfmu_eq

#ifdef HAS_QN_V3
    integer, parameter :: nbmoments = 2
    real(RKIND)   , dimension(:,:,:), pointer :: rho
    real(RKIND)   , dimension(:)    , pointer :: fmu_1D
    real(RKIND)   , dimension(:)    , pointer :: fmu_v_1D
    real(RKIND)   , dimension(:)    , pointer :: scoefvpar
    real(RKIND)   , dimension(:,:)  , pointer :: J0_intdv_f
    real(RKIND)   , dimension(:,:)  , pointer :: J0_intvdv_f
    real(RKIND)   , dimension(:,:,:), pointer :: M0_loc
    complex(CKIND), dimension(:)    , pointer :: crhs
    complex(CKIND), dimension(:,:)  , pointer :: Acomp
    real(RKIND)   , dimension(:)    , pointer :: ne0

    integer     :: no_operation
    integer     :: ierr, N2, N3
    integer     :: ir, itheta, iphi, ivpar, imu, iphiloc
    logical     :: case4D
    real(RKIND) :: fmu_tmp
    real(RKIND) :: Bnorm_ij, scalprod_mu0Jb_ij
    real(RKIND) :: coeff_intdmu
    real(RKIND) :: coeff_intdth, coeff_intdthdphi

    !-> variable for parallelization
    integer     :: nbreqr, nbbcast_r, nbbcast_s
    integer     :: local_id, dep, base, global_id, startbuf
    integer     :: l_istart, l_iend, l_jstart, l_jend
    integer     :: iproc, itag_phi, nbelements, tid
    
    real(RKIND)   , dimension(:)    , pointer :: ldiag
    real(RKIND)   , dimension(:)    , pointer :: udiag
    real(RKIND)   , dimension(:)    , pointer :: diag
    real(RKIND)   , dimension(:)    , pointer :: tmpv
    complex(CKIND), dimension(:)    , pointer :: prhs
    complex(CKIND), dimension(:,:)  , pointer :: fft_rho
    !-> \int rho(r,theta,phi) dtheta dphi for local values of phi
    real(RKIND)   , dimension(:)    , pointer :: intdtheta_rho_loc
    !-> \int rho dtheta
    real(RKIND)   , dimension(:,:)  , pointer :: intdtheta_rho 
    !-> <rho>_{theta,phi} with 
    !->   <.>_{theta,phi} = 1/(Ltheta*Lphi)*\int \int . dtheta dphi
    real(RKIND)   , dimension(:)    , pointer :: rho_thetaphi_avg
    !-> <Phi>_{theta,phi}
    real(RKIND)   , dimension(:)    , pointer :: Phi_thetaphi_avg

    real(RKIND) :: icoef
    integer     :: isign, midn2, N1, i, j, k, err
    logical     :: stopit
    real(RKIND) :: hatZs, sqrtAs_on_Zs

    !*** variables for the diagonal filter ***
    integer     :: N2m1, N3m1
    real(RKIND) :: ktheta, ktheta0
    real(RKIND) :: ktheta_min, ktheta_max 
    real(RKIND) :: kphi
    real(RKIND) :: coef_fftinv
    real(RKIND) :: mask
    complex(CKIND),dimension(:,:), pointer :: FFT2D_Phi
    real(RKIND)                            :: qmin, qmax, qloc

#ifdef TIMER
    integer(TIMEPREC) :: tdeb, tfin, tth1, tth2
    real(RKIND)       :: tcmm(12), tbar(12), tcmp(12), tthd
    tcmp(1:4) = 0.
    tcmm(1:4) = 0.

    call clck_time(tdeb)
    tth1 = tdeb
#endif

    do tid = 1, Nbthread
       call FTR_ALLOC_2(Romp1_1Nrp1_1Nthetap1(tid)%val,1,Nr+1,1,Ntheta+1)
       call FTR_ALLOC_2(Romp2_1Nrp1_1Nthetap1(tid)%val,1,Nr+1,1,Ntheta+1)
       
       call FTR_ALLOC_2_ZERO(Comp1_1Nrp1_1Ntheta(tid)%val,1,Nr+1,1,Ntheta)
       call FTR_ALLOC_2_ZERO(Comp2_1Nrp1_1Ntheta(tid)%val,1,Nr+1,1,Ntheta)
    enddo
    
    do tid = 0, MAX(Nmu,Nbthread)
       call FTR_ALLOC_2(Romp10_1Nrp1_1Nthetap1(tid)%val,1,Nr+1,1,Ntheta+1)
    enddo

    call FTR_ALLOC_1_ZERO(RRecvBufMPI_A,0,nbmoments*dom_r*dom_theta*Nbproc_loc*(Nmu+1)*dom_mapphi)

    call FTR_ALLOC_3(Rarray1_NrNthetaNphi,0,geom%Nr,0,geom%Ntheta,0,geom%Nphi)

    case4D = .false.
    if (geom%Nmu.eq.0) case4D = .true.

    !--> compute sqrt{hat_As}/hat_Zs
    hatZs        = hat_Zs(pSpecies_id)
    sqrtAs_on_Zs = sqrt(hat_As(pSpecies_id))/hatZs

    !*** initialize local variables ***
    N2   = size(geom%thetag)
    N3   = size(geom%phig)
    N2m1 = N2-1
    N3m1 = N3-1

    !*** initialize local variables ***
    ne0 => init_prof%ne0
    N1  = size(geom%rg)
    tid = 1
    rho_thetaphi_avg => Romp1_0Nr(tid)%val
    Phi_thetaphi_avg => Romp2_0Nr(tid)%val
    intdtheta_rho    => Rarray2_NrNphi
    M0_loc           => Rarray1_NrNthetaDomphi
    rho              => Rarray1_NrNthetaNphi

#ifdef TIMER
    do ir = 1,4
      tcmm(ir) = 0
      tcmp(ir) = 0
      tbar(ir) = 0
    end do
#endif

#ifdef TIMER
    call ppbarrier()
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tbar(1))
    tdeb = tfin
#endif

    !*****************************************************
    !*** Computation of the RHS as:                    ***
    !***  rhos(r,theta,phi)= (1/ne0(r))*hat_Zs *       ***
    !***          (nGs(r,theta,phi)-nGs_eq(r,theta))   ***
    !*** with Jv = 2*pi*Bstar being the jacobian       ***
    !*** in velocity and                               *** 
    !***   nGs_eq = \int J0.feq_s Jv dvpar dmu and     ***
    !***   nGs    = \int J0.f_s Jv dvpar dmu           ***
    !***                                               ***
    !*** Rk: We use the fact that                      ***
    !***     rhos = (1/ne0(r))*hat_Zs *                ***
    !***            \int dmu (I0*B_norm +              ***
    !***              I1*(sqrt{hat_As}/hat_Zs)*        ***
    !***              mu0*vec_J.vec_b/(e*B_norm))      ***
    !***   where I0 = J0.(\int (f_s-feq_s) dvpar) and  ***
    !***         I1 = J0.(\int (f_s-feq_s)*vpar dvpar) *** 
    !*** This by using the fact that                   ***
    !***    Bstar = B_norm + (sqrt{hat_As}/hat_Zs)*    ***
    !***            vpar*mu0*vec_J.vec_b/B_norm        ***
    !******************************************************
!$OMP PARALLEL private(tid,imu,iphi,itheta,ir,ivpar,&
!$OMP scoefvpar,fmu_tmp,fmu_1D,fmu_v_1D)  default(shared)
#ifdef _OPENMP
    tid = 1+omp_get_thread_num()
#else
    tid = 1
#endif
    scoefvpar => Romp_scoefvpar(tid)%val
    fmu_1D    => Romp1_0Nvpar(tid)%val
    fmu_v_1D  => Romp2_0Nvpar(tid)%val

!$OMP BARRIER
!$OMP DO SCHEDULE(STATIC)
    do iphi = 0,f%n3-1
      do itheta = f%jstart,f%jend
        do ir = f%istart,f%iend
          do ivpar = 0,f%n4
            fmu_tmp         = f%values(ir,itheta,iphi,ivpar) - &
              Sfmu_eq(ir,itheta,ivpar)
            fmu_1D(ivpar)   = fmu_tmp
            fmu_v_1D(ivpar) = fmu_tmp*geom%vparg(ivpar)
          end do
          !*** computation of the integrals in vpar direction ***
          if (integration_CS) then
            !-> \int (f-feq) dvpar
            call compute_omp_vpar_integral_CS(f%n4,f%h4,fmu_1D, &
              f%BCvpar_left,f%BCvpar_right, &
              f%nspline1d_vpar(tid),scoefvpar, &
              Rarray_PNrPNthetaNphi_nbM2(0,ir,itheta,iphi))
            !-> \int (f-feq) vpar dvpar
            call compute_omp_vpar_integral_CS(f%n4,f%h4,fmu_v_1D, &
              f%BCvpar_left,f%BCvpar_right, &
              f%nspline1d_vpar(tid),scoefvpar, &
              Rarray_PNrPNthetaNphi_nbM2(1,ir,itheta,iphi))
          else
            !-> \int (f-feq) dvpar
            call compute_vpar_integral_colloc(fmu_1D(0:f%n4), &
              geom,Rarray_PNrPNthetaNphi_nbM2(0,ir,itheta,iphi))
            !-> \int (f-feq) vpar dvpar
            call compute_vpar_integral_colloc(fmu_v_1D(0:f%n4), &
              geom,Rarray_PNrPNthetaNphi_nbM2(1,ir,itheta,iphi))
          end if
        enddo
      enddo
    end do
!$OMP END DO
!$OMP BARRIER
!$OMP END PARALLEL


#ifdef TIMER
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tcmp(1))
    tdeb = tfin
#endif

    !-> message receive
    if (iphistart .ne. -1) then
      nbreqr  = 0
      iphiloc = 0
      do iphi = 0,f%n3-1
        if (moments_mapphi(iphi).eq.pInSpecies_id) then
          do imu = 0, Nmu
            do local_id = 0,Nbproc_loc-1
              global_id  = imu*Nbproc_loc + local_id
              startbuf   = dom_theta * (local_id + Nbproc_loc * &
                (imu + (Nmu+1) * iphiloc)) 
              itag_phi   = 4000 + iphi + pSpecies_id*f%n3
              nbelements = nbmoments * dom_r * dom_theta
              call MPI_IRECV( &
                RRecvBufMPI_A(startbuf*dom_r*nbmoments), &
                nbelements,MPI_REAL8,global_id,itag_phi, &
                mpi_comm_species,moments_reqr(nbreqr),ierr)
              nbreqr = nbreqr + 1
            enddo
          enddo
          iphiloc = iphiloc + 1
        endif
      enddo
    end if
    !-> message sending 
    do iphi = 0,f%n3-1
      itag_phi   = 4000 + iphi + pSpecies_id*f%n3
      nbelements = nbmoments*dom_r*dom_theta
      call MPI_SEND(& 
        Rarray_PNrPNthetaNphi_nbM2(0,istart,jstart,iphi),&
        nbelements,MPI_REAL8,moments_mapphi(iphi),itag_phi,&
        mpi_comm_species,ierr)
    end do
    if (iphistart .ne. -1) &
      call MPI_WAITALL(nbreqr, moments_reqr, &
              moments_status, ierr)

#ifdef TIMER
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tcmm(1))
    tdeb = tfin
#endif
   
    tid    = 1
    stopit = .false.
    if (iphistart .ne. -1) then
   do tid = 1, Nbthread
     call FTR_ALLOC_2(J0%omp_diag(tid)%val,1,J0%N1,1,J0%N2)
     call FTR_ALLOC_2(J0%omp_ldiag(tid)%val,1,J0%N1-1,1,J0%N2)
     call FTR_ALLOC_2(J0%omp_udiag(tid)%val,1,J0%N1-1,1,J0%N2)
   enddo
!$OMP PARALLEL private(intdtheta_rho_loc, &
!$OMP nbbcast_s,tid,imu,iphi, &
!$OMP coeff_intdmu,J0_intdv_f,J0_intvdv_f, &
!$OMP nbelements,local_id,base,dep,l_istart, &
!$OMP l_iend,l_jstart,l_jend,startbuf,itheta,ir, &
!$OMP iphiloc,Bnorm_ij,scalprod_mu0Jb_ij,Acomp,isign,&
!$OMP j,i,err,midn2,icoef)  default(shared)
!$OMP BARRIER
#ifdef _OPENMP
      tid = 1+omp_get_thread_num()
#else
      tid = 1
#endif
      J0_intdv_f        => Romp1_1Nrp1_1Nthetap1(tid)%val
      J0_intvdv_f       => Romp2_1Nrp1_1Nthetap1(tid)%val
      Acomp             => Comp1_1Nrp1_1Ntheta(tid)%val
      intdtheta_rho_loc => Romp3_0Nr(tid)%val

      err   = 0
      isign = 1      
      midn2 = geom%Ntheta/2
      icoef = 1.0_RKIND/geom%Ntheta
      
      do iphi = 0,geom%Nphi-1
        if (moments_mapphi(iphi) .eq. pInSpecies_id) then
!$OMP BARRIER
          iphiloc = iphi - iphistart
          !*** Parallelization in variable imu in order ***
          !***  to speed up computation                 ***
          !*** whenever Nmu is large                    ***
!$OMP DO SCHEDULE(STATIC,1)
          do imu = 0,Nmu
            coeff_intdmu = geom%coeff_intdmu(imu)
            do itheta = 0,geom%Ntheta
              do ir = 0,geom%Nr
                J0_intdv_f(1+ir,1+itheta)  = 0._RKIND
                J0_intvdv_f(1+ir,1+itheta) = 0._RKIND
              end do
            end do
            do local_id = 0,Nbproc_loc-1
              base     = (local_id/Nbproc_r)
              dep      = mod(local_id,Nbproc_r)              
              l_istart = dep *  dom_r
              l_iend   = l_istart + dom_r - 1
              l_jstart = base * dom_theta
              l_jend   = l_jstart + dom_theta - 1
              startbuf = dom_theta * (local_id + Nbproc_loc * &
                (imu + (Nmu+1) * iphiloc)) 
              !*** J0_intdv_f  = (\int (f-feq) dvpar) and  ***
              !*** J0_intvdv_f = (\int (f-feq) vpar dvpar) *** 
              do itheta = l_jstart, l_jend
                do ir = l_istart, l_iend
                  J0_intdv_f(1+ir,1+itheta)  = &
                    RRecvBufMPI_A( &
                    nbmoments*(ir-l_istart)+0+ &
                    (startbuf+itheta-l_jstart)*dom_r*nbmoments) 
                  J0_intvdv_f(1+ir,1+itheta) = &
                    RRecvBufMPI_A( &
                    nbmoments*(ir-l_istart)+1+ &
                    (startbuf+itheta-l_jstart)*dom_r*nbmoments) 
                enddo
              enddo
            enddo
            !*** computation of the gyroaverages ***             
            !-> J0.(\int (f-feq) dvpar)
            call omp_compute_gyrofunction_2D(J0,geom,imu, &
              Acomp,J0_intdv_f,tid)
            if (.not.case4D) then
              !-> J0.(\int (f-feq) vpar dvpar)
              call omp_compute_gyrofunction_2D(J0,geom,imu, &
                Acomp,J0_intvdv_f,tid)
            end if
            !******************************************************
            !*** computation of the integrals in mu             ***
            !***     rhos = (1/ne0(r))*hat_Zs *                 ***
            !***            \int dmu (I0*B_norm +               ***
            !***              I1*(sqrt{hat_As}/hat_Zs)*         ***
            !***              mu0*vec_J.vec_b/(e*B_norm))       ***
            !***   where I0 = J0.(\int (f_s-feq_s) dvpar) and   ***
            !***         I1 = J0.(\int (f_s-feq_s)*vpar dvpar)  *** 
            !******************************************************
            !->  Sum the contribution into       
            !    Romp10_1Nrp1_1Nthetap1(tid)%val belonging 
            !    to thread 'tid'      
            if (.not.case4D) then
              do itheta = 0,f%n2-1
                do ir = 0,f%n1
                  Bnorm_ij = init_magnet%B_norm(ir,itheta) 
                  call compute_scalprod_mu0Jb(geom,init_magnet, &
                    init_curr,ir,itheta,scalprod_mu0Jb_ij)
                  Romp10_1Nrp1_1Nthetap1(imu)%val(1+ir,1+itheta) = &
                    TWOPI*coeff_intdmu * &
                    (J0_intdv_f(1+ir,1+itheta)*Bnorm_ij + &
                    J0_intvdv_f(1+ir,1+itheta)*sqrtAs_on_Zs * &
                    scalprod_mu0Jb_ij/Bnorm_ij)
                end do
              end do
            else
              do itheta = 0,f%n2-1
                do ir = 0,f%n1
                  Romp10_1Nrp1_1Nthetap1(imu)%val(1+ir,1+itheta) = &
                    coeff_intdmu * J0_intdv_f(1+ir,1+itheta)
                end do
              end do
            end if
          end do
!$OMP END DO
!$OMP BARRIER
!$OMP MASTER
          do itheta = 0,geom%Ntheta
            do ir = 0,geom%Nr
              M0_loc(ir,itheta,iphi) = 0._RKIND
            end do
          end do
          do ir = 0,f%n1
            intdtheta_rho_loc(ir) = 0._RKIND
          end do
          !*** compute rhos = 1/ne0(r)*hat_Zs*[nGs-nGs_eq]  ***
          !***  AND compute the theta averaged value of rho ***
          do itheta = 0,f%n2-1
            do ir = 0,f%n1
              !*** Sum the contribution of each thread   ***
              !*** 'Romp10_1Nrp1_1Nthetap1(tid)%val'     ***
              !***  into M0_loc data structure           *** 
              do j = 0,Nmu
                !-> M0 = \int J0.f Jv dvpar dmu
                M0_loc(ir,itheta,iphi) = M0_loc(ir,itheta,iphi) + &
                  Romp10_1Nrp1_1Nthetap1(j)%val(1+ir,1+itheta)
              end do
              M0_loc(ir,itheta,iphi) = hatZs* &
                M0_loc(ir,itheta,iphi)/ne0(ir)
              intdtheta_rho_loc(ir)  = intdtheta_rho_loc(ir) + &
                M0_loc(ir,itheta,iphi)
            end do
          end do
          !*** Store the integral of rho over theta      ***
          !***  into the boundaries of M0_loc            ***
          !***  in order to send it to other processors  ***
          do ir = 0,f%n1
            M0_loc(ir,f%n2,iphi) = intdtheta_rho_loc(ir)
          end do
!$OMP END MASTER
        end if
        no_operation = 0
!$OMP BARRIER
      end do
!$OMP BARRIER
!$OMP END PARALLEL
   do tid = 1, Nbthread
     call FTR_DEALLOC(J0%omp_diag(tid)%val)
     call FTR_DEALLOC(J0%omp_ldiag(tid)%val)
     call FTR_DEALLOC(J0%omp_udiag(tid)%val)
   enddo
    end if

#ifdef TIMER
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tcmp(2))
    tdeb = tfin
#endif

    !*** distribution of \int rho dtheta ***
    if (iphistart .ne. -1) then
      !-> message receive
      nbbcast_r = 0
      nbbcast_s = 0
      do iphi = 0,f%n3-1
        itag_phi   = 7000 + iphi + pSpecies_id*f%n3
        nbelements = (Nr+1)
        call MPI_IRECV(intdtheta_rho(0,iphi),nbelements,MPI_REAL8, &
          moments_mapphi(iphi),itag_phi,mpi_comm_species, &
          bcast_reqr(nbbcast_r),ierr)
        if (moments_mapphi(iphi) .eq. pInSpecies_id) then
          nbelements = (Nr+1)
          itag_phi   = 7000 + iphi + pSpecies_id*f%n3
          do iproc = 0,MIN(Nbproc_species,Nphi)-1
            call MPI_SEND(M0_loc(0,f%n2,iphi),nbelements, &
              MPI_REAL8,iproc,itag_phi,mpi_comm_species,ierr)
            nbbcast_s = nbbcast_s + 1
          enddo
        endif
        call MPI_WAIT(bcast_reqr(nbbcast_r), moments_status(:,0), ierr)
        nbbcast_r = nbbcast_r + 1
      enddo
    endif

#ifdef TIMER
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tcmm(2))
    tdeb = tfin
#endif

    !**********************************************
    !*  Poisson solver                            *
    !*   -> compute pthis%Phi_NrNthPNphi_Cworld   *
    !**********************************************
    if (iphistart .ne. -1) then
      do ir = 0,geom%Nr
        rho_thetaphi_avg(ir) = 0._RKIND
      end do
      do iphi = 0,geom%Nphi-1
        !--> compute \int dphi of \int rho dtheta
        do ir = 0,geom%Nr
          rho_thetaphi_avg(ir) = rho_thetaphi_avg(ir) + &
            intdtheta_rho(ir,iphi)
        end do
      enddo
      do iphi = 0,geom%Nphi-1
         if (moments_mapphi(iphi) .eq. pInSpecies_id) then 
            !*** restore periodic conditions in theta ***
            do ir = 0,geom%Nr
               M0_loc(ir,f%n2,iphi) = M0_loc(ir,0,iphi)
            end do
            do itheta = 0,geom%Ntheta
               do ir = 0,geom%Nr
                  rho(ir,itheta,iphi) = M0_loc(ir,itheta,iphi)
               end do
            end do
         end if
      end do
      coeff_intdthdphi = 1._RKIND/(geom%Nphi*geom%Ntheta)
      do ir = 0,geom%Nr
        rho_thetaphi_avg(ir) = rho_thetaphi_avg(ir)*coeff_intdthdphi
      end do

      !*** Compute Phi in adding                       ***  
      !***    [Phi(r,theta,phi)-<Phi>_{theta}(r,phi)]  ***
      !*** (which is stored into rho)                  ***
      !*** to <Phi>_{theta}(r,phi) that is             ***
      !*** found thanks to                             *** 
      tid   = 1
      ldiag => Romp1_1Nrp1(tid)%val
      udiag => Romp2_1Nrp1(tid)%val
      diag  => Romp3_1Nrp1(tid)%val
      tmpv  => Romp4_1Nrp1(tid)%val
    
      N1 = f%n1+1
      do ir = 1,N1-2
         Phi_thetaphi_avg(ir) = rho_thetaphi_avg(ir)
      end do
      do ir = 1,N1-3
        ldiag(ir) = pthis%L_ldiag(ir) 
        udiag(ir) = pthis%L_udiag(ir) 
      end do
      if (QN_coef_Phi00==1) then 
        !--> case lambda = 1
        do ir = 1,N1-2
          diag(ir) = pthis%L_diag(ir)
        end do
      else
        !--> case lambda = 0
        do ir = 1,N1-2
          diag(ir) = pthis%L_invZ02Te_diag(ir)
        end do
      end if
      call lusolver(N1-2,udiag,ldiag,diag,Phi_thetaphi_avg,tmpv,ierr) 
      if (Phi00_BCrmin_Neumann) then
        Phi_thetaphi_avg(0)    = Phi_thetaphi_avg(1)
        Phi_thetaphi_avg(N1-1) = 0._RKIND
      else
        if (geom%rg(0).eq.0_RKIND) then 
          Phi_thetaphi_avg(0) = Phi_thetaphi_avg(1)
        else
          Phi_thetaphi_avg(0) = 0._RKIND
        end if
        Phi_thetaphi_avg(N1-1) = 0._RKIND
      end if

!$OMP PARALLEL private(nbbcast_s,tid,imu,iphi, &
!$OMP coeff_intdmu,nbelements,local_id,base,dep, &
!$OMP l_istart,l_iend,l_jstart,l_jend,startbuf, &
!$OMP itheta,ir,iphiloc, &
!$OMP isign,ldiag,udiag,diag,tmpv,prhs,fft_rho, &
!$OMP j,i,err,midn2,icoef) default(shared)
!$OMP BARRIER
#ifdef _OPENMP
      tid = 1+omp_get_thread_num()
#else
      tid = 1
#endif
      ldiag   => Romp1_1Nrp1(tid)%val
      udiag   => Romp2_1Nrp1(tid)%val
      diag    => Romp3_1Nrp1(tid)%val
      tmpv    => Romp4_1Nrp1(tid)%val
      prhs    => Comp_0Nr(tid)%val
      fft_rho => Comp2_1Nrp1_1Ntheta(tid)%val

      N1       = f%n1+1
      err      = 0
      isign    = 1
      midn2    = geom%Ntheta/2
      icoef    = 1.0_RKIND/geom%Ntheta

!$OMP DO SCHEDULE(STATIC,1)
      do iphi = 0,geom%Nphi-1
        if (moments_mapphi(iphi) .eq. pInSpecies_id) then
          !*** the big following block computes          ***
          !***  [Phi(r,theta,phi)-<Phi>_{theta}(r,phi)]  ***
          do j = 1,geom%Ntheta
            do i = 1,N1
              !-> in Fourier for the modes > 0   
              !   \hat{rho(i,k)}^m = \hat{rho(i,k)-mrho(i)}^m
              fft_rho(i,j) = M0_loc(i-1,j-1,iphi) - &
                real(QN_coef_Phi00)*rho_thetaphi_avg(i-1)
            end do
          end do
          call fourrow(fft_rho,1)

          do j = 1,midn2+1 ! n2 -> theta 
            do ir = 1,N1-3
              ldiag(ir) = pthis%ldiag(j,ir) 
              udiag(ir) = pthis%udiag(j,ir) 
            end do
            do ir = 1,N1-2
              diag (ir) = pthis%diag (j,ir) 
              prhs (ir) = fft_rho(1+ir,j) 
            end do
            call lusolvec(N1-2,udiag,ldiag,diag,prhs,tmpv,err) 
            if (err.ne.0) then
!$OMP CRITICAL
               stopit = .true.
!$OMP END CRITICAL
            endif
            do ir = 2,N1-1
              fft_rho(ir,j) = prhs (ir-1)            
            end do
            if ((j .ne. 1) .and. (j.ne.midn2+1)) then
              do ir = 2,N1-1
                fft_rho(ir,geom%Ntheta+2-j) = conjg(prhs(ir-1))
              end do
            endif
            !*** vanishing Dirichlet boundary conditions ***
            !***  in r direction                         ***
            fft_rho(1,j)  = 0._CKIND
            if (j .ne. 1) fft_rho(1,geom%Ntheta+2-j) = 0._CKIND
            fft_rho(N1,j) = 0._CKIND
            if (j .ne. 1) fft_rho(N1,geom%Ntheta+2-j) = 0._CKIND
          enddo
          
          call fourrow(fft_rho,-1)
          do j = 1,geom%Ntheta
            do i = 1,N1
              M0_loc(i-1,j-1,iphi) = icoef*real(fft_rho(i,j)) + &
                real(QN_coef_Phi00)*Phi_thetaphi_avg(i-1)
            enddo
          end do
          
        endif
      end do
!$OMP END DO
!$OMP BARRIER
!$OMP END PARALLEL
    endif

#ifdef TIMER
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tcmp(3))
    tdeb = tfin
#endif

    !*** distribution of Phi ***
    !-> message receive
    nbbcast_r = 0
    do iphi = 0,f%n3-1
      itag_phi   = 9000 + iphi + pSpecies_id*f%n3
      nbelements = (Nr+1)*(Ntheta+1)
      call MPI_IRECV(pthis%Phi_NrNthPNphi_Cworld(0,0,iphi), &
        nbelements,MPI_REAL8,moments_mapphi(iphi), &
        itag_phi,mpi_comm_species,bcast_reqr(nbbcast_r),ierr)
      nbbcast_r = nbbcast_r + 1
    enddo
    !-> message sending
    if (iphistart .ne. -1) then
      nbbcast_s = 0
      do iphi = 0,geom%Nphi-1
        if (moments_mapphi(iphi) .eq. pInSpecies_id) then
          nbelements = (Nr+1)*(Ntheta+1)
          itag_phi   = 9000 + iphi + pSpecies_id*f%n3
          do iproc = 0,Nbproc_species-1
            call MPI_SEND(M0_loc(0,0,iphi),nbelements,MPI_REAL8, &
              iproc,itag_phi,mpi_comm_species, &
              ierr)
            nbbcast_s = nbbcast_s+1
          enddo
        endif
      end do
    endif
    call MPI_WAITALL(nbbcast_r, bcast_reqr(0:nbbcast_r-1), &
            moments_status, ierr)

    !*** computation of the mode (0,0) of Phi ***
    do ir = 0,geom%Nr
      Phi00_diag(ir) = 0._RKIND
      do iphi = 0, geom%Nphi-1
        Phi00_diag(ir) = Phi00_diag(ir) + &
          pthis%Phi_NrNthPNphi_Cworld(ir,geom%Ntheta,iphi)
      enddo
    end do

#ifdef TIMER
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tcmm(3))
    tdeb = tfin
#endif

    !**********************************************************
    !*** Diagonal filter [ m = n*q(r) +- delta_m ]          ***
    !*** with m0 = n*q(r), then FFT_PHI is put equal to 0   ***
    !***  if abs(m-m0)>delta_m                              ***
    !*** Rk : this test is equivalent to                    ***
    !***  ktheta = kphi*q(r)*(Lphi/Ltheta) +-               ***
    !***            2*pi*delta_m/Ltheta                     *** 
    !***   with ktheta = 2*pi*m/Ltheta and                  ***
    !***        kphi   = 2*pi*n/Lphi                        ***
    !**********************************************************
    if ((FFTpotential_filter.eq.1).or. &
        (FFTpotential_filter.eq.2).or. &
        (single_n).or.(single_m)) then

   do tid = 1, Nbthread
   call FTR_ALLOC_2(omp_data_i(tid)%val,1,N2m1,1,N3m1)
   enddo

      N2m1        = size(geom%thetag)-1
      N3m1        = size(geom%phig)-1
      ktheta_min  = minval(pthis%kth)
      ktheta_max  = maxval(pthis%kth)
      coef_fftinv = 1._RKIND/(geom%Ntheta*geom%Nphi)
!$OMP PARALLEL private(isign,tid,k,j,i,mask, &
!$OMP  ktheta,ktheta0,kphi,FFT2D_Phi, &
!$OMP  qmin,qmax,qloc) &
!$OMP default(shared)
!$OMP BARRIER
#ifdef _OPENMP
      tid = 1+omp_get_thread_num()
#else
      tid = 1
#endif 
      FFT2D_Phi   => omp_tmp(tid)%val   ! FFT2D_Phi(k,j)

!$OMP DO SCHEDULE(STATIC)
      do i = 1,N1 
        !*** Perform 2D FFT of Phi ***
        isign = 1
        do k = 1,N3m1
          do j = 1,N2m1
            omp_data_i(tid)%val(j,k) = &
              pthis%Phi_NrNthPNphi_Cworld(i-1,j-1,k-1)
          end do
        end do
        call fourrow(omp_data_i(tid)%val,isign)
        do k = 1,N3m1
          do j = 1,N2m1
            FFT2D_Phi(k,j) = omp_data_i(tid)%val(j,k)
          end do
        end do
        call fourrow(FFT2D_Phi,isign)
        if (FFTpotential_filter.eq.1) then
          !*** diagonal filter ***
          do j = 1,N2m1
            ktheta = pthis%kth(j)
            do k = 1,N3m1
              kphi    = pthis%kphi(k)
              ktheta0 = -kphi*geom%Lphi / &
                (init_prof%iota(i-1)*geom%Ltheta)
              ktheta0 = max(ktheta0,ktheta_min)
              ktheta0 = min(ktheta0,ktheta_max)
              if ( abs(ktheta-ktheta0).gt.FFTpotential_filter_deltam ) then
                FFT2D_Phi(k,j) = 0._CKIND
              end if
            end do
          end do
        else if (FFTpotential_filter.eq.2) then
          qmax = maxval(1._RKIND/init_prof%iota)
          qmin = minval(1._RKIND/init_prof%iota)
          do j = 1,N2m1
            ktheta = pthis%kth(j)
            do k = 1,N3m1
              kphi    = pthis%kphi(k)
              if (kphi .eq. 0) then
                if (ktheta .ne. 0) FFT2D_Phi(k,j) = 0._CKIND
              else
                qloc = - ktheta/kphi
                if ((qloc .gt. qmax).or.(qloc.lt.qmin)) then
                  FFT2D_Phi(k,j) = 0._CKIND
                end if
              end if
            end do
          end do
        end if
        !*** the single n mode is retained ***
        if (single_n) then
          do k = 1,N3/2
            mask =  ON - min(ON,abs(ON*(k-1-n)))
            do j = 1,N2m1
              FFT2D_Phi(k,j) = mask*FFT2D_Phi(k,j)
            end do
          end do
          do k = N3/2+1,N3m1
            mask =  ON - min(ON,abs(ON*(N3m1-k+1-n)))
            do j = 1,N2m1
              FFT2D_Phi(k,j) = mask*FFT2D_Phi(k,j)
            end do
          end do
        end if
        !*** the single m mode is retained ***
        if (single_m) then
          do j = 1,N2/2
            mask =  ON - min(ON,abs(ON*(j-1-m)))
            do k = 1,N3m1
              FFT2D_Phi(k,j) = mask*FFT2D_Phi(k,j)
            end do
          end do
          do j = N2/2+1,N2m1
            mask =  ON - min(ON,abs(ON*(N3m1-j+1-m)))
            do k = 1,N3m1
              FFT2D_Phi(k,j) = mask*FFT2D_Phi(k,j)
            end do
          end do
        end if
        !*** Perform 2D inverse FFT of Phi ***
        isign = -1
        do k = 1,N3m1
          do j = 1,N2m1
            omp_data_i(tid)%val(j,k) = FFT2D_Phi(k,j)
          end do
        end do
        call fourrow(omp_data_i(tid)%val,isign)
        do k = 1,N3m1
          do j = 1,N2m1
            FFT2D_Phi(k,j) = omp_data_i(tid)%val(j,k)
          end do
        end do
        call fourrow(FFT2D_Phi,isign)
        do k = 0,geom%Nphi-1
          do j = 0,geom%Ntheta-1
            pthis%Phi_NrNthPNphi_Cworld(i-1,j,k) = &
              dreal(FFT2D_Phi(k+1,j+1))*coef_fftinv
          enddo
        enddo
      enddo
!$OMP END DO
!$OMP BARRIER
!$OMP END PARALLEL

   do tid = 1, Nbthread
     call FTR_DEALLOC(omp_data_i(tid)%val)
     call FTR_DEALLOC(omp_tmp(tid)%val)
   enddo

    end if

    !*** duplicate periodic value ***
    do i = 1,N1 
      do j = 0,geom%Ntheta-1
        pthis%Phi_NrNthPNphi_Cworld(i-1,j,geom%Nphi) = &
          pthis%Phi_NrNthPNphi_Cworld(i-1,j,0)
      end do
      do iphi = 0,geom%Nphi
        pthis%Phi_NrNthPNphi_Cworld(i-1,geom%Ntheta,iphi) = &
          pthis%Phi_NrNthPNphi_Cworld(i-1,0,iphi)
      end do
    end do

    !----------------------------------------------------------
    !-> Saving of the Fourier (0,0) mode of Phi
    !    Phi00 = 1/(Ltheta*Lphi)\int Phi dtheta dphi 
    !----------------------------------------------------------
    Phi00_diag = Phi_thetaphi_avg

    !----------------------------------------------------------
    !-> Compute the flux surface average of Phi
    !    <Phi>_{flux surf} = \int Phi Js(r,theta) dtheta dphi /
    !           \int Js(r,theta) dtheta dphi
    !----------------------------------------------------------
    call compute_fluxsurf_average(H=pthis%Phi_NrNthPNphi_Cworld, &
      geom=geom,H_Sflux_avg=Phi_FSavg_diag)


    !**********************************************
    !*** computation of the number of electrons ***
    !***  and the number of ions                ***
    !**********************************************
    call compute_nbparticles(geom,init_prof,rho, &
      pthis%Phi_NrNthPNphi_Cworld,Phi00_diag, &
      nbelectrons_diag,nbions_diag)

    do tid = 1, Nbthread
       call FTR_DEALLOC(Romp1_1Nrp1_1Nthetap1(tid)%val)
       call FTR_DEALLOC(Romp2_1Nrp1_1Nthetap1(tid)%val)
       call FTR_DEALLOC(Comp1_1Nrp1_1Ntheta(tid)%val)
       call FTR_DEALLOC(Comp2_1Nrp1_1Ntheta(tid)%val)
    enddo
    
    do tid = 0, MAX(Nmu,Nbthread)
       call FTR_DEALLOC(Romp10_1Nrp1_1Nthetap1(tid)%val)
    enddo
    
    call FTR_DEALLOC(RRecvBufMPI_A)
    
    call FTR_DEALLOC(Rarray1_NrNthetaNphi)

#ifdef TIMER
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tcmp(4))
    tdeb = tfin
    if (pglobal_id .eq. 0) then
      write(6,'(I4,A,3F13.6,A,4F13.6,A,F13.6)') pglobal_id, &
        " Temps poisson3A, tcomm ", tcmm(1:3), &
        " tcomp ",tcmp(1:4)," tbarrier ",tbar(1)
    end if
    call clck_ldiff(tth1,tfin,tthd)
    if (pglobal_id .eq. 0) then
      write(6,'(I4,A,F13.6)') pglobal_id, &
        " Temps poisson3B,  tcomp ",tthd
    end if
#endif

    if (stopit) then
       print*, 'solve_poisson : problem in Poisson solving'
       stop
    end if
#endif 
    ! HAS_QN_V3
  end subroutine solve_poisson_FFT1DMPI


  !-------------------------------------------------------
  ! Choice of the QN solver
  !-------------------------------------------------------
  subroutine solve_poisson(pthis,geom,init_prof, &
    init_magnet,init_curr,J0,f,Sfmu_eq)
    use globals, only : solve_QN, iter_glob, QN_version, distrib3D 
    use gyroaverage_class
    use clock_module

    type(poisson)      , intent(inout) :: pthis
    type(geometry)     , intent(in)    :: geom
    type(init_profile) , intent(in)    :: init_prof
    type(init_magnetic), intent(in)    :: init_magnet
    type(init_current) , intent(in)    :: init_curr
    type(J0operator)   , intent(inout) :: J0
    type(fdistribu5d)  , intent(inout) :: f
    real(RKIND), &
      dimension(:,:,:) , pointer       :: Sfmu_eq

!R3 #include "code_instrumentation/r3_info.h" !R3

!R3 call r3_info_begin (r3_info_index_0, 'CPU_poisson')      !R3
    call Write_dyn_balise_open('CPU_poisson') !R4

#ifdef TIMER
    call ppbarrier()
#endif 
    call clck_time(bclock_poisson)
    if (solve_QN) then
      select case(QN_version)                          
      case (1)                                         
        call solve_poisson_FFT2Dhybrid(pthis,geom, &
          init_prof,init_magnet,init_curr,J0,f,Sfmu_eq)
      case (2)
        call solve_poisson_FFT1Dhybrid(pthis,geom, &
          init_prof,init_magnet,init_curr,J0,f,Sfmu_eq)
      case (3)
        call solve_poisson_FFT1DMPI(pthis,geom, &
          init_prof,init_magnet,init_curr,J0,f,Sfmu_eq)
      end select
    else
      if (distrib3D) then
        pthis%Phi_NrNthPNphi_Cworld = 0._RKIND
        pthis%Phi_NrNthPNphi_Cmu    = 0._RKIND
      else
        pthis%Phi_NrNthPNphi_Cworld = 0._RKIND
      end if
    end if

    call clck_time(eclock_poisson)  
    if (iter_glob .eq. -1) then
      call clck_diff(bclock_poisson,eclock_poisson, &
        global_time_init)
    else
      call clck_diff(bclock_poisson,eclock_poisson, &
        global_time_poisson)
    end if
    call Write_dyn_balise_close('CPU_poisson') !R4
!R3 call r3_info_end (r3_info_index_0) !R3
  end subroutine solve_poisson


  !---------------------------------------------------------- 
  ! Computes the number of electrons and ions as:
  !  . nbelectrons = nbions_eq + 
  !         \int delta_ne(r,theta,phi) Js dr dtheta dphi
  !     with 
  !       --> nbions_eq = \int J0.feq dtau
  !        where dtau = Js Jv dr dtheta d3v
  !
  !       --> delta_ne(r,theta,phi) =  ne0(r)/(Te*Z02) * 
  !                               (Phi-lambda*<Phi>_{FS})
  !     where: . lambda = 1 if zonal flows 
  !                    = 0 otherwise
  !            . <Phi>_{FS} = \int Phi Js dtheta dphi /
  !                           \int Js(r,theta) dtheta dphi
  !
  !  . nions = \int J0.feq dtau computed as
  !          = \int J0.(f-feq) dtau + \int J0.feq dtau
  !          = \int QN_RHS*ne0 Js dr dtheta dphi + nbions_eq
  !---------------------------------------------------------- 
  subroutine compute_nbparticles(geom,init_prof,QN_RHS, &
    Phi,Phi_avg,Snbelectrons,Snbions)
    use OMPutils_module
    use geometry_class
    use init_profile_class
    use globals, only : nbions_eq, QN_coef_Phi00, &
      Rarray1_Nr, Nbthread, outputproc, diag_targ
    use globals, only : Z0
    use globals, only : mpi_comm_species, pInSpecies_id 
    use coord_system_class, only : jacobian_space
    use mpiMod    
    type(geometry)               , intent(in)  :: geom
    type(init_profile)           , intent(in)  :: init_prof
    real(RKIND), dimension(:,:,:), pointer     :: QN_RHS
    real(RKIND), dimension(:,:,:), pointer     :: Phi
    real(RKIND), &
             dimension(0:), optional, intent(in)  :: Phi_avg
    real(RKIND)                     , intent(out) :: Snbelectrons
    real(RKIND)                     , intent(out) :: Snbions

    !--> local variables 
    integer     :: status_spec, status
    integer     :: ierr, request_send, request_recv
    integer, dimension(MPI_STATUS_SIZE) :: status_tmp
    integer     :: ir, itheta, iphi, tid
    real(RKIND) :: jacob_space_tmp
    real(RKIND) :: coeff_intdr_tmp, coeff_intdtheta_tmp
    real(RKIND) :: coeff_intdphi_tmp, intdspace
    real(RKIND), dimension(:), pointer :: Phi_avg_tmp
    real(RKIND) :: nbspecies_tmp(2), nbspecies_loc(2)
    real(RKIND) :: nbions_loc(1:Nbthread), nbelec_loc(1:Nbthread)
    real(RKIND) :: Z02

    Z02 = Z0*Z0

    !*** initialisation of <Phi>_avg ***
    Phi_avg_tmp => Rarray1_Nr
    if ( present(Phi_avg) .and. (QN_coef_Phi00.eq.1) ) then
      do ir = 0,geom%Nr
        Phi_avg_tmp(ir) = Phi_avg(ir)
      end do
    else
      do ir = 0,geom%Nr
        Phi_avg_tmp(ir) = 0._RKIND
      end do
    end if

    !*** computation of the number of electrons ***
    !***  and the number of ions                ***
#ifdef _OPENMP
!$OMP PARALLEL private(tid,ir,itheta,iphi, &
!$OMP coeff_intdr_tmp,coeff_intdtheta_tmp, &
!$OMP coeff_intdphi_tmp,jacob_space_tmp,intdspace) &
!$OMP default(shared) 
!$OMP BARRIER
    tid = 1+omp_get_thread_num()
#else
    tid = 1
#endif
    nbelec_loc(tid) = 0._RKIND
    nbions_loc(tid) = 0._RKIND

    if (iphistart .ne. -1) then
!$OMP DO SCHEDULE(STATIC)
      do iphi = 0,geom%Nphi-1
        if (moments_mapphi(iphi) .eq. pInSpecies_id) then
          coeff_intdphi_tmp = geom%coeff_intdphi(iphi)
          do itheta = 0,geom%Ntheta-1
            coeff_intdtheta_tmp = geom%coeff_intdtheta(itheta)
            do ir = 0,geom%Nr
              coeff_intdr_tmp = geom%coeff_intdr(ir)
              jacob_space_tmp = jacobian_space(ir,itheta)
              intdspace       = jacob_space_tmp * &
                coeff_intdr_tmp*coeff_intdtheta_tmp * &
                coeff_intdphi_tmp
              !--> number of electrons
              nbelec_loc(tid) = nbelec_loc(tid) + &
                init_prof%ne0(ir)/(init_prof%Te0(ir)*Z02) * &
                (Phi(ir,itheta,iphi)-Phi_avg_tmp(ir)) * &
                intdspace
              !--> number of ions ( using the fact that
              !-->  rhoi = ne0*\int (J0f-J0.feq) Jv d3v )
              nbions_loc(tid)  = nbions_loc(tid) + &
                QN_RHS(ir,itheta,iphi)*init_prof%ne0(ir) * &
                intdspace
            end do
          end do
        end if
      end do
!$OMP END DO
    end if
!$OMP BARRIER
!$OMP END PARALLEL

    nbspecies_loc(1) = SUM(nbelec_loc(1:Nbthread))
    nbspecies_loc(2) = SUM(nbions_loc(1:Nbthread))
    nbspecies_tmp(1) = 0._RKIND
    nbspecies_tmp(2) = 0._RKIND
    call MY_Reduce(nbspecies_loc,nbspecies_tmp,2, &
      MPI_REAL8,MPI_SUM,diag_targ(1),mpi_comm_species,ierr)
    nbspecies_loc(1:2) = nbspecies_tmp(1:2)

    ! processor with id 'diag_targ(1)' send the computed
    ! densities to processor with id 'outputproc'.
    if (pInSpecies_id .eq. outputproc) &
         call MPI_IRECV(nbspecies_tmp,2,MPI_REAL8,diag_targ(1), &
         5000,mpi_comm_species,request_recv,ierr)
    if (pInSpecies_id .eq. diag_targ(1)) &
         call  MPI_ISEND(nbspecies_loc,2,MPI_REAL8,outputproc, &
         5000,mpi_comm_species,request_send,ierr)
    if (pInSpecies_id .eq. outputproc) &
         call MPI_wait(request_recv,status_tmp, ierr) 
    if (pInSpecies_id .eq. diag_targ(1)) &
         call MPI_wait(request_send,status_tmp, ierr) 
    status_spec  = 0
    Snbelectrons = nbions_eq + nbspecies_tmp(1)
    Snbions      = nbions_eq + nbspecies_tmp(2)
    if (Snbelectrons.ne.Snbelectrons .or. Snbions.ne.Snbions) then
       if (pInSpecies_id.eq.outputproc) then
          print *,' NAN detected in compute_nbparticles', &
            'so STOP the run !'
          status_spec = 1
       end if
    end if
    call MY_Allreduce(status_spec,status,1, &
      MPI_INTEGER,MPI_SUM,MPI_COMM_WORLD,ierr)
    if (status.gt.1) then
      call ppexit()
      call exit(-1)
    end if
  end subroutine compute_nbparticles
end module poisson_class

