!**************************************************************
!  Copyright Euratom-CEA
!  Authors : 
!     Virginie Grandgirard (virginie.grandgirard@cea.fr)
!     Chantal Passeron (chantal.passeron@cea.fr)
!     Guillaume Latu (guillaume.latu@cea.fr)
!     Xavier Garbet (xavier.garbet@cea.fr)
!     Philippe Ghendrih (philippe.ghendrih@cea.fr)
!     Yanick Sarazin (yanick.sarazin@cea.fr)
!  
!  This code GYSELA (for GYrokinetic SEmi-LAgrangian) 
!  is a 5D gyrokinetic global full-f code for simulating 
!  the plasma turbulence in a tokamak.
!  
!  This software is governed by the CeCILL-B license 
!  under French law and abiding by the rules of distribution 
!  of free software.  You can  use, modify and redistribute 
!  the software under the terms of the CeCILL-B license as 
!  circulated by CEA, CNRS and INRIA at the following URL
!  "http://www.cecill.info". 
!**************************************************************

#include "config-F.inc"

!----------------------------------------
! file : read_data.F90
! date : 16/09/2005
! used for the reading of :
! - mesh data
! - the equilibrium data
! - restart file (if in restart mode)
!----------------------------------------

!------------------------------------
! parallel data reading
!------------------------------------
subroutine parallel_data
  use prec_const
  use globals
  use MPIutils_module
  use mpiMod
  use Io_Unit_module, only: IU_open, IU_close
  
  implicit none
  
  integer:: io_unit
  integer            :: ierr
  integer            :: position
  integer, parameter :: lbuffer = 1024
  character*1        :: buffer(lbuffer)

  if (pglobal_id.eq.outputproc) then    
    call IU_open(UNIT=io_unit,FILE='DATA') 
    read(io_unit,PARALLEL)
    write(uout_res,PARALLEL)
    position = 0
    CALL MPI_PACK(Nspecies,1,MPI_INTEGER,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(Nbproc_r,1,MPI_INTEGER,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(Nbproc_theta,1,MPI_INTEGER,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(Nbproc_mu,1,MPI_INTEGER,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(bloc_phi,1,MPI_INTEGER,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(Nbthread,1,MPI_INTEGER,buffer,lbuffer,position, &
      MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(transpose4D,1,MPI_LOGICAL,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(large_platform,1,MPI_LOGICAL,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(distrib3D,1,MPI_LOGICAL,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_BCAST(buffer,lbuffer,MPI_PACKED,0,MPI_COMM_WORLD,ierr)
    call IU_close(io_unit)
  else
    CALL MPI_BCAST(buffer,lbuffer,MPI_PACKED,0,MPI_COMM_WORLD,ierr)
    position = 0
    CALL MPI_UNPACK(buffer,lbuffer,position,Nspecies,1, &
      MPI_INTEGER,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,Nbproc_r,1, &
      MPI_INTEGER,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,Nbproc_theta,1, &
      MPI_INTEGER,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,Nbproc_mu,1, &
      MPI_INTEGER,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,bloc_phi,1, &
      MPI_INTEGER,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,Nbthread,1, &
      MPI_INTEGER,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,transpose4D,1, &
      MPI_LOGICAL,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,large_platform, &
      1,MPI_LOGICAL,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,distrib3D,1, &
      MPI_LOGICAL,MPI_COMM_WORLD,ierr)
  end if

  ! For optimization of the OpenMP parallelization of 2D advection,
  ! bloc_phi must be at least equal to the number of threads.
  if (bloc_phi .lt. Nbthread) then
    bloc_phi = Nbthread
    if (pglobal_id.eq.0) then
      write(6,*) 'ATTENTION replacement : bloc_phi = ',bloc_phi
    end if
  end if

  ! For parallel diagnostics we need to know
  ! very early the maximum number of species MAXNSPECIES
  if (Nspecies .gt. MAXNSPECIES) then
    if (pglobal_id.eq.0) then
      write(6,*) 'ERROR Nspecies .gt. MAXNSPECIES'
    end if
    call ppexit()
  end if

  ! If the number of processors in theta direction is greater
  ! than 1, the transposition is required
  if (Nbproc_theta.gt.1) then
    if (.not.transpose4D) then
      transpose4D = .true.
      if (pglobal_id.eq.0) &
        write(6,*) 'ATTENTION transpose4D is forced to .true.'
    end if
  end if

  if (distrib3D) then
    transp_Bstar = .true.
  else
    transp_Bstar = .false.
  end if
end subroutine parallel_data


!------------------------------------
! mesh data reading
!------------------------------------
subroutine mesh_data
  use prec_const
  use globals
  use mpiMod
  use Io_Unit_module, only: IU_open, IU_close
  
  implicit none
  
  integer:: io_unit
  integer            :: ierr
  integer            :: position
  integer, parameter :: lbuffer = 1024
  character*1        :: buffer(lbuffer)

  if (pglobal_id.eq.outputproc) then
    call IU_open(UNIT=io_unit,FILE='DATA') 
    read(io_unit,MESH)
    write(uout_res,MESH)
    position = 0
    CALL MPI_PACK(CylindricalGeometry,1,MPI_LOGICAL,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(Nr,1,MPI_INTEGER,buffer,lbuffer,position, &
      MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(Ntheta,1,MPI_INTEGER,buffer,lbuffer,position, &
      MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(Nphi,1,MPI_INTEGER,buffer,lbuffer,position, &
      MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(Nvpar,1,MPI_INTEGER,buffer,lbuffer,position, &
      MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(Nmu,1,MPI_INTEGER,buffer,lbuffer,position, &
      MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(a,1,MPI_REAL8,buffer,lbuffer,position, &
      MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(rhomin,1,MPI_REAL8,buffer,lbuffer,position, &
      MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(rhomax,1,MPI_REAL8,buffer,lbuffer,position, &
      MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(Ltheta,1,MPI_REAL8,buffer,lbuffer,position, &
      MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(Lphi,1,MPI_REAL8,buffer,lbuffer,position, &
      MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(aspect_ratio,1,MPI_REAL8,buffer, &
      lbuffer,position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(nb_vth0,1,MPI_REAL8,buffer,lbuffer,position, &
      MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(Lmu,1,MPI_REAL8,buffer,lbuffer,position, &
      MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(mumin,1,MPI_REAL8,buffer,lbuffer,position, &
      MPI_COMM_WORLD,ierr)
    CALL MPI_BCAST(buffer,lbuffer,MPI_PACKED,0,MPI_COMM_WORLD,ierr)
    call IU_close(io_unit)
  else
    CALL MPI_BCAST(buffer,lbuffer,MPI_PACKED,0,MPI_COMM_WORLD,ierr)
    position = 0
    CALL MPI_UNPACK(buffer,lbuffer,position,CylindricalGeometry, &
      1,MPI_LOGICAL,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,Nr,1,MPI_INTEGER, &
      MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,Ntheta,1,MPI_INTEGER, &
      MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,Nphi,1,MPI_INTEGER, &
      MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,Nvpar,1,MPI_INTEGER, &
      MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,Nmu,1,MPI_INTEGER, &
      MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,a,1,MPI_REAL8, &
      MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,rhomin,1,MPI_REAL8, &
      MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,rhomax,1,MPI_REAL8, &
      MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,Ltheta,1,MPI_REAL8, &
      MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,Lphi,1,MPI_REAL8, &
      MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,aspect_ratio,1, &
      MPI_REAL8,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,nb_vth0,1, &
      MPI_REAL8,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,Lmu,1, &
      MPI_REAL8,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,mumin,1, &
      MPI_REAL8,MPI_COMM_WORLD,ierr)
  end if
end subroutine mesh_data


!----------------------------------------
! equilibrium data reading
!----------------------------------------
subroutine equil_data
  use prec_const
  use MPIutils_module, only : ppexit
  use globals
  use mpiMod
  use Io_Unit_module, only: IU_open, IU_close
  use wrap_alloc_module, only: write_param, write_expr
 
  implicit none
  
  integer:: io_unit
  integer            :: ierr
  integer            :: position
  integer, parameter :: lbuffer = 1024
  character*1        :: buffer(lbuffer)
  integer            :: ispec
  real(RKIND)        :: sum_all_species, max_frac_species

  if (pglobal_id.eq.outputproc) then
    call IU_open(UNIT=io_unit,FILE='DATA') 
    read(io_unit,EQUIL)
    write(uout_res,EQUIL)
    position = 0
    CALL MPI_PACK(adiabatic_electrons,1,MPI_LOGICAL,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(canonical_profile,1,MPI_INTEGER,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(A0,1,MPI_REAL8,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(Z0,1,MPI_REAL8,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(As,Nspecies,MPI_REAL8,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(Zs,Nspecies,MPI_REAL8,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(species_frac,Nspecies,MPI_REAL8,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(tau0,1,MPI_REAL8,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(read_ns0,Nspecies,MPI_LOGICAL,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(read_Ts0,Nspecies,MPI_LOGICAL,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(read_Te0,1,MPI_LOGICAL,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(read_q,1,MPI_LOGICAL,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(read_Vpar0,1,MPI_LOGICAL,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(profile_choice,Nspecies,MPI_INTEGER,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(rpeak,1,MPI_REAL8,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(kappa_ns0,Nspecies,MPI_REAL8,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(kappa_Ts0,Nspecies,MPI_REAL8,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(kappa_Te0,1,MPI_REAL8,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(deltar_ns0,Nspecies,MPI_REAL8,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(deltar_Ts0,Nspecies,MPI_REAL8,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(deltar_Te0,1,MPI_REAL8,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(q_profile,1,MPI_INTEGER,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(q_param1,1,MPI_REAL8,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(q_param2,1,MPI_REAL8,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(q_param3,1,MPI_REAL8,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(q_param4,1,MPI_REAL8,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(magnetic_drift,1,MPI_LOGICAL,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(perturb_amplitude,Nspecies,MPI_REAL8,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(perturb_choice,1,MPI_INTEGER,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(m,1,MPI_INTEGER,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(n,1,MPI_INTEGER,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(FFTpotential_filter,1,MPI_INTEGER,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(FFTpotential_filter_deltam,1,MPI_REAL8,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(TF_ripple,1,MPI_LOGICAL,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(delta_ripple,1,MPI_REAL8,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(N_ripple,1,MPI_INTEGER,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_BCAST(buffer,lbuffer,MPI_PACKED,0,MPI_COMM_WORLD,ierr)
    call IU_close(io_unit)
  else
    CALL MPI_BCAST(buffer,lbuffer,MPI_PACKED,0,MPI_COMM_WORLD,ierr)
    position = 0
    CALL MPI_UNPACK(buffer,lbuffer,position,adiabatic_electrons, &
      1,MPI_LOGICAL,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,canonical_profile, &
      1,MPI_INTEGER,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,A0, &
      1,MPI_REAL8,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,Z0, &
      1,MPI_REAL8,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,As, &
      Nspecies,MPI_REAL8,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,Zs, &
      Nspecies,MPI_REAL8,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,species_frac, &
      Nspecies,MPI_REAL8,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,tau0, &
      1,MPI_REAL8,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,read_ns0, &
      Nspecies,MPI_LOGICAL,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,read_Ts0, &
      Nspecies,MPI_LOGICAL,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,read_Te0, &
      1,MPI_LOGICAL,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,read_q, &
      1,MPI_LOGICAL,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,read_Vpar0, &
      1,MPI_LOGICAL,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,profile_choice, &
      Nspecies,MPI_INTEGER,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,rpeak, &
      1,MPI_REAL8,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,kappa_ns0, &
      Nspecies,MPI_REAL8,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,kappa_Ts0, &
      Nspecies,MPI_REAL8,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,kappa_Te0, &
      1,MPI_REAL8,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,deltar_ns0, &
      Nspecies,MPI_REAL8,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,deltar_Ts0, &
      Nspecies,MPI_REAL8,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,deltar_Te0, &
      1,MPI_REAL8,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,q_profile, &
      1,MPI_INTEGER,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,q_param1, &
      1,MPI_REAL8,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,q_param2, &
      1,MPI_REAL8,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,q_param3, &
      1,MPI_REAL8,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,q_param4, &
      1,MPI_REAL8,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,magnetic_drift, &
      1,MPI_LOGICAL,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,perturb_amplitude, &
      Nspecies,MPI_REAL8,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,perturb_choice,1, &
      MPI_INTEGER,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,m,1, &
      MPI_INTEGER,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,n,1, &
      MPI_INTEGER,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,FFTpotential_filter,1, &
      MPI_INTEGER,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,FFTpotential_filter_deltam, &
      1,MPI_REAL8,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,TF_ripple, &
      1,MPI_LOGICAL,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,delta_ripple, &
      1,MPI_REAL8,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,N_ripple, &
      1,MPI_INTEGER,MPI_COMM_WORLD,ierr)
  endif
  call MPI_BARRIER(MPI_COMM_WORLD,ierr)

  !*** Define the majoritary species                  ***
  !***  (i.e the one with the maximum 'species_frac') ***
  major_species    = 0
  max_frac_species = species_frac(0) 
  do ispec = 1,Nspecies
    if (species_frac(ispec).gt.max_frac_species) then
      major_species    = ispec
      max_frac_species = species_frac(ispec)
    end if
  end do

  !*** Test if adiabatic_electrons is equal to .true. ***
  if (.not.adiabatic_electrons) then
    if (pglobal_id.eq.0) &
      print*,' The case of kinetic electrons is not treated at the moment'
    call ppexit()
    call exit(-1)
  end if

  !*** Definition of the coefficient according to the case treated ***
  K_Bdrift = 0._RKIND
  if (q_profile .ne. 0) then
    if (magnetic_drift) then
      K_Bdrift = 1._RKIND
    end if
  else
    !-> if no magnetic shear, rbar = r
    if (canonical_profile.ne.0) then
      if (pglobal_id.eq.0) &
        print*,' The case q_profile = 0', &
          ' and canonical_profile not equal to 0 is not possible.'
      canonical_profile = 0
      if (pglobal_id.eq.0) &
        print*,'BECAREFUL: canonical_profile has been put equal to', &
        canonical_profile
    end if
    !-> if no magnetic shear, no B curvature can be taken 
    !->   into account
    if (magnetic_drift) then
      if (pglobal_id.eq.0) &
        print*,' The q_profile = 0', &
        ' and  case magnetic_drift = true is not possible'
      call ppexit()
      call exit(-1)
    end if
    !-> if no magnetic shear, the diagonal filter 
    !->  (FFTpotential_filter=1) is not possible
    if ((FFTpotential_filter.eq.1).or.(FFTpotential_filter.eq.2)) then
      if (pglobal_id.eq.0) &
        print*,' The case q_profile = 0', &
        ' and FFTpotential_filter = 1 is not possible.'
      call ppexit()
      call exit(-1)
    end if
  end if

  call write_param("m", m)
  call write_param("n", n)
  !*** Test coherence of (m,n) modes chosen ***
  if ((m.ge.Ntheta/2).or.(n.ge.Nphi/2)) then
    m = min(Ntheta/3,m)
    n = min(Nphi/3,n)
    call write_param("Ntheta", Ntheta)
    call write_param("Nphi", Nphi)
	call write_expr("m", "min(Ntheta/3,m)")
	call write_expr("n", "min(Nphi/3,n)")
    if (pglobal_id.eq.0) &
      print*,'BECAREFUL: initial m or n non consistent with', &
      ' Ntheta or Nphi', &
      ' changing to m =',m,'and n =',n
  end if

  !*** No canonical profile when initial parallel velocity is imposed
  if ((canonical_profile.ne.0).and.read_Vpar0) then
    if (pglobal_id.eq.0) &
      print*,' The case read_Vpar0 = true', &
      ' and canonical_profile not equal to 0 is not possible.'
    canonical_profile = 0
    if (pglobal_id.eq.0) &
      print*,'BECAREFUL: canonical_profile has been put equal to', &
      canonical_profile,' (local maxwellian)'
  end if
end subroutine equil_data


!----------------------------------------
! collisions data reading
!----------------------------------------
subroutine collisions_data
  use prec_const
  use globals
  use MPIutils_module, only : ppexit
  use mpiMod
  use Io_Unit_module, only: IU_open, IU_close
  
  implicit none

  integer:: io_unit
  integer            :: ierr
  integer            :: position
  integer, parameter :: lbuffer = 1024
  character*1        :: buffer(lbuffer)

  if (pglobal_id.eq.outputproc) then
    call IU_open(UNIT=io_unit,FILE='DATA')
    read(io_unit,COLLISION_OPERATOR)
    write(uout_res,COLLISION_OPERATOR)
    position = 0
    CALL MPI_PACK(collisions,1,MPI_LOGICAL,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(version_coll,1,MPI_INTEGER,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(energy_transfer,1,MPI_LOGICAL,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(moment_transfer,1,MPI_LOGICAL,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(nustar,1,MPI_REAL8,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(coll_FMrefresh,1,MPI_INTEGER,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(coll_nbstep,1,MPI_INTEGER,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_BCAST(buffer,lbuffer,MPI_PACKED,0,MPI_COMM_WORLD,ierr)
    call IU_close(io_unit)
  else
    CALL MPI_BCAST(buffer,lbuffer,MPI_PACKED,0,MPI_COMM_WORLD,ierr)
    position = 0
    CALL MPI_UNPACK(buffer,lbuffer,position,collisions, &
      1,MPI_LOGICAL,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,version_coll, &
      1,MPI_INTEGER,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,energy_transfer, &
      1,MPI_LOGICAL,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,moment_transfer, &
      1,MPI_LOGICAL,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,nustar, &
      1,MPI_REAL8,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,coll_FMrefresh, &
      1,MPI_INTEGER,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,coll_nbstep, &
      1,MPI_INTEGER,MPI_COMM_WORLD,ierr)
  end if
  call MPI_BARRIER(MPI_COMM_WORLD,ierr)

  !*** if "nustar" = 0, then "collisions" is set to false ***
  if (nustar.eq.0._RKIND) then
    collisions = .false.
  end if
  !*** Verification of the coherence in the collision ***
  !***  operator version                              ***
  if ((version_coll.eq.1).and.(coll_FMrefresh.ne.0)) then
    coll_FMrefresh = 0
    if (pglobal_id.eq.0) &
      print*,"The variable 'coll_FMrefresh' has been forced to 0"
  end if
  if (version_coll.gt.3) then
    if (pglobal_id.eq.0) then
      print*,'*******************************', &
        '************************'
      print*,'******  version_coll = ', version_coll, &
        '             ******'
      print*,'******     is not yet implemented', &
        '                ******' 
      print*,'******   change integer "version_coll".', &
        '          ******'
      print*,'******          CODE IS HALTED', &
        '                   ******'
      print*,'*******************************', &
        '************************'
    end if
    call ppexit()
    call exit(-1)
  end if

  !*** verification of the coherence of the energy ***
  !***  transfer operator                          ***
  if ( (energy_transfer).or.(moment_transfer) ) then
    if (.not.collisions) then
      if (pglobal_id.eq.0) &
        print*,'The species transfer operators are not possible ', & 
        'without collisions'
      call ppexit()
      call exit(-1)
    end if
    if (Nspecies.ne.2) then
      if (pglobal_id.eq.0) then
        print*,'The species transfer operators are only applicable ', &
          'for two species and not possible with ',Nspecies, ' species'
        print*,' ===> The logical variable energy_transfer is forced ', &
          'to .false.'
        print*,' ===> The logical variable moment_transfer is forced ', &
          'to .false.'
      end if
      energy_transfer = .false.
      moment_transfer = .false.
    end if
    if (asynchrone_writing) then
      if (pglobal_id.eq.0) then
        print*,'The asynchron writting is not treated at the moment ', & 
        'with the species transfer operators'
        print*,'====> The logical variable asynchrone_writing is forced ', &
          'to .false.'
      end if
      asynchrone_writing = .false.
    end if
  end if
end subroutine collisions_data


!----------------------------------------
! dissipation data reading 
!----------------------------------------
subroutine buffer_region_data
  use prec_const
  use globals
  use mpiMod
  use Io_Unit_module, only: IU_open, IU_close
  
  implicit none
  
  integer:: io_unit
  integer            :: ierr, ispec
  integer            :: position
  integer, parameter :: lbuffer = 1024
  character*1        :: buffer(lbuffer)
  
  if (pglobal_id.eq.outputproc) then
    call IU_open(UNIT=io_unit,FILE='DATA') 
    read(io_unit,BUFFER_REGION)
    write(uout_res,BUFFER_REGION)
    position = 0
    CALL MPI_PACK(coef_krook,1,MPI_REAL8,buffer,lbuffer,position, &
      MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(coef_diffr,Nspecies,MPI_REAL8,buffer,lbuffer,position, &
      MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(coef_diffth,Nspecies,MPI_REAL8,buffer,lbuffer,position, &
      MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(coef_nustar,1,MPI_REAL8,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(buffer_asymmetry,1,MPI_LOGICAL,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(buffer_location,1,MPI_REAL8,buffer,lbuffer,position, &
      MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(buffer_stiffness,1,MPI_REAL8,buffer,lbuffer,position, &
      MPI_COMM_WORLD,ierr)
    CALL MPI_BCAST(buffer,lbuffer,MPI_PACKED,0,MPI_COMM_WORLD,ierr)
    call IU_close(io_unit)
  else
    CALL MPI_BCAST(buffer,lbuffer,MPI_PACKED,0,MPI_COMM_WORLD,ierr)
    position = 0
    CALL MPI_UNPACK(buffer,lbuffer,position,coef_krook, &
      1,MPI_REAL8,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,coef_diffr, &
      Nspecies,MPI_REAL8,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,coef_diffth, &
      Nspecies,MPI_REAL8,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,coef_nustar, &
      1,MPI_REAL8,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,buffer_asymmetry, &
      1,MPI_LOGICAL,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,buffer_location, &
      1,MPI_REAL8,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,buffer_stiffness, &
      1,MPI_REAL8,MPI_COMM_WORLD,ierr)
  end if
  call MPI_BARRIER(MPI_COMM_WORLD,ierr)

  !-> if transpose4D='false', poloidal diffusion cannot be treated
  if (.not.transpose4D) then
    do ispec = 0,Nspecies-1
      if (coef_diffth(ispec).ne.0._RKIND) then
        coef_diffth(ispec) = 0._RKIND
        if (pglobal_id.eq.0) then
          print*,' The case transpose4D = false', &
            ' and coef_diffth > 0 is not possible'
          print*,'BE CAREFUL: coef_diffth has been put equal to', &
            coef_diffth(ispec), ' for species ', ispec
        end if
      end if
    end do
  end if
end subroutine buffer_region_data


!----------------------------------------
! reading of the source data
! in the case of flux-driven simulations
!----------------------------------------
subroutine source_data
  use prec_const
  use globals, only : pglobal_id, Nspecies, &
    major_species, outputproc, uout_res, SOURCE, &
    atleast1_source, atleast1_fp_srce, &
    flux_driven, heat_source, heat_version, &
    Zs, Sce0, TS, rhoS, LS, &
    dens_source, Sce_dens, TS_dens, rhoS_dens, LS_dens, &
    mom_source, Sce_mom, TS_mom, rhoS_mom, LS_mom, &
    vor_source, Sce_vor, TS_vor, rhoS_vor, LS_vor, &
    fp_source, fast_part, Sce_fp, Sce_fpplus, TS_fp, &
    v0_fp, rhoS_fp, LS_fp, alpha_trapp, rflux_driven
  use MPIutils_module
  use mpiMod
  use Io_Unit_module, only: IU_open, IU_close
  
  implicit none

  integer:: io_unit
  integer            :: ierr
  integer            :: position
  integer, parameter :: lbuffer = 1024
  character*1        :: buffer(lbuffer)

  integer     :: ispec
  integer     :: ierr_srce, impur_species, ispec1, ispec2
  logical     :: atleast1_dens_srce
  real(RKIND) :: Sce_dens_major

  if (pglobal_id.eq.outputproc) then
    call IU_open(UNIT=io_unit,FILE='DATA')
    read(io_unit,SOURCE)
    write(uout_res,SOURCE)
    position = 0
    CALL MPI_PACK(flux_driven,1,MPI_LOGICAL,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(heat_source,Nspecies,MPI_LOGICAL,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(heat_version,Nspecies,MPI_INTEGER,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(Sce0,Nspecies,MPI_REAL8,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(TS,Nspecies,MPI_REAL8,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(rhoS,Nspecies,MPI_REAL8,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(LS,Nspecies,MPI_REAL8,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(dens_source,Nspecies,MPI_LOGICAL,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(Sce_dens,Nspecies,MPI_REAL8,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(TS_dens,Nspecies,MPI_REAL8,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(rhoS_dens,Nspecies,MPI_REAL8,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(LS_dens,Nspecies,MPI_REAL8,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(mom_source,Nspecies,MPI_LOGICAL,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(Sce_mom,Nspecies,MPI_REAL8,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(TS_mom,Nspecies,MPI_REAL8,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(rhoS_mom,Nspecies,MPI_REAL8,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(LS_mom,Nspecies,MPI_REAL8,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(vor_source,Nspecies,MPI_LOGICAL,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(Sce_vor,Nspecies,MPI_REAL8,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(TS_vor,Nspecies,MPI_REAL8,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(rhoS_vor,Nspecies,MPI_REAL8,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(LS_vor,Nspecies,MPI_REAL8,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(fp_source,Nspecies,MPI_LOGICAL,buffer,lbuffer, &
       position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(fast_part,Nspecies,MPI_LOGICAL,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(Sce_fp,Nspecies,MPI_REAL8,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(Sce_fpplus,Nspecies,MPI_REAL8,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(TS_fp,Nspecies,MPI_REAL8,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(v0_fp,Nspecies,MPI_REAL8,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(rhoS_fp,Nspecies,MPI_REAL8,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(LS_fp,Nspecies,MPI_REAL8,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(alpha_trapp,Nspecies,MPI_REAL8,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_BCAST(buffer,lbuffer,MPI_PACKED,0,MPI_COMM_WORLD,ierr)
    call IU_close(io_unit)
  else
    CALL MPI_BCAST(buffer,lbuffer,MPI_PACKED,0,MPI_COMM_WORLD,ierr)
    position = 0
    CALL MPI_UNPACK(buffer,lbuffer,position,flux_driven, &
      1,MPI_LOGICAL,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,heat_source, &
      Nspecies,MPI_LOGICAL,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,heat_version, &
      Nspecies,MPI_INTEGER,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,Sce0, &
      Nspecies,MPI_REAL8,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,TS, &
      Nspecies,MPI_REAL8,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,rhoS, &
      Nspecies,MPI_REAL8,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,LS, &
      Nspecies,MPI_REAL8,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,dens_source, &
      Nspecies,MPI_LOGICAL,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,Sce_dens, &
      Nspecies,MPI_REAL8,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,TS_dens, &
      Nspecies,MPI_REAL8,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,rhoS_dens, &
      Nspecies,MPI_REAL8,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,LS_dens, &
      Nspecies,MPI_REAL8,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,mom_source, &
      Nspecies,MPI_LOGICAL,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,Sce_mom, &
      Nspecies,MPI_REAL8,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,TS_mom, &
      Nspecies,MPI_REAL8,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,rhoS_mom, &
      Nspecies,MPI_REAL8,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,LS_mom, &
      Nspecies,MPI_REAL8,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,vor_source, &
      Nspecies,MPI_LOGICAL,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,Sce_vor, &
      Nspecies,MPI_REAL8,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,TS_vor, &
      Nspecies,MPI_REAL8,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,rhoS_vor, &
      Nspecies,MPI_REAL8,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,LS_vor, &
      Nspecies,MPI_REAL8,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,fp_source, &
      Nspecies,MPI_LOGICAL,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,fast_part, &
      Nspecies,MPI_LOGICAL,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,Sce_fp, &
      Nspecies,MPI_REAL8,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,Sce_fpplus, &
      Nspecies,MPI_REAL8,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,TS_fp, &
      Nspecies,MPI_REAL8,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,v0_fp, &
      Nspecies,MPI_REAL8,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,rhoS_fp, &
      Nspecies,MPI_REAL8,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,LS_fp, &
      Nspecies,MPI_REAL8,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,alpha_trapp, &
      Nspecies,MPI_REAL8,MPI_COMM_WORLD,ierr)
  end if
  call MPI_BARRIER(MPI_COMM_WORLD,ierr)

  !*** Check if there is at least one source ***
  atleast1_source    = .false.
  atleast1_fp_srce   = .false.
  atleast1_dens_srce = .false.
  do ispec = 0,Nspecies-1
    if ( heat_source(ispec) .or. dens_source(ispec) .or. &
      mom_source(ispec) .or. vor_source(ispec) .or. &
      fp_source(ispec) ) &
    atleast1_source = .true.
    if (fp_source(ispec)) then
      atleast1_fp_srce = .true.
      fast_part(ispec) = .false.
    end if
    if (dens_source(ispec)) then
      atleast1_dens_srce = .true.
    end if
  end do

  !**** Definition of related parameters
  if (flux_driven) rflux_driven = 1.0_RKIND
  do ispec = 0,Nspecies-1
    if (.not.heat_source(ispec)) then
      Sce0(ispec)         = 0._RKIND
      TS(ispec)           = 1._RKIND
      rhoS(ispec)         = 0.5_RKIND
      LS(ispec)           = 1._RKIND
    end if
    if (.not.dens_source(ispec)) then
      Sce_dens(ispec)  = 0._RKIND
      TS_dens(ispec)   = 1._RKIND
      rhoS_dens(ispec) = 0.5_RKIND
      LS_dens(ispec)   = 1._RKIND
    end if
    if (.not.mom_source(ispec)) then
      Sce_mom(ispec)  = 0._RKIND
      TS_mom(ispec)   = 1._RKIND
      rhoS_mom(ispec) = 0.5_RKIND
      LS_mom(ispec)   = 1._RKIND
    end if
    if (.not.vor_source(ispec)) then
      Sce_vor(ispec)  = 0._RKIND
      TS_vor(ispec)   = 1._RKIND
      rhoS_vor(ispec) = 0.5_RKIND
      LS_vor(ispec)   = 1._RKIND
    end if
    if (.not.fp_source(ispec)) then
      Sce_fp(ispec)      = 0._RKIND
      Sce_fpplus(ispec)  = 0.5_RKIND
      TS_fp(ispec)       = 1._RKIND
      rhoS_fp(ispec)     = 0.5_RKIND
      LS_fp(ispec)       = 1._RKIND
      v0_fp(ispec)       = 0._RKIND
      alpha_trapp(ispec) = 0._RKIND
    end if
  end do

  !*** The source of density is actually only ***
  !*** designed as a source of impurity which ***
  !*** must be compensated by a negative      ***
  !*** source of the majority species such    ***
  !*** that:                                  ***
  !***    sum_s Zs \int S_ns d3v = 0          ***
  !***                                        ***
  !*** Rk: Actually this option is only       ***
  !***     available for 2 species, i.e       ***
  !***     the major species + one impurity   ***
  !*** Therefore, considering s as the major  ***
  !*** species and s' as the impurity:        ***
  !***  1) The shape and temperature of both  ***
  !***     density sources must be the same   ***
  !***  2) The source intensities are linked  ***
  !***     by S0_ns = -(Zs'/Zs) S0_ns'        ***
  ierr_srce = 0
  if (atleast1_dens_srce) then
    if (Nspecies.gt.2) then
      ierr_srce = 1
      if (pglobal_id.eq.0) then
        print*,'ERROR: The case of more than 2', &
          'species is actually not treated when ', &
          'a source of density is injected'
      end if
    else
      impur_species = 1
      if (major_species==1) impur_species = 0
      ispec1 = major_species
      ispec2 = impur_species
      if ( (TS_dens(ispec1).ne.TS_dens(ispec2)) .or. &
        (LS_dens(ispec1).ne.LS_dens(ispec2)) .or. &
        (rhoS_dens(ispec1).ne.rhoS_dens(ispec2))) then
        if (pglobal_id.eq.0) then
          print*,'ERROR: Ts_dens, LS_dens and rhoS_dens ', &
            'must be the same for both species'
        end if
        ierr_srce = 1
      else
        !-> rk: hat_Zs(impur)/hat_Zs(major) = Zs(impur)/Zs(major)
        Sce_dens_major = -Zs(ispec2)/Zs(ispec1) * &
          Sce_dens(ispec2) 
        if (Sce_dens(ispec1).ne.Sce_dens_major) then
          if (pglobal_id.eq.0) then 
            print*,'WARNING: The value Sce_dens for ', &
              'the major species which was equal to ', &
              Sce_dens(ispec1), ' has been replaced by ', &
              Sce_dens_major, 'to ensure Sce_dens(major) = ', &
              '-Zimpur/Zmajor Sce_dens(impur)'
          end if
          Sce_dens(ispec1) = Sce_dens_major
        end if
      end if
    end if
  end if

  if (ierr_srce==1) then
    call ppexit()
    call exit(-1)
  end if
end subroutine source_data

!----------------------------------------
! Initial fast particles data reading
!----------------------------------------
subroutine fast_particles_data
  use prec_const
  use globals
  use mpiMod
  use Io_Unit_module, only: IU_open, IU_close
  
  implicit none

  integer:: io_unit
  integer            :: ierr
  integer            :: position
  integer, parameter :: lbuffer = 1024
  character*1        :: buffer(lbuffer)
  integer            :: ierr_namelist

  integer :: ispec

  if (pglobal_id.eq.outputproc) then
    call IU_open(UNIT=io_unit,FILE='DATA')
    read(io_unit,FAST_PARTICLES,iostat=ierr_namelist)
    call IU_close(io_unit)
  end if
  call MPI_BCAST(ierr_namelist,1,MPI_INTEGER, &
    outputproc,MPI_COMM_WORLD,ierr)

  if (ierr_namelist.eq.0) then
    if (pglobal_id.eq.outputproc) then
      call IU_open(UNIT=io_unit,FILE='DATA')
      read(io_unit,FAST_PARTICLES)
      write(uout_res,FAST_PARTICLES)
      position = 0
      CALL MPI_PACK(bump_on_tail,Nspecies,MPI_LOGICAL, &
        buffer,lbuffer,position,MPI_COMM_WORLD,ierr)
      CALL MPI_PACK(fp_rshape,Nspecies,MPI_LOGICAL, &
        buffer,lbuffer,position,MPI_COMM_WORLD,ierr)
      CALL MPI_PACK(fp_lambda,Nspecies,MPI_LOGICAL, &
        buffer,lbuffer,position,MPI_COMM_WORLD,ierr)
      CALL MPI_PACK(fp_mom,Nspecies,MPI_LOGICAL,buffer,lbuffer, &
        position,MPI_COMM_WORLD,ierr)
      CALL MPI_PACK(fp_energy,Nspecies,MPI_LOGICAL, &
        buffer,lbuffer,position,MPI_COMM_WORLD,ierr)
      CALL MPI_PACK(fp_momgauss,Nspecies,MPI_LOGICAL, &
        buffer,lbuffer,position,MPI_COMM_WORLD,ierr)
      CALL MPI_PACK(Eb_fp,Nspecies,MPI_REAL8,buffer,lbuffer, &
        position,MPI_COMM_WORLD,ierr)
      CALL MPI_PACK(Emin_fp,Nspecies,MPI_REAL8,buffer,lbuffer, &
        position,MPI_COMM_WORLD,ierr)
      CALL MPI_PACK(n_fp,Nspecies,MPI_REAL8,buffer,lbuffer, &
        position,MPI_COMM_WORLD,ierr)
      CALL MPI_PACK(dn_fp,Nspecies,MPI_REAL8,buffer,lbuffer, &
        position,MPI_COMM_WORLD,ierr)
      CALL MPI_PACK(E_critfp,Nspecies,MPI_REAL8,buffer,lbuffer, &
        position,MPI_COMM_WORLD,ierr)
      CALL MPI_PACK(rho_0fp,Nspecies,MPI_REAL8,buffer,lbuffer, &
        position,MPI_COMM_WORLD,ierr)
      CALL MPI_PACK(deltarhofp,Nspecies,MPI_REAL8,buffer,lbuffer, &
        position,MPI_COMM_WORLD,ierr)
      CALL MPI_PACK(Lambda_0fp,Nspecies,MPI_REAL8,buffer,lbuffer, &
        position,MPI_COMM_WORLD,ierr)
      CALL MPI_PACK(deltaLambdafp,Nspecies,MPI_REAL8, &
        buffer,lbuffer,position,MPI_COMM_WORLD,ierr)
      CALL MPI_PACK(psi_0fp,Nspecies,MPI_REAL8,buffer,lbuffer, &
        position,MPI_COMM_WORLD,ierr)
      CALL MPI_PACK(deltapsifp,Nspecies,MPI_REAL8,buffer,lbuffer, &
        position,MPI_COMM_WORLD,ierr)
      CALL MPI_PACK(zetabar,Nspecies,MPI_REAL8,buffer,lbuffer, &
        position,MPI_COMM_WORLD,ierr)
      CALL MPI_BCAST(buffer,lbuffer,MPI_PACKED,0, &
        MPI_COMM_WORLD,ierr)
      call IU_close(io_unit)
    else
      CALL MPI_BCAST(buffer,lbuffer,MPI_PACKED,0, &
        MPI_COMM_WORLD,ierr)
      position = 0
      CALL MPI_UNPACK(buffer,lbuffer,position,bump_on_tail, &
        Nspecies,MPI_LOGICAL,MPI_COMM_WORLD,ierr)
      CALL MPI_UNPACK(buffer,lbuffer,position,fp_rshape, &
        Nspecies,MPI_LOGICAL,MPI_COMM_WORLD,ierr)
      CALL MPI_UNPACK(buffer,lbuffer,position,fp_lambda, &
        Nspecies,MPI_LOGICAL,MPI_COMM_WORLD,ierr)
      CALL MPI_UNPACK(buffer,lbuffer,position,fp_mom, &
        Nspecies,MPI_LOGICAL,MPI_COMM_WORLD,ierr)
      CALL MPI_UNPACK(buffer,lbuffer,position,fp_energy, &
        Nspecies,MPI_LOGICAL,MPI_COMM_WORLD,ierr)
      CALL MPI_UNPACK(buffer,lbuffer,position,fp_momgauss, &
        Nspecies,MPI_LOGICAL,MPI_COMM_WORLD,ierr)
      CALL MPI_UNPACK(buffer,lbuffer,position,Eb_fp, &
        Nspecies,MPI_REAL8,MPI_COMM_WORLD,ierr)
      CALL MPI_UNPACK(buffer,lbuffer,position,Emin_fp, &
        Nspecies,MPI_REAL8,MPI_COMM_WORLD,ierr)
      CALL MPI_UNPACK(buffer,lbuffer,position,n_fp, &
        Nspecies,MPI_REAL8,MPI_COMM_WORLD,ierr)
      CALL MPI_UNPACK(buffer,lbuffer,position,dn_fp, &
        Nspecies,MPI_REAL8,MPI_COMM_WORLD,ierr)
      CALL MPI_UNPACK(buffer,lbuffer,position,E_critfp, &
        Nspecies,MPI_REAL8,MPI_COMM_WORLD,ierr)
      CALL MPI_UNPACK(buffer,lbuffer,position,rho_0fp, &
        Nspecies,MPI_REAL8,MPI_COMM_WORLD,ierr)
      CALL MPI_UNPACK(buffer,lbuffer,position,deltarhofp, &
        Nspecies,MPI_REAL8,MPI_COMM_WORLD,ierr)
      CALL MPI_UNPACK(buffer,lbuffer,position,Lambda_0fp, &
        Nspecies,MPI_REAL8,MPI_COMM_WORLD,ierr)
      CALL MPI_UNPACK(buffer,lbuffer,position,deltaLambdafp, &
        Nspecies,MPI_REAL8,MPI_COMM_WORLD,ierr)
      CALL MPI_UNPACK(buffer,lbuffer,position,psi_0fp, &
        Nspecies,MPI_REAL8,MPI_COMM_WORLD,ierr)
      CALL MPI_UNPACK(buffer,lbuffer,position,deltapsifp, &
        Nspecies,MPI_REAL8,MPI_COMM_WORLD,ierr)
      CALL MPI_UNPACK(buffer,lbuffer,position,zetabar, &
        Nspecies,MPI_REAL8,MPI_COMM_WORLD,ierr)
    end if
  else
    if (pglobal_id.eq.0) then
      print*,'WARNING : The input data file does not contains ', &
        'a FAST_PARTICLES namelist, so the values are ', &
        'initialized to the default values'
      print*, ' '
    end if
  end if
  call MPI_BARRIER(MPI_COMM_WORLD,ierr)

  !**** Definition of related parameters
  do ispec = 0,Nspecies-1
    if (fp_mom(ispec)) fp_rshape(ispec) = .false.
    if (.not.fast_part(ispec)) then
      bump_on_tail(ispec) = .true.
      fp_rshape(ispec)    = .false.
      n_fp(ispec)         = 0._RKIND
      dn_fp(ispec)        = 0._RKIND
      rho_0fp(ispec)      = 0.5_RKIND
      deltarhofp(ispec)   = 0.1_RKIND
      zetabar(ispec)      = 4._RKIND
    end if
  end do
end subroutine fast_particles_data

!----------------------------------------
! algorithm data reading
!----------------------------------------
subroutine algorithm_data
  use prec_const
  use globals
  use MPIutils_module, only : ppexit
  use mpiMod
  use Io_Unit_module, only: IU_open, IU_close
  
  implicit none
  
  integer:: io_unit
  integer            :: ierr
  integer            :: position
  integer, parameter :: lbuffer = 1024
  character*1        :: buffer(lbuffer)
  
  if (pglobal_id.eq.outputproc) then
    call IU_open(UNIT=io_unit,FILE='DATA') 
    read(io_unit,ALGORITHM)
    write(uout_res,ALGORITHM)
    position = 0
    CALL MPI_PACK(Vlasov_scheme,5,MPI_CHARACTER,buffer,lbuffer, &      
      position,MPI_COMM_WORLD,ierr)                        
    CALL MPI_PACK(limiter,1,MPI_INTEGER,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(deltat,1,MPI_REAL8,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(nbiter,1,MPI_INTEGER,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(dt_diag,1,MPI_REAL8,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_BCAST(buffer,lbuffer,MPI_PACKED,0,MPI_COMM_WORLD,ierr)
    call IU_close(io_unit)
  else
    CALL MPI_BCAST(buffer,lbuffer,MPI_PACKED,0,MPI_COMM_WORLD,ierr)
    position = 0
    CALL MPI_UNPACK(buffer,lbuffer,position,Vlasov_scheme,5, &       
      MPI_CHARACTER, MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,limiter, &
      1,MPI_INTEGER,MPI_COMM_WORLD,ierr) 
    CALL MPI_UNPACK(buffer,lbuffer,position,deltat, &
      1,MPI_REAL8,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,nbiter, &
      1,MPI_INTEGER,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,dt_diag, &
      1,MPI_REAL8,MPI_COMM_WORLD,ierr)
  end if
  call MPI_BARRIER(MPI_COMM_WORLD,ierr)

  !-> defined is the scheme is conservative or not
  if ( (Vlasov_scheme=="PPM2D").or.  &
    (Vlasov_scheme=="PPM1D").or.  &
    (Vlasov_scheme=="PSM1D").or.  &
    (Vlasov_scheme=="PSM2D") ) then
    conservative_scheme = .true.
  end if

  !->  The PSM scheme is not possible without transposition of 
  !    the distribution function
  if (conservative_scheme) then
    if (.not.transpose4D) then
      transpose4D = .true.
      if (pglobal_id.eq.0) &
        write(6,*) 'ATTENTION transpose4D is forced to .true.', &
        ' because conservative scheme'
    end if
    if (transp_Bstar) then
      transp_Bstar = .false.
      if (pglobal_id.eq.0) &
        write(6,*) 'ATTENTION transp_Bstar is forced to .false.', &
        ' because conservative scheme'
    end if

    if (distrib3D) then
      write(6,*) 'ERROR conservative scheme not allowed if the ', &
        '3D data are distibuted, i.e if distrib3D=.true.'
      call ppexit()
    end if
  end if

  !-> checking of 'dt_diag' which must be larger than 'deltat' 
  !-> and a multiple of 'deltat'
  call check_dt_diag('deltat',deltat,'dt_diag',dt_diag)
end subroutine algorithm_data


!----------------------------------------
! test data reading 
!----------------------------------------
subroutine test_data
  use prec_const
  use globals
  use MPIutils_module, only : ppexit
  use mpiMod
  use Io_Unit_module, only: IU_open, IU_close
  
  implicit none
  
  integer:: io_unit
  integer            :: ierr
  integer            :: position
  integer, parameter :: lbuffer = 1024
  character*1        :: buffer(lbuffer)
  
  if (pglobal_id.eq.outputproc) then
    call IU_open(UNIT=io_unit,FILE='DATA') 
    read(io_unit,TEST)
    write(uout_res,TEST)
    position = 0
    !--> Tests for general numerics
    CALL MPI_PACK(integral_vperp,1,MPI_LOGICAL,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(integration_scheme,1,MPI_INTEGER,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    !--> Tests for QN solver
    CALL MPI_PACK(solve_QN,1,MPI_LOGICAL,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(QN_version,1,MPI_INTEGER,buffer, &     
      lbuffer,position,MPI_COMM_WORLD,ierr)              
    CALL MPI_PACK(QN_comm_version,1,MPI_INTEGER,buffer, &
      lbuffer,position,MPI_COMM_WORLD,ierr)              
    CALL MPI_PACK(QN_coef_polarization,1,MPI_REAL8, &
      buffer,lbuffer,position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(QN_coef_Phi00,1,MPI_INTEGER,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(Phi00_BCrmin_Neumann,1,MPI_LOGICAL,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(QN_species_impact,Nspecies,MPI_REAL8,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    !--> Tests for Vlasov solver
    CALL MPI_PACK(adaptive_deltat,1,MPI_REAL8,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(advec2D_deltat,1,MPI_REAL8,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(reduced_deltat,1,MPI_REAL8,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(reduced_begin,1,MPI_REAL8,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(reduced_end,1,MPI_REAL8,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(hffiltertype,1,MPI_INTEGER,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(hffilterfreq,1,MPI_INTEGER,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(hffilterfref,1,MPI_INTEGER,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(deltaF_in_diff,1,MPI_LOGICAL,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(avoidtaylor,1,MPI_LOGICAL,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(deltaF_in_advec,1,MPI_LOGICAL,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(split_linear,1,MPI_LOGICAL,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(clampv,1,MPI_LOGICAL,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(cancel_nonlin,1,MPI_LOGICAL,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(diagf,1,MPI_LOGICAL,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    !--> Tests for RHS of Boltzmann equation
    CALL MPI_PACK(RHS_only,1,MPI_LOGICAL,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    !--> Tests for equilibrium definition
    CALL MPI_PACK(feq_choice,1,MPI_INTEGER,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(canonical_vpar0,1,MPI_INTEGER,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(hvpar_in_fperturb,1,MPI_LOGICAL,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    !--> Tests with simplified physics
    CALL MPI_PACK(gyroaverage,1,MPI_LOGICAL,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(BstareqB,1,MPI_LOGICAL,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(single_m,1,MPI_LOGICAL,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(single_n,1,MPI_LOGICAL,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(Rosenbluth_Hinton,1,MPI_LOGICAL,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    !--> Tests for computer sciences
    CALL MPI_PACK(memory_test,1,MPI_LOGICAL,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(asynchrone_writing,1,MPI_LOGICAL,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(ask_for_checksum,1,MPI_INTEGER,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(modulo_restart_iter,1,MPI_INTEGER,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(checkoverflow,1,MPI_LOGICAL,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(cmpr,1,MPI_INTEGER,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(transpose_version,1,MPI_INTEGER,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_BCAST(buffer,lbuffer,MPI_PACKED,0,MPI_COMM_WORLD,ierr)
    call IU_close(io_unit)
  else
    CALL MPI_BCAST(buffer,lbuffer,MPI_PACKED,0,MPI_COMM_WORLD,ierr)
    position = 0
    !--> Tests for general numerics
    CALL MPI_UNPACK(buffer,lbuffer,position,integral_vperp,1, &
      MPI_LOGICAL,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,integration_scheme,1, &
      MPI_INTEGER,MPI_COMM_WORLD,ierr)
    !--> Tests for QN solver
    CALL MPI_UNPACK(buffer,lbuffer,position,solve_QN, &
      1,MPI_LOGICAL,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position, &            
      QN_version,1,MPI_INTEGER,MPI_COMM_WORLD,ierr)       
    CALL MPI_UNPACK(buffer,lbuffer,position, &            
      QN_comm_version,1,MPI_INTEGER,MPI_COMM_WORLD,ierr)  
    CALL MPI_UNPACK(buffer,lbuffer,position, &
      QN_coef_polarization,1,MPI_REAL8,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,QN_coef_Phi00,1, &
      MPI_INTEGER,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,Phi00_BCrmin_Neumann, &
      1,MPI_LOGICAL,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,QN_species_impact, &
      Nspecies,MPI_REAL8,MPI_COMM_WORLD,ierr)
    !--> Tests for Vlasov solver
    CALL MPI_UNPACK(buffer,lbuffer,position,adaptive_deltat, &
      1,MPI_REAL8,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,advec2D_deltat, &
      1,MPI_REAL8,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,reduced_deltat, &
      1,MPI_REAL8,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,reduced_begin, &
      1,MPI_REAL8,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,reduced_end, &
      1,MPI_REAL8,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,hffiltertype, &
      1,MPI_INTEGER,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,hffilterfreq, &
      1,MPI_INTEGER,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,hffilterfref, &
       1,MPI_INTEGER,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,deltaF_in_diff, &
      1,MPI_LOGICAL,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,avoidtaylor, &
      1,MPI_LOGICAL,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,deltaF_in_advec, &
      1,MPI_LOGICAL,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,split_linear, &
      1,MPI_LOGICAL,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,clampv, &
      1,MPI_LOGICAL,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,cancel_nonlin, &
      1,MPI_LOGICAL,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,diagf, &
      1,MPI_LOGICAL,MPI_COMM_WORLD,ierr)
    !--> Tests for RHS of Boltzmann equation
    CALL MPI_UNPACK(buffer,lbuffer,position,RHS_only,1,MPI_LOGICAL,&
      MPI_COMM_WORLD,ierr)
    !--> Tests for equilibrium definition
    CALL MPI_UNPACK(buffer,lbuffer,position,feq_choice, &
      1,MPI_INTEGER,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,canonical_vpar0, &
      1,MPI_INTEGER,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,hvpar_in_fperturb, &
      1,MPI_LOGICAL,MPI_COMM_WORLD,ierr)
    !--> Tests with simplified physics
    CALL MPI_UNPACK(buffer,lbuffer,position,gyroaverage, &
      1,MPI_LOGICAL,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,BstareqB, &
      1,MPI_LOGICAL,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,single_m, &
      1,MPI_LOGICAL,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,single_n, &
      1,MPI_LOGICAL,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,Rosenbluth_Hinton, &
      1,MPI_LOGICAL,MPI_COMM_WORLD,ierr)
    !--> Tests for computer sciences
    CALL MPI_UNPACK(buffer,lbuffer,position,memory_test, &
      1,MPI_LOGICAL,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,asynchrone_writing, &
      1,MPI_LOGICAL,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,ask_for_checksum, &
      1,MPI_INTEGER,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,modulo_restart_iter, &
       1,MPI_INTEGER,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,checkoverflow, &
      1,MPI_LOGICAL,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,cmpr, &
      1,MPI_INTEGER,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,transpose_version, &
      1,MPI_INTEGER,MPI_COMM_WORLD,ierr)
  end if
  call MPI_BARRIER(MPI_COMM_WORLD,ierr)
  
  !*** force "perturb_choice" to case 3 if "single_n" is true  ***
  !***                                4 if "single_m" is true  ***
  !***                                1 if "single_m" and      ***
  !***                                    "single_m" are true  ***
  !***  => avoids bad loops on "n" and "m" in subroutine       ***
  !***    init_fperturb                                        ***
  if (single_n) then
    perturb_choice = 3
  end if
  if (single_m) then
    perturb_choice = 4
  end if  
  if (single_n.and.single_m) then
    perturb_choice = 1
  end if

  if (Nspecies.gt.1) then
    if (QN_version.lt.4) then
      if (pglobal_id.eq.0) then
        print*, 'The version ',QN_version, &
          ' of the poisson solver is not treated', &
          ' at the moment for 2 species'
      end if
      call ppexit()
      call exit(-1)
    end if
  end if

  if (distrib3D) then
    if (QN_version.ne.5) then
      if (pglobal_id.eq.0) then
        print*, 'The version ',QN_version, &
          ' of the poisson solver cannot be used', &
          ' when the 3D data are distributed'
      end if
      call ppexit()
      call exit(-1)
    end if
  else
    if (QN_version.eq.5) then
      if (pglobal_id.eq.0) then
        print*, 'The version ',QN_version, &
          ' of the poisson solver cannot be used', &
          ' when the 3D data are not distributed'
      end if
      call ppexit()
      call exit(-1)
    end if
  end if

  !*** force adpative_dt=.true. in the case of ***
  !***  conservative schemes                   ***
  if (conservative_scheme) then
    adaptive_deltat = .true.
  end if

  !*** force Dirichlet boundary conditions for Phi00      ***
  !***  in the case of QN_version>=4 because it is        ***
  !***  not possible to apply                             ***
  !***  \partial_r <Phi>_{FS} (rmin) = 0 without applying ***
  !***  \partial_r Phi(rmin) = \partial_r Phi00(rmin) = 0 ***
  if ((QN_version.ge.4) .and. (Phi00_BCrmin_Neumann) &
       .and. (rhomin .gt. .01_RKIND)) then
    Phi00_BCrmin_Neumann = .false.
    if (pglobal_id.eq.0) &
      write(6,*) 'ATTENTION Phi00_BCrmin_Neumann ', &
      'is forced to .false. because QN_version>=4 and rhomin', &
      'is greater than the value .01'
  end if

  !*** The splitting between linear and non-linear ***
  !***  advection is only available at the moment  ***
  !***  for the BSL algorithm                      ***
  if (split_linear) then
    if (RHS_only) then
      if (pglobal_id.eq.0) &
        print*,' RHS_only = .true. is not available ', &
        'with split_linear = .true.'
      call ppexit()
      call exit(-1)
    end if
    if (Vlasov_scheme.ne."BSL") then
      if (pglobal_id.eq.0) &
        print*,' split_linear = .true. is only available ', &
        'for BSL algorithm'
      call ppexit()
      call exit(-1)
    end if
  end if

  !*** The case 'avoidtaylor=.true' is only possible ***
  !***  for the case transpose4D = .true.            ***
  if ( (avoidtaylor) .and. (.not.transpose4D) ) then
    transpose4D = .true.
    if (pglobal_id.eq.0) &
      write(6,*) 'ATTENTION transpose4D has been forced ', &
      'to .true. because avoidtaylor is equal to .true.'
  end if
end subroutine test_data


!----------------------------------------
! output data reading 
!----------------------------------------
subroutine output_data
  use prec_const
  use globals
  use MPIutils_module, only : ppexit
  use mpiMod
  use Io_Unit_module, only: IU_open, IU_close
  
  implicit none
  
  integer:: io_unit
  integer            :: ierr
  integer            :: position
  integer, parameter :: lbuffer = 1024
  character*1        :: buffer(lbuffer)
  
  if (pglobal_id.eq.outputproc) then
    call IU_open(UNIT=io_unit,FILE='DATA') 
    read(io_unit,OUTPUT)
    write(uout_res,OUTPUT)
    position = 0
    CALL MPI_PACK(integration_CS,1,MPI_LOGICAL,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(diag_level,1,MPI_INTEGER,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(Phi3D_saving,1,MPI_LOGICAL,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(Moments3D_saving,1,MPI_LOGICAL,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(f5D_saving,1,MPI_LOGICAL,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(Phi3D_dt_diag,1,MPI_REAL8,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(Moments3D_dt_diag,1,MPI_REAL8,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(f5D_dt_diag,1,MPI_REAL8,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(Sp_saving,1,MPI_LOGICAL,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(Sp_dt_diag,1,MPI_REAL8,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(doubleSP_m,nb_fulldiag_max,MPI_INTEGER,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(doubleSP_n,nb_fulldiag_max,MPI_INTEGER,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(r_diags,N_rdiags_Spectr,MPI_INTEGER, &
      buffer,lbuffer,position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(rst_saving,1,MPI_LOGICAL,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(enable_dyn_trace,1,MPI_LOGICAL,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(iter_beg_write_trace,1,MPI_INTEGER,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_PACK(iter_end_write_trace,1,MPI_INTEGER,buffer,lbuffer, &
      position,MPI_COMM_WORLD,ierr)
    CALL MPI_BCAST(buffer,lbuffer,MPI_PACKED,0,MPI_COMM_WORLD,ierr)
    call IU_close(io_unit)
  else
    CALL MPI_BCAST(buffer,lbuffer,MPI_PACKED,0,MPI_COMM_WORLD,ierr)
    position = 0
    CALL MPI_UNPACK(buffer,lbuffer,position,integration_CS, &
      1,MPI_LOGICAL,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,diag_level, &
      1,MPI_INTEGER,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,Phi3D_saving, &
      1,MPI_LOGICAL,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,Moments3D_saving, &
      1,MPI_LOGICAL,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,f5D_saving, &
      1,MPI_LOGICAL,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,Phi3D_dt_diag, &
      1,MPI_REAL8,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,Moments3D_dt_diag, &
      1,MPI_REAL8,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,f5D_dt_diag, &
      1,MPI_REAL8,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,Sp_saving, &
      1,MPI_LOGICAL,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,Sp_dt_diag, &
      1,MPI_REAL8,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,doubleSP_m, &
      nb_fulldiag_max,MPI_INTEGER,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,doubleSP_n, &
      nb_fulldiag_max,MPI_INTEGER,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,r_diags, &
      N_rdiags_Spectr,MPI_INTEGER,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,rst_saving, &
      1,MPI_LOGICAL,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,enable_dyn_trace, &
      1,MPI_LOGICAL,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,iter_beg_write_trace,&
      1,MPI_INTEGER,MPI_COMM_WORLD,ierr)
    CALL MPI_UNPACK(buffer,lbuffer,position,iter_end_write_trace,&
      1,MPI_INTEGER,MPI_COMM_WORLD,ierr)
  end if
  call MPI_BARRIER(MPI_COMM_WORLD,ierr)

  !-> checking of the validity of the elapsed time between two savings
  call check_dt_diag('dt_diag',dt_diag,'Phi3D_dt_diag',Phi3D_dt_diag)
  call check_dt_diag('dt_diag',dt_diag,'Moments3D_dt_diag',Moments3D_dt_diag)
  call check_dt_diag('dt_diag',dt_diag,'f5D_dt_diag',f5D_dt_diag)
  call check_dt_diag('dt_diag',dt_diag,'Sp_dt_diag',Sp_dt_diag)
end subroutine output_data


!----------------------------------------
!  reading of the input file .dat
!  -> corresponding to the reading of all
!     previous namelist
!----------------------------------------
subroutine read_input()
  use globals
  use MPIutils_module, only : ppexit
  use Io_Unit_module, only: IU_open, IU_close
  
  implicit none

  !*** open output file results ***
  if (pglobal_id.eq.outputproc) &
    call IU_open(unit=uout_res,file=file_name_res,status='OLD', &
    position='APPEND',form='FORMATTED')

  call parallel_data
  call mesh_data
  call equil_data
  call collisions_data                                 
  call source_data                                     
  call fast_particles_data                                 
  call buffer_region_data
  call algorithm_data
  call test_data
  call output_data

  if (modulo_restart_iter == -1) &
    modulo_restart_iter = nbiter + 1

  !*** Verification of the number of points in the mesh ***
  call verif_mesh_number  

  !*** Verification of the compatibility between the number ***
  !*** of processors and the number of points               ***
  call verif_parallel_data

  !*** close output file results ***
  if (pglobal_id.eq.outputproc) &
    call IU_close(uout_res)
end subroutine read_input


!****************************************************************
!  UTILS USED FOR COMPATIBILIY DATA CHECKING
!****************************************************************
!-----------------------------------------------------
! Verification of the compatibility between the number
! of processors and the number of points
!-----------------------------------------------------
subroutine verif_parallel_data
  use globals
  use MPIutils_module, only : ppexit
  implicit none

  !*** tests on the number of processors in each direction ***
  !-> Verification of Nbproc_tot
  if (Nbproc_tot.ne.(Nspecies*Nbproc_r*Nbproc_theta*Nbproc_mu)) then
    if (pglobal_id.eq.0) then
      print*,' Nbproc_tot not equal to ', &
        'Nspecies*Nbproc_r*Nbproc_theta*Nbproc_mu'
      print*,' => Nbproc_tot   = ',Nbproc_tot
      print*,'    Nspecies     = ',Nspecies
      print*,'    Nbproc_r     = ',Nbproc_r
      print*,'    Nbproc_theta = ',Nbproc_theta
      print*,'    Nbproc_mu    = ',Nbproc_mu
    end if
    call ppexit()
    call exit(-1)
  end if

  !-> Verification of Nbproc_r
  if (Nbproc_r.gt.(Nr+1)) then
    if (pglobal_id.eq.0) then
      print*,' Nbproc_r greater than Nr+1 '
      print*,' => Nbproc_r = ',Nbproc_r
      print*,'    Nr       = ',Nr
    end if
    call ppexit()
    call exit(-1)
  end if
  if (mod((Nr+1),Nbproc_r) .ne. 0) then
    if (pglobal_id.eq.0) &
      print *,"Nbproc_r does not divide (Nr+1)"
    call ppexit()
    call exit(-1)
  endif
  
  !-> Verification of Nbproc_theta
  if (Nbproc_theta.gt.(Ntheta+1)) then
    if (pglobal_id.eq.0) then
      print*,' Nbproc_theta greater than Ntheta+1 '
      print*,' => Nbproc_theta = ',Nbproc_theta
      print*,'    Ntheta       = ',Ntheta
    end if
    call ppexit()
    call exit(-1)
  end if
  if (mod((Ntheta),Nbproc_theta) .ne. 0) then 
    if (pglobal_id.eq.0) &
      print *,"Nbproc_theta does not divide Ntheta"
    call ppexit()
    call exit(-1)
  endif

  !-> Verification of bloc_phi
  if (mod((Nphi),bloc_phi) .ne. 0) then 
    if (pglobal_id.eq.0) &
      print *,"bloc_phi does not divide (Nphi)"
    call ppexit()
    call exit(-1)
  endif

  !-> Verification of Nbproc_mu
  if (Nbproc_mu.gt.(Nmu+1)) then
    if (pglobal_id.eq.0) then
      print*,' Nbproc_mu greater than Nmu+1 '
      print*,' => Nbproc_mu = ',Nbproc_mu
      print*,'    Nmu       = ',Nmu
    end if
    call ppexit()
    call exit(-1)
  end if
end subroutine verif_parallel_data


!----------------------------------------
! verification of the mesh number
!----------------------------------------
subroutine verif_mesh_number
  use prec_const
  use globals
  use MPIutils_module
  implicit none

  integer     :: pow    
  real(RKIND) :: rem1, rem2, divr, divtheta, divphi, divvpar      

  if (clampv) then
    vstart = 1
    vend   = Nvpar-vstart
  else
    vstart = 0
    vend   = Nvpar
  end if

  if ((Nr.eq.0).or.(Ntheta.eq.0) &
    .or.(Nphi.eq.0).or.(Nvpar.eq.0)) then
    print*,'Problem with one of the dimension which is equal to 0'
    stop
  end if

  !------------------------------------------------------
  ! verify the number of points in (r,theta) direction :  
  !  - Nr+1 must be a power of 2                       
  !     (for the parallel transposition)               
  !  -  Ntheta must be a power of 2                    
  !     (for Fast Fourier Transform)                   
  !------------------------------------------------------
  !*** -> (Nr+1) verification ***
  rem1     = modulo(real(log(real(Nr+1))),real(log(2._RKIND))) 
  divr     = log(real(Nr+1))/log(2._RKIND)
  if (rem1.ne.0) then
    pow = int(divr)
    Nr  = max(1,2**pow-1)
    if (pglobal_id.eq.0) then
      write(6,*) 'ATTENTION changment : Nr = ',Nr
    endif
  endif

  !*** -> Ntheta verification  ***
  rem2     = modulo(real(log(real(Ntheta))),real(log(2._RKIND))) 
  divtheta = log(real(Ntheta))/log(2._RKIND)
  if (rem2.ne.0) then
    pow    = int(divtheta)
    Ntheta = max(1,2**pow)
    if (pglobal_id.eq.0) then
      write(6,*) 'ATTENTION changment : Ntheta = ',Ntheta
    endif
  end if

  !-------------------------------------------------------------
  !  verify the number of points in (phi,vpar) direction : 
  !   -  Nphi must be a power of 2                         
  !      (for Fast Fourier Transform)                      
  !-------------------------------------------------------------
  !*** -> Nphi verification ***
  rem1    = modulo(real(log(real(Nphi))),real(log(2._RKIND))) 
  divphi  = log(real(Nphi))/log(2._RKIND)
  if (rem1.ne.0) then
    pow   = int(divphi)
    Nphi  = max(1,2**pow)
    if (pglobal_id.eq.0) then
      write(6,*) 'ATTENTION replacement : Nphi = ',Nphi
    end if
  end if

  if (((Nr+1)/Nbproc_r).lt.32) then
    if (pglobal_id.eq.0) then
      print*, "Attention (Nr+1)/Nbproc_r ", &
        "must be equal to or greater than 32 "
    end if
    call ppexit()
    call exit(-1)
  end if
  if ((Ntheta/Nbproc_theta).lt.32) then
    if (pglobal_id.eq.0) then
      print*, "Attention Ntheta/Nbproc_theta ", &
        "must be equal to or greater than 32 "
    end if
    call ppexit()
    call exit(-1)
  end if

  !-------------------------------------------------------------
  !  verify the number of points in mu direction
  !-------------------------------------------------------------
  if ((Lmu.eq.0).and.(Nmu.ne.0)) then
    if (pglobal_id.eq.0) then
      print*, 'Lmu=0 and Nmu not equal to 0 is not possible'
    end if
    call ppexit()
    call exit(-1)
  end if
end subroutine verif_mesh_number



!------------------------------------------------------------------
! Verification of the compatibility between the elapsed time 
!  between two diagnostics saving 'Sdt_diag' and the time step
!  'Sdeltat', i.e:
!    - 'Sdt_diag' must be larger than 'Sdeltat' 
!    - 'Sdt_diag' must be a multiple of 'Sdeltat'
!------------------------------------------------------------------
subroutine check_dt_diag(Sdeltat_name,Sdeltat,Sdt_diag_name,Sdt_diag)
  use prec_const
  use globals
  use Io_Unit_module, only: IU_open, IU_close
  
  implicit none
  
  character(LEN=*), intent(in)    :: Sdeltat_name
  real(RKIND)     , intent(in)    :: Sdeltat
  character(LEN=*), intent(in)    :: Sdt_diag_name
  real(RKIND)     , intent(inout) :: Sdt_diag


  character(LEN=100) :: parameter_file
  integer:: udata
  real(RKIND)        :: dt_diag_old
  character(LEN=100) :: Sdt_diag_name_eq, Sdeltat_name_eq
  
  !--> write in a string character : "<Sdeltat_name> = " 
  Sdeltat_name_eq = trim(Sdeltat_name)//" = "
  !--> write in a string character : "<Sdt_diag_name> = " 
  Sdt_diag_name_eq = trim(Sdt_diag_name)//" = "

  !--> test if Sdt_diag < Sdeltat
  if (Sdt_diag.lt.Sdeltat) then 
    dt_diag_old = Sdt_diag
    Sdt_diag    = Sdeltat
    if (pglobal_id.eq.0) then
      write(6,*) '----------------------------------------------'
      write(6,'(A,A,1pe12.3,A,A,1pe12.3)') '  WARNING: ', &
        trim(Sdt_diag_name_eq), &
        dt_diag_old, ' was smaller than ', &
        trim(Sdeltat_name_eq), Sdeltat
      write(6,'(A,A,1pe12.3)') '--> so now ', &
        trim(Sdt_diag_name_eq), Sdt_diag
      write(6,*) '----------------------------------------------'
      write(parameter_file,'(A,'//numfmt_rst//',A)') &
        "gysela_data", nb_restart_diag, ".out"
      call IU_open(udata,file = parameter_file,status = 'UNKNOWN', &
        position = 'APPEND', form = 'FORMATTED')
      write(udata,'(A,A,1pe12.3,A,A,1pe12.3)') '  WARNING: ', &
        trim(Sdt_diag_name_eq), &
        dt_diag_old, ' was smaller than ', &
        trim(Sdeltat_name_eq), Sdeltat
      write(udata,'(A,A,1pe12.3)') '--> so now ', &
        trim(Sdt_diag_name_eq), Sdt_diag
      call IU_close(udata)
    end if
  end if

  !--> test if Sdt_diag is a multiple of Sdeltat
  if (mod(Sdt_diag,Sdeltat).ne.0._RKIND) then
    dt_diag_old = Sdt_diag
    Sdt_diag    = Sdeltat*floor(Sdt_diag/Sdeltat)
    if (pglobal_id.eq.0) then
      write(6,*) '----------------------------------------------'
      write(6,'(A,A,1pe12.3,A,A,1pe12.3)') '  WARNING: ', &
        trim(Sdt_diag_name_eq), &
        dt_diag_old, ' was not a multiple of ', &
        trim(Sdeltat_name_eq), Sdeltat
      write(6,'(A,A,1pe12.3)') '--> so now ', &
        trim(Sdt_diag_name_eq), Sdt_diag
      write(6,*) '----------------------------------------------'
      write(parameter_file,'(A,'//numfmt_rst//',A)') &
        "gysela_data", nb_restart_diag, ".out"
      call IU_open(udata,file = parameter_file,status = 'UNKNOWN', &
        position = 'APPEND', form = 'FORMATTED')
      write(udata,'(A,A,1pe12.3,A,A,1pe12.3)') '  WARNING: ', &
        trim(Sdt_diag_name_eq), &
        dt_diag_old, ' was not a multiple of ', &
        trim(Sdeltat_name_eq), Sdeltat
      write(udata,'(A,A,1pe12.3)') '--> so now ', &
        trim(Sdt_diag_name_eq), Sdt_diag
      call IU_close(udata)
    end if
  end if
end subroutine check_dt_diag
