!**************************************************************
!  Copyright Euratom-CEA
!  Authors : 
!     Virginie Grandgirard (virginie.grandgirard@cea.fr)
!     Chantal Passeron (chantal.passeron@cea.fr)
!     Guillaume Latu (guillaume.latu@cea.fr)
!     Xavier Garbet (xavier.garbet@cea.fr)
!     Philippe Ghendrih (philippe.ghendrih@cea.fr)
!     Yanick Sarazin (yanick.sarazin@cea.fr)
!  
!  This code GYSELA (for GYrokinetic SEmi-LAgrangian) 
!  is a 5D gyrokinetic global full-f code for simulating 
!  the plasma turbulence in a tokamak.
!  
!  This software is governed by the CeCILL-B license 
!  under French law and abiding by the rules of distribution 
!  of free software.  You can  use, modify and redistribute 
!  the software under the terms of the CeCILL-B license as 
!  circulated by CEA, CNRS and INRIA at the following URL
!  "http://www.cecill.info". 
!**************************************************************

#include "config-F.inc"

!----------------------------------------
! file : read_write.F90
! date : 16/09/2005
! used for the reading of :
! - mesh data
! - the equilibrium data
! - restart file (if in restart mode)
!----------------------------------------
module read_write_module
#ifndef NO_HDF5
  use HDF5
  use HDF5_module
#endif
  use prec_const
  implicit none

  !-------------------------------------
  !  PUBLIC VARIABLES
  !-------------------------------------  
  ! Output file names
  character(LEN=100), public       :: parameter_file = ''
  character(LEN=100), public       :: file_list
  integer           , public, save :: upar       = 24
  integer           , public, save :: ulist      = 25
  integer           , public, save :: uout_CL    = 26
  integer           , public, save :: uout_rprof = 27
  integer           , public, save :: uout_r3    = 28
!R3 logical           , public       :: file_r3_open !R3

  !******************************
  contains
  !******************************
  !------------------------------------------
  ! write parameters in the text result file
  !------------------------------------------
  subroutine Write_parameters()
    use globals
    use Io_Unit_module, only: IU_open, IU_close
    
    implicit none

    if (pglobal_id.eq.0) then
      write(parameter_file,'(A,'//numfmt_rst//',A)') &
        "gysela_data", nb_restart_diag, ".out"
      call IU_open(upar,file = parameter_file,status = 'UNKNOWN', &
        position='APPEND',form = 'FORMATTED')
      write(upar,PARALLEL)
      write(upar,MESH)
      write(upar,EQUIL)
      write(upar,COLLISION_OPERATOR)                 
      write(upar,BUFFER_REGION)                            
      write(upar,SOURCE)                          
      write(upar,FAST_PARTICLES)                         
      write(upar,ALGORITHM)
      write(upar,TEST)
      write(upar,OUTPUT)
      call IU_close(upar)
    end if
  end subroutine Write_parameters


  !-----------------------------------------------------------
  ! Read the restart files
  !  Output : - sEnstransf_norst is equal to .true. if
  !          the fluid moments mandatory for the computation
  !          of the energy transfer operator are readable
  !-----------------------------------------------------------
  subroutine read_f5d_rst(filename,argname, &
    f5d,Snb_diag,SPhi3D_nb_diag, &
    SMoments3D_nb_diag, Sf5D_nb_diag, &
    SSp_nb_diag, &                                  
    Stime_restart,Snb_restart,Snbiter_prev, &
    read_fields,sEntransf_read) 
    use prec_const
    use globals, only : deltat, mu_id, &
      pSpecies_id, species_frac, Nspecies, &
      pInSpecies_id
    use globals, only : collisions, &
      Tmean_loc, Vmean_loc
    use globals, only : energy_transfer, &
      moment_transfer
    use species_transf_module, only : Ns_FSavg, &
      Vs_FSavg, Ts_FSavg, Tscoll_FSavg
    use geometry_class
    use fdistribu5d_class
    implicit none
    character(LEN=*) , intent(in)    :: filename
    character(LEN=*) , intent(in)    :: argname
    type(fdistribu5d), intent(inout) :: f5d
    integer          , intent(inout) :: Snb_diag
    integer          , intent(inout) :: SPhi3D_nb_diag
    integer          , intent(inout) :: SMoments3D_nb_diag
    integer          , intent(inout) :: Sf5D_nb_diag
    integer          , intent(inout) :: SSp_nb_diag    
    real(RKIND)      , intent(inout) :: Stime_restart
    integer          , intent(out)   :: Snb_restart
    integer          , intent(out)   :: Snbiter_prev
    logical          , intent(in)    :: read_fields
    logical, optional, intent(inout) :: sEntransf_read

    character(LEN=100) :: structname
    integer            :: ierr
#ifndef NO_HDF5
    logical            :: fexist, var_exist
    character(LEN=100) :: filename_tmp
    integer(HID_T)     :: file_id  
    logical            :: exist_Ns, exist_Vs, exist_Ts, exist_Tscoll
    integer            :: ispec, ir, itheta, iphi, ivpar
    real(RKIND)        :: cs, previous_cs, newcs_on_oldcs

    filename_tmp = filename
    inquire(file=filename_tmp,exist=fexist)
    if (.not.(fexist)) then
      write(6,'(3A)') ' Error : the file ',filename_tmp, &
        ' does not exist'
    end if

    ierr   = 0
    call HDF5_open(trim(filename_tmp),file_id)
    call HDF5_load(file_id,deltat,"deltat")
    call HDF5_load(file_id,Snbiter_prev,"nbiter_prev")
    call HDF5_load(file_id,Snb_restart,"nb_restart")
    call HDF5_load(file_id,Snb_diag,"nb_diag")
    call HDF5_load(file_id,SPhi3D_nb_diag, &
      "Phi3D_nb_diag")
    call HDF5_load(file_id,SMoments3D_nb_diag, &
      "Moments3D_nb_diag")
    call HDF5_load(file_id,Sf5D_nb_diag,"f5D_nb_diag")
    call HDF5_load(file_id,SSp_nb_diag,"Spectra_nb_diag")
    call HDF5_load(file_id,Stime_restart,"initial_time")
    write(structname,'(A,A)') argname,"%mu_value"
    call HDF5_load(file_id,f5d%mu_value, &
      structname(1:len_trim(structname)))
    write(structname,'(A,A)') argname,"%values"
    if (read_fields) then
      call HDF5_load(file_id, &
        f5d%values,structname(1:len_trim(structname)),ierr)
      if (collisions) then
        call HDF5_load(file_id,Tmean_loc,"Tmean_loc")
        call HDF5_load(file_id,Vmean_loc,"Vmean_loc")
        if ( (energy_transfer).or.(moment_transfer) ) then
          sEntransf_read = .true.
          call HDF5_load(file_id=file_id, &
            data=Ns_FSavg,dsetname="Ns_FSavg",var_exist=exist_Ns)
          call HDF5_load(file_id=file_id, &
            data=Vs_FSavg,dsetname="Vs_FSavg",var_exist=exist_Vs)
          call HDF5_load(file_id=file_id, &
            data=Ts_FSavg,dsetname="Ts_FSavg",var_exist=exist_Ts)
          call HDF5_load(file_id=file_id, &
            data=Tscoll_FSavg,dsetname="Tscoll_FSavg", &
            var_exist=exist_Tscoll)
          if (any((/exist_Ns,exist_Vs,exist_Ts,exist_TScoll/) &
            .eqv. .false.)) &
            sEntransf_read = .false.
        end if
      endif
    end if

    !*** Reading of the previous concentration of the species s ***
    call H5Lexists_f(file_id,"cs",var_exist,ierr)
    if (var_exist) then
      call HDF5_load(file_id,previous_cs,"cs",ierr)
    else
      previous_cs = 1.0_RKIND
      print*,'pglobal_id = ',pglobal_id, &
        'The specie concentration (cs) is not present ', &
        'in HDF5 restart file. So it is set equal to 1.'
    end if
    call HDF5_close(file_id)
    if (ierr.ne.0) then
      print*,'pglobal_id = ',pglobal_id, &
        ' ==> error in reading of the 4D structure '
      call exit(-1)
    end if

    !*** the results are multiplied by the species ***
    !***  concentration 'species_frac' to allow a  ***
    !***  restart with a different value of this   ***
    !***  concentration                            *** 
    cs = species_frac(pSpecies_id)
    if (previous_cs .ne. cs) then
      newcs_on_oldcs = cs/previous_cs
!$OMP PARALLEL default(none) &
!$OMP private(ir,itheta,iphi,ivpar) &
!$OMP shared(f5d,cs,newcs_on_oldcs)
!$OMP BARRIER

!$OMP DO SCHEDULE(static)
      !--> For f5d values
      do ivpar = 0,f5d%n4
        do iphi = 0,f5d%n3
          do itheta = f5d%jstart_buf,f5d%jend_buf
            do ir = f5d%istart_buf,f5d%iend_buf
              f5d%values(ir,itheta,iphi,ivpar) = &
                f5d%values(ir,itheta,iphi,ivpar) * newcs_on_oldcs
            end do
          end do
        end do
      end do
!$OMP END DO
!$OMP BARRIER
!$OMP END PARALLEL

      !--> For energy transfer operator
      if (collisions) then
        if ( ((energy_transfer).or.(moment_transfer)) &
          .and.(sEntransf_read)) then
          do ir = 0,f5d%n1
            do ispec = 0,Nspecies-1
              Ns_FSavg(ispec,ir) = Ns_FSavg(ispec,ir) * &
                newcs_on_oldcs
            end do
          end do
        end if
      end if
    endif
#else 
    Snb_restart = 0
    Snbiter_prev = 0
#endif
  end subroutine read_f5d_rst


  !------------------------------------------
  ! Write the restart file
  !------------------------------------------
  subroutine write_f5d_rst(argname,initial_time, &
    Snb_diag,SPhi3D_nb_diag,SMoments3D_nb_diag, Sf5D_nb_diag, &
    SSp_nb_diag, f5d,pfn_restart_save) 
    use globals, only : rst_saving, &
      deltat, nbiter_prev, nb_restart, num_restart, &
      pSpecies_id, Nbproc_species, species_frac, Nspecies
    use globals, only : collisions, &
      Tmean_loc, Vmean_loc
    use globals, only : Tmean_loc_restart_save, &
      Vmean_loc_restart_save
    use globals, only : energy_transfer, &
      moment_transfer
    use species_transf_module, only : Ns_FSavg, & 
      Vs_FSavg, Ts_FSavg, Tscoll_FSavg
    use species_transf_module, only : Ns_FSavg_restart_save, &
      Vs_FSavg_restart_save, Ts_FSavg_restart_save, &
      Tscoll_FSavg_restart_save
    use globals, only : stop_gysela, &
      ask_for_checksum, h5_lock, asynchrone_writing
    use prec_const
    use clock_module
    use fdistribu5d_class
    use mpiMod
    implicit none

!R3 #include "code_instrumentation/r3_info.h" !R3
    character(LEN=*) , intent(in) :: argname
    real(RKIND)      , intent(in) :: initial_time
    integer          , intent(in) :: Snb_diag
    integer          , intent(in) :: SPhi3d_nb_diag
    integer          , intent(in) :: SMoments3D_nb_diag
    integer          , intent(in) :: Sf5D_nb_diag
    integer          , intent(in) :: SSp_nb_diag     
    type(fdistribu5d), pointer    :: f5d
    type(fdistribu5d), pointer    :: pfn_restart_save
    integer                       :: ierr

#ifndef NO_HDF5
    integer        :: pInSpecies_id_tmp
    integer        :: inumber_buf, jnumber_buf
    integer(HID_T) :: file_id   ! HDF5 file identifier
    integer        :: i, j, k, l, icount_open
    integer        :: ispec, ir, itheta, iphi, ivpar
#endif
    character(LEN=100), save :: structname, structname1, structname2
    character(LEN=100), save :: filename
    ! list of data name for C asynchrone writing
    character(LEN=100), save, dimension(1:18) :: dataname
    integer           , save, dimension(1:7)  :: int_array
    real(RKIND)       , save, dimension(1:4)  :: real_array
    integer           , save, dimension(4)    :: dim_4D
    integer           , save, dimension(3)    :: dim_3D
    integer           , save, dimension(2)    :: dim_2D
    integer           , save :: activate_collision_writing
    integer           , save :: activate_sptransfer_writing
    integer           , save :: stop_gysela_tmp
    real(RKIND)              :: dummy_real(0:1)

    if (rst_saving) then                     
!R3      call r3_info_begin(r3_info_index_0, &   !R3
!R3        'CPU_write_f5d_res')                  !R3
      call Write_dyn_balise_open('CPU_write_f5d_res') !R4
      call clck_time(bclock_writerst)
    end if

#ifndef NO_HDF5
    if (rst_saving) then
      !*** definition of the file name saving ***
      pInSpecies_id_tmp = pglobal_id-Nbproc_species*pSpecies_id
      write(filename,'(A,I1,A,I1,A,A,A,I5.5,A)') &
        "gys_rst.sp",pSpecies_id,".n",num_restart,".", &
        trim(argname),".p",pInSpecies_id_tmp,".h5"

      !*** 5D results writting ***
      if (.not.asynchrone_writing) then
        call Write_dyn_balise_open('CPU_write_f5d_Header') !R4
        !*****************************
        !***   SYNCHRONE saving    ***
        !*****************************
        stop_gysela = 0
        !*** writting of the restart files ***
        if (pglobal_id.eq.0) &
          write(*,*) '=============> restart synchrone writing : ', &
          trim(filename)
        ierr        = 1
        icount_open = 0
        do while ( (ierr.ne.0).and.(icount_open.lt.5) ) 
          call HDF5_create(trim(filename),file_id,ierr)
          icount_open = icount_open+1
        end do
        if (ierr.ne.0) then
          print*,'pglobal_id = ',pglobal_id, &
            ' ==> error for opening of ',filename
        end if
        call HDF5_save(file_id,deltat,"deltat")
        call HDF5_save(file_id,nbiter_prev,"nbiter_prev")
        call HDF5_save(file_id,nb_restart,"nb_restart")
        call HDF5_save(file_id,Snb_diag,"nb_diag")
        call HDF5_save(file_id, &
          SPhi3D_nb_diag,"Phi3D_nb_diag")
        call HDF5_save(file_id,SMoments3D_nb_diag, &
          "Moments3D_nb_diag")
        call HDF5_save(file_id,Sf5D_nb_diag,"f5D_nb_diag")
        call HDF5_save(file_id,SSp_nb_diag, &
          "Spectra_nb_diag")
        call HDF5_save(file_id,initial_time,"initial_time")
        call HDF5_save(file_id, &
          species_frac(pSpecies_id),"cs")
        write(structname,'(A,A)') argname,"%mu_value"
        call HDF5_save(file_id, &
          f5d%mu_value,structname(1:len_trim(structname)))
        call Write_dyn_balise_close('CPU_write_f5d_Header') !R4
        call Write_dyn_balise_open('CPU_write_f5d_4Ddata') !R4
        inumber_buf = f5d%iend_buf-f5d%istart_buf+1
        jnumber_buf = f5d%jend_buf-f5d%jstart_buf+1
        write(structname,'(A,A)') argname,"%values"
        call HDF5_save(file_id, &
          f5d%values, &      
          inumber_buf,jnumber_buf,f5d%n3+1, &
          f5d%n4+1,structname(1:len_trim(structname)), &
          ask_for_checksum)
        call Write_dyn_balise_close('CPU_write_f5d_4Ddata') !R4
        if (collisions) then
          call Write_dyn_balise_open('CPU_write_f5d_3Ddata') !R4
          inumber_buf = f5d%iend-f5d%istart+1
          jnumber_buf = f5d%jend-f5d%jstart+1
          call HDF5_save(file_id, &
            Tmean_loc, & 
            inumber_buf,jnumber_buf,f5d%n3+1, &
            "Tmean_loc",ask_for_checksum)
          call HDF5_save(file_id, &
            Vmean_loc, & 
            inumber_buf,jnumber_buf,f5d%n3+1,"Vmean_loc", &
            ask_for_checksum)
          call Write_dyn_balise_close('CPU_write_f5d_3Ddata') !R4
          if ( (energy_transfer).or.(moment_transfer) ) then
            call Write_dyn_balise_open('CPU_write_f5d_2Ddata') !R4
            call HDF5_save(file_id, &
              Ns_FSavg,Nspecies,f5d%n1+1,"Ns_FSavg", & 
              ask_for_checksum)
            call HDF5_save(file_id, &
              Vs_FSavg,Nspecies,f5d%n1+1,"Vs_FSavg", & 
              ask_for_checksum)
            call HDF5_save(file_id, &
              Ts_FSavg,Nspecies,f5d%n1+1,"Ts_FSavg", & 
              ask_for_checksum)
            call HDF5_save(file_id, &
              Tscoll_FSavg,Nspecies,f5d%n1+1, & 
              "Tscoll_FSavg",ask_for_checksum)
            call Write_dyn_balise_close('CPU_write_f5d_2Ddata') !R4
          end if
        end if
        call HDF5_close(file_id)
      else
        !*****************************
        !***   ASYNCHRONE saving   ***
        !*****************************
        call MPI_BARRIER(MPI_COMM_WORLD,ierr)
!OT!        call clck_time(bclock_writerst)
!$OMP PARALLEL default(none) &
!$OMP  private(i,j,k,l) &
!$OMP  shared(pfn_restart_save,f5d)
!$OMP DO SCHEDULE(static)
        do l = 0,f5d%n4
          do k = 0,f5d%n3
            do j = f5d%jstart_buf,f5d%jend_buf
              do i = f5d%istart_buf,f5d%iend_buf
                pfn_restart_save%values(i,j,k,l) = &
                  f5d%values(i,j,k,l)
              enddo
            enddo
          enddo
        enddo
!$OMP ENDDO
!$OMP END PARALLEL
!OT!          call clck_time(eclock_writerst)
!OT!          call clck_diff(bclock_writerst, &
!OT!            eclock_writerst,global_time_writerst)
!OT!          call clck_ldiff(bclock_writerst,eclock_writerst,temp)
!OT!          write(*,*) 'pglobal_id = ',pglobal_id, &
!OT!            ' time for f5D copy = ', temp
        pfn_restart_save%muid       = f5d%muid
        pfn_restart_save%mu_value   = f5d%mu_value
        pfn_restart_save%n1         = f5d%n1
        pfn_restart_save%n2         = f5d%n2
        pfn_restart_save%n3         = f5d%n3
        pfn_restart_save%n4         = f5d%n4
        pfn_restart_save%istart     = f5d%istart
        pfn_restart_save%istart_buf = f5d%istart_buf
        pfn_restart_save%iend       = f5d%iend
        pfn_restart_save%iend_buf   = f5d%iend_buf
        pfn_restart_save%jstart     = f5d%jstart
        pfn_restart_save%jstart_buf = f5d%jstart_buf
        pfn_restart_save%jend       = f5d%jend
        pfn_restart_save%jend_buf   = f5d%jend_buf
        if (pglobal_id.eq.0) &
          write(*,*) '=============> restart asynchrone writing : ', &
          trim(filename)
        filename     = trim(adjustl(filename))//char(0)
        structname1  = trim(adjustl(argname//"%mu_value"))
        structname2  = trim(adjustl(argname//"%values"))
        ! real data name
        dataname(1)  = "deltat"
        dataname(2)  = "initial_time"
        dataname(3)  = structname1
        dataname(4)  = "cs"
        ! integer data name
        dataname(5)  = "nbiter_prev"
        dataname(6)  = "nb_restart"
        dataname(7)  = "nb_diag"
        dataname(8)  = "Phi3D_nb_diag"
        dataname(9)  = "Moments3D_nb_diag"
        dataname(10) = "f5D_nb_diag"
        dataname(11) = "Spectra_nb_diag"
        dataname(12) = structname2
        do i = 1, 12
          dataname(i) = trim(adjustl(dataname(i)))//char(0)
        enddo
        int_array(1:7) = (/nbiter_prev,nb_restart,Snb_diag, &
          SPhi3D_nb_diag,SMoments3D_nb_diag, &
          Sf5D_nb_diag,SSp_nb_diag/)
        real_array(1:4) = (/deltat,initial_time,f5d%mu_value, &
          species_frac(pSpecies_id)/)
        DIM_4D = (/f5d%iend_buf-f5d%istart_buf+1, &
          f5d%jend_buf-f5d%jstart_buf+1,f5d%n3+1,f5d%n4+1/)
        activate_collision_writing  = 0
        activate_sptransfer_writing = 0
        DIM_3D = 0
        DIM_2D = 0
        if (collisions) then
          DIM_3D = (/f5d%iend-f5d%istart+1, &
            f5d%jend-f5d%jstart+1,f5d%n3+1/)
          activate_collision_writing = 1
          dataname(13) = "Tmean_loc"
          dataname(14) = "Vmean_loc"
          do i = 13,14
            dataname(i) = trim(adjustl(dataname(i)))//char(0)
          enddo
!$OMP PARALLEL default(none) &
!$OMP  private(ir,itheta,iphi) &
!$OMP  shared(f5d, &
!$OMP  Tmean_loc_restart_save,Vmean_loc_restart_save, &
!$OMP  Tmean_loc,Vmean_loc)
!$OMP DO SCHEDULE(static)
          do iphi = 0,f5d%n3
            do itheta = f5d%jstart,f5d%jend
              do ir = f5d%istart,f5d%iend
                Tmean_loc_restart_save(ir,itheta,iphi) = &
                  Tmean_loc(ir,itheta,iphi)
                Vmean_loc_restart_save(ir,itheta,iphi) = &
                  Vmean_loc(ir,itheta,iphi)
              enddo
            enddo
          enddo
!$OMP ENDDO
!$OMP END PARALLEL
          if ( (.not.energy_transfer).and.(.not.moment_transfer) ) then
            call asynch_restart_writing(f5d%values(f5d%istart_buf, &
              f5d%jstart_buf,0,0),DIM_2D,DIM_3D,DIM_4D, &
              filename,dataname,int_array,real_array, &
              activate_collision_writing, &
              activate_sptransfer_writing, &
              h5_lock,ask_for_checksum,stop_gysela, &
              Tmean_loc_restart_save(f5d%istart,f5d%jstart,0), &
              Vmean_loc_restart_save(f5d%istart,f5d%jstart,0), &
              dummy_real,dummy_real,dummy_real,dummy_real)
          else
!$OMP PARALLEL default(none) &
!$OMP  private(ir,ispec) &
!$OMP  shared(f5d,Nspecies,Ns_FSavg,Vs_FSavg, &
!$OMP  Ts_FSavg,Tscoll_FSavg, &
!$OMP  Ns_FSavg_restart_save,Vs_FSavg_restart_save, &
!$OMP  Ts_FSavg_restart_save,Tscoll_FSavg_restart_save)
!$OMP DO SCHEDULE(static)
            do ir = 0,f5d%n1
              do ispec = 0,Nspecies-1
                Ns_FSavg_restart_save(ispec,ir) = &
                  Ns_FSavg(ispec,ir)
                Vs_FSavg_restart_save(ispec,ir) = &
                  Vs_FSavg(ispec,ir)
                Ts_FSavg_restart_save(ispec,ir) = &
                  Ts_FSavg(ispec,ir)
                Tscoll_FSavg_restart_save(ispec,ir) = &
                  Tscoll_FSavg(ispec,ir)
              enddo
            enddo
!$OMP ENDDO
!$OMP END PARALLEL
            DIM_2D = (/Nspecies,f5d%n1+1/)
            activate_sptransfer_writing = 1
            dataname(15) = "Ns_FSavg"; dataname(16) = "Vs_FSavg"
            dataname(17) = "Ts_FSavg"; dataname(18) = "Tscoll_FSavg"
            do i = 15,18
              dataname(i) = trim(adjustl(dataname(i)))//char(0)
            enddo
            call asynch_restart_writing(f5d%values(f5d%istart_buf, &
              f5d%jstart_buf,0,0),DIM_2D,DIM_3D,DIM_4D, &
              filename,dataname,int_array,real_array, &
              activate_collision_writing, &
              activate_sptransfer_writing, &
              h5_lock,ask_for_checksum,stop_gysela, &
              Tmean_loc_restart_save(f5d%istart,f5d%jstart,0), &
              Vmean_loc_restart_save(f5d%istart,f5d%jstart,0), &
              Ns_FSavg_restart_save(0:,0:), &
              Vs_FSavg_restart_save(0:,0:), &
              Ts_FSavg_restart_save(0:,0:), &
              Tscoll_FSavg_restart_save(0:,0:))
          end if
        else
          call asynch_restart_writing(f5d%values(f5d%istart_buf, &
            f5d%jstart_buf,0,0),DIM_2D,DIM_3D,DIM_4D,filename, &
            dataname,int_array,real_array, &
            activate_collision_writing, &
            activate_sptransfer_writing, &
            h5_lock,ask_for_checksum,stop_gysela, &
            dummy_real,dummy_real, &
            dummy_real,dummy_real,dummy_real,dummy_real)
        endif
      endif
    endif
#endif

    !*** if any failure on writing restart file, stop gysela ***
    call MY_Allreduce(stop_gysela,stop_gysela_tmp,1, &
      MPI_INTEGER,MPI_MAX,MPI_COMM_WORLD,ierr)
    stop_gysela = stop_gysela_tmp

    if (rst_saving) then
!R3      call r3_info_end (r3_info_index_0)  !R3
      call Write_dyn_balise_close('CPU_write_f5d_res') !R4
      call clck_time(eclock_writerst)
      call clck_diff(bclock_writerst,eclock_writerst, &
        global_time_writerst)
    end if
  end subroutine write_f5d_rst


  !------------------------------------------------ 
  !  check the HDF5 files (filename), and checksum 
  !    of this file if necessary
  !------------------------------------------------
  subroutine check_hdf5_files(filename, &
    time_restart,flag,array4D,Tmean_loc,Vmean_loc, &
    Ns_FSavg,Vs_FSavg,Ts_FSavg,Tscoll_FSavg,time0)
#ifndef NO_HDF5
    use HDF5
#endif
    use globals, only : ask_for_checksum, &
      collisions, energy_transfer, moment_transfer, pglobal_id
    use globals, only : mpi_comm_species
    use mpiMod
    character(LEN=*)               , intent(in)        :: filename
    real(RKIND)                    , intent(out)       :: time_restart
    integer                        , intent(out)       :: flag
    real(RKIND), dimension(:,:,:,:), pointer           :: array4D
    real(RKIND), dimension(:,:,:)  , pointer, optional :: Tmean_loc
    real(RKIND), dimension(:,:,:)  , pointer, optional :: Vmean_loc
    real(RKIND), dimension(:,:)    , pointer, optional :: Ns_FSavg
    real(RKIND), dimension(:,:)    , pointer, optional :: Vs_FSavg
    real(RKIND), dimension(:,:)    , pointer, optional :: Ts_FSavg
    real(RKIND), dimension(:,:)    , pointer, optional :: Tscoll_FSavg
    real(RKIND)                 , intent(in), optional :: time0

#ifndef NO_HDF5
    integer(HID_T) :: file_id, dataset, property, error_hdf
    integer        :: err, flag_tmp, isenable
    logical        :: exis, noerror, var_exist
    integer(HSIZE_T), dimension(1) :: dim      ! dataset dimensions
    integer(HSIZE_T), dimension(4) :: dim4D    ! dataset dimensions
    integer(HSIZE_T), dimension(3) :: dim3D    ! dataset dimensions
    integer(HSIZE_T), dimension(2) :: dim2D    ! dataset dimensions

    dim4D = (/size(array4D,1),size(array4D,2), &
      size(array4D,3),size(array4D,4)/)
    if (collisions) then
      dim3D = (/size(Tmean_loc,1),size(Tmean_loc,2), &
        size(Tmean_loc,3)/)
      if ( (energy_transfer).or.(moment_transfer) ) &
        dim2D = (/size(Ns_FSavg,1),size(Ns_FSavg,2)/)  
    end if

    !*** Test if the file <filename> exists ***
    noerror = .true.
    inquire(file = filename, exist=exis)
    if (.not.exis) then
      flag = 1
      write(6,*) 'file ', filename, &
        ' doesnt exist, skip from restart'
      noerror = .false.
    endif
    if (noerror) then
      call HDF5_open(filename,file_id,err)
      if (err < 0)  then
        flag = 1
        write(6,*) 'file ', filename, &
          ' cant be opened, skip from restart'
        noerror = .false.
      end if
    endif

    !*** Test if the variable 'initial_time' is readable ***
    !***  -> if yes: read 'init_time' valu e             *** 
    if (noerror) then
      call H5Dopen_f(file_id,"initial_time",dataset,err)
      if (err < 0)  then
        call H5Fclose_f(file_id,err)
        write(6,*) 'file ', filename, &
          ' dont contain initial_time variable, skip from restart'
        flag    = 1
        noerror = .false.
      endif
    end if
    if (noerror) then
      call H5Dread_f(dataset,H5T_NATIVE_DOUBLE,time_restart,dim,err)
      if (err < 0)  then
        call H5Dclose_f(dataset,err)
        call H5Fclose_f(file_id,err)
        write(6,*) 'failed to read integer from ', &
          filename, ' , skip from restart'
        flag    = 1
        noerror = .false.
      end if
    endif

    !*** Test if the first files (denoted by .n0.) ***
    !***  are more recent than the second ones     ***
    !***  (denoted with .n1.)                      ***
    if (present(time0).and.noerror) then
      if (time0 > time_restart) then
        call H5Dclose_f(dataset,err)
        call H5Fclose_f(file_id,err)
        write(6,'(A,1pe12.3,A,A)') &
          'first file more recent than second ( time =', &
          time_restart ,'), ignoring it : ', filename
        flag = -1
        return
      endif
    endif
    if (noerror) then
      call H5Dclose_f(dataset,err)
      call H5Fclose_f(file_id,err)
      if (err < 0)  then
        write(6,*) 'unable to close ', &
          filename, ' , skip from restart'
        flag    = 1
        noerror = .false.
      endif
    end if
    if (noerror) then
      call H5Fopen_f(filename, H5F_ACC_RDONLY_F, file_id, err)
      if (err < 0)  then
        write(6,*) 'file ', filename, &
          ' cant be opened, skip from restart'
        call H5Fclose_f(file_id,err)
        flag    = 1
        noerror = .false.
      endif
    end if

    !*** Try to access to f4d field         ***
    !***  (checksum is implicit if present) ***
    if (noerror) then
      call H5Dopen_f(file_id, "fn%values",dataset,err)
      if (err < 0)  then
        write(6,*) 'file ', filename, &
          ' dont contain fn%values or checksum failed, skip from restart'
        call H5Dclose_f(dataset,err)
        call H5Fclose_f(file_id,err)
        flag    = 1
        noerror = .false.
      endif
    end if
    if (noerror) then
      call H5Dread_f(dataset,H5T_NATIVE_DOUBLE,array4D,dim4D,err)
      if (err < 0)  then
        call H5Pclose_f(property,err)
        call H5Dclose_f(dataset,err)
        call H5Fclose_f(file_id,err)
        flag    = 1
        noerror = .false.
      endif
    end if

    !*** try to access to f3d fields (Tmean_loc and Vmean_loc) ***
    !***  which are saved when collisions are applied          ***
    !*** (checksum is implicit if present)                     ***
    if (collisions) then
      !---> Test if 'Tmean_loc' is readable 
      if (noerror) then
        call H5Dclose_f(dataset,err)
        call H5Dopen_f(file_id, "Tmean_loc",dataset,err)
        if (err < 0)  then
          write(6,*) 'file ', filename, &
            ' dont contain Tmean_loc or checksum failed, skip from restart'
          call H5Fclose_f(file_id,err)
          flag    = 1
          noerror = .false.
        endif
      end if
      if (noerror) then
        call H5Dread_f(dataset,H5T_NATIVE_DOUBLE, &
          Tmean_loc,dim3D,err)
        if (err < 0)  then
          call H5Dclose_f(dataset,err)
          call H5Fclose_f(file_id,err)
          flag    = 1
          noerror = .false.
        endif
      end if
      !---> Test if 'Vmean_loc' is readable 
      if (noerror) then
        call H5Dclose_f(dataset,err)
        call H5Dopen_f(file_id, "Vmean_loc",dataset,err)
        if (err < 0)  then
          write(6,*) 'file ', filename, &
            ' dont contain Vmean_loc or checksum failed, skip from restart'
          call H5Fclose_f(file_id,err)
          flag    = 1
          noerror = .false.
        endif
      end if
      if (noerror) then
        call H5Dread_f(dataset,H5T_NATIVE_DOUBLE, &
          Vmean_loc,dim3D,err)
        if (err < 0)  then
          call H5Dclose_f(dataset,err)
          call H5Fclose_f(file_id,err)
          flag    = 1
          noerror = .false.
        endif
      end if

      !*** Try to access to 2D fields (Ns_FSavg, Vs_FSavg, ***
      !***  Ts_FSavg and Tscoll_FSavg) which are required  ***
      !***  for energy transfer operator                   ***
      !*** (checksum is implicit if present)               ***
      if ( (energy_transfer).or.(moment_transfer) ) then
        !---> Test if 'Ns_FSavg' is readable 
        if (noerror) then
          call H5Lexists_f(file_id,"Ns_FSavg",var_exist,err)
          if (var_exist) then
            call H5Dclose_f(dataset,err)
            call H5Dopen_f(file_id,"Ns_FSavg",dataset,err)
            if (err < 0)  then
              write(6,*) 'file ', filename, &
                ' checksum failed for Ns_FSavg, skip from restart'
              call H5Fclose_f(file_id,err)
              flag    = 1
              noerror = .false.
            end if
            if (noerror) then
              call H5Dread_f(dataset,H5T_NATIVE_DOUBLE, &
                Ns_FSavg,dim2D,err)
              if (err < 0)  then
                call H5Dclose_f(dataset,err)
                call H5Fclose_f(file_id,err)
                flag    = 1
                noerror = .false.
              endif
            end if
          else
            write(6,*) 'file ', filename, ' dont contain Ns_FSavg'
          endif
        end if

        !---> Test if 'Vs_FSavg' is readable 
        if (noerror) then
          call H5Lexists_f(file_id,"Vs_FSavg",var_exist,err)
          if (var_exist) then
            call H5Dclose_f(dataset,err)
            call H5Dopen_f(file_id,"Vs_FSavg",dataset,err)
            if (err < 0)  then
              write(6,*) 'file ', filename, &
                ' checksum failed for Vs_FSavg, skip from restart'
              call H5Fclose_f(file_id,err)
              flag    = 1
              noerror = .false.
            end if
            if (noerror) then
              call H5Dread_f(dataset,H5T_NATIVE_DOUBLE, &
                Vs_FSavg,dim2D,err)
              if (err < 0)  then
                call H5Dclose_f(dataset,err)
                call H5Fclose_f(file_id,err)
                flag    = 1
                noerror = .false.
              endif
            end if
          else
            write(6,*) 'file ', filename, ' dont contain Vs_FSavg'
          endif
        end if

        !---> Test if 'Ts_FSavg' is readable 
        if (noerror) then
          call H5Lexists_f(file_id,"Ts_FSavg",var_exist,err)
          if (var_exist) then
            call H5Dclose_f(dataset,err)
            call H5Dopen_f(file_id,"Ts_FSavg",dataset,err)
            if (err < 0)  then
              write(6,*) 'file ', filename, &
                ' checksum failed for Ts_FSavg, skip from restart'
              call H5Fclose_f(file_id,err)
              flag    = 1
              noerror = .false.
            end if
            if (noerror) then
              call H5Dread_f(dataset,H5T_NATIVE_DOUBLE, &
                Ts_FSavg,dim2D,err)
              if (err < 0)  then
                call H5Dclose_f(dataset,err)
                call H5Fclose_f(file_id,err)
                flag    = 1
                noerror = .false.
              endif
            end if
          else
            write(6,*) 'file ', filename, ' dont contain Ts_FSavg'
          endif
        end if


        !---> Test if 'Tscoll_FSavg' is readable 
        if (noerror) then
          call H5Lexists_f(file_id,"Tscoll_FSavg",var_exist,err)
          if (var_exist) then
            call H5Dclose_f(dataset,err)
            call H5Dopen_f(file_id,"Tscoll_FSavg",dataset,err)
            if (err < 0)  then
              write(6,*) 'file ', filename, &
                ' checksum failed for Tscoll_FSavg, skip from restart'
              call H5Fclose_f(file_id,err)
              flag    = 1
              noerror = .false.
            end if
            if (noerror) then
              call H5Dread_f(dataset,H5T_NATIVE_DOUBLE, &
                Tscoll_FSavg,dim2D,err)
              if (err < 0)  then
                call H5Dclose_f(dataset,err)
                call H5Fclose_f(file_id,err)
                flag    = 1
                noerror = .false.
              endif
            end if
          else
            write(6,*) 'file ', filename, ' dont contain Tscoll_FSavg'
          endif
        end if
      endif   !end : if (energy_transfer)
    end if    !end : if (collisions)

    if (.not.(noerror)) then
      time_restart = -1._RKIND
      if (pglobal_id.eq.0) &
        write(*,*) 'Problem during the restart file reading'
    end if
    if (noerror) then 
      call H5Dclose_f(dataset,err)
      call H5Fclose_f(file_id,err)
    endif
    call MY_Allreduce(flag,flag_tmp,1,MPI_INTEGER, &
      MPI_MAX,mpi_comm_species,err)
    flag = flag_tmp
#else 
    time_restart = -1._RKIND
    flag = 1
#endif
  end subroutine check_hdf5_files


  !--------------------------------------------------
  ! Test if the simulation is a restart case or not
  ! Output :
  !   - Sinit_mode (logical) : equal to .true.
  !     if the restart status is not the same
  !     for all species 
  !       => a new species is added 
  !       => all is re-initialized at time t=0
  !--------------------------------------------------
  subroutine testif_restart(Sinit_mode,Sdiff_init_species)
    use globals        , only : restart, &
      pglobal_id, major_species, Nspecies
    use MPIutils_module, only : ppexit
    use mpiMod
    implicit none
    logical, intent(out) :: Sinit_mode
    logical, intent(out) :: Sdiff_init_species

    integer :: ispec
    logical :: exist_n0, exist_n1
    character(LEN=100), save :: filename

    Sinit_mode = .true.
    do ispec = 0,Nspecies-1
      restart(ispec) = .true.
      write(filename,'(A,I1,A)') &
        "gys_rst.sp",ispec,".n0.fn.p00000.h5"
      inquire(file = filename, exist=exist_n0)
      write(filename,'(A,I1,A)') &
        "gys_rst.sp",ispec,".n1.fn.p00000.h5"
      inquire(file = filename, exist=exist_n1)
      if ((.not.exist_n0).and.(.not.exist_n1)) then
        restart(ispec)  = .false.
      else
        Sinit_mode   = .false.
      end if
    enddo

#ifdef NO_HDF5
    do ispec = 0,Nspecies-1
      if (restart(ispec)) then
        if (pglobal_id.eq.0) then
          print*,'RESTART NOT POSSIBLE : ', &
            'The restart files are not saved in the mode NO_HDF5'
        end if
        call ppexit()
        call exit(-1)
      end if
    end do
#endif
    !*** Test if the restart status is the same          ***
    !*** for all species                                 ***
    !*** Rk : The case 'restart' for major species while ***
    !***      'init' for the minoritary species is not   ***
    !***      treated                                    ***
    Sdiff_init_species   = .false.
    if (Nspecies.gt.1) then
      do ispec = 0,Nspecies-2
        if ( restart(ispec).neqv.restart(ispec+1) ) &
          Sdiff_init_species = .true.
      end do
      if (Sdiff_init_species) then
        if (.not.restart(major_species)) then
          if (pglobal_id.eq.0) then
            print*,'The case <init> for the major species', &
              ' and <restart> for minoritary species is not treated'
          end if
          call ppexit()
          call exit(-1)
        else
          Sinit_mode = .true.
        end if
      end if
    end if
  end subroutine testif_restart


  !-----------------------------------------------
  ! end of the restart
  !-----------------------------------------------
  subroutine end_restart()
    use globals, only : pglobal_id, &
      num_restart, nb_restart
    use mpiMod
    use Io_Unit_module, only: IU_open, IU_close
    
    implicit none

    if (pglobal_id.eq.0) then
      write(file_list,'(A)')"file_list.out"
      call IU_open(ulist,file = file_list, status="replace", &
        form = 'FORMATTED')
      write(ulist,*) nb_restart
      call IU_close(ulist)
    end if
  end subroutine end_restart


  !-----------------------------------------------------
  !  - initialisation of the distribution function at 
  !      the time t
  !  These functions are read into one set 
  !  of restart files as:
  !     . gys_rst.n<0 or 1>.fn.p<num_proc>.h5
  !  (used in predictor-corrector case)
  !----------------------------------------------------- 
  subroutine init_fn_rst(fn, &
    Snb_diag,SPhi3D_nb_diag,SMoments3D_nb_diag, &
    Sf5D_nb_diag,sSP_nb_diag, &
    Stime_restart,Snum_restart, &
    Snb_restart,Snbiter_prev,sEntransf_read)
    use globals            , only : pInSpecies_id, &
      pSpecies_id, Nbproc_species, asynchrone_writing, &
      Tmean_loc, Vmean_loc, &
      collisions, energy_transfer, moment_transfer
    use species_transf_module, only : Ns_FSavg, Vs_FSavg, &
      Ts_FSavg, Tscoll_FSavg
    use clock_module, only : clck_time, clck_diff, &
      bclock_init, eclock_init, global_time_init
    use fdistribu5d_class
    use mpiMod
    implicit none
    type(fdistribu5d) , intent(out)   :: fn
    integer           , intent(inout) :: Snb_diag
    integer           , intent(inout) :: SPhi3D_nb_diag
    integer           , intent(inout) :: SMoments3D_nb_diag
    integer           , intent(inout) :: Sf5D_nb_diag
    integer           , intent(inout) :: sSP_nb_diag
    real(RKIND)       , intent(inout) :: Stime_restart
    integer           , intent(out)   :: Snum_restart
    integer           , intent(out)   :: Snb_restart
    integer           , intent(out)   :: Snbiter_prev
    logical           , intent(out)   :: sEntransf_read

    character(LEN=50) :: restart_file
    logical           :: readed_n0, readed_n1
    integer           :: flag, flag_tmp
    real(RKIND)       :: time_restart1, time_restart2
    integer           :: err
    integer           :: pInSpecies_id_tmp 

    call clck_time(bclock_init)

    sEntransf_read = .false.
    !*** check the most recent file ***
    !---> Reading of the all the files indexed by '.n0.', i.e
    !--->   'gys_rst.sp<pSpecies_id>.n0.fn.p*.h5'
    pInSpecies_id_tmp = pglobal_id-Nbproc_species*pSpecies_id
    write(restart_file(1:),'(A,I1,A,I5.5,A)') &
      "gys_rst.sp",pSpecies_id,".n0.fn.p",pInSpecies_id_tmp,".h5"
    flag          =  0
    time_restart1 = -1._RKIND
    time_restart2 = -1._RKIND
    readed_n0     = .false.
    readed_n1     = .false.
    if (collisions) then
      if ( (energy_transfer).or.(moment_transfer) ) then
        call check_hdf5_files(trim(restart_file(1:)),time_restart1, &
          flag,fn%values,Tmean_loc,Vmean_loc, &
          Ns_FSavg,Vs_FSavg,Ts_FSavg,Tscoll_FSavg)
      else
        call check_hdf5_files(trim(restart_file(1:)),time_restart1, &
          flag,fn%values,Tmean_loc,Vmean_loc)
      end if
    else
      call check_hdf5_files(trim(restart_file(1:)), &
        time_restart1,flag,fn%values)
    endif
    if (flag == 0) then
      Snum_restart   = 0
      Stime_restart  = time_restart1
      readed_n0      = .true.
      if (pInSpecies_id_tmp.eq.0) &
        write(*,*) trim(restart_file(1:)), &
        ' was successfully readed, time_restart = ', time_restart1
    endif

    !---> Reading of the all the files indexed by '.n1.', i.e
    !--->   'gys_rst.sp<pSpecies_id>.n1.fn.p*.h5'
    !---> Chek if these files are more recent than the files indexed
    !--->   by '.n0.'
    write(restart_file(1:),'(A,I1,A,I5.5,A)') &
      "gys_rst.sp",pSpecies_id,".n1.fn.p",pInSpecies_id_tmp,".h5"
    flag = 0
    if (collisions) then
      if ( (energy_transfer).or.(moment_transfer) ) then
        if (readed_n0) then
          call check_hdf5_files(trim(restart_file(1:)),time_restart2,flag, &
            fn%values,Tmean_loc,Vmean_loc, &
            Ns_FSavg,Vs_FSavg,Ts_FSavg,Tscoll_FSavg,time0=time_restart1)
        else
          call check_hdf5_files(trim(restart_file(1:)), &
            time_restart2,flag,fn%values,Tmean_loc,Vmean_loc, &
            Ns_FSavg,Vs_FSavg,Ts_FSavg,Tscoll_FSavg)
        endif
      else
        if (readed_n0) then
          call check_hdf5_files(trim(restart_file(1:)),time_restart2,flag, &
            fn%values,Tmean_loc,Vmean_loc,time0=time_restart1)
        else
          call check_hdf5_files(trim(restart_file(1:)), &
            time_restart2,flag,fn%values,Tmean_loc,Vmean_loc)
        endif
      end if
    else
      if (readed_n0) then
        call check_hdf5_files(trim(restart_file(1:)), &
          time_restart2,flag,fn%values,time0=time_restart1)
      else
        call check_hdf5_files(trim(restart_file(1:)), &
          time_restart2,flag,fn%values)
      endif
    endif
    if (flag == 0) then
      write(*,*) trim(restart_file(1:)), &
        ' was successfully readed, time_restart = ', time_restart2
      if (time_restart2 > Stime_restart) then
        Snum_restart  = 1
        Stime_restart = time_restart2
        readed_n1     = .true.
      endif
    endif
    if (Stime_restart .eq. -1._RKIND) then
      write(6,*) 'n0 and n1 are corrupted'
      stop
    endif

    !*** reading of the distribution function fn ***
    write(restart_file(1:),'(A,I1,A,I1,A,I5.5,A)') &
      "gys_rst.sp",pSpecies_id,".n",Snum_restart,".fn.p", &
      pInSpecies_id_tmp,".h5"
    if ((readed_n0.and..not.(time_restart2 < time_restart1)) .or. &
      (.not.readed_n0.and.(time_restart2 < time_restart1)) .or. &
      (time_restart2 > time_restart1)) then
      call read_f5d_rst( &
        restart_file(1:len_trim(restart_file)), &
        "fn",fn,Snb_diag, &
        SPhi3D_nb_diag,SMoments3D_nb_diag,Sf5D_nb_diag, &
        sSP_nb_diag, Stime_restart,Snb_restart,Snbiter_prev, &
        read_fields=.false., &
        sEntransf_read=sEntransf_read)
      write(*,*) 'no F3D & F4D re-reading'
    else
      call read_f5d_rst( &
        restart_file(1:len_trim(restart_file)), &
        "fn",fn,Snb_diag,SPhi3D_nb_diag, &
        SMoments3D_nb_diag,Sf5D_nb_diag, &
        sSP_nb_diag, Stime_restart,Snb_restart,Snbiter_prev, &
        read_fields=.true., &
        sEntransf_read=sEntransf_read)
    endif

    Snum_restart = mod(Snum_restart+1,2)

    call clck_time(eclock_init)
    call clck_diff(bclock_init,eclock_init, &
      global_time_init)
  end subroutine init_fn_rst
end module read_write_module
