// *******************************************************************
//  Copyright Euratom-CEA
//  Authors : 
//     Virginie Grandgirard (virginie.grandgirard@cea.fr)
//     Chantal Passeron (chantal.passeron@cea.fr)
//     Guillaume Latu (guillaume.latu@cea.fr)
//     Xavier Garbet (xavier.garbet@cea.fr)
//     Philippe Ghendrih (philippe.ghendrih@cea.fr)
//     Yanick Sarazin (yanick.sarazin@cea.fr)
//  
//  This code GYSELA (for GYrokinetic SEmi-LAgrangian) 
//  is a 5D gyrokinetic global full-f code for simulating 
//  the plasma turbulence in a tokamak.
//  
//  This software is governed by the CeCILL-B license 
//  under French law and abiding by the rules of distribution 
//  of free software.  You can  use, modify and redistribute 
//  the software under the terms of the CeCILL-B license as 
//  circulated by CEA, CNRS and INRIA at the following URL
//  "http://www.cecill.info". 
// *******************************************************************

#include "config-c.h"

/** \file
 * C/fortran interface for subroutines
 */

#ifndef NO_HDF5

#include <pthread.h>
#include <hdf5.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <sys/time.h>
#include <sys/resource.h>


#define MSRATE 1000000LL
#define MAXSEC 1000000000LL

// data transfered to sub-thread for restart writing, 
//   containing all the restart informations
// dim_4D : the dimension in each direction of the 4D-field 
//   (transposed to perform compatibility with fortran)
// dim_3D : the dimension in each direction of the 3D-field 
//   (transposed to perform compatibility with fortran)
// dim_2D : the dimension in each direction of the 2D-field 
//   (transposed to perform compatibility with fortran)
// lock_f5d : to indicate to gysela if the asynchrone thread 
// had finish to write
// int_array : list of all integer to write in the restart file :
//             [nbiter_prev,nb_restart,Snb_diag,
//              SPhi3D_nb_diag,SMoments3D_nb_diag,
//              Sf5D_nb_diag,SSp_nb_diag]
// collision : integer to know if collisions is activated 
//             ( = 1 : activated, 0 else)
// species_transf : integer to know if energy transfer is activated 
//             ( = 1 : activated, 0 else)
// checksum : integer to know if checksum is activated 
//             ( = 1 : activated, 0 else)
// field_4D       : pointer to the 4D-field in gysela code
// field_3D_Vmean : pointer to the 3D-field Vmean in gysela code
// field_3D_Tmean : pointer to the 3D-field Tmean in gysela code
// field_2D_Ns_FSavg : pointer to the 2D-field Ns_FSavg in gysela code
// field_2D_Vs_FSavg : pointer to the 2D-field Vs_FSavg in gysela code
// field_2D_Ts_FSavg : pointer to the 2D-field Ts_FSavg in gysela code
// field_2D_Tscoll_FSavg : pointer to the 2D-field 
//                         Tscoll_FSavg in gysela code
// real_array : list of all real to write in the restart file :
//             [deltat,initial_time,f5d%mu_value,cs]
//              where cs is the concentration of the species considered
// filename  : HDF5 filename
// data_name : all the data names (100 char max for each name, 17 names) : 
//              - 4 x for real dataset
//              - 7 x for integer dataset
//              - 2 x for field_3D dataset
//              - 1 x for field_4D dataset
//              - 4 x for field_2D dataset

typedef struct _thread_data_restart_t {
  int     dim_4D[4],          dim_3D[3],       *lock_f5d, int_array[7]  ;
  int     dim_2D[2],         *species_transfer                           ;
  int    *collision,         *checksum,        *stop_gysela             ;
  double *field_3D_Tmean,    *field_3D_Vmean                            ;
  double *field_2D_Ns_FSavg, *field_2D_Vs_FSavg                         ;
  double *field_2D_Ts_FSavg, *field_2D_Tscoll_FSavg                     ;
  double *field_4D,         real_array[4]                               ;
  char   *filename,         data_name[18][100]                          ;
} thread_data_restart_t;

// global declarations :
thread_data_restart_t data_restart              ;
pthread_t             threads, threads_restart  ;

void printrestart (void * ptr){
// local declaration
  thread_data_restart_t *data_restart = (thread_data_restart_t *)ptr;
  pthread_mutex_t mutex1 = PTHREAD_MUTEX_INITIALIZER;
  int cmpr = 1, error, i;
  hsize_t dim_var[1], dim_2D[2], dim_3D[3], dim_4D[4];
  hid_t file_id, dataset_id, datatype_id, dataspace_id, property;
  herr_t status;
#ifdef TIMER
  struct timeval tv2;

  long long int dt1, dt2, nb_periods;
  double p_difftime;
  gettimeofday(&tv2, (struct timezone*)0);
  dt1 = (((tv2).tv_sec)%MAXSEC  + 0.000001*(tv2).tv_usec) * MSRATE ;
#endif
//    creating file
  file_id = H5Fcreate(data_restart->filename, H5F_ACC_TRUNC, H5P_DEFAULT, H5P_DEFAULT);
//    real saving : 1 -> 4 : deltat, initial_time, f5d%mu_value, cs
  dim_var[0] = 1;
  for (i=0;i<4;i++){
    dataspace_id = H5Screate_simple(1, dim_var, NULL);
    dataset_id   = H5Dcreate2(file_id,data_restart->data_name[i],H5T_NATIVE_DOUBLE,dataspace_id,H5P_DEFAULT,H5P_DEFAULT,H5P_DEFAULT);
    status       = H5Dwrite(dataset_id, H5T_NATIVE_DOUBLE, H5S_ALL, H5S_ALL, H5P_DEFAULT, &data_restart->real_array[i]);
  if (status < 0) *data_restart->stop_gysela = 1;
    status       = H5Sclose(dataspace_id);
  if (status < 0) *data_restart->stop_gysela = 1;
    status       = H5Dclose(dataset_id);
  if (status < 0) *data_restart->stop_gysela = 1;
  }
//    integer saving : 1 -> 7 : nbiter_prev, nb_restart, nb_diag, Phi3D_nb_diag, Moments3D_nb_diag
//                              f5D_nb_diag, Spectra_nb_diag
  for (i=0;i<7;i++){
    dataspace_id = H5Screate_simple(1, dim_var, NULL);
    dataset_id   = H5Dcreate2(file_id,data_restart->data_name[i+4],H5T_NATIVE_INT,dataspace_id,H5P_DEFAULT,H5P_DEFAULT,H5P_DEFAULT);
    status       = H5Dwrite(dataset_id, H5T_NATIVE_INT, H5S_ALL, H5S_ALL, H5P_DEFAULT, &data_restart->int_array[i]);
  if (status < 0) *data_restart->stop_gysela = 1;
    status       = H5Sclose(dataspace_id);
  if (status < 0) *data_restart->stop_gysela = 1;
    status       = H5Dclose(dataset_id);
  if (status < 0) *data_restart->stop_gysela = 1;
  }
// 4D saving
// recuperation of fields dimensions
  for (i=0;i<4;i++)
    dim_4D[i] = data_restart->dim_4D[i];
  dataspace_id   = H5Screate_simple(4,dim_4D,NULL);
  property       = H5Pcreate(H5P_DATASET_CREATE);
  status         = H5Pset_chunk(property,4,dim_4D);
  if (status < 0) *data_restart->stop_gysela = 1;
  status         = H5Pset_deflate(property,cmpr);
  if (status < 0) *data_restart->stop_gysela = 1;
  status         = H5Pset_shuffle(property);
  if (status < 0) *data_restart->stop_gysela = 1;
  if (*data_restart->checksum == 1){
    status = H5Pset_fletcher32(property);
    if (status < 0) *data_restart->stop_gysela = 1;
  }
  dataset_id     = H5Dcreate2(file_id,data_restart->data_name[11],H5T_NATIVE_DOUBLE,dataspace_id,H5P_DEFAULT,property,H5P_DEFAULT);
  status         = H5Dwrite(dataset_id, H5T_NATIVE_DOUBLE, H5S_ALL, H5S_ALL, H5P_DEFAULT, data_restart->field_4D);
  if (status < 0) *data_restart->stop_gysela = 1;
  status         = H5Pclose(property);
  if (status < 0) *data_restart->stop_gysela = 1;
  status         = H5Sclose(dataspace_id);
  if (status < 0) *data_restart->stop_gysela = 1;
  status         = H5Dclose(dataset_id);
  if (status < 0) *data_restart->stop_gysela = 1;
// 3D saving, if collision
  if (*data_restart->collision == 1) {
    for (i=0;i<3;i++)
      dim_3D[i] = data_restart->dim_3D[i];
    for (i=0;i<2;i++){
      dataspace_id   = H5Screate_simple(3,dim_3D,NULL);
      property       = H5Pcreate(H5P_DATASET_CREATE);
      status         = H5Pset_chunk(property,3,dim_3D);
      if (status < 0) *data_restart->stop_gysela = 1;
      status         = H5Pset_deflate(property,cmpr);
      if (status < 0) *data_restart->stop_gysela = 1;
      if (*data_restart->checksum == 1){
        status = H5Pset_fletcher32(property);
        if (status < 0) *data_restart->stop_gysela = 1;
      }
      status         = H5Pset_shuffle(property);
      if (status < 0) *data_restart->stop_gysela = 1;
      dataset_id     = H5Dcreate2(file_id,data_restart->data_name[12+i],H5T_NATIVE_DOUBLE,dataspace_id,H5P_DEFAULT,property,H5P_DEFAULT);
      status = dataset_id;
      if (status < 0) *data_restart->stop_gysela = 1;
      switch (i){
        case 0 :
          status     = H5Dwrite(dataset_id, H5T_NATIVE_DOUBLE, H5S_ALL, H5S_ALL, H5P_DEFAULT, data_restart->field_3D_Tmean);
          break;
        case 1 :
          status     = H5Dwrite(dataset_id, H5T_NATIVE_DOUBLE, H5S_ALL, H5S_ALL, H5P_DEFAULT, data_restart->field_3D_Vmean);
          break;
      }
      if (status < 0) *data_restart->stop_gysela = 1;
      status         = H5Pclose(property);
      if (status < 0) *data_restart->stop_gysela = 1;
      status         = H5Sclose(dataspace_id);
      if (status < 0) *data_restart->stop_gysela = 1;
      status         = H5Dclose(dataset_id);
      if (status < 0) *data_restart->stop_gysela = 1;
    }
// 2D saving, if energy transfer
    if (*data_restart->species_transfer == 1) {
      for (i=0;i<2;i++)
        dim_2D[i] = data_restart->dim_2D[i];
      for (i=0;i<4;i++){
        dataspace_id   = H5Screate_simple(2,dim_2D,NULL);
        property       = H5Pcreate(H5P_DATASET_CREATE);
        status         = H5Pset_chunk(property,2,dim_2D);
        if (status < 0) *data_restart->stop_gysela = 1;
        status         = H5Pset_deflate(property,cmpr);
        if (status < 0) *data_restart->stop_gysela = 1;
        if (*data_restart->checksum == 1){
          status = H5Pset_fletcher32(property);
          if (status < 0) *data_restart->stop_gysela = 1;
        }
        status         = H5Pset_shuffle(property);
        if (status < 0) *data_restart->stop_gysela = 1;
        dataset_id     = H5Dcreate2(file_id,data_restart->data_name[14+i],H5T_NATIVE_DOUBLE,dataspace_id,H5P_DEFAULT,property,H5P_DEFAULT);
        status = dataset_id;
        if (status < 0) *data_restart->stop_gysela = 1;
        switch (i){
          case 0 :
            status     = H5Dwrite(dataset_id, H5T_NATIVE_DOUBLE, H5S_ALL, H5S_ALL, H5P_DEFAULT, data_restart->field_2D_Ns_FSavg);
            break;
          case 1 :
            status     = H5Dwrite(dataset_id, H5T_NATIVE_DOUBLE, H5S_ALL, H5S_ALL, H5P_DEFAULT, data_restart->field_2D_Vs_FSavg);
            break;
          case 2 :
            status     = H5Dwrite(dataset_id, H5T_NATIVE_DOUBLE, H5S_ALL, H5S_ALL, H5P_DEFAULT, data_restart->field_2D_Ts_FSavg);
            break;
          case 3 :
            status     = H5Dwrite(dataset_id, H5T_NATIVE_DOUBLE, H5S_ALL, H5S_ALL, H5P_DEFAULT, data_restart->field_2D_Tscoll_FSavg);
            break;
        }
        if (status < 0) *data_restart->stop_gysela = 1;
        status         = H5Pclose(property);
        if (status < 0) *data_restart->stop_gysela = 1;
        status         = H5Sclose(dataspace_id);
        if (status < 0) *data_restart->stop_gysela = 1;
        status         = H5Dclose(dataset_id);
        if (status < 0) *data_restart->stop_gysela = 1;
      }
    }
  }
//     close file
  status         = H5Fclose(file_id);
  if (status < 0) *data_restart->stop_gysela = 1;
#ifdef TIMER
  gettimeofday(&tv2, (struct timezone*)0);
  dt2 = (((tv2).tv_sec)%MAXSEC  + 0.000001*(tv2).tv_usec) * MSRATE ;
  nb_periods  = dt2 - dt1;
  if (dt2 < dt1) nb_periods = nb_periods + MSRATE * MAXSEC;
  p_difftime = (float) nb_periods/MSRATE;
  fprintf(stderr,"in thread time : %f \n ", p_difftime);
#endif
  // unlock
  *data_restart->lock_f5d = 0;
  return;
}


void FORTSYM(asynch_restart_writing, ASYNCH_RESTART_WRITING)(double *field_4D, int nelements_2D[2], int nelements_3D[3],int nelements_4D[4], char filename[],  char data_name[18][100], int int_array[], double real_array[], int *collision, int *species_transfer, int *lock, int *checksum, int *stop_gysela, double *field_3D_Tmean, double *field_3D_Vmean, double *field_2D_Ns_FSavg, double *field_2D_Vs_FSavg, double *field_2D_Ts_FSavg, double *field_2D_Tscoll_FSavg) {
// local declaration :
  int i, j, rc;
#ifdef TIMER
  struct timeval tv;
  long long int dtt1, dtt2, nb_periods_t;
  double p_difftime_thread;
  gettimeofday(&tv, (struct timezone*)0);
  dtt1 = (((tv).tv_sec)%MAXSEC  + 0.000001*(tv).tv_usec) * MSRATE ;
#endif
  data_restart.field_4D         = field_4D                  ;
  data_restart.collision        = collision                 ;
  data_restart.species_transfer = species_transfer          ;
  data_restart.checksum         = checksum                  ;
  data_restart.stop_gysela      = stop_gysela               ;
// recuperate and transpose dimension of array from fortran 
// to perform the contiguous of writting in C formalism
  for (i=0;i<2;i++)
    data_restart.dim_2D[i]        = nelements_2D[1-i]       ;
  for (i=0;i<3;i++)
    data_restart.dim_3D[i]        = nelements_3D[2-i]       ;
  for (i=0;i<4;i++)
    data_restart.dim_4D[i]        = nelements_4D[3-i]       ;
// recuperate the filename
  data_restart.filename           = filename                ;
  for (i=0;i<12;i++)
    for (j=0;j<100;j++)
       data_restart.data_name[i][j]  = data_name[i][j]      ;
// real and integer data recuperation :
  for (i=0;i<4;i++)
    data_restart.real_array[i]  = real_array[i]             ;
  for (i=0;i<7;i++)
    data_restart.int_array[i]  = int_array[i]               ;
    data_restart.lock_f5d      = lock                       ;
  if (*collision == 1) {
     data_restart.field_3D_Tmean = field_3D_Tmean    ;
     data_restart.field_3D_Vmean = field_3D_Vmean    ;
// recuperate the dataset names for collision
     for (i=12;i<14;i++)
       for (j=0;j<100;j++)
         data_restart.data_name[i][j]  = data_name[i][j]    ;
     if (*species_transfer == 1) {
       data_restart.field_2D_Ns_FSavg      = field_2D_Ns_FSavg        ;
       data_restart.field_2D_Vs_FSavg      = field_2D_Vs_FSavg        ;
       data_restart.field_2D_Ts_FSavg      = field_2D_Ts_FSavg        ;
       data_restart.field_2D_Tscoll_FSavg  = field_2D_Tscoll_FSavg    ;
// recuperate the dataset names for energy transfer operator
       for (i=14;i<18;i++)
         for (j=0;j<100;j++)
           data_restart.data_name[i][j]  = data_name[i][j]    ;
     }
    }
  *lock               = 1                                   ;
// launch a new thread (threads_restart) with all the parameters (data_restart)
  rc = pthread_create(&threads_restart, 0, (void*(*)(void*)) &printrestart, &data_restart);
#ifdef TIMER
  gettimeofday(&tv, (struct timezone*)0);
  dtt2 = (((tv).tv_sec)%MAXSEC  + 0.000001*(tv).tv_usec) * MSRATE ;
  nb_periods_t  = dtt2 - dtt1;
  if (dtt2 < dtt1) nb_periods_t = nb_periods_t + MSRATE * MAXSEC;
  p_difftime_thread = (float) nb_periods_t/MSRATE;
  fprintf(stderr,"launching asynchrone writing : %f \n ", p_difftime_thread);
#endif  
  return;
}

void FORTSYM(thread_restart_barrier, ASYNCH_RESTART_BARRIER)(int *lock) {
  void *ret                          ;
// wait for threads_restart to finish to write
   (void)pthread_join(threads_restart,&ret) ; // else, wait for...
}

#else
void FORTSYM(asynch_restart_writing, ASYNCH_RESTART_WRITING)(double *field_4D, int nelements_2D[2], int nelements_3D[3],int nelements_4D[4], char filename[],  char data_name[18][100], int int_array[], double real_array[], int *collision, int *species_transfer, int *lock, int *checksum, int *stop_gysela, double *field_3D_Tmean, double *field_3D_Vmean, double *field_2D_Ns_FSavg, double *field_2D_Vs_FSavg, double *field_2D_Ts_FSavg, double *field_2D_Tscoll_FSavg) {}

void FORTSYM(thread_restart_barrier, ASYNCH_RESTART_BARRIER)(int *lock) {}
#endif
