#!/bin/bash

#**************************************************************
#  Copyright Euratom-CEA
#  Authors : 
#     Virginie Grandgirard (virginie.grandgirard@cea.fr)
#     Chantal Passeron (chantal.passeron@cea.fr)
#     Guillaume Latu (guillaume.latu@cea.fr)
#     Xavier Garbet (xavier.garbet@cea.fr)
#     Philippe Ghendrih (philippe.ghendrih@cea.fr)
#     Yanick Sarazin (yanick.sarazin@cea.fr)
#  
#  This code GYSELA (for GYrokinetic SEmi-LAgrangian) 
#  is a 5D gyrokinetic global full-f code for simulating 
#  the plasma turbulence in a tokamak.
#  
#  This software is governed by the CeCILL-B license 
#  under French law and abiding by the rules of distribution 
#  of free software.  You can  use, modify and redistribute 
#  the software under the terms of the CeCILL-B license as 
#  circulated by CEA, CNRS and INRIA at the following URL
#  "http://www.cecill.info". 
#**************************************************************

# F: Fortran type
F_l="logical"
F_c="character"
F_i="integer"
F_f="real(SKIND)"
F_d="real(RKIND)"

# H: HDF5 type constant
H_c="H5T_NATIVE_CHARACTER"
H_i="H5T_NATIVE_INTEGER"
H_f="H5T_NATIVE_REAL"
H_d="H5T_NATIVE_DOUBLE"

H5TYPES="c i f d"
ALLTYPES="${H5TYPES} l"

MAXDIM=7

cat<<EOF

#include "config-F.inc"

module HDF5_module
  
  use HDF5
  use mpiMod
  use prec_const
  
  implicit none
  
  private
  public:: HDF5_load, HDF5_save
  public:: HDF5_create, HDF5_open, HDF5_close
  
  interface HDF5_save
EOF
echo '    module procedure HDF5_save_s0d'
for T in ${ALLTYPES}; do # T: type
  if [ "${T}" != c ]; then # no 0D character saving, 0D character is used for string saving
    echo "    module procedure HDF5_save_${T}0d"
  fi
  for D in $(seq 1 ${MAXDIM}); do # D: dimensions of array
    echo "    module procedure HDF5_save_${T}${D}d"
    echo "    module procedure HDF5_save_nodim_${T}${D}d"
  done
done
cat<<EOF
  endinterface HDF5_save
  
  interface HDF5_load
EOF
for T in ${H5TYPES}; do # T: type 
  for D in $(seq 0 ${MAXDIM}); do # D: dimensions of array
    echo "    module procedure HDF5_load_${T}${D}d"
  done
done
cat<<EOF
  endinterface HDF5_load

contains

  !-------------------------------------------------------------------
  !> creates a new file and open it for HDF5 writing
  subroutine HDF5_create(filename, file_id, ierr)
    
    character(LEN=*), intent(in):: filename !< file name
    integer(HID_T), intent(out):: file_id !< file identifier
    integer, optional, intent(out):: ierr !< HDF5 error code

    integer :: ierr_HDF5
    
    ! ensure HDF5 is initialized
    call H5open_f(ierr_HDF5)
    
    call H5Fcreate_f(trim(filename)//char(0), H5F_ACC_TRUNC_F, &
            file_id, ierr_HDF5)
    
    if (present(ierr)) ierr = ierr_HDF5
    
  end subroutine HDF5_create
  !-------------------------------------------------------------------


  !-------------------------------------------------------------------
  !> opens an existring file for HDF5 reading
  subroutine HDF5_open(filename,file_id,ierr)
    
    character(LEN=*), intent(in):: filename !< file name
    integer(HID_T), intent(out):: file_id !< file identifier
    integer, optional, intent(out):: ierr !< HDF5 error code

    integer :: ierr_HDF5

    ! ensure HDF5 is initialized
    call H5open_f(ierr_HDF5)
    
    call H5Fopen_f(trim(filename)//char(0), H5F_ACC_RDONLY_F, &
            file_id, ierr_HDF5)
    
    if (present(ierr)) ierr = ierr_HDF5
    
  end subroutine HDF5_open
  !-------------------------------------------------------------------


  !-------------------------------------------------------------------
  !> closes a file opened for HDF5 reading or writing
  subroutine HDF5_close(file_id,ierr)
    
    integer(HID_T), intent(in):: file_id !< file identifier
    integer, optional, intent(out) :: ierr !< HDF5 error code
    
    integer :: ierr_HDF5
    
    call H5Fclose_f(file_id,ierr_HDF5)
    
    if (present(ierr)) ierr = ierr_HDF5
    
  end subroutine HDF5_close
  !-------------------------------------------------------------------
  
  !-------------------------------------------------------------------
  !> HDF5 saving for character string
  subroutine HDF5_save_s0d(file_id, data, dsetname)
    
    integer(HID_T), intent(in):: file_id !< file identifier
    character(LEN=*), intent(in):: data !< the actual data to save
    character(LEN=*), intent(in):: dsetname  !< dataset name
    
    integer:: ierr_HDF5 ! error flag
    integer(HID_T):: dataset ! dataset identifier
    integer(HID_T):: dataspace ! dataspace identifier
    integer(HID_T):: type_id ! type identifier
    
    CALL H5Tcopy_f(H5T_NATIVE_CHARACTER, type_id, ierr_HDF5)
    CALL H5Tset_size_f(type_id, int(len(data),SIZE_T), ierr_HDF5)
    CALL H5Screate_f(H5S_SCALAR_F, dataspace, ierr_HDF5)
    CALL H5Dcreate_f(file_id, dsetname, type_id, dataspace, dataset, &
            ierr_HDF5)
    CALL H5Dwrite_f(dataset, type_id, data, (/ 1_HSIZE_T /), ierr_HDF5)
    call H5Dclose_f(dataset, ierr_HDF5)
    call H5Sclose_f(dataspace, ierr_HDF5)
    
  end subroutine HDF5_save_s0d
  !-------------------------------------------------------------------
EOF




for D in $(seq 1 ${MAXDIM}); do # D: dimensions of array
  A="(:" # A: assumed shape array descriptor
  S="(dim1" # S: Array size
  I="(ii1" # I: indexed array
  if (( D > 1 )); then
    for N in $(seq 2 ${D}); do
      A="${A},:"
      S="${S},dim${N}"
      I="${I},ii${N}"
    done
  fi
  A="${A})"
  S="${S})"
  I="${I})"
  cat<<EOF
  
  !-------------------------------------------------------------------
  !> HDF5 logical saving with explicit array dimensions, forwards to
  !> character saving (.TRUE. => 'T', .FALSE. => 'F' )
  subroutine HDF5_save_l${D}d(file_id, ldata, &
EOF
  for X in $(seq 1 ${D}); do
    echo "          dim${X}, &"
  done
  cat<<EOF
          dsetname, checksum)
    
    integer(HID_T), intent(in):: file_id   ! file identifier
    logical, intent(IN):: ldata${A}
EOF
  for X in $(seq 1 ${D}); do
    echo "    integer, intent(in):: dim${X}"
  done
  cat<<EOF
    character(LEN=*), intent(in):: dsetname  !< dataset name
    integer, optional, intent(in):: checksum
    
    character, pointer:: cdata${A} ! the character data that corresponds
                                   ! to the original logical data
EOF
  for X in $(seq 1 ${D}); do
    echo "    integer:: ii${X}"
  done
  cat<<EOF
    
    allocate( cdata${S} )
    
EOF
  for X in $(seq 1 ${D}); do
    echo "    do ii${X}=1, dim${X}"
  done
  cat<<EOF
      if ( ldata${I} ) then
        cdata${I} = 'T'
      else
        cdata${I} = 'F'
      endif
EOF
  for X in $(seq 1 ${D}); do
    echo "    enddo"
  done
  cat<<EOF
    
    call HDF5_save_c${D}d(file_id, cdata, &
EOF
  for X in $(seq 1 ${D}); do
    echo "          dim${X}, &"
  done
  cat<<EOF
          dsetname, checksum)
    
    deallocate(cdata)
    
  end subroutine HDF5_save_l${D}d
  !-------------------------------------------------------------------
  
  !-------------------------------------------------------------------
  !> HDF5 logical saving with implicit array dimensions, forwards to
  !> character saving (.TRUE. => 'T', .FALSE. => 'F' )
  subroutine HDF5_save_nodim_l${D}d(file_id, data, dsetname, &
          checksum)
    
    integer(HID_T), intent(in):: file_id   ! file identifier
    logical, intent(IN):: data${A}
    character(LEN=*), intent(in):: dsetname  !< dataset name
    integer, optional, intent(in):: checksum
    
    call HDF5_save_l${D}d(file_id, data, &
EOF
      for X in $(seq 1 ${D}); do
        echo "          size(data, ${X}), &"
      done
      cat<<EOF
          dsetname, checksum)
    
  end subroutine HDF5_save_nodim_l${D}d
  !-------------------------------------------------------------------
EOF
done



for D in $(seq 1 ${MAXDIM}); do # D: dimensions of array
  A="(:" # A: assumed shape array descriptor
  if (( D > 1 )); then
    for N in $(seq 2 ${D} ); do
      A="${A},:"
    done
  fi
  A="${A})"
  for T in ${H5TYPES}; do # T: type
    F="$(eval "echo \${F_${T}}")" # F: Fortran type
    H="$(eval "echo \${H_${T}}")" # H: HDF5 type constant
    cat<<EOF
  
  !-------------------------------------------------------------------
  !> HDF5 saving with explicit array dimensions
  subroutine HDF5_save_${T}${D}d(file_id, data, &
EOF
    for X in $(seq 1 ${D}); do
      echo "          dim${X}, &"
    done
    cat<<EOF
          dsetname, checksum)
    
    use globals, only: cmpr
    
    integer(HID_T), intent(in):: file_id   ! file identifier
    ${F}, intent(IN):: data${A}
EOF
    for X in $(seq 1 ${D}); do
      echo "    integer, intent(in):: dim${X}"
    done
    cat<<EOF
    character(LEN=*), intent(in):: dsetname  !< dataset name
    integer, optional, intent(in):: checksum
    
    integer:: ierr_HDF5  ! error flag
    integer:: max_dim_size
    integer(HSIZE_T), dimension(${D}):: dim_chunk ! chunk dimensions
    integer(HSIZE_T), dimension(${D}):: dim ! dataset dimensions
    integer(HID_T):: dataset ! dataset identifier
    integer(HID_T):: dataspace ! dataspace identifier
    integer(HID_T):: property ! Property list identifier 

    !*** Create and initialize dataspaces for datasets ***
EOF
    for X in $(seq 1 ${D}); do
      echo "    dim(${X}) = dim${X}"
    done
    cat<<EOF
    dim_chunk = dim
    do while( kind(data)*product(dim_chunk) >= (2_8**32) )
      max_dim_size = maxloc(dim_chunk, 1) !! 2**32 = 4Giga
      dim_chunk(max_dim_size) = dim_chunk(max_dim_size)/2
    enddo
    
    call H5Screate_simple_f(${D}, dim_chunk, dataspace, ierr_HDF5)
    
    !*** Creates a new property dataset ***
    call H5Pcreate_f(H5P_DATASET_CREATE_F, property, ierr_HDF5)
    if ( cmpr /= -1 ) then
      call H5Pset_chunk_f(property, ${D}, dim, ierr_HDF5)
    else if ( present(checksum) .or. &
            kind(data)*product(dim)>=(2_8**32) ) then
      if ( checksum == 1 .or. kind(data)*product(dim)>=(2_8**32) ) then
          call H5Pset_chunk_f(property, ${D}, dim, ierr_HDF5)
        else
          call H5Pset_layout_f(property, H5D_CONTIGUOUS_F, ierr_HDF5)
        end if
      else
        call H5Pset_layout_f(property, H5D_CONTIGUOUS_F, ierr_HDF5)
      end if
    if ( cmpr /= -1 ) then
        call H5Pset_deflate_f(property, cmpr, ierr_HDF5)
    end if
    if ( present(checksum) ) then
      if ( checksum == 1 ) call H5Pset_fletcher32_f(property, ierr_HDF5)
    endif

    call H5Dcreate_f(file_id, trim(dsetname), ${H}, dataspace, &
            dataset, ierr_HDF5, property)

    call H5Dwrite_f(dataset, ${H}, data, dim, ierr_HDF5)
    
    call H5Pclose_f(property, ierr_HDF5)
    call H5Sclose_f(dataspace, ierr_HDF5)
    call H5Dclose_f(dataset, ierr_HDF5)
    
  end subroutine HDF5_save_${T}${D}d
  !-------------------------------------------------------------------
  
  !-------------------------------------------------------------------
  !> HDF5 saving with implicit array dimensions
  subroutine HDF5_save_nodim_${T}${D}d(file_id, data, dsetname, &
          checksum)
    
    integer(HID_T), intent(in):: file_id   ! file identifier
    ${F}, intent(IN):: data${A}
    character(LEN=*), intent(in):: dsetname  !< dataset name
    integer, optional, intent(in):: checksum
    
    call HDF5_save_${T}${D}d(file_id, data, &
EOF
    for X in $(seq 1 ${D}); do
      echo "          size(data, ${X}), &"
    done
    cat<<EOF
          dsetname, checksum)
    
  end subroutine HDF5_save_nodim_${T}${D}d
  !-------------------------------------------------------------------
EOF
  done
done


for T in ${ALLTYPES}; do # T: type
  F="$(eval "echo \${F_${T}}")" # F: Fortran type
  if [ "${T}" != c ]; then
    cat<<EOF
  
  !-------------------------------------------------------------------
  !> HDF5 scalar saving
  subroutine HDF5_save_${T}0d(file_id, data, dsetname)
    
    integer(HID_T), intent(in):: file_id !< file identifier
    ${F}, intent(IN):: data
    character(LEN=*), intent(in):: dsetname !< dataset name
    
    call HDF5_save_${T}1d(file_id, (/ data /), 1, &
            dsetname)
    
  end subroutine HDF5_save_${T}0d
  !-------------------------------------------------------------------
EOF
  fi
done


for D in $(seq 0 ${MAXDIM}); do # D: dimensions of array
  A=''
  if (( D > 0 )); then
    A="(:" # A: assumed shape array descriptor
    if (( D > 1 )); then
      for N in $(seq 2 ${D} ); do
        A="${A},:"
      done
    fi
    A="${A})"
  fi
  for T in ${H5TYPES}; do # T: type
    F="$(eval "echo \${F_${T}}")" # F: Fortran type
    H="$(eval "echo \${H_${T}}")" # H: HDF5 type constant
    cat<<EOF
  
  !-------------------------------------------------------------------
  !> HDF5 loading
  subroutine HDF5_load_${T}${D}d(file_id, data, dsetname, ierr, &
          var_exist)
    
    integer(HID_T), intent(in):: file_id !< file identifier
    ${F}, intent(INOUT):: data${A}
    character(LEN=*), intent(in):: dsetname !< dataset name
    integer, optional, intent(out):: ierr
    logical, optional, intent(out):: var_exist

    integer:: ierr_HDF5 ! error flag
    integer(HSIZE_T), dimension($((D>0?D:1))):: dim ! dataset dimensions
    integer(HID_T):: dataset ! dataset identifier
    logical:: var_exist_HDF5
    
    call H5Lexists_f(file_id, trim(dsetname), var_exist_HDF5, &
            ierr_HDF5)
    
    if ( var_exist_HDF5 ) then
      call H5Dopen_f(file_id, trim(dsetname), dataset, ierr_HDF5)
EOF
    if (( D == 0 )); then
      echo '      dim(1) = 1_HSIZE_T'
    else
      echo '      dim = shape(data)'
    fi
    cat<<EOF
      call H5Dread_f(dataset, ${H}, data, dim, ierr_HDF5)
      
      call H5Dclose_f(dataset, ierr_HDF5)
    elseif ( .not. present(var_exist) ) then
      write(*,'(A,A,A)') 'The variable ',trim(dsetname), &
        ' does not exist in the HDF5 file'
      stop
    end if
    
    if (present(ierr)) ierr = ierr_HDF5
    if (present(var_exist)) var_exist = var_exist_HDF5
    
  end subroutine HDF5_load_${T}${D}d
  !-------------------------------------------------------------------
EOF
  done
done
cat<<EOF

endmodule HDF5_module

EOF
