!**************************************************************
!  Copyright Euratom-CEA
!  Authors : 
!     Rozar Fabien (fabien.rozar@cea.fr)
!
!  This code GYSELA (for GYrokinetic Semi-LAgrangian) 
!  is a 5D gyrokinetic global full-f code for simulating 
!  the plasma turbulence in a tokamak.
!  
!  This software is governed by the CeCILL-B license 
!  under French law and abiding by the rules of distribution 
!  of free software.  You can  use, modify and redistribute 
!  the software under the terms of the CeCILL-B license as 
!  circulated by CEA, CNRS and INRIA at the following URL
!  "http://www.cecill.info". 
!**************************************************************
  
#include "config-F.inc"
#include "system_tools/iounit.inc"
  
!---------------------------------------------
! file : wrap_alloc.F90
! date : 15/06/2012
!  manage allocation of array along the code
!---------------------------------------------
module wrap_alloc_module
  use prec_const
  use globals, only : pglobal_id, memory_test, mem_dyn_write, &
       enable_dyn_trace
  ! use globals, only : pglobal_id, mem_dyn_write
  use prec_const

  implicit none

  ! used for memory size calculation
  integer(KIND=8), public :: nb_dyn_allocate  = 0
  integer(KIND=8), public :: max_dyn_allocate = 0

  !*** interface definition for loc_take ***
  interface ftr_alloc
     module procedure ftr_alloc_1d_i, ftr_alloc_1d_d, ftr_alloc_1d_c, &
          ftr_alloc_2d_c, ftr_alloc_2d_d, &
          ftr_alloc_2d_i, ftr_alloc_3d_c, ftr_alloc_3d_d, ftr_alloc_3d_i, &
          ftr_alloc_4d_c, ftr_alloc_4d_d, ftr_alloc_5d_d
  end interface

  !*** interface definition for loc_drop ***
  interface ftr_dealloc
     module procedure ftr_dealloc_1d_i, ftr_dealloc_1d_d, ftr_dealloc_1d_c, &
          ftr_dealloc_2d_c, ftr_dealloc_2d_d, &
          ftr_dealloc_2d_i, ftr_dealloc_3d_c, ftr_dealloc_3d_d, ftr_dealloc_3d_i, &
          ftr_dealloc_4d_c, ftr_dealloc_4d_d, ftr_dealloc_5d_d
  end interface

  !*** surdefinition for temporary allocation ***
  interface temp_allocate
     module procedure temp_allocate1d_i, temp_allocate1d_d, &
          temp_allocate1d_c, temp_allocate2d_i, temp_allocate2d_d, &
          temp_allocate2d_c, temp_allocate3d_i, temp_allocate3d_d, &
          temp_allocate3d_c, temp_allocate4d_d, temp_allocate4d_c, &
          temp_allocate5d_d   
  end interface

  !*** surdefinition for temporary deallocation ***
  interface temp_deallocate
     module procedure temp_deallocate1d_i, temp_deallocate1d_d, &
          temp_deallocate1d_c, temp_deallocate2d_i, temp_deallocate2d_d, &
          temp_deallocate2d_c, temp_deallocate3d_d, temp_deallocate3d_c, &
          temp_deallocate3d_i, temp_deallocate4d_d, temp_deallocate4d_c, &
          temp_deallocate5d_d
  end interface

  character(LEN=100), parameter          :: &
       dyn_memory_filename = "gysela_mem.out"
  character(LEN=100), parameter          :: &
       memory_leak_filename = "gysela_mem_leak.out"
  integer           , save               :: uout_dyn_mem  = 31
  integer           , save               :: uout_mem_leak = 32

  !******************************
  contains
  !******************************  
  !---------------------------------------- 
  ! constructor for dynamic trace
  !----------------------------------------
  subroutine new_dyn_allocate()

    use Io_Unit_module, only: IU_open, IU_close

    LOGICAL :: file_exists
    if (pglobal_id.eq.0) then       
       ! file_exists will be TRUE if the file
       ! exists and FALSE otherwise
       INQUIRE(FILE=dyn_memory_filename, EXIST=file_exists)   

       ! If the file "dyn_memory_filename" doesn't exist
       ! it is create
       if (.not.file_exists) then
          call IU_open(uout_dyn_mem, file = dyn_memory_filename, status = 'UNKNOWN', &
               form = 'FORMATTED')        
          write(uout_dyn_mem,*) ' '
          write(uout_dyn_mem,*) &
               '*****************************************************************'
          write(uout_dyn_mem,*) &
               '***           GYSELA dynamic memory required in Bytes         ***'
          write(uout_dyn_mem,*) &
               '*****************************************************************'
          write(uout_dyn_mem,*) ' '
          call IU_close(uout_dyn_mem)
       end if
    end if
  end subroutine new_dyn_allocate

  !----------------------------------------
  ! constructor
  !----------------------------------------
  subroutine new_trace_mem_leak()

    use Io_Unit_module, only: IU_open, IU_close

    if (pglobal_id.eq.0) then
       call IU_open(uout_mem_leak, file = memory_leak_filename, status = 'UNKNOWN', &
            form = 'FORMATTED')
       write(uout_mem_leak,*) ' '
       write(uout_mem_leak,*) &
            '*********************************************************'
       write(uout_mem_leak,*) &
            '***           GYSELA memory required in Bytes         ***'
       write(uout_mem_leak,*) &
            '*********************************************************'
       write(uout_mem_leak,*) ' '
       call IU_close(uout_mem_leak)
    end if
  end subroutine new_trace_mem_leak


  !******************************  
  ! TRACE BLOCK  
  !******************************  
  !-------------------------------------------
  ! Write memory in the file dyn_memory_filename
  !-------------------------------------------
  subroutine write_dyn_memory(size_array,type_name,var_name, arg_prefix)

    use Io_Unit_module, only: IU_open, IU_close

    integer(KIND=8)    , intent(in) :: size_array
    character*(*), intent(in) :: type_name
    character*(*), intent(in) :: var_name
    character*(*), optional, intent(in) :: arg_prefix

    character(len=4) :: prefix

    if (present(arg_prefix)) then
       prefix = arg_prefix
    else 
       prefix = '->'
    endif

    if (mem_dyn_write.and.enable_dyn_trace) then
       if (pglobal_id.eq.0) then
          call IU_open(uout_dyn_mem, file = dyn_memory_filename, status = 'OLD', &
               position = 'APPEND', form = 'FORMATTED')
          write(uout_dyn_mem,'(A5,A6,I12,A12,A17,A90)') &
               prefix, '  +  ', size_array,' Bytes for ',type_name//" ",adjustr(var_name)
          UFLUSH(uout_dyn_mem)
          call IU_close(uout_dyn_mem)
       end if
    end if
  end subroutine write_dyn_memory


  !-------------------------------------------
  ! Write memory deallocation in the file dyn_memory_filename
  !-------------------------------------------
  subroutine write_dyn_dealloc_memory(size_array,type_name,var_name, arg_prefix)

    use Io_Unit_module, only: IU_open, IU_close

    integer(KIND=8)    , intent(in) :: size_array
    character*(*), intent(in) :: type_name
    character*(*), intent(in) :: var_name
    character*(*), optional, intent(in) :: arg_prefix

    character(len=4) :: prefix

    if (present(arg_prefix)) then
       prefix = arg_prefix
    else 
       prefix = '->'
    endif

    if (mem_dyn_write.and.enable_dyn_trace) then
       if (pglobal_id.eq.0) then
          call IU_open(uout_dyn_mem, file = dyn_memory_filename, status = 'OLD', &
               position = 'APPEND', form = 'FORMATTED')
          write(uout_dyn_mem,'(A5,A6,I12,A12,A17,A90)') &
               prefix, '  -  ', &
               size_array,' Bytes for ',type_name//" ",adjustr(var_name)
          UFLUSH(uout_dyn_mem)
          call IU_close(uout_dyn_mem)
       end if
    end if
  end subroutine write_dyn_dealloc_memory


  !-------------------------------------------
  ! Write the addresse of a variable at an allocation to check which the memory leak
  !-------------------------------------------
  subroutine write_alloc_mem_leak(var_name, loc, size_array)

    use Io_Unit_module, only: IU_open, IU_close

    character*(*), intent(in) :: var_name
    integer(8),    intent(in) :: loc
    integer(KIND=8)    , intent(in) :: size_array

    if (pglobal_id.eq.0) then
       call IU_open(uout_mem_leak, file = memory_leak_filename, status = 'OLD', &
            position = 'APPEND', form = 'FORMATTED')
       write(uout_mem_leak,*) "allocate : ", var_name, " : ", loc, " : ", size_array
       call IU_close(uout_mem_leak)
    end if
  end subroutine write_alloc_mem_leak


  !-------------------------------------------
  ! Write the addresse of a variable at an deallocation to check which the memory leak
  !-------------------------------------------
  subroutine write_dealloc_mem_leak(loc)

    use Io_Unit_module, only: IU_open, IU_close

    integer(8),    intent(in) :: loc

    if (pglobal_id.eq.0) then
       call IU_open(uout_mem_leak, file = memory_leak_filename, status = 'OLD', &
            position = 'APPEND', form = 'FORMATTED')
       write(uout_mem_leak,*) "deallocate : ", loc
       call IU_close(uout_mem_leak)
    end if
  end subroutine write_dealloc_mem_leak

  !-------------------------------------------
  ! Write the name of a subroutine in the trace file in the beginning of a subroutine
  !-------------------------------------------
  subroutine write_dyn_balise_open(chaine)

    use Io_Unit_module, only: IU_open, IU_close

    character*(*), intent(in)  :: chaine

    if (mem_dyn_write.and.enable_dyn_trace) then
       if (pglobal_id.eq.0) then
          call IU_open(uout_dyn_mem, file = dyn_memory_filename, status = 'OLD', &
               position = 'APPEND', form = 'FORMATTED')
          write(uout_dyn_mem,'(A3,A)') ">> ", chaine
          UFLUSH(uout_dyn_mem)
          call IU_close(uout_dyn_mem)
       end if
    end if
  end subroutine write_dyn_balise_open


  !-------------------------------------------
  ! Write the name of a subroutine in the trace file in the ending of a subroutine
  !-------------------------------------------
  subroutine write_dyn_balise_close(chaine)

    use Io_Unit_module, only: IU_open, IU_close

    character*(*), intent(in)  :: chaine

    if (mem_dyn_write.and.enable_dyn_trace) then
       if (pglobal_id.eq.0) then
          call IU_open(uout_dyn_mem, file = dyn_memory_filename, status = 'OLD', &
               position = 'APPEND', form = 'FORMATTED')
          write(uout_dyn_mem,'(A3,A)') "<< ", chaine
          UFLUSH(uout_dyn_mem)
          call IU_close(uout_dyn_mem)
       end if
    end if
  end subroutine write_dyn_balise_close


  ! Delete whitespaces in the expression of the array size
  character(256) function sweep_blanks(in_str)
    character(*), intent(in) :: in_str
    character(256) :: out_str
    character :: ch
    integer :: j

    out_str = " "
    do j=1, len_trim(in_str)
       ch = in_str(j:j)
       if (ch .ne. " ") then
          out_str = trim(out_str) // ch
       endif
       sweep_blanks = out_str 
    end do
  end function sweep_blanks


  ! Write the expression associated with the allocation
  subroutine write_trm_alloc(size_array,var_type,var_name,expr,file_str,line_str)

    use Io_Unit_module, only: IU_open, IU_close

    integer(KIND=8),           intent(in) :: size_array
    character*(*),             intent(in) :: var_type
    character*(*),             intent(in) :: var_name
    character*(*)                         :: expr
    character*(*),             intent(in) :: file_str
    character*(*),             intent(in) :: line_str

    if (mem_dyn_write.and.enable_dyn_trace) then
       if (pglobal_id.eq.0) then
          call IU_open(uout_dyn_mem, file = dyn_memory_filename, status = 'OLD', &
               position = 'APPEND', form = 'FORMATTED')
          expr=sweep_blanks(expr)
          write(uout_dyn_mem,'(A11,I14,A1,A,A1,A12,A1,A48,A1,A,A1,A6)') &
               'loc_alloc: ', size_array, " ", expr, " ", var_type, " ", var_name, " ", file_str, " ", line_str
          UFLUSH(uout_dyn_mem)
          call IU_close(uout_dyn_mem)
       end if
    end if
  end subroutine write_trm_alloc

  !----------------------------------------
  ! trm_alloc-block
  !----------------------------------------
  subroutine ftr_alloc_5d_d(array, dim1, dim2, &
       dim3, dim4, dim5, dim6, dim7, dim8, dim9, dim10, array_name, file_str, line, &
       dim1_str, dim2_str, dim3_str, dim4_str, dim5_str, dim6_str, &
       dim7_str, dim8_str, dim9_str, dim10_str, pfillzero)
    real(RKIND), dimension(:,:,:,:,:), pointer, &
         intent(inout) :: array
    integer         , intent(in) :: dim1,dim2,dim3,dim4,dim5,dim6,dim7,dim8,dim9,dim10
    character(len=*), intent(in) :: dim1_str,dim2_str,dim3_str,dim4_str,dim5_str,dim6_str,dim7_str,dim8_str,dim9_str,dim10_str
    character*(*)   , intent(in) :: file_str
    integer         , intent(in) :: line
    character*(*)   , intent(in) :: array_name
    logical         , intent(in) :: pfillzero

    character(len=8) :: line_str
    integer(KIND=8) :: size_array 
    integer   :: err

    logical          :: fillzero 

    write(line_str, '(i5)') line

    fillzero = pfillzero

    size_array = (dim2-dim1+1_8) * &
         (dim4-dim3+1_8) * &
         (dim6-dim5+1_8) * &
         (dim8-dim7+1_8) * &
         (dim10-dim9+1_8)
    size_array = size_array * KIND(0._RKIND)

    call write_trm_alloc(size_array, &
         "5D double", array_name,"("//dim2_str//"-"//dim1_str//"+1)*("// & 
         dim4_str//"-"//dim3_str//"+1)*("//dim6_str//"-"//dim5_str//"+1)*("// &
         dim8_str//"-"//dim7_str//"+1)*("//dim10_str//"-"//dim9_str//"+1)", &
         file_str,line_str)

    call temp_allocate(array, &
         dim1, dim2, & 
         dim3, dim4, & 
         dim5, dim6, & 
         dim7, dim8, & 
         dim9, dim10, & 
         trim(adjustl(adjustr(array_name)//" "//file_str//" "//line_str)), &
         oth_prefix=.true.,pfillzero=fillzero)
  end subroutine ftr_alloc_5d_d

  subroutine ftr_alloc_4d_d(array, dim1, dim2, &
       dim3, dim4, dim5, dim6, dim7, dim8, array_name, file_str, line, &
       dim1_str, dim2_str, dim3_str, dim4_str, dim5_str, dim6_str, &
       dim7_str, dim8_str, pfillzero)
    real(RKIND), dimension(:,:,:,:), pointer, &
         intent(inout) :: array
    integer         , intent(in) :: dim1,dim2,dim3,dim4,dim5,dim6,dim7,dim8
    character(len=*), intent(in) :: dim1_str,dim2_str,dim3_str,dim4_str,dim5_str,dim6_str,dim7_str,dim8_str
    character*(*)   , intent(in) :: file_str
    integer         , intent(in) :: line
    character*(*)   , intent(in) :: array_name
    logical         , intent(in) :: pfillzero

    character(len=8) :: line_str
    integer(KIND=8)   :: size_array 
    logical          :: fillzero 

    write(line_str, '(i5)') line

    fillzero = pfillzero

    size_array = (dim2-dim1+1_8) * &
         (dim4-dim3+1_8) * &
         (dim6-dim5+1_8) * &
         (dim8-dim7+1_8)
    size_array = size_array * KIND(0.0D0)

    call write_trm_alloc(size_array, &
         "4D double", array_name,"("//dim2_str//"-"//dim1_str//"+1)*("// & 
         dim4_str//"-"//dim3_str//"+1)*("//dim6_str//"-"//dim5_str//"+1)*("//dim8_str//"-"//dim7_str//"+1)", &
         file_str,line_str)

    call temp_allocate(array, &
         dim1, dim2, & 
         dim3, dim4, & 
         dim5, dim6, & 
         dim7, dim8, & 
         trim(adjustl(adjustr(array_name)//" "//file_str//" "//line_str)), &
         oth_prefix=.true.,pfillzero=fillzero)
  end subroutine ftr_alloc_4d_d

  subroutine ftr_alloc_4d_c(array, dim1, dim2, &
       dim3, dim4, dim5, dim6, dim7, dim8, array_name, file_str, line, &
       dim1_str, dim2_str, dim3_str, dim4_str, dim5_str, dim6_str, &
       dim7_str, dim8_str, pfillzero)
    complex(CKIND), dimension(:,:,:,:), pointer, &
         intent(inout) :: array
    integer         , intent(in) :: dim1,dim2,dim3,dim4,dim5,dim6,dim7,dim8
    character(len=*), intent(in) :: dim1_str,dim2_str,dim3_str,dim4_str,dim5_str,dim6_str,dim7_str,dim8_str
    character*(*)   , intent(in) :: file_str
    integer         , intent(in) :: line
    character*(*)   , intent(in) :: array_name
    logical         , intent(in) :: pfillzero

    character(len=8) :: line_str
    logical          :: fillzero 
    integer(KIND=8)   :: size_array 

    write(line_str, '(i5)') line

    fillzero = pfillzero

    size_array = (dim2-dim1+1_8) * &
         (dim4-dim3+1_8) * &
         (dim6-dim5+1_8) * &
         (dim8-dim7+1_8)
    size_array = size_array * 2 * KIND(0.0D0)

    call write_trm_alloc(size_array, &
         "4D complex", array_name,"("//dim2_str//"-"//dim1_str//"+1)*("// & 
         dim4_str//"-"//dim3_str//"+1)*("//dim6_str//"-"//dim5_str//"+1)*("//dim8_str//"-"//dim7_str//"+1)", &
         file_str,line_str)

    call temp_allocate(array, &
         dim1, dim2, & 
         dim3, dim4, & 
         dim5, dim6, & 
         dim7, dim8, & 
         trim(adjustl(adjustr(array_name)//" "//file_str//" "//line_str)), &
         oth_prefix=.true.,pfillzero=fillzero)
  end subroutine ftr_alloc_4d_c

  subroutine ftr_alloc_3d_d(array, dim1, dim2, &
       dim3, dim4, dim5, dim6, array_name, file_str, line, &
       dim1_str, dim2_str, dim3_str, dim4_str, dim5_str, dim6_str, pfillzero)
    real(RKIND), dimension(:,:,:), pointer, &
         intent(inout)    :: array
    integer         , intent(in) :: dim1,dim2,dim3,dim4,dim5,dim6
    character(len=*), intent(in) :: dim1_str,dim2_str,dim3_str,dim4_str,dim5_str,dim6_str
    character*(*)   , intent(in) :: file_str
    integer         , intent(in) :: line
    character*(*)   , intent(in) :: array_name
    logical         , intent(in) :: pfillzero

    character(len=8) :: line_str
    logical          :: fillzero 
    integer(KIND=8)   :: size_array 

    write(line_str, '(i5)') line

    fillzero = pfillzero

    size_array = (dim2-dim1+1_8) * &
         (dim4-dim3+1_8) * &
         (dim6-dim5+1_8)
    size_array = size_array * KIND(0.0D0)

    call write_trm_alloc(size_array, &
         "3D double", array_name,"("//dim2_str//"-"//dim1_str//"+1)*("// & 
         dim4_str//"-"//dim3_str//"+1)*("//dim6_str//"-"//dim5_str//"+1)", &
         file_str,line_str)

    call temp_allocate(array, &
         dim1, dim2, & 
         dim3, dim4, & 
         dim5, dim6, & 
         trim(adjustl(adjustr(array_name)//" "//file_str//" "//line_str)), &
         oth_prefix=.true.,pfillzero=fillzero)
  end subroutine ftr_alloc_3d_d

  subroutine ftr_alloc_3d_c(array, dim1, dim2, &
       dim3, dim4, dim5, dim6, array_name, file_str, line, &
       dim1_str, dim2_str, dim3_str, dim4_str, dim5_str, dim6_str, pfillzero)
    complex(CKIND), dimension(:,:,:), pointer, &
         intent(inout)    :: array
    integer         , intent(in) :: dim1,dim2,dim3,dim4,dim5,dim6
    character(len=*), intent(in) :: dim1_str,dim2_str,dim3_str,dim4_str,dim5_str,dim6_str
    character*(*)   , intent(in) :: file_str
    integer         , intent(in) :: line
    character*(*)   , intent(in) :: array_name
    logical         , intent(in) :: pfillzero

    character(len=8) :: line_str
    logical          :: fillzero 
    integer(KIND=8)   :: size_array 

    write(line_str, '(i5)') line

    fillzero = pfillzero

    size_array = (dim2-dim1+1_8) * &
         (dim4-dim3+1_8) * &
         (dim6-dim5+1_8)
    size_array = size_array * 2 * KIND(0.0D0)

    call write_trm_alloc(size_array, &
         "3D complex", array_name,"("//dim2_str//"-"//dim1_str//"+1)*("// & 
         dim4_str//"-"//dim3_str//"+1)*("//dim6_str//"-"//dim5_str//"+1)", &
         file_str,line_str)

    call temp_allocate(array, &
         dim1, dim2, & 
         dim3, dim4, & 
         dim5, dim6, & 
         trim(adjustl(adjustr(array_name)//" "//file_str//" "//line_str)), &
         oth_prefix=.true.,pfillzero=fillzero)
  end subroutine ftr_alloc_3d_c

  subroutine ftr_alloc_3d_i(array, dim1, dim2, &
       dim3, dim4, dim5, dim6, array_name, file_str, line, &
       dim1_str, dim2_str, dim3_str, dim4_str, dim5_str, dim6_str,pfillzero)
    integer, dimension(:,:,:), pointer, &
         intent(inout)    :: array
    integer         , intent(in) :: dim1,dim2,dim3,dim4,dim5,dim6
    character(len=*), intent(in) :: dim1_str,dim2_str,dim3_str,dim4_str,dim5_str,dim6_str
    character*(*)   , intent(in) :: file_str
    integer         , intent(in) :: line
    character*(*)   , intent(in) :: array_name
    logical         , intent(in) :: pfillzero

    character(len=8) :: line_str
    logical          :: fillzero 
    integer(KIND=8)   :: size_array 

    write(line_str, '(i5)') line

    fillzero = pfillzero

    size_array = (dim2-dim1+1_8) * &
         (dim4-dim3+1_8) * &
         (dim6-dim5+1_8)
    size_array = size_array * KIND(0)

    call write_trm_alloc(size_array, &
         "3D integer", array_name,"("//dim2_str//"-"//dim1_str//"+1)*("// & 
         dim4_str//"-"//dim3_str//"+1)*("//dim6_str//"-"//dim5_str//"+1)", &
         file_str,line_str)

    call temp_allocate(array, &
         dim1, dim2, & 
         dim3, dim4, & 
         dim5, dim6, & 
         trim(adjustl(adjustr(array_name)//" "//file_str//" "//line_str)), &
         oth_prefix=.true.,pfillzero=fillzero)
  end subroutine ftr_alloc_3d_i

  subroutine ftr_alloc_2d_i(array, dim1, dim2, &
       dim3, dim4, array_name, file_str, line, &
       dim1_str, dim2_str, dim3_str, dim4_str, pfillzero)
    integer, dimension(:,:), pointer, &
         intent(inout)    :: array
    integer         , intent(in) :: dim1,dim2,dim3,dim4
    character(len=*), intent(in) :: dim1_str,dim2_str,dim3_str,dim4_str
    character*(*)   , intent(in) :: file_str
    integer         , intent(in) :: line
    character*(*)   , intent(in) :: array_name
    logical         , intent(in) :: pfillzero

    character(len=8) :: line_str
    logical          :: fillzero 
    integer(KIND=8)   :: size_array 

    write(line_str, '(i5)') line

    fillzero = pfillzero

    size_array = (dim2-dim1+1_8) * &
         (dim4-dim3+1_8)
    size_array = size_array * KIND(0)

    call write_trm_alloc(size_array, &
         "2D integer", array_name, "("//dim2_str//"-"//dim1_str//"+1)*("// & 
         dim4_str//"-"//dim3_str//"+1)", &
         file_str,line_str)

    call temp_allocate(array, &
         dim1, dim2, & 
         dim3, dim4, & 
         trim(adjustl(adjustr(array_name)//" "//file_str//" "//line_str)), &
         oth_prefix=.true.,pfillzero=fillzero)
  end subroutine ftr_alloc_2d_i

  subroutine ftr_alloc_2d_d(array, dim1, dim2, &
       dim3, dim4, array_name, file_str, line, &
       dim1_str, dim2_str, dim3_str, dim4_str, pfillzero)
    real(RKIND), dimension(:,:), pointer, &
         intent(inout)    :: array
    integer         , intent(in) :: dim1,dim2,dim3,dim4
    character(len=*), intent(in) :: dim1_str,dim2_str,dim3_str,dim4_str
    character*(*)   , intent(in) :: file_str
    integer         , intent(in) :: line
    character*(*)   , intent(in) :: array_name
    logical         , intent(in) :: pfillzero

    character(len=8) :: line_str
    logical          :: fillzero 
    integer(KIND=8)   :: size_array 

    write(line_str, '(i5)') line

    fillzero = pfillzero

    size_array = (dim2-dim1+1_8) * &
         (dim4-dim3+1_8)
    size_array = size_array * KIND(0.0D0)

    call write_trm_alloc(size_array, &
         "2D double", array_name, "("//dim2_str//"-"//dim1_str//"+1)*("// & 
         dim4_str//"-"//dim3_str//"+1)", &
         file_str,line_str)

    call temp_allocate(array, &
         dim1, dim2, & 
         dim3, dim4, & 
         trim(adjustl(adjustr(array_name)//" "//file_str//" "//line_str)), &
         oth_prefix=.true.,pfillzero=fillzero)
  end subroutine ftr_alloc_2d_d

  subroutine ftr_alloc_2d_c(array, dim1, dim2, &
       dim3, dim4, array_name, file_str, line, &
       dim1_str, dim2_str, dim3_str, dim4_str, pfillzero)
    complex(CKIND), dimension(:,:), pointer, &
         intent(inout)    :: array
    integer         , intent(in) :: dim1,dim2,dim3,dim4
    character(len=*), intent(in) :: dim1_str,dim2_str,dim3_str,dim4_str
    character*(*)   , intent(in) :: file_str
    integer         , intent(in) :: line
    character*(*)   , intent(in) :: array_name
    logical         , intent(in) :: pfillzero

    character(len=8) :: line_str
    logical          :: fillzero 
    integer(KIND=8)   :: size_array 

    write(line_str, '(i5)') line

    fillzero = pfillzero

    size_array = (dim2-dim1+1_8) * &
         (dim4-dim3+1_8)
    size_array = size_array * 2 * KIND(0.0D0)

    call write_trm_alloc(size_array, &
         "2D complex", array_name, "("//dim2_str//"-"//dim1_str//"+1)*("// & 
         dim4_str//"-"//dim3_str//"+1)", &
         file_str,line_str)

    call temp_allocate(array, &
         dim1, dim2, & 
         dim3, dim4, & 
         trim(adjustl(adjustr(array_name)//" "//file_str//" "//line_str)), &
         oth_prefix=.true.,pfillzero=fillzero)
  end subroutine ftr_alloc_2d_c

  subroutine ftr_alloc_1d_i(array, dim1, dim2, &
       array_name, file_str, line, &
       dim1_str, dim2_str, pfillzero)
    integer, dimension(:), pointer, &
         intent(inout)    :: array
    integer         , intent(in) :: dim1,dim2
    character(len=*), intent(in) :: dim1_str,dim2_str
    character*(*)   , intent(in) :: file_str
    integer         , intent(in) :: line
    character*(*)   , intent(in) :: array_name
    logical         , intent(in) :: pfillzero

    character(len=8) :: line_str
    logical          :: fillzero 
    integer(KIND=8)   :: size_array 

    write(line_str, '(i5)') line

    fillzero = pfillzero

    size_array = (dim2-dim1+1_8)
    size_array = size_array * KIND(0)

    call write_trm_alloc(size_array, &
         "1D integer", array_name, "("//dim2_str//"-"//dim1_str//"+1)", &
         file_str,line_str)

    call temp_allocate(array, &
         dim1, dim2, & 
         trim(adjustl(adjustr(array_name)//" "//file_str//" "//line_str)), &
         oth_prefix=.true.,pfillzero=fillzero)
  end subroutine ftr_alloc_1d_i

  subroutine ftr_alloc_1d_d(array, dim1, dim2, &
       array_name, file_str, line, &
       dim1_str, dim2_str, pfillzero)
    real(RKIND), dimension(:), pointer, &
         intent(inout)    :: array
    integer         , intent(in) :: dim1,dim2
    character(len=*), intent(in) :: dim1_str,dim2_str
    character*(*)   , intent(in) :: file_str
    integer         , intent(in) :: line
    character*(*)   , intent(in) :: array_name
    logical         , intent(in) :: pfillzero

    character(len=8) :: line_str
    logical          :: fillzero 
    integer(KIND=8)   :: size_array 

    write(line_str, '(i5)') line

    fillzero = pfillzero

    size_array = (dim2-dim1+1_8)
    size_array = size_array * KIND(0.0D0)

    call write_trm_alloc(size_array, &
         "1D double", array_name, "("//dim2_str//"-"//dim1_str//"+1)", &
         file_str,line_str)

    call temp_allocate(array, &
         dim1, dim2, & 
         trim(adjustl(adjustr(array_name)//" "//file_str//" "//line_str)), &
         oth_prefix=.true.,pfillzero=fillzero)
  end subroutine ftr_alloc_1d_d

  subroutine ftr_alloc_1d_c(array, dim1, dim2, &
       array_name, file_str, line, &
       dim1_str, dim2_str, pfillzero)
    complex(CKIND), dimension(:), pointer, &
         intent(inout)    :: array
    integer         , intent(in) :: dim1,dim2
    character(len=*), intent(in) :: dim1_str,dim2_str
    character*(*)   , intent(in) :: file_str
    integer         , intent(in) :: line
    character*(*)   , intent(in) :: array_name
    logical         , intent(in) :: pfillzero

    character(len=8) :: line_str
    logical          :: fillzero 
    integer(KIND=8)   :: size_array 

    write(line_str, '(i5)') line

    fillzero = pfillzero

    size_array = (dim2-dim1+1_8)
    size_array = size_array * 2 * KIND(0.0D0)

    call write_trm_alloc(size_array, &
         "1D complex", array_name, "("//dim2_str//"-"//dim1_str//"+1)", &
         file_str,line_str)

    call temp_allocate(array, &
         dim1, dim2, & 
         trim(adjustl(adjustr(array_name)//" "//file_str//" "//line_str)), &
         oth_prefix=.true.,pfillzero=fillzero)
  end subroutine ftr_alloc_1d_c

  subroutine Write_dyn_dealloc_memory_(size_array,type_name,var_name, arg_prefix)
    use Io_Unit_module, only : IU_open, IU_close
    integer(KIND=8)    , intent(in) :: size_array
    character*(*), intent(in) :: type_name
    character*(*), intent(in) :: var_name
    character*(*), optional, intent(in) :: arg_prefix

    character(len=4) :: prefix

    if (present(arg_prefix)) then
       prefix = arg_prefix
    else 
       prefix = '->'
    endif

    if (mem_dyn_write.and.enable_dyn_trace) then
       if (pglobal_id.eq.0) then
          call IU_open(uout_dyn_mem, file = dyn_memory_filename, status = 'OLD', &
               position = 'APPEND', form = 'FORMATTED')
          write(uout_dyn_mem,'(A5,A6,I12,A12,A17,A60)') &
               prefix, '  -  ', &
               size_array,' Bytes for ',type_name//" ",adjustr(var_name)
          UFLUSH(uout_dyn_mem)
          call IU_close(uout_dyn_mem)
       end if
    end if
  end subroutine Write_dyn_dealloc_memory_

  !----------------------------------------
  ! loc_drop-block
  !----------------------------------------
  subroutine ftr_dealloc_5d_d(array, array_name, file_str, line)
    real(RKIND), dimension(:,:,:,:,:), pointer, &
         intent(inout)    :: array
    character*(*)      :: array_name

    character*(*), intent(in) :: file_str
    integer      , intent(in) :: line
    character(len=8)          :: line_str
    integer(KIND=8) :: size_array 

    write(line_str, '(i5)') line

    size_array = int(size(array),kind=8) * KIND(0)
    call Write_dyn_dealloc_memory_(size_array,'double array 5D', &
         trim(adjustl(adjustr(array_name)//" "//file_str//" "//line_str)), '#>')

    deallocate(array)

  end subroutine ftr_dealloc_5d_d

  subroutine ftr_dealloc_4d_d(array, array_name, file_str, line)
    real(RKIND), dimension(:,:,:,:), pointer, &
         intent(inout)    :: array
    character*(*)      :: array_name

    character*(*), intent(in) :: file_str
    integer      , intent(in) :: line
    character(len=8)          :: line_str

    write(line_str, '(i5)') line

    if (associated(array)) then    
       call temp_deallocate(array, trim(adjustl(adjustr(array_name)//" "//file_str//" "//line_str)), &
            oth_prefix=.true.)
    end if
  end subroutine ftr_dealloc_4d_d

  subroutine ftr_dealloc_4d_c(array, array_name, file_str, line)
    complex(CKIND), dimension(:,:,:,:), pointer, &
         intent(inout)    :: array
    character*(*)      :: array_name

    character*(*), intent(in) :: file_str
    integer      , intent(in) :: line
    character(len=8)          :: line_str

    write(line_str, '(i5)') line

    if (associated(array)) then    
       call temp_deallocate(array, trim(adjustl(adjustr(array_name)//" "//file_str//" "//line_str)), &
            oth_prefix=.true.)
    end if
  end subroutine ftr_dealloc_4d_c

  subroutine ftr_dealloc_3d_d(array, array_name, file_str, line)
    real(RKIND), dimension(:,:,:), pointer, &
         intent(inout)    :: array
    character*(*)      :: array_name

    character*(*), intent(in) :: file_str
    integer      , intent(in) :: line
    character(len=8)          :: line_str

    write(line_str, '(i5)') line

    if (associated(array)) then    
       call temp_deallocate(array, trim(adjustl(adjustr(array_name)//" "//file_str//" "//line_str)), &
            oth_prefix=.true.)
    end if
  end subroutine ftr_dealloc_3d_d

  subroutine ftr_dealloc_3d_c(array, array_name, file_str, line)
    complex(CKIND), dimension(:,:,:), pointer, &
         intent(inout)    :: array
    character*(*)      :: array_name

    character*(*), intent(in) :: file_str
    integer      , intent(in) :: line
    character(len=8)          :: line_str

    write(line_str, '(i5)') line

    if (associated(array)) then    
       call temp_deallocate(array, trim(adjustl(adjustr(array_name)//" "//file_str//" "//line_str)), &
            oth_prefix=.true.)
    end if
  end subroutine ftr_dealloc_3d_c

  subroutine ftr_dealloc_3d_i(array, array_name, file_str, line)
    integer, dimension(:,:,:), pointer, &
         intent(inout)    :: array
    character*(*)      :: array_name

    character*(*), intent(in) :: file_str
    integer      , intent(in) :: line
    character(len=8)          :: line_str

    write(line_str, '(i5)') line

    if (associated(array)) then    
       call temp_deallocate(array, trim(adjustl(adjustr(array_name)//" "//file_str//" "//line_str)), &
            oth_prefix=.true.)
    end if
  end subroutine ftr_dealloc_3d_i

  subroutine ftr_dealloc_2d_c(array, array_name, file_str, line)
    complex(CKIND), dimension(:,:), pointer, &
         intent(inout)    :: array
    character*(*)      :: array_name

    character*(*), intent(in) :: file_str
    integer      , intent(in) :: line
    character(len=8)          :: line_str

    write(line_str, '(i5)') line

    if (associated(array)) then    
       call temp_deallocate(array, trim(adjustl(adjustr(array_name)//" "//file_str//" "//line_str)), &
            oth_prefix=.true.)
    end if
  end subroutine ftr_dealloc_2d_c


  subroutine ftr_dealloc_2d_d(array, array_name, file_str, line)
    real(RKIND), dimension(:,:), pointer, &
         intent(inout)    :: array
    character*(*)      :: array_name

    character*(*), intent(in) :: file_str
    integer      , intent(in) :: line
    character(len=8)          :: line_str

    write(line_str, '(i5)') line

    if (associated(array)) then
       call temp_deallocate(array, trim(adjustl(adjustr(array_name)//" "//file_str//" "//line_str)), &
            oth_prefix=.true.)
    end if
  end subroutine ftr_dealloc_2d_d

  subroutine ftr_dealloc_2d_i(array, array_name, file_str, line)
    integer, dimension(:,:), pointer, &
         intent(inout)    :: array
    character*(*)      :: array_name

    character*(*), intent(in) :: file_str
    integer      , intent(in) :: line
    character(len=8)          :: line_str

    write(line_str, '(i5)') line

    if (associated(array)) then    
       call temp_deallocate(array, trim(adjustl(adjustr(array_name)//" "//file_str//" "//line_str)), &
            oth_prefix=.true.)
    end if
  end subroutine ftr_dealloc_2d_i

  subroutine ftr_dealloc_1d_c(array, array_name, file_str, line)
    complex(CKIND), dimension(:), pointer, &
         intent(inout)    :: array
    character*(*)      :: array_name

    character*(*), intent(in) :: file_str
    integer      , intent(in) :: line
    character(len=8)          :: line_str

    write(line_str, '(i5)') line

    if (associated(array)) then    
       call temp_deallocate(array, trim(adjustl(adjustr(array_name)//" "//file_str//" "//line_str)), &
            oth_prefix=.true.)
    end if
  end subroutine ftr_dealloc_1d_c

  subroutine ftr_dealloc_1d_d(array, array_name, file_str, line)
    real(RKIND), dimension(:), pointer, &
         intent(inout)    :: array
    character*(*)      :: array_name

    character*(*), intent(in) :: file_str
    integer      , intent(in) :: line
    character(len=8)          :: line_str

    write(line_str, '(i5)') line

    if (associated(array)) then    
       call temp_deallocate(array, trim(adjustl(adjustr(array_name)//" "//file_str//" "//line_str)), &
            oth_prefix=.true.)
    end if
  end subroutine ftr_dealloc_1d_d

  subroutine ftr_dealloc_1d_i(array, array_name, file_str, line)
    integer, dimension(:), pointer, &
         intent(inout)    :: array
    character*(*)      :: array_name

    character*(*), intent(in) :: file_str
    integer      , intent(in) :: line
    character(len=8)          :: line_str

    write(line_str, '(i5)') line

    if (associated(array)) then    
       call temp_deallocate(array, trim(adjustl(adjustr(array_name)//" "//file_str//" "//line_str)), &
            oth_prefix=.true.)
    end if
  end subroutine ftr_dealloc_1d_i

  !---------------------------------------- 
  ! memory allocation for a 1D array
  !----------------------------------------
  subroutine temp_allocate1d_i(array1d,begin_dim1,end_dim1,var_name,&
       oth_prefix,pfillzero)
    integer, dimension(:)    , pointer    :: array1d
    integer                  , intent(in) :: begin_dim1
    integer                  , intent(in) :: end_dim1
    character*(*)            , intent(in) :: var_name
    logical , optional       , intent(in) :: oth_prefix
    logical , optional       , intent(in) :: pfillzero

    integer   :: err
    integer   :: i1
    integer(KIND=8) :: size_array 

    logical :: fillzero 
    if (present(pfillzero)) then
       fillzero = pfillzero
    else 
       fillzero = .true.
    end if

    size_array = (end_dim1-begin_dim1+1_8) * KIND(0)
    if (present(oth_prefix).AND.oth_prefix) then 
       call Write_dyn_memory(size_array,'int array 1D',var_name, '#>')
    else 
       call Write_dyn_memory(size_array,'int array 1D',var_name)
    endif

    if (.not.memory_test) then 
       NULLIFY(array1d)
       allocate(array1d(begin_dim1:end_dim1),stat=err)
       if (err.eq.0) then
          if (fillzero) then
             do i1 = begin_dim1,end_dim1
                array1d(i1) = 0
             end do
          end if
       else
          if (pglobal_id.eq.0) then
             print*,'problem for allocating ',var_name
             print*,'-> required memory (in Bytes) = ',size_array
             print*,'-> memory already allocated (in Bytes) = ',nb_dyn_allocate
          end if
          stop
       end if
    end if
    nb_dyn_allocate  = nb_dyn_allocate + size_array
    max_dyn_allocate = max(max_dyn_allocate,nb_dyn_allocate)
  end subroutine temp_allocate1d_i

  !---------------------------------------- 
  subroutine temp_allocate1d_d(array1d,begin_dim1,end_dim1,var_name,&
       oth_prefix,pfillzero)
    real(RKIND), dimension(:), pointer    :: array1d
    integer                  , intent(in) :: begin_dim1
    integer                  , intent(in) :: end_dim1
    character*(*)            , intent(in) :: var_name
    logical , optional       , intent(in) :: oth_prefix
    logical , optional       , intent(in) :: pfillzero

    integer   :: err
    integer   :: i1
    integer(KIND=8) :: size_array 

    logical :: fillzero 
    if (present(pfillzero)) then
       fillzero = pfillzero
    else 
       fillzero = .true.
    end if

    size_array = (end_dim1-begin_dim1+1_8) * KIND(0.0D0)
    if (present(oth_prefix).AND.oth_prefix) then 
       call Write_dyn_memory(size_array,'double array 1D',var_name, '#>')
    else 
       call Write_dyn_memory(size_array,'double array 1D',var_name)
    endif

    if (.not.memory_test) then 
       NULLIFY(array1d)
       allocate(array1d(begin_dim1:end_dim1),stat=err)
       if (err.eq.0) then
          if (fillzero) then
             do i1 = begin_dim1,end_dim1
                array1d(i1) = 0._RKIND
             end do
          end if
       else
          if (pglobal_id.eq.0) then
             print*,'problem for allocating ',var_name
             print*,'-> required memory (in Bytes) = ',size_array
             print*,'-> memory already allocated (in Bytes) = ',nb_dyn_allocate
          end if
          stop
       end if
    end if
    nb_dyn_allocate  = nb_dyn_allocate + size_array
    max_dyn_allocate = max(max_dyn_allocate,nb_dyn_allocate)
  end subroutine temp_allocate1d_d

  !---------------------------------------- 
  subroutine temp_allocate1d_c(array1d,begin_dim1,end_dim1,var_name,&
       oth_prefix,pfillzero)
    complex(CKIND), dimension(:), pointer    :: array1d
    integer                     , intent(in) :: begin_dim1
    integer                     , intent(in) :: end_dim1
    character*(*)               , intent(in) :: var_name
    logical , optional          , intent(in) :: oth_prefix
    logical , optional          , intent(in) :: pfillzero

    integer   :: err
    integer   :: i1
    integer(KIND=8) :: size_array 

    logical :: fillzero 
    if (present(pfillzero)) then
       fillzero = pfillzero
    else 
       fillzero = .true.
    end if

    size_array = (end_dim1-begin_dim1+1_8) * 2*KIND(0.0D0)
    if (present(oth_prefix).AND.oth_prefix) then 
       call Write_dyn_memory(size_array,'complex array 1D',var_name, '#>')
    else 
       call Write_dyn_memory(size_array,'complex array 1D',var_name)
    endif

    if (.not.memory_test) then 
       NULLIFY(array1d)
       allocate(array1d(begin_dim1:end_dim1),stat=err)
       if (err.eq.0) then
          if (fillzero) then
             do i1 = begin_dim1,end_dim1
                array1d(i1) = 0._CKIND
             end do
          end if
       else
          if (pglobal_id.eq.0) then
             print*,'problem for allocating ',var_name
             print*,'-> required memory (in Bytes) = ',size_array
             print*,'-> memory already allocated (in Bytes) = ',nb_dyn_allocate
          end if
          stop
       end if
    end if
    nb_dyn_allocate  = nb_dyn_allocate + size_array
    max_dyn_allocate = max(max_dyn_allocate,nb_dyn_allocate)
  end subroutine temp_allocate1d_c


  !---------------------------------------- 
  ! memory allocation for a 2D array
  !----------------------------------------
  subroutine temp_allocate2d_i(array2d,begin_dim1,end_dim1, &
       begin_dim2,end_dim2,var_name,oth_prefix,pwrite,pfillzero)
    integer, dimension(:,:), pointer    :: array2d
    integer                , intent(in) :: begin_dim1
    integer                , intent(in) :: end_dim1
    integer                , intent(in) :: begin_dim2
    integer                , intent(in) :: end_dim2
    character*(*)          , intent(in) :: var_name
    logical , optional     , intent(in) :: oth_prefix
    logical , optional     , intent(in) :: pwrite
    logical , optional     , intent(in) :: pfillzero

    integer   :: i1, i2, err
    integer(KIND=8) :: size_array 

    logical :: fillzero 
    if (present(pfillzero)) then
       fillzero = pfillzero
    else 
       fillzero = .true.
    end if

    size_array = (end_dim1-begin_dim1+1_8) * &
         (end_dim2-begin_dim2+1_8) * KIND(0)
    if (present(pwrite)) then
       if (pwrite) then
          if (present(oth_prefix).AND.oth_prefix) then 
             call Write_dyn_memory(size_array,'integer array 2D',var_name, '#>')
          else 
             call Write_dyn_memory(size_array,'integer array 2D',var_name)
          endif
       endif
    else
       if (present(oth_prefix).AND.oth_prefix) then 
          call Write_dyn_memory(size_array,'integer array 2D',var_name, '#>')
       else 
          call Write_dyn_memory(size_array,'integer array 2D',var_name)
       endif
    endif

    if (.not.memory_test) then 
       NULLIFY(array2d)
       allocate(array2d(begin_dim1:end_dim1,begin_dim2:end_dim2), &
            stat=err)
       if (err.eq.0) then
          if (fillzero) then
             do i2 = begin_dim2,end_dim2
                do i1 = begin_dim1,end_dim1
                   array2d(i1,i2) = 0
                end do
             end do
          end if
       else
          if (pglobal_id.eq.0) then
             print*,'problem for allocating ',var_name
             print*,'-> required memory (in Bytes) = ',size_array
             print*,'-> memory already allocated (in Bytes) = ',nb_dyn_allocate
          end if
          stop
       end if
    end if
    nb_dyn_allocate  = nb_dyn_allocate + size_array
    max_dyn_allocate = max(max_dyn_allocate,nb_dyn_allocate)
  end subroutine temp_allocate2d_i

  !----------------------------------------
  subroutine temp_allocate2d_d(array2d,begin_dim1,end_dim1, &
       begin_dim2,end_dim2,var_name,oth_prefix,pwrite,pfillzero)
    real(RKIND), dimension(:,:), pointer    :: array2d
    integer                    , intent(in) :: begin_dim1
    integer                    , intent(in) :: end_dim1  
    integer                    , intent(in) :: begin_dim2
    integer                    , intent(in) :: end_dim2
    character*(*)              , intent(in) :: var_name
    logical , optional         , intent(in) :: oth_prefix
    logical , optional         , intent(in) :: pwrite
    logical , optional         , intent(in) :: pfillzero

    integer   :: i1, i2, err
    integer(KIND=8) :: size_array 

    logical :: fillzero 
    if (present(pfillzero)) then
       fillzero = pfillzero
    else 
       fillzero = .true.
    end if

    size_array = (end_dim1-begin_dim1+1_8) * &
         (end_dim2-begin_dim2+1_8) * KIND(0.0D0)
    if (present(pwrite)) then
       if (pwrite) then
          if (present(oth_prefix).AND.oth_prefix) then 
             call Write_dyn_memory(size_array,'double array 2D',var_name, '#>')
          else 
             call Write_dyn_memory(size_array,'double array 2D',var_name)
          endif
       endif
    else
       if (present(oth_prefix).AND.oth_prefix) then 
          call Write_dyn_memory(size_array,'double array 2D',var_name, '#>')
       else 
          call Write_dyn_memory(size_array,'double array 2D',var_name)
       endif
    endif

    if (.not.memory_test) then 
       NULLIFY(array2d)
       allocate(array2d(begin_dim1:end_dim1,begin_dim2:end_dim2), &
            stat=err)
       if (err.eq.0) then
          if (fillzero) then
             do i2 = begin_dim2,end_dim2
                do i1 = begin_dim1,end_dim1
                   array2d(i1,i2) = 0._RKIND
                end do
             end do
          end if
       else
          if (pglobal_id.eq.0) then
             print*,'problem for allocating ',var_name
             print*,'-> required memory (in Bytes) = ',size_array
             print*,'-> memory already allocated (in Bytes) = ',nb_dyn_allocate
          end if
          stop
       end if
    end if
    nb_dyn_allocate  = nb_dyn_allocate + size_array
    max_dyn_allocate = max(max_dyn_allocate,nb_dyn_allocate)
  end subroutine temp_allocate2d_d

  !----------------------------------------  
  subroutine temp_allocate2d_c(array2d,begin_dim1,end_dim1, &
       begin_dim2,end_dim2,var_name,oth_prefix,pfillzero, pwrite)
    complex(CKIND), dimension(:,:), pointer    :: array2d
    integer                       , intent(in) :: begin_dim1
    integer                       , intent(in) :: end_dim1  
    integer                       , intent(in) :: begin_dim2
    integer                       , intent(in) :: end_dim2
    character*(*)                 , intent(in) :: var_name
    logical , optional            , intent(in) :: oth_prefix
    logical , optional            , intent(in) :: pfillzero
    logical , optional            , intent(in) :: pwrite

    integer   :: err
    integer   :: i1, i2
    integer(KIND=8) :: size_array 

    logical :: fillzero 
    if (present(pfillzero)) then
       fillzero = pfillzero
    else 
       fillzero = .true.
    end if

    size_array = (end_dim1-begin_dim1+1_8) * &
         (end_dim2-begin_dim2+1_8) * 2*KIND(0.0D0)
    if (present(pwrite)) then
       if (pwrite) then
          if (present(oth_prefix).AND.oth_prefix) then 
             call Write_dyn_memory(size_array,'complex array 2D',var_name, '#>')
          else 
             call Write_dyn_memory(size_array,'complex array 2D',var_name)
          endif
       endif
    else
       if (present(oth_prefix).AND.oth_prefix) then 
          call Write_dyn_memory(size_array,'complex array 2D',var_name, '#>')
       else 
          call Write_dyn_memory(size_array,'complex array 2D',var_name)
       endif
    endif

    if (.not.memory_test) then 
       NULLIFY(array2d)
       allocate(array2d(begin_dim1:end_dim1,begin_dim2:end_dim2), &
            stat=err)
       if (err.eq.0) then
          if (fillzero) then
             do i2 = begin_dim2,end_dim2
                do i1 = begin_dim1,end_dim1
                   array2d(i1,i2) = 0._CKIND
                end do
             end do
          end if
       else
          if (pglobal_id.eq.0) then
             print*,'problem for allocating ',var_name
             print*,'-> required memory (in Bytes) = ',size_array
             print*,'-> memory already allocated (in Bytes) = ',nb_dyn_allocate
          end if
          stop
       end if
    end if
    nb_dyn_allocate  = nb_dyn_allocate + size_array
    max_dyn_allocate = max(max_dyn_allocate,nb_dyn_allocate)
  end subroutine temp_allocate2d_c

  !---------------------------------------- 
  ! memory allocation for a 3D array
  !----------------------------------------
  subroutine temp_allocate3d_d(array3d,begin_dim1,end_dim1, &
       begin_dim2,end_dim2,begin_dim3,end_dim3,var_name,oth_prefix,pfillzero)
    real(RKIND), dimension(:,:,:), pointer    :: array3d
    integer                      , intent(in) :: begin_dim1
    integer                      , intent(in) :: end_dim1  
    integer                      , intent(in) :: begin_dim2
    integer                      , intent(in) :: end_dim2
    integer                      , intent(in) :: begin_dim3
    integer                      , intent(in) :: end_dim3    
    character*(*)                , intent(in) :: var_name
    logical , optional           , intent(in) :: oth_prefix
    logical , optional           , intent(in) :: pfillzero

    integer   :: i1, i2, i3, err
    integer(KIND=8) :: size_array 

    logical :: fillzero 
    if (present(pfillzero)) then
       fillzero = pfillzero
    else 
       fillzero = .true.
    end if

    size_array = (end_dim1-begin_dim1+1_8) * &
         (end_dim2-begin_dim2+1_8)* &
         (end_dim3-begin_dim3+1_8)* KIND(0.0D0)
    if (present(oth_prefix).AND.oth_prefix) then 
       call Write_dyn_memory(size_array,'double array 3D',var_name, '#>')
    else 
       call Write_dyn_memory(size_array,'double array 3D',var_name)
    endif

    if (.not.memory_test) then 
       NULLIFY(array3d)
       allocate(array3d(begin_dim1:end_dim1,begin_dim2:end_dim2, &
            begin_dim3:end_dim3),stat=err)
       if (err.eq.0) then
          if (fillzero) then
             do i3 = begin_dim3,end_dim3
                do i2 = begin_dim2,end_dim2
                   do i1 = begin_dim1,end_dim1
                      array3d(i1,i2,i3) = 0._RKIND
                   end do
                end do
             end do
          end if
       else
          if (pglobal_id.eq.0) then
             print*,'problem for allocating ',var_name
             print*,'-> required memory (in Bytes) = ',size_array
             print*,'-> memory already allocated (in Bytes) = ',nb_dyn_allocate
          end if
          stop
       end if
    end if
    nb_dyn_allocate  = nb_dyn_allocate + size_array
    max_dyn_allocate = max(max_dyn_allocate,nb_dyn_allocate)
  end subroutine temp_allocate3d_d

  !----------------------------------------
  subroutine temp_allocate3d_c(array3d,begin_dim1,end_dim1, &
       begin_dim2,end_dim2,begin_dim3,end_dim3,var_name,oth_prefix,pfillzero)
    complex(CKIND), dimension(:,:,:), pointer    :: array3d
    integer                         , intent(in) :: begin_dim1
    integer                         , intent(in) :: end_dim1  
    integer                         , intent(in) :: begin_dim2
    integer                         , intent(in) :: end_dim2
    integer                         , intent(in) :: begin_dim3
    integer                         , intent(in) :: end_dim3
    character*(*)                   , intent(in) :: var_name
    logical , optional              , intent(in) :: oth_prefix
    logical , optional              , intent(in) :: pfillzero

    integer   :: i1, i2, i3, err
    integer(KIND=8) :: size_array 

    logical :: fillzero 
    if (present(pfillzero)) then
       fillzero = pfillzero
    else 
       fillzero = .true.
    end if

    size_array = (end_dim1-begin_dim1+1_8) * &
         (end_dim2-begin_dim2+1_8)* &
         (end_dim3-begin_dim3+1_8)* 2*KIND(0.0D0)
    if (present(oth_prefix).AND.oth_prefix) then 
       call Write_dyn_memory(size_array,'complex array 3D',var_name, '#>')
    else 
       call Write_dyn_memory(size_array,'complex array 3D',var_name)
    endif

    if (.not.memory_test) then 
       NULLIFY(array3d)
       allocate(array3d(begin_dim1:end_dim1,begin_dim2:end_dim2, &
            begin_dim3:end_dim3),stat=err)
       if (err.eq.0) then
          if (fillzero) then
             do i3 = begin_dim3,end_dim3
                do i2 = begin_dim2,end_dim2
                   do i1 = begin_dim1,end_dim1
                      array3d(i1,i2,i3) = 0._CKIND
                   end do
                end do
             end do
          end if
       else
          if (pglobal_id.eq.0) then
             print*,'problem for allocating ',var_name
             print*,'-> required memory (in Bytes) = ',size_array
             print*,'-> memory already allocated (in Bytes) = ',nb_dyn_allocate
          end if
          stop
       end if
    end if
    nb_dyn_allocate  = nb_dyn_allocate + size_array
    max_dyn_allocate = max(max_dyn_allocate,nb_dyn_allocate)
  end subroutine temp_allocate3d_c

  !----------------------------------------
  subroutine temp_allocate3d_i(array3d,begin_dim1,end_dim1, &
       begin_dim2,end_dim2,begin_dim3,end_dim3,var_name,oth_prefix,pfillzero)
    integer, dimension(:,:,:), pointer    :: array3d
    integer                  , intent(in) :: begin_dim1
    integer                  , intent(in) :: end_dim1  
    integer                  , intent(in) :: begin_dim2
    integer                  , intent(in) :: end_dim2
    integer                  , intent(in) :: begin_dim3
    integer                  , intent(in) :: end_dim3
    character*(*)            , intent(in) :: var_name
    logical , optional       , intent(in) :: oth_prefix
    logical , optional           , intent(in) :: pfillzero

    integer   :: i1, i2, i3, err
    integer(KIND=8) :: size_array 

    logical :: fillzero 
    if (present(pfillzero)) then
       fillzero = pfillzero
    else 
       fillzero = .true.
    end if

    size_array = (end_dim1-begin_dim1+1_8) * &
         (end_dim2-begin_dim2+1_8)* &
         (end_dim3-begin_dim3+1_8)* KIND(0)
    if (present(oth_prefix).AND.oth_prefix) then 
       call Write_dyn_memory(size_array,'int array 3D',var_name, '#>')
    else 
       call Write_dyn_memory(size_array,'int array 3D',var_name)
    endif

    if (.not.memory_test) then 
       NULLIFY(array3d)
       allocate(array3d(begin_dim1:end_dim1,begin_dim2:end_dim2, &
            begin_dim3:end_dim3),stat=err)
       if (err.eq.0) then
          if (fillzero) then
             do i3 = begin_dim3,end_dim3
                do i2 = begin_dim2,end_dim2
                   do i1 = begin_dim1,end_dim1
                      array3d(i1,i2,i3) = 0
                   end do
                end do
             end do
          end if
       else
          if (pglobal_id.eq.0) then
             print*,'problem for allocating ',var_name
             print*,'-> required memory (in Bytes) = ',size_array
             print*,'-> memory already allocated (in Bytes) = ',nb_dyn_allocate
          end if
          stop
       end if
    end if
    nb_dyn_allocate  = nb_dyn_allocate + size_array
    max_dyn_allocate = max(max_dyn_allocate,nb_dyn_allocate)
  end subroutine temp_allocate3d_i

  !---------------------------------------- 
  ! memory allocation for a 4D array
  !----------------------------------------
  subroutine temp_allocate4d_d(array4d,begin_dim1,end_dim1, &
       begin_dim2,end_dim2, begin_dim3,end_dim3, &
       begin_dim4,end_dim4,var_name,oth_prefix,pfillzero)
    real(RKIND), dimension(:,:,:,:), pointer    :: array4d
    integer                        , intent(in) :: begin_dim1
    integer                        , intent(in) :: end_dim1
    integer                        , intent(in) :: begin_dim2
    integer                        , intent(in) :: end_dim2
    integer                        , intent(in) :: begin_dim3
    integer                        , intent(in) :: end_dim3
    integer                        , intent(in) :: begin_dim4
    integer                        , intent(in) :: end_dim4
    character*(*)                  , intent(in) :: var_name
    logical , optional             , intent(in) :: oth_prefix
    logical , optional             , intent(in) :: pfillzero

    integer   :: err
    integer   :: i1, i2, i3, i4
    integer(KIND=8) :: size_array 

    logical :: fillzero 
    if (present(pfillzero)) then
       fillzero = pfillzero
    else 
       fillzero = .true.
    end if

    size_array = (end_dim1-begin_dim1+1_8) * &
         (end_dim2-begin_dim2+1_8) * &
         (end_dim3-begin_dim3+1_8) * &
         (end_dim4-begin_dim4+1_8) * KIND(0.0D0)
    if (present(oth_prefix).AND.oth_prefix) then 
       call Write_dyn_memory(size_array,'double array 4D',var_name, '#>')
    else 
       call Write_dyn_memory(size_array,'double array 4D',var_name)
    endif

    if (.not.memory_test) then 
       NULLIFY(array4d)
       allocate(array4d(begin_dim1:end_dim1,begin_dim2:end_dim2, &
            begin_dim3:end_dim3,begin_dim4:end_dim4),stat=err)
       if (err.eq.0) then
          if (fillzero) then
             do i4 = begin_dim4,end_dim4
                do i3 = begin_dim3,end_dim3
                   do i2 = begin_dim2,end_dim2
                      do i1 = begin_dim1,end_dim1
                         array4d(i1,i2,i3,i4) = 0._RKIND
                      end do
                   end do
                end do
             end do
          end if
       else
          if (pglobal_id.eq.0) then
             print*,'problem for allocating ',var_name
             print*,'-> required memory (in Bytes) = ',size_array
             print*,'-> memory already allocated (in Bytes) = ',nb_dyn_allocate
          end if
          stop
       end if
    end if
    nb_dyn_allocate  = nb_dyn_allocate + size_array
    max_dyn_allocate = max(max_dyn_allocate,nb_dyn_allocate)
  end subroutine temp_allocate4d_d


  subroutine temp_allocate4d_c(array4d,begin_dim1,end_dim1, &
       begin_dim2,end_dim2, begin_dim3,end_dim3, &
       begin_dim4,end_dim4,var_name,oth_prefix,pfillzero)
    complex(CKIND), dimension(:,:,:,:), pointer    :: array4d
    integer                           , intent(in) :: begin_dim1
    integer                           , intent(in) :: end_dim1
    integer                           , intent(in) :: begin_dim2
    integer                           , intent(in) :: end_dim2
    integer                           , intent(in) :: begin_dim3
    integer                           , intent(in) :: end_dim3
    integer                           , intent(in) :: begin_dim4
    integer                           , intent(in) :: end_dim4
    character*(*)                     , intent(in) :: var_name
    logical , optional                , intent(in) :: oth_prefix
    logical , optional                , intent(in) :: pfillzero

    integer   :: err
    integer   :: i1, i2, i3, i4
    integer(KIND=8) :: size_array 

    logical :: fillzero 
    if (present(pfillzero)) then
       fillzero = pfillzero
    else 
       fillzero = .true.
    end if

    size_array = (end_dim1-begin_dim1+1_8) * &
         (end_dim2-begin_dim2+1_8) * &
         (end_dim3-begin_dim3+1_8) * &
         (end_dim4-begin_dim4+1_8) * 2*KIND(0.0D0)
    if (present(oth_prefix).AND.oth_prefix) then 
       call Write_dyn_memory(size_array,'complex array 4D',var_name, '#>')
    else 
       call Write_dyn_memory(size_array,'complex array 4D',var_name)
    endif

    if (.not.memory_test) then 
       NULLIFY(array4d)
       allocate(array4d(begin_dim1:end_dim1,begin_dim2:end_dim2, &
            begin_dim3:end_dim3,begin_dim4:end_dim4),stat=err)
       if (err.eq.0) then
          if (fillzero) then
             do i4 = begin_dim4,end_dim4
                do i3 = begin_dim3,end_dim3
                   do i2 = begin_dim2,end_dim2
                      do i1 = begin_dim1,end_dim1
                         array4d(i1,i2,i3,i4) = 0._CKIND
                      end do
                   end do
                end do
             end do
          end if
       else
          if (pglobal_id.eq.0) then
             print*,'problem for allocating ',var_name
             print*,'-> required memory (in Bytes) = ',size_array
             print*,'-> memory already allocated (in Bytes) = ',nb_dyn_allocate
          end if
          stop
       end if
    end if
    nb_dyn_allocate  = nb_dyn_allocate + size_array
    max_dyn_allocate = max(max_dyn_allocate,nb_dyn_allocate)
  end subroutine temp_allocate4d_c

  !---------------------------------------- 
  ! memory allocation for a 5D array
  !----------------------------------------
  subroutine temp_allocate5d_d(array5d,begin_dim1,end_dim1, &
       begin_dim2,end_dim2,begin_dim3,end_dim3,begin_dim4,end_dim4, &
       begin_dim5,end_dim5,var_name,oth_prefix,pfillzero)
    real(RKIND), dimension(:,:,:,:,:), pointer    :: array5d
    integer                          , intent(in) :: begin_dim1
    integer                          , intent(in) :: end_dim1
    integer                          , intent(in) :: begin_dim2
    integer                          , intent(in) :: end_dim2
    integer                          , intent(in) :: begin_dim3
    integer                          , intent(in) :: end_dim3
    integer                          , intent(in) :: begin_dim4
    integer                          , intent(in) :: end_dim4
    integer                          , intent(in) :: begin_dim5
    integer                          , intent(in) :: end_dim5
    character*(*)                    , intent(in) :: var_name
    logical , optional               , intent(in) :: oth_prefix
    logical , optional               , intent(in) :: pfillzero

    integer   :: err
    integer   :: i1, i2, i3, i4, i5
    integer(KIND=8) :: size_array 

    logical :: fillzero 
    if (present(pfillzero)) then
       fillzero = pfillzero
    else 
       fillzero = .true.
    end if

    size_array = (end_dim1-begin_dim1+1_8) * &
         (end_dim2-begin_dim2+1_8) * &
         (end_dim3-begin_dim3+1_8) * &
         (end_dim4-begin_dim4+1_8) * &
         (end_dim5-begin_dim5+1_8) * KIND(0.0D0)
    if (present(oth_prefix).AND.oth_prefix) then 
       call Write_dyn_memory(size_array,'double array 5D',var_name, '#>')
    else 
       call Write_dyn_memory(size_array,'double array 5D',var_name)
    endif

    if (.not.memory_test) then 
       NULLIFY(array5d)
       allocate(array5d(begin_dim1:end_dim1,begin_dim2:end_dim2, &
            begin_dim3:end_dim3,begin_dim4:end_dim4, &
            begin_dim5:end_dim5),stat=err)
       if (err.eq.0) then
          if (fillzero) then
             do i5 = begin_dim5,end_dim5
                do i4 = begin_dim4,end_dim4
                   do i3 = begin_dim3,end_dim3
                      do i2 = begin_dim2,end_dim2
                         do i1 = begin_dim1,end_dim1
                            array5d(i1,i2,i3,i4,i5) = 0._RKIND
                         end do
                      end do
                   end do
                end do
             end do
          end if
       else
          if (pglobal_id.eq.0) then
             print*,'problem for allocating ',var_name
             print*,'-> required memory (in Bytes) = ',size_array
             print*,'-> memory already allocated (in Bytes) = ',nb_dyn_allocate
          end if
          stop
       end if
    end if
    nb_dyn_allocate  = nb_dyn_allocate + size_array
    max_dyn_allocate = max(max_dyn_allocate,nb_dyn_allocate)
  end subroutine temp_allocate5d_d


  !---------------------------------------- 
  ! memory deallocation of array 1D
  !----------------------------------------
  subroutine temp_deallocate1d_i(array1d,var_name,oth_prefix)
    integer, dimension(:), pointer      :: array1d
    character*(*), optional, intent(in) :: var_name
    logical      , optional, intent(in) :: oth_prefix

    integer(KIND=8) :: size_array 

    if (associated(array1d)) then
       size_array = int(size(array1d),kind=8) * KIND(0)
       nb_dyn_allocate = nb_dyn_allocate - size_array
       if (present(oth_prefix).AND.oth_prefix) then 
          call Write_dyn_dealloc_memory(size_array,'int array 1D',var_name, '#>')
       else 
          call Write_dyn_dealloc_memory(size_array,'int array 1D',var_name)
       endif

       if (.not.memory_test) &
            deallocate(array1d)
    end if
  end subroutine temp_deallocate1d_i

  !---------------------------------------- 
  subroutine temp_deallocate1d_d(array1d,var_name,oth_prefix)
    real(RKIND), dimension(:), pointer  :: array1d
    character*(*), optional, intent(in) :: var_name
    logical      , optional, intent(in) :: oth_prefix

    integer(KIND=8) :: size_array 

    if (associated(array1d)) then
       size_array = int(size(array1d),kind=8) * KIND(0.0D0)
       nb_dyn_allocate = nb_dyn_allocate - size_array
       if (present(oth_prefix).AND.oth_prefix) then 
          call Write_dyn_dealloc_memory(size_array,'double array 1D',var_name, '#>')
       else 
          call Write_dyn_dealloc_memory(size_array,'double array 1D',var_name)
       endif

       if (.not.memory_test) &
            deallocate(array1d)
    end if
  end subroutine temp_deallocate1d_d

  !---------------------------------------- 
  subroutine temp_deallocate1d_c(array1d,var_name,oth_prefix)
    complex(CKIND), dimension(:), pointer :: array1d
    character*(*), optional, intent(in) :: var_name
    logical      , optional, intent(in) :: oth_prefix

    integer(KIND=8)   :: size_array

    if (associated(array1d)) then
       size_array = int(size(array1d),kind=8) * 2*KIND(0.0D0)
       nb_dyn_allocate = nb_dyn_allocate - size_array
       if (present(oth_prefix).AND.oth_prefix) then 
          call Write_dyn_dealloc_memory(size_array,'complex array 1D',var_name, '#>')
       else 
          call Write_dyn_dealloc_memory(size_array,'complex array 1D',var_name)
       endif

       if (.not.memory_test) &
            deallocate(array1d)
    end if
  end subroutine temp_deallocate1d_c


  !---------------------------------------- 
  ! memory deallocation of array 2D
  !----------------------------------------
  subroutine temp_deallocate2d_i(array2d,var_name,oth_prefix, pwrite)
    integer, dimension(:,:), pointer    :: array2d
    character*(*), intent(in), optional :: var_name
    logical      , intent(in), optional :: oth_prefix
    logical      , intent(in), optional :: pwrite

    integer(KIND=8)   :: size_array

    if (associated(array2d)) then
       size_array = int(size(array2d),kind=8) * KIND(0)
       nb_dyn_allocate = nb_dyn_allocate - size_array
       if (present(pwrite)) then
          if (pwrite) then
             if (present(oth_prefix).AND.oth_prefix) then 
                call Write_dyn_dealloc_memory(size_array,'int array 2D',var_name, '#>')
             else 
                call Write_dyn_dealloc_memory(size_array,'int array 2D',var_name)
             endif
          endif
       else
          if (present(oth_prefix).AND.oth_prefix) then 
             call Write_dyn_dealloc_memory(size_array,'int array 2D',var_name, '#>')
          else 
             call Write_dyn_dealloc_memory(size_array,'int array 2D',var_name)
          endif
       endif

       if (.not.memory_test) &
            deallocate(array2d)
    end if
  end subroutine temp_deallocate2d_i

  !----------------------------------------
  subroutine temp_deallocate2d_d(array2d,var_name,oth_prefix, pwrite)
    real(RKIND), dimension(:,:), pointer :: array2d
    character*(*), intent(in), optional  :: var_name
    logical      , intent(in), optional  :: oth_prefix
    logical      , intent(in), optional  :: pwrite

    integer(KIND=8)   :: size_array

    if (associated(array2d)) then
       size_array = int(size(array2d),kind=8) * KIND(0.0D0)
       nb_dyn_allocate = nb_dyn_allocate - size_array
       if (present(pwrite)) then
          if (pwrite) then
             if (present(oth_prefix).AND.oth_prefix) then 
                call Write_dyn_dealloc_memory(size_array,'double array 2D',var_name, '#>')
             else 
                call Write_dyn_dealloc_memory(size_array,'double array 2D',var_name)
             endif
          endif
       else
          if (present(oth_prefix).AND.oth_prefix) then 
             call Write_dyn_dealloc_memory(size_array,'double array 2D',var_name, '#>')
          else 
             call Write_dyn_dealloc_memory(size_array,'double array 2D',var_name)
          endif
       endif

       if (.not.memory_test) &
            deallocate(array2d)
    end if
  end subroutine temp_deallocate2d_d

  !----------------------------------------
  subroutine temp_deallocate2d_c(array2d,var_name,oth_prefix, pwrite)
    complex(CKIND), dimension(:,:), pointer :: array2d
    character*(*), intent(in), optional     :: var_name
    logical      , intent(in), optional     :: oth_prefix
    logical      , intent(in), optional     :: pwrite

    integer(KIND=8)   :: size_array

    if (associated(array2d)) then
       size_array = int(size(array2d),kind=8) * 2*KIND(0.0D0)
       nb_dyn_allocate = nb_dyn_allocate - size_array
       if (present(pwrite)) then
          if (pwrite) then
             if (present(oth_prefix).AND.oth_prefix) then 
                call Write_dyn_dealloc_memory(size_array,'complex array 2D',var_name, '#>')
             else 
                call Write_dyn_dealloc_memory(size_array,'complex array 2D',var_name)
             endif
          endif
       else
          if (present(oth_prefix).AND.oth_prefix) then 
             call Write_dyn_dealloc_memory(size_array,'complex array 2D',var_name, '#>')
          else 
             call Write_dyn_dealloc_memory(size_array,'complex array 2D',var_name)
          endif
       endif

       if (.not.memory_test) &
            deallocate(array2d)
    end if
  end subroutine temp_deallocate2d_c


  !---------------------------------------- 
  ! memory deallocation of array 3D
  !----------------------------------------
  subroutine temp_deallocate3d_d(array3d,var_name, oth_prefix)
    real(RKIND), dimension(:,:,:), pointer :: array3d
    character*(*), intent(in), optional    :: var_name
    logical      , intent(in), optional    :: oth_prefix

    integer(KIND=8)   :: size_array

    if (associated(array3d)) then
       size_array = int(size(array3d),kind=8) * KIND(0.0D0)
       nb_dyn_allocate = nb_dyn_allocate - size_array
       if (present(oth_prefix).AND.oth_prefix) then 
          call Write_dyn_dealloc_memory(size_array,'double array 3D',var_name, '#>')
       else 
          call Write_dyn_dealloc_memory(size_array,'double array 3D',var_name)
       endif
       if (.not.memory_test) &
            deallocate(array3d)
    end if
  end subroutine temp_deallocate3d_d

  subroutine temp_deallocate3d_i(array3d,var_name, oth_prefix)
    integer, dimension(:,:,:), pointer  :: array3d
    character*(*), intent(in), optional :: var_name
    logical      , intent(in), optional :: oth_prefix

    integer(KIND=8)   :: size_array

    if (associated(array3d)) then
       size_array = int(size(array3d),kind=8) * KIND(0)
       nb_dyn_allocate = nb_dyn_allocate - size_array
       if (present(oth_prefix).AND.oth_prefix) then 
          call Write_dyn_dealloc_memory(size_array,'int array 3D',var_name, '#>')
       else 
          call Write_dyn_dealloc_memory(size_array,'int array 3D',var_name)
       endif
       if (.not.memory_test) &
            deallocate(array3d)
    end if
  end subroutine temp_deallocate3d_i

  subroutine temp_deallocate3d_c(array3d,var_name, oth_prefix)
    complex(CKIND), dimension(:,:,:), pointer :: array3d
    character*(*) , intent(in), optional      :: var_name
    logical       , intent(in), optional      :: oth_prefix

    integer(KIND=8)   :: size_array

    if (associated(array3d)) then
       size_array = int(size(array3d),kind=8) * 2*KIND(0.0D0)
       nb_dyn_allocate = nb_dyn_allocate - size_array
       if (present(oth_prefix).AND.oth_prefix) then 
          call Write_dyn_dealloc_memory(size_array,'complex array 3D',var_name, '#>')
       else 
          call Write_dyn_dealloc_memory(size_array,'complex array 3D',var_name)
       endif
       if (.not.memory_test) &
            deallocate(array3d)
    end if
  end subroutine temp_deallocate3d_c


  !---------------------------------------- 
  ! memory deallocation of array 4D
  !----------------------------------------
  subroutine temp_deallocate4d_d(array4d,var_name, oth_prefix)
    real(RKIND), dimension(:,:,:,:), pointer :: array4d
    character*(*), intent(in), optional      :: var_name
    logical      , intent(in), optional      :: oth_prefix

    integer(KIND=8)   :: size_array

    if (associated(array4d)) then
       size_array = int(size(array4d),kind=8) * KIND(0.0D0)
       nb_dyn_allocate = nb_dyn_allocate - size_array
       if (present(oth_prefix).AND.oth_prefix) then 
          call Write_dyn_dealloc_memory(size_array,'double array 4D',var_name, '#>')
       else 
          call Write_dyn_dealloc_memory(size_array,'double array 4D',var_name)
       endif
       if (.not.memory_test) &
            deallocate(array4d)
    end if
  end subroutine temp_deallocate4d_d

  subroutine temp_deallocate4d_c(array4d,var_name, oth_prefix)
    complex(CKIND), dimension(:,:,:,:), pointer :: array4d
    character*(*), intent(in), optional         :: var_name
    logical      , intent(in), optional         :: oth_prefix

    integer(KIND=8)   :: size_array

    if (associated(array4d)) then
       size_array = int(size(array4d),kind=8) * 2*KIND(0.0D0)
       nb_dyn_allocate = nb_dyn_allocate - size_array
       if (present(oth_prefix).AND.oth_prefix) then 
          call Write_dyn_dealloc_memory(size_array,'double array 4D',var_name, '#>')
       else 
          call Write_dyn_dealloc_memory(size_array,'double array 4D',var_name)
       endif
       if (.not.memory_test) &
            deallocate(array4d)
    end if
  end subroutine temp_deallocate4d_c

  !---------------------------------------- 
  ! memory deallocation of array 5D
  !----------------------------------------
  subroutine temp_deallocate5d_d(array5d,var_name, oth_prefix)
    real(RKIND), dimension(:,:,:,:,:), pointer :: array5d
    character*(*), intent(in), optional        :: var_name
    logical      , intent(in), optional        :: oth_prefix

    integer(KIND=8)   :: size_array

    if (associated(array5d)) then
       size_array = int(size(array5d),kind=8) * KIND(0.0D0)
       nb_dyn_allocate = nb_dyn_allocate - size_array
       if (present(oth_prefix).AND.oth_prefix) then 
          call Write_dyn_dealloc_memory(size_array,'double array 5D',var_name, '#>')
       else 
          call Write_dyn_dealloc_memory(size_array,'double array 5D',var_name)
       endif
       if (.not.memory_test) &
            deallocate(array5d)
    end if
  end subroutine temp_deallocate5d_d



  !***********************************************
  !  function for program analysis
  !***********************************************
  subroutine print_dyn_memory_size()

    use globals, only : uout_res, outputproc, file_name_res
    use Io_Unit_module, only: IU_open, IU_close

    implicit none

    integer :: iprint
    integer :: uout = 55

    integer(KIND=8)   :: base
    real(RKIND)       :: peak_mem
    character(len=48) :: peak_mem_str

    base         = 1024
    peak_mem     = REAL(max_dyn_allocate)
    peak_mem_str = '  Memory peak per MPI process = '

    if (pglobal_id.eq.outputproc) then
       print *, "max_dyn_allocate ", max_dyn_allocate 
       print *, "base**4 ", base**4 
       do iprint = 1,2
          if (iprint.eq.1) then
             call IU_open(uout, file = file_name_res, status = 'OLD', &
                  position = 'APPEND', form = 'FORMATTED')
          else
             call IU_open(uout, file = dyn_memory_filename, status = 'OLD', &
                  position = 'APPEND', form = 'FORMATTED')
          end if

          if (max_dyn_allocate .gt. base**4 ) then
             write(uout,'(A,1f10.2,A)') &
                  peak_mem_str, &
                  peak_mem/REAL(base**4, RKIND), ' TBytes'
          else if (max_dyn_allocate .gt. base**3 ) then
             write(uout,'(A,1f10.2,A)') &
                  peak_mem_str, &
                  peak_mem/REAL(base**3, RKIND), ' GBytes'
          else if (max_dyn_allocate .gt. (base * base) ) then
             write(uout,'(A,1f10.2,A)') &
                  peak_mem_str, &
                  peak_mem/REAL((base * base), RKIND), ' MBytes'
          else 
             write(uout,'(A,1f10.2,A)') &
                  peak_mem_str, &
                  peak_mem/REAL(base, RKIND), ' KBytes'
          end if

          call IU_close(uout)
       end do
    end if
  end subroutine print_dyn_memory_size

  subroutine write_expr(param_str, expr)
    use Io_Unit_module, only : IU_open, IU_close
    character*(*), intent(in)    :: param_str
    character*(*), intent(in)    :: expr

    character(len=256)           :: tmp_param_str
    character(len=256)           :: tmp_expr

    tmp_param_str = sweep_blanks(param_str)
    tmp_expr = sweep_blanks(expr)

    if (mem_dyn_write.and.enable_dyn_trace) then
       if (pglobal_id.eq.0) then
          call IU_open(uout_dyn_mem, file = dyn_memory_filename, status = 'OLD', &
               position = 'APPEND', form = 'FORMATTED')
          write(uout_dyn_mem,'(A7,A20,A3,A128)') &
               "expr: ", tmp_param_str, "= ", tmp_expr
          UFLUSH(uout_dyn_mem)
          call IU_close(uout_dyn_mem)
       end if
    end if
  end subroutine write_expr

  ! Write param in gysela_dyn_mem.out
  subroutine write_param(param_str, param_val)
    use Io_Unit_module, only : IU_open, IU_close
    character*(*), intent(in)    :: param_str
    integer      , intent(in)    :: param_val

    character(len=256)           :: tmp_param_str

    tmp_param_str = sweep_blanks(param_str)

    if (mem_dyn_write.and.enable_dyn_trace) then
       if (pglobal_id.eq.0) then
          call IU_open(uout_dyn_mem, file = dyn_memory_filename, status = 'OLD', &
               position = 'APPEND', form = 'FORMATTED')
          write(uout_dyn_mem,'(A7,A20,A2,I8)') &
               "param: ", tmp_param_str, "=", param_val
          UFLUSH(uout_dyn_mem)
          call IU_close(uout_dyn_mem)
       end if
    end if
  end subroutine write_param

  ! Write the param in gysela_dyn_mem.out
  subroutine write_param_2(param_str, param_val)
    use Io_Unit_module, only : IU_open, IU_close
    character*(*), intent(in)    :: param_str
    real(RKIND)  , intent(in)    :: param_val

    character(len=256)           :: tmp_param_str

    tmp_param_str = sweep_blanks(param_str)

    if (mem_dyn_write.and.enable_dyn_trace) then
       if (pglobal_id.eq.0) then
          call IU_open(uout_dyn_mem, file = dyn_memory_filename, status = 'OLD', &
               position = 'APPEND', form = 'FORMATTED')
          write(uout_dyn_mem,'(A7,A20,A2,F8.4)') &
               "param: ", tmp_param_str, "=", param_val
          UFLUSH(uout_dyn_mem)
          call IU_close(uout_dyn_mem)
       end if
    end if
  end subroutine write_param_2

  ! Write array in gysela_dyn_mem.out
  subroutine write_array(param_str, param_val)
    use Io_Unit_module, only : IU_open, IU_close
    character*(*), intent(in)    :: param_str
    character*(*), intent(in)    :: param_val

    character(len=256)           :: tmp_param_str

    ! There are more than 1 argument in this arg.
    ! Each argument are separate by a ";"
    tmp_param_str = sweep_blanks(param_str)

    if (mem_dyn_write.and.enable_dyn_trace) then
       if (pglobal_id.eq.0) then
          call IU_open(uout_dyn_mem, file = dyn_memory_filename, status = 'OLD', &
               position = 'APPEND', form = 'FORMATTED')
          write(uout_dyn_mem,'(A7,A20,A2,A80)') &
               "array: ", tmp_param_str, "=", param_val
          UFLUSH(uout_dyn_mem)
          call IU_close(uout_dyn_mem)
       end if
    end if
  end subroutine write_array

end module wrap_alloc_module
