!**************************************************************
!  Copyright Euratom-CEA
!  Authors : 
!     Virginie Grandgirard (virginie.grandgirard@cea.fr)
!     Chantal Passeron (chantal.passeron@cea.fr)
!     Guillaume Latu (guillaume.latu@cea.fr)
!     Xavier Garbet (xavier.garbet@cea.fr)
!     Philippe Ghendrih (philippe.ghendrih@cea.fr)
!     Yanick Sarazin (yanick.sarazin@cea.fr)
!  
!  This code GYSELA (for GYrokinetic SEmi-LAgrangian) 
!  is a 5D gyrokinetic global full-f code for simulating 
!  the plasma turbulence in a tokamak.
!  
!  This software is governed by the CeCILL-B license 
!  under French law and abiding by the rules of distribution 
!  of free software.  You can  use, modify and redistribute 
!  the software under the terms of the CeCILL-B license as 
!  circulated by CEA, CNRS and INRIA at the following URL
!  "http://www.cecill.info". 
!**************************************************************

#include "config-F.inc"

!---------------------------------------------------------
! file : advec1D_BSL.F90
! date : 31/10/2000
!   1D advection in phi and vparallel directions used in
!   a Backward Semi-Lagrangian method
!---------------------------------------------------------
module advec1D_BSL_module
  use globals, only : Nr, Ntheta, Nphi, Nvpar, Nbthread
  use wrap_alloc_module
  use integration_module
  use clock_module
  use geometry_class
  use init_magnetic_class
  use fdistribu5d_class
  use interpolation_module
  use MPIutils_module  
  implicit none
      
  include "gyrokinetic/Bstar_inline.h"
  include "gyrokinetic/velocities_inline.h"
  !******************************
  contains
  !******************************
#include "gyrokinetic/Bstar_inline.F90"
#include "gyrokinetic/velocities_inline.F90"

  !------------------------------------------------------------
  ! splitting in phi direction 
  !  - Solving of : df/dt+vpar/R*df/dphi=0
  !     . dphi = vpar bstar_gradphi + vExB_gradphi + vD_gradphi
  !     . f**(r,theta,f,vpar,mu) = f*(r,theta,phi-dphi,vpar,mu)
  !  with bstar_gradphi = 1/Bstar* [B_gradphi + 
  !                   mi*vpar*mu0*J_gradphi/(e*B_norm)]
  !------------------------------------------------------------
  subroutine advec1D_BSL_phi(geom,dt,init_magnet,init_curr,f,&
       onlynl,onlylin)
    use globals          , only : istart, iend, jstart, jend, &
      dJ0Phidr_PNrPNthNphi, dJ0Phidtheta_PNrPNthNphi, &
      hat_As, pSpecies_id, plocal_id, vstart, vend
#ifdef _OPENMP
    use OMPutils_module  , only : omp_get_thread_num
#endif
    use OMPutils_module  , only : Romp_scoefphi, Romp1_0Nphi, &
      Romp2_0Nphi, Romp3_0Nphi, Romp1_0Nvpar
!R3 #include "code_instrumentation/r3_info.h" !R3
    type(geometry)     , intent(in)    :: geom
    real(RKIND)        , intent(in)    :: dt
    type(init_magnetic), intent(in)    :: init_magnet
    type(init_current) , intent(in)    :: init_curr
    type(fdistribu5d)  , intent(inout) :: f
    logical ,optional  , intent(in)    :: onlynl
    logical ,optional  , intent(in)    :: onlylin

    integer     :: ir, itheta, iphi, ivpar
    real(RKIND) :: phimin, phimax, phistar, disp_phi
    real(RKIND) :: vparl
    real(RKIND) :: bstar_gradphi_tmp
    real(RKIND) :: vExB_gradphi_tmp, vD_gradphi_tmp
    real(RKIND) :: finterpol, nbLphi
    real(RKIND) :: inv_sqrtAs
    integer     :: tid
    
    real(RKIND), dimension(:), pointer :: rhsphi
    real(RKIND), dimension(:), pointer :: phig
    real(RKIND), dimension(:), pointer :: scoefphi

    real(RKIND), dimension(:), pointer :: vExB_gradphi_1D
    real(RKIND), dimension(:), pointer :: vD_gradphi_1D

    real(RKIND)                        :: dphi_tmp
    integer                            :: Nphi_tmp
    logical                            :: l_onlynl, l_onlylin
#ifdef TIMER
    real(RKIND)                        :: tbar1, tbar2, tadvec, tcalc, tmin, tmax
    integer                            :: idmin, idmax
    character(40)                      :: hmin, hmax
    integer(TIMEPREC)                  :: t1, t2
#endif
!R3 call r3_info_begin (r3_info_index_0, 'CPU_advec1d_sl_phi') !R3
    call Write_dyn_balise_open('CPU_advec1d_sl_phi') !R4

    call clck_time(bclock_advec1Dphi)
#ifdef TIMER
    call clck_time(t1)
    call ppbarrier
    call clck_time(t2)
    call clck_ldiff(t1,t2,tbar1)
    t1 = t2
#endif

    l_onlynl = .false.
    if (present(onlynl))  l_onlynl = onlynl
    l_onlylin = .false.
    if (present(onlylin)) l_onlylin = onlylin

    !*** computation of 1/sqrt{hat_As} ***
    inv_sqrtAs = 1._RKIND/sqrt(hat_As(pSpecies_id))

#ifdef _OPENMP
!$OMP PARALLEL private(ir,itheta,iphi,ivpar, &
!$OMP   phimin,phimax,phistar,disp_phi,vparl, &
!$OMP   bstar_gradphi_tmp,vExB_gradphi_tmp,vD_gradphi_tmp, &
!$OMP   vExB_gradphi_1D, vD_gradphi_1D, nbLphi, &
!$OMP   finterpol,rhsphi,scoefphi,phig,tid,Nphi_tmp,&
!$OMP   dphi_tmp) default(shared)
!$OMP BARRIER
    tid = 1+omp_get_thread_num()
#else
    tid = 1
#endif
    
    dphi_tmp   = geom%dphi
    Nphi_tmp   = geom%Nphi
    scoefphi             => Romp_scoefphi(tid)%val
    phig                 => Romp1_0Nphi(tid)%val
    rhsphi               => Romp2_0Nphi(tid)%val
    vExB_gradphi_1D      => Romp3_0Nphi(tid)%val
    vD_gradphi_1D        => Romp1_0Nvpar(tid)%val

    !-> initialisation of phig
    do iphi = 0,Nphi
       phig(iphi) = geom%phig(iphi)
    end do

    phimin = geom%phig(0)
    phimax = geom%phig(Nphi)

!$OMP DO SCHEDULE(static)
    do itheta = jstart,jend 
      do ir = f%istart_modif,f%iend_modif
        !-> computation of vD_gradphi
        call compute_vDgradphi(geom,init_magnet, &
          init_curr,ir,ir,itheta,itheta,0,f%n4,.false., &
          vD_gradphi_1D)
        do ivpar = vstart, vend
          vparl          = geom%vparg(ivpar)
          vD_gradphi_tmp = vD_gradphi_1D(ivpar) 
          do iphi = 0,f%n3-1
            rhsphi(iphi) = f%values(ir,itheta,iphi,ivpar)
          end do
          rhsphi(Nphi) = rhsphi(0)
          !*** cubic spline computation ***
          call period_omp_spline_coef(f%pspline1d_phi, &
            rhsphi,scoefphi)
          !*** splitting in phi direction ***
          !-> computation of bstar.gradphi
#ifdef NO_PRECOMP_BSTAR 
          call compute_bstar_contravariant(geom, &
            init_magnet,init_curr,ir,itheta,ivpar, &
            Sbstar_gradx3=bstar_gradphi_tmp)
#else
          call precomp_bstar_gradphi(ir,itheta,ivpar, &
            bstar_gradphi_tmp)
#endif
          !-> computation of vExB_gradphi
          call compute_vExBgradphi(geom,init_magnet,init_curr, &
            dJ0Phidr_PNrPNthNphi,dJ0Phidtheta_PNrPNthNphi, &
            ir,ir,itheta,itheta,0,f%n3,ivpar,ivpar,.false., &
            vExB_gradphi_1D)
          do iphi = 0,f%n3-1
            vExB_gradphi_tmp = vExB_gradphi_1D(iphi) 
            if (l_onlylin) then
              disp_phi = dt * &
                (inv_sqrtAs*vparl*bstar_gradphi_tmp + &
                vD_gradphi_tmp)
            else
              if (l_onlynl) then
                disp_phi = dt * vExB_gradphi_tmp
              else
                disp_phi = dt * &
                  (inv_sqrtAs*vparl*bstar_gradphi_tmp + &
                  vExB_gradphi_tmp + vD_gradphi_tmp)
              end if
            end if
            phistar = geom%phig(iphi) - disp_phi      
            if (phistar.gt.phimax) then
              nbLphi  = abs(floor((phistar-phimin)/geom%Lphi))
              phistar = phistar - nbLphi*geom%Lphi  
              if (phistar.gt.phimax) phistar = phimax
              if (phistar.lt.phimin) phistar = phimin
            elseif (phistar.lt.phimin) then
              nbLphi  = -abs(floor((phistar-phimin)/geom%Lphi))
              phistar = phistar - nbLphi*geom%Lphi  
              if (phistar.gt.phimax) phistar = phimax
              if (phistar.lt.phimin) phistar = phimin
            endif
            !*** f interpolation ***
            call interpol1d_phi(scoefphi, &
              Nphi_tmp,dphi_tmp,phig,phistar,finterpol)
            rhsphi(iphi) = finterpol
         end do
         do iphi = 0,f%n3-1
            f%values(ir,itheta,iphi,ivpar) = rhsphi(iphi)
         end do
        end do
      end do
    end do
!$OMP END DO
!$OMP BARRIER
!$OMP END PARALLEL


#ifdef TIMER
    call clck_time(t2)
    call clck_ldiff(t1,t2,tcalc)
    t1=t2
    call pp_reduce_scalar(tcalc,tmin,idmin,hmin,tmax,idmax,hmax)
#endif
    call ppbarrier
#ifdef TIMER
    call clck_time(t2)
    call clck_ldiff(t1,t2,tbar2)
#endif
    call clck_time(eclock_advec1Dphi)
    call clck_diff(bclock_advec1Dphi,eclock_advec1Dphi, &
      global_time_advec1Dphi)
#ifdef TIMER
    call clck_ldiff(bclock_advec1Dphi,eclock_advec1Dphi,tadvec)
    if (pglobal_id .eq. 0) then
      write(6,'(I4,A,F13.7,A,F13.7,A,F13.7,A,F13.7,A,F13.7)') &
        pglobal_id, " Temps advection1D_phi,  barrier_beg ", &
        tbar1," barrier_end ",tbar2," advec ",tadvec
    end if
    if (plocal_id .eq. 0) then
      write(6,'(I6,A,2(1X,A,1X,I6,1X,F13.7))') pglobal_id, &
        " adv_phi", hmin, idmin, tmin, hmax, idmax, tmax 
    end if
#endif

    call check_overflow_4D(f,"After advec1D_BSL_phi")

    call Write_dyn_balise_close('CPU_advec1d_sl_phi') !R4
!R3 call r3_info_end (r3_info_index_0) !R3
  end subroutine advec1D_BSL_phi



  !------------------------------------------------------------
  ! splitting in v parallel direction
  !  - Solving of : df/dt+(dvpar/dt)*df/dvpar=0 as:
  !     . f*(r,theta,f,vpar,mu) = fn(r,theta,phi,vpar-dpar,mu)
  !    with
  !     . dpar = dt*[-(hat_Zs/sqrt{hat_As})*gradpar(J0.Phi) -
  !              (1/sqrt{hat_As}*mu*gradpar B + 
  !              K_Bdrift*vpar/B*(vExB.grad_B) ]
  !    where
  !        -> gradpar_J0Phi = vec_bstar.grad_J0Phi
  !        -> gradpar_B     = vec_bstar.grad_B
  !        -> vExB.grad_B   = vExB_gradr*dB/dr + 
  !                          vExB_gradtheta*dB/dtheta
  ! Rk : K_Bdrift=1 if magnetic_drift=.true. and equal 
  !       to 0 otherwise. 
  !------------------------------------------------------------
  subroutine advec1D_BSL_vpar(geom,dt,init_magnet,init_curr,f, &
       onlynl,onlylin)
    use globals        , only : K_Bdrift, &
      pSpecies_id, hat_As, hat_Zs, &
      istart, iend, jstart, jend, &
      dJ0Phidr_PNrPNthNphi, dJ0Phidtheta_PNrPNthNphi, &
      dJ0Phidphi_PNrPNthNphi, plocal_id, vstart, vend
#ifdef _OPENMP
    use OMPutils_module, only : omp_get_thread_num
#endif
    use OMPutils_module, only : Romp_scoefvpar, &
      Romp1_0Nvpar, Romp2_0Nvpar, Romp3_0Nvpar, &
      Romp4_0Nvpar, Romp5_0Nvpar, Romp6_0Nvpar
!R3 #include "code_instrumentation/r3_info.h" !R3
    type(geometry)       , intent(in)    :: geom
    real(RKIND)          , intent(in)    :: dt
    type(init_magnetic)  , intent(in)    :: init_magnet
    type(init_current)   , intent(in)    :: init_curr
    type(fdistribu5d)    , intent(inout) :: f
    logical , optional   , intent(in)    :: onlynl
    logical , optional   , intent(in)    :: onlylin

    integer     :: ir, itheta, iphi, ivpar
    real(RKIND) :: mu_value, vstar, disp_vpar, vparl
    real(RKIND) :: inv_Bij, hatZs, inv_sqrtAs
    real(RKIND) :: finterpol
    real(RKIND) :: dBdr_tmp, dBdtheta_tmp
    real(RKIND) :: gradpar_J0Phi_tmp, gradpar_B_tmp
    real(RKIND) :: vExB_gradB_tmp, dvpar_tmp
    integer     :: tid, Nvpar_tmp
    logical     :: l_onlynl, l_onlylin

    real(RKIND), dimension(:), pointer :: rhsvpar
    real(RKIND), dimension(:), pointer :: scoefvpar
    real(RKIND), dimension(:), pointer :: vparg

    !-> components of ExB drift velocity
    real(RKIND), dimension(:), pointer :: vExB_gradr_1D
    real(RKIND), dimension(:), pointer :: vExB_gradtheta_1D
    !-> gradpar.B
    real(RKIND), dimension(:), pointer :: gradpar_B_1D
    !-> gradpar.Phi
    real(RKIND), dimension(:), pointer :: gradpar_J0Phi_1D

#ifdef TIMER
    real(RKIND)                        :: tbar, tadvec, tcalc, tmin, tmax
    integer                            :: idmin, idmax
    character(40)                      :: hmin, hmax
    integer(TIMEPREC)                  :: t1, t2
#endif

#ifdef TIMER
    call clck_time(t1)
#endif
    call ppbarrier
#ifdef TIMER
    call clck_time(t2)
    call clck_ldiff(t1,t2,tbar)
    t1 = t2
#endif

!R3 call r3_info_begin (r3_info_index_0, 'CPU_advec1d_sl_vpar') !R3
    call Write_dyn_balise_open('CPU_advec1d_sl_vpar') !R4

    call clck_time(bclock_advec1Dvpar)

    l_onlynl = .false.			      
    if (present(onlynl))  l_onlynl = onlynl  
    l_onlylin = .false.		      
    if (present(onlylin)) l_onlylin = onlylin

    !*** initialisation of mu value ***
    mu_value = f%mu_value

    !*** computation of 1/sqrt{hat_As} ***
    hatZs      = hat_Zs(pSpecies_id)
    inv_sqrtAs = 1._RKIND/sqrt(hat_As(pSpecies_id))
 
#ifdef _OPENMP
!$OMP PARALLEL private(ir,itheta,iphi,ivpar,vstar,disp_vpar,vparl, &
!$OMP inv_Bij,finterpol,dBdr_tmp,dBdtheta_tmp, &
!$OMP gradpar_J0Phi_tmp,gradpar_B_tmp,vExB_gradB_tmp, &
!$OMP vExB_gradr_1D,vExB_gradtheta_1D, &
!$OMP gradpar_B_1D,gradpar_J0Phi_1D, &
!$OMP rhsvpar,scoefvpar,vparg,tid,Nvpar_tmp,dvpar_tmp) &
!$OMP default(shared)
!$OMP BARRIER
    tid = 1+omp_get_thread_num()
#else
    tid = 1
#endif

    dvpar_tmp = geom%dvpar
    Nvpar_tmp = geom%Nvpar
    scoefvpar => Romp_scoefvpar(tid)%val
    vparg     => Romp1_0Nvpar(tid)%val
    rhsvpar   => Romp2_0Nvpar(tid)%val

    vExB_gradr_1D     => Romp3_0Nvpar(tid)%val
    vExB_gradtheta_1D => Romp4_0Nvpar(tid)%val
    gradpar_B_1D      => Romp5_0Nvpar(tid)%val
    gradpar_J0Phi_1D  => Romp6_0Nvpar(tid)%val

    do ivpar = 0,Nvpar
       vparg(ivpar) = geom%vparg(ivpar)
    end do

!$OMP DO SCHEDULE(static)
    do iphi = 0,Nphi-1
      !*** vpar advection ***
      do itheta = jstart,jend
        do ir = f%istart_modif,f%iend_modif 
          inv_Bij      = 1._RKIND/init_magnet%B_norm(ir,itheta)
          dBdr_tmp     = init_magnet%dBdr(ir,itheta)
          dBdtheta_tmp = init_magnet%dBdtheta(ir,itheta)
          !-> computation of vExB contravariant components
          call compute_vExBgradr(geom,init_magnet,init_curr, &
            dJ0Phidtheta_PNrPNthNphi,dJ0Phidphi_PNrPNthNphi, &
            ir,ir,itheta,itheta,iphi,iphi,0,Nvpar,.false., &
            vExB_gradr_1D)
          call compute_vExBgradtheta(geom,init_magnet,init_curr, &
            dJ0Phidr_PNrPNthNphi,dJ0Phidphi_PNrPNthNphi, &
            ir,ir,itheta,itheta,iphi,iphi,0,Nvpar,.false., &
            vExB_gradtheta_1D)
          !-> computation of gradpar.B
          call compute_gradpar_B(geom,init_magnet,init_curr, &
            ir,ir,itheta,itheta,iphi,iphi,0,Nvpar,.false., &
            gradpar_B_1D)
          !-> computation of gradpar.Phi
          call compute_gradpar_J0Phi(geom, &
            init_magnet,init_curr,ir,ir,itheta,itheta, &
            iphi,iphi,0,Nvpar,.false., gradpar_J0Phi_1D)
          !-> cubic spline computation 
          do ivpar = 0,Nvpar
            rhsvpar(ivpar) = f%values(ir,itheta,iphi,ivpar)
          end do
          call natural_omp_spline_coef(f%nspline1d_vpar(tid), &
            rhsvpar,f%BCvpar_left,f%BCvpar_right,scoefvpar)
          !*** splitting in vpar direction ***          
          do ivpar = 0,Nvpar
            vparl             = vparg(ivpar)
            gradpar_B_tmp     = gradpar_B_1D(ivpar)
            gradpar_J0Phi_tmp = gradpar_J0Phi_1D(ivpar)
            !-> computation of vExB.gradB
            vExB_gradB_tmp = vExB_gradr_1D(ivpar)*dBdr_tmp + &
              vExB_gradtheta_1D(ivpar)*dBdtheta_tmp
            !-> computation of vpar displacement
            if (l_onlylin) then
              disp_vpar = dt * &
                (-inv_sqrtAs*mu_value*gradpar_B_tmp)
            else
              if (l_onlynl) then
                disp_vpar = dt * &
                  (-(hatZs*inv_sqrtAs)*gradpar_J0Phi_tmp + &
                  K_Bdrift*vparl*inv_Bij*vExB_gradB_tmp)
              else
                disp_vpar = dt * &
                  (-inv_sqrtAs*mu_value*gradpar_B_tmp - &
                  (hatZs*inv_sqrtAs)*gradpar_J0Phi_tmp + &
                  K_Bdrift*vparl*inv_Bij*vExB_gradB_tmp)
              end if
            end if
            if (ivpar .lt. vstart .or. ivpar .gt. vend) disp_vpar = 0._RKIND
            vstar       = vparg(ivpar) - disp_vpar       
            if (vstar .lt. vparg(0)) vstar = vparg(0)
            if (vstar .gt. vparg(Nvpar_tmp)) vstar = vparg(Nvpar_tmp)
            !-> f interpolation 
            call interpol1d_vpar(scoefvpar, &
              Nvpar_tmp,dvpar_tmp,vparg,vstar,finterpol)
            rhsvpar(ivpar) = finterpol
          end do
          do ivpar = vstart,vend
             f%values(ir,itheta,iphi,ivpar) = rhsvpar(ivpar)
          end do
        end do
      end do
    end do
!$OMP END DO
!$OMP BARRIER
!$OMP END PARALLEL

#ifdef TIMER
    call clck_time(t2)
    call clck_ldiff(t1,t2,tcalc)
    call pp_reduce_scalar(tcalc,tmin,idmin,hmin,tmax,idmax,hmax)
    t1 = t2
#endif
    call ppbarrier
#ifdef TIMER
    call clck_time(t2)
    call clck_ldiff(t1,t2,tbar)
#endif
    call clck_time(eclock_advec1Dvpar)
    call clck_diff(bclock_advec1Dvpar,eclock_advec1Dvpar, &
      global_time_advec1Dvpar)

#ifdef TIMER
    call clck_ldiff(bclock_advec1Dvpar,eclock_advec1Dvpar,tadvec)
    if (pglobal_id .eq. 0) then
      write(6,'(I4,A,F13.7,A,F13.7,A,F13.7,A,F13.7,A,F13.7)') &
        pglobal_id, " Temps advection1D_vpar, barrier ", &
        tbar," advec ",tadvec
    end if
    if (plocal_id .eq. 0) then
      write(6,'(I6,A,2(1X,A,1X,I6,1X,F13.7))') pglobal_id, " adv_vpar", hmin, idmin, tmin, hmax, idmax, tmax 
    end if
#endif

    call check_overflow_4D(f,"After advec1D_BSL_vpar")

    call Write_dyn_balise_close('CPU_advec1d_sl_vpar') !R4
!R3 call r3_info_end (r3_info_index_0) !R3
  end subroutine advec1D_BSL_vpar
end module advec1D_BSL_module
