!**************************************************************
!  Copyright Euratom-CEA
!  Authors : 
!     Virginie Grandgirard (virginie.grandgirard@cea.fr)
!     Chantal Passeron (chantal.passeron@cea.fr)
!     Guillaume Latu (guillaume.latu@cea.fr)
!     Xavier Garbet (xavier.garbet@cea.fr)
!     Philippe Ghendrih (philippe.ghendrih@cea.fr)
!     Yanick Sarazin (yanick.sarazin@cea.fr)
!  
!  This code GYSELA (for GYrokinetic SEmi-LAgrangian) 
!  is a 5D gyrokinetic global full-f code for simulating 
!  the plasma turbulence in a tokamak.
!  
!  This software is governed by the CeCILL-B license 
!  under French law and abiding by the rules of distribution 
!  of free software.  You can  use, modify and redistribute 
!  the software under the terms of the CeCILL-B license as 
!  circulated by CEA, CNRS and INRIA at the following URL
!  "http://www.cecill.info". 
!**************************************************************

#include "config-F.inc"

!---------------------------------------------------------
! file : advec1D_PPM2.F90
! date : 18/10/2009
!   1D advection in r and theta directions used in
!   PPM2 Colella-Sekora JCP 2008
! developed by 
!  Jean-Philippe Braeunig (braeunig@math.u-strasbg.fr)
!---------------------------------------------------------
module advec1D_PPM2_module
  use globals, only : Nr, Ntheta, Nphi, Nvpar, Nbthread
  use wrap_alloc_module
  use integration_module
  use clock_module
  use geometry_class
  use init_profile_class
  use init_magnetic_class
  use fdistribu5d_class
  use interpolation_module
  use velocities_edges_module
  
  implicit none

  integer    , parameter, private :: countmax = 50
  real(RKIND), parameter, private :: Kppm2    = 1.5_RKIND

  include "gyrokinetic/Bstar_inline.h"
  include "gyrokinetic/velocities_inline.h"
  !******************************
  contains
  !******************************
#include "gyrokinetic/Bstar_inline.F90"
#include "gyrokinetic/velocities_inline.F90"

  !******************************  
  !--------------------------------------------------------
  ! splitting in r parallel direction
  !  - Resolution of : df/dt+(dr/dt)*df/dr=0
  !         dr/dt = vpar*bstar_gradr +  vExB_gradr 
  !                 + vD_gradr   
  !
  !  . f*(r,theta,f,vpar,mu) = fn(r-dr,theta,phi,vpar,mu)
  !--------------------------------------------------------
  subroutine advec1D_PPM2_r(geom,dt,init_prof, &
    init_magnet,init_curr,f,maxalp_r)
    use MPIutils_module, only : f4d_transp
    use globals        , only : istart, iend, &
      jstart, jend, kstart, kend, lstart, lend, &
      deltat, errmax, limiter
#ifdef _OPENMP
    use OMPutils_module, only : omp_get_thread_num
#endif
    use OMPutils_module, only : Romp_alpha, Romp_rhs, &
      Romp1_0Nrp1, Romp2_0Nrp1, & 
      Romp1_0Nr, Romp2_0Nr, Romp3_0Nr, Romp4_0Nr, &
      Romp5_0Nr
    use velocities_edges_module, only : Phiedges_rtheta, &
      Phiedges_rphi, comp_vExBgradr_edges
!R3 #include "code_instrumentation/r3_info.h" !R3
    type(geometry)     , intent(in)    :: geom
    real(RKIND)        , intent(in)    :: dt
    type(init_profile) , intent(in)    :: init_prof
    type(init_magnetic), intent(in)    :: init_magnet
    type(init_current) , intent(in)    :: init_curr
    type(fdistribu5d)  , intent(inout) :: f
    real(RKIND)        , intent(out)   :: maxalp_r
#ifdef HAS_ADV_PPM

    integer     :: ir, itheta, iphi, ivpar
    real(RKIND) :: rmin, rmax, rstar
    real(RKIND) :: depr, xr, rn
    real(RKIND) :: sum, err
    real(RKIND) :: bstar_gradr_left, bstar_gradr_right
    real(RKIND) :: bstar_gradr_edge
    real(RKIND) :: vD_gradr_left, vD_gradr_right
    real(RKIND) :: vD_gradr_edge
    real(RKIND) :: vparl
    real(RKIND) :: finterpol, Fcentre
    integer     :: tid, ileft, iright, irpos
    integer     :: ig, id, igg, idd, iggg, iddd
    integer     :: k, count
    character(LEN=50) :: subp

    ! array for the edges of the cell (0:geom%Nr+1)
    real(RKIND), dimension(:), pointer :: alpha  
    ! array for the edges of the cell (-1:geom%Nr+2)
    real(RKIND), dimension(:), pointer :: udt   

    !-> radial components of ExB drift velocity
    real(RKIND), dimension(:), pointer :: vExB_gradr_1D
    !-> radial components of curvature drift velocity
    real(RKIND), dimension(:), pointer :: vD_gradr_1D

    real(RKIND) :: dp, dpc, lim, flux_upw, xx
    real(RKIND) :: Mdd, Mdd_c, Mdd_l, Mdd_r, Mdd_lim
    real(RKIND) :: Ndd, Ndd_c, Ndd_l, Ndd_r, Ndd_lim
    real(RKIND), dimension(1:Nbthread) :: Romp_maxalp_r
    !-> array of size (0:Nr)
    real(RKIND), dimension(:), pointer :: ar, al, delta, a6
    !-> array of size (0:Nr+1)
    real(RKIND), dimension(:), pointer :: flux

!R3 call r3_info_begin (r3_info_index_0, 'CPU_advec1d_PPM2_r') !R3
    call Write_dyn_balise_open('CPU_advec1d_PPM2_r') !R4
    call clck_time(bclock_advec1Dr)

#ifdef _OPENMP
!$OMP PARALLEL default(none) &
!$OMP private(tid,ir,itheta,iphi,ivpar, &
!$OMP  rmin,rmax,alpha,udt,sum,depr,vparl, &
!$OMP  bstar_gradr_left,bstar_gradr_right, &
!$OMP  vD_gradr_left,vD_gradr_right, &
!$OMP  bstar_gradr_edge,vD_gradr_edge, &
!$OMP  vExB_gradr_1D,vD_gradr_1D, &
!$OMP  ileft,iright,irpos,rn,rstar,err,count,xr, &
!$OMP  finterpol,Fcentre,iddd,idd,id,ig,igg,iggg,xx, &
!$OMP  Mdd,Mdd_c,Mdd_l,Mdd_r,Mdd_lim, &
!$OMP  Ndd,Ndd_c,Ndd_l,Ndd_r,Ndd_lim, &
!$OMP  ar,al,delta,a6,flux,subp ) & 
!$OMP shared(kstart,kend,lstart,lend,Nr,Ntheta, &
!$OMP  Romp_alpha,Romp_rhs, Romp_maxalp_r, &
!$OMP  Romp1_0Nrp1,Romp2_0Nrp1,Romp1_0Nr, & 
!$OMP  Romp2_0Nr,Romp3_0Nr,Romp4_0Nr,Romp5_0Nr, &
!$OMP  f,geom,init_magnet,init_curr, &
!$OMP  Phiedges_rtheta,Phiedges_rphi,dt, &
!$OMP  errmax,f4D_transp)
!$OMP BARRIER
    tid = 1+omp_get_thread_num()
#else
    tid = 1
#endif

    subp = " advec1D_PPM2_r "//char(0)

    rmin = f%hhspline2d_rtheta(tid)%xgrid(0)
    rmax = f%hhspline2d_rtheta(tid)%xgrid(Nr+1)

    alpha         => Romp_alpha(tid)%val
    udt           => Romp_rhs(tid)%val
    vExB_gradr_1D => Romp1_0Nrp1(tid)%val
    vD_gradr_1D   => Romp1_0Nr(tid)%val  
    ar            => Romp2_0Nr(tid)%val
    al            => Romp3_0Nr(tid)%val
    delta         => Romp4_0Nr(tid)%val
    a6            => Romp5_0Nr(tid)%val
    flux          => Romp2_0Nrp1(tid)%val

    Romp_maxalp_r(tid) = 0._RKIND
    alpha              = 0._RKIND

!$OMP DO SCHEDULE(static)
    do iphi = kstart,kend
      do ivpar = lstart,lend
        vparl = geom%vparg(ivpar)
        do itheta = 0,Ntheta-1
!baoter (ATTENTION prendre en compte correctement le jacobien)
          !*** Jacobian factor ***
          f4d_transp(:,itheta,iphi,ivpar) = &
            f4d_transp(:,itheta,iphi,ivpar)*geom%rg(:)
!eaoter
          !-> compute vExB.gradr
          call comp_vExBgradr_edges(geom, &
            init_magnet,init_curr, &
            Phiedges_rtheta,Phiedges_rphi, &
            0,Nr+1,itheta,itheta, &
            iphi,iphi,ivpar,ivpar, &
            vExB_gradr_1D)
          !-> compute vD.gradr 
          call compute_vDgradr(geom, &
            init_magnet,init_curr,0,Nr, &
            itheta,itheta,iphi,iphi,ivpar, &
            ivpar,.true.,vD_gradr_1D)

          do ir = 0,Nr+1
            ileft  = max(ir-1,0)
            iright = min(ir,Nr) 
            !-> compute bstar.gradr at the edge
#ifdef NO_PRECOMP_BSTAR 
            call compute_bstar_contravariant(geom, &
              init_magnet,init_curr,ileft,itheta,ivpar, &
              Sbstar_gradx1=bstar_gradr_left)
            call compute_bstar_contravariant(geom, &
              init_magnet,init_curr,iright,itheta,ivpar, &
              Sbstar_gradx1=bstar_gradr_right)
#else
            call precomp_transp_bstar_gradr(ileft, &
              itheta,ivpar,bstar_gradr_left)
            call precomp_transp_bstar_gradr(iright, &
              itheta,ivpar,bstar_gradr_right)
#endif
            bstar_gradr_edge = 0.5_RKIND * &
              (bstar_gradr_left+bstar_gradr_right)
            !-> compute vD_gradr at the edge
            vD_gradr_left  = vD_gradr_1D(ileft)
            vD_gradr_right = vD_gradr_1D(iright)
            vD_gradr_edge  = 0.5_RKIND * &
              (vD_gradr_left+vD_gradr_right)
            !-> computation of alpha
            alpha(ir)      = vparl*bstar_gradr_edge + &
              vExB_gradr_1D(ir) + vD_gradr_edge  
          end do
          alpha              = dt*alpha
          Romp_maxalp_r(tid) = max(Romp_maxalp_r(tid), &
            maxval(abs(alpha(0:Nr+1))))
          
          depr = 0._RKIND 
          do ir = 0,Nr+1
            ileft  = max(ir-1,0)
            iright = min(ir,Nr)
            depr   = alpha(ir)
            rn     = &
              f%hhspline2d_rtheta(tid)%xgrid(ir) - depr
            rstar  = 0.5_RKIND * &
              (f%hhspline2d_rtheta(tid)%xgrid(ir)+rn)
            err    = 1._RKIND
            count  = 0
            do while (err>errmax.and.count<countmax)
              !*** Inside the domain ***
              if ((rstar.gt.rmin).and.(rstar.lt.rmax)) then
                !-> array position location
                call locate(rstar, &
                  f%hhspline2d_rtheta(tid)%xgrid(:), &
                  geom%Nr+1,geom%dr,subp,irpos)
                !-> relative position in the cell 
                xr   = (rstar - &
                  f%hhspline2d_rtheta(tid)%xgrid(irpos)) / &
                  geom%dr
                depr = alpha(irpos  )*(1._RKIND-xr) + &
                  alpha(irpos+1)*xr 
                !-> Out of the domain by the left
              elseif (rstar.le.rmin) then
                depr = alpha(0)
                !-> Out of the domain by the right
              elseif (rstar.ge.rmax) then
                depr = alpha(Nr+1)
              endif
              err   = rstar
              rn    = f%hhspline2d_rtheta(tid)%xgrid(ir)-depr
              rstar = 0.5_RKIND * &
                (f%hhspline2d_rtheta(tid)%xgrid(ir)+rn)
              err   = abs(rstar-err)/geom%dr
              count = count+1
            enddo !while
            udt(ir)=depr
          enddo

          !*** PPM2 ***
          do ir = 0,Nr+1
            iddd = min(ir+2,Nr)
            idd  = min(ir+1,Nr)
            id   = min(ir  ,Nr)
            ig   = max(ir-1,0)
            igg  = max(ir-2,0)
            iggg = max(ir-3,0)

            flux(ir) = &
              ( 37._RKIND*(f4d_transp(ig,itheta,iphi,ivpar) &
              + f4d_transp(id,itheta,iphi,ivpar)) &
              - 8._RKIND*(f4d_transp(igg,itheta,iphi,ivpar) &
              + f4d_transp(idd,itheta,iphi,ivpar)) &
              + f4d_transp(iggg,itheta,iphi,ivpar) &
              + f4d_transp(iddd,itheta,iphi,ivpar) )/60._RKIND
       
            !--> second derivative of f at the nodes
            if ( (f4d_transp(ig,itheta,iphi,ivpar)-flux(ir)) * &
              (f4d_transp(id,itheta,iphi,ivpar)-flux(ir)) &
              > 0._RKIND) then
              Ndd = 3._RKIND * &
                ( f4d_transp(ig,itheta,iphi,ivpar) - &
                2._RKIND*flux(ir) + &
                f4d_transp(id,itheta,iphi,ivpar) )/(geom%dr**2)
              Ndd_l = ( f4d_transp(igg,itheta,iphi,ivpar) - &
                2._RKIND*f4d_transp(ig,itheta,iphi,ivpar) + &
                f4d_transp(id,itheta,iphi,ivpar) )/(geom%dr**2)
              Ndd_r = ( f4d_transp(ig,itheta,iphi,ivpar) - &
                2._RKIND*f4d_transp(id,itheta,iphi,ivpar) + &
                f4d_transp(idd,itheta,iphi,ivpar))/(geom%dr**2)

              if ( Ndd.ge.0._RKIND .and. Ndd_l.ge.0._RKIND &
                .and. Ndd_r.ge.0._RKIND) then
                Ndd_lim = min( Ndd, Kppm2*Ndd_l, &
                  Kppm2*Ndd_r )
              elseif ( Ndd.lt.0._RKIND .and. &
                Ndd_l.lt.0._RKIND .and. &
                Ndd_r.lt.0._RKIND) then
                Ndd_lim = max( Ndd, Kppm2*Ndd_l, &
                  Kppm2*Ndd_r )
              else 
                Ndd_lim = 0._RKIND
              endif
              flux(ir) = 0.5_RKIND * &
                ( f4d_transp(ig,itheta,iphi,ivpar) + &
                f4d_transp(id,itheta,iphi,ivpar) ) - &
                (geom%dr**2)*Ndd_lim/3._RKIND
            endif
          enddo

          do ir = 0,Nr
            al(ir)    = flux(ir)
            ar(ir)    = flux(ir+1)
            delta(ir) = ar(ir)-al(ir)
            a6(ir)    = 6._RKIND * &
              ( f4d_transp(ir,itheta,iphi,ivpar) - &
              (al(ir)+ar(ir))/2._RKIND )

            !--> Local min or max 
            if ( (ar(ir)-f4d_transp(ir,itheta,iphi,ivpar)) * &
              (f4d_transp(ir,itheta,iphi,ivpar)-al(ir)) &
              > 0._RKIND ) then
              !----> Non mono L
              if ( delta(ir)*a6(ir) > delta(ir)**2 ) then
                al(ir) =  3._RKIND * &
                  f4d_transp(ir,itheta,iphi,ivpar) - 2._RKIND*ar(ir)
              endif
              !----> Non mono R
              if ( -delta(ir)**2 > delta(ir)*a6(ir) ) then
                ar(ir) =  3._RKIND * &
                  f4d_transp(ir,itheta,iphi,ivpar) - 2._RKIND*al(ir)
              endif
            else 
              iddd = min(ir+2,Nr)
              idd  = min(ir+1,Nr)
              id   = min(ir  ,Nr)
              ig   = max(ir-1,0)
              igg  = max(ir-2,0)

              !----> Limitation
              Mdd  = -2._RKIND*a6(ir)/(geom%dr**2)

              !----> second derivative of f at the cells
              Mdd_c = ( f4d_transp(ig,itheta,iphi,ivpar) - &
                2._RKIND*f4d_transp(id,itheta,iphi,ivpar) + &
                f4d_transp(idd ,itheta,iphi,ivpar)) / (geom%dr**2)
              Mdd_l = ( f4d_transp(igg,itheta,iphi,ivpar) - &
                2._RKIND*f4d_transp(ig,itheta,iphi,ivpar) + &
                f4d_transp(id  ,itheta,iphi,ivpar)) / (geom%dr**2)
              Mdd_r = ( f4d_transp(id,itheta,iphi,ivpar) - &
                2._RKIND*f4d_transp(idd,itheta,iphi,ivpar) + &
                f4d_transp(iddd,itheta,iphi,ivpar)) / (geom%dr**2)

              if ( Mdd>=0._RKIND .and. Mdd_c>=0._RKIND .and. &
                Mdd_l>=0._RKIND .and. Mdd_r>=0._RKIND) then
                Mdd_lim = min( Mdd,Kppm2*Mdd_c, &
                  Kppm2*Mdd_l,Kppm2*Mdd_r )
              elseif ( Mdd<0._RKIND .and. Mdd_c<0._RKIND .and. &
                Mdd_l<0._RKIND .and. Mdd_r<0._RKIND) then
                Mdd_lim = max( Mdd,Kppm2*Mdd_c, &
                  Kppm2*Mdd_l,Kppm2*Mdd_r )
              else
                Mdd_lim = 0._RKIND
              endif
              if (Mdd/=0._RKIND) then
                al(ir) = f4d_transp(ir,itheta,iphi,ivpar) + &
                  (al(ir)-f4d_transp(ir,itheta,iphi,ivpar)) * &
                  Mdd_lim/Mdd
                ar(ir) = f4d_transp(ir,itheta,iphi,ivpar) + &
                  (ar(ir)-f4d_transp(ir,itheta,iphi,ivpar)) * &
                  Mdd_lim/Mdd
              else
                al(ir) = f4d_transp(ir,itheta,iphi,ivpar)
                ar(ir) = f4d_transp(ir,itheta,iphi,ivpar)
              endif
            endif
            delta(ir) = ar(ir)-al(ir)
            a6(ir)    = 6._RKIND * &
              (f4d_transp(ir,itheta,iphi,ivpar) - &
              (al(ir)+ar(ir))/2._RKIND)
            ! end limitation
          enddo

          do ir = 1,Nr
            xx = udt(ir)/geom%dr
            if (abs(xx)>1.d0) xx = sign(1.d0,xx)
            if (udt(ir)>=0._RKIND) then
              flux(ir) = ar(ir-1) + &
                0.5_RKIND*xx*(a6(ir-1) - delta(ir-1)) - &
                a6(ir-1)*(xx**2/3._RKIND) 
            else
              flux(ir) = al(ir) - &
                0.5_RKIND*xx*(a6(ir) + delta(ir)) - &
                a6(ir)*(xx**2/3._RKIND) 
            endif
            flux(ir) = udt(ir)*flux(ir)
          enddo

          !*** PPM2 boundary conditions ***
          xx = udt(0)/geom%dr 
          if (abs(xx)>1.d0) xx = sign(1.d0,xx)
          if (udt(0)<=0._RKIND) then
            flux(0) = al(0) - 0.5_RKIND*xx*(a6(0) + delta(0)) &
              - a6(0)*(xx**2/3._RKIND)
          else
            flux(0) = f4d_transp(0,itheta,iphi,ivpar)
          endif
          flux(0) = udt(0)*flux(0)

          xx = udt(Nr+1)/geom%dr 
          if (abs(xx)>1.d0) xx = sign(1.d0,xx)
          if (udt(Nr+1)>=0._RKIND) then
            flux(Nr+1) = ar(Nr) + 0.5_RKIND*xx* &
              (a6(Nr)-delta(Nr)) - a6(Nr)*(xx**2/3._RKIND) 
          else   
            flux(Nr+1) = f4d_transp(Nr,itheta,iphi,ivpar)
          endif
          flux(Nr+1) = udt(Nr+1)*flux(Nr+1)

          do ir = 0,Nr
            f4d_transp(ir,itheta,iphi,ivpar) = &
              f4d_transp(ir,itheta,iphi,ivpar) - &
              (1._RKIND/geom%dr)*(flux(ir+1)-flux(ir))
          enddo

!baoter (ATTENTION prendre en compte correctement le jacobien)
          !*** Jacobian factor ***
          f4d_transp(:,itheta,iphi,ivpar) = &
            f4d_transp(:,itheta,iphi,ivpar)/geom%rg(:)
!eaoter
        end do
      end do
    end do
!$OMP END DO
!$OMP BARRIER
!$OMP END PARALLEL

    maxalp_r = maxval(Romp_maxalp_r(1:Nbthread))/(geom%dr*dt)

    call clck_time(eclock_advec1Dr)
    call clck_diff(bclock_advec1Dr,eclock_advec1Dr, &
      global_time_advec1Dr)

    call Write_dyn_balise_close('CPU_advec1d_PPM2_r') !R4
!R3 call r3_info_end (r3_info_index_0) !R3
#else 
    ! HAS_ADV_PPM
    maxalp_r = 0._RKIND
#endif 
    ! HAS_ADV_PPM
  end subroutine advec1D_PPM2_r


  !------------------------------------------------------------
  ! splitting in theta direction 
  !  - Solving of : df/dt+dtheta/dt*df/dtheta=0
  !           dtheta/dt = vparl*b_gradtheta_tmp +
  !                       vExB_gradtheta_tmp + 
  !                       vD_gradtheta_tmp
  !------------------------------------------------------------
  subroutine advec1D_PPM2_theta(geom,dt,init_prof, &
    init_magnet,init_curr,f,maxalp_theta)
    use MPIutils_module, only : f4d_transp
    use globals        , only : istart, iend, &
      jstart, jend, kstart, kend, lstart, lend, &
      deltat, errmax, limiter
#ifdef _OPENMP
    use OMPutils_module, only : omp_get_thread_num
#endif
    use OMPutils_module, only : Romp_alpha, Romp_rhs, &
      Romp_vect, Romp1_0Nthetap1, Romp1_0Ntheta, &
      Romp2_0Ntheta, Romp3_0Ntheta, Romp4_0Ntheta, &
      Romp5_0Ntheta, Romp6_0Ntheta
    use velocities_edges_module, only : Phiedges_rtheta, &
      Phiedges_thetaphi, comp_vExBgradtheta_edges
!R3 #include "code_instrumentation/r3_info.h" !R3
    type(geometry)     , intent(in)    :: geom
    real(RKIND)        , intent(in)    :: dt
    type(init_profile) , intent(in)    :: init_prof
    type(init_magnetic), intent(in)    :: init_magnet
    type(init_current) , intent(in)    :: init_curr
    type(fdistribu5d)  , intent(inout) :: f
    real(RKIND)        , intent(out)   :: maxalp_theta

#ifdef HAS_ADV_PPM
    integer     :: ir, itheta, iphi, ivpar
    real(RKIND) :: thetamin, thetamax, thetastar
    real(RKIND) :: sum, thetan, err
    real(RKIND) :: vparl
    real(RKIND) :: finterpol, Fcentre,rfootl,rfootr,divu
    real(RKIND) :: out, xtheta, deptheta,xx
    integer     :: tid, ileft, iright, ithetapos
    integer     :: k, count
    integer     :: iup, ig, id, idd, igg, iddd, iggg
    real(RKIND) :: der_bc1, der_bc2
    real(RKIND) :: bstar_gradtheta_left, bstar_gradtheta_right
    real(RKIND) :: bstar_gradtheta_edge
    real(RKIND) :: vD_gradtheta_left, vD_gradtheta_right
    real(RKIND) :: vD_gradtheta_edge
    character(LEN=50) :: subp

    !-> array of size (0:geom%Ntheta)
    real(RKIND), dimension(:), pointer :: alpha
    !-> array of size (-1:geom%Ntheta+1)
    real(RKIND), dimension(:), pointer :: udt
    !-> array of size (-10:10)
    real(RKIND), dimension(:), pointer :: vect

    !-> poloidal components of ExB drift velocity
    real(RKIND), dimension(:), pointer :: vExB_gradtheta_1D
    !-> poloidal components of curvature drift velocity
    real(RKIND), dimension(:), pointer :: vD_gradtheta_1D

    real(RKIND) :: Mdd, Mdd_c, Mdd_l, Mdd_r, Mdd_lim
    real(RKIND) :: Ndd, Ndd_c, Ndd_l, Ndd_r, Ndd_lim
    real(RKIND), dimension(1:Nbthread) :: Romp_maxalp_th
    !--> array of size (0:Ntheta)
    real(RKIND), dimension(:), pointer :: ar, al, delta, a6
    !--> array of size (0:Ntheta+1)
    real(RKIND), dimension(:), pointer :: flux

!R3 call r3_info_begin(r3_info_index_0,'CPU_advec1d_PPM2_theta') !R3
    call Write_dyn_balise_open('CPU_advec1d_PPM2_theta') !R4
    call clck_time(bclock_advec1Dtheta)

!baremettre
#ifdef _OPENMP
!$OMP PARALLEL default(none) & 
!$OMP private(tid,ir,itheta,iphi,ivpar, &
!$OMP  alpha,udt,vect,thetamin,thetamax, &
!$OMP  sum,der_bc1,deptheta,ileft,iright,ithetapos, &
!$OMP  vD_gradtheta_left,vD_gradtheta_right, &
!$OMP  bstar_gradtheta_left,bstar_gradtheta_right, &
!$OMP  bstar_gradtheta_edge,vD_gradtheta_edge, &
!$OMP  vExB_gradtheta_1D,vD_gradtheta_1D, &
!$OMP  vparl,thetan,thetastar,out,err,count,xtheta, &
!$OMP  finterpol,Fcentre,iddd,idd,id,ig,igg,iggg,xx, &
!$OMP  Mdd,Mdd_c,Mdd_l,Mdd_r,Mdd_lim, &
!$OMP  Ndd,Ndd_c,Ndd_l,Ndd_r,Ndd_lim, &
!$OMP  ar,al,delta,a6,flux,subp ) &
!$OMP shared(kstart,kend,lstart,lend,Nr,Ntheta, &
!$OMP  Romp_alpha,Romp_rhs, &
!$OMP  Romp_vect,Romp1_0Ntheta,Romp2_0Ntheta, &
!$OMP  Romp3_0Ntheta,Romp4_0Ntheta,Romp5_0Ntheta, &
!$OMP  Romp6_0Ntheta,Romp1_0Nthetap1, &  
!$OMP  Romp_maxalp_th,f,geom,init_magnet,init_curr, &
!$OMP  Phiedges_rtheta,Phiedges_thetaphi,dt, &
!$OMP  errmax,f4D_transp)
!$OMP BARRIER
    tid = 1+omp_get_thread_num()
#else
    tid = 1
#endif
!earemettre

    subp = " advec1D_PPM2_theta "//char(0)

    alpha             => Romp_alpha(tid)%val
    udt               => Romp_rhs(tid)%val
    vect              => Romp_vect(tid)%val
    vExB_gradtheta_1D => Romp1_0Ntheta(tid)%val
    vD_gradtheta_1D   => Romp2_0Ntheta(tid)%val
    ar                => Romp3_0Ntheta(tid)%val  
    al                => Romp4_0Ntheta(tid)%val
    delta             => Romp5_0Ntheta(tid)%val
    a6                => Romp6_0Ntheta(tid)%val
    flux              => Romp1_0Nthetap1(tid)%val
      
    thetamin = f%hhspline2d_rtheta(tid)%ygrid(0)
    thetamax = f%hhspline2d_rtheta(tid)%ygrid(Ntheta)

    Romp_maxalp_th(tid) = 0._RKIND
    alpha               = 0._RKIND

!$OMP DO SCHEDULE(static)
    do iphi = kstart,kend
      do ivpar = lstart,lend
        vparl = geom%vparg(ivpar)
!baoter (ask JPB)
!VG!        do ir = f%rstart_modif, f%rend_modif !0,Nr
        do ir = 0,Nr
!eaoter
!baoter (utiliser le jacobian spatial)
!VG!           !*** Jacobian factor ***
!VG!          f4d_transp(ir,:,iphi,ivpar)= &
!VG!            f4d_transp(ir,:,iphi,ivpar)*geom%rg(ir)
!eaoter
          !-> compute vExB.gradtheta
          call comp_vExBgradtheta_edges(geom, &
            init_magnet,init_curr, &
            Phiedges_rtheta,Phiedges_thetaphi, &
            ir,ir,0,Ntheta,iphi,iphi,ivpar,ivpar, &
            vExB_gradtheta_1D)
          !-> compute vD.gradtheta
          call compute_vDgradtheta(geom, &
            init_magnet,init_curr,ir,ir, &
            0,Ntheta,iphi,iphi,ivpar,ivpar,.true., &
            vD_gradtheta_1D)
          do itheta = 0,Ntheta-1
            ileft = itheta-1
            if (ileft<0) ileft = Ntheta-1
            iright = itheta
            !-> compute bstar.gradtheta at the edge
#ifdef NO_PRECOMP_BSTAR 
            call compute_bstar_contravariant(geom, &
              init_magnet,init_curr,ir,ileft,ivpar, &
              Sbstar_gradx2=bstar_gradtheta_left)
            call compute_bstar_contravariant(geom, &
              init_magnet,init_curr,ir,iright,ivpar, &
              Sbstar_gradx2=bstar_gradtheta_right)
#else
            call precomp_transp_bstar_gradtheta(ir, &
              ileft,ivpar,bstar_gradtheta_left)
            call precomp_transp_bstar_gradtheta(ir, &
              iright,ivpar,bstar_gradtheta_right)
#endif
            bstar_gradtheta_edge = 0.5_RKIND * &
              (bstar_gradtheta_left+bstar_gradtheta_right)
            !-> compute vD_gradtheta at the edge
            vD_gradtheta_left  = vD_gradtheta_1D(ileft)
            vD_gradtheta_right = vD_gradtheta_1D(iright)
            vD_gradtheta_edge  = 0.5_RKIND * &
              (vD_gradtheta_left+vD_gradtheta_right)
            !-> computation of alpha
            alpha(itheta)      = &
              vparl*bstar_gradtheta_edge + &
              vExB_gradtheta_1D(itheta) + &
              vD_gradtheta_edge  
            alpha(itheta)      = &
              vparl*bstar_gradtheta_edge + &
              vExB_gradtheta_1D(itheta) + &
              vD_gradtheta_edge  
          end do
          !*** periodicity in theta direction ***
          alpha(Ntheta)       = alpha(0)
          alpha               = dt*alpha

          Romp_maxalp_th(tid) = max(Romp_maxalp_th(tid), &
            maxval(abs(alpha(0:Ntheta))))

          deptheta            = 0._RKIND
          !*** splitting in theta direction ***
          do itheta = 0,Ntheta-1
            deptheta    = alpha(itheta)
            thetan      = &
              f%hhspline2d_rtheta(tid)%ygrid(itheta) - deptheta
            thetastar   = 0.5_RKIND * &
              (f%hhspline2d_rtheta(tid)%ygrid(itheta)+thetan )
            out         = 0._RKIND
            if (thetastar.gt.thetamax) then
              out       = abs(floor((thetastar-thetamin)/ &
                geom%Ltheta))
              thetastar = thetastar-out*geom%Ltheta  
            elseif (thetastar.lt.thetamin) then
              out       = -abs(floor((thetastar-thetamin)/ &
                geom%Ltheta))
              thetastar = thetastar-out*geom%Ltheta  
            endif

            err   = 1._RKIND
            count = 0
            do while (err>errmax.and.count<countmax)
              if ( (thetastar>=thetamin) &
                .and. (thetastar <= thetamax)) then
                !*** array position location ***
                call locate(thetastar, &
                  f%hhspline2d_rtheta(tid)%ygrid(0:Ntheta), &
                  geom%Ntheta,geom%dtheta,subp,ithetapos)
              else
                print*, "error locate theta", thetastar
                deptheta  = alpha(itheta)
                thetan    = &
                  f%hhspline2d_rtheta(tid)%ygrid(itheta) - deptheta
                thetastar = thetan
              endif
              ! relative position in the cell 
              xtheta     = (thetastar - &
                f%hhspline2d_rtheta(tid)%ygrid(ithetapos) ) &
                /geom%dtheta
              ileft      = ithetapos
              iright     = ithetapos+1
              deptheta   = alpha(ileft) * &
                (1._RKIND-xtheta) + alpha(iright)*xtheta 
              err        = thetastar
              thetan     = &
                f%hhspline2d_rtheta(tid)%ygrid(itheta) - deptheta
              thetastar  = 0.5_RKIND * &
                (f%hhspline2d_rtheta(tid)%ygrid(itheta)+thetan)
              out        = 0._RKIND
              if (thetastar.gt.thetamax) then
                out       = abs(floor((thetastar-thetamin)/ &
                  geom%Ltheta))
                thetastar = thetastar - out*geom%Ltheta  
              elseif (thetastar.lt.thetamin) then
                out       = -abs(floor((thetastar-thetamin)/ &
                  geom%Ltheta))
                thetastar = thetastar - out*geom%Ltheta  
              endif

              err   = abs(thetastar-err)/geom%dtheta
              count = count+1
            enddo ! while
            udt(itheta) = deptheta
          enddo
          udt(Ntheta) = udt(0)
          !--> PPM2 algorithm
          do itheta = 0,Ntheta
            iddd = itheta+2
            idd  = itheta+1
            id   = itheta  
            ig   = itheta-1
            igg  = itheta-2
            iggg = itheta-3
            if (iddd>Ntheta-1) iddd = iddd-Ntheta
            if (idd >Ntheta-1) idd  = idd-Ntheta
            if (id  >Ntheta-1) id   = id -Ntheta
            if (ig  <0       ) ig   = ig +Ntheta
            if (igg <0       ) igg  = igg+Ntheta
            if (iggg<0       ) iggg = iggg+Ntheta

            flux(itheta) = &
              ( 37._RKIND*(f4d_transp(ir,ig,iphi,ivpar) + &
              f4d_transp(ir,id,iphi,ivpar)) - &
              8._RKIND*(f4d_transp(ir,igg,iphi,ivpar) + &
              f4d_transp(ir,idd,iphi,ivpar)) + &
              f4d_transp(ir,iggg,iphi,ivpar) + &
              f4d_transp(ir,iddd,iphi,ivpar) )/60._RKIND

            !--> second derivative of f at the nodes
            if ( (f4d_transp(ir,ig,iphi,ivpar)-flux(itheta)) * &
              (f4d_transp(ir,id,iphi,ivpar)-flux(itheta)) &
              > 0._RKIND) then
              Ndd = 3._RKIND * &
                ( f4d_transp(ir,ig,iphi,ivpar) - &
                2._RKIND*flux(itheta) + &
                f4d_transp(ir,id,iphi,ivpar) )/(geom%dtheta**2)
              Ndd_l = ( f4d_transp(ir,igg,iphi,ivpar) - &
                2._RKIND*f4d_transp(ir,ig,iphi,ivpar) + &
                f4d_transp(ir,id,iphi,ivpar) )/(geom%dtheta**2)
              Ndd_r = ( f4d_transp(ir,ig,iphi,ivpar) - &
                2._RKIND*f4d_transp(ir,id,iphi,ivpar) + &
                f4d_transp(ir,idd,iphi,ivpar))/(geom%dtheta**2)

              if ( Ndd.ge.0._RKIND .and. Ndd_l.ge.0._RKIND .and. &
                Ndd_r.ge.0._RKIND) then
                Ndd_lim = min( Ndd,Kppm2*Ndd_l, &
                  Kppm2*Ndd_r )
              elseif ( Ndd.lt.0._RKIND .and. &
                Ndd_l.lt.0._RKIND .and. Ndd_r.lt.0._RKIND) then
                Ndd_lim = max( Ndd,Kppm2*Ndd_l, &
                  Kppm2*Ndd_r )
              else 
                Ndd_lim = 0._RKIND
              endif
              flux(itheta) = 0.5_RKIND * &
                ( f4d_transp(ir,ig,iphi,ivpar) + &
                f4d_transp(ir,id,iphi,ivpar)) - &
                (geom%dtheta**2)*Ndd_lim/3._RKIND
            endif
          enddo

          do itheta = 0,Ntheta-1
            al(itheta)    = flux(itheta)
            ar(itheta)    = flux(itheta+1)
            delta(itheta) = ar(itheta)-al(itheta)
            a6(itheta)    = 6._RKIND * &
              (f4d_transp(ir,itheta,iphi,ivpar) - &
              (al(itheta)+ar(itheta))/2._RKIND)

            !--> Local min or max
            if ( (ar(itheta)-f4d_transp(ir,itheta,iphi,ivpar)) * &
              (f4d_transp(ir,itheta,iphi,ivpar)-al(itheta)) &
              > 0._RKIND ) then
              !----> Non mono L
              if ( delta(itheta)*a6(itheta) > delta(itheta)**2 ) &
                then
                al(itheta) = 3._RKIND * &
                  f4d_transp(ir,itheta,iphi,ivpar) - &
                  2._RKIND*ar(itheta)
              endif
              !----> Non mono R
              if ( -delta(itheta)**2 > delta(itheta)*a6(itheta) ) &
                then
                ar(itheta) = 3._RKIND * &
                  f4d_transp(ir,itheta,iphi,ivpar) - &
                  2._RKIND*al(itheta)
              endif
            else
              !----> Limitation
              iddd = itheta+2
              idd  = itheta+1
              id   = itheta  
              ig   = itheta-1
              igg  = itheta-2

              if (iddd>Ntheta-1) iddd = iddd-Ntheta
              if (idd >Ntheta-1) idd  = idd-Ntheta
              if (id  >Ntheta-1) id   = id -Ntheta
              if (ig  <0       ) ig   = ig +Ntheta
              if (igg <0       ) igg  = igg+Ntheta

              Mdd = -2._RKIND*a6(itheta)/(geom%dtheta**2)

              !---> second derivative of f at the cells
              Mdd_c = ( f4d_transp(ir,ig,iphi,ivpar) - &
                2._RKIND*f4d_transp(ir,id ,iphi,ivpar) + &
                f4d_transp(ir,idd ,iphi,ivpar)) / (geom%dtheta**2)
              Mdd_l = ( f4d_transp(ir,igg,iphi,ivpar) - & 
                2._RKIND*f4d_transp(ir,ig,iphi,ivpar) + &
                f4d_transp(ir,id  ,iphi,ivpar)) / (geom%dtheta**2)
              Mdd_r = ( f4d_transp(ir,id,iphi,ivpar) - &
                2._RKIND*f4d_transp(ir,idd,iphi,ivpar) + &
                f4d_transp(ir,iddd,iphi,ivpar)) / (geom%dtheta**2)

              if ( Mdd>=0._RKIND .and. Mdd_c>=0._RKIND .and. &
                Mdd_l>=0._RKIND .and. Mdd_r>=0._RKIND) then
                Mdd_lim = min( Mdd,Kppm2*Mdd_c, &
                  Kppm2*Mdd_l,Kppm2*Mdd_r )
              elseif ( Mdd<0._RKIND .and. Mdd_c<0._RKIND .and. &
                Mdd_l<0._RKIND .and. Mdd_r<0._RKIND) then
                Mdd_lim = max( Mdd,Kppm2*Mdd_c, &
                  Kppm2*Mdd_l,Kppm2*Mdd_r )
              else
                Mdd_lim = 0._RKIND
              endif
              if (Mdd/=0._RKIND) then
                al(itheta) = f4d_transp(ir,itheta,iphi,ivpar) + &
                  (al(itheta)-f4d_transp(ir,itheta,iphi,ivpar)) * &
                  Mdd_lim/Mdd
                ar(itheta) = f4d_transp(ir,itheta,iphi,ivpar) + &
                  (ar(itheta)-f4d_transp(ir,itheta,iphi,ivpar)) * &
                  Mdd_lim/Mdd
              else
                al(itheta) = f4d_transp(ir,itheta,iphi,ivpar)
                ar(itheta) = f4d_transp(ir,itheta,iphi,ivpar)
              endif
            endif
            delta(itheta) = ar(itheta)-al(itheta)
            a6(itheta)    = 6._RKIND * &
              (f4d_transp(ir,itheta,iphi,ivpar) - &
              (al(itheta)+ar(itheta))/2._RKIND)
            ! end Limitation  
          enddo

          do itheta = 1,Ntheta-1
            xx = udt(itheta)/geom%dtheta
            if (udt(itheta)>=0._RKIND) then
              flux(itheta) = ar(itheta-1) + &
                0.5_RKIND*xx*(a6(itheta-1) - &
                delta(itheta-1)) - &
                a6(itheta-1)*(xx**2/3._RKIND) 
            else
              flux(itheta) = al(itheta) - &
                0.5_RKIND*xx*(a6(itheta) + &
                delta(itheta)) - &
                a6(itheta)*(xx**2/3._RKIND) 
            endif
            flux(itheta) = udt(itheta)*flux(itheta)
          enddo

          xx = udt(0)/geom%dtheta  !! there is udt(Ntheta) = udt(0)
          if (udt(0).le.0._RKIND) then
            flux(0) = al(0) - 0.5_RKIND*xx*(a6(0) + delta(0)) - &
              a6(0)*(xx**2/3._RKIND)
            flux(Ntheta) =  flux(0)
          else   
            flux(Ntheta) = ar(Ntheta-1) + &
              0.5_RKIND*xx*(a6(Ntheta-1) - &
              delta(Ntheta-1)) - &
              a6(Ntheta-1)*(xx**2/3._RKIND)
            flux(0) = flux(Ntheta)
          endif
          flux(0)      = udt(0)*flux(0)
          flux(Ntheta) = udt(0)*flux(Ntheta)

          do itheta = 0,Ntheta-1
            f4d_transp(ir,itheta,iphi,ivpar) =  &
              f4d_transp(ir,itheta,iphi,ivpar) - &
              (1._RKIND/geom%dtheta)*(flux(itheta+1)-flux(itheta))
          enddo

!baoter (utiliser le jacobian spatial)
!          !*** Jacobian factor ***
!          f4d_transp(ir,:,iphi,ivpar)= &
!            f4d_transp(ir,:,iphi,ivpar)/geom%rg(ir)
!eaoter
        end do
      end do
    end do
!$OMP END DO
!$OMP BARRIER
!$OMP END PARALLEL

    maxalp_theta = maxval(Romp_maxalp_th(1:Nbthread))/(geom%dtheta*dt)

    call clck_time(eclock_advec1Dtheta)
    call clck_diff(bclock_advec1Dtheta,eclock_advec1Dtheta, &
      global_time_advec1Dtheta)

    call Write_dyn_balise_close('CPU_advec1d_PPM2_theta') !R4
!R3 call r3_info_end (r3_info_index_0) !R3
#else 
    ! HAS_ADV_PPM
    maxalp_theta = 0._RKIND
#endif 
    ! HAS_ADV_PPM
   end subroutine advec1D_PPM2_theta
end module advec1D_PPM2_module
