!**************************************************************
!  Copyright Euratom-CEA
!  Authors : 
!     Virginie Grandgirard (virginie.grandgirard@cea.fr)
!     Chantal Passeron (chantal.passeron@cea.fr)
!     Guillaume Latu (guillaume.latu@cea.fr)
!     Xavier Garbet (xavier.garbet@cea.fr)
!     Philippe Ghendrih (philippe.ghendrih@cea.fr)
!     Yanick Sarazin (yanick.sarazin@cea.fr)
!  
!  This code GYSELA (for GYrokinetic SEmi-LAgrangian) 
!  is a 5D gyrokinetic global full-f code for simulating 
!  the plasma turbulence in a tokamak.
!  
!  This software is governed by the CeCILL-B license 
!  under French law and abiding by the rules of distribution 
!  of free software.  You can  use, modify and redistribute 
!  the software under the terms of the CeCILL-B license as 
!  circulated by CEA, CNRS and INRIA at the following URL
!  "http://www.cecill.info". 
!**************************************************************

#include "config-F.inc"

!---------------------------------------------------------
! file : advec1D_PSM.F90
! date : 18/10/2009
!   1D advection in phi and vparallel directions used in
!   PSM Conservative Semi-Lagrangian method
! developed by 
!  Jean-Philippe Braeunig (braeunig@math.u-strasbg.fr)
!---------------------------------------------------------
module advec1D_PSM_module
  use globals, only : Nr, Ntheta, Nphi, Nvpar, Nbthread
  use wrap_alloc_module
  use integration_module
  use clock_module
  use geometry_class
  use init_magnetic_class
  use fdistribu5d_class
  use interpolation_module
  use velocities_edges_module

  implicit none

  integer, parameter, private :: countmax = 50
  real(RKIND)       , save    :: Klim     = 5._RKIND

  include "gyrokinetic/Bstar_inline.h"
  include "gyrokinetic/velocities_inline.h"
  !******************************
  contains
  !******************************
#include "gyrokinetic/Bstar_inline.F90"
#include "gyrokinetic/velocities_inline.F90"

  !******************************  
  !--------------------------------------------------------
  ! splitting in r parallel direction
  !  - Solving of : df/dt+(dr/dt)*df/dr=0
  !         dr/dt = (1/sqrt{hat_As})*vpar*bstar_gradr 
  !                 +  vExB_gradr + vD_gradr   
  !
  !  . f*(r,theta,f,vpar,mu) = fn(r-dr,theta,phi,vpar,mu)
  !--------------------------------------------------------
  subroutine advec1D_PSM_r(geom,dt,init_magnet,init_curr,f,maxalp_r)
    use MPIutils_module, only : f4d_transp
    use globals        , only : istart, iend, &
      jstart, jend, kstart, kend, lstart, lend, &
      errmax, limiter, pSpecies_id, hat_As
#ifdef _OPENMP
    use OMPutils_module, only : omp_get_thread_num
#endif
    use OMPutils_module, only : Romp_alpha, Romp_rhs, &
      Romp1_0Nr, Romp1_0Nrp1 
    use velocities_edges_module, only : Phiedges_rtheta, &
      Phiedges_rphi, comp_vExBgradr_edges
    use globals, only : transpose4D, conservative_scheme, &
         dom_r, dom_theta
!R3 #include "code_instrumentation/r3_info.h" !R3
    type(geometry)     , intent(in)    :: geom
    real(RKIND)        , intent(in)    :: dt
    type(init_magnetic), intent(in)    :: init_magnet
    type(init_current) , intent(in)    :: init_curr
    type(fdistribu5d)  , intent(inout) :: f
    real(RKIND)        , intent(out)   :: maxalp_r
#ifdef HAS_ADV_PSM
    integer     :: ir, itheta, iphi, ivpar
    real(RKIND) :: rmin, rmax, rstar
    real(RKIND) :: depr, xr, rn
    real(RKIND) :: sum, err
    real(RKIND) :: bstar_gradr_left, bstar_gradr_right
    real(RKIND) :: bstar_gradr_edge
    real(RKIND) :: vD_gradr_left, vD_gradr_right
    real(RKIND) :: vD_gradr_edge
    real(RKIND) :: vparl
    real(RKIND) :: finterpol, Fcentre
    real(RKIND) :: inv_sqrtAs
    integer     :: tid, ileft, iright, irpos
    integer     :: k, count
    character(LEN=50) :: subp

    ! array for the edges of the cell (0:geom%Nr+1)
    real(RKIND), dimension(:), pointer :: alpha  
    ! array for the edges of the cell (-1:geom%Nr+2)
    real(RKIND), dimension(:), pointer :: rhsr   

    !-> radial components of ExB drift velocity
    real(RKIND), dimension(:), pointer :: vExB_gradr_1D
    !-> radial components of curvature drift velocity
    real(RKIND), dimension(:), pointer :: vD_gradr_1D

    real(RKIND) :: dp, dpc, lim, flux_upw
    real(RKIND), dimension(0:Nr+1)     :: flux
    real(RKIND), dimension(1:Nbthread) :: Romp_maxalp_r

    integer :: min_i1, min_i2, max_i1, max_i2

    subp = " advec1D_PSM_r "//char(0)
!R3 call r3_info_begin (r3_info_index_0, 'CPU_advec1d_PSM_r') !R3
    call Write_dyn_balise_open('CPU_advec1d_PSM_r') !R4

    call clck_time(bclock_advec1Dr)

    if (transpose4D) then
       if (conservative_scheme) then
          min_i1 = -1
          max_i1 = (f%n1)+2
          min_i2 = -1
          max_i2 = (f%n2)+1

          call write_param("min_i1", -1)
          call write_expr("max_i1" , "(fthis%n1)+2")
          call write_param("min_i2", -1)
          call write_expr("max_i2" , "(fthis%n2)+1")
       else
          min_i1 = -2
          max_i1 = (f%n1+2)+1
          min_i2 = -2
          max_i2 = (f%n2+1)+1

          call write_param("min_i1", -2)
          call write_expr("max_i1" , "(fthis%n1+2)+1")
          call write_param("min_i2", -2)
          call write_expr("max_i2" , "(fthis%n2+1)+1")
       endif
    else
       if (conservative_scheme) then
          min_i1 = -1
          max_i1 = (dom_r+1)+2
          min_i2 = -1
          max_i2 = (dom_theta+1)+1

          call write_param("min_i1", -1)
          call write_expr("max_i1" , "(dom_r+1)+2")
          call write_param("min_i2", -1)
          call write_expr("max_i2" , "(dom_theta+1)+1")
       else
          min_i1 = -2
          max_i1 = (dom_r+1)+1
          min_i2 = -2
          max_i2 = (dom_theta+1)+1

          call write_param("min_i1", -2)
          call write_expr("max_i1" , "(dom_r+1)+1")
          call write_param("min_i2", -2)
          call write_expr("max_i2" , "(dom_theta+1)+1")
       endif
    endif

    do tid = 1, Nbthread
       call FTR_ALLOC_2(f%hhspline2d_rtheta(tid)%coef,min_i1, max_i1, min_i2, max_i2)
       call FTR_ALLOC_2(f%hhspline2d_rtheta(tid)%bcoef,min_i1, max_i1, min_i2, max_i2)
       call FTR_ALLOC_2(f%hhspline2d_rtheta(tid)%aux,min_i1, max_i1, min_i2, max_i2)
       call FTR_ALLOC_2(f%hhspline2d_rtheta(tid)%auy,min_i1, max_i1, min_i2, max_i2)
    enddo
    
    !*** computation of 1/sqrt{hat_As} ***
    inv_sqrtAs = 1._RKIND/sqrt(hat_As(pSpecies_id))

#ifdef _OPENMP
!$OMP PARALLEL private(tid,ir,itheta,iphi,ivpar, &
!$OMP rmin,rmax,alpha,rhsr,sum,depr,vparl, &
!$OMP bstar_gradr_left,bstar_gradr_right, &
!$OMP vD_gradr_left,vD_gradr_right, &
!$OMP bstar_gradr_edge,vD_gradr_edge, &
!$OMP vExB_gradr_1D,vD_gradr_1D, &
!$OMP ileft,iright,irpos,rn,rstar,err,count,xr, &
!$OMP finterpol,Fcentre) default(shared)
!$OMP BARRIER
    tid = 1+omp_get_thread_num()
#else
    tid = 1
#endif

    rmin = f%hhspline2d_rtheta(tid)%xgrid(0)
    rmax = f%hhspline2d_rtheta(tid)%xgrid(Nr+1)

    alpha         => Romp_alpha(tid)%val
    rhsr          => Romp_rhs(tid)%val
    vExB_gradr_1D => Romp1_0Nrp1(tid)%val
    vD_gradr_1D   => Romp1_0Nr(tid)%val    

    Romp_maxalp_r(tid) = 0._RKIND
    alpha              = 0._RKIND

!$OMP DO SCHEDULE(static)
    do iphi = kstart,kend
      do ivpar = lstart,lend
        vparl = geom%vparg(ivpar)
        do itheta = 0,Ntheta-1
!baoter (ATTENTION prendre en compte correctement le jacobien)
          !*** Jacobian factor ***
          f4d_transp(:,itheta,iphi,ivpar)= &
            f4d_transp(:,itheta,iphi,ivpar)*geom%rg(:)
!eaoter
          !*** Values of the primitives ***
          rhsr(-1:Nr+2) = 0._RKIND
          do ir = 1,Nr+1
            rhsr(ir) = rhsr(ir-1) &
              + f4d_transp(ir-1,itheta,iphi,ivpar) * geom%dr
          end do
          sum = rhsr(Nr+1)-rhsr(0) 
          !************************************************ 
          !*** Natural conditions :                     ***
          !***  Second derivatives G and D = 0          ***
          !***     rhsr(-1)   = 0._RKIND                ***
          !***     rhsr(Nr+2) = 0._RKIND                *** 
          !************************************************
          call hermite2_1D_r(f%hhspline2d_rtheta(tid), &
            rhsr(-1:Nr+2),-1,Nr+2)
          !-> compute vExB.gradr
          ! WARNING: transposition of Bstar data is not taken 
          ! into account. You must set transp_Bstar to .false.
          call comp_vExBgradr_edges(geom, &
            init_magnet,init_curr, &
            Phiedges_rtheta,Phiedges_rphi, &
            0,Nr+1,itheta,itheta, &
            iphi,iphi,ivpar,ivpar,vExB_gradr_1D)
          !-> compute vD.gradr 
          call compute_vDgradr(geom, &
            init_magnet,init_curr, &
            0,Nr,itheta,itheta,iphi,iphi, &
            ivpar,ivpar,.true.,vD_gradr_1D)
          do ir = 0,Nr+1
            ileft  = max(ir-1,0)
            iright = min(ir,Nr) 
            !-> compute bstar.gradr at the edge
#ifdef NO_PRECOMP_BSTAR 
            call compute_bstar_contravariant(geom, &
              init_magnet,init_curr,ileft,itheta,ivpar, &
              Sbstar_gradx1=bstar_gradr_left)
            call compute_bstar_contravariant(geom, &
              init_magnet,init_curr,iright,itheta,ivpar, &
              Sbstar_gradx1=bstar_gradr_right)
#else
            call precomp_transp_bstar_gradr(ileft,itheta, &
                 ivpar,bstar_gradr_left)
            call precomp_transp_bstar_gradr(iright,itheta, &
                 ivpar,bstar_gradr_right)
#endif
            bstar_gradr_edge = 0.5_RKIND * &
              (bstar_gradr_left+bstar_gradr_right)
            !-> compute vD_gradr at the edge
            vD_gradr_left  = vD_gradr_1D(ileft)
            vD_gradr_right = vD_gradr_1D(iright)
            vD_gradr_edge  = 0.5_RKIND * &
              (vD_gradr_left+vD_gradr_right)
            !-> computation of alpha
            alpha(ir)      = &
              inv_sqrtAs*vparl*bstar_gradr_edge + &
              vExB_gradr_1D(ir) + vD_gradr_edge  
          end do
          alpha              = dt*alpha
          Romp_maxalp_r(tid) = max(Romp_maxalp_r(tid), &
            maxval(abs(alpha(0:Nr+1))))

          depr = 0._RKIND 
          do ir = 0,Nr+1
            ileft  = max(ir-1,0)
            iright = min(ir,Nr)
            depr   = alpha(ir)
            rn     = &
              f%hhspline2d_rtheta(tid)%xgrid(ir) - depr
            rstar  = 0.5_RKIND * &
              (f%hhspline2d_rtheta(tid)%xgrid(ir)+rn)
            err    = 1._RKIND
            count  = 0
            do while (err>errmax.and.count<countmax)
              !*** Inside the domain ***
              if (rstar.gt.rmin.and.rstar.lt.rmax) then
                !-> array position location
                call locate(rstar, &
                  f%hhspline2d_rtheta(tid)%xgrid(:), &
                  geom%Nr+1,geom%dr,subp,irpos)
                !-> relative position in the cell 
                xr   = (rstar - &
                  f%hhspline2d_rtheta(tid)%xgrid(irpos)) / &
                  geom%dr
                depr = alpha(irpos)*(1._RKIND-xr) + &
                  alpha(irpos+1)*xr 
                !-> Out of the domain by the left
              elseif (rstar.le.rmin) then
                depr = alpha(0)
                !-> Out of the domain by the right
              elseif (rstar.ge.rmax) then
                depr = alpha(Nr+1)
              end if
              err   = rstar
              rn    = f%hhspline2d_rtheta(tid)%xgrid(ir)-depr
              rstar = 0.5_RKIND * &
                (f%hhspline2d_rtheta(tid)%xgrid(ir)+rn)
              err   = abs(rstar-err)/geom%dr
              count = count+1
            enddo !while
            !*** interpolation of the distribution function ***
            if (rn.gt.rmin.and.rn.lt.rmax) then
              !-> inside the domain
              call interpol1d_r(geom%Nr+1,geom%dr, &
                f%hhspline2d_rtheta(tid)%bcoef, &
                f%hhspline2d_rtheta(tid)%xgrid,rn,finterpol)
              !-> Outside the domain by the left
              !->   characteristic foot exterior of the domain
              !->   primitive derivative at the 
              !->   boundary = function at the boundary
            elseif (rn.le.rmin) then
              finterpol = f4d_transp(0,itheta,iphi,ivpar) 
              finterpol = rhsr(0) - finterpol * (rmin-rn)
            elseif (rn.ge.rmax) then
              !-> Outside the domain by the right
              !->   characteristic foot exterior of the domain
              finterpol = f4d_transp(Nr,itheta,iphi,ivpar) 
              finterpol = rhsr(Nr+1) - finterpol * (rmax-rn)
            end if
            !*** Entropic limiter ***
            if (limiter == 1) then
              Fcentre = rhsr(ir)-depr*0.5_RKIND * &
                ( f4d_transp(iright,itheta,iphi,ivpar)  + &
                f4d_transp(ileft ,itheta,iphi,ivpar)  )
              if ( (finterpol-Fcentre) * &
                ( f4d_transp(iright,itheta,iphi,ivpar)  - &
                f4d_transp(ileft ,itheta,iphi,ivpar) ) &
                < 0._RKIND ) then
                finterpol = Fcentre
              end if
            end if

            ! *** SLS limiter ***
            if (limiter == 2) then
              flux(ir) = rhsr(ir)-finterpol
              dpc      = f4d_transp(iright,itheta,iphi,ivpar) - &
                f4d_transp(ileft,itheta,iphi,ivpar)
              if (abs(dpc)>1.e-10_RKIND) then
!baoter (ask JPB)
!VG!                dpc = sign(1.e-10_RKIND,dpc)
!eaoter
                if (alpha(ir) > 0._RKIND ) then
                  dp       = f4d_transp(ileft,itheta,iphi,ivpar) - &
                    f4d_transp(max(ileft-1,0),itheta,iphi,ivpar)
                  flux_upw = alpha(ir) * &
                    f4d_transp(ileft,itheta,iphi,ivpar)
                else
                  dp       = f4d_transp(min(iright+1,Nr), &
                    itheta,iphi,ivpar) - &
                    f4d_transp(iright,itheta,iphi,ivpar)
                  flux_upw = alpha(ir) * &
                    f4d_transp(iright,itheta,iphi,ivpar)
                end if
                lim      = max(0._RKIND, &
                  min(Klim*abs(dp/dpc),1._RKIND))
                flux(ir) = lim*flux(ir)+(1._RKIND-lim)*flux_upw
              end if
              ! Prim
              rhsr(ir) = rhsr(ir)-flux(ir)
            else
              rhsr(ir) = finterpol
            end if !sls
          end do ! 0,Nr+1
!baoter (ask JPB)
!VG!          do ir = f%rstart_modif,f%rend_modif !0,Nr
          do ir = 0,Nr
!eaoter
            f4d_transp(ir,itheta,iphi,ivpar) =  &
              (rhsr(ir+1) - rhsr(ir))/geom%dr            
          enddo
!baoter (ATTENTION prendre en compte correctement le jacobien)
          !*** Jacobian factor ***
          f4d_transp(:,itheta,iphi,ivpar) = &
            f4d_transp(:,itheta,iphi,ivpar)/geom%rg(:)
!eaoter
        end do
      end do
    end do
!$OMP END DO
!$OMP BARRIER
!$OMP END PARALLEL

    maxalp_r = maxval(Romp_maxalp_r(1:Nbthread))/(geom%dr*dt)

    do tid = 1, Nbthread
       call FTR_DEALLOC(f%hhspline2d_rtheta(tid)%coef)
       call FTR_DEALLOC(f%hhspline2d_rtheta(tid)%bcoef)
       call FTR_DEALLOC(f%hhspline2d_rtheta(tid)%aux)
       call FTR_DEALLOC(f%hhspline2d_rtheta(tid)%auy)
    enddo
    
    call clck_time(eclock_advec1Dr)
    call clck_diff(bclock_advec1Dr,eclock_advec1Dr, &
      global_time_advec1Dr)

    call Write_dyn_balise_close('CPU_advec1d_PSM_r') !R4
!R3 call r3_info_end (r3_info_index_0) !R3
#else 
    ! HAS_ADV_PSM
    maxalp_r = 0._RKIND
#endif 
    ! HAS_ADV_PSM
  end subroutine advec1D_PSM_r


  !------------------------------------------------------------
  ! splitting in theta direction 
  !  - Solving of : df/dt+dtheta/dt*df/dtheta=0
  !      dtheta/dt = (1/sqrt{hat_As})*vparl*b_gradtheta_tmp +
  !                  vExB_gradtheta_tmp + vD_gradtheta_tmp
  !------------------------------------------------------------
  subroutine advec1D_PSM_theta(geom,dt,init_magnet,init_curr,f,maxalp_theta)
    use MPIutils_module, only : f4d_transp
    use globals        , only : istart, iend, &
      jstart, jend, kstart, kend, lstart, lend, &
      errmax, limiter, pSpecies_id, hat_As
#ifdef _OPENMP
    use OMPutils_module, only : omp_get_thread_num
#endif
    use OMPutils_module, only : Romp_alpha, Romp_rhs, &
      Romp_vect, Romp1_0Ntheta, Romp2_0Ntheta 
    use velocities_edges_module, only : Phiedges_rtheta, &
      Phiedges_thetaphi, comp_vExBgradtheta_edges
    use globals, only : transpose4D, conservative_scheme, &
         dom_r, dom_theta
!R3 #include "code_instrumentation/r3_info.h" !R3
    type(geometry)     , intent(in)    :: geom
    real(RKIND)        , intent(in)    :: dt
    type(init_magnetic), intent(in)    :: init_magnet
    type(init_current) , intent(in)    :: init_curr
    type(fdistribu5d)  , intent(inout) :: f
    real(RKIND)        , intent(out)   :: maxalp_theta

#ifdef HAS_ADV_PSM
    integer     :: ir, itheta, iphi, ivpar
    real(RKIND) :: thetamin, thetamax, thetastar
    real(RKIND) :: sum, thetan, err
    real(RKIND) :: vparl
    real(RKIND) :: finterpol, Fcentre
    real(RKIND) :: out, xtheta, deptheta
    integer     :: tid, ileft, iright, ithetapos
    integer     :: k, count, iup
    real(RKIND) :: der_bc1, der_bc2
    real(RKIND) :: bstar_gradtheta_left, bstar_gradtheta_right
    real(RKIND) :: bstar_gradtheta_edge
    real(RKIND) :: vD_gradtheta_left, vD_gradtheta_right
    real(RKIND) :: vD_gradtheta_edge
    real(RKIND) :: inv_sqrtAs
    character(LEN=50) :: subp

    !-> array of size (0:geom%Ntheta)
    real(RKIND), dimension(:), pointer :: alpha
    !-> array of size (-1:geom%Ntheta+1)
    real(RKIND), dimension(:), pointer :: rhstheta 
    !-> array of size (-10:10)
    real(RKIND), dimension(:), pointer :: vect

    !-> poloidal components of ExB drift velocity
    real(RKIND), dimension(:), pointer :: vExB_gradtheta_1D
    !-> poloidal components of curvature drift velocity
    real(RKIND), dimension(:), pointer :: vD_gradtheta_1D

    real(RKIND), dimension(0:Ntheta+1) :: flux
    real(RKIND) :: dp,dpc,lim,flux_upw
    real(RKIND), dimension(1:Nbthread) :: Romp_maxalp_th

    integer :: min_i1, min_i2, max_i1, max_i2

!R3 call r3_info_begin(r3_info_index_0,'CPU_advec1d_PSM_theta') !R3
    call Write_dyn_balise_open('CPU_advec1d_PSM_theta') !R4
    call clck_time(bclock_advec1Dtheta)
    subp = " advec1D_PSM_theta "//char(0)

    if (transpose4D) then
       if (conservative_scheme) then
          min_i1 = -1
          max_i1 = (f%n1)+2
          min_i2 = -1
          max_i2 = (f%n2)+1

          call write_param("min_i1", -1)
          call write_expr("max_i1" , "(fthis%n1)+2")
          call write_param("min_i2", -1)
          call write_expr("max_i2" , "(fthis%n2)+1")
       else
          min_i1 = -2
          max_i1 = (f%n1+2)+1
          min_i2 = -2
          max_i2 = (f%n2+1)+1

          call write_param("min_i1", -2)
          call write_expr("max_i1" , "(fthis%n1+2)+1")
          call write_param("min_i2", -2)
          call write_expr("max_i2" , "(fthis%n2+1)+1")
       endif
    else
       if (conservative_scheme) then
          min_i1 = -1
          max_i1 = (dom_r+1)+2
          min_i2 = -1
          max_i2 = (dom_theta+1)+1

          call write_param("min_i1", -1)
          call write_expr("max_i1" , "(dom_r+1)+2")
          call write_param("min_i2", -1)
          call write_expr("max_i2" , "(dom_theta+1)+1")
       else
          min_i1 = -2
          max_i1 = (dom_r+1)+1
          min_i2 = -2
          max_i2 = (dom_theta+1)+1

          call write_param("min_i1", -2)
          call write_expr("max_i1" , "(dom_r+1)+1")
          call write_param("min_i2", -2)
          call write_expr("max_i2" , "(dom_theta+1)+1")
       endif
    endif

    do tid = 1, Nbthread
       call FTR_ALLOC_2(f%hhspline2d_rtheta(tid)%coef,min_i1, max_i1, min_i2, max_i2)
       call FTR_ALLOC_2(f%hhspline2d_rtheta(tid)%bcoef,min_i1, max_i1, min_i2, max_i2)
       call FTR_ALLOC_2(f%hhspline2d_rtheta(tid)%aux,min_i1, max_i1, min_i2, max_i2)
       call FTR_ALLOC_2(f%hhspline2d_rtheta(tid)%auy,min_i1, max_i1, min_i2, max_i2)
    enddo

    !*** computation of 1/sqrt{hat_As} ***
    inv_sqrtAs = 1._RKIND/sqrt(hat_As(pSpecies_id))

#ifdef _OPENMP
!$OMP PARALLEL private(tid,ir,itheta,iphi,ivpar, &
!$OMP  alpha,rhstheta,vect,thetamin,thetamax, &
!$OMP  sum,der_bc1,deptheta,ileft,iright,ithetapos, &
!$OMP  vD_gradtheta_left,vD_gradtheta_right, &
!$OMP  bstar_gradtheta_left,bstar_gradtheta_right, &
!$OMP  bstar_gradtheta_edge,vD_gradtheta_edge, &
!$OMP  vExB_gradtheta_1D,vD_gradtheta_1D, &
!$OMP  vparl,thetan,thetastar,out,err,count,xtheta, &
!$OMP  finterpol,Fcentre) default(shared)
!$OMP BARRIER
    tid = 1+omp_get_thread_num()
#else
    tid = 1
#endif

    alpha             => Romp_alpha(tid)%val
    rhstheta          => Romp_rhs(tid)%val
    vect              => Romp_vect(tid)%val
    vExB_gradtheta_1D => Romp1_0Ntheta(tid)%val
    vD_gradtheta_1D   => Romp2_0Ntheta(tid)%val
      
    thetamin = f%hhspline2d_rtheta(tid)%ygrid(0)
    thetamax = f%hhspline2d_rtheta(tid)%ygrid(Ntheta)

    Romp_maxalp_th(tid) = 0._RKIND
    alpha               = 0._RKIND

!$OMP DO SCHEDULE(static)
    do iphi = kstart,kend
      do ivpar = lstart,lend
        vparl = geom%vparg(ivpar)
!baoter (ask JPB)
!VG!        do ir = f%rstart_modif, f%rend_modif !0,Nr
        do ir = 0,Nr
!eaoter
           !*** Jacobian factor ***
!baoter (utiliser le jacobian spatial)
!VG!          f4d_transp(ir,:,iphi,ivpar)= &
!VG!            f4d_transp(ir,:,iphi,ivpar)*geom%rg(ir)
!eaoter
          !*** Values of the primitives ***
          rhstheta(0) = 0._RKIND
          do itheta = 1,Ntheta
            rhstheta(itheta) = rhstheta(itheta-1) + &
              f4d_transp(ir,itheta-1,iphi,ivpar)*geom%dtheta
          end do
          sum = rhstheta(Ntheta) 
          !*** Derivatives of G and D by local splines
          vect(-10:-1) = rhstheta(Ntheta-10:Ntheta-1)-sum
          vect(1:10)   = rhstheta(1:10)

          der_bc1            = derive10(vect,geom%dtheta)    
          rhstheta(-1)       = der_bc1
          rhstheta(Ntheta+1) = rhstheta(-1)

          !*** cubic spline computation ***
          call hermite1D_theta(f%hhspline2d_rtheta(tid),rhstheta)

          ! WARNING: transposition of Bstar data is not taken 
          ! into account. You must set transp_Bstar to .false.
          !-> compute vExB.gradtheta
          call comp_vExBgradtheta_edges(geom, &
            init_magnet,init_curr, &
            Phiedges_rtheta,Phiedges_thetaphi, &
            ir,ir,0,Ntheta,iphi,iphi,ivpar,ivpar, &
            vExB_gradtheta_1D)
          !-> compute vD.gradtheta
          call compute_vDgradtheta(geom, &
            init_magnet,init_curr, &
            ir,ir,0,Ntheta,iphi,iphi, &
            ivpar,ivpar,.true.,vD_gradtheta_1D)
          do itheta = 0,Ntheta-1
            ileft = itheta-1
            if (ileft<0) ileft = Ntheta-1
            iright = itheta
            !-> compute bstar.gradtheta at the edge
#ifdef NO_PRECOMP_BSTAR 
            call compute_bstar_contravariant(geom, &
              init_magnet,init_curr,ir,ileft,ivpar, &
              Sbstar_gradx2=bstar_gradtheta_left)
            call compute_bstar_contravariant(geom, &
              init_magnet,init_curr,ir,iright,ivpar, &
              Sbstar_gradx2=bstar_gradtheta_right)
#else
            call precomp_transp_bstar_gradtheta(ir, &
              ileft,ivpar,bstar_gradtheta_left)
            call precomp_transp_bstar_gradtheta(ir, &
              iright,ivpar,bstar_gradtheta_right)
#endif
            bstar_gradtheta_edge = 0.5_RKIND * &
              (bstar_gradtheta_left+bstar_gradtheta_right)
            !-> compute vD_gradtheta at the edge
            vD_gradtheta_left  = vD_gradtheta_1D(ileft)
            vD_gradtheta_right = vD_gradtheta_1D(iright)
            vD_gradtheta_edge  = 0.5_RKIND * &
              (vD_gradtheta_left+vD_gradtheta_right)
            !-> computation of alpha
            alpha(itheta)      = &
              vparl*bstar_gradtheta_edge + &
              vExB_gradtheta_1D(itheta) + &
              vD_gradtheta_edge  
          end do
          !*** periodicity in theta direction ***
          alpha(Ntheta)       = alpha(0)
          alpha               = dt*alpha
          Romp_maxalp_th(tid) = max(Romp_maxalp_th(tid), &
            maxval(abs(alpha(0:Ntheta))))
          deptheta            = 0._RKIND
          !*** splitting in theta direction ***
          do itheta = 0,Ntheta-1
            deptheta  = alpha(itheta)
            thetan    = &
              f%hhspline2d_rtheta(tid)%ygrid(itheta) - deptheta
            thetastar = 0.5_RKIND * &
              (f%hhspline2d_rtheta(tid)%ygrid(itheta)+thetan )
            out = 0._RKIND
            if (thetastar.gt.thetamax) then
              out       = abs(floor((thetastar-thetamin)/ &
                geom%Ltheta))
              thetastar = thetastar-out*geom%Ltheta  
            elseif (thetastar.lt.thetamin) then
              out       = -abs(floor((thetastar-thetamin)/ &
                geom%Ltheta))
              thetastar = thetastar-out*geom%Ltheta  
            end if

            err   = 1._RKIND
            count = 0
            do while (err>errmax.and.count<countmax)
              if ( (thetastar>=thetamin) &
                .and. (thetastar <= thetamax)) then
                !*** array position location ***
                call locate(thetastar, &
                  f%hhspline2d_rtheta(tid)%ygrid(0:Ntheta), &
                  geom%Ntheta,geom%dtheta,subp,ithetapos)
              else
                print*, "error locate theta", thetastar
                deptheta  = alpha(itheta)
                thetan    = &
                  f%hhspline2d_rtheta(tid)%ygrid(itheta) - deptheta
                thetastar = thetan
              end if
              ! relative position in the cell 
              xtheta    = (thetastar - &
                f%hhspline2d_rtheta(tid)%ygrid(ithetapos) ) &
                /geom%dtheta
              ileft     = ithetapos
              iright    = ithetapos+1
              deptheta  = alpha(ileft) * &
                (1._RKIND-xtheta) + alpha(iright)*xtheta 
              err       = thetastar
              thetan    = &
                f%hhspline2d_rtheta(tid)%ygrid(itheta) - deptheta
              thetastar = 0.5_RKIND * &
                (f%hhspline2d_rtheta(tid)%ygrid(itheta)+thetan)
              out       = 0._RKIND
              if (thetastar.gt.thetamax) then
                out = abs(floor((thetastar-thetamin)/ &
                  geom%Ltheta))
                thetastar = thetastar - out*geom%Ltheta  
              elseif (thetastar.lt.thetamin) then
                out       = -abs(floor((thetastar-thetamin)/ &
                  geom%Ltheta))
                thetastar = thetastar - out*geom%Ltheta  
              end if

              err   = abs(thetastar-err)/geom%dtheta
              count = count+1
            enddo ! while

            out = 0._RKIND
            if (thetan.gt.thetamax) then
              out    = abs(floor((thetan-thetamin)/geom%Ltheta))
              thetan = thetan - out*geom%Ltheta  
            elseif (thetan.lt.thetamin) then
              out    = -abs(floor((thetan-thetamin)/geom%Ltheta))
              thetan = thetan - out*geom%Ltheta  
            end if

            !*** f interpolation ***
            call interpol1d_theta(geom%Ntheta,geom%dtheta, &
              f%hhspline2d_rtheta(tid)%bcoef, &
              f%hhspline2d_rtheta(tid)%ygrid,thetan,finterpol)

            !**************************************************
            ! primitive interpolation, total mass "sum" 
            !  is added or substracted if particule
            !  goes out of the domain trajectory wise
            !**************************************************
            finterpol = finterpol + out*sum

            !*** Entropic limiter ***
            if (limiter==1) then
              iright = itheta
              ileft  = iright-1
              if (iright==0) ileft = Ntheta-1
              Fcentre = rhstheta(itheta) - &
                deptheta*0.5_RKIND * &
                ( f4d_transp(ir,iright,iphi,ivpar) + &
                f4d_transp(ir,ileft,iphi,ivpar) )
              if ( (finterpol-Fcentre) * &
                ( f4d_transp(ir,iright,iphi,ivpar) - &
                f4d_transp(ir,ileft,iphi,ivpar) ) < 0._RKIND ) then
                finterpol = Fcentre
              end if
            end if

            ! *** SLS limiter ***
            if (limiter==2) then
              flux(itheta) = rhstheta(itheta)-finterpol
              iright       = itheta
              ileft        = iright-1
              if (ileft<0) ileft = ileft+Ntheta
              dpc = f4d_transp(ir,iright,iphi,ivpar) - &
                f4d_transp(ir,ileft ,iphi,ivpar)
              if (abs(dpc)>1.e-10_RKIND) then
                if (alpha(itheta) > 0._RKIND ) then
                  iup = ileft-1
                  if (iup<0) iup = iup + Ntheta
                  dp       = f4d_transp(ir,ileft,iphi,ivpar) - &
                    f4d_transp(ir,iup,iphi,ivpar)
                  flux_upw = alpha(itheta) * &
                    f4d_transp(ir,ileft,iphi,ivpar)
                else
                  iup = iright+1
                  if (iup>Ntheta-1) iup = iup - Ntheta
                  dp       = f4d_transp(ir,iup,iphi,ivpar) - &
                    f4d_transp(ir,iright,iphi,ivpar)
                  flux_upw = alpha(itheta) * &
                    f4d_transp(ir,iright,iphi,ivpar)
                end if
                lim          = max(0._RKIND, &
                  min( Klim*abs(dp/dpc) ,1._RKIND))
                flux(itheta) = lim*flux(itheta) + &
                  (1._RKIND-lim)*flux_upw
              end if
              !end upwind
              rhstheta(itheta) = rhstheta(itheta)-flux(itheta)
            else
              rhstheta(itheta) = finterpol
            end if !sls

          end do !0,Ntheta-1

          rhstheta(Ntheta) = rhstheta(0)+sum
          do itheta = 0,Ntheta-1
            f4d_transp(ir,itheta,iphi,ivpar) =  &
              (rhstheta(itheta+1)-rhstheta(itheta)) &
              /geom%dtheta
          enddo
          !*** Jacobian factor ***
!baoter (utiliser le jacobian spatial)
!VG!          f4d_transp(ir,:,iphi,ivpar)= &
!VG!            f4d_transp(ir,:,iphi,ivpar)/geom%rg(ir)
!eaoter
        end do
      end do
    end do
!$OMP END DO
!$OMP BARRIER
!$OMP END PARALLEL

    maxalp_theta = maxval(Romp_maxalp_th(1:Nbthread)) / &
      (geom%dtheta*dt)

    do tid = 1, Nbthread
       call FTR_DEALLOC(f%hhspline2d_rtheta(tid)%coef)
       call FTR_DEALLOC(f%hhspline2d_rtheta(tid)%bcoef)
       call FTR_DEALLOC(f%hhspline2d_rtheta(tid)%aux)
       call FTR_DEALLOC(f%hhspline2d_rtheta(tid)%auy)
    enddo

    call clck_time(eclock_advec1Dtheta)
    call clck_diff(bclock_advec1Dtheta,eclock_advec1Dtheta, &
      global_time_advec1Dtheta)
      
    call Write_dyn_balise_close('CPU_advec1d_PSM_theta') !R4
!R3 call r3_info_end (r3_info_index_0) !R3
#else 
    ! HAS_ADV_PSM
    maxalp_theta = 0._RKIND
#endif 
    ! HAS_ADV_PSM
  end subroutine advec1D_PSM_theta


  !------------------------------------------------------------
  ! splitting in phi direction 
  !  - Solving of : df/dt+vpar/R*df/dphi=0
  !     . dphi = 1/sqrt{hat_As}*vpar b_gradphi 
  !              + vExB_gradphi + vD_gradphi
  !     . f**(r,theta,f,vpar,mu) = f*(r,theta,phi-dphi,vpar,mu)
  !------------------------------------------------------------
  subroutine advec1D_PSM_phi(geom,dt,init_magnet,init_curr,f,maxalp_phi)
    use globals        , only : istart, iend, &
      jstart, jend, kstart, kend, lstart, lend, &
      errmax, limiter, pSpecies_id, hat_As
#ifdef _OPENMP
    use OMPutils_module, only : omp_get_thread_num
#endif
    use OMPutils_module, only : Romp_alpha, Romp_rhs, &
      Romp_vect, Romp1_0Nphi, Romp1_0Nvpar
    use velocities_edges_module, only : Phiedges_rphi, &
      Phiedges_thetaphi, comp_vExBgradphi_edges
    use globals        , only : transpose4D, conservative_scheme, &
         dom_r, dom_theta
!R3 #include "code_instrumentation/r3_info.h" !R3
    type(geometry)     , intent(in)    :: geom
    real(RKIND)        , intent(in)    :: dt
    type(init_magnetic), intent(in)    :: init_magnet
    type(init_current) , intent(in)    :: init_curr
    type(fdistribu5d)  , intent(inout) :: f
    real(RKIND)        , intent(out)   :: maxalp_phi
#ifdef HAS_ADV_PSM

    integer     :: ir, itheta, iphi, ivpar, ileft, iright
    real(RKIND) :: xphi, phimin, phimax, phistar
    real(RKIND) :: phin, depphi, err
    real(RKIND) :: sum, out 
    real(RKIND) :: vparl
    real(RKIND) :: finterpol, Fcentre
    real(RKIND) :: der_bc1
    integer     :: tid
    integer     :: count, iphipos
    real(RKIND) :: bstar_gradphi_edge
    real(RKIND) :: vD_gradphi_edge
    real(RKIND) :: inv_sqrtAs
    character(LEN=50) :: subp

    !-> array of size (0:geom%Nphi)
    real(RKIND), dimension(:), pointer :: alpha
    !-> array of size (-1:Nphi+1)
    real(RKIND), dimension(:), pointer :: rhsphi
    !-> array of size (-10:10)
    real(RKIND), dimension(:), pointer :: vect

    !-> toroidal components of ExB drift velocity
    real(RKIND), dimension(:), pointer :: vExB_gradphi_1D
    !-> toroidal components of curvature drift velocity
    real(RKIND), dimension(:), pointer :: vD_gradphi_1D

    real(RKIND), dimension(0:Nphi+1) :: flux
    real(RKIND) :: dp,dpc,lim,flux_upw
    real(RKIND), dimension(1:Nbthread) :: Romp_maxalp_phi

    integer :: min_i1, min_i2, max_i1, max_i2
    integer     :: min_i3, max_i3, min_i4, max_i4

    if (transpose4D) then
       if (conservative_scheme) then
          min_i1 = -1
          max_i1 = (f%n1)+2
          min_i2 = -1
          max_i2 = (f%n2)+1

          call write_param("min_i1", -1)
          call write_expr("max_i1" , "(fthis%n1)+2")
          call write_param("min_i2", -1)
          call write_expr("max_i2" , "(fthis%n2)+1")

          min_i3 = -1
          max_i3 = (f%n3)+2
          min_i4 = -1
          max_i4 = (f%n4)+1

          call write_param("min_i3", -1)
          call write_expr("max_i3" , "(fthis%n3)+2")
          call write_param("min_i4", -1)
          call write_expr("max_i4" , "(fthis%n4)+1")
       else
          min_i1 = -2
          max_i1 = (f%n1+2)+1
          min_i2 = -2
          max_i2 = (f%n2+1)+1

          call write_param("min_i1", -2)
          call write_expr("max_i1" , "(fthis%n1+2)+1")
          call write_param("min_i2", -2)
          call write_expr("max_i2" , "(fthis%n2+1)+1")
       endif
    else
       if (conservative_scheme) then
          min_i1 = -1
          max_i1 = (dom_r+1)+2
          min_i2 = -1
          max_i2 = (dom_theta+1)+1

          call write_param("min_i1", -1)
          call write_expr("max_i1" , "(dom_r+1)+2")
          call write_param("min_i2", -1)
          call write_expr("max_i2" , "(dom_theta+1)+1")
       else
          min_i1 = -2
          max_i1 = (dom_r+1)+1
          min_i2 = -2
          max_i2 = (dom_theta+1)+1

          call write_param("min_i1", -2)
          call write_expr("max_i1" , "(dom_r+1)+1")
          call write_param("min_i2", -2)
          call write_expr("max_i2" , "(dom_theta+1)+1")
       endif
    endif

    do tid = 1, Nbthread
       call FTR_ALLOC_2(f%hhspline2d_vparphi(tid)%coef,min_i4, max_i4, min_i3, max_i3)
       call FTR_ALLOC_2(f%hhspline2d_vparphi(tid)%bcoef,min_i4, max_i4, min_i3, max_i3)
       call FTR_ALLOC_2(f%hhspline2d_vparphi(tid)%aux,min_i4, max_i4, min_i3, max_i3)
       call FTR_ALLOC_2(f%hhspline2d_vparphi(tid)%auy,min_i4, max_i4, min_i3, max_i3)

       call FTR_ALLOC_2(f%hhspline2d_rtheta(tid)%coef,min_i1, max_i1, min_i2, max_i2)
       call FTR_ALLOC_2(f%hhspline2d_rtheta(tid)%bcoef,min_i1, max_i1, min_i2, max_i2)
       call FTR_ALLOC_2(f%hhspline2d_rtheta(tid)%aux,min_i1, max_i1, min_i2, max_i2)
       call FTR_ALLOC_2(f%hhspline2d_rtheta(tid)%auy,min_i1, max_i1, min_i2, max_i2)
    enddo

    subp = " advec1D_PSM_phi "//char(0)

!R3 call r3_info_begin (r3_info_index_0, 'CPU_advec1d_PSM_phi') !R3
    call Write_dyn_balise_open('CPU_advec1d_PSM_phi') !R4
    call clck_time(bclock_advec1Dphi)

    !*** computation of 1/sqrt{hat_As} ***
    inv_sqrtAs = 1._RKIND/sqrt(hat_As(pSpecies_id))

#ifdef _OPENMP
!$OMP PARALLEL private(tid,ir,itheta,iphi,ivpar, &
!$OMP  alpha,rhsphi,vect,depphi,iphipos, &
!$OMP  phin,phimin,phimax,phistar,xphi, &
!$OMP  sum,der_bc1,vparl, &
!$OMP  bstar_gradphi_edge,vD_gradphi_edge, &
!$OMP  vExB_gradphi_1D,vD_gradphi_1D, &
!$OMP  ileft,iright,out,count,err, &
!$OMP  finterpol,Fcentre) default(shared)
!$OMP BARRIER
    tid = 1+omp_get_thread_num()
#else
    tid = 1
#endif

    alpha           => Romp_alpha(tid)%val
    rhsphi          => Romp_rhs(tid)%val
    vect            => Romp_vect(tid)%val
    vExB_gradphi_1D => Romp1_0Nphi(tid)%val
    vD_gradphi_1D   => Romp1_0Nvpar(tid)%val

    phimin = f%hhspline2d_vparphi(tid)%ygrid(0)
    phimax = f%hhspline2d_vparphi(tid)%ygrid(Nphi)

    Romp_maxalp_phi(tid) = 0._RKIND
    alpha                = 0._RKIND

!$OMP DO SCHEDULE(static)
    do itheta = jstart,jend
      do ir = f%istart_modif,f%iend_modif
        !-> compute vD.gradphi
        call compute_vDgradphi(geom,init_magnet,init_curr, &
          ir,ir,itheta,itheta,0,Nvpar,.false.,vD_gradphi_1D)
        do ivpar = 0,Nvpar
          vparl = geom%vparg(ivpar)
          !*** Jacobian factor ***
!baoter (utiliser le jacobian spatial)
!VG!          f%values(ir,itheta,:,ivpar)= &
!VG!            f%values(ir,itheta,:,ivpar)*geom%rg(ir)
!eaoter
          !*** Values of the primitives ***
          rhsphi(0) = 0._RKIND
          do iphi = 1,Nphi
            rhsphi(iphi) = rhsphi(iphi-1) &
              + f%values(ir,itheta,iphi-1,ivpar)*geom%dphi
          end do
          sum = rhsphi(Nphi) 

          !*** Derivatives of G and D by local splines ***
          vect(-10:-1 )  = rhsphi(Nphi-10:Nphi-1)-sum
          vect(  1: 10)  = rhsphi(1:10)
          der_bc1        = derive10(vect,geom%dphi)
          rhsphi(-1)     = der_bc1
          rhsphi(Nphi+1) = rhsphi(-1)

          !*** cubic spline computation ***
          call hermite1D_theta(f%hhspline2d_vparphi(tid),rhsphi)

          !-> compute vExB.gradphi
          call comp_vExBgradphi_edges(geom, &
            init_magnet,init_curr, &
            Phiedges_rphi, Phiedges_thetaphi, &
            ir,ir,itheta,itheta,0,Nphi, &
            ivpar,ivpar,vExB_gradphi_1D)
          !-> compute bstar.gradphi at the edge
#ifdef NO_PRECOMP_BSTAR 
            call compute_bstar_contravariant(geom, &
              init_magnet,init_curr,ir,itheta,ivpar, &
              Sbstar_gradx3=bstar_gradphi_edge)
#else
            call precomp_bstar_gradphi(ir,itheta,ivpar, &
              bstar_gradphi_edge)
#endif
          do iphi = 0,Nphi-1
            ileft = iphi-1
            if (ileft<0) ileft = Nphi-1
            iright = iphi
            !-> compute vD_gradphi at the edge
            vD_gradphi_edge  = vD_gradphi_1D(ivpar)
            !-> computation of alpha
            alpha(iphi)      = &
              inv_sqrtAs*vparl*bstar_gradphi_edge + &
              vExB_gradphi_1D(iphi) + &
              vD_gradphi_edge  
          end do
          !*** periodicity in phi direction ***
          alpha(Nphi)          = alpha(0)
          alpha                = dt*alpha
          Romp_maxalp_phi(tid) = max(Romp_maxalp_phi(tid), &
            maxval(abs(alpha(0:Nphi))))

          !*** splitting in phi direction ***
          do iphi = 0,Nphi-1
            depphi  = alpha(iphi)
            phin    = &
              f%hhspline2d_vparphi(tid)%ygrid(iphi) - depphi
            phistar = 0.5_RKIND * &
              (f%hhspline2d_vparphi(tid)%ygrid(iphi)+phin )
            
            out = 0._RKIND
            if (phistar.ge.phimax) then
              out = abs(floor((phistar-phimin)/geom%Lphi))
              phistar = phistar - out*geom%Lphi  
            elseif (phistar.lt.phimin) then
              out = -abs(floor((phistar-phimin)/geom%Lphi))
              phistar = phistar - out*geom%Lphi  
            end if
            phistar = MIN(MAX(phistar,phimin),phimax)

            err   = 1._RKIND
            count = 0
            do while (err>errmax.and.count<countmax)
              if ( (phistar>=phimin).and.(phistar<=phimax) ) then
                !*** array position location ***
                call locate(phistar, &
                  f%hhspline2d_vparphi(tid)%ygrid(0:Nphi), &
                  geom%Nphi,geom%dphi,subp,iphipos)
              else
                print*, "error locate phi", phistar
                depphi  = alpha(iphi)
                phin    = &
                  f%hhspline2d_vparphi(tid)%ygrid(iphi) - depphi
                phistar = phin
              end if
              ! relative position in the cell 
              xphi    = (phistar - &
                f%hhspline2d_vparphi(tid)%ygrid(iphipos) ) &
                /geom%dphi
              ileft   = iphipos
              iright  = iphipos+1
              depphi  = alpha(ileft) * &
                (1._RKIND-xphi) + alpha(iright)*xphi 
              err     = phistar
              phin    = &
                f%hhspline2d_vparphi(tid)%ygrid(iphi) - depphi
              phistar = 0.5_RKIND * &
                (f%hhspline2d_vparphi(tid)%ygrid(iphi)+phin)

              out     = 0._RKIND
              if (phistar.gt.phimax) then
                out = abs(floor((phistar-phimin)/ &
                  geom%Lphi))
                phistar = phistar - out*geom%Lphi  
              elseif (phistar.lt.phimin) then
                out     = -abs(floor((phistar-phimin)/geom%Lphi))
                phistar = phistar - out*geom%Lphi  
              end if
              err   = abs(phistar-err)/geom%dphi
              count = count+1
            enddo ! while

            out = 0._RKIND
            if (phin.gt.phimax) then
              out  = abs(floor((phin-phimin)/geom%Lphi))
              phin = phin - out*geom%Lphi  
            elseif (phin.lt.phimin) then
              out  = -abs(floor((phin-phimin)/geom%Lphi))
              phin = phin - out*geom%Lphi  
            end if

            !*** f interpolation ***
            call interpol1d_theta(geom%Nphi,geom%dphi, &
              f%hhspline2d_vparphi(tid)%bcoef, &
              f%hhspline2d_vparphi(tid)%ygrid, &
              phin,finterpol)
            !**********************************************
            ! primitive interpolation, total mass "sum" 
            !  is added or substracted if particule
            !  goes out of the domain trajectory wise
            !**********************************************
            finterpol    = finterpol + out*sum
            rhsphi(iphi) = finterpol
          end do

          rhsphi(Nphi) = rhsphi(0)+sum
          do iphi = 0,Nphi-1
            f%values(ir,itheta,iphi,ivpar) =  &
              (rhsphi(iphi+1)-rhsphi(iphi)) &
              /geom%dphi
          enddo
          !*** Jacobian factor ***
!baoter (utiliser le jacobian spatial)
!VG!          f%values(ir,itheta,:,ivpar)= &
!VG!            f%values(ir,itheta,:,ivpar)/geom%rg(ir)
!eaoter
        end do
      end do
    end do
!$OMP END DO
!$OMP BARRIER
!$OMP END PARALLEL

    maxalp_phi = maxval(Romp_maxalp_phi(1:Nbthread))/(geom%dphi*dt)

    do tid = 1, Nbthread
       call FTR_DEALLOC(f%hhspline2d_vparphi(tid)%coef)
       call FTR_DEALLOC(f%hhspline2d_vparphi(tid)%bcoef)
       call FTR_DEALLOC(f%hhspline2d_vparphi(tid)%aux)
       call FTR_DEALLOC(f%hhspline2d_vparphi(tid)%auy)

       call FTR_DEALLOC(f%hhspline2d_rtheta(tid)%coef)
       call FTR_DEALLOC(f%hhspline2d_rtheta(tid)%bcoef)
       call FTR_DEALLOC(f%hhspline2d_rtheta(tid)%aux)
       call FTR_DEALLOC(f%hhspline2d_rtheta(tid)%auy)
    enddo
    
    call clck_time(eclock_advec1Dphi)
    call clck_diff(bclock_advec1Dphi,eclock_advec1Dphi, &
      global_time_advec1Dphi)

    call Write_dyn_balise_close('CPU_advec1d_PSM_phi') !R4
!R3 call r3_info_end (r3_info_index_0) !R3
#else 
    ! HAS_ADV_PSM
    maxalp_phi = 0._RKIND
#endif 
    ! HAS_ADV_PSM
  end subroutine advec1D_PSM_phi


  !----------------------------------------------------------
  ! splitting in v parallel direction
  !  - Solving of : df/dt+(dvpar/dt)*df/dvpar=0
  !    . dpar = dt*[ -(hat_Zs/sqrt{hat_As})*gradpar_J0Phi - 
  !             (1/sqrt{hat_As}*mu*gradpar_B + 
  !             K_Bdrift*vpar/B*(vExB.grad_B) ]
  !   with 
  !    -> gradpar_J0Phi = vec_bstar.grad_J0Phi
  !    -> gradpar_B     = vec_bstar.grad_B
  !    -> vExB.grad_B   = vExB_gradr*dB/dr + 
  !                       vExB_gradtheta*dB/dtheta
  !    . f*(r,theta,f,vpar,mu) = fn(r,theta,phi,vpar-dpar,mu)
  !
  ! Rk : K_Bdrift=1 if magnetic_drift=.true. and equal 
  !       to 0 otherwise. 
  !----------------------------------------------------------
  subroutine advec1D_PSM_vpar(geom,dt,init_prof, &
    init_magnet,init_curr,f,maxalp_vpar)
    use globals        , only : K_Bdrift, &
      pSpecies_id, hat_As, hat_Zs, &
      istart, iend, jstart, jend,  &
      dJ0Phidr_PNrPNthNphi, dJ0Phidtheta_PNrPNthNphi, &
      dJ0Phidphi_PNrPNthNphi
    use coord_system_class, only : inv_R
#ifdef _OPENMP
    use OMPutils_module, only : omp_get_thread_num
#endif
    use OMPutils_module, only : Romp_alpha, &
      Romp_rhs, Romp_vect, &
      Romp1_0Nvpar, Romp2_0Nvpar, Romp3_0Nvpar, &
      Romp4_0Nvpar, Romp5_0Nvpar, Romp6_0Nvpar
    use globals        , only : transpose4D, conservative_scheme, &
         dom_r, dom_theta

!R3 #include "code_instrumentation/r3_info.h" !R3
    type(geometry)     , intent(in)    :: geom
    real(RKIND)        , intent(in)    :: dt
    type(init_profile) , intent(in)    :: init_prof
    type(init_magnetic), intent(in)    :: init_magnet
    type(init_current) , intent(in)    :: init_curr
    type(fdistribu5d)  , intent(inout) :: f
    real(RKIND)        , intent(out)   :: maxalp_vpar

#ifdef HAS_ADV_PSM
    integer     :: tid
    integer     :: ir, itheta, iphi, ivpar, ileft, iright
    real(RKIND) :: mu_value, vstar, vparl
    real(RKIND) :: iotai, inv_Bij, inv_Rij
    real(RKIND) :: hatZs, inv_sqrtAs
    real(RKIND) :: sum
    real(RKIND) :: vparmin, vparmax
    real(RKIND) :: finterpol
    real(RKIND) :: dBdr_tmp, dBdtheta_tmp
    real(RKIND) :: gradpar_J0Phi_tmp, gradpar_B_tmp
    real(RKIND) :: vExB_gradB_tmp
    real(RKIND) :: depvpar

    !-> array of size (0:geom%Nvpar)
    real(RKIND), dimension(:), pointer :: alpha
    !-> array of size (-1:Nvpar+2)
    real(RKIND), dimension(:), pointer :: rhsvpar

    !-> components of ExB drift velocity
    real(RKIND), dimension(:), pointer :: vExB_gradr_1D
    real(RKIND), dimension(:), pointer :: vExB_gradtheta_1D
    !-> gradpar.B
    real(RKIND), dimension(:), pointer :: gradpar_B_1D
    !-> gradpar.Phi
    real(RKIND), dimension(:), pointer :: gradpar_J0Phi_1D

    real(RKIND), dimension(0:Nvpar+1) :: flux
    real(RKIND) :: dp,dpc,lim,flux_upw
    real(RKIND), dimension(1:Nbthread) :: Romp_maxalp_vpar

    integer :: min_i1, min_i2, max_i1, max_i2

!R3 call r3_info_begin (r3_info_index_0, 'CPU_advec1d_PSM_vpar') !R3
    call Write_dyn_balise_open('CPU_advec1d_PSM_vpar') !R4
    call clck_time(bclock_advec1Dvpar)
    
    if (transpose4D) then
       if (conservative_scheme) then
          min_i1 = -1
          max_i1 = (f%n1)+2
          min_i2 = -1
          max_i2 = (f%n2)+1

          call write_param("min_i1", -1)
          call write_expr("max_i1" , "(fthis%n1)+2")
          call write_param("min_i2", -1)
          call write_expr("max_i2" , "(fthis%n2)+1")
       else
          min_i1 = -2
          max_i1 = (f%n1+2)+1
          min_i2 = -2
          max_i2 = (f%n2+1)+1

          call write_param("min_i1", -2)
          call write_expr("max_i1" , "(fthis%n1+2)+1")
          call write_param("min_i2", -2)
          call write_expr("max_i2" , "(fthis%n2+1)+1")
       endif
    else
       if (conservative_scheme) then
          min_i1 = -1
          max_i1 = (dom_r+1)+2
          min_i2 = -1
          max_i2 = (dom_theta+1)+1

          call write_param("min_i1", -1)
          call write_expr("max_i1" , "(dom_r+1)+2")
          call write_param("min_i2", -1)
          call write_expr("max_i2" , "(dom_theta+1)+1")
       else
          min_i1 = -2
          max_i1 = (dom_r+1)+1
          min_i2 = -2
          max_i2 = (dom_theta+1)+1

          call write_param("min_i1", -2)
          call write_expr("max_i1" , "(dom_r+1)+1")
          call write_param("min_i2", -2)
          call write_expr("max_i2" , "(dom_theta+1)+1")
       endif
    endif

    do tid = 1, Nbthread
       call FTR_ALLOC_2(f%hhspline2d_vparphi(tid)%coef,min_i1, max_i1, min_i2, max_i2)
       call FTR_ALLOC_2(f%hhspline2d_vparphi(tid)%bcoef,min_i1, max_i1, min_i2, max_i2)
       call FTR_ALLOC_2(f%hhspline2d_vparphi(tid)%aux,min_i1, max_i1, min_i2, max_i2)
       call FTR_ALLOC_2(f%hhspline2d_vparphi(tid)%auy,min_i1, max_i1, min_i2, max_i2)

       call FTR_ALLOC_2(f%hhspline2d_rtheta(tid)%coef,min_i1, max_i1, min_i2, max_i2)
       call FTR_ALLOC_2(f%hhspline2d_rtheta(tid)%bcoef,min_i1, max_i1, min_i2, max_i2)
       call FTR_ALLOC_2(f%hhspline2d_rtheta(tid)%aux,min_i1, max_i1, min_i2, max_i2)
       call FTR_ALLOC_2(f%hhspline2d_rtheta(tid)%auy,min_i1, max_i1, min_i2, max_i2)
    enddo

    !*** initialisation of mu value ***
    mu_value = f%mu_value

    !*** computation of 1/sqrt{hat_As} ***
    hatZs      = hat_Zs(pSpecies_id)
    inv_sqrtAs = 1._RKIND/sqrt(hat_As(pSpecies_id))

#ifdef _OPENMP
!$OMP PARALLEL private(tid,ir,itheta,iphi,ivpar, &
!$OMP alpha,rhsvpar,sum,iotai,inv_Bij,inv_Rij, &
!$OMP dBdr_tmp,dBdtheta_tmp,gradpar_J0Phi_tmp,gradpar_B_tmp, &
!$OMP vExB_gradB_tmp,vExB_gradr_1D,vExB_gradtheta_1D, &
!$OMP gradpar_B_1D,gradpar_J0Phi_1D,vparmax,vparmin, &
!$OMP vparl,ileft,iright,vstar, &
!$OMP depvpar,finterpol ) default(shared)
!$OMP BARRIER
    tid = 1+omp_get_thread_num()
#else
    tid = 1
#endif

    alpha             => Romp_alpha(tid)%val
    rhsvpar           => Romp_rhs(tid)%val
    vExB_gradr_1D     => Romp3_0Nvpar(tid)%val
    vExB_gradtheta_1D => Romp4_0Nvpar(tid)%val
    gradpar_B_1D      => Romp5_0Nvpar(tid)%val
    gradpar_J0Phi_1D  => Romp6_0Nvpar(tid)%val

    vparmax = f%hhspline2d_vparphi(tid)%xgrid(Nvpar+1)
    vparmin = f%hhspline2d_vparphi(tid)%xgrid(0)

    Romp_maxalp_vpar(tid) = 0._RKIND
    alpha                 = 0._RKIND

!$OMP DO SCHEDULE(static)
    do iphi = 0,Nphi-1
      !*** vpar advection ***
      do itheta = jstart,jend  
        do ir = f%istart_modif,f%iend_modif
          !*** Values of the primitives ***
          rhsvpar(-1:Nvpar+2) = 0._RKIND
          do ivpar = 1,Nvpar+1
            rhsvpar(ivpar) = rhsvpar(ivpar-1) + &
              f%values(ir,itheta,iphi,ivpar-1)*geom%dvpar
          end do
          sum = rhsvpar(Nvpar+1)
          !************************************************ 
          !*** Natural conditions :                     ***  
          !***  Second derivatives G and D = 0          ***
          !***     rhsr(-1)      = 0._RKIND             ***
          !***     rhsr(Nvpar+2) = 0._RKIND             *** 
          !************************************************     
          call hermite2_1D_r(f%hhspline2d_vparphi(tid), &
            rhsvpar(-1:Nvpar+2),-1,Nvpar+2)
          iotai        = init_prof%iota(ir)
          inv_Rij      = inv_R(ir,itheta)
          inv_Bij      = 1._RKIND/init_magnet%B_norm(ir,itheta)
          dBdr_tmp     = init_magnet%dBdr(ir,itheta)
          dBdtheta_tmp = init_magnet%dBdtheta(ir,itheta)
          !-> computation of vExB contravariant components
          call compute_vExBgradr(geom,init_magnet,init_curr, &
            dJ0Phidtheta_PNrPNthNphi,dJ0Phidphi_PNrPNthNphi, &
            ir,ir,itheta,itheta,iphi,iphi,0,Nvpar,.false., &
            vExB_gradr_1D)
          call compute_vExBgradtheta(geom,init_magnet,init_curr, &
            dJ0Phidr_PNrPNthNphi,dJ0Phidphi_PNrPNthNphi, &
            ir,ir,itheta,itheta,iphi,iphi,0,Nvpar,.false., &
            vExB_gradtheta_1D)
          !-> computation of gradpar.B
          call compute_gradpar_B(geom,init_magnet,init_curr, &
            ir,ir,itheta,itheta,iphi,iphi, &
            0,Nvpar,.false.,gradpar_B_1D)
          !-> computation of gradpar.Phi
          call compute_gradpar_J0Phi(geom, &
            init_magnet,init_curr,ir,ir,itheta,itheta, &
            iphi,iphi,0,Nvpar,.false.,gradpar_J0Phi_1D)
          do ivpar = 0,Nvpar
            vparl             = geom%vparg(ivpar)
            gradpar_B_tmp     = gradpar_B_1D(ivpar)
            gradpar_J0Phi_tmp = gradpar_J0Phi_1D(ivpar)
            !-> computation of vExB.gradB
            vExB_gradB_tmp    = &
              vExB_gradr_1D(ivpar)*dBdr_tmp + &
              vExB_gradtheta_1D(ivpar)*dBdtheta_tmp
            !-> computation of vpar displacement
            alpha(ivpar) = &
              - inv_sqrtAs*mu_value*gradpar_B_tmp - &
              (hatZs*inv_sqrtAs)*gradpar_J0Phi_tmp + &
              K_Bdrift*vparl*inv_Bij*vExB_gradB_tmp
          end do
          alpha                 = dt*alpha
          Romp_maxalp_vpar(tid) = max(Romp_maxalp_vpar(tid), &
            maxval(abs(alpha(0:Nvpar))))

          do ivpar = 0,Nvpar+1
            !*** splitting in vpar direction ***
            ileft   = max(ivpar-1,0)
            iright  = min(ivpar,Nvpar)
            depvpar = 0.5_RKIND*(alpha(ileft)+alpha(iright)) *  &
              (1._RKIND-(alpha(iright)-alpha(ileft))/geom%dvpar)
            vstar   = f%hhspline2d_vparphi(tid)%xgrid(ivpar) - &
              depvpar
            if (vstar>vparmin.and.vstar<vparmax) then
              !*** f interpolation ***
              call interpol1d_r(geom%Nvpar+1,geom%dvpar, &
                f%hhspline2d_vparphi(tid)%bcoef, &
                f%hhspline2d_vparphi(tid)%xgrid,vstar,finterpol)
            elseif (vstar<=vparmin) then
              finterpol = f%values(ir,itheta,iphi,0)
              finterpol = rhsvpar(0) - finterpol*(vparmin-vstar)
            elseif &
              (vstar>=vparmax) then
              finterpol = f%values(ir,itheta,iphi,Nvpar)
              finterpol = rhsvpar(Nvpar+1) - &
                finterpol * (vparmax-vstar)
            end if
            !-> no entropic limiter because too diffusive
            rhsvpar(ivpar) = finterpol
          end do ! 0,Nvpar+1
          do ivpar = 0,Nvpar
            f%values(ir,itheta,iphi,ivpar) =  &
              ( rhsvpar(ivpar+1)-rhsvpar(ivpar) )/geom%dvpar
          enddo
        end do
      end do
    end do
!$OMP END DO
!$OMP BARRIER
!$OMP END PARALLEL

    maxalp_vpar = maxval(Romp_maxalp_vpar(1:Nbthread))/(geom%dvpar*dt)

    do tid = 1, Nbthread
       call FTR_DEALLOC(f%hhspline2d_vparphi(tid)%coef)
       call FTR_DEALLOC(f%hhspline2d_vparphi(tid)%bcoef)
       call FTR_DEALLOC(f%hhspline2d_vparphi(tid)%aux)
       call FTR_DEALLOC(f%hhspline2d_vparphi(tid)%auy)

       call FTR_DEALLOC(f%hhspline2d_rtheta(tid)%coef)
       call FTR_DEALLOC(f%hhspline2d_rtheta(tid)%bcoef)
       call FTR_DEALLOC(f%hhspline2d_rtheta(tid)%aux)
       call FTR_DEALLOC(f%hhspline2d_rtheta(tid)%auy)
    enddo

    call clck_time(eclock_advec1Dvpar)
    call clck_diff(bclock_advec1Dvpar, &
      eclock_advec1Dvpar,global_time_advec1Dvpar)

    call Write_dyn_balise_close('CPU_advec1d_PSM_vpar') !R4
!R3 call r3_info_end (r3_info_index_0) !R3
#else 
    ! HAS_ADV_PSM
    maxalp_vpar = 0._RKIND
#endif 
    ! HAS_ADV_PSM
  end subroutine advec1D_PSM_vpar
end module advec1D_PSM_module
