!**************************************************************
!  Copyright Euratom-CEA
!  Authors : 
!     Virginie Grandgirard (virginie.grandgirard@cea.fr)
!     Chantal Passeron (chantal.passeron@cea.fr)
!     Guillaume Latu (guillaume.latu@cea.fr)
!     Xavier Garbet (xavier.garbet@cea.fr)
!     Philippe Ghendrih (philippe.ghendrih@cea.fr)
!     Yanick Sarazin (yanick.sarazin@cea.fr)
!  
!  This code GYSELA (for GYrokinetic SEmi-LAgrangian) 
!  is a 5D gyrokinetic global full-f code for simulating 
!  the plasma turbulence in a tokamak.
!  
!  This software is governed by the CeCILL-B license 
!  under French law and abiding by the rules of distribution 
!  of free software.  You can  use, modify and redistribute 
!  the software under the terms of the CeCILL-B license as 
!  circulated by CEA, CNRS and INRIA at the following URL
!  "http://www.cecill.info". 
!**************************************************************

#include "config-F.inc"

!---------------------------------------------------------
! file : advec2D_PPM2.F90
! date : 18/10/2009
!   2D advection in (r,theta) directions used in
!   PPM2 Colella-Sekora JCP 2008
! developed by 
!  Jean-Philippe Braeunig (braeunig@math.u-strasbg.fr)
!
! Rk: This PPM2 scheme is such that the discrete divergence
!     is equal to 0
!---------------------------------------------------------
module advec2D_PPM2_module
  use globals, only : Nr, Ntheta, Nphi, Nvpar, Nbthread
  use wrap_alloc_module
  use integration_module
  use clock_module
  use geometry_class
  use init_profile_class
  use init_magnetic_class
  use fdistribu5d_class
  use interpolation_module
  use velocities_edges_module
  use wrap_alloc_module

  implicit none

  real(RKIND), parameter, private :: Kppm2 = 1.5_RKIND

  include "gyrokinetic/Bstar_inline.h"
  include "gyrokinetic/velocities_inline.h"
  !******************************
  contains
  !******************************
#include "gyrokinetic/Bstar_inline.F90"
#include "gyrokinetic/velocities_inline.F90"

  !******************************  
  !--------------------------------------------------------
  ! splitting in r parallel direction
  !  - Resolution of : df/dt+(dr/dt)*df/dr=0
  !         dr/dt = vpar*bstar_gradr +  vExB_gradr 
  !                 + vD_gradr   
  !
  !  . f*(r,theta,f,vpar,mu) = fn(r-dr,theta,phi,vpar,mu)
  !--------------------------------------------------------
  subroutine advec2D_PPM2(geom,dt,coord_sys,init_prof, &
    init_magnet,init_curr,f,maxalp_r,maxalp_theta)
    use MPIutils_module, only : f4d_transp
    use globals        , only : istart, iend, &
      jstart, jend, kstart, kend, lstart, lend, &
      deltat
#ifdef _OPENMP
    use OMPutils_module, only : omp_get_thread_num
#endif
    use OMPutils_module, only : Romp_alpha, Romp_rhs, &
      Romp1_0Nrp1, Romp1_0Nr, Romp2_0Nr, Romp3_0Nr, &
      Romp4_0Nr, Romp5_0Nr, &
      Romp1_0Ntheta, Romp2_0Ntheta, Romp3_0Ntheta, &
      Romp4_0Ntheta, Romp5_0Ntheta, Romp6_0Ntheta, &
      Romp1_0Nr_0Ntheta, Romp2_0Nr_0Ntheta, &
      Romp1_0Nrp1_0Ntheta, Romp2_0Nrp1_0Ntheta
    use velocities_edges_module, only : Phiedges_rtheta, &
      Phiedges_rphi, comp_vExBgradr_edges, &
      Phiedges_thetaphi, comp_vExBgradtheta_edges

!R3 #include "code_instrumentation/r3_info.h" !R3
    type(geometry)     , intent(in)    :: geom
    real(RKIND)        , intent(in)    :: dt
    type(coord_system) , intent(in)    :: coord_sys
    type(init_profile) , intent(in)    :: init_prof
    type(init_magnetic), intent(in)    :: init_magnet
    type(init_current) , intent(in)    :: init_curr
    type(fdistribu5d)  , intent(inout) :: f
    real(RKIND)        , intent(out)   :: maxalp_r,maxalp_theta

#ifdef HAS_ADV_PPM
    integer     :: ir, itheta, iphi, ivpar,ix,iy,m
    real(RKIND) :: rmin, rmax, rstar
    real(RKIND) :: depr
    real(RKIND) :: sum, err
    real(RKIND) :: bstar_gradr_left, bstar_gradr_right
    real(RKIND) :: bstar_gradr_edge
    real(RKIND) :: vD_gradr_left, vD_gradr_right
    real(RKIND) :: vD_gradr_edge
    real(RKIND) :: vparl
    real(RKIND) :: finterpol, Fcentre, out
    integer     :: irpos, ig, id, igg, idd, iggg, iddd
    integer     :: k

    integer     :: tid, ileft, iright, ithetapos
    real(RKIND) :: thetamin, thetamax, thetastar
    real(RKIND) :: xtheta, deptheta
    real(RKIND) :: bstar_gradtheta_left, bstar_gradtheta_right
    real(RKIND) :: bstar_gradtheta_edge
    real(RKIND) :: vD_gradtheta_left, vD_gradtheta_right
    real(RKIND) :: vD_gradtheta_edge

    ! array for the edges of the cell (0:geom%Nr+1)
    real(RKIND), dimension(:), pointer :: alpha  

    !-> radial components of ExB drift velocity
    real(RKIND), dimension(:), pointer :: vExB_gradr_1D
    !-> radial components of curvature drift velocity
    real(RKIND), dimension(:), pointer :: vD_gradr_1D
    !-> poloidal components of ExB drift velocity
    real(RKIND), dimension(:), pointer :: vExB_gradtheta_1D
    !-> poloidal components of curvature drift velocity
    real(RKIND), dimension(:), pointer :: vD_gradtheta_1D

    real(RKIND) :: xx, dp, dpc, lim
    !-> array of size (0:Nr+1,0:Ntheta)
    real(RKIND), dimension(:,:), pointer :: fluxr 
    real(RKIND), dimension(:,:), pointer :: udtr
    !-> array of size (0:Nr,0:Ntheta)
    real(RKIND), dimension(:,:), pointer :: fluxtheta 
    real(RKIND), dimension(:,:), pointer :: udttheta
    !-> array of size (0:max(Nr,Ntheta))
    real(RKIND), dimension(:), pointer   :: ar,al,delta,a6

    real(RKIND), dimension(1:Nbthread) :: Romp_maxalp_r
    real(RKIND), dimension(1:Nbthread) :: Romp_maxalp_th
    real(RKIND) :: Mdd,Mdd_c,Mdd_l,Mdd_r,Mdd_lim
    real(RKIND) :: Ndd,Ndd_c,Ndd_l,Ndd_r,Ndd_lim

!R3 call r3_info_begin (r3_info_index_0, 'CPU_advec1d_PPM2_r') !R3
    call Write_dyn_balise_open('CPU_advec1d_PPM2_r') !R4
    call clck_time(bclock_advec1Dr)

   do tid = 1, Nbthread
   call FTR_ALLOC_2(Romp1_0Nr_0Ntheta(tid)%val,0,Nr,0,Ntheta)
   call FTR_ALLOC_2(Romp2_0Nr_0Ntheta(tid)%val,0,Nr,0,Ntheta)
   enddo

#ifdef _OPENMP
!$OMP PARALLEL default(none) &
!$OMP private(tid,ir,itheta,iphi,ivpar, &
!$OMP  rmin,rmax,thetamin,thetamax, &
!$OMP  depr,deptheta,vparl,alpha,sum, &
!$OMP  dp,dpc,lim,out, &
!$OMP  vExB_gradtheta_1D, vD_gradtheta_1D, &
!$OMP  vExB_gradr_1D,vD_gradr_1D, &
!$OMP  bstar_gradr_left,bstar_gradr_right, &
!$OMP  bstar_gradtheta_left,bstar_gradtheta_right, & 
!$OMP  vD_gradr_left,vD_gradr_right, &
!$OMP  vD_gradtheta_left,vD_gradtheta_right, & 
!$OMP  bstar_gradr_edge,bstar_gradtheta_edge, &
!$OMP  vD_gradr_edge,vD_gradtheta_edge, &
!$OMP  fluxr,udtr,fluxtheta,udttheta, &
!$OMP  ileft,iright,irpos,rstar,err,xx, &
!$OMP  ig, id, igg, idd, iggg, iddd, &
!$OMP  Mdd,Mdd_c,Mdd_l,Mdd_r,Mdd_lim, &
!$OMP  Ndd,Ndd_c,Ndd_l,Ndd_r,Ndd_lim, &
!$OMP  ar,al,delta,a6, &
!$OMP  finterpol,Fcentre) &
!$OMP shared(f,geom,init_magnet,init_curr,dt, &
!$OMP  kstart,kend,lstart,lend,Nr,Ntheta, &
!$OMP  Romp_alpha,Romp_rhs,Romp1_0Nrp1,Romp1_0Nr, &
!$OMP  Romp2_0Nr,Romp3_0Nr,Romp4_0Nr,Romp5_0Nr, &
!$OMP  Romp1_0Ntheta,Romp2_0Ntheta,Romp3_0Ntheta, &
!$OMP  Romp4_0Ntheta,Romp5_0Ntheta,Romp6_0Ntheta, &
!$OMP  Romp1_0Nr_0Ntheta,Romp2_0Nr_0Ntheta, &
!$OMP  Romp1_0Nrp1_0Ntheta,Romp2_0Nrp1_0Ntheta, &
!$OMP  Romp_maxalp_r,Romp_maxalp_th, &
!$OMP  Phiedges_rtheta,Phiedges_thetaphi,Phiedges_rphi, &
!$OMP  f4D_transp)
!$OMP BARRIER
    tid = 1+omp_get_thread_num()
#else
    tid = 1
#endif

    rmin = f%hhspline2d_rtheta(tid)%xgrid(0)
    rmax = f%hhspline2d_rtheta(tid)%xgrid(Nr+1)

    alpha             => Romp_alpha(tid)%val
    vExB_gradr_1D     => Romp1_0Nrp1(tid)%val
    vD_gradr_1D       => Romp1_0Nr(tid)%val      
    vExB_gradtheta_1D => Romp1_0Ntheta(tid)%val
    vD_gradtheta_1D   => Romp2_0Ntheta(tid)%val
      
    fluxr     => Romp1_0Nrp1_0Ntheta(tid)%val
    udtr      => Romp2_0Nrp1_0Ntheta(tid)%val
    fluxtheta => Romp1_0Nr_0Ntheta(tid)%val
    udttheta  => Romp2_0Nr_0Ntheta(tid)%val

    thetamin = f%hhspline2d_rtheta(tid)%ygrid(0)
    thetamax = f%hhspline2d_rtheta(tid)%ygrid(Ntheta)

    Romp_maxalp_r (tid) = 0._RKIND
    Romp_maxalp_th(tid) = 0._RKIND
    alpha               = 0._RKIND
    
    if (geom%Ntheta.ge.geom%Nr) then
      ar    => Romp3_0Ntheta(tid)%val  
      al    => Romp4_0Ntheta(tid)%val
      delta => Romp5_0Ntheta(tid)%val
      a6    => Romp6_0Ntheta(tid)%val
    else
      ar    => Romp2_0Nr(tid)%val  
      al    => Romp3_0Nr(tid)%val
      delta => Romp4_0Nr(tid)%val
      a6    => Romp5_0Nr(tid)%val
    end if

!$OMP DO SCHEDULE(static)
    do iphi = kstart,kend
      do ivpar = lstart,lend
        vparl = geom%vparg(ivpar)
        !********************************
        !*  Velocity in theta direction *
        !********************************
        do ir = 0,Nr
          !-> compute vExB.gradtheta
          call comp_vExBgradtheta_edges(geom, &
            init_magnet,init_curr, &
            Phiedges_rtheta,Phiedges_thetaphi, &
            ir,ir,0,Ntheta,iphi,iphi,ivpar,ivpar, &
            vExB_gradtheta_1D)
          !-> compute vD.gradtheta
          call compute_vDgradtheta(geom, &
            init_magnet,init_curr, &
            ir,ir,0,Ntheta,iphi,iphi, &
            ivpar,ivpar,.true.,vD_gradtheta_1D)
          do itheta = 0,Ntheta-1
            ileft = itheta-1
            if (ileft<0) ileft = Ntheta-1
            iright = itheta
            !-> compute bstar.gradtheta at the edge
#ifdef NO_PRECOMP_BSTAR 
            call compute_bstar_contravariant(geom, &
              init_magnet,init_curr,ir,ileft,ivpar, &
              Sbstar_gradx2=bstar_gradtheta_left)
            call compute_bstar_contravariant(geom, &
              init_magnet,init_curr,ir,iright,ivpar, &
              Sbstar_gradx2=bstar_gradtheta_right)
#else
            call precomp_transp_bstar_gradtheta(ir, &
              ileft,ivpar,bstar_gradtheta_left)
            call precomp_transp_bstar_gradtheta(ir, &
              iright,ivpar,bstar_gradtheta_right)
#endif
            bstar_gradtheta_edge = 0.5_RKIND * &
              (bstar_gradtheta_left+bstar_gradtheta_right)
            !-> compute vD_gradtheta at the edge
            vD_gradtheta_left  = vD_gradtheta_1D(ileft)
            vD_gradtheta_right = vD_gradtheta_1D(iright)
            vD_gradtheta_edge  = 0.5_RKIND * &
              (vD_gradtheta_left+vD_gradtheta_right)
            !-> computation of alpha
            alpha(itheta)      = &
              vparl*bstar_gradtheta_edge + &
              vExB_gradtheta_1D(itheta) + &
              vD_gradtheta_edge  
            alpha(itheta)      = &
              vparl*bstar_gradtheta_edge + &
              vExB_gradtheta_1D(itheta) + &
              vD_gradtheta_edge  
          end do
          !*** periodicity in theta direction ***
          alpha(Ntheta)       = alpha(0)
          alpha               = dt*alpha

          Romp_maxalp_th(tid) = max(Romp_maxalp_th(tid), &
            maxval(abs(alpha(0:Ntheta-1))))

          !*** splitting in theta direction ***
          deptheta = 0._RKIND
          do itheta = 0,Ntheta-1
            deptheta            = alpha(itheta)
            udttheta(ir,itheta) = deptheta
          enddo
          !--> theta velocity at the cell faces
          udttheta(ir,Ntheta) = udttheta(ir,0)
        enddo

        !*******************************
        !*   Velocity in r direction   *
        !*******************************
        do itheta = 0,Ntheta-1
          !-> compute vExB.gradr
          call comp_vExBgradr_edges(geom, &
            init_magnet,init_curr, &
            Phiedges_rtheta,Phiedges_rphi, &
            0,Nr+1,itheta,itheta, &
            iphi,iphi,ivpar,ivpar,vExB_gradr_1D)
          !-> compute vD.gradr 
          call compute_vDgradr(geom, &
            init_magnet,init_curr, &
            0,Nr,itheta,itheta,iphi,iphi, &
            ivpar,ivpar,.true.,vD_gradr_1D)

          do ir = 0,Nr+1
            ileft  = max(ir-1,0)
            iright = min(ir,Nr) 
            !-> compute bstar.gradr at the edge
#ifdef NO_PRECOMP_BSTAR 
            call compute_bstar_contravariant(geom, &
              init_magnet,init_curr,ileft,itheta,ivpar, &
              Sbstar_gradx1=bstar_gradr_left)
            call compute_bstar_contravariant(geom, &
              init_magnet,init_curr,iright,itheta,ivpar, &
              Sbstar_gradx1=bstar_gradr_right)
#else
            call precomp_transp_bstar_gradr(ileft, &
              itheta,ivpar,bstar_gradr_left)
            call precomp_transp_bstar_gradr(iright, &
              itheta,ivpar,bstar_gradr_right)
#endif
            bstar_gradr_edge = 0.5_RKIND * &
              (bstar_gradr_left+bstar_gradr_right)
            !-> compute vD_gradr at the edge
            vD_gradr_left    = vD_gradr_1D(ileft)
            vD_gradr_right   = vD_gradr_1D(iright)
            vD_gradr_edge    = 0.5_RKIND * &
              (vD_gradr_left+vD_gradr_right)
            !-> computation of alpha
            alpha(ir)        = vparl*bstar_gradr_edge + &
              vExB_gradr_1D(ir) + vD_gradr_edge  
          end do
          alpha              = dt*alpha
          Romp_maxalp_r(tid) = max(Romp_maxalp_r(tid), &
            maxval(abs(alpha(0:Nr+1))))

          depr = 0._RKIND 
          do ir = 0,Nr+1
            ileft           = max(ir-1,0)
            iright          = min(ir,Nr)
            depr            = alpha(ir)
            udtr(ir,itheta) = depr
          enddo
        enddo
        !--> r velocity at the cell faces
        udtr(:,Ntheta) = udtr(:,0)

!baremettre
!VG!          !--> Checking of the 'div(udt)=0' property
!VG!          call compute_displacement_divergence( &
!VG!            jacobian_space,geom%Nr,geom%Ntheta, &
!VG!            geom%dr,geom%dtheta,udtr,udttheta)
!earemettre
        
        !*******************************
        ! FLUX R
        !*******************************
        do itheta = 0,Ntheta-1
          !*** Jacobian factor ***
!baoter (ATTENTION prendre en compte correctement le jacobien)
!VG!          f4d_transp%values(:,itheta,iphi,ivpar)= &
!VG!            f4d_transp%values(:,itheta,iphi,ivpar)*geom%rg(:)
!eaoter

          !*** PPM2 ***
          do ir = 0,Nr+1
            iddd = min(ir+2,Nr)
            idd  = min(ir+1,Nr)
            id   = min(ir  ,Nr)
            ig   = max(ir-1,0)
            igg  = max(ir-2,0)
            iggg = max(ir-3,0)

            fluxr(ir,itheta) = ( 37._RKIND * &
              (f4d_transp(ig,itheta,iphi,ivpar) &
              + f4d_transp(id  ,itheta,iphi,ivpar)) &
              - 8._RKIND*(f4d_transp(igg ,itheta,iphi,ivpar) &
              + f4d_transp(idd ,itheta,iphi,ivpar)) &
              + f4d_transp(iggg,itheta,iphi,ivpar) &
              + f4d_transp(iddd,itheta,iphi,ivpar) )/60._RKIND

            !----> second derivative of f at the nodes
            if ( (  f4d_transp(ig,itheta,iphi,ivpar)-fluxr(ir,itheta)) &
              * (f4d_transp(id,itheta,iphi,ivpar) - &
              fluxr(ir,itheta)) > 0._RKIND) then
              Ndd  = 3._RKIND*( f4d_transp(ig,itheta,iphi,ivpar) &
                - 2._RKIND*fluxr(ir,itheta)    &
                + f4d_transp(id,itheta,iphi,ivpar) )/(geom%dr**2)
              Ndd_l = ( f4d_transp(igg,itheta,iphi,ivpar) &
                - 2._RKIND*f4d_transp(ig,itheta,iphi,ivpar) &
                + f4d_transp(id ,itheta,iphi,ivpar) )/(geom%dr**2)
              Ndd_r = ( f4d_transp(ig ,itheta,iphi,ivpar)&
                - 2._RKIND*f4d_transp(id,itheta,iphi,ivpar) &
                + f4d_transp(idd,itheta,iphi,ivpar))/(geom%dr**2)

              if ( Ndd>=0._RKIND.and.Ndd_l>=0._RKIND.and. &
                Ndd_r>=0._RKIND) then
                Ndd_lim = min( Ndd, Kppm2*Ndd_l, Kppm2*Ndd_r )
              elseif ( Ndd<0._RKIND .and. Ndd_l<0._RKIND .and. &
                Ndd_r<0._RKIND) then
                Ndd_lim = max( Ndd, Kppm2*Ndd_l, Kppm2*Ndd_r )
              else 
                Ndd_lim = 0._RKIND
              endif
              fluxr(ir,itheta) = 0.5_RKIND * ( &
                f4d_transp(ig,itheta,iphi,ivpar) &
                + f4d_transp(id,itheta,iphi,ivpar)) &
                - (geom%dr**2)*Ndd_lim/3._RKIND
            endif
          enddo

          do ir = 0,Nr
            al(ir)    = fluxr(ir,itheta)
            ar(ir)    = fluxr(ir+1,itheta)
            delta(ir) = ar(ir)-al(ir)
            a6(ir)    = 6._RKIND*(f4d_transp(ir,itheta,iphi,ivpar)-(al(ir)+ar(ir))/2._RKIND)

            !*** limiter definition ***
            !---> local minimum or local maximum
            if (  (ar(ir)-f4d_transp(ir,itheta,iphi,ivpar)) &
              *(f4d_transp(ir,itheta,iphi,ivpar)-al(ir)) > 0._RKIND ) then
              if (  delta(ir)*a6(ir) > delta(ir)**2 ) then! Non mono L
                al(ir) =  3._RKIND*f4d_transp(ir,itheta,iphi,ivpar) - 2._RKIND*ar(ir)
              endif
              if ( -delta(ir)**2 > delta(ir)*a6(ir) ) then! Non mono R
                ar(ir) =  3._RKIND*f4d_transp(ir,itheta,iphi,ivpar) - 2._RKIND*al(ir)
              endif
            else 
              iddd = min(ir+2,Nr)
              idd  = min(ir+1,Nr)
              id   = min(ir  ,Nr)
              ig   = max(ir-1,0)
              igg  = max(ir-2,0)
              !---> Limitation
              Mdd   = -2._RKIND*a6(ir)/(geom%dr**2)
              !----> second derivative of f at the cells  
              Mdd_c = ( f4d_transp(ig,itheta,iphi,ivpar) &
                - 2._RKIND*f4d_transp(id,itheta,iphi,ivpar) &
                + f4d_transp(idd,itheta,iphi,ivpar)) / (geom%dr**2)
              Mdd_l = ( f4d_transp(igg,itheta,iphi,ivpar) &
                - 2._RKIND*f4d_transp(ig,itheta,iphi,ivpar)  &
                + f4d_transp(id,itheta,iphi,ivpar)) / (geom%dr**2)
              Mdd_r = ( f4d_transp(id,itheta,iphi,ivpar) &
                - 2._RKIND*f4d_transp(idd,itheta,iphi,ivpar) &
                + f4d_transp(iddd,itheta,iphi,ivpar)) / (geom%dr**2)

              if ( Mdd>=0._RKIND  .and.Mdd_c>=0._RKIND.and. &
                Mdd_l>=0._RKIND.and.Mdd_r>=0._RKIND) then
                Mdd_lim = min( Mdd,Kppm2*Mdd_c, &
                  Kppm2*Mdd_l,Kppm2*Mdd_r )
              elseif ( Mdd<0._RKIND  .and.Mdd_c<0._RKIND.and. &
                Mdd_l<0._RKIND.and.Mdd_r<0._RKIND) then
                Mdd_lim = max( Mdd,Kppm2*Mdd_c, &
                  Kppm2*Mdd_l,Kppm2*Mdd_r )
              else
                Mdd_lim = 0._RKIND
              endif
              if (Mdd/=0._RKIND) then
                al(ir) = f4d_transp(ir,itheta,iphi,ivpar) + &
                  (al(ir)-f4d_transp(ir,itheta,iphi,ivpar))*Mdd_lim/Mdd
                ar(ir) = f4d_transp(ir,itheta,iphi,ivpar) + &
                  (ar(ir)-f4d_transp(ir,itheta,iphi,ivpar))*Mdd_lim/Mdd
              else
                al(ir) = f4d_transp(ir,itheta,iphi,ivpar)
                ar(ir) = f4d_transp(ir,itheta,iphi,ivpar)
              endif
            endif
            delta(ir) = ar(ir)-al(ir)
            a6(ir)    = 6._RKIND*(f4d_transp(ir,itheta,iphi,ivpar)-(al(ir)+ar(ir))/2._RKIND)
          enddo

          do ir = 1,Nr
            xx = udtr(ir,itheta)/geom%dr
            if (abs(xx)>1.d0) xx=sign(1.d0,xx)
            if (udtr(ir,itheta)>=0._RKIND) then
              fluxr(ir,itheta) = ar(ir-1) &
                + 0.5_RKIND*xx*(a6(ir-1) - delta(ir-1)) &
                - a6(ir-1)*(xx**2/3._RKIND) 
            else
              fluxr(ir,itheta) = al(ir) &
                - 0.5_RKIND*xx*(a6(ir) + delta(ir)) &
                - a6(ir)*(xx**2/3._RKIND) 
            endif
            fluxr(ir,itheta) = xx*geom%dr*fluxr(ir,itheta)
          enddo

          !*** PPM2 boundary conditions ***
          xx = udtr(0,itheta)/geom%dr 
          if (abs(xx)>1.d0) xx = sign(1.d0,xx)
          if (udtr(0,itheta)<=0._RKIND) then
            fluxr(0,itheta) = al(0) &
              - 0.5_RKIND*xx*(a6(0) + delta(0)) &
              - a6(0)*(xx**2/3._RKIND)
          else
            fluxr(0,itheta) = f4d_transp(0,itheta,iphi,ivpar)
          endif
          fluxr(0,itheta) = xx*geom%dr*fluxr(0,itheta)

          xx = udtr(Nr+1,itheta)/geom%dr 
          if (abs(xx)>1.d0) xx=sign(1.d0,xx)
          if (udtr(Nr+1,itheta)>=0._RKIND) then
            fluxr(Nr+1,itheta) = ar(Nr) &
              + 0.5_RKIND*xx*(a6(Nr) - delta(Nr)) &
              - a6(Nr)*(xx**2/3._RKIND) 
          else   
            fluxr(Nr+1,itheta) = f4d_transp(Nr,itheta,iphi,ivpar)
          endif
          fluxr(Nr+1,itheta) = xx*geom%dr*fluxr(Nr+1,itheta)

          !*** Jacobian factor ***
!baoter (ATTENTION prendre en compte correctement le jacobien)
!VG!          f4d_transp(:,itheta,iphi,ivpar) = &
!VG!            f4d_transp(:,itheta,iphi,ivpar)/geom%rg(:)
!eaoter
        end do
        fluxr(:,Ntheta)=fluxr(:,0)

        !*******************************
        ! FLUX THETA
        !*******************************
        do ir = 0,Nr
          !--> PPM2
          do itheta = 0,Ntheta
            iddd = itheta+2
            idd  = itheta+1
            id   = itheta  
            ig   = itheta-1
            igg  = itheta-2
            iggg = itheta-3
            if (iddd>Ntheta-1) iddd = iddd-Ntheta
            if (idd >Ntheta-1) idd  = idd-Ntheta
            if (id  >Ntheta-1) id   = id -Ntheta
            if (ig  <0       ) ig   = ig +Ntheta
            if (igg <0       ) igg  = igg+Ntheta
            if (iggg<0       ) iggg = iggg+Ntheta

            fluxtheta(ir,itheta) = ( 37._RKIND*(f4d_transp(ir,ig,iphi,ivpar) &
              + f4d_transp(ir,id,iphi,ivpar)) &
              - 8._RKIND*(f4d_transp(ir,igg,iphi,ivpar) &
              + f4d_transp(ir,idd,iphi,ivpar)) &
              + f4d_transp(ir,iggg,iphi,ivpar) &
              + f4d_transp(ir,iddd,iphi,ivpar) )/60._RKIND

            !----> second derivative of f at the nodes
            if ( (f4d_transp(ir,ig,iphi,ivpar)-fluxtheta(ir,itheta)) &
              * (f4d_transp(ir,id,iphi,ivpar)-fluxtheta(ir,itheta)) > 0._RKIND) then
              Ndd   = 3._RKIND*( f4d_transp(ir,ig,iphi,ivpar) &
                - 2._RKIND*fluxtheta(ir,itheta)    &
                + f4d_transp(ir,id,iphi,ivpar) )/(geom%dtheta**2)
              Ndd_l = ( f4d_transp(ir,igg,iphi,ivpar) &
                - 2._RKIND*f4d_transp(ir,ig,iphi,ivpar) &
                + f4d_transp(ir,id,iphi,ivpar) )/(geom%dtheta**2)
              Ndd_r = ( f4d_transp(ir,ig,iphi,ivpar) &
                - 2._RKIND*f4d_transp(ir,id,iphi,ivpar) &
                + f4d_transp(ir,idd,iphi,ivpar))/(geom%dtheta**2)

              if ( Ndd>=0._RKIND.and.Ndd_l>=0._RKIND &
                .and.Ndd_r>=0._RKIND) then
                Ndd_lim = min( Ndd , Kppm2*Ndd_l &
                  , Kppm2*Ndd_r )
              elseif ( Ndd<0._RKIND.and.Ndd_l<0._RKIND.and. &
                Ndd_r<0._RKIND) then
                Ndd_lim = max( Ndd , Kppm2*Ndd_l &
                  , Kppm2*Ndd_r )
              else 
                Ndd_lim = 0._RKIND
              endif
              fluxtheta(ir,itheta) = 0.5_RKIND*( f4d_transp(ir,ig,iphi,ivpar) &
                + f4d_transp(ir,id,iphi,ivpar)) &
                - (geom%dtheta**2)*Ndd_lim/3._RKIND
            endif
          enddo

          do itheta = 0,Ntheta-1
            al(itheta)    = fluxtheta(ir,itheta)
            ar(itheta)    = fluxtheta(ir,itheta+1)
            delta(itheta) = ar(itheta)-al(itheta)
            a6(itheta)    = 6._RKIND*(f4d_transp(ir,itheta,iphi,ivpar) &
              - (al(itheta)+ar(itheta))/2._RKIND)

            !*** limiter definition ***
            !min ou max local
            if ( (ar(itheta)-f4d_transp(ir,itheta,iphi,ivpar)) &
              *(f4d_transp(ir,itheta,iphi,ivpar)-al(itheta)) > 0._RKIND ) then
              if (  delta(itheta)*a6(itheta) > delta(itheta)**2 ) then! Non mono L
                al(itheta) =  3._RKIND*f4d_transp(ir,itheta,iphi,ivpar) &
                  - 2._RKIND*ar(itheta)
              endif
              if ( -delta(itheta)**2 > delta(itheta)*a6(itheta) ) then! Non mono R
                ar(itheta) =  3._RKIND*f4d_transp(ir,itheta,iphi,ivpar) &
                  - 2._RKIND*al(itheta)
              endif
            else
              !----> Limitation
              iddd = itheta+2
              idd  = itheta+1
              id   = itheta  
              ig   = itheta-1
              igg  = itheta-2

              if (iddd>Ntheta-1) iddd = iddd-Ntheta
              if (idd >Ntheta-1) idd  = idd-Ntheta
              if (id  >Ntheta-1) id   = id -Ntheta
              if (ig  <0       ) ig   = ig +Ntheta
              if (igg <0       ) igg  = igg+Ntheta


              Mdd   = -2._RKIND*a6(itheta)/(geom%dtheta**2)
              !----> second derivative of f at the cells
              Mdd_c = (f4d_transp(ir,ig ,iphi,ivpar)&
                - 2._RKIND*f4d_transp(ir,id ,iphi,ivpar) &
                + f4d_transp(ir,idd ,iphi,ivpar)) / (geom%dtheta**2)
              Mdd_l = (f4d_transp(ir,igg,iphi,ivpar) & 
                - 2._RKIND*f4d_transp(ir,ig ,iphi,ivpar)  &
                + f4d_transp(ir,id  ,iphi,ivpar)) / (geom%dtheta**2)
              Mdd_r = (f4d_transp(ir,id ,iphi,ivpar) &
                - 2._RKIND*f4d_transp(ir,idd,iphi,ivpar) &
                + f4d_transp(ir,iddd,iphi,ivpar)) / (geom%dtheta**2)

              if ( Mdd>=0._RKIND  .and.Mdd_c>=0._RKIND.and. &
                Mdd_l>=0._RKIND.and.Mdd_r>=0._RKIND) then
                Mdd_lim = min( Mdd,Kppm2*Mdd_c, &
                  Kppm2*Mdd_l,Kppm2*Mdd_r )
              elseif ( Mdd<0._RKIND  .and.Mdd_c<0._RKIND.and. &
                Mdd_l<0._RKIND.and.Mdd_r<0._RKIND) then
                Mdd_lim = max( Mdd,Kppm2*Mdd_c, &
                  Kppm2*Mdd_l,Kppm2*Mdd_r )
              else
                Mdd_lim = 0._RKIND
              endif
              if (Mdd/=0._RKIND) then
                al(itheta) = f4d_transp(ir,itheta,iphi,ivpar) + &
                  (al(itheta)-f4d_transp(ir,itheta,iphi,ivpar)) &
                  *Mdd_lim/Mdd
                ar(itheta) = f4d_transp(ir,itheta,iphi,ivpar) + &
                  (ar(itheta)-f4d_transp(ir,itheta,iphi,ivpar)) &
                  *Mdd_lim/Mdd
              else
                al(itheta) = f4d_transp(ir,itheta,iphi,ivpar)
                ar(itheta) = f4d_transp(ir,itheta,iphi,ivpar)
              endif
            endif
            delta(itheta) = ar(itheta)-al(itheta)
            a6(itheta)    = 6._RKIND*(f4d_transp(ir,itheta,iphi,ivpar) &
              - (al(itheta)+ar(itheta))/2._RKIND)
            ! end Limitation  
          enddo

          do itheta=1,Ntheta-1
            xx = udttheta(ir,itheta)/geom%dtheta
            if (abs(xx)>1.d0) xx=sign(1.d0,xx)
            if (udttheta(ir,itheta)>=0._RKIND) then
              fluxtheta(ir,itheta) = ar(itheta-1) &
                + 0.5_RKIND*xx*(a6(itheta-1) &
                - delta(itheta-1)) &
                - a6(itheta-1)*(xx**2/3._RKIND) 
            else
              fluxtheta(ir,itheta) = al(itheta) &
                - 0.5_RKIND*xx*(a6(itheta) &
                + delta(itheta)) &
                - a6(itheta)*(xx**2/3._RKIND) 
            endif
            fluxtheta(ir,itheta) = xx*geom%dtheta*fluxtheta(ir,itheta)
          enddo

          xx = udttheta(ir,0)/geom%dtheta  !! on a udttheta(Ntheta) = udttheta(0)
          if (abs(xx)>1.d0) xx=sign(1.d0,xx)
          if (udttheta(ir,0)<=0._RKIND) then
            fluxtheta(ir,0) = al(0) - 0.5_RKIND*xx*(a6(0) + delta(0)) &
              - a6(0)*(xx**2/3._RKIND)
            fluxtheta(ir,Ntheta) =  fluxtheta(ir,0)
          else   
            fluxtheta(ir,Ntheta) = ar(Ntheta-1) &
              + 0.5_RKIND*xx*(a6(Ntheta-1) &
              - delta(Ntheta-1)) &
              - a6(Ntheta-1)*(xx**2/3._RKIND)
            fluxtheta(ir,0) = fluxtheta(ir,Ntheta)
          endif
          fluxtheta(ir,0)      = xx*geom%dtheta*fluxtheta(ir,0)
          fluxtheta(ir,Ntheta) = xx*geom%dtheta*fluxtheta(ir,Ntheta)

          do itheta = 0,Ntheta-1
            f4d_transp(ir,itheta,iphi,ivpar) = &
              f4d_transp(ir,itheta,iphi,ivpar) &
              -( f%hhspline2d_rtheta(tid)%xgrid(ir+1)*fluxr(ir+1,itheta) &
                -f%hhspline2d_rtheta(tid)%xgrid(ir)*fluxr(ir,itheta)) &
              /geom%rg(ir)/geom%dr &
              -(fluxtheta(ir,itheta+1)-fluxtheta(ir,itheta)) &
              /geom%dtheta
          enddo
        enddo
      end do
    end do
!$OMP END DO
!$OMP BARRIER
!$OMP END PARALLEL
    maxalp_r     = maxval(Romp_maxalp_r(1:Nbthread))/(geom%dr*dt)
    maxalp_theta = maxval(Romp_maxalp_th(1:Nbthread))/(geom%dtheta*dt)

   do tid = 1, Nbthread
   call FTR_DEALLOC(Romp1_0Nr_0Ntheta(tid)%val)
   call FTR_DEALLOC(Romp2_0Nr_0Ntheta(tid)%val)
   enddo

    call clck_time(eclock_advec1Dr)
    call clck_diff(bclock_advec1Dr,eclock_advec1Dr, &
      global_time_advec1Dr)
    call Write_dyn_balise_close('CPU_advec1d_PPM2_r') !R4
!R3 call r3_info_end (r3_info_index_0) !R3
#else 
    ! HAS_ADV_PPM
    maxalp_theta = 0._RKIND
    maxalp_r     = 0._RKIND
#endif 
    ! HAS_ADV_PPM
  end subroutine advec2D_PPM2


  !--------------------------------------------------------
  ! Chek if the displacment is at null divergence
  !  div u = 1/Js * \sum_{i=1}^3 \partial_i(Js*u.gradxi)
  !--------------------------------------------------------
  subroutine compute_displacement_divergence(Js_cell, &
    Nr,Ntheta,dr,dtheta,udtr_node,udttheta_node)
    real(RKIND), dimension(:,:), intent(in)  :: Js_cell
    integer                    , intent(in)  :: Nr, Ntheta
    real(RKIND)                , intent(in)  :: dr, dtheta
    real(RKIND), dimension(:,:), intent(in)  :: udtr_node
    real(RKIND), dimension(:,:), intent(in)  :: udttheta_node

#ifdef HAS_ADV_PPM
    integer     :: ir, itheta
    integer     :: ir_left, ir_right
    integer     :: itheta_left, itheta_right
    real(RKIND) :: Js_left, Js_right, Js_bottom, Js_top
    real(RKIND), dimension(0:Nr,0:Ntheta) :: div_udt_cell

    do ir = 0,Nr
      do itheta = 0,Ntheta
        ir_left      = max(ir-1,0)
        ir_right     = min(ir+1,Nr)
        itheta_left  = itheta-1
        if (itheta_left.eq.0) itheta_left = Ntheta-1
        itheta_right = itheta+1
        if (itheta_left.eq.Ntheta) itheta_left = 1
        Js_left   = 0.5_RKIND*(Js_cell(ir_left,itheta)+Js_cell(ir,itheta))
        Js_right  = 0.5_RKIND*(Js_cell(ir,itheta)+Js_cell(ir_right,itheta))
        Js_bottom = 0.5_RKIND*(Js_cell(ir,itheta_left)+Js_cell(ir,itheta))
        Js_top    = 0.5_RKIND*(Js_cell(ir,itheta)+Js_cell(ir,itheta_right))
        !*** divergence computation ***
        div_udt_cell(ir,itheta) = ( (Js_right*udtr_node(ir+1,itheta) - &
          Js_left*udtr_node(ir,itheta))/dr + &
          (Js_top*udttheta_node(ir,itheta+1) - &
          Js_bottom*udttheta_node(ir,itheta))/dtheta ) / &
          Js_cell(ir,itheta)
      enddo
    enddo
    print*,'div u = ', maxval(abs(div_udt_cell))
#endif 
    ! HAS_ADV_PPM
  end subroutine compute_displacement_divergence

end module advec2D_PPM2_module
