!**************************************************************
!  Copyright Euratom-CEA
!  Authors : 
!     Virginie Grandgirard (virginie.grandgirard@cea.fr)
!     Chantal Passeron (chantal.passeron@cea.fr)
!     Guillaume Latu (guillaume.latu@cea.fr)
!     Xavier Garbet (xavier.garbet@cea.fr)
!     Philippe Ghendrih (philippe.ghendrih@cea.fr)
!     Yanick Sarazin (yanick.sarazin@cea.fr)
!  
!  This code GYSELA (for GYrokinetic SEmi-LAgrangian) 
!  is a 5D gyrokinetic global full-f code for simulating 
!  the plasma turbulence in a tokamak.
!  
!  This software is governed by the CeCILL-B license 
!  under French law and abiding by the rules of distribution 
!  of free software.  You can  use, modify and redistribute 
!  the software under the terms of the CeCILL-B license as 
!  circulated by CEA, CNRS and INRIA at the following URL
!  "http://www.cecill.info". 
!**************************************************************

#include "config-F.inc"

!---------------------------------------------------------
! file : advec2D_PSM.F90
! date : 18/10/2009
!   2D advection in (r,theta) directions used in
!   PSM scheme
! developed by 
!  Jean-Philippe Braeunig (braeunig@math.u-strasbg.fr)
!
! Rk: This PSM scheme is such that the discrete divergence
!     is equal to 0
!---------------------------------------------------------
module advec2D_PSM_module
  use globals, only : Nr, Ntheta, Nphi, Nvpar, Nbthread
  use wrap_alloc_module
  use integration_module
  use clock_module
  use geometry_class
  use init_profile_class
  use init_magnetic_class
  use fdistribu5d_class
  use interpolation_module
  use velocities_edges_module  
  implicit none

  real(RKIND), parameter, private :: Klim = 5._RKIND

  include "gyrokinetic/Bstar_inline.h"
  include "gyrokinetic/velocities_inline.h"
  !******************************
  contains
  !******************************
#include "gyrokinetic/Bstar_inline.F90"
#include "gyrokinetic/velocities_inline.F90"


  !--------------------------------------------------------
  ! splitting in r parallel direction
  !  - Resolution of : df/dt+(dr/dt)*df/dr=0
  !         dr/dt = vpar*bstar_gradr +  vExB_gradr 
  !                 + vD_gradr   
  !
  !  . f*(r,theta,f,vpar,mu) = fn(r-dr,theta,phi,vpar,mu)
  !--------------------------------------------------------
  subroutine advec2D_PSM(geom,dt,init_prof, &
    init_magnet,init_curr,f,maxalp_r,maxalp_theta)
    use MPIutils_module, only : f4d_transp
    use globals        , only : istart, iend, &
      jstart, jend, kstart, kend, lstart, lend, &
      deltat, limiter
#ifdef _OPENMP
    use OMPutils_module, only : omp_get_thread_num
#endif
    use OMPutils_module, only : Romp_alpha, Romp_rhs, Romp_vect, &
      Romp1_0Nr, Romp1_0Nrp1, Romp1_0Ntheta, Romp2_0Ntheta, &
      Romp1_0Nr_0Ntheta, Romp2_0Nr_0Ntheta, &
      Romp1_0Nrp1_0Ntheta, Romp2_0Nrp1_0Ntheta
    use velocities_edges_module, only : Phiedges_rtheta, &
      Phiedges_rphi, comp_vExBgradr_edges, &
      Phiedges_thetaphi, comp_vExBgradtheta_edges
    use globals, only : transpose4D, conservative_scheme, &
         dom_r, dom_theta
!R3 #include "code_instrumentation/r3_info.h" !R3
    type(geometry)     , intent(in)    :: geom
    real(RKIND)        , intent(in)    :: dt
    type(init_profile) , intent(in)    :: init_prof
    type(init_magnetic), intent(in)    :: init_magnet
    type(init_current) , intent(in)    :: init_curr
    type(fdistribu5d)  , intent(inout) :: f
    real(RKIND)        , intent(out)   :: maxalp_r
    real(RKIND)        , intent(out)   :: maxalp_theta

#ifdef HAS_ADV_PSM
    integer     :: ir, itheta, iphi, ivpar,ix,iy,m,iup
    real(RKIND) :: rmin, rmax, rstar
    real(RKIND) :: xr, rn
    real(RKIND) :: sum, err
    real(RKIND) :: bstar_gradr_left, bstar_gradr_right
    real(RKIND) :: bstar_gradr_edge
    real(RKIND) :: vD_gradr_left, vD_gradr_right
    real(RKIND) :: vD_gradr_edge
    real(RKIND) :: vparl
    real(RKIND) :: finterpol, Fcentre,out
    integer     :: irpos,ig,id,igg,idd,iggg,iddd
    integer     :: k

    real(RKIND) :: thetamin, thetamax, thetastar
    real(RKIND) :: thetan 
    real(RKIND) :: xtheta
    integer     :: tid, ileft, iright, ithetapos
    real(RKIND) :: der_bc1
    real(RKIND) :: bstar_gradtheta_left, bstar_gradtheta_right
    real(RKIND) :: bstar_gradtheta_edge
    real(RKIND) :: vD_gradtheta_left, vD_gradtheta_right
    real(RKIND) :: vD_gradtheta_edge

    !-> array for the edges of the cell (0:geom%Nr+1)
    real(RKIND), dimension(:), pointer :: alpha 
    !-> array for the edges of the cell (-1:max(geom%Nr,geom%Ntheta+1)  
    real(RKIND), dimension(:), pointer :: rhs
    !-> array of size (-10:10)
    real(RKIND), dimension(:), pointer :: vect


    !-> radial components of ExB drift velocity
    real(RKIND), dimension(:), pointer :: vExB_gradr_1D
    !-> radial components of curvature drift velocity
    real(RKIND), dimension(:), pointer :: vD_gradr_1D
    !-> poloidal components of ExB drift velocity
    real(RKIND), dimension(:), pointer :: vExB_gradtheta_1D
    !-> poloidal components of curvature drift velocity
    real(RKIND), dimension(:), pointer :: vD_gradtheta_1D

    real(RKIND) :: xx, dp, dpc, lim, flux_upw
    !-> array of size (0:Nr+1,0:Ntheta)
    real(RKIND), dimension(:,:), pointer :: fluxr 
    real(RKIND), dimension(:,:), pointer :: udtr
    !-> array of size (0:Nr,0:Ntheta)
    real(RKIND), dimension(:,:), pointer :: fluxtheta
    real(RKIND), dimension(:,:), pointer :: udttheta
    real(RKIND), dimension(1:Nbthread)   :: Romp_maxalp_r
    real(RKIND), dimension(1:Nbthread)   :: Romp_maxalp_th

    integer :: min_i1, min_i2, max_i1, max_i2

!R3 call r3_info_begin (r3_info_index_0, 'CPU_advec1d_PSM_r') !R3
    call Write_dyn_balise_open('CPU_advec1d_PSM_r') !R4
    call clck_time(bclock_advec1Dr)

    if (transpose4D) then
       if (conservative_scheme) then
          min_i1 = -1
          max_i1 = (f%n1)+2
          min_i2 = -1
          max_i2 = (f%n2)+1

          call write_param("min_i1", -1)
          call write_expr("max_i1" , "(fthis%n1)+2")
          call write_param("min_i2", -1)
          call write_expr("max_i2" , "(fthis%n2)+1")
       else
          min_i1 = -2
          max_i1 = (f%n1+2)+1
          min_i2 = -2
          max_i2 = (f%n2+1)+1

          call write_param("min_i1", -2)
          call write_expr("max_i1" , "(fthis%n1+2)+1")
          call write_param("min_i2", -2)
          call write_expr("max_i2" , "(fthis%n2+1)+1")
       endif
    else
       if (conservative_scheme) then
          min_i1 = -1
          max_i1 = (dom_r+1)+2
          min_i2 = -1
          max_i2 = (dom_theta+1)+1

          call write_param("min_i1", -1)
          call write_expr("max_i1" , "(dom_r+1)+2")
          call write_param("min_i2", -1)
          call write_expr("max_i2" , "(dom_theta+1)+1")
       else
          min_i1 = -2
          max_i1 = (dom_r+1)+1
          min_i2 = -2
          max_i2 = (dom_theta+1)+1

          call write_param("min_i1", -2)
          call write_expr("max_i1" , "(dom_r+1)+1")
          call write_param("min_i2", -2)
          call write_expr("max_i2" , "(dom_theta+1)+1")
       endif
    endif

    do tid = 1, Nbthread
      call FTR_ALLOC_2(f%hhspline2d_rtheta(tid)%coef,min_i1, max_i1, min_i2, max_i2)
      call FTR_ALLOC_2(f%hhspline2d_rtheta(tid)%bcoef,min_i1, max_i1, min_i2, max_i2)
      call FTR_ALLOC_2(f%hhspline2d_rtheta(tid)%aux,min_i1, max_i1, min_i2, max_i2)
      call FTR_ALLOC_2(f%hhspline2d_rtheta(tid)%auy,min_i1, max_i1, min_i2, max_i2)

      call FTR_ALLOC_2(Romp1_0Nr_0Ntheta(tid)%val,0,Nr,0,Ntheta)
      call FTR_ALLOC_2(Romp2_0Nr_0Ntheta(tid)%val,0,Nr,0,Ntheta)
   enddo

#ifdef _OPENMP
!$OMP PARALLEL default(none) &
!$OMP private(tid,ir,itheta,iphi,ivpar, &
!$OMP  rmin,rmax,thetamin,thetamax, &
!$OMP  vparl,alpha,rhs,vect,sum, &
!$OMP  dp,dpc,flux_upw,lim,thetan,der_bc1,out,iup, &
!$OMP  vExB_gradtheta_1D, vD_gradtheta_1D, &
!$OMP  vExB_gradr_1D,vD_gradr_1D, &
!$OMP  bstar_gradr_left,bstar_gradr_right, &
!$OMP  bstar_gradtheta_left,bstar_gradtheta_right, & 
!$OMP  vD_gradr_left,vD_gradr_right, &
!$OMP  vD_gradtheta_left,vD_gradtheta_right, & 
!$OMP  bstar_gradr_edge,bstar_gradtheta_edge, &
!$OMP  vD_gradr_edge,vD_gradtheta_edge, &
!$OMP  fluxr,udtr,fluxtheta,udttheta, &
!$OMP  ileft,iright,irpos,rn,rstar,err,xr, &
!$OMP  finterpol,Fcentre) &
!$OMP shared(f,geom,init_magnet,init_curr,dt, &
!$OMP  kstart,kend,lstart,lend,Nr,Ntheta, &
!$OMP  Romp_alpha,Romp_rhs,Romp1_0Nrp1,Romp1_0Nr, &
!$OMP  Romp_vect,Romp1_0Ntheta,Romp2_0Ntheta, &
!$OMP  Romp1_0Nr_0Ntheta,Romp2_0Nr_0Ntheta, &
!$OMP  Romp1_0Nrp1_0Ntheta,Romp2_0Nrp1_0Ntheta, &
!$OMP  Romp_maxalp_r,Romp_maxalp_th, &
!$OMP  Phiedges_rtheta,Phiedges_thetaphi,Phiedges_rphi, &
!$OMP  f4D_transp,limiter)
!$OMP BARRIER
    tid = 1+omp_get_thread_num()
#else
    tid = 1
#endif

    rmin = f%hhspline2d_rtheta(tid)%xgrid(0)
    rmax = f%hhspline2d_rtheta(tid)%xgrid(Nr+1)

    alpha         => Romp_alpha(tid)%val
    rhs           => Romp_rhs(tid)%val
    vExB_gradr_1D => Romp1_0Nrp1(tid)%val
    vD_gradr_1D   => Romp1_0Nr(tid)%val    
    vect          => Romp_vect(tid)%val
  
    vExB_gradtheta_1D => Romp1_0Ntheta(tid)%val
    vD_gradtheta_1D   => Romp2_0Ntheta(tid)%val
      
    fluxr     => Romp1_0Nrp1_0Ntheta(tid)%val
    udtr      => Romp2_0Nrp1_0Ntheta(tid)%val
    fluxtheta => Romp1_0Nr_0Ntheta(tid)%val
    udttheta  => Romp2_0Nr_0Ntheta(tid)%val

    thetamin = f%hhspline2d_rtheta(tid)%ygrid(0)
    thetamax = f%hhspline2d_rtheta(tid)%ygrid(Ntheta)

    Romp_maxalp_r(tid)  = 0._RKIND
    Romp_maxalp_th(tid) = 0._RKIND
    alpha               = 0._RKIND

!$OMP DO SCHEDULE(static)
    do iphi = kstart,kend
      do ivpar = lstart,lend
        vparl = geom%vparg(ivpar)
        !********************************
        !*  Velocity in theta direction *
        !********************************
        do ir = 0,Nr
          !-> compute vExB.gradtheta
          call comp_vExBgradtheta_edges(geom, &
            init_magnet,init_curr, &
            Phiedges_rtheta,Phiedges_thetaphi, &
            ir,ir,0,Ntheta,iphi,iphi,ivpar,ivpar, &
            vExB_gradtheta_1D)
          !-> compute vD.gradtheta
          call compute_vDgradtheta(geom, &
            init_magnet,init_curr, &
            ir,ir,0,Ntheta,iphi,iphi, &
            ivpar,ivpar,.true.,vD_gradtheta_1D)
          do itheta = 0,Ntheta-1
            ileft = itheta-1
            if (ileft<0) ileft = Ntheta-1
            iright = itheta
            !-> compute bstar.gradtheta at the edge
#ifdef NO_PRECOMP_BSTAR 
            call compute_bstar_contravariant(geom, &
              init_magnet,init_curr,ir,ileft,ivpar, &
              Sbstar_gradx2=bstar_gradtheta_left)
            call compute_bstar_contravariant(geom, &
              init_magnet,init_curr,ir,iright,ivpar, &
              Sbstar_gradx2=bstar_gradtheta_right)
#else
            call precomp_transp_bstar_gradtheta(ir, &
              ileft,ivpar,bstar_gradtheta_left)
            call precomp_transp_bstar_gradtheta(ir, &
              iright,ivpar,bstar_gradtheta_right)
#endif
            bstar_gradtheta_edge = 0.5_RKIND * &
              (bstar_gradtheta_left+bstar_gradtheta_right)
            !-> compute vD_gradtheta at the edge
            vD_gradtheta_left  = vD_gradtheta_1D(ileft)
            vD_gradtheta_right = vD_gradtheta_1D(iright)
            vD_gradtheta_edge  = 0.5_RKIND * &
              (vD_gradtheta_left+vD_gradtheta_right)
            !-> computation of alpha
            alpha(itheta)      = &
              vparl*bstar_gradtheta_edge + &
              vExB_gradtheta_1D(itheta) + &
              vD_gradtheta_edge  
            alpha(itheta)      = &
              vparl*bstar_gradtheta_edge + &
              vExB_gradtheta_1D(itheta) + &
              vD_gradtheta_edge  
          end do
          !*** periodicity in theta direction ***
          alpha(Ntheta) = alpha(0)
          alpha         = dt*alpha

          Romp_maxalp_th(tid) = max(Romp_maxalp_th(tid), &
            maxval(abs(alpha(0:Ntheta-1))))

          !*** splitting in theta direction ***
          do itheta = 0,Ntheta-1
            udttheta(ir,itheta) = alpha(itheta)
          enddo
          udttheta(ir,Ntheta) = udttheta(ir,0)
        enddo

        !*******************************
        !*   Velocity in r direction   *
        !*******************************
        do itheta = 0,Ntheta-1
          !-> compute vExB.gradr
          call comp_vExBgradr_edges(geom, &
            init_magnet,init_curr, &
            Phiedges_rtheta,Phiedges_rphi, &
            0,Nr+1,itheta,itheta, &
            iphi,iphi,ivpar,ivpar,vExB_gradr_1D)
          !-> compute vD.gradr 
          call compute_vDgradr(geom, &
            init_magnet,init_curr, &
            0,Nr,itheta,itheta,iphi,iphi, &
            ivpar,ivpar,.true.,vD_gradr_1D)
          do ir = 0,Nr+1
            ileft  = max(ir-1,0)
            iright = min(ir,Nr) 
            !-> compute bstar.gradr at the edge
#ifdef NO_PRECOMP_BSTAR 
            call compute_bstar_contravariant(geom, &
              init_magnet,init_curr,ileft,itheta,ivpar, &
              Sbstar_gradx1=bstar_gradr_left)
            call compute_bstar_contravariant(geom, &
              init_magnet,init_curr,iright,itheta,ivpar, &
              Sbstar_gradx1=bstar_gradr_right)
#else
            call precomp_transp_bstar_gradr(ileft, &
              itheta,ivpar,bstar_gradr_left)
            call precomp_transp_bstar_gradr(iright, &
              itheta,ivpar,bstar_gradr_right)
#endif
            bstar_gradr_edge = 0.5_RKIND * &
              (bstar_gradr_left+bstar_gradr_right)
            !-> compute vD_gradr at the edge
            vD_gradr_left  = vD_gradr_1D(ileft)
            vD_gradr_right = vD_gradr_1D(iright)
            vD_gradr_edge  = 0.5_RKIND * &
              (vD_gradr_left+vD_gradr_right)
            !-> computation of alpha
            alpha(ir)      = vparl*bstar_gradr_edge + &
              vExB_gradr_1D(ir) + vD_gradr_edge  
          end do
          alpha              = dt*alpha
          Romp_maxalp_r(tid) = max(Romp_maxalp_r(tid), &
            maxval(abs(alpha(0:Nr+1))))

          do ir = 0,Nr+1
            ileft           = max(ir-1,0)
            iright          = min(ir,Nr)
            udtr(ir,itheta) = alpha(ir)
          enddo
        enddo
        !--> r velocity at the cell faces
        udtr(:,Ntheta) = udtr(:,0)

        !=============================================================
        do itheta = 0,Ntheta-1
          !*** Jacobian factor ***
!baoter (ATTENTION prendre en compte correctement le jacobien)
!VG!          f4d_transp(:,itheta,iphi,ivpar)= &
!VG!            f4d_transp(:,itheta,iphi,ivpar)*geom%rg(:)
!eaoter

          !*** Values of the primitives ***
          rhs(-1:Nr+2) = 0._RKIND
          do ir = 1,Nr+1
            rhs(ir) = rhs(ir-1) &
              + f4d_transp(ir-1,itheta,iphi,ivpar) * geom%dr
          end do
          sum = rhs(Nr+1)-rhs(0) 
          !************************************************ 
          !*** Natural conditions :                     ***
          !***  Second derivatives G and D = 0          ***
          !***     rhs(-1)   = 0._RKIND                ***
          !***     rhs(Nr+2) = 0._RKIND                *** 
          !************************************************
          call hermite2_1D_r(f%hhspline2d_rtheta(tid), &
            rhs(-1:Nr+2),-1,Nr+2)

          do ir = 0,Nr+1
            rn    = f%hhspline2d_rtheta(tid)%xgrid(ir)-udtr(ir,itheta)

            !*** interpolation of the distribution function ***
            if (rn.gt.rmin.and.rn.lt.rmax) then
              !-> inside the domain
              call interpol1d_r(geom%Nr+1,geom%dr, &
                f%hhspline2d_rtheta(tid)%bcoef, &
                f%hhspline2d_rtheta(tid)%xgrid,rn,finterpol)
              !-> Outside the domain by the left
              !->   characteristic foot exterior of the domain
              !->   primitive derivative at the 
              !->   boundary = function at the boundary
            elseif (rn.le.rmin) then
              finterpol = f4d_transp(0,itheta,iphi,ivpar) 
              finterpol = rhs(0) - finterpol * (rmin-rn)
            elseif (rn.ge.rmax) then
              !-> Outside the domain by the right
              !->   characteristic foot exterior of the domain
              finterpol = f4d_transp(Nr,itheta,iphi,ivpar) 
              finterpol = rhs(Nr+1) - finterpol * (rmax-rn)
            endif

            fluxr(ir,itheta) = rhs(ir)-finterpol

            ! *** SLS limiter ***
            if (limiter == 2) then
              ileft  = max(ir-1,0)
              iright = min(ir,Nr)
              dpc = f4d_transp(iright,itheta,iphi,ivpar) - &
                f4d_transp(ileft ,itheta,iphi,ivpar)
              if (abs(dpc)>1.e-10_RKIND) then
                ! dpc = sign(1.e-10_RKIND,dpc)
                if (udtr(ir,itheta) > 0._RKIND ) then
                  dp = f4d_transp(ileft,itheta,iphi,ivpar) - &
                    f4d_transp(max(ileft-1,0),itheta,iphi,ivpar)
                  flux_upw = udtr(ir,itheta) * &
                    f4d_transp(ileft,itheta,iphi,ivpar)
                else
                  dp = f4d_transp(min(iright+1,Nr), &
                    itheta,iphi,ivpar) - &
                    f4d_transp(iright,itheta,iphi,ivpar)
                  flux_upw = udtr(ir,itheta) * &
                    f4d_transp(iright,itheta,iphi,ivpar)
                endif
                lim = max(0._RKIND, &
                  min(Klim*abs(dp/dpc),1._RKIND))
                fluxr(ir,itheta) = lim*fluxr(ir,itheta) + &
                  (1._RKIND-lim)*flux_upw
              endif
            endif !sls
          end do ! 0,Nr+1

          !*** Jacobian factor ***
!baoter (ATTENTION prendre en compte correctement le jacobien)
!VG!          f4d_transp(:,itheta,iphi,ivpar)= &
!VG!            f4d_transp(:,itheta,iphi,ivpar)/geom%rg(:)
!eaoter
        end do
        fluxr(:,Ntheta) = fluxr(:,0)

        !=============================================================
        do ir = 0,Nr
          !*** Values of the primitives ***
          rhs(0) = 0._RKIND
          do itheta = 1,Ntheta
            rhs(itheta) = rhs(itheta-1) + &
              f4d_transp(ir,itheta-1,iphi,ivpar)*geom%dtheta
          end do
          sum = rhs(Ntheta) 
          !*** Derivatives of G and D by local splines
          do itheta = -10,-1
            vect(itheta) = rhs(Ntheta+itheta)-sum
          end do
          do itheta = 1,10
            vect(itheta)   = rhs(itheta)
          end do
          der_bc1       = derive10(vect,geom%dtheta)    
          rhs(-1)       = der_bc1
          rhs(Ntheta+1) = rhs(-1)

          !*** cubic spline computation ***
          call hermite1D_theta(f%hhspline2d_rtheta(tid),rhs)

          do itheta = 0,Ntheta-1                
            thetan = f%hhspline2d_rtheta(tid)%ygrid(itheta) - &
              udttheta(ir,itheta)
            out    = 0._RKIND
            if (thetan.gt.thetamax) then
              out    = abs(floor((thetan-thetamin)/geom%Ltheta))
              thetan = thetan - out*geom%Ltheta  
            elseif (thetan.lt.thetamin) then
              out    = -abs(floor((thetan-thetamin)/geom%Ltheta))
              thetan = thetan - out*geom%Ltheta  
            endif

            !*** f interpolation ***
            call interpol1d_theta(geom%Ntheta,geom%dtheta, &
              f%hhspline2d_rtheta(tid)%bcoef, &
              f%hhspline2d_rtheta(tid)%ygrid,thetan,finterpol)

            !**************************************************
            ! primitive interpolation, total mass "sum" 
            !  is added or substracted if particule
            !  goes out of the domain trajectory wise
            !**************************************************
            finterpol            = finterpol + out*sum
            fluxtheta(ir,itheta) = rhs(itheta)-finterpol

            ! *** SLS limiter ***
            if (limiter==2) then
              iright = itheta
              ileft  = iright-1
              if (ileft<0) ileft = ileft+Ntheta
              dpc = f4d_transp(ir,iright,iphi,ivpar) - &
                f4d_transp(ir,ileft ,iphi,ivpar)
              if (abs(dpc)>1.e-10_RKIND) then
                if (udttheta(ir,itheta) > 0._RKIND ) then
                  iup = ileft-1
                  if (iup<0) iup = iup + Ntheta
                  dp       = f4d_transp(ir,ileft,iphi,ivpar) - &
                    f4d_transp(ir,iup,iphi,ivpar)
                  flux_upw = udttheta(ir,itheta) * &
                    f4d_transp(ir,ileft,iphi,ivpar)
                else
                  iup = iright+1
                  if (iup>Ntheta-1) iup = iup - Ntheta
                  dp       = f4d_transp(ir,iup,iphi,ivpar) - &
                    f4d_transp(ir,iright,iphi,ivpar)
                  flux_upw = udttheta(ir,itheta) * &
                    f4d_transp(ir,iright,iphi,ivpar)
                endif
                lim = max(0._RKIND, &
                  min( Klim*abs(dp/dpc),1._RKIND))
                fluxtheta(ir,itheta) = lim*fluxtheta(ir,itheta) + &
                  (1._RKIND-lim)*flux_upw
              endif
            endif !sls
          end do !0,Ntheta-1
          fluxtheta(ir,Ntheta) = fluxtheta(ir,0)
          !=======================================================================

          do itheta = 0,Ntheta-1
            f4d_transp(ir,itheta,iphi,ivpar) =  &
              f4d_transp(ir,itheta,iphi,ivpar) - &
              ( f%hhspline2d_rtheta(tid)%xgrid(ir+1)*fluxr(ir+1,itheta) - &
              f%hhspline2d_rtheta(tid)%xgrid(ir)*fluxr(ir,itheta) )/ &
              (geom%rg(ir)*geom%dr) - &
              (fluxtheta(ir,itheta+1)-fluxtheta(ir,itheta))/geom%dtheta
          end do
        end do  ! ir = 0,Nr
      end do
    end do
!$OMP END DO
!$OMP BARRIER
!$OMP END PARALLEL

    maxalp_r     = maxval(Romp_maxalp_r(1:Nbthread))/(geom%dr*dt)
    maxalp_theta = maxval(Romp_maxalp_th(1:Nbthread))/(geom%dtheta*dt)

    do tid = 1, Nbthread
       call FTR_DEALLOC(f%hhspline2d_rtheta(tid)%coef)
       call FTR_DEALLOC(f%hhspline2d_rtheta(tid)%bcoef)
       call FTR_DEALLOC(f%hhspline2d_rtheta(tid)%aux)
       call FTR_DEALLOC(f%hhspline2d_rtheta(tid)%auy)

       call FTR_DEALLOC(Romp1_0Nr_0Ntheta(tid)%val)
       call FTR_DEALLOC(Romp2_0Nr_0Ntheta(tid)%val)
    enddo
    
    call clck_time(eclock_advec1Dr)
    call clck_diff(bclock_advec1Dr,eclock_advec1Dr, &
      global_time_advec1Dr)
    call Write_dyn_balise_close('CPU_advec1d_PSM_r') !R4
!R3 call r3_info_end (r3_info_index_0) !R3
#else 
    ! HAS_ADV_PSM
    maxalp_theta = 0._RKIND
    maxalp_r     = 0._RKIND
#endif 
    ! HAS_ADV_PSM
  end subroutine advec2D_PSM
end module advec2D_PSM_module
