!**************************************************************
!  Copyright Euratom-CEA
!  Authors : 
!     Virginie Grandgirard (virginie.grandgirard@cea.fr)
!     Chantal Passeron (chantal.passeron@cea.fr)
!     Guillaume Latu (guillaume.latu@cea.fr)
!     Xavier Garbet (xavier.garbet@cea.fr)
!     Philippe Ghendrih (philippe.ghendrih@cea.fr)
!     Yanick Sarazin (yanick.sarazin@cea.fr)
!  
!  This code GYSELA (for GYrokinetic SEmi-LAgrangian) 
!  is a 5D gyrokinetic global full-f code for simulating 
!  the plasma turbulence in a tokamak.
!  
!  This software is governed by the CeCILL-B license 
!  under French law and abiding by the rules of distribution 
!  of free software.  You can  use, modify and redistribute 
!  the software under the terms of the CeCILL-B license as 
!  circulated by CEA, CNRS and INRIA at the following URL
!  "http://www.cecill.info". 
!**************************************************************

#include "config-F.inc"

!---------------------------------------------------------
! file : advec_conservative.F90
! date : 18/10/2009
!  Used to choose between the different available
!  conservative advections in r and theta directions, i.e
!   - 1D in r + 1D in theta with PPM2
!   - 2D in (r,theta) with divergence-free PPM2
!   - 1D  in r + 1D in theta with PSM 
!   - 2D in (r,theta) with divergence-free PSM
!  PSM being the Parabolic SPline Method and
!  PPM2 Colella-Sekora JCP 2008
!  Jean-Philippe Braeunig (braeunig@math.u-strasbg.fr)
!---------------------------------------------------------
module advec_conserv_module
  use globals, only : Vlasov_scheme
  use advec1D_PSM_module
  use advec2D_PSM_module
  use advec1D_PPM2_module
  use advec2D_PPM2_module

  implicit none

  !******************************
  contains
  !******************************

  !-----------------------------------------------------
  ! 2D advection solving (in cartesian coordinates)
  !  - compute the characteristic feet and 
  !  - interpole the distribution function at this point
  ! (using the explicit taylor development)
  ! -> developped by G. Latu for local splines
  !----------------------------------------------------- 
  subroutine advec2D_rtheta_conservative(geom,dt,nbsubit, &
    coord_sys,init_prof,init_magnet,init_curr,f,Sfmu_eq, &
    activatefilter,maxalp_rtheta)
    use globals
    use MPIutils_module
    use fdistribu5d_class
    use init_profile_class
    use init_magnetic_class
    use utils_module
    use clock_module
    use filter_module
!R3 #include "code_instrumentation/r3_info.h" !R3
    type(geometry)     , intent(in)    :: geom
    real(RKIND)        , intent(in)    :: dt
    integer            , intent(in)    :: nbsubit
    type(coord_system) , intent(in)    :: coord_sys
    type(init_profile) , intent(in)    :: init_prof
    type(init_magnetic), intent(in)    :: init_magnet
    type(init_current) , intent(in)    :: init_curr
    type(fdistribu5d)  , intent(inout) :: f
    real(RKIND), &
      dimension(:,:,:) , pointer       :: Sfmu_eq
    logical            , intent(in)    :: activatefilter
    real(RKIND)        , intent(out)   :: maxalp_rtheta

    integer     :: ir, itheta, iphi, ivpar, zz, k, q
    integer     :: iter_save
    real(RKIND) :: rfeet, tfeet, xfeet, yfeet
    real(RKIND) :: finterpol, outofdom_percent
    real(RKIND) :: maxalp_r, maxalp_theta 

    integer           :: tid
    integer           :: subit
    integer(TIMEPREC) :: tdeb, tfin
    real(RKIND)       :: tcomm2, tadvect

    tcomm2  = 0._RKIND
    tadvect = 0._RKIND

    if (Vlasov_scheme=="PPM2D") then ! PPM2
      !*** PPM2 conservative method in 2D ***
      !*** in (r,theta) cross-section     ***
      !--> divergence-free scheme
      do subit = 1, nbsubit
        call clck_time(tdeb)
        call advec2D_PPM2(geom,dt,coord_sys,init_prof, &
          init_magnet,init_curr,f,maxalp_r,maxalp_theta)
        call clck_time(tfin)
        call clck_diff(tdeb,tfin,tadvect)
      end do
    elseif (Vlasov_scheme=="PPM1D") then ! PPM2
      !*** PPM2 conservative method splitted in 1D ***
      !*** in r and theta directions               ***
      do subit = 1, nbsubit
        call clck_time(tdeb)
        call advec1D_PPM2_r(geom,0.5_RKIND*dt,init_prof, &
          init_magnet,init_curr,f,maxalp_r)
        call advec1D_PPM2_theta(geom,dt,init_prof, &
          init_magnet,init_curr,f,maxalp_theta)
        call advec1D_PPM2_r(geom,0.5_RKIND*dt,init_prof, &
          init_magnet,init_curr,f,maxalp_r)
        call clck_time(tfin)
        call clck_diff(tdeb,tfin,tadvect)
      end do
    elseif  (Vlasov_scheme=="PSM1D") then ! PSM
      !*** PSM conservative method splitted in 1D in r and theta ***
      do subit = 1, nbsubit
        call clck_time(tdeb)
        call advec1D_PSM_r(geom,0.5_RKIND*dt,&
          init_magnet,init_curr,f,maxalp_r)
        call advec1D_PSM_theta(geom,dt, &
          init_magnet,init_curr,f,maxalp_theta)
        call advec1D_PSM_r(geom,0.5_RKIND*dt, &
          init_magnet,init_curr,f,maxalp_r)
        call clck_time(tfin)
        call clck_diff(tdeb,tfin,tadvect)
      end do
    elseif  (Vlasov_scheme=="PSM2D") then ! PSM
      !*** PSM conservative method splitted in 2D (r,theta) ***
      !--> divergence-free scheme
      do subit = 1, nbsubit
        call clck_time(tdeb)
        call advec2D_PSM(geom,dt,init_prof, &
          init_magnet,init_curr,f,maxalp_r,maxalp_theta)
        call clck_time(tfin)
        call clck_diff(tdeb,tfin,tadvect)
      end do
    end if 
    maxalp_rtheta = max(maxalp_theta,maxalp_r)

    if (activatefilter) call hffilter_loc(Sfmu_eq)

#ifdef TIMER    
    if (pglobal_id .eq. 0) then
      write(*,'(I4,A,F13.7,A,F13.7,A,F13.7,A,F13.7,A,F13.7)') &
        pglobal_id, " Temps advection2D,  comm2 ",tcomm2, &
        " advec2D ",tadvect
    end if
#endif

    !*** counter for particles out of domain ***
    if (plocal_id.eq.0) then
      iter_save                      = max(iter_run,0)
      nbpart_rleft_muid(iter_save)   = nbpart_rleft_iter(mu_id)
      nbpart_rright_muid(iter_save)  = nbpart_rright_iter(mu_id)
      if ((nbpart_rleft_iter(mu_id) + &
        nbpart_rright_iter(mu_id)).ne.0) then
        outofdom_percent = ((nbpart_rright_iter(mu_id) + &
          nbpart_rleft_iter(mu_id)) / &
          float(Nbpoints_muid))*100._RKIND
        write(6,'(A,I4,A,f6.2,A)') &
          '      -> WARNING : particles out ', &
          'of domain in r for mu_id     (',mu_id,') = ', &
          outofdom_percent,' %'
      end if
    end if
  end subroutine advec2D_rtheta_conservative
end module advec_conserv_module
