!**************************************************************
!  Copyright Euratom-CEA
!  Authors : 
!     Virginie Grandgirard (virginie.grandgirard@cea.fr)
!     Chantal Passeron (chantal.passeron@cea.fr)
!     Guillaume Latu (guillaume.latu@cea.fr)
!     Xavier Garbet (xavier.garbet@cea.fr)
!     Philippe Ghendrih (philippe.ghendrih@cea.fr)
!     Yanick Sarazin (yanick.sarazin@cea.fr)
!  
!  This code GYSELA (for GYrokinetic SEmi-LAgrangian) 
!  is a 5D gyrokinetic global full-f code for simulating 
!  the plasma turbulence in a tokamak.
!  
!  This software is governed by the CeCILL-B license 
!  under French law and abiding by the rules of distribution 
!  of free software.  You can  use, modify and redistribute 
!  the software under the terms of the CeCILL-B license as 
!  circulated by CEA, CNRS and INRIA at the following URL
!  "http://www.cecill.info". 
!**************************************************************

#include "config-F.inc"
#include "code_instrumentation/debug.inc"

!-------------------------------------------------
! file : boltzmann_solving.F90
! date :  16/09/2005
!  Global algorithm for solving the Boltzmann 
!  system, i.e Vlasov + RHS where the RHS included
!    - Diffusion & Krook operators
!    - collision operator
!    - source terms
!  which is divided into :
!   - time-splitting algorithm and 
!   - leap-frog or predictor-corrector algorithm
!-------------------------------------------------
module Boltz_solving_module
  use globals
  use Debug_module
  use prec_const
  use geometry_class
  use init_profile_class
  use init_magnetic_class
  use init_current_class
  use sources_module
  use fdistribu5d_class
  use poisson_class
  use diag3D_module
  use diagnostics_module
  use sources_module
  use vlasov_solving_module

  implicit none
      
  !******************************
  contains
  !******************************
  
  !-------------------------------------------------------------
  ! TIME-SPLITTING between t and t+dt for the
  !  for treating only the RHS of the equation, including
  !  treatment of Krook & Diffusion operators, collision
  !  operator and source term i.e :
  !     coll/2,krook/2,Diffusion/2,source,
  !     Diffusion/2,krook/2,coll/2
  !------------------------------------------------------------
  subroutine timesplit_RHSonly(geom,init_prof,init_magnet, &
    init_curr,init_srce,dt,fn,Sfmu_eq,activatefilter)
    use globals, only : transp_velocity, advec2D_deltat, &
      coef_diffr, coef_diffth
    use globals, only : atleast1_source 
    use MPIutils_module
    use globals, only : coef_krook
    use krook_module
    use diffusion_module
    use collisions_vpar_module, only : AdvecDiff_coll_vpar
    use species_transf_module , only : advance_species_transfer
    use sources_module
    use mpiMod

    type(geometry)     , intent(in)    :: geom
    type(init_profile) , intent(in)    :: init_prof
    type(init_magnetic), intent(in)    :: init_magnet
    type(init_current) , intent(in)    :: init_curr
    type(init_sources) , intent(in)    :: init_srce
    real(RKIND)        , intent(in)    :: dt
    type(fdistribu5d)  , intent(inout) :: fn
    real(RKIND), &
      dimension(:,:,:) , pointer       :: Sfmu_eq
    logical            , intent(in)    :: activatefilter
    integer                            :: ierr

    !*** step 1 : collisions in the v parallel ***
    !***  direction (on 0.5*dt)                ***
    if (collisions) then
      if (mod(iter_glob,coll_nbstep).eq.0) then
        call AdvecDiff_coll_vpar(geom, &
          coll_nbstep*0.5_RKIND*dt, &
          init_prof,init_magnet,init_curr,fn)
        if ( (energy_transfer).or.(moment_transfer)) &
          call advance_species_transfer(geom,init_prof,init_magnet,0.5*dt,fn)
      end if
    end if
    !*** step 2 : Krook & Diffusion operators (on 0.5*dt) ***
    if (coef_krook.ne.0._RKIND) &
      call advance_Krook(Sfmu_eq,0.5_RKIND*dt,fn)
    if (coef_diffr(pSpecies_id).ne.0._RKIND) &
      call advance_Diff_r(Sfmu_eq,0.5_RKIND*dt,fn)
    !*** step 3 : add source (on 0.5*dt) ***
    if (atleast1_source) &
      call advance_sources(init_srce, 0.5_RKIND*dt,geom,init_prof, &
      init_magnet,init_curr,fn)
    !*** step 4 : Diffusion in (theta) (on dt) ***
    if (transpose4D .and. (coef_diffth(pSpecies_id).ne.0._RKIND)) then
      !*** Transpose the 4D function from f%valuesues to f4D_transp ***

      select case ( transpose_version )
      case ( TRANSPOSE_SENDIRECV )
         call FTR_ALLOC_4_ZERO(f4D_send,0,dom_r-1,0,dom_theta-1,0,dom_phi-1,0,dom_vpar-1)
         call FTR_ALLOC_4_ZERO(f4D_recv,0,dom_r-1,0,dom_theta-1,0,dom_phi-1,0,dom_vpar-1)
      case ( TRANSPOSE_ALLTOALLW )
         ! nothing specific to allocate
      case default
         print *, pglobal_id, 'Invalid value for transpose_version', &
              transpose_version
         call MPI_abort(MPI_COMM_WORLD, 1, ierr)
      end select

      call pptranspose_forward(fn%values,timing=.true.)
      call advance_Diff_th(Sfmu_eq,dt,f4D_transp)
      !*** Transpose the 4D function from f4D_transp to f%valuesues ***
      call pptranspose_backward(fn%values,timing=.true.)

      select case ( transpose_version )
      case ( TRANSPOSE_SENDIRECV )
         call FTR_DEALLOC(f4D_send)
         call FTR_DEALLOC(f4D_recv)
      case ( TRANSPOSE_ALLTOALLW )
         ! nothing specific to allocate
      case default
         print *, pglobal_id, 'Invalid value for transpose_version', &
              transpose_version
         call MPI_abort(MPI_COMM_WORLD, 1, ierr)
      end select
    end if
    !*** step 5 : add source (on 0.5*dt) ***
    if (atleast1_source) &
      call advance_sources(init_srce,0.5_RKIND*dt,geom,init_prof, &
      init_magnet,init_curr,fn)
    !*** step 6 : Diffusion & Krook operators (on 0.5*dt) ***
    if (coef_diffr(pSpecies_id).ne.0._RKIND) &
      call advance_Diff_r(Sfmu_eq,0.5_RKIND*dt,fn)
    if (coef_krook.ne.0._RKIND) &
      call advance_Krook(Sfmu_eq,0.5_RKIND*dt,fn)
    !*** step 7 : collisions in the v parallel ***
    !***  direction (on 0.5*dt)                ***
    if (collisions) then
      if (mod(iter_glob,coll_nbstep).eq.0) then
        if ( (energy_transfer).or.(moment_transfer) ) &
          call advance_species_transfer(geom,init_prof,init_magnet,0.5*dt,fn)
        call AdvecDiff_coll_vpar(geom, &
          coll_nbstep*0.5_RKIND*dt, &
          init_prof,init_magnet,init_curr,fn)
      end if
    end if
  end subroutine timesplit_RHSonly


  !-------------------------------------------------------------
  ! TIME-SPLITTING between t and t+dt for the Boltzmann
  !  equation, i.e : Vlasov + RHS terms 
  !  where the RHS terms can be :
  !   - Krook & Diffusion operators
  !   - collision operator
  !   - source terms
  ! The global shift sequence is the following:
  !     vpar/2,phi/2,coll/2,krook/2,Diffusion/2,(r,theta),
  !     coll/2,Diffusion/2,krook/2,phi/2,vpar/2 
  !
  ! Case of SEMI-LAGRANGIAN method IN ALL 
  !   DIRECTIONS (r,theta,phi,vpar)
  !------------------------------------------------------------
  subroutine timesplit_BSL(geom,init_prof,init_magnet, &
    init_curr,init_srce,dt,fn,Sfmu_eq,activatefilter, &
    onlynl,onlylin)
    use globals, only : transp_velocity, advec2D_deltat, &
      coef_diffr, coef_diffth
    use globals, only : atleast1_source 
    use globals, only : coef_krook
    use krook_module
    use MPIutils_module
    use advec1D_BSL_module
    use advec2D_BSL_module
    use diffusion_module
    use collisions_vpar_module, only : AdvecDiff_coll_vpar
    use species_transf_module , only : advance_species_transfer
    use sources_module
    use filter_module
    use mpiMod

    type(geometry)     , intent(in)    :: geom
    type(init_profile) , intent(in)    :: init_prof
    type(init_magnetic), intent(in)    :: init_magnet
    type(init_current) , intent(in)    :: init_curr
    type(init_sources) , intent(in)    :: init_srce 
    real(RKIND)        , intent(in)    :: dt
    type(fdistribu5d)  , intent(inout) :: fn
    real(RKIND), &
      dimension(:,:,:) , pointer       :: Sfmu_eq
    logical            , intent(in)    :: activatefilter
    logical ,optional  , intent(in)    :: onlynl
    logical ,optional  , intent(in)    :: onlylin
    
    integer     :: ivpar, nbsubit
    real(RKIND) :: subdt
    logical     :: l_onlynl, l_onlylin
    integer     :: ierr
    
    PRINT_CKSUM(fn%values)

    if (advec2D_deltat .ne. 0._RKIND) then
       nbsubit = MAX(1, nint(dt / advec2D_deltat))
       subdt   = dt / nbsubit
    else
       nbsubit = 1
       subdt   = dt
    end if

    l_onlynl  = .false.
    l_onlylin = .false.
    if (split_linear) then
      if (present(onlynl))  l_onlynl  = onlynl
      if (present(onlylin)) l_onlylin = onlylin
    end if
    
    PRINT_CKSUM(fn%values)

    if ((.not. split_linear) .or. &
      (split_linear .and. .not. l_onlylin)) then
      !*** step 1 : splitting in v parallel ***
      !***  direction (on 0.5*dt)           ***
      call advec1D_BSL_vpar(geom,0.5_RKIND*dt, &
        init_magnet,init_curr,fn,onlynl=split_linear)
    
      PRINT_CKSUM(fn%values)
      
      !*** step 2 : collisions in the v parallel ***
      !**   direction (on 0.5*dt)                ***
      if (collisions) then
        if (mod(iter_glob,coll_nbstep).eq.0) then
          call AdvecDiff_coll_vpar(geom, &
            coll_nbstep*0.5_RKIND*dt, &
            init_prof,init_magnet,init_curr,fn)
          
          PRINT_CKSUM(fn%values)
          
          if ( (energy_transfer).or.(moment_transfer) ) &
            call advance_species_transfer(geom, &
            init_prof,init_magnet,0.5*dt,fn)
        end if
      end if
    
      PRINT_CKSUM(fn%values)
      
      !*** step 3 : splitting in phi direction (on 0.5*dt) ***
      call advec1D_BSL_phi(geom,0.5_RKIND*dt, &
        init_magnet,init_curr,fn,onlynl=split_linear)
    
      PRINT_CKSUM(fn%values)
      
      !*** step 4 : Krook & Diffusion operators (on 0.5*dt) ***
      if (coef_krook.ne.0._RKIND) then
        call advance_Krook(Sfmu_eq,0.5_RKIND*dt,fn)
      endif
      if (coef_diffr(pSpecies_id).ne.0._RKIND) then
        call advance_Diff_r(Sfmu_eq,0.5_RKIND*dt,fn)
      endif
    
      PRINT_CKSUM(fn%values)
      
      !*** step 5 : add source (on 0.5*dt) ***
      if (atleast1_source) then
        call advance_sources(init_srce,0.5_RKIND*dt,geom, &
        init_prof,init_magnet,init_curr,fn)
      endif
    
      PRINT_CKSUM(fn%values)
      
      !*** Transpose back and forth distribution function ***
      !*** coll5D_transp data structure stores            ***
      !***   (vpar=*,mu=*) slices                         ***
      if (transp_velocity) then
        call pptranspvel_forward(fn%values)   
        call pptranspvel_backward(fn%values)   
      end if
      !*** step 6 : splitting in (r,theta) (on dt) ***
      !***          and Diffusion in theta         ***
      if (transpose4D) then
        !*** Transpose the 4D function from f%values ***
        !***  to f4D_transp                          ***

        select case ( transpose_version )
        case ( TRANSPOSE_SENDIRECV )
           call FTR_ALLOC_4_ZERO(f4D_send,0,dom_r-1,0,dom_theta-1,0,dom_phi-1,0,dom_vpar-1)
           call FTR_ALLOC_4_ZERO(f4D_recv,0,dom_r-1,0,dom_theta-1,0,dom_phi-1,0,dom_vpar-1)
        case ( TRANSPOSE_ALLTOALLW )
           ! nothing specific to allocate
        case default
           print *, pglobal_id, 'Invalid value for transpose_version', &
                transpose_version
           call MPI_abort(MPI_COMM_WORLD, 1, ierr)
        end select

        call pptranspose_forward(fn%values,timing=.true.)
        call FTR_DEALLOC(fn%values)
      
        PRINT_CKSUM(f4D_transp)
        
        if (coef_diffth(pSpecies_id).ne.0._RKIND) &
          call advance_Diff_th(Sfmu_eq,0.5_RKIND*dt,f4D_transp)   
      
        PRINT_CKSUM(f4D_transp)
         
        if (avoidtaylor) then
          call avoidtaylor_advec2D(geom,subdt, &
            nbsubit,init_prof,init_magnet,init_curr, &
            fn,Sfmu_eq,activatefilter,onlynl=split_linear)
        else
          call transpose_advec2D_xy(geom,subdt, &
            nbsubit,init_prof,init_magnet,init_curr, &
            fn,Sfmu_eq,activatefilter)
        end if        
      
        PRINT_CKSUM(f4D_transp)
          
        if (coef_diffth(pSpecies_id).ne.0._RKIND) then
           call advance_Diff_th(Sfmu_eq,0.5_RKIND*dt,f4D_transp)
        end if

        PRINT_CKSUM(f4D_transp)

        !*** Transpose the 4D function from f4D_transp ***
        !***   to f%values                             ***
        call FTR_ALLOC_4(fn%values,fn%istart_buf,fn%iend_buf,fn%jstart_buf,fn%jend_buf,0,fn%n3,0,fn%n4)
        call pptranspose_backward(fn%values,timing=.true.)

        select case ( transpose_version )
        case ( TRANSPOSE_SENDIRECV )
           call FTR_DEALLOC(f4D_send)
           call FTR_DEALLOC(f4D_recv)
        case ( TRANSPOSE_ALLTOALLW )
           ! nothing specific to allocate
        case default
           print *, pglobal_id, 'Invalid value for transpose_version', &
                transpose_version
           call MPI_abort(MPI_COMM_WORLD, 1, ierr)
        end select
      else
        call local_advec2D_xy(geom,dt, &
          init_prof,init_magnet,init_curr,fn)
      end if
    
      PRINT_CKSUM(fn%values)
      
      call check_overflow_4D(fn,"After advec2D")
      !*** Transpose back and forth distribution function ***
      !*** coll5D_transp data structure stores            ***
      !***   (vpar=*,mu=*) slices                         ***
      if (transp_velocity) then
        call pptranspvel_forward(fn%values)   
        call pptranspvel_backward(fn%values)   
      end if
    
      PRINT_CKSUM(fn%values)
      
      !*** step 7 : add source (on 0.5*dt) ***
      if (atleast1_source) &
        call advance_sources(init_srce,0.5_RKIND*dt,geom, &
        init_prof,init_magnet,init_curr,fn)
    
      PRINT_CKSUM(fn%values)
      
      !*** step 8 : Diffusion & Krook operators (on 0.5*dt) ***
      if (coef_diffr(pSpecies_id).ne.0._RKIND) &
        call advance_Diff_r(Sfmu_eq,0.5_RKIND*dt,fn)
      if (coef_krook.ne.0._RKIND) &
        call advance_Krook(Sfmu_eq,0.5_RKIND*dt,fn)
    
      PRINT_CKSUM(fn%values)
      
      !*** step 9 : splitting in phi direction (on 0.5*dt) ***
      call advec1D_BSL_phi(geom,0.5_RKIND*dt, &
        init_magnet,init_curr,fn,onlynl=split_linear)
    
      PRINT_CKSUM(fn%values)
      
      !*** step 10 : collisions in the v parallel ***
      !***  direction (on 0.5*dt)                 ***
      if (collisions) then
        if (mod(iter_glob,coll_nbstep).eq.0) then
          if ( (energy_transfer).or.(moment_transfer) ) &
            call advance_species_transfer(geom, &
            init_prof,init_magnet,0.5*dt,fn)
          call AdvecDiff_coll_vpar(geom, &
            coll_nbstep*0.5_RKIND*dt, &
            init_prof,init_magnet,init_curr,fn)
        end if
      end if
    
      PRINT_CKSUM(fn%values)
      
      !*** step 11 : splitting in v parallel ***
      !***  direction (on 0.5*dt)            ***
      call advec1D_BSL_vpar(geom,0.5_RKIND*dt, &
        init_magnet,init_curr,fn,onlynl=split_linear)
    
      PRINT_CKSUM(fn%values)
      
    end if

    !!*** perform the LINEAR part alone if split_linear=.true. ***
    if (split_linear .and. .not. l_onlynl) then
      if ((geom%Nmu .ne. 0) .or. (mumin.ne.0._RKIND)) then
        call advec1D_BSL_vpar(geom,0.5_RKIND*dt, &
          init_magnet,init_curr,fn,onlylin=.true.)
      end if
    
      PRINT_CKSUM(fn%values)
      
      call advec1D_BSL_phi(geom,0.5_RKIND*dt, &
        init_magnet,init_curr,fn,onlylin=.true.)
    
      PRINT_CKSUM(fn%values)
      
      if (transpose4D .and. avoidtaylor) then
        !*** Transpose the 4D function from f%values ***
        !***   to f4D_transp                         ***

         select case ( transpose_version )
         case ( TRANSPOSE_SENDIRECV )
            call FTR_ALLOC_4_ZERO(f4D_send,0,dom_r-1,0,dom_theta-1,0,dom_phi-1,0,dom_vpar-1)
            call FTR_ALLOC_4_ZERO(f4D_recv,0,dom_r-1,0,dom_theta-1,0,dom_phi-1,0,dom_vpar-1)
         case ( TRANSPOSE_ALLTOALLW )
            ! nothing specific to allocate
         case default
            print *, pglobal_id, 'Invalid value for transpose_version', &
                 transpose_version
            call MPI_abort(MPI_COMM_WORLD, 1, ierr)
         end select

        call pptranspose_forward(fn%values,timing=.true.)           
        call avoidtaylor_advec2D(geom,dt, &
          1,init_prof,init_magnet,init_curr, &
          fn,Sfmu_eq,.false.,onlylin=.true.)
        !*** Transpose the 4D function from f4D_transp ***
        !***   to f%values                             ***
        call pptranspose_backward(fn%values,timing=.true.)

        select case ( transpose_version )
        case ( TRANSPOSE_SENDIRECV )
           call FTR_DEALLOC(f4D_send)
           call FTR_DEALLOC(f4D_recv)
        case ( TRANSPOSE_ALLTOALLW )
           ! nothing specific to allocate
        case default
           print *, pglobal_id, 'Invalid value for transpose_version', &
                transpose_version
           call MPI_abort(MPI_COMM_WORLD, 1, ierr)
        end select
      else
        print *, "split_linear option requires ", &
          "(transpose4D .and. avoidtaylor)"
        call ppexit()
      end if
    
      PRINT_CKSUM(fn%values)
      
      call advec1D_BSL_phi(geom,0.5_RKIND*dt, &
        init_magnet,init_curr,fn,onlylin=.true.)
    
      PRINT_CKSUM(fn%values)
      
      if ((geom%Nmu .ne. 0) .or. (mumin.ne.0._RKIND)) then
        call advec1D_BSL_vpar(geom,0.5_RKIND*dt, &
          init_magnet,init_curr,fn,onlylin=.true.)
      end if
    
      PRINT_CKSUM(fn%values)
      
    end if
    
    PRINT_CKSUM(fn%values)

  end subroutine timesplit_BSL


  !-------------------------------------------------------------
  ! TIME-SPLITTING between t and t+dt for the Boltzmann
  !  equation, i.e : Vlasov + RHS terms 
  !  where the RHS terms can be :
  !   - Krook & Diffusion operators
  !   - collision operator
  !   - source terms
  ! The global shift sequence is the following:
  !     vpar/2,phi/2,coll/2,krook/2,Diffusion/2,(r,theta),
  !     coll/2,Diffusion/2,krook/2,phi/2,vpar/2 
  !
  ! Case of Parabolic Spline Method (PSM) IN ALL 
  !   DIRECTIONS (r,theta,phi,vpar)
  !------------------------------------------------------------
  subroutine timesplit_conserv(geom, &
    coord_sys,init_prof,init_magnet, &
    init_curr,init_srce,dt,fn,Sfmu_eq,activatefilter, &
    maxalprth,maxalpphi,maxalpvpar)
    use globals, only : transp_velocity, advec2D_deltat, &
      coef_diffr, coef_diffth
    use globals, only : atleast1_source
    use globals, only : coef_krook
    use krook_module
    use MPIutils_module
    use advec1D_PSM_module
    use advec_conserv_module
    use diffusion_module
    use collisions_vpar_module, only : AdvecDiff_coll_vpar
    use species_transf_module , only : advance_species_transfer
    use sources_module
    use filter_module
    use mpiMod

    type(geometry)       , intent(in)    :: geom
    type(coord_system)   , intent(in)    :: coord_sys
    type(init_profile)   , intent(in)    :: init_prof
    type(init_magnetic)  , intent(in)    :: init_magnet
    type(init_current)   , intent(in)    :: init_curr
    type(init_sources)   , intent(in)    :: init_srce
    real(RKIND)          , intent(in)    :: dt
    type(fdistribu5d)    , intent(inout) :: fn
    real(RKIND), &       
      dimension(:,:,:)   , pointer       :: Sfmu_eq
    logical              , intent(in)    :: activatefilter
    real(RKIND)          , intent(out)   :: maxalprth
    real(RKIND)          , intent(out)   :: maxalpphi
    real(RKIND)          , intent(out)   :: maxalpvpar
    
    integer     :: ivpar, nbsubit
    real(RKIND) :: subdt
    integer     :: ierr

    if (advec2D_deltat .ne. 0._RKIND) then
       nbsubit = MAX(1, nint(dt / advec2D_deltat))
       subdt   = dt / nbsubit
    else
       nbsubit = 1
       subdt   = dt
    end if

    !*** step 1 : splitting in v parallel ***
    !***  direction (on 0.5*dt)           ***
    call advec1D_PSM_vpar(geom,0.5_RKIND*dt, &
      init_prof,init_magnet,init_curr,fn,maxalpvpar)
    !*** step 2 : collisions in the v parallel ***
    !**   direction (on 0.5*dt)                ***
    if (collisions) then
      if (mod(iter_glob,coll_nbstep).eq.0) then
        call AdvecDiff_coll_vpar(geom, &
          coll_nbstep*0.5_RKIND*dt, &
          init_prof,init_magnet,init_curr,fn)
        if ( (energy_transfer).or.(moment_transfer) ) &
          call advance_species_transfer(geom, &
          init_prof,init_magnet,0.5*dt,fn)
      end if
    end if
    !*** step 3 : splitting in phi direction (on 0.5*dt) ***
    call advec1D_PSM_phi(geom,0.5_RKIND*dt, &
      init_magnet,init_curr,fn,maxalpphi)
    !*** step 4 : Krook & Diffusion operators (on 0.5*dt) ***
    if (coef_krook.ne.0._RKIND) &
      call advance_Krook(Sfmu_eq,0.5_RKIND*dt,fn)
    if (coef_diffr(pSpecies_id).ne.0._RKIND) &  
      call advance_Diff_r(Sfmu_eq,0.5_RKIND*dt,fn)
    !*** step 5 : add source (on 0.5*dt) ***
    if (atleast1_source) &
      call advance_sources(init_srce,0.5_RKIND*dt,geom, &
      init_prof,init_magnet,init_curr,fn)
    !*** Transpose back and forth distribution function ***
    !*** coll5D_transp data structure stores            ***
    !***   (vpar=*,mu=*) slices                         ***
    if (transp_velocity) then
      call pptranspvel_forward(fn%values)
      call pptranspvel_backward(fn%values)
    end if
    !*** step 6 : splitting in (r,theta) (on dt) ***

    select case ( transpose_version )
    case ( TRANSPOSE_SENDIRECV )
       call FTR_ALLOC_4_ZERO(f4D_send,0,dom_r-1,0,dom_theta-1,0,dom_phi-1,0,dom_vpar-1)
       call FTR_ALLOC_4_ZERO(f4D_recv,0,dom_r-1,0,dom_theta-1,0,dom_phi-1,0,dom_vpar-1)
    case ( TRANSPOSE_ALLTOALLW )
       ! nothing specific to allocate
    case default
       print *, pglobal_id, 'Invalid value for transpose_version', &
            transpose_version
       call MPI_abort(MPI_COMM_WORLD, 1, ierr)
    end select

    call pptranspose_forward(fn%values,timing=.true.)
    if (coef_diffth(pSpecies_id).ne.0._RKIND) then
      call advance_Diff_th(Sfmu_eq,0.5_RKIND*dt,f4D_transp)
    endif
    call advec2D_rtheta_conservative(geom,subdt, &
      nbsubit,coord_sys,init_prof,init_magnet,init_curr, &
      fn,Sfmu_eq,activatefilter,maxalprth)
    if (coef_diffth(pSpecies_id).ne.0._RKIND) &
      call advance_Diff_th(Sfmu_eq,0.5_RKIND*dt,f4D_transp)
    call pptranspose_backward(fn%values,timing=.true.)   

    select case ( transpose_version )
    case ( TRANSPOSE_SENDIRECV )
       call FTR_DEALLOC(f4D_send)
       call FTR_DEALLOC(f4D_recv)
    case ( TRANSPOSE_ALLTOALLW )
       ! nothing specific to allocate
    case default
       print *, pglobal_id, 'Invalid value for transpose_version', &
            transpose_version
       call MPI_abort(MPI_COMM_WORLD, 1, ierr)
    end select
    !*** Transpose back and forth distribution function ***
    !*** coll5D_transp data structure stores            ***
    !***   (vpar=*,mu=*) slices                         ***
    if (transp_velocity) then
      call pptranspvel_forward(fn%values)   
      call pptranspvel_backward(fn%values)   
    end if
    !*** step 7 : add source (on 0.5*dt) ***
    if (atleast1_source) &
      call advance_sources(init_srce,0.5_RKIND*dt,geom, &
      init_prof,init_magnet,init_curr,fn)
    !*** step 8 : Diffusion & Krook operators (on 0.5*dt) ***
    if (coef_diffr(pSpecies_id).ne.0._RKIND) &
      call advance_Diff_r(Sfmu_eq,0.5_RKIND*dt,fn)
    if (coef_krook.ne.0._RKIND) &
      call advance_Krook(Sfmu_eq,0.5_RKIND*dt,fn)
    !*** step 9 : splitting in phi direction (on 0.5*dt) ***
    call advec1D_PSM_phi(geom,0.5_RKIND*dt, &
      init_magnet,init_curr,fn,maxalpphi)
    !*** step 10 : collisions in the v parallel ***
    !***  direction (on 0.5*dt)                 ***
    if (collisions) then
      if (mod(iter_glob,coll_nbstep).eq.0) then
        if ( (energy_transfer).or.(moment_transfer) ) then
          call advance_species_transfer(geom, &
               init_prof,init_magnet,0.5*dt,fn)
        endif
        call AdvecDiff_coll_vpar(geom, &
          coll_nbstep*0.5_RKIND*dt, &
          init_prof,init_magnet,init_curr,fn)
      end if
    end if
    !*** step 11 : splitting in v parallel ***
    !***  direction (on 0.5*dt)            ***
    call advec1D_PSM_vpar(geom,0.5_RKIND*dt, &
      init_prof,init_magnet,init_curr,fn,maxalpvpar)

    !-> maximum of the normalised displacments in phi and vpar
    !-> (multiplied by 2, because advection performed on 1/2 dt)
    maxalpvpar = 2._RKIND*maxalpvpar
    maxalpphi  = 2._RKIND*maxalpphi
  end subroutine timesplit_conserv


  !****************************************************
  !   PREDICTOR-CORRECTOR ALGORITHM FOR 
  !    THE BSL AND PSM METHODS
  !****************************************************
  !---------------------------------------------------------
  ! step 0 : (step only used for the initialisation)
  !  - f initialisation,
  !  - computation of the corresponding Phi,
  !  (used in predictor-corrector case)  
  !--------------------------------------------------------- 
  subroutine step0_predcorr(geom,coord_sys, &
    init_prof,init_magnet,init_curr,init_srce,Sdiag, &
    dt,fnp1,SJ0ftmp,Sfmu_eq,SJ0fmu_eq,poiss,J0,sEntransf_read)
    
    use globals, only : mu_id, collisions, &
      energy_transfer, moment_transfer, &
      flux_driven, heat_source, fp_source, &
      dens_source, mom_source, vor_source, uout_res, outputproc
    use globals, only : pSpecies_id, major_species
    use Io_Unit_module, only: IU_open, IU_close
    use coord_system_class
    use poisson_class
    use QNsolver_module
    use gyroaverage_class
    use collisions_vpar_module , only : init_collisions_vpar
    use species_transf_module  , only : init_moments_Entransf
    use efield_module   
    
!R3 #include "code_instrumentation/r3_info.h" !R3
    type(geometry)     , intent(in)    :: geom
    type(coord_system) , intent(in)    :: coord_sys
    type(init_profile) , intent(in)    :: init_prof
    type(init_magnetic), intent(in)    :: init_magnet
    type(init_current) , intent(in)    :: init_curr
    type(init_sources) , intent(in)    :: init_srce
    type(diagnostics)  , intent(inout) :: Sdiag
    real(RKIND)        , intent(in)    :: dt
    type(fdistribu5d)  , intent(inout) :: fnp1
    type(fdistribu5d)  , intent(inout) :: SJ0ftmp
    real(RKIND), &
      dimension(:,:,:) , pointer       :: Sfmu_eq
    real(RKIND), &
      dimension(:,:,:) , pointer       :: SJ0fmu_eq
    type(poisson)      , intent(inout) :: poiss
    type(J0operator)   , intent(inout) :: J0
    logical            , intent(in)    :: sEntransf_read

!R3 call r3_info_begin (r3_info_index_0, 'CPU_step0') !R3
    call Write_dyn_balise_open('CPU_step0') !R4

    if (pglobal_id.eq.outputproc) then
      call IU_open(unit=uout_res,file=file_name_res,status='OLD',&
        position='APPEND', form='FORMATTED')
      write(uout_res,*) ' '
      write(uout_res,*) ' '
      write(uout_res,*) '**************************', &
        '*****************************************'
      write(uout_res,*) '******    ', &
        '  SOLVING WITH PREDICTOR-CORRECTOR ALGORITHM       ******'
      if (geom%Nmu.eq.0) then
        if (geom%Lmu.eq.0._RKIND) then
          write(uout_res,*) &
            '******                   NO GYROAVERAGE', &
            '                      ******'
        else
          write(uout_res,*) '******           ', &
            'GYROAVERAGE WITH ONE VALUE OF MU            ******'
        end if
      else
        write(uout_res,*) '******              ', &
          'WITH GYROAVERAGE                         ******'
      end if
      write(uout_res,*) '******                  ', &
        '                                     ******'
      if (collisions) then
        write(uout_res,*) '******              ', &
          'WITH COLLISIONS                          ******'
      else
        write(uout_res,*) '******              ', &
          'WITHOUT COLLISIONS                       ******'
      end if
      if (energy_transfer) then
        write(uout_res,*) '******              ', &
          'WITH ENERGY TRANSFER                     ******'
      else
        write(uout_res,*) '******              ', &
          'WITHOUT ENERGY TRANSFER                  ******'
      end if
      if (moment_transfer) then
        write(uout_res,*) '******              ', &
          'WITH MOMENTUM TRANSFER                   ******'
      else
        write(uout_res,*) '******              ', &
          'WITHOUT MOMENTUM TRANSFER                ******'
      end if
      if (flux_driven) then
        write(uout_res,*) '******              ', &
          'WITH FLUX DRIVEN CONDITIONS              ******'
      else
        write(uout_res,*) '******         WITH ', &
          'PRESCRIBED TEMPERATURE AT THE EDGE      ******'
      end if
      if (heat_source(major_species)) then
        write(uout_res,*) '******              ', &
          'WITH HEAT SOURCE                         ******'
      end if
      if (dens_source(major_species)) then
        write(uout_res,*) '******              ', &
          'WITH DENSITY SOURCE                      ******'
      end if
      if (mom_source(major_species)) then
        write(uout_res,*) '******              ', &
          'WITH MOMENTUM SOURCE                     ******'
      end if
      if (vor_source(major_species)) then
        write(uout_res,*) '******              ', &
          'WITH VORTICITY SOURCE                    ******'
      end if
      write(uout_res,*) '********************************', &
        '***********************************'
      write(uout_res,*) ' '
      call IU_close(uout_res)
    end if

    !*** initialisation of the collision operator ***
    if (collisions) then
      call init_collisions_vpar(geom,init_prof, &
        init_magnet,init_curr,fnp1)
      if ( (energy_transfer).or.(moment_transfer) ) &
        call init_moments_Entransf(geom, &
        init_magnet,init_curr,fnp1)
    end if

    !*** computation of Phi(0) by solving   ***
    !***   the quasi-neutrality equation     ***
    if (QN_version.ge.4) then
      call solve_quasineutrality(geom,J0,fnp1,Sfmu_eq, &
        init_prof,init_magnet,init_curr,poiss)
    else
      call solve_poisson(poiss,geom,init_prof, &
        init_magnet,init_curr,J0,fnp1,Sfmu_eq)
    end if

    !*** computation of the velocity components at time t=-1 ***
    call comp_dJ0Phidx(J0,geom,mu_id,poiss, &
      dJ0Phidr_PNrPNthNphi,dJ0Phidtheta_PNrPNthNphi, &
      dJ0Phidphi_PNrPNthNphi,dJ0Phidr_NrNthPNphi, &
      dJ0Phidtheta_NrNthPNphi,dJ0Phidphi_NrNthPNphi)

    call Write_dyn_balise_close('CPU_step0') !R4
!R3 call r3_info_end (r3_info_index_0) !R3
  end subroutine step0_predcorr


  !------------------------------------------------------------
  ! Solve the Vlasov equation for the distribution function f
  !  between t and t+dt by using a predictor-corrector
  !  scheme
  !---------------------------------------------------------  
  subroutine predcorr(geom, &
    coord_sys,init_prof, &
    init_magnet,init_curr,init_srce,dt, &
    Siter_time,Stime_save,fn,fnp1,Sfmu_eq,J0,poiss)
    
    use Debug_module
    use globals
    use gyroaverage_class
    use Io_Unit_module, only: IU_open, IU_close
    use poisson_class
    use QNsolver_module
    use efield_module         
    use collisions_vpar_module , only : refresh_collisions_vpar
    use species_transf_module  , only : compute_moments_Entransf
    use filter_module
    use mpiMod
    
    type(geometry)     , intent(in)    :: geom
    type(coord_system) , intent(in)    :: coord_sys
    type(init_profile) , intent(in)    :: init_prof
    type(init_magnetic), intent(in)    :: init_magnet
    type(init_current) , intent(in)    :: init_curr
    type(init_sources) , intent(in)    :: init_srce
    real(RKIND)        , intent(inout) :: dt
    real(RKIND)        , intent(in)    :: Siter_time
    real(RKIND)        , intent(in)    :: Stime_save
    type(fdistribu5d)  , pointer       :: fn
    type(fdistribu5d)  , pointer       :: fnp1
    real(RKIND), &
      dimension(:,:,:) , pointer       :: Sfmu_eq
    type(J0operator)   , intent(inout) :: J0
    type(poisson)      , intent(inout) :: poiss

    integer            :: ir, itheta, iphi, ivpar, ipred, ierr
    integer(TIMEPREC)  :: tdeb, tfin
    real(RKIND)        :: predratio, effectivedt, coeffadapt
    integer            :: itadapt, nbadapt
    integer, parameter :: MAXPRED = 2
    logical            :: activatefilter

    type(fdistribu5d) , pointer :: swap
    !-> for time step adaptation
    real(RKIND) :: maxalprth, maxalpphi, maxalpvpar, maxalp
    real(RKIND) :: dt_muid_reduce, dt_muid_diag, dt_lin
    integer     :: dt_direction_reduce

    maxalprth  = 0._RKIND
    maxalpphi  = 0._RKIND
    maxalpvpar = 0._RKIND
    
    PRINT_CKSUM(fn%values)
    PRINT_CKSUM(fnp1%values)
    
    call FTR_ALLOC_4(f4D_transp,0,Nr,0,Ntheta,kstart,kend,lstart,lend)

    !*** open file results ***
    if (pglobal_id.eq.outputproc) then
      call IU_open(unit=uout_res,file=file_name_res,status='OLD',&
        position='APPEND',form='FORMATTED')
    end if

    !*** refresh collision operator ***
    if ((collisions).and.(coll_FMrefresh.ne.0)) then
      if (mod(iter_glob,coll_FMrefresh).eq.0) then 
        if (pglobal_id.eq.outputproc) &
          write(uout_res,'(A17,I5,A38)') '---> iteration = ', &
          iter_glob,' Update of the collision operator'
        call refresh_collisions_vpar(geom, &
          init_prof,init_magnet,init_curr,fnp1)
        if ( (energy_transfer).or.(moment_transfer) ) &
          call compute_moments_Entransf(geom, &
          init_magnet,init_curr,fnp1)
      end if
    end if

    if (cancel_nonlin) then
      call predcorr_unperturbed(geom, &
        coord_sys,init_prof, &
        init_magnet,init_curr,init_srce,dt, &
        Siter_time,Stime_save,fn,fnp1,Sfmu_eq,J0,poiss)
    else 
      if (split_linear) then
        call predcorr_splitlinear(geom, &
          coord_sys,init_prof, &
          init_magnet,init_curr,init_srce,dt, &
          Siter_time,Stime_save,fn,fnp1,Sfmu_eq,J0,poiss)
      else
        call predcorr_standard(geom, &
          coord_sys,init_prof, &
          init_magnet,init_curr,init_srce,dt, &
          Siter_time,Stime_save,maxalprth,maxalpphi,&
          maxalpvpar,fn,fnp1,Sfmu_eq,J0,poiss)
      end if
    end if

    if (adaptive_deltat) then
      !*** computation of the adaptive time step ***
      !-> local maximum displacement 
      maxalp = max(maxalpphi/8._RKIND,maxalprth/0.5_RKIND)
      maxalp = max(maxalp,maxalpvpar/8._RKIND)
      !-> give the direction of the biggest normalized displacement
      if (maxalp == maxalprth/0.5_RKIND) dt_direction_reduce = 1
      if (maxalp == maxalpphi/8._RKIND)  dt_direction_reduce = 3
      if (maxalp == maxalpvpar/8._RKIND) dt_direction_reduce = 4
      !-> local adaptive time step
      dt = min(dt_diag,1._RKIND/maxalp)
      if ( ((Siter_time+dt).gt.Stime_save) .and. &
        .not.(approx_equal(Stime_save,Siter_time)) ) then
        dt = Stime_save - Siter_time
      end if
      !-> global adaptive time step
      dt_muid_reduce = dt
      call MY_Allreduce(dt_muid_reduce,dt_muid_diag,1, &
        MPI_REAL8,MPI_MIN,MPI_COMM_WORLD,ierr)

      dt = dt_muid_diag
      !-> print the direction responsible of 
      !->  the time step changing 
      !->  (most restrictive CFL condition)
      if (dt.eq.dt_muid_reduce) then
        write(6,'(A,I8,A,I2,A,E15.5)') &
          "pglobal_id = ", pglobal_id, &
          " ==> dt_direction_reduce = ", dt_direction_reduce,&
          " new dt ", dt
      end if
      if (pglobal_id .eq. cible) then
        write(6,'(A,E15.5)') "Chosen new dt ", dt
      end if
    end if
    if (pglobal_id.eq.outputproc) &
      call IU_close(uout_res)
    
    call FTR_DEALLOC(f4D_transp)

    PRINT_CKSUM(fn%values)
    PRINT_CKSUM(fnp1%values)
  end subroutine predcorr


  !------------------------------------------------------------
  ! Solve the Vlasov equation for the distribution function f
  !  between t and t+dt by using a predictor-corrector
  !  scheme
  !---------------------------------------------------------  
  subroutine predcorr_standard(geom, &
    coord_sys,init_prof, &
    init_magnet,init_curr,init_srce,dt, &
    Siter_time,Stime_save,maxalprth,maxalpphi,maxalpvpar, &
    fn,fnp1,Sfmu_eq,J0,poiss)
    use globals
    use gyroaverage_class
    use poisson_class
    use QNsolver_module
    use efield_module         
    use collisions_vpar_module , only : refresh_collisions_vpar
    use species_transf_module  , only : compute_moments_Entransf
    use filter_module
    use mpiMod
    type(geometry)     , intent(in)    :: geom
    type(coord_system) , intent(in)    :: coord_sys
    type(init_profile) , intent(in)    :: init_prof
    type(init_magnetic), intent(in)    :: init_magnet
    type(init_current) , intent(in)    :: init_curr
    type(init_sources) , intent(in)    :: init_srce
    real(RKIND)        , intent(inout) :: dt
    real(RKIND)        , intent(in)    :: Siter_time
    real(RKIND)        , intent(in)    :: Stime_save
    real(RKIND)        , intent(out)   :: maxalprth
    real(RKIND)        , intent(out)   :: maxalpphi
    real(RKIND)        , intent(out)   :: maxalpvpar
    type(fdistribu5d)  , pointer       :: fn
    type(fdistribu5d)  , pointer       :: fnp1
    real(RKIND), &
      dimension(:,:,:) , pointer       :: Sfmu_eq
    type(J0operator)   , intent(inout) :: J0
    type(poisson)      , intent(inout) :: poiss

    integer            :: ir, itheta, iphi, ivpar, ipred, ierr
    integer(TIMEPREC)  :: tdeb, tfin
    real(RKIND)        :: predratio, effectivedt, coeffadapt
    integer            :: itadapt, nbadapt
    integer, parameter :: MAXPRED = 2
    logical            :: activatefilter

    type(fdistribu5d) , pointer :: swap
    
    PRINT_CKSUM(fn%values)
    PRINT_CKSUM(fnp1%values)

    call Write_dyn_balise_open('predcorr_standard') !R4
	
    if ((reduced_deltat .ne. 0._RKIND) .and. &
      (Siter_time .ge. reduced_begin) .and. &
      (Siter_time .le. reduced_end)) then
      nbadapt    = MAX(1, nint(dt / reduced_deltat))
      coeffadapt = 1._RKIND / nbadapt
    else
      nbadapt    = 1
      coeffadapt = 1._RKIND
    end if

    maxalprth  = 0._RKIND
    maxalpphi  = 0._RKIND
    maxalpvpar = 0._RKIND

    do itadapt = 1,nbadapt
      swap => fn
      fn   => fnp1
      fnp1 => swap

      do ipred = MAXPRED,1,-1
        activatefilter = (itadapt .eq. nbadapt) .and. &
          (ipred .eq. 1)
        predratio   = (2._RKIND ** (ipred-1))
        predratio   = coeffadapt / predratio
        effectivedt = predratio * dt
        !*** time-splitting iteration to compute    ***
        !***  f(tn+1/(2^ipred))                     ***
        !*** computation of the velocity components ***
        !***   at time t=-1                         ***
        call comp_dJ0Phidx(J0,geom,mu_id,poiss, &
          dJ0Phidr_PNrPNthNphi,dJ0Phidtheta_PNrPNthNphi, &
          dJ0Phidphi_PNrPNthNphi,dJ0Phidr_NrNthPNphi, &
          dJ0Phidtheta_NrNthPNphi,dJ0Phidphi_NrNthPNphi)

        PRINT_CKSUM(fn%values)
        PRINT_CKSUM(fnp1%values)

        do ivpar = 0,geom%Nvpar
          do iphi = 0,geom%Nphi
            do itheta = fn%jstart_buf,fn%jend_buf
              do ir = fn%istart_buf,fn%iend_buf
                fnp1%values(ir,itheta,iphi,ivpar) = &
                  fn%values(ir,itheta,iphi,ivpar)
              end do
            end do
          end do
        end do
    
        PRINT_CKSUM(fn%values)
        PRINT_CKSUM(fnp1%values)
        
        if (.not.RHS_only) then
          if (Vlasov_scheme=="BSL") then
            call timesplit_BSL(geom, &
              init_prof,init_magnet, &
              init_curr,init_srce,effectivedt,fnp1,Sfmu_eq, &
              activatefilter)
          else
            call timesplit_conserv(geom, &
              coord_sys,init_prof,init_magnet, &
              init_curr,init_srce,effectivedt,fnp1,Sfmu_eq, &
              activatefilter,maxalprth,maxalpphi,maxalpvpar)
          end if
        else
          call timesplit_RHSonly(geom,init_prof,init_magnet, &
            init_curr,init_srce,effectivedt,fnp1,Sfmu_eq, &
            activatefilter=.true.)
        end if

        if (activatefilter) then
          call hffilter(fnp1,Sfmu_eq)
        end if
        !*** computation of Phi(tn+1) by solving ***
        !***  the quasi-neutrality equation      ***
        if (QN_version.ge.4) then
          call solve_quasineutrality(geom,J0,fnp1,Sfmu_eq, &
            init_prof,init_magnet,init_curr,poiss)
        else
          call solve_poisson(poiss,geom,init_prof, &
            init_magnet,init_curr,J0,fnp1,Sfmu_eq)
        end if
      end do
    end do
	
    call Write_dyn_balise_close('predcorr_standard') !R4
  end subroutine predcorr_standard


  !------------------------------------------------------------
  ! Solve the Vlasov equation for the distribution function f
  !  between t and t+dt by using a predictor-corrector
  !  scheme
  !---------------------------------------------------------  
  subroutine predcorr_splitlinear(geom, &
    coord_sys,init_prof, &
    init_magnet,init_curr,init_srce,dt, &
    Siter_time,Stime_save,fn,fnp1,Sfmu_eq,J0,poiss)
    use globals
    use gyroaverage_class
    use poisson_class
    use QNsolver_module
    use efield_module         
    use collisions_vpar_module , only : refresh_collisions_vpar
    use species_transf_module  , only : compute_moments_Entransf
    use filter_module
    use mpiMod
    type(geometry)     , intent(in)    :: geom
    type(coord_system) , intent(in)    :: coord_sys
    type(init_profile) , intent(in)    :: init_prof
    type(init_magnetic), intent(in)    :: init_magnet
    type(init_current) , intent(in)    :: init_curr
    type(init_sources) , intent(in)    :: init_srce
    real(RKIND)        , intent(inout) :: dt
    real(RKIND)        , intent(in)    :: Siter_time
    real(RKIND)        , intent(in)    :: Stime_save
    type(fdistribu5d)  , pointer       :: fn
    type(fdistribu5d)  , pointer       :: fnp1
    real(RKIND), &
      dimension(:,:,:) , pointer       :: Sfmu_eq
    type(J0operator)   , intent(inout) :: J0
    type(poisson)      , intent(inout) :: poiss

    integer            :: ir, itheta, iphi, ivpar, ipred, ierr
    integer(TIMEPREC)  :: tdeb, tfin
    real(RKIND)        :: predratio, effectivedt, coeffadapt
    integer            :: itadapt, nbadapt
    integer, parameter :: MAXPRED = 2
    logical            :: activatefilter

    type(fdistribu5d) , pointer :: swap
    !-> for time step adaptation
    real(RKIND) :: dt_lin

    if ((reduced_deltat .ne. 0._RKIND) .and. &
      (Siter_time .ge. reduced_begin) .and. &
      (Siter_time .le. reduced_end)) then
      nbadapt    = MAX(1, nint(dt / reduced_deltat))
      coeffadapt = 1._RKIND / nbadapt
    else
      nbadapt    = 1
      coeffadapt = 1._RKIND
    end if
    
    do itadapt = 1,nbadapt
      swap => fn
      fn   => fnp1
      fnp1 => swap

      !*** Linear advection on 0.5 dt ***
      dt_lin = coeffadapt*.5_RKIND*dt
      call timesplit_BSL(geom, &
        init_prof,init_magnet, &
        init_curr,init_srce,dt_lin,fn,Sfmu_eq, &
        activatefilter,onlylin=.true.)

      !*** computation of Phi(tn+1) by solving ***
      !***  the quasi-neutrality equation      ***
      if (QN_version.ge.4) then
        call solve_quasineutrality(geom,J0,fn,Sfmu_eq, &
          init_prof,init_magnet,init_curr,poiss)
      else
        call solve_poisson(poiss,geom,init_prof, &
          init_magnet,init_curr,J0,fn,Sfmu_eq)
      end if

      !*** Non-linear advection on dt ***
      do ipred = MAXPRED,1,-1
        activatefilter = (itadapt .eq. nbadapt) .and. &
          (ipred .eq. 1)
        predratio   = (2._RKIND ** (ipred-1))
        predratio   = coeffadapt / predratio
        effectivedt = predratio * dt
        !*** time-splitting iteration to compute    ***
        !***  f(tn+1/(2^ipred))                     ***
        !*** computation of the velocity components ***
        !***  at time t=-1                          ***
        call comp_dJ0Phidx(J0,geom,mu_id,poiss, &
          dJ0Phidr_PNrPNthNphi,dJ0Phidtheta_PNrPNthNphi, &
          dJ0Phidphi_PNrPNthNphi,dJ0Phidr_NrNthPNphi, &
          dJ0Phidtheta_NrNthPNphi,dJ0Phidphi_NrNthPNphi)

        do ivpar = 0,geom%Nvpar
          do iphi = 0,geom%Nphi
            do itheta = fn%jstart_buf,fn%jend_buf
              do ir = fn%istart_buf,fn%iend_buf
                fnp1%values(ir,itheta,iphi,ivpar) = &
                  fn%values(ir,itheta,iphi,ivpar)
              end do
            end do
          end do
        end do
        if ( (Vlasov_scheme=="BSL") .and. (.not.RHS_only) ) then
          call timesplit_BSL(geom, &
            init_prof,init_magnet,init_curr,init_srce, &
            effectivedt,fnp1,Sfmu_eq, &
            activatefilter,onlynl=.true.)
        else
          if (pglobal_id.eq.0) &
            print*,' split_linear = .true. is only available ', &
            'for BSL algorithm'
          call ppexit()
          call exit(-1)          
        end if

        !*** Linear advection on 0.5 dt ***
        if (ipred .eq. 1) then
          call timesplit_BSL(geom, &
            init_prof,init_magnet, &
            init_curr,init_srce,dt_lin,fnp1,Sfmu_eq, &
            activatefilter,onlylin=.true.)
        end if

        if (activatefilter) then
          call hffilter(fnp1,Sfmu_eq)
        end if

        !*** computation of Phi(tn+1) by solving ***
        !***  the quasi-neutrality equation      ***
        if (QN_version.ge.4) then
          call solve_quasineutrality(geom,J0,fnp1,Sfmu_eq, &
            init_prof,init_magnet,init_curr,poiss)
        else
          call solve_poisson(poiss,geom,init_prof, &
            init_magnet,init_curr,J0,fnp1,Sfmu_eq)
        end if
      end do
    end do
  end subroutine predcorr_splitlinear


  !------------------------------------------------------------
  ! Solve the Vlasov equation for the distribution function f
  !  between t and t+dt by using a predictor-corrector
  !  scheme
  !---------------------------------------------------------  
  subroutine predcorr_unperturbed(geom, &
    coord_sys,init_prof, &
    init_magnet,init_curr,init_srce,dt, &
    Siter_time,Stime_save,fn,fnp1,Sfmu_eq,J0,poiss)
    use globals
    use gyroaverage_class
    use poisson_class
    use QNsolver_module
    use efield_module         
    use collisions_vpar_module , only : refresh_collisions_vpar
    use species_transf_module  , only : compute_moments_Entransf
    use filter_module
    use mpiMod
    type(geometry)     , intent(in)    :: geom
    type(coord_system) , intent(in)    :: coord_sys
    type(init_profile) , intent(in)    :: init_prof
    type(init_magnetic), intent(in)    :: init_magnet
    type(init_current) , intent(in)    :: init_curr
    type(init_sources) , intent(in)    :: init_srce
    real(RKIND)        , intent(inout) :: dt
    real(RKIND)        , intent(in)    :: Siter_time
    real(RKIND)        , intent(in)    :: Stime_save
    type(fdistribu5d)  , pointer       :: fn
    type(fdistribu5d)  , pointer       :: fnp1
    real(RKIND), &
      dimension(:,:,:) , pointer       :: Sfmu_eq
    type(J0operator)   , intent(inout) :: J0
    type(poisson)      , intent(inout) :: poiss

    integer            :: ir, itheta, iphi, ivpar, ipred, ierr
    integer(TIMEPREC)  :: tdeb, tfin
    real(RKIND)        :: predratio, effectivedt, coeffadapt
    integer            :: itadapt, nbadapt
    integer, parameter :: MAXPRED = 2
    logical            :: activatefilter

    type(fdistribu5d) , pointer :: swap
    !-> for time step adaptation
    real(RKIND) :: dt_lin

    if ((reduced_deltat .ne. 0._RKIND) .and. &
      (Siter_time .ge. reduced_begin) .and. &
      (Siter_time .le. reduced_end)) then
      nbadapt    = MAX(1, nint(dt / reduced_deltat))
      coeffadapt = 1._RKIND / nbadapt
    else
      nbadapt    = 1
      coeffadapt = 1._RKIND
    end if
    
    do itadapt = 1,nbadapt
      dt_lin = coeffadapt*dt
      call timesplit_BSL(geom, &
        init_prof,init_magnet, &
        init_curr,init_srce,dt_lin,fn,Sfmu_eq, &
        activatefilter,onlylin=.true.)
      
      !*** computation of Phi(tn+1) by solving ***
      !***  the quasi-neutrality equation      ***
      if (QN_version.ge.4) then
        call solve_quasineutrality(geom,J0,fn,Sfmu_eq, &
          init_prof,init_magnet,init_curr,poiss)
      else
        call solve_poisson(poiss,geom,init_prof, &
          init_magnet,init_curr,J0,fn,Sfmu_eq)
      end if
    end do

    swap => fn
    fn   => fnp1
    fnp1 => swap
  end subroutine predcorr_unperturbed

end module Boltz_solving_module
