!**************************************************************
!  Copyright Euratom-CEA
!  Authors : 
!     Virginie Grandgirard (virginie.grandgirard@cea.fr)
!     Chantal Passeron (chantal.passeron@cea.fr)
!     Guillaume Latu (guillaume.latu@cea.fr)
!     Xavier Garbet (xavier.garbet@cea.fr)
!     Philippe Ghendrih (philippe.ghendrih@cea.fr)
!     Yanick Sarazin (yanick.sarazin@cea.fr)
!  
!  This code GYSELA (for GYrokinetic SEmi-LAgrangian) 
!  is a 5D gyrokinetic global full-f code for simulating 
!  the plasma turbulence in a tokamak.
!  
!  This software is governed by the CeCILL-B license 
!  under French law and abiding by the rules of distribution 
!  of free software.  You can  use, modify and redistribute 
!  the software under the terms of the CeCILL-B license as 
!  circulated by CEA, CNRS and INRIA at the following URL
!  "http://www.cecill.info". 
!**************************************************************

#include "config-F.inc"

!-------------------------------------------------
! file : vlasov_solving.F90
! date :  16/09/2005
!  Global algorithm for solving the vlasov system
!  which is divided into :
!   - time-splitting algorithm and 
!   - leap-frog or predictor algorithm
!-------------------------------------------------
module vlasov_solving_module
  use globals
  use prec_const
  use geometry_class
  use init_profile_class
  use init_magnetic_class
  use init_current_class
  use fdistribu5d_class
  use poisson_class
  use diagnostics_module
  use clock_module
  implicit none
      
  !******************************
  contains
  !******************************
  
  !-----------------------------------------------------
  !  Mesh and profile initialisations
  !   - grid definition in (r,theta,phi)
  !   - initialisation of the density and temperature
  !     profiles
  !   - grid definition in vparallel direction
  !  (step only used for the initialisation)
  !-----------------------------------------------------
  subroutine init_mesh_profile(geom,init_prof)
    use globals, only : memory_test, &
      rhomin, rhomax, Ltheta, Lphi, nb_vth0
!R3 #include "code_instrumentation/r3_info.h" !R3
    type(geometry)    , intent(inout) :: geom
    type(init_profile), intent(inout) :: init_prof

    real(RKIND) :: rmin, rmax, thetamin, phimin, vparmin
    real(RKIND) :: Lr, Lvpar

!R3 call r3_info_begin (r3_info_index_1, 'Init_mesh') !R3
    call Write_dyn_balise_open('Init_mesh') !R4
    !*** (r,theta,phi) mesh definition ***
    rmin     = a*rhomin
    rmax     = a*rhomax
    Lr       = abs(rmax-rmin)
    thetamin = 0._RKIND
    phimin   = 0._RKIND
	
	call write_param_2("rhomin", rhomin)
	call write_param_2("rhomax", rhomax)
	call write_param_2("a", a)
	call write_expr("rmin", "a*rhomin")
	call write_expr("rmax", "a*rhomax")
	call write_expr("Lr", "abs(rmax-rmin)")
	
    call new_space_coord(geom,rmin,thetamin,phimin, &
      Lr,Ltheta,Lphi,Nr,Ntheta,Nphi)

    ! *** Definition of the major radius ***
    R0 = a*aspect_ratio
    
    !*** velocity mesh definition ***
    Lvpar   = 2_RKIND*nb_vth0
    vparmin = -nb_vth0
    call new_velocity_coord(geom,vparmin,mumin,Lvpar,Lmu,Nvpar,Nmu)

    !*** profile definition *** 
    call init_all_profile(init_prof,geom)
    call Write_dyn_balise_close('Init_mesh') !R4
!R3 call r3_info_end (r3_info_index_1) !R3
  end subroutine init_mesh_profile


  !--------------------------------------------------------
  !  - initialisation of the distribution function 
  ! at the initial time (with the equilibrium Maxwellian)
  !-------------------------------------------------------- 
  subroutine initialisation_fstart(geom,init_prof, &
    Sfmu_eq,fstart)
    use Pcross_section_module
    type(geometry)    , intent(in)    :: geom
    type(init_profile), intent(in)    :: init_prof
    real(RKIND), &
      dimension(:,:,:), pointer       :: Sfmu_eq
    type(fdistribu5d) , intent(inout) :: fstart

    !*** distribution function initialisation ***
    call clck_time(bclock_init)
    call f_initialisation(fstart,geom,init_prof,Sfmu_eq)
    call clck_time(eclock_init)
    call clck_diff(bclock_init,eclock_init, &
      global_time_init)
  end subroutine initialisation_fstart


  !-------------------------------------------------------------
  ! TIME-SPLITTING between t and t+dt for the
  !  SEMI-LAGRANGIAN method IN ALL DIRECTIONS
  ! Computation at the time t, with the shift sequence :
  !     vpar/2,phi/2,(r,theta),phi/2,vpar/2 
  ! to keep a shift symmetry
  !------------------------------------------------------------
  subroutine timesplit_vlasov_BSL(geom, &
    init_prof,init_magnet,init_curr, &
    dt,fn,Sfmu_eq,activatefilter)
    use globals, only : transp_velocity, advec2D_deltat
    use MPIutils_module
    use advec1D_BSL_module
    use advec2D_BSL_module
    use sources_module
    use filter_module
    use diffusion_module, only : advance_diff_r , &
      advance_diff_th
    use mpiMod

    type(geometry)       , intent(in)    :: geom
    type(init_profile)   , intent(in)    :: init_prof
    type(init_magnetic)  , intent(in)    :: init_magnet
    type(init_current)   , intent(in)    :: init_curr
    real(RKIND)          , intent(in)    :: dt
    type(fdistribu5d)    , intent(inout) :: fn
    real(RKIND), &
      dimension(:,:,:)   , pointer       :: Sfmu_eq
    logical              , intent(in)    :: activatefilter
    
    integer     :: ivpar, nbsubit
    real(RKIND) :: subdt
    integer     :: ierr

    if (advec2D_deltat .ne. 0._RKIND) then
       nbsubit = MAX(1, nint(dt / advec2D_deltat))
       subdt   = dt / nbsubit
    else
       nbsubit = 1
       subdt   = dt
    end if

    !*** step 1 : splitting in v parallel ***
    !***  direction (on 0.5*dt)           ***
    call advec1D_BSL_vpar(geom,0.5_RKIND*dt, &
      init_magnet,init_curr,fn)
    !*** step 2 : splitting in phi direction (on 0.5*dt) ***
    call advec1D_BSL_phi(geom,0.5_RKIND*dt, &
      init_magnet,init_curr,fn)
    !*** step 3 : Diffusion operator in r (on 0.5*dt) ***
    if (coef_diffr(pSpecies_id).ne.0._RKIND) &
      call advance_Diff_r(Sfmu_eq,0.5_RKIND*dt,fn)
    !*** step 4 : splitting in (r,theta) (on dt)          ***
    !***        + 2 Diffusion operator in theta (on dt/2) ***
    if (transpose4D) then
      call FTR_ALLOC_4(f4D_transp,0,Nr,0,Ntheta,kstart,kend,lstart,lend)
      !*** Transpose the 4D function from f%values to f4D_transp ***

      select case ( transpose_version )
      case ( TRANSPOSE_SENDIRECV )
         call FTR_ALLOC_4_ZERO(f4D_send,0,dom_r-1,0,dom_theta-1,0,dom_phi-1,0,dom_vpar-1)
         call FTR_ALLOC_4_ZERO(f4D_recv,0,dom_r-1,0,dom_theta-1,0,dom_phi-1,0,dom_vpar-1)
      case ( TRANSPOSE_ALLTOALLW )
         ! nothing specific to allocate
      case default
         print *, pglobal_id, 'Invalid value for transpose_version', &
              transpose_version
         call MPI_abort(MPI_COMM_WORLD, 1, ierr)
      end select
      
      call pptranspose_forward(fn%values,timing=.true.)
      !--> theta diffusion operator
      if (coef_diffth(pSpecies_id).ne.0._RKIND) &
        call advance_Diff_th(Sfmu_eq,0.5_RKIND*dt,f4D_transp)
      !--> 2D advection in (r,theta)
      call transpose_advec2D_xy(geom,subdt, &
        nbsubit,init_prof,init_magnet,init_curr, &
        fn,Sfmu_eq,activatefilter)
      !--> theta diffusion operator
      if (coef_diffth(pSpecies_id).ne.0._RKIND) &
        call advance_Diff_th(Sfmu_eq,0.5_RKIND*dt,f4D_transp)
      !*** Transpose the 4D function from f4D_transp to f%values ***
      call pptranspose_backward(fn%values,timing=.true.)

      select case ( transpose_version )
      case ( TRANSPOSE_SENDIRECV )
         call FTR_DEALLOC(f4D_send)
         call FTR_DEALLOC(f4D_recv)
      case ( TRANSPOSE_ALLTOALLW )
         ! nothing specific to allocate
      case default
         print *, pglobal_id, 'Invalid value for transpose_version', &
              transpose_version
         call MPI_abort(MPI_COMM_WORLD, 1, ierr)
      end select

      call FTR_DEALLOC(f4D_transp)
    else
      call local_advec2D_xy(geom,dt, &
        init_prof,init_magnet,init_curr,fn)
      if (activatefilter) call hffilter(fn,Sfmu_eq)
    end if
    !*** step 5 : Diffusion operator in r (on 0.5*dt) ***
    if (coef_diffr(pSpecies_id).ne.0._RKIND) &
      call advance_Diff_r(Sfmu_eq,0.5_RKIND*dt,fn)
    !*** step 6 : splitting in phi direction (on 0.5*dt) ***
    call advec1D_BSL_phi(geom,0.5_RKIND*dt, &
      init_magnet,init_curr,fn)
    !*** step 7 : splitting in v parallel ***
    !***  direction (on 0.5*dt)           ***
    call advec1D_BSL_vpar(geom,0.5_RKIND*dt, &
      init_magnet,init_curr,fn)

  end subroutine timesplit_vlasov_BSL


  !****************************************************
  !   PREDICTOR-CORRECTOR ALGORITHM FOR THE BSL METHOD
  !****************************************************
  !---------------------------------------------------------
  ! step 0 : (step only used for the initialisation)
  !  - f initialisation,
  !  - computation of the corresponding Phi,
  !  (used in predictor-corrector case)  
  !--------------------------------------------------------- 
  subroutine step0_predcorr_vlasov(geom,coord_sys, &
    init_prof,init_magnet,init_curr,init_srce &                                  
    ,Sdiag,dt,fnp1,SJ0ftmp,Sfmu_eq,SJ0fmu_eq,poiss,J0)
    use globals, only : mu_id, uout_res, outputproc, &
      pSpecies_id
    use coord_system_class
    use poisson_class
    use gyroaverage_class
    use efield_module 
    use sources_module                        
!R3 #include "code_instrumentation/r3_info.h" !R3
    type(geometry)     , intent(in)    :: geom
    type(coord_system) , intent(in)    :: coord_sys
    type(init_profile) , intent(in)    :: init_prof
    type(init_magnetic), intent(in)    :: init_magnet
    type(init_current) , intent(in)    :: init_curr
    type(init_sources) , intent(in)    :: init_srce  
    type(diagnostics)  , intent(inout) :: Sdiag
    real(RKIND)        , intent(in)    :: dt
    type(fdistribu5d)  , intent(inout) :: fnp1
    type(fdistribu5d)  , intent(inout) :: SJ0ftmp
    real(RKIND), &
      dimension(:,:,:) , pointer       :: Sfmu_eq
    real(RKIND), &
      dimension(:,:,:) , pointer       :: SJ0fmu_eq
    type(poisson)      , intent(inout) :: poiss
    type(J0operator)   , intent(inout) :: J0

!R3 call r3_info_begin (r3_info_index_0, 'CPU_step0') !R3
    call Write_dyn_balise_open('CPU_step0') !R4

    if (pglobal_id.eq.outputproc) then
      write(uout_res,*) ' '
      write(uout_res,*) ' '
      write(uout_res,*) '**************************', &
        '*****************************************'
      write(uout_res,*) '******    ', &
        '  SOLVING WITH PREDICTOR-CORRECTOR ALGORITHM       ******'
      if (geom%Nmu.eq.0) then
        if (geom%Lmu.eq.0._RKIND) then
          write(uout_res,*) &
            '******                   NO GYROAVERAGE', &
            '                      ******'
        else
          write(uout_res,*) '******           ', &
            'GYROAVERAGE WITH ONE VALUE OF MU            ******'
        end if
      else
        write(uout_res,*) '******                 ', &
          ' WITH GYROAVERAGE                     ******'
      end if
      write(uout_res,*) '******                  ', &
        '                                     ******'
      write(uout_res,*) &
        '******       WITH PRESCRIBED TEMPERATURE', &
        ' AT THE EDGE         ******'
      write(uout_res,*) '********************************', &
        '***********************************'
      write(uout_res,*) ' '
    end if

    !*** computation of Phi(-1) by solving   ***
    !***   the quasi-neutrality equation     ***
    call solve_poisson(poiss,geom,init_prof, &
      init_magnet,init_curr,J0,fnp1,Sfmu_eq)

    !*** computation of the velocity components at time t=-1 ***
    call comp_dJ0Phidx(J0,geom,mu_id,poiss, &
      dJ0Phidr_PNrPNthNphi,dJ0Phidtheta_PNrPNthNphi, &
      dJ0Phidphi_PNrPNthNphi,dJ0Phidr_NrNthPNphi, &
      dJ0Phidtheta_NrNthPNphi,dJ0Phidphi_NrNthPNphi)
	
    call Write_dyn_balise_close('CPU_step0') !R4
!R3 call r3_info_end (r3_info_index_0) !R3
  end subroutine step0_predcorr_vlasov


  !------------------------------------------------------------
  ! Solve the Vlasov equation for the distribution function f
  !  between t and t+dt by using a predictor-corrector
  !  scheme
  !---------------------------------------------------------  
  subroutine global_predcorr_vlasov(geom,init_prof, &
    init_magnet,init_curr,dt,iter_time,time_save, &
    fn,fnp1,Sfmu_eq,J0,poiss)
    use globals
    use gyroaverage_class
    use poisson_class
    use efield_module
    use mpiMod
    type(geometry)     , intent(in)    :: geom
    type(init_profile) , intent(in)    :: init_prof
    type(init_magnetic), intent(in)    :: init_magnet
    type(init_current) , intent(in)    :: init_curr
    real(RKIND)        , intent(in)    :: iter_time
    real(RKIND)        , intent(in)    :: time_save
    real(RKIND)        , intent(inout) :: dt
    type(fdistribu5d)  , pointer       :: fn
    type(fdistribu5d)  , pointer       :: fnp1
    real(RKIND), &
      dimension(:,:,:) , pointer       :: Sfmu_eq
    type(J0operator)   , intent(inout) :: J0
    type(poisson)      , intent(inout) :: poiss

    integer            :: ir, itheta, iphi, ivpar, ipred, ierr
    integer(TIMEPREC)  :: tdeb, tfin
    real(RKIND)        :: predratio, effectivedt, coeffadapt
    integer            :: itadapt, nbadapt
    integer, parameter :: MAXPRED = 2
    logical            :: activatefilter
    type(fdistribu5d) , pointer :: swap
    !-> for time step adaptation
    real(RKIND) :: dt_muid_reduce, dt_muid_diag
    integer     :: dt_direction_reduce

    if ((reduced_deltat .ne. 0._RKIND) .and. &
      (iter_time .ge. reduced_begin) .and. &
      (iter_time .le. reduced_end)) then
      nbadapt    = MAX(1,nint(dt/reduced_deltat))
      coeffadapt = 1._RKIND / nbadapt
    else
      nbadapt    = 1
      coeffadapt = 1._RKIND
    end if
    
    do itadapt = 1,nbadapt
      swap => fn
      fn   => fnp1
      fnp1 => swap
      do ipred = MAXPRED,1,-1
        activatefilter = (itadapt .eq. nbadapt) .and. &
          (ipred .eq. 1)
        predratio   = (2._RKIND ** (ipred-1))
        predratio   = coeffadapt / predratio
        effectivedt = predratio * dt
        !*** time-splitting iteration to compute ***
        !***  f(tn+1/(2^ipred))                  ***
        call comp_dJ0Phidx(J0,geom,mu_id,poiss, &
          dJ0Phidr_PNrPNthNphi,dJ0Phidtheta_PNrPNthNphi, &
          dJ0Phidphi_PNrPNthNphi,dJ0Phidr_NrNthPNphi, &
          dJ0Phidtheta_NrNthPNphi,dJ0Phidphi_NrNthPNphi)
        do ivpar = 0,geom%Nvpar
          do iphi = 0,geom%Nphi
            do itheta = fn%jstart_buf,fn%jend_buf
              do ir = fn%istart_buf,fn%iend_buf
                fnp1%values(ir,itheta,iphi,ivpar) = &
                  fn%values(ir,itheta,iphi,ivpar)
              end do
            end do
          end do
        end do
        call timesplit_vlasov_BSL(geom,init_prof,init_magnet, &
          init_curr,effectivedt,fnp1,Sfmu_eq,activatefilter)

        !*** computation of Phi(tn+1) by solving ***
        !***  the quasi-neutrality equation      ***
        call solve_poisson(poiss,geom,init_prof, &
          init_magnet,init_curr,J0,fnp1,Sfmu_eq)
      end do
    end do

  end subroutine global_predcorr_vlasov
end module vlasov_solving_module
