!**************************************************************
!  Copyright Euratom-CEA
!  Authors : 
!     Virginie Grandgirard (virginie.grandgirard@cea.fr)
!     Chantal Passeron (chantal.passeron@cea.fr)
!     Guillaume Latu (guillaume.latu@cea.fr)
!     Xavier Garbet (xavier.garbet@cea.fr)
!     Philippe Ghendrih (philippe.ghendrih@cea.fr)
!     Yanick Sarazin (yanick.sarazin@cea.fr)
!  
!  This code GYSELA (for GYrokinetic SEmi-LAgrangian) 
!  is a 5D gyrokinetic global full-f code for simulating 
!  the plasma turbulence in a tokamak.
!  
!  This software is governed by the CeCILL-B license 
!  under French law and abiding by the rules of distribution 
!  of free software.  You can  use, modify and redistribute 
!  the software under the terms of the CeCILL-B license as 
!  circulated by CEA, CNRS and INRIA at the following URL
!  "http://www.cecill.info". 
!**************************************************************

#include "config-F.inc"
#include "code_instrumentation/debug.inc"

!----------------------------------------------------------------
! file   : collisions.F90
! date   : 31/05/2007
! author : G. Dif-Pradalier
!  Treatment of the collision operator
!  We shall solve the following equation:
!     df/dt = d/dvpar (Dcoll df/dvpar - Vcoll f)
!  - The operator is advected in vpar only
!  - Dcoll and Vcoll depend on the energy
!  - Dcoll and Vcoll can have a radial envelope
!
! date :  01 March 2009
!
! The choice of the collision operator depends on 
! the value of 'version_coll'
!   - = 1 : case where the parallel velocity is forced equal to 0
!           (developed by G. Dif-Pradalier) 
!   - = 2 : case where there is no more constrainst 
!           on the parallel velocity. The collision operator 
!           is refreshed every 'refresh_nbstep' iterations, 
!           by computing the mean temperature (Tmean) and 
!           the mean velocity (Vmean) which depend on 
!           (r,theta,phi). (developed by J. Abiteboul)
!   - = 3 : case 2 above with additionally nustar and the collision
!           frequency having the correct radial dependence (so far
!           only correct for ion-ion collisions).
!           (developed by G. Dif-Pradalier)
!
!  In the second case, computation of the mean 
!  velocity 'Vmean' and mean temperature 'Tmean' as :
!  . Tmean = P^{-1}[Imean0*Imean2 - Imean1*Imean1] and
!  . Vmean = P^{-1}[Imean4*Imean1 - Imean3*Imean2] with
!  . P  = Imean0*Imean4 - Imean1*Imean3.
!   where the 5 integrals are defined as:
!     Imean0=<Dpar> ; Imean1=<vpar*Dpar> ; 
!     Imean2=<vpar^2*Dpar> ; 
!     Imean3=<1/Bstar*d/dvpar(Bstar*Dpar)> 
!   and Imean4=<1/Bstar*d/dvpar(Bstar*vpar*Dpar>
!   The brackets <.> represents the following integral 
!    in velocity:
!  . <.> = \int \int . Jv dvpar dmu where Jv corresponds
!    to the jacobian in velocity
!-----------------------------------------------------------------
module collisions_vpar_module
  use Debug_module
  use prec_const
  use MPIutils_module
  use geometry_class
  use spline1d_class
  use fdistribu5d_class
  use wrap_alloc_module
  use init_profile_class
  use init_magnetic_class
  use clock_module
  implicit none

  !-------------------------------------
  !  PRIVATE VARIABLES
  !-------------------------------------
  ! nustar coefficient for each species
  real(RKIND), dimension(:), pointer, public :: nustars

  ! used for the collision operator
  real(RKIND), dimension(:,:), pointer, private :: AA_coll_rvpar
  real(RKIND), dimension(:,:), pointer, private :: BB_coll_rvpar
  real(RKIND), dimension(:,:), pointer, private :: CC_coll_rvpar
  real(RKIND), dimension(:)  , pointer, private :: Dcoll_begin
  real(RKIND), dimension(:)  , pointer, private :: Dcoll_end
  real(RKIND), dimension(:)  , pointer, private :: Vcoll_begin
  real(RKIND), dimension(:)  , pointer, private :: Vcoll_1
  real(RKIND), dimension(:)  , pointer, private :: Vcoll_nvm1
  real(RKIND), dimension(:)  , pointer, private :: Vcoll_end
  real(RKIND), dimension(:)  , pointer, private :: Dsemi_begin
  real(RKIND), dimension(:)  , pointer, private :: Dsemi_nvm1

  ! used for computation of  Tmean, Vmean and Pmean
  integer, private                                :: Fmean_BCvpar_left
  integer, private                                :: Fmean_BCvpar_right
  type(pspline1d), dimension(:), pointer, private :: Fmean_pspline1d_theta
  type(pspline1d), dimension(:), pointer, private :: Fmean_pspline1d_phi
  type(nspline1d), dimension(:), pointer, private :: Fmean_nspline1d_vpar
  !--> for H(r,theta)
  real(RKIND), dimension(:,:)  , pointer, private :: Hrtheta
  !--> for d(Bstar*Dpar)/dvpar and vpar*d(Bstar*Dpar)/dvpar
  real(RKIND), dimension(:,:,:), pointer, private :: deriv_BstarDpar
  real(RKIND), dimension(:,:,:), pointer, private :: deriv_vBstarDpar
  !-> 5 temporary integrals for computation of Vmean(r,theta,phi) 
  !  and Tmean(r,theta,phi) (used for version_coll = 2 & 3)
  real(RKIND), dimension(:,:,:), pointer, private :: Imean0
  real(RKIND), dimension(:,:,:), pointer, private :: Imean1
  real(RKIND), dimension(:,:,:), pointer, private :: Imean2
  real(RKIND), dimension(:,:,:), pointer, private :: Imean3
  real(RKIND), dimension(:,:,:), pointer, private :: Imean4

  !-------------------------------------
  !  PUBLIC VARIABLES
  !------------------------------------- 
  public
  ! -> variable for initial collision parameter saving
  real(RKIND), dimension(:,:), pointer :: Dcoll_rvpar
  real(RKIND), dimension(:,:), pointer :: Dsemi_rvpar
  real(RKIND), dimension(:,:), pointer :: Vcoll_rvpar
  real(RKIND), dimension(:)  , pointer :: collision_freqr_loc
  real(RKIND), dimension(:)  , pointer :: envelope_r_loc
  real(RKIND), dimension(:)  , pointer :: collision_freqr_diag
  real(RKIND), dimension(:)  , pointer :: nustarr_diag
  real(RKIND), dimension(:)  , pointer :: envelope_r_diag

  !-> Vmean(r) and Tmean(r)
  real(RKIND), dimension(:)  , pointer :: Vmean_r_avg_diag
  real(RKIND), dimension(:)  , pointer :: Tmean_r_avg_diag
  real(RKIND), dimension(:)  , pointer :: Vmean_r_diag
  real(RKIND), dimension(:)  , pointer :: Tmean_r_diag
  real(RKIND), dimension(:,:), pointer :: Vmean_thetaphi_diag
  real(RKIND), dimension(:,:), pointer :: Tmean_thetaphi_diag

  include "gyrokinetic/Bstar_inline.h"
  include "gyrokinetic/velocities_inline.h"
  !******************************
  contains
  !******************************
#include "gyrokinetic/Bstar_inline.F90"
#include "gyrokinetic/velocities_inline.F90"

  !-------------------------------------------------
  ! Constructor 
  !-------------------------------------------------
  subroutine new_collisions_vpar(geom,init_prof,init_magnet)
    use globals, only : istart, iend, version_coll
    use globals, only : Nspecies, hat_Zs, nustar
    implicit none
!R3 #include "code_instrumentation/r3_info.h" !R3
    type(geometry)     , intent(in) :: geom
    type(init_profile) , intent(in) :: init_prof
    type(init_magnetic), intent(in) :: init_magnet

    integer     :: ispec
    real(RKIND) :: Ts_on_Tsprime
    real(RKIND) :: hatZsprime_on_hatZs

!R3 call r3_info_begin (r3_info_index_1, 'Init_collvpar') !R3
    call Write_dyn_balise_open('Init_collvpar') !R4
    !-> for the computation of Tmean, Vmean and Pmean
    call new_fmaxwellian(geom,init_magnet)

    !-> for collision operator 
    call FTR_ALLOC_1_ZERO(nustars,0,Nspecies-1)
    if (version_coll.eq.1) then
      call FTR_ALLOC_2_ZERO(AA_coll_rvpar,istart,iend,0,geom%Nvpar)
      call FTR_ALLOC_2_ZERO(BB_coll_rvpar,istart,iend,0,geom%Nvpar)
      call FTR_ALLOC_2_ZERO(CC_coll_rvpar,istart,iend,0,geom%Nvpar)
      call FTR_ALLOC_1_ZERO(Dcoll_begin,istart,iend)
      call FTR_ALLOC_1_ZERO(Dcoll_end,istart,iend)
      call FTR_ALLOC_1_ZERO(Vcoll_begin,istart,iend)
      call FTR_ALLOC_1_ZERO(Vcoll_1,istart,iend)
      call FTR_ALLOC_1_ZERO(Vcoll_nvm1,istart,iend)
      call FTR_ALLOC_1_ZERO(Vcoll_end,istart,iend)
      call FTR_ALLOC_1_ZERO(Dsemi_begin,istart,iend)
      call FTR_ALLOC_1_ZERO(Dsemi_nvm1,istart,iend)
      call FTR_ALLOC_2_ZERO(Dcoll_rvpar,istart,iend,0,geom%Nvpar)
      call FTR_ALLOC_2_ZERO(Dsemi_rvpar,istart,iend,0,geom%Nvpar-1)
      call FTR_ALLOC_2_ZERO(Vcoll_rvpar,istart,iend,0,geom%Nvpar)
      call FTR_ALLOC_1_ZERO(collision_freqr_loc,istart,iend)
    end if
    call FTR_ALLOC_1_ZERO(envelope_r_loc,istart,iend)
    call FTR_ALLOC_1_ZERO(collision_freqr_diag,0,geom%Nr)
    call FTR_ALLOC_1_ZERO(nustarr_diag,0,geom%Nr)
    call FTR_ALLOC_1_ZERO(envelope_r_diag,0,geom%Nr)

!baoter (need to be check)
!VG!    !*** nustars(0) = nustar corresponding to the main ion species ***
!VG!    !***   while nustar for the other species is defined           ***
!VG!    !***   in function of the main ion species as                  ***
!VG!    !*** nustar_sprime = ns(rp)/nsprime(rp)*(Zsprime/Zs)^4 *       ***   
!VG!    !***                  (Ts/Tsprime)^2*nustars                   ***
!VG!    nustars(0) = nustar
!VG!    do ispec = 1,Nspecies-1
!VG!      nustars(ispec) = (init_prof%ns0_rp(ispec)/init_prof%ns0_rp(0)) * &
!VG!        (init_prof%Ts0_rp(0)/init_prof%Ts0_rp(ispec)) * &
!VG!        (hat_Zs(ispec)/hat_Zs(0))**4 * nustars(0)
!VG!    end do

    !*** nustars(0) = nustar corresponding to the      ***
    !***   main ion species while nustar for the other ***
    !***   species is defined in function of the main  ***
    !***    ion species as                             ***
    !*** nustar_sprime = (Zsprime/Zs)^2 *              ***
    !***                 (Ts/Tsprime)^2*nustars        ***
    if (.not.memory_test) then
      nustars(0)    = nustar
      do ispec = 1,Nspecies-1
        Ts_on_Tsprime       = init_prof%Ts0_rp(0) / &
          init_prof%Ts0_rp(ispec) 
        hatZsprime_on_hatZs = hat_Zs(ispec)/hat_Zs(0)
        nustars(ispec) = hatZsprime_on_hatZs*hatZsprime_on_hatZs * &
          Ts_on_Tsprime*Ts_on_Tsprime * nustars(0)
      end do
    end if
!eaoter
    call Write_dyn_balise_close('Init_collvpar') !R4
!R3 call r3_info_end (r3_info_index_1) !R3
  end subroutine new_collisions_vpar


  !-------------------------------------------------
  ! Destructor 
  !-------------------------------------------------
  subroutine delete_collisions_vpar
    use globals, only : version_coll
    !-> for the computation of Tmean, Vmean and Pmean
    call delete_fmaxwellian()

    !-> for collision operator 
    call FTR_DEALLOC(nustars)
    if (version_coll.eq.1) then
      call FTR_DEALLOC(AA_coll_rvpar)
      call FTR_DEALLOC(BB_coll_rvpar)
      call FTR_DEALLOC(CC_coll_rvpar)
      call FTR_DEALLOC(Dcoll_begin)
      call FTR_DEALLOC(Dcoll_end)
      call FTR_DEALLOC(Vcoll_begin)
      call FTR_DEALLOC(Vcoll_1)
      call FTR_DEALLOC(Vcoll_nvm1)
      call FTR_DEALLOC(Vcoll_end)
      call FTR_DEALLOC(Dsemi_begin)
      call FTR_DEALLOC(Dsemi_nvm1)
      call FTR_DEALLOC(Dcoll_rvpar)
      call FTR_DEALLOC(Dsemi_rvpar)
      call FTR_DEALLOC(Vcoll_rvpar)
      call FTR_DEALLOC(collision_freqr_loc)
    end if
    call FTR_DEALLOC(envelope_r_loc)
    call FTR_DEALLOC(collision_freqr_diag)
    call FTR_DEALLOC(nustarr_diag)
    call FTR_DEALLOC(envelope_r_diag)
  end subroutine delete_collisions_vpar


  !-----------------------------------------------------------
  ! Constructor for the computation of Tmean, Vmean and Pmean
  !-----------------------------------------------------------
  subroutine new_fmaxwellian(geom,init_magnet)
    use globals, only : BC_Hermite, Nbthread, &
      istart, iend, Nr, jstart, jend, Ntheta, Nphi, memory_test, &
      version_coll, Tmean_loc, Vmean_loc, &
      Tmean_loc_restart_save, Vmean_loc_restart_save, &
      asynchrone_writing
    use geometry_class
    use wrap_alloc_module
    use OMPutils_module, only : Romp1_0Nvpar, Romp2_0Nvpar
    type(geometry)     , intent(in) :: geom
    type(init_magnetic), intent(in) :: init_magnet

    integer :: ithread, ir, itheta

    !*** initialisation of the splines ***
    Fmean_BCvpar_left  = BC_Hermite
    Fmean_BCvpar_right = BC_Hermite
    allocate(Fmean_pspline1d_theta(1:Nbthread))
    allocate(Fmean_pspline1d_phi(1:Nbthread))
    allocate(Fmean_nspline1d_vpar(1:Nbthread))
    do ithread = 1,Nbthread
       call write_expr("n", "geom%Ntheta")
       call new_spline1d_period(Fmean_pspline1d_theta(ithread),&
            geom%Ntheta,geom%dtheta)
       call write_expr("n", "geom%Nphi")
       call new_spline1d_period(Fmean_pspline1d_phi(ithread),&
            geom%Nphi,geom%dphi)
       call write_expr("n", "geom%Nvpar")
       call new_spline1d_natural(Fmean_nspline1d_vpar(ithread),&
            geom%Nvpar,geom%dvpar)
    end do

    !*** array allocations ***
    if ((version_coll.eq.2).or.(version_coll.eq.3)) then
      !--> for integrals to compute Pmean(r,theta,phi), 
      ! Tmean(r,theta,phi) and Vmean(r,theta,phi)
      call FTR_ALLOC_3_ZERO(Imean0,istart,iend,jstart,jend,0,Nphi)
      call FTR_ALLOC_3_ZERO(Imean1,istart,iend,jstart,jend,0,Nphi)
      call FTR_ALLOC_3_ZERO(Imean2,istart,iend,jstart,jend,0,Nphi)
      call FTR_ALLOC_3_ZERO(Imean3,istart,iend,jstart,jend,0,Nphi)
      call FTR_ALLOC_3_ZERO(Imean4,istart,iend,jstart,jend,0,Nphi)
    end if

    !--> for Tmean(r,theta,phi) and Vmean(r,theta,phi)
    call FTR_ALLOC_3_ZERO(Vmean_loc,istart,iend,jstart,jend,0,Nphi)
    call FTR_ALLOC_3_ZERO(Tmean_loc,istart,iend,jstart,jend,0,Nphi)
    if (asynchrone_writing) then
      call FTR_ALLOC_3_ZERO(Vmean_loc_restart_save,istart,iend,jstart,jend,0,Nphi)
      call FTR_ALLOC_3_ZERO(Tmean_loc_restart_save,istart,iend,jstart,jend,0,Nphi)
    endif
    call FTR_ALLOC_1_ZERO(Vmean_r_avg_diag,0,Nr)
    call FTR_ALLOC_1_ZERO(Tmean_r_avg_diag,0,Nr)
    if ((version_coll.eq.2).or.(version_coll.eq.3)) then
      call FTR_ALLOC_1_ZERO(Vmean_r_diag,0,Nr)
      call FTR_ALLOC_1_ZERO(Tmean_r_diag,0,Nr)
      call FTR_ALLOC_2_ZERO(Vmean_thetaphi_diag,0,Ntheta,0,Nphi)
      call FTR_ALLOC_2_ZERO(Tmean_thetaphi_diag,0,Ntheta,0,Nphi)
    end if
  end subroutine new_fmaxwellian


  !-----------------------------------------------------------
  ! Destructor for the computation of Tmean, Vmean and Pmean
  !-----------------------------------------------------------
  subroutine delete_fmaxwellian()
    use globals, only : Nbthread, version_coll, &
      Tmean_loc, Vmean_loc, asynchrone_writing, &
      Tmean_loc_restart_save, Vmean_loc_restart_save

    integer :: ithread

    !*** deallocation of the splines ***
    do ithread = 1,Nbthread    
      call del_spline1d_natural(Fmean_nspline1d_vpar(ithread))
      call del_spline1d_period(Fmean_pspline1d_phi(ithread))
      call del_spline1d_period(Fmean_pspline1d_theta(ithread))
    end do
    deallocate(Fmean_nspline1d_vpar)
    deallocate(Fmean_pspline1d_phi)
    deallocate(Fmean_pspline1d_theta)
    !*** deallocation of the array for radial profile saving ***
    if ((version_coll.eq.2).or.(version_coll.eq.3)) then
      call FTR_DEALLOC(Imean0)
      call FTR_DEALLOC(Imean1)
      call FTR_DEALLOC(Imean2)
      call FTR_DEALLOC(Imean3)
      call FTR_DEALLOC(Imean4)
    end if
    call FTR_DEALLOC(Vmean_loc)
    call FTR_DEALLOC(Tmean_loc)
    if (asynchrone_writing) then
      call FTR_DEALLOC(Vmean_loc_restart_save)
      call FTR_DEALLOC(Tmean_loc_restart_save)
    endif
    call FTR_DEALLOC(Vmean_r_avg_diag)
    call FTR_DEALLOC(Tmean_r_avg_diag)
    if ((version_coll.eq.2).or.(version_coll.eq.3)) then
      call FTR_DEALLOC(Vmean_r_diag)
      call FTR_DEALLOC(Tmean_r_diag)
      call FTR_DEALLOC(Vmean_thetaphi_diag)
      call FTR_DEALLOC(Tmean_thetaphi_diag)
    end if
  end subroutine delete_fmaxwellian


  !----------------------------------------------------------------
  ! Initialisation of the collision operator, according to
  !  the value of 'version_coll':
  !   - = 1 : case where the parallel velocity is forced equal to 0
  !           (developed by G. Dif-Pradalier)
  !   - = 2 : case where there is no more constrainst 
  !           on the parallel velocity. The collision operator 
  !           is refreshed every 'refresh_nbstep' iterations, 
  !           by computing the mean temperature (Tmean) and 
  !           the mean velocity (Vmean) which depend on 
  !           (r,theta,phi). (developed by J. Abiteboul)
  !   - = 3 : case 2 above with additionally nustar and 
  !           the collision
  !           frequency having the correct radial dependence (so far
  !           only correct for ion-ion collisions).
  !           (developed by G. Dif-Pradalier)
  !----------------------------------------------------------------
  subroutine init_collisions_vpar(geom,init_prof, &
    init_magnet,init_curr,f)
    use globals, only : pSpecies_id, &
      restart, istart, iend, jstart, jend, Nphi, &
      version_coll, coef_nustar, &
      Nvpar, Nr, TGCseq_r, Tmean_loc, Vmean_loc, &
      dom_r, mpi_comm_mu
    use mpiMod
    implicit none
    type(geometry)     , intent(in) :: geom
    type(init_profile) , intent(in) :: init_prof
    type(init_magnetic), intent(in) :: init_magnet
    type(init_current) , intent(in) :: init_curr
    type(fdistribu5d)  , intent(in) :: f

    integer     :: ir, itheta, iphi, ilim, ierr
    real(RKIND) :: correct_loc, correct_glob

    real(RKIND), dimension(:,:,:), pointer :: &
      Tmean_tmp, Vmean_tmp

    ! used in compute_DcollVcoll
    real(RKIND), dimension(:), pointer :: Tprofile_init
    real(RKIND), dimension(:), pointer :: Vprofile_init

    call clck_time(bclock_init)
    allocate(Tmean_tmp(istart:iend,jstart:jend,0:Nphi))
    allocate(Vmean_tmp(istart:iend,jstart:jend,0:Nphi))
    allocate(Tprofile_init(istart:iend))
    allocate(Vprofile_init(istart:iend))

    !*** Initialisation of the radial profile for ***
    !***  the collisions                          ***
    if (coef_nustar.ne.0._RKIND) then
      call init_Collisions_vpar_RadialEnvelope(geom, &
        envelope_r_loc)
    else
      envelope_r_loc(istart:iend) = 1._RKIND 
    end if

    !*** choice of the version of the collision operator in use ***
    select case (version_coll)
    case (1)
      !*** initialisation of T and V ***
      do ir = istart,iend
        Tprofile_init(ir) = TGCseq_r(ir)
        Vprofile_init(ir) = 0._RKIND
      end do
      !*** compute Dcoll and Vcoll ***
      call compute_DcollVcoll_vers1(geom,init_prof,init_magnet, &
        Dcoll_rvpar,Dsemi_rvpar,Vcoll_rvpar,Tprofile_init, &
        Vprofile_init)
      !***  initialisation of the Crank-Nicolson matrix ***
      call construct_CrankNic_vers1(geom,Dcoll_rvpar, &
        Dsemi_rvpar,Vcoll_rvpar, AA_coll_rvpar, &
        BB_coll_rvpar,CC_coll_rvpar)
      !*** reconstruction of the global P, V and T ***
      !***  profiles for diagnostic saving         ***
      call save_glob_coll()

    case (2,3)
      !*** initialisation of T and V ***
      if (.not.restart(pSpecies_id)) then 
        do iphi = 0,Nphi
          do itheta = jstart,jend
            do ir = istart,iend
              Tmean_loc(ir,itheta,iphi) = TGCseq_r(ir)
              Vmean_loc(ir,itheta,iphi) = 0._RKIND
              Tmean_tmp(ir,itheta,iphi) = 0._RKIND
              Vmean_tmp(ir,itheta,iphi) = 0._RKIND
            end do
          end do
        end do
        !*** computation of the mean value of T, V and P ***
        call compute_mean_TandV_vers2(geom,init_prof, &
          init_magnet,init_curr,f)

        !*** We iterate the computation until it converges
        ilim         = 0
        correct_glob = 1._RKIND
        do while ((correct_glob.ge.1e-10_RKIND).and.(ilim.lt.20))
          call compute_mean_TandV_vers2(geom,init_prof, &
            init_magnet,init_curr,f)
          correct_loc = &
            max(maxval(abs(Tmean_tmp-Tmean_loc)), & 
            maxval(abs(Vmean_tmp-Vmean_loc)))
          !--> Communicate between processors with same mu value
          call MY_Allreduce(correct_loc,correct_glob,1, &
            MPI_REAL8,MPI_MAX,mpi_comm_mu,ierr)
          do iphi = 0,Nphi
            do itheta = jstart,jend
              do ir = istart,iend
                Tmean_tmp(ir,itheta,iphi) = &
                  Tmean_loc(ir,itheta,iphi)
                Vmean_tmp(ir,itheta,iphi) = &
                  Vmean_loc(ir,itheta,iphi)
              end do
            end do
          end do
          ilim = ilim+1
        end do
        if (pglobal_id.eq.0) then
          write(6,*) 'Number of iterations necessary to compute ', &
            'the reference Maxwellian for the collision operator', &
            ilim, 'to obtain a precision of ', correct_glob
        end if
      end if
      !*** reconstruction of the global P, V and T ***
      !***  profiles for diagnostic saving         ***
      call save_glob_coll_vers2(geom,init_prof)
    end select
    deallocate(Tmean_tmp)
    deallocate(Vmean_tmp)
    deallocate(Tprofile_init)
    deallocate(Vprofile_init)
    call clck_time(eclock_init)
    call clck_diff(bclock_init,eclock_init, &
      global_time_init)
  end subroutine init_collisions_vpar


  !------------------------------------------------------------
  ! Recalculation of the collision operator 
  !  after that Tmean and Vmean have been computed:
  !  - new values for Dcoll_rvpar, Vcoll_rvpar and Dsemi_rvpar
  !  - construction of the new Crank-Nicolson matrix
  !------------------------------------------------------------
  subroutine refresh_collisions_vpar(geom,init_prof, &
    init_magnet,init_curr,f)
    use globals, only : version_coll
    use clock_module
    implicit none
    type(geometry)     , intent(in) :: geom
    type(init_profile) , intent(in) :: init_prof
    type(init_magnetic), intent(in) :: init_magnet
    type(init_current) , intent(in) :: init_curr
    type(fdistribu5d)  , intent(in) :: f
!R3 #include "code_instrumentation/r3_info.h" !R3

!R3 call r3_info_begin (r3_info_index_0, 'CPU_refresh_coll_vpar') !R3
    call Write_dyn_balise_open('CPU_refresh_coll_vpar') !R4

    call clck_time(bclock_coll_refresh)

    !*** choice of the version of the collision operator in use ***
    if ((version_coll.eq.2).or.(version_coll.eq.3)) then
      !*** computation of the mean value of T, V and P ***
      call compute_mean_TandV_vers2(geom,init_prof, &
        init_magnet,init_curr,f)

      !*** reconstruction of the global P, V and T ***
      !***  profiles for diagnostic saving         ***
      call save_glob_coll_vers2(geom,init_prof)
    else
      !*** not yet implemented ***
      if (pglobal_id.eq.0) then
        print*,'*******************************', &
          '************************'
        print*,'******  version_coll = ', version_coll, &
          '        ******'
        print*,'******     is not yet implemented', &
          '                ******' 
        print*,'******   change integer "version_coll".', &
          '          ******'
        print*,'******          CODE IS HALTED', &
          '                   ******'
        print*,'*******************************', &
          '************************'
        stop
      end if
    end if

    call clck_time(eclock_coll_refresh)
    call clck_diff(bclock_coll_refresh,eclock_coll_refresh,&
      global_time_coll_refresh)
    call Write_dyn_balise_close('CPU_refresh_coll_vpar') !R4
!R3 call r3_info_end (r3_info_index_0) !R3
  end subroutine refresh_collisions_vpar


  !--------------------------------------------------------
  ! initialisation of a collision radial envelope used to 
  ! magnify the collisions (create physical buffer zones)
  ! only if 'coef_nustar' not equal to 0
  !  - at position parametrised by buffer_stiffness
  !     and buffer_stiffness from the edges of 
  !    the simulation box
  !  - with magnitude parametrised by coef_nustar
  !--------------------------------------------------------
  subroutine init_Collisions_vpar_RadialEnvelope(geom,envelope)
    use globals, only : istart, iend, Nr, &
      buffer_asymmetry, buffer_stiffness, buffer_location, &
      coef_nustar
    use globals, only : pSpecies_id, rflux_driven
    use init_profile_class
    use geometry_class
    implicit none
    type(geometry)          , intent(in)  :: geom
    real(RKIND), &
      dimension(istart:iend), intent(out) :: envelope

    integer     :: ir
    real(RKIND) :: rho_loc
    real(RKIND) :: asymmetric, cst
    real(RKIND) :: mask, shape
    
    !*** asymmetric=0.5 if flux_driven and buffer_asymmetry=true,
    !*** asymmetric=1.0 otherwise
    !*** => leads to larger buffer region at the inner boundary
    !***    when the source is active
    if (buffer_asymmetry) then
      asymmetric = 1._RKIND-rflux_driven/2._RKIND
    else
      asymmetric = 1._RKIND
    end if
    cst = 1.e-7_RKIND

    !*** Smooth magnification at the edges       ***
    !*** => sharpness is controled by            ***
    !***     buffer_stiffness                    ***
    !*** => size of buffer zones is governed by  ***
    !***     buffer_location                     ***
    do ir = istart,iend
      rho_loc       = (geom%rg(ir)-geom%rg(0))/geom%Lr
      shape         = ON + HF*(tanh((rho_loc- &
        (ON-asymmetric*buffer_location))/buffer_stiffness) - &
        tanh((rho_loc-buffer_location)/buffer_stiffness))
      mask          = 0.5_RKIND*(1._RKIND + &
        sign(1._RKIND,shape-cst))
      envelope(ir) = (mask*shape) * &
        (coef_nustar-1._RKIND)+1._RKIND
    end do
  end subroutine init_Collisions_vpar_RadialEnvelope


  !--------------------------------------------------------------
  !  Compute Collisional terms: Dcoll, Dsemi and  Vcoll
  !    for each value of mu*B
  !    - Dcoll & Vcoll are defined on vpar grid values
  !    - Dsemi on vpar/2 grid values (for Crank-Nicolson scheme)
  !      such that Dsemi(j) = Dcoll(j+1/2)
  !
  !  vpar is normalised by vT=sqrt(T/m)
  !  vpar0 = vpar/sqrt(2)
  !--------------------------------------------------------------
  subroutine compute_DcollVcoll_vers1(geom,init_prof,init_magnet, &
    Dcoll,Dsemi,Vcoll,Tprofile,Vprofile)
    use globals, only : a, R0, istart, iend, &
      Nr, jstart, jend, Nphi, Nvpar, mu_id, coef_nustar
    use globals, only : hat_As, pSpecies_id, major_species, &
      version_coll
    use geometry_class
    use init_profile_class
    use init_magnetic_class
    implicit none
    real(RKIND), &
      dimension(istart:iend,0:Nvpar)  , intent(out) :: Dcoll
    real(RKIND), &
      dimension(istart:iend,0:Nvpar)  , intent(out) :: Vcoll
    real(RKIND), &
      dimension(istart:iend,0:Nvpar-1), intent(out) :: Dsemi
    type(init_profile)                , intent(in)  :: init_prof
    type(init_magnetic)               , intent(in)  :: init_magnet
    type(geometry)                    , intent(in)  :: geom
    real(RKIND), &
      dimension(istart:iend)          , intent(in)  :: Tprofile
    real(RKIND), &
      dimension(istart:iend)          , intent(in)  :: Vprofile

#ifdef HAS_COLL_V1
    real(RKIND) :: inv_sqrtAs, nustar_sp
    real(RKIND) :: q_on_qpic, rpic_on_r
    real(RKIND) :: n_on_npic, Tpic_on_T, nustar_prof
    real(RKIND) :: derf
    real(RKIND) :: gauss, gauss_semi, temp
    real(RKIND) :: cst, dx, x, rhostar, normnustar
    real(RKIND) :: epstemp, energy_even, energy_odd
    integer     :: ix, Nx, ivpar, ivend, ir
    integer     :: ieven, iodd, ifill, isymm, iMiddle
    real(RKIND), dimension(istart:iend)           :: AD, vT, Baxis
    real(RKIND), dimension(istart:iend,0:Nvpar  ) :: phi
    real(RKIND), dimension(istart:iend,0:Nvpar  ) :: phiprime
    real(RKIND), dimension(istart:iend,0:Nvpar  ) :: psi
    real(RKIND), dimension(istart:iend,0:2*Nvpar) :: vtemp

    !*** Initialisation of nustar value depending ***
    !***  on each species                         ***
    nustar_sp = nustars(pSpecies_id)

    !***  Initialisation: null array ***
    collision_freqr_loc = 0._RKIND
    !***  Normalisations:                              ***
    !***  - vparg is normalised to thermal speed       ***
    !***    = rho_i / tau_Larmor                       ***
    !***    = sqrt(Tir_glob)(r))                       ***
    !***  - phi, psi are given in eqs.                 ***
    !***     (2.13.1) & (2.13.2) in Wesson             ***
    !***  - AD = uT^3/Psi(1/sqrt(2)) *                 ***
    !***      nustar(r)*rhostar/q*a/R*epsilon^(3/2)    ***
    !***                                               ***
    !***  values of theta are taken on the equatorial  ***
    !***   plane (theta=0)                             ***
    cst         = 2._RKIND/sqrt(pi)
    rhostar     = 1._RKIND/a
    normnustar  = 3._RKIND/cst
    inv_sqrtAs  = 1._RKIND/sqrt(hat_As(pSpecies_id))
    nustar_prof = 1._RKIND
    do ir = istart,iend
      vT(ir)      = sqrt(Tprofile(ir))
      Baxis(ir)   = init_magnet%B_norm(ir,0)
      epstemp     = geom%rg(ir)*inv_R(ir,0)
      !--> Compute the radial dependence of nustar
      if (version_coll==3) then
        !---!! computed with the INITIAL profile !!---
        !---!!  of temperature                   !!---
        q_on_qpic   = 1._RKIND/(init_prof%q_rp*init_prof%iota(ir))
        rpic_on_r   = init_prof%rpeak / geom%rg(ir)
        n_on_npic   = init_prof%ns0(major_species,ir) / &
          init_prof%ns0_rp(major_species)
        Tpic_on_T   = init_prof%Ts0_rp(major_species) / &
          init_prof%Ts0(major_species,ir) ! Tprofile(ir)
        nustar_prof = q_on_qpic*rpic_on_r*sqrt(rpic_on_r) * &
          n_on_npic*Tpic_on_T
      end if
      !--> AD(r)   = (1/sqrt(hat_As)) * 
      !-->           (3*sqrt(pi)/2)*nustar(r)*vT(r)^3 * 
      !-->           eps^(3/2)/(q R0)
      !*** BECAREFUL: collision_freqr is not consistent with    ***
      !***         the definition of nustar for version_coll=3. *** 
      !***         It should be defined with Poisson logarithm. ***
      !--> nuss(r) = (1/sqrt(hat_As))*nustar*vT(r)*eps^(3/2)/(q R0)
      if (init_prof%iota(ir).ne.0._RKIND) then
        AD(ir)                   = inv_sqrtAs*normnustar * &
          nustar_sp*nustar_prof *vT(ir)**3*epstemp * &
          sqrt(epstemp)* init_prof%iota(ir)/R0
        collision_freqr_loc(ir)  = inv_sqrtAs*nustar_sp*vT(ir) * &
          epstemp*sqrt(epstemp)* init_prof%iota(ir)/R0
      else
        ! specific case if iota = 0 (slab geometry)
        AD(ir)                   = inv_sqrtAs*normnustar * &
          nustar_sp*nustar_prof *vT(ir)**3*epstemp*sqrt(epstemp)/R0
        collision_freqr_loc(ir)  = inv_sqrtAs*nustar_sp*vT(ir) * &
          epstemp*sqrt(epstemp)/R0
      end if
    end do

    !*** The diffusion and the drag operators are function of *** 
    !*** the energy                                           ***
    !*** Construction of a temporary vector for the total     ***
    !*** velocity. B is taken on the magnetic axis.           ***
    !*** Contains vpar_j and vpar_j+1/2 ; normalised to vT    ***
    do ivpar = 0,Nvpar-1
      ieven = 2*ivpar
      iodd  = ieven + 1
      do ir = istart,iend
        energy_even     = 0.5_RKIND*geom%vparg(ivpar)**2 + &
          geom%mug(mu_id)*Baxis(ir)
        vtemp(ir,ieven) = sqrt(energy_even)/vT(ir)
        energy_odd      = 0.125_RKIND*(geom%vparg(ivpar) + &
          geom%vparg(ivpar+1))**2 + geom%mug(mu_id)*Baxis(ir)
        vtemp(ir,iodd)  = sqrt(energy_odd)/vT(ir)
      end do
    end do
    ivend = 2*Nvpar
    do ir = istart,iend
      energy_even     = 0.5_RKIND*geom%vparg(Nvpar)**2 + &
        geom%mug(mu_id)*Baxis(ir)
      vtemp(ir,ivend) = sqrt(energy_even)/vT(ir)
    end do
    !*** Two cases: can vpar_j or vpar_j+1/2 vanish ?          ***
    !***                                                       ***
    !***  when mu = 0                                          ***
    !***  - if Nvpar is odd (even number of vpar grid points)  ***
    !***      vtemp(Nvpar)=0 ==> pb with Dsemi                 ***
    !***  - if Nvpar is even (odd number of vpar grid points)  ***
    !***      vtemp(Nvpar)=0 ==> pb with Dcoll & Vcoll         ***
    !***                                                       ***
    !*** Here, uses the even part of vtemp for Dcoll & Vcoll   ***
    ifill = 0
    do ivpar = 0,Nvpar-1,2
      do ir = istart,iend
        !*** computes the integral                      ***
        !*** phi(x) = 2/sqrt(pi) int_0^x exp(-y^2) dy   ***
        !***  with x = v = sqrt(E/T)                    ***
        !*** using the derf FORTRAN function ***
        phi(ir,ifill)      = derf(vtemp(ir,ivpar))
        !*** computes phi'(x) = 2/sqrt(pi) exp(-x^2) ***
        !***  with x = sqrt(E) / vT                  ***
        phiprime(ir,ifill) = cst * &
          exp(-vtemp(ir,ivpar)*vtemp(ir,ivpar))
        !*** computes psi(x) = (phi(x) - x phi'(x)) / (2 x^2) ***
        !*** with x = sqrt(E) / vT                            ***
        psi(ir,ifill)      = (phi(ir,ifill) - &
          vtemp(ir,ivpar)*phiprime(ir,ifill)) /&
          (2._RKIND*vtemp(ir,ivpar)*vtemp(ir,ivpar))
        !*** computes Dcoll = AD (phi(v) - psi(v)) / v   ***
        !*** computes Vcoll = -AD (1+m/m1) vpar/vT**2 *  ***
        !***                  (phi(v) - psi(v)) / v      ***
        Dcoll(ir,ifill)    = AD(ir) * (phi(ir,ifill) - &
          psi(ir,ifill)) / (2._RKIND*vtemp(ir,ivpar))
        Vcoll(ir,ifill)    = - Dcoll(ir,ifill) * &
          (geom%vparg(ifill) - Vprofile(ir)) / (vT(ir)**2)
        !*** One completes by symmetry        ***
        !***  - Dcoll, Dsemi are even in vpar ***
        !***  - Vcoll is not symmetric        ***
        isymm              = Nvpar-ifill
        Dcoll(ir,isymm)    = Dcoll(ir,ifill)
        Vcoll(ir,isymm)    = -Dcoll(ir,ifill) * &
          (geom%vparg(isymm) - Vprofile(ir)) / (vT(ir)**2)
      end do
      ifill = ifill+1
    end do

    !*** Here, uses the odd part of vtemp for Dsemi ***
    ifill = 0
    do ivpar = 1,Nvpar-1,2
      do ir = istart,iend
        !*** computes the integral                           ***
        !***   phi(x) = 2/sqrt(pi) int_0^x exp(-y^2) dy      ***
        !***    with x = vpar / (sqrt(2)*vT)                 ***
        !*** using the derf FORTRAN function, which already  ***
        !*** contains the 2/sqrt(pi) normalisation factor    *** 
        phi(ir,ifill)      = derf(vtemp(ir,ivpar))
        !*** computes phi'(x) = 2/sqrt(pi) exp(-x^2)
        !*** with x = sqrt(E) / vT
        phiprime(ir,ifill) = cst * &
          exp(-vtemp(ir,ivpar)*vtemp(ir,ivpar))
        !*** computes psi(x) = (phi(x) - x phi'(x)) / (2 x^2)
        !*** with x = sqrt(E) / vT
        psi(ir,ifill)      = (phi(ir,ifill) - &
          vtemp(ir,ivpar)*phiprime(ir,ifill)) / &
          (2._RKIND*vtemp(ir,ivpar)*vtemp(ir,ivpar))
        !*** Dcoll = AD psi(vpar / (sqrt(2)*vT)) / vpar      ***
        !*** Vcoll = -AD (1+m/m1) psi(vpar/(sqrt(2)*vT1)) /  ***
        !***         (2 vT1^2)                               ***
        Dsemi(ir,ifill)    = AD(ir) * &
          (phi(ir,ifill) - psi(ir,ifill)) / &
          (2._RKIND*vtemp(ir,ivpar))
        !*** One completes by symmetry        ***
        !***  - Dcoll, Dsemi are even in vpar ***
        isymm              = Nvpar-1-ifill
        Dsemi(ir,isymm)    = Dsemi(ir,ifill)
      end do
      ifill = ifill+1
    end do

    !*** Three values lack: at grid point where vtemp(Nvpar)=0 ***
    !***   but only when mu = 0                                ***
    if (geom%mug(mu_id).eq.0) then
      if (mod(Nvpar,2).eq.0) then
        iMiddle = Nvpar/2
        do ir = istart,iend
          Dcoll(ir,iMiddle) = AD(ir) * 2._RKIND/(3._RKIND*sqrt(pi))
          Vcoll(ir,iMiddle) = (Dcoll(ir,iMiddle) *  Vprofile(ir)) /&
          (vT(ir)**2)
        end do
      else
        temp    = Nvpar/2
        iMiddle = floor(temp)
        do ir = istart,iend
          Dsemi(ir,iMiddle) = AD(ir) * 2._RKIND/(3._RKIND*sqrt(pi))
        end do
      end if
    else
      !*** No problem when mu != 0 ***
      if (mod(Nvpar,2).eq.0) then
        ivpar   = Nvpar
        iMiddle = Nvpar/2
        ifill   = iMiddle
        do ir = istart,iend
          phi(ir,ifill)      = derf(vtemp(ir,ivpar))
          phiprime(ir,ifill) = cst * &
            exp(-vtemp(ir,ivpar)*vtemp(ir,ivpar))
          psi(ir,ifill)      = (phi(ir,ifill) - &
            vtemp(ir,ivpar)*phiprime(ir,ifill)) / &
            (2._RKIND*vtemp(ir,ivpar)*vtemp(ir,ivpar))
          Dcoll(ir,ifill)    = AD(ir) * &
            (phi(ir,ifill) - psi(ir,ifill)) / &
            (2._RKIND*vtemp(ir,ivpar))
          Vcoll(ir,ifill) = (Dcoll(ir,ifill) * &
            Vprofile(ir)) /  (vT(ir)**2)
        end do
      else
        ivpar   = Nvpar
        temp    = Nvpar/2
        iMiddle = floor(temp)
        ifill   = iMiddle
        do ir = istart,iend
          phi(ir,ifill)      = derf(vtemp(ir,ivpar))
          phiprime(ir,ifill) = cst * &
            exp(-vtemp(ir,ivpar)*vtemp(ir,ivpar))
          psi(ir,ifill)      = (phi(ir,ifill) - &
            vtemp(ir,ivpar)*phiprime(ir,ifill)) / &
            (2._RKIND*vtemp(ir,ivpar)*vtemp(ir,ivpar))
          Dsemi(ir,ifill)    = AD(ir) * &
            (phi(ir,ifill) - psi(ir,ifill)) / &
            (2._RKIND*vtemp(ir,ivpar))
        end do
      end if
    end if

    !*** if 'coef_nustar' not equal to 0, the magnitude of ***
    !***   D and V is modified at the radial edges         ***
    !***     providing collisional buffer zones            ***
    if (coef_nustar.ne.0._RKIND) then
      do ir = istart,iend
        do ivpar = 0,Nvpar-1
          Dcoll(ir,ivpar) = Dcoll(ir,ivpar)*envelope_r_loc(ir)
          Dsemi(ir,ivpar) = Dsemi(ir,ivpar)*envelope_r_loc(ir)
          Vcoll(ir,ivpar) = Vcoll(ir,ivpar)*envelope_r_loc(ir)
        end do
        Dcoll(ir,Nvpar) = Dcoll(ir,Nvpar)*envelope_r_loc(ir)
        Vcoll(ir,Nvpar) = Vcoll(ir,Nvpar)*envelope_r_loc(ir)
      end do
    end if
#else 
    ! HAS_COLL_V1
    Dcoll = 0._RKIND
    Dsemi = 0._RKIND
    Vcoll = 0._RKIND
#endif 
    ! HAS_COLL_V1
  end subroutine compute_DcollVcoll_vers1


  !------------------------------------------------------
  !  Initialize matrix coefficients A, B and C used 
  !   in Crank-Nicholson and solved using tridiag in 
  !   AdvecDiff_coll_vpar
  !
  !  Here, dt is given by argument of 
  !  step0_collisions_vpar in gysela.F90
  !------------------------------------------------------
  subroutine construct_CrankNic_vers1(geom, &
    Dcoll,Dsemi,Vcoll,AA,BB,CC)
    use prec_const
    use globals, only : istart, iend, jstart, jend, Nphi, Nvpar
    use MPIutils_module
    use geometry_class
    implicit none
    type(geometry)                    , intent(in)  :: geom
    real(RKIND) , &
      dimension(istart:iend,0:Nvpar)  , intent(in)  :: Dcoll
    real(RKIND) , &
      dimension(istart:iend,0:Nvpar)  , intent(in)  :: Vcoll
    real(RKIND) , &
      dimension(istart:iend,0:Nvpar-1), intent(in)  :: Dsemi
    real(RKIND) , &
      dimension(istart:iend,0:Nvpar)  , intent(out) :: AA
    real(RKIND) , &
      dimension(istart:iend,0:Nvpar)  , intent(out) :: BB
    real(RKIND) , &
      dimension(istart:iend,0:Nvpar)  , intent(out) :: CC

#ifdef HAS_COLL_V1
    real(RKIND), dimension(istart:iend) :: coefa, coefb, coefc
    real(RKIND)                         :: alpha, beta
    real(RKIND)                         :: dvpar, dvpar2
    integer                             :: ivpar, ir
    
    !*** Minimal version: without saving Dcoll, Vcoll & Dsemi  ***
    !***  we need these values for computing AA_coll_rvpar,    ***
    !***  BB_coll_rvpar and CC_coll_rvpar for the CN scheme    ***
    do ir = istart,iend
      Dcoll_begin(ir)  = Dcoll(ir,0)
      Dcoll_end(ir)    = Dcoll(ir,Nvpar)
      Vcoll_begin(ir)  = Vcoll(ir,0)
      Vcoll_1(ir)      = Vcoll(ir,1)
      Vcoll_nvm1(ir)   = Vcoll(ir,Nvpar-1)
      Vcoll_end(ir)    = Vcoll(ir,Nvpar)
      Dsemi_begin(ir)  = Dsemi(ir,0)       ! <==> D(1/2)
      Dsemi_nvm1(ir)   = Dsemi(ir,Nvpar-1) ! <==> D(Nvpar-1/2)
    end do
    dvpar  = geom%dvpar
    dvpar2 = dvpar**2
    alpha  = 1._RKIND/(2._RKIND*dvpar2) 
    beta   = 1._RKIND/(4._RKIND*dvpar)  
    do ir = istart,iend
      do ivpar = 1,Nvpar-1
        coefa(ir)    = Dsemi(ir,ivpar-1)*alpha + &
          Vcoll(ir,ivpar-1)*beta
        coefb(ir)    = (Dsemi(ir,ivpar-1) + &
          Dsemi(ir,ivpar))*alpha
        coefc(ir)    = Dsemi(ir,ivpar)*alpha - &
          Vcoll(ir,ivpar+1)*beta
        AA(ir,ivpar) = - coefa(ir)
        BB(ir,ivpar) = coefb(ir) 
        CC(ir,ivpar) = - coefc(ir)
      end do
      ! vanishing gradient at ivpar = 0
      coefb(ir) = Dcoll_begin(ir)*2._RKIND*alpha
      AA(ir,0)  = - (2._RKIND*Dcoll_begin(ir) - &
        Dsemi_begin(ir))*alpha - &
        (2._RKIND*Vcoll_begin(ir)-Vcoll_1(ir))*beta
      BB(ir,0)  = coefb(ir) 
      CC(ir,0)  = - Dsemi_begin(ir)*alpha + Vcoll_1(ir)*beta
      ! vanishing gradient at ivpar = Nvpar
      coefb(ir)    = Dcoll_end(ir)*2._RKIND*alpha
      AA(ir,Nvpar) = - Dsemi_nvm1(ir)*alpha - Vcoll_nvm1(ir)*beta
      BB(ir,Nvpar) = coefb(ir) 
      CC(ir,Nvpar) = - (2._RKIND*Dcoll_end(ir) - &
        Dsemi_nvm1(ir))*alpha + &
        (2._RKIND*Vcoll_end(ir)-Vcoll_nvm1(ir))*beta
    end do
#else 
    ! HAS_COLL_V1
    AA = 0._RKIND
    BB = 0._RKIND
    CC = 0._RKIND
#endif 
    ! HAS_COLL_V1
  end subroutine construct_CrankNic_vers1


  !-------------------------------------------------------------- 
  ! Solves for F the equation M(A,B,C) * F = R , where A,B and C 
  ! are input vectors defined in init_CrankNich
  !  - diffusion + advection on dt
  !  - with a Crank-Nicolson scheme
  !
  ! Call for this subroutine in timesplitting_BSL (resol_steps.F90)
  !--------------------------------------------------------------- 
  subroutine AdvecDiff_coll_vers1(geom,dt,f)
    use prec_const
    use globals, only : istart, iend, jstart, jend, Nphi, Nvpar
    use MPIutils_module
#ifdef _OPENMP
    use OMPutils_module, only : omp_get_thread_num
#endif
    use OMPutils_module, only : Romp1_0Nvpar, Romp2_0Nvpar, &
      Romp3_0Nvpar, Romp4_0Nvpar, Romp5_0Nvpar, Romp6_0Nvpar
    use clock_module
    use geometry_class
    use utils_module
    implicit none
    type(geometry)     , intent(in)    :: geom
    real(RKIND)        , intent(in)    :: dt
    type(fdistribu5d)  , intent(inout) :: f

#ifdef HAS_COLL_V1
    integer     :: tid
    integer     :: ir, itheta, iphi, ivpar
    real(RKIND) :: tempf0, tempf1, tempfN, tempfNm1

    real(RKIND), dimension(:), pointer :: fvpar
    real(RKIND), dimension(:), pointer :: fcoll
    real(RKIND), dimension(:), pointer :: AAtri, BBtri, CCtri, RR

    call clck_time(bclock_collisions_vpar)
#ifdef _OPENMP
!$OMP PARALLEL private(tid,ir,itheta,iphi,ivpar,fvpar,fcoll, &
!$OMP AAtri,BBtri,CCtri,RR, &
!$OMP tempf0,tempf1,tempfN,tempfNm1) default(shared)
!$OMP BARRIER
    tid = 1+omp_get_thread_num()
#else
    tid = 1
#endif
   
    fvpar => Romp1_0Nvpar(tid)%val
    fcoll => Romp2_0Nvpar(tid)%val
    AAtri => Romp3_0Nvpar(tid)%val
    BBtri => Romp4_0Nvpar(tid)%val
    CCtri => Romp5_0Nvpar(tid)%val
    RR    => Romp6_0Nvpar(tid)%val

!$OMP DO SCHEDULE(static)
    do iphi = 0,Nphi-1
      !*** vpar advection & diffusion ***
      !*** A,B and C defining Crank-Nicolson tridiagonal matrix ***
      do itheta = jstart,jend  
        do ir = f%istart_modif,f%iend_modif 
          do ivpar = 0,Nvpar
            fvpar(ivpar) = f%values(ir,itheta,iphi,ivpar)
          end do
          do ivpar = 1,Nvpar-1
            AAtri(ivpar) = dt*AA_coll_rvpar(ir,ivpar)
            BBtri(ivpar) = ON + dt*BB_coll_rvpar(ir,ivpar)
            CCtri(ivpar) = dt*CC_coll_rvpar(ir,ivpar)
            fcoll(ivpar) = fvpar(ivpar)
            RR(ivpar)    = &
              - dt*AA_coll_rvpar(ir,ivpar)*fvpar(ivpar-1) + &
              (ON - dt*BB_coll_rvpar(ir,ivpar))*fvpar(ivpar) - &
              dt*CC_coll_rvpar(ir,ivpar)*fvpar(ivpar+1)
          end do
          ! tempory arrays in vpar for tridiag solve 
          !  at each radial position
          AAtri(0)     = dt*AA_coll_rvpar(ir,0)
          AAtri(Nvpar) = dt*AA_coll_rvpar(ir,Nvpar)
          BBtri(0)     = ON + dt*BB_coll_rvpar(ir,0)
          BBtri(Nvpar) = ON + dt*BB_coll_rvpar(ir,Nvpar)
          CCtri(0)     = dt*CC_coll_rvpar(ir,0)
          CCtri(Nvpar) = dt*CC_coll_rvpar(ir,Nvpar)
          ! vanishing gradient at ivpar = 0
          tempf0 = fvpar(0)
          tempf1 = fvpar(1)
          RR(0)  = &
            - dt*AA_coll_rvpar(ir,0)*(2._RKIND*tempf0 - tempf1)+ &
            (ON - dt*BB_coll_rvpar(ir,0))*tempf0 - &
            dt*CC_coll_rvpar(ir,0)*tempf1
          ! vanishing gradient at ivpar = Nvpar
          tempfNm1  = fvpar(Nvpar-1)
          tempfN    = fvpar(Nvpar)
          RR(Nvpar) = - dt* AA_coll_rvpar(ir,Nvpar)*tempfNm1 + &
            (ON - dt*BB_coll_rvpar(ir,Nvpar))*tempfN - &
            dt*CC_coll_rvpar(ir,Nvpar)*(2._RKIND*tempfN - tempfNm1)
          ! fcoll at 0 and Nvpar
          fcoll(0)     = fvpar(0)
          fcoll(Nvpar) = fvpar(Nvpar)
          !*** matrix inversion ***
          call tridiag(AAtri,BBtri,CCtri,RR,fcoll,Nvpar)
          do ivpar = 0,Nvpar
            f%values(ir,itheta,iphi,ivpar) = fcoll(ivpar)
          end do
        end do
      end do
    end do
!$OMP END DO
!$OMP BARRIER
!$OMP END PARALLEL

    call clck_time(eclock_collisions_vpar)
    call clck_diff(bclock_collisions_vpar,eclock_collisions_vpar, &
      global_time_collisions_vpar)
#endif 
    ! HAS_COLL_V1
  end subroutine AdvecDiff_coll_vers1


  !--------------------------------------------------------
  ! Reconstruction of the global radial profiles necessary 
  !  for the output saving in the file 
  !  "collisions_<num_diag>.h5"
  !--------------------------------------------------------
  subroutine save_glob_coll()
    use globals        , only : istart, iend, jstart, &
      dom_r, Nr, Rarray_PNr_nbM6, Rarray_Nr_nbM6, &
      mpi_comm_mu, mpi_comm_column, &
      Tmean_loc, Vmean_loc
    use MPIutils_module, only : comm_gather2D
    use mpiMod

#ifdef HAS_COLL_V1
    integer :: ir, nbval, nbelem, ierr

    !*** saving of the radial profiles of P, T and V ***
    !*** averaged in theta and phi                   ***
    nbval  = 6
    nbelem = nbval*dom_r
    do ir = istart,iend
      Rarray_PNr_nbM6(0,ir) = collision_freqr_loc(ir)
      Rarray_PNr_nbM6(1,ir) = envelope_r_loc(ir)
      Rarray_PNr_nbM6(2,ir) = Tmean_loc(ir,jstart,0)
      Rarray_PNr_nbM6(3,ir) = Vmean_loc(ir,jstart,0)
    end do
    if (jstart.eq.0) then 
!VG!      call MPI_ALLGATHER(Rarray_PNr_nbM6(0,istart), &
!VG!        nbelem,MPI_REAL8,Rarray_Nr_nbM6(0,0), &
!VG!        nbelem,MPI_REAL8,mpi_comm_column,ierr)       
      call comm_gather2D(Rarray_PNr_nbM6(0:,istart:),nbelem,0, &
        dom_r,0,Rarray_Nr_nbM6(0:,0:),mpi_comm_column)
    end if
    nbelem = nbval*(Nr+1)
    call MPI_BCAST(Rarray_Nr_nbM6,nbelem, &
      MPI_REAL8,0,mpi_comm_mu,ierr)
    do ir = 0,Nr
      collision_freqr_diag(ir) = Rarray_Nr_nbM6(0,ir)
      envelope_r_diag(ir)      = Rarray_Nr_nbM6(1,ir)
      Tmean_r_avg_diag(ir)     = Rarray_Nr_nbM6(2,ir)
      Vmean_r_avg_diag(ir)     = Rarray_Nr_nbM6(3,ir)
    end do
#endif 
    ! HAS_COLL_V1
  end subroutine save_glob_coll


  !********************************************************
  ! Subroutines for version 2 of the collision operator
  !********************************************************
  !--------------------------------------------------------
  !  Compute Collisional terms: Dcoll, Dsemi and  Vcoll
  !    for each value of mu*B
  !    - Dcoll & Vcoll are defined on vpar grid values
  !    - Dsemi on vpar/2 grid values 
  !      (for Crank-Nicolson scheme)
  !      such that Dsemi(j) = Dcoll(j+1/2)
  !
  !  vpar is normalised by vT=sqrt(T/m)
  !  vpar0 = vpar/sqrt(2)
  !--------------------------------------------------------
  subroutine compute_DcollVcoll_vers2(geom,init_prof,init_magnet, &
    ir_coll,itheta_coll,Tloc,Vloc,nustar_prof,Dcoll,Dsemi,Vcoll)
    use globals, only : a, R0, coef_nustar, Nvpar, mu_id
    use globals, only : hat_As, pSpecies_id, major_species
    use geometry_class
    use init_profile_class
    use init_magnetic_class
    implicit none
    type(init_profile)                , intent(in)  :: init_prof
    type(init_magnetic)               , intent(in)  :: init_magnet
    type(geometry)                    , intent(in)  :: geom
    integer                           , intent(in)  :: ir_coll
    integer                           , intent(in)  :: itheta_coll
    real(RKIND)                       , intent(in)  :: Tloc, Vloc
    real(RKIND)                                     :: q_on_qpic, rpic_on_r
    real(RKIND)                                     :: n_on_npic, Tpic_on_T
    real(RKIND)                       , intent(in)  :: nustar_prof
    real(RKIND), dimension(0:Nvpar  ) , intent(out) :: Dcoll, Vcoll
    real(RKIND), dimension(0:Nvpar-1) , intent(out) :: Dsemi

#ifdef HAS_COLL_V2
    real(RKIND) :: inv_sqrtAs, nustar_sp
    real(RKIND) :: derf
    real(RKIND) :: gauss, gauss_semi, temp
    real(RKIND) :: cst, dx, x, rhostar, normnustar
    real(RKIND) :: epstemp, energy_even, energy_odd
    integer     :: ix, Nx, ivpar, ivend
    integer     :: ieven, iodd, ifill, isymm, iMiddle
    real(RKIND) :: AD, vT, Bij
    real(RKIND), dimension(0:Nvpar  ) :: phi, phiprime, psi
    real(RKIND), dimension(0:2*Nvpar) :: vtemp

    !*** initialisation of nustar value depending on each species ***
    nustar_sp = nustars(pSpecies_id)

    !***  Normalisations:                         ***
    !***  - vparg is normalised to thermal speed  ***
    !***     = rho_i / tau_Larmor                 ***
    !***     = sqrt(Tir_glob)(r))                 ***
    !***  - phi, psi are given in                 ***
    !***     eqs. (2.13.1) & (2.13.2) in Wesson   ***
    !***  - AD = uT^3/Psi(1/sqrt(2))*nustar *     ***
    !***         rhostar/q*a/R*epsilon^(3/2)      ***
    !***                                          ***
    !***  values of theta are taken on the        ***
    !***  equatorial plane (theta=0)              ***
    cst        = 2._RKIND/sqrt(pi)
    rhostar    = 1._RKIND/a
    normnustar = 3._RKIND/cst

    !--> AD(r)   = (1/sqrt(hat_As)) * 
    !--->          (3*sqrt(pi)/2)*nustar(r)*vT(r)^3*eps^(3/2)/(q R0)
    inv_sqrtAs = 1._RKIND/sqrt(hat_As(pSpecies_id))
    vT         = sqrt(Tloc)
    Bij        = init_magnet%B_norm(ir_coll,itheta_coll)
    epstemp    = geom%rg(ir_coll)*inv_R(ir_coll,0)
    if (init_prof%iota(ir_coll).ne.0._RKIND) then
      AD = inv_sqrtAs * normnustar * nustar_sp * &
        nustar_prof*vT**3*epstemp* &
        sqrt(epstemp)* init_prof%iota(ir_coll)/R0
    else  
      ! specific case if iota = 0 (slab geometry)
      AD = inv_sqrtAs * normnustar * nustar_sp * &
        nustar_prof*vT**3*epstemp * sqrt(epstemp)/R0
    end if

    !*** The diffusion and the drag operators are function of *** 
    !*** the energy                                           ***
    !*** Construction of a temporary vector for the total     ***
    !*** velocity. B is taken on the magnetic axis.           ***
    !*** Contains vpar_j and vpar_j+1/2 ; normalised to vT    ***
    do ivpar = 0,Nvpar-1
      ieven = 2*ivpar
      iodd  = ieven + 1
      energy_even  = 0.5_RKIND*geom%vparg(ivpar)**2 + &
        geom%mug(mu_id)*Bij
      vtemp(ieven) = sqrt(energy_even)/vT
      energy_odd   = 0.125_RKIND*(geom%vparg(ivpar) + &
        geom%vparg(ivpar+1))**2 + geom%mug(mu_id)*Bij
      vtemp(iodd)  = sqrt(energy_odd)/vT
    end do
    ivend = 2*Nvpar
    energy_even    = 0.5_RKIND*geom%vparg(Nvpar)**2 + &
      geom%mug(mu_id)*Bij
    vtemp(ivend)   = sqrt(energy_even)/vT
    
    !*** Two cases: can vpar_j or vpar_j+1/2 vanish ?            ***
    !***                                                         ***
    !***  when mu = 0                                            ***
    !***  - if Nvpar is odd (even number of vpar grid points)    ***
    !***      vtemp(Nvpar)=0 ==> pb with Dsemi                   ***
    !***  - if Nvpar is even (odd number of vpar grid points)    ***
    !***      vtemp(Nvpar)=0 ==> pb with Dcoll & Vcoll           ***
    !***                                                         ***
    !*** Here, uses the even part of vtemp for Dcoll & Vcoll     ***
    ifill = 0
    do ivpar = 0,Nvpar-1,2
      !*** computes the integral                      ***
      !***  phi(x) = 2/sqrt(pi) int_0^x exp(-y^2) dy  ***
      !***   with x = v = sqrt(E/T)                   ***
      !*** using the derf FORTRAN function            *** 
      phi(ifill)      = derf(vtemp(ivpar))
      !*** computes phi'(x) = 2/sqrt(pi) exp(-x^2)  ***
      !***  with x = sqrt(E) / vT                   ***
      phiprime(ifill) = cst*exp(-vtemp(ivpar)*vtemp(ivpar))
      !*** computes psi(x) = (phi(x) - x phi'(x)) / (2 x^2) ***
      !***  with x = sqrt(E) / vT                           ***
      psi(ifill)      = (phi(ifill) - &
        vtemp(ivpar)*phiprime(ifill)) / &
        (2._RKIND*vtemp(ivpar)*vtemp(ivpar))
      !*** computes Dcoll = AD (phi(v) - psi(v)) / v    ***
      !*** computes Vcoll = -AD (1+m/m1) vpar/vT**2 *   ***
      !***                  (phi(v) - psi(v)) / v       ***
      Dcoll(ifill)    = AD * (phi(ifill) - psi(ifill)) / &
        (2._RKIND*vtemp(ivpar))
      Vcoll(ifill)    = - Dcoll(ifill) * &
        (geom%vparg(ifill) - Vloc) / (vT**2)
      !*** One completes by symmetry        ***
      !***  - Dcoll, Dsemi are even in vpar ***
      !***  - Vcoll is not symmetric        ***
      isymm              = Nvpar-ifill
      Dcoll(isymm)    = Dcoll(ifill)
      Vcoll(isymm)    = -Dcoll(ifill) * &
        (geom%vparg(isymm) - Vloc) / (vT**2)
      ifill           = ifill+1
    end do

    !*** Here, uses the odd part of vtemp for Dsemi ***
    ifill = 0
    do ivpar = 1,Nvpar-1,2
      !*** computes the integral                            ***
      !***  phi(x) = 2/sqrt(pi) int_0^x exp(-y^2) dy        ***
      !***   with x = vpar / (sqrt(2)*vT)                   ***
      !*** using the derf FORTRAN function, which already   ***
      !*** contains the 2/sqrt(pi) normalisation factor     ***
      phi(ifill)      = derf(vtemp(ivpar))
      !*** computes phi'(x) = 2/sqrt(pi) exp(-x^2)          ***
      !***  with x = sqrt(E) / vT                           ***
      phiprime(ifill) = cst*exp(-vtemp(ivpar)*vtemp(ivpar))
      !*** computes psi(x) = (phi(x) - x phi'(x)) / (2 x^2) ***
      !***  with x = sqrt(E) / vT                           ***
      psi(ifill)      = (phi(ifill) - &
        vtemp(ivpar)*phiprime(ifill)) /&
        (2._RKIND*vtemp(ivpar)*vtemp(ivpar))
      !*** computes Dcoll = AD psi(vpar/(sqrt(2)*vT))/vpar  ***
      !*** computes Vcoll = -AD (1+m/m1) *                  ***
      !***             psi(vpar/(sqrt(2)*vT1)) / (2 vT1^2)  ***
      Dsemi(ifill)    = AD * (phi(ifill) - psi(ifill)) / &
        (2._RKIND*vtemp(ivpar))
      !*** One completes by symmetry        ***
      !***  - Dcoll, Dsemi are even in vpar ***
      isymm           = Nvpar-1-ifill
      Dsemi(isymm)    = Dsemi(ifill)
      ifill           = ifill+1
    end do

    !*** Three values lack: at grid point where vtemp(Nvpar)=0 ***
    !***   but only when mu = 0                                ***
    if (geom%mug(mu_id).eq.0) then
      if (mod(Nvpar,2).eq.0) then
        iMiddle = Nvpar/2
        Dcoll(iMiddle) = AD * 2._RKIND/(3._RKIND*sqrt(pi))
        Vcoll(iMiddle) = (Dcoll(iMiddle) *  Vloc) /&
          (vT**2)
      else
        temp    = Nvpar/2
        iMiddle = floor(temp)
        Dsemi(iMiddle) = AD * 2._RKIND/(3._RKIND*sqrt(pi))
      end if
    else
      !*** No problem when mu != 0 ***
      if (mod(Nvpar,2).eq.0) then
        ivpar   = Nvpar
        iMiddle = Nvpar/2
        ifill   = iMiddle
        phi(ifill)      = derf(vtemp(ivpar))
        phiprime(ifill) = cst*exp(-vtemp(ivpar)*vtemp(ivpar))
        psi(ifill)      = (phi(ifill) - &
          vtemp(ivpar)*phiprime(ifill)) /&
          (2._RKIND*vtemp(ivpar)*vtemp(ivpar))
        Dcoll(ifill)    = AD * (phi(ifill) - psi(ifill)) / &
          (2._RKIND*vtemp(ivpar))
        Vcoll(ifill)    = (Dcoll(ifill) *  Vloc) /  (vT**2)
      else
        ivpar   = Nvpar
        temp    = Nvpar/2
        iMiddle = floor(temp)
        ifill   = iMiddle
        phi(ifill)      = derf(vtemp(ivpar))
        phiprime(ifill) = cst*exp(-vtemp(ivpar)*vtemp(ivpar))
        psi(ifill)      = (phi(ifill) - &
          vtemp(ivpar)*phiprime(ifill)) /&
          (2._RKIND*vtemp(ivpar)*vtemp(ivpar))
        Dsemi(ifill)    = AD * (phi(ifill) - psi(ifill)) / &
          (2._RKIND*vtemp(ivpar))
        end if
    end if

    !*** if 'coef_nustar' not equal to 0, the magnitude of ***
    !***   D and V is modified at the radial edges         ***
    !***     providing collisional buffer zones            ***
    if (coef_nustar.ne.0._RKIND) then
      do ivpar = 0,Nvpar-1
        Dcoll(ivpar) = Dcoll(ivpar)*envelope_r_loc(ir_coll)
        Dsemi(ivpar) = Dsemi(ivpar)*envelope_r_loc(ir_coll)
        Vcoll(ivpar) = Vcoll(ivpar)*envelope_r_loc(ir_coll)
      end do
      Dcoll(Nvpar) = Dcoll(Nvpar)*envelope_r_loc(ir_coll)
      Vcoll(Nvpar) = Vcoll(Nvpar)*envelope_r_loc(ir_coll)
    end if
#else 
    ! HAS_COLL_V2
    Dcoll = 0._RKIND
    Dsemi = 0._RKIND
    Vcoll = 0._RKIND
#endif 
    ! HAS_COLL_V2
  end subroutine compute_DcollVcoll_vers2


  !---------------------------------------------------
  !  Initialize matrix coefficients A, B and C used 
  !   in Crank-Nicholson and solved using tridiag 
  !   in AdvecDiff_coll_vpar
  !
  !  Here, dt is given by argument of 
  !   step0_collisions_vpar in gysela.F90
  !---------------------------------------------------
  subroutine construct_CrankNic_vers2(geom,dt,Dcoll, &
    Dsemi,Vcoll,Bstar,AA,BB,CC)
    use prec_const
    use MPIutils_module
    use geometry_class
    use globals, only : Nvpar
    implicit none
    type(geometry)                     , intent(in)  :: geom
    real(RKIND)                        , intent(in)  :: dt
    real(RKIND) , dimension(0:Nvpar  ) , intent(in)  :: Dcoll
    real(RKIND) , dimension(0:Nvpar  ) , intent(in)  :: Vcoll
    real(RKIND) , dimension(0:Nvpar-1) , intent(in)  :: Dsemi
    real(RKIND) , dimension(0:Nvpar  ) , intent(in)  :: Bstar
    real(RKIND) , dimension(0:Nvpar  ) , intent(out) :: AA
    real(RKIND) , dimension(0:Nvpar  ) , intent(out) :: BB
    real(RKIND) , dimension(0:Nvpar  ) , intent(out) :: CC

#ifdef HAS_COLL_V2
    real(RKIND) :: coefa, coefb, coefc
    real(RKIND) :: alpha, beta
    real(RKIND) :: dvpar, dvpar2
    integer     :: ivpar

    real(RKIND) :: Bstar_msemi, Bstar_psemi

    !***  We compute the matrix for the Crank-Nicolson scheme  ***
    !***   using the values of Dcoll, Dsemi and Vcoll          ***
    !***  Note: B(i+1/2) = 1/2* (B(i)+B(i+1))                  ***
    dvpar  = geom%dvpar
    dvpar2 = dvpar**2
    alpha  = dt/(2._RKIND*dvpar2)
    beta   = dt/(4._RKIND*dvpar)
    do ivpar = 1,Nvpar-1
      Bstar_msemi = 0.5_RKIND*(Bstar(ivpar-1)+Bstar(ivpar))
      Bstar_psemi = 0.5_RKIND*(Bstar(ivpar)+Bstar(ivpar+1))
      coefa       = Bstar_msemi*Dsemi(ivpar-1)*alpha + &
        Bstar(ivpar-1)*Vcoll(ivpar-1)*beta
      coefb       = (Bstar_msemi*Dsemi(ivpar-1) + &
        Bstar_psemi*Dsemi(ivpar))*alpha
      coefc       = Bstar_psemi*Dsemi(ivpar)*alpha - &
        Bstar(ivpar+1)*Vcoll(ivpar+1)*beta
      AA(ivpar)   = - coefa
      BB(ivpar)   = Bstar(ivpar) + coefb
      CC(ivpar)   = - coefc
    end do

    ! to compute the matrix coefficients at ivpar=0 and ivpar=Nvpar:
    ! boundary conditions for Dcoll and Vcoll are vanishing gradients
    !  => D(-1/2)  = 2*D(0) - D(1/2)   ; V(-1)  = 2*V(0) - V(1)
    !     D(N+1/2) = 2*D(N) - D(N-1/2) ; V(N+1) = 2*V(N) - V(N-1)
    ! we also use the fact that B* is linear in vpar:
    ! B*(-1/2)  = 3/2*B*(0) - 1/2*B*(1) ; B*(-1) = 2*B*(0) - B*(1) ;  
    ! B*(N+1/2) = 3/2*B*(N) - 1/2*B*(N-1) ; B*(N+1) = 2*B*(N) - B*(N-1)

    ! ivpar = 0
    Bstar_msemi = 1.5_RKIND*Bstar(0)-0.5_RKIND*Bstar(1)
    Bstar_psemi = 0.5_RKIND*(Bstar(0)+Bstar(1))
    coefa       =  Bstar_msemi*(2*Dcoll(0)-Dsemi(0))*alpha + &
      (2._RKIND*Bstar(0)-Bstar(1))*(2*Vcoll(0)-Vcoll(1))*beta
    coefb       = (Bstar_psemi*Dsemi(0) + &
      Bstar_msemi*(2*Dcoll(0)-Dsemi(0)))*alpha
    coefc       = Bstar_psemi*Dsemi(0)*alpha - &
        Bstar(1)*Vcoll(1)*beta
    AA(0)       = - coefa
    BB(0)       = Bstar(0) + coefb
    CC(0)       = - coefc

    ! ivpar = N
    Bstar_msemi = 0.5_RKIND*(Bstar(Nvpar-1)+Bstar(Nvpar)) 
    Bstar_psemi = 1.5_RKIND*Bstar(Nvpar)-0.5_RKIND*Bstar(Nvpar-1)
    coefa       =  Bstar_msemi*Dsemi(Nvpar-1)*alpha + &
      Bstar(Nvpar-1)*Vcoll(Nvpar-1)*beta
    coefb       = (Bstar_msemi*Dsemi(Nvpar-1) + &
      Bstar_psemi*(2*Dcoll(Nvpar)-Dsemi(Nvpar-1)))*alpha
    coefc       = Bstar_psemi * &
      (2._RKIND*Dcoll(Nvpar)-Dsemi(Nvpar-1))*alpha - &
      (2._RKIND*Bstar(Nvpar)-Bstar(Nvpar-1))* &
      (2._RKIND*Vcoll(Nvpar)-Vcoll(Nvpar-1))*beta
    AA(Nvpar)   = - coefa
    BB(Nvpar)   = Bstar(Nvpar) + coefb
    CC(Nvpar)   = - coefc
#else 
    ! HAS_COLL_V2
    AA = 0._RKIND
    BB = 0._RKIND
    CC = 0._RKIND
#endif 
    ! HAS_COLL_V2
  end subroutine construct_CrankNic_vers2


  !----------------------------------------------------------
  ! Solves for F the equation M(A,B,C) * F = R , 
  !  where A,B and C are input vectors defined in 
  !  init_CrankNich
  !   - diffusion + advection on dt
  !   - with a Crank-Nicolson scheme
  !
  ! Call for this subroutine in timesplitting_BSL 
  !   (resol_steps.F90)
  !----------------------------------------------------------
  subroutine AdvecDiff_coll_vers2(geom,dt,init_prof, &
    init_magnet, init_curr, f)
    use globals, only : Nr, Nvpar, Nphi, mu_id, &
      Tmean_loc, Vmean_loc, &
      major_species, pSpecies_id, hat_Zs, hat_As, &
      version_coll
#ifdef _OPENMP
    use OMPutils_module, only : omp_get_thread_num
#endif
    use OMPutils_module, only : Romp1_0Nvpar, Romp2_0Nvpar, &
      Romp3_0Nvpar, Romp4_0Nvpar, Romp5_0Nvpar, Romp6_0Nvpar, &
      Romp7_0Nvpar, Romp8_0Nvpar, Romp9_0Nvpar, Romp10_0Nvpar
    use clock_module
    use geometry_class
    use utils_module
    use init_profile_class
    use init_magnetic_class
    implicit none
    type(geometry)     , intent(in)    :: geom
    type(init_profile) , intent(in)    :: init_prof
    type(init_magnetic), intent(in)    :: init_magnet
    type(init_current) , intent(in)    :: init_curr
    real(RKIND)        , intent(in)    :: dt
    type(fdistribu5d)  , intent(inout) :: f

#ifdef HAS_COLL_V2
    integer     :: tid
    integer     :: ir, itheta, iphi, ivpar
    real(RKIND) :: tempf0, tempf1, tempfN, tempfNm1
    real(RKIND) :: q_on_qpic, rpic_on_r, Epic_on_E, epstemp
    real(RKIND) :: Tpic_on_T, nustar_prof, epstemp_rp
    real(RKIND) :: Coul_logarithm_ii, norm_nuii, hatZsprime_on_hatZs
    real(RKIND) :: hatAsprime_on_hatAs, vT
    real(RKIND) :: Coul_logarithm_ii_rp, norm_nuii_rp
    real(RKIND) :: collision_freqr_loc_rp, vT_rp
    real(RKIND) :: collision_freqr_omp, collision_freqr_omp_rp

    real(RKIND), dimension(:), pointer :: fvpar
    real(RKIND), dimension(:), pointer :: fcoll
    real(RKIND), dimension(:), pointer :: AAtri, BBtri, CCtri, RR
    real(RKIND), dimension(:), pointer :: Dcoll_tmp
    real(RKIND), dimension(:), pointer :: Dsemi_tmp
    real(RKIND), dimension(:), pointer :: Vcoll_tmp
    real(RKIND), dimension(:), pointer :: Bstar_tmp

    real(RKIND) :: Tloc, Vloc

    call clck_time(bclock_collisions_vpar)

    PRINT_CKSUM(f%values)
    PRINT_CKSUM(f%jstart)
    PRINT_CKSUM(f%jend)
    PRINT_CKSUM(f%istart_modif)
    PRINT_CKSUM(f%iend_modif)
    PRINT_CKSUM(Nvpar)
    PRINT_CKSUM(Nphi)
    PRINT_CKSUM(Tmean_loc)
    PRINT_CKSUM(Vmean_loc)

#ifdef _OPENMP
!$OMP PARALLEL private(tid,ir,itheta,iphi,ivpar,fvpar,fcoll, &
!$OMP AAtri,BBtri,CCtri,RR, &
!$OMP Tloc,Vloc,Dcoll_tmp,Dsemi_tmp,Vcoll_tmp, Bstar_tmp, &
!$OMP q_on_qpic,rpic_on_r,Epic_on_E,epstemp,epstemp_rp, &
!$OMP Tpic_on_T,nustar_prof, &
!$OMP hatZsprime_on_hatZs,hatAsprime_on_hatAs,vT,vT_rp, &
!$OMP Coul_logarithm_ii,Coul_logarithm_ii_rp,norm_nuii, &
!$OMP norm_nuii_rp,collision_freqr_omp,collision_freqr_omp_rp, &
!$OMP tempf0,tempf1,tempfN,tempfNm1) default(shared)
!$OMP BARRIER
    tid = 1+omp_get_thread_num()
#else
    tid = 1
#endif
   
    fvpar     => Romp1_0Nvpar(tid)%val
    fcoll     => Romp2_0Nvpar(tid)%val
    AAtri     => Romp3_0Nvpar(tid)%val
    BBtri     => Romp4_0Nvpar(tid)%val
    CCtri     => Romp5_0Nvpar(tid)%val
    RR        => Romp6_0Nvpar(tid)%val
    Dcoll_tmp => Romp7_0Nvpar(tid)%val
    Dsemi_tmp => Romp8_0Nvpar(tid)%val
    Vcoll_tmp => Romp9_0Nvpar(tid)%val
    Bstar_tmp => Romp10_0Nvpar(tid)%val

!$OMP DO SCHEDULE(static)
    do iphi = 0,Nphi-1
      !*** vpar advection & diffusion                           ***
      !*** A,B and C defining Crank-Nicolson tridiagonal matrix ***
      do itheta = f%jstart,f%jend  
        do ir = f%istart_modif,f%iend_modif 
          do ivpar = 0,Nvpar
            fvpar(ivpar) = f%values(ir,itheta,iphi,ivpar)
            !-> Compute Bstar
#ifdef NO_PRECOMP_BSTAR 
            call compute_Bstar(geom, init_magnet, init_curr, &
              ir, itheta, ivpar, Bstar_tmp(ivpar))
#else
            call precomp_Bstar(ir,itheta,ivpar,Bstar_tmp(ivpar))
#endif
          end do
          !-> Compute Dcoll, Vcoll and Dsemi
          Tloc = Tmean_loc(ir,itheta,iphi)
          Vloc = Vmean_loc(ir,itheta,iphi)
          !-> Compute the radial dependence of nustar
          nustar_prof = 1._RKIND
          if (version_coll==3) then
            !--!! computed with the INITIAL profile of temperature !!--
            q_on_qpic   = 1._RKIND / (init_prof%q_rp * init_prof%iota(ir))
            rpic_on_r   = init_prof%rpeak / geom%rg(ir)
            epstemp     = geom%rg(ir)*inv_R(ir,itheta)
            epstemp_rp  = geom%rg(int(Nr/2)+1)*inv_R(int(Nr/2)+1,itheta)
            Epic_on_E   = epstemp_rp / epstemp
            !--!! not computed from Tmean_loc !!--
            Tpic_on_T   = init_prof%Ts0_rp(major_species) / &
              init_prof%Ts0(major_species,ir)
            !--!! needed for radial profile of the collision frequency !!--
            ! properly define nustar(r) and the collision frequency(r) in the
            ! toroidal case when (version_coll==3), as in Eq.(12) [Dif-Pradalier
            ! PoP 2011]. So far, this is done for ion-ion collisions only
            ! NB: proper definitions of the Coulomb logarithm are required
            ! as density and temperatures are non-dimensional so far...
            hatZsprime_on_hatZs = hat_Zs(pSpecies_id)/hat_Zs(major_species)
            hatAsprime_on_hatAs = hat_As(pSpecies_id)/hat_As(major_species)
            vT                  = sqrt( init_prof%Ts0(major_species,ir) )
            vT_rp               = sqrt( init_prof%Ts0_rp(major_species) )
            ! Coulomb_logarithm: should be the electron density,
            ! profile is right but magnitude is wrong: should be log(n_e/10e20)
            ! and temperature should be in keV
            !---!! computed with INITIAL profiles of temperature & density !!---
            Coul_logarithm_ii    = 17.3_RKIND - &
              0.5_RKIND*log(init_prof%ns0(major_species,ir)) + &
              1.5_RKIND*log(init_prof%Ts0(major_species,ir))
            Coul_logarithm_ii_rp = 17.3_RKIND - &
              0.5_RKIND*log(init_prof%ns0_rp(major_species)) + &
              1.5_RKIND*log(init_prof%Ts0_rp(major_species))
            norm_nuii            = Coul_logarithm_ii * &
              4._RKIND*sqrt(pi)/3._RKIND
            norm_nuii_rp         = Coul_logarithm_ii_rp * &
              4._RKIND*sqrt(pi)/3._RKIND
            ! collision_freqr is correct for the ion-ion collisions
            collision_freqr_omp    = norm_nuii * &
              init_prof%ns0(major_species,ir) * &
              hatZsprime_on_hatZs**4 / ( hatAsprime_on_hatAs**2 * (vT)**3 )
            collision_freqr_omp_rp = norm_nuii_rp * &
              init_prof%ns0_rp(major_species) * &
              hatZsprime_on_hatZs**4 / ( hatAsprime_on_hatAs**2 * (vT_rp)**3 )
            ! at last...
            nustar_prof = q_on_qpic*sqrt(Tpic_on_T) * &
              Epic_on_E*sqrt(Epic_on_E) * &
              collision_freqr_omp / collision_freqr_omp_rp
          end if
          call compute_DcollVcoll_vers2(geom,init_prof, &
            init_magnet,ir,itheta,Tloc,Vloc,nustar_prof, &
            Dcoll_tmp,Dsemi_tmp,Vcoll_tmp)
          !-> Construct matrix for Crank-Nicolson
          call construct_CrankNic_vers2(geom,dt,Dcoll_tmp, &
            Dsemi_tmp,Vcoll_tmp,Bstar_tmp,AAtri,BBtri,CCtri)
          !-> Construct right-hand-side of the equation
          do ivpar = 1,Nvpar-1
            fcoll(ivpar) = fvpar(ivpar)
            RR(ivpar)    = - AAtri(ivpar)*fvpar(ivpar-1) + &
              (2._RKIND*Bstar_tmp(ivpar) - &
              BBtri(ivpar))*fvpar(ivpar) - &
              CCtri(ivpar)*fvpar(ivpar+1)
          end do
          !*** Boundary conditions ***
          !-> Vanishing second derivative for the right-hand side
          RR(0)        = (2._RKIND*Bstar_tmp(0) - BBtri(0) -&
            2._RKIND*AAtri(0))*fvpar(0) -&
            (CCtri(0) - AAtri(0))*fvpar(1)
          RR(Nvpar)    = (2._RKIND*Bstar_tmp(Nvpar) &
            - BBtri(Nvpar) - &
            2._RKIND*CCtri(Nvpar))*fvpar(Nvpar) - &
            (AAtri(Nvpar) - CCtri(Nvpar))*fvpar(Nvpar-1)
          !*** matrix inversion ***
          call tridiag(AAtri,BBtri,CCtri,RR,fcoll,Nvpar)
          do ivpar = 0,Nvpar
            f%values(ir,itheta,iphi,ivpar) = fcoll(ivpar)
          end do
        end do
      end do
    end do
!$OMP END DO
!$OMP BARRIER
!$OMP END PARALLEL

    PRINT_CKSUM(f%values)

    call clck_time(eclock_collisions_vpar)
    call clck_diff(bclock_collisions_vpar,eclock_collisions_vpar, &
      global_time_collisions_vpar)
#endif 
    ! HAS_COLL_V2
  end subroutine AdvecDiff_coll_vers2


  !-------------------------------------------------------
  ! Computation of the 5 useful integrals:
  !   . Imean0 = <Dpar> 
  !   . Imean1 = <vpar*Dpar>
  !   . Imean2 = <vpar^2*Dpar>  
  !   . Imean3 = <1/Bstar*d/dvpar(Bstar*Dpar)> 
  !   . Imean4 = <1/Bstar*d/dvpar(Bstar*vpar*Dpar>
  ! where the integral average <.> corresponds to:
  !  where <.> = \int Bstar(r,theta,vpar) dmu dvpar F_i .
  !   with F_i the ion distribution function
  !-------------------------------------------------------
  subroutine compute_mean_integrals_vers2(geom, &
    init_prof,init_magnet,init_curr,f, &
    sImean0,sImean1,sImean2,sImean3,sImean4)
    use globals, only : istart, iend, jstart, jend, &
      Nr, Ntheta, Nphi, Nmu, &
      Nbproc_loc, Nbproc_r, dom_r, dom_theta, Nbthread, &
      Tmean_loc, Vmean_loc, &
      Rarray_send_nbM5, RRecvBufMPI_C, &
      integration_CS, mu_id, plocal_id, &
      mpi_comm_species, mpi_comm_intermu, &
      pSpecies_id, major_species, hat_Zs, hat_As, &
      version_coll
    use OMPutils_module
    use clock_module
    use integration_module, only : compute_vpar_integral_colloc, &
      compute_omp_vpar_integral_CS
    use utils_module      , only : deriv1
    use mpiMod
    type(geometry)                 , intent(in)  :: geom
    type(init_profile)             , intent(in)  :: init_prof
    type(init_magnetic)            , intent(in)  :: init_magnet
    type(init_current)             , intent(in)  :: init_curr
    type(fdistribu5d)              , intent(in)  :: f
    real(RKIND), &
      dimension(:,:,:), pointer :: sImean0
    real(RKIND), &
      dimension(:,:,:), pointer :: sImean1
    real(RKIND), &
      dimension(:,:,:), pointer :: sImean2
    real(RKIND), &
      dimension(:,:,:), pointer :: sImean3
    real(RKIND), &
      dimension(:,:,:), pointer :: sImean4

#ifdef HAS_COLL_V2
    integer, parameter :: nbmoments = 5
    integer            :: ir, itheta, iphi, ivpar, imu
    real(RKIND)        :: mu_value, coeff_intdmu, jacob_vel_tmp
    real(RKIND)        :: vpar, fval, Dpar_tmp
    real(RKIND)        :: Bstar_tmp, inv_Bstar

    !-> I0 = Dpar*f ; I1 = Dpar*vpar*f and I2 = Dpar*vpar^2*f
    real(RKIND), dimension(:)  , pointer :: I0_tmp, I1_tmp, I2_tmp
    !-> I3 = 1/Bstar*deriv_BstarDpar*f ; 
    !-> I4 = 1/Bstar*deriv_vBstarDpar*f
    !->   with deriv_BstarDpar  = d(Bstar*Dpar)/dvpar and
    !->        deriv_vBstarDpar = d(Bstar*vpar*Dpar)/dvpar 
    real(RKIND)                          :: deriv_BstarDpar_tmp
    real(RKIND)                          :: deriv_vBstarDpar_tmp
    real(RKIND), dimension(:)  , pointer :: BstarDpar, vBstarDpar
    real(RKIND), dimension(:)  , pointer :: Dcoll_tmp
    real(RKIND), dimension(:)  , pointer :: Vcoll_tmp
    real(RKIND), dimension(:)  , pointer :: Dsemi_tmp
    real(RKIND)                          :: Tloc, Vloc
    real(RKIND)                          :: q_on_qpic, rpic_on_r
    real(RKIND)                          :: Tpic_on_T, nustar_prof
    real(RKIND), dimension(:)  , pointer :: deriv_BstarDpar
    real(RKIND), dimension(:)  , pointer :: deriv_vBstarDpar
    real(RKIND), dimension(:)  , pointer :: I3_tmp, I4_tmp

    !-> for radial profile of nustar (version_coll==3)
    real(RKIND) :: Coul_logarithm_ii, norm_nuii, hatZsprime_on_hatZs
    real(RKIND) :: hatAsprime_on_hatAs, vT, Epic_on_E, epstemp
    real(RKIND) :: Coul_logarithm_ii_rp, norm_nuii_rp, epstemp_rp
    real(RKIND) :: collision_freqr_loc_rp, vT_rp
    real(RKIND) :: collision_freqr_omp, collision_freqr_omp_rp

    !-> integrals in vpar direction, i.e \int . dvpar
    real(RKIND), dimension(:,:), pointer :: int_I0_dvpar
    real(RKIND), dimension(:,:), pointer :: int_I1_dvpar
    real(RKIND), dimension(:,:), pointer :: int_I2_dvpar
    real(RKIND), dimension(:,:), pointer :: int_I3_dvpar
    real(RKIND), dimension(:,:), pointer :: int_I4_dvpar

    !-> for cubic splines in vpar directions
    real(RKIND), dimension(:), pointer :: scoefvpar

    ! -> variable for parallelization
    integer :: ierr, itag
    integer :: nbelem, nbreqr, local_id
    integer :: dep, base, global_id, startbuf
    integer :: l_istart, l_iend, l_jstart, l_jend
    integer :: nbelements, tid, Nbthread_loc

    ! -> variables for time perfomance analysis
    integer(TIMEPREC) :: tdeb, tfin
    real(RKIND)       :: tcomm1, tcomm2, tcomm3, timu, tivpar
    real(RKIND)       :: I0sum, I1sum, I2sum, I3sum, I4sum
    real(RKIND)       :: I0perc, I1perc, I2perc, I3perc, I4perc
    real(RKIND)       :: percent_bias, eps, local_bias
    call clck_time(tdeb)
    
    call write_param("nbmoments", nbmoments)
    call FTR_ALLOC_1_ZERO(RRecvBufMPI_C,0,nbmoments*dom_r*dom_theta*Nphi*(Nmu+1))

    percent_bias = 0._RKIND
    mu_value  = f%mu_value
#ifdef _OPENMP
!$OMP PARALLEL private(tid,iphi,itheta,ir,ivpar, &
!$OMP BstarDpar,vBstarDpar,deriv_BstarDpar,deriv_vBstarDpar, &
!$OMP scoefvpar,fval,vpar,inv_Bstar,jacob_vel_tmp, &
!$OMP I0_tmp,I1_tmp,I2_tmp,I3_tmp,I4_tmp, &
!$OMP Bstar_tmp,Dpar_tmp,deriv_BstarDpar_tmp, &
!$OMP deriv_vBstarDpar_tmp,Dcoll_tmp,Dsemi_tmp, &
!$OMP Vcoll_tmp,Tloc,Vloc,epstemp,epstemp_rp, &
!$OMP q_on_qpic,rpic_on_r,Epic_on_E,Tpic_on_T,nustar_prof, &
!$OMP hatZsprime_on_hatZs,hatAsprime_on_hatAs,vT,vT_rp, &
!$OMP Coul_logarithm_ii,Coul_logarithm_ii_rp,norm_nuii, &
!$OMP norm_nuii_rp,collision_freqr_omp,collision_freqr_omp_rp, &
!$OMP I0sum,I1sum,I2sum,I3sum,I4sum,&
!$OMP eps,I0perc,I1perc,I2perc,I3perc,I4perc,local_bias), &
!$OMP default(shared)
!$OMP BARRIER
    tid = 1+omp_get_thread_num()
#else
    tid = 1
#endif
    eps        = 100._RKIND*epsilon(1._RKIND)
    local_bias = 0._RKIND
    BstarDpar        => Romp1_0Nvpar(tid)%val
    vBstarDpar       => Romp2_0Nvpar(tid)%val
    deriv_BstarDpar  => Romp3_0Nvpar(tid)%val
    deriv_vBstarDpar => Romp4_0Nvpar(tid)%val

    I0_tmp           => Romp5_0Nvpar(tid)%val
    I1_tmp           => Romp6_0Nvpar(tid)%val
    I2_tmp           => Romp7_0Nvpar(tid)%val
    I3_tmp           => Romp8_0Nvpar(tid)%val
    I4_tmp           => Romp9_0Nvpar(tid)%val

    Dcoll_tmp        => Romp10_0Nvpar(tid)%val
    Dsemi_tmp        => Romp11_0Nvpar(tid)%val
    Vcoll_tmp        => Romp12_0Nvpar(tid)%val

    scoefvpar        => Romp_scoefvpar(tid)%val
!$OMP DO SCHEDULE(STATIC)
    do iphi = 0,f%n3-1
      do itheta = f%jstart,f%jend
        do ir = f%istart,f%iend
          Tloc = Tmean_loc(ir,itheta,iphi)
          Vloc = Vmean_loc(ir,itheta,iphi)
          !-> Compute the radial dependence of nustar
          nustar_prof = 1._RKIND
          if (version_coll==3) then
            !--!! computed with the INITIAL profile of !!--
            !--!!  temperature                         !!--
            q_on_qpic   = 1._RKIND / &
              (init_prof%q_rp * init_prof%iota(ir))
            rpic_on_r   = init_prof%rpeak / geom%rg(ir)
            epstemp     = geom%rg(ir)*inv_R(ir,itheta)
            epstemp_rp  = geom%rg(int(Nr/2)+1) * &
              inv_R(int(Nr/2)+1,itheta)
            Epic_on_E   = epstemp_rp / epstemp
            !--!! not computed from Tmean_loc !!--
            Tpic_on_T   = init_prof%Ts0_rp(major_species) / &
              init_prof%Ts0(major_species,ir)
            !--!! needed for radial profile of !!--
            !--!!   the collision frequency    !!--
            ! properly define nustar(r) and the collision 
            !  frequency(r) in the toroidal case when 
            !  (version_coll==3), as in Eq.(12) [Dif-Pradalier
            !  PoP 2011]. So far, this is done for ion-ion 
            !  collisions only
            ! NB: proper definitions of the Coulomb logarithm 
            !     are required as density and temperatures are 
            !     non-dimensional so far...
            hatZsprime_on_hatZs = hat_Zs(pSpecies_id) / &
              hat_Zs(major_species)
            hatAsprime_on_hatAs = hat_As(pSpecies_id) / &
              hat_As(major_species)
            vT    = sqrt( init_prof%Ts0(major_species,ir) )
            vT_rp = sqrt( init_prof%Ts0_rp(major_species) )
            ! Coulomb_logarithm: should be the electron density,
            !  profile is right but magnitude is wrong: 
            !  should be log(n_e/10e20) and temperature 
            !  should be in keV
            !--!! computed with INITIAL profiles of !!--
            !--!!  temperature & density            !!--
            Coul_logarithm_ii    = 17.3_RKIND - &
                 0.5_RKIND*log(init_prof%ns0(major_species,ir)) + &
                 1.5_RKIND*log(init_prof%Ts0(major_species,ir))
            Coul_logarithm_ii_rp = 17.3_RKIND - &
                 0.5_RKIND*log(init_prof%ns0_rp(major_species)) + &
                 1.5_RKIND*log(init_prof%Ts0_rp(major_species))
            norm_nuii            = Coul_logarithm_ii * &
              4._RKIND*sqrt(pi)/3._RKIND
            norm_nuii_rp         = Coul_logarithm_ii_rp * &
              4._RKIND*sqrt(pi)/3._RKIND
            ! collision_freqr is correct for the ion-ion collisions
            collision_freqr_omp    = norm_nuii * &
              init_prof%ns0(major_species,ir) * &
              hatZsprime_on_hatZs**4 / &
              ( hatAsprime_on_hatAs**2 * (vT)**3 )
            collision_freqr_omp_rp = norm_nuii_rp * &
              init_prof%ns0_rp(major_species) * &
              hatZsprime_on_hatZs**4 / &
              ( hatAsprime_on_hatAs**2 * (vT_rp)**3 )
            ! at last...
            nustar_prof = q_on_qpic*sqrt(Tpic_on_T) * &
              Epic_on_E*sqrt(Epic_on_E) * &
              collision_freqr_omp / collision_freqr_omp_rp
          end if
          call compute_DcollVcoll_vers2(geom,init_prof, &
            init_magnet,ir,itheta,Tloc,Vloc,nustar_prof, &
            Dcoll_tmp,Dsemi_tmp,Vcoll_tmp)
          do ivpar = 0,f%n4
#ifdef NO_PRECOMP_BSTAR 
            call compute_Bstar(geom,init_magnet, &
              init_curr,ir,itheta,ivpar,Bstar_tmp)
#else
            call precomp_Bstar(ir,itheta,ivpar,Bstar_tmp)
#endif
            Dpar_tmp          = Dcoll_tmp(ivpar)
            BstarDpar(ivpar)  = Bstar_tmp * Dpar_tmp
            vBstarDpar(ivpar) = geom%vparg(ivpar) * &
              Bstar_tmp * Dpar_tmp
          end do
          call deriv1(BstarDpar,deriv_BstarDpar, &
            geom%Nvpar,geom%dvpar,0)
          call deriv1(vBstarDpar,deriv_vBstarDpar, &
            geom%Nvpar,geom%dvpar,0)
          do ivpar = 0,f%n4
            vpar = geom%vparg(ivpar)
#ifdef NO_PRECOMP_BSTAR 
            call compute_Bstar(geom,init_magnet, &
              init_curr,ir,itheta,ivpar,Bstar_tmp)
#else
            call precomp_Bstar(ir,itheta,ivpar,Bstar_tmp)
#endif
            !-> jacobian_velocity = 2*pi*Bstar
            call compute_jacobian_velocity(geom,init_magnet, &
              init_curr,ir,itheta,ivpar,jacob_vel_tmp)
            Dpar_tmp             = Dcoll_tmp(ivpar)
            inv_Bstar            = 1._RKIND/Bstar_tmp
            deriv_BstarDpar_tmp  = deriv_BstarDpar(ivpar)
            deriv_vBstarDpar_tmp = deriv_vBstarDpar(ivpar)
            fval                 = f%values(ir,itheta,iphi,ivpar)
            I0_tmp(ivpar)        = Dpar_tmp*fval * &
              jacob_vel_tmp
            I1_tmp(ivpar)        = Dpar_tmp*vpar*fval * &
              jacob_vel_tmp
            I2_tmp(ivpar)        = Dpar_tmp*vpar*vpar*fval * &
              jacob_vel_tmp
            I3_tmp(ivpar)        = inv_Bstar * &
              deriv_BstarDpar_tmp*fval*jacob_vel_tmp
            I4_tmp(ivpar)        = inv_Bstar * &
              deriv_vBstarDpar_tmp*fval*jacob_vel_tmp
          end do
          !*** computation of the integrals in vpar direction ***
          if (integration_CS) then
            !*** -> \int I0 dvpar ***
            call compute_omp_vpar_integral_CS(f%n4,f%h4,I0_tmp,&
              Fmean_BCvpar_left, Fmean_BCvpar_right,&
              Fmean_nspline1d_vpar(tid),scoefvpar,&
              Rarray_send_nbM5(0,ir,itheta,iphi))
            !*** -> \int I1 dvpar ***
            call compute_omp_vpar_integral_CS(f%n4,f%h4,I1_tmp,&
              Fmean_BCvpar_left,Fmean_BCvpar_right,&
              Fmean_nspline1d_vpar(tid),scoefvpar,&
              Rarray_send_nbM5(1,ir,itheta,iphi))
            !*** -> \int I2 dvpar ***
            call compute_omp_vpar_integral_CS(f%n4,f%h4,I2_tmp,&
              Fmean_BCvpar_left,Fmean_BCvpar_right,&
              Fmean_nspline1d_vpar(tid),scoefvpar,&
              Rarray_send_nbM5(2,ir,itheta,iphi))
            !*** -> \int I3 dvpar ***
            call compute_omp_vpar_integral_CS(f%n4,f%h4,I3_tmp,&
              Fmean_BCvpar_left,Fmean_BCvpar_right,&
              Fmean_nspline1d_vpar(tid),scoefvpar,&
              Rarray_send_nbM5(3,ir,itheta,iphi))
            !*** -> \int I4 dvpar ***
            call compute_omp_vpar_integral_CS(f%n4,f%h4,I4_tmp,&
              Fmean_BCvpar_left,Fmean_BCvpar_right,&
              Fmean_nspline1d_vpar(tid),scoefvpar,&
              Rarray_send_nbM5(4,ir,itheta,iphi))
          else
            !*** -> \int I0 dvpar ***
            call compute_vpar_integral_colloc( &
              I0_tmp(0:f%n4),geom, &
              Rarray_send_nbM5(0,ir,itheta,iphi))
            !*** -> \int I1 dvpar ***
            call compute_vpar_integral_colloc( &
              I1_tmp(0:f%n4),geom, &
              Rarray_send_nbM5(1,ir,itheta,iphi))
            !*** -> \int I2 dvpar ***
            call compute_vpar_integral_colloc( &
              I2_tmp(0:f%n4),geom, &
              Rarray_send_nbM5(2,ir,itheta,iphi))
            !*** -> \int I3 dvpar ***
            call compute_vpar_integral_colloc( &
              I3_tmp(0:f%n4),geom, &
              Rarray_send_nbM5(3,ir,itheta,iphi))
            !*** -> \int I4 dvpar ***
            call compute_vpar_integral_colloc( &
              I4_tmp(0:f%n4),geom, &
              Rarray_send_nbM5(4,ir,itheta,iphi))
          end if
          
          ! Recompute the same quantitities done in the
          ! previous loop, but constraining fval to be
          ! positive for the second species
          if (pSpecies_id .ne. major_species) then
            do ivpar = 0,f%n4
              vpar = geom%vparg(ivpar)
#ifdef NO_PRECOMP_BSTAR 
              call compute_Bstar(geom,init_magnet, &
                init_curr,ir,itheta,ivpar,Bstar_tmp)
#else
              call precomp_Bstar(ir,itheta,ivpar,Bstar_tmp)
#endif
              !-> jacobian_velocity = 2*pi*Bstar
              call compute_jacobian_velocity(geom,init_magnet, &
                init_curr,ir,itheta,ivpar,jacob_vel_tmp)
              Dpar_tmp             = Dcoll_tmp(ivpar)
              inv_Bstar            = 1._RKIND/Bstar_tmp
              deriv_BstarDpar_tmp  = deriv_BstarDpar(ivpar)
              deriv_vBstarDpar_tmp = deriv_vBstarDpar(ivpar)
              fval = MAX(eps,f%values(ir,itheta,iphi,ivpar))
              I0_tmp(ivpar)        = Dpar_tmp*fval * &
                jacob_vel_tmp
              I1_tmp(ivpar)        = Dpar_tmp*vpar*fval * &
                jacob_vel_tmp
              I2_tmp(ivpar)        = Dpar_tmp*vpar*vpar*fval * &
                jacob_vel_tmp
              I3_tmp(ivpar)        = inv_Bstar * &
                deriv_BstarDpar_tmp*fval*jacob_vel_tmp
              I4_tmp(ivpar)        = inv_Bstar * &
                deriv_vBstarDpar_tmp*fval*jacob_vel_tmp
            end do
            !*** computation of the integrals in vpar direction ***
            if (integration_CS) then
              !*** -> \int I0 dvpar ***
              call compute_omp_vpar_integral_CS(f%n4,f%h4,I0_tmp,&
                Fmean_BCvpar_left, Fmean_BCvpar_right,&
                Fmean_nspline1d_vpar(tid),scoefvpar,&
                I0sum)
              !*** -> \int I1 dvpar ***
              call compute_omp_vpar_integral_CS(f%n4,f%h4,I1_tmp,&
                Fmean_BCvpar_left,Fmean_BCvpar_right,&
                Fmean_nspline1d_vpar(tid),scoefvpar,&
                I1sum)
              !*** -> \int I2 dvpar ***
              call compute_omp_vpar_integral_CS(f%n4,f%h4,I2_tmp,&
                Fmean_BCvpar_left,Fmean_BCvpar_right,&
                Fmean_nspline1d_vpar(tid),scoefvpar,&
                I2sum)
              !*** -> \int I3 dvpar ***
              call compute_omp_vpar_integral_CS(f%n4,f%h4,I3_tmp,&
                Fmean_BCvpar_left,Fmean_BCvpar_right,&
                Fmean_nspline1d_vpar(tid),scoefvpar,&
                I3sum)
              !*** -> \int I4 dvpar ***
              call compute_omp_vpar_integral_CS(f%n4,f%h4,I4_tmp,&
                Fmean_BCvpar_left,Fmean_BCvpar_right,&
                Fmean_nspline1d_vpar(tid),scoefvpar,&
                I4sum)
            else
              !*** -> \int I0 dvpar ***
              call compute_vpar_integral_colloc(I0_tmp(0:f%n4), &
                geom,I0sum)
              !*** -> \int I1 dvpar ***
              call compute_vpar_integral_colloc(I1_tmp(0:f%n4), &
                geom,I1sum)
              !*** -> \int I2 dvpar ***
              call compute_vpar_integral_colloc(I2_tmp(0:f%n4), &
                geom,I2sum)
              !*** -> \int I3 dvpar ***
              call compute_vpar_integral_colloc(I3_tmp(0:f%n4), &
                geom,I3sum)
              !*** -> \int I4 dvpar ***
              call compute_vpar_integral_colloc(I4_tmp(0:f%n4), &
                geom,I4sum)
            end if
            I0perc = abs(Rarray_send_nbM5(0,ir,itheta,iphi) / &
              I0sum - 1._RKIND)
            I1perc = abs(Rarray_send_nbM5(1,ir,itheta,iphi) / &
              I1sum - 1._RKIND)
            I2perc = abs(Rarray_send_nbM5(2,ir,itheta,iphi) / &
              I2sum - 1._RKIND)
            I3perc = abs(Rarray_send_nbM5(3,ir,itheta,iphi) / &
              I3sum - 1._RKIND) 
            I4perc = abs(Rarray_send_nbM5(4,ir,itheta,iphi) / &
              I4sum - 1._RKIND) 
            local_bias = MAX(MAX(MAX(MAX(MAX(local_bias,I0perc), &
              I1perc),I2perc),I3perc),I4perc)
            Rarray_send_nbM5(0,ir,itheta,iphi) = I0sum
            Rarray_send_nbM5(1,ir,itheta,iphi) = I1sum
            Rarray_send_nbM5(2,ir,itheta,iphi) = I2sum
            Rarray_send_nbM5(3,ir,itheta,iphi) = I3sum
            Rarray_send_nbM5(4,ir,itheta,iphi) = I4sum
          end if
        end do
      end do
    end do
!$OMP END DO
!$OMP CRITICAL
    percent_bias = MAX(percent_bias,local_bias)
!$OMP END CRITICAL
!$OMP BARRIER
!$OMP END PARALLEL
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tivpar)

    if (.not.(percent_bias.lt.10)) &
      print *, pglobal_id, "percent_bias", percent_bias

    call clck_time(tdeb)
    nbreqr = 0
	RRecvBufMPI_C = 0._RKIND
    if (mu_id .eq. 0) then
      do imu = 0, Nmu
        global_id = imu*Nbproc_loc + plocal_id
        startbuf  = dom_theta * Nphi * imu
        nbelem    = nbmoments * dom_r * dom_theta * Nphi
        itag      = 20 
        call MPI_IRECV( &
          RRecvBufMPI_C(startbuf*dom_r*nbmoments),nbelem, &
          MPI_REAL8,global_id,itag,mpi_comm_species, &
          moments_reqr(nbreqr),ierr)
        nbreqr = nbreqr + 1
      enddo
    endif
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tcomm1)

    !*** Parallel communication to send                   ***
    !***   the five integrals in vpar                     ***
    !***   . Imean0  = <Dpar>                             ***
    !***   . Imean1  = <vpar*Dpar>                        ***
    !***   . Imean2  = <vpar^2*Dpar>                      ***
    !***   . Imean3  = <1/Bstar*d/dvpar(Bstar*Dpar)>      ***
    !***   . Imean4  = <1/Bstar*d/dvpar(Bstar*vpar*Dpar>  ***
    !*** The parallelism is such that                     ***
    !***  (r=block,theta=block,phi=*)                     ***
    !*** RK : These 5 integrals are saved in an           ***
    !***  unique array                                    ***
    call clck_time(tdeb)
    nbelem = nbmoments * dom_r * dom_theta * Nphi
    itag   = 20 
    call MPI_SSEND(Rarray_send_nbM5(0,istart,jstart,0),nbelem, &
      MPI_REAL8,plocal_id,itag,mpi_comm_species, ierr)
    if (mu_id .eq. 0) then
      call MPI_WAITALL(nbreqr,moments_reqr(0:nbreqr-1), &
        moments_status(:,0:nbreqr-1),ierr)
    endif
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tcomm2)

    do tid = 1, Nbthread
       call FTR_ALLOC_2(Romp1_1Nrp1_1Nthetap1(tid)%val,1,Nr+1,1,Ntheta+1)
       call FTR_ALLOC_2(Romp2_1Nrp1_1Nthetap1(tid)%val,1,Nr+1,1,Ntheta+1)
       call FTR_ALLOC_2(Romp3_1Nrp1_1Nthetap1(tid)%val,1,Nr+1,1,Ntheta+1)
       call FTR_ALLOC_2(Romp4_1Nrp1_1Nthetap1(tid)%val,1,Nr+1,1,Ntheta+1)
       call FTR_ALLOC_2(Romp5_1Nrp1_1Nthetap1(tid)%val,1,Nr+1,1,Ntheta+1)
    enddo

    do tid = 0, MAX(Nmu,Nbthread)
       call FTR_ALLOC_2(Romp10_1Nrp1_1Nthetap1(tid)%val,1,Nr+1,1,Ntheta+1)
       call FTR_ALLOC_2(Romp11_1Nrp1_1Nthetap1(tid)%val,1,Nr+1,1,Ntheta+1)
       call FTR_ALLOC_2(Romp12_1Nrp1_1Nthetap1(tid)%val,1,Nr+1,1,Ntheta+1)
       call FTR_ALLOC_2(Romp13_1Nrp1_1Nthetap1(tid)%val,1,Nr+1,1,Ntheta+1)
       call FTR_ALLOC_2(Romp14_1Nrp1_1Nthetap1(tid)%val,1,Nr+1,1,Ntheta+1)
    enddo

    call clck_time(tdeb)
    if (mu_id .eq. 0) then
#ifdef _OPENMP
!$OMP PARALLEL private(tid,ir,itheta,iphi,ivpar,imu, &
!$OMP coeff_intdmu,local_id,base,dep, &
!$OMP l_iend,l_istart,startbuf, &
!$OMP l_jstart,l_jend,int_I0_dvpar, &
!$OMP int_I1_dvpar,int_I2_dvpar, &
!$OMP int_I3_dvpar,int_I4_dvpar) default(shared) 
!$OMP BARRIER
      tid = 1+omp_get_thread_num()
#else
      tid = 1
#endif

      int_I0_dvpar => Romp1_1Nrp1_1Nthetap1(tid)%val
      int_I1_dvpar => Romp2_1Nrp1_1Nthetap1(tid)%val
      int_I2_dvpar => Romp3_1Nrp1_1Nthetap1(tid)%val
      int_I3_dvpar => Romp4_1Nrp1_1Nthetap1(tid)%val
      int_I4_dvpar => Romp5_1Nrp1_1Nthetap1(tid)%val

      l_istart = f%istart
      l_jstart = f%jstart
      do iphi = 0,geom%Nphi-1
!$OMP DO SCHEDULE(STATIC)
        do itheta = f%jstart, f%jend
          do imu = 0,Nmu
            coeff_intdmu = geom%coeff_intdmu(imu)
            startbuf = dom_theta * Nphi * imu +iphi*dom_theta
            do ir = f%istart, f%iend
              int_I0_dvpar(1+ir,1+itheta) = &
                RRecvBufMPI_C(nbmoments*(ir-l_istart)+0+&
                (startbuf+itheta-l_jstart)*nbmoments*dom_r) 
              int_I1_dvpar(1+ir,1+itheta) = &
                RRecvBufMPI_C(nbmoments*(ir-l_istart)+1+&
                (startbuf+itheta-l_jstart)*nbmoments*dom_r) 
              int_I2_dvpar(1+ir,1+itheta) = &
                RRecvBufMPI_C(nbmoments*(ir-l_istart)+2+&
                (startbuf+itheta-l_jstart)*nbmoments*dom_r) 
              int_I3_dvpar(1+ir,1+itheta) = &
                RRecvBufMPI_C(nbmoments*(ir-l_istart)+3+&
                (startbuf+itheta-l_jstart)*nbmoments*dom_r) 
              int_I4_dvpar(1+ir,1+itheta) = &
                RRecvBufMPI_C(nbmoments*(ir-l_istart)+4+&
                (startbuf+itheta-l_jstart)*nbmoments*dom_r) 
            enddo

            !*** computation of the integrals in mu ***
            do ir = f%istart, f%iend
              !*** -> \int (\int I0 Jv dvpar) dmu
              Romp10_1Nrp1_1Nthetap1(imu)%val(1+ir,1+itheta) = &
                coeff_intdmu*int_I0_dvpar(1+ir,1+itheta)
              !*** -> \int (\int I1 Jv dvpar) dmu
              Romp11_1Nrp1_1Nthetap1(imu)%val(1+ir,1+itheta) = &
                coeff_intdmu*int_I1_dvpar(1+ir,1+itheta)
              !*** -> \int (\int I2 Jv dvpar) dmu ***
              Romp12_1Nrp1_1Nthetap1(imu)%val(1+ir,1+itheta) = &
                coeff_intdmu*int_I2_dvpar(1+ir,1+itheta)
              !*** ->  \int (\int I3 Jv dvpar) dmu ***
              Romp13_1Nrp1_1Nthetap1(imu)%val(1+ir,1+itheta) = &
                coeff_intdmu*int_I3_dvpar(1+ir,1+itheta)
              !*** ->  \int (\int I4 Jv dvpar) dmu ***
              Romp14_1Nrp1_1Nthetap1(imu)%val(1+ir,1+itheta) = &
                coeff_intdmu*int_I4_dvpar(1+ir,1+itheta)
            end do
          end do

          do ir = f%istart, f%iend
            sImean0(ir,itheta,iphi) = 0._RKIND
            sImean1(ir,itheta,iphi) = 0._RKIND
            sImean2(ir,itheta,iphi) = 0._RKIND
            sImean3(ir,itheta,iphi) = 0._RKIND
            sImean4(ir,itheta,iphi) = 0._RKIND
          end do

          !*** computation of the integrals in mu ***
          do imu = 0,Nmu
            do ir = f%istart, f%iend
              !*** -> Imean0 = \int (\int I0 Jv dvpar) dmu
              sImean0(ir,itheta,iphi) = sImean0(ir,itheta,iphi) + &
                Romp10_1Nrp1_1Nthetap1(imu)%val(1+ir,1+itheta)
              !*** -> Imean1 = \int (\int I1 Jv dvpar) dmu
              sImean1(ir,itheta,iphi) = sImean1(ir,itheta,iphi) + &
                Romp11_1Nrp1_1Nthetap1(imu)%val(1+ir,1+itheta)
              !*** -> Imean2 = \int (\int I2 Jv dvpar) dmu
              sImean2(ir,itheta,iphi) = sImean2(ir,itheta,iphi) + &
                Romp12_1Nrp1_1Nthetap1(imu)%val(1+ir,1+itheta)
              !*** -> Imean3 = \int (\int I3 Jv dvpar) dmu
              sImean3(ir,itheta,iphi) = sImean3(ir,itheta,iphi) + &
                Romp13_1Nrp1_1Nthetap1(imu)%val(1+ir,1+itheta)
              !*** -> Imean4 = \int (\int I4 Jv dvpar) dmu
              sImean4(ir,itheta,iphi) = sImean4(ir,itheta,iphi) + &
                Romp14_1Nrp1_1Nthetap1(imu)%val(1+ir,1+itheta)
            end do
          end do
        end do
!$OMP END DO
!$OMP BARRIER
      end do
!$OMP END PARALLEL
    end if
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,timu)

    do tid = 1, Nbthread
       call FTR_DEALLOC(Romp1_1Nrp1_1Nthetap1(tid)%val)
       call FTR_DEALLOC(Romp2_1Nrp1_1Nthetap1(tid)%val)
       call FTR_DEALLOC(Romp3_1Nrp1_1Nthetap1(tid)%val)
       call FTR_DEALLOC(Romp4_1Nrp1_1Nthetap1(tid)%val)
       call FTR_DEALLOC(Romp5_1Nrp1_1Nthetap1(tid)%val)
    enddo

    do tid = 0, MAX(Nmu,Nbthread)
       call FTR_DEALLOC(Romp10_1Nrp1_1Nthetap1(tid)%val)
       call FTR_DEALLOC(Romp11_1Nrp1_1Nthetap1(tid)%val)
       call FTR_DEALLOC(Romp12_1Nrp1_1Nthetap1(tid)%val)
       call FTR_DEALLOC(Romp13_1Nrp1_1Nthetap1(tid)%val)
       call FTR_DEALLOC(Romp14_1Nrp1_1Nthetap1(tid)%val)
    enddo

   call FTR_DEALLOC(RRecvBufMPI_C)
	
    !*** distribution of the five integrals ***
    !***   Imean<num>(r=*,theta=*,phi=*)    ***
    call clck_time(tfin)
    call clck_ldiff(tdeb,tfin,tcomm3)
    
    !*** periodic conditions in phi ***
    do itheta = f%jstart, f%jend
      do ir = f%istart, f%iend
        sImean0(ir,itheta,f%n3) = sImean0(ir,itheta,0)
        sImean1(ir,itheta,f%n3) = sImean1(ir,itheta,0)
        sImean2(ir,itheta,f%n3) = sImean2(ir,itheta,0)
        sImean3(ir,itheta,f%n3) = sImean3(ir,itheta,0)
        sImean4(ir,itheta,f%n3) = sImean4(ir,itheta,0)
      end do
    end do

    nbelements = dom_r*dom_theta*(Nphi+1)
    call MPI_BCAST(sImean0(f%istart,f%jstart,0),nbelements, &
      MPI_REAL8,0,mpi_comm_intermu,ierr)    
    call MPI_BCAST(sImean1(f%istart,f%jstart,0),nbelements, &
      MPI_REAL8,0,mpi_comm_intermu,ierr)  
    call MPI_BCAST(sImean2(f%istart,f%jstart,0),nbelements, &
      MPI_REAL8,0,mpi_comm_intermu,ierr)     
    call MPI_BCAST(sImean3(f%istart,f%jstart,0),nbelements, &
      MPI_REAL8,0,mpi_comm_intermu,ierr) 
    call MPI_BCAST(sImean4(f%istart,f%jstart,0),nbelements, &
      MPI_REAL8,0,mpi_comm_intermu,ierr)

#ifdef TIMER    
    if (pglobal_id .eq. 0) then
      write(6,'(I4,A,F13.7,A,F13.7,A,F13.7,A,F13.7,A,F13.7)') &
        pglobal_id," Temps comp_mean, ivpar ", &
        tivpar," comm1 ",tcomm1," comm2 ", tcomm2, &
        " timu ", timu," comm3 ",tcomm3
    end if
#endif
#endif 
    ! HAS_COLL_V2
  end subroutine compute_mean_integrals_vers2

  !----------------------------------------------------------
  ! Computation of the mean velocity (Vmean) 
  !   and the mean temperature (Tmean) as
  !  . Tmean = P^{-1}[Imean0*Imean2 - Imean1*Imean1] and
  !  . Vmean = P^{-1}[Imean4*Imean1 - Imean3*Imean2] with
  !  . P  = Imean0*Imean4 - Imean1*Imean3.
  !   where the 5 integrals are defined as:
  !     Imean0=<Dpar> ; Imean1=<vpar*Dpar> ; 
  !     Imean2=<vpar^2*Dpar> ; 
  !     Imean3=<1/Bstar*d/dvpar(Bstar*Dpar)> 
  !   and Imean4=<1/Bstar*d/dvpar(Bstar*vpar*Dpar>
  !----------------------------------------------------------
  subroutine compute_mean_TandV_vers2(geom,init_prof, &
    init_magnet,init_curr,f)
    use clock_module
    use globals, only : istart, iend, jstart, jend, Nphi, &
      Tmean_loc, Vmean_loc
    type(geometry)     , intent(in) :: geom
    type(init_profile) , intent(in) :: init_prof
    type(init_magnetic), intent(in) :: init_magnet
    type(init_current) , intent(in) :: init_curr
    type(fdistribu5d)  , intent(in) :: f

#ifdef HAS_COLL_V2

    integer     :: ir, itheta, iphi
    real(RKIND) :: Imean0_tmp, Imean1_tmp
    real(RKIND) :: Imean2_tmp, Imean3_tmp, Imean4_tmp
    real(RKIND) :: Pmean_tmp

    call clck_time(bclock_meanTandV)

    !*** computation of the 5 integrals:             ***
    !***   Imean0, Imean1, Imean2, Imean3 and Imean4 ***
    call compute_mean_integrals_vers2(geom,init_prof, &
      init_magnet,init_curr,f,Imean0,Imean1,Imean2, &
      Imean3,Imean4)

!$OMP PARALLEL default(none) &
!$OMP private(ir,itheta,iphi, &
!$OMP  Imean0_tmp,Imean1_tmp,Imean2_tmp,Imean3_tmp, &
!$OMP  Imean4_tmp,Pmean_tmp) &
!$OMP shared(istart,iend,jstart,jend,Nphi, &
!$OMP  Imean0,Imean1,Imean2,Imean3,Imean4, &
!$OMP  Tmean_loc,Vmean_loc) 
!$OMP BARRIER

!$OMP DO SCHEDULE(STATIC)
    !*** computation of Pmean, Tmean and Vmean ***
    do iphi = 0,Nphi
      do itheta = jstart,jend
        do ir = istart,iend
          Imean0_tmp = Imean0(ir,itheta,iphi)
          Imean1_tmp = Imean1(ir,itheta,iphi)
          Imean2_tmp = Imean2(ir,itheta,iphi)
          Imean3_tmp = Imean3(ir,itheta,iphi)
          Imean4_tmp = Imean4(ir,itheta,iphi)
          Pmean_tmp  = Imean0_tmp*Imean4_tmp-Imean1_tmp*Imean3_tmp
          Tmean_loc(ir,itheta,iphi) = (Imean0_tmp*Imean2_tmp - &
            Imean1_tmp*Imean1_tmp)/Pmean_tmp
          Vmean_loc(ir,itheta,iphi) = (Imean4_tmp*Imean1_tmp - &
            Imean3_tmp*Imean2_tmp)/Pmean_tmp
        end do
      end do
    end do
!$OMP END DO
!$OMP END PARALLEL

    call clck_time(eclock_meanTandV)
    call clck_diff(bclock_meanTandV,eclock_meanTandV, &
      global_time_meanTandV)
#endif 
    ! HAS_COLL_V2
  end subroutine compute_mean_TandV_vers2


  !------------------------------------------------------------
  ! Reconstruction of the global radial profiles 
  !  necessary for the output saving in the file 
  !  "collisions_<num_diag>.h5"
  !------------------------------------------------------------
  subroutine save_glob_coll_vers2(geom,init_prof)
    use globals        , only : istart, iend, Nr, &
      jstart, jend, Ntheta, Nphi, R0, &
      mpi_comm_row, mpi_comm_column, mpi_comm_mu, &
      dom_r, Nbproc_r, dom_theta, Nbproc_theta, &
      cible_istart, cible_jstart, &
      Rarray_PNr_nbM5, Rarray_Nr_nbM5, &
      Rarray_PNthetaNphi_nbM2, Rarray_NthetaNphi_nbM2, &
      Tmean_loc, Vmean_loc, mu_id, jstart_id, &
      hat_As, hat_Zs, version_coll, major_species
    use globals        , only : hat_As, pSpecies_id
    use MPIutils_module, only : comm_gather2D
    use mpiMod
    type(init_profile), intent(in)  :: init_prof
    type(geometry)    , intent(in)  :: geom

#ifdef HAS_COLL_V2
    integer     :: ir, itheta, iphi 
    integer     :: ir_diag, itheta_diag, iphi_diag 
    integer     :: nbval, nbelem, ierr    
    integer     :: base, dep, j, iproc_theta
    integer     :: cible_istart_loc, cible_jstart_loc
    real(RKIND) :: inv_sqrtAs, nustar_sp
    real(RKIND) :: epstemp, inv_LthetaLphi
    real(RKIND) :: q_on_qpic, rpic_on_r, epstemp_rp
    real(RKIND) :: Epic_on_E, Tpic_on_T, nustar_prof
    real(RKIND) :: Coul_logarithm_ii, norm_nuii, hatZsprime_on_hatZs
    real(RKIND) :: hatAsprime_on_hatAs, vT
    real(RKIND) :: Coul_logarithm_ii_rp, norm_nuii_rp
    real(RKIND) :: collision_freqr_loc_rp, vT_rp
    real(RKIND), dimension(istart:iend) :: Vmean_avg_loc
    real(RKIND), dimension(istart:iend) :: Tmean_avg_loc
    real(RKIND), dimension(istart:iend) :: Vmean_avg_tmp
    real(RKIND), dimension(istart:iend) :: Tmean_avg_tmp

    !*** Initialisation of nustar value depending ***
    !***  on each species                         ***
    nustar_sp = nustars(pSpecies_id)

    !*** computation of the radial profiles of V and T ***
    !*** averaged in phi and locally averaged in theta ***
    Vmean_avg_loc(istart:iend) = 0._RKIND
    Tmean_avg_loc(istart:iend) = 0._RKIND
    inv_LthetaLphi             = 1._RKIND/(geom%Ltheta*geom%Lphi) 
    do iphi = 0,geom%Nphi
      do itheta = jstart,jend
        do ir = istart,iend
          Vmean_avg_loc(ir) = Vmean_avg_loc(ir) + &
            Vmean_loc(ir,itheta,iphi) * inv_LthetaLphi * &
            geom%coeff_intdtheta(itheta) * geom%coeff_intdphi(iphi)
          Tmean_avg_loc(ir) = Tmean_avg_loc(ir) + &
            Tmean_loc(ir,itheta,iphi) * inv_LthetaLphi * &
            geom%coeff_intdtheta(itheta) * geom%coeff_intdphi(iphi)
        end do
      end do
    end do

    !*** computation of the global integral in theta direction ***
    call MY_Allreduce(Vmean_avg_loc,Vmean_avg_tmp,dom_r, &
      MPI_REAL8,MPI_SUM,mpi_comm_row,ierr)
    call MY_Allreduce(Tmean_avg_loc,Tmean_avg_tmp,dom_r, &
      MPI_REAL8,MPI_SUM,mpi_comm_row,ierr)

    itheta_diag = 0
    iphi_diag   = int(Nphi/3)
    nbval  = 5
    nbelem = nbval*dom_r
    if ( (jstart.le.itheta_diag).and.(itheta_diag.le.jend) ) then
      !*** saving of the data in an array distributed in r *** 
      do ir = istart,iend
        !-> radial envelope
        Rarray_PNr_nbM5(0,ir) = envelope_r_loc(ir)
        !-> radial profiles of P, V and T at a fixed point in
        !->  theta and phi
        Rarray_PNr_nbM5(1,ir) = Vmean_loc(ir,itheta_diag,iphi_diag)
        Rarray_PNr_nbM5(2,ir) = Tmean_loc(ir,itheta_diag,iphi_diag)
        !-> radial profiles of P, V and T averaged in theta and phi
        Rarray_PNr_nbM5(3,ir) = Vmean_avg_tmp(ir)
        Rarray_PNr_nbM5(4,ir) = Tmean_avg_tmp(ir)
      end do

      call comm_gather2D(Rarray_PNr_nbM5(0:,istart:),nbelem,0, &
        dom_r,cible_jstart,Rarray_Nr_nbM5(0:,0:),mpi_comm_column)
    end if

    nbval  = 5
    nbelem           = nbval*(Nr+1)
    base             = itheta_diag/dom_theta
    dep              = cible_jstart
    cible_jstart_loc = base * Nbproc_r + dep

    call MPI_BCAST(Rarray_Nr_nbM5,nbelem,MPI_REAL8, &
      cible_jstart_loc,mpi_comm_mu,ierr)

    !*** nuss(r) = (1/sqrt(hat_As))*nustar*vT(r)* ***
    !***            eps^(3/2)/(q R0)              ***
    inv_sqrtAs = 1._RKIND/sqrt(hat_As(pSpecies_id))
    do ir = 0,Nr
      envelope_r_diag(ir)      = Rarray_Nr_nbM5(0,ir)
      Vmean_r_diag(ir)         = Rarray_Nr_nbM5(1,ir)
      Tmean_r_diag(ir)         = Rarray_Nr_nbM5(2,ir)
      Vmean_r_avg_diag(ir)     = Rarray_Nr_nbM5(3,ir)
      Tmean_r_avg_diag(ir)     = Rarray_Nr_nbM5(4,ir)
      epstemp                  = geom%rg(ir)*inv_R(ir,0)
      if (init_prof%iota(ir).eq.0._RKIND) then
        ! specific case if iota = 0 (slab geometry)
        collision_freqr_diag(ir) = inv_sqrtAs*nustar_sp * &
          sqrt(Tmean_r_diag(ir))*epstemp* &
          sqrt(epstemp)/R0
      elseif ((init_prof%iota(ir).ne.0._RKIND) &
        .and.(version_coll==3)) then
        ! properly define nustar(r) and the collision 
        !  frequency(r) in the toroidal case when 
        !  (version_coll==3), as in Eq.(12) [Dif-Pradalier
        !  PoP 2011]. So far, this is done for 
        !  ion-ion collisions only
        ! NB: proper definitions of the Coulomb 
        !     logarithm are required as density and 
        !     temperatures are non-dimensional so far...
        hatZsprime_on_hatZs = hat_Zs(pSpecies_id) / &
          hat_Zs(major_species)
        hatAsprime_on_hatAs = hat_As(pSpecies_id) / &
          hat_As(major_species)
        vT    = sqrt( init_prof%Ts0(major_species,ir) )
        vT_rp = sqrt( init_prof%Ts0_rp(major_species) )
        ! Coulomb_logarithm: should be the electron density,
        !  profile is right but magnitude is wrong: 
        !  should be log(n_e/10e20) and temperature should be in keV
        !---!! computed with INITIAL profiles of  !!---
        !---!!  temperature & density             !!---
        Coul_logarithm_ii    = 17.3_RKIND - &
          0.5_RKIND*log(init_prof%ns0(major_species,ir)) + &
          1.5_RKIND*log(init_prof%Ts0(major_species,ir))
        Coul_logarithm_ii_rp = 17.3_RKIND - &
          0.5_RKIND*log(init_prof%ns0_rp(major_species)) + &
          1.5_RKIND*log(init_prof%Ts0_rp(major_species))
        norm_nuii            = Coul_logarithm_ii * &
          4._RKIND*sqrt(pi)/3._RKIND
        norm_nuii_rp         = Coul_logarithm_ii_rp * &
          4._RKIND*sqrt(pi)/3._RKIND
        ! collision_freqr is correct for the ion-ion collisions
        collision_freqr_diag(ir) = norm_nuii * &
          init_prof%ns0(major_species,ir) * &
          hatZsprime_on_hatZs**4 / &
          ( hatAsprime_on_hatAs**2 * vT**3 )
        collision_freqr_loc_rp  = norm_nuii_rp * &
          init_prof%ns0_rp(major_species) * &
          hatZsprime_on_hatZs**4 / &
          ( hatAsprime_on_hatAs**2 * (vT_rp)**3 )
        !---!! computed with the INITIAL profile of !!---
        !---!!  temperature                         !!---
        q_on_qpic   = 1._RKIND/(init_prof%q_rp*init_prof%iota(ir))
        rpic_on_r   = init_prof%rpeak / geom%rg(ir)
        epstemp_rp  = geom%rg(int(Nr/2)+1)*inv_R(int(Nr/2)+1,0)
        Epic_on_E   = epstemp_rp / epstemp
        Tpic_on_T   = init_prof%Ts0_rp(major_species) / &
          init_prof%Ts0(major_species,ir) ! Tprofile(ir)
        nustar_prof = q_on_qpic*sqrt(Tpic_on_T) * &
          Epic_on_E*sqrt(Epic_on_E) * &
          collision_freqr_diag(ir) / collision_freqr_loc_rp
        !--> nuss(r) = (1/sqrt(hat_As))*nustar*vT(r) * 
        !-->           eps^(3/2)/(q R0)
        ! collisionality, with its radial profile, 
        !  as used in AD computation
        nustarr_diag(ir) = nustar_sp * nustar_prof
      else
        collision_freqr_diag(ir) = inv_sqrtAs* nustar_sp * &
          sqrt(Tmean_r_diag(ir))*epstemp* &
          sqrt(epstemp)*init_prof%iota(ir)/R0
      end if
    end do

    ir_diag = int(Nr/2)
    nbval   = 2
    if ( (ir_diag.ge.istart).and.(ir_diag.le.iend) ) then
      do iphi = 0,Nphi
        do itheta = jstart,jend 
          Rarray_PNthetaNphi_nbM2((itheta-jstart)*nbval+0,iphi) = &
            Vmean_loc(ir_diag,itheta,iphi)
          Rarray_PNthetaNphi_nbM2((itheta-jstart)*nbval+1,iphi) = &
            Tmean_loc(ir_diag,itheta,iphi)
        end do
      end do
      nbelem = nbval*dom_theta*(Nphi+1)
      call comm_gather2D(Rarray_PNthetaNphi_nbM2(0:,0:), &
        nbelem,0,(Nphi+1),cible_istart, &
        Rarray_NthetaNphi_nbM2(0:,0:),mpi_comm_row)
    end if
    nbelem           = nbval*(Ntheta)*(Nphi+1)
    base             = cible_istart
    dep              = ir_diag/dom_r
    cible_istart_loc = base * Nbproc_r + dep

    call MPI_BCAST(Rarray_NthetaNphi_nbM2,nbelem,MPI_REAL8, &
      cible_istart_loc,mpi_comm_mu,ierr)

    do iphi = 0,Nphi
      do iproc_theta = 0,Nbproc_theta-1
        do j = 0,dom_theta-1
          itheta = iproc_theta*dom_theta+j
          Vmean_thetaphi_diag(itheta,iphi) = &
            Rarray_NthetaNphi_nbM2(j*nbval+0,iphi+(Nphi+1) * &
            iproc_theta)
          Tmean_thetaphi_diag(itheta,iphi) = &
            Rarray_NthetaNphi_nbM2(j*nbval+1,iphi+(Nphi+1) * &
            iproc_theta)
        end do
      end do
    end do
    !*** periodicity in theta direction ***
    Vmean_thetaphi_diag(Ntheta,0:Nphi) = &
      Vmean_thetaphi_diag(0,0:Nphi)
    Tmean_thetaphi_diag(Ntheta,0:Nphi) = &
      Tmean_thetaphi_diag(0,0:Nphi)
#endif 
    ! HAS_COLL_V2
  end subroutine save_glob_coll_vers2


  !-------------------------------------------------------
  ! Choice of the advection for the collision operator
  !-------------------------------------------------------
  subroutine AdvecDiff_coll_vpar(geom,dt,init_prof,&
    init_magnet,init_curr,f)
    use globals, only : version_coll
    type(geometry)     , intent(in)    :: geom
    type(init_profile) , intent(in)    :: init_prof
    type(init_magnetic), intent(in)    :: init_magnet
    type(init_current) , intent(in)    :: init_curr
    real(RKIND)        , intent(in)    :: dt
    type(fdistribu5d)  , intent(inout) :: f
!R3 #include "code_instrumentation/r3_info.h" !R3
    
!R3 call r3_info_begin (r3_info_index_0, 'CPU_collisions')      !R3
    call Write_dyn_balise_open('CPU_collisions') !R4
    call check_overflow_4D(f,"Before Diff_coll_vpar")
    
    PRINT_CKSUM(version_coll)
    
    if (version_coll.eq.1) then
      call AdvecDiff_coll_vers1(geom,dt,f)
    else
      call AdvecDiff_coll_vers2(geom,dt,init_prof,init_magnet, &
        init_curr,f)
    end if
    call check_overflow_4D(f,"After Diff_coll_vpar")
    call Write_dyn_balise_close('CPU_collisions') !R4
!R3 call r3_info_end (r3_info_index_0) !R3
  end subroutine AdvecDiff_coll_vpar

end module collisions_vpar_module
