!**************************************************************
!  Copyright Euratom-CEA
!  Authors : 
!     Virginie Grandgirard (virginie.grandgirard@cea.fr)
!     Chantal Passeron (chantal.passeron@cea.fr)
!     Guillaume Latu (guillaume.latu@cea.fr)
!     Xavier Garbet (xavier.garbet@cea.fr)
!     Philippe Ghendrih (philippe.ghendrih@cea.fr)
!     Yanick Sarazin (yanick.sarazin@cea.fr)
!  
!  This code GYSELA (for GYrokinetic SEmi-LAgrangian) 
!  is a 5D gyrokinetic global full-f code for simulating 
!  the plasma turbulence in a tokamak.
!  
!  This software is governed by the CeCILL-B license 
!  under French law and abiding by the rules of distribution 
!  of free software.  You can  use, modify and redistribute 
!  the software under the terms of the CeCILL-B license as 
!  circulated by CEA, CNRS and INRIA at the following URL
!  "http://www.cecill.info". 
!**************************************************************

#include "config-F.inc"

!------------------------------------------------
! file : diffusion.F90
! date : 2012-02-28
!
!------------------------------------------------
module diffusion_module
  use prec_const

  implicit none

  real(RKIND), dimension(:), pointer, public :: diff_coeffDr       => NULL()
  real(RKIND), dimension(:), pointer, public :: diff_coeffrDr_hf   => NULL()
  real(RKIND), dimension(:), pointer, public :: diff_coeffDth	   => NULL()
  real(RKIND), dimension(:), pointer, public :: diff_A, diff_Ath   => NULL()
  real(RKIND), dimension(:), pointer, public :: diff_B, diff_Bth   => NULL()
  real(RKIND), dimension(:), pointer, public :: diff_C, diff_Cth   => NULL()

  !******************************
  contains
  !******************************

  !----------------------------------------------
  ! Initialisation of the diffusion coefficients
  !----------------------------------------------
  subroutine new_diffusion(sNr)
    use wrap_alloc_module
    integer, intent(in) :: sNr

    !--> for diffusion definition
    call FTR_ALLOC_1_ZERO(diff_coeffDr,0,sNr)
    call FTR_ALLOC_1_ZERO(diff_coeffrDr_hf,0,sNr+1)
    call FTR_ALLOC_1_ZERO(diff_coeffDth,0,sNr)
    call FTR_ALLOC_1_ZERO(diff_A,0,sNr)
    call FTR_ALLOC_1_ZERO(diff_B,0,sNr)
    call FTR_ALLOC_1_ZERO(diff_C,0,sNr)
    call FTR_ALLOC_1_ZERO(diff_Ath,0,sNr)
    call FTR_ALLOC_1_ZERO(diff_Bth,0,sNr)
    call FTR_ALLOC_1_ZERO(diff_Cth,0,sNr)
  end subroutine new_diffusion


  !----------------------------------------------
  ! Delete the diffusion coefficients
  !----------------------------------------------
  subroutine del_diffusion()
    use wrap_alloc_module

    call FTR_DEALLOC(diff_coeffDr)
    call FTR_DEALLOC(diff_coeffrDr_hf)
    call FTR_DEALLOC(diff_coeffDth)
    call FTR_DEALLOC(diff_A)
    call FTR_DEALLOC(diff_B)
    call FTR_DEALLOC(diff_C)
    call FTR_DEALLOC(diff_Ath)
    call FTR_DEALLOC(diff_Bth)
    call FTR_DEALLOC(diff_Cth)
  end subroutine del_diffusion


  !----------------------------------------------------
  ! initialization of a dissipation coefficient profile
  !----------------------------------------------------
  subroutine init_dissip_coeff(geom,init_prof)
    use globals, only : pSpecies_id, coef_diffr, coef_diffth
    use geometry_class
    use init_profile_class
    type(geometry)    , intent(in) :: geom
    type(init_profile), intent(in) :: init_prof

#ifdef HAS_DIFF
    integer :: ir, Nr_loc

    Nr_loc = geom%Nr

    !*** Taking into account of the buffer region profile ***
    do ir = 0,Nr_loc
      diff_coeffDr(ir)  = coef_diffr(pSpecies_id)*init_prof%buffer_r(ir)
      diff_coeffDth(ir) = coef_diffth(pSpecies_id)*init_prof%buffer_r(ir)
    end do

    !*** Cylindrical operator: computation at (i - 1/2)      ***
    !*** Let us define chi(r) = diff_coeffDr(r), then        ***
    !***  D(i-1/2) = 1/4*(r_i*chi_i + r_i*chi_{i-1} +        ***
    !***                  r_{i-1}*chi_i + r_{i-1}*chi_{i-1}) *** 
    do ir = 1,Nr_loc
      diff_coeffrDr_hf(ir) = &
        (geom%rg(ir)*diff_coeffDr(ir) + &
        geom%rg(ir)*diff_coeffDr(ir-1) + &
        geom%rg(ir-1)*diff_coeffDr(ir) + &
        geom%rg(ir-1)*diff_coeffDr(ir-1))/4._RKIND
    end do
    !*** Boundary condition = vanishing gradient        *** 
    !*** Using the assumptions chi_{-1} = 2*chi_0-chi_1 ***
    !***  and chi_{N+1} = 2*chi_N-chi_{N-1}             ***
    !***  D(-1/2)  = 1/2(r_0-dr/2)(3*chi_0-chi_1)       ***
    !***  D(N+1/2) = 1/2(r_N+dr/2)(3*chi_N-chi_{N-1})   ***
    diff_coeffrDr_hf(0)    = &
      1.5_RKIND*diff_coeffDr(0) * &
      (geom%rg(0)-0.5_RKIND*geom%dr) - &
      (0.5_RKIND*geom%rg(0)-0.25_RKIND*geom%dr) * &
      diff_coeffDr(1)
    diff_coeffrDr_hf(Nr_loc+1) = &
      1.5_RKIND*diff_coeffDr(Nr_loc)* &
      (geom%rg(Nr_loc)+0.5_RKIND*geom%dr) - &
      (0.5_RKIND*geom%rg(Nr_loc)+0.25_RKIND*geom%dr) * &
      diff_coeffDr(Nr_loc-1)
#endif 
    ! HAS_DIFF
  end subroutine init_dissip_coeff


  !----------------------------------------------------
  ! initialization of the matrix coefficients A, B, C
  ! used to solve the diffusive transport with 
  ! the Crank-Nicolson numerical scheme
  !----------------------------------------------------
  subroutine init_diffus_matrix(geom)
    use globals, only : pSpecies_id, flux_driven    
    use geometry_class
    type(geometry), intent(in) :: geom

#ifdef HAS_DIFF
    real(RKIND) :: alpha0, alpha, coef_A, coef_B, coef_C
    real(RKIND) :: alphath0, alphath
    real(RKIND) :: coef_Ath, coef_Bth, coef_Cth
    integer     :: ir, Nr_loc

    Nr_loc   = geom%Nr
    alpha0   = 1._RKIND/(2._RKIND*geom%dr*geom%dr)
    alphath0 = 1._RKIND/(2._RKIND*geom%dtheta*geom%dtheta)

    do ir = 0,Nr_loc
      alpha        = alpha0/geom%rg(ir)
      coef_A       = alpha*diff_coeffrDr_hf(ir)
      coef_C       = alpha*diff_coeffrDr_hf(ir+1)
      coef_B       = coef_A+coef_C
      diff_A(ir)   = -coef_A
      diff_B(ir)   =  coef_B
      diff_C(ir)   = -coef_C
      alphath      = alphath0/(geom%rg(ir)*geom%rg(ir))
      coef_Ath     = alphath*diff_coeffDth(ir)
      coef_Cth     = coef_Ath
      coef_Bth     = coef_Ath+coef_Cth
      diff_Ath(ir) = -coef_Ath
      diff_Bth(ir) =  coef_Bth
      diff_Cth(ir) = -coef_Cth
    end do
    ! Vanishing gradient at rmin      
    !  if the system is flux driven
    if (flux_driven) then
      diff_C(0) = diff_A(0) + diff_C(0)
    else
      diff_B(0) = 0._RKIND
      diff_C(0) = 0._RKIND
    end if                                       
    diff_B(Nr_loc) = 0._RKIND
    diff_A(Nr_loc) = 0._RKIND
#endif 
    ! HAS_DIFF
  end subroutine init_diffus_matrix


  !-----------------------------------------------------------
  !  Radial diffusion operator (cylindrical geometry):
  !    d/dt F  =  1/r d/dr [ r D(r) d/dr (F-Feq) ]
  !  Numerical scheme: Crank-Nicolson
  !-----------------------------------------------------------
  subroutine advance_Diff_r(Sfmu_eq,dt,f)
    use clock_module
    use globals        , only : pSpecies_id, flux_driven
    use globals        , only : inumber, deltaF_in_diff
#ifdef _OPENMP
    use OMPutils_module, only : omp_get_thread_num
#endif
    use OMPutils_module, only : Romp1_0Nr, Romp2_0Nr, &
      Romp3_0Nr, Romp4_0Nr, Romp5_0Nr
    use fdistribu5d_class
!R3 #include "code_instrumentation/r3_info.h" !R3
    real(RKIND), dimension(:,:,:), pointer       :: Sfmu_eq
    real(RKIND)                  , intent(in)    :: dt
    type(fdistribu5d)            , intent(inout) :: f

#ifdef HAS_DIFF
    integer     :: tid
    real(RKIND) :: test_istart, test_iend
    integer     :: ir, ir0, itheta, iphi, ivpar

    real(RKIND), dimension(:), pointer :: floc  ! (0:inumber-1)
    real(RKIND), dimension(:), pointer :: AA    ! (0:inumber-1)
    real(RKIND), dimension(:), pointer :: BB    ! (0:inumber-1)
    real(RKIND), dimension(:), pointer :: CC    ! (0:inumber-1)
    real(RKIND), dimension(:), pointer :: RR    ! (0:inumber-1)

!R3 call r3_info_begin (r3_info_index_0, 'CPU_Diffus_r')    !R3
    call Write_dyn_balise_open('CPU_Diffus_r') !R4
    call clck_time(bclock_advance_diffr)

    ! =1 if istart=inner boundary, =0 otherwise
    test_istart = 0.5_RKIND*dfloat((1-sign(1,f%istart-1)))
    ! =1 if iend=outer boundary, =0 otherwise
    test_iend   = 0.5_RKIND*dfloat((1+sign(1,f%iend-f%n1)))

#ifdef _OPENMP
!$OMP PARALLEL private(tid,ir,ir0,itheta,iphi,ivpar,floc, &
!$OMP AA,BB,CC,RR) default(shared)
!$OMP BARRIER
    tid = 1+omp_get_thread_num()
#else
    tid = 1
#endif

    floc => Romp1_0Nr(tid)%val
    AA   => Romp2_0Nr(tid)%val
    BB   => Romp3_0Nr(tid)%val
    CC   => Romp4_0Nr(tid)%val
    RR   => Romp5_0Nr(tid)%val
    do ir = f%istart, f%iend
      ir0     = ir-f%istart
      AA(ir0) = dt*diff_A(ir)
      BB(ir0) = 1._RKIND + dt*diff_B(ir)
      CC(ir0) = dt*diff_C(ir)
    end do

!$OMP DO SCHEDULE(static)
    do ivpar = 0, f%n4
      do iphi = 0, f%n3-1
        do itheta = f%jstart, f%jend
          do ir = f%istart+1, f%iend-1
            ir0       = ir-f%istart
            floc(ir0) = f%values(ir,itheta,iphi,ivpar)
            RR(ir0)   = &
              -AA(ir0)*f%values(ir-1,itheta,iphi,ivpar) + &
              (2._RKIND-BB(ir0))*f%values(ir,itheta,iphi,ivpar) - &
              CC(ir0)*f%values(ir+1,itheta,iphi,ivpar)
            if (deltaF_in_diff) then
              RR(ir0)   = RR(ir0) + 2._RKIND * &
                (AA(ir0)*Sfmu_eq(ir-1,itheta,ivpar) + &
                (BB(ir0)-1._RKIND)*Sfmu_eq(ir,itheta,ivpar) + &
                CC(ir0)*Sfmu_eq(ir+1,itheta,ivpar))
            end if
          end do
          if (flux_driven .and. test_istart .eq. 1._RKIND) then
            RR(0) = (2._RKIND-BB(0)) * &
              f%values(f%istart,itheta,iphi,ivpar) - &
              CC(0)*f%values(f%istart+1,itheta,iphi,ivpar)
            if (deltaF_in_diff) then
              RR(0) = RR(0) - &
                2._RKIND*((1._RKIND-BB(0)) * &
                Sfmu_eq(f%istart,itheta,ivpar)- &
                CC(0)*Sfmu_eq(f%istart+1,itheta,ivpar))
            end if
          else
            RR(0) = f%values(f%istart,itheta,iphi,ivpar)
          end if
          RR(inumber-1) = f%values(f%iend,itheta,iphi,ivpar)
          call TRIDIAG(AA,BB,CC,RR,floc,inumber-1)
          do ir = f%istart, f%iend
            ir0 = ir-f%istart
            f%values(ir,itheta,iphi,ivpar) = floc(ir0)
          end do
        end do
      end do
    end do
!$OMP END DO
!$OMP BARRIER
!$OMP END PARALLEL

     call clck_time(eclock_advance_diffr)
     call clck_diff(bclock_advance_diffr,eclock_advance_diffr, &
       global_time_advance_diffr)
    call Write_dyn_balise_close('CPU_Diffus_r') !R4
!R3 call r3_info_end (r3_info_index_0)      !R3
#endif 
    ! HAS_DIFF
   end subroutine advance_Diff_r


  !-----------------------------------------------------------
  !  Poloidal diffusion operator (cylindrical geometry):
  !    d/dt F  =  D_th(r)/r**2 d**2/dtheta**2 (F-Feq)
  !  Numerical scheme: Crank-Nicolson
  !-----------------------------------------------------------
  subroutine advance_Diff_th(Sfmu_eq,dt,ftransp)
    use clock_module
    use globals, only : jnumber, Nr, Ntheta, &
      kstart, kend, lstart, lend, deltaF_in_diff
    use MPIutils_module
#ifdef _OPENMP
    use OMPutils_module, only : omp_get_thread_num
#endif
    use OMPutils_module, only : Romp1_0Ntheta, Romp2_0Ntheta, &
      Romp3_0Ntheta, Romp4_0Ntheta, Romp5_0Ntheta
    use utils_module   , only : cyclic 
    use wrap_alloc_module
!R3 #include "code_instrumentation/r3_info.h" !R3
    real(RKIND), dimension(:,:,:)  , pointer    :: Sfmu_eq
    real(RKIND)                    , intent(in) :: dt
    real(RKIND), dimension(:,:,:,:), pointer    :: ftransp

#ifdef HAS_DIFF
    integer     :: tid
    integer     :: ir, itheta, iphi, ivpar, imix

    real(RKIND), dimension(:), pointer :: floc
    real(RKIND), dimension(:), pointer :: AA  
    real(RKIND), dimension(:), pointer :: BB  
    real(RKIND), dimension(:), pointer :: CC  
    real(RKIND), dimension(:), pointer :: RR  
    real(RKIND) :: AA_loc, BB_loc, CC_loc

!R3 call r3_info_begin (r3_info_index_0, 'CPU_Diffus_th')    !R3
    call Write_dyn_balise_open('CPU_Diffus_th') !R4
    call clck_time(bclock_advance_diffth)

#ifdef _OPENMP
!$OMP PARALLEL private(tid,ir,itheta,iphi,ivpar,imix,floc, &
!$OMP AA,BB,CC,RR,AA_loc,BB_loc,CC_loc) default(shared)
!$OMP BARRIER
    tid = 1+omp_get_thread_num()
#else
    tid = 1
#endif

    floc => Romp1_0Ntheta(tid)%val
    AA   => Romp2_0Ntheta(tid)%val
    BB   => Romp3_0Ntheta(tid)%val
    CC   => Romp4_0Ntheta(tid)%val
    RR   => Romp5_0Ntheta(tid)%val

!$OMP DO SCHEDULE(static,1)
    do imix = 0, (lend-lstart+1)*(kend-kstart+1)-1
      ! Variable imix encapsulates ivpar and iphi together 
      ! in order to parallelize on these two variables 
      ! simultaneously (kind of nested loops parallelization).
      ! Useful on BG/Q machine with many threads (turing).
      ivpar = lstart + mod(imix,lend-lstart+1)
      iphi  = kstart + (imix/(lend-lstart+1))
        do ir = 0, Nr
          AA_loc = dt*diff_Ath(ir)
          BB_loc = 1._RKIND + dt*diff_Bth(ir)
          CC_loc = dt*diff_Cth(ir)
          do itheta = 1, Ntheta-1
            floc(itheta) = ftransp(ir,itheta,iphi,ivpar)
            AA(itheta)   = AA_loc
            BB(itheta)   = BB_loc
            CC(itheta)   = CC_loc
            RR(itheta)   = &
              -AA_loc*ftransp(ir,itheta-1,iphi,ivpar) + &
              (2._RKIND-BB_loc)*ftransp(ir,itheta,iphi,ivpar) - &
              CC_loc*ftransp(ir,itheta+1,iphi,ivpar)
            if (deltaF_in_diff) then
              RR(itheta)   = RR(itheta) + 2._RKIND * &
                (AA_loc*Sfmu_eq(ir,itheta-1,ivpar) + &
                (BB_loc-1._RKIND)*Sfmu_eq(ir,itheta,ivpar) + &
                CC_loc*Sfmu_eq(ir,itheta+1,ivpar))
            end if
          end do
          floc(0) = ftransp(ir,0,iphi,ivpar)
          AA(0)   = AA_loc
          BB(0)   = BB_loc
          CC(0)   = CC_loc
          RR(0)   = -AA_loc*ftransp(ir,Ntheta-1,iphi,ivpar) + &
              (2._RKIND-BB_loc)*ftransp(ir,0,iphi,ivpar) - &
              CC_loc*ftransp(ir,1,iphi,ivpar)
          if (deltaF_in_diff) then
            RR(0) = RR(0) + 2._RKIND * &
              (AA_loc*Sfmu_eq(ir,Ntheta-1,ivpar) + &
              (BB_loc-1._RKIND)*Sfmu_eq(ir,0,ivpar) + &
              CC_loc*Sfmu_eq(ir,1,ivpar))
          end if
          call cyclic(AA,BB,CC,AA_loc,AA_loc,RR,floc,Ntheta-1)
          do itheta = 0, Ntheta-1
            ftransp(ir,itheta,iphi,ivpar) = floc(itheta)
          end do
          ftransp(ir,Ntheta,iphi,ivpar) = floc(0)
        end do
    end do
!$OMP END DO
!$OMP BARRIER
!$OMP END PARALLEL

     call clck_time(eclock_advance_diffth)
     call clck_diff(bclock_advance_diffth,eclock_advance_diffth, &
       global_time_advance_diffth)
    call Write_dyn_balise_close('CPU_Diffus_th') !R4
!R3 call r3_info_end (r3_info_index_0)      !R3
#endif 
    ! HAS_DIFF
  end subroutine advance_Diff_th

end module diffusion_module
