!**************************************************************
!  Copyright Euratom-CEA
!  Authors : 
!     Virginie Grandgirard (virginie.grandgirard@cea.fr)
!     Chantal Passeron (chantal.passeron@cea.fr)
!     Guillaume Latu (guillaume.latu@cea.fr)
!     Xavier Garbet (xavier.garbet@cea.fr)
!     Philippe Ghendrih (philippe.ghendrih@cea.fr)
!     Yanick Sarazin (yanick.sarazin@cea.fr)
!  
!  This code GYSELA (for GYrokinetic SEmi-LAgrangian) 
!  is a 5D gyrokinetic global full-f code for simulating 
!  the plasma turbulence in a tokamak.
!  
!  This software is governed by the CeCILL-B license 
!  under French law and abiding by the rules of distribution 
!  of free software.  You can  use, modify and redistribute 
!  the software under the terms of the CeCILL-B license as 
!  circulated by CEA, CNRS and INRIA at the following URL
!  "http://www.cecill.info". 
!**************************************************************

#include "config-F.inc"

!--------------------------------------------------------
! file : krook_operator
! date : 2012-02-27
!  Krook operator at the RHS of the Vlasov equation, i.e:
!    df/dt = -nu*(f-feq)
! The analytical solution of this equation is:
!  f(t+dt) = exp(-nu*dt)f(t)+feq(1-exp(-nu*dt))
!--------------------------------------------------------
module krook_module
  use fdistribu5d_class
  implicit none

  !--> radial profile of Krook coefficient
  real(RKIND), dimension(:), pointer, public :: nu_r => NULL()

  !******************************
  contains
  !******************************

  !----------------------------------------------
  ! Initialisation of the Krook operator
  !----------------------------------------------
  subroutine init_Krook(sNr,Scoef_krook,buffer_shape)
    use globals, only : pSpecies_id, rflux_driven
    integer                   , intent(in) :: sNr
    real(RKIND)               , intent(in) :: Scoef_krook
    real(RKIND), dimension(0:), intent(in) :: buffer_shape 
    integer :: ir 
    
    call FTR_ALLOC_1_ZERO(nu_r,0,sNr)
    if (.not.memory_test) then
      do ir = 0,sNr
        nu_r(ir) = Scoef_krook*buffer_shape(ir) * &
          (1._RKIND-rflux_driven)
      end do
    end if
  end subroutine init_Krook


  !----------------------------------------------
  ! Delete the Krook operator
  !----------------------------------------------
  subroutine delete_Krook()
    call FTR_DEALLOC(nu_r)
  end subroutine delete_Krook


  !----------------------------------------------
  ! Treatment of the Krook operator
  !   df/dt = - nu (f-feq)
  ! The analytical solution of this equation is:
  !  f(t+dt) = exp(-nu*dt)f(t)+feq(1-exp(-nu*dt))
  !----------------------------------------------
  subroutine advance_Krook(Sfmu_eq,dt,f)
#ifdef _OPENMP
    use OMPutils_module, only : omp_get_thread_num
#endif
    use clock_module
    real(RKIND), dimension(:,:,:), pointer       :: Sfmu_eq
    real(RKIND)                  , intent(in)    :: dt
    type(fdistribu5d)            , intent(inout) :: f

#ifdef HAS_KROOK
    integer     :: tid
    integer     :: ir, itheta, iphi, ivpar
    real(RKIND) :: exp_coeff
!R3 #include "r3_info.h" !R3

!R3 call r3_info_begin (r3_info_index_0, 'CPU_Krook')      !R3
    call Write_dyn_balise_open('CPU_Krook') !R4
    call clck_time(bclock_advance_Krook)
#ifdef _OPENMP
!$OMP PARALLEL private(tid,ir,itheta,iphi,ivpar, &
!$OMP exp_coeff) default(shared)
!$OMP BARRIER
    tid = 1+omp_get_thread_num()
#else
    tid = 1
#endif

    !*** computation for a fixed value of mu ***
!$OMP DO SCHEDULE(static)
    do ivpar = 0,f%n4
      do iphi = 0,f%n3-1
        do itheta = f%jstart,f%jend
          do ir = f%istart,f%iend
            exp_coeff = exp(-nu_r(ir)*dt)
            f%values(ir,itheta,iphi,ivpar) = &
              exp_coeff*f%values(ir,itheta,iphi,ivpar) + &
              Sfmu_eq(ir,itheta,ivpar)*(1._RKIND-exp_coeff)
          end do
        end do
      end do
    end do
!$OMP END DO
!$OMP BARRIER
!$OMP END PARALLEL
    call clck_time(eclock_advance_Krook)
    call clck_diff(bclock_advance_Krook,eclock_advance_Krook, &
      global_time_advance_Krook)
    call Write_dyn_balise_close('CPU_Krook') !R4
!R3 call r3_info_end (r3_info_index_0)      !R3
#endif 
    !HAS_KROOK
  end subroutine advance_Krook
end module krook_module
