!**************************************************************
!  Copyright Euratom-CEA
!  Authors : 
!     Virginie Grandgirard (virginie.grandgirard@cea.fr)
!     Chantal Passeron (chantal.passeron@cea.fr)
!     Guillaume Latu (guillaume.latu@cea.fr)
!     Xavier Garbet (xavier.garbet@cea.fr)
!     Philippe Ghendrih (philippe.ghendrih@cea.fr)
!     Yanick Sarazin (yanick.sarazin@cea.fr)
!  
!  This code GYSELA (for GYrokinetic SEmi-LAgrangian) 
!  is a 5D gyrokinetic global full-f code for simulating 
!  the plasma turbulence in a tokamak.
!  
!  This software is governed by the CeCILL-B license 
!  under French law and abiding by the rules of distribution 
!  of free software.  You can  use, modify and redistribute 
!  the software under the terms of the CeCILL-B license as 
!  circulated by CEA, CNRS and INRIA at the following URL
!  "http://www.cecill.info". 
!**************************************************************

#include "config-F.inc"

!------------------------------------------------
! file : sources.F90
! date : 03/02/2009
!  Initialization of the source terms and 
!  solving of df/dt = S_mu(r,theta,vpar)
!------------------------------------------------
module sources_module
  use prec_const
  use wrap_alloc_module
  use geometry_class
  use init_magnetic_class

  use wrap_alloc_module
  implicit none

  !*** init_sources type definition ***
  type :: init_sources
     !--> for heat source
     integer                                :: heat_version
     real(RKIND)                            :: Sce0, rhoS, LS, TS
     real(RKIND), dimension(:)    , pointer :: rshape
     real(RKIND), dimension(:)    , pointer :: dshapedr
     !--> for density source
     real(RKIND)                            :: Sce_dens
     real(RKIND)                            :: rhoS_dens
     real(RKIND)                            :: LS_dens
     real(RKIND)                            :: TS_dens  
     real(RKIND), dimension(:)    , pointer :: rshape_dens
     real(RKIND), dimension(:)    , pointer :: dshapedr_dens
     !--> for moment source
     real(RKIND)                            :: Sce_mom
     real(RKIND)                            :: rhoS_mom
     real(RKIND)                            :: LS_mom
     real(RKIND)                            :: TS_mom  
     real(RKIND), dimension(:)    , pointer :: rshape_mom
     real(RKIND), dimension(:)    , pointer :: dshapedr_mom
     !--> for vorticity source
     real(RKIND)                            :: Sce_vor
     real(RKIND)                            :: rhoS_vor
     real(RKIND)                            :: LS_vor
     real(RKIND)                            :: TS_vor  
     real(RKIND), dimension(:)    , pointer :: rshape_vor
     real(RKIND), dimension(:)    , pointer :: dshapedr_vor
     !--> for fast particle source
     real(RKIND)                            :: Sce_fp
     real(RKIND)                            :: Sce_fpplus
     real(RKIND)                            :: rhoS_fp
     real(RKIND)                            :: LS_fp
     real(RKIND)                            :: TS_fp
     real(RKIND)                            :: v0_fp
     real(RKIND)                            :: alpha_trapp
     real(RKIND), dimension(:)    , pointer :: rshape_fp
     real(RKIND), dimension(:)    , pointer :: dshapedr_fp
     !--> for total source and its derivatives in r,
     !-->  theta and vpar
     real(RKIND), dimension(:,:,:), pointer :: Sce_mu, dSdr_mu
     real(RKIND), dimension(:,:,:), pointer :: dSdtheta_mu
     real(RKIND), dimension(:,:,:), pointer :: dSdvpar_mu
  end type init_sources

  real(RKIND), dimension(:,:), pointer, private :: mu0_Jpar

  include "gyrokinetic/Bstar_inline.h"                               
  include "gyrokinetic/velocities_inline.h" 
  !******************************
  contains
  !******************************
#include "gyrokinetic/Bstar_inline.F90"                              
#include "gyrokinetic/velocities_inline.F90"                         


  !---------------------------------------------------- 
  ! Constructor : Memory allocation for source and 
  ! derivatives
  !----------------------------------------------------
  subroutine new_init_sources(ithis,geom)
    use globals, only : &
      heat_version, Sce0, LS, rhoS, TS, &
      Sce_dens, TS_dens, rhoS_dens, LS_dens, &
      Sce_mom, TS_mom, rhoS_mom, LS_mom, &
      Sce_vor, TS_vor, rhoS_vor, LS_vor, &
      Sce_fp, Sce_fpplus, TS_fp, v0_fp, &
      rhoS_fp, LS_fp, alpha_trapp
    use globals, only : pSpecies_id, &
      istart, iend, jstart, jend
    type(init_sources), intent(out) :: ithis
    type(geometry)    , intent(in)  :: geom
    integer :: Nr_loc, Ntheta_loc,Nvpar_loc
    
    Nr_loc     = geom%Nr
    Ntheta_loc = geom%Ntheta
    Nvpar_loc  = geom%Nvpar

    !*** variable initialization and memory allocation ***
    !--> for heat source
    ithis%heat_version = heat_version(pSpecies_id)
    ithis%Sce0         = Sce0(pSpecies_id)
    ithis%rhoS         = rhoS(pSpecies_id)
    ithis%LS           = LS(pSpecies_id)
    ithis%TS           = TS(pSpecies_id)
    !--> for source of density
    ithis%Sce_dens     = Sce_dens(pSpecies_id)
    ithis%rhoS_dens    = rhoS_dens(pSpecies_id)
    ithis%LS_dens      = LS_dens(pSpecies_id)
    ithis%TS_dens      = TS_dens(pSpecies_id)
    !--> for source of momentum
    ithis%Sce_mom      = Sce_mom(pSpecies_id)
    ithis%rhoS_mom     = rhoS_mom(pSpecies_id)
    ithis%LS_mom       = LS_mom(pSpecies_id)
    ithis%TS_mom       = TS_mom(pSpecies_id)
    !--> for vorticity source
    ithis%Sce_vor      = Sce_vor(pSpecies_id)
    ithis%rhoS_vor     = rhoS_vor(pSpecies_id)
    ithis%LS_vor       = LS_vor(pSpecies_id)
    ithis%TS_vor       = TS_vor(pSpecies_id)
    !--> for source of fast particles
    ithis%Sce_fp       = Sce_fp(pSpecies_id)
    ithis%Sce_fpplus   = Sce_fpplus(pSpecies_id)
    ithis%rhoS_fp      = rhoS_fp(pSpecies_id)
    ithis%LS_fp        = LS_fp(pSpecies_id)
    ithis%TS_fp        = TS_fp(pSpecies_id)
    ithis%v0_fp        = v0_fp(pSpecies_id)
    ithis%alpha_trapp  = alpha_trapp(pSpecies_id)

    !*** memory allocation ****
    !--> For the radial shapes of sources and their derivatives 
    call FTR_ALLOC_1_ZERO(ithis%rshape,0,Nr_loc)
    call FTR_ALLOC_1_ZERO(ithis%dshapedr,0,Nr_loc)
    call FTR_ALLOC_1_ZERO(ithis%rshape_dens,0,Nr_loc)
    call FTR_ALLOC_1_ZERO(ithis%dshapedr_dens,0,Nr_loc)
    call FTR_ALLOC_1_ZERO(ithis%rshape_mom,0,Nr_loc)
    call FTR_ALLOC_1_ZERO(ithis%dshapedr_mom,0,Nr_loc)
    call FTR_ALLOC_1_ZERO(ithis%rshape_vor,0,Nr_loc)
    call FTR_ALLOC_1_ZERO(ithis%dshapedr_vor,0,Nr_loc)
    call FTR_ALLOC_1_ZERO(ithis%rshape_fp,0,Nr_loc)
    call FTR_ALLOC_1_ZERO(ithis%dshapedr_fp,0,Nr_loc)
    !--> For the 3D saving of the total source and 
    !-->  its derivative in r, theta and vpar 
    call FTR_ALLOC_3_ZERO(ithis%Sce_mu,istart,iend,jstart,jend,0,Nvpar_loc)
    call FTR_ALLOC_3_ZERO(ithis%dSdr_mu,istart,iend,jstart,jend,0,Nvpar_loc)
    call FTR_ALLOC_3_ZERO(ithis%dSdtheta_mu,istart,iend,jstart,jend,0,Nvpar_loc)
    call FTR_ALLOC_3_ZERO(ithis%dSdvpar_mu,istart,iend,jstart,jend,0,Nvpar_loc)
    !--> for the local computation of mu0 J.b
    call FTR_ALLOC_2_ZERO(mu0_Jpar,0,Nr_loc,0,Ntheta_loc)
  end subroutine new_init_sources


  !---------------------------------------------------- 
  ! destructor
  !---------------------------------------------------- 
  subroutine del_init_sources(ithis)
    type(init_sources), intent(inout) :: ithis

    call FTR_DEALLOC(ithis%rshape)
    call FTR_DEALLOC(ithis%dshapedr)
    call FTR_DEALLOC(ithis%rshape_dens)
    call FTR_DEALLOC(ithis%dshapedr_dens)
    call FTR_DEALLOC(ithis%rshape_mom)
    call FTR_DEALLOC(ithis%dshapedr_mom)
    call FTR_DEALLOC(ithis%rshape_vor)
    call FTR_DEALLOC(ithis%dshapedr_vor)   
    call FTR_DEALLOC(ithis%rshape_fp)
    call FTR_DEALLOC(ithis%dshapedr_fp)
    call FTR_DEALLOC(ithis%Sce_mu)
    call FTR_DEALLOC(ithis%dSdr_mu)
    call FTR_DEALLOC(ithis%dSdtheta_mu)
    call FTR_DEALLOC(ithis%dSdvpar_mu)
    call FTR_DEALLOC(mu0_Jpar)
  end subroutine del_init_sources


  !----------------------------------------------------
  ! Compute the radial shape of the source and its
  !  derivative 
  !----------------------------------------------------
  subroutine compute_source_radialshape(geom,SrhoS, &
    sLS,Srshape, Sdshape_dr)
    type(geometry)               , intent(in)    :: geom
    real(RKIND)                  , intent(in)    :: SrhoS
    real(RKIND)                  , intent(in)    :: sLS
    real(RKIND)   , dimension(0:), intent(inout) :: Srshape
    real(RKIND)   , dimension(0:), intent(inout) :: Sdshape_dr

#ifdef HAS_SRC    
    integer     :: ir
    real(RKIND) :: ri, rho, Lr_loc, rmin
    real(RKIND) :: int_rshape

    Lr_loc = geom%Lr
    rmin   = geom%rg(0)

    do ir = 0,geom%Nr
      ri  = geom%rg(ir)
      rho = (ri-rmin)/Lr_loc
      Srshape(ir) = -0.5_RKIND* &
        (tanh((rho-SrhoS-3._RKIND*sLS)/sLS) + &
        tanh((SrhoS-3._RKIND*sLS-rho)/sLS)) 
      Sdshape_dr(ir) = -0.5_RKIND/(Lr_loc*sLS)* &
       (1/cosh((rho-SrhoS-3._RKIND*sLS)/sLS)**2 - &
       1/cosh((SrhoS-3._RKIND*sLS-rho)/sLS)**2)
    end do
    !*** Computation of the radial integral of the "shape" ***
    int_rshape     = 0._RKIND
    do ir = 0,geom%Nr
      int_rshape = int_rshape + geom%rg(ir)*Srshape(ir)
    end do
    int_rshape = (int_rshape - &
      0.5_RKIND*(geom%rg(0)*Srshape(0)+ &
      geom%rg(geom%Nr)*Srshape(geom%Nr)))*geom%dr
    !*** Normalisation of the radial shapes such that ***
    !***   int{r shape dr}=1                          ***
    Srshape    = Srshape / int_rshape
    Sdshape_dr = Sdshape_dr / int_rshape
#endif 
    ! HAS_SRC
  end subroutine compute_source_radialshape


  !-----------------------------------------------------------
  ! initialization of the radial shape of the heat source 
  !  and computation of the contribution of the heat source
  !  to the total source and to its first derivatives
  !  in r, theta and vpar 
  ! INPUT/OUTPUT : Sce_tot_mu, dSdr_tot_mu, dSdtheta_mu and
  !                dSdvpar_tot_mu
  !-----------------------------------------------------------
  subroutine initialize_heat_source(geom,init_prof, &
    init_magnet,init_curr,heat_version_spec, &
    Sce0_spec,rhoS_spec,LS_spec,TS_spec, &
    Sce_heat_rshape,Sce_heat_drshapedr, &
    Sce_tot_mu,dSdr_tot_mu,dSdtheta_tot_mu,dSdvpar_tot_mu)
    use globals     , only : &
      istart, iend, jstart, jend, mu_id
    use globals     , only : pSpecies_id, hat_As, hat_Zs
    use globals     , only : Rarray1_NrNtheta, &
      Rarray2_NrNtheta, Rarray3_NrNtheta, &
      Rarray1_Nr, Rarray1_Ntheta
    use utils_module, only : deriv1
    use init_magnetic_class
    use init_profile_class
    implicit none
    type(geometry)                 , intent(in)    :: geom
    type(init_profile)             , intent(in)    :: init_prof
    type(init_magnetic)            , intent(in)    :: init_magnet
    type(init_current)             , intent(in)    :: init_curr
    integer                        , intent(in)    :: heat_version_spec
    real(RKIND)                    , intent(in)    :: Sce0_spec
    real(RKIND)                    , intent(in)    :: rhoS_spec
    real(RKIND)                    , intent(in)    :: LS_spec
    real(RKIND)                    , intent(in)    :: TS_spec
    real(RKIND), dimension(0:)     , intent(inout) :: Sce_heat_rshape
    real(RKIND), dimension(0:)     , intent(inout) :: Sce_heat_drshapedr
    !---> arrays of size (istart:iend,jstart:jend,0:Nvpar)
    real(RKIND), &
      dimension(istart:,jstart:,0:), intent(inout) :: Sce_tot_mu
    real(RKIND), &
      dimension(istart:,jstart:,0:), intent(inout) :: dSdr_tot_mu
    real(RKIND), &
      dimension(istart:,jstart:,0:), intent(inout) :: dSdtheta_tot_mu
    real(RKIND), &
      dimension(istart:,jstart:,0:), intent(inout) :: dSdvpar_tot_mu

#ifdef HAS_SRC    
    integer     :: ir, itheta, ivpar, imu
    real(RKIND) :: sqrthatAS_on_hatZS
    real(RKIND) :: Bij, dBdr_ij, dBdtheta_ij
    real(RKIND) :: lambda, my_mu_val, mu_val, my_vpar, vpar, energy
    real(RKIND) :: Jparij, Jparij_p, heat_coef
    real(RKIND) :: Sjacobian_velocity, int_dvpardmu_tmp
    real(RKIND) :: Sce_heat_Norm_tmp 
    real(RKIND) :: inv_TS, sqrt_2TS
    real(RKIND) :: energy_onTS, exp_mEonTS

    real(RKIND), dimension(:,:), pointer :: Sce_heat_rth
    real(RKIND), dimension(:,:), pointer :: dSdvpar_rth
    real(RKIND), dimension(:,:), pointer :: Sce_heat_Norm
    real(RKIND), dimension(:)  , pointer :: dSdr_heat 
    real(RKIND), dimension(:)  , pointer :: dSdtheta_heat 

    Sce_heat_rth  => Rarray1_NrNtheta
    dSdvpar_rth   => Rarray2_NrNtheta
    Sce_heat_Norm => Rarray3_NrNtheta
    dSdr_heat     => Rarray1_Nr
    dSdtheta_heat => Rarray1_Ntheta

    !*** variable initialization ***
    Sce_heat_Norm = 0._RKIND
    Sce_heat_rth  = 0._RKIND
    dSdvpar_rth   = 0._RKIND
    dSdr_heat     = 0._RKIND
    dSdtheta_heat = 0._RKIND

    lambda             = 0.5_RKIND+float(ipow)
    inv_TS             = 1._RKIND/TS_spec
    sqrt_2TS           = sqrt(2._RKIND*TS_spec)
    sqrthatAS_on_hatZS = sqrt(hat_As(pSpecies_id)) / &
      hat_Zs(pSpecies_id)

    !*** Computation of the radial shapes ***
    call compute_source_radialshape(geom, &
      rhoS_spec,LS_spec,Sce_heat_rshape, &
      Sce_heat_drshapedr)

    !*** Computation of the integral in mu,vpar  ***
    !*** of the source to have a good            ***
    !*** normalisation                           ***
    if (heat_version_spec .eq. 0) then
      Sce_heat_Norm(0:geom%Nr,0:geom%Ntheta) = &
        sqrt(2._RKIND*pi*TS_spec) * &
        pi*TS_spec**(1+ipow) * &
        (1._RKIND+2._RKIND*float(ipow))
    else
      do itheta = 0,geom%Ntheta
        do ir = 0,geom%Nr
          Bij              = init_magnet%B_norm(ir,itheta)
          Jparij           = mu0_Jpar(ir,itheta)
          Jparij_p         = Jparij * sqrthatAS_on_hatZS * &
            sqrt(2._RKIND*TS_spec)/(Bij*Bij)
          heat_coef        = Jparij_p/(2._RKIND-Jparij_p*Jparij_p)
          int_dvpardmu_tmp = 0._RKIND
          do imu = 0,geom%Nmu
            my_mu_val = geom%mug(imu)
            mu_val    = my_mu_val*Bij*inv_TS
            do ivpar = 0,geom%Nvpar
              my_vpar     = geom%vparg(ivpar)
              vpar        = my_vpar/sqrt_2TS
              energy      = 0.5_RKIND*my_vpar*my_vpar + my_mu_val*Bij
              energy_onTS = energy*inv_TS
              exp_mEonTS  = exp(-energy_onTS)
              call compute_noprecomp_jacobian_velocity(geom, &
                init_magnet,init_curr,ir,itheta,ivpar,Sjacobian_velocity)
              int_dvpardmu_tmp = int_dvpardmu_tmp + &
                energy*exp_mEonTS * &
                (vpar*vpar-0.5_RKIND-heat_coef*(2._RKIND-mu_val) * &
                (2._RKIND*vpar-Jparij_p))*SJacobian_velocity * &
                geom%coeff_intdvpar(ivpar)*geom%coeff_intdmu(imu)
            end do
          end do
          Sce_heat_Norm(ir,itheta) = int_dvpardmu_tmp
        end do
      end do
    end if

    !*** Computation of the source with the good normalisation ***
    my_mu_val = geom%mug(mu_id)
    if (heat_version_spec.eq.0) then
      do ivpar = 0,geom%Nvpar
        my_vpar = geom%vparg(ivpar)
        do itheta = jstart,jend
          do ir = istart,iend
            Bij               = init_magnet%B_norm(ir,itheta)
            dBdr_ij           = init_magnet%dBdr(ir,itheta)
            dBdtheta_ij       = init_magnet%dBdtheta(ir,itheta)
            mu_val            = my_mu_val*inv_TS
            energy            = 0.5_RKIND*my_vpar*my_vpar + my_mu_val*Bij
            energy_onTS       = energy*inv_TS
            exp_mEonTS        = exp(-energy_onTS)
            Sce_heat_norm_tmp = Sce_heat_Norm(ir,itheta)
            Sce_tot_mu(ir,itheta,ivpar) = &
              Sce_tot_mu(ir,itheta,ivpar) + &            
              Sce0_spec*exp_mEonTS*Sce_heat_rshape(ir) * &
              (energy_onTS-lambda)/Sce_heat_Norm_tmp
            dSdr_tot_mu(ir,itheta,ivpar) = &
              dSdr_tot_mu(ir,itheta,ivpar) + &
              Sce0_spec*exp_mEonTS*(Sce_heat_drshapedr(ir) * &
              (energy_onTS-lambda) + &
              Sce_heat_rshape(ir)*mu_val*dBdr_ij * &
              (lambda+1._RKIND-energy_onTS)) / &
              Sce_heat_Norm_tmp
            dSdtheta_tot_mu(ir,itheta,ivpar) = &
              dSdtheta_tot_mu(ir,itheta,ivpar) + &
              Sce0_spec*exp_mEonTS* &
              Sce_heat_rshape(ir)*mu_val*dBdtheta_ij* &
              (lambda+1._RKIND-energy_onTS)/Sce_heat_Norm_tmp
            mu_val = my_mu_val*Bij*inv_TS
            vpar   = my_vpar/sqrt_2TS
            dSdvpar_tot_mu(ir,itheta,ivpar) = &
              dSdvpar_tot_mu(ir,itheta,ivpar) + &
              Sce0_spec*exp_mEonTS*Sce_heat_rshape(ir) * &
              vpar*inv_TS*(lambda+1._RKIND-energy_onTS) / &
              Sce_heat_Norm_tmp
          end do
        end do
      end do
    else
      do ivpar = 0,geom%Nvpar
        my_vpar = geom%vparg(ivpar)
        do itheta = 0,geom%Ntheta
          do ir = 0,geom%Nr
            Bij         = init_magnet%B_norm(ir,itheta)
            energy      = 0.5_RKIND*my_vpar*my_vpar + my_mu_val*Bij
            energy_onTS = energy*inv_TS  
            exp_mEonTS  = exp(-energy_onTS)
            Jparij      = mu0_Jpar(ir,itheta)
            Jparij_p    = Jparij*sqrthatAS_on_hatZS * &
              sqrt_2TS/(Bij*Bij)
            !*** Heating Source ***
            heat_coef         = Jparij_p / &
              (2._RKIND-Jparij_p*Jparij_p)
            mu_val            = my_mu_val*Bij*inv_TS
            vpar              = my_vpar/sqrt_2TS
            Sce_heat_Norm_tmp = Sce_heat_Norm(ir,itheta)
            Sce_heat_rth(ir,itheta) = &
              Sce0_spec*exp_mEonTS*Sce_heat_rshape(ir) * &
              (vpar*vpar-0.5_RKIND-heat_coef*(2._RKIND-mu_val)* &
              (2._RKIND*vpar-Jparij_p))/Sce_heat_Norm_tmp
            dSdvpar_rth(ir,itheta) = &
              Sce0_spec*exp_mEonTS/sqrt_2TS * &
              Sce_heat_rshape(ir)*(2._RKIND*vpar - &
              2._RKIND*heat_coef*(2._RKIND-mu_val) - &
              2_RKIND*vpar * &
              (vpar*vpar-0.5_RKIND-heat_coef*(2._RKIND-mu_val) * &
              (2._RKIND*vpar-Jparij_p)) )/Sce_heat_Norm_tmp
          end do
        end do

        do itheta = jstart,jend
          call deriv1(Sce_heat_rth(0:geom%Nr,itheta), &
            dSdr_heat,geom%Nr,geom%dr,0)
          do ir = istart,iend
            !*** Heat source ***
            Sce_tot_mu(ir,itheta,ivpar) = &
              Sce_tot_mu(ir,itheta,ivpar) + Sce_heat_rth(ir,itheta)
            !*** Derivative in r ***
            dSdr_tot_mu(ir,itheta,ivpar) = &
              dSdr_tot_mu(ir,itheta,ivpar) + dSdr_heat(ir)
            !*** Derivative in vpar ***
            dSdvpar_tot_mu(ir,itheta,ivpar) = &
              dSdvpar_tot_mu(ir,itheta,ivpar) + &
              dSdvpar_rth(ir,itheta)
          end do
        end do

        !*** Derivative in theta ***
        do ir = istart,iend
          call deriv1(Sce_heat_rth(ir,0:geom%Ntheta-1), &
            dSdtheta_heat(0:geom%Ntheta-1),geom%Ntheta-1, &
            geom%dtheta,1)
          !--> periodicity in theta
          dSdtheta_heat(geom%Ntheta) = dSdtheta_heat(0)
          do itheta = jstart,jend
            dSdtheta_tot_mu(ir,itheta,ivpar) = &
              dSdtheta_tot_mu(ir,itheta,ivpar) + &
              dSdtheta_heat(itheta)
          end do
        end do
      end do
    end if
#endif 
    ! HAS_SRC
  end subroutine initialize_heat_source


  !-----------------------------------------------------------
  ! initialization of the radial shape of the density source 
  !  and computation of the contribution of the density 
  !  source to the total source and to its first derivatives
  !  in r, theta and vpar 
  ! INPUT/OUTPUT : Sce_tot_mu, dSdr_tot_mu, dSdtheta_mu and
  !                dSdvpar_tot_mu
  !-----------------------------------------------------------
  subroutine initialize_density_source(geom,init_prof, &
    init_magnet,init_curr,Sce_dens_spec,rhoS_dens_spec, &
    LS_dens_spec,TS_dens_spec,Sce_dens_rshape, &
    Sce_dens_drshapedr,Sce_tot_mu,dSdr_tot_mu, &
    dSdtheta_tot_mu,dSdvpar_tot_mu)
    use globals     , only : &
      istart, iend, jstart, jend, mu_id
    use globals     , only : pSpecies_id, hat_As, hat_Zs
    use globals     , only : Rarray1_NrNtheta, &
      Rarray2_NrNtheta, Rarray3_NrNtheta, &
      Rarray1_Nr, Rarray1_Ntheta
    use utils_module, only : deriv1
    use init_magnetic_class
    use init_profile_class
    implicit none
    type(geometry)                 , intent(in)    :: geom
    type(init_profile)             , intent(in)    :: init_prof
    type(init_magnetic)            , intent(in)    :: init_magnet
    type(init_current)             , intent(in)    :: init_curr
    real(RKIND)                    , intent(in)    :: Sce_dens_spec
    real(RKIND)                    , intent(in)    :: rhoS_dens_spec
    real(RKIND)                    , intent(in)    :: LS_dens_spec
    real(RKIND)                    , intent(in)    :: TS_dens_spec
    real(RKIND), dimension(0:)     , intent(inout) :: Sce_dens_rshape
    real(RKIND), dimension(0:)     , intent(inout) :: Sce_dens_drshapedr
    !---> arrays of size (istart:iend,jstart:jend,0:Nvpar)
    real(RKIND), &
      dimension(istart:,jstart:,0:), intent(inout) :: Sce_tot_mu
    real(RKIND), &
      dimension(istart:,jstart:,0:), intent(inout) :: dSdr_tot_mu
    real(RKIND), &
      dimension(istart:,jstart:,0:), intent(inout) :: dSdtheta_tot_mu
    real(RKIND), &
      dimension(istart:,jstart:,0:), intent(inout) :: dSdvpar_tot_mu

#ifdef HAS_SRC    
    integer     :: ir, itheta, ivpar, imu
    real(RKIND) :: sqrthatAS_on_hatZS
    real(RKIND) :: Bij
    real(RKIND) :: my_mu_val, mu_val, my_vpar, vpar, energy

    real(RKIND) :: Sjacobian_velocity, int_dvpardmu_tmp
    real(RKIND) :: Sce_dens_Norm_tmp 
    real(RKIND) :: inv_TS_dens, sqrt_2TS_dens
    real(RKIND) :: energy_onTS_dens, exp_mEonT_dens

    real(RKIND), dimension(:,:), pointer :: Sce_dens_Norm
    real(RKIND), dimension(:,:), pointer :: Sce_dens_rth
    real(RKIND), dimension(:,:), pointer :: dSdvpar_rth
    real(RKIND), dimension(:)  , pointer :: dSdr_dens 
    real(RKIND), dimension(:)  , pointer :: dSdtheta_dens 

    Sce_dens_Norm => Rarray1_NrNtheta 
    Sce_dens_rth  => Rarray2_NrNtheta
    dSdvpar_rth   => Rarray3_NrNtheta
    dSdr_dens     => Rarray1_Nr
    dSdtheta_dens => Rarray1_Ntheta

    !*** variable initialization ***
    Sce_dens_Norm = 0._RKIND
    Sce_dens_rth  = 0._RKIND
    dSdvpar_rth   = 0._RKIND
    dSdr_dens     = 0._RKIND
    dSdtheta_dens = 0._RKIND

    inv_TS_dens        = 1._RKIND/TS_dens_spec
    sqrt_2TS_dens      = sqrt(2._RKIND*TS_dens_spec)
    sqrthatAS_on_hatZS = sqrt(hat_As(pSpecies_id)) / &
      hat_Zs(pSpecies_id)

    !*** Computation of the radial shapes ***
    call compute_source_radialshape(geom, &
      rhoS_dens_spec,LS_dens_spec, &
      Sce_dens_rshape,Sce_dens_drshapedr)

    !*** Computation of the integral in mu,vpar  ***
    !*** of the source to have a good            ***
    !*** normalisation                           ***
    do itheta = 0,geom%Ntheta
      do ir = 0,geom%Nr
        Bij              = init_magnet%B_norm(ir,itheta)
        int_dvpardmu_tmp = 0._RKIND
        do imu = 0,geom%Nmu
          my_mu_val = geom%mug(imu)
          mu_val    = my_mu_val*Bij*inv_TS_dens
          do ivpar = 0,geom%Nvpar
            my_vpar = geom%vparg(ivpar)
            vpar    = my_vpar/sqrt_2TS_dens
            energy  = 0.5_RKIND*my_vpar*my_vpar + &
              my_mu_val*Bij
            energy_onTS_dens = energy*inv_TS_dens
            exp_mEonT_dens   = exp(-energy_onTS_dens)
            call compute_noprecomp_jacobian_velocity(geom, &
              init_magnet,init_curr, &
              ir,itheta,ivpar,Sjacobian_velocity)
            int_dvpardmu_tmp = int_dvpardmu_tmp + &
              exp_mEonT_dens*(2.5_RKIND-mu_val-vpar*vpar) * &
              SJacobian_velocity*geom%coeff_intdvpar(ivpar) * &
              geom%coeff_intdmu(imu)
          end do
        end do
        Sce_dens_Norm(ir,itheta) = int_dvpardmu_tmp
      end do
    end do

    !*** Computation of the source with the good normalisation ***
    my_mu_val = geom%mug(mu_id)
    do ivpar = 0,geom%Nvpar
      my_vpar = geom%vparg(ivpar)
      do itheta = 0,geom%Ntheta
        do ir = 0,geom%Nr
          Bij              = init_magnet%B_norm(ir,itheta)
          energy           = 0.5_RKIND*my_vpar*my_vpar + &
            my_mu_val*Bij
          energy_onTS_dens  = energy*inv_TS_dens
          exp_mEonT_dens    = exp(-energy_onTS_dens)
          mu_val            = my_mu_val*Bij*inv_TS_dens
          vpar              = my_vpar/sqrt_2TS_dens
          Sce_dens_Norm_tmp = Sce_dens_Norm(ir,itheta)
          Sce_dens_rth(ir,itheta) = &
            Sce_dens_spec*exp_mEonT_dens*Sce_dens_rshape(ir) * &
            (2.5_RKIND-mu_val-vpar*vpar) / &
            Sce_dens_Norm_tmp
          dSdvpar_rth(ir,itheta) = -2._RKIND*vpar * &
            Sce_dens_spec/sqrt_2TS_dens * &
            exp_mEonT_dens*Sce_dens_rshape(ir) * &
            ( 3.5_RKIND-mu_val-vpar*vpar  )/Sce_dens_Norm_tmp
        end do ! Loop over all r
      end do ! Loop over all theta

      do itheta = jstart,jend
        call deriv1(Sce_dens_rth(0:geom%Nr,itheta), &
          dSdr_dens,geom%Nr,geom%dr,0)
        do ir = istart,iend
          !*** Density Source ***
          Sce_tot_mu(ir,itheta,ivpar) = &
            Sce_tot_mu(ir,itheta,ivpar) + Sce_dens_rth(ir,itheta)
          !*** Derivative in r ***
          dSdr_tot_mu(ir,itheta,ivpar) = &
            dSdr_tot_mu(ir,itheta,ivpar) + dSdr_dens(ir)
          !*** Derivative in vpar ***
          dSdvpar_tot_mu(ir,itheta,ivpar) = &
            dSdvpar_tot_mu(ir,itheta,ivpar) + dSdvpar_rth(ir,itheta)
        end do
      end do

      !*** Derivative in theta ***
      do ir = istart,iend
        call deriv1(Sce_dens_rth(ir,0:geom%Ntheta-1), &
          dSdtheta_dens(0:geom%Ntheta-1),geom%Ntheta-1, &
          geom%dtheta,1)
        !--> periodicity in theta
        dSdtheta_dens(geom%Ntheta)  = dSdtheta_dens(0)
        do itheta = jstart,jend
          dSdtheta_tot_mu(ir,itheta,ivpar) = &
            dSdtheta_tot_mu(ir,itheta,ivpar) + dSdtheta_dens(itheta)
        end do
      end do
    end do ! Loop over all vpar
#endif 
    ! HAS_SRC
  end subroutine initialize_density_source


  !-----------------------------------------------------------
  ! initialization of the radial shape of the moment source 
  !  and computation of the contribution of the moment source
  !  to the total source and to its first derivatives
  !  in r, theta and vpar 
  ! INPUT/OUTPUT : Sce_tot_mu, dSdr_tot_mu, dSdtheta_mu and
  !                dSdvpar_tot_mu
  !-----------------------------------------------------------
  subroutine initialize_moment_source(geom,init_prof, &
    init_magnet,init_curr,Sce_mom_spec,rhoS_mom_spec, &
    LS_mom_spec,TS_mom_spec,Sce_mom_rshape, &
    Sce_mom_drshapedr,Sce_tot_mu,dSdr_tot_mu, &
    dSdtheta_tot_mu,dSdvpar_tot_mu)
    use globals     , only : &
      istart, iend, jstart, jend, mu_id
    use globals     , only : pSpecies_id, hat_As, hat_Zs
    use globals     , only : Rarray1_NrNtheta, &
      Rarray2_NrNtheta, Rarray3_NrNtheta, &
      Rarray1_Nr, Rarray1_Ntheta
    use utils_module, only : deriv1
    use init_magnetic_class
    use init_profile_class
    implicit none
    type(geometry)                 , intent(in)    :: geom
    type(init_profile)             , intent(in)    :: init_prof
    type(init_magnetic)            , intent(in)    :: init_magnet
    type(init_current)             , intent(in)    :: init_curr
    real(RKIND)                    , intent(in)    :: Sce_mom_spec
    real(RKIND)                    , intent(in)    :: rhoS_mom_spec
    real(RKIND)                    , intent(in)    :: LS_mom_spec
    real(RKIND)                    , intent(in)    :: TS_mom_spec
    real(RKIND), dimension(0:)     , intent(inout) :: Sce_mom_rshape
    real(RKIND), dimension(0:)     , intent(inout) :: Sce_mom_drshapedr
    !---> arrays of size (istart:iend,jstart:jend,0:Nvpar)
    real(RKIND), &
      dimension(istart:,jstart:,0:), intent(inout) :: Sce_tot_mu
    real(RKIND), &
      dimension(istart:,jstart:,0:), intent(inout) :: dSdr_tot_mu
    real(RKIND), &
      dimension(istart:,jstart:,0:), intent(inout) :: dSdtheta_tot_mu
    real(RKIND), &
      dimension(istart:,jstart:,0:), intent(inout) :: dSdvpar_tot_mu

#ifdef HAS_SRC    
    integer     :: ir, itheta, ivpar, imu
    real(RKIND) :: sqrthatAS_on_hatZS
    real(RKIND) :: Bij
    real(RKIND) :: my_mu_val, mu_val, my_vpar, vpar, energy
    real(RKIND) :: Jparij, Jparij_mom
    real(RKIND) :: Sjacobian_velocity, int_dvpardmu_tmp
    real(RKIND) :: Sce_mom_Norm_tmp 
    real(RKIND) :: inv_TS_mom, sqrt_2TS_mom
    real(RKIND) :: energy_onTS_mom, exp_mEonT_mom

    real(RKIND), dimension(:,:), pointer :: Sce_mom_Norm
    real(RKIND), dimension(:,:), pointer :: Sce_mom_rth
    real(RKIND), dimension(:,:), pointer :: dSdvpar_rth
    real(RKIND), dimension(:)  , pointer :: dSdr_mom 
    real(RKIND), dimension(:)  , pointer :: dSdtheta_mom 

    Sce_mom_Norm => Rarray1_NrNtheta 
    Sce_mom_rth  => Rarray2_NrNtheta
    dSdvpar_rth  => Rarray3_NrNtheta
    dSdr_mom     => Rarray1_Nr
    dSdtheta_mom => Rarray1_Ntheta

    !*** variable initialization ***
    Sce_mom_Norm = 0._RKIND
    Sce_mom_rth  = 0._RKIND
    dSdvpar_rth  = 0._RKIND
    dSdr_mom     = 0._RKIND
    dSdtheta_mom = 0._RKIND

    inv_TS_mom         = 1._RKIND/TS_mom_spec
    sqrt_2TS_mom       = sqrt(2._RKIND*TS_mom_spec)
    sqrthatAS_on_hatZS = sqrt(hat_As(pSpecies_id)) / &
      hat_Zs(pSpecies_id)

    !*** Computation of the radial shapes ***
    call compute_source_radialshape(geom, &
      rhoS_mom_spec,LS_mom_spec, &
      Sce_mom_rshape,Sce_mom_drshapedr)

    !*** Computation of the integral in mu,vpar  ***
    !*** of the source to have a good            ***
    !*** normalisation                           ***
    do itheta = 0,geom%Ntheta
      do ir = 0,geom%Nr
        Bij              = init_magnet%B_norm(ir,itheta)
        Jparij           = mu0_Jpar(ir,itheta)
        Jparij_mom       = Jparij*sqrt_2TS_mom/(Bij*Bij)
        int_dvpardmu_tmp = 0._RKIND
        do imu = 0,geom%Nmu
          my_mu_val = geom%mug(imu)
          mu_val    = my_mu_val*Bij*inv_TS_mom
          do ivpar = 0,geom%Nvpar
            my_vpar         = geom%vparg(ivpar)
            vpar            = my_vpar/sqrt_2TS_mom
            energy          = 0.5_RKIND*my_vpar*my_vpar + &
              my_mu_val*Bij
            energy_onTS_mom = energy*inv_TS_mom
            exp_mEonT_mom   = exp(-energy_onTS_mom)
            call compute_noprecomp_jacobian_velocity(geom, &
              init_magnet,init_curr, &
              ir,itheta,ivpar,Sjacobian_velocity)
            int_dvpardmu_tmp = int_dvpardmu_tmp + &
              my_vpar*exp_mEonT_mom * &
              (2._RKIND*vpar*(2._RKIND-mu_val)+Jparij_mom*(&
              mu_val-1._RKIND-2._RKIND*vpar*vpar)) * &
              SJacobian_velocity*geom%coeff_intdvpar(ivpar) * &
              geom%coeff_intdmu(imu)
          end do
        end do
        Sce_mom_Norm(ir,itheta) = int_dvpardmu_tmp
      end do
    end do

    !*** Computation of the source with the good normalisation ***
    my_mu_val = geom%mug(mu_id)
    do ivpar = 0,geom%Nvpar
      my_vpar = geom%vparg(ivpar)
      do itheta = 0,geom%Ntheta
        do ir = 0,geom%Nr
          Bij              = init_magnet%B_norm(ir,itheta)
          energy           = 0.5_RKIND*my_vpar*my_vpar + &
            my_mu_val*Bij
          energy_onTS_mom  = energy*inv_TS_mom
          exp_mEonT_mom    = exp(-energy_onTS_mom)
          Jparij           = mu0_Jpar(ir,itheta)
          mu_val           = my_mu_val*Bij*inv_TS_mom
          vpar             = my_vpar/sqrt_2TS_mom
          Jparij_mom       = Jparij*sqrt_2TS_mom/(Bij*Bij)
          Sce_mom_Norm_tmp = Sce_mom_Norm(ir,itheta)
          Sce_mom_rth(ir,itheta) = &
            Sce_mom_spec*exp_mEonT_mom*Sce_mom_rshape(ir) * &
            (2._RKIND*vpar*(2._RKIND-mu_val)+Jparij_mom * &
            (mu_val-1._RKIND-2._RKIND*vpar*vpar)) / &
            Sce_mom_Norm_tmp
          dSdvpar_rth(ir,itheta) = &
            Sce_mom_spec/sqrt_2TS_mom * &
            exp_mEonT_mom*Sce_mom_rshape(ir) * &
            (4._RKIND-2._RKIND*mu_val - &
            4._RKIND*Jparij_mom*vpar-2._RKIND*vpar * &
            (2._RKIND*vpar*(2._RKIND-mu_val) + &
            Jparij_mom*(mu_val-1._RKIND-&
            2._RKIND*vpar*vpar)) )/Sce_mom_Norm_tmp
        end do ! Loop over all r
      end do ! Loop over all theta

      do itheta = jstart,jend
        call deriv1(Sce_mom_rth(0:geom%Nr,itheta), &
          dSdr_mom,geom%Nr,geom%dr,0)
        do ir = istart,iend
          !*** Moment Source ***
          Sce_tot_mu(ir,itheta,ivpar) = &
            Sce_tot_mu(ir,itheta,ivpar) + Sce_mom_rth(ir,itheta)
          !*** Derivative in r ***
          dSdr_tot_mu(ir,itheta,ivpar) = &
            dSdr_tot_mu(ir,itheta,ivpar) + dSdr_mom(ir)
          !*** Derivative in vpar ***
          dSdvpar_tot_mu(ir,itheta,ivpar) = &
            dSdvpar_tot_mu(ir,itheta,ivpar) + dSdvpar_rth(ir,itheta)
        end do
      end do

      !*** Derivative in theta ***
      do ir = istart,iend
        call deriv1(Sce_mom_rth(ir,0:geom%Ntheta-1), &
          dSdtheta_mom(0:geom%Ntheta-1),geom%Ntheta-1, &
          geom%dtheta,1)
        !--> periodicity in theta
        dSdtheta_mom(geom%Ntheta)  = dSdtheta_mom(0)
        do itheta = jstart,jend
          dSdtheta_tot_mu(ir,itheta,ivpar) = &
            dSdtheta_tot_mu(ir,itheta,ivpar) + dSdtheta_mom(itheta)
        end do
      end do
    end do ! Loop over all vpar
#endif 
    ! HAS_SRC
  end subroutine initialize_moment_source


  !---------------------------------------------------------------
  ! initialization of the radial shape of the vorticity source 
  !  and computation of the contribution of the vorticity source
  !  to the total source and to its first derivatives
  !  in r, theta and vpar 
  ! INPUT/OUTPUT : Sce_tot_mu, dSdr_tot_mu, dSdtheta_mu and
  !                dSdvpar_tot_mu
  !---------------------------------------------------------------
  subroutine initialize_vorticity_source(geom, &
    init_prof,init_magnet,init_curr, &
    Sce_vor_spec,rhoS_vor_spec,LS_vor_spec,TS_vor_spec, &
    Sce_vor_rshape,Sce_vor_drshapedr,Sce_tot_mu, &
    dSdr_tot_mu,dSdtheta_tot_mu,dSdvpar_tot_mu)
    use globals     , only : &
      istart, iend, jstart, jend, mu_id
    use globals     , only : pSpecies_id, hat_As, hat_Zs
    use globals     , only : Rarray1_NrNtheta, &
      Rarray2_NrNtheta, Rarray3_NrNtheta, &
      Rarray1_Nr, Rarray1_Ntheta
    use utils_module, only : deriv1
    use init_magnetic_class
    use init_profile_class
    implicit none
    type(geometry)                 , intent(in)    :: geom
    type(init_profile)             , intent(in)    :: init_prof
    type(init_magnetic)            , intent(in)    :: init_magnet
    type(init_current)             , intent(in)    :: init_curr
    real(RKIND)                    , intent(in)    :: Sce_vor_spec
    real(RKIND)                    , intent(in)    :: rhoS_vor_spec
    real(RKIND)                    , intent(in)    :: LS_vor_spec
    real(RKIND)                    , intent(in)    :: TS_vor_spec
    real(RKIND), dimension(0:)     , intent(inout) :: Sce_vor_rshape
    real(RKIND), dimension(0:)     , intent(inout) :: Sce_vor_drshapedr
    !---> arrays of size (istart:iend,jstart:jend,0:Nvpar)
    real(RKIND), &
      dimension(istart:,jstart:,0:), intent(inout) :: Sce_tot_mu
    real(RKIND), &
      dimension(istart:,jstart:,0:), intent(inout) :: dSdr_tot_mu
    real(RKIND), &
      dimension(istart:,jstart:,0:), intent(inout) :: dSdtheta_tot_mu
    real(RKIND), &
      dimension(istart:,jstart:,0:), intent(inout) :: dSdvpar_tot_mu

#ifdef HAS_SRC    
    integer     :: ir, itheta, ivpar, imu
    real(RKIND) :: sqrthatAS_on_hatZS
    real(RKIND) :: Bij
    real(RKIND) :: my_mu_val, mu_val, my_vpar, vpar, energy
    real(RKIND) :: Sjacobian_velocity, int_dvpardmu_tmp
    real(RKIND) :: Sce_vor_Norm_tmp 
    real(RKIND) :: inv_TS_vor, sqrt_2TS_vor
    real(RKIND) :: energy_onTS_vor, exp_mEonTs_vor

    real(RKIND), dimension(:,:), pointer :: Sce_vor_Norm
    real(RKIND), dimension(:,:), pointer :: Sce_vor_rth
    real(RKIND), dimension(:,:), pointer :: dSdvpar_rth
    real(RKIND), dimension(:)  , pointer :: dSdr_vor 
    real(RKIND), dimension(:)  , pointer :: dSdtheta_vor 

    Sce_vor_Norm => Rarray1_NrNtheta
    Sce_vor_rth  => Rarray2_NrNtheta
    dSdvpar_rth  => Rarray3_NrNtheta
    dSdr_vor     => Rarray1_Nr
    dSdtheta_vor => Rarray1_Ntheta

    !*** variable initialization ***
    Sce_vor_Norm = 0._RKIND
    Sce_vor_rth  = 0._RKIND
    dSdvpar_rth  = 0._RKIND
    dSdr_vor     = 0._RKIND
    dSdtheta_vor = 0._RKIND

    inv_TS_vor         = 1._RKIND/TS_vor_spec
    sqrt_2TS_vor       = sqrt(2._RKIND*TS_vor_spec)
    sqrthatAS_on_hatZS = sqrt(hat_As(pSpecies_id)) / &
      hat_Zs(pSpecies_id)

    !*** Computation of the radial shapes ***
    call compute_source_radialshape(geom, &
      rhoS_vor_spec,LS_vor_spec, &
      Sce_vor_rshape,Sce_vor_drshapedr)

    !*** Computation of the integral in mu,vpar  ***
    !*** of the source to have a good            ***
    !*** normalisation                           ***
    do itheta = 0,geom%Ntheta
      do ir = 0,geom%Nr
        Sce_vor_Norm(ir,itheta) = sqrt(2._RKIND) * &
          pi*sqrt(pi)*TS_vor_spec*TS_vor_spec * &
          sqrt(TS_vor_spec)
      end do
    end do

    !*** Computation of the source with the good normalisation ***
    my_mu_val = geom%mug(mu_id)
    do ivpar = 0,geom%Nvpar
      my_vpar = geom%vparg(ivpar)
      vpar    = my_vpar/sqrt_2TS_vor
      do itheta = 0,geom%Ntheta
        do ir = 0,geom%Nr
          Bij              = init_magnet%B_norm(ir,itheta)
          energy           = 0.5_RKIND*my_vpar*my_vpar + &
            my_mu_val*Bij
          energy_onTS_vor  = energy*inv_TS_vor
          exp_mEonTs_vor   = exp(-energy_onTS_vor)
          mu_val           = my_mu_val*Bij*inv_TS_vor
          Sce_vor_Norm_tmp = Sce_vor_Norm(ir,itheta)
          Sce_vor_rth(ir,itheta) = &
            Sce_vor_spec*exp_mEonTs_vor*Sce_vor_rshape(ir) * &
            (1._RKIND-mu_val+(2._RKIND*vpar*vpar-1._RKIND)) / &
            Sce_vor_Norm_tmp
          dSdvpar_rth(ir,itheta) = &
            Sce_vor_spec/sqrt_2TS_vor * &
            exp_mEonTs_vor*Sce_vor_rshape(ir) * &
            (4._RKIND*vpar - 2._RKIND*vpar * &
            (1._RKIND-mu_val+ &
            (2._RKIND*vpar*vpar-1._RKIND)) ) / &
            Sce_vor_Norm_tmp
        end do
      end do

      do itheta = jstart,jend
        call deriv1(Sce_vor_rth(0:geom%Nr,itheta), &
          dSdr_vor,geom%Nr,geom%dr,0)
        do ir = istart,iend
          !*** Vorticity Source ***
          Sce_tot_mu(ir,itheta,ivpar) = &
            Sce_tot_mu(ir,itheta,ivpar) + Sce_vor_rth(ir,itheta)
          !*** Derivative in r ***
          dSdr_tot_mu(ir,itheta,ivpar) = &
            dSdr_tot_mu(ir,itheta,ivpar) + dSdr_vor(ir)
          !*** Derivative in vpar ***
          dSdvpar_tot_mu(ir,itheta,ivpar) = &
            dSdvpar_tot_mu(ir,itheta,ivpar) + dSdvpar_rth(ir,itheta)
        end do
      end do

      !*** Derivative in theta ***
      do ir = istart,iend
        call deriv1(Sce_vor_rth(ir,0:geom%Ntheta-1), &
          dSdtheta_vor(0:geom%Ntheta-1),geom%Ntheta-1, &
          geom%dtheta,1)
        !--> periodicity in theta
        dSdtheta_vor(geom%Ntheta)  = dSdtheta_vor(0)
        do itheta = jstart,jend
          dSdtheta_tot_mu(ir,itheta,ivpar) = &
            dSdtheta_tot_mu(ir,itheta,ivpar) + dSdtheta_vor(itheta)
        end do
      end do
    end do ! Loop over all vpar
#endif 
    ! HAS_SRC
  end subroutine initialize_vorticity_source

  !-----------------------------------------------------------
  ! initialization of the radial shape of the fast
  !  particle source and computation of the 
  !  contribution of the heat source
  !  to the total source and to its first derivatives
  !  in r, theta and vpar 
  ! INPUT/OUTPUT : Sce_tot_mu, dSdr_tot_mu, dSdtheta_mu and
  !                dSdvpar_tot_mu
  !-----------------------------------------------------------
  subroutine initialize_fp_source(geom,init_prof, &
    init_magnet,init_curr,Sce_fp_spec,Sce_fpplus_spec, &
    rhoS_fp_spec,LS_fp_spec,TS_fp_spec,v0_fp_spec, &
    alpha_trapp_spec,Sce_fp_rshape,Sce_fp_drshapedr, &
    Sce_tot_mu,dSdr_tot_mu,dSdtheta_tot_mu,dSdvpar_tot_mu)
    use globals     , only : mu_id, pSpecies_id, &
      major_species, istart, iend, jstart, jend
    use globals     , only : pSpecies_id, hat_As, hat_Zs
    use globals     , only : Rarray1_NrNtheta, &
      Rarray2_NrNtheta, Rarray3_NrNtheta, &
      Rarray1_Nr, Rarray1_Ntheta
    use utils_module, only : deriv1
    use init_magnetic_class
    use init_profile_class
    implicit none
    type(geometry)                 , intent(in)    :: geom
    type(init_profile)             , intent(in)    :: init_prof
    type(init_magnetic)            , intent(in)    :: init_magnet
    type(init_current)             , intent(in)    :: init_curr
    real(RKIND)                    , intent(in)    :: Sce_fp_spec
    real(RKIND)                    , intent(in)    :: Sce_fpplus_spec
    real(RKIND)                    , intent(in)    :: rhoS_fp_spec
    real(RKIND)                    , intent(in)    :: LS_fp_spec
    real(RKIND)                    , intent(in)    :: TS_fp_spec
    real(RKIND)                    , intent(in)    :: v0_fp_spec
    real(RKIND)                    , intent(in)    :: alpha_trapp_spec
    real(RKIND), dimension(0:)     , intent(inout) :: Sce_fp_rshape
    real(RKIND), dimension(0:)     , intent(inout) :: Sce_fp_drshapedr
    !---> arrays of size (istart:iend,jstart:jend,0:Nvpar)
    real(RKIND), &
      dimension(istart:,jstart:,0:), intent(inout) :: Sce_tot_mu
    real(RKIND), &
      dimension(istart:,jstart:,0:), intent(inout) :: dSdr_tot_mu
    real(RKIND), &
      dimension(istart:,jstart:,0:), intent(inout) :: dSdtheta_tot_mu
    real(RKIND), &
      dimension(istart:,jstart:,0:), intent(inout) :: dSdvpar_tot_mu

#ifdef HAS_SRC    
    integer           :: ir, itheta, ivpar, imu
    integer           :: irSfp
    real(RKIND)       :: sqrthatAS_on_hatZS
    real(RKIND)       :: lambda, Bij
    real(RKIND)       :: my_mu_val, mu_val, my_vpar, vpar, energy_fp
    real(RKIND)       :: Jparij, Jparij_fp, heat_coef_fp
    real(RKIND)       :: vpar0, v1, v2, Ti
    real(RKIND)       :: inv_dvparH, half_inv_dvparH
    real(RKIND)       :: Sjacobian_velocity, int_dvpardmu_tmp
    real(RKIND)       :: Sce_fp_Norm_tmp 
    real(RKIND)       :: inv_TS_fp, sqrt_2TS_fp
    real(RKIND)       :: energy_onTS_fp, exp_mEonTS_fp
    real(RKIND)       :: exp_vpar0_pow2, exp_2vparvpar0
    real(RKIND)       :: exp_m2vparvpar0
    real(RKIND)       :: vpar_mvpar0, vpar_pvpar0
    real(RKIND)       :: vpar_pv1_ondvparH, vpar_pv2_ondvparH
    real(RKIND)       :: vpar_mv1_ondvparH, vpar_mv2_ondvparH
    real(RKIND)       :: one_malpha_trapp
    character(LEN=50) :: subp

    real(RKIND), dimension(:,:), pointer :: Sce_fp_rth
    real(RKIND), dimension(:,:), pointer :: dSdvpar_fp_rth
    real(RKIND), dimension(:,:), pointer :: Sce_fp_Norm
    real(RKIND), dimension(:)  , pointer :: dSdr_fp 
    real(RKIND), dimension(:)  , pointer :: dSdtheta_fp 

    Sce_fp_Norm    => Rarray1_NrNtheta
    Sce_fp_rth     => Rarray2_NrNtheta
    dSdvpar_fp_rth => Rarray3_NrNtheta
    dSdr_fp        => Rarray1_Nr
    dSdtheta_fp    => Rarray1_Ntheta

    !*** Computation of Ti = Ts_fp/Ts(irSfp) ***
    subp = " initialize_fp_source"//char(0)
    call locate(geom%Lr*rhoS_fp_spec+geom%rg(0), &
      geom%rg,geom%Nr,geom%dr,subp,irSfp)
    Ti  = TS_fp_spec/init_prof%Ts0(pSpecies_id,irSfp)
    !*** variable initialization ***
    Sce_fp_Norm     = 0._RKIND
    Sce_fp_rth      = 0._RKIND
    dSdvpar_fp_rth  = 0._RKIND
    dSdr_fp         = 0._RKIND
    dSdtheta_fp     = 0._RKIND

    sqrthatAS_on_hatZS = sqrt(hat_As(pSpecies_id)) / &
      hat_Zs(pSpecies_id)
    inv_TS_fp          = 1._RKIND/TS_fp_spec
    sqrt_2TS_fp        = sqrt(2._RKIND*TS_fp_spec)
    vpar0              = v0_fp_spec/sqrt_2TS_fp
    v1                 = vpar0 - sqrt(0.5_RKIND)
    v2                 = vpar0 + sqrt(0.5_RKIND)
    inv_dvparH         = 1._RKIND/(6._RKIND*geom%dvpar/sqrt_2TS_fp)
    half_inv_dvparH    = 0.5_RKIND*inv_dvparH 
    exp_vpar0_pow2     = exp(-vpar0**2)
    one_malpha_trapp   = 1._RKIND-alpha_trapp_spec

    !*** Computation of the radial shapes ***
    call compute_source_radialshape(geom,rhoS_fp_spec, &
      LS_fp_spec,Sce_fp_rshape,Sce_fp_drshapedr)

    !*** Computation of the integral in mu,vpar  ***
    !*** of the source to have a good            ***
    !*** normalisation                           ***
    do itheta = 0,geom%Ntheta
      do ir = 0,geom%Nr
        Bij              = init_magnet%B_norm(ir,itheta)
        Jparij           = mu0_Jpar(ir,itheta)
        Jparij_fp        = Jparij*sqrt_2TS_fp/(Bij*Bij)
        heat_coef_fp     = Jparij_fp/ &
          (2._RKIND-Jparij_fp*Jparij_fp* &
          (1._RKIND+2._RKIND*vpar0**2))            
        int_dvpardmu_tmp = 0._RKIND
        do imu = 0,geom%Nmu
          my_mu_val = geom%mug(imu)
          mu_val    = my_mu_val*inv_TS_fp*Ti*Bij
          do ivpar = 0,geom%Nvpar
            my_vpar           = geom%vparg(ivpar)
            vpar              = my_vpar/sqrt_2TS_fp
            exp_2vparvpar0    = exp(2._RKIND*vpar*vpar0)
            exp_m2vparvpar0   = exp(-2._RKIND*vpar*vpar0)
            vpar_mvpar0       = vpar-vpar0
            vpar_pvpar0       = vpar+vpar0
            vpar_pv1_ondvparH = (vpar+v1)*inv_dvparH
            vpar_pv2_ondvparH = (vpar+v2)*inv_dvparH
            vpar_mv1_ondvparH = (vpar-v1)*inv_dvparH
            vpar_mv2_ondvparH = (vpar-v2)*inv_dvparH
            call compute_noprecomp_jacobian_velocity(geom, &
              init_magnet,init_curr, &
              ir,itheta,ivpar,Sjacobian_velocity)
            energy_fp        = 0.5_RKIND*my_vpar*my_vpar + &
              my_mu_val*Ti*Bij
            energy_onTS_fp   = energy_fp*inv_Ts_fp 
            exp_mEonTs_fp    = exp(-energy_onTS_fp)
            int_dvpardmu_tmp = int_dvpardmu_tmp + &
              TS_fp_spec*(vpar*vpar+mu_val/Ti)* &
              exp_mEonTs_fp*exp_vpar0_pow2*( &
              Sce_fpplus_spec*exp_2vparvpar0* &
              (vpar_mvpar0**2-0.5_RKIND-heat_coef_fp*(2._RKIND-mu_val)* &
              (2._RKIND*vpar_mvpar0-Jparij_fp))* &
              (0.5_RKIND*(1+tanh(vpar_mv1_ondvparH)) + &
              alpha_trapp_spec*0.5_RKIND*(1+tanh(-vpar_mv1_ondvparH)) + &
              one_malpha_trapp*0.5_RKIND * &
              (1+tanh(vpar_mv2_ondvparH))) + &
              (1._RKIND-Sce_fpplus_spec)*exp_m2vparvpar0 * &
              (vpar_pvpar0**2-0.5_RKIND-heat_coef_fp*(2._RKIND-mu_val) * &
              (2._RKIND*vpar_pvpar0-Jparij_fp))* &
              (0.5_RKIND*(1+tanh(-vpar_pv1_ondvparH)) + &
              alpha_trapp_spec*0.5_RKIND*(1+tanh(vpar_pv1_ondvparH)) + &
              one_malpha_trapp*0.5_RKIND * &
              (1+tanh(-vpar_pv2_ondvparH)))) * &
              SJacobian_velocity*geom%coeff_intdvpar(ivpar) * &
              geom%coeff_intdmu(imu)
          end do
        end do
        Sce_fp_Norm(ir,itheta) = int_dvpardmu_tmp
      end do
    end do

    !*** Computation of the fast particles source at v=vS_fp ***
    !***   with the good normalisation                       ***
    my_mu_val = geom%mug(mu_id)
    do ivpar = 0,geom%Nvpar
      my_vpar = geom%vparg(ivpar)
      do itheta = 0,geom%Ntheta
        do ir = 0,geom%Nr
          Bij               = init_magnet%B_norm(ir,itheta)
          Jparij            = mu0_Jpar(ir,itheta)
          Jparij_fp         = Jparij*sqrthatAS_on_hatZS * &
            sqrt_2TS_fp/(Bij*Bij)
          heat_coef_fp      = Jparij_fp/ &
            (2._RKIND-Jparij_fp*Jparij_fp* &
            (1._RKIND+2._RKIND*vpar0**2))
          energy_fp         = 0.5_RKIND*my_vpar*my_vpar + &
            my_mu_val*Ti*Bij
          energy_onTS_fp    = energy_fp*inv_Ts_fp 
          exp_mEonTs_fp     = exp(-energy_onTS_fp)
          mu_val            = my_mu_val*inv_TS_fp*Ti*Bij
          vpar              = my_vpar/sqrt_2TS_fp
          exp_2vparvpar0    = exp(2._RKIND*vpar*vpar0)
          exp_m2vparvpar0   = exp(-2._RKIND*vpar*vpar0)
          vpar_mvpar0       = vpar-vpar0
          vpar_pvpar0       = vpar+vpar0
          vpar_pv1_ondvparH = (vpar+v1)*inv_dvparH
          vpar_pv2_ondvparH = (vpar+v2)*inv_dvparH
          vpar_mv1_ondvparH = (vpar-v1)*inv_dvparH
          vpar_mv2_ondvparH = (vpar-v2)*inv_dvparH
          Sce_fp_Norm_tmp   = Sce_fp_Norm(ir,itheta)
          Sce_fp_rth(ir,itheta) = &
            Sce_fp_spec*Sce_fp_rshape(ir)*exp_mEonTs_fp*exp_vpar0_pow2 * &
            (Sce_fpplus_spec*exp_2vparvpar0 * &
            (vpar_mvpar0**2-0.5_RKIND-heat_coef_fp*(2._RKIND-mu_val) * &
            (2._RKIND*vpar_mvpar0-Jparij_fp)) * &
            (0.5_RKIND*(1+tanh(vpar_mv1_ondvparH)) + &
            alpha_trapp_spec*0.5_RKIND*(1+tanh(-vpar_mv1_ondvparH)) + &
            one_malpha_trapp*0.5_RKIND * &
            (1+tanh(vpar_mv2_ondvparH))) + &
            (1._RKIND-Sce_fpplus_spec)*exp_m2vparvpar0 * &
            (vpar_pvpar0**2-0.5_RKIND-heat_coef_fp*(2._RKIND-mu_val) * &
            (2._RKIND*vpar_pvpar0-Jparij_fp)) * &
            (0.5_RKIND*(1+tanh(-vpar_pv1_ondvparH)) + &
            alpha_trapp_spec*0.5_RKIND*(1+tanh(vpar_pv1_ondvparH)) + &
            one_malpha_trapp*0.5_RKIND * &
            (1+tanh(-vpar_pv2_ondvparH))))/Sce_fp_Norm_tmp
          dSdvpar_fp_rth(ir,itheta) = &
            Sce_fp_spec*Sce_fp_rshape(ir)* &
            exp_mEonTs_fp/sqrt_2TS_fp * &
            exp_vpar0_pow2* &
            ( & ! Opening main parethesis
            Sce_fpplus_spec* &
            ( &
            exp_2vparvpar0*(2._RKIND*vpar_mvpar0 - &
            2._RKIND*heat_coef_fp*(2._RKIND-mu_val)) + &
            2._RKIND*vpar0*exp_2vparvpar0 * &
            (vpar_mvpar0**2 - 0.5_RKIND-heat_coef_fp*(2._RKIND-mu_val) * &
            (2._RKIND*vpar_mvpar0-Jparij_fp)) &
            ) * &
            ( &
            0.5_RKIND*(1+tanh(vpar_mv1_ondvparH)) + &
            alpha_trapp_spec*0.5_RKIND*(1+tanh(-vpar_mv1_ondvparH)) + &
            one_malpha_trapp*0.5_RKIND*(1+tanh(vpar_mv2_ondvparH)) &
            ) + &              
            (1._RKIND-Sce_fpplus_spec)* &
            ( &
            exp_m2vparvpar0 * &
            (2._RKIND*vpar_pvpar0 - &
            2._RKIND*heat_coef_fp*(2._RKIND-mu_val)) - &
            2._RKIND*vpar0*exp_m2vparvpar0 * &
            (vpar_pvpar0**2 - 0.5_RKIND-heat_coef_fp*(2._RKIND-mu_val) * &
            (2._RKIND*vpar_pvpar0-Jparij_fp)) &
            ) * &
            (0.5_RKIND*(1+tanh(-vpar_pv1_ondvparH)) + &
            alpha_trapp_spec*0.5_RKIND*(1+tanh(vpar_pv1_ondvparH)) + &
            one_malpha_trapp*0.5_RKIND*(1+tanh(-vpar_pv2_ondvparH)) &
            ) &
            - 2._RKIND*vpar * &
            ( &
            Sce_fpplus_spec*exp_2vparvpar0 * &
            (vpar_mvpar0**2-0.5_RKIND-heat_coef_fp*(2._RKIND-mu_val) * &
            (2._RKIND*vpar_mvpar0-Jparij_fp)) * &
            ( &
            0.5_RKIND*(1+tanh(vpar_mv1_ondvparH)) + &
            alpha_trapp_spec*0.5_RKIND*(1+tanh(-vpar_mv1_ondvparH)) + &
            one_malpha_trapp*0.5_RKIND*(1+tanh(vpar_mv2_ondvparH)) &
            ) + &
            (1._RKIND-Sce_fpplus_spec)*exp_m2vparvpar0 * &
            (vpar_pvpar0**2-0.5_RKIND-heat_coef_fp*(2._RKIND-mu_val) * &
            (2._RKIND*vpar_pvpar0-Jparij_fp)) * &
            ( &
            0.5_RKIND*(1+tanh(-vpar_pv1_ondvparH)) + &
            alpha_trapp_spec*0.5_RKIND*(1+tanh(vpar_pv1_ondvparH)) + &
            one_malpha_trapp*0.5_RKIND*(1+tanh(-vpar_pv2_ondvparH)) &
            ) + &
            Sce_fpplus_spec*exp_2vparvpar0 * &
            (vpar_mvpar0**2-0.5_RKIND-heat_coef_fp*(2._RKIND-mu_val) * &
            (2._RKIND*vpar_mvpar0-Jparij_fp)) * &
            ( &
            half_inv_dvparH*(1-(tanh(vpar_mv1_ondvparH))**2) - &
            alpha_trapp_spec*half_inv_dvparH*(1-(tanh(-vpar_mv1_ondvparH))**2) + &
            one_malpha_trapp*half_inv_dvparH*(1-(tanh(vpar_mv2_ondvparH))**2) &
            ) + &
            (1._RKIND-Sce_fpplus_spec)*exp_m2vparvpar0 * &
            (vpar_pvpar0**2-0.5_RKIND-heat_coef_fp*(2._RKIND-mu_val) * &
            (2._RKIND*vpar_pvpar0-Jparij_fp)) * &
            ( &
            -half_inv_dvparH*(1-(tanh(-vpar_pv1_ondvparH))**2) + &
            alpha_trapp_spec*half_inv_dvparH*(1-(tanh(vpar_pv1_ondvparH))**2) - &
            one_malpha_trapp*half_inv_dvparH*(1-(tanh(-vpar_pv2_ondvparH))**2)) &
            ) &
            ) & ! Closing main parenthesis
            / Sce_fp_Norm_tmp
        end do ! Loop over all r
      end do ! Loop over all theta

      do itheta = jstart,jend
        call deriv1(Sce_fp_rth(0:geom%Nr,itheta), &
          dSdr_fp,geom%Nr,geom%dr,0)
        do ir = istart,iend
          !*** Fast particles source ***
          Sce_tot_mu(ir,itheta,ivpar) = &
            Sce_tot_mu(ir,itheta,ivpar) + Sce_fp_rth(ir,itheta)
          !*** Derivative in r ***
          dSdr_tot_mu(ir,itheta,ivpar) = &
            dSdr_tot_mu(ir,itheta,ivpar) + dSdr_fp(ir)
          !*** deriv in vpar ***
          dSdvpar_tot_mu(ir,itheta,ivpar) = &
            dSdvpar_tot_mu(ir,itheta,ivpar) + dSdvpar_fp_rth(ir,itheta)
        end do
      end do

      !*** Derivative in theta ***
      do ir = istart,iend
        call deriv1(Sce_fp_rth(ir,0:geom%Ntheta-1), &
          dSdtheta_fp(0:geom%Ntheta-1),geom%Ntheta-1, &
          geom%dtheta,1)
        !--> periodicity in theta
        dSdtheta_fp(geom%Ntheta) = dSdtheta_fp(0)
        do itheta = jstart,jend
          dSdtheta_tot_mu(ir,itheta,ivpar) = &
            dSdtheta_tot_mu(ir,itheta,ivpar) + dSdtheta_fp(itheta)
        end do
      end do
    end do ! Loop over all vpar
#endif 
    ! HAS_SRC
  end subroutine initialize_fp_source

  !----------------------------------------------------
  ! initialization of the source terms
  !----------------------------------------------------
  subroutine initialize_sources(ithis,init_prof, &
    init_magnet,init_curr,geom)
    use globals, only : mu_id, pSpecies_id, &
      istart, iend, jstart, jend, &
      heat_source, fp_source, &
      dens_source, mom_source, vor_source
    use init_magnetic_class
    use init_profile_class
    use utils_module, only : deriv1
!R3 #include "code_instrumentation/r3_info.h" !R3
    type(init_sources) , intent(inout) :: ithis
    type(geometry)     , intent(in)    :: geom
    type(init_profile) , intent(in)    :: init_prof
    type(init_magnetic), intent(in)    :: init_magnet
    type(init_current) , intent(in)    :: init_curr

#ifdef HAS_SRC    
    integer :: ir, itheta, ivpar

!R3    call r3_info_begin (r3_info_index_1, 'Init_sources') !R3
    call Write_dyn_balise_open('Init_sources') !R4

    !*** initialization of the heat source ***
    do ivpar = 0,geom%Nvpar
      do itheta = jstart,jend
        do ir = istart,iend
          ithis%Sce_mu(ir,itheta,ivpar)      = 0._RKIND
          ithis%dSdr_mu(ir,itheta,ivpar)     = 0._RKIND 
          ithis%dSdtheta_mu(ir,itheta,ivpar) = 0._RKIND 
          ithis%dSdvpar_mu(ir,itheta,ivpar)  = 0._RKIND 
        end do
      end do
    end do

    !*** computation of mu0 vec_J.vec_b ***
    do itheta = 0,geom%Ntheta
      do ir = 0,geom%Nr
        call compute_scalprod_mu0Jb(geom, &
          init_magnet,init_curr,ir,itheta,mu0_Jpar(ir,itheta))
      end do
    end do
    
    if (heat_source(pSpecies_id)) then
      call initialize_heat_source(geom,init_prof, &
      init_magnet,init_curr,ithis%heat_version, &
      ithis%Sce0,ithis%rhoS,ithis%LS,ithis%TS, &
      ithis%rshape,ithis%dshapedr,ithis%Sce_mu, &
      ithis%dSdr_mu,ithis%dSdtheta_mu,ithis%dSdvpar_mu)
    end if

    if (dens_source(pSpecies_id)) then
      call initialize_density_source(geom,init_prof, &
      init_magnet,init_curr,ithis%Sce_dens, &
      ithis%rhoS_dens,ithis%LS_dens,ithis%TS_dens, &
      ithis%rshape_dens,ithis%dshapedr_dens,ithis%Sce_mu, &
      ithis%dSdr_mu,ithis%dSdtheta_mu,ithis%dSdvpar_mu)
    end if

    if (mom_source(pSpecies_id)) then
      call initialize_moment_source(geom,init_prof, &
      init_magnet,init_curr,ithis%Sce_mom, &
      ithis%rhoS_mom,ithis%LS_mom,ithis%TS_mom, &
      ithis%rshape_mom,ithis%dshapedr_mom,ithis%Sce_mu, &
      ithis%dSdr_mu,ithis%dSdtheta_mu,ithis%dSdvpar_mu)
    end if

    if (vor_source(pSpecies_id)) then
      call initialize_vorticity_source(geom,init_prof, &
      init_magnet,init_curr,ithis%Sce_vor,ithis%rhoS_vor, &
      ithis%LS_vor,ithis%TS_vor,ithis%rshape_vor, &
      ithis%dshapedr_vor,ithis%Sce_mu, &
      ithis%dSdr_mu,ithis%dSdtheta_mu,ithis%dSdvpar_mu)
    end if

    if (fp_source(pSpecies_id)) then
      call initialize_fp_source(geom,init_prof, &
      init_magnet,init_curr,ithis%Sce_fp, &
      ithis%Sce_fpplus,ithis%rhoS_fp,ithis%LS_fp, &
      ithis%TS_fp,ithis%v0_fp,ithis%alpha_trapp, &
      ithis%rshape_fp,ithis%dshapedr_fp,ithis%Sce_mu, &
      ithis%dSdr_mu,ithis%dSdtheta_mu,ithis%dSdvpar_mu)
    end if
    call Write_dyn_balise_close('Init_sources') !R4
!R3    call r3_info_end (r3_info_index_1) !R3
#endif 
    ! HAS_SRC
  end subroutine initialize_sources


  !------------------------------------------------------
  ! Routine to compute effective sources from source%S_mu
  ! and write them in the file sources.h5
  !------------------------------------------------------
  subroutine diag_sources(ithis,geom,init_magnet,init_curr,J0)
    use globals, only : mu_id, dom_r, dom_theta, istart, iend, &
      jstart, jend, mpi_comm_species, mpi_comm_mu, plocal_id, &
      Nbproc_r, Nbproc_theta, mpi_comm_intermu, vor_source, &
      numfmt_rst, nb_restart_diag, R0, pSpecies_id, nbproc_r, &
      nbproc_theta, nbproc_loc
    use OMPutils_module, only : Comp1_1Nrp1_1Ntheta, &
      Romp1_1Nrp1_1Nthetap1
    use MPIutils_module
    use gyroaverage_class
#ifndef NO_HDF5
  use HDF5
  use HDF5_module
#endif
    use MpiMod
    use globals, only: Nbthread
#include "code_instrumentation/r3_info.h" 
    type(init_sources) , intent(in)    :: ithis
    type(geometry)     , intent(in)    :: geom 
    type(init_magnetic), intent(in)    :: init_magnet
    type(init_current) , intent(in)    :: init_curr
    type(J0operator)   , intent(inout) :: J0

#ifdef HAS_SRC    
    integer     :: ir, itheta, ivpar, imu
    integer     :: ivpar_pass_m, ivpar_pass_p
    integer     :: target_mu = 0
    integer     :: target_pid
    integer     :: nbelem, nbelem2, p_theta, p_r, ierr, tid
    real(RKIND) :: coeff_intdmu_tmp
    real(RKIND) :: vpar, Bij, mu_val, SJacobian_velocity
    REAL(rkind) :: pass_coef, R_tmp, vpar_pass, coeff_mult
    real(RKIND), dimension(:,:), pointer :: &
      Mass_Sce, Mass_Sce_trapp, Mass_Sce_pass, &
      Energy_Sce, Moment_Sce, Vorticity_Sce, buff, &
      E_Sce_par, E_Sce_perp, E_Sce_par_trapp, E_Sce_par_pass, &
      E_Sce_perp_trapp, E_Sce_perp_pass
    complex(CKIND), dimension(:,:), pointer :: Acomp
    real(RKIND),    dimension(:,:), pointer :: buff_pointer
    real(RKIND), &
      dimension(:,:), pointer :: Mass_Sce_glob, &
      Mass_Sce_trapp_glob, Mass_Sce_pass_glob, &
      Energy_Sce_glob, Moment_Sce_glob, Vorticity_Sce_glob, &
      Vorticity_Sce_temp, E_Sce_par_glob, E_Sce_perp_glob, &
      E_Sce_par_trapp_glob, E_Sce_par_pass_glob, &
      E_Sce_perp_trapp_glob, E_Sce_perp_pass_glob
    integer, dimension(:),   pointer :: mapvpar, loc_reqr
    integer, dimension(:,:), pointer :: loc_status
    integer :: nbreqr, i, locnvpar, itag_vpar, ilocv
    integer :: local_id, locir, locith, indir
    real(RKIND), dimension(:), pointer :: buffrecv
    real(RKIND), dimension(:), pointer :: buffsend
    integer, parameter :: NBFIELDS = 11
#ifndef NO_HDF5
    integer(HID_T)    :: file_id        
    character(LEN=30) :: sources_file    
    character(LEN=3)  :: num_species
#endif
    character(LEN=50) :: subp

!R3    call r3_info_begin (r3_info_index_1, 'CPU_diag_sources') !R3
    call Write_dyn_balise_open('CPU_diag_sources') !R4


    do tid = 1, Nbthread
      call FTR_ALLOC_2(Romp1_1Nrp1_1Nthetap1(tid)%val,1,Nr+1,1,Ntheta+1)
      call FTR_ALLOC_2_ZERO(Comp1_1Nrp1_1Ntheta(tid)%val,1,Nr+1,1,Ntheta)

      call FTR_ALLOC_2(J0%omp_diag(tid)%val,1,J0%N1,1,J0%N2)
      call FTR_ALLOC_2(J0%omp_ldiag(tid)%val,1,J0%N1-1,1,J0%N2)
      call FTR_ALLOC_2(J0%omp_udiag(tid)%val,1,J0%N1-1,1,J0%N2)
    enddo

    call FTR_ALLOC_2_ZERO(Mass_Sce,0,dom_r-1,0,dom_theta-1)
    call FTR_ALLOC_2_ZERO(Mass_Sce_trapp,0,dom_r-1,0,dom_theta-1)
    call FTR_ALLOC_2_ZERO(Mass_Sce_pass,0,dom_r-1,0,dom_theta-1)
    call FTR_ALLOC_2_ZERO(Energy_Sce,0,dom_r-1,0,dom_theta-1)
    call FTR_ALLOC_2_ZERO(Moment_Sce,0,dom_r-1,0,dom_theta-1)
    call FTR_ALLOC_2_ZERO(Vorticity_Sce,0,dom_r-1,0,dom_theta-1)
    call FTR_ALLOC_2_ZERO(buff,0,dom_r-1,0,dom_theta-1)
    call FTR_ALLOC_2_ZERO(E_Sce_par,0,dom_r-1,0,dom_theta-1)
    call FTR_ALLOC_2_ZERO(E_Sce_perp,0,dom_r-1,0,dom_theta-1)
    call FTR_ALLOC_2_ZERO(E_Sce_par_trapp,0,dom_r-1,0,dom_theta-1)
    call FTR_ALLOC_2_ZERO(E_Sce_par_pass,0,dom_r-1,0,dom_theta-1)
    call FTR_ALLOC_2_ZERO(E_Sce_perp_trapp,0,dom_r-1,0,dom_theta-1)
    call FTR_ALLOC_2_ZERO(E_Sce_perp_pass,0,dom_r-1,0,dom_theta-1)

    call FTR_ALLOC_2_ZERO(Mass_Sce_glob,0,geom%Nr,0,geom%Ntheta)
    call FTR_ALLOC_2_ZERO(Mass_Sce_trapp_glob,0,geom%Nr,0,geom%Ntheta)
    call FTR_ALLOC_2_ZERO(Mass_Sce_pass_glob,0,geom%Nr,0,geom%Ntheta)
    call FTR_ALLOC_2_ZERO(Energy_Sce_glob,0,geom%Nr,0,geom%Ntheta)
    call FTR_ALLOC_2_ZERO(Moment_Sce_glob,0,geom%Nr,0,geom%Ntheta)
    call FTR_ALLOC_2_ZERO(Vorticity_Sce_glob,0,geom%Nr,0,geom%Ntheta)
    call FTR_ALLOC_2_ZERO(Vorticity_Sce_temp,0,geom%Nr,0,geom%Ntheta)
    call FTR_ALLOC_2_ZERO(E_Sce_par_glob,0,geom%Nr,0,geom%Ntheta)
    call FTR_ALLOC_2_ZERO(E_Sce_perp_glob,0,geom%Nr,0,geom%Ntheta)
    call FTR_ALLOC_2_ZERO(E_Sce_par_trapp_glob,0,geom%Nr,0,geom%Ntheta)
    call FTR_ALLOC_2_ZERO(E_Sce_par_pass_glob,0,geom%Nr,0,geom%Ntheta)
    call FTR_ALLOC_2_ZERO(E_Sce_perp_trapp_glob,0,geom%Nr,0,geom%Ntheta)
    call FTR_ALLOC_2_ZERO(E_Sce_perp_pass_glob,0,geom%Nr,0,geom%Ntheta)

    Mass_Sce              = 0._RKIND
    Mass_Sce_trapp        = 0._RKIND
    Mass_Sce_pass         = 0._RKIND
    Energy_Sce            = 0._RKIND
    E_Sce_par             = 0._RKIND
    E_Sce_par_trapp       = 0._RKIND
    E_Sce_par_pass        = 0._RKIND
    E_Sce_perp            = 0._RKIND
    E_Sce_perp_trapp      = 0._RKIND
    E_Sce_perp_pass       = 0._RKIND
    Moment_Sce            = 0._RKIND
    Mass_Sce_trapp_glob   = 0._RKIND
    Mass_Sce_pass_glob    = 0._RKIND
    Vorticity_Sce         = 0._RKIND
    Mass_Sce_glob         = 0._RKIND
    Energy_Sce_glob       = 0._RKIND
    E_Sce_par_glob        = 0._RKIND
    E_Sce_par_trapp_glob  = 0._RKIND
    E_Sce_par_pass_glob   = 0._RKIND
    E_Sce_perp_glob       = 0._RKIND
    E_Sce_perp_trapp_glob = 0._RKIND
    E_Sce_perp_pass_glob  = 0._RKIND
    Moment_Sce_glob       = 0._RKIND
    Vorticity_Sce_glob    = 0._RKIND
    subp                  = " diag_sources "//char(0)

    target_pid = target_mu*Nbproc_loc
    tid = 1

    Acomp        => Comp1_1Nrp1_1Ntheta(tid)%val
    buff_pointer => Romp1_1Nrp1_1Nthetap1(tid)%val

    mu_val  = geom%mug(mu_id)
    nbelem  = dom_r*dom_theta
    nbelem2 = (geom%Nr+1)*(geom%Ntheta+1)

    locnvpar = ceiling((1.*(geom%Nvpar+1))/Nbproc_loc)
    call FTR_ALLOC_1_ZERO(buffrecv,0,MAX((1+geom%Nr)*(1+geom%Ntheta)*locNvpar,(1+geom%Nr)*(1+geom%Ntheta)*NBFIELDS)-1)
    call FTR_ALLOC_1_ZERO(buffsend,0,(1+geom%Nr)*(1+geom%Ntheta)*NBFIELDS-1)
    call FTR_ALLOC_1_ZERO(loc_reqr,0,Nbproc_loc*locnvpar-1)
    call FTR_ALLOC_2_ZERO(loc_status,1,MPI_STATUS_SIZE,0,Nbproc_loc*locnvpar-1)
    call FTR_ALLOC_1_ZERO(mapvpar,0,geom%Nvpar)

    do i = 0, nbproc_loc-1
      do ivpar = i*locnvpar, MIN((i+1)*locnvpar-1, geom%Nvpar)
        mapvpar(ivpar) = i
      end do
    end do

    nbreqr = 0
    ilocv  = 0
    do ivpar = 0,geom%Nvpar
      if (mapvpar(ivpar) .eq. plocal_id) then
        do local_id = 0, Nbproc_loc-1
          itag_vpar = ivpar + (geom%Nvpar+1)*local_id + 3000
          call MPI_IRECV( &
            buffrecv(local_id*nbelem+ilocv*nbelem2), &
            nbelem,MPI_REAL8,local_id,itag_vpar, &
            mpi_comm_mu,loc_reqr(nbreqr),ierr)
          nbreqr = nbreqr + 1
        end do
        ilocv = ilocv + 1
      end if
    end do

    do ivpar = 0,geom%Nvpar
      itag_vpar = ivpar + (geom%Nvpar+1)*plocal_id + 3000
      call MPI_SEND(& 
        ithis%Sce_mu(istart,jstart,ivpar),&
        nbelem,MPI_REAL8,mapvpar(ivpar),itag_vpar,&
        mpi_comm_mu,ierr)
    end do

    if (nbreqr .ne. 0) then
      call MPI_WAITALL(nbreqr, loc_reqr(0:nbreqr-1), &
              loc_status(:,0:nbreqr-1), ierr)
    end if

    ilocv  = 0
    do ivpar = 0,geom%Nvpar
      if (mapvpar(ivpar) .eq. plocal_id) then
        do p_theta = 0, Nbproc_theta-1
          do p_r = 0, Nbproc_r-1
            do itheta = 0, dom_theta-1
              do ir = 0, dom_r-1
                Vorticity_Sce_temp(p_r*dom_r+ir, &
                     p_theta*dom_theta+itheta) = &
                     buffrecv(ir+dom_r*itheta + &
                     (p_theta*Nbproc_r+p_r)*nbelem+ilocv*nbelem2)
             enddo
            enddo
          enddo
        end do
        do ir = 0,geom%Nr
          Vorticity_Sce_temp(ir,geom%Ntheta) = &
            Vorticity_Sce_temp(ir,0)
        end do

        !*** Gyroaverage for the vorticity source ***
        do ir = 0,geom%Nr
          do itheta = 0,geom%Ntheta
            buff_pointer(ir+1,itheta+1) = &
              Vorticity_Sce_temp(ir,itheta)
          end do
        end do
        call omp_compute_gyrofunction_2D(J0,geom,mu_id, &
          Acomp,buff_pointer,tid)
        
        coeff_mult = geom%coeff_intdmu(mu_id) * &
          geom%coeff_intdvpar(ivpar)
        do ir = 0,geom%Nr
          do itheta = 0,geom%Ntheta
            call compute_noprecomp_jacobian_velocity( &
              geom,init_magnet,init_curr, &
              ir,itheta,ivpar,Sjacobian_velocity)
            Vorticity_Sce_glob(ir,itheta) = &
              Vorticity_Sce_glob(ir,itheta) + &
              buff_pointer(ir+1,itheta+1) * &
              SJacobian_velocity * coeff_mult
          end do
        end do
        ilocv = ilocv + 1

      end if   ! if (mapvpar(ivpar) .eq. plocal_id) then
    end do   ! do ivpar = 0,geom%Nvpar

    call MY_Reduce(Vorticity_Sce_glob,Vorticity_Sce_temp,nbelem2, &
      MPI_REAL8,MPI_SUM,target_pid,mpi_comm_species,ierr)

    do ir = 0,geom%Nr
      do itheta = 0,geom%Ntheta
        Vorticity_Sce_glob(ir,itheta) = &
          Vorticity_Sce_temp(ir,itheta)
      end do
    end do

    do ivpar = 0,geom%Nvpar
      vpar = geom%vparg(ivpar)
      !*** Computing integral in vpar ***
      do ir = istart,iend
        do itheta = jstart,jend
          call compute_jacobian_velocity(geom,init_magnet, &
            init_curr,ir,&
            itheta,ivpar,Sjacobian_velocity)

          Mass_Sce(ir-istart,itheta-jstart) = &
            Mass_Sce(ir-istart,itheta-jstart) + &
            ithis%Sce_mu(ir,itheta,ivpar) * &
            geom%coeff_intdvpar(ivpar)*SJacobian_velocity

          Energy_Sce(ir-istart,itheta-jstart) = &
            Energy_Sce(ir-istart,itheta-jstart) + &
            (0.5_RKIND*vpar*vpar+mu_val * &
            init_magnet%B_norm(ir,itheta)) * &
            ithis%Sce_mu(ir,itheta,ivpar) * &
            geom%coeff_intdvpar(ivpar)*SJacobian_velocity

          E_Sce_par(ir-istart,itheta-jstart) = &
            E_Sce_par(ir-istart,itheta-jstart) + &
            0.5_RKIND*vpar*vpar * &
            ithis%Sce_mu(ir,itheta,ivpar) * &
            geom%coeff_intdvpar(ivpar)*SJacobian_velocity

          E_Sce_perp(ir-istart,itheta-jstart) = &
            E_Sce_perp(ir-istart,itheta-jstart) + &
            mu_val*init_magnet%B_norm(ir,itheta) * &
            ithis%Sce_mu(ir,itheta,ivpar) * &
            geom%coeff_intdvpar(ivpar)*SJacobian_velocity
          
          Moment_Sce(ir-istart,itheta-jstart) = &
            Moment_Sce(ir-istart,itheta-jstart) + &
            vpar*ithis%Sce_mu(ir,itheta,ivpar) * &
            geom%coeff_intdvpar(ivpar)*SJacobian_velocity

        end do   ! do ir = istart,iend
      end do   ! do itheta = jstart,jend
    end do   ! do ivpar = 0,geom%Nvpar

    do ir = istart,iend
      do itheta = jstart,jend
        Bij       = init_magnet%B_norm(ir,itheta)
        R_tmp     = R0+geom%rg(ir)*geom%cos_theta(itheta)
        pass_coef = SQRT(2._RKIND*geom%rg(ir)/R_tmp)
        vpar_pass = pass_coef*SQRT(2._RKIND*mu_val*Bij)
        !*** find the index ivpar_pass_m corresponding to ***
        !***  vpar = -                                    ***
        call locate(-vpar_pass,geom%vparg,geom%Nvpar, &
          geom%dvpar,subp,ivpar_pass_m)
        !*** find the index ivpar_pass_p corresponding to ***
        !***  vpar = -                                    ***
        call locate(vpar_pass,geom%vparg,geom%Nvpar, &
          geom%dvpar,subp,ivpar_pass_p)
        !*** Computing integral in vpar ***
        do ivpar = 0,ivpar_pass_m
          vpar = geom%vparg(ivpar)           
          call compute_jacobian_velocity(geom,init_magnet, &
            init_curr,ir,itheta,ivpar,Sjacobian_velocity)          
          E_Sce_par_pass(ir-istart,itheta-jstart) = &
            E_Sce_par_pass(ir-istart,itheta-jstart) + &
            0.5_RKIND*vpar*vpar * &
            ithis%Sce_mu(ir,itheta,ivpar) * &
            geom%coeff_intdvpar(ivpar)*SJacobian_velocity

          E_Sce_perp_pass(ir-istart,itheta-jstart) = &
            E_Sce_perp_pass(ir-istart,itheta-jstart) + &
            mu_val*Bij* &
            ithis%Sce_mu(ir,itheta,ivpar) * &
            geom%coeff_intdvpar(ivpar)*SJacobian_velocity

          Mass_Sce_pass(ir-istart,itheta-jstart) = &
            Mass_Sce_pass(ir-istart,itheta-jstart) + &
            ithis%Sce_mu(ir,itheta,ivpar) * &
            geom%coeff_intdvpar(ivpar)*SJacobian_velocity
        end do
        
        do ivpar = ivpar_pass_p,geom%Nvpar
          vpar = geom%vparg(ivpar)           
          call compute_jacobian_velocity(geom,init_magnet, &
            init_curr,ir,&
            itheta,ivpar,Sjacobian_velocity)          
          E_Sce_par_pass(ir-istart,itheta-jstart) = &
            E_Sce_par_pass(ir-istart,itheta-jstart) + &
            0.5_RKIND*vpar*vpar * &
            ithis%Sce_mu(ir,itheta,ivpar) * &
            geom%coeff_intdvpar(ivpar)*SJacobian_velocity

          E_Sce_perp_pass(ir-istart,itheta-jstart) = &
            E_Sce_perp_pass(ir-istart,itheta-jstart) + &
            mu_val*Bij* &
            ithis%Sce_mu(ir,itheta,ivpar) * &
            geom%coeff_intdvpar(ivpar)*SJacobian_velocity

          Mass_Sce_pass(ir-istart,itheta-jstart) = &
            Mass_Sce_pass(ir-istart,itheta-jstart) + &
            ithis%Sce_mu(ir,itheta,ivpar) * &
            geom%coeff_intdvpar(ivpar)*SJacobian_velocity

        end do
        
        do ivpar = ivpar_pass_m+1,ivpar_pass_p-1
          vpar = geom%vparg(ivpar)           
          call compute_jacobian_velocity(geom,init_magnet, &
            init_curr,ir,&
            itheta,ivpar,Sjacobian_velocity)          
          E_Sce_par_trapp(ir-istart,itheta-jstart) = &
            E_Sce_par_trapp(ir-istart,itheta-jstart) + &
            0.5_RKIND*vpar*vpar * &
            ithis%Sce_mu(ir,itheta,ivpar) * &
            geom%coeff_intdvpar(ivpar)*SJacobian_velocity

          E_Sce_perp_trapp(ir-istart,itheta-jstart) = &
            E_Sce_perp_trapp(ir-istart,itheta-jstart) + &
            mu_val*Bij* &
            ithis%Sce_mu(ir,itheta,ivpar) * &
            geom%coeff_intdvpar(ivpar)*SJacobian_velocity

          Mass_Sce_trapp(ir-istart,itheta-jstart) = &
            Mass_Sce_trapp(ir-istart,itheta-jstart) + &
            ithis%Sce_mu(ir,itheta,ivpar) * &
            geom%coeff_intdvpar(ivpar)*SJacobian_velocity

        end do
      end do   ! do itheta = jstart,jend
    end do   ! do ir = istart,iend

    !*** Integral in mu direction ***
    coeff_intdmu_tmp = geom%coeff_intdmu(mu_id)
    do ir = 0, dom_r-1
      do itheta = 0, dom_theta-1
        buffsend( 0+NBFIELDS*(ir+dom_r*itheta)) = &
          coeff_intdmu_tmp*Mass_Sce(ir,itheta) 
        buffsend( 1+NBFIELDS*(ir+dom_r*itheta)) = &
          coeff_intdmu_tmp*Mass_Sce_trapp(ir,itheta) 
        buffsend( 2+NBFIELDS*(ir+dom_r*itheta)) = &
          coeff_intdmu_tmp*Mass_Sce_pass(ir,itheta)
        buffsend( 3+NBFIELDS*(ir+dom_r*itheta)) = &
          coeff_intdmu_tmp*Energy_Sce(ir,itheta)
        buffsend( 4+NBFIELDS*(ir+dom_r*itheta)) = &
          coeff_intdmu_tmp*E_Sce_par(ir,itheta)
        buffsend( 5+NBFIELDS*(ir+dom_r*itheta)) = &
          coeff_intdmu_tmp*E_Sce_par_trapp(ir,itheta)
        buffsend( 6+NBFIELDS*(ir+dom_r*itheta)) = &
          coeff_intdmu_tmp*E_Sce_par_pass(ir,itheta)
        buffsend( 7+NBFIELDS*(ir+dom_r*itheta)) = &
          coeff_intdmu_tmp*E_Sce_perp(ir,itheta)
        buffsend( 8+NBFIELDS*(ir+dom_r*itheta)) = &
          coeff_intdmu_tmp*E_Sce_perp_trapp(ir,itheta)
        buffsend( 9+NBFIELDS*(ir+dom_r*itheta)) = &
          coeff_intdmu_tmp*E_Sce_perp_pass (ir,itheta)
        buffsend(10+NBFIELDS*(ir+dom_r*itheta)) = &
          coeff_intdmu_tmp*Moment_Sce(ir,itheta)
      end do
    end do
    
    nbelem  = dom_r*dom_theta*NBFIELDS
    nbelem2 = -1
    nbreqr = 0
    if (plocal_id .eq. 0) then
      do local_id = 0, Nbproc_loc-1
        itag_vpar = local_id + 12000
        call MPI_IRECV(buffrecv(local_id*nbelem), &
          nbelem,MPI_REAL8,local_id,itag_vpar, &
          mpi_comm_mu,loc_reqr(nbreqr),ierr)
        nbreqr = nbreqr + 1
      end do
    end if

    itag_vpar = plocal_id + 12000
    call MPI_SEND(buffsend(0), &
      nbelem,MPI_REAL8,0,itag_vpar, &
      mpi_comm_mu,ierr)

    if (nbreqr .ne. 0) then
      call MPI_WAITALL(nbreqr, loc_reqr(0:nbreqr-1), &
              loc_status(:,0:nbreqr-1), ierr)
    end if

    nbelem  = Nbproc_loc*dom_r*dom_theta*NBFIELDS
    nbelem2 = -1
    if (plocal_id .eq. 0) then
      call MY_Reduce(buffrecv(0),buffsend(0),nbelem, &
        MPI_REAL8,MPI_SUM,target_mu,mpi_comm_intermu,ierr)

      do p_theta = 0, Nbproc_theta-1
        do p_r = 0, Nbproc_r-1
          do itheta = 0, dom_theta-1
            do ir = 0, dom_r-1
              locir  = p_r*dom_r+ir
              locith = p_theta*dom_theta+itheta
              indir  = NBFIELDS*(ir+dom_r * &
                (itheta+dom_theta*(p_theta*Nbproc_r+p_r)))
              Mass_Sce_glob(locir,locith)         = &
                buffsend( 0+indir) 
              Mass_Sce_trapp_glob(locir,locith)   = &
                buffsend( 1+indir) 
              Mass_Sce_pass_glob(locir,locith)    = &
                buffsend( 2+indir) 
              Energy_Sce_glob(locir,locith)       = &
                buffsend( 3+indir) 
              E_Sce_par_glob(locir,locith)        = &
                buffsend( 4+indir) 
              E_Sce_par_trapp_glob(locir,locith)  = &
                buffsend( 5+indir) 
              E_Sce_par_pass_glob(locir,locith)   = &
                buffsend( 6+indir) 
              E_Sce_perp_glob(locir,locith)       = &
                buffsend( 7+indir) 
              E_Sce_perp_trapp_glob(locir,locith) = &
                buffsend( 8+indir) 
              E_Sce_perp_pass_glob(locir,locith)  = &
                buffsend( 9+indir) 
              Moment_Sce_glob(locir,locith)       = &
                buffsend(10+indir) 
            end do
          end do
        end do
      end do
      do ir = 0, geom%Nr
        Mass_Sce_glob(ir,geom%Ntheta)         = Mass_Sce_glob(ir,0)
        Mass_Sce_trapp_glob(ir,geom%Ntheta)   = &
          Mass_Sce_trapp_glob(ir,0)
        Mass_Sce_pass_glob(ir,geom%Ntheta)    = &
          Mass_Sce_pass_glob(ir,0)
        Energy_Sce_glob(ir,geom%Ntheta)       = &
          Energy_Sce_glob(ir,0)
        E_Sce_par_glob(ir,geom%Ntheta)        = &
          E_Sce_par_glob(ir,0)
        E_Sce_par_trapp_glob(ir,geom%Ntheta)  = &
          E_Sce_par_trapp_glob(ir,0)
        E_Sce_par_pass_glob(ir,geom%Ntheta)   = &
          E_Sce_par_pass_glob(ir,0)
        E_Sce_perp_glob(ir,geom%Ntheta)       = &
          E_Sce_perp_glob(ir,0)
        E_Sce_perp_trapp_glob(ir,geom%Ntheta) = &
          E_Sce_perp_trapp_glob(ir,0)
        E_Sce_perp_pass_glob(ir,geom%Ntheta)  = &
          E_Sce_perp_pass_glob(ir,0)
        Moment_Sce_glob(ir,geom%Ntheta)       = &
          Moment_Sce_glob(ir,0)
      end do
    end if  ! if (plocal_id .eq. 0) then

    !*** Creation of (r,theta) profiles of the sources ***
    if (mu_id .eq. target_mu) then

      !** HDF5 writing ***
#ifndef NO_HDF5
      write(num_species,'(A2,I1)') "sp", pSpecies_id+1
      if (pSpecies_id .eq. 0) then
        write(sources_file,'(A,'//numfmt_rst//',A)') &
          "sources", nb_restart_diag, ".h5"
      else
        write(sources_file,'((2A),'//numfmt_rst//',A)') &
          num_species, "/sources", nb_restart_diag, ".h5"
      end if

      if (plocal_id.eq.0) then
        Call HDF5_create(trim(sources_file),file_id)
        Call HDF5_save(file_id,Mass_Sce_glob(0:,0:), &
          geom%Nr+1,geom%Ntheta+1,'Mass_Sce'//char(0))
        Call HDF5_save(file_id, &
          Mass_Sce_trapp_glob(0:,0:),geom%Nr+1,geom%Ntheta+1, &
          'Mass_trapp_Sce'//char(0))
        Call HDF5_save(file_id, &
          Mass_Sce_pass_glob(0:,0:), &
          geom%Nr+1,geom%Ntheta+1,'Mass_pass_Sce'//char(0))
        Call HDF5_save(file_id,Energy_Sce_glob(0:,0:), &
          geom%Nr+1,geom%Ntheta+1,'Energy_Sce'//char(0))
        Call HDF5_save(file_id,E_Sce_par_glob(0:,0:), &
          geom%Nr+1,geom%Ntheta+1,'E_par_Sce'//char(0))
        Call HDF5_save(file_id, &
          E_Sce_par_trapp_glob(0:,0:), &
          geom%Nr+1,geom%Ntheta+1,'E_par_trapp_Sce'//char(0))
        Call HDF5_save(file_id, &
          E_Sce_par_pass_glob(0:,0:), &
          geom%Nr+1,geom%Ntheta+1,'E_par_pass_Sce'//char(0))
        Call HDF5_save(file_id, &
          E_Sce_perp_glob(0:,0:),geom%Nr+1,geom%Ntheta+1, &
          'E_perp_Sce'//char(0))
        Call HDF5_save(file_id, &
          E_Sce_perp_trapp_glob(0:,0:),geom%Nr+1,geom%Ntheta+1, &
          'E_perp_trapp_Sce'//char(0))
        Call HDF5_save(file_id, &
          E_Sce_perp_pass_glob(0:,0:),geom%Nr+1,geom%Ntheta+1, &
          'E_perp_pass_Sce'//char(0))
        Call HDF5_save(file_id,Moment_Sce_glob(0:,0:), &
          geom%Nr+1,geom%Ntheta+1,'Moment_Sce'//char(0))
        Call HDF5_save(file_id, &
          Vorticity_Sce_glob(0:,0:),geom%Nr+1,geom%Ntheta+1, &
          'Vorticity_Sce'//char(0))
        Call HDF5_close(file_id)
      end if
#endif
    end if  ! if (mu_id .eq. target_mu) then
    
    ! FR_DEBUG
    call FTR_DEALLOC(Mass_Sce)
    call FTR_DEALLOC(Mass_Sce_trapp)
    call FTR_DEALLOC(Mass_Sce_pass)
    call FTR_DEALLOC(Energy_Sce)
    call FTR_DEALLOC(Moment_Sce)
    call FTR_DEALLOC(Vorticity_Sce)
    call FTR_DEALLOC(buff)
    call FTR_DEALLOC(E_Sce_par)
    call FTR_DEALLOC(E_Sce_perp)
    call FTR_DEALLOC(E_Sce_par_trapp)
    call FTR_DEALLOC(E_Sce_par_pass)
    call FTR_DEALLOC(E_Sce_perp_trapp)
    call FTR_DEALLOC(E_Sce_perp_pass)

    call FTR_DEALLOC(Mass_Sce_glob)
    call FTR_DEALLOC(Mass_Sce_trapp_glob)
    call FTR_DEALLOC(Mass_Sce_pass_glob)
    call FTR_DEALLOC(Energy_Sce_glob)
    call FTR_DEALLOC(Moment_Sce_glob)
    call FTR_DEALLOC(Vorticity_Sce_glob)
    call FTR_DEALLOC(Vorticity_Sce_temp)
    call FTR_DEALLOC(E_Sce_par_glob)
    call FTR_DEALLOC(E_Sce_perp_glob)
    call FTR_DEALLOC(E_Sce_par_trapp_glob)
    call FTR_DEALLOC(E_Sce_par_pass_glob)
    call FTR_DEALLOC(E_Sce_perp_trapp_glob)
    call FTR_DEALLOC(E_Sce_perp_pass_glob)

    call FTR_DEALLOC(buffrecv)
    call FTR_DEALLOC(buffsend)
    call FTR_DEALLOC(mapvpar)
    call FTR_DEALLOC(loc_reqr)
    call FTR_DEALLOC(loc_status)

    do tid = 1, Nbthread
       call FTR_DEALLOC(Romp1_1Nrp1_1Nthetap1(tid)%val)
       call FTR_DEALLOC(Comp1_1Nrp1_1Ntheta(tid)%val)

       call FTR_DEALLOC(J0%omp_diag(tid)%val)
       call FTR_DEALLOC(J0%omp_ldiag(tid)%val)
       call FTR_DEALLOC(J0%omp_udiag(tid)%val)
    enddo

    call Write_dyn_balise_close('CPU_diag_sources') !R4
!R3    call r3_info_end (r3_info_index_1) !R3
#endif 
    ! HAS_SRC
  end subroutine diag_sources 


  !---------------------------------------------- 
  ! Initialization of all the source properties
  !----------------------------------------------
  subroutine init_source_properties(ithis,init_prof, &
    init_magnet,init_curr,geom,J0)
    use globals, only : memory_test
    use init_profile_class
    use init_magnetic_class
    use gyroaverage_class
    type(init_sources) , intent(out) :: ithis
    type(geometry)     , intent(in)  :: geom
    type(init_profile) , intent(in)  :: init_prof
    type(init_magnetic), intent(in)  :: init_magnet
    type(init_current) , intent(in)  :: init_curr
    type(J0operator)   , intent(in)  :: J0

#ifdef HAS_SRC    
    !*** profile object construction ***
    call new_init_sources(ithis,geom)      

    !*** initialization of source profiles ***
    if (.not.memory_test) then
      call initialize_sources(ithis, &
        init_prof,init_magnet,init_curr,geom)
    end if
#else 
    ! HAS_SRC
    ithis%heat_version = 0
#endif 
    ! HAS_SRC
  end subroutine init_source_properties


  !------------------------------------------------------------
  ! Treatment of the Source term
  !   df/dt = S_mu(r,theta,vpar)
  ! The integral of the source is performed 
  ! along the characteristics:
  !  f(r,theta,vpar,t+dt) = f(r,theta,vpar,t) +
  !   dt*S_mu(r,theta,vpar) - dt*dt/2*
  !   [dr/dt*dSdr_mu+dtheta/dt*dSdtheta_mu+dvpar/dt*dSdvpar_mu]
  !------------------------------------------------------------
  subroutine advance_sources(ithis,dt_loc,geom, &
    init_prof,init_magnet,init_curr,f)
    use globals, only : &
      K_Bdrift, pglobal_id, RHS_only, &
      dJ0Phidr_PNrPNthNphi, dJ0Phidtheta_PNrPNthNphi, &
      dJ0Phidphi_PNrPNthNphi
    use globals, only : pSpecies_id, hat_As, hat_Zs
    use fdistribu5d_class
    use init_profile_class
    use init_magnetic_class
    use clock_module
#ifdef _OPENMP
    use OMPutils_module, only : omp_get_thread_num
#endif
    use OMPutils_module, only : Romp1_0Nr, Romp2_0Nr, &
      Romp3_0Nr, Romp4_0Nr, Romp5_0Nr, Romp6_0Nr
!R3 #include "code_instrumentation/r3_info.h" !R3
    type(init_sources) , intent(in)    :: ithis
    real(RKIND)        , intent(in)    :: dt_loc
    type(geometry)     , intent(in)    :: geom
    type(init_profile) , intent(in)    :: init_prof
    type(init_magnetic), intent(in)    :: init_magnet
    type(init_current) , intent(in)    :: init_curr
    type(fdistribu5d)  , intent(inout) :: f

#ifdef HAS_SRC    
    integer     :: ir, itheta, iphi, ivpar
    integer     :: tid
    real(RKIND) :: Sce_tmp, dSdr_tmp, dSdtheta_tmp, dSdvpar_tmp
    real(RKIND) :: deltaf, HFdt2, RHS_only_true
    real(RKIND) :: mu_value, vparl
    real(RKIND) :: inv_Bij, hatZs, inv_sqrtAs
    real(RKIND) :: drdt, dthetadt, dvpardt
    real(RKIND) :: dBdr_tmp, dBdtheta_tmp
    real(RKIND) :: bstar_gradr_tmp, bstar_gradtheta_tmp
    real(RKIND) :: vExB_gradr_tmp, vExB_gradtheta_tmp
    real(RKIND) :: vD_gradr_tmp, vD_gradtheta_tmp
    real(RKIND) :: gradpar_B_tmp, gradpar_J0Phi_tmp
    real(RKIND) :: vExB_gradB_tmp

    !-> components of ExB drift velocity  
    real(RKIND), dimension(:), pointer :: vExB_gradr_1D
    real(RKIND), dimension(:), pointer :: vExB_gradtheta_1D

    !-> components of the curvature drift velocity
    real(RKIND), dimension(:), pointer :: vD_gradr_1D
    real(RKIND), dimension(:), pointer :: vD_gradtheta_1D

    !-> gradpar.B
    real(RKIND), dimension(:), pointer :: gradpar_B_1D

    !-> gradpar.Phi
    real(RKIND), dimension(:), pointer :: gradpar_J0Phi_1D

!R3    call r3_info_begin (r3_info_index_0, 'CPU_advance_source') !R3
    call Write_dyn_balise_open('CPU_advance_source') !R4
    call clck_time(bclock_advance_source)

    RHS_only_true = 0._RKIND
    if (RHS_only) RHS_only_true = 1._RKIND

    HFdt2 = 0.5_RKIND*dt_loc*dt_loc

    !*** computation of 1/sqrt{hat_As} ***
    hatZs      = hat_Zs(pSpecies_id)
    inv_sqrtAs = 1._RKIND/sqrt(hat_As(pSpecies_id))

    !*** initialisation of mu value ***
    mu_value = f%mu_value

#ifdef _OPENMP
!$OMP PARALLEL private(ir,itheta,iphi,ivpar, &
!$OMP Sce_tmp,dSdr_tmp,dSdtheta_tmp,dSdvpar_tmp, &
!$OMP inv_Bij,dBdr_tmp, dBdtheta_tmp, &
!$OMP vExB_gradr_tmp,vExB_gradtheta_tmp, &
!$OMP bstar_gradr_tmp,bstar_gradtheta_tmp, &
!$OMP gradpar_B_tmp,gradpar_J0Phi_tmp,vExB_gradB_tmp, &
!$OMP vparl,vD_gradr_tmp,vD_gradtheta_tmp, &
!$OMP vExB_gradr_1D,vExB_gradtheta_1D, &
!$OMP gradpar_B_1D,gradpar_J0Phi_1D, &
!$OMP vD_gradr_1D,vD_gradtheta_1D, &
!$OMP drdt,dthetadt,dvpardt,deltaf,tid) default(shared)
!$OMP BARRIER
    tid = 1+omp_get_thread_num()
#else
    tid = 1
#endif

    vExB_gradr_1D     => Romp1_0Nr(tid)%val
    vExB_gradtheta_1D => Romp2_0Nr(tid)%val
    vD_gradr_1D       => Romp3_0Nr(tid)%val
    vD_gradtheta_1D   => Romp4_0Nr(tid)%val
    gradpar_J0Phi_1D  => Romp5_0Nr(tid)%val
    gradpar_B_1D      => Romp6_0Nr(tid)%val

!$OMP DO SCHEDULE(static)
    do ivpar = 0,f%n4
      vparl = geom%vparg(ivpar)
      do iphi = 0,f%n3-1
        do itheta = f%jstart,f%jend
          !-> computation of vExB contravariant components
          call compute_vExBgradr(geom,init_magnet,init_curr, &
            dJ0Phidtheta_PNrPNthNphi,dJ0Phidphi_PNrPNthNphi, &
            f%istart_modif,f%iend_modif, &
            itheta,itheta,iphi,iphi,ivpar,ivpar,&
            .false.,vExB_gradr_1D(f%istart_modif:))
          call compute_vExBgradtheta(geom,init_magnet,init_curr, &
            dJ0Phidr_PNrPNthNphi,dJ0Phidphi_PNrPNthNphi, &
            f%istart_modif,f%iend_modif, &
            itheta,itheta,iphi,iphi,ivpar,ivpar,&
            .false.,vExB_gradtheta_1D(f%istart_modif:))
          !-> computation of vD contravariant components
          call compute_vDgradr(geom,init_magnet,init_curr, &
            f%istart_modif,f%iend_modif, &
            itheta,itheta,iphi,iphi,ivpar,ivpar,&
            .false.,vD_gradr_1D(f%istart_modif:))
          call compute_vDgradtheta(geom,init_magnet,init_curr, &
            f%istart_modif,f%iend_modif, &
            itheta,itheta,iphi,iphi,ivpar,ivpar,&
            .false.,vD_gradtheta_1D(f%istart_modif:))
          !-> computation of gradpar.B
          call compute_gradpar_B(geom,init_magnet,init_curr, &
            f%istart_modif,f%iend_modif, &
            itheta,itheta,iphi,iphi,ivpar,ivpar,&
            .false.,gradpar_B_1D(f%istart_modif:))
          !-> computation of gradpar.Phi
          call compute_gradpar_J0Phi(geom, &
            init_magnet,init_curr, &
            f%istart_modif,f%iend_modif,itheta,itheta, &
            iphi,iphi,ivpar,ivpar,.false., &
            gradpar_J0Phi_1D(f%istart_modif:))
          !-> computation of the displacements
          do ir = f%istart_modif,f%iend_modif
            inv_Bij      = 1._RKIND / &
              init_magnet%B_norm(ir,itheta)
            !-> dB/dr and dB/dtheta
            dBdr_tmp     = init_magnet%dBdr(ir,itheta)
            dBdtheta_tmp = init_magnet%dBdtheta(ir,itheta)

            !-> compute bstar.gradr and bstar.gradtheta
#ifdef NO_PRECOMP_BSTAR 
            call compute_bstar_contravariant(geom, &
              init_magnet,init_curr,ir,itheta,ivpar, &
              Sbstar_gradx1=bstar_gradr_tmp, &
              Sbstar_gradx2=bstar_gradtheta_tmp)
#else
            call precomp_bstar_gradr(ir,itheta,ivpar, &
              bstar_gradr_tmp)
            call precomp_bstar_gradtheta(ir,itheta,ivpar, &
              bstar_gradtheta_tmp)
#endif
            !-> vD.gradr and vD.gradtheta
            vD_gradr_tmp       = vD_gradr_1D(ir)
            vD_gradtheta_tmp   = vD_gradtheta_1D(ir)
            !-> gradpar.B
            gradpar_B_tmp      = gradpar_B_1D(ir)
            !-> gradpar(J0.Phi)
            gradpar_J0Phi_tmp  = gradpar_J0Phi_1D(ir)
            !-> computation of vExB.gradB
            vExB_gradr_tmp     = vExB_gradr_1D(ir)
            vExB_gradtheta_tmp = vExB_gradtheta_1D(ir)
            vExB_gradB_tmp     = vExB_gradr_tmp*dBdr_tmp + &
              vExB_gradtheta_tmp*dBdtheta_tmp
            !-> dr/dt, dtheta/dt and dvpar/dt
            drdt               = vparl*bstar_gradr_tmp + &
              vExB_gradr_tmp + vD_gradr_tmp
            dthetadt           = vparl*bstar_gradtheta_tmp + &
              vExB_gradtheta_tmp + vD_gradtheta_tmp
            dvpardt            = &
              - inv_sqrtAs*mu_value*gradpar_B_tmp - &
              (hatZs*inv_sqrtAs)*gradpar_J0Phi_tmp + &
              K_Bdrift*vparl*inv_Bij*vExB_gradB_tmp
            !-> source terms
            Sce_tmp      = ithis%Sce_mu(ir,itheta,ivpar)
            dSdr_tmp     = ithis%dSdr_mu(ir,itheta,ivpar)
            dSdtheta_tmp = ithis%dSdtheta_mu(ir,itheta,ivpar)
            dSdvpar_tmp  = ithis%dSdvpar_mu(ir,itheta,ivpar)
            !->  deltaf = dt*S_mu(r,theta,vpar) - dt*dt/2*
            !      [ dr/dt*dSdr_mu + dtheta/dt*dSdtheta_mu + 
            !        dvpar/dt*dSdvpar_mu ]
            deltaf = dt_loc*Sce_tmp
            deltaf = deltaf - &
              (1._RKIND-RHS_only_true)*HFdt2* &
              (drdt*dSdr_tmp + dthetadt*dSdtheta_tmp + &
              dvpardt*dSdvpar_tmp)
            f%values(ir,itheta,iphi,ivpar) = &
              f%values(ir,itheta,iphi,ivpar) + deltaf
          end do
        end do
      end do
    end do
!$OMP END DO
!$OMP BARRIER
!$OMP END PARALLEL

    call clck_time(eclock_advance_source)
    call clck_diff(bclock_advance_source,eclock_advance_source, &
      global_time_advance_source)
    call Write_dyn_balise_close('CPU_advance_source') !R4
!R3    call r3_info_end (r3_info_index_0) !R3
#endif 
    ! HAS_SRC
  end subroutine advance_sources
end module sources_module
