!**************************************************************
!  Copyright Euratom-CEA
!  Authors : 
!     Virginie Grandgirard (virginie.grandgirard@cea.fr)
!     Chantal Passeron (chantal.passeron@cea.fr)
!     Guillaume Latu (guillaume.latu@cea.fr)
!     Xavier Garbet (xavier.garbet@cea.fr)
!     Philippe Ghendrih (philippe.ghendrih@cea.fr)
!     Yanick Sarazin (yanick.sarazin@cea.fr)
!  
!  This code GYSELA (for GYrokinetic SEmi-LAgrangian) 
!  is a 5D gyrokinetic global full-f code for simulating 
!  the plasma turbulence in a tokamak.
!  
!  This software is governed by the CeCILL-B license 
!  under French law and abiding by the rules of distribution 
!  of free software.  You can  use, modify and redistribute 
!  the software under the terms of the CeCILL-B license as 
!  circulated by CEA, CNRS and INRIA at the following URL
!  "http://www.cecill.info". 
!**************************************************************

#include "config-F.inc"

!----------------------------------------------------------------
! file   : species_transfer.F90
! date   : 2012-03-22
! author : V. Grandgirard
!  
!-----------------------------------------------------------------
module species_transf_module
  use prec_const
  use OMPutils_module
  use MPIutils_module
  use geometry_class
  use spline1d_class
  use fdistribu5d_class
  use wrap_alloc_module
  use init_profile_class
  use init_magnetic_class
  use clock_module
  implicit none

  !-------------------------------------
  !  PRIVATE VARIABLES
  !-------------------------------------
  integer, parameter, private :: nb_moments_Entransf = 4
  !--> temporary arrays
  real(RKIND), dimension(:,:), pointer, private :: fmoments_loc
  real(RKIND), dimension(:,:), pointer, private :: fmoments_glob
  real(RKIND), dimension(:,:), pointer, private :: fmoments_sp0
  real(RKIND), dimension(:,:), pointer, private :: fmoments_sp1
  real(RKIND), dimension(:,:), pointer, private :: fmoments_sp2
  real(RKIND), dimension(:,:), pointer, private :: fmoments_sp3
  !--> Nbar = sum_ispec Ns_FSavg / Nspecies
  real(RKIND), dimension(:)  , pointer, private :: Nbar
  !--> Vbar = sum_ispec Vs_FSavg / Nspecies
  real(RKIND), dimension(:)  , pointer, private :: Vbar
  !--> Tbar = sum_ispec Ts_FSavg / Nspecies
  real(RKIND), dimension(:)  , pointer, private :: Tbar

  !-------------------------------------
  !  PUBLIC VARIABLES
  !------------------------------------- 
  !-> For the third moments of f (specific at each species)
  real(RKIND), dimension(:,:), pointer, public  :: Ns_FSavg
  real(RKIND), dimension(:,:), pointer, public  :: Vs_FSavg
  real(RKIND), dimension(:,:), pointer, public  :: Ts_FSavg
  real(RKIND), dimension(:,:), pointer, public  :: Tscoll_FSavg
  ! -> for asynchrone, need field saving
  real(RKIND), dimension(:,:), pointer, public  :: Ns_FSavg_restart_save
  real(RKIND), dimension(:,:), pointer, public  :: Vs_FSavg_restart_save
  real(RKIND), dimension(:,:), pointer, public  :: Ts_FSavg_restart_save
  real(RKIND), dimension(:,:), pointer, public  :: Tscoll_FSavg_restart_save


  !******************************
  contains
  !******************************

  !-------------------------------------------------
  ! Constructor 
  !-------------------------------------------------
  subroutine new_species_transfer(Nr)
    use globals, only : Nspecies, asynchrone_writing
    integer, intent(in) :: Nr
#ifdef HAS_SP_TRANSF
    !--> temporary arrays
    call FTR_ALLOC_2_ZERO(fmoments_loc,1,nb_moments_Entransf,0,Nr)
    call FTR_ALLOC_2_ZERO(fmoments_glob,1,nb_moments_Entransf,0,Nr)
    call FTR_ALLOC_2_ZERO(fmoments_sp0,1,nb_moments_Entransf,0,Nr)
    call FTR_ALLOC_2_ZERO(fmoments_sp1,1,nb_moments_Entransf,0,Nr)
    !-> For the third moments of f (specific at each species)
    call FTR_ALLOC_2_ZERO(Ns_FSavg,0,Nspecies-1,0,Nr)
    call FTR_ALLOC_2_ZERO(Vs_FSavg,0,Nspecies-1,0,Nr)
    call FTR_ALLOC_2_ZERO(Ts_FSavg,0,Nspecies-1,0,Nr)
    call FTR_ALLOC_2_ZERO(Tscoll_FSavg,0,Nspecies-1,0,Nr)
    if (asynchrone_writing) then
      call FTR_ALLOC_2_ZERO(Ns_FSavg_restart_save,0,Nspecies-1,0,Nr)
      call FTR_ALLOC_2_ZERO(Vs_FSavg_restart_save,0,Nspecies-1,0,Nr)
      call FTR_ALLOC_2_ZERO(Ts_FSavg_restart_save,0,Nspecies-1,0,Nr)
      call FTR_ALLOC_2_ZERO(Tscoll_FSavg_restart_save,0,Nspecies-1,0,Nr)
    endif
#endif 
    ! HAS_SP_TRANSF
  end subroutine new_species_transfer


  !-------------------------------------------------
  ! Destructor 
  !-------------------------------------------------
  subroutine delete_species_transfer()
    use globals, only : Nspecies, asynchrone_writing

#ifdef HAS_SP_TRANSF
    !--> temporary arrays
    call FTR_DEALLOC(fmoments_loc)
    call FTR_DEALLOC(fmoments_glob)
    call FTR_DEALLOC(fmoments_sp0)
    call FTR_DEALLOC(fmoments_sp1)
    !-> For the third moments of f (specific at each species)
    call FTR_DEALLOC(Ns_FSavg)
    call FTR_DEALLOC(Vs_FSavg)
    call FTR_DEALLOC(Ts_FSavg)
    call FTR_DEALLOC(Tscoll_FSavg)
    if (asynchrone_writing) then
      call FTR_DEALLOC(Ns_FSavg_restart_save)
      call FTR_DEALLOC(Vs_FSavg_restart_save)
      call FTR_DEALLOC(Ts_FSavg_restart_save)
      call FTR_DEALLOC(Tscoll_FSavg_restart_save)
    endif
#endif 
    ! HAS_SP_TRANSF
  end subroutine delete_species_transfer


  !----------------------------------------------------------------
  ! Initialisation of the 4 moments mandatory for the
  !  energy transfer operator computation
  !----------------------------------------------------------------
  subroutine init_moments_Entransf(geom,init_magnet,init_curr,f, &
    sEntransf_read)
    use globals, only : pglobal_id, restart, pSpecies_id, Nspecies 
    use MpiMod
    type(geometry)     , intent(in) :: geom
    type(init_magnetic), intent(in) :: init_magnet
    type(init_current) , intent(in) :: init_curr
    type(fdistribu5d)  , intent(in) :: f
    logical, optional  , intent(in) :: sEntransf_read

#ifdef HAS_SP_TRANSF
    integer :: ispec
    logical :: atleast_one_init

    atleast_one_init = .false.
    do ispec = 0,Nspecies-1
      if (.not.restart(ispec)) atleast_one_init = .true.
    end do
    if (atleast_one_init) then
      if (pglobal_id.eq.0) &
        print*,'Computation of the moments for energy transfer', &
        ' operator for species',pSpecies_id
      call compute_moments_Entransf(geom,init_magnet,init_curr,f)
    end if
#endif 
    ! HAS_SP_TRANSF
  end subroutine init_moments_Entransf


  !----------------------------------------------------------------
  ! Compute the three fluid moments 
  !   . <Ns>_FS   = < \int f Jv dvpar dmu >_FS, 
  !   . <V||s>_FS = < \int f vpar Jv dvpar dmu >_FS / <Ns>_FS and
  !   . <Ts>_FS   = 2/3 < \int f Epar_s Jv dvpar dmu >_FS / <Ns>_FS
  !      where Epars = 0.5*(vpar-<V||s>_FS)^2+mu*B
  ! i.e <Ts>_FS = 2/3 < \int f E Jv dvpar dmu >_FS / <Ns>_FS
  !                -1/3 <V||s>_FS*<V||s>_FS
  ! with the energy E = 0.5*vpar^2+mu*B
  !
  ! Compute the flux surface average of Ts,coll (which
  !     corresponds to the temperature appearing in 
  !     collision operator computation)
  !   . <Ts,coll>_FS
  !
  ! Rk: <.>_FS corresponds to the flux surface average,
  !      i.e <.>_FS = \int . Js dtheta dphi /
  !                   \int Js dtheta dphi 
  !      with Js the jacobian in space
  !----------------------------------------------------------------
  subroutine compute_moments_Entransf(geom, &
    init_magnet,init_curr,f)
    use globals               , only : mu_id, pglobal_id, &
      mpi_comm_species, Nspecies, Nbproc_species, pSpecies_id, &
      pInSpecies_id, Nbthread, Tmean_loc
    use coord_system_class    , only : intdthetadphi_Js
    use clock_module          , only : bclock_species_transfer, &
      eclock_species_transfer, global_time_species_transfer
#ifdef _OPENMP
    use OMPutils_module, only : omp_get_thread_num
#endif
    use OMPutils_module, only : Romp1_1nbmom_0Nr
    use MpiMod
    type(geometry)     , intent(in)  :: geom
    type(init_magnetic), intent(in)  :: init_magnet
    type(init_current) , intent(in)  :: init_curr
    type(fdistribu5d)  , intent(in)  :: f

#ifdef HAS_SP_TRANSF
    integer     :: tid, ithread
    integer     :: imom, ir, itheta, iphi, ivpar, imu
    integer     :: nb_elements, iproc_bcast, ierr
    real(RKIND) :: Bij, vpar, mu_value, energy, fval
    real(RKIND) :: jacob_space_tmp, jacob_vel_tmp
    real(RKIND) :: inv_indthetadphi_Js
    real(RKIND) :: coeff_intdtheta_tmp, coeff_intdphi_tmp
    real(RKIND) :: coeff_intdvpar_tmp, coeff_intdmu_tmp
    real(RKIND) :: coeff_int_space, coeff_int_volume
    real(RKIND) :: one_third, two_third
    real(RKIND) :: Ns_tmp, inv_Ns_tmp, Vs_FSavg_tmp

    call clck_time(bclock_species_transfer)

    !***********************************************
    !*** Computation of the third moments of f   ***
    !***    (specific at each species)           ***
    !***********************************************
    mu_value         = f%mu_value
    coeff_intdmu_tmp = geom%coeff_intdmu(mu_id)

#ifdef _OPENMP
!$OMP PARALLEL default(none) &
!$OMP private(tid,imom,ir,itheta,iphi,ivpar,fmoments_loc, &
!$OMP  vpar,coeff_intdvpar_tmp,coeff_intdphi_tmp, &
!$OMP  coeff_intdtheta_tmp,coeff_int_volume,coeff_int_space, &
!$OMP  Bij,energy,fval,jacob_vel_tmp,jacob_space_tmp, &
!$OMP  inv_indthetadphi_Js) &
!$OMP shared(f,geom,init_magnet,init_curr,mu_value, &
!$OMP  Romp1_1nbmom_0Nr,coeff_intdmu_tmp,jacobian_space, &
!$OMP  intdthetadphi_Js,Tmean_loc)
!$OMP BARRIER
    tid = 1+omp_get_thread_num()
#else
    tid = 1
#endif
    !*** initialisation of the values to 0 ***
    do ir = 0,geom%Nr
      do imom = 1,nb_moments_Entransf
        Romp1_1nbmom_0Nr(tid)%val(imom,ir) = 0._RKIND
      end do
    end do
!$OMP DO SCHEDULE(static)
    do ivpar = 0,f%n4
      coeff_intdvpar_tmp = geom%coeff_intdvpar(ivpar)
      vpar               = geom%vparg(ivpar)
      do iphi = 0,f%n3-1
        coeff_intdphi_tmp = geom%coeff_intdphi(iphi)
        do itheta = f%jstart,f%jend
          coeff_intdtheta_tmp = geom%coeff_intdtheta(itheta)
          do ir = f%istart,f%iend
            Bij    = init_magnet%B_norm(ir,itheta)
            energy = 0.5*vpar*vpar+mu_value*Bij 
            fval   = f%values(ir,itheta,iphi,ivpar)
            !-> jacobian_velocity = 2*pi*Bstar
            call compute_jacobian_velocity(geom,init_magnet, &
              init_curr,ir,itheta,ivpar,jacob_vel_tmp)
            !-> jacobian in space
            jacob_space_tmp     = jacobian_space(ir,itheta)
            !-> compute the elements of volume in space and velocity
            coeff_int_volume    = jacob_space_tmp * &
              coeff_intdtheta_tmp*coeff_intdphi_tmp * &
              jacob_vel_tmp*coeff_intdvpar_tmp*coeff_intdmu_tmp 
            inv_indthetadphi_Js = 1._RKIND/intdthetadphi_Js(ir)
            !-> <Ns>_FS = < \int f Jv dvpar dmu >_FS
            Romp1_1nbmom_0Nr(tid)%val(1,ir) = &
              Romp1_1nbmom_0Nr(tid)%val(1,ir) + &
              fval*coeff_int_volume*inv_indthetadphi_Js
            !-> < \int f vpar Jv dvpar dmu >_FS
            Romp1_1nbmom_0Nr(tid)%val(2,ir) = &
              Romp1_1nbmom_0Nr(tid)%val(2,ir) + &
              vpar*fval*coeff_int_volume*inv_indthetadphi_Js
            !-> < \int f E Jv dvpar dmu >_FS with E the energy
            Romp1_1nbmom_0Nr(tid)%val(3,ir) = &
              Romp1_1nbmom_0Nr(tid)%val(3,ir) + &
              energy*fval*coeff_int_volume*inv_indthetadphi_Js
          end do
        end do
      end do
    end do
!$OMP END DO
!$OMP BARRIER

    !*** Computation of <Ts,coll>_FS ***
!$OMP DO SCHEDULE(static)
    do iphi = 0,f%n3-1
      coeff_intdphi_tmp = geom%coeff_intdphi(iphi)
      do itheta = f%jstart,f%jend
        coeff_intdtheta_tmp = geom%coeff_intdtheta(itheta)
        do ir = f%istart,f%iend
          !-> jacobian in space
          jacob_space_tmp       = jacobian_space(ir,itheta)
          !-> compute the elements of volume in space
          coeff_int_space     = jacob_space_tmp * &
            coeff_intdtheta_tmp*coeff_intdphi_tmp
          inv_indthetadphi_Js = 1._RKIND/intdthetadphi_Js(ir)
          !-> <Ts,coll>_FS
          Romp1_1nbmom_0Nr(tid)%val(4,ir) = &
            Romp1_1nbmom_0Nr(tid)%val(4,ir) + &
              Tmean_loc(ir,itheta,iphi)*coeff_int_space*inv_indthetadphi_Js
        end do
      end do
    end do
!$OMP END DO
!$OMP BARRIER
!$OMP END PARALLEL

    !*** fill an unique array containing all the       ***
    !***  radial profiles of the fluid moments         ***
    !*** Rq: each processor fills the part istart:iend ***
    !***     and all the others values are put equal   ***
    !***     to 0                                      ***
    fmoments_loc = 0._RKIND
    do ithread = 1,Nbthread
      do ir = f%istart,f%iend
        do imom = 1,nb_moments_Entransf
          fmoments_loc(imom,ir) = fmoments_loc(imom,ir) + & 
            Romp1_1nbmom_0Nr(ithread)%val(imom,ir)
        end do
      end do
    end do

    !*** MY_Reduce ****
    nb_elements = nb_moments_Entransf*(geom%Nr+1)
    call MY_Reduce(fmoments_loc,fmoments_glob,nb_elements, &
      MPI_REAL8,MPI_SUM,0,mpi_comm_species,ierr)

    one_third = 1._RKIND/3._RKIND
    two_third = 2._RKIND/3._RKIND

    !--> Broadcast the moments associated to the first species
    if ( (pSpecies_id.eq.0).and.(pInSpecies_id.eq.0) ) &
      fmoments_sp0 = fmoments_glob
    iproc_bcast = 0
    call MPI_BCAST(fmoments_sp0,nb_elements,MPI_REAL8,0, &
      MPI_COMM_WORLD,ierr)
    do ir = 0,geom%Nr
      Ns_tmp             = fmoments_sp0(1,ir)
      inv_Ns_tmp         = 1._RKIND/Ns_tmp
      Ns_FSavg(0,ir)     = Ns_tmp
      Vs_FSavg_tmp       = fmoments_sp0(2,ir) * inv_Ns_tmp
      Vs_FSavg(0,ir)     = Vs_FSavg_tmp
      Ts_FSavg(0,ir)     = two_third * &
        fmoments_sp0(3,ir)*inv_Ns_tmp - &
        one_third*Vs_FSavg_tmp*Vs_FSavg_tmp
      Tscoll_FSavg(0,ir) = fmoments_sp0(4,ir) 
    end do

    !--> Broadcast the moments associated to the second species
    if ( (pSpecies_id.eq.1).and.(pInSpecies_id.eq.0) ) &
      fmoments_sp1 = fmoments_glob
    iproc_bcast = Nbproc_species 
    call MPI_BCAST(fmoments_sp1,nb_elements,MPI_REAL8,iproc_bcast, &
      MPI_COMM_WORLD,ierr)
    do ir = 0,geom%Nr
      Ns_tmp             = fmoments_sp1(1,ir)
      inv_Ns_tmp         = 1._RKIND/Ns_tmp
      Ns_FSavg(1,ir)     = Ns_tmp
      Vs_FSavg_tmp       = fmoments_sp1(2,ir) * inv_Ns_tmp
      Vs_FSavg(1,ir)     = Vs_FSavg_tmp
      Ts_FSavg(1,ir)     = two_third * &
        fmoments_sp1(3,ir)*inv_Ns_tmp - &
        one_third*Vs_FSavg_tmp*Vs_FSavg_tmp
      Tscoll_FSavg(1,ir) = fmoments_sp1(4,ir) 
    end do

    call clck_time(eclock_species_transfer)
    call clck_diff(bclock_species_transfer, &
      eclock_species_transfer,global_time_species_transfer)
#endif 
    ! HAS_SP_TRANSF
  end subroutine compute_moments_Entransf


  !----------------------------------------------------------------------
  ! Solve the equations 
  !   df_s/dt  = C1_{ss'}(Epar_s) + C2_{ss'}(E_s)    (for species s, 
  !                                             which is consider
  !                                             as the major species)
  !   df_s'/dt = C1_{s's}(Epar_s') + C2_{s's}(E_s')  (for species s')
  ! with the reduced collision operators for energy transfer
  ! are defined by
  !   C1_{ss'}(Epar_s) = - mu_ss'^E deltaT_ss'/<T>_ss'*(Epar_s/<T>_ss'-3/2) * 
  !                    exp(-Epar_s/<T>_ss')/(2*pi*<T_ss')^(3/2)
  !   C1_{s's}(Epar_s') = - mu_ss'^E deltaT_ss'/<T>_ss'*(Epar_s'/<T>_ss'-3/2) * 
  !                     exp(-Epar_s'/<T>_ss')/(2*pi*<T>_ss')^(3/2)
  ! with 
  !   mu_ss'^E = 8*epsilon^(3/2)/(3*sqrt(pi)) 1/hat_As' * 
  !            (hat_Zs'/hat_Zs)^2 * n_s' * sqrt{<Ts,coll>_FS}/(q R0) *
  !            vTs/vT>*nustar_s
  ! and the reduced collision operators for momentum transfer are
  ! defined by
  !   C2_{ss'}(E_s)  = - mu^vGpar deltaV_ss'*vGpar_s*exp(-E_s/<T>ss')
  !   C2_{s's}(E_s') = - mu^vGpar deltaV_s's*vGpar_s'*exp(-E_s'/<T>ss')
  ! with
  !   mu_vGpar = (2*pi)^{-3/2} epsilon^{3/2}/(qR_0) *
  !                  n_s'/vT>*sqrt{T_s}*sqrt{<Ts,coll>_FS} /
  !                  <T>_ss'^{5/2} * (hat_Zs'/hat_ZS)^2 * nustar_s 
  ! where
  !   . epsilon is the aspect ratio
  !   . Epar_s = (vGpar_s-Vbar)^2/2 + mu*B
  !   . E_s   = vGpar_s^2/2 + mu*B
  !   . deltaT_ss' = T_s - T_s'
  !   . deltaV_ss' = hat_As^{-1/2}*V_s - hat_As'^{-1/2}*V_s'
  !   . <T>_ss'    = (T_s+T_s')/2
  !   . <V>_ss'    = (V_s+V_s')/2
  !   . <Ts,coll>_FS is the flux surface average of Ts,coll which
  !                  corresponds to the temperature appearing in 
  !                  collision operator computation
  !   . vTs = sqrt{T_s}
  !   . vT> = max(hat_As^{-1/2}*vTs,hat_As'^{-1/2}*vTs') 
  !           with vTs' = sqrt{T_s'}
  ! Rk: These equations are solved as
  !      f_s(tn+1) - f_s(tn)   = (C1_{ss'}(Epar_s) -
  !                               C2_{ss'}(E_s))* delta t
  !      f_s'(tn+1) - f_s'(tn) = (C1_{s's}(Epar_s') -
  !                               C2_{s's}(E_s'))* delta t
  !----------------------------------------------------------------------
  subroutine advance_species_transfer(geom,init_prof,init_magnet,deltat,f)
    use globals              , only : major_species, pSpecies_id, &
      Nspecies, hat_As, hat_Zs, R0, energy_transfer, moment_transfer
    use clock_module          , only : bclock_species_transfer, &
      eclock_species_transfer, global_time_species_transfer
    use collisions_vpar_module, only : nustars
    implicit none
    type(geometry)     , intent(in)    :: geom
    type(init_profile) , intent(in)    :: init_prof
    type(init_magnetic), intent(in)    :: init_magnet
    real(RKIND)        , intent(in)    :: deltat  
    type(fdistribu5d)  , intent(inout) :: f
    
#ifdef HAS_SP_TRANSF
    integer     :: ispec_major, ispec_impur
    integer     :: ir, itheta, iphi, ivpar
    real(RKIND) :: mu_value, Bij, vpar
    real(RKIND) :: inv_Tavg, Epar_s, Epars_onTavg, Es_onTavg
    real(RKIND) :: hatZs, hatZsprime, hatZsprime_on_hatZs
    real(RKIND) :: hatAs, hatAsprime
    real(RKIND) :: sqrt_hatAs, sqrt_hatAsprime
    real(RKIND) :: invsqrt_hatAs, invsqrt_hatAsprime
    real(RKIND) :: epstemp, epstemp_pow3on2, one_on_qR0
    real(RKIND) :: Tavg_tmp, Tavg_pow5on2, sqrt_Tscoll 
    real(RKIND) :: C1_ssprime, C2_ssprime
    real(RKIND) :: Ts, Tsprime, Ns, Nsprime, Vs, Vsprime
    real(RKIND) :: vTs, vTmax, vTs_on_vTmax, Vavg_tmp
    real(RKIND) :: coeff_mu_E, coeff_mu_vGpar
    real(RKIND), dimension(f%istart:f%iend) :: Navg_species
    real(RKIND), dimension(f%istart:f%iend) :: Vavg_species
    real(RKIND), dimension(f%istart:f%iend) :: Tavg_species
    real(RKIND), dimension(f%istart:f%iend) :: DeltaT_species
    real(RKIND), dimension(f%istart:f%iend) :: DeltaV_species
    real(RKIND), dimension(f%istart:f%iend) :: mu_ssprime_E
    real(RKIND), dimension(f%istart:f%iend) :: mu_vGpar

    call clck_time(bclock_species_transfer)

    ispec_major = 0
    ispec_impur = 1
    if (major_species.ne.0) then
      ispec_major = 1
      ispec_impur = 0
    end if
    
    mu_ssprime_E(f%istart:f%iend) = 0._RKIND
    mu_vGpar(f%istart:f%iend)     = 0._RKIND
    coeff_mu_E     = 0._RKIND
    coeff_mu_vGpar = 0._RKIND
    if (energy_transfer) coeff_mu_E     = 1._RKIND
    if (moment_transfer) coeff_mu_vGpar = 1._RKIND

    !*** computation of N>_ss', <V>_ss', <T>_ss' ***
    !***  mu_ss'^E and mu_ss'^vGpar              ***
    hatZs               = hat_Zs(ispec_major)
    hatZsprime          = hat_Zs(ispec_impur)
    hatZsprime_on_hatZs = hatZsprime/hatZs
    hatAs               = hat_As(ispec_major)
    hatAsprime          = hat_As(ispec_impur)
    sqrt_hatAs          = sqrt(hatAs)
    sqrt_hatAsprime     = sqrt(hatAsprime)
    invsqrt_hatAs       = 1._RKIND/sqrt_hatAs
    invsqrt_hatAsprime  = 1._RKIND/sqrt_hatAsprime
    do ir = f%istart,f%iend
      Ts      = Ts_FSavg(ispec_major,ir)
      Tsprime = Ts_FSavg(ispec_impur,ir)
      Ns      = Ns_FSavg(ispec_major,ir)
      Nsprime = Ns_FSavg(ispec_impur,ir)      
      Vs      = Vs_FSavg(ispec_major,ir)
      Vsprime = Vs_FSavg(ispec_impur,ir)
      !--> <N>_ss' = 0.5*(N_s+N_s')
      Navg_species(ir) = 0.5_RKIND*(Ns+Nsprime)
      !--> <V>_ss' = 0.5*(V_s+V_s')
      Vavg_species(ir) = 0.5_RKIND*(Vs+Vsprime)
      !--> <T>_ss' = 0.5*(T_s+T_s')
      Tavg_tmp         = 0.5_RKIND*(Ts+Tsprime) 
      Tavg_species(ir) = Tavg_tmp
      !--> Delta T_ss' = T_s-T_s'
      if (pSpecies_id.eq.major_species) then
        DeltaT_species(ir) = Ts-Tsprime
        DeltaV_species(ir) = invsqrt_hatAs*Vs - &
          invsqrt_hatAsprime*Vsprime
      else
        DeltaT_species(ir) = Tsprime-Ts
        DeltaV_species(ir) = invsqrt_hatAsprime*Vsprime - &
          invsqrt_hatAs*Vs
      end if
      epstemp              = geom%rg(ir)*inv_R(ir,0)
      epstemp_pow3on2      = epstemp*sqrt(epstemp)
      vTs                  = sqrt(Ts)
      vTmax                = max(invsqrt_hatAs*vTs, &
        invsqrt_hatAsprime*sqrt(Tsprime))
      vTs_on_vTmax         = vTs/vTmax
      one_on_qR0           = init_prof%iota(ir)/R0 
      Tavg_pow5on2         = Tavg_tmp*Tavg_tmp*sqrt(Tavg_tmp)
      sqrt_Tscoll          = sqrt(abs(Tscoll_FSavg(ispec_major,ir)))
      !--> mu_ss'^E = 8*epsilon^(3/2)/(3*sqrt(pi)) *
      !-->          1/hat_As'*(hat_Zs'/hat_Zs)^2 * 
      !-->          n_s'*sqrt{Ts,coll}/(q R0)*vTs/vT> * nustar_s
      mu_ssprime_E(ir)     = coeff_mu_E * &
        8._RKIND*epstemp_pow3on2/ &
        (3._RKIND*sqrt(PI)*hatAsprime) * &
        hatZsprime_on_hatZs*hatZsprime_on_hatZs * &
        Nsprime*sqrt_Tscoll*vTs_on_vTmax*one_on_qR0 * &
        nustars(ispec_major)
      !--> mu^vGpar = (2*pi)^{-3/2}*epsilon^{3/2}/(qR_0)*n_s'/vT> * 
      !-->             sqrt{T_s}*sqrt{<Ts,coll>_FS} /<T>_ss'^{5/2} * 
      !-->             (hat_Zs'/hat_ZS)^2 * nustar_s 
      mu_vGpar(ir) = coeff_mu_vGpar * &
        epstemp_pow3on2*one_on_qR0 * &
        hatZsprime_on_hatZs*hatZsprime_on_hatZs * &        
        Nsprime*sqrt_Tscoll*vTs_on_vTmax / &
        (TWOPI*sqrt(TWOPI)*Tavg_pow5on2) * &
        nustars(ispec_major)
    end do
    
    !***********************************************************************
    ! Computation of the energy transfer operator
    !  C1_{ss'}(Epar_s) = - mu_ss' deltaT_ss'/<T>_ss' * 
    !    (Epar_s/<T>_ss'-3/2)*exp(-Epar_s/<T>_ss') / (2*pi*<T_ss')^(3/2)
    ! and the momentum transfer operator
    !   C2_{ss'}(E_s) = - sqrt(hat_As)*mu^vGpar *
    !                      deltaV_ss'*vGpar_s*exp(-E_s/<T>ss')
    !***********************************************************************
#ifdef _OPENMP
!$OMP PARALLEL default(none) &
!$OMP private(ir,itheta,iphi,ivpar, &
!$OMP  vpar,Bij,Vavg_tmp,Tavg_tmp,inv_Tavg, &
!$OMP  Epar_s,Epars_onTavg, Es_onTavg, &
!$OMP  C1_ssprime,C2_ssprime) &
!$OMP shared(pSpecies_id,hat_As,invsqrt_hatAs, &
!$OMP  geom,init_magnet,mu_value, &
!$OMP  deltat,f,Vavg_species,Tavg_species, &
!$OMP  DeltaT_species,DeltaV_species, &
!$OMP  mu_ssprime_E,mu_vGpar)
!$OMP BARRIER
#endif
    mu_value      = f%mu_value
    invsqrt_hatAs = 1._RKIND/sqrt(hat_As(pspecies_id))
 
!$OMP DO SCHEDULE(STATIC)
    do ivpar = 0,f%n4
      vpar = geom%vparg(ivpar)
      do iphi = 0,f%n3-1
        do itheta = f%jstart,f%jend
          do ir = f%istart,f%iend
            Bij          = init_magnet%B_norm(ir,itheta)
            Tavg_tmp     = Tavg_species(ir)
            inv_Tavg     = 1._RKIND/Tavg_tmp
            Vavg_tmp     = Vavg_species(ir)
            Epar_s       = 0.5_RKIND*(vpar-Vavg_tmp) * &
              (vpar-Vavg_tmp) + mu_value*Bij
            Epars_onTavg = Epar_s*inv_Tavg
            Es_onTavg    = (0.5*vpar*vpar + mu_value*Bij) * &
              inv_Tavg 
            C1_ssprime   = -mu_ssprime_E(ir)* &
              DeltaT_species(ir)*inv_Tavg * &
              (Epars_onTavg-1.5_RKIND)*exp(-Epars_onTavg) / &
              ((TWOPI*Tavg_tmp)*sqrt(TWOPI*Tavg_tmp))
            C2_ssprime   = -invsqrt_hatAs * &
              mu_vGpar(ir)*DeltaV_species(ir) * &
              vpar*exp(-Es_onTavg)

            f%values(ir,itheta,iphi,ivpar) = &
              f%values(ir,itheta,iphi,ivpar) + &
              (C1_ssprime+C2_ssprime)*deltat
          end do
        end do
      end do
    end do
!$OMP END DO
!$OMP END PARALLEL

    call clck_time(eclock_species_transfer)
    call clck_diff(bclock_species_transfer, &
      eclock_species_transfer,global_time_species_transfer)
#endif 
    ! HAS_SP_TRANSF
  end subroutine advance_species_transfer
end module species_transf_module
